# /*
#  * Copyright (C) 2012 Technische Universitaet Muenchen
#  * This file is part of the Sierpinski project. For conditions of distribution and
#  * use, please see the copyright notice in the file 'copyright.txt' at the root
#  * directory of this package and the copyright notice at
#  * http://www5.in.tum.de/sierpinski
#  *
#  * Created on: Jun 5, 2012
#  * Author: Martin Schreiber <martin.schreiber@in.tum.de>
#  */
# 
# Computation of DG matrices:
restart;
# Setup:
with(LinearAlgebra); with(ArrayTools); with(orthopoly); with(CodeGeneration); with(FileTools); with(Physics);
kernelopts(assertlevel = 1);
libname := cat(currentdir(), "/../lib/"), libname;
# Setup:
# Basis functions:
if type(progArgBasisFunctionType, integer) then configBasisFunctionType := progArgBasisFunctionType else configBasisFunctionType := 0 end if;
# Degree of basis functions:
if type(progArgDegreeOfBasisFunctions, integer) then configDegreeOfBasisFunctions := progArgDegreeOfBasisFunctions else configDegreeOfBasisFunctions := 2 end if;
# Create 2D or 3D basis functions and matrices
if type(progArgNumberOfDimensions, integer) then configNumberOfDimensions := progArgNumberOfDimensions else configNumberOfDimensions := 2 end if;
# Use Gauss quadrature instead of GL
configUseGaussQuadratureRules := 0;
# Output directory for matrix files.
if configNumberOfDimensions = 2 then configOutputDir := cat(currentdir(), "/dg_matrices_2d_triangle/") else configOutputDir := cat(currentdir(), "/dg_matrices_3d_prism/") end if;
# Cell domain
OmegaTriangle := [y = 0 .. 1-x, x = 0 .. 1]; if configNumberOfDimensions = 2 then Omega := [y = 0 .. 1-x, x = 0 .. 1]; basisFunctionParameters := x, y else Omega := [y = 0 .. 1-x, x = 0 .. 1, z = 0 .. 1]; basisFunctionParameters := x, y, z end if;
# Kramers rule to compute inverse
kramerInverse := proc (M) local MInv; MInv := simplify(simplify(LinearAlgebra[Adjoint](M))/simplify(LinearAlgebra[Determinant](M))); ASSERT(LinearAlgebra[Equal](simplify(Physics:-`.`(M, MInv)), LinearAlgebra[IdentityMatrix](RowDimension(M)))); return MInv end proc;
# Vandermonde Matrix for 2D basis functions
getVandermondeMatrix2D := proc (nodalCoords2D, degreeOfBasisFunctions) local numberOfBasisFunctions2D, Mret, c, i, m, n; numberOfBasisFunctions2D := (1/2)*(degreeOfBasisFunctions+1)*(degreeOfBasisFunctions+2); Mret := Matrix(numberOfBasisFunctions2D, numberOfBasisFunctions2D, proc (i, j) options operator, arrow; Pi end proc); c := 1; for i to numberOfBasisFunctions2D do c := 1; for m from 0 to degreeOfBasisFunctions do for n from 0 to degreeOfBasisFunctions-m do Mret[i, c] := nodalCoords2D[i][1]^n*nodalCoords2D[i][2]^m; c := c+1 end do end do end do; return Mret end proc;
getVandermondeMatrix3D := proc (nodalCoords3D, degreeOfBasisFunctions) local numberOfBasisFunctions3D, Mret, c, i, m, n, o; numberOfBasisFunctions3D := (1/6)*(degreeOfBasisFunctions+1)*(degreeOfBasisFunctions+2)*(degreeOfBasisFunctions+3); Mret := Matrix(numberOfBasisFunctions3D, numberOfBasisFunctions3D, proc (i, j) options operator, arrow; Pi end proc); for i to numberOfBasisFunctions3D do c := 1; for m from 0 to degreeOfBasisFunctions do for n from 0 to degreeOfBasisFunctions-m do for o from 0 to degreeOfBasisFunctions-m-n do Mret[i, c] := nodalCoords3D[i][1]^n*nodalCoords3D[i][2]^m*nodalCoords3D[i][3]^o; c := c+1 end do end do end do end do; return Mret end proc;
getBasisFunctionsFromVandermondeMatrix1D := proc (MatV, degreeOfBasisFunctions) local numberOfBasisFunctions, Vinv, funcs, i, m, n, c; numberOfBasisFunctions := degreeOfBasisFunctions+1; Vinv := LinearAlgebra[MatrixInverse](MatV); funcs := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); for i to numberOfBasisFunctions do for m from 0 to degreeOfBasisFunctions do funcs[i] := funcs[i]+Vinv[m+1, i]*z^m end do end do; return funcs end proc;
getBasisFunctionsFromVandermondeMatrix2D := proc (MatV, degreeOfBasisFunctions) local numberOfBasisFunctions2D, Vinv, funcs, i, m, n, c; numberOfBasisFunctions2D := (1/2)*(degreeOfBasisFunctions+1)*(degreeOfBasisFunctions+2); Vinv := LinearAlgebra[MatrixInverse](MatV); funcs := Vector(numberOfBasisFunctions2D, proc (i) options operator, arrow; 0 end proc); for i to numberOfBasisFunctions2D do c := 1; for m from 0 to degreeOfBasisFunctions do for n from 0 to degreeOfBasisFunctions-m do funcs[i] := funcs[i]+Vinv[c, i]*x^n*y^m; c := c+1 end do end do end do; return funcs end proc;
getBasisFunctionsFromVandermondeMatrix3D := proc (MatV, degreeOfBasisFunctions) local numberOfBasisFunctions3D, Vinv, funcs, i, m, n, o, c; numberOfBasisFunctions3D := (1/6)*(degreeOfBasisFunctions+1)*(degreeOfBasisFunctions+2)*(degreeOfBasisFunctions+3); Vinv := LinearAlgebra[MatrixInverse](MatV); funcs := Vector(numberOfBasisFunctions3D, proc (i) options operator, arrow; 0 end proc); for i to numberOfBasisFunctions3D do c := 1; for m from 0 to degreeOfBasisFunctions do for n from 0 to degreeOfBasisFunctions-m do for o from 0 to degreeOfBasisFunctions-m-n do funcs[i] := funcs[i]+Vinv[c, i]*x^n*y^m*z^o; c := c+1 end do end do end do end do; return funcs end proc;
# 1D Gauss quadrature sampling points:
quadGauss1DNumberOfSamplingPoints := ceil((2*configDegreeOfBasisFunctions+1)*(1/2)); quadGauss1DSamplingPointsX := eval(solve(P(quadGauss1DNumberOfSamplingPoints, x), x)); quadGauss1DSamplingPointsX := sort(Vector([quadGauss1DSamplingPointsX]), proc (a, b) options operator, arrow; evalf(Re(a)) < evalf(Re(b)) end proc); quadGauss1DSamplingPoints := simplify([seq((i+1)*(1/2), `in`(i, quadGauss1DSamplingPointsX))]); Pdiff := proc (n, x) options operator, arrow; (unapply(diff(P(n, diffx), diffx), diffx))(x) end proc; Pw := proc (x) options operator, arrow; 2/((1-x^2)*Pdiff(quadGauss1DNumberOfSamplingPoints, x)^2) end proc; quadGauss1DWeightsX := simplify([seq(Pw(i), `in`(i, quadGauss1DSamplingPointsX))]); quadGauss1DWeights := simplify([seq((1/2)*w, `in`(w, quadGauss1DWeightsX))]); quadGauss1DPhiTmp := Matrix(configDegreeOfBasisFunctions+1, configDegreeOfBasisFunctions+1, proc (i, j) options operator, arrow; quadGauss1DSamplingPoints[i]^(j-1) end proc); quadGauss1DPhiTmp2 := getBasisFunctionsFromVandermondeMatrix1D(quadGauss1DPhiTmp, configDegreeOfBasisFunctions); quadGauss1DPhi := Vector(quadGauss1DNumberOfSamplingPoints, proc (i) options operator, arrow; unapply(eval(quadGauss1DPhiTmp2[i]), z) end proc);
# 2D Gauss quadrature sampling points
if configUseGaussQuadratureRules = 1 then numberOfBasisFunctions2D := (1/2)*(configDegreeOfBasisFunctions+1)*(configDegreeOfBasisFunctions+2); i := 1; do if Dunavant:-getDunavantOrderNum(i) >= numberOfBasisFunctions2D then break end if; i := i+1 end do; quadGauss2DNumberOfBasisFunctions := Dunavant:-getDunavantOrderNum(i); quadGauss2DSamplingPoints := identify(Dunavant:-getDunavantCoords(i)); quadGauss2DWeights := identify(Dunavant:-getDunavantWeights(i)); quadGauss2DNumberOfSamplingPoints := quadGauss2DNumberOfBasisFunctions; quadGauss2DPhiTmp := getVandermondeMatrix2D(quadGauss2DSamplingPoints, configDegreeOfBasisFunctions); quadGauss2DPhiTmp2 := simplify(getBasisFunctionsFromVandermondeMatrix2D(quadGauss2DPhiTmp, configDegreeOfBasisFunctions)); quadGauss2DPhi := Vector(quadGauss2DNumberOfSamplingPoints, proc (i) options operator, arrow; unapply(eval(quadGauss2DPhiTmp2[i]), z) end proc) end if;
# 1D Gauss Lobatto nodal points:
quadGL1DNumberOfSamplingPoints := configDegreeOfBasisFunctions+1; Pdiff := proc (n, x) options operator, arrow; (unapply(diff(P(n, diffx), diffx), diffx))(x) end proc; if quadGL1DNumberOfSamplingPoints = 1 then quadGL1DSamplingPointsX := [0]; quadGL1DWeightsX := [2] else quadGL1DSamplingPointsXTmp := eval(solve(Pdiff(quadGL1DNumberOfSamplingPoints-1, x), x)); quadGL1DSamplingPointsX := [-1, sort([quadGL1DSamplingPointsXTmp], proc (a, b) options operator, arrow; evalf(Re(a)) < evalf(Re(b)) end proc)[], 1]; quadGL1DWeightsX := simplify([seq(2/(quadGL1DNumberOfSamplingPoints*(quadGL1DNumberOfSamplingPoints-1)*P(quadGL1DNumberOfSamplingPoints-1, i)^2), `in`(i, convert(quadGL1DSamplingPointsX, list)))]) end if; quadGL1DSamplingPoints := simplify([seq((i+1)*(1/2), `in`(i, quadGL1DSamplingPointsX))]); quadGL1DWeights := simplify([seq((1/2)*i, `in`(i, quadGL1DWeightsX))]); quadGL1DPhiTmp := Matrix(quadGL1DNumberOfSamplingPoints, quadGL1DNumberOfSamplingPoints, proc (i, j) options operator, arrow; quadGL1DSamplingPoints[i]^(j-1) end proc); quadGL1DPhiTmp2 := getBasisFunctionsFromVandermondeMatrix1D(quadGL1DPhiTmp, quadGL1DNumberOfSamplingPoints-1); quadGL1DPhi := Vector(quadGL1DNumberOfSamplingPoints, proc (i) options operator, arrow; unapply(eval(quadGL1DPhiTmp2[i]), z) end proc);
# 2D Gauss Lobatto nodal points
quadGL2DNumberOfBasisFunctions := (1/2)*(configDegreeOfBasisFunctions+1)*(configDegreeOfBasisFunctions+2); quadGL2DNumberOfSamplingPoints := quadGL2DNumberOfBasisFunctions; quadGL2DSamplingPointsTmp := GaussLobattoPoints:-Nodes2D(configDegreeOfBasisFunctions); quadGL2DSamplingPointsTmp2 := simplify([GaussLobattoPoints:-xytoxieta(quadGL2DSamplingPointsTmp[1], quadGL2DSamplingPointsTmp[2])]); quadGL2DSamplingPoints := [seq([quadGL2DSamplingPointsTmp2[1][i], quadGL2DSamplingPointsTmp2[2][i]], i = 1 .. quadGL2DNumberOfSamplingPoints)]; quadGL2DPhiTmp := getVandermondeMatrix2D(quadGL2DSamplingPoints, configDegreeOfBasisFunctions); quadGL2DPhiTmp2 := simplify(getBasisFunctionsFromVandermondeMatrix2D(quadGL2DPhiTmp, configDegreeOfBasisFunctions)); quadGL2DPhi := Vector(quadGL2DNumberOfSamplingPoints, proc (i) options operator, arrow; unapply(quadGL2DPhiTmp2[i], x, y) end proc);
# Basis functions
# 
if basisFunctionType = -1 then basisFunctionDescription := "using manally specified basis functions"; if degreeOfBasisFunctions = 0 then basisFunctionArray := Vector([1]) elif degreeOfBasisFunctions = 1 then basisFunctionArray := Vector([2*x+2*y-1, 1-2*x, 1-2*y]) end if; numberOfBasisFunctions := (1/2)*(degreeOfBasisFunctions+1)*(degreeOfBasisFunctions+2) end if;
#   -2: NODAL, Dunavant elements, usually not used;
if configBasisFunctionType = -2 then if numberOfDimensions = 2 then numberOfBasisFunctions2D := (1/2)*(degreeOfBasisFunctions+1)*(degreeOfBasisFunctions+2); i := 1; do if Dunavant:-getDunavantOrderNum(i) >= numberOfBasisFunctions2D then break end if; i := i+1 end do; numberOfBasisFunctions2D := Dunavant:-getDunavantOrderNum(i); nodalCoords2D := identify(Dunavant:-getDunavantCoords(i)); numberOfBasisFunctions := numberOfBasisFunctions2D; nodalCoords2D := [seq(nodalCoords2D[i], i = 1 .. numberOfBasisFunctions)]; nodalCoords := nodalCoords2D; NodalMatrix := Matrix(numberOfBasisFunctions, numberOfBasisFunctions, proc (i, j) options operator, arrow; 0 end proc); c := 1; basisFunctionArray := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); for i to numberOfBasisFunctions do c := 1; for m from 0 to degreeOfBasisFunctions do for n from 0 to degreeOfBasisFunctions-m do if c <= numberOfBasisFunctions then NodalMatrix[i, c] := nodalCoords2D[i][1]^n*nodalCoords2D[i][2]^m end if; c := c+1 end do end do end do; CoeffMatrix := 1/NodalMatrix; for i to numberOfBasisFunctions do c := 1; for m from 0 to degreeOfBasisFunctions do for n from 0 to degreeOfBasisFunctions-m do if c <= numberOfBasisFunctions then basisFunctionArray[i] := basisFunctionArray[i]+CoeffMatrix[c, i]*x^n*y^m end if; c := c+1 end do end do end do else return "3D nodal basis functions not yet available!!!"; numberOfBasisFunctions2D := dunavantOrderNum[i]; nodalCoords2D := identify(dunavantCoords[i]); numberOfBasisFunctions := numberOfBasisFunctions2D*numberOf1DGaussSamplingPoints; nodalCoords3D := Matrix(numberOfBasisFunctions, 3, proc (i, j) options operator, arrow; 0 end proc); c := 1; for i to numberOfBasisFunctions2D do for j to numberOf1DGaussSamplingPoints do nodalCoords3D[c, 1] := nodalCoords2D[i, 1]; nodalCoords3D[c, 2] := nodalCoords2D[i, 2]; nodalCoords3D[c, 3] := samplingPoints0to1[j]; c := c+1 end do end do; nodalCoords := nodalCoords3D; NodalMatrix := Matrix(numberOfBasisFunctions, numberOfBasisFunctions, proc (i, j) options operator, arrow; 0 end proc); c := 1; for i to numberOfBasisFunctions do c := 1; for m from 0 to degreeOfBasisFunctions do for n from 0 to degreeOfBasisFunctions-m do for o from 0 to degreeOfBasisFunctions-m-n do if c <= numberOfBasisFunctions then NodalMatrix[i, c] := nodalCoords3D[i][1]^m*nodalCoords3D[i][2]^n*nodalCoords3D[i][3]^o end if; c := c+1 end do end do end do end do; CoeffMatrix := 1/NodalMatrix; basisFunctionArray := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); for i to numberOfBasisFunctions do c := 1; for m from 0 to degreeOfBasisFunctions do for n from 0 to degreeOfBasisFunctions-m do for o from 0 to degreeOfBasisFunctions-m-n do if c <= numberOfBasisFunctions then basisFunctionArray[i] := basisFunctionArray[i]+CoeffMatrix[c, i]*x^m*y^n*z^o end if; c := c+1 end do end do end do end do end if; if numberOfDimensions = 2 then A := simplify(Matrix(numberOfBasisFunctions, numberOfBasisFunctions, proc (i, j) options operator, arrow; (unapply(basisFunctionArray[j], x, y))(nodalCoords2D[i, 1], nodalCoords2D[i, 2]) end proc)) else A := simplify(Matrix(numberOfBasisFunctions, numberOfBasisFunctions, proc (i, j) options operator, arrow; (unapply(basisFunctionArray[j], x, y, z))(nodalCoords3D[i, 1], nodalCoords3D[i, 2], nodalCoords3D[i, 3]) end proc)) end if; A; nodalCoords2D end if;
# 0: Nodal, Gauss-Lobatto;
if configBasisFunctionType = 0 then basisFunctionDescription := "Gauss-Lobatto nodal basis"; numberOfBasisFunctions2D := (1/2)*(configDegreeOfBasisFunctions+1)*(configDegreeOfBasisFunctions+2); n := GaussLobattoPoints:-Nodes2D(configDegreeOfBasisFunctions); k := simplify([GaussLobattoPoints:-xytoxieta(n[1], n[2])]); nodalCoords2D := [seq([k[1][i], k[2][i]], i = 1 .. numberOfBasisFunctions2D)] end if;
if `and`(configBasisFunctionType = 0, configNumberOfDimensions = 2) then N := getVandermondeMatrix2D(nodalCoords2D, configDegreeOfBasisFunctions); basisFunctionArray := simplify(getBasisFunctionsFromVandermondeMatrix2D(N, configDegreeOfBasisFunctions)); numberOfBasisFunctions := numberOfBasisFunctions2D; nodalCoords := nodalCoords2D; A := simplify(Matrix(numberOfBasisFunctions, numberOfBasisFunctions, proc (i, j) options operator, arrow; (unapply(basisFunctionArray[j], x, y))(nodalCoords2D[i, 1], nodalCoords2D[i, 2]) end proc)); ASSERT(LinearAlgebra[Equal](A, LinearAlgebra[IdentityMatrix](numberOfBasisFunctions, numberOfBasisFunctions)), "CHECK FAILED") end if;
if `and`(configBasisFunctionType = 0, configNumberOfDimensions = 3) then N := getVandermondeMatrix2D(nodalCoords2D, configDegreeOfBasisFunctions); basisFunctionArray2D := getBasisFunctionsFromVandermondeMatrix2D(N, configDegreeOfBasisFunctions); M1D := Matrix(configDegreeOfBasisFunctions+1, configDegreeOfBasisFunctions+1, proc (i, j) options operator, arrow; quadGL1DSamplingPoints[i]^(j-1) end proc); basisFunctionArray1D := getBasisFunctionsFromVandermondeMatrix1D(M1D, configDegreeOfBasisFunctions); numberOfBasisFunctions := numberOfBasisFunctions2D*quadGL1DNumberOfSamplingPoints; nodalCoords3D := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); basisFunctionArray := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); c := 1; for j to quadGL1DNumberOfSamplingPoints do for i to numberOfBasisFunctions2D do nodalCoords3D[c] := [quadGL2DSamplingPoints[i, 1], quadGL2DSamplingPoints[i, 2], quadGL1DSamplingPoints[j]]; basisFunctionArray[c] := simplify(expand(basisFunctionArray2D[i]*basisFunctionArray1D[j])); c := c+1 end do end do; nodalCoords := nodalCoords3D; A := simplify(Matrix(numberOfBasisFunctions, numberOfBasisFunctions, proc (i, j) options operator, arrow; (unapply(basisFunctionArray[j], x, y, z))(nodalCoords3D[i][1], nodalCoords3D[i][2], nodalCoords3D[i][3]) end proc)) end if;
# 1; JACOBI*POLYNOMIALS;
if configBasisFunctionType = 1 then basisFunctionDescription := "using jacobi basis functions" end if;
if `and`(configBasisFunctionType = 1, configNumberOfDimensions = 2) then numberOfBasisFunctions := (1/2)*(configDegreeOfBasisFunctions+1)*(configDegreeOfBasisFunctions+2); c := 1; basisFunctionArray := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions-m do Pp := proc (m, a, b, x) options operator, arrow; P(m, a, b, 2*x-1) end proc; basis := Pp(m, 0, 0, x/(1-y))*(1-y)^m*Pp(n, 2*m+1, 0, y); normalize := 1/sqrt(int(basis*basis, Omega)); basisFunctionArray[c] := simplify(basis*normalize); c := c+1 end do end do end if;
if `and`(configBasisFunctionType = 1, configNumberOfDimensions = 3) then numberOfBasisFunctions := (configDegreeOfBasisFunctions+1)*(configDegreeOfBasisFunctions+2)*(configDegreeOfBasisFunctions+3)/(2*3); c := 1; basisFunctionArray := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions-m do for o from 0 to configDegreeOfBasisFunctions-m-n do Pp := proc (m, a, b, x) options operator, arrow; P(m, a, b, 2*x-1) end proc; basis := Pp(m, 0, 0, x/(1-y))*(1-y)^m*Pp(n, 2*m+1, 0, y)*Pp(o, 0, 0, z); normalize := 1/sqrt(int(basis*basis, Omega)); basisFunctionArray[c] := simplify(basis*normalize); c := c+1 end do end do end do end if;
# 2: MONOMIALS
if `and`(configBasisFunctionType = 2, configNumberOfDimensions = 2) then numberOfBasisFunctions := (1/2)*(configDegreeOfBasisFunctions+1)*(configDegreeOfBasisFunctions+2); c := 1; basisFunctionArray := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions-m do basis := x^m*y^n; normalize := 1; basisFunctionArray[c] := simplify(basis*normalize); c := c+1 end do end do end if;
if `and`(configBasisFunctionType = 2, configNumberOfDimensions = 3) then numberOfBasisFunctions := (configDegreeOfBasisFunctions+1)*(configDegreeOfBasisFunctions+2)*(configDegreeOfBasisFunctions+3)/(2*3); c := 1; basisFunctionArray := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions-m do for o from 0 to configDegreeOfBasisFunctions-m-n do basis := x^m*y^n*z^o; normalize := 1/sqrt(int(basis*basis, Omega)); basisFunctionArray[c] := simplify(basis*normalize); c := c+1 end do end do end do end if;
# Show basis functions
print("Basis function description", basisFunctionDescription); print("Basis functions: ", basisFunctionArray); print("Degree of basis functions: ", configDegreeOfBasisFunctions);
# Computation of DG matrices:
# chlid basis functions, domain specific settings, etc.
phi := proc (id) options operator, arrow; unapply(basisFunctionArray[id], basisFunctionParameters) end proc; phin := ArrayNumElems(basisFunctionArray); if numberOfDimensions = 2 then leftChildOmega1 := [y = 0 .. x, x = 0 .. 1/2]; leftChildOmega2 := [y = 0 .. 1-x, x = 1/2 .. 1]; intLeftChild := proc (poly) options operator, arrow; int(poly, leftChildOmega1)+int(poly, leftChildOmega2) end proc; rightChildOmega1 := [y = x .. 1/2, x = 0 .. 1/2]; rightChildOmega2 := [y = 1/2 .. 1-x, x = 1/2 .. 1]; intRightChild := proc (poly) options operator, arrow; int(poly, rightChildOmega1)+int(poly, rightChildOmega2) end proc else leftChildOmega1 := [y = 0 .. x, x = 0 .. 1/2, z = 0 .. 1]; leftChildOmega2 := [y = 0 .. 1-x, x = 1/2 .. 1, z = 0 .. 1]; intLeftChild := proc (poly) options operator, arrow; int(poly, leftChildOmega1)+int(poly, leftChildOmega2) end proc; rightChildOmega1 := [y = x .. 1/2, x = 0 .. 1/2, z = 0 .. 1]; rightChildOmega2 := [y = 1/2 .. 1-x, x = 1/2 .. 1, z = 0 .. 1]; intRightChild := proc (poly) options operator, arrow; int(poly, rightChildOmega1)+int(poly, rightChildOmega2) end proc end if;
# Matrices for single timestep:
OUTMMass := simplify(Matrix(phin, phin, proc (i, j) options operator, arrow; simplify(int(simplify((phi(j))(basisFunctionParameters)*(phi(i))(basisFunctionParameters)), Omega)) end proc));
NULL;
OUTMInvMass := kramerInverse(OUTMMass);
# Stiffness matrix for single timestep:
# 
;
if configNumberOfDimensions = 2 then OUTMStiffnessToGLNodes := simplify(Matrix(quadGL2DNumberOfSamplingPoints, phin, proc (i, j) options operator, arrow; simplify((phi(j))(quadGL2DSamplingPoints[i][1], quadGL2DSamplingPoints[i][2])) end proc)); OUTMStiffnessU := simplify(Matrix(phin, quadGL2DNumberOfSamplingPoints, proc (i, j) options operator, arrow; simplify(int(simplify((quadGL2DPhi(j))(basisFunctionParameters)*(diff((phi(i))(basisFunctionParameters), x))), Omega)) end proc)); OUTMStiffnessV := simplify(Matrix(phin, quadGL2DNumberOfSamplingPoints, proc (i, j) options operator, arrow; simplify(int(simplify((quadGL2DPhi(j))(basisFunctionParameters)*(diff((phi(i))(basisFunctionParameters), y))), Omega)) end proc)) end if;
if configNumberOfDimensions = 3 then quadGL3DNumberOfBasisFunctions := quadGL1DNumberOfSamplingPoints*quadGL2DNumberOfSamplingPoints; quadGL3DSamplingPoints := Vector(quadGL3DNumberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); quadGL3DBasisFunctionArray := Vector(quadGL3DNumberOfBasisFunctions, proc (i) options operator, arrow; 0 end proc); c := 1; for j to quadGL1DNumberOfSamplingPoints do for i to quadGL2DNumberOfSamplingPoints do quadGL3DSamplingPoints[c] := [quadGL2DSamplingPoints[i, 1], quadGL2DSamplingPoints[i, 2], quadGL1DSamplingPoints[j]]; quadGL3DBasisFunctionArray[c] := simplify(expand((quadGL2DPhi(i))(x, y)*(quadGL1DPhi(j))(z))); c := c+1 end do end do; OUTMStiffnessToGLNodes := simplify(Matrix(phin, phin, proc (i, j) options operator, arrow; simplify((phi(j))(quadGL3DSamplingPoints[i][1], quadGL3DSamplingPoints[i][2], quadGL3DSamplingPoints[i][3])) end proc)); OUTMStiffnessU := simplify(Matrix(phin, phin, proc (i, j) options operator, arrow; simplify(int(simplify((phi(j))(basisFunctionParameters)*(diff((phi(i))(basisFunctionParameters), x))), Omega)) end proc)); OUTMStiffnessV := simplify(Matrix(phin, phin, proc (i, j) options operator, arrow; simplify(int(simplify((phi(j))(basisFunctionParameters)*(diff((phi(i))(basisFunctionParameters), y))), Omega)) end proc)); OUTMStiffnessW := simplify(Matrix(phin, phin, proc (i, j) options operator, arrow; simplify(int(simplify((phi(j))(basisFunctionParameters)*(diff((phi(i))(basisFunctionParameters), z))), Omega)) end proc)) end if;
# Flux Matrices, Edges & Quad-Faces:
# Setup for Edge Quadrature
if configUseGaussQuadratureRules = 1 then quadEdgeNumberOfSamplingPoints := quadGauss1DNumberOfSamplingPoints; quadEdgeSamplingPoints := quadGauss1DSamplingPoints; quadEdgeWeights := quadGauss1DWeights; quadEdgePhi := quadGauss1DPhi else quadEdgeNumberOfSamplingPoints := quadGL1DNumberOfSamplingPoints; quadEdgeSamplingPoints := quadGL1DSamplingPoints; quadEdgeWeights := quadGL1DWeights; quadEdgePhi := quadGL1DPhi end if; quadEdgeHypSamplingPoints := seq([1-x, x], `in`(x, quadEdgeSamplingPoints)); quadEdgeRightSamplingPoints := seq([0, 1-x], `in`(x, quadEdgeSamplingPoints)); quadEdgeLeftSamplingPoints := seq([x, 0], `in`(x, quadEdgeSamplingPoints));
# Setup for Face quadrature
if configNumberOfDimensions = 3 then quadFaceNumberOfSamplingPoints := quadEdgeNumberOfSamplingPoints*quadEdgeNumberOfSamplingPoints; quadFaceHypSamplingPoints := Vector(quadFaceNumberOfSamplingPoints, proc (i) options operator, arrow; pi end proc); quadFaceRightSamplingPoints := Vector(quadFaceNumberOfSamplingPoints, proc (i) options operator, arrow; pi end proc); quadFaceLeftSamplingPoints := Vector(quadFaceNumberOfSamplingPoints, proc (i) options operator, arrow; pi end proc); quadFaceWeights := Vector(quadFaceNumberOfSamplingPoints, proc (i) options operator, arrow; pi end proc); quadFacePhi := Vector(quadFaceNumberOfSamplingPoints, proc (i) options operator, arrow; pi end proc); c := 1; for i to quadEdgeNumberOfSamplingPoints do for j to quadEdgeNumberOfSamplingPoints do quadFaceHypSamplingPoints[c] := [quadEdgeHypSamplingPoints[i][1], quadEdgeHypSamplingPoints[i][2], quadEdgeSamplingPoints[j]]; quadFaceRightSamplingPoints[c] := [quadEdgeRightSamplingPoints[i][1], quadEdgeRightSamplingPoints[i][2], quadEdgeSamplingPoints[j]]; quadFaceLeftSamplingPoints[c] := [quadEdgeLeftSamplingPoints[i][1], quadEdgeLeftSamplingPoints[i][2], quadEdgeSamplingPoints[j]]; quadFaceWeights[c] := quadEdgeWeights[i]*quadEdgeWeights[j]; quadFacePhiX[c] := (quadEdgePhi(i))(a)*(quadEdgePhi(j))(b); c := c+1 end do end do; quadFacePhi := Vector(quadFaceNumberOfSamplingPoints, proc (i) options operator, arrow; unapply(eval(quadFacePhiX[i]), a, b) end proc); quadFaceHypSamplingPoints; quadFaceRightSamplingPoints; quadFaceLeftSamplingPoints; quadFaceWeights; quadFacePhi end if;
if configNumberOfDimensions = 2 then quadEdgeOrFaceNumberOfSamplingPoints := quadEdgeNumberOfSamplingPoints else quadEdgeOrFaceNumberOfSamplingPoints := quadFaceNumberOfSamplingPoints end if;
# 2D DOFs to Face
if configNumberOfDimensions = 2 then OUTMDofsToHypFace := Matrix(quadEdgeNumberOfSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; simplify(apply(phi(j), quadEdgeHypSamplingPoints[i][1], quadEdgeHypSamplingPoints[i][2])) end proc); OUTMDofsToRightFace := Matrix(quadEdgeNumberOfSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; simplify(apply(phi(j), quadEdgeRightSamplingPoints[i][1], quadEdgeRightSamplingPoints[i][2])) end proc); OUTMDofsToLeftFace := Matrix(quadEdgeNumberOfSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; simplify(apply(phi(j), quadEdgeLeftSamplingPoints[i][1], quadEdgeLeftSamplingPoints[i][2])) end proc) end if;
# 3D DOFs to Face
if configNumberOfDimensions = 3 then OUTMDofsToHypFace := Matrix(quadEdgeOrFaceNumberOfSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; simplify(apply(phi(j), quadFaceHypSamplingPoints[i][1], quadFaceHypSamplingPoints[i][2], quadFaceHypSamplingPoints[i][3])) end proc); OUTMDofsToRightFace := Matrix(quadEdgeOrFaceNumberOfSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; simplify(apply(phi(j), quadFaceRightSamplingPoints[i][1], quadFaceRightSamplingPoints[i][2], quadFaceHypSamplingPoints[i][3])) end proc); OUTMDofsToLeftFace := Matrix(quadEdgeOrFaceNumberOfSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; simplify(apply(phi(j), quadFaceLeftSamplingPoints[i][1], quadFaceLeftSamplingPoints[i][2], quadFaceHypSamplingPoints[i][3])) end proc) end if;
# 2D Flux to DOFs
# 
if configNumberOfDimensions = 2 then OUTMHypFluxesToDofs := Matrix(numberOfBasisFunctions, quadEdgeNumberOfSamplingPoints, proc (i, j) options operator, arrow; simplify(-(int((quadEdgePhi(j))(u)*(phi(i))(1-u, u), u = 0 .. 1))*sqrt(2)) end proc); OUTMRightFluxesToDofs := Matrix(numberOfBasisFunctions, quadEdgeNumberOfSamplingPoints, proc (i, j) options operator, arrow; simplify(-(int((quadEdgePhi(j))(u)*(phi(i))(0, 1-u), u = 0 .. 1))) end proc); OUTMLeftFluxesToDofs := Matrix(numberOfBasisFunctions, quadEdgeNumberOfSamplingPoints, proc (i, j) options operator, arrow; simplify(-(int((quadEdgePhi(j))(u)*(phi(i))(u, 0), u = 0 .. 1))) end proc) end if;
# 3D Flux to DOFs:
if configNumberOfDimensions = 3 then OUTMHypFluxesToDofs := Matrix(numberOfBasisFunctions, quadFaceNumberOfSamplingPoints, proc (i, j) options operator, arrow; simplify(-(int((quadFacePhi(j))(inta, intb)*(phi(i))(1-inta, inta, intb), inta = 0 .. 1, intb = 0 .. 1))*sqrt(2)) end proc); OUTMRightFluxesToDofs := Matrix(numberOfBasisFunctions, quadFaceNumberOfSamplingPoints, proc (i, j) options operator, arrow; simplify(-(int((quadFacePhi(j))(inta, intb)*(phi(i))(0, 1-inta, intb), inta = 0 .. 1, intb = 0 .. 1))) end proc); OUTMLeftFluxesToDofs := Matrix(numberOfBasisFunctions, quadFaceNumberOfSamplingPoints, proc (i, j) options operator, arrow; simplify(-(int((quadFacePhi(j))(inta, intb)*(phi(i))(inta, 0, intb), inta = 0 .. 1, intb = 0 .. 1))) end proc) end if;
# Show
OUTMDofsToHypFace; OUTMDofsToRightFace; OUTMDofsToLeftFace;
evalf(OUTMHypFluxesToDofs); evalf(OUTMRightFluxesToDofs); evalf(OUTMLeftFluxesToDofs);
# TOP/BOTTOM Flux Matrices, Triangle Faces:
getNodalDenseDofMasks := proc (fluxToDofsMatrix, numberOfSamplingPoints) local dofsMask, scanMask, i, c; dofsMask := Vector(numberOfBasisFunctions, proc (i) options operator, arrow; eval(evalb(0 < add(fluxToDofsMatrix[j, i]*fluxToDofsMatrix[j, i], j = 1 .. numberOfBasisFunctions)), [true = 1, false = 0]) end proc); scanMask := Vector(numberOfSamplingPoints); c := 1; for i to numberOfBasisFunctions do if 0 < dofsMask[i] then scanMask[c] := i; c := c+1 end if end do; return scanMask end proc;
if configNumberOfDimensions = 3 then numberOfTopBottomFaceNodalSamplingPoints := numberOfBasisFunctions2D end if;
if configNumberOfDimensions = 3 then OUTMDofsToTopFace := Matrix(quadGL2DNumberOfSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; Pi end proc); OUTMDofsToBottomFace := Matrix(quadGL2DNumberOfSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; Pi end proc); for i to quadGL2DNumberOfSamplingPoints do for j to numberOfBasisFunctions do OUTMDofsToTopFace[i, j] := simplify(apply(phi(j), quadGL2DSamplingPoints[i][1], quadGL2DSamplingPoints[i][2], 1)); OUTMDofsToBottomFace[i, j] := simplify(apply(phi(j), quadGL2DSamplingPoints[i][1], quadGL2DSamplingPoints[i][2], 0)) end do end do end if;
if `and`(`and`(configNumberOfDimensions = 3, configBasisFunctionType = 0), false) then TopFluxesToDofs := Matrix(numberOfBasisFunctions, numberOfBasisFunctions, proc (i, j) options operator, arrow; simplify(int((phi(i))(x, y, 1)*(phi(j))(x, y, 1), Omega)) end proc); BottomFluxesToDofs := Matrix(numberOfBasisFunctions, numberOfBasisFunctions, proc (i, j) options operator, arrow; simplify(int((phi(i))(x, y, 0)*(phi(j))(x, y, 0), Omega)) end proc); NodalScanMaskTop := getNodalDenseDofMasks(TopFluxesToDofs, numberOfTopBottomFaceNodalSamplingPoints); NodalScanMaskBottom := getNodalDenseDofMasks(BottomFluxesToDofs, numberOfTopBottomFaceNodalSamplingPoints); OUTMDofsToTopFace := Matrix(numberOfTopBottomFaceNodalSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; eval(evalb(NodalScanMaskTop[i] = j), [true = 1, false = 0]) end proc); OUTMDofsToBottomFace := Matrix(numberOfTopBottomFaceNodalSamplingPoints, numberOfBasisFunctions, proc (i, j) options operator, arrow; eval(evalb(NodalScanMaskBottom[i] = j), [true = 1, false = 0]) end proc) end if;
if `and`(`and`(configNumberOfDimensions = 3, configBasisFunctionType <> 0), false) then d := (1/2)*configDegreeOfBasisFunctions*(configDegreeOfBasisFunctions+1); orderNum := identify(Dunavant:-getDunavantOrderNum(d+1)); coords := identify(Dunavant:-getDunavantCoords(d+1)); OUTMDofsToTopFace := Matrix(orderNum, numberOfBasisFunctions, proc (i, j) options operator, arrow; Pi end proc); OUTMDofsToBottomFace := Matrix(orderNum, numberOfBasisFunctions, proc (i, j) options operator, arrow; Pi end proc); for i to orderNum do for j to numberOfBasisFunctions do OUTMDofsToTopFace[i, j] := simplify(apply(phi(j), coords[i][1], coords[i][2], 0)); OUTMDofsToBottomFace[i, j] := simplify(apply(phi(j), coords[i][1], coords[i][2], 1)) end do end do end if;
evalf(OUTMDofsToTopFace); evalf(OUTMDofsToBottomFace);
if configNumberOfDimensions = 3 then OUTMTopFluxesToDofs := Matrix(numberOfBasisFunctions, quadGL2DNumberOfSamplingPoints, proc (i, j) options operator, arrow; Pi end proc); OUTMBottomFluxesToDofs := Matrix(numberOfBasisFunctions, quadGL2DNumberOfSamplingPoints, proc (i, j) options operator, arrow; Pi end proc); for i to numberOfBasisFunctions do for j to quadGL2DNumberOfSamplingPoints do OUTMTopFluxesToDofs[i, j] := simplify(-(int((quadGL2DPhi(j))(x, y)*(phi(i))(x, y, 1), OmegaTriangle))); OUTMBottomFluxesToDofs[i, j] := simplify(-(int((quadGL2DPhi(j))(x, y)*(phi(i))(x, y, 0), OmegaTriangle))) end do end do; OUTMTopFluxesToDofs; OUTMBottomFluxesToDofs end if;
if `and`(`and`(configNumberOfDimensions = 3, configBasisFunctionType = 0), false) then OUTMTopFluxesToDofs := Matrix(numberOfBasisFunctions, numberOfTopBottomFaceNodalSamplingPoints, proc (i, j) options operator, arrow; -TopFluxesToDofs[i, NodalScanMaskTop[j]] end proc); OUTMBottomFluxesToDofs := Matrix(numberOfBasisFunctions, numberOfTopBottomFaceNodalSamplingPoints, proc (i, j) options operator, arrow; -BottomFluxesToDofs[i, NodalScanMaskBottom[j]] end proc) end if;
if `and`(`and`(configNumberOfDimensions = 3, configBasisFunctionType <> 0), false) then OUTMTopFluxesToDofs := Matrix(numberOfBasisFunctions, quadGauss2DNumberOfSamplingPoints, proc (i, j) options operator, arrow; Pi end proc); OUTMBottomFluxesToDofs := Matrix(numberOfBasisFunctions, quadGauss2DNumberOfSamplingPoints, proc (i, j) options operator, arrow; Pi end proc); for i to numberOfBasisFunctions do for j to quadGauss2DNumberOfSamplingPoints do OUTMTopFluxesToDofs[i, j] := simplify(-apply(phi(i), quadGauss2DSamplingPoints[j][1], quadGauss2DSamplingPoints[j][2], 0)*quadGauss2DWeights[j]); OUTMBottomFluxesToDofs[i, j] := simplify(-apply(phi(i), quadGauss2DSamplingPoints[j][1], quadGauss2DSamplingPoints[j][2], 1)*quadGauss2DWeights[j]) end do end do end if;
# Adaptivity Matrices:
# 
# Helper Functions
rotate := proc (x, y, angle) options operator, arrow; cos(angle)*x-sin(angle)*y, sin(angle)*x+cos(angle)*y end proc; rotationMatrix := proc (angle) options operator, arrow; Matrix(3, 3, [cos(angle), -sin(angle), 0, sin(angle), cos(angle), 0, 0, 0, 1]) end proc; VCoeffs := Vector(3, [x, y, 1]);
# Left child basis functions:
MTranslate := Matrix(3, 3, [1, 0, -1/2, 0, 1, -1/2, 0, 0, 1]); MScale := Matrix(3, 3, [sqrt(2), 0, 0, 0, sqrt(2), 0, 0, 0, 1]); leftToParentTransform := Physics:-`.`(1/Physics:-`.`(Physics:-`.`(rotationMatrix((3/4)*Pi), MScale), MTranslate), VCoeffs); parentToLeftTransform := Physics:-`.`(Physics:-`.`(Physics:-`.`(rotationMatrix((3/4)*Pi), MScale), MTranslate), VCoeffs); if numberOfDimensions = 2 then leftToParentPhi := proc (i) options operator, arrow; unapply((phi(i))(leftToParentTransform[1], leftToParentTransform[2]), basisFunctionParameters) end proc; parentToLeftPhi := proc (i) options operator, arrow; unapply((phi(i))(parentToLeftTransform[1], parentToLeftTransform[2]), basisFunctionParameters) end proc else leftToParentPhi := proc (i) options operator, arrow; unapply((phi(i))(leftToParentTransform[1], leftToParentTransform[2], z), basisFunctionParameters) end proc; parentToLeftPhi := proc (i) options operator, arrow; unapply((phi(i))(parentToLeftTransform[1], parentToLeftTransform[2], z), basisFunctionParameters) end proc end if;
ASSERT(LinearAlgebra[Equal]((unapply(parentToLeftTransform, x, y))(0, 0), Vector([1, 0, 1])), "CHECK FAILED"); ASSERT(LinearAlgebra[Equal]((unapply(parentToLeftTransform, x, y))(1, 0), Vector([0, 1, 1])), "CHECK FAILED"); ASSERT(LinearAlgebra[Equal]((unapply(parentToLeftTransform, x, y))(1/2, 1/2), Vector([0, 0, 1])), "CHECK FAILED");
# Right Child basis functions:
MTranslate := Matrix(3, 3, [1, 0, -1/2, 0, 1, -1/2, 0, 0, 1]); MScale := Matrix(3, 3, [sqrt(2), 0, 0, 0, sqrt(2), 0, 0, 0, 1]); VCoeffs := Vector(3, [x, y, 1]); rightToParentTransform := Physics:-`.`(1/Physics:-`.`(Physics:-`.`(rotationMatrix(-(3/4)*Pi), MScale), MTranslate), VCoeffs); parentToRightTransform := Physics:-`.`(Physics:-`.`(Physics:-`.`(rotationMatrix(-(3/4)*Pi), MScale), MTranslate), VCoeffs); if numberOfDimensions = 2 then rightToParentPhi := proc (i) options operator, arrow; unapply((phi(i))(rightToParentTransform[1], rightToParentTransform[2]), basisFunctionParameters) end proc; parentToRightPhi := proc (i) options operator, arrow; unapply((phi(i))(parentToRightTransform[1], parentToRightTransform[2]), basisFunctionParameters) end proc else rightToParentPhi := proc (i) options operator, arrow; unapply((phi(i))(rightToParentTransform[1], rightToParentTransform[2], z), basisFunctionParameters) end proc; parentToRightPhi := proc (i) options operator, arrow; unapply((phi(i))(parentToRightTransform[1], parentToRightTransform[2], z), basisFunctionParameters) end proc end if;
ASSERT(LinearAlgebra[Equal]((unapply(parentToRightTransform, x, y))(0, 0), Vector([0, 1, 1])), "CHECK FAILED"); ASSERT(LinearAlgebra[Equal]((unapply(parentToRightTransform, x, y))(0, 1), Vector([1, 0, 1])), "CHECK FAILED"); ASSERT(LinearAlgebra[Equal]((unapply(parentToRightTransform, x, y))(1/2, 1/2), Vector([0, 0, 1])), "CHECK FAILED");
# Setup system of equations in Matrix Form:
# Each line represents a monomial coefficient for all basis functions
# 
# Multiplying this matrix with the basis function weights leads to the polynomial representing the sum of all weighted basis function.
getEqMatrix := proc (basisFunctions) local i, c, m, n, o, M; M := Matrix(phin, phin, proc (i, j) options operator, arrow; Pi end proc); for i to phin do c := 1; if configNumberOfDimensions = 2 then for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions-m do if c <= phin then M[c, i] := coeff(coeff((basisFunctions(i))(basisFunctionParameters), x, n), y, m) end if; c := c+1 end do end do else if configBasisFunctionType = 1 or configBasisFunctionType = 2 then for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions-m do for o from 0 to configDegreeOfBasisFunctions-m-n do if c <= phin then M[c, i] := coeff(coeff(coeff((basisFunctions(i))(basisFunctionParameters), x, o), y, n), z, m) end if; c := c+1 end do end do end do else for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions do for o from 0 to configDegreeOfBasisFunctions-n do if c <= phin then M[c, i] := coeff(coeff(coeff((basisFunctions(i))(basisFunctionParameters), x, o), y, n), z, m) end if; c := c+1 end do end do end do end if end if end do; return simplify(M) end proc;
xgetIntMonomialTimesPhiMatrix := proc (basisFunctions) local i, c, m, n, o, M; M := Matrix(phin, phin, proc (i, j) options operator, arrow; Pi end proc); for i to phin do c := 1; if configNumberOfDimensions = 2 then for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions-m do if c <= phin then M[i, c] := int((unapply(x^n*y^m, x, y)*basisFunctions(i))(basisFunctionParameters), Omega) end if; c := c+1 end do end do else if configBasisFunctionType = 1 or configBasisFunctionType = 2 then for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions-m do for o from 0 to configDegreeOfBasisFunctions-m-n do if c <= phin then M[i, c] := int((unapply(x^o*y^n*z^m, x, y, z)*basisFunctions(i))(basisFunctionParameters), Omega) end if; c := c+1 end do end do end do else for m from 0 to configDegreeOfBasisFunctions do for n from 0 to configDegreeOfBasisFunctions do for o from 0 to configDegreeOfBasisFunctions-n do if c <= phin then M[i, c] := int((unapply(x^o*y^n*z^m, x, y, z)*basisFunctions(i))(basisFunctionParameters), Omega) end if; c := c+1 end do end do end do end if end if end do; return simplify(M) end proc;
# Now we compute the matrices which relate the different 'weights' to the polynomial coefficients.
# Each line represents a monomial.
MParent := getEqMatrix(phi);
MParent := getEqMatrix(phi); MLeftToParentChild := getEqMatrix(leftToParentPhi); MRightToParentChild := getEqMatrix(rightToParentPhi); MParentToLeftChild := getEqMatrix(parentToLeftPhi); MParentToRightChild := getEqMatrix(parentToRightPhi);
MParentInv := kramerInverse(MParent); MParentToLeftChildInv := kramerInverse(MParentToLeftChild); MParentToRightChildInv := kramerInverse(MParentToRightChild);
OUTMAdaptiveRefineLeft := simplify(Physics:-`.`(MParentToLeftChildInv, MParent)); OUTMAdaptiveRefineRight := simplify(Physics:-`.`(MParentToRightChildInv, MParent));
# Next, we compute the coarsening matrices
# At first, we compute the parent projection for both childs
OUTMAdaptiveCoarsenLeft := simplify((1/2)*Physics:-`.`(MParentInv, MParentToLeftChild)); OUTMAdaptiveCoarsenRight := simplify((1/2)*Physics:-`.`(MParentInv, MParentToRightChild));
ASSERT(LinearAlgebra[Equal](simplify(Physics:-`.`(OUTMAdaptiveCoarsenLeft, OUTMAdaptiveRefineLeft)+Physics:-`.`(OUTMAdaptiveCoarsenRight, OUTMAdaptiveRefineRight)), LinearAlgebra[IdentityMatrix](phin, phin)), "CHECK FAILED");
simplify(Physics:-`.`(Physics:-`.`(Physics:-`.`(MParent, OUTMAdaptiveRefineLeft), MParentInv), Vector(phin, proc (i) options operator, arrow; KroneckerDelta[i, 2] end proc)));
simplify(Physics:-`.`(Physics:-`.`(Physics:-`.`(MParent, OUTMAdaptiveRefineRight), MParentInv), Vector(phin, proc (i) options operator, arrow; KroneckerDelta[i, 2] end proc)));
# Helper functions to convert Polynomials to DOFs and vice versa
MIntMonomialTimesPhi := getIntMonomialTimesPhiMatrix(phi); MInverseIntMonomialTimesPhi := 1/MIntMonomialTimesPhi;
# Get Surface Polynomial from DOFs
OUTMConvDofsToPoly := MParent;
# Get DOFs from Polynomial
OUTMConvPolyToDofs := MParentInv;
# Validate matrices
# 
m := Vector(phin, proc (i) options operator, arrow; i*i end proc); dofs := simplify(Physics:-`.`(OUTMConvPolyToDofs, m)); identify(evalf(simplify(expand(add(dofs[k]*(phi(k))(basisFunctionParameters), k = 1 .. phin)))));
# Get 0st order representation
OUTMConvDofsTo0stOrder := Matrix(1, phin, proc (i, j) options operator, arrow; int(2*(phi(j))(basisFunctionParameters), Omega) end proc);
# Export Matrices
# Filename prefix:
exportMatrixMarket := proc (filename, matrix) ExportMatrix(filename, matrix, target = MatrixMarket) end proc;
if configOutputDir <> "" then if `not`(Exists(configOutputDir)) then mkdir(configOutputDir) end if; dirName := cat(configOutputDir, "type_", configBasisFunctionType, "_degree_", configDegreeOfBasisFunctions); if `not`(Exists(dirName)) then mkdir(dirName) end if; dirName := cat(dirName, "/"); exportMatrixMarket(cat(dirName, "timestep_inv_mass.mtx"), evalf(OUTMInvMass)); exportMatrixMarket(cat(dirName, "timestep_stiffness_dofs_to_nodes.mtx"), evalf(OUTMStiffnessToGLNodes)); exportMatrixMarket(cat(dirName, "timestep_stiffness_u.mtx"), evalf(OUTMStiffnessU)); exportMatrixMarket(cat(dirName, "timestep_stiffness_v.mtx"), evalf(OUTMStiffnessV)); if configNumberOfDimensions = 3 then exportMatrixMarket(cat(dirName, "timestep_stiffness_w.mtx"), evalf(OUTMStiffnessW)) end if; "edge faces"; exportMatrixMarket(cat(dirName, "timestep_dofs_to_hyp_face.mtx"), evalf(OUTMDofsToHypFace)); exportMatrixMarket(cat(dirName, "timestep_dofs_to_right_face.mtx"), evalf(OUTMDofsToRightFace)); exportMatrixMarket(cat(dirName, "timestep_dofs_to_left_face.mtx"), evalf(OUTMDofsToLeftFace)); exportMatrixMarket(cat(dirName, "timestep_hyp_fluxes_to_dofs.mtx"), evalf(OUTMHypFluxesToDofs)); exportMatrixMarket(cat(dirName, "timestep_right_fluxes_to_dofs.mtx"), evalf(OUTMRightFluxesToDofs)); exportMatrixMarket(cat(dirName, "timestep_left_fluxes_to_dofs.mtx"), evalf(OUTMLeftFluxesToDofs)); if configNumberOfDimensions = 3 then "top and bottom face"; exportMatrixMarket(cat(dirName, "timestep_dofs_to_top_face.mtx"), evalf(OUTMDofsToTopFace)); exportMatrixMarket(cat(dirName, "timestep_dofs_to_bottom_face.mtx"), evalf(OUTMDofsToBottomFace)); exportMatrixMarket(cat(dirName, "timestep_top_fluxes_to_dofs.mtx"), evalf(OUTMTopFluxesToDofs)); exportMatrixMarket(cat(dirName, "timestep_bottom_fluxes_to_dofs.mtx"), evalf(OUTMBottomFluxesToDofs)) end if; "adaptive"; exportMatrixMarket(cat(dirName, "adaptivity_refine_left.mtx"), evalf(OUTMAdaptiveRefineLeft)); exportMatrixMarket(cat(dirName, "adaptivity_refine_right.mtx"), evalf(OUTMAdaptiveRefineRight)); exportMatrixMarket(cat(dirName, "adaptivity_coarsen_left.mtx"), evalf(OUTMAdaptiveCoarsenLeft)); exportMatrixMarket(cat(dirName, "adaptivity_coarsen_right.mtx"), evalf(OUTMAdaptiveCoarsenRight)); "conversion matrices"; exportMatrixMarket(cat(dirName, "convert_poly_to_dofs.mtx"), evalf(OUTMConvPolyToDofs)); exportMatrixMarket(cat(dirName, "convert_dofs_to_poly.mtx"), evalf(OUTMConvDofsToPoly)); exportMatrixMarket(cat(dirName, "convert_dofs_to_0th_order.mtx"), evalf(OUTMConvDofsTo0stOrder)) end if;

