#! /usr/bin/python2

#
# This file is part of the Sierpinski project. For conditions of distribution and
# use, please see the copyright notice in the file 'copyright.txt' at the root
# directory of this package and the copyright notice at
# http://www5.in.tum.de/sierpinski
#

# This is the main python program to create all different kinds of traversator codes.
# The type of Code is specified on the end of this file

# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# !!! not all combinations for the configuration class scg.CConfig are possible!  !!!
# !!! e. g. it's not possible to combine the adaptivity part with the computation !!!
# !!! traversals                                                                  !!!
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import sys
import math

# import our code generator class
import libs.sierpinski_code_generator as scg



###########################################################################################
######################################## TEST SET #########################################
###########################################################################################


###########################################################################################
############# CODE GENERATION CONFIGURATION STARTS HERE ###################################
###########################################################################################

#####################################################################################################################
#####################################################################################################################
# LEAF ONLY
#####################################################################################################################
#####################################################################################################################

# debug output is written to STDERR! STDOUT is redirected to the C++ header file
sys.stderr.write("VERTICES DEPTH\n\n")
# VERTICES DEPTH
if True:
	sys.stdout = open("../traversators/vertexCoords/auto/CTraversator_VertexCoords_Depth_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()

	# create a config class with default parameters
	config = scg.CConfig()

	# kernel parameters
	config.kernelParams_vertexCoords = True
	config.kernelParams_depth = True
	config.kernelOpCellCall = True

	# forward/backward code
	config.createForwardCode = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	# restore stdout
	sys.stdout = sys.__stdout__

sys.stderr.write("VERTICES\n\n")
# VERTICES
if True:
	sys.stdout = open("../traversators/vertexCoords/auto/CTraversator_VertexCoords_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.kernelParams_vertexCoords = True
	config.kernelOpCellCall = True

	config.createForwardCode = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


sys.stderr.write("VERTICES & ELEMENT DATA\n\n")
# VERTICES & ELEMENT DATA
if True:
	sys.stdout = open("../traversators/vertexCoords/auto/CTraversator_VertexCoords_CellData_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	config = scg.CConfig()

	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelOpCellCall = True

	config.createForwardCode = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


sys.stderr.write("VERTICES & ELEMENT DATA & DEPTH\n\n")
# VERTICES & ELEMENT DATA
if True:
	sys.stdout = open("../traversators/vertexCoords/auto/CTraversator_VertexCoords_CellData_Depth_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()

	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	config.kernelOpCellCall = True

	config.createForwardCode = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


sys.stderr.write("VERTICES & NORMALS & CELL DATA & DEPTH\n\n")
# VERTICES & ELEMENT DATA
if True:
	sys.stdout = open("../traversators/vertexCoords/auto/CTraversator_VertexCoords_Normals_CellData_Depth_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()

	config.kernelParams_vertexCoords = True
	config.kernelParams_normals = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	config.kernelOpCellCall = True

	config.createForwardCode = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__





sys.stderr.write("VERTICES & NORMALS & ELEMENT DATA\n\n")
# VERTICES & NORMALS & ELEMENT DATA
if True:
	sys.stdout = open("../traversators/vertexCoords/auto/CTraversator_VertexCoords_Normals_CellData_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.kernelParams_vertexCoords = True
	config.kernelParams_normals = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = False
	config.kernelOpCellCall = True

	config.createForwardCode = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


#####################################################################################################################
#####################################################################################################################
# ADAPTIVITY
#####################################################################################################################
#####################################################################################################################



sys.stderr.write("ADAPTIVE with ELEMENTS and NORMALS and DEPTH and DEPTH LIMITERS min/max PARALLEL PART (first)\n\n")
if True:
	#
	# FIRST TRAVERSAL
	#
	# ADAPTIVE with ELEMENTS and NORMALS and DEPTH and DEPTH LIMITERS min/max (first traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_Normals_Depth_DepthLimiters_Cell_FirstTraversal_autogenerated.hpp", 'w')

	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.createForwardCode = True

	config.adaptivity = True
	config.adaptivity_create_first_traversal = True
	config.adaptivity_depthLimiterMax = True
	config.adaptivity_depthLimiterMin = True

	config.createReversedStructureStack = True
	config.assumeKVHToBeEqual = True

	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	
	config.varname_element_data_input_stream = 'element_lifo_in'

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)


	#
	# MIDDLE TRAVERSAL
	#
	# ADAPTIVE with ELEMENTS and NORMALS and DEPTH and DEPTH LIMITERS min/max (middle traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_Normals_Depth_DepthLimiters_Cell_MiddleTraversal_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	config = scg.CConfig()

	config.createForwardCode = True
	config.createBackwardCode = True

	config.adaptivity = True
	config.adaptivity_create_middle_traversal = True

	config.createReversedStructureStack = False
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)


	#
	# LAST TRAVERSAL
	#
	# ADAPTIVE with ELEMENTS and NORMALS and DEPTH and DEPTH LIMITERS min/max (last traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_Normals_Depth_DepthLimiters_Cell_LastTraversal_autogenerated.hpp", 'w')
	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()

	config = scg.CConfig()

	config.createBackwardCode = True

	config.adaptivity = True
	config.adaptivity_create_last_traversal = True
	config.adaptivity_updateMaxDepth = True

 	config.kernelParams_normals = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	
	config.varname_element_data_input_stream = 'element_fifo_in'

	config.createReversedStructureStack = False
	config.assumeKVHToBeEqual = True

	# return value of methods has to be of type 'char' to spread coarsening information
	config.adaptivity_char_return_variable = True

	# output refinement information to the communication stacks
	config.adaptivity_last_traversal_edge_comm_adaptive_information = True

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


sys.stderr.write("ADAPTIVE with ELEMENTS and VERTICES and NORMALS and DEPTH and DEPTH LIMITERS min/max PARALLEL PART (first)\n\n")
if True:
	#
	# FIRST TRAVERSAL
	#
	# ADAPTIVE with ELEMENTS and VERTICES and NORMALS and DEPTH and DEPTH LIMITERS min/max (first traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Cell_FirstTraversal_autogenerated.hpp", 'w')

	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.createForwardCode = True

	config.adaptivity = True
	config.adaptivity_create_first_traversal = True
	config.adaptivity_depthLimiterMax = True
	config.adaptivity_depthLimiterMin = True

	config.createReversedStructureStack = True
	config.assumeKVHToBeEqual = True

 	config.kernelParams_normals = True
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	
	config.varname_element_data_input_stream = 'element_lifo_in'

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)


	#
	# MIDDLE TRAVERSAL
	#
	# ADAPTIVE with ELEMENTS and VERTICES and NORMALS and DEPTH and DEPTH LIMITERS min/max (middle traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Cell_MiddleTraversal_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	config = scg.CConfig()

	config.createForwardCode = True
	config.createBackwardCode = True

	config.adaptivity = True
	config.adaptivity_create_middle_traversal = True

	config.createReversedStructureStack = False
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)


	#
	# LAST TRAVERSAL
	#
	# ADAPTIVE with ELEMENTS and VERTICES and NORMALS and DEPTH and DEPTH LIMITERS min/max (last traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_VertexCoords_Normals_Depth_DepthLimiters_Cell_LastTraversal_autogenerated.hpp", 'w')
	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()

	config = scg.CConfig()

	config.createBackwardCode = True

	config.adaptivity = True
	config.adaptivity_create_last_traversal = True
	config.adaptivity_updateMaxDepth = True

 	config.kernelParams_normals = True
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	
#	config.adaptiveKernelUnmodifiedElementActionCall = True
	
	config.varname_element_data_input_stream = 'element_fifo_in'

	config.createReversedStructureStack = False
	config.assumeKVHToBeEqual = True

	# return value of methods has to be of type 'char' to spread coarsening information
	config.adaptivity_char_return_variable = True

	# output refinement information to the communication stacks
	config.adaptivity_last_traversal_edge_comm_adaptive_information = True

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


sys.stderr.write("ADAPTIVE with VERTICES, ELEMENT DATA, DEPTH LIMITERS min/max PARALLEL PART (first)\n\n")
if True:
	#	
	# FIRST TRAVERSAL
	#
	# ADAPTIVE with VERTICES DEPTH LIMITERS min/max ELEMENT DATA PARALLEL PART (first traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_VertexCoords_DepthLimiters_Cell_FirstTraversal_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()

	config.createForwardCode = True

	config.kernelParams_cellData = True
	config.kernelParams_vertexCoords = True

	config.adaptivity = True
	config.adaptivity_depthLimiterMax = True
	config.adaptivity_depthLimiterMin = True

	config.createReversedStructureStack = True
	config.adaptivity_create_first_traversal = True
	
	config.varname_element_data_input_stream = 'element_lifo_in'

	# !!!
	# the refinement check is executed even for element triangles which are already on the finest level.
	# This is handy to run the setup routines in the should_refine() calls.
	# !!!
	config.adaptivity_callShouldRefineForAllLeaves = True
	
	# assume that K, V and H types are equal and can be replaced by 'U' (unified)
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)


	#
	# MIDDLE TRAVERSAL
	#
	# ADAPTIVE with ELEMENTS and NORMALS and DEPTH and DEPTH LIMITERS min/max (middle traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_VertexCoords_DepthLimiters_Cell_MiddleTraversal_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	
	config.adaptivity = True
	config.adaptivity_create_middle_traversal = True
	
	config.createForwardCode = True
	config.createBackwardCode = True

	config.createReversedStructureStack = False

	# assume that K, V and H types are equal and can be replaced by 'U' (unified)
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)	
	scgrun.outputRecursionTable(config)


	#
	# LAST TRAVERSAL
	#
	# ADAPTIVE with ELEMENTS and NORMALS and DEPTH and DEPTH LIMITERS min/max (last traversal only)
	sys.stdout = open("../traversators/adaptive/auto/CAdaptiveConformingGrid_VertexCoords_DepthLimiters_Cell_LastTraversal_autogenerated.hpp", 'w')
	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()

	config.adaptivity = True

	config.createForwardCode = False
	config.createBackwardCode = True

	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	config.kernelParams_vertexCoords = True
	config.kernelParams_normals = True

	config.adaptivity_updateMaxDepth = True

	config.createReversedStructureStack = True
	config.adaptivity_create_last_traversal = True
	
	# return value of methods has to be of type 'char' to spread coarsening information
	config.adaptivity_char_return_variable = True
	# output refinement information to the communication stacks
	config.adaptivity_last_traversal_edge_comm_adaptive_information = True

	# assume that K, V and H types are equal and can be replaced by 'U' (unified)
	config.assumeKVHToBeEqual = True
	
	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


#####################################################################################################################
#####################################################################################################################
# VERTEX DATA COMM
#####################################################################################################################
#####################################################################################################################


sys.stderr.write("VERTEX DATA COMM + DEPTH + recursionMethod Table\n\n")
# COMPUTATION: VERTEX DATA COMM + NORMALS + DEPTH
if True:
	sys.stdout = open("../traversators/vertexData/auto/CTraversator_VertexDataComm_Depth_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True

	config.vertexDataComm = True
	config.vertexDataCommMode = 0

	config.createForwardCode = True

	config.createReversedStructureStack = True

	scgrun.createCode(config)
	
	# backup recursion table
	recTable = scgrun.recursionTable
	
	sys.stdout = open("../traversators/vertexData/auto/CTraversator_VertexDataComm_Depth_autogenerated.hpp", 'a')
	
	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	config = scg.CConfig()
	config.createBackwardCode = True
	config.vertexDataComm = True
	config.vertexDataCommMode = 1

	# restore recursion table
	scgrun.recursionTable = recTable
	
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	config.kernelOpCellCall = True

	config.varname_element_data_input_stream = 'element_data_fifo'

	config.methodRecursionTable = True

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__



sys.stderr.write("3 PHASE VERTEX DATA COMM + DEPTH + recursionMethod Table\n\n")
# COMPUTATION: 3 PHASE VERTEX DATA COMM + NORMALS + DEPTH
if False:
	#
	# CELL DATA TO VERTEX DATA
	#
	sys.stderr.write(" + Cell to VertexData\n")
	sys.stdout = open("../traversators/vertexData/auto/CTraversator_VertexData_CellData_3Pass_1stPass.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True

	config.vertexDataComm = True
	config.vertexDataCommMode = 0

	config.createForwardCode = True

	config.createReversedStructureStack = True

	scgrun.createCode(config)
	
	# backup recursion table
	recTable = scgrun.recursionTable
	

	#
	# VERTEX DATA TO VERTEX DATA
	#
	sys.stderr.write(" + Vertex to Vertex Data\n")
	sys.stdout = open("../traversators/vertexData/auto/CTraversator_VertexData_CellData_3Pass_2ndPass.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True

	config.vertexDataComm = True
	config.vertexDataCommMode = 3

	config.createBackwardCode = True

	scgrun.createCode(config)
	
	# backup recursion table
	recTable = scgrun.recursionTable
	

	#
	# VERTEX DATA TO CELL DATA
	#
	sys.stderr.write(" + Vertex to Cell Data\n")
	sys.stdout = open("../traversators/vertexData/auto/CTraversator_VertexData_CellData_3Pass_3rdPass.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	config = scg.CConfig()
	config.createForwardCode = True
	config.vertexDataComm = True
	config.vertexDataCommMode = 2

	# restore recursion table
	scgrun.recursionTable = recTable
	
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_depth = True
	config.kernelOpCellCall = True

	config.varname_element_data_input_stream = 'element_data_fifo'

	config.methodRecursionTable = True

	scgrun.createCode(config)
	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


#####################################################################################################################
#####################################################################################################################
# EDGE COMM
#####################################################################################################################
#####################################################################################################################




sys.stderr.write("EDGE COMM + NORMALS + DEPTH + recursionMethod Table\n\n")
# COMPUTATION: EDGE COMM + NORMALS + DEPTH
if True:
	sys.stdout = open("../traversators/edgeComm/auto/CTraversator_EdgeComm_Normals_Depth_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.edgeComm = True

	config.kernelParams_cellData = True
	config.kernelParams_normals = True
	config.kernelParams_depth = True

	config.createForwardCode = True
	config.createReversedStructureStack = True
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)
	
	# backup recursion table
	recTable = scgrun.recursionTable
	
	sys.stdout = open("../traversators/edgeComm/auto/CTraversator_EdgeComm_Normals_Depth_autogenerated.hpp", 'a')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	# restore recursion table
	scgrun.recursionTable = recTable
	
	config = scg.CConfig()
	config.edgeComm = True
	
	config.kernelParams_cellData = True
	config.kernelParams_normals = True
	config.kernelParams_depth = True
	config.kernelOpCellCall = True

	config.createBackwardCode = True

	config.varname_element_data_input_stream = 'element_data_fifo'

	config.methodRecursionTable = True
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


sys.stderr.write("COMPUTATION: EDGE COMM + NORMALS + VERTICES\n\n")
# COMPUTATION: EDGE COMM + NORMALS + VERTICES
if True:
	sys.stdout = open("../traversators/edgeComm/auto/CTraversator_EdgeComm_VertexCoords_Normals_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()
	config.edgeComm = True
	
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_normals = True
	
	config.createForwardCode = True
	config.createReversedStructureStack = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	# backup recursion table
	recTable = scgrun.recursionTable
	
	sys.stdout = open("../traversators/edgeComm/auto/CTraversator_EdgeComm_VertexCoords_Normals_autogenerated.hpp", 'a')
	
	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	# restore recursion table
	scgrun.recursionTable = recTable
	
	config = scg.CConfig()
	config.edgeComm = True
	
	config.kernelParams_vertexCoords = True
	config.kernelParams_cellData = True
	config.kernelParams_normals = True
	config.kernelOpCellCall = True

	config.createBackwardCode = True

	config.varname_element_data_input_stream = 'element_data_fifo'
	config.methodRecursionTable = True

	# assume that K, V and H types are equal and can be replaced by 'U' (unified)
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)
	
	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__



#####################################################################################################################
#####################################################################################################################
# FLUX COMM
#####################################################################################################################
#####################################################################################################################

sys.stderr.write("FLUX COMM + NORMALS + DEPTH + recursionMethod Table\n\n")
# COMPUTATION: EDGE COMM + NORMALS + DEPTH
if True:
	sys.stdout = open("../traversators/fluxComm/auto/CTraversator_FluxComm_Normals_Depth_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	config = scg.CConfig()

	config.fluxComm = True

	config.kernelParams_cellData = True
	config.kernelParams_normals = True
	config.kernelParams_depth = True

	config.createForwardCode = True

	config.createReversedStructureStack = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)
	
	# backup recursion table
	recTable = scgrun.recursionTable
	
	sys.stdout = open("../traversators/fluxComm/auto/CTraversator_FluxComm_Normals_Depth_autogenerated.hpp", 'a')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()
	
	# restore recursion table
	scgrun.recursionTable = recTable
	
	config = scg.CConfig()
	config.fluxComm = True

	config.createBackwardCode = True

	config.kernelParams_cellData = True
	config.kernelParams_normals = True
	config.kernelParams_depth = True
	config.kernelOpCellCall = True

	config.varname_element_data_input_stream = 'element_data_fifo'
	config.methodRecursionTable = True
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


sys.stderr.write("FLUX COMM + DEPTH + RETURN ADAPTIVITY INFORMATION + recursionMethod Table\n\n")
# COMPUTATION: EDGE COMM + DEPTH
if True:
	sys.stdout = open("../traversators/fluxComm/auto/CTraversator_FluxComm_Depth_RetAdapt_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()

	config = scg.CConfig()

	config.fluxComm = True

	config.kernelParams_cellData = True
	config.kernelParams_normals = False
	config.kernelParams_depth = True

	config.createForwardCode = True

	config.createReversedStructureStack = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)
	
	# backup recursion table
	recTable = scgrun.recursionTable

	sys.stdout = open("../traversators/fluxComm/auto/CTraversator_FluxComm_Depth_RetAdapt_autogenerated.hpp", 'a')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()

	# restore recursion table
	scgrun.recursionTable = recTable

	config = scg.CConfig()
	config.fluxComm = True

	config.createBackwardCode = True

	config.kernelParams_cellData = True
	config.kernelParams_normals = False
	config.kernelParams_depth = True
	config.kernelOpCellCall = True
	config.kernelReturnAdaptivity = True
    
	config.varname_element_data_input_stream = 'element_data_fifo'
	config.varname_adaptivity_data_output_stream = 'adaptivity_output_stream'
	config.methodRecursionTable = True
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__


sys.stderr.write("FLUX COMM + NORMALS + DEPTH + RETURN ADAPTIVITY INFORMATION + recursionMethod Table\n\n")
# COMPUTATION: EDGE COMM + NORMALS + DEPTH
if True:
	sys.stdout = open("../traversators/fluxComm/auto/CTraversator_FluxComm_Normals_Depth_RetAdapt_autogenerated.hpp", 'w')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()

	config = scg.CConfig()

	config.fluxComm = True

	config.kernelParams_cellData = True
	config.kernelParams_normals = True
	config.kernelParams_depth = True

	config.createForwardCode = True

	config.createReversedStructureStack = True

	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)
	
	# backup recursion table
	recTable = scgrun.recursionTable

	sys.stdout = open("../traversators/fluxComm/auto/CTraversator_FluxComm_Normals_Depth_RetAdapt_autogenerated.hpp", 'a')

	# create a fresh instance
	scgrun = scg.CSierpinskiCodeGenerator()

	# restore recursion table
	scgrun.recursionTable = recTable

	config = scg.CConfig()
	config.fluxComm = True

	config.createBackwardCode = True

	config.kernelParams_cellData = True
	config.kernelParams_normals = True
	config.kernelParams_depth = True
	config.kernelOpCellCall = True
	config.kernelReturnAdaptivity = True
    
	config.varname_element_data_input_stream = 'element_data_fifo'
	config.varname_adaptivity_data_output_stream = 'adaptivity_output_stream'
	config.methodRecursionTable = True
	config.assumeKVHToBeEqual = True

	scgrun.createCode(config)

	scgrun.outputRecursionTable(config)

	sys.stdout = sys.__stdout__
