
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION2D_MATRICES_CONST_HPP_
#define CCOMPUTATION2D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (3)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (10)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (10)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS (8)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (13)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*3+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_2D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[10],
		T io_hv[10]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{10.01343146347375,3.555936538441215,3.555937098156799,-2.443949673945655,1.221975100338516,1.221974836791108,-1.692409305965417,-2.712038407274934,4.404446223441276,-2.712038613236915},
			{3.555936538441217,10.01343417086613,3.555936391892739,1.221975581963253,-2.443951207983593,1.221975965780719,4.404448518811391,-1.692410447499592,-2.712038761013454,4.404448776516672},
			{3.555937098156799,3.555936391892738,10.01343367527705,1.221975646054362,1.221976509869574,-2.443951048927955,-2.712039122432838,4.404445702989177,-1.692409835195898,-1.692410602929889},
			{-2.443949673945654,1.221975581963253,1.221975646054363,38.49545879518467,0.4199242800956394,0.4199241693029314,-0.5815858743757831,-0.9319749716238978,1.513560482333602,-0.9319749598961085},
			{1.221975100338516,-2.443951207983592,1.221976509869574,0.4199242800956392,38.49546134962688,0.4199219182622189,1.513558163691415,-0.5815859893830884,-0.9319745106740631,1.513562152147228},
			{1.221974836791108,1.221975965780719,-2.443951048927956,0.4199241693029314,0.4199219182622188,38.49545988947659,-0.9319731523868369,1.513561523012706,-0.581585805429657,-0.581585839144423},
			{-1.692409305965416,4.404448518811394,-2.712039122432836,-0.5815858743757829,1.513558163691415,-0.9319731523868361,21.36325840413494,1.388221761191283,1.388221169100666,-2.678985707761788},
			{-2.712038407274936,-1.692410447499594,4.404445702989176,-0.9319749716238981,-0.5815859893830885,1.513561523012706,1.388221761191283,21.3632568199595,1.388221789541477,0.7080263342095845},
			{4.404446223441278,-2.712038761013454,-1.692409835195899,1.513560482333602,-0.9319745106740633,-0.5815858054296573,1.388221169100667,1.388221789541476,21.36325634891856,1.970958521976558},
			{-2.712038613236918,4.404448776516673,-1.692410602929889,-0.9319749598961082,1.513562152147227,-0.5815858391444229,-2.67898570776179,0.7080263342095842,1.970958521976558,21.36326241046861}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{10.01343146347375,3.555936538441215,3.555937098156799,-2.443949673945655,1.221975100338516,1.221974836791108,-1.692409305965417,-2.712038407274934,4.404446223441276,-2.712038613236915},
			{3.555936538441217,10.01343417086613,3.555936391892739,1.221975581963253,-2.443951207983593,1.221975965780719,4.404448518811391,-1.692410447499592,-2.712038761013454,4.404448776516672},
			{3.555937098156799,3.555936391892738,10.01343367527705,1.221975646054362,1.221976509869574,-2.443951048927955,-2.712039122432838,4.404445702989177,-1.692409835195898,-1.692410602929889},
			{-2.443949673945654,1.221975581963253,1.221975646054363,38.49545879518467,0.4199242800956394,0.4199241693029314,-0.5815858743757831,-0.9319749716238978,1.513560482333602,-0.9319749598961085},
			{1.221975100338516,-2.443951207983592,1.221976509869574,0.4199242800956392,38.49546134962688,0.4199219182622189,1.513558163691415,-0.5815859893830884,-0.9319745106740631,1.513562152147228},
			{1.221974836791108,1.221975965780719,-2.443951048927956,0.4199241693029314,0.4199219182622188,38.49545988947659,-0.9319731523868369,1.513561523012706,-0.581585805429657,-0.581585839144423},
			{-1.692409305965416,4.404448518811394,-2.712039122432836,-0.5815858743757829,1.513558163691415,-0.9319731523868361,21.36325840413494,1.388221761191283,1.388221169100666,-2.678985707761788},
			{-2.712038407274936,-1.692410447499594,4.404445702989176,-0.9319749716238981,-0.5815859893830885,1.513561523012706,1.388221761191283,21.3632568199595,1.388221789541477,0.7080263342095845},
			{4.404446223441278,-2.712038761013454,-1.692409835195899,1.513560482333602,-0.9319745106740633,-0.5815858054296573,1.388221169100667,1.388221789541476,21.36325634891856,1.970958521976558},
			{-2.712038613236918,4.404448776516673,-1.692410602929889,-0.9319749598961082,1.513562152147227,-0.5815858391444229,-2.67898570776179,0.7080263342095842,1.970958521976558,21.36326241046861}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.196550164,.4793816273e-1,-.6424538824,.1017586202,-.2089231459e-1,-.4523178164,-.1165631648,.5765375751,-.3681999783,.9177441691e-1},
			{-1.614157763,-.8606081071,2.108557451,-.1613326088,-.9592368542e-1,1.099069750,.8388856257,-.9931250716,.4243741011,.2701346789},
			{-.4939467924,.8570090710,-.5988882699,-.6431307239e-1,.1109579345,-.5170067994,-.2531039171,-.3454402975,.2366351205,-.3619090993},
			{.4116664741,-.3141941727,-.7707910669e-1,.2332852119,-.1225866642e-1,-.2828216365e-1,.1417392661,.6437806545e-1,-.9433499315e-1,.1117823332},
			{-.2959893810,-.8562776001e-1,.3965061501,-.7123702935e-1,-.1911527625,.1400168279,.2047874035e-1,-.2146465232,.1365311129,-.5763669031e-1},
			{-.3591310055e-1,.1298337006,-.1292028211,-.6344574504e-2,.1924813370e-1,-.8327498158e-1,-.5929460679e-1,-.1688283697e-1,.2203178642e-1,-.5414565976e-1},
			{.6433838849,.4457654845,-1.216746222,.7461356425e-1,.8418363048e-1,-.4132928908,-.6199578990,.5975247067,-.1733826073,-.1958158126},
			{.2848632824,-.7577047154,.6800633411,.4271752329e-1,-.1067172984,.4773099294,.3056284364,.1683813182,-.1559978518,.3180119185},
			{-1.022345454,.5101078166,.4326126220,-.3745944684,-.3474159621e-1,.2505216067,-.2131210396,-.3416844211,.4326093507,-.1221960989},
			{.9258886745,.2748060201e-1,-.9533692907,.2254468336,.2472966162,-.4727434602,-.4469144601e-1,.5049574983,-.4602660349,-.2102989041e-7}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.196550164,.4793816273e-1,-.6424538824,.1017586202,-.2089231459e-1,-.4523178164,-.1165631648,.5765375751,-.3681999783,.9177441691e-1},
			{-1.614157763,-.8606081071,2.108557451,-.1613326088,-.9592368542e-1,1.099069750,.8388856257,-.9931250716,.4243741011,.2701346789},
			{-.4939467924,.8570090710,-.5988882699,-.6431307239e-1,.1109579345,-.5170067994,-.2531039171,-.3454402975,.2366351205,-.3619090993},
			{.4116664741,-.3141941727,-.7707910669e-1,.2332852119,-.1225866642e-1,-.2828216365e-1,.1417392661,.6437806545e-1,-.9433499315e-1,.1117823332},
			{-.2959893810,-.8562776001e-1,.3965061501,-.7123702935e-1,-.1911527625,.1400168279,.2047874035e-1,-.2146465232,.1365311129,-.5763669031e-1},
			{-.3591310055e-1,.1298337006,-.1292028211,-.6344574504e-2,.1924813370e-1,-.8327498158e-1,-.5929460679e-1,-.1688283697e-1,.2203178642e-1,-.5414565976e-1},
			{.6433838849,.4457654845,-1.216746222,.7461356425e-1,.8418363048e-1,-.4132928908,-.6199578990,.5975247067,-.1733826073,-.1958158126},
			{.2848632824,-.7577047154,.6800633411,.4271752329e-1,-.1067172984,.4773099294,.3056284364,.1683813182,-.1559978518,.3180119185},
			{-1.022345454,.5101078166,.4326126220,-.3745944684,-.3474159621e-1,.2505216067,-.2131210396,-.3416844211,.4326093507,-.1221960989},
			{.9258886745,.2748060201e-1,-.9533692907,.2254468336,.2472966162,-.4727434602,-.4469144601e-1,.5049574983,-.4602660349,-.2102989041e-7}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.052802072,-.6973086860,-.1196673474,.6988407343,-.1439101562,-.8456864235e-1,.3262866284,-.2580407541e-1,.6142652878e-1,.3619090881},
			{-1.685580211,1.254675024,-.1711293401,-.6911086538,.4743431337,.5882170241,-.8429021664,.8355349005e-1,.6675742418,-.9177438926e-1},
			{-.2344371164,.3541879483,.2464576610,-.1374772071,-.2065459391,-.4977903149,.2238063038,-.5269680023,.3302707068e-1,-.2701346645},
			{.1971134137,-.1423120799,-.1951911635e-1,.1104793775,-.2480795644e-1,-.1530000035e-1,.6218762146e-1,.1756684490e-1,-.2560880700e-1,.5414565701e-1},
			{-.3050594429,.1912399183,.9342635488e-1,-.1478498856,-.1117594088,.6686491302e-1,-.1963366736,-.6693481936e-1,.1514891566,-.1117823336},
			{-.8227822257e-1,-.1286918288,.1960810361,.8910825944e-2,-.1913624085e-1,.1325983776,.6992114560e-1,-.5355542637e-1,.4127095947e-1,.5763667963e-1},
			{.6740174926,-.3057624077,-.2886300575,.2836168264,.7884249473e-1,-.2036448620,.2177233347,.1636411322,-.2591683549,.1221960935},
			{.5089731701,-.3620400861,-.1933626420e-1,.3893313950e-1,.3817128319e-1,.1773912651,.1769300339e-1,.4243595358,-.2462367082,.1958158156},
			{-1.078920380,.7619007810,.1097977345,-.6370886093,.1402496060,.8352885947e-1,-.3386452082,-.6055014888e-1,.8118344680e-1,-.3180118993},
			{.9533692458,-.9258886590,-.2748057619e-1,.4727434581,-.2254468468,-.2472966239,.4602660439,.4469142960e-1,-.5049574849,-.6140337311e-8}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.052802072,-.6973086860,-.1196673474,.6988407343,-.1439101562,-.8456864235e-1,.3262866284,-.2580407541e-1,.6142652878e-1,.3619090881},
			{-1.685580211,1.254675024,-.1711293401,-.6911086538,.4743431337,.5882170241,-.8429021664,.8355349005e-1,.6675742418,-.9177438926e-1},
			{-.2344371164,.3541879483,.2464576610,-.1374772071,-.2065459391,-.4977903149,.2238063038,-.5269680023,.3302707068e-1,-.2701346645},
			{.1971134137,-.1423120799,-.1951911635e-1,.1104793775,-.2480795644e-1,-.1530000035e-1,.6218762146e-1,.1756684490e-1,-.2560880700e-1,.5414565701e-1},
			{-.3050594429,.1912399183,.9342635488e-1,-.1478498856,-.1117594088,.6686491302e-1,-.1963366736,-.6693481936e-1,.1514891566,-.1117823336},
			{-.8227822257e-1,-.1286918288,.1960810361,.8910825944e-2,-.1913624085e-1,.1325983776,.6992114560e-1,-.5355542637e-1,.4127095947e-1,.5763667963e-1},
			{.6740174926,-.3057624077,-.2886300575,.2836168264,.7884249473e-1,-.2036448620,.2177233347,.1636411322,-.2591683549,.1221960935},
			{.5089731701,-.3620400861,-.1933626420e-1,.3893313950e-1,.3817128319e-1,.1773912651,.1769300339e-1,.4243595358,-.2462367082,.1958158156},
			{-1.078920380,.7619007810,.1097977345,-.6370886093,.1402496060,.8352885947e-1,-.3386452082,-.6055014888e-1,.8118344680e-1,-.3180118993},
			{.9533692458,-.9258886590,-.2748057619e-1,.4727434581,-.2254468468,-.2472966239,.4602660439,.4469142960e-1,-.5049574849,-.6140337311e-8}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[10],
		T o_output[8]
	) {
		static const T matrix[8][10] = 
		{
			{-1.45425377932711,1.32618123974878,.214929690789362,1.72546667234598,.0749963295610948,-.0493986167919454,-.52096553277996,.0893348435663733,-.178270751151675,-.228020095960901},
			{.901617182972597,-.878907514262425,-.154299052712732,1.32719741075668,-.237190144283086,.0181176435188071,.36814359005261,-.0122740843570015,-1.08754546709488,.755140435409436},
			{2.76631253535879,-2.68655311455725,-.48375671187014,.80135987241094,-.485537526595039,-.00945790998173296,1.10972468290512,.179128119201145,-1.71166846020543,1.52044851333359},
			{2.54784227751401,-2.5501417334435,-.585993500196011,.346798469815348,-.450179502588528,-.132679980584154,1.09020785565277,.767598933424162,-1.39841366170664,1.36496084211255},
			{.780302598025293,-.765805601220523,-.602789949818311,.0717332396365517,-.183222802955387,-.124571448359104,.427867625824713,1.37006771146173,-.432974482491953,.45939310989698},
			{-.643005715312524,1.06666564380352,-.827657213013815,-.0332833473253481,.0759860981521238,.263661688539869,-.244989278148133,1.44637652504956,.319061272858217,-.422815674603482},
			{-.559184637570565,1.76597055534194,-1.33837529716116,-.0424577813150609,.158513281569754,.99206941222702,-.469111858018255,.880750738644108,.343501557358165,-.731675971075964},
			{.368586029145061,1.57122965010233,-1.85295852760903,-.0264099052189327,.111658320369302,1.66581597019539,-.354409845276218,.172073747789449,-.0456839969593226,-.609901442538053}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 8; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[10],
		T o_output[8]
	) {
		static const T matrix[8][10] = 
		{
			{-1.45425377932711,1.32618123974878,.214929690789362,1.72546667234598,.0749963295610948,-.0493986167919454,-.52096553277996,.0893348435663733,-.178270751151675,-.228020095960901},
			{.901617182972597,-.878907514262425,-.154299052712732,1.32719741075668,-.237190144283086,.0181176435188071,.36814359005261,-.0122740843570015,-1.08754546709488,.755140435409436},
			{2.76631253535879,-2.68655311455725,-.48375671187014,.80135987241094,-.485537526595039,-.00945790998173296,1.10972468290512,.179128119201145,-1.71166846020543,1.52044851333359},
			{2.54784227751401,-2.5501417334435,-.585993500196011,.346798469815348,-.450179502588528,-.132679980584154,1.09020785565277,.767598933424162,-1.39841366170664,1.36496084211255},
			{.780302598025293,-.765805601220523,-.602789949818311,.0717332396365517,-.183222802955387,-.124571448359104,.427867625824713,1.37006771146173,-.432974482491953,.45939310989698},
			{-.643005715312524,1.06666564380352,-.827657213013815,-.0332833473253481,.0759860981521238,.263661688539869,-.244989278148133,1.44637652504956,.319061272858217,-.422815674603482},
			{-.559184637570565,1.76597055534194,-1.33837529716116,-.0424577813150609,.158513281569754,.99206941222702,-.469111858018255,.880750738644108,.343501557358165,-.731675971075964},
			{.368586029145061,1.57122965010233,-1.85295852760903,-.0264099052189327,.111658320369302,1.66581597019539,-.354409845276218,.172073747789449,-.0456839969593226,-.609901442538053}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 8; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[10],
		T o_output[8]
	) {
		static const T matrix[8][10] = 
		{
			{.136238006771799,-.905026226343549,.855645365157245,-.061171775023035,1.80763727482326,.00459888177778651,-.432860546853416,-.33552263668457,.158481745582043,-.22802008920756},
			{.1063063675187,-.917274894712635,.679379143532882,.0571071372659607,1.05507075503056,-.00405298238561168,-.244413410198392,-.245992766476126,-.241269790385193,.755140440809851},
			{.0409630703466698,-.895962207967806,.451001852692577,.0690460542762892,.25344249350708,-.0161241086859924,-.0140519380454184,-.126817111625959,-.281946622110677,1.52044851761324},
			{-.114933838641156,-.739912335428009,.266553225860955,-.0622041723066864,-.145086497875149,-.0287703389043051,.125597218611059,-.0198798029625122,.353675695323054,1.36496084632275},
			{-.444249450430285,-.32622864510776,.182185148199971,-.100852011776358,-.0938162473078612,-.0413927496187602,.0799487290695668,.0542549849480529,1.2307571273412,.459393114682234},
			{-.973574506362864,.375422165711069,.194155058565812,.241830824758655,.11908488572855,-.0545512699657424,-.153022370217456,.0988758653826939,1.57459501609323,-.422815669693946},
			{-1.59088292224111,1.20318902343184,.256104520240868,.954291439404765,.221213087038666,-.0673796136424686,-.473431870973027,.125941445012019,1.10263085872931,-.731675967000855},
			{-2.06344072573252,1.83764375032395,.312654127890014,1.63432547259165,.19337761091637,-.0766386977395684,-.726653301497992,.141607355522673,.357025847311682,-.609901439586253}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 8; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[10],
		T o_output[8]
	) {
		static const T matrix[8][10] = 
		{
			{.136238006771799,-.905026226343549,.855645365157245,-.061171775023035,1.80763727482326,.00459888177778651,-.432860546853416,-.33552263668457,.158481745582043,-.22802008920756},
			{.1063063675187,-.917274894712635,.679379143532882,.0571071372659607,1.05507075503056,-.00405298238561168,-.244413410198392,-.245992766476126,-.241269790385193,.755140440809851},
			{.0409630703466698,-.895962207967806,.451001852692577,.0690460542762892,.25344249350708,-.0161241086859924,-.0140519380454184,-.126817111625959,-.281946622110677,1.52044851761324},
			{-.114933838641156,-.739912335428009,.266553225860955,-.0622041723066864,-.145086497875149,-.0287703389043051,.125597218611059,-.0198798029625122,.353675695323054,1.36496084632275},
			{-.444249450430285,-.32622864510776,.182185148199971,-.100852011776358,-.0938162473078612,-.0413927496187602,.0799487290695668,.0542549849480529,1.2307571273412,.459393114682234},
			{-.973574506362864,.375422165711069,.194155058565812,.241830824758655,.11908488572855,-.0545512699657424,-.153022370217456,.0988758653826939,1.57459501609323,-.422815669693946},
			{-1.59088292224111,1.20318902343184,.256104520240868,.954291439404765,.221213087038666,-.0673796136424686,-.473431870973027,.125941445012019,1.10263085872931,-.731675967000855},
			{-2.06344072573252,1.83764375032395,.312654127890014,1.63432547259165,.19337761091637,-.0766386977395684,-.726653301497992,.141607355522673,.357025847311682,-.609901439586253}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 8; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[10],
		T o_output[8]
	) {
		static const T matrix[8][10] = 
		{
			{.776953679829385,.685465564650613,-1.37556209400805,-.00717427718582053,.0209988307739315,1.73723983109591,-.0961080517716217,-.247417655554891,-.266375731459099,-.228020096370379},
			{.939984561275504,-1.71258571660202,.641011771672815,.034936511000544,-.21501951926035,1.28820791831646,.601862270909948,-.858549768800081,-.474988470363239,.755140441850394},
			{.975721629867719,-3.62131168822314,2.24159276929892,.0623798553302253,-.478871328737102,.722855910166654,1.41566990956488,-1.25059373153752,-.587891850478623,1.52044852474798},
			{.737612879067618,-3.40268846552148,2.07678263375027,.0417054687438725,-.554089143978086,.276322663635704,1.87768658543182,-.984490437069049,-.433803036111242,1.36496085205056},
			{.340725637124317,-1.55078069903093,.621762112284199,-.01767331420709,-.266401500287098,.0480138046448678,1.743680345325,-.293663908629057,-.0850555914960167,.459393114271799},
			{.048237755650778,.0448533753610167,-.497088414299526,-.0763821350003243,.39419905756096,-.0114524825777734,1.10251137688036,.190842775774665,.227094364146653,-.42281567349681},
			{.00359688925636023,.171490737841793,-.306677008860344,-.10515758714779,1.21796230595075,-.00467980786731078,.285697435187406,.12162143376829,.347821569515127,-.731675967644279},
			{.102171927666744,-.59438301103978,.579068229500308,-.1081291952931,1.85411298420521,.00508059295826713,-.323943449747462,-.230636099361268,.326559455490798,-.609901434379722}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 8; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[10],
		T o_output[8]
	) {
		static const T matrix[8][10] = 
		{
			{.776953679829385,.685465564650613,-1.37556209400805,-.00717427718582053,.0209988307739315,1.73723983109591,-.0961080517716217,-.247417655554891,-.266375731459099,-.228020096370379},
			{.939984561275504,-1.71258571660202,.641011771672815,.034936511000544,-.21501951926035,1.28820791831646,.601862270909948,-.858549768800081,-.474988470363239,.755140441850394},
			{.975721629867719,-3.62131168822314,2.24159276929892,.0623798553302253,-.478871328737102,.722855910166654,1.41566990956488,-1.25059373153752,-.587891850478623,1.52044852474798},
			{.737612879067618,-3.40268846552148,2.07678263375027,.0417054687438725,-.554089143978086,.276322663635704,1.87768658543182,-.984490437069049,-.433803036111242,1.36496085205056},
			{.340725637124317,-1.55078069903093,.621762112284199,-.01767331420709,-.266401500287098,.0480138046448678,1.743680345325,-.293663908629057,-.0850555914960167,.459393114271799},
			{.048237755650778,.0448533753610167,-.497088414299526,-.0763821350003243,.39419905756096,-.0114524825777734,1.10251137688036,.190842775774665,.227094364146653,-.42281567349681},
			{.00359688925636023,.171490737841793,-.306677008860344,-.10515758714779,1.21796230595075,-.00467980786731078,.285697435187406,.12162143376829,.347821569515127,-.731675967644279},
			{.102171927666744,-.59438301103978,.579068229500308,-.1081291952931,1.85411298420521,.00508059295826713,-.323943449747462,-.230636099361268,.326559455490798,-.609901434379722}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 8; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[8],
		T o_output[10]
	) {
		static const T matrix[10][8] = 
		{
			{.104094590346149,-.141776721078137,-.613634778917128,-.653409853364533,-.200113409945121,.142634161872485,.0879301835515604,-.0263831610799882},
			{-.0949272367992553,.138205690681529,.595942217456873,.653999563008246,.196395565775554,-.236612142752328,-.277693814609261,-.112467651170763},
			{-.0153845349649324,.0242630843468947,.107308895558184,.150281644360358,.154589197375555,.183594313525759,.210455684282643,.132633630802123},
			{-.123507842281073,-.208697993643045,-.177760929704989,-.0889386047593935,-.0183964313670215,.00738304845043946,.00667636457332789,.001890404758733},
			{-.0053681911053648,.0372974712145967,.107703923175648,.115451307708537,.0469886169441115,-.0168555475725761,-.024925759771127,-.00799243383982093},
			{.00353592258223542,-.00284894758026933,.00209799233690204,.0340265986725407,.0319471156139029,-.058486515853891,-.156000075202356,-.119238081739623},
			{.0372903921516954,-.057889525698158,-.246163675199138,-.279590523081356,-.109729289413606,.0543445252884263,.0737664968067963,.0253684385649251},
			{-.00639453311167032,.00193006462968555,-.0397349332076028,-.196855476870741,-.351362307763876,-.320841165933044,-.138495532439186,-.0123169329453331},
			{.0127605106313839,.171013411523225,.379689309769675,.358631801388526,.111038974277361,-.0707754785942782,-.0540146365965548,.00327003238118161},
			{.0163215381092027,-.118743671824099,-.337272117813443,-.350052619647759,-.117814194082433,.0937907050239907,.115053951976384,.0436563698269361}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 8; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[8],
		T o_output[10]
	) {
		static const T matrix[10][8] = 
		{
			{.104094590346149,-.141776721078137,-.613634778917128,-.653409853364533,-.200113409945121,.142634161872485,.0879301835515604,-.0263831610799882},
			{-.0949272367992553,.138205690681529,.595942217456873,.653999563008246,.196395565775554,-.236612142752328,-.277693814609261,-.112467651170763},
			{-.0153845349649324,.0242630843468947,.107308895558184,.150281644360358,.154589197375555,.183594313525759,.210455684282643,.132633630802123},
			{-.123507842281073,-.208697993643045,-.177760929704989,-.0889386047593935,-.0183964313670215,.00738304845043946,.00667636457332789,.001890404758733},
			{-.0053681911053648,.0372974712145967,.107703923175648,.115451307708537,.0469886169441115,-.0168555475725761,-.024925759771127,-.00799243383982093},
			{.00353592258223542,-.00284894758026933,.00209799233690204,.0340265986725407,.0319471156139029,-.058486515853891,-.156000075202356,-.119238081739623},
			{.0372903921516954,-.057889525698158,-.246163675199138,-.279590523081356,-.109729289413606,.0543445252884263,.0737664968067963,.0253684385649251},
			{-.00639453311167032,.00193006462968555,-.0397349332076028,-.196855476870741,-.351362307763876,-.320841165933044,-.138495532439186,-.0123169329453331},
			{.0127605106313839,.171013411523225,.379689309769675,.358631801388526,.111038974277361,-.0707754785942782,-.0540146365965548,.00327003238118161},
			{.0163215381092027,-.118743671824099,-.337272117813443,-.350052619647759,-.117814194082433,.0937907050239907,.115053951976384,.0436563698269361}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 8; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[8],
		T o_output[10]
	) {
		static const T matrix[10][8] = 
		{
			{-.00689558700631383,-.0118202599888947,-.00642519370165688,.0208423197182862,.080561035722907,.152708396451657,.176891094971095,.104439542193928},
			{.0458072400985804,.101992269982154,.140534649547464,.134177102590675,.0591589196270397,-.0588862141967161,-.13378320983686,-.093010993534226},
			{-.0433078639492565,-.0755405183624456,-.0707411392464508,-.048337266213479,-.0330377994122565,-.0304538661014528,-.028476394069675,-.0158247598657252},
			{.00309616462393304,-.00634977212993765,-.0108301030490586,.0112802222770543,.0182886945961844,-.0379319684524605,-.106108158732416,-.0827201877062657},
			{-.091492237737142,-.117313862962601,-.0397532972805173,.0263102599832379,.0170128157579876,-.0186788600383275,-.0245967975651416,-.00978766625219707},
			{-.000232769035468907,.000450653207766819,.00252912002682659,.00521726768144552,.0075062395180826,.00855654796466611,.0074919740914405,.00387901159768857},
			{.0219089197879117,.0271764534970979,.00220409317615605,-.0227760372138291,-.0144980537676211,.0240020672526013,.0526411346050896,.0367790250506059},
			{.0169822327019335,.0273520629385043,.0198916853640507,.00360504107563016,-.00983870160404794,-.0155090080437392,-.0140034944111629,-.0071673526637556},
			{-.00802143756700706,.0268269127841043,.0442242645694705,-.064136219634369,-.223187825681996,-.246980460557323,-.122602095492219,-.0180706019705966},
			{.0115410699376412,-.0839644561924366,-.238487402345228,-.247524581953833,-.0833072164454612,.0663200427821505,.0813554292131623,.030869715005355}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 8; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[8],
		T o_output[10]
	) {
		static const T matrix[10][8] = 
		{
			{-.00689558700631383,-.0118202599888947,-.00642519370165688,.0208423197182862,.080561035722907,.152708396451657,.176891094971095,.104439542193928},
			{.0458072400985804,.101992269982154,.140534649547464,.134177102590675,.0591589196270397,-.0588862141967161,-.13378320983686,-.093010993534226},
			{-.0433078639492565,-.0755405183624456,-.0707411392464508,-.048337266213479,-.0330377994122565,-.0304538661014528,-.028476394069675,-.0158247598657252},
			{.00309616462393304,-.00634977212993765,-.0108301030490586,.0112802222770543,.0182886945961844,-.0379319684524605,-.106108158732416,-.0827201877062657},
			{-.091492237737142,-.117313862962601,-.0397532972805173,.0263102599832379,.0170128157579876,-.0186788600383275,-.0245967975651416,-.00978766625219707},
			{-.000232769035468907,.000450653207766819,.00252912002682659,.00521726768144552,.0075062395180826,.00855654796466611,.0074919740914405,.00387901159768857},
			{.0219089197879117,.0271764534970979,.00220409317615605,-.0227760372138291,-.0144980537676211,.0240020672526013,.0526411346050896,.0367790250506059},
			{.0169822327019335,.0273520629385043,.0198916853640507,.00360504107563016,-.00983870160404794,-.0155090080437392,-.0140034944111629,-.0071673526637556},
			{-.00802143756700706,.0268269127841043,.0442242645694705,-.064136219634369,-.223187825681996,-.246980460557323,-.122602095492219,-.0180706019705966},
			{.0115410699376412,-.0839644561924366,-.238487402345228,-.247524581953833,-.0833072164454612,.0663200427821505,.0813554292131623,.030869715005355}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 8; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[8],
		T o_output[10]
	) {
		static const T matrix[10][8] = 
		{
			{-.0393249418872754,-.104517369553325,-.153045179908154,-.133760114824425,-.061787831583125,-.00756625226494136,-.000399939976821812,-.0051713573438354},
			{-.0346943378935191,.190423291624017,.568014771695437,.617049963166621,.281221505557339,-.00703540097040323,-.0190681438402154,.0300842611017119},
			{.0696230686764805,-.071274430440696,-.351601274540444,-.376607691431514,-.112751517622278,.0779699695773312,.0340995752367153,-.0293091146422881},
			{.000363120789231028,-.00388460872824636,-.00978448759299643,-.00756294859579786,.00320491223073097,.0119807916879719,.0116925265052733,.0054728800848884},
			{-.0010628404515272,.0239081315603922,.0751125591726019,.100479573533425,.0483097520108983,-.0618314320778367,-.135425858761273,-.0938445717540383},
			{-.0879291226435904,-.143236504733123,-.113382351515694,-.0501088745402918,-.00870691416149512,.00179635994822412,.00052035025728781,-.000257150494326289},
			{.00486443870328052,-.0669213772017063,-.222052529499926,-.340503237401605,-.31620229232248,-.172932573041675,-.0317668455888258,.0163961606293954},
			{.0125228635621091,.0954625928577346,.196159782438273,.17852935820801,.053253568711632,-.029934323539148,-.0135231501265476,.0116734773770315},
			{.0134824126994415,.0528142136964019,.0922127902762967,.0786666631889227,.0154241418606299,-.035620505637109,-.0386744602169855,-.016528567845558},
			{.0115410703001821,-.0839644563081351,-.238487403464336,-.247524582992525,-.083307216371032,.0663200433786422,.0813554292847049,.0308697147418303}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 8; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[8],
		T o_output[10]
	) {
		static const T matrix[10][8] = 
		{
			{-.0393249418872754,-.104517369553325,-.153045179908154,-.133760114824425,-.061787831583125,-.00756625226494136,-.000399939976821812,-.0051713573438354},
			{-.0346943378935191,.190423291624017,.568014771695437,.617049963166621,.281221505557339,-.00703540097040323,-.0190681438402154,.0300842611017119},
			{.0696230686764805,-.071274430440696,-.351601274540444,-.376607691431514,-.112751517622278,.0779699695773312,.0340995752367153,-.0293091146422881},
			{.000363120789231028,-.00388460872824636,-.00978448759299643,-.00756294859579786,.00320491223073097,.0119807916879719,.0116925265052733,.0054728800848884},
			{-.0010628404515272,.0239081315603922,.0751125591726019,.100479573533425,.0483097520108983,-.0618314320778367,-.135425858761273,-.0938445717540383},
			{-.0879291226435904,-.143236504733123,-.113382351515694,-.0501088745402918,-.00870691416149512,.00179635994822412,.00052035025728781,-.000257150494326289},
			{.00486443870328052,-.0669213772017063,-.222052529499926,-.340503237401605,-.31620229232248,-.172932573041675,-.0317668455888258,.0163961606293954},
			{.0125228635621091,.0954625928577346,.196159782438273,.17852935820801,.053253568711632,-.029934323539148,-.0135231501265476,.0116734773770315},
			{.0134824126994415,.0528142136964019,.0922127902762967,.0786666631889227,.0154241418606299,-.035620505637109,-.0386744602169855,-.016528567845558},
			{.0115410703001821,-.0839644563081351,-.238487403464336,-.247524582992525,-.083307216371032,.0663200433786422,.0813554292847049,.0308697147418303}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 8; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.4950154827e-3,.4979200234,.5632705854e-2,-.8993514245e-4,-.4470419751e-3,-.1770892152e-3,.1951962908e-3,-.1177982527e-2,.2113254103e-3,-.2562216676e-2},
			{-.1281141660,-.2838618647,.1050485042,-.5740150454e-1,-.8067496632e-1,-.1757166247e-1,.6413128635e-1,.1104039889e-1,.4097598670,.4776441091},
			{.3636314458,-5.308226489,3.294098734,.1415318972,.3835830171,-.1457440800,.1078714613,-.9819756352,-.6628918831,3.308121535},
			{.7413706065e-1,-.5931377528,.5333969803,.1268229038e-1,-.6197535211e-1,.3797220948,.1830849646,-.2001348350,-.7031616992e-1,.2425407198},
			{-.7323025910,.5168281488,.1209164195,.3948856607,.1069437875,-.3242423823e-1,-.2032058944,.6063540984e-1,.6262668914,-.3585435938},
			{5.535508427,-42.89042954,36.90564798,.7785369157,6.017043476,-1.003687260,-3.063422492,-13.89394497,-4.702699789,16.81744727},
			{-1.632837563,-3.134234566,1.791634164,-.1068573714,-.3443262408,-.3210806162,.6547234565,.1307590946,1.401688857,2.060530788},
			{1.022460692,-7.758938445,6.725713673,.1236608155,2.259628211,-.9571488741e-1,-1.316298742,-2.554081879,-.7979402353,2.891510802},
			{.5918762156e-1,-.6845178477e-1,.2194085493,-.8488708245e-4,.7357863397e-2,-.1161932079e-1,-.9505895631e-2,.3093727130,-.1699154828e-1,.1132669124e-1},
			{.9849025854,-.6924882517,-.1092209654,.1388908979,-.8552654630e-1,-.5817171720e-2,.2582914530,.5390718559e-1,-.3326184541,.2896792686}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.4950154827e-3,.4979200234,.5632705854e-2,-.8993514245e-4,-.4470419751e-3,-.1770892152e-3,.1951962908e-3,-.1177982527e-2,.2113254103e-3,-.2562216676e-2},
			{-.1281141660,-.2838618647,.1050485042,-.5740150454e-1,-.8067496632e-1,-.1757166247e-1,.6413128635e-1,.1104039889e-1,.4097598670,.4776441091},
			{.3636314458,-5.308226489,3.294098734,.1415318972,.3835830171,-.1457440800,.1078714613,-.9819756352,-.6628918831,3.308121535},
			{.7413706065e-1,-.5931377528,.5333969803,.1268229038e-1,-.6197535211e-1,.3797220948,.1830849646,-.2001348350,-.7031616992e-1,.2425407198},
			{-.7323025910,.5168281488,.1209164195,.3948856607,.1069437875,-.3242423823e-1,-.2032058944,.6063540984e-1,.6262668914,-.3585435938},
			{5.535508427,-42.89042954,36.90564798,.7785369157,6.017043476,-1.003687260,-3.063422492,-13.89394497,-4.702699789,16.81744727},
			{-1.632837563,-3.134234566,1.791634164,-.1068573714,-.3443262408,-.3210806162,.6547234565,.1307590946,1.401688857,2.060530788},
			{1.022460692,-7.758938445,6.725713673,.1236608155,2.259628211,-.9571488741e-1,-1.316298742,-2.554081879,-.7979402353,2.891510802},
			{.5918762156e-1,-.6845178477e-1,.2194085493,-.8488708245e-4,.7357863397e-2,-.1161932079e-1,-.9505895631e-2,.3093727130,-.1699154828e-1,.1132669124e-1},
			{.9849025854,-.6924882517,-.1092209654,.1388908979,-.8552654630e-1,-.5817171720e-2,.2582914530,.5390718559e-1,-.3326184541,.2896792686}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.7000634912,-1.685154308,-.6654054918,-.5338408486,.9256352574,-.1242356958e-1,1.769371557,.2413296780,1.297420291,-1.536996054},
			{.2730529330,-1.293625148,.7136446862,.7563684239e-2,-.2317468802,.6853506119e-1,.9441295150,-.3298992044,-.1365862042,.4849315585},
			{.2868254785e-2,.4998418584,.1337634241e-2,-.5906832877e-3,-.1595138647e-3,.3613023169e-4,-.3284285352e-2,-.5290909477e-3,.1251160462e-2,-.7714607282e-3},
			{1.816631343,-2.891972461,.6260680303,-6.253397703,12.00126132,.4402956750e-1,.2671831493,-.3887420399,16.93900600,-21.66006720},
			{.4830978084,-1.116774265,.5391184310,.2176216014e-1,-.1374614579,.5851045057,.3900113458,-.4909391690,-.2576157732,.4836964137},
			{.2706939016e-1,-.1077592549e-1,-.1897173064e-2,.3039698058,.2515251023e-1,.1306715462e-2,.1266539972e-2,-.3504390005e-2,.2447785717,-.8736604111e-1},
			{-.9416524900e-1,-.2558006803e-2,.2799166265,-.3564667680e-2,.1769475654e-1,.3341709151e-1,-.3785572809e-1,.2868573685,.4067762761e-1,-.2041981915e-1},
			{.1178627081e-2,.2066660683e-3,-.1299171970e-2,-.5766829992e-2,-.1257757669e-1,.4016587058e-4,.1023726354e-2,.3374059824e-3,.2458860523e-1,.4922683894},
			{3.440689837,-7.590270143,-1.640908069,-2.441593877,2.560514215,.8423160601e-3,5.435180613,.4769898667,5.332123365,-5.073568128},
			{2.590325858,-6.780286064,1.214522230,-.2015873803,-.8898164422,.3191395899,3.731892388,-1.100305025,-.5710565810,2.187171425}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.7000634912,-1.685154308,-.6654054918,-.5338408486,.9256352574,-.1242356958e-1,1.769371557,.2413296780,1.297420291,-1.536996054},
			{.2730529330,-1.293625148,.7136446862,.7563684239e-2,-.2317468802,.6853506119e-1,.9441295150,-.3298992044,-.1365862042,.4849315585},
			{.2868254785e-2,.4998418584,.1337634241e-2,-.5906832877e-3,-.1595138647e-3,.3613023169e-4,-.3284285352e-2,-.5290909477e-3,.1251160462e-2,-.7714607282e-3},
			{1.816631343,-2.891972461,.6260680303,-6.253397703,12.00126132,.4402956750e-1,.2671831493,-.3887420399,16.93900600,-21.66006720},
			{.4830978084,-1.116774265,.5391184310,.2176216014e-1,-.1374614579,.5851045057,.3900113458,-.4909391690,-.2576157732,.4836964137},
			{.2706939016e-1,-.1077592549e-1,-.1897173064e-2,.3039698058,.2515251023e-1,.1306715462e-2,.1266539972e-2,-.3504390005e-2,.2447785717,-.8736604111e-1},
			{-.9416524900e-1,-.2558006803e-2,.2799166265,-.3564667680e-2,.1769475654e-1,.3341709151e-1,-.3785572809e-1,.2868573685,.4067762761e-1,-.2041981915e-1},
			{.1178627081e-2,.2066660683e-3,-.1299171970e-2,-.5766829992e-2,-.1257757669e-1,.4016587058e-4,.1023726354e-2,.3374059824e-3,.2458860523e-1,.4922683894},
			{3.440689837,-7.590270143,-1.640908069,-2.441593877,2.560514215,.8423160601e-3,5.435180613,.4769898667,5.332123365,-5.073568128},
			{2.590325858,-6.780286064,1.214522230,-.2015873803,-.8898164422,.3191395899,3.731892388,-1.100305025,-.5710565810,2.187171425}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.1399963824,.6066373186,-.1500974604,-.5301262542e-1,-.4056576076e-1,.1575431901e-1,-.8843295445e-1,.5288654342e-2,.2014328850,.3629992388},
			{.9980310026,.2924601440e-2,.3115311091e-2,-.4407708503e-3,.2726146197e-3,-.5500013738e-3,-.1154421722e-2,.1015692083e-2,-.2448578657e-2,-.7654510784e-3},
			{.7025456159,-.1315631293,-.1055475837,.6653621745e-1,.1974878725e-2,.2579197294e-1,.5072486143e-1,-.4812672441e-1,.4581365691,-.2047268080e-1},
			{.4181649274e-1,-.3811120635,.2920165591,.4023103814e-1,.9452045314,-.7329653079e-3,-.1280528648,-.1066543470,-.1570488168,.4543324353},
			{1.638343036,-1.368183417,-.1704607967,.1363340614,-.1936805227,-.1169850622,.5238216155,.6326090150,-.7243942729,.6425963428},
			{-.8257675286,.6818471792,.1277018379,1.120145752,.1002201052,-.2933133878e-1,-.2686978346,.5216222775e-1,.3663393696,-.3246197716},
			{2.392311234,-2.050662096,-.3415608147,.3448039064,-.3065615554,-.5701246243e-1,.8080119633,.3253439474,-1.107767856,.9930937346},
			{-.3216179944,.1897104759e-2,.7798123431e-1,.1493632536e-1,.1056798012e-1,-.2050449058e-1,-.9366112062e-2,.3456459358e-1,1.244124064,-.3258270956e-1},
			{-.6719967917e-1,.8714671097,-.2693680159,.2781806232e-1,.1056746163,.5338541089e-2,.5502818178e-1,.7743308311e-1,-.8123541137e-1,.2750435099},
			{.4714077012,.5160632366,.2709475896,-.3054946501e-1,.2367285576e-1,-.1915082842e-1,-.1278403002,-.2352434188e-2,.1373537197e-1,-.1159337298}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.1399963824,.6066373186,-.1500974604,-.5301262542e-1,-.4056576076e-1,.1575431901e-1,-.8843295445e-1,.5288654342e-2,.2014328850,.3629992388},
			{.9980310026,.2924601440e-2,.3115311091e-2,-.4407708503e-3,.2726146197e-3,-.5500013738e-3,-.1154421722e-2,.1015692083e-2,-.2448578657e-2,-.7654510784e-3},
			{.7025456159,-.1315631293,-.1055475837,.6653621745e-1,.1974878725e-2,.2579197294e-1,.5072486143e-1,-.4812672441e-1,.4581365691,-.2047268080e-1},
			{.4181649274e-1,-.3811120635,.2920165591,.4023103814e-1,.9452045314,-.7329653079e-3,-.1280528648,-.1066543470,-.1570488168,.4543324353},
			{1.638343036,-1.368183417,-.1704607967,.1363340614,-.1936805227,-.1169850622,.5238216155,.6326090150,-.7243942729,.6425963428},
			{-.8257675286,.6818471792,.1277018379,1.120145752,.1002201052,-.2933133878e-1,-.2686978346,.5216222775e-1,.3663393696,-.3246197716},
			{2.392311234,-2.050662096,-.3415608147,.3448039064,-.3065615554,-.5701246243e-1,.8080119633,.3253439474,-1.107767856,.9930937346},
			{-.3216179944,.1897104759e-2,.7798123431e-1,.1493632536e-1,.1056798012e-1,-.2050449058e-1,-.9366112062e-2,.3456459358e-1,1.244124064,-.3258270956e-1},
			{-.6719967917e-1,.8714671097,-.2693680159,.2781806232e-1,.1056746163,.5338541089e-2,.5502818178e-1,.7743308311e-1,-.8123541137e-1,.2750435099},
			{.4714077012,.5160632366,.2709475896,-.3054946501e-1,.2367285576e-1,-.1915082842e-1,-.1278403002,-.2352434188e-2,.1373537197e-1,-.1159337298}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{-.6336264080e-1,-1.135754819,1.664552346,.3210331344e-1,-.1482634610,.2104632170,.4519073527,-.3871851132,-.8519491265e-1,.4607347118},
			{.4716762623e-2,.6726488544e-2,.9926276496,-.3104062577e-3,.8414196410e-3,-.1249169678e-2,-.2673305457e-2,.2299376267e-2,-.3915143001e-3,-.2587308856e-2},
			{-.6784257086e-1,.1105363257,.5538424796,.2806057303e-1,-.1147880340,.8903393740e-2,.5922049243,-.1622164194,-.6698926094e-1,.1182885870},
			{.3797708307,-.3081878614,-.8780150501e-1,.8381009252e-2,-.4790023917e-1,1.230553750,.1268289709,-.2821162673,-.1693324624,.1498037680},
			{1.103297561,-.9287832601,-.7481549834e-1,.7358347446e-1,-.1279413307,-.1199736642,.3482777909,.7809694467,-.4866508787,.4320363431},
			{-.1907857902e-1,.2172572895,-.2454577215,-.4727628721e-1,1.034727298,-.2748406196e-2,.1922444923,.9538093504e-1,.1667070084,-.3917560294},
			{.2700215195,.4878335311,.5005634674,-.1928937878e-1,.1944937474e-1,-.2618743263e-1,-.1025110355,-.1188399753e-1,-.1538691007e-2,-.1164573642},
			{.1031702025,-.4734999780,.1155713535,-.1525297738e-1,-.5907379311e-1,.8205137317e-2,1.289383899,-.5661229510e-1,.7068743713e-2,.8103970839e-1},
			{-.5567655583,.5752477523,-.2110643668e-1,-.2009854232e-1,.7891768764e-1,.4988803473,-.2133873767,.6672589846,.2586376782,-.2675845463},
			{.5978865906e-1,-.5377386777e-1,-.5926490512e-2,.3033920142e-2,-.7804677394e-2,-.1399934976e-1,.2083876183e-1,.9995794779,-.2732448950e-1,.2558804115e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{-.6336264080e-1,-1.135754819,1.664552346,.3210331344e-1,-.1482634610,.2104632170,.4519073527,-.3871851132,-.8519491265e-1,.4607347118},
			{.4716762623e-2,.6726488544e-2,.9926276496,-.3104062577e-3,.8414196410e-3,-.1249169678e-2,-.2673305457e-2,.2299376267e-2,-.3915143001e-3,-.2587308856e-2},
			{-.6784257086e-1,.1105363257,.5538424796,.2806057303e-1,-.1147880340,.8903393740e-2,.5922049243,-.1622164194,-.6698926094e-1,.1182885870},
			{.3797708307,-.3081878614,-.8780150501e-1,.8381009252e-2,-.4790023917e-1,1.230553750,.1268289709,-.2821162673,-.1693324624,.1498037680},
			{1.103297561,-.9287832601,-.7481549834e-1,.7358347446e-1,-.1279413307,-.1199736642,.3482777909,.7809694467,-.4866508787,.4320363431},
			{-.1907857902e-1,.2172572895,-.2454577215,-.4727628721e-1,1.034727298,-.2748406196e-2,.1922444923,.9538093504e-1,.1667070084,-.3917560294},
			{.2700215195,.4878335311,.5005634674,-.1928937878e-1,.1944937474e-1,-.2618743263e-1,-.1025110355,-.1188399753e-1,-.1538691007e-2,-.1164573642},
			{.1031702025,-.4734999780,.1155713535,-.1525297738e-1,-.5907379311e-1,.8205137317e-2,1.289383899,-.5661229510e-1,.7068743713e-2,.8103970839e-1},
			{-.5567655583,.5752477523,-.2110643668e-1,-.2009854232e-1,.7891768764e-1,.4988803473,-.2133873767,.6672589846,.2586376782,-.2675845463},
			{.5978865906e-1,-.5377386777e-1,-.5926490512e-2,.3033920142e-2,-.7804677394e-2,-.1399934976e-1,.2083876183e-1,.9995794779,-.2732448950e-1,.2558804115e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.000000006678052,0.2492867462770891,0.06214388170514502,0.01549164605322355,0.5014265110308812,0.1249989827878944,0.03116058958302124,0.2514285449611203,0.06267780356834177,0.1260729376986461},
			{1.000000002727766,0.2492867456812384,0.0621438815273546,0.01549164599022445,0.2492867457527619,0.06214388140720038,0.01549164594555115,0.06214388146785528,0.01549164592094434,0.01549164593222891},
			{1.000000003576573,0.5014265100496312,0.2514285446585445,0.1260729375549567,0.2492867458951233,0.124998982573936,0.06267780349997333,0.06214388154328702,0.03116058947820461,0.01549164598373243},
			{1.000000006950642,0.06308901532490006,0.003980223939943512,0.000251108433567182,0.8738219716002563,0.05512856710496877,0.003478006995493655,0.763564836837001,0.04817255308189202,0.6672197306714707},
			{1.000000000445023,0.06308901458801877,0.003980223798460351,0.0002511084180032569,0.06308901456325133,0.003980223769587246,0.0002511084050078769,0.003980223745597828,0.0002511083970989716,0.0002511083717996743},
			{0.9999999985620476,0.8738219689893114,0.7635648357521585,0.6672197300443169,0.06308901409666628,0.05512856629294544,0.0481725526432219,0.003980223644618228,0.003478006714843057,0.0002511083604818387},
			{1.000000000161897,0.3103524510692248,0.09631864390110025,0.02989272720994166,0.05314504983478008,0.01649369646636187,0.005118859127037292,0.002824396303404175,0.0008765583067880602,0.0001501026629394386},
			{1.000000005702582,0.6365024994350827,0.4051354313154271,0.2578697143150596,0.3103524521141743,0.1975401105953079,0.1257347740045679,0.09631864421075328,0.06130705744372503,0.02989272734321535},
			{1.00000000508223,0.05314505038816113,0.00282439644395872,0.0001501027108452568,0.6365024998244311,0.03382695712711142,0.001797735339921156,0.4051354314361555,0.02153094270228698,0.2578697144028026},
			{1.000000002301832,0.05314505010497485,0.002824396397559775,0.0001501027097547125,0.3103524514281697,0.01649369653006277,0.0008765583372290419,0.09631864389649525,0.005118859135323841,0.02989272715310334}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{1.000000006678052,0.2492867462770891,0.06214388170514502,0.01549164605322355,0.5014265110308812,0.1249989827878944,0.03116058958302124,0.2514285449611203,0.06267780356834177,0.1260729376986461},
			{1.000000002727766,0.2492867456812384,0.0621438815273546,0.01549164599022445,0.2492867457527619,0.06214388140720038,0.01549164594555115,0.06214388146785528,0.01549164592094434,0.01549164593222891},
			{1.000000003576573,0.5014265100496312,0.2514285446585445,0.1260729375549567,0.2492867458951233,0.124998982573936,0.06267780349997333,0.06214388154328702,0.03116058947820461,0.01549164598373243},
			{1.000000006950642,0.06308901532490006,0.003980223939943512,0.000251108433567182,0.8738219716002563,0.05512856710496877,0.003478006995493655,0.763564836837001,0.04817255308189202,0.6672197306714707},
			{1.000000000445023,0.06308901458801877,0.003980223798460351,0.0002511084180032569,0.06308901456325133,0.003980223769587246,0.0002511084050078769,0.003980223745597828,0.0002511083970989716,0.0002511083717996743},
			{0.9999999985620476,0.8738219689893114,0.7635648357521585,0.6672197300443169,0.06308901409666628,0.05512856629294544,0.0481725526432219,0.003980223644618228,0.003478006714843057,0.0002511083604818387},
			{1.000000000161897,0.3103524510692248,0.09631864390110025,0.02989272720994166,0.05314504983478008,0.01649369646636187,0.005118859127037292,0.002824396303404175,0.0008765583067880602,0.0001501026629394386},
			{1.000000005702582,0.6365024994350827,0.4051354313154271,0.2578697143150596,0.3103524521141743,0.1975401105953079,0.1257347740045679,0.09631864421075328,0.06130705744372503,0.02989272734321535},
			{1.00000000508223,0.05314505038816113,0.00282439644395872,0.0001501027108452568,0.6365024998244311,0.03382695712711142,0.001797735339921156,0.4051354314361555,0.02153094270228698,0.2578697144028026},
			{1.000000002301832,0.05314505010497485,0.002824396397559775,0.0001501027097547125,0.3103524514281697,0.01649369653006277,0.0008765583372290419,0.09631864389649525,0.005118859135323841,0.02989272715310334}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.1434916523,-.9004988097,.9030150961,-.1067453851,2.025016611,.6887840715e-2,-.4840826434,-.3594495672,.3079487054,-.5355835002},
			{-2.265454245,16.70409693,-17.47811933,-.9238674571e-1,-8.773089995,-.8041738950e-1,8.180764798,6.947881488,1.081296853,-4.224572364},
			{9.417271768,-65.82037606,59.44258085,1.162378865,8.367595220,-.5840800059,-5.766002909,-23.51280944,-7.365646221,24.65908794},
			{-6.577128526,51.45031916,-44.87319054,-.9840122000,-1.533261333,2.517273589,-2.220722120,16.87587121,5.779366347,-20.43451559},
			{-.3697276700,-.2378553637,-2.431893641,2.479189888,-11.30779609,-.1172879445,2.633254871,1.227548245,-8.085375399,16.20994310},
			{-4.826837002,30.86403973,34.85547730,-3.807684718,38.45529561,2.129234571,-46.14403174,-18.77992157,16.08919042,-48.83476262},
			{-16.90178689,16.69830880,-57.64972520,.3102882567,-22.35517921,-5.786060334,29.86183199,39.71875072,3.913267871,12.19030399},
			{.2657870504,.4414875731,2.332202047,-9.437592078,18.28111325,.1023729643,-2.687194351,-1.139948583,28.48623061,-36.64445849},
			{39.37055855,-74.69000466,-22.53375735,8.729262281,-36.02789597,-.5323176822,48.86180115,8.862117521,-45.53361422,73.49385036},
			{-2.230099397,2.705096812,-.4749974040,8.897158303,-8.826093353,-.7106493971e-1,-.2570037770,.4173833622,-20.59489498,20.43451538}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[10],
		T o_output[10]
	) {
		static const T matrix[10][10] = 
		{
			{.1434916523,-.9004988097,.9030150961,-.1067453851,2.025016611,.6887840715e-2,-.4840826434,-.3594495672,.3079487054,-.5355835002},
			{-2.265454245,16.70409693,-17.47811933,-.9238674571e-1,-8.773089995,-.8041738950e-1,8.180764798,6.947881488,1.081296853,-4.224572364},
			{9.417271768,-65.82037606,59.44258085,1.162378865,8.367595220,-.5840800059,-5.766002909,-23.51280944,-7.365646221,24.65908794},
			{-6.577128526,51.45031916,-44.87319054,-.9840122000,-1.533261333,2.517273589,-2.220722120,16.87587121,5.779366347,-20.43451559},
			{-.3697276700,-.2378553637,-2.431893641,2.479189888,-11.30779609,-.1172879445,2.633254871,1.227548245,-8.085375399,16.20994310},
			{-4.826837002,30.86403973,34.85547730,-3.807684718,38.45529561,2.129234571,-46.14403174,-18.77992157,16.08919042,-48.83476262},
			{-16.90178689,16.69830880,-57.64972520,.3102882567,-22.35517921,-5.786060334,29.86183199,39.71875072,3.913267871,12.19030399},
			{.2657870504,.4414875731,2.332202047,-9.437592078,18.28111325,.1023729643,-2.687194351,-1.139948583,28.48623061,-36.64445849},
			{39.37055855,-74.69000466,-22.53375735,8.729262281,-36.02789597,-.5323176822,48.86180115,8.862117521,-45.53361422,73.49385036},
			{-2.230099397,2.705096812,-.4749974040,8.897158303,-8.826093353,-.7106493971e-1,-.2570037770,.4173833622,-20.59489498,20.43451538}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[10],
		T o_output[1]
	) {
		static const T matrix[1][10] = 
		{
			{.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[10],
		T o_output[1]
	) {
		static const T matrix[1][10] = 
		{
			{.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 3900;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[10][10] = 
		{
			{10.01343146347375,3.555936538441215,3.555937098156799,-2.443949673945655,1.221975100338516,1.221974836791108,-1.692409305965417,-2.712038407274934,4.404446223441276,-2.712038613236915},
			{3.555936538441217,10.01343417086613,3.555936391892739,1.221975581963253,-2.443951207983593,1.221975965780719,4.404448518811391,-1.692410447499592,-2.712038761013454,4.404448776516672},
			{3.555937098156799,3.555936391892738,10.01343367527705,1.221975646054362,1.221976509869574,-2.443951048927955,-2.712039122432838,4.404445702989177,-1.692409835195898,-1.692410602929889},
			{-2.443949673945654,1.221975581963253,1.221975646054363,38.49545879518467,0.4199242800956394,0.4199241693029314,-0.5815858743757831,-0.9319749716238978,1.513560482333602,-0.9319749598961085},
			{1.221975100338516,-2.443951207983592,1.221976509869574,0.4199242800956392,38.49546134962688,0.4199219182622189,1.513558163691415,-0.5815859893830884,-0.9319745106740631,1.513562152147228},
			{1.221974836791108,1.221975965780719,-2.443951048927956,0.4199241693029314,0.4199219182622188,38.49545988947659,-0.9319731523868369,1.513561523012706,-0.581585805429657,-0.581585839144423},
			{-1.692409305965416,4.404448518811394,-2.712039122432836,-0.5815858743757829,1.513558163691415,-0.9319731523868361,21.36325840413494,1.388221761191283,1.388221169100666,-2.678985707761788},
			{-2.712038407274936,-1.692410447499594,4.404445702989176,-0.9319749716238981,-0.5815859893830885,1.513561523012706,1.388221761191283,21.3632568199595,1.388221789541477,0.7080263342095845},
			{4.404446223441278,-2.712038761013454,-1.692409835195899,1.513560482333602,-0.9319745106740633,-0.5815858054296573,1.388221169100667,1.388221789541476,21.36325634891856,1.970958521976558},
			{-2.712038613236918,4.404448776516673,-1.692410602929889,-0.9319749598961082,1.513562152147227,-0.5815858391444229,-2.67898570776179,0.7080263342095842,1.970958521976558,21.36326241046861}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[10][10] = 
		{
			{1.196550164,.4793816273e-1,-.6424538824,.1017586202,-.2089231459e-1,-.4523178164,-.1165631648,.5765375751,-.3681999783,.9177441691e-1},
			{-1.614157763,-.8606081071,2.108557451,-.1613326088,-.9592368542e-1,1.099069750,.8388856257,-.9931250716,.4243741011,.2701346789},
			{-.4939467924,.8570090710,-.5988882699,-.6431307239e-1,.1109579345,-.5170067994,-.2531039171,-.3454402975,.2366351205,-.3619090993},
			{.4116664741,-.3141941727,-.7707910669e-1,.2332852119,-.1225866642e-1,-.2828216365e-1,.1417392661,.6437806545e-1,-.9433499315e-1,.1117823332},
			{-.2959893810,-.8562776001e-1,.3965061501,-.7123702935e-1,-.1911527625,.1400168279,.2047874035e-1,-.2146465232,.1365311129,-.5763669031e-1},
			{-.3591310055e-1,.1298337006,-.1292028211,-.6344574504e-2,.1924813370e-1,-.8327498158e-1,-.5929460679e-1,-.1688283697e-1,.2203178642e-1,-.5414565976e-1},
			{.6433838849,.4457654845,-1.216746222,.7461356425e-1,.8418363048e-1,-.4132928908,-.6199578990,.5975247067,-.1733826073,-.1958158126},
			{.2848632824,-.7577047154,.6800633411,.4271752329e-1,-.1067172984,.4773099294,.3056284364,.1683813182,-.1559978518,.3180119185},
			{-1.022345454,.5101078166,.4326126220,-.3745944684,-.3474159621e-1,.2505216067,-.2131210396,-.3416844211,.4326093507,-.1221960989},
			{.9258886745,.2748060201e-1,-.9533692907,.2254468336,.2472966162,-.4727434602,-.4469144601e-1,.5049574983,-.4602660349,-.2102989041e-7}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[10][10] = 
		{
			{1.052802072,-.6973086860,-.1196673474,.6988407343,-.1439101562,-.8456864235e-1,.3262866284,-.2580407541e-1,.6142652878e-1,.3619090881},
			{-1.685580211,1.254675024,-.1711293401,-.6911086538,.4743431337,.5882170241,-.8429021664,.8355349005e-1,.6675742418,-.9177438926e-1},
			{-.2344371164,.3541879483,.2464576610,-.1374772071,-.2065459391,-.4977903149,.2238063038,-.5269680023,.3302707068e-1,-.2701346645},
			{.1971134137,-.1423120799,-.1951911635e-1,.1104793775,-.2480795644e-1,-.1530000035e-1,.6218762146e-1,.1756684490e-1,-.2560880700e-1,.5414565701e-1},
			{-.3050594429,.1912399183,.9342635488e-1,-.1478498856,-.1117594088,.6686491302e-1,-.1963366736,-.6693481936e-1,.1514891566,-.1117823336},
			{-.8227822257e-1,-.1286918288,.1960810361,.8910825944e-2,-.1913624085e-1,.1325983776,.6992114560e-1,-.5355542637e-1,.4127095947e-1,.5763667963e-1},
			{.6740174926,-.3057624077,-.2886300575,.2836168264,.7884249473e-1,-.2036448620,.2177233347,.1636411322,-.2591683549,.1221960935},
			{.5089731701,-.3620400861,-.1933626420e-1,.3893313950e-1,.3817128319e-1,.1773912651,.1769300339e-1,.4243595358,-.2462367082,.1958158156},
			{-1.078920380,.7619007810,.1097977345,-.6370886093,.1402496060,.8352885947e-1,-.3386452082,-.6055014888e-1,.8118344680e-1,-.3180118993},
			{.9533692458,-.9258886590,-.2748057619e-1,.4727434581,-.2254468468,-.2472966239,.4602660439,.4469142960e-1,-.5049574849,-.6140337311e-8}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[8][10] = 
		{
			{-1.45425377932711,1.32618123974878,.214929690789362,1.72546667234598,.0749963295610948,-.0493986167919454,-.52096553277996,.0893348435663733,-.178270751151675,-.228020095960901},
			{.901617182972597,-.878907514262425,-.154299052712732,1.32719741075668,-.237190144283086,.0181176435188071,.36814359005261,-.0122740843570015,-1.08754546709488,.755140435409436},
			{2.76631253535879,-2.68655311455725,-.48375671187014,.80135987241094,-.485537526595039,-.00945790998173296,1.10972468290512,.179128119201145,-1.71166846020543,1.52044851333359},
			{2.54784227751401,-2.5501417334435,-.585993500196011,.346798469815348,-.450179502588528,-.132679980584154,1.09020785565277,.767598933424162,-1.39841366170664,1.36496084211255},
			{.780302598025293,-.765805601220523,-.602789949818311,.0717332396365517,-.183222802955387,-.124571448359104,.427867625824713,1.37006771146173,-.432974482491953,.45939310989698},
			{-.643005715312524,1.06666564380352,-.827657213013815,-.0332833473253481,.0759860981521238,.263661688539869,-.244989278148133,1.44637652504956,.319061272858217,-.422815674603482},
			{-.559184637570565,1.76597055534194,-1.33837529716116,-.0424577813150609,.158513281569754,.99206941222702,-.469111858018255,.880750738644108,.343501557358165,-.731675971075964},
			{.368586029145061,1.57122965010233,-1.85295852760903,-.0264099052189327,.111658320369302,1.66581597019539,-.354409845276218,.172073747789449,-.0456839969593226,-.609901442538053}
		};
		for (int j = 0; j < 8; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[8][10] = 
		{
			{.136238006771799,-.905026226343549,.855645365157245,-.061171775023035,1.80763727482326,.00459888177778651,-.432860546853416,-.33552263668457,.158481745582043,-.22802008920756},
			{.1063063675187,-.917274894712635,.679379143532882,.0571071372659607,1.05507075503056,-.00405298238561168,-.244413410198392,-.245992766476126,-.241269790385193,.755140440809851},
			{.0409630703466698,-.895962207967806,.451001852692577,.0690460542762892,.25344249350708,-.0161241086859924,-.0140519380454184,-.126817111625959,-.281946622110677,1.52044851761324},
			{-.114933838641156,-.739912335428009,.266553225860955,-.0622041723066864,-.145086497875149,-.0287703389043051,.125597218611059,-.0198798029625122,.353675695323054,1.36496084632275},
			{-.444249450430285,-.32622864510776,.182185148199971,-.100852011776358,-.0938162473078612,-.0413927496187602,.0799487290695668,.0542549849480529,1.2307571273412,.459393114682234},
			{-.973574506362864,.375422165711069,.194155058565812,.241830824758655,.11908488572855,-.0545512699657424,-.153022370217456,.0988758653826939,1.57459501609323,-.422815669693946},
			{-1.59088292224111,1.20318902343184,.256104520240868,.954291439404765,.221213087038666,-.0673796136424686,-.473431870973027,.125941445012019,1.10263085872931,-.731675967000855},
			{-2.06344072573252,1.83764375032395,.312654127890014,1.63432547259165,.19337761091637,-.0766386977395684,-.726653301497992,.141607355522673,.357025847311682,-.609901439586253}
		};
		for (int j = 0; j < 8; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[8][10] = 
		{
			{.776953679829385,.685465564650613,-1.37556209400805,-.00717427718582053,.0209988307739315,1.73723983109591,-.0961080517716217,-.247417655554891,-.266375731459099,-.228020096370379},
			{.939984561275504,-1.71258571660202,.641011771672815,.034936511000544,-.21501951926035,1.28820791831646,.601862270909948,-.858549768800081,-.474988470363239,.755140441850394},
			{.975721629867719,-3.62131168822314,2.24159276929892,.0623798553302253,-.478871328737102,.722855910166654,1.41566990956488,-1.25059373153752,-.587891850478623,1.52044852474798},
			{.737612879067618,-3.40268846552148,2.07678263375027,.0417054687438725,-.554089143978086,.276322663635704,1.87768658543182,-.984490437069049,-.433803036111242,1.36496085205056},
			{.340725637124317,-1.55078069903093,.621762112284199,-.01767331420709,-.266401500287098,.0480138046448678,1.743680345325,-.293663908629057,-.0850555914960167,.459393114271799},
			{.048237755650778,.0448533753610167,-.497088414299526,-.0763821350003243,.39419905756096,-.0114524825777734,1.10251137688036,.190842775774665,.227094364146653,-.42281567349681},
			{.00359688925636023,.171490737841793,-.306677008860344,-.10515758714779,1.21796230595075,-.00467980786731078,.285697435187406,.12162143376829,.347821569515127,-.731675967644279},
			{.102171927666744,-.59438301103978,.579068229500308,-.1081291952931,1.85411298420521,.00508059295826713,-.323943449747462,-.230636099361268,.326559455490798,-.609901434379722}
		};
		for (int j = 0; j < 8; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[10][8] = 
		{
			{.104094590346149,-.141776721078137,-.613634778917128,-.653409853364533,-.200113409945121,.142634161872485,.0879301835515604,-.0263831610799882},
			{-.0949272367992553,.138205690681529,.595942217456873,.653999563008246,.196395565775554,-.236612142752328,-.277693814609261,-.112467651170763},
			{-.0153845349649324,.0242630843468947,.107308895558184,.150281644360358,.154589197375555,.183594313525759,.210455684282643,.132633630802123},
			{-.123507842281073,-.208697993643045,-.177760929704989,-.0889386047593935,-.0183964313670215,.00738304845043946,.00667636457332789,.001890404758733},
			{-.0053681911053648,.0372974712145967,.107703923175648,.115451307708537,.0469886169441115,-.0168555475725761,-.024925759771127,-.00799243383982093},
			{.00353592258223542,-.00284894758026933,.00209799233690204,.0340265986725407,.0319471156139029,-.058486515853891,-.156000075202356,-.119238081739623},
			{.0372903921516954,-.057889525698158,-.246163675199138,-.279590523081356,-.109729289413606,.0543445252884263,.0737664968067963,.0253684385649251},
			{-.00639453311167032,.00193006462968555,-.0397349332076028,-.196855476870741,-.351362307763876,-.320841165933044,-.138495532439186,-.0123169329453331},
			{.0127605106313839,.171013411523225,.379689309769675,.358631801388526,.111038974277361,-.0707754785942782,-.0540146365965548,.00327003238118161},
			{.0163215381092027,-.118743671824099,-.337272117813443,-.350052619647759,-.117814194082433,.0937907050239907,.115053951976384,.0436563698269361}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 8; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[10][8] = 
		{
			{-.00689558700631383,-.0118202599888947,-.00642519370165688,.0208423197182862,.080561035722907,.152708396451657,.176891094971095,.104439542193928},
			{.0458072400985804,.101992269982154,.140534649547464,.134177102590675,.0591589196270397,-.0588862141967161,-.13378320983686,-.093010993534226},
			{-.0433078639492565,-.0755405183624456,-.0707411392464508,-.048337266213479,-.0330377994122565,-.0304538661014528,-.028476394069675,-.0158247598657252},
			{.00309616462393304,-.00634977212993765,-.0108301030490586,.0112802222770543,.0182886945961844,-.0379319684524605,-.106108158732416,-.0827201877062657},
			{-.091492237737142,-.117313862962601,-.0397532972805173,.0263102599832379,.0170128157579876,-.0186788600383275,-.0245967975651416,-.00978766625219707},
			{-.000232769035468907,.000450653207766819,.00252912002682659,.00521726768144552,.0075062395180826,.00855654796466611,.0074919740914405,.00387901159768857},
			{.0219089197879117,.0271764534970979,.00220409317615605,-.0227760372138291,-.0144980537676211,.0240020672526013,.0526411346050896,.0367790250506059},
			{.0169822327019335,.0273520629385043,.0198916853640507,.00360504107563016,-.00983870160404794,-.0155090080437392,-.0140034944111629,-.0071673526637556},
			{-.00802143756700706,.0268269127841043,.0442242645694705,-.064136219634369,-.223187825681996,-.246980460557323,-.122602095492219,-.0180706019705966},
			{.0115410699376412,-.0839644561924366,-.238487402345228,-.247524581953833,-.0833072164454612,.0663200427821505,.0813554292131623,.030869715005355}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 8; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[10][8] = 
		{
			{-.0393249418872754,-.104517369553325,-.153045179908154,-.133760114824425,-.061787831583125,-.00756625226494136,-.000399939976821812,-.0051713573438354},
			{-.0346943378935191,.190423291624017,.568014771695437,.617049963166621,.281221505557339,-.00703540097040323,-.0190681438402154,.0300842611017119},
			{.0696230686764805,-.071274430440696,-.351601274540444,-.376607691431514,-.112751517622278,.0779699695773312,.0340995752367153,-.0293091146422881},
			{.000363120789231028,-.00388460872824636,-.00978448759299643,-.00756294859579786,.00320491223073097,.0119807916879719,.0116925265052733,.0054728800848884},
			{-.0010628404515272,.0239081315603922,.0751125591726019,.100479573533425,.0483097520108983,-.0618314320778367,-.135425858761273,-.0938445717540383},
			{-.0879291226435904,-.143236504733123,-.113382351515694,-.0501088745402918,-.00870691416149512,.00179635994822412,.00052035025728781,-.000257150494326289},
			{.00486443870328052,-.0669213772017063,-.222052529499926,-.340503237401605,-.31620229232248,-.172932573041675,-.0317668455888258,.0163961606293954},
			{.0125228635621091,.0954625928577346,.196159782438273,.17852935820801,.053253568711632,-.029934323539148,-.0135231501265476,.0116734773770315},
			{.0134824126994415,.0528142136964019,.0922127902762967,.0786666631889227,.0154241418606299,-.035620505637109,-.0386744602169855,-.016528567845558},
			{.0115410703001821,-.0839644563081351,-.238487403464336,-.247524582992525,-.083307216371032,.0663200433786422,.0813554292847049,.0308697147418303}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 8; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[10][10] = 
		{
			{.4950154827e-3,.4979200234,.5632705854e-2,-.8993514245e-4,-.4470419751e-3,-.1770892152e-3,.1951962908e-3,-.1177982527e-2,.2113254103e-3,-.2562216676e-2},
			{-.1281141660,-.2838618647,.1050485042,-.5740150454e-1,-.8067496632e-1,-.1757166247e-1,.6413128635e-1,.1104039889e-1,.4097598670,.4776441091},
			{.3636314458,-5.308226489,3.294098734,.1415318972,.3835830171,-.1457440800,.1078714613,-.9819756352,-.6628918831,3.308121535},
			{.7413706065e-1,-.5931377528,.5333969803,.1268229038e-1,-.6197535211e-1,.3797220948,.1830849646,-.2001348350,-.7031616992e-1,.2425407198},
			{-.7323025910,.5168281488,.1209164195,.3948856607,.1069437875,-.3242423823e-1,-.2032058944,.6063540984e-1,.6262668914,-.3585435938},
			{5.535508427,-42.89042954,36.90564798,.7785369157,6.017043476,-1.003687260,-3.063422492,-13.89394497,-4.702699789,16.81744727},
			{-1.632837563,-3.134234566,1.791634164,-.1068573714,-.3443262408,-.3210806162,.6547234565,.1307590946,1.401688857,2.060530788},
			{1.022460692,-7.758938445,6.725713673,.1236608155,2.259628211,-.9571488741e-1,-1.316298742,-2.554081879,-.7979402353,2.891510802},
			{.5918762156e-1,-.6845178477e-1,.2194085493,-.8488708245e-4,.7357863397e-2,-.1161932079e-1,-.9505895631e-2,.3093727130,-.1699154828e-1,.1132669124e-1},
			{.9849025854,-.6924882517,-.1092209654,.1388908979,-.8552654630e-1,-.5817171720e-2,.2582914530,.5390718559e-1,-.3326184541,.2896792686}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[10][10] = 
		{
			{.7000634912,-1.685154308,-.6654054918,-.5338408486,.9256352574,-.1242356958e-1,1.769371557,.2413296780,1.297420291,-1.536996054},
			{.2730529330,-1.293625148,.7136446862,.7563684239e-2,-.2317468802,.6853506119e-1,.9441295150,-.3298992044,-.1365862042,.4849315585},
			{.2868254785e-2,.4998418584,.1337634241e-2,-.5906832877e-3,-.1595138647e-3,.3613023169e-4,-.3284285352e-2,-.5290909477e-3,.1251160462e-2,-.7714607282e-3},
			{1.816631343,-2.891972461,.6260680303,-6.253397703,12.00126132,.4402956750e-1,.2671831493,-.3887420399,16.93900600,-21.66006720},
			{.4830978084,-1.116774265,.5391184310,.2176216014e-1,-.1374614579,.5851045057,.3900113458,-.4909391690,-.2576157732,.4836964137},
			{.2706939016e-1,-.1077592549e-1,-.1897173064e-2,.3039698058,.2515251023e-1,.1306715462e-2,.1266539972e-2,-.3504390005e-2,.2447785717,-.8736604111e-1},
			{-.9416524900e-1,-.2558006803e-2,.2799166265,-.3564667680e-2,.1769475654e-1,.3341709151e-1,-.3785572809e-1,.2868573685,.4067762761e-1,-.2041981915e-1},
			{.1178627081e-2,.2066660683e-3,-.1299171970e-2,-.5766829992e-2,-.1257757669e-1,.4016587058e-4,.1023726354e-2,.3374059824e-3,.2458860523e-1,.4922683894},
			{3.440689837,-7.590270143,-1.640908069,-2.441593877,2.560514215,.8423160601e-3,5.435180613,.4769898667,5.332123365,-5.073568128},
			{2.590325858,-6.780286064,1.214522230,-.2015873803,-.8898164422,.3191395899,3.731892388,-1.100305025,-.5710565810,2.187171425}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[10][10] = 
		{
			{.1399963824,.6066373186,-.1500974604,-.5301262542e-1,-.4056576076e-1,.1575431901e-1,-.8843295445e-1,.5288654342e-2,.2014328850,.3629992388},
			{.9980310026,.2924601440e-2,.3115311091e-2,-.4407708503e-3,.2726146197e-3,-.5500013738e-3,-.1154421722e-2,.1015692083e-2,-.2448578657e-2,-.7654510784e-3},
			{.7025456159,-.1315631293,-.1055475837,.6653621745e-1,.1974878725e-2,.2579197294e-1,.5072486143e-1,-.4812672441e-1,.4581365691,-.2047268080e-1},
			{.4181649274e-1,-.3811120635,.2920165591,.4023103814e-1,.9452045314,-.7329653079e-3,-.1280528648,-.1066543470,-.1570488168,.4543324353},
			{1.638343036,-1.368183417,-.1704607967,.1363340614,-.1936805227,-.1169850622,.5238216155,.6326090150,-.7243942729,.6425963428},
			{-.8257675286,.6818471792,.1277018379,1.120145752,.1002201052,-.2933133878e-1,-.2686978346,.5216222775e-1,.3663393696,-.3246197716},
			{2.392311234,-2.050662096,-.3415608147,.3448039064,-.3065615554,-.5701246243e-1,.8080119633,.3253439474,-1.107767856,.9930937346},
			{-.3216179944,.1897104759e-2,.7798123431e-1,.1493632536e-1,.1056798012e-1,-.2050449058e-1,-.9366112062e-2,.3456459358e-1,1.244124064,-.3258270956e-1},
			{-.6719967917e-1,.8714671097,-.2693680159,.2781806232e-1,.1056746163,.5338541089e-2,.5502818178e-1,.7743308311e-1,-.8123541137e-1,.2750435099},
			{.4714077012,.5160632366,.2709475896,-.3054946501e-1,.2367285576e-1,-.1915082842e-1,-.1278403002,-.2352434188e-2,.1373537197e-1,-.1159337298}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[10][10] = 
		{
			{-.6336264080e-1,-1.135754819,1.664552346,.3210331344e-1,-.1482634610,.2104632170,.4519073527,-.3871851132,-.8519491265e-1,.4607347118},
			{.4716762623e-2,.6726488544e-2,.9926276496,-.3104062577e-3,.8414196410e-3,-.1249169678e-2,-.2673305457e-2,.2299376267e-2,-.3915143001e-3,-.2587308856e-2},
			{-.6784257086e-1,.1105363257,.5538424796,.2806057303e-1,-.1147880340,.8903393740e-2,.5922049243,-.1622164194,-.6698926094e-1,.1182885870},
			{.3797708307,-.3081878614,-.8780150501e-1,.8381009252e-2,-.4790023917e-1,1.230553750,.1268289709,-.2821162673,-.1693324624,.1498037680},
			{1.103297561,-.9287832601,-.7481549834e-1,.7358347446e-1,-.1279413307,-.1199736642,.3482777909,.7809694467,-.4866508787,.4320363431},
			{-.1907857902e-1,.2172572895,-.2454577215,-.4727628721e-1,1.034727298,-.2748406196e-2,.1922444923,.9538093504e-1,.1667070084,-.3917560294},
			{.2700215195,.4878335311,.5005634674,-.1928937878e-1,.1944937474e-1,-.2618743263e-1,-.1025110355,-.1188399753e-1,-.1538691007e-2,-.1164573642},
			{.1031702025,-.4734999780,.1155713535,-.1525297738e-1,-.5907379311e-1,.8205137317e-2,1.289383899,-.5661229510e-1,.7068743713e-2,.8103970839e-1},
			{-.5567655583,.5752477523,-.2110643668e-1,-.2009854232e-1,.7891768764e-1,.4988803473,-.2133873767,.6672589846,.2586376782,-.2675845463},
			{.5978865906e-1,-.5377386777e-1,-.5926490512e-2,.3033920142e-2,-.7804677394e-2,-.1399934976e-1,.2083876183e-1,.9995794779,-.2732448950e-1,.2558804115e-1}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[10][10] = 
		{
			{1.000000006678052,0.2492867462770891,0.06214388170514502,0.01549164605322355,0.5014265110308812,0.1249989827878944,0.03116058958302124,0.2514285449611203,0.06267780356834177,0.1260729376986461},
			{1.000000002727766,0.2492867456812384,0.0621438815273546,0.01549164599022445,0.2492867457527619,0.06214388140720038,0.01549164594555115,0.06214388146785528,0.01549164592094434,0.01549164593222891},
			{1.000000003576573,0.5014265100496312,0.2514285446585445,0.1260729375549567,0.2492867458951233,0.124998982573936,0.06267780349997333,0.06214388154328702,0.03116058947820461,0.01549164598373243},
			{1.000000006950642,0.06308901532490006,0.003980223939943512,0.000251108433567182,0.8738219716002563,0.05512856710496877,0.003478006995493655,0.763564836837001,0.04817255308189202,0.6672197306714707},
			{1.000000000445023,0.06308901458801877,0.003980223798460351,0.0002511084180032569,0.06308901456325133,0.003980223769587246,0.0002511084050078769,0.003980223745597828,0.0002511083970989716,0.0002511083717996743},
			{0.9999999985620476,0.8738219689893114,0.7635648357521585,0.6672197300443169,0.06308901409666628,0.05512856629294544,0.0481725526432219,0.003980223644618228,0.003478006714843057,0.0002511083604818387},
			{1.000000000161897,0.3103524510692248,0.09631864390110025,0.02989272720994166,0.05314504983478008,0.01649369646636187,0.005118859127037292,0.002824396303404175,0.0008765583067880602,0.0001501026629394386},
			{1.000000005702582,0.6365024994350827,0.4051354313154271,0.2578697143150596,0.3103524521141743,0.1975401105953079,0.1257347740045679,0.09631864421075328,0.06130705744372503,0.02989272734321535},
			{1.00000000508223,0.05314505038816113,0.00282439644395872,0.0001501027108452568,0.6365024998244311,0.03382695712711142,0.001797735339921156,0.4051354314361555,0.02153094270228698,0.2578697144028026},
			{1.000000002301832,0.05314505010497485,0.002824396397559775,0.0001501027097547125,0.3103524514281697,0.01649369653006277,0.0008765583372290419,0.09631864389649525,0.005118859135323841,0.02989272715310334}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[10][10] = 
		{
			{.1434916523,-.9004988097,.9030150961,-.1067453851,2.025016611,.6887840715e-2,-.4840826434,-.3594495672,.3079487054,-.5355835002},
			{-2.265454245,16.70409693,-17.47811933,-.9238674571e-1,-8.773089995,-.8041738950e-1,8.180764798,6.947881488,1.081296853,-4.224572364},
			{9.417271768,-65.82037606,59.44258085,1.162378865,8.367595220,-.5840800059,-5.766002909,-23.51280944,-7.365646221,24.65908794},
			{-6.577128526,51.45031916,-44.87319054,-.9840122000,-1.533261333,2.517273589,-2.220722120,16.87587121,5.779366347,-20.43451559},
			{-.3697276700,-.2378553637,-2.431893641,2.479189888,-11.30779609,-.1172879445,2.633254871,1.227548245,-8.085375399,16.20994310},
			{-4.826837002,30.86403973,34.85547730,-3.807684718,38.45529561,2.129234571,-46.14403174,-18.77992157,16.08919042,-48.83476262},
			{-16.90178689,16.69830880,-57.64972520,.3102882567,-22.35517921,-5.786060334,29.86183199,39.71875072,3.913267871,12.19030399},
			{.2657870504,.4414875731,2.332202047,-9.437592078,18.28111325,.1023729643,-2.687194351,-1.139948583,28.48623061,-36.64445849},
			{39.37055855,-74.69000466,-22.53375735,8.729262281,-36.02789597,-.5323176822,48.86180115,8.862117521,-45.53361422,73.49385036},
			{-2.230099397,2.705096812,-.4749974040,8.897158303,-8.826093353,-.7106493971e-1,-.2570037770,.4173833622,-20.59489498,20.43451538}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][10] = 
		{
			{.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634,.3449406634}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
