
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION2D_MATRICES_CONST_HPP_
#define CCOMPUTATION2D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (4)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (15)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (15)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS (10)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (27)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*4+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_2D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[15],
		T io_hv[15]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{224.99996,-857.32140,1626.65301,-1428.70571,469.574280,-857.32140,1328.15664,-1106.67069,354.964788,1626.65301,-1106.67069,336.749168,-1428.70571,354.964788,469.574280},
			{-857.32140,4900.00002,-11156.51558,10887.64438,-3834.057902,2450.00002,-6072.83788,6325.14980,-2318.620276,-3718.83852,4216.76653,-1649.727254,2721.911098,-1159.310137,-766.811580},
			{1626.65301,-11156.51558,28224.,-29511.21822,10911.92009,-3718.83852,11522.39975,-13715.57480,5499.090835,4704.,-6857.787398,3130.137376,-2951.121822,1649.727251,727.4613394},
			{-1428.70571,10887.64438,-29511.21822,32400.,-12423.76754,2721.911098,-9638.323502,12548.46604,-5366.563145,-2951.121822,5019.386416,-2545.584412,1620.,-1073.312629,-354.9647869},
			{469.574280,-3834.057902,10911.92009,-12423.76754,4900.,-766.811580,2969.848480,-4124.318126,1852.025918,727.4613394,-1374.772709,752.9940236,-354.9647869,264.5751311,70.},
			{-857.32140,2450.00002,-3718.83852,2721.911098,-766.811580,4900.00002,-6072.83788,4216.76653,-1159.310137,-11156.51558,6325.14980,-1649.727254,10887.64438,-2318.620276,-3834.057902},
			{1328.15664,-6072.83788,11522.39975,-9638.323502,2969.848480,-6072.83788,12544.,-11198.71993,3591.991091,11522.39975,-11198.71993,3833.619699,-9638.323502,3591.991091,2969.848480},
			{-1106.67069,6325.14980,-13715.57480,12548.46604,-4124.318126,4216.76653,-11198.71993,11664.,-4156.921939,-6857.787398,8748.,-3549.242172,5019.386416,-2494.153164,-1374.772709},
			{354.964788,-2318.620276,5499.090835,-5366.563145,1852.025918,-1159.310137,3591.991091,-4156.921939,1600.,1649.727251,-2494.153164,1138.419957,-1073.312629,640.,264.5751311},
			{1626.65301,-3718.83852,4704.,-2951.121822,727.4613394,-11156.51558,11522.39975,-6857.787398,1649.727251,28224.,-13715.57480,3130.137376,-29511.21822,5499.090835,10911.92009},
			{-1106.67069,4216.76653,-6857.787398,5019.386416,-1374.772709,6325.14980,-11198.71993,8748.,-2494.153164,-13715.57480,11664.,-3549.242172,12548.46604,-4156.921939,-4124.318126},
			{336.749168,-1649.727254,3130.137376,-2545.584412,752.9940236,-1649.727254,3833.619699,-3549.242172,1138.419957,3130.137376,-3549.242172,1296.,-2545.584412,1138.419957,752.9940236},
			{-1428.70571,2721.911098,-2951.121822,1620.,-354.9647869,10887.64438,-9638.323502,5019.386416,-1073.312629,-29511.21822,12548.46604,-2545.584412,32400.,-5366.563145,-12423.76754},
			{354.964788,-1159.310137,1649.727251,-1073.312629,264.5751311,-2318.620276,3591.991091,-2494.153164,640.,5499.090835,-4156.921939,1138.419957,-5366.563145,1600.,1852.025918},
			{469.574280,-766.811580,727.4613394,-354.9647869,70.,-3834.057902,2969.848480,-1374.772709,264.5751311,10911.92009,-4124.318126,752.9940236,-12423.76754,1852.025918,4900.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{224.99996,-857.32140,1626.65301,-1428.70571,469.574280,-857.32140,1328.15664,-1106.67069,354.964788,1626.65301,-1106.67069,336.749168,-1428.70571,354.964788,469.574280},
			{-857.32140,4900.00002,-11156.51558,10887.64438,-3834.057902,2450.00002,-6072.83788,6325.14980,-2318.620276,-3718.83852,4216.76653,-1649.727254,2721.911098,-1159.310137,-766.811580},
			{1626.65301,-11156.51558,28224.,-29511.21822,10911.92009,-3718.83852,11522.39975,-13715.57480,5499.090835,4704.,-6857.787398,3130.137376,-2951.121822,1649.727251,727.4613394},
			{-1428.70571,10887.64438,-29511.21822,32400.,-12423.76754,2721.911098,-9638.323502,12548.46604,-5366.563145,-2951.121822,5019.386416,-2545.584412,1620.,-1073.312629,-354.9647869},
			{469.574280,-3834.057902,10911.92009,-12423.76754,4900.,-766.811580,2969.848480,-4124.318126,1852.025918,727.4613394,-1374.772709,752.9940236,-354.9647869,264.5751311,70.},
			{-857.32140,2450.00002,-3718.83852,2721.911098,-766.811580,4900.00002,-6072.83788,4216.76653,-1159.310137,-11156.51558,6325.14980,-1649.727254,10887.64438,-2318.620276,-3834.057902},
			{1328.15664,-6072.83788,11522.39975,-9638.323502,2969.848480,-6072.83788,12544.,-11198.71993,3591.991091,11522.39975,-11198.71993,3833.619699,-9638.323502,3591.991091,2969.848480},
			{-1106.67069,6325.14980,-13715.57480,12548.46604,-4124.318126,4216.76653,-11198.71993,11664.,-4156.921939,-6857.787398,8748.,-3549.242172,5019.386416,-2494.153164,-1374.772709},
			{354.964788,-2318.620276,5499.090835,-5366.563145,1852.025918,-1159.310137,3591.991091,-4156.921939,1600.,1649.727251,-2494.153164,1138.419957,-1073.312629,640.,264.5751311},
			{1626.65301,-3718.83852,4704.,-2951.121822,727.4613394,-11156.51558,11522.39975,-6857.787398,1649.727251,28224.,-13715.57480,3130.137376,-29511.21822,5499.090835,10911.92009},
			{-1106.67069,4216.76653,-6857.787398,5019.386416,-1374.772709,6325.14980,-11198.71993,8748.,-2494.153164,-13715.57480,11664.,-3549.242172,12548.46604,-4156.921939,-4124.318126},
			{336.749168,-1649.727254,3130.137376,-2545.584412,752.9940236,-1649.727254,3833.619699,-3549.242172,1138.419957,3130.137376,-3549.242172,1296.,-2545.584412,1138.419957,752.9940236},
			{-1428.70571,2721.911098,-2951.121822,1620.,-354.9647869,10887.64438,-9638.323502,5019.386416,-1073.312629,-29511.21822,12548.46604,-2545.584412,32400.,-5366.563145,-12423.76754},
			{354.964788,-1159.310137,1649.727251,-1073.312629,264.5751311,-2318.620276,3591.991091,-2494.153164,640.,5499.090835,-4156.921939,1138.419957,-5366.563145,1600.,1852.025918},
			{469.574280,-766.811580,727.4613394,-354.9647869,70.,-3834.057902,2969.848480,-1374.772709,264.5751311,10911.92009,-4124.318126,752.9940236,-12423.76754,1852.025918,4900.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,2.,1.936491673,1.673320053,1.449137675,1.581138830,1.673320053,1.527525232,1.296148140,1.449137675,1.095445115},
			{3.162277659,3.872983346,3.674234614,3.346640106,3.030457633,1.936491673,3.,3.240370349,3.207134903,1.224744872,2.160246899,2.535462764,.8366600264,1.603567452,.6060915266},
			{3.415650256,5.019960159,5.291502622,5.163977793,4.909902528,1.673320053,3.240370349,4.,4.330127020,.8819171036,2.,2.738612788,.5163977793,1.299038106,.3273268354},
			{3.549647870,5.796550700,6.546536711,6.708203931,6.614378278,1.449137675,3.207134903,4.330127020,5.,.6546536711,1.732050808,2.635231383,.3354101966,1.,.1889822365},
			{2.581988897,1.581138830,1.,.6831300514,.4948716594,3.162277660,2.449489743,1.763834207,1.309307342,3.,2.645751311,2.070196679,2.732520206,2.618614683,2.474358298},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,3.346640106,4.320493799,4.,3.464101616,2.645751311,4.,4.107919181,2.065591118,3.464101616,1.636634177},
			{3.741657386,4.582575696,4.140393357,3.535533905,2.988071523,3.055050464,5.070925527,5.477225575,5.270462768,2.070196679,4.107919181,5.,1.414213562,3.162277660,.9960238407},
			{2.645751311,1.296148140,.6831300514,.4000000000,.2535462764,3.888444420,2.509980079,1.549193338,1.006230590,4.098780308,3.098386677,2.121320343,4.,3.354101966,3.803194146},
			{3.549647870,2.898275350,1.963961013,1.341640786,.9449111824,4.347413025,4.810702356,3.897114318,3.,3.927922025,5.196152424,4.743416490,3.354101966,5.,2.834733546},
			{2.683281572,1.095445115,.4948716594,.2535462764,.1428571429,4.381780461,2.424366106,1.309307341,.7559289460,4.948716594,3.273268354,1.992047682,5.070925527,3.779644731,5.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,2.,1.936491673,1.673320053,1.449137675,1.581138830,1.673320053,1.527525232,1.296148140,1.449137675,1.095445115},
			{3.162277659,3.872983346,3.674234614,3.346640106,3.030457633,1.936491673,3.,3.240370349,3.207134903,1.224744872,2.160246899,2.535462764,.8366600264,1.603567452,.6060915266},
			{3.415650256,5.019960159,5.291502622,5.163977793,4.909902528,1.673320053,3.240370349,4.,4.330127020,.8819171036,2.,2.738612788,.5163977793,1.299038106,.3273268354},
			{3.549647870,5.796550700,6.546536711,6.708203931,6.614378278,1.449137675,3.207134903,4.330127020,5.,.6546536711,1.732050808,2.635231383,.3354101966,1.,.1889822365},
			{2.581988897,1.581138830,1.,.6831300514,.4948716594,3.162277660,2.449489743,1.763834207,1.309307342,3.,2.645751311,2.070196679,2.732520206,2.618614683,2.474358298},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,3.346640106,4.320493799,4.,3.464101616,2.645751311,4.,4.107919181,2.065591118,3.464101616,1.636634177},
			{3.741657386,4.582575696,4.140393357,3.535533905,2.988071523,3.055050464,5.070925527,5.477225575,5.270462768,2.070196679,4.107919181,5.,1.414213562,3.162277660,.9960238407},
			{2.645751311,1.296148140,.6831300514,.4000000000,.2535462764,3.888444420,2.509980079,1.549193338,1.006230590,4.098780308,3.098386677,2.121320343,4.,3.354101966,3.803194146},
			{3.549647870,2.898275350,1.963961013,1.341640786,.9449111824,4.347413025,4.810702356,3.897114318,3.,3.927922025,5.196152424,4.743416490,3.354101966,5.,2.834733546},
			{2.683281572,1.095445115,.4948716594,.2535462764,.1428571429,4.381780461,2.424366106,1.309307341,.7559289460,4.948716594,3.273268354,1.992047682,5.070925527,3.779644731,5.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,2.,1.936491673,1.673320053,1.449137675,1.581138830,1.673320053,1.527525232,1.296148140,1.449137675,1.095445115},
			{2.581988897,3.162277660,3.,2.732520206,2.474358298,1.581138830,2.449489743,2.645751311,2.618614683,1.,1.763834207,2.070196679,.6831300514,1.309307342,.4948716594},
			{2.645751311,3.888444420,4.098780308,4.,3.803194146,1.296148140,2.509980079,3.098386677,3.354101966,.6831300514,1.549193338,2.121320343,.4000000000,1.006230590,.2535462764},
			{2.683281572,4.381780461,4.948716594,5.070925527,5.,1.095445115,2.424366106,3.273268354,3.779644731,.4948716594,1.309307341,1.992047682,.2535462764,.7559289460,.1428571429},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,1.936491673,1.224744872,.8366600264,.6060915266,3.872983346,3.,2.160246899,1.603567452,3.674234614,3.240370349,2.535462764,3.346640106,3.207134903,3.030457633},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,3.346640106,4.320493799,4.,3.464101616,2.645751311,4.,4.107919181,2.065591118,3.464101616,1.636634177},
			{3.549647870,4.347413025,3.927922025,3.354101966,2.834733546,2.898275350,4.810702356,5.196152424,5.,1.963961013,3.897114318,4.743416490,1.341640786,3.,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,1.673320053,.8819171036,.5163977793,.3273268354,5.019960159,3.240370349,2.,1.299038106,5.291502622,4.,2.738612788,5.163977793,4.330127020,4.909902528},
			{3.741657386,3.055050464,2.070196679,1.414213562,.9960238407,4.582575696,5.070925527,4.107919181,3.162277660,4.140393357,5.477225575,5.,3.535533905,5.270462768,2.988071523},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,1.449137675,.6546536711,.3354101966,.1889822365,5.796550700,3.207134903,1.732050808,1.,6.546536711,4.330127020,2.635231383,6.708203931,5.,6.614378278},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,2.,1.936491673,1.673320053,1.449137675,1.581138830,1.673320053,1.527525232,1.296148140,1.449137675,1.095445115},
			{2.581988897,3.162277660,3.,2.732520206,2.474358298,1.581138830,2.449489743,2.645751311,2.618614683,1.,1.763834207,2.070196679,.6831300514,1.309307342,.4948716594},
			{2.645751311,3.888444420,4.098780308,4.,3.803194146,1.296148140,2.509980079,3.098386677,3.354101966,.6831300514,1.549193338,2.121320343,.4000000000,1.006230590,.2535462764},
			{2.683281572,4.381780461,4.948716594,5.070925527,5.,1.095445115,2.424366106,3.273268354,3.779644731,.4948716594,1.309307341,1.992047682,.2535462764,.7559289460,.1428571429},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,1.936491673,1.224744872,.8366600264,.6060915266,3.872983346,3.,2.160246899,1.603567452,3.674234614,3.240370349,2.535462764,3.346640106,3.207134903,3.030457633},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,3.346640106,4.320493799,4.,3.464101616,2.645751311,4.,4.107919181,2.065591118,3.464101616,1.636634177},
			{3.549647870,4.347413025,3.927922025,3.354101966,2.834733546,2.898275350,4.810702356,5.196152424,5.,1.963961013,3.897114318,4.743416490,1.341640786,3.,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,1.673320053,.8819171036,.5163977793,.3273268354,5.019960159,3.240370349,2.,1.299038106,5.291502622,4.,2.738612788,5.163977793,4.330127020,4.909902528},
			{3.741657386,3.055050464,2.070196679,1.414213562,.9960238407,4.582575696,5.070925527,4.107919181,3.162277660,4.140393357,5.477225575,5.,3.535533905,5.270462768,2.988071523},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,1.449137675,.6546536711,.3354101966,.1889822365,5.796550700,3.207134903,1.732050808,1.,6.546536711,4.330127020,2.635231383,6.708203931,5.,6.614378278},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,.0451952183653594,.000932318622844532,1.66187980881431e-005,2.74870870898337e-007,3.41890639763464,.172756623071377,.00486900211054309,.000110027806742895,5.33523806427656,.368327957423731,.0131603422009587,7.19422103343416,.629639404683915,9.00134910987265},
			{1.414213562,.233717104755142,.0249321870529784,.00229823078236534,.000196571591120254,3.23038451124486,.844111317574503,.123027723654541,.014376841680633,4.7630793830775,1.70046113382978,.314193514488106,6.0685481480974,2.74657165283107,7.17423567654394},
			{1.414213562,.555278916264744,.140734880491406,.0308216365902836,.00626330408481417,2.90882269973526,1.80585734803576,.625327304106952,.173615533771693,3.86201434387853,3.27577043476888,1.43801908896753,4.43071582310304,4.76431769725595,4.71658607960963},
			{1.414213562,.981387965414959,.439603191878621,.170154848292167,.0611113213546646,2.48271365058504,2.72409573525399,1.66715454931021,.818062705137864,2.81340752869054,4.21756481949244,3.27222005765247,2.75487590413568,5.23550287189747,2.50302559510192},
			{1.414213562,1.4741928888764,.991945941760336,.576746649127296,.31115444533374,1.9899087271236,3.27976309580122,3.01515438576158,2.22245957594899,1.80736427869175,4.06994367638345,4.74332322095564,1.41847476579742,4.04940821803081,1.03297875354965},
			{1.414213562,1.9899087271236,1.80736427869175,1.41847476579742,1.03297875354965,1.4741928888764,3.27976309580122,4.06994367638345,4.04940821803081,.991945941760336,3.01515438576158,4.74332322095564,.576746649127296,2.22245957594899,.31115444533374},
			{1.414213562,2.48271365058504,2.81340752869054,2.75487590413568,2.50302559510192,.981387965414959,2.72409573525399,4.21756481949244,5.23550287189747,.439603191878621,1.66715454931021,3.27222005765247,.170154848292167,.818062705137864,.0611113213546646},
			{1.414213562,2.90882269973526,3.86201434387853,4.43071582310304,4.71658607960963,.555278916264744,1.80585734803576,3.27577043476888,4.76431769725595,.140734880491406,.625327304106952,1.43801908896753,.0308216365902836,.173615533771693,.00626330408481417},
			{1.414213562,3.23038451124486,4.7630793830775,6.0685481480974,7.17423567654394,.233717104755142,.844111317574503,1.70046113382978,2.74657165283107,.0249321870529784,.123027723654541,.314193514488106,.00229823078236534,.014376841680633,.000196571591120254},
			{1.414213562,3.41890639763464,5.33523806427656,7.19422103343416,9.00134910987265,.0451952183653594,.172756623071377,.368327957423731,.629639404683915,.000932318622844532,.00486900211054309,.0131603422009587,1.66187980881431e-005,.000110027806742895,2.74870870898337e-007}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,.0451952183653594,.000932318622844532,1.66187980881431e-005,2.74870870898337e-007,3.41890639763464,.172756623071377,.00486900211054309,.000110027806742895,5.33523806427656,.368327957423731,.0131603422009587,7.19422103343416,.629639404683915,9.00134910987265},
			{1.414213562,.233717104755142,.0249321870529784,.00229823078236534,.000196571591120254,3.23038451124486,.844111317574503,.123027723654541,.014376841680633,4.7630793830775,1.70046113382978,.314193514488106,6.0685481480974,2.74657165283107,7.17423567654394},
			{1.414213562,.555278916264744,.140734880491406,.0308216365902836,.00626330408481417,2.90882269973526,1.80585734803576,.625327304106952,.173615533771693,3.86201434387853,3.27577043476888,1.43801908896753,4.43071582310304,4.76431769725595,4.71658607960963},
			{1.414213562,.981387965414959,.439603191878621,.170154848292167,.0611113213546646,2.48271365058504,2.72409573525399,1.66715454931021,.818062705137864,2.81340752869054,4.21756481949244,3.27222005765247,2.75487590413568,5.23550287189747,2.50302559510192},
			{1.414213562,1.4741928888764,.991945941760336,.576746649127296,.31115444533374,1.9899087271236,3.27976309580122,3.01515438576158,2.22245957594899,1.80736427869175,4.06994367638345,4.74332322095564,1.41847476579742,4.04940821803081,1.03297875354965},
			{1.414213562,1.9899087271236,1.80736427869175,1.41847476579742,1.03297875354965,1.4741928888764,3.27976309580122,4.06994367638345,4.04940821803081,.991945941760336,3.01515438576158,4.74332322095564,.576746649127296,2.22245957594899,.31115444533374},
			{1.414213562,2.48271365058504,2.81340752869054,2.75487590413568,2.50302559510192,.981387965414959,2.72409573525399,4.21756481949244,5.23550287189747,.439603191878621,1.66715454931021,3.27222005765247,.170154848292167,.818062705137864,.0611113213546646},
			{1.414213562,2.90882269973526,3.86201434387853,4.43071582310304,4.71658607960963,.555278916264744,1.80585734803576,3.27577043476888,4.76431769725595,.140734880491406,.625327304106952,1.43801908896753,.0308216365902836,.173615533771693,.00626330408481417},
			{1.414213562,3.23038451124486,4.7630793830775,6.0685481480974,7.17423567654394,.233717104755142,.844111317574503,1.70046113382978,2.74657165283107,.0249321870529784,.123027723654541,.314193514488106,.00229823078236534,.014376841680633,.000196571591120254},
			{1.414213562,3.41890639763464,5.33523806427656,7.19422103343416,9.00134910987265,.0451952183653594,.172756623071377,.368327957423731,.629639404683915,.000932318622844532,.00486900211054309,.0131603422009587,1.66187980881431e-005,.000110027806742895,2.74870870898337e-007}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,0.,0.,0.,0.,.0451952183653594,0.,0.,0.,.000932318622844532,0.,0.,1.66187980881431e-005,0.,2.74870870898337e-007},
			{1.414213562,0.,0.,0.,0.,.233717104755142,0.,0.,0.,.0249321870529784,0.,0.,.00229823078236534,0.,.000196571591120254},
			{1.414213562,0.,0.,0.,0.,.555278916264744,0.,0.,0.,.140734880491406,0.,0.,.0308216365902836,0.,.00626330408481417},
			{1.414213562,0.,0.,0.,0.,.981387965414959,0.,0.,0.,.439603191878621,0.,0.,.170154848292167,0.,.0611113213546646},
			{1.414213562,0.,0.,0.,0.,1.4741928888764,0.,0.,0.,.991945941760336,0.,0.,.576746649127296,0.,.31115444533374},
			{1.414213562,0.,0.,0.,0.,1.9899087271236,0.,0.,0.,1.80736427869175,0.,0.,1.41847476579742,0.,1.03297875354965},
			{1.414213562,0.,0.,0.,0.,2.48271365058504,0.,0.,0.,2.81340752869054,0.,0.,2.75487590413568,0.,2.50302559510192},
			{1.414213562,0.,0.,0.,0.,2.90882269973526,0.,0.,0.,3.86201434387853,0.,0.,4.43071582310304,0.,4.71658607960963},
			{1.414213562,0.,0.,0.,0.,3.23038451124486,0.,0.,0.,4.7630793830775,0.,0.,6.0685481480974,0.,7.17423567654394},
			{1.414213562,0.,0.,0.,0.,3.41890639763464,0.,0.,0.,5.33523806427656,0.,0.,7.19422103343416,0.,9.00134910987265}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,0.,0.,0.,0.,.0451952183653594,0.,0.,0.,.000932318622844532,0.,0.,1.66187980881431e-005,0.,2.74870870898337e-007},
			{1.414213562,0.,0.,0.,0.,.233717104755142,0.,0.,0.,.0249321870529784,0.,0.,.00229823078236534,0.,.000196571591120254},
			{1.414213562,0.,0.,0.,0.,.555278916264744,0.,0.,0.,.140734880491406,0.,0.,.0308216365902836,0.,.00626330408481417},
			{1.414213562,0.,0.,0.,0.,.981387965414959,0.,0.,0.,.439603191878621,0.,0.,.170154848292167,0.,.0611113213546646},
			{1.414213562,0.,0.,0.,0.,1.4741928888764,0.,0.,0.,.991945941760336,0.,0.,.576746649127296,0.,.31115444533374},
			{1.414213562,0.,0.,0.,0.,1.9899087271236,0.,0.,0.,1.80736427869175,0.,0.,1.41847476579742,0.,1.03297875354965},
			{1.414213562,0.,0.,0.,0.,2.48271365058504,0.,0.,0.,2.81340752869054,0.,0.,2.75487590413568,0.,2.50302559510192},
			{1.414213562,0.,0.,0.,0.,2.90882269973526,0.,0.,0.,3.86201434387853,0.,0.,4.43071582310304,0.,4.71658607960963},
			{1.414213562,0.,0.,0.,0.,3.23038451124486,0.,0.,0.,4.7630793830775,0.,0.,6.0685481480974,0.,7.17423567654394},
			{1.414213562,0.,0.,0.,0.,3.41890639763464,0.,0.,0.,5.33523806427656,0.,0.,7.19422103343416,0.,9.00134910987265}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,3.41890639763464,5.33523806427656,7.19422103343416,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.23038451124486,4.7630793830775,6.0685481480974,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.90882269973526,3.86201434387853,4.43071582310304,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.48271365058504,2.81340752869054,2.75487590413568,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.9899087271236,1.80736427869175,1.41847476579742,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.4741928888764,.991945941760336,.576746649127296,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.981387965414959,.439603191878621,.170154848292167,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.555278916264744,.140734880491406,.0308216365902836,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.233717104755142,.0249321870529784,.00229823078236534,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0451952183653594,.000932318622844532,1.66187980881431e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[15],
		T o_output[10]
	) {
		static const T matrix[10][15] = 
		{
			{1.414213562,3.41890639763464,5.33523806427656,7.19422103343416,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.23038451124486,4.7630793830775,6.0685481480974,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.90882269973526,3.86201434387853,4.43071582310304,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.48271365058504,2.81340752869054,2.75487590413568,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.9899087271236,1.80736427869175,1.41847476579742,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.4741928888764,.991945941760336,.576746649127296,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.981387965414959,.439603191878621,.170154848292167,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.555278916264744,.140734880491406,.0308216365902836,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.233717104755142,.0249321870529784,.00229823078236534,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0451952183653594,.000932318622844532,1.66187980881431e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 10; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0666713464862914,-.149451343590541,-.219086364157122,-.269266719261963,-.295524224714757,-.295524224714757,-.269266719261963,-.219086364157122,-.149451343590541,-.0666713464862914},
			{-.00213067258194671,-.0246987698646587,-.0860223958122111,-.186856585689445,-.308057935549154,-.415825621622801,-.472709481205777,-.450627405966319,-.341380763300079,-.161180106771333},
			{-4.395300651663e-005,-.0026347851214796,-.0218022893353363,-.0837005897640987,-.207284149381534,-.37767982241688,-.535673702440208,-.598293428384306,-.503352981601333,-.251523189241307},
			{-7.83472648479441e-007,-.000242872567024931,-.00477480946005413,-.0323975380897769,-.120521122714454,-.296414676313609,-.524529262214111,-.686395213472642,-.641312386098096,-.339162638412016},
			{-1.29584466981531e-008,-2.0773304102391e-005,-.000970295120694341,-.0116356153302479,-.0650210679957301,-.215858660481792,-.476576882004331,-.730681505704966,-.758159297392565,-.424357452903382},
			{-.161180106771333,-.341380763300079,-.450627405966319,-.472709481205777,-.415825621622801,-.308057935549154,-.186856585689445,-.0860223958122111,-.0246987698646587,-.00213067258194671},
			{-.00814439698359798,-.0892040451843217,-.279758822139498,-.518668708114341,-.685362855842359,-.685362855842359,-.518668708114341,-.279758822139498,-.0892040451843217,-.00814439698359798},
			{-.000229543072776175,-.0130013309753251,-.0968741136939265,-.317426764825887,-.630068319043131,-.85048466602198,-.80302570427492,-.5074740147261,-.179701431147967,-.0173643652676472},
			{-5.18712464632104e-006,-.00151931671591861,-.026896076418173,-.155759403364294,-.46442111746321,-.846193429133371,-.99684143833616,-.738075966372677,-.290252359754073,-.0296835697358091},
			{-.251523189241307,-.503352981601333,-.598293428384306,-.535673702440208,-.37767982241688,-.207284149381534,-.0837005897640987,-.0218022893353363,-.0026347851214796,-4.395300651663e-005},
			{-.0173643652676472,-.179701431147967,-.5074740147261,-.80302570427492,-.85048466602198,-.630068319043131,-.317426764825887,-.0968741136939265,-.0130013309753251,-.000229543072776175},
			{-.000620428030017239,-.0332033605988752,-.222774255663801,-.62303175619127,-.991198892706415,-.991198892706415,-.62303175619127,-.222774255663801,-.0332033605988752,-.000620428030017239},
			{-.339162638412016,-.641312386098096,-.686395213472642,-.524529262214111,-.296414676313609,-.120521122714454,-.0323975380897769,-.00477480946005413,-.000242872567024931,-7.83472648479441e-007},
			{-.0296835697358091,-.290252359754073,-.738075966372677,-.99684143833616,-.846193429133371,-.46442111746321,-.155759403364294,-.026896076418173,-.00151931671591861,-5.18712464632104e-006},
			{-.424357452903382,-.758159297392565,-.730681505704966,-.476576882004331,-.215858660481792,-.0650210679957301,-.0116356153302479,-.000970295120694341,-2.0773304102391e-005,-1.29584466981531e-008}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0666713464862914,-.149451343590541,-.219086364157122,-.269266719261963,-.295524224714757,-.295524224714757,-.269266719261963,-.219086364157122,-.149451343590541,-.0666713464862914},
			{-.00213067258194671,-.0246987698646587,-.0860223958122111,-.186856585689445,-.308057935549154,-.415825621622801,-.472709481205777,-.450627405966319,-.341380763300079,-.161180106771333},
			{-4.395300651663e-005,-.0026347851214796,-.0218022893353363,-.0837005897640987,-.207284149381534,-.37767982241688,-.535673702440208,-.598293428384306,-.503352981601333,-.251523189241307},
			{-7.83472648479441e-007,-.000242872567024931,-.00477480946005413,-.0323975380897769,-.120521122714454,-.296414676313609,-.524529262214111,-.686395213472642,-.641312386098096,-.339162638412016},
			{-1.29584466981531e-008,-2.0773304102391e-005,-.000970295120694341,-.0116356153302479,-.0650210679957301,-.215858660481792,-.476576882004331,-.730681505704966,-.758159297392565,-.424357452903382},
			{-.161180106771333,-.341380763300079,-.450627405966319,-.472709481205777,-.415825621622801,-.308057935549154,-.186856585689445,-.0860223958122111,-.0246987698646587,-.00213067258194671},
			{-.00814439698359798,-.0892040451843217,-.279758822139498,-.518668708114341,-.685362855842359,-.685362855842359,-.518668708114341,-.279758822139498,-.0892040451843217,-.00814439698359798},
			{-.000229543072776175,-.0130013309753251,-.0968741136939265,-.317426764825887,-.630068319043131,-.85048466602198,-.80302570427492,-.5074740147261,-.179701431147967,-.0173643652676472},
			{-5.18712464632104e-006,-.00151931671591861,-.026896076418173,-.155759403364294,-.46442111746321,-.846193429133371,-.99684143833616,-.738075966372677,-.290252359754073,-.0296835697358091},
			{-.251523189241307,-.503352981601333,-.598293428384306,-.535673702440208,-.37767982241688,-.207284149381534,-.0837005897640987,-.0218022893353363,-.0026347851214796,-4.395300651663e-005},
			{-.0173643652676472,-.179701431147967,-.5074740147261,-.80302570427492,-.85048466602198,-.630068319043131,-.317426764825887,-.0968741136939265,-.0130013309753251,-.000229543072776175},
			{-.000620428030017239,-.0332033605988752,-.222774255663801,-.62303175619127,-.991198892706415,-.991198892706415,-.62303175619127,-.222774255663801,-.0332033605988752,-.000620428030017239},
			{-.339162638412016,-.641312386098096,-.686395213472642,-.524529262214111,-.296414676313609,-.120521122714454,-.0323975380897769,-.00477480946005413,-.000242872567024931,-7.83472648479441e-007},
			{-.0296835697358091,-.290252359754073,-.738075966372677,-.99684143833616,-.846193429133371,-.46442111746321,-.155759403364294,-.026896076418173,-.00151931671591861,-5.18712464632104e-006},
			{-.424357452903382,-.758159297392565,-.730681505704966,-.476576882004331,-.215858660481792,-.0650210679957301,-.0116356153302479,-.000970295120694341,-2.0773304102391e-005,-1.29584466981531e-008}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.00150661303158024,-.0174646676628736,-.060827019428775,-.132127558885229,-.217829855282603,-.294033116918165,-.334256079779963,-.318641694631351,-.241392652759809,-.113971546520449},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.10794689696449e-005,-.00186307442685916,-.0154165466384746,-.0591852546271217,-.146572027698836,-.267059963618762,-.378778507598704,-.423057340461501,-.355924306714669,-.177853752785117},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-5.53998822760152e-007,-.000171736839152821,-.00337630014896868,-.0229085188830744,-.0852213031700891,-.209596827719864,-.370898198340224,-.485354710148538,-.453476337188524,-.239824201609527},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-9.16300553632585e-009,-1.46889442023262e-005,-.000686102259776195,-.00822762250546704,-.0459768381119019,-.152635122644788,-.336990745110908,-.516669847707885,-.536099580547344,-.300066032674195}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.00150661303158024,-.0174646676628736,-.060827019428775,-.132127558885229,-.217829855282603,-.294033116918165,-.334256079779963,-.318641694631351,-.241392652759809,-.113971546520449},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.10794689696449e-005,-.00186307442685916,-.0154165466384746,-.0591852546271217,-.146572027698836,-.267059963618762,-.378778507598704,-.423057340461501,-.355924306714669,-.177853752785117},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-5.53998822760152e-007,-.000171736839152821,-.00337630014896868,-.0229085188830744,-.0852213031700891,-.209596827719864,-.370898198340224,-.485354710148538,-.453476337188524,-.239824201609527},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-9.16300553632585e-009,-1.46889442023262e-005,-.000686102259776195,-.00822762250546704,-.0459768381119019,-.152635122644788,-.336990745110908,-.516669847707885,-.536099580547344,-.300066032674195}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{-.113971546520449,-.241392652759809,-.318641694631351,-.334256079779963,-.294033116918165,-.217829855282603,-.132127558885229,-.060827019428775,-.0174646676628736,-.00150661303158024},
			{-.177853752785117,-.355924306714669,-.423057340461501,-.378778507598704,-.267059963618762,-.146572027698836,-.0591852546271217,-.0154165466384746,-.00186307442685916,-3.10794689696449e-005},
			{-.239824201609527,-.453476337188524,-.485354710148538,-.370898198340224,-.209596827719864,-.0852213031700891,-.0229085188830744,-.00337630014896868,-.000171736839152821,-5.53998822760152e-007},
			{-.300066032674195,-.536099580547344,-.516669847707885,-.336990745110908,-.152635122644788,-.0459768381119019,-.00822762250546704,-.000686102259776195,-1.46889442023262e-005,-9.16300553632585e-009},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[10],
		T o_output[15]
	) {
		static const T matrix[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{-.113971546520449,-.241392652759809,-.318641694631351,-.334256079779963,-.294033116918165,-.217829855282603,-.132127558885229,-.060827019428775,-.0174646676628736,-.00150661303158024},
			{-.177853752785117,-.355924306714669,-.423057340461501,-.378778507598704,-.267059963618762,-.146572027698836,-.0591852546271217,-.0154165466384746,-.00186307442685916,-3.10794689696449e-005},
			{-.239824201609527,-.453476337188524,-.485354710148538,-.370898198340224,-.209596827719864,-.0852213031700891,-.0229085188830744,-.00337630014896868,-.000171736839152821,-5.53998822760152e-007},
			{-.300066032674195,-.536099580547344,-.516669847707885,-.336990745110908,-.152635122644788,-.0459768381119019,-.00822762250546704,-.000686102259776195,-1.46889442023262e-005,-9.16300553632585e-009},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 10; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,3.354101966},
			{0.,-.5000000000,0.,0.,0.,-.5000000000,-1.936491673,0.,0.,-1.581138830,-4.183300132,0.,-3.240370350,-7.245688375,-5.477225576},
			{0.,0.,.5000000000,0.,0.,0.,1.224744872,2.645751311,0.,.5000000000,5.291502622,7.245688375,2.049390154,13.74772709,5.196152424},
			{0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,-3.354101966,0.,-1.936491673,-10.60660172,-.5000000000,-10.06230590,-2.535462764},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,4.183300132,0.,2.645751311,.5000000000},
			{0.,.5000000000,0.,0.,0.,-.5000000000,1.936491673,0.,0.,-1.581138830,4.183300132,0.,-3.240370350,7.245688375,-5.477225576},
			{0.,0.,-.4082482906,0.,0.,0.,0.,-2.160246899,0.,.4082482906,0.,-5.916079782,1.673320053,0.,4.242640686},
			{0.,0.,0.,.3872983346,0.,0.,0.,.5000000000,2.598076212,0.,-.5000000000,2.738612788,-.3872983346,-2.598076212,-1.963961012},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,-1.,0.,0.,0.,0.,1.,.3779644731},
			{0.,0.,.5000000000,0.,0.,0.,-1.224744872,2.645751311,0.,.5000000000,-5.291502622,7.245688375,2.049390154,-13.74772709,5.196152424},
			{0.,0.,0.,-.3872983346,0.,0.,0.,.5000000000,-2.598076212,0.,.5000000000,2.738612788,-.3872983346,2.598076212,-1.963961012},
			{0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3585685828},
			{0.,0.,0.,.5000000000,0.,0.,0.,-1.936491673,3.354101966,0.,1.936491673,-10.60660172,-.5000000000,10.06230590,-2.535462764},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,1.,0.,0.,0.,0.,-1.,.3779644731},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,-2.645751311,0.,0.,4.183300132,0.,-2.645751311,.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,3.354101966},
			{0.,-.5000000000,0.,0.,0.,-.5000000000,-1.936491673,0.,0.,-1.581138830,-4.183300132,0.,-3.240370350,-7.245688375,-5.477225576},
			{0.,0.,.5000000000,0.,0.,0.,1.224744872,2.645751311,0.,.5000000000,5.291502622,7.245688375,2.049390154,13.74772709,5.196152424},
			{0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,-3.354101966,0.,-1.936491673,-10.60660172,-.5000000000,-10.06230590,-2.535462764},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,4.183300132,0.,2.645751311,.5000000000},
			{0.,.5000000000,0.,0.,0.,-.5000000000,1.936491673,0.,0.,-1.581138830,4.183300132,0.,-3.240370350,7.245688375,-5.477225576},
			{0.,0.,-.4082482906,0.,0.,0.,0.,-2.160246899,0.,.4082482906,0.,-5.916079782,1.673320053,0.,4.242640686},
			{0.,0.,0.,.3872983346,0.,0.,0.,.5000000000,2.598076212,0.,-.5000000000,2.738612788,-.3872983346,-2.598076212,-1.963961012},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,-1.,0.,0.,0.,0.,1.,.3779644731},
			{0.,0.,.5000000000,0.,0.,0.,-1.224744872,2.645751311,0.,.5000000000,-5.291502622,7.245688375,2.049390154,-13.74772709,5.196152424},
			{0.,0.,0.,-.3872983346,0.,0.,0.,.5000000000,-2.598076212,0.,.5000000000,2.738612788,-.3872983346,2.598076212,-1.963961012},
			{0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3585685828},
			{0.,0.,0.,.5000000000,0.,0.,0.,-1.936491673,3.354101966,0.,1.936491673,-10.60660172,-.5000000000,10.06230590,-2.535462764},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,1.,0.,0.,0.,0.,-1.,.3779644731},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,-2.645751311,0.,0.,4.183300132,0.,-2.645751311,.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.5000000000,1.224744872,1.936491673,2.645751311,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,.5000000000,1.936491673,4.183300132,7.245688375,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,2.049390154,5.196152424,0.,-1.224744872,-5.291502622,-13.74772709,.5000000000,2.645751311,7.245688375,0.,0.,0.},
			{0.,0.,0.,-.5000000000,-2.535462764,0.,0.,1.936491673,10.06230590,0.,-1.936491673,-10.60660172,.5000000000,3.354101966,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,-2.645751311,0.,0.,4.183300132,0.,-2.645751311,.5000000000},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,-.5000000000,-1.936491673,-4.183300132,-7.245688375,0.,0.,0.,0.,0.,0.},
			{0.,0.,.4082482906,1.673320053,4.242640686,0.,0.,0.,0.,-.4082482906,-2.160246899,-5.916079782,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,0.,0.,.5000000000,2.598076212,0.,.5000000000,2.738612788,-.3872983346,-2.598076212,0.},
			{0.,0.,0.,0.,.3779644731,0.,0.,0.,-1.,0.,0.,0.,0.,1.,-.3779644731},
			{0.,0.,.5000000000,2.049390154,5.196152424,0.,1.224744872,5.291502622,13.74772709,.5000000000,2.645751311,7.245688375,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,0.,0.,-.5000000000,-2.598076212,0.,.5000000000,2.738612788,.3872983346,2.598076212,0.},
			{0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3585685828},
			{0.,0.,0.,-.5000000000,-2.535462764,0.,0.,-1.936491673,-10.06230590,0.,-1.936491673,-10.60660172,-.5000000000,-3.354101966,0.},
			{0.,0.,0.,0.,.3779644731,0.,0.,0.,1.,0.,0.,0.,0.,-1.,-.3779644731},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,4.183300132,0.,2.645751311,.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.5000000000,1.224744872,1.936491673,2.645751311,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,.5000000000,1.936491673,4.183300132,7.245688375,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,2.049390154,5.196152424,0.,-1.224744872,-5.291502622,-13.74772709,.5000000000,2.645751311,7.245688375,0.,0.,0.},
			{0.,0.,0.,-.5000000000,-2.535462764,0.,0.,1.936491673,10.06230590,0.,-1.936491673,-10.60660172,.5000000000,3.354101966,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,-2.645751311,0.,0.,4.183300132,0.,-2.645751311,.5000000000},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,-.5000000000,-1.936491673,-4.183300132,-7.245688375,0.,0.,0.,0.,0.,0.},
			{0.,0.,.4082482906,1.673320053,4.242640686,0.,0.,0.,0.,-.4082482906,-2.160246899,-5.916079782,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,0.,0.,.5000000000,2.598076212,0.,.5000000000,2.738612788,-.3872983346,-2.598076212,0.},
			{0.,0.,0.,0.,.3779644731,0.,0.,0.,-1.,0.,0.,0.,0.,1.,-.3779644731},
			{0.,0.,.5000000000,2.049390154,5.196152424,0.,1.224744872,5.291502622,13.74772709,.5000000000,2.645751311,7.245688375,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,0.,0.,-.5000000000,-2.598076212,0.,.5000000000,2.738612788,.3872983346,2.598076212,0.},
			{0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3585685828},
			{0.,0.,0.,-.5000000000,-2.535462764,0.,0.,-1.936491673,-10.06230590,0.,-1.936491673,-10.60660172,-.5000000000,-3.354101966,0.},
			{0.,0.,0.,0.,.3779644731,0.,0.,0.,1.,0.,0.,0.,0.,-1.,-.3779644731},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,4.183300132,0.,2.645751311,.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,1.224744872,2.371708244,2.561737692,2.218529919,.9682458365,2.561737692,3.507803799,.6614378278,2.218529919,.4192627457},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,.5000000000,0.,-1.045825033,-1.811422094,.7905694150,1.045825033,0.,.8100925875,1.811422094,.6846531970},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,-.6123724358,-.6614378278,0.,.2500000000,-.6614378278,-1.811422094,.5123475385,0.,.6495190530},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,.4841229182,.8385254914,0.,-.4841229182,0.,.1250000000,-.8385254914,.3169328455},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,-.3307189139,0.,0.,.5229125166,0.,-.3307189139,.6250000000e-1},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-.7905694150,-3.137475099,-5.728219620,-.8100925875,-3.622844188,-.6846531970},
			{0.,0.,.2041241452,.4183300132,.5303300858,0.,0.,.5400617248,1.403121520,-.2041241452,-.5400617248,0.,-.4183300132,-1.403121520,-.5303300858},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,.1250000000,0.,0.,.1250000000,.6846531969,-.9682458365e-1,0.,-.2454951265},
			{0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,-.1250000000,0.,0.,0.,0.,.1250000000,-.4724555913e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,.6123724358,1.984313483,3.436931772,.2500000000,1.984313483,5.434266281,.5123475385,3.436931772,.6495190530},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,-.1250000000,-.6495190530,0.,.1250000000,0.,.9682458365e-1,.6495190530,.2454951265},
			{0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4482107285e-1},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,-.4841229182,-1.677050983,0.,-.4841229182,-2.651650429,-.1250000000,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,.3307189139,0.,0.,.5229125166,0.,.3307189139,.6250000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,1.224744872,2.371708244,2.561737692,2.218529919,.9682458365,2.561737692,3.507803799,.6614378278,2.218529919,.4192627457},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,.5000000000,0.,-1.045825033,-1.811422094,.7905694150,1.045825033,0.,.8100925875,1.811422094,.6846531970},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,-.6123724358,-.6614378278,0.,.2500000000,-.6614378278,-1.811422094,.5123475385,0.,.6495190530},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,.4841229182,.8385254914,0.,-.4841229182,0.,.1250000000,-.8385254914,.3169328455},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,-.3307189139,0.,0.,.5229125166,0.,-.3307189139,.6250000000e-1},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-.7905694150,-3.137475099,-5.728219620,-.8100925875,-3.622844188,-.6846531970},
			{0.,0.,.2041241452,.4183300132,.5303300858,0.,0.,.5400617248,1.403121520,-.2041241452,-.5400617248,0.,-.4183300132,-1.403121520,-.5303300858},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,.1250000000,0.,0.,.1250000000,.6846531969,-.9682458365e-1,0.,-.2454951265},
			{0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,-.1250000000,0.,0.,0.,0.,.1250000000,-.4724555913e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,.6123724358,1.984313483,3.436931772,.2500000000,1.984313483,5.434266281,.5123475385,3.436931772,.6495190530},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,-.1250000000,-.6495190530,0.,.1250000000,0.,.9682458365e-1,.6495190530,.2454951265},
			{0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4482107285e-1},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,-.4841229182,-1.677050983,0.,-.4841229182,-2.651650429,-.1250000000,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,.3307189139,0.,0.,.5229125166,0.,.3307189139,.6250000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,1.224744872,2.371708244,2.561737692,2.218529919,.9682458365,2.561737692,3.507803799,.6614378278,2.218529919,.4192627457},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-.7905694150,-3.137475099,-5.728219620,-.8100925875,-3.622844188,-.6846531970},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,.6123724358,1.984313483,3.436931772,.2500000000,1.984313483,5.434266281,.5123475385,3.436931772,.6495190530},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,-.4841229182,-1.677050983,0.,-.4841229182,-2.651650429,-.1250000000,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,.3307189139,0.,0.,.5229125166,0.,.3307189139,.6250000000e-1},
			{0.,.5000000000,.7905694150,.8100925875,.6846531970,-.5000000000,0.,1.045825033,1.811422094,-.7905694150,-1.045825033,0.,-.8100925875,-1.811422094,-.6846531970},
			{0.,0.,-.2041241452,-.4183300132,-.5303300858,0.,0.,-.5400617248,-1.403121520,.2041241452,.5400617248,0.,.4183300132,1.403121520,.5303300858},
			{0.,0.,0.,.9682458365e-1,.2454951265,0.,0.,.1250000000,.6495190530,0.,-.1250000000,0.,-.9682458365e-1,-.6495190530,-.2454951265},
			{0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,-.1250000000,0.,0.,0.,0.,.1250000000,.4724555913e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,-.6123724358,-.6614378278,0.,.2500000000,-.6614378278,-1.811422094,.5123475385,0.,.6495190530},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,.1250000000,0.,0.,.1250000000,.6846531969,-.9682458365e-1,0.,-.2454951265},
			{0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4482107285e-1},
			{0.,0.,0.,.1250000000,.3169328455,0.,0.,-.4841229182,-.8385254914,0.,.4841229182,0.,-.1250000000,.8385254914,-.3169328455},
			{0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,-.3307189139,0.,0.,.5229125166,0.,-.3307189139,.6250000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,1.224744872,2.371708244,2.561737692,2.218529919,.9682458365,2.561737692,3.507803799,.6614378278,2.218529919,.4192627457},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-.7905694150,-3.137475099,-5.728219620,-.8100925875,-3.622844188,-.6846531970},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,.6123724358,1.984313483,3.436931772,.2500000000,1.984313483,5.434266281,.5123475385,3.436931772,.6495190530},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,-.4841229182,-1.677050983,0.,-.4841229182,-2.651650429,-.1250000000,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,.3307189139,0.,0.,.5229125166,0.,.3307189139,.6250000000e-1},
			{0.,.5000000000,.7905694150,.8100925875,.6846531970,-.5000000000,0.,1.045825033,1.811422094,-.7905694150,-1.045825033,0.,-.8100925875,-1.811422094,-.6846531970},
			{0.,0.,-.2041241452,-.4183300132,-.5303300858,0.,0.,-.5400617248,-1.403121520,.2041241452,.5400617248,0.,.4183300132,1.403121520,.5303300858},
			{0.,0.,0.,.9682458365e-1,.2454951265,0.,0.,.1250000000,.6495190530,0.,-.1250000000,0.,-.9682458365e-1,-.6495190530,-.2454951265},
			{0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,-.1250000000,0.,0.,0.,0.,.1250000000,.4724555913e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,-.6123724358,-.6614378278,0.,.2500000000,-.6614378278,-1.811422094,.5123475385,0.,.6495190530},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,.1250000000,0.,0.,.1250000000,.6846531969,-.9682458365e-1,0.,-.2454951265},
			{0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4482107285e-1},
			{0.,0.,0.,.1250000000,.3169328455,0.,0.,-.4841229182,-.8385254914,0.,.4841229182,0.,-.1250000000,.8385254914,-.3169328455},
			{0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,-.3307189139,0.,0.,.5229125166,0.,-.3307189139,.6250000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[15],
		T o_output[15]
	) {
		static const T matrix[15][15] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 15; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[15],
		T o_output[1]
	) {
		static const T matrix[1][15] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[15],
		T o_output[1]
	) {
		static const T matrix[1][15] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 15; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 7875;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[15][15] = 
		{
			{224.99996,-857.32140,1626.65301,-1428.70571,469.574280,-857.32140,1328.15664,-1106.67069,354.964788,1626.65301,-1106.67069,336.749168,-1428.70571,354.964788,469.574280},
			{-857.32140,4900.00002,-11156.51558,10887.64438,-3834.057902,2450.00002,-6072.83788,6325.14980,-2318.620276,-3718.83852,4216.76653,-1649.727254,2721.911098,-1159.310137,-766.811580},
			{1626.65301,-11156.51558,28224.,-29511.21822,10911.92009,-3718.83852,11522.39975,-13715.57480,5499.090835,4704.,-6857.787398,3130.137376,-2951.121822,1649.727251,727.4613394},
			{-1428.70571,10887.64438,-29511.21822,32400.,-12423.76754,2721.911098,-9638.323502,12548.46604,-5366.563145,-2951.121822,5019.386416,-2545.584412,1620.,-1073.312629,-354.9647869},
			{469.574280,-3834.057902,10911.92009,-12423.76754,4900.,-766.811580,2969.848480,-4124.318126,1852.025918,727.4613394,-1374.772709,752.9940236,-354.9647869,264.5751311,70.},
			{-857.32140,2450.00002,-3718.83852,2721.911098,-766.811580,4900.00002,-6072.83788,4216.76653,-1159.310137,-11156.51558,6325.14980,-1649.727254,10887.64438,-2318.620276,-3834.057902},
			{1328.15664,-6072.83788,11522.39975,-9638.323502,2969.848480,-6072.83788,12544.,-11198.71993,3591.991091,11522.39975,-11198.71993,3833.619699,-9638.323502,3591.991091,2969.848480},
			{-1106.67069,6325.14980,-13715.57480,12548.46604,-4124.318126,4216.76653,-11198.71993,11664.,-4156.921939,-6857.787398,8748.,-3549.242172,5019.386416,-2494.153164,-1374.772709},
			{354.964788,-2318.620276,5499.090835,-5366.563145,1852.025918,-1159.310137,3591.991091,-4156.921939,1600.,1649.727251,-2494.153164,1138.419957,-1073.312629,640.,264.5751311},
			{1626.65301,-3718.83852,4704.,-2951.121822,727.4613394,-11156.51558,11522.39975,-6857.787398,1649.727251,28224.,-13715.57480,3130.137376,-29511.21822,5499.090835,10911.92009},
			{-1106.67069,4216.76653,-6857.787398,5019.386416,-1374.772709,6325.14980,-11198.71993,8748.,-2494.153164,-13715.57480,11664.,-3549.242172,12548.46604,-4156.921939,-4124.318126},
			{336.749168,-1649.727254,3130.137376,-2545.584412,752.9940236,-1649.727254,3833.619699,-3549.242172,1138.419957,3130.137376,-3549.242172,1296.,-2545.584412,1138.419957,752.9940236},
			{-1428.70571,2721.911098,-2951.121822,1620.,-354.9647869,10887.64438,-9638.323502,5019.386416,-1073.312629,-29511.21822,12548.46604,-2545.584412,32400.,-5366.563145,-12423.76754},
			{354.964788,-1159.310137,1649.727251,-1073.312629,264.5751311,-2318.620276,3591.991091,-2494.153164,640.,5499.090835,-4156.921939,1138.419957,-5366.563145,1600.,1852.025918},
			{469.574280,-766.811580,727.4613394,-354.9647869,70.,-3834.057902,2969.848480,-1374.772709,264.5751311,10911.92009,-4124.318126,752.9940236,-12423.76754,1852.025918,4900.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,2.,1.936491673,1.673320053,1.449137675,1.581138830,1.673320053,1.527525232,1.296148140,1.449137675,1.095445115},
			{3.162277659,3.872983346,3.674234614,3.346640106,3.030457633,1.936491673,3.,3.240370349,3.207134903,1.224744872,2.160246899,2.535462764,.8366600264,1.603567452,.6060915266},
			{3.415650256,5.019960159,5.291502622,5.163977793,4.909902528,1.673320053,3.240370349,4.,4.330127020,.8819171036,2.,2.738612788,.5163977793,1.299038106,.3273268354},
			{3.549647870,5.796550700,6.546536711,6.708203931,6.614378278,1.449137675,3.207134903,4.330127020,5.,.6546536711,1.732050808,2.635231383,.3354101966,1.,.1889822365},
			{2.581988897,1.581138830,1.,.6831300514,.4948716594,3.162277660,2.449489743,1.763834207,1.309307342,3.,2.645751311,2.070196679,2.732520206,2.618614683,2.474358298},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,3.346640106,4.320493799,4.,3.464101616,2.645751311,4.,4.107919181,2.065591118,3.464101616,1.636634177},
			{3.741657386,4.582575696,4.140393357,3.535533905,2.988071523,3.055050464,5.070925527,5.477225575,5.270462768,2.070196679,4.107919181,5.,1.414213562,3.162277660,.9960238407},
			{2.645751311,1.296148140,.6831300514,.4000000000,.2535462764,3.888444420,2.509980079,1.549193338,1.006230590,4.098780308,3.098386677,2.121320343,4.,3.354101966,3.803194146},
			{3.549647870,2.898275350,1.963961013,1.341640786,.9449111824,4.347413025,4.810702356,3.897114318,3.,3.927922025,5.196152424,4.743416490,3.354101966,5.,2.834733546},
			{2.683281572,1.095445115,.4948716594,.2535462764,.1428571429,4.381780461,2.424366106,1.309307341,.7559289460,4.948716594,3.273268354,1.992047682,5.070925527,3.779644731,5.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[15][15] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,2.,1.936491673,1.673320053,1.449137675,1.581138830,1.673320053,1.527525232,1.296148140,1.449137675,1.095445115},
			{2.581988897,3.162277660,3.,2.732520206,2.474358298,1.581138830,2.449489743,2.645751311,2.618614683,1.,1.763834207,2.070196679,.6831300514,1.309307342,.4948716594},
			{2.645751311,3.888444420,4.098780308,4.,3.803194146,1.296148140,2.509980079,3.098386677,3.354101966,.6831300514,1.549193338,2.121320343,.4000000000,1.006230590,.2535462764},
			{2.683281572,4.381780461,4.948716594,5.070925527,5.,1.095445115,2.424366106,3.273268354,3.779644731,.4948716594,1.309307341,1.992047682,.2535462764,.7559289460,.1428571429},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,1.936491673,1.224744872,.8366600264,.6060915266,3.872983346,3.,2.160246899,1.603567452,3.674234614,3.240370349,2.535462764,3.346640106,3.207134903,3.030457633},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,3.346640106,4.320493799,4.,3.464101616,2.645751311,4.,4.107919181,2.065591118,3.464101616,1.636634177},
			{3.549647870,4.347413025,3.927922025,3.354101966,2.834733546,2.898275350,4.810702356,5.196152424,5.,1.963961013,3.897114318,4.743416490,1.341640786,3.,.9449111824},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,1.673320053,.8819171036,.5163977793,.3273268354,5.019960159,3.240370349,2.,1.299038106,5.291502622,4.,2.738612788,5.163977793,4.330127020,4.909902528},
			{3.741657386,3.055050464,2.070196679,1.414213562,.9960238407,4.582575696,5.070925527,4.107919181,3.162277660,4.140393357,5.477225575,5.,3.535533905,5.270462768,2.988071523},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,1.449137675,.6546536711,.3354101966,.1889822365,5.796550700,3.207134903,1.732050808,1.,6.546536711,4.330127020,2.635231383,6.708203931,5.,6.614378278},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[10][15] = 
		{
			{1.414213562,.0451952183653594,.000932318622844532,1.66187980881431e-005,2.74870870898337e-007,3.41890639763464,.172756623071377,.00486900211054309,.000110027806742895,5.33523806427656,.368327957423731,.0131603422009587,7.19422103343416,.629639404683915,9.00134910987265},
			{1.414213562,.233717104755142,.0249321870529784,.00229823078236534,.000196571591120254,3.23038451124486,.844111317574503,.123027723654541,.014376841680633,4.7630793830775,1.70046113382978,.314193514488106,6.0685481480974,2.74657165283107,7.17423567654394},
			{1.414213562,.555278916264744,.140734880491406,.0308216365902836,.00626330408481417,2.90882269973526,1.80585734803576,.625327304106952,.173615533771693,3.86201434387853,3.27577043476888,1.43801908896753,4.43071582310304,4.76431769725595,4.71658607960963},
			{1.414213562,.981387965414959,.439603191878621,.170154848292167,.0611113213546646,2.48271365058504,2.72409573525399,1.66715454931021,.818062705137864,2.81340752869054,4.21756481949244,3.27222005765247,2.75487590413568,5.23550287189747,2.50302559510192},
			{1.414213562,1.4741928888764,.991945941760336,.576746649127296,.31115444533374,1.9899087271236,3.27976309580122,3.01515438576158,2.22245957594899,1.80736427869175,4.06994367638345,4.74332322095564,1.41847476579742,4.04940821803081,1.03297875354965},
			{1.414213562,1.9899087271236,1.80736427869175,1.41847476579742,1.03297875354965,1.4741928888764,3.27976309580122,4.06994367638345,4.04940821803081,.991945941760336,3.01515438576158,4.74332322095564,.576746649127296,2.22245957594899,.31115444533374},
			{1.414213562,2.48271365058504,2.81340752869054,2.75487590413568,2.50302559510192,.981387965414959,2.72409573525399,4.21756481949244,5.23550287189747,.439603191878621,1.66715454931021,3.27222005765247,.170154848292167,.818062705137864,.0611113213546646},
			{1.414213562,2.90882269973526,3.86201434387853,4.43071582310304,4.71658607960963,.555278916264744,1.80585734803576,3.27577043476888,4.76431769725595,.140734880491406,.625327304106952,1.43801908896753,.0308216365902836,.173615533771693,.00626330408481417},
			{1.414213562,3.23038451124486,4.7630793830775,6.0685481480974,7.17423567654394,.233717104755142,.844111317574503,1.70046113382978,2.74657165283107,.0249321870529784,.123027723654541,.314193514488106,.00229823078236534,.014376841680633,.000196571591120254},
			{1.414213562,3.41890639763464,5.33523806427656,7.19422103343416,9.00134910987265,.0451952183653594,.172756623071377,.368327957423731,.629639404683915,.000932318622844532,.00486900211054309,.0131603422009587,1.66187980881431e-005,.000110027806742895,2.74870870898337e-007}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[10][15] = 
		{
			{1.414213562,0.,0.,0.,0.,.0451952183653594,0.,0.,0.,.000932318622844532,0.,0.,1.66187980881431e-005,0.,2.74870870898337e-007},
			{1.414213562,0.,0.,0.,0.,.233717104755142,0.,0.,0.,.0249321870529784,0.,0.,.00229823078236534,0.,.000196571591120254},
			{1.414213562,0.,0.,0.,0.,.555278916264744,0.,0.,0.,.140734880491406,0.,0.,.0308216365902836,0.,.00626330408481417},
			{1.414213562,0.,0.,0.,0.,.981387965414959,0.,0.,0.,.439603191878621,0.,0.,.170154848292167,0.,.0611113213546646},
			{1.414213562,0.,0.,0.,0.,1.4741928888764,0.,0.,0.,.991945941760336,0.,0.,.576746649127296,0.,.31115444533374},
			{1.414213562,0.,0.,0.,0.,1.9899087271236,0.,0.,0.,1.80736427869175,0.,0.,1.41847476579742,0.,1.03297875354965},
			{1.414213562,0.,0.,0.,0.,2.48271365058504,0.,0.,0.,2.81340752869054,0.,0.,2.75487590413568,0.,2.50302559510192},
			{1.414213562,0.,0.,0.,0.,2.90882269973526,0.,0.,0.,3.86201434387853,0.,0.,4.43071582310304,0.,4.71658607960963},
			{1.414213562,0.,0.,0.,0.,3.23038451124486,0.,0.,0.,4.7630793830775,0.,0.,6.0685481480974,0.,7.17423567654394},
			{1.414213562,0.,0.,0.,0.,3.41890639763464,0.,0.,0.,5.33523806427656,0.,0.,7.19422103343416,0.,9.00134910987265}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[10][15] = 
		{
			{1.414213562,3.41890639763464,5.33523806427656,7.19422103343416,9.00134910987265,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.23038451124486,4.7630793830775,6.0685481480974,7.17423567654394,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.90882269973526,3.86201434387853,4.43071582310304,4.71658607960963,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.48271365058504,2.81340752869054,2.75487590413568,2.50302559510192,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.9899087271236,1.80736427869175,1.41847476579742,1.03297875354965,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.4741928888764,.991945941760336,.576746649127296,.31115444533374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.981387965414959,.439603191878621,.170154848292167,.0611113213546646,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.555278916264744,.140734880491406,.0308216365902836,.00626330408481417,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.233717104755142,.0249321870529784,.00229823078236534,.000196571591120254,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0451952183653594,.000932318622844532,1.66187980881431e-005,2.74870870898337e-007,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 10; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[15][10] = 
		{
			{-.0666713464862914,-.149451343590541,-.219086364157122,-.269266719261963,-.295524224714757,-.295524224714757,-.269266719261963,-.219086364157122,-.149451343590541,-.0666713464862914},
			{-.00213067258194671,-.0246987698646587,-.0860223958122111,-.186856585689445,-.308057935549154,-.415825621622801,-.472709481205777,-.450627405966319,-.341380763300079,-.161180106771333},
			{-4.395300651663e-005,-.0026347851214796,-.0218022893353363,-.0837005897640987,-.207284149381534,-.37767982241688,-.535673702440208,-.598293428384306,-.503352981601333,-.251523189241307},
			{-7.83472648479441e-007,-.000242872567024931,-.00477480946005413,-.0323975380897769,-.120521122714454,-.296414676313609,-.524529262214111,-.686395213472642,-.641312386098096,-.339162638412016},
			{-1.29584466981531e-008,-2.0773304102391e-005,-.000970295120694341,-.0116356153302479,-.0650210679957301,-.215858660481792,-.476576882004331,-.730681505704966,-.758159297392565,-.424357452903382},
			{-.161180106771333,-.341380763300079,-.450627405966319,-.472709481205777,-.415825621622801,-.308057935549154,-.186856585689445,-.0860223958122111,-.0246987698646587,-.00213067258194671},
			{-.00814439698359798,-.0892040451843217,-.279758822139498,-.518668708114341,-.685362855842359,-.685362855842359,-.518668708114341,-.279758822139498,-.0892040451843217,-.00814439698359798},
			{-.000229543072776175,-.0130013309753251,-.0968741136939265,-.317426764825887,-.630068319043131,-.85048466602198,-.80302570427492,-.5074740147261,-.179701431147967,-.0173643652676472},
			{-5.18712464632104e-006,-.00151931671591861,-.026896076418173,-.155759403364294,-.46442111746321,-.846193429133371,-.99684143833616,-.738075966372677,-.290252359754073,-.0296835697358091},
			{-.251523189241307,-.503352981601333,-.598293428384306,-.535673702440208,-.37767982241688,-.207284149381534,-.0837005897640987,-.0218022893353363,-.0026347851214796,-4.395300651663e-005},
			{-.0173643652676472,-.179701431147967,-.5074740147261,-.80302570427492,-.85048466602198,-.630068319043131,-.317426764825887,-.0968741136939265,-.0130013309753251,-.000229543072776175},
			{-.000620428030017239,-.0332033605988752,-.222774255663801,-.62303175619127,-.991198892706415,-.991198892706415,-.62303175619127,-.222774255663801,-.0332033605988752,-.000620428030017239},
			{-.339162638412016,-.641312386098096,-.686395213472642,-.524529262214111,-.296414676313609,-.120521122714454,-.0323975380897769,-.00477480946005413,-.000242872567024931,-7.83472648479441e-007},
			{-.0296835697358091,-.290252359754073,-.738075966372677,-.99684143833616,-.846193429133371,-.46442111746321,-.155759403364294,-.026896076418173,-.00151931671591861,-5.18712464632104e-006},
			{-.424357452903382,-.758159297392565,-.730681505704966,-.476576882004331,-.215858660481792,-.0650210679957301,-.0116356153302479,-.000970295120694341,-2.0773304102391e-005,-1.29584466981531e-008}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.00150661303158024,-.0174646676628736,-.060827019428775,-.132127558885229,-.217829855282603,-.294033116918165,-.334256079779963,-.318641694631351,-.241392652759809,-.113971546520449},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.10794689696449e-005,-.00186307442685916,-.0154165466384746,-.0591852546271217,-.146572027698836,-.267059963618762,-.378778507598704,-.423057340461501,-.355924306714669,-.177853752785117},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-5.53998822760152e-007,-.000171736839152821,-.00337630014896868,-.0229085188830744,-.0852213031700891,-.209596827719864,-.370898198340224,-.485354710148538,-.453476337188524,-.239824201609527},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-9.16300553632585e-009,-1.46889442023262e-005,-.000686102259776195,-.00822762250546704,-.0459768381119019,-.152635122644788,-.336990745110908,-.516669847707885,-.536099580547344,-.300066032674195}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[15][10] = 
		{
			{-.0471437611988572,-.105678058482433,-.154917453720137,-.190400323087757,-.208967183245572,-.208967183245572,-.190400323087757,-.154917453720137,-.105678058482433,-.0471437611988572},
			{-.113971546520449,-.241392652759809,-.318641694631351,-.334256079779963,-.294033116918165,-.217829855282603,-.132127558885229,-.060827019428775,-.0174646676628736,-.00150661303158024},
			{-.177853752785117,-.355924306714669,-.423057340461501,-.378778507598704,-.267059963618762,-.146572027698836,-.0591852546271217,-.0154165466384746,-.00186307442685916,-3.10794689696449e-005},
			{-.239824201609527,-.453476337188524,-.485354710148538,-.370898198340224,-.209596827719864,-.0852213031700891,-.0229085188830744,-.00337630014896868,-.000171736839152821,-5.53998822760152e-007},
			{-.300066032674195,-.536099580547344,-.516669847707885,-.336990745110908,-.152635122644788,-.0459768381119019,-.00822762250546704,-.000686102259776195,-1.46889442023262e-005,-9.16300553632585e-009},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 10; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[15][15] = 
		{
			{.5000000000,0.,0.,0.,0.,1.224744872,0.,0.,0.,1.936491673,0.,0.,2.645751311,0.,3.354101966},
			{0.,-.5000000000,0.,0.,0.,-.5000000000,-1.936491673,0.,0.,-1.581138830,-4.183300132,0.,-3.240370350,-7.245688375,-5.477225576},
			{0.,0.,.5000000000,0.,0.,0.,1.224744872,2.645751311,0.,.5000000000,5.291502622,7.245688375,2.049390154,13.74772709,5.196152424},
			{0.,0.,0.,-.5000000000,0.,0.,0.,-1.936491673,-3.354101966,0.,-1.936491673,-10.60660172,-.5000000000,-10.06230590,-2.535462764},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,4.183300132,0.,2.645751311,.5000000000},
			{0.,.5000000000,0.,0.,0.,-.5000000000,1.936491673,0.,0.,-1.581138830,4.183300132,0.,-3.240370350,7.245688375,-5.477225576},
			{0.,0.,-.4082482906,0.,0.,0.,0.,-2.160246899,0.,.4082482906,0.,-5.916079782,1.673320053,0.,4.242640686},
			{0.,0.,0.,.3872983346,0.,0.,0.,.5000000000,2.598076212,0.,-.5000000000,2.738612788,-.3872983346,-2.598076212,-1.963961012},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,-1.,0.,0.,0.,0.,1.,.3779644731},
			{0.,0.,.5000000000,0.,0.,0.,-1.224744872,2.645751311,0.,.5000000000,-5.291502622,7.245688375,2.049390154,-13.74772709,5.196152424},
			{0.,0.,0.,-.3872983346,0.,0.,0.,.5000000000,-2.598076212,0.,.5000000000,2.738612788,-.3872983346,2.598076212,-1.963961012},
			{0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3585685828},
			{0.,0.,0.,.5000000000,0.,0.,0.,-1.936491673,3.354101966,0.,1.936491673,-10.60660172,-.5000000000,10.06230590,-2.535462764},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,1.,0.,0.,0.,0.,-1.,.3779644731},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,-2.645751311,0.,0.,4.183300132,0.,-2.645751311,.5000000000}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[15][15] = 
		{
			{.5000000000,1.224744872,1.936491673,2.645751311,3.354101966,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,.5000000000,1.936491673,4.183300132,7.245688375,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,2.049390154,5.196152424,0.,-1.224744872,-5.291502622,-13.74772709,.5000000000,2.645751311,7.245688375,0.,0.,0.},
			{0.,0.,0.,-.5000000000,-2.535462764,0.,0.,1.936491673,10.06230590,0.,-1.936491673,-10.60660172,.5000000000,3.354101966,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,-2.645751311,0.,0.,4.183300132,0.,-2.645751311,.5000000000},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,-.5000000000,-1.936491673,-4.183300132,-7.245688375,0.,0.,0.,0.,0.,0.},
			{0.,0.,.4082482906,1.673320053,4.242640686,0.,0.,0.,0.,-.4082482906,-2.160246899,-5.916079782,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,0.,0.,.5000000000,2.598076212,0.,.5000000000,2.738612788,-.3872983346,-2.598076212,0.},
			{0.,0.,0.,0.,.3779644731,0.,0.,0.,-1.,0.,0.,0.,0.,1.,-.3779644731},
			{0.,0.,.5000000000,2.049390154,5.196152424,0.,1.224744872,5.291502622,13.74772709,.5000000000,2.645751311,7.245688375,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,0.,0.,-.5000000000,-2.598076212,0.,.5000000000,2.738612788,.3872983346,2.598076212,0.},
			{0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,0.,-1.,0.,0.,.3585685828},
			{0.,0.,0.,-.5000000000,-2.535462764,0.,0.,-1.936491673,-10.06230590,0.,-1.936491673,-10.60660172,-.5000000000,-3.354101966,0.},
			{0.,0.,0.,0.,.3779644731,0.,0.,0.,1.,0.,0.,0.,0.,-1.,-.3779644731},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,2.645751311,0.,0.,4.183300132,0.,2.645751311,.5000000000}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[15][15] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,1.224744872,2.371708244,2.561737692,2.218529919,.9682458365,2.561737692,3.507803799,.6614378278,2.218529919,.4192627457},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,.5000000000,0.,-1.045825033,-1.811422094,.7905694150,1.045825033,0.,.8100925875,1.811422094,.6846531970},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,-.6123724358,-.6614378278,0.,.2500000000,-.6614378278,-1.811422094,.5123475385,0.,.6495190530},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,.4841229182,.8385254914,0.,-.4841229182,0.,.1250000000,-.8385254914,.3169328455},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,-.3307189139,0.,0.,.5229125166,0.,-.3307189139,.6250000000e-1},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-.7905694150,-3.137475099,-5.728219620,-.8100925875,-3.622844188,-.6846531970},
			{0.,0.,.2041241452,.4183300132,.5303300858,0.,0.,.5400617248,1.403121520,-.2041241452,-.5400617248,0.,-.4183300132,-1.403121520,-.5303300858},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,.1250000000,0.,0.,.1250000000,.6846531969,-.9682458365e-1,0.,-.2454951265},
			{0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,-.1250000000,0.,0.,0.,0.,.1250000000,-.4724555913e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,.6123724358,1.984313483,3.436931772,.2500000000,1.984313483,5.434266281,.5123475385,3.436931772,.6495190530},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,-.1250000000,-.6495190530,0.,.1250000000,0.,.9682458365e-1,.6495190530,.2454951265},
			{0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4482107285e-1},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,-.4841229182,-1.677050983,0.,-.4841229182,-2.651650429,-.1250000000,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,.4724555913e-1,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.1250000000,-.4724555913e-1},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,.3307189139,0.,0.,.5229125166,0.,.3307189139,.6250000000e-1}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[15][15] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,1.224744872,2.371708244,2.561737692,2.218529919,.9682458365,2.561737692,3.507803799,.6614378278,2.218529919,.4192627457},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-.7905694150,-3.137475099,-5.728219620,-.8100925875,-3.622844188,-.6846531970},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,.6123724358,1.984313483,3.436931772,.2500000000,1.984313483,5.434266281,.5123475385,3.436931772,.6495190530},
			{0.,0.,0.,-.1250000000,-.3169328455,0.,0.,-.4841229182,-1.677050983,0.,-.4841229182,-2.651650429,-.1250000000,-1.677050983,-.3169328455},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,.3307189139,0.,0.,.5229125166,0.,.3307189139,.6250000000e-1},
			{0.,.5000000000,.7905694150,.8100925875,.6846531970,-.5000000000,0.,1.045825033,1.811422094,-.7905694150,-1.045825033,0.,-.8100925875,-1.811422094,-.6846531970},
			{0.,0.,-.2041241452,-.4183300132,-.5303300858,0.,0.,-.5400617248,-1.403121520,.2041241452,.5400617248,0.,.4183300132,1.403121520,.5303300858},
			{0.,0.,0.,.9682458365e-1,.2454951265,0.,0.,.1250000000,.6495190530,0.,-.1250000000,0.,-.9682458365e-1,-.6495190530,-.2454951265},
			{0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,-.1250000000,0.,0.,0.,0.,.1250000000,.4724555913e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,0.,-.6123724358,-.6614378278,0.,.2500000000,-.6614378278,-1.811422094,.5123475385,0.,.6495190530},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,0.,0.,.1250000000,0.,0.,.1250000000,.6846531969,-.9682458365e-1,0.,-.2454951265},
			{0.,0.,0.,0.,.4482107285e-1,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.4482107285e-1},
			{0.,0.,0.,.1250000000,.3169328455,0.,0.,-.4841229182,-.8385254914,0.,.4841229182,0.,-.1250000000,.8385254914,-.3169328455},
			{0.,0.,0.,0.,-.4724555913e-1,0.,0.,0.,.1250000000,0.,0.,0.,0.,-.1250000000,.4724555913e-1},
			{0.,0.,0.,0.,.6250000000e-1,0.,0.,0.,-.3307189139,0.,0.,.5229125166,0.,-.3307189139,.6250000000e-1}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[15][15] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[15][15] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980}
		};
		for (int j = 0; j < 15; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][15] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 15; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
