
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION2D_MATRICES_CONST_HPP_
#define CCOMPUTATION2D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (5)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (21)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (21)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS (12)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (52)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*5+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_2D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[21],
		T io_hv[21]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{440.9984,-2400.5005,6831.94254,-10000.9403,7231.4440,-2047.25768,-2400.5005,5578.25780,-7746.69470,5466.45789,-1525.93588,6831.94254,-7746.69470,5185.9373,-1412.742036,-10000.9403,5466.45789,-1412.742036,7231.4440,-1525.93588,-2047.25768},
			{-2400.5005,19600.0001,-66939.09355,108876.4438,-84349.27383,25073.68342,9800.0001,-36437.02740,63251.49791,-51009.64615,15574.01681,-22313.03111,42167.66538,-36293.99952,11534.99024,27219.1110,-25504.82298,8651.24268,-16869.85472,6229.60672,4178.947267},
			{6831.94254,-66939.09355,254016.0000,-442668.2732,360093.3630,-111005.9285,-22313.03111,103701.5977,-205733.6219,181469.9975,-59099.23855,42336.00000,-102866.8110,103294.5334,-36476.84197,-44266.82730,54440.99929,-21886.10519,24006.22418,-11819.84771,-5285.996592},
			{-10000.9403,108876.4438,-442668.2732,810000.,-683307.2148,216661.3874,27219.1110,-144574.8525,313711.6510,-295160.9730,100931.1647,-44266.82730,125484.6604,-140007.1426,53396.62909,40500.,-59032.19459,26698.31455,-19523.06328,11534.99025,3868.953347},
			{7231.4440,-84349.27383,360093.3630,-683307.2148,592900.,-192267.9080,-16869.85472,98004.99978,-226837.4970,224095.1360,-79615.57635,24006.22418,-75612.49898,91112.27685,-36854.85042,-19523.06328,32013.59086,-15794.93589,8470.,-5686.826882,-1525.935778},
			{-2047.25768,25073.68342,-111005.9285,216661.3874,-192267.9080,63504.,4178.947267,-25895.98888,62934.44589,-64595.94539,23666.54347,-5285.996592,17981.27025,-22980.41043,9738.205169,3868.953347,-6920.994149,3651.826938,-1525.935778,1126.978260,252.},
			{-2400.5005,9800.0001,-22313.03111,27219.1110,-16869.85472,4178.947267,19600.0001,-36437.02740,42167.66538,-25504.82298,6229.60672,-66939.09355,63251.49791,-36293.99952,8651.24268,108876.4438,-51009.64615,11534.99024,-84349.27383,15574.01681,25073.68342},
			{5578.25780,-36437.02740,103701.5977,-144574.8525,98004.99978,-25895.98888,-36437.02740,112896.0000,-167980.7989,118535.7060,-32169.55082,103701.5977,-167980.7989,126509.4500,-35739.86008,-144574.8525,118535.7060,-35739.86008,98004.99978,-32169.55082,-25895.98888},
			{-7746.69470,63251.49791,-205733.6219,313711.6510,-226837.4970,62934.44589,42167.66538,-167980.7989,291600.,-228630.7067,67012.23766,-102866.8110,218700.,-195208.3195,62041.27659,125484.6604,-137178.4240,49633.02127,-75612.49898,33506.11883,17981.27025},
			{5466.45789,-51009.64615,181469.9975,-295160.9730,224095.1360,-64595.94539,-25504.82298,118535.7060,-228630.7067,193600.,-60183.71873,54440.99929,-137178.4240,137748.8148,-47759.39698,-59032.19459,77440.,-31839.59798,32013.59086,-17195.34821,-6920.994149},
			{-1525.93588,15574.01681,-59099.23855,100931.1647,-79615.57635,23666.54347,6229.60672,-32169.55082,67012.23766,-60183.71873,19600.,-11819.84771,33506.11883,-36704.11420,13609.55547,11534.99025,-17195.34821,7776.888839,-5686.826882,3500.,1126.978260},
			{6831.94254,-22313.03111,42336.00000,-44266.82730,24006.22418,-5285.996592,-66939.09355,103701.5977,-102866.8110,54440.99929,-11819.84771,254016.0000,-205733.6219,103294.5334,-21886.10519,-442668.2732,181469.9975,-36476.84197,360093.3630,-59099.23855,-111005.9285},
			{-7746.69470,42167.66538,-102866.8110,125484.6604,-75612.49898,17981.27025,63251.49791,-167980.7989,218700.,-137178.4240,33506.11883,-205733.6219,291600.,-195208.3195,49633.02127,313711.6510,-228630.7067,62041.27659,-226837.4970,67012.23766,62934.44589},
			{5185.9373,-36293.99952,103294.5334,-140007.1426,91112.27685,-22980.41043,-36293.99952,126509.4500,-195208.3195,137748.8148,-36704.11420,103294.5334,-195208.3195,156816.,-45308.54223,-140007.1426,137748.8148,-45308.54223,91112.27685,-36704.11420,-22980.41043},
			{-1412.742036,11534.99024,-36476.84197,53396.62909,-36854.85042,9738.205169,8651.24268,-35739.86008,62041.27659,-47759.39698,13609.55547,-21886.10519,49633.02127,-45308.54223,14400.,26698.31455,-31839.59798,12000.,-15794.93589,7776.888839,3651.826938},
			{-10000.9403,27219.1110,-44266.82730,40500.,-19523.06328,3868.953347,108876.4438,-144574.8525,125484.6604,-59032.19459,11534.99025,-442668.2732,313711.6510,-140007.1426,26698.31455,810000.,-295160.9730,53396.62909,-683307.2148,100931.1647,216661.3874},
			{5466.45789,-25504.82298,54440.99929,-59032.19459,32013.59086,-6920.994149,-51009.64615,118535.7060,-137178.4240,77440.,-17195.34821,181469.9975,-228630.7067,137748.8148,-31839.59798,-295160.9730,193600.,-47759.39698,224095.1360,-60183.71873,-64595.94539},
			{-1412.742036,8651.24268,-21886.10519,26698.31455,-15794.93589,3651.826938,11534.99024,-35739.86008,49633.02127,-31839.59798,7776.888839,-36476.84197,62041.27659,-45308.54223,12000.,53396.62909,-47759.39698,14400.,-36854.85042,13609.55547,9738.205169},
			{7231.4440,-16869.85472,24006.22418,-19523.06328,8470.,-1525.935778,-84349.27383,98004.99978,-75612.49898,32013.59086,-5686.826882,360093.3630,-226837.4970,91112.27685,-15794.93589,-683307.2148,224095.1360,-36854.85042,592900.,-79615.57635,-192267.9080},
			{-1525.93588,6229.60672,-11819.84771,11534.99025,-5686.826882,1126.978260,15574.01681,-32169.55082,33506.11883,-17195.34821,3500.,-59099.23855,67012.23766,-36704.11420,7776.888839,100931.1647,-60183.71873,13609.55547,-79615.57635,19600.,23666.54347},
			{-2047.25768,4178.947267,-5285.996592,3868.953347,-1525.935778,252.,25073.68342,-25895.98888,17981.27025,-6920.994149,1126.978260,-111005.9285,62934.44589,-22980.41043,3651.826938,216661.3874,-64595.94539,9738.205169,-192267.9080,23666.54347,63504.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{440.9984,-2400.5005,6831.94254,-10000.9403,7231.4440,-2047.25768,-2400.5005,5578.25780,-7746.69470,5466.45789,-1525.93588,6831.94254,-7746.69470,5185.9373,-1412.742036,-10000.9403,5466.45789,-1412.742036,7231.4440,-1525.93588,-2047.25768},
			{-2400.5005,19600.0001,-66939.09355,108876.4438,-84349.27383,25073.68342,9800.0001,-36437.02740,63251.49791,-51009.64615,15574.01681,-22313.03111,42167.66538,-36293.99952,11534.99024,27219.1110,-25504.82298,8651.24268,-16869.85472,6229.60672,4178.947267},
			{6831.94254,-66939.09355,254016.0000,-442668.2732,360093.3630,-111005.9285,-22313.03111,103701.5977,-205733.6219,181469.9975,-59099.23855,42336.00000,-102866.8110,103294.5334,-36476.84197,-44266.82730,54440.99929,-21886.10519,24006.22418,-11819.84771,-5285.996592},
			{-10000.9403,108876.4438,-442668.2732,810000.,-683307.2148,216661.3874,27219.1110,-144574.8525,313711.6510,-295160.9730,100931.1647,-44266.82730,125484.6604,-140007.1426,53396.62909,40500.,-59032.19459,26698.31455,-19523.06328,11534.99025,3868.953347},
			{7231.4440,-84349.27383,360093.3630,-683307.2148,592900.,-192267.9080,-16869.85472,98004.99978,-226837.4970,224095.1360,-79615.57635,24006.22418,-75612.49898,91112.27685,-36854.85042,-19523.06328,32013.59086,-15794.93589,8470.,-5686.826882,-1525.935778},
			{-2047.25768,25073.68342,-111005.9285,216661.3874,-192267.9080,63504.,4178.947267,-25895.98888,62934.44589,-64595.94539,23666.54347,-5285.996592,17981.27025,-22980.41043,9738.205169,3868.953347,-6920.994149,3651.826938,-1525.935778,1126.978260,252.},
			{-2400.5005,9800.0001,-22313.03111,27219.1110,-16869.85472,4178.947267,19600.0001,-36437.02740,42167.66538,-25504.82298,6229.60672,-66939.09355,63251.49791,-36293.99952,8651.24268,108876.4438,-51009.64615,11534.99024,-84349.27383,15574.01681,25073.68342},
			{5578.25780,-36437.02740,103701.5977,-144574.8525,98004.99978,-25895.98888,-36437.02740,112896.0000,-167980.7989,118535.7060,-32169.55082,103701.5977,-167980.7989,126509.4500,-35739.86008,-144574.8525,118535.7060,-35739.86008,98004.99978,-32169.55082,-25895.98888},
			{-7746.69470,63251.49791,-205733.6219,313711.6510,-226837.4970,62934.44589,42167.66538,-167980.7989,291600.,-228630.7067,67012.23766,-102866.8110,218700.,-195208.3195,62041.27659,125484.6604,-137178.4240,49633.02127,-75612.49898,33506.11883,17981.27025},
			{5466.45789,-51009.64615,181469.9975,-295160.9730,224095.1360,-64595.94539,-25504.82298,118535.7060,-228630.7067,193600.,-60183.71873,54440.99929,-137178.4240,137748.8148,-47759.39698,-59032.19459,77440.,-31839.59798,32013.59086,-17195.34821,-6920.994149},
			{-1525.93588,15574.01681,-59099.23855,100931.1647,-79615.57635,23666.54347,6229.60672,-32169.55082,67012.23766,-60183.71873,19600.,-11819.84771,33506.11883,-36704.11420,13609.55547,11534.99025,-17195.34821,7776.888839,-5686.826882,3500.,1126.978260},
			{6831.94254,-22313.03111,42336.00000,-44266.82730,24006.22418,-5285.996592,-66939.09355,103701.5977,-102866.8110,54440.99929,-11819.84771,254016.0000,-205733.6219,103294.5334,-21886.10519,-442668.2732,181469.9975,-36476.84197,360093.3630,-59099.23855,-111005.9285},
			{-7746.69470,42167.66538,-102866.8110,125484.6604,-75612.49898,17981.27025,63251.49791,-167980.7989,218700.,-137178.4240,33506.11883,-205733.6219,291600.,-195208.3195,49633.02127,313711.6510,-228630.7067,62041.27659,-226837.4970,67012.23766,62934.44589},
			{5185.9373,-36293.99952,103294.5334,-140007.1426,91112.27685,-22980.41043,-36293.99952,126509.4500,-195208.3195,137748.8148,-36704.11420,103294.5334,-195208.3195,156816.,-45308.54223,-140007.1426,137748.8148,-45308.54223,91112.27685,-36704.11420,-22980.41043},
			{-1412.742036,11534.99024,-36476.84197,53396.62909,-36854.85042,9738.205169,8651.24268,-35739.86008,62041.27659,-47759.39698,13609.55547,-21886.10519,49633.02127,-45308.54223,14400.,26698.31455,-31839.59798,12000.,-15794.93589,7776.888839,3651.826938},
			{-10000.9403,27219.1110,-44266.82730,40500.,-19523.06328,3868.953347,108876.4438,-144574.8525,125484.6604,-59032.19459,11534.99025,-442668.2732,313711.6510,-140007.1426,26698.31455,810000.,-295160.9730,53396.62909,-683307.2148,100931.1647,216661.3874},
			{5466.45789,-25504.82298,54440.99929,-59032.19459,32013.59086,-6920.994149,-51009.64615,118535.7060,-137178.4240,77440.,-17195.34821,181469.9975,-228630.7067,137748.8148,-31839.59798,-295160.9730,193600.,-47759.39698,224095.1360,-60183.71873,-64595.94539},
			{-1412.742036,8651.24268,-21886.10519,26698.31455,-15794.93589,3651.826938,11534.99024,-35739.86008,49633.02127,-31839.59798,7776.888839,-36476.84197,62041.27659,-45308.54223,12000.,53396.62909,-47759.39698,14400.,-36854.85042,13609.55547,9738.205169},
			{7231.4440,-16869.85472,24006.22418,-19523.06328,8470.,-1525.935778,-84349.27383,98004.99978,-75612.49898,32013.59086,-5686.826882,360093.3630,-226837.4970,91112.27685,-15794.93589,-683307.2148,224095.1360,-36854.85042,592900.,-79615.57635,-192267.9080},
			{-1525.93588,6229.60672,-11819.84771,11534.99025,-5686.826882,1126.978260,15574.01681,-32169.55082,33506.11883,-17195.34821,3500.,-59099.23855,67012.23766,-36704.11420,7776.888839,100931.1647,-60183.71873,13609.55547,-79615.57635,19600.,23666.54347},
			{-2047.25768,4178.947267,-5285.996592,3868.953347,-1525.935778,252.,25073.68342,-25895.98888,17981.27025,-6920.994149,1126.978260,-111005.9285,62934.44589,-22980.41043,3651.826938,216661.3874,-64595.94539,9738.205169,-192267.9080,23666.54347,63504.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,.9476070828,2.,1.936491673,1.673320053,1.449137675,1.271348312,1.581138830,1.673320053,1.527525232,1.373213125,1.296148140,1.449137675,1.373213125,1.095445115,1.271348312,.9476070828},
			{3.162277659,3.872983346,3.674234614,3.346640106,3.030457633,2.752549839,1.936491673,3.,3.240370349,3.207134903,3.077444275,1.224744872,2.160246899,2.535462764,2.659215782,.8366600264,1.603567452,1.994411836,.6060915266,1.230977710,.4587583064},
			{3.415650256,5.019960159,5.291502622,5.163977793,4.909902528,4.624812309,1.673320053,3.240370349,4.,4.330127020,4.432026303,.8819171036,2.,2.738612788,3.191423693,.5163977793,1.299038106,1.914854215,.3273268354,.8864052607,.2202291575},
			{3.549647870,5.796550700,6.546536711,6.708203931,6.614378278,6.408327913,1.449137675,3.207134903,4.330127020,5.,5.373546315,.6546536711,1.732050808,2.635231383,3.316624790,.3354101966,1.,1.658312395,.1889822365,.6141195788,.1144344270},
			{3.633180424,6.356741560,7.538168182,8.010409897,8.124038403,8.049844717,1.271348312,3.077444275,4.432026303,5.373546315,6.,.5025445453,1.477342101,2.427520781,3.240370350,.2288688541,.7676494735,1.388730150,.1160576915,.4285714286,.6388765648e-1},
			{2.581988897,1.581138830,1.,.6831300514,.4948716594,.3745745885,3.162277660,2.449489743,1.763834207,1.309307342,1.005089091,3.,2.645751311,2.070196679,1.628430445,2.732520206,2.618614683,2.171240595,2.474358298,2.512722727,2.247447532},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,1.321374945,3.346640106,4.320493799,4.,3.464101616,2.954684202,2.645751311,4.,4.107919181,3.829708432,2.065591118,3.464101616,3.829708432,1.636634177,2.954684202,1.321374945},
			{3.741657386,4.582575696,4.140393357,3.535533905,2.988071523,2.533114026,3.055050464,5.070925527,5.477225575,5.270462768,4.855041564,2.070196679,4.107919181,5.,5.244044240,1.414213562,3.162277660,4.195235392,.9960238407,2.427520781,.7237468643},
			{3.924283374,5.492852500,5.428101485,4.944132325,4.387482194,3.864367134,2.746426250,5.318431565,6.382847385,6.633249580,6.480740700,1.628430445,3.829708432,5.244044240,6.,.9888264647,2.653299832,4.,.6267831705,1.851640200,.4140393357},
			{2.645751311,1.296148140,.6831300514,.4000000000,.2535462764,.1705887719,3.888444420,2.509980079,1.549193338,1.006230590,.6866065626,4.098780308,3.098386677,2.121320343,1.483239697,4.,3.354101966,2.472066162,3.803194146,3.433032812,3.582364211},
			{3.549647870,2.898275350,1.963961013,1.341640786,.9449111824,.6866065625,4.347413025,4.810702356,3.897114318,3.,2.302948420,3.927922025,5.196152424,4.743416490,3.979949748,3.354101966,5.,4.974937185,2.834733546,4.605896842,2.403122968},
			{3.924283374,4.119639374,3.256860892,2.472066162,1.880349512,1.449137675,4.119639374,5.983235509,5.744562647,4.974937185,4.166190450,3.256860892,5.744562647,6.292853088,6.,2.472066162,4.974937185,6.,1.880349512,4.166190450,1.449137675},
			{2.683281572,1.095445115,.4948716594,.2535462764,.1428571429,.8650429586e-1,4.381780461,2.424366106,1.309307341,.7559289460,.4642307659,4.948716594,3.273268354,1.992047682,1.253566341,5.070925527,3.779644731,2.507132682,5.,4.062019202,4.844240569},
			{3.633180424,2.542696624,1.507633636,.9154754167,.5802884574,.3833259389,5.085393248,4.923910840,3.545621043,2.456478316,1.714285714,5.025445453,5.909368405,4.855041564,3.703280400,4.577377083,6.141195788,5.554920600,4.062019202,6.,3.577708763},
			{2.708012801,.9476070828,.3745745885,.1705887719,.8650429586e-1,.4761904762e-1,4.738035416,2.293791533,1.101145788,.5721721350,.3194382825,5.618618828,3.303437362,1.809367160,1.035098339,5.970607017,4.005204947,2.415229459,6.055300709,4.472135954,6.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,.9476070828,2.,1.936491673,1.673320053,1.449137675,1.271348312,1.581138830,1.673320053,1.527525232,1.373213125,1.296148140,1.449137675,1.373213125,1.095445115,1.271348312,.9476070828},
			{3.162277659,3.872983346,3.674234614,3.346640106,3.030457633,2.752549839,1.936491673,3.,3.240370349,3.207134903,3.077444275,1.224744872,2.160246899,2.535462764,2.659215782,.8366600264,1.603567452,1.994411836,.6060915266,1.230977710,.4587583064},
			{3.415650256,5.019960159,5.291502622,5.163977793,4.909902528,4.624812309,1.673320053,3.240370349,4.,4.330127020,4.432026303,.8819171036,2.,2.738612788,3.191423693,.5163977793,1.299038106,1.914854215,.3273268354,.8864052607,.2202291575},
			{3.549647870,5.796550700,6.546536711,6.708203931,6.614378278,6.408327913,1.449137675,3.207134903,4.330127020,5.,5.373546315,.6546536711,1.732050808,2.635231383,3.316624790,.3354101966,1.,1.658312395,.1889822365,.6141195788,.1144344270},
			{3.633180424,6.356741560,7.538168182,8.010409897,8.124038403,8.049844717,1.271348312,3.077444275,4.432026303,5.373546315,6.,.5025445453,1.477342101,2.427520781,3.240370350,.2288688541,.7676494735,1.388730150,.1160576915,.4285714286,.6388765648e-1},
			{2.581988897,1.581138830,1.,.6831300514,.4948716594,.3745745885,3.162277660,2.449489743,1.763834207,1.309307342,1.005089091,3.,2.645751311,2.070196679,1.628430445,2.732520206,2.618614683,2.171240595,2.474358298,2.512722727,2.247447532},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,1.321374945,3.346640106,4.320493799,4.,3.464101616,2.954684202,2.645751311,4.,4.107919181,3.829708432,2.065591118,3.464101616,3.829708432,1.636634177,2.954684202,1.321374945},
			{3.741657386,4.582575696,4.140393357,3.535533905,2.988071523,2.533114026,3.055050464,5.070925527,5.477225575,5.270462768,4.855041564,2.070196679,4.107919181,5.,5.244044240,1.414213562,3.162277660,4.195235392,.9960238407,2.427520781,.7237468643},
			{3.924283374,5.492852500,5.428101485,4.944132325,4.387482194,3.864367134,2.746426250,5.318431565,6.382847385,6.633249580,6.480740700,1.628430445,3.829708432,5.244044240,6.,.9888264647,2.653299832,4.,.6267831705,1.851640200,.4140393357},
			{2.645751311,1.296148140,.6831300514,.4000000000,.2535462764,.1705887719,3.888444420,2.509980079,1.549193338,1.006230590,.6866065626,4.098780308,3.098386677,2.121320343,1.483239697,4.,3.354101966,2.472066162,3.803194146,3.433032812,3.582364211},
			{3.549647870,2.898275350,1.963961013,1.341640786,.9449111824,.6866065625,4.347413025,4.810702356,3.897114318,3.,2.302948420,3.927922025,5.196152424,4.743416490,3.979949748,3.354101966,5.,4.974937185,2.834733546,4.605896842,2.403122968},
			{3.924283374,4.119639374,3.256860892,2.472066162,1.880349512,1.449137675,4.119639374,5.983235509,5.744562647,4.974937185,4.166190450,3.256860892,5.744562647,6.292853088,6.,2.472066162,4.974937185,6.,1.880349512,4.166190450,1.449137675},
			{2.683281572,1.095445115,.4948716594,.2535462764,.1428571429,.8650429586e-1,4.381780461,2.424366106,1.309307341,.7559289460,.4642307659,4.948716594,3.273268354,1.992047682,1.253566341,5.070925527,3.779644731,2.507132682,5.,4.062019202,4.844240569},
			{3.633180424,2.542696624,1.507633636,.9154754167,.5802884574,.3833259389,5.085393248,4.923910840,3.545621043,2.456478316,1.714285714,5.025445453,5.909368405,4.855041564,3.703280400,4.577377083,6.141195788,5.554920600,4.062019202,6.,3.577708763},
			{2.708012801,.9476070828,.3745745885,.1705887719,.8650429586e-1,.4761904762e-1,4.738035416,2.293791533,1.101145788,.5721721350,.3194382825,5.618618828,3.303437362,1.809367160,1.035098339,5.970607017,4.005204947,2.415229459,6.055300709,4.472135954,6.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,.9476070828,2.,1.936491673,1.673320053,1.449137675,1.271348312,1.581138830,1.673320053,1.527525232,1.373213125,1.296148140,1.449137675,1.373213125,1.095445115,1.271348312,.9476070828},
			{2.581988897,3.162277660,3.,2.732520206,2.474358298,2.247447532,1.581138830,2.449489743,2.645751311,2.618614683,2.512722727,1.,1.763834207,2.070196679,2.171240595,.6831300514,1.309307342,1.628430445,.4948716594,1.005089091,.3745745885},
			{2.645751311,3.888444420,4.098780308,4.,3.803194146,3.582364211,1.296148140,2.509980079,3.098386677,3.354101966,3.433032812,.6831300514,1.549193338,2.121320343,2.472066162,.4000000000,1.006230590,1.483239697,.2535462764,.6866065626,.1705887719},
			{2.683281572,4.381780461,4.948716594,5.070925527,5.,4.844240569,1.095445115,2.424366106,3.273268354,3.779644731,4.062019202,.4948716594,1.309307341,1.992047682,2.507132682,.2535462764,.7559289460,1.253566341,.1428571429,.4642307659,.8650429586e-1},
			{2.708012801,4.738035416,5.618618828,5.970607017,6.055300709,6.,.9476070828,2.293791533,3.303437362,4.005204947,4.472135954,.3745745885,1.101145788,1.809367160,2.415229459,.1705887719,.5721721350,1.035098339,.8650429586e-1,.3194382825,.4761904762e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,1.936491673,1.224744872,.8366600264,.6060915266,.4587583064,3.872983346,3.,2.160246899,1.603567452,1.230977710,3.674234614,3.240370349,2.535462764,1.994411836,3.346640106,3.207134903,2.659215782,3.030457633,3.077444275,2.752549839},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,1.321374945,3.346640106,4.320493799,4.,3.464101616,2.954684202,2.645751311,4.,4.107919181,3.829708432,2.065591118,3.464101616,3.829708432,1.636634177,2.954684202,1.321374945},
			{3.549647870,4.347413025,3.927922025,3.354101966,2.834733546,2.403122968,2.898275350,4.810702356,5.196152424,5.,4.605896842,1.963961013,3.897114318,4.743416490,4.974937185,1.341640786,3.,3.979949748,.9449111824,2.302948420,.6866065625},
			{3.633180424,5.085393248,5.025445453,4.577377083,4.062019202,3.577708763,2.542696624,4.923910840,5.909368405,6.141195788,6.,1.507633636,3.545621043,4.855041564,5.554920600,.9154754167,2.456478316,3.703280400,.5802884574,1.714285714,.3833259389},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,1.673320053,.8819171036,.5163977793,.3273268354,.2202291575,5.019960159,3.240370349,2.,1.299038106,.8864052607,5.291502622,4.,2.738612788,1.914854215,5.163977793,4.330127020,3.191423693,4.909902528,4.432026303,4.624812309},
			{3.741657386,3.055050464,2.070196679,1.414213562,.9960238407,.7237468643,4.582575696,5.070925527,4.107919181,3.162277660,2.427520781,4.140393357,5.477225575,5.,4.195235392,3.535533905,5.270462768,5.244044240,2.988071523,4.855041564,2.533114026},
			{3.924283374,4.119639374,3.256860892,2.472066162,1.880349512,1.449137675,4.119639374,5.983235509,5.744562647,4.974937185,4.166190450,3.256860892,5.744562647,6.292853088,6.,2.472066162,4.974937185,6.,1.880349512,4.166190450,1.449137675},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,1.449137675,.6546536711,.3354101966,.1889822365,.1144344270,5.796550700,3.207134903,1.732050808,1.,.6141195788,6.546536711,4.330127020,2.635231383,1.658312395,6.708203931,5.,3.316624790,6.614378278,5.373546315,6.408327913},
			{3.924283374,2.746426250,1.628430445,.9888264647,.6267831705,.4140393357,5.492852500,5.318431565,3.829708432,2.653299832,1.851640200,5.428101485,6.382847385,5.244044240,4.,4.944132325,6.633249580,6.,4.387482194,6.480740700,3.864367134},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.633180424,1.271348312,.5025445453,.2288688541,.1160576915,.6388765648e-1,6.356741560,3.077444275,1.477342101,.7676494735,.4285714286,7.538168182,4.432026303,2.427520781,1.388730150,8.010409897,5.373546315,3.240370350,8.124038403,6.,8.049844717},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,2.,1.581138830,1.296148140,1.095445115,.9476070828,2.,1.936491673,1.673320053,1.449137675,1.271348312,1.581138830,1.673320053,1.527525232,1.373213125,1.296148140,1.449137675,1.373213125,1.095445115,1.271348312,.9476070828},
			{2.581988897,3.162277660,3.,2.732520206,2.474358298,2.247447532,1.581138830,2.449489743,2.645751311,2.618614683,2.512722727,1.,1.763834207,2.070196679,2.171240595,.6831300514,1.309307342,1.628430445,.4948716594,1.005089091,.3745745885},
			{2.645751311,3.888444420,4.098780308,4.,3.803194146,3.582364211,1.296148140,2.509980079,3.098386677,3.354101966,3.433032812,.6831300514,1.549193338,2.121320343,2.472066162,.4000000000,1.006230590,1.483239697,.2535462764,.6866065626,.1705887719},
			{2.683281572,4.381780461,4.948716594,5.070925527,5.,4.844240569,1.095445115,2.424366106,3.273268354,3.779644731,4.062019202,.4948716594,1.309307341,1.992047682,2.507132682,.2535462764,.7559289460,1.253566341,.1428571429,.4642307659,.8650429586e-1},
			{2.708012801,4.738035416,5.618618828,5.970607017,6.055300709,6.,.9476070828,2.293791533,3.303437362,4.005204947,4.472135954,.3745745885,1.101145788,1.809367160,2.415229459,.1705887719,.5721721350,1.035098339,.8650429586e-1,.3194382825,.4761904762e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.162277659,1.936491673,1.224744872,.8366600264,.6060915266,.4587583064,3.872983346,3.,2.160246899,1.603567452,1.230977710,3.674234614,3.240370349,2.535462764,1.994411836,3.346640106,3.207134903,2.659215782,3.030457633,3.077444275,2.752549839},
			{3.415650256,3.346640106,2.645751311,2.065591118,1.636634177,1.321374945,3.346640106,4.320493799,4.,3.464101616,2.954684202,2.645751311,4.,4.107919181,3.829708432,2.065591118,3.464101616,3.829708432,1.636634177,2.954684202,1.321374945},
			{3.549647870,4.347413025,3.927922025,3.354101966,2.834733546,2.403122968,2.898275350,4.810702356,5.196152424,5.,4.605896842,1.963961013,3.897114318,4.743416490,4.974937185,1.341640786,3.,3.979949748,.9449111824,2.302948420,.6866065625},
			{3.633180424,5.085393248,5.025445453,4.577377083,4.062019202,3.577708763,2.542696624,4.923910840,5.909368405,6.141195788,6.,1.507633636,3.545621043,4.855041564,5.554920600,.9154754167,2.456478316,3.703280400,.5802884574,1.714285714,.3833259389},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.415650256,1.673320053,.8819171036,.5163977793,.3273268354,.2202291575,5.019960159,3.240370349,2.,1.299038106,.8864052607,5.291502622,4.,2.738612788,1.914854215,5.163977793,4.330127020,3.191423693,4.909902528,4.432026303,4.624812309},
			{3.741657386,3.055050464,2.070196679,1.414213562,.9960238407,.7237468643,4.582575696,5.070925527,4.107919181,3.162277660,2.427520781,4.140393357,5.477225575,5.,4.195235392,3.535533905,5.270462768,5.244044240,2.988071523,4.855041564,2.533114026},
			{3.924283374,4.119639374,3.256860892,2.472066162,1.880349512,1.449137675,4.119639374,5.983235509,5.744562647,4.974937185,4.166190450,3.256860892,5.744562647,6.292853088,6.,2.472066162,4.974937185,6.,1.880349512,4.166190450,1.449137675},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.549647870,1.449137675,.6546536711,.3354101966,.1889822365,.1144344270,5.796550700,3.207134903,1.732050808,1.,.6141195788,6.546536711,4.330127020,2.635231383,1.658312395,6.708203931,5.,3.316624790,6.614378278,5.373546315,6.408327913},
			{3.924283374,2.746426250,1.628430445,.9888264647,.6267831705,.4140393357,5.492852500,5.318431565,3.829708432,2.653299832,1.851640200,5.428101485,6.382847385,5.244044240,4.,4.944132325,6.633249580,6.,4.387482194,6.480740700,3.864367134},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.633180424,1.271348312,.5025445453,.2288688541,.1160576915,.6388765648e-1,6.356741560,3.077444275,1.477342101,.7676494735,.4285714286,7.538168182,4.432026303,2.427520781,1.388730150,8.010409897,5.373546315,3.240370350,8.124038403,6.,8.049844717},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,.0319379183519753,.000465578153648717,5.86464816800813e-006,6.8546484667622e-008,7.65361974719019e-010,3.43216369764802,.122554596919714,.00244089433681864,3.89785419960172e-005,5.51740069821357e-007,5.37669458746301,.262307293803493,.00662303395559898,.000128085303901421,7.27823584467216,.450140785461579,.0137645079246526,9.14177937313726,.68472894504081,10.9691712594614},
			{1.414213562,.166073783576779,.0125887190510939,.000824565926330002,5.01144631725624e-005,2.90963994598978e-006,3.29802783242322,.612365059546825,.0634197097850823,.00526617557576491,.000387613365377294,4.96464287844228,1.2594400120878,.165355473459674,.0166285878943821,6.45780747384476,2.07683410541886,.330223979417038,7.79428043908953,3.03569721712443,8.98680735137037},
			{1.414213562,.398540259080393,.0724976248699598,.0113956405123916,.00166206328995312,.000231576690958128,3.06556135691961,1.36595759702526,.339486246269225,.067649468233801,.0119492023316399,4.2894282422279,2.61131941894657,.822758173894117,.198554676115027,5.18623489409549,4.0025799484364,1.5272774292324,5.81833621896758,5.43816471058122,6.23568135536041},
			{1.414213562,.714786270724957,.23320173943763,.0657431866024243,.0171974308138302,.00429748986298374,2.74931534527504,2.19713005267825,.979365349166059,.350018452610409,.110884150015163,3.45007465931297,3.76697803717186,2.12867470143779,.921342735777177,3.7410698100085,5.17830525859385,3.543802987655,3.76406389114815,6.30978459469567,3.6179005979755},
			{1.414213562,1.09494796295235,.547225517982271,.236321146004057,.0946962283839188,.0362494356624709,2.36915365304765,2.90029193047496,1.98037740107218,1.08420509999131,.526147189418459,2.56192161158569,4.28496925234002,3.70920808974538,2.4592936780291,2.39387697960251,5.07587527588423,5.32120685033247,2.07554270972356,5.32974583920556,1.71909544841957},
			{1.414213562,1.51514018159912,1.04781596872157,.626153181688297,.347192252628355,.183906859961414,1.94896143440088,3.30149822057238,3.11944010216456,2.36319452013471,1.58691847335237,1.73374759666501,4.0126111958998,4.80639993236566,4.40969381865368,1.33269841309056,3.91021223386622,5.67229573503973,.95054294779362,3.37758202463881,.647664476709677},
			{1.414213562,1.94896143440088,1.73374759666501,1.33269841309056,.95054294779362,.647664476709677,1.51514018159912,3.30149822057238,4.0126111958998,3.91021223386622,3.37758202463881,1.04781596872157,3.11944010216456,4.80639993236566,5.67229573503973,.626153181688297,2.36319452013471,4.40969381865368,.347192252628355,1.58691847335237,.183906859961414},
			{1.414213562,2.36915365304765,2.56192161158569,2.39387697960251,2.07554270972356,1.71909544841957,1.09494796295235,2.90029193047496,4.28496925234002,5.07587527588423,5.32974583920556,.547225517982271,1.98037740107218,3.70920808974538,5.32120685033247,.236321146004057,1.08420509999131,2.4592936780291,.0946962283839188,.526147189418459,.0362494356624709},
			{1.414213562,2.74931534527504,3.45007465931297,3.7410698100085,3.76406389114815,3.6179005979755,.714786270724957,2.19713005267825,3.76697803717186,5.17830525859385,6.30978459469567,.23320173943763,.979365349166059,2.12867470143779,3.543802987655,.0657431866024243,.350018452610409,.921342735777177,.0171974308138302,.110884150015163,.00429748986298374},
			{1.414213562,3.06556135691961,4.2894282422279,5.18623489409549,5.81833621896758,6.23568135536041,.398540259080393,1.36595759702526,2.61131941894657,4.0025799484364,5.43816471058122,.0724976248699598,.339486246269225,.822758173894117,1.5272774292324,.0113956405123916,.067649468233801,.198554676115027,.00166206328995312,.0119492023316399,.000231576690958128},
			{1.414213562,3.29802783242322,4.96464287844228,6.45780747384476,7.79428043908953,8.98680735137037,.166073783576779,.612365059546825,1.2594400120878,2.07683410541886,3.03569721712443,.0125887190510939,.0634197097850823,.165355473459674,.330223979417038,.000824565926330002,.00526617557576491,.0166285878943821,5.01144631725624e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,3.43216369764802,5.37669458746301,7.27823584467216,9.14177937313726,10.9691712594614,.0319379183519753,.122554596919714,.262307293803493,.450140785461579,.68472894504081,.000465578153648717,.00244089433681864,.00662303395559898,.0137645079246526,5.86464816800813e-006,3.89785419960172e-005,.000128085303901421,6.8546484667622e-008,5.51740069821357e-007,7.65361974719019e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,.0319379183519753,.000465578153648717,5.86464816800813e-006,6.8546484667622e-008,7.65361974719019e-010,3.43216369764802,.122554596919714,.00244089433681864,3.89785419960172e-005,5.51740069821357e-007,5.37669458746301,.262307293803493,.00662303395559898,.000128085303901421,7.27823584467216,.450140785461579,.0137645079246526,9.14177937313726,.68472894504081,10.9691712594614},
			{1.414213562,.166073783576779,.0125887190510939,.000824565926330002,5.01144631725624e-005,2.90963994598978e-006,3.29802783242322,.612365059546825,.0634197097850823,.00526617557576491,.000387613365377294,4.96464287844228,1.2594400120878,.165355473459674,.0166285878943821,6.45780747384476,2.07683410541886,.330223979417038,7.79428043908953,3.03569721712443,8.98680735137037},
			{1.414213562,.398540259080393,.0724976248699598,.0113956405123916,.00166206328995312,.000231576690958128,3.06556135691961,1.36595759702526,.339486246269225,.067649468233801,.0119492023316399,4.2894282422279,2.61131941894657,.822758173894117,.198554676115027,5.18623489409549,4.0025799484364,1.5272774292324,5.81833621896758,5.43816471058122,6.23568135536041},
			{1.414213562,.714786270724957,.23320173943763,.0657431866024243,.0171974308138302,.00429748986298374,2.74931534527504,2.19713005267825,.979365349166059,.350018452610409,.110884150015163,3.45007465931297,3.76697803717186,2.12867470143779,.921342735777177,3.7410698100085,5.17830525859385,3.543802987655,3.76406389114815,6.30978459469567,3.6179005979755},
			{1.414213562,1.09494796295235,.547225517982271,.236321146004057,.0946962283839188,.0362494356624709,2.36915365304765,2.90029193047496,1.98037740107218,1.08420509999131,.526147189418459,2.56192161158569,4.28496925234002,3.70920808974538,2.4592936780291,2.39387697960251,5.07587527588423,5.32120685033247,2.07554270972356,5.32974583920556,1.71909544841957},
			{1.414213562,1.51514018159912,1.04781596872157,.626153181688297,.347192252628355,.183906859961414,1.94896143440088,3.30149822057238,3.11944010216456,2.36319452013471,1.58691847335237,1.73374759666501,4.0126111958998,4.80639993236566,4.40969381865368,1.33269841309056,3.91021223386622,5.67229573503973,.95054294779362,3.37758202463881,.647664476709677},
			{1.414213562,1.94896143440088,1.73374759666501,1.33269841309056,.95054294779362,.647664476709677,1.51514018159912,3.30149822057238,4.0126111958998,3.91021223386622,3.37758202463881,1.04781596872157,3.11944010216456,4.80639993236566,5.67229573503973,.626153181688297,2.36319452013471,4.40969381865368,.347192252628355,1.58691847335237,.183906859961414},
			{1.414213562,2.36915365304765,2.56192161158569,2.39387697960251,2.07554270972356,1.71909544841957,1.09494796295235,2.90029193047496,4.28496925234002,5.07587527588423,5.32974583920556,.547225517982271,1.98037740107218,3.70920808974538,5.32120685033247,.236321146004057,1.08420509999131,2.4592936780291,.0946962283839188,.526147189418459,.0362494356624709},
			{1.414213562,2.74931534527504,3.45007465931297,3.7410698100085,3.76406389114815,3.6179005979755,.714786270724957,2.19713005267825,3.76697803717186,5.17830525859385,6.30978459469567,.23320173943763,.979365349166059,2.12867470143779,3.543802987655,.0657431866024243,.350018452610409,.921342735777177,.0171974308138302,.110884150015163,.00429748986298374},
			{1.414213562,3.06556135691961,4.2894282422279,5.18623489409549,5.81833621896758,6.23568135536041,.398540259080393,1.36595759702526,2.61131941894657,4.0025799484364,5.43816471058122,.0724976248699598,.339486246269225,.822758173894117,1.5272774292324,.0113956405123916,.067649468233801,.198554676115027,.00166206328995312,.0119492023316399,.000231576690958128},
			{1.414213562,3.29802783242322,4.96464287844228,6.45780747384476,7.79428043908953,8.98680735137037,.166073783576779,.612365059546825,1.2594400120878,2.07683410541886,3.03569721712443,.0125887190510939,.0634197097850823,.165355473459674,.330223979417038,.000824565926330002,.00526617557576491,.0166285878943821,5.01144631725624e-005,.000387613365377294,2.90963994598978e-006},
			{1.414213562,3.43216369764802,5.37669458746301,7.27823584467216,9.14177937313726,10.9691712594614,.0319379183519753,.122554596919714,.262307293803493,.450140785461579,.68472894504081,.000465578153648717,.00244089433681864,.00662303395559898,.0137645079246526,5.86464816800813e-006,3.89785419960172e-005,.000128085303901421,6.8546484667622e-008,5.51740069821357e-007,7.65361974719019e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,.0319379183519753,0.,0.,0.,0.,.000465578153648717,0.,0.,0.,5.86464816800813e-006,0.,0.,6.8546484667622e-008,0.,7.65361974719019e-010},
			{1.414213562,0.,0.,0.,0.,0.,.166073783576779,0.,0.,0.,0.,.0125887190510939,0.,0.,0.,.000824565926330002,0.,0.,5.01144631725624e-005,0.,2.90963994598978e-006},
			{1.414213562,0.,0.,0.,0.,0.,.398540259080393,0.,0.,0.,0.,.0724976248699598,0.,0.,0.,.0113956405123916,0.,0.,.00166206328995312,0.,.000231576690958128},
			{1.414213562,0.,0.,0.,0.,0.,.714786270724957,0.,0.,0.,0.,.23320173943763,0.,0.,0.,.0657431866024243,0.,0.,.0171974308138302,0.,.00429748986298374},
			{1.414213562,0.,0.,0.,0.,0.,1.09494796295235,0.,0.,0.,0.,.547225517982271,0.,0.,0.,.236321146004057,0.,0.,.0946962283839188,0.,.0362494356624709},
			{1.414213562,0.,0.,0.,0.,0.,1.51514018159912,0.,0.,0.,0.,1.04781596872157,0.,0.,0.,.626153181688297,0.,0.,.347192252628355,0.,.183906859961414},
			{1.414213562,0.,0.,0.,0.,0.,1.94896143440088,0.,0.,0.,0.,1.73374759666501,0.,0.,0.,1.33269841309056,0.,0.,.95054294779362,0.,.647664476709677},
			{1.414213562,0.,0.,0.,0.,0.,2.36915365304765,0.,0.,0.,0.,2.56192161158569,0.,0.,0.,2.39387697960251,0.,0.,2.07554270972356,0.,1.71909544841957},
			{1.414213562,0.,0.,0.,0.,0.,2.74931534527504,0.,0.,0.,0.,3.45007465931297,0.,0.,0.,3.7410698100085,0.,0.,3.76406389114815,0.,3.6179005979755},
			{1.414213562,0.,0.,0.,0.,0.,3.06556135691961,0.,0.,0.,0.,4.2894282422279,0.,0.,0.,5.18623489409549,0.,0.,5.81833621896758,0.,6.23568135536041},
			{1.414213562,0.,0.,0.,0.,0.,3.29802783242322,0.,0.,0.,0.,4.96464287844228,0.,0.,0.,6.45780747384476,0.,0.,7.79428043908953,0.,8.98680735137037},
			{1.414213562,0.,0.,0.,0.,0.,3.43216369764802,0.,0.,0.,0.,5.37669458746301,0.,0.,0.,7.27823584467216,0.,0.,9.14177937313726,0.,10.9691712594614}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,.0319379183519753,0.,0.,0.,0.,.000465578153648717,0.,0.,0.,5.86464816800813e-006,0.,0.,6.8546484667622e-008,0.,7.65361974719019e-010},
			{1.414213562,0.,0.,0.,0.,0.,.166073783576779,0.,0.,0.,0.,.0125887190510939,0.,0.,0.,.000824565926330002,0.,0.,5.01144631725624e-005,0.,2.90963994598978e-006},
			{1.414213562,0.,0.,0.,0.,0.,.398540259080393,0.,0.,0.,0.,.0724976248699598,0.,0.,0.,.0113956405123916,0.,0.,.00166206328995312,0.,.000231576690958128},
			{1.414213562,0.,0.,0.,0.,0.,.714786270724957,0.,0.,0.,0.,.23320173943763,0.,0.,0.,.0657431866024243,0.,0.,.0171974308138302,0.,.00429748986298374},
			{1.414213562,0.,0.,0.,0.,0.,1.09494796295235,0.,0.,0.,0.,.547225517982271,0.,0.,0.,.236321146004057,0.,0.,.0946962283839188,0.,.0362494356624709},
			{1.414213562,0.,0.,0.,0.,0.,1.51514018159912,0.,0.,0.,0.,1.04781596872157,0.,0.,0.,.626153181688297,0.,0.,.347192252628355,0.,.183906859961414},
			{1.414213562,0.,0.,0.,0.,0.,1.94896143440088,0.,0.,0.,0.,1.73374759666501,0.,0.,0.,1.33269841309056,0.,0.,.95054294779362,0.,.647664476709677},
			{1.414213562,0.,0.,0.,0.,0.,2.36915365304765,0.,0.,0.,0.,2.56192161158569,0.,0.,0.,2.39387697960251,0.,0.,2.07554270972356,0.,1.71909544841957},
			{1.414213562,0.,0.,0.,0.,0.,2.74931534527504,0.,0.,0.,0.,3.45007465931297,0.,0.,0.,3.7410698100085,0.,0.,3.76406389114815,0.,3.6179005979755},
			{1.414213562,0.,0.,0.,0.,0.,3.06556135691961,0.,0.,0.,0.,4.2894282422279,0.,0.,0.,5.18623489409549,0.,0.,5.81833621896758,0.,6.23568135536041},
			{1.414213562,0.,0.,0.,0.,0.,3.29802783242322,0.,0.,0.,0.,4.96464287844228,0.,0.,0.,6.45780747384476,0.,0.,7.79428043908953,0.,8.98680735137037},
			{1.414213562,0.,0.,0.,0.,0.,3.43216369764802,0.,0.,0.,0.,5.37669458746301,0.,0.,0.,7.27823584467216,0.,0.,9.14177937313726,0.,10.9691712594614}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,3.43216369764802,5.37669458746301,7.27823584467216,9.14177937313726,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.29802783242322,4.96464287844228,6.45780747384476,7.79428043908953,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.06556135691961,4.2894282422279,5.18623489409549,5.81833621896758,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.74931534527504,3.45007465931297,3.7410698100085,3.76406389114815,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.36915365304765,2.56192161158569,2.39387697960251,2.07554270972356,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94896143440088,1.73374759666501,1.33269841309056,.95054294779362,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.51514018159912,1.04781596872157,.626153181688297,.347192252628355,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.09494796295235,.547225517982271,.236321146004057,.0946962283839188,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.714786270724957,.23320173943763,.0657431866024243,.0171974308138302,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.398540259080393,.0724976248699598,.0113956405123916,.00166206328995312,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.166073783576779,.0125887190510939,.000824565926330002,5.01144631725624e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319379183519753,.000465578153648717,5.86464816800813e-006,6.8546484667622e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,3.43216369764802,5.37669458746301,7.27823584467216,9.14177937313726,10.9691712594614,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.29802783242322,4.96464287844228,6.45780747384476,7.79428043908953,8.98680735137037,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.06556135691961,4.2894282422279,5.18623489409549,5.81833621896758,6.23568135536041,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.74931534527504,3.45007465931297,3.7410698100085,3.76406389114815,3.6179005979755,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,2.36915365304765,2.56192161158569,2.39387697960251,2.07554270972356,1.71909544841957,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.94896143440088,1.73374759666501,1.33269841309056,.95054294779362,.647664476709677,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.51514018159912,1.04781596872157,.626153181688297,.347192252628355,.183906859961414,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,1.09494796295235,.547225517982271,.236321146004057,.0946962283839188,.0362494356624709,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.714786270724957,.23320173943763,.0657431866024243,.0171974308138302,.00429748986298374,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.398540259080393,.0724976248699598,.0113956405123916,.00166206328995312,.000231576690958128,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.166073783576779,.0125887190510939,.000824565926330002,5.01144631725624e-005,2.90963994598978e-006,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,.0319379183519753,.000465578153648717,5.86464816800813e-006,6.8546484667622e-008,7.65361974719019e-010,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0471753389703784,-.106939314428477,-.16007833916176,-.203167424021368,-.233492536723722,-.249147045798746,-.249147045798746,-.233492536723722,-.203167424021368,-.16007833916176,-.106939314428477,-.0471753389703784},
			{-.00106538514687764,-.0125580867190646,-.0451117599514472,-.102686955615835,-.180780459320273,-.266927648101013,-.343355485018646,-.391157114392381,-.394969565584768,-.346998489861205,-.249388667069849,-.114490123771661},
			{-1.55307570187141e-005,-.000951927643970909,-.00820618588879868,-.0335020098285882,-.0903492073046712,-.184597475250337,-.305440495868931,-.422983905496475,-.49564139368943,-.485531016712943,-.375413954291232,-.179355789242482},
			{-1.95632945798191e-007,-6.2351625798004e-005,-.00128990080067243,-.00944473608571088,-.0390176033630978,-.110311638598761,-.234786231237538,-.395239038357481,-.537446066425504,-.587042784922414,-.488323349567532,-.242787406459338},
			{-2.28657377825139e-009,-3.78953113999148e-006,-.000188133064231669,-.00247060119508884,-.0156347408665733,-.0611660970770162,-.1674605403216,-.342680727393616,-.540749420569421,-.658591900934981,-.589384113245925,-.30495149536164},
			{-2.55309463461056e-011,-2.20019740479395e-007,-2.62127397542194e-005,-.000617381962823848,-.00598493248162457,-.0323995272485823,-.114101360142285,-.283829803146731,-.519751446470867,-.70583223156883,-.67956003419324,-.365909637710141},
			{-.114490123771661,-.249388667069849,-.346998489861205,-.394969565584768,-.391157114392381,-.343355485018646,-.266927648101013,-.180780459320273,-.102686955615835,-.0451117599514472,-.0125580867190646,-.00106538514687764},
			{-.00408817649919769,-.0463055237009092,-.15461612676984,-.315642039364805,-.478850251422369,-.581636713176573,-.581636713176573,-.478850251422369,-.315642039364805,-.15461612676984,-.0463055237009092,-.00408817649919769},
			{-8.14233583693626e-005,-.00479564081714783,-.0384272898398098,-.140696667326168,-.326968539425373,-.549562885319267,-.706916021549165,-.707466232023107,-.541168069885518,-.295581718785172,-.0952357232347048,-.00875004725280186},
			{-1.30024628505366e-006,-.000398215107369438,-.00765741102002616,-.0502840230429645,-.179006767997741,-.416332404701323,-.688876031299047,-.838048103539602,-.743920839043482,-.453061947209493,-.157045032847724,-.0150157972585881},
			{-1.8404946397784e-008,-2.93103592333472e-005,-.00135255983533517,-.0159297348836293,-.086869087656615,-.279573085688301,-.595040668219595,-.87996318863769,-.906470363461013,-.615559345412792,-.229551877993115,-.0228411895742278},
			{-.179355789242482,-.375413954291232,-.485531016712943,-.49564139368943,-.422983905496475,-.305440495868931,-.184597475250337,-.0903492073046712,-.0335020098285882,-.00820618588879868,-.000951927643970909,-1.55307570187141e-005},
			{-.00875004725280186,-.0952357232347048,-.295581718785172,-.541168069885518,-.707466232023107,-.706916021549165,-.549562885319267,-.326968539425373,-.140696667326168,-.0384272898398098,-.00479564081714783,-8.14233583693626e-005},
			{-.000220931180479551,-.0125037698934495,-.0931300374131473,-.30580767082361,-.612405671197932,-.846760613546111,-.846760613546111,-.612405671197932,-.30580767082361,-.0931300374131473,-.0125037698934495,-.000220931180479551},
			{-4.272669834208e-006,-.00125741248435335,-.0224748960288392,-.132361076996858,-.406039607141487,-.776871483017825,-.999308564498337,-.878553366086448,-.509106504991363,-.172876318507379,-.0249707165147866,-.000459156484007277},
			{-.242787406459338,-.488323349567532,-.587042784922414,-.537446066425504,-.395239038357481,-.234786231237538,-.110311638598761,-.0390176033630978,-.00944473608571088,-.00128990080067243,-6.2351625798004e-005,-1.95632945798191e-007},
			{-.0150157972585881,-.157045032847724,-.453061947209493,-.743920839043482,-.838048103539602,-.688876031299047,-.416332404701323,-.179006767997741,-.0502840230429645,-.00765741102002616,-.000398215107369438,-1.30024628505366e-006},
			{-.000459156484007277,-.0249707165147866,-.172876318507379,-.509106504991363,-.878553366086448,-.999308564498337,-.776871483017825,-.406039607141487,-.132361076996858,-.0224748960288392,-.00125741248435335,-4.272669834208e-006},
			{-.30495149536164,-.589384113245925,-.658591900934981,-.540749420569421,-.342680727393616,-.1674605403216,-.0611660970770162,-.0156347408665733,-.00247060119508884,-.000188133064231669,-3.78953113999148e-006,-2.28657377825139e-009},
			{-.0228411895742278,-.229551877993115,-.615559345412792,-.906470363461013,-.87996318863769,-.595040668219595,-.279573085688301,-.086869087656615,-.0159297348836293,-.00135255983533517,-2.93103592333472e-005,-1.8404946397784e-008},
			{-.365909637710141,-.67956003419324,-.70583223156883,-.519751446470867,-.283829803146731,-.114101360142285,-.0323995272485823,-.00598493248162457,-.000617381962823848,-2.62127397542194e-005,-2.20019740479395e-007,-2.55309463461056e-011}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0471753389703784,-.106939314428477,-.16007833916176,-.203167424021368,-.233492536723722,-.249147045798746,-.249147045798746,-.233492536723722,-.203167424021368,-.16007833916176,-.106939314428477,-.0471753389703784},
			{-.00106538514687764,-.0125580867190646,-.0451117599514472,-.102686955615835,-.180780459320273,-.266927648101013,-.343355485018646,-.391157114392381,-.394969565584768,-.346998489861205,-.249388667069849,-.114490123771661},
			{-1.55307570187141e-005,-.000951927643970909,-.00820618588879868,-.0335020098285882,-.0903492073046712,-.184597475250337,-.305440495868931,-.422983905496475,-.49564139368943,-.485531016712943,-.375413954291232,-.179355789242482},
			{-1.95632945798191e-007,-6.2351625798004e-005,-.00128990080067243,-.00944473608571088,-.0390176033630978,-.110311638598761,-.234786231237538,-.395239038357481,-.537446066425504,-.587042784922414,-.488323349567532,-.242787406459338},
			{-2.28657377825139e-009,-3.78953113999148e-006,-.000188133064231669,-.00247060119508884,-.0156347408665733,-.0611660970770162,-.1674605403216,-.342680727393616,-.540749420569421,-.658591900934981,-.589384113245925,-.30495149536164},
			{-2.55309463461056e-011,-2.20019740479395e-007,-2.62127397542194e-005,-.000617381962823848,-.00598493248162457,-.0323995272485823,-.114101360142285,-.283829803146731,-.519751446470867,-.70583223156883,-.67956003419324,-.365909637710141},
			{-.114490123771661,-.249388667069849,-.346998489861205,-.394969565584768,-.391157114392381,-.343355485018646,-.266927648101013,-.180780459320273,-.102686955615835,-.0451117599514472,-.0125580867190646,-.00106538514687764},
			{-.00408817649919769,-.0463055237009092,-.15461612676984,-.315642039364805,-.478850251422369,-.581636713176573,-.581636713176573,-.478850251422369,-.315642039364805,-.15461612676984,-.0463055237009092,-.00408817649919769},
			{-8.14233583693626e-005,-.00479564081714783,-.0384272898398098,-.140696667326168,-.326968539425373,-.549562885319267,-.706916021549165,-.707466232023107,-.541168069885518,-.295581718785172,-.0952357232347048,-.00875004725280186},
			{-1.30024628505366e-006,-.000398215107369438,-.00765741102002616,-.0502840230429645,-.179006767997741,-.416332404701323,-.688876031299047,-.838048103539602,-.743920839043482,-.453061947209493,-.157045032847724,-.0150157972585881},
			{-1.8404946397784e-008,-2.93103592333472e-005,-.00135255983533517,-.0159297348836293,-.086869087656615,-.279573085688301,-.595040668219595,-.87996318863769,-.906470363461013,-.615559345412792,-.229551877993115,-.0228411895742278},
			{-.179355789242482,-.375413954291232,-.485531016712943,-.49564139368943,-.422983905496475,-.305440495868931,-.184597475250337,-.0903492073046712,-.0335020098285882,-.00820618588879868,-.000951927643970909,-1.55307570187141e-005},
			{-.00875004725280186,-.0952357232347048,-.295581718785172,-.541168069885518,-.707466232023107,-.706916021549165,-.549562885319267,-.326968539425373,-.140696667326168,-.0384272898398098,-.00479564081714783,-8.14233583693626e-005},
			{-.000220931180479551,-.0125037698934495,-.0931300374131473,-.30580767082361,-.612405671197932,-.846760613546111,-.846760613546111,-.612405671197932,-.30580767082361,-.0931300374131473,-.0125037698934495,-.000220931180479551},
			{-4.272669834208e-006,-.00125741248435335,-.0224748960288392,-.132361076996858,-.406039607141487,-.776871483017825,-.999308564498337,-.878553366086448,-.509106504991363,-.172876318507379,-.0249707165147866,-.000459156484007277},
			{-.242787406459338,-.488323349567532,-.587042784922414,-.537446066425504,-.395239038357481,-.234786231237538,-.110311638598761,-.0390176033630978,-.00944473608571088,-.00128990080067243,-6.2351625798004e-005,-1.95632945798191e-007},
			{-.0150157972585881,-.157045032847724,-.453061947209493,-.743920839043482,-.838048103539602,-.688876031299047,-.416332404701323,-.179006767997741,-.0502840230429645,-.00765741102002616,-.000398215107369438,-1.30024628505366e-006},
			{-.000459156484007277,-.0249707165147866,-.172876318507379,-.509106504991363,-.878553366086448,-.999308564498337,-.776871483017825,-.406039607141487,-.132361076996858,-.0224748960288392,-.00125741248435335,-4.272669834208e-006},
			{-.30495149536164,-.589384113245925,-.658591900934981,-.540749420569421,-.342680727393616,-.1674605403216,-.0611660970770162,-.0156347408665733,-.00247060119508884,-.000188133064231669,-3.78953113999148e-006,-2.28657377825139e-009},
			{-.0228411895742278,-.229551877993115,-.615559345412792,-.906470363461013,-.87996318863769,-.595040668219595,-.279573085688301,-.086869087656615,-.0159297348836293,-.00135255983533517,-2.93103592333472e-005,-1.8404946397784e-008},
			{-.365909637710141,-.67956003419324,-.70583223156883,-.519751446470867,-.283829803146731,-.114101360142285,-.0323995272485823,-.00598493248162457,-.000617381962823848,-2.62127397542194e-005,-2.20019740479395e-007,-2.55309463461056e-011}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.000753341062131352,-.008879908280122,-.0318988313813435,-.072610642674515,-.127831088725108,-.188746350108197,-.242788991878333,-.276589848169184,-.279285658260975,-.245364985307081,-.176344417682688,-.0809567429191864},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.09819036077906e-005,-.000673114492428343,-.00580264969117775,-.0236894983394228,-.0638865371768165,-.130530126573866,-.215979045934882,-.299094787988234,-.350471390607008,-.343322274484695,-.265457752901419,-.126823694851883},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.38333382633896e-007,-4.40892574314062e-005,-.000912097603454061,-.00667843693448535,-.0275896119309721,-.0780021077175622,-.166018936281095,-.279476204285956,-.380031758191769,-.415101934174787,-.345296751982027,-.171676621539384},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.61685182471146e-009,-2.67960316731249e-006,-.000133030165518713,-.00174697885911579,-.011055431291765,-.0432509620332832,-.118412483673806,-.24231186618589,-.382367582308209,-.465694799308523,-.416757503309762,-.215633270360082},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.80531052962039e-011,-1.55577450528929e-007,-1.85352060385767e-005,-.000436554972610175,-.00423198634381684,-.0229099254307549,-.0806818455205033,-.200697978560844,-.367519772428025,-.499098757454024,-.480521508528172,-.258737186194613}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-.000753341062131352,-.008879908280122,-.0318988313813435,-.072610642674515,-.127831088725108,-.188746350108197,-.242788991878333,-.276589848169184,-.279285658260975,-.245364985307081,-.176344417682688,-.0809567429191864},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.09819036077906e-005,-.000673114492428343,-.00580264969117775,-.0236894983394228,-.0638865371768165,-.130530126573866,-.215979045934882,-.299094787988234,-.350471390607008,-.343322274484695,-.265457752901419,-.126823694851883},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.38333382633896e-007,-4.40892574314062e-005,-.000912097603454061,-.00667843693448535,-.0275896119309721,-.0780021077175622,-.166018936281095,-.279476204285956,-.380031758191769,-.415101934174787,-.345296751982027,-.171676621539384},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.61685182471146e-009,-2.67960316731249e-006,-.000133030165518713,-.00174697885911579,-.011055431291765,-.0432509620332832,-.118412483673806,-.24231186618589,-.382367582308209,-.465694799308523,-.416757503309762,-.215633270360082},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-1.80531052962039e-011,-1.55577450528929e-007,-1.85352060385767e-005,-.000436554972610175,-.00423198634381684,-.0229099254307549,-.0806818455205033,-.200697978560844,-.367519772428025,-.499098757454024,-.480521508528172,-.258737186194613}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{-.0809567429191864,-.176344417682688,-.245364985307081,-.279285658260975,-.276589848169184,-.242788991878333,-.188746350108197,-.127831088725108,-.072610642674515,-.0318988313813435,-.008879908280122,-.000753341062131352},
			{-.126823694851883,-.265457752901419,-.343322274484695,-.350471390607008,-.299094787988234,-.215979045934882,-.130530126573866,-.0638865371768165,-.0236894983394228,-.00580264969117775,-.000673114492428343,-1.09819036077906e-005},
			{-.171676621539384,-.345296751982027,-.415101934174787,-.380031758191769,-.279476204285956,-.166018936281095,-.0780021077175622,-.0275896119309721,-.00667843693448535,-.000912097603454061,-4.40892574314062e-005,-1.38333382633896e-007},
			{-.215633270360082,-.416757503309762,-.465694799308523,-.382367582308209,-.24231186618589,-.118412483673806,-.0432509620332832,-.011055431291765,-.00174697885911579,-.000133030165518713,-2.67960316731249e-006,-1.61685182471146e-009},
			{-.258737186194613,-.480521508528172,-.499098757454024,-.367519772428025,-.200697978560844,-.0806818455205033,-.0229099254307549,-.00423198634381684,-.000436554972610175,-1.85352060385767e-005,-1.55577450528929e-007,-1.80531052962039e-011},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{-.0809567429191864,-.176344417682688,-.245364985307081,-.279285658260975,-.276589848169184,-.242788991878333,-.188746350108197,-.127831088725108,-.072610642674515,-.0318988313813435,-.008879908280122,-.000753341062131352},
			{-.126823694851883,-.265457752901419,-.343322274484695,-.350471390607008,-.299094787988234,-.215979045934882,-.130530126573866,-.0638865371768165,-.0236894983394228,-.00580264969117775,-.000673114492428343,-1.09819036077906e-005},
			{-.171676621539384,-.345296751982027,-.415101934174787,-.380031758191769,-.279476204285956,-.166018936281095,-.0780021077175622,-.0275896119309721,-.00667843693448535,-.000912097603454061,-4.40892574314062e-005,-1.38333382633896e-007},
			{-.215633270360082,-.416757503309762,-.465694799308523,-.382367582308209,-.24231186618589,-.118412483673806,-.0432509620332832,-.011055431291765,-.00174697885911579,-.000133030165518713,-2.67960316731249e-006,-1.61685182471146e-009},
			{-.258737186194613,-.480521508528172,-.499098757454024,-.367519772428025,-.200697978560844,-.0806818455205033,-.0229099254307549,-.00423198634381684,-.000436554972610175,-1.85352060385767e-005,-1.55577450528929e-007,-1.80531052962039e-011},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,3.354101966,0.,4.062019202},
			{0.,-.5000000000,0.,0.,0.,0.,-.5000000000,-1.936491673,0.,0.,0.,-1.581138830,-4.183300132,0.,0.,-3.240370350,-7.245688375,0.,-5.477225576,-11.12429773,-8.291561975},
			{0.,0.,.5000000000,0.,0.,0.,0.,1.224744872,2.645751311,0.,0.,.5000000000,5.291502622,7.245688375,0.,2.049390154,13.74772709,15.19868416,5.196152424,28.14249455,10.48808848},
			{0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,-3.354101966,0.,0.,-1.936491673,-10.60660172,-11.12429773,-.5000000000,-10.06230590,-33.37289319,-2.535462764,-30.89729531,-7.676494735},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,2.645751311,4.062019202,0.,0.,4.183300132,17.54992877,0.,2.645751311,26.32489316,.5000000000,16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-3.354101966,0.,0.,0.,-7.245688375,0.,0.,-7.245688375,0.,-3.354101966,-.5000000000},
			{0.,.5000000000,0.,0.,0.,0.,-.5000000000,1.936491673,0.,0.,0.,-1.581138830,4.183300132,0.,0.,-3.240370350,7.245688375,0.,-5.477225576,11.12429773,-8.291561975},
			{0.,0.,-.4082482906,0.,0.,0.,0.,0.,-2.160246899,0.,0.,.4082482906,0.,-5.916079782,0.,1.673320053,0.,-12.40967365,4.242640686,0.,8.563488387},
			{0.,0.,0.,.3872983346,0.,0.,0.,0.,.5000000000,2.598076212,0.,0.,-.5000000000,2.738612788,8.616843970,-.3872983346,-2.598076212,8.616843970,-1.963961012,-7.977647347,-5.946187254},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,-1.,-3.070597895,0.,0.,0.,-6.633249580,0.,1.,0.,.3779644731,6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,1.500000000,0.,0.,0.,1.080123450,0.,0.,-1.080123450,0.,-1.500000000,-.3726779962},
			{0.,0.,.5000000000,0.,0.,0.,0.,-1.224744872,2.645751311,0.,0.,.5000000000,-5.291502622,7.245688375,0.,2.049390154,-13.74772709,15.19868416,5.196152424,-28.14249455,10.48808848},
			{0.,0.,0.,-.3872983346,0.,0.,0.,0.,.5000000000,-2.598076212,0.,0.,.5000000000,2.738612788,-8.616843970,-.3872983346,2.598076212,8.616843970,-1.963961012,7.977647347,-5.946187254},
			{0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,2.913024938,0.,0.,-1.,0.,0.,0.,-6.292853088,.3585685828,0.,2.171240594},
			{0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,-.4629100500,0.,0.,0.,1.,0.,0.,1.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,-1.936491673,3.354101966,0.,0.,1.936491673,-10.60660172,11.12429773,-.5000000000,10.06230590,-33.37289319,-2.535462764,30.89729531,-7.676494735},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,1.,-3.070597895,0.,0.,0.,6.633249580,0.,-1.,0.,.3779644731,-6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,.3450327798,0.,0.,0.,0.,-.4629100500,0.,0.,0.,-1.,0.,0.,1.,0.,.4629100500,-.3450327798},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,-2.645751311,4.062019202,0.,0.,4.183300132,-17.54992877,0.,-2.645751311,26.32489316,.5000000000,-16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,1.500000000,0.,0.,0.,-1.080123450,0.,0.,-1.080123450,0.,1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,-3.354101966,0.,0.,0.,7.245688375,0.,0.,-7.245688375,0.,3.354101966,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,1.224744872,0.,0.,0.,0.,1.936491673,0.,0.,0.,2.645751311,0.,0.,3.354101966,0.,4.062019202},
			{0.,-.5000000000,0.,0.,0.,0.,-.5000000000,-1.936491673,0.,0.,0.,-1.581138830,-4.183300132,0.,0.,-3.240370350,-7.245688375,0.,-5.477225576,-11.12429773,-8.291561975},
			{0.,0.,.5000000000,0.,0.,0.,0.,1.224744872,2.645751311,0.,0.,.5000000000,5.291502622,7.245688375,0.,2.049390154,13.74772709,15.19868416,5.196152424,28.14249455,10.48808848},
			{0.,0.,0.,-.5000000000,0.,0.,0.,0.,-1.936491673,-3.354101966,0.,0.,-1.936491673,-10.60660172,-11.12429773,-.5000000000,-10.06230590,-33.37289319,-2.535462764,-30.89729531,-7.676494735},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,2.645751311,4.062019202,0.,0.,4.183300132,17.54992877,0.,2.645751311,26.32489316,.5000000000,16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-3.354101966,0.,0.,0.,-7.245688375,0.,0.,-7.245688375,0.,-3.354101966,-.5000000000},
			{0.,.5000000000,0.,0.,0.,0.,-.5000000000,1.936491673,0.,0.,0.,-1.581138830,4.183300132,0.,0.,-3.240370350,7.245688375,0.,-5.477225576,11.12429773,-8.291561975},
			{0.,0.,-.4082482906,0.,0.,0.,0.,0.,-2.160246899,0.,0.,.4082482906,0.,-5.916079782,0.,1.673320053,0.,-12.40967365,4.242640686,0.,8.563488387},
			{0.,0.,0.,.3872983346,0.,0.,0.,0.,.5000000000,2.598076212,0.,0.,-.5000000000,2.738612788,8.616843970,-.3872983346,-2.598076212,8.616843970,-1.963961012,-7.977647347,-5.946187254},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,-1.,-3.070597895,0.,0.,0.,-6.633249580,0.,1.,0.,.3779644731,6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,1.500000000,0.,0.,0.,1.080123450,0.,0.,-1.080123450,0.,-1.500000000,-.3726779962},
			{0.,0.,.5000000000,0.,0.,0.,0.,-1.224744872,2.645751311,0.,0.,.5000000000,-5.291502622,7.245688375,0.,2.049390154,-13.74772709,15.19868416,5.196152424,-28.14249455,10.48808848},
			{0.,0.,0.,-.3872983346,0.,0.,0.,0.,.5000000000,-2.598076212,0.,0.,.5000000000,2.738612788,-8.616843970,-.3872983346,2.598076212,8.616843970,-1.963961012,7.977647347,-5.946187254},
			{0.,0.,0.,0.,.3585685828,0.,0.,0.,0.,0.,2.913024938,0.,0.,-1.,0.,0.,0.,-6.292853088,.3585685828,0.,2.171240594},
			{0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,-.4629100500,0.,0.,0.,1.,0.,0.,1.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,-1.936491673,3.354101966,0.,0.,1.936491673,-10.60660172,11.12429773,-.5000000000,10.06230590,-33.37289319,-2.535462764,30.89729531,-7.676494735},
			{0.,0.,0.,0.,-.3779644731,0.,0.,0.,0.,1.,-3.070597895,0.,0.,0.,6.633249580,0.,-1.,0.,.3779644731,-6.141195788,2.288688541},
			{0.,0.,0.,0.,0.,.3450327798,0.,0.,0.,0.,-.4629100500,0.,0.,0.,-1.,0.,0.,1.,0.,.4629100500,-.3450327798},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,-2.645751311,4.062019202,0.,0.,4.183300132,-17.54992877,0.,-2.645751311,26.32489316,.5000000000,-16.24807681,3.027650356},
			{0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,1.500000000,0.,0.,0.,-1.080123450,0.,0.,-1.080123450,0.,1.500000000,-.3726779962},
			{0.,0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,-3.354101966,0.,0.,0.,7.245688375,0.,0.,-7.245688375,0.,3.354101966,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.5000000000,1.224744872,1.936491673,2.645751311,3.354101966,4.062019202,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,-8.291561975,.5000000000,1.936491673,4.183300132,7.245688375,11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,2.049390154,5.196152424,10.48808848,0.,-1.224744872,-5.291502622,-13.74772709,-28.14249455,.5000000000,2.645751311,7.245688375,15.19868416,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5000000000,-2.535462764,-7.676494735,0.,0.,1.936491673,10.06230590,30.89729531,0.,-1.936491673,-10.60660172,-33.37289319,.5000000000,3.354101966,11.12429773,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,3.027650356,0.,0.,0.,-2.645751311,-16.24807681,0.,0.,4.183300132,26.32489316,0.,-2.645751311,-17.54992877,.5000000000,4.062019202,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,3.354101966,0.,0.,0.,-7.245688375,0.,0.,7.245688375,0.,-3.354101966,.5000000000},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,-8.291561975,-.5000000000,-1.936491673,-4.183300132,-7.245688375,-11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.4082482906,1.673320053,4.242640686,8.563488387,0.,0.,0.,0.,0.,-.4082482906,-2.160246899,-5.916079782,-12.40967365,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,-5.946187254,0.,0.,.5000000000,2.598076212,7.977647347,0.,.5000000000,2.738612788,8.616843970,-.3872983346,-2.598076212,-8.616843970,0.,0.,0.},
			{0.,0.,0.,0.,.3779644731,2.288688541,0.,0.,0.,-1.,-6.141195788,0.,0.,0.,0.,0.,1.,6.633249580,-.3779644731,-3.070597895,0.},
			{0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,1.500000000,0.,0.,0.,-1.080123450,0.,0.,-1.080123450,0.,1.500000000,-.3726779962},
			{0.,0.,.5000000000,2.049390154,5.196152424,10.48808848,0.,1.224744872,5.291502622,13.74772709,28.14249455,.5000000000,2.645751311,7.245688375,15.19868416,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,-5.946187254,0.,0.,-.5000000000,-2.598076212,-7.977647347,0.,.5000000000,2.738612788,8.616843970,.3872983346,2.598076212,8.616843970,0.,0.,0.},
			{0.,0.,0.,0.,.3585685828,2.171240594,0.,0.,0.,0.,0.,0.,0.,-1.,-6.292853088,0.,0.,0.,.3585685828,2.913024938,0.},
			{0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,.4629100500,0.,0.,0.,1.,0.,0.,-1.,0.,-.4629100500,.3450327798},
			{0.,0.,0.,-.5000000000,-2.535462764,-7.676494735,0.,0.,-1.936491673,-10.06230590,-30.89729531,0.,-1.936491673,-10.60660172,-33.37289319,-.5000000000,-3.354101966,-11.12429773,0.,0.,0.},
			{0.,0.,0.,0.,.3779644731,2.288688541,0.,0.,0.,1.,6.141195788,0.,0.,0.,0.,0.,-1.,-6.633249580,-.3779644731,-3.070597895,0.},
			{0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,-.4629100500,0.,0.,0.,1.,0.,0.,1.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,0.,.5000000000,3.027650356,0.,0.,0.,2.645751311,16.24807681,0.,0.,4.183300132,26.32489316,0.,2.645751311,17.54992877,.5000000000,4.062019202,0.},
			{0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,-1.500000000,0.,0.,0.,-1.080123450,0.,0.,1.080123450,0.,1.500000000,.3726779962},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-3.354101966,0.,0.,0.,-7.245688375,0.,0.,-7.245688375,0.,-3.354101966,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.5000000000,1.224744872,1.936491673,2.645751311,3.354101966,4.062019202,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,-8.291561975,.5000000000,1.936491673,4.183300132,7.245688375,11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.5000000000,2.049390154,5.196152424,10.48808848,0.,-1.224744872,-5.291502622,-13.74772709,-28.14249455,.5000000000,2.645751311,7.245688375,15.19868416,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5000000000,-2.535462764,-7.676494735,0.,0.,1.936491673,10.06230590,30.89729531,0.,-1.936491673,-10.60660172,-33.37289319,.5000000000,3.354101966,11.12429773,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,3.027650356,0.,0.,0.,-2.645751311,-16.24807681,0.,0.,4.183300132,26.32489316,0.,-2.645751311,-17.54992877,.5000000000,4.062019202,0.},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,3.354101966,0.,0.,0.,-7.245688375,0.,0.,7.245688375,0.,-3.354101966,.5000000000},
			{0.,-.5000000000,-1.581138830,-3.240370350,-5.477225576,-8.291561975,-.5000000000,-1.936491673,-4.183300132,-7.245688375,-11.12429773,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.4082482906,1.673320053,4.242640686,8.563488387,0.,0.,0.,0.,0.,-.4082482906,-2.160246899,-5.916079782,-12.40967365,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,-5.946187254,0.,0.,.5000000000,2.598076212,7.977647347,0.,.5000000000,2.738612788,8.616843970,-.3872983346,-2.598076212,-8.616843970,0.,0.,0.},
			{0.,0.,0.,0.,.3779644731,2.288688541,0.,0.,0.,-1.,-6.141195788,0.,0.,0.,0.,0.,1.,6.633249580,-.3779644731,-3.070597895,0.},
			{0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,1.500000000,0.,0.,0.,-1.080123450,0.,0.,-1.080123450,0.,1.500000000,-.3726779962},
			{0.,0.,.5000000000,2.049390154,5.196152424,10.48808848,0.,1.224744872,5.291502622,13.74772709,28.14249455,.5000000000,2.645751311,7.245688375,15.19868416,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.3872983346,-1.963961012,-5.946187254,0.,0.,-.5000000000,-2.598076212,-7.977647347,0.,.5000000000,2.738612788,8.616843970,.3872983346,2.598076212,8.616843970,0.,0.,0.},
			{0.,0.,0.,0.,.3585685828,2.171240594,0.,0.,0.,0.,0.,0.,0.,-1.,-6.292853088,0.,0.,0.,.3585685828,2.913024938,0.},
			{0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,.4629100500,0.,0.,0.,1.,0.,0.,-1.,0.,-.4629100500,.3450327798},
			{0.,0.,0.,-.5000000000,-2.535462764,-7.676494735,0.,0.,-1.936491673,-10.06230590,-30.89729531,0.,-1.936491673,-10.60660172,-33.37289319,-.5000000000,-3.354101966,-11.12429773,0.,0.,0.},
			{0.,0.,0.,0.,.3779644731,2.288688541,0.,0.,0.,1.,6.141195788,0.,0.,0.,0.,0.,-1.,-6.633249580,-.3779644731,-3.070597895,0.},
			{0.,0.,0.,0.,0.,-.3450327798,0.,0.,0.,0.,-.4629100500,0.,0.,0.,1.,0.,0.,1.,0.,-.4629100500,-.3450327798},
			{0.,0.,0.,0.,.5000000000,3.027650356,0.,0.,0.,2.645751311,16.24807681,0.,0.,4.183300132,26.32489316,0.,2.645751311,17.54992877,.5000000000,4.062019202,0.},
			{0.,0.,0.,0.,0.,-.3726779962,0.,0.,0.,0.,-1.500000000,0.,0.,0.,-1.080123450,0.,0.,1.080123450,0.,1.500000000,.3726779962},
			{0.,0.,0.,0.,0.,-.5000000000,0.,0.,0.,0.,-3.354101966,0.,0.,0.,-7.245688375,0.,0.,-7.245688375,0.,-3.354101966,-.5000000000}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,.2538762001,1.224744872,2.371708244,2.561737692,2.218529919,1.703053324,.9682458365,2.561737692,3.507803799,3.679015663,.6614378278,2.218529919,3.679015663,.4192627457,1.703053324,.2538762001},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5182226234,.5000000000,0.,-1.045825033,-1.811422094,-2.085805824,.7905694150,1.045825033,0.,-1.501951855,.8100925875,1.811422094,1.501951855,.6846531970,2.085805824,.5182226234},
			{0.,0.,.2500000000,.5123475385,.6495190530,.6555055300,0.,-.6123724358,-.6614378278,0.,.8794529546,.2500000000,-.6614378278,-1.811422094,-1.899835520,.5123475385,0.,-1.899835520,.6495190530,.8794529546,.6555055300},
			{0.,0.,0.,-.1250000000,-.3169328455,-.4797809210,0.,0.,.4841229182,.8385254914,.6436936521,0.,-.4841229182,0.,1.390537216,.1250000000,-.8385254914,-1.390537216,.3169328455,-.6436936521,.4797809210},
			{0.,0.,0.,0.,.6250000000e-1,.1892281472,0.,0.,0.,-.3307189139,-.7616286003,0.,0.,.5229125166,.5484352742,0.,-.3307189139,.5484352742,.6250000000e-1,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,.2096313728,0.,0.,0.,-.4528555234,0.,0.,.4528555234,0.,-.2096313728,.3125000000e-1},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5182226234,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-3.476343041,-.7905694150,-3.137475099,-5.728219620,-7.509759277,-.8100925875,-3.622844188,-7.509759277,-.6846531970,-3.476343041,-.5182226234},
			{0.,0.,.2041241452,.4183300132,.5303300858,.5352180242,0.,0.,.5400617248,1.403121520,2.154210993,-.2041241452,-.5400617248,0.,1.551209206,-.4183300132,-1.403121520,-1.551209206,-.5303300858,-2.154210993,-.5352180242},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,-.3716367033,0.,0.,.1250000000,0.,-.4986029592,0.,.1250000000,.6846531969,1.077105496,-.9682458365e-1,0.,1.077105496,-.2454951265,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,.4724555913e-1,.1430430338,0.,0.,0.,-.1250000000,-.1919123684,0.,0.,0.,-.4145780988,0.,.1250000000,.4145780988,-.4724555913e-1,.1919123684,-.1430430338},
			{0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,-.6750771565e-1,0.,0.,-.6750771565e-1,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,.6555055300,0.,.6123724358,1.984313483,3.436931772,4.397264773,.2500000000,1.984313483,5.434266281,9.499177600,.5123475385,3.436931772,9.499177600,.6495190530,4.397264773,.6555055300},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,-.3716367033,0.,0.,-.1250000000,-.6495190530,-1.495808878,0.,.1250000000,0.,-1.077105496,.9682458365e-1,.6495190530,1.077105496,.2454951265,1.495808878,.3716367033},
			{0.,0.,0.,0.,.4482107285e-1,.1357025371,0.,0.,0.,0.,.1820640586,0.,0.,-.1250000000,-.3933033180,0.,0.,-.3933033180,.4482107285e-1,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,.2893187812e-1,0.,0.,0.,.6250000000e-1,0.,0.,-.6250000000e-1,0.,-.2893187812e-1,.2156454873e-1},
			{0.,0.,0.,-.1250000000,-.3169328455,-.4797809210,0.,0.,-.4841229182,-1.677050983,-3.218468261,0.,-.4841229182,-2.651650429,-6.952686082,-.1250000000,-1.677050983,-6.952686082,-.3169328455,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,.4724555913e-1,.1430430338,0.,0.,0.,.1250000000,.5757371053,0.,0.,0.,.4145780988,0.,-.1250000000,-.4145780988,-.4724555913e-1,-.5757371053,-.1430430338},
			{0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,.6250000000e-1,0.,0.,.6250000000e-1,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,.6250000000e-1,.1892281472,0.,0.,0.,.3307189139,1.269381000,0.,0.,.5229125166,2.742176371,0.,.3307189139,2.742176371,.6250000000e-1,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,-.9375000000e-1,0.,0.,0.,-.6750771565e-1,0.,0.,.6750771565e-1,0.,.9375000000e-1,.2329237477e-1},
			{0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,-.2096313728,0.,0.,0.,-.4528555234,0.,0.,-.4528555234,0.,-.2096313728,-.3125000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,.2538762001,1.224744872,2.371708244,2.561737692,2.218529919,1.703053324,.9682458365,2.561737692,3.507803799,3.679015663,.6614378278,2.218529919,3.679015663,.4192627457,1.703053324,.2538762001},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5182226234,.5000000000,0.,-1.045825033,-1.811422094,-2.085805824,.7905694150,1.045825033,0.,-1.501951855,.8100925875,1.811422094,1.501951855,.6846531970,2.085805824,.5182226234},
			{0.,0.,.2500000000,.5123475385,.6495190530,.6555055300,0.,-.6123724358,-.6614378278,0.,.8794529546,.2500000000,-.6614378278,-1.811422094,-1.899835520,.5123475385,0.,-1.899835520,.6495190530,.8794529546,.6555055300},
			{0.,0.,0.,-.1250000000,-.3169328455,-.4797809210,0.,0.,.4841229182,.8385254914,.6436936521,0.,-.4841229182,0.,1.390537216,.1250000000,-.8385254914,-1.390537216,.3169328455,-.6436936521,.4797809210},
			{0.,0.,0.,0.,.6250000000e-1,.1892281472,0.,0.,0.,-.3307189139,-.7616286003,0.,0.,.5229125166,.5484352742,0.,-.3307189139,.5484352742,.6250000000e-1,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,.2096313728,0.,0.,0.,-.4528555234,0.,0.,.4528555234,0.,-.2096313728,.3125000000e-1},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5182226234,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-3.476343041,-.7905694150,-3.137475099,-5.728219620,-7.509759277,-.8100925875,-3.622844188,-7.509759277,-.6846531970,-3.476343041,-.5182226234},
			{0.,0.,.2041241452,.4183300132,.5303300858,.5352180242,0.,0.,.5400617248,1.403121520,2.154210993,-.2041241452,-.5400617248,0.,1.551209206,-.4183300132,-1.403121520,-1.551209206,-.5303300858,-2.154210993,-.5352180242},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,-.3716367033,0.,0.,.1250000000,0.,-.4986029592,0.,.1250000000,.6846531969,1.077105496,-.9682458365e-1,0.,1.077105496,-.2454951265,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,.4724555913e-1,.1430430338,0.,0.,0.,-.1250000000,-.1919123684,0.,0.,0.,-.4145780988,0.,.1250000000,.4145780988,-.4724555913e-1,.1919123684,-.1430430338},
			{0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,-.6750771565e-1,0.,0.,-.6750771565e-1,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,.6555055300,0.,.6123724358,1.984313483,3.436931772,4.397264773,.2500000000,1.984313483,5.434266281,9.499177600,.5123475385,3.436931772,9.499177600,.6495190530,4.397264773,.6555055300},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,-.3716367033,0.,0.,-.1250000000,-.6495190530,-1.495808878,0.,.1250000000,0.,-1.077105496,.9682458365e-1,.6495190530,1.077105496,.2454951265,1.495808878,.3716367033},
			{0.,0.,0.,0.,.4482107285e-1,.1357025371,0.,0.,0.,0.,.1820640586,0.,0.,-.1250000000,-.3933033180,0.,0.,-.3933033180,.4482107285e-1,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,.2893187812e-1,0.,0.,0.,.6250000000e-1,0.,0.,-.6250000000e-1,0.,-.2893187812e-1,.2156454873e-1},
			{0.,0.,0.,-.1250000000,-.3169328455,-.4797809210,0.,0.,-.4841229182,-1.677050983,-3.218468261,0.,-.4841229182,-2.651650429,-6.952686082,-.1250000000,-1.677050983,-6.952686082,-.3169328455,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,.4724555913e-1,.1430430338,0.,0.,0.,.1250000000,.5757371053,0.,0.,0.,.4145780988,0.,-.1250000000,-.4145780988,-.4724555913e-1,-.5757371053,-.1430430338},
			{0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,.6250000000e-1,0.,0.,.6250000000e-1,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,.6250000000e-1,.1892281472,0.,0.,0.,.3307189139,1.269381000,0.,0.,.5229125166,2.742176371,0.,.3307189139,2.742176371,.6250000000e-1,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,-.9375000000e-1,0.,0.,0.,-.6750771565e-1,0.,0.,.6750771565e-1,0.,.9375000000e-1,.2329237477e-1},
			{0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,-.2096313728,0.,0.,0.,-.4528555234,0.,0.,-.4528555234,0.,-.2096313728,-.3125000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,.2538762001,1.224744872,2.371708244,2.561737692,2.218529919,1.703053324,.9682458365,2.561737692,3.507803799,3.679015663,.6614378278,2.218529919,3.679015663,.4192627457,1.703053324,.2538762001},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5182226234,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-3.476343041,-.7905694150,-3.137475099,-5.728219620,-7.509759277,-.8100925875,-3.622844188,-7.509759277,-.6846531970,-3.476343041,-.5182226234},
			{0.,0.,.2500000000,.5123475385,.6495190530,.6555055300,0.,.6123724358,1.984313483,3.436931772,4.397264773,.2500000000,1.984313483,5.434266281,9.499177600,.5123475385,3.436931772,9.499177600,.6495190530,4.397264773,.6555055300},
			{0.,0.,0.,-.1250000000,-.3169328455,-.4797809210,0.,0.,-.4841229182,-1.677050983,-3.218468261,0.,-.4841229182,-2.651650429,-6.952686082,-.1250000000,-1.677050983,-6.952686082,-.3169328455,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,.6250000000e-1,.1892281472,0.,0.,0.,.3307189139,1.269381000,0.,0.,.5229125166,2.742176371,0.,.3307189139,2.742176371,.6250000000e-1,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,-.2096313728,0.,0.,0.,-.4528555234,0.,0.,-.4528555234,0.,-.2096313728,-.3125000000e-1},
			{0.,.5000000000,.7905694150,.8100925875,.6846531970,.5182226234,-.5000000000,0.,1.045825033,1.811422094,2.085805824,-.7905694150,-1.045825033,0.,1.501951855,-.8100925875,-1.811422094,-1.501951855,-.6846531970,-2.085805824,-.5182226234},
			{0.,0.,-.2041241452,-.4183300132,-.5303300858,-.5352180242,0.,0.,-.5400617248,-1.403121520,-2.154210993,.2041241452,.5400617248,0.,-1.551209206,.4183300132,1.403121520,1.551209206,.5303300858,2.154210993,.5352180242},
			{0.,0.,0.,.9682458365e-1,.2454951265,.3716367033,0.,0.,.1250000000,.6495190530,1.495808878,0.,-.1250000000,0.,1.077105496,-.9682458365e-1,-.6495190530,-1.077105496,-.2454951265,-1.495808878,-.3716367033},
			{0.,0.,0.,0.,-.4724555913e-1,-.1430430338,0.,0.,0.,-.1250000000,-.5757371053,0.,0.,0.,-.4145780988,0.,.1250000000,.4145780988,.4724555913e-1,.5757371053,.1430430338},
			{0.,0.,0.,0.,0.,.2329237477e-1,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,.6750771565e-1,0.,0.,-.6750771565e-1,0.,-.9375000000e-1,-.2329237477e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,.6555055300,0.,-.6123724358,-.6614378278,0.,.8794529546,.2500000000,-.6614378278,-1.811422094,-1.899835520,.5123475385,0.,-1.899835520,.6495190530,.8794529546,.6555055300},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,-.3716367033,0.,0.,.1250000000,0.,-.4986029592,0.,.1250000000,.6846531969,1.077105496,-.9682458365e-1,0.,1.077105496,-.2454951265,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,.4482107285e-1,.1357025371,0.,0.,0.,0.,.1820640586,0.,0.,-.1250000000,-.3933033180,0.,0.,-.3933033180,.4482107285e-1,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,.6250000000e-1,0.,0.,.6250000000e-1,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,.1250000000,.3169328455,.4797809210,0.,0.,-.4841229182,-.8385254914,-.6436936521,0.,.4841229182,0.,-1.390537216,-.1250000000,.8385254914,1.390537216,-.3169328455,.6436936521,-.4797809210},
			{0.,0.,0.,0.,-.4724555913e-1,-.1430430338,0.,0.,0.,.1250000000,.1919123684,0.,0.,0.,.4145780988,0.,-.1250000000,-.4145780988,.4724555913e-1,-.1919123684,.1430430338},
			{0.,0.,0.,0.,0.,.2156454873e-1,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,-.6250000000e-1,0.,0.,.6250000000e-1,0.,.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,.6250000000e-1,.1892281472,0.,0.,0.,-.3307189139,-.7616286003,0.,0.,.5229125166,.5484352742,0.,-.3307189139,.5484352742,.6250000000e-1,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,-.6750771565e-1,0.,0.,-.6750771565e-1,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,.3125000000e-1,0.,0.,0.,0.,-.2096313728,0.,0.,0.,.4528555234,0.,0.,-.4528555234,0.,.2096313728,-.3125000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,1.224744872,.9682458365,.6614378278,.4192627457,.2538762001,1.224744872,2.371708244,2.561737692,2.218529919,1.703053324,.9682458365,2.561737692,3.507803799,3.679015663,.6614378278,2.218529919,3.679015663,.4192627457,1.703053324,.2538762001},
			{0.,-.5000000000,-.7905694150,-.8100925875,-.6846531970,-.5182226234,-.5000000000,-1.936491673,-3.137475099,-3.622844188,-3.476343041,-.7905694150,-3.137475099,-5.728219620,-7.509759277,-.8100925875,-3.622844188,-7.509759277,-.6846531970,-3.476343041,-.5182226234},
			{0.,0.,.2500000000,.5123475385,.6495190530,.6555055300,0.,.6123724358,1.984313483,3.436931772,4.397264773,.2500000000,1.984313483,5.434266281,9.499177600,.5123475385,3.436931772,9.499177600,.6495190530,4.397264773,.6555055300},
			{0.,0.,0.,-.1250000000,-.3169328455,-.4797809210,0.,0.,-.4841229182,-1.677050983,-3.218468261,0.,-.4841229182,-2.651650429,-6.952686082,-.1250000000,-1.677050983,-6.952686082,-.3169328455,-3.218468261,-.4797809210},
			{0.,0.,0.,0.,.6250000000e-1,.1892281472,0.,0.,0.,.3307189139,1.269381000,0.,0.,.5229125166,2.742176371,0.,.3307189139,2.742176371,.6250000000e-1,1.269381000,.1892281472},
			{0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,-.2096313728,0.,0.,0.,-.4528555234,0.,0.,-.4528555234,0.,-.2096313728,-.3125000000e-1},
			{0.,.5000000000,.7905694150,.8100925875,.6846531970,.5182226234,-.5000000000,0.,1.045825033,1.811422094,2.085805824,-.7905694150,-1.045825033,0.,1.501951855,-.8100925875,-1.811422094,-1.501951855,-.6846531970,-2.085805824,-.5182226234},
			{0.,0.,-.2041241452,-.4183300132,-.5303300858,-.5352180242,0.,0.,-.5400617248,-1.403121520,-2.154210993,.2041241452,.5400617248,0.,-1.551209206,.4183300132,1.403121520,1.551209206,.5303300858,2.154210993,.5352180242},
			{0.,0.,0.,.9682458365e-1,.2454951265,.3716367033,0.,0.,.1250000000,.6495190530,1.495808878,0.,-.1250000000,0.,1.077105496,-.9682458365e-1,-.6495190530,-1.077105496,-.2454951265,-1.495808878,-.3716367033},
			{0.,0.,0.,0.,-.4724555913e-1,-.1430430338,0.,0.,0.,-.1250000000,-.5757371053,0.,0.,0.,-.4145780988,0.,.1250000000,.4145780988,.4724555913e-1,.5757371053,.1430430338},
			{0.,0.,0.,0.,0.,.2329237477e-1,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,.6750771565e-1,0.,0.,-.6750771565e-1,0.,-.9375000000e-1,-.2329237477e-1},
			{0.,0.,.2500000000,.5123475385,.6495190530,.6555055300,0.,-.6123724358,-.6614378278,0.,.8794529546,.2500000000,-.6614378278,-1.811422094,-1.899835520,.5123475385,0.,-1.899835520,.6495190530,.8794529546,.6555055300},
			{0.,0.,0.,-.9682458365e-1,-.2454951265,-.3716367033,0.,0.,.1250000000,0.,-.4986029592,0.,.1250000000,.6846531969,1.077105496,-.9682458365e-1,0.,1.077105496,-.2454951265,-.4986029592,-.3716367033},
			{0.,0.,0.,0.,.4482107285e-1,.1357025371,0.,0.,0.,0.,.1820640586,0.,0.,-.1250000000,-.3933033180,0.,0.,-.3933033180,.4482107285e-1,.1820640586,.1357025371},
			{0.,0.,0.,0.,0.,-.2156454873e-1,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,.6250000000e-1,0.,0.,.6250000000e-1,0.,-.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,.1250000000,.3169328455,.4797809210,0.,0.,-.4841229182,-.8385254914,-.6436936521,0.,.4841229182,0.,-1.390537216,-.1250000000,.8385254914,1.390537216,-.3169328455,.6436936521,-.4797809210},
			{0.,0.,0.,0.,-.4724555913e-1,-.1430430338,0.,0.,0.,.1250000000,.1919123684,0.,0.,0.,.4145780988,0.,-.1250000000,-.4145780988,.4724555913e-1,-.1919123684,.1430430338},
			{0.,0.,0.,0.,0.,.2156454873e-1,0.,0.,0.,0.,-.2893187812e-1,0.,0.,0.,-.6250000000e-1,0.,0.,.6250000000e-1,0.,.2893187812e-1,-.2156454873e-1},
			{0.,0.,0.,0.,.6250000000e-1,.1892281472,0.,0.,0.,-.3307189139,-.7616286003,0.,0.,.5229125166,.5484352742,0.,-.3307189139,.5484352742,.6250000000e-1,-.7616286003,.1892281472},
			{0.,0.,0.,0.,0.,-.2329237477e-1,0.,0.,0.,0.,.9375000000e-1,0.,0.,0.,-.6750771565e-1,0.,0.,-.6750771565e-1,0.,.9375000000e-1,-.2329237477e-1},
			{0.,0.,0.,0.,0.,.3125000000e-1,0.,0.,0.,0.,-.2096313728,0.,0.,0.,.4528555234,0.,0.,-.4528555234,0.,.2096313728,-.3125000000e-1}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8703882798e-1},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.8703882798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1054092553,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8703882798e-1},
			{0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.7453559923e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.},
			{0.,0.,.1825741858,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.3450327798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1259881577e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.},
			{0.,0.,0.,.1336306209,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1992047682e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6006249750e-2,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1054092553,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1297498240e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.8703882798e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,11.48912529,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,11.48912529}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.414213562,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,7.483314774,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,11.48912529,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.477225575,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,28.98275350,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,79.37253933,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.483314774,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,50.19960159,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,166.4932431,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,77.07139548,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,11.48912529}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}
