/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Aug 26, 2012
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CHYPERBOLIC_CONFIG_TSUNAMI_HPP_
#define CHYPERBOLIC_CONFIG_TSUNAMI_HPP_

#include "libsierpi/grid/CBoundaryConditions.hpp"

#define SIMULATION_DATASET_SIZE	4

// bathymetry
#define SIMULATION_DATASET_DEFAULT_VALUE_0	(10)
// displacement
#define SIMULATION_DATASET_DEFAULT_VALUE_1	(0.2)
// not used
#define SIMULATION_DATASET_DEFAULT_VALUE_2	(0)
// not used
#define SIMULATION_DATASET_DEFAULT_VALUE_3	(0)


#define DEFAULT_PARAMETER_VISUALIZATION_TRANSLATE_X		(0)
#define DEFAULT_PARAMETER_VISUALIZATION_TRANSLATE_Y		(0)
#define DEFAULT_PARAMETER_VISUALIZATION_TRANSLATE_Z		(0)

#define DEFAULT_PARAMETER_VISUALIZATION_SCALE_X		(1)
#define DEFAULT_PARAMETER_VISUALIZATION_SCALE_Y		(1)
#define DEFAULT_PARAMETER_VISUALIZATION_SCALE_Z		(0.5)

#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE == 1
#	define DEFAULT_PARAMETER_VISUALIZATION_DATA0_SCALE		(0.2)
#	define DEFAULT_PARAMETER_VISUALIZATION_DATA1_SCALE		(1.0/30000.0)
#else
#	define DEFAULT_PARAMETER_VISUALIZATION_DATA0_SCALE		(1.0)
#	define DEFAULT_PARAMETER_VISUALIZATION_DATA1_SCALE		(1.0/1000.0)
#endif

/*
 * default boundary condition
 */
//#define DEFAULT_BOUNDARY_CONDITION	(BOUNDARY_CONDITION_OUTFLOW)
#define DEFAULT_BOUNDARY_CONDITION	(BOUNDARY_CONDITION_DIRICHLET)
//#define DEFAULT_BOUNDARY_CONDITION	(BOUNDARY_CONDITION_VELOCITY_ZERO)



#if SIMULATION_HYPERBOLIC_ADAPTIVITY_MODE == 1

#	define SIMULATION_REFINE_PARAMETER_0	(0.1)
#	define SIMULATION_COARSEN_PARAMETER_0	(0.01)

#	define SIMULATION_REFINE_PARAMETER_1	(0.1)
#	define SIMULATION_COARSEN_PARAMETER_1	(0.02)

#elif SIMULATION_HYPERBOLIC_ADAPTIVITY_MODE == 2

#	define SIMULATION_REFINE_PARAMETER_0	100
#	define SIMULATION_REFINE_PARAMETER_1	100

#	define SIMULATION_COARSEN_PARAMETER_0	20
#	define SIMULATION_COARSEN_PARAMETER_1	20

#elif SIMULATION_HYPERBOLIC_ADAPTIVITY_MODE == 3

#	define SIMULATION_REFINE_PARAMETER_0	0.1
#	define SIMULATION_REFINE_PARAMETER_1	1

#	define SIMULATION_COARSEN_PARAMETER_0	0.01
#	define SIMULATION_COARSEN_PARAMETER_1	0.1

#else

#	error "invalid adaptivity mode!"

#endif



#ifndef SIMULATION_TSUNAMI_ENABLE_BATHYMETRY_KERNELS
#	error "SIMULATION_TSUNAMI_ENABLE_BATHYMETRY_KERNELS not defined"
#	define SIMULATION_TSUNAMI_ENABLE_BATHYMETRY_KERNELS 1
#endif


/**
 * tsunami dry threshold
 */
#ifndef SIMULATION_TSUNAMI_DRY_THRESHOLD
#	error "SIMULATION_TSUNAMI_DRY_THRESHOLD not defined"
#	define SIMULATION_TSUNAMI_DRY_THRESHOLD		(0.1)
#endif


/**
 * tsunami zero threshold
 */
#ifndef SIMULATION_TSUNAMI_ZERO_THRESHOLD
#	error "SIMULATION_TSUNAMI_ZERO_THRESHOLD not defined"
#	define SIMULATION_TSUNAMI_ZERO_THRESHOLD		(0.000000001)
#endif





#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==0
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"lax friedrich"
#endif
#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==1
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"rusanov"
#endif
#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==2
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"fwave"
#endif
#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==3
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"augumented riemann"
#endif
#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==4
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"hybrid"
#endif
#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==5
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"augumented riemann (geoclaw)"
#endif
#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==6
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"augumented riemann (geoclaw)"
#endif
#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==7
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"velocity upwinding"
#endif
#if SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID==10
	#define SIMULATION_HYPERBOLIC_FLUX_SOLVER_ID_STRING		"benchmark velocity upwinding"
#endif



#endif /* CCONFIGTSUNAMI_HPP_ */
