
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION2D_MATRICES_CONST_HPP_
#define CCOMPUTATION2D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (5)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (21)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (21)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS (12)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (52)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*5+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_2D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[21],
		T io_hv[21]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,7.838367178,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,7.668115805,10.73312629,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.828427124,7.314285714,11.33763824,12.92995257,9.938586932,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,4.381780461,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,2.449489743,9.142857143,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.533303756,6.337932448,13.55261854,16.66751699,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,-2.993325910,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,3.346640106,-5.074680381,1.521277658,0.,0.,0.,0.,0.,0.,0.,4.364357806,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.320493798,5.935526613,-.3207134902,-5.966396197,2.530240385,0.,0.,0.,0.,0.,0.,1.792842914,9.200874127,19.80226324,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,10.95445115,-3.833259389,.5532833353,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.857142857,9.972922524,-7.273098319,1.490788040,0.,0.,0.,0.,0.,4.320493800,21.25401924,0.,0.,0.,0.},
			{5.416025602,-4.376809636,2.010178182,-.5158119621,.5766953058e-1,0.,0.,0.,0.,0.,0.,15.73213272,-4.036867140,.4513354669,0.,0.,0.,0.,21.79908256,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,7.838367178,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,7.668115805,10.73312629,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.828427124,7.314285714,11.33763824,12.92995257,9.938586932,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,4.381780461,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,2.449489743,9.142857143,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.533303756,6.337932448,13.55261854,16.66751699,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,-2.993325910,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,3.346640106,-5.074680381,1.521277658,0.,0.,0.,0.,0.,0.,0.,4.364357806,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.320493798,5.935526613,-.3207134902,-5.966396197,2.530240385,0.,0.,0.,0.,0.,0.,1.792842914,9.200874127,19.80226324,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,10.95445115,-3.833259389,.5532833353,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.857142857,9.972922524,-7.273098319,1.490788040,0.,0.,0.,0.,0.,4.320493800,21.25401924,0.,0.,0.,0.},
			{5.416025602,-4.376809636,2.010178182,-.5158119621,.5766953058e-1,0.,0.,0.,0.,0.,0.,15.73213272,-4.036867140,.4513354669,0.,0.,0.,0.,21.79908256,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,-2.828427124,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,8.854377448,-3.098386677,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.715476066,-4.618802155,12.72792206,-3.265986323,20.08316045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,3.919183589,4.200000000,0.,0.,0.,-5.091168823,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,3.834057902,5.366563145,4.647580015,0.,0.,6.640783086,-6.196773354,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.657142857,5.668819119,6.464976283,4.969293464,0.,-6.730254569,9.455027815,-6.998542123,17.21413752,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,-.9128709293,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,2.529822128,-1.807392228,0.,0.,0.,2.190890230,6.260990336,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,1.649572198,-2.571428571,0.,0.,1.224744872,4.571428571,7.423074892,0.,0.,-7.377111136,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.825741857,3.836097790,3.659207005,.5216405308,-3.207665141,0.,.7666518779,3.168966225,6.776309270,8.333758491,0.,8.636804253,-8.748177653,15.06237033,0.,0.,0.,0.,0.,0.,0.},
			{2.645751311,-1.496662955,.3055050464,0.,0.,0.,5.184592560,-1.058300524,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,1.673320053,-2.537340190,.7606388294,0.,0.,2.898275350,5.747048932,-2.195775165,0.,0.,2.182178903,8.504200642,0.,0.,10.06230590,0.,0.,0.,0.,0.},
			{2.160246899,2.967763307,-.1603567451,-2.983198099,1.265120192,0.,1.814229470,5.416047582,5.345224839,-3.286878676,0.,.8964214566,4.600437064,9.901131622,0.,-9.525793445,13.38877363,0.,0.,0.,0.},
			{2.683281572,-1.897366596,.6639400021,-.9583148473e-1,0.,0.,5.477225576,-1.916629694,.2766416676,0.,0.,7.423074892,-1.071428571,0.,0.,8.874119673,0.,0.,0.,0.,0.},
			{2.449489743,.9897433188,-2.727411870,1.516350793,-.2869022921,0.,3.428571429,4.986461263,-3.636549159,.7453940198,0.,3.049339172,8.346248498,-2.245365598,0.,2.160246900,10.62700962,0.,12.04989627,0.,0.},
			{2.708012801,-2.188404818,1.005089091,-.2579059811,.2883476528e-1,0.,5.685642496,-2.611298058,.6700593942,-.7491491772e-1,0.,7.866066360,-2.018433570,.2256677334,0.,9.552971226,-1.068054652,0.,10.89954128,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,-2.828427124,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,8.854377448,-3.098386677,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.715476066,-4.618802155,12.72792206,-3.265986323,20.08316045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,3.919183589,4.200000000,0.,0.,0.,-5.091168823,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,3.834057902,5.366563145,4.647580015,0.,0.,6.640783086,-6.196773354,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.657142857,5.668819119,6.464976283,4.969293464,0.,-6.730254569,9.455027815,-6.998542123,17.21413752,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,-.9128709293,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,2.529822128,-1.807392228,0.,0.,0.,2.190890230,6.260990336,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,1.649572198,-2.571428571,0.,0.,1.224744872,4.571428571,7.423074892,0.,0.,-7.377111136,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.825741857,3.836097790,3.659207005,.5216405308,-3.207665141,0.,.7666518779,3.168966225,6.776309270,8.333758491,0.,8.636804253,-8.748177653,15.06237033,0.,0.,0.,0.,0.,0.,0.},
			{2.645751311,-1.496662955,.3055050464,0.,0.,0.,5.184592560,-1.058300524,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,1.673320053,-2.537340190,.7606388294,0.,0.,2.898275350,5.747048932,-2.195775165,0.,0.,2.182178903,8.504200642,0.,0.,10.06230590,0.,0.,0.,0.,0.},
			{2.160246899,2.967763307,-.1603567451,-2.983198099,1.265120192,0.,1.814229470,5.416047582,5.345224839,-3.286878676,0.,.8964214566,4.600437064,9.901131622,0.,-9.525793445,13.38877363,0.,0.,0.,0.},
			{2.683281572,-1.897366596,.6639400021,-.9583148473e-1,0.,0.,5.477225576,-1.916629694,.2766416676,0.,0.,7.423074892,-1.071428571,0.,0.,8.874119673,0.,0.,0.,0.,0.},
			{2.449489743,.9897433188,-2.727411870,1.516350793,-.2869022921,0.,3.428571429,4.986461263,-3.636549159,.7453940198,0.,3.049339172,8.346248498,-2.245365598,0.,2.160246900,10.62700962,0.,12.04989627,0.,0.},
			{2.708012801,-2.188404818,1.005089091,-.2579059811,.2883476528e-1,0.,5.685642496,-2.611298058,.6700593942,-.7491491772e-1,0.,7.866066360,-2.018433570,.2256677334,0.,9.552971226,-1.068054652,0.,10.89954128,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,-1.94468190274016,2.27090372268066,-2.44801058376252,2.49586968540814,-2.4264683064468,3.43216369764802,-4.00974905232829,4.32546942586304,-4.4144312457315,4.2977488551218,5.37669458746301,-5.80779112496784,5.9385809984625,-5.79724384742923,7.27823584467214,-7.46210387346585,7.3119743919632,9.14177937313724,-8.99870040642757,10.9691712594618},
			{1.414213562,-1.71235176911142,1.5663330178964,-1.07607439139602,.363617646552178,.420994417051474,3.29802783242322,-3.07100849759875,2.20598407348367,-.923194488569861,-.507379758856801,4.96464287844228,-3.80884940873901,2.02329330850018,.0162698697481229,6.45780747384475,-4.10479708970297,1.32544341327556,7.79428043908965,-4.03553981216457,8.98680735137078},
			{1.414213562,-1.30970802258292,.519224820053661,.518731848959601,-1.30425364587756,1.47252576661999,3.06556135691961,-1.5947619510804,-.444911232504622,2.09440936915946,-2.63186985254377,4.2894282422279,-.964148801340397,-2.01722273728899,3.46190525493913,5.18623489409548,.205483038001612,-3.63947886955732,5.81833621896757,1.69243379052584,6.23568135536058},
			{1.414213562,-.761953862859853,-.55104212557113,1.37646295656649,-1.03415883911242,-.174795669385988,2.74931534527504,.106757773657142,-2.26282036321899,2.15044627554921,-.119501749587716,3.45007465931297,1.77035093065653,-3.42380239520456,1.33170919606025,3.7410698100085,3.58481468643821,-3.59724489879787,3.76406389114812,5.2355975909823,3.61790059797515},
			{1.414213562,-.103494496994541,-1.29720437850508,.991754894783212,.589436063368126,-1.41318416081787,2.36915365304765,1.68415540504794,-2.32831086437971,-.381749900992996,2.46800676549626,2.56192161158569,3.58714638367868,-1.52344991306454,-2.69130383032704,2.3938769796025,4.93737708601548,.508166439915615,2.07554270972352,5.63164748396125,1.71909544842},
			{1.414213562,.624299774465593,-1.43546196207946,-.337640235075449,1.50389998701775,-.0107903422167659,1.94896143440088,2.83314651369184,-.637278791777431,-2.63766838605638,.808065843843811,1.73374759666501,4.12740556141282,1.98940127224618,-2.96832239843811,1.33269841309056,4.37531622836173,4.52786743647191,.950542947793622,3.96849611187728,.64766447670966},
			{1.414213562,1.37570022553441,-.821946062819915,-1.61733246050546,.359308464240942,1.71165917287845,1.51514018159912,3.36446686788906,1.91964817870762,-1.95173675811295,-2.75500656898115,1.04781596872157,3.55510980002441,4.5904502888745,1.24891795372526,.626153181688305,2.8447312409815,5.48872263996492,.347192252628357,1.9734504193592,.183906859961432},
			{1.414213562,2.10349449699454,.504794589391635,-1.67800190212113,-1.79846700964276,.309494957156941,1.09494796295235,3.24473228828999,4.05017972204568,1.63294619261043,-2.28560008620039,.547225517982272,2.39319484717323,4.87179403320197,5.49730994822616,.23632114600406,1.36209257652084,3.7763572186341,.0946962283839243,.676667877319742,.0362494356624604},
			{1.414213562,2.76195386285985,2.32621648423849,.160883133644535,-2.12140648579363,-2.7058281817046,.714786270724957,2.59853682213183,4.7549644652192,5.36731919542456,3.25729414076559,.233201739437629,1.22672710731592,3.30495822058058,5.89679444984458,.0657431866024218,.451524561772794,1.59193561065057,.0171974308138245,.145628931559436,.00429748986296868},
			{1.414213562,3.30970802258292,4.2909622271481,3.64701232940425,1.49376547749553,-1.29952592474851,.398540259080393,1.67165845871656,3.84755914106515,6.3113399865174,7.98777931164405,.07249762486996,.434860760346183,1.40707187748205,3.23180916099,.0113956405123903,.0887334889355507,.368089956036703,.00166206328995187,.015902832140972,.000231576691010652},
			{1.414213562,3.71235176911142,5.99558490979395,7.62488741580146,8.14599447640013,7.34718311089616,.166073783576779,.764836159193308,2.02218309922343,4.02885967082133,6.68010475273478,.0125887190510937,.0823390061849577,.299004079114616,.796298871799331,.000824565926335141,.00697737063218211,.0321260785532915,5.01144631619021e-005,.000520025886860232,2.90963993885675e-006},
			{1.414213562,3.94468190274016,7.0795491341205,10.545740291788,14.1178337840267,17.5926352342147,.0319379183519753,.154660030302817,.43925822557731,.961797763577864,1.79684767953686,.000465578153648566,.00319092434145458,.0123137463285649,.0353622090543065,5.86464816370859e-006,5.19109369727387e-005,.000252650701904178,6.85464935253131e-008,7.43273958578616e-007,7.65349815116598e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,-1.94468190274016,2.27090372268066,-2.44801058376252,2.49586968540814,-2.4264683064468,3.43216369764802,-4.00974905232829,4.32546942586304,-4.4144312457315,4.2977488551218,5.37669458746301,-5.80779112496784,5.9385809984625,-5.79724384742923,7.27823584467214,-7.46210387346585,7.3119743919632,9.14177937313724,-8.99870040642757,10.9691712594618},
			{1.414213562,-1.71235176911142,1.5663330178964,-1.07607439139602,.363617646552178,.420994417051474,3.29802783242322,-3.07100849759875,2.20598407348367,-.923194488569861,-.507379758856801,4.96464287844228,-3.80884940873901,2.02329330850018,.0162698697481229,6.45780747384475,-4.10479708970297,1.32544341327556,7.79428043908965,-4.03553981216457,8.98680735137078},
			{1.414213562,-1.30970802258292,.519224820053661,.518731848959601,-1.30425364587756,1.47252576661999,3.06556135691961,-1.5947619510804,-.444911232504622,2.09440936915946,-2.63186985254377,4.2894282422279,-.964148801340397,-2.01722273728899,3.46190525493913,5.18623489409548,.205483038001612,-3.63947886955732,5.81833621896757,1.69243379052584,6.23568135536058},
			{1.414213562,-.761953862859853,-.55104212557113,1.37646295656649,-1.03415883911242,-.174795669385988,2.74931534527504,.106757773657142,-2.26282036321899,2.15044627554921,-.119501749587716,3.45007465931297,1.77035093065653,-3.42380239520456,1.33170919606025,3.7410698100085,3.58481468643821,-3.59724489879787,3.76406389114812,5.2355975909823,3.61790059797515},
			{1.414213562,-.103494496994541,-1.29720437850508,.991754894783212,.589436063368126,-1.41318416081787,2.36915365304765,1.68415540504794,-2.32831086437971,-.381749900992996,2.46800676549626,2.56192161158569,3.58714638367868,-1.52344991306454,-2.69130383032704,2.3938769796025,4.93737708601548,.508166439915615,2.07554270972352,5.63164748396125,1.71909544842},
			{1.414213562,.624299774465593,-1.43546196207946,-.337640235075449,1.50389998701775,-.0107903422167659,1.94896143440088,2.83314651369184,-.637278791777431,-2.63766838605638,.808065843843811,1.73374759666501,4.12740556141282,1.98940127224618,-2.96832239843811,1.33269841309056,4.37531622836173,4.52786743647191,.950542947793622,3.96849611187728,.64766447670966},
			{1.414213562,1.37570022553441,-.821946062819915,-1.61733246050546,.359308464240942,1.71165917287845,1.51514018159912,3.36446686788906,1.91964817870762,-1.95173675811295,-2.75500656898115,1.04781596872157,3.55510980002441,4.5904502888745,1.24891795372526,.626153181688305,2.8447312409815,5.48872263996492,.347192252628357,1.9734504193592,.183906859961432},
			{1.414213562,2.10349449699454,.504794589391635,-1.67800190212113,-1.79846700964276,.309494957156941,1.09494796295235,3.24473228828999,4.05017972204568,1.63294619261043,-2.28560008620039,.547225517982272,2.39319484717323,4.87179403320197,5.49730994822616,.23632114600406,1.36209257652084,3.7763572186341,.0946962283839243,.676667877319742,.0362494356624604},
			{1.414213562,2.76195386285985,2.32621648423849,.160883133644535,-2.12140648579363,-2.7058281817046,.714786270724957,2.59853682213183,4.7549644652192,5.36731919542456,3.25729414076559,.233201739437629,1.22672710731592,3.30495822058058,5.89679444984458,.0657431866024218,.451524561772794,1.59193561065057,.0171974308138245,.145628931559436,.00429748986296868},
			{1.414213562,3.30970802258292,4.2909622271481,3.64701232940425,1.49376547749553,-1.29952592474851,.398540259080393,1.67165845871656,3.84755914106515,6.3113399865174,7.98777931164405,.07249762486996,.434860760346183,1.40707187748205,3.23180916099,.0113956405123903,.0887334889355507,.368089956036703,.00166206328995187,.015902832140972,.000231576691010652},
			{1.414213562,3.71235176911142,5.99558490979395,7.62488741580146,8.14599447640013,7.34718311089616,.166073783576779,.764836159193308,2.02218309922343,4.02885967082133,6.68010475273478,.0125887190510937,.0823390061849577,.299004079114616,.796298871799331,.000824565926335141,.00697737063218211,.0321260785532915,5.01144631619021e-005,.000520025886860232,2.90963993885675e-006},
			{1.414213562,3.94468190274016,7.0795491341205,10.545740291788,14.1178337840267,17.5926352342147,.0319379183519753,.154660030302817,.43925822557731,.961797763577864,1.79684767953686,.000465578153648566,.00319092434145458,.0123137463285649,.0353622090543065,5.86464816370859e-006,5.19109369727387e-005,.000252650701904178,6.85464935253131e-008,7.43273958578616e-007,7.65349815116598e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.40022577929605,4.1644090826311,-4.80864541143983,5.37622900930936,-5.88936380548032,5.17702934684997,-5.97791857375092,6.68351614602547,-7.32142511479571,-6.67435503772231,7.46215578440288,-8.17438210188192,7.8090615264033,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.13195404884644,3.83584465679206,-4.42925189189755,4.95205415939119,-5.42470353822284,3.97724364248013,-4.59252537503604,5.13459946240545,-5.62467189980939,-3.67768288095471,4.11177446033512,-4.5042232468252,2.21014640326355,-2.42109408164425,.161112776439225},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-2.66702109783921,3.26642040979696,-3.77173740709444,4.21693061735794,-4.61941604516583,2.13132645129358,-2.46104380066281,2.75153061636385,-3.01415077313091,.104423971232213,-.116749549066047,.127892723210502,-2.94703586087065,3.22831603852647,4.81780334972074},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-2.03452907455009,2.49177904847469,-2.87725860980962,3.21687291987534,-3.52390772571971,.0953780462518065,-.110133081351697,.123132528211026,-.134884926556002,2.8024998850757,-3.1332901246654,3.43234736114769,-3.77550789188145,4.13586167809416,1.3916398503358},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-1.2742056900953,1.56057688324205,-1.80199896789671,2.01469609360342,-2.20698899398908,-1.62700976129611,1.87870904723461,-2.10046056933792,2.30093927043685,3.19783168117628,-3.57528450949468,3.91652795089531,-.49602746920004,.543370868210304,-3.52996795992507},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-.433821252801768,.531320354197217,-.613515899259544,.68593162794344,-.751400451068813,-2.60976045384023,3.01349180095816,-3.3691862575809,3.69075862820297,1.36899683098764,-1.53058498738027,1.67667184761831,3.00982432351404,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,.433821252801768,-.531320354197217,.613515899259544,-.68593162794344,.751400451068813,-2.60976045384023,3.01349180095816,-3.36918625758091,3.69075862820298,-1.36899683098765,1.53058498738027,-1.67667184761832,3.00982432351407,-3.297097353229,2.50312955199098},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,1.2742056900953,-1.56057688324205,1.80199896789671,-2.01469609360342,2.20698899398908,-1.62700976129612,1.87870904723461,-2.10046056933793,2.30093927043686,-3.1978316811763,3.5752845094947,-3.91652795089533,-.496027469200023,.543370868210286,3.52996795992515},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,2.03452907455009,-2.49177904847469,2.87725860980962,-3.21687291987534,3.52390772571971,.0953780462518019,-.110133081351691,.12313252821102,-.134884926555995,-2.8024998850757,3.13329012466539,-3.43234736114768,-3.77550789188148,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,2.66702109783921,-3.26642040979696,3.77173740709444,-4.21693061735794,4.61941604516583,2.13132645129358,-2.46104380066281,2.75153061636384,-3.0141507731309,-.104423971232209,.116749549066043,-.127892723210497,-2.94703586087065,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,3.13195404884644,-3.83584465679206,4.42925189189755,-4.95205415939119,5.42470353822284,3.97724364248012,-4.59252537503603,5.13459946240544,-5.62467189980938,3.67768288095471,-4.11177446033511,4.5042232468252,2.21014640326349,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,3.40022577929605,-4.1644090826311,4.80864541143983,-5.37622900930936,5.88936380548032,5.17702934684997,-5.97791857375093,6.68351614602547,-7.32142511479571,6.67435503772228,-7.46215578440284,8.17438210188188,7.8090615264033,-8.55439830449838,8.51143945889045}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.40022577929605,4.1644090826311,-4.80864541143983,5.37622900930936,-5.88936380548032,5.17702934684997,-5.97791857375092,6.68351614602547,-7.32142511479571,-6.67435503772231,7.46215578440288,-8.17438210188192,7.8090615264033,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.13195404884644,3.83584465679206,-4.42925189189755,4.95205415939119,-5.42470353822284,3.97724364248013,-4.59252537503604,5.13459946240545,-5.62467189980939,-3.67768288095471,4.11177446033512,-4.5042232468252,2.21014640326355,-2.42109408164425,.161112776439225},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-2.66702109783921,3.26642040979696,-3.77173740709444,4.21693061735794,-4.61941604516583,2.13132645129358,-2.46104380066281,2.75153061636385,-3.01415077313091,.104423971232213,-.116749549066047,.127892723210502,-2.94703586087065,3.22831603852647,4.81780334972074},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-2.03452907455009,2.49177904847469,-2.87725860980962,3.21687291987534,-3.52390772571971,.0953780462518065,-.110133081351697,.123132528211026,-.134884926556002,2.8024998850757,-3.1332901246654,3.43234736114769,-3.77550789188145,4.13586167809416,1.3916398503358},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-1.2742056900953,1.56057688324205,-1.80199896789671,2.01469609360342,-2.20698899398908,-1.62700976129611,1.87870904723461,-2.10046056933792,2.30093927043685,3.19783168117628,-3.57528450949468,3.91652795089531,-.49602746920004,.543370868210304,-3.52996795992507},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-.433821252801768,.531320354197217,-.613515899259544,.68593162794344,-.751400451068813,-2.60976045384023,3.01349180095816,-3.3691862575809,3.69075862820297,1.36899683098764,-1.53058498738027,1.67667184761831,3.00982432351404,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,.433821252801768,-.531320354197217,.613515899259544,-.68593162794344,.751400451068813,-2.60976045384023,3.01349180095816,-3.36918625758091,3.69075862820298,-1.36899683098765,1.53058498738027,-1.67667184761832,3.00982432351407,-3.297097353229,2.50312955199098},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,1.2742056900953,-1.56057688324205,1.80199896789671,-2.01469609360342,2.20698899398908,-1.62700976129612,1.87870904723461,-2.10046056933793,2.30093927043686,-3.1978316811763,3.5752845094947,-3.91652795089533,-.496027469200023,.543370868210286,3.52996795992515},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,2.03452907455009,-2.49177904847469,2.87725860980962,-3.21687291987534,3.52390772571971,.0953780462518019,-.110133081351691,.12313252821102,-.134884926555995,-2.8024998850757,3.13329012466539,-3.43234736114768,-3.77550789188148,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,2.66702109783921,-3.26642040979696,3.77173740709444,-4.21693061735794,4.61941604516583,2.13132645129358,-2.46104380066281,2.75153061636384,-3.0141507731309,-.104423971232209,.116749549066043,-.127892723210497,-2.94703586087065,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,3.13195404884644,-3.83584465679206,4.42925189189755,-4.95205415939119,5.42470353822284,3.97724364248012,-4.59252537503603,5.13459946240544,-5.62467189980938,3.67768288095471,-4.11177446033511,4.5042232468252,2.21014640326349,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,3.40022577929605,-4.1644090826311,4.80864541143983,-5.37622900930936,5.88936380548032,5.17702934684997,-5.97791857375093,6.68351614602547,-7.32142511479571,6.67435503772228,-7.46215578440284,8.17438210188188,7.8090615264033,-8.55439830449838,8.51143945889045}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,3.94468190274016,7.0795491341205,10.545740291788,14.1178337840267,17.5926352342147,-.0319379183519753,-.154660030302817,-.43925822557731,-.961797763577864,-1.79684767953686,.000465578153648587,.00319092434145473,.0123137463285655,.0353622090543082,-5.86464816610367e-006,-5.19109369939388e-005,-.000252650702007359,6.85464900528793e-008,7.43273920925785e-007,-7.65370303624714e-010},
			{1.414213562,3.71235176911142,5.99558490979395,7.62488741580146,8.14599447640013,7.34718311089616,-.166073783576779,-.764836159193308,-2.02218309922343,-4.02885967082133,-6.68010475273478,.0125887190510936,.0823390061849571,.299004079114613,.796298871799325,-.000824565926326729,-.00697737063211093,-.0321260785529637,5.01144631645353e-005,.000520025886887555,-2.90963996691883e-006},
			{1.414213562,3.30970802258292,4.2909622271481,3.64701232940425,1.49376547749553,-1.29952592474851,-.398540259080393,-1.67165845871656,-3.84755914106515,-6.3113399865174,-7.98777931164405,.0724976248699598,.434860760346183,1.40707187748205,3.23180916098999,-.0113956405123911,-.0887334889355571,-.36808995603673,.00166206328995187,.015902832140972,-.00023157669095963},
			{1.414213562,2.76195386285985,2.32621648423849,.160883133644535,-2.12140648579363,-2.7058281817046,-.714786270724957,-2.59853682213183,-4.7549644652192,-5.36731919542456,-3.25729414076559,.233201739437629,1.22672710731592,3.30495822058058,5.89679444984459,-.0657431866024218,-.451524561772794,-1.59193561065057,.017197430813833,.145628931559507,-.00429748986298909},
			{1.414213562,2.10349449699454,.504794589391635,-1.67800190212113,-1.79846700964276,.309494957156941,-1.09494796295235,-3.24473228828999,-4.05017972204568,-1.63294619261043,2.28560008620039,.547225517982271,2.39319484717323,4.87179403320196,5.49730994822615,-.236321146004056,-1.36209257652082,-3.77635721863404,.0946962283839159,.676667877319681,-.0362494356624808},
			{1.414213562,1.37570022553441,-.821946062819915,-1.61733246050546,.359308464240942,1.71165917287845,-1.51514018159912,-3.36446686788906,-1.91964817870762,1.95173675811295,2.75500656898115,1.04781596872157,3.55510980002442,4.59045028887451,1.24891795372526,-.626153181688296,-2.84473124098146,-5.48872263996485,.347192252628353,1.97345041935918,-.183906859961412},
			{1.414213562,.624299774465593,-1.43546196207946,-.337640235075449,1.50389998701775,-.0107903422167659,-1.94896143440088,-2.83314651369184,.637278791777431,2.63766838605638,-.808065843843811,1.73374759666501,4.12740556141283,1.98940127224618,-2.96832239843812,-1.33269841309056,-4.37531622836174,-4.52786743647192,.95054294779362,3.96849611187727,-.647664476709678},
			{1.414213562,-.103494496994541,-1.29720437850508,.991754894783212,.589436063368126,-1.41318416081787,-2.36915365304765,-1.68415540504794,2.32831086437971,.381749900992996,-2.46800676549626,2.56192161158569,3.58714638367868,-1.52344991306454,-2.69130383032704,-2.39387697960251,-4.93737708601551,-.508166439915618,2.07554270972356,5.63164748396134,-1.71909544841957},
			{1.414213562,-.761953862859853,-.55104212557113,1.37646295656649,-1.03415883911242,-.174795669385988,-2.74931534527504,-.106757773657142,2.26282036321899,-2.15044627554921,.119501749587716,3.45007465931297,1.77035093065653,-3.42380239520456,1.33170919606025,-3.74106981000851,-3.58481468643821,3.59724489879787,3.76406389114815,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.30970802258292,.519224820053661,.518731848959601,-1.30425364587756,1.47252576661999,-3.06556135691961,1.5947619510804,.444911232504622,-2.09440936915946,2.63186985254377,4.2894282422279,-.964148801340396,-2.01722273728898,3.46190525493913,-5.18623489409549,-.205483038001613,3.63947886955733,5.81833621896758,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.71235176911142,1.5663330178964,-1.07607439139602,.363617646552178,.420994417051474,-3.29802783242322,3.07100849759875,-2.20598407348367,.923194488569861,.507379758856801,4.96464287844228,-3.80884940873902,2.02329330850018,.0162698697481229,-6.45780747384476,4.10479708970298,-1.32544341327556,7.79428043908953,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.94468190274016,2.27090372268066,-2.44801058376252,2.49586968540814,-2.4264683064468,-3.43216369764802,4.00974905232829,-4.32546942586304,4.4144312457315,-4.2977488551218,5.37669458746301,-5.80779112496784,5.9385809984625,-5.79724384742923,-7.27823584467216,7.46210387346587,-7.31197439196322,9.14177937313726,-8.9987004064276,-10.9691712594614}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[21],
		T o_output[12]
	) {
		static const T matrix[12][21] = 
		{
			{1.414213562,3.94468190274016,7.0795491341205,10.545740291788,14.1178337840267,17.5926352342147,-.0319379183519753,-.154660030302817,-.43925822557731,-.961797763577864,-1.79684767953686,.000465578153648587,.00319092434145473,.0123137463285655,.0353622090543082,-5.86464816610367e-006,-5.19109369939388e-005,-.000252650702007359,6.85464900528793e-008,7.43273920925785e-007,-7.65370303624714e-010},
			{1.414213562,3.71235176911142,5.99558490979395,7.62488741580146,8.14599447640013,7.34718311089616,-.166073783576779,-.764836159193308,-2.02218309922343,-4.02885967082133,-6.68010475273478,.0125887190510936,.0823390061849571,.299004079114613,.796298871799325,-.000824565926326729,-.00697737063211093,-.0321260785529637,5.01144631645353e-005,.000520025886887555,-2.90963996691883e-006},
			{1.414213562,3.30970802258292,4.2909622271481,3.64701232940425,1.49376547749553,-1.29952592474851,-.398540259080393,-1.67165845871656,-3.84755914106515,-6.3113399865174,-7.98777931164405,.0724976248699598,.434860760346183,1.40707187748205,3.23180916098999,-.0113956405123911,-.0887334889355571,-.36808995603673,.00166206328995187,.015902832140972,-.00023157669095963},
			{1.414213562,2.76195386285985,2.32621648423849,.160883133644535,-2.12140648579363,-2.7058281817046,-.714786270724957,-2.59853682213183,-4.7549644652192,-5.36731919542456,-3.25729414076559,.233201739437629,1.22672710731592,3.30495822058058,5.89679444984459,-.0657431866024218,-.451524561772794,-1.59193561065057,.017197430813833,.145628931559507,-.00429748986298909},
			{1.414213562,2.10349449699454,.504794589391635,-1.67800190212113,-1.79846700964276,.309494957156941,-1.09494796295235,-3.24473228828999,-4.05017972204568,-1.63294619261043,2.28560008620039,.547225517982271,2.39319484717323,4.87179403320196,5.49730994822615,-.236321146004056,-1.36209257652082,-3.77635721863404,.0946962283839159,.676667877319681,-.0362494356624808},
			{1.414213562,1.37570022553441,-.821946062819915,-1.61733246050546,.359308464240942,1.71165917287845,-1.51514018159912,-3.36446686788906,-1.91964817870762,1.95173675811295,2.75500656898115,1.04781596872157,3.55510980002442,4.59045028887451,1.24891795372526,-.626153181688296,-2.84473124098146,-5.48872263996485,.347192252628353,1.97345041935918,-.183906859961412},
			{1.414213562,.624299774465593,-1.43546196207946,-.337640235075449,1.50389998701775,-.0107903422167659,-1.94896143440088,-2.83314651369184,.637278791777431,2.63766838605638,-.808065843843811,1.73374759666501,4.12740556141283,1.98940127224618,-2.96832239843812,-1.33269841309056,-4.37531622836174,-4.52786743647192,.95054294779362,3.96849611187727,-.647664476709678},
			{1.414213562,-.103494496994541,-1.29720437850508,.991754894783212,.589436063368126,-1.41318416081787,-2.36915365304765,-1.68415540504794,2.32831086437971,.381749900992996,-2.46800676549626,2.56192161158569,3.58714638367868,-1.52344991306454,-2.69130383032704,-2.39387697960251,-4.93737708601551,-.508166439915618,2.07554270972356,5.63164748396134,-1.71909544841957},
			{1.414213562,-.761953862859853,-.55104212557113,1.37646295656649,-1.03415883911242,-.174795669385988,-2.74931534527504,-.106757773657142,2.26282036321899,-2.15044627554921,.119501749587716,3.45007465931297,1.77035093065653,-3.42380239520456,1.33170919606025,-3.74106981000851,-3.58481468643821,3.59724489879787,3.76406389114815,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.30970802258292,.519224820053661,.518731848959601,-1.30425364587756,1.47252576661999,-3.06556135691961,1.5947619510804,.444911232504622,-2.09440936915946,2.63186985254377,4.2894282422279,-.964148801340396,-2.01722273728898,3.46190525493913,-5.18623489409549,-.205483038001613,3.63947886955733,5.81833621896758,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.71235176911142,1.5663330178964,-1.07607439139602,.363617646552178,.420994417051474,-3.29802783242322,3.07100849759875,-2.20598407348367,.923194488569861,.507379758856801,4.96464287844228,-3.80884940873902,2.02329330850018,.0162698697481229,-6.45780747384476,4.10479708970298,-1.32544341327556,7.79428043908953,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.94468190274016,2.27090372268066,-2.44801058376252,2.49586968540814,-2.4264683064468,-3.43216369764802,4.00974905232829,-4.32546942586304,4.4144312457315,-4.2977488551218,5.37669458746301,-5.80779112496784,5.9385809984625,-5.79724384742923,-7.27823584467216,7.46210387346587,-7.31197439196322,9.14177937313726,-8.9987004064276,-10.9691712594614}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 12; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0471753389703784,-.106939314428477,-.16007833916176,-.203167424021368,-.233492536723722,-.249147045798746,-.249147045798746,-.233492536723722,-.203167424021368,-.16007833916176,-.106939314428477,-.0471753389703784},
			{.0648707029602942,.129483784537873,.148249097989688,.109463102050698,.0170873715800573,-.109985117239921,-.2423620138609,-.347295683640528,-.396785228458322,-.374634056214685,-.280718813313607,-.13158670712415},
			{-.0757528111090398,-.118442209516973,-.0587723445986148,.079163297629637,.214173834111807,.252890452071526,.144805168577106,-.083343684650135,-.334186733367935,-.485704652268982,-.453371228180024,-.236159614755104},
			{.0816607421928192,.0813700708167378,-.0587165440093283,-.197744131905342,-.163742854978433,.0594832841179061,.284931226398005,.277045088013018,-.0231126420431299,-.412814367136931,-.576575034174456,-.351784826794551},
			{-.083257226162066,-.0274958626198417,.147631703568395,.14856835834851,-.0973183437761787,-.264947422944133,-.063300653277769,.296934377775289,.304763511236575,-.169083017610384,-.615979854522674,-.470942728759879},
			{.0809421348181849,-.0318345513886016,-.166678842080749,.025111331707413,.23332257818713,.00190097306183677,-.301549099493064,-.0510989036970105,.388722153430519,.147096561093246,-.555575724598485,-.586855162769535},
			{-.114490123783502,-.249388667095643,-.346998489897094,-.394969565625618,-.391157114432837,-.343355485054158,-.26692764812862,-.18078045933897,-.102686955626456,-.0451117599561129,-.0125580867203635,-.00106538514698783},
			{.133757217306151,.232222029374965,.180515058932322,-.0153369352769492,-.278061056920914,-.499125523307164,-.592730124604984,-.535718786284836,-.373308562838685,-.189219165271344,-.0578350093026769,-.00515914961555458},
			{-.144289018132812,-.166811032438666,.050360605399021,.325079179263335,.384413800291541,.112271677012304,-.33819126429111,-.668701504869756,-.683103250662319,-.435514757768147,-.152912459621155,-.0146527768052979},
			{.14725660667997,.0698096725195432,-.237071388762475,-.308935398296182,.0630284952156476,.464687444293257,.343844423679318,-.26960620296335,-.771074782139231,-.71439621956824,-.3046523541133,-.0320836517985888},
			{-.143364315256757,.0383667962154668,.297907873310876,.017167748400487,-.407478174094171,-.142359840909465,.485359330372213,.377362073254748,-.467946339429932,-.904156542888517,-.505132917252383,-.0599392486348982},
			{-.179355789242482,-.375413954291231,-.485531016712943,-.49564139368943,-.422983905496475,-.305440495868931,-.184597475250337,-.0903492073046713,-.0335020098285881,-.0082061858887987,-.000951927643970896,-1.55307570187091e-005},
			{.193736308438081,.288015724966542,.109134393057063,-.254330496941975,-.592252776234179,-.727139754226691,-.626316369393508,-.395126415458444,-.176232920497942,-.0492229675323306,-.00622627098487373,-.000106442860825517},
			{-.198099197414935,-.152996410947114,.228334442676318,.491867092554372,.251527912283571,-.350479915627058,-.80871599529964,-.804353442629352,-.474793812063302,-.159269954185708,-.0226099452664103,-.000410761975881334},
			{.193384472331989,-.00123028710976741,-.391861638259147,-.191314758984308,.444345447527085,.522939940635989,-.220026328987702,-.907628719438496,-.84713976015901,-.365816490950941,-.0602141413953284,-.00117961264325513},
			{-.242787406459337,-.488323349567532,-.587042784922414,-.537446066425504,-.395239038357479,-.234786231237537,-.110311638598763,-.0390176033630983,-.00944473608571052,-.00128990080067228,-6.23516257983926e-005,-1.95632945654767e-007},
			{.248920876546637,.31039455298989,-.0232591344869702,-.514998289242352,-.815181476789607,-.77081506036106,-.501166445756355,-.224887145281521,-.0648664986069148,-.0100439635929164,-.000527611424168513,-1.73164514361146e-006},
			{-.243912856990113,-.100226736373669,.411961635922747,.516784026765791,-.0839003911851568,-.797690550622854,-.966967837799883,-.623492271451269,-.228699162378071,-.041665014660205,-.00242929420722927,-8.42792264012015e-006},
			{-.304951495361639,-.589384113245934,-.65859190093498,-.540749420569416,-.34268072739361,-.1674605403216,-.0611660970770167,-.0156347408665742,-.00247060119508803,-.000188133064231527,-3.78953113918537e-006,-2.28657407372626e-009},
			{.300178666892258,.305157489809165,-.191570776483383,-.752151516427833,-.929808404899219,-.699144109902361,-.347669796815464,-.111720678797647,-.0209212071410637,-.00180008099494654,-3.9323064981717e-005,-2.47941342577084e-008},
			{-.365909637710154,-.679560034193271,-.705832231568849,-.519751446470816,-.283829803146802,-.114101360142282,-.0323995272485856,-.00598493248162283,-.000617381962821684,-2.62127397601647e-005,-2.20019739940013e-007,-2.55305407260627e-011}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0471753389703784,-.106939314428477,-.16007833916176,-.203167424021368,-.233492536723722,-.249147045798746,-.249147045798746,-.233492536723722,-.203167424021368,-.16007833916176,-.106939314428477,-.0471753389703784},
			{.0648707029602942,.129483784537873,.148249097989688,.109463102050698,.0170873715800573,-.109985117239921,-.2423620138609,-.347295683640528,-.396785228458322,-.374634056214685,-.280718813313607,-.13158670712415},
			{-.0757528111090398,-.118442209516973,-.0587723445986148,.079163297629637,.214173834111807,.252890452071526,.144805168577106,-.083343684650135,-.334186733367935,-.485704652268982,-.453371228180024,-.236159614755104},
			{.0816607421928192,.0813700708167378,-.0587165440093283,-.197744131905342,-.163742854978433,.0594832841179061,.284931226398005,.277045088013018,-.0231126420431299,-.412814367136931,-.576575034174456,-.351784826794551},
			{-.083257226162066,-.0274958626198417,.147631703568395,.14856835834851,-.0973183437761787,-.264947422944133,-.063300653277769,.296934377775289,.304763511236575,-.169083017610384,-.615979854522674,-.470942728759879},
			{.0809421348181849,-.0318345513886016,-.166678842080749,.025111331707413,.23332257818713,.00190097306183677,-.301549099493064,-.0510989036970105,.388722153430519,.147096561093246,-.555575724598485,-.586855162769535},
			{-.114490123783502,-.249388667095643,-.346998489897094,-.394969565625618,-.391157114432837,-.343355485054158,-.26692764812862,-.18078045933897,-.102686955626456,-.0451117599561129,-.0125580867203635,-.00106538514698783},
			{.133757217306151,.232222029374965,.180515058932322,-.0153369352769492,-.278061056920914,-.499125523307164,-.592730124604984,-.535718786284836,-.373308562838685,-.189219165271344,-.0578350093026769,-.00515914961555458},
			{-.144289018132812,-.166811032438666,.050360605399021,.325079179263335,.384413800291541,.112271677012304,-.33819126429111,-.668701504869756,-.683103250662319,-.435514757768147,-.152912459621155,-.0146527768052979},
			{.14725660667997,.0698096725195432,-.237071388762475,-.308935398296182,.0630284952156476,.464687444293257,.343844423679318,-.26960620296335,-.771074782139231,-.71439621956824,-.3046523541133,-.0320836517985888},
			{-.143364315256757,.0383667962154668,.297907873310876,.017167748400487,-.407478174094171,-.142359840909465,.485359330372213,.377362073254748,-.467946339429932,-.904156542888517,-.505132917252383,-.0599392486348982},
			{-.179355789242482,-.375413954291231,-.485531016712943,-.49564139368943,-.422983905496475,-.305440495868931,-.184597475250337,-.0903492073046713,-.0335020098285881,-.0082061858887987,-.000951927643970896,-1.55307570187091e-005},
			{.193736308438081,.288015724966542,.109134393057063,-.254330496941975,-.592252776234179,-.727139754226691,-.626316369393508,-.395126415458444,-.176232920497942,-.0492229675323306,-.00622627098487373,-.000106442860825517},
			{-.198099197414935,-.152996410947114,.228334442676318,.491867092554372,.251527912283571,-.350479915627058,-.80871599529964,-.804353442629352,-.474793812063302,-.159269954185708,-.0226099452664103,-.000410761975881334},
			{.193384472331989,-.00123028710976741,-.391861638259147,-.191314758984308,.444345447527085,.522939940635989,-.220026328987702,-.907628719438496,-.84713976015901,-.365816490950941,-.0602141413953284,-.00117961264325513},
			{-.242787406459337,-.488323349567532,-.587042784922414,-.537446066425504,-.395239038357479,-.234786231237537,-.110311638598763,-.0390176033630983,-.00944473608571052,-.00128990080067228,-6.23516257983926e-005,-1.95632945654767e-007},
			{.248920876546637,.31039455298989,-.0232591344869702,-.514998289242352,-.815181476789607,-.77081506036106,-.501166445756355,-.224887145281521,-.0648664986069148,-.0100439635929164,-.000527611424168513,-1.73164514361146e-006},
			{-.243912856990113,-.100226736373669,.411961635922747,.516784026765791,-.0839003911851568,-.797690550622854,-.966967837799883,-.623492271451269,-.228699162378071,-.041665014660205,-.00242929420722927,-8.42792264012015e-006},
			{-.304951495361639,-.589384113245934,-.65859190093498,-.540749420569416,-.34268072739361,-.1674605403216,-.0611660970770167,-.0156347408665742,-.00247060119508803,-.000188133064231527,-3.78953113918537e-006,-2.28657407372626e-009},
			{.300178666892258,.305157489809165,-.191570776483383,-.752151516427833,-.929808404899219,-.699144109902361,-.347669796815464,-.111720678797647,-.0209212071410637,-.00180008099494654,-3.9323064981717e-005,-2.47941342577084e-008},
			{-.365909637710154,-.679560034193271,-.705832231568849,-.519751446470816,-.283829803146802,-.114101360142282,-.0323995272485856,-.00598493248162283,-.000617381962821684,-2.62127397601647e-005,-2.20019739940013e-007,-2.55305407260627e-011}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{.0471753389703784,.106939314428477,.16007833916176,.203167424021368,.233492536723722,.249147045798746,.249147045798746,.233492536723722,.203167424021368,.16007833916176,.106939314428477,.0471753389703784},
			{-.057777754465245,-.130973376908003,-.196055124926603,-.248828260626037,-.285968786885903,-.305141566591389,-.305141566591389,-.285968786885903,-.248828260626037,-.196055124926603,-.130973376908003,-.057777754465245},
			{.0667160041638562,.151235028775734,.226384958224997,.287322126407624,.33020831206047,.352347131100821,.352347131100821,.33020831206047,.287322126407624,.226384958224997,.151235028775734,.0667160041638562},
			{-.0745907602644774,-.169085902496444,-.253106077890569,-.32123590311126,-.369184116329077,-.393936068492185,-.393936068492185,-.369184116329077,-.32123590311126,-.253106077890569,-.169085902496444,-.0745907602644774},
			{.0817100839813177,.18522432596281,.277263816688425,.35189630093549,.404420936894292,.43153534198653,.43153534198653,.404420936894292,.35189630093549,.277263816688425,.18522432596281,.0817100839813177},
			{.080203401857055,.167464509402566,.213466153925738,.20667501558646,.148758759444076,.0540426417701356,-.0540426417701356,-.148758759444076,-.20667501558646,-.213466153925738,-.167464509402566,-.080203401857055},
			{-.0982287050422223,-.20510129892574,-.261441577102187,-.25312416525451,-.182191527610292,-.0661884483104899,.0661884483104899,.182191527610292,.25312416525451,.261441577102187,.20510129892574,.0982287050422223},
			{.113424738636514,.236830580375279,.301886729940981,.292282609999163,.210376655093866,.0764278369255381,-.0764278369255381,-.210376655093866,-.292282609999163,-.301886729940981,-.236830580375279,-.113424738636514},
			{-.126812712948275,-.264784638408991,-.337519624793518,-.326781892267586,-.235208250811417,-.0854489193610162,.0854489193610162,.235208250811417,.326781892267586,.337519624793518,.264784638408991,.126812712948275},
			{.138916366921706,.290057038677642,.369734224203667,.357971627561736,.257657729363923,.0936046012978199,-.0936046012978199,-.257657729363923,-.357971627561736,-.369734224203667,-.290057038677642,-.138916366921706},
			{-.122114057148622,-.212661854220922,-.170589599267302,-.00968885598258523,.189947318219643,.325107053658344,.325107053658344,.189947318219644,-.00968885598258477,-.170589599267302,-.212661854220921,-.122114057148622},
			{.14100516752701,.245560757550869,.196979902107224,.01118772721888,-.219332270602307,-.375401289873734,-.375401289873734,-.219332270602308,.0111877272188795,.196979902107224,.245560757550869,.14100516752701},
			{-.157648569851374,-.274545273187232,-.22023022561013,-.0125082592849363,.245220933311432,.419711401411007,.419711401411007,.245220933311433,-.0125082592849357,-.220230225610129,-.274545273187232,-.157648569851374},
			{.172695355768364,.300749278425367,.241250124872966,.0137021115338472,-.268626073550765,-.459770804486501,-.459770804486501,-.268626073550765,.0137021115338465,.241250124872965,.300749278425367,.172695355768364},
			{.157432480656601,.196644442987321,-.00835800794176406,-.284688341235506,-.373334915626667,-.170540758074208,.170540758074208,.373334915626668,.284688341235505,.00835800794176374,-.196644442987321,-.1574324806566},
			{-.176014864289488,-.219855170936379,.00934453695618862,.31829124166993,.41740112481547,.190670363974853,-.190670363974853,-.417401124815472,-.31829124166993,-.00934453695618827,.219855170936379,.176014864289487},
			{.192814623264837,.240839273024148,-.0102364273612059,-.348670585855459,-.457240023201953,-.208868918804014,.208868918804014,.457240023201955,.348670585855458,.0102364273612055,-.240839273024148,-.192814623264836},
			{-.184197562274308,-.118175770575784,.235878303029161,.383530106382951,.0579093560340826,-.374944419288365,-.374944419288369,.0579093560340806,.383530106382954,.235878303029161,-.118175770575781,-.184197562274308},
			{.201778319851171,.12945507062894,-.258391734868296,-.420136181623543,-.0634365212000975,.41073103263393,.410731032633934,-.0634365212000954,-.420136181623546,-.258391734868296,.129455070628936,.20177831985117},
			{.20076502079953,-.00861464492903962,-.385612979315631,-.141367941779104,.412110586758183,.311823666565035,-.311823666565045,-.412110586758192,.141367941779136,.38561297931557,.0086146449290491,-.200765020799505}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{.0471753389703784,.106939314428477,.16007833916176,.203167424021368,.233492536723722,.249147045798746,.249147045798746,.233492536723722,.203167424021368,.16007833916176,.106939314428477,.0471753389703784},
			{-.057777754465245,-.130973376908003,-.196055124926603,-.248828260626037,-.285968786885903,-.305141566591389,-.305141566591389,-.285968786885903,-.248828260626037,-.196055124926603,-.130973376908003,-.057777754465245},
			{.0667160041638562,.151235028775734,.226384958224997,.287322126407624,.33020831206047,.352347131100821,.352347131100821,.33020831206047,.287322126407624,.226384958224997,.151235028775734,.0667160041638562},
			{-.0745907602644774,-.169085902496444,-.253106077890569,-.32123590311126,-.369184116329077,-.393936068492185,-.393936068492185,-.369184116329077,-.32123590311126,-.253106077890569,-.169085902496444,-.0745907602644774},
			{.0817100839813177,.18522432596281,.277263816688425,.35189630093549,.404420936894292,.43153534198653,.43153534198653,.404420936894292,.35189630093549,.277263816688425,.18522432596281,.0817100839813177},
			{.080203401857055,.167464509402566,.213466153925738,.20667501558646,.148758759444076,.0540426417701356,-.0540426417701356,-.148758759444076,-.20667501558646,-.213466153925738,-.167464509402566,-.080203401857055},
			{-.0982287050422223,-.20510129892574,-.261441577102187,-.25312416525451,-.182191527610292,-.0661884483104899,.0661884483104899,.182191527610292,.25312416525451,.261441577102187,.20510129892574,.0982287050422223},
			{.113424738636514,.236830580375279,.301886729940981,.292282609999163,.210376655093866,.0764278369255381,-.0764278369255381,-.210376655093866,-.292282609999163,-.301886729940981,-.236830580375279,-.113424738636514},
			{-.126812712948275,-.264784638408991,-.337519624793518,-.326781892267586,-.235208250811417,-.0854489193610162,.0854489193610162,.235208250811417,.326781892267586,.337519624793518,.264784638408991,.126812712948275},
			{.138916366921706,.290057038677642,.369734224203667,.357971627561736,.257657729363923,.0936046012978199,-.0936046012978199,-.257657729363923,-.357971627561736,-.369734224203667,-.290057038677642,-.138916366921706},
			{-.122114057148622,-.212661854220922,-.170589599267302,-.00968885598258523,.189947318219643,.325107053658344,.325107053658344,.189947318219644,-.00968885598258477,-.170589599267302,-.212661854220921,-.122114057148622},
			{.14100516752701,.245560757550869,.196979902107224,.01118772721888,-.219332270602307,-.375401289873734,-.375401289873734,-.219332270602308,.0111877272188795,.196979902107224,.245560757550869,.14100516752701},
			{-.157648569851374,-.274545273187232,-.22023022561013,-.0125082592849363,.245220933311432,.419711401411007,.419711401411007,.245220933311433,-.0125082592849357,-.220230225610129,-.274545273187232,-.157648569851374},
			{.172695355768364,.300749278425367,.241250124872966,.0137021115338472,-.268626073550765,-.459770804486501,-.459770804486501,-.268626073550765,.0137021115338465,.241250124872965,.300749278425367,.172695355768364},
			{.157432480656601,.196644442987321,-.00835800794176406,-.284688341235506,-.373334915626667,-.170540758074208,.170540758074208,.373334915626668,.284688341235505,.00835800794176374,-.196644442987321,-.1574324806566},
			{-.176014864289488,-.219855170936379,.00934453695618862,.31829124166993,.41740112481547,.190670363974853,-.190670363974853,-.417401124815472,-.31829124166993,-.00934453695618827,.219855170936379,.176014864289487},
			{.192814623264837,.240839273024148,-.0102364273612059,-.348670585855459,-.457240023201953,-.208868918804014,.208868918804014,.457240023201955,.348670585855458,.0102364273612055,-.240839273024148,-.192814623264836},
			{-.184197562274308,-.118175770575784,.235878303029161,.383530106382951,.0579093560340826,-.374944419288365,-.374944419288369,.0579093560340806,.383530106382954,.235878303029161,-.118175770575781,-.184197562274308},
			{.201778319851171,.12945507062894,-.258391734868296,-.420136181623543,-.0634365212000975,.41073103263393,.410731032633934,-.0634365212000954,-.420136181623546,-.258391734868296,.129455070628936,.20177831985117},
			{.20076502079953,-.00861464492903962,-.385612979315631,-.141367941779104,.412110586758183,.311823666565035,-.311823666565045,-.412110586758192,.141367941779136,.38561297931557,.0086146449290491,-.200765020799505}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{-.0930458529460421,-.198498176553059,-.264906281682713,-.280569525791552,-.245575133043822,-.171375823548283,-.0777712222504629,.0120825963201004,.0774021017701836,.104827942520953,.0915588621245824,.0458705139756637},
			{-.166990065079791,-.320581869925543,-.343445053363858,-.236305705409389,-.0589328846007312,.102392716678746,.178820553604284,.151443770493135,.055976904589773,-.0415583234228769,-.0837512895502642,-.0535653264432772},
			{-.248749436479337,-.407700116420065,-.291903838296748,-.0163431059155229,.195900460376747,.201476802304676,.042061033555921,-.11578368309555,-.139826216573226,-.041518866425881,.0575373288449655,.0577428645460353},
			{-.33300679714446,-.435563532322195,-.119559748367329,.215500345510458,.209964312147707,-.0447603211980575,-.187346119471122,-.0688144608361341,.105053693685699,.104391378738875,-.0194425109181931,-.0588717492175104},
			{-.41496926527815,-.39285136242986,.104012975865696,.274868070760673,-.0361323813248869,-.213227413168496,.00134419094323235,.164983977283574,.017756392939621,-.117859739546712,-.0225104271688505,.0572347324287539},
			{.000753341062131352,.008879908280122,.0318988313813435,.072610642674515,.127831088725108,.188746350108197,.242788991878333,.276589848169184,.279285658260975,.245364985307081,.176344417682688,.0809567429191864},
			{.00364806967735219,.0408955272571209,.133798154858527,.263969016188599,.378810386491097,.419123490411159,.352935042100669,.196618858880805,.0108448509340889,-.12764342224366,-.164205771668619,-.0945806353648701},
			{.0103610778435682,.108125437139903,.307955438564179,.483026940855864,.472843368743711,.239137336348973,-.0793880641607704,-.27182160500271,-.22986569210915,-.0356103255868759,.117953212229241,.102027743185548},
			{.0226865677588688,.215421745563086,.505154411463456,.545232207417426,.190640374422976,-.243134723730331,-.328583643091346,-.0445678763884421,.21845031514984,.167634786669938,-.0493628928459047,-.104126145189406},
			{.0423834491801445,.357182911283934,.639335222899324,.330888029929621,-.266835281031443,-.343200873908896,.100663608902278,.288130580163515,-.0121394313148915,-.210652677442557,-.0271294217835161,.101373879524963},
			{-1.09819036077875e-005,-.000673114492428329,-.00580264969117775,-.0236894983394228,-.0638865371768165,-.130530126573866,-.215979045934882,-.299094787988234,-.350471390607008,-.343322274484695,-.265457752901419,-.126823694851883},
			{-7.52664687184791e-005,-.00440263843607072,-.0348058941414186,-.12461549318528,-.279396567870308,-.442872552083127,-.514165451219659,-.41878595436223,-.17983881909766,.0771696694116863,.20365787226593,.136992257494756},
			{-.000290452578522664,-.0159876456159174,-.112620864614273,-.335729924086801,-.568763773603908,-.571848564179541,-.247826724944201,.177857092386487,.347802556495951,.16145683275228,-.108184799649367,-.140077285802758},
			{-.000834112099439185,-.0425778277151948,-.25867132148952,-.599018269179214,-.641790422483938,-.155582109307832,.369774378274552,.314199679218665,-.135279963454564,-.277088021773013,-.000869944358352441,.136743471798207},
			{1.38333382588974e-007,4.40892574312313e-005,.00091209760345402,.00667843693448509,.0275896119309721,.0780021077175621,.166018936281095,.279476204285956,.380031758191769,.415101934174787,.345296751982027,.171676621539384},
			{1.22445802447951e-006,.000373077615955666,.00710215476841621,.0458675410488779,.159019225472198,.354378192390965,.545048556365818,.576420350287669,.36415878271881,.0164466917246055,-.219482093320422,-.176013639831463},
			{5.95944125415059e-006,.00171777040786467,.0294616144122425,.161714728611881,.440875613276901,.683749515477967,.56405239778266,.0593265355668802,-.365421489831386,-.291300866426529,.0708710049647147,.172472435241796},
			{-1.61685195173712e-009,-2.67960316688328e-006,-.000133030165518613,-.00174697885911607,-.0110554312917647,-.043250962033283,-.118412483673806,-.242311866185891,-.382367582308209,-.465694799308523,-.416757503309762,-.215633270360082},
			{-1.7532099583758e-008,-2.78056059144079e-005,-.00127284947854753,-.0147935274439646,-.0789984495974143,-.245839671006818,-.494369541269015,-.657473828481949,-.531851437886183,-.13546099516431,.215778930430848,.212258370983052},
			{1.80533017556786e-011,1.55577451647998e-007,1.85352060386969e-005,.000436554972610718,.00423198634381799,.0229099254307547,.0806818455205034,.200697978560845,.367519772428026,.499098757454024,.480521508528172,.258737186194613}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[12],
		T o_output[21]
	) {
		static const T matrix[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{-.0930458529460421,-.198498176553059,-.264906281682713,-.280569525791552,-.245575133043822,-.171375823548283,-.0777712222504629,.0120825963201004,.0774021017701836,.104827942520953,.0915588621245824,.0458705139756637},
			{-.166990065079791,-.320581869925543,-.343445053363858,-.236305705409389,-.0589328846007312,.102392716678746,.178820553604284,.151443770493135,.055976904589773,-.0415583234228769,-.0837512895502642,-.0535653264432772},
			{-.248749436479337,-.407700116420065,-.291903838296748,-.0163431059155229,.195900460376747,.201476802304676,.042061033555921,-.11578368309555,-.139826216573226,-.041518866425881,.0575373288449655,.0577428645460353},
			{-.33300679714446,-.435563532322195,-.119559748367329,.215500345510458,.209964312147707,-.0447603211980575,-.187346119471122,-.0688144608361341,.105053693685699,.104391378738875,-.0194425109181931,-.0588717492175104},
			{-.41496926527815,-.39285136242986,.104012975865696,.274868070760673,-.0361323813248869,-.213227413168496,.00134419094323235,.164983977283574,.017756392939621,-.117859739546712,-.0225104271688505,.0572347324287539},
			{.000753341062131352,.008879908280122,.0318988313813435,.072610642674515,.127831088725108,.188746350108197,.242788991878333,.276589848169184,.279285658260975,.245364985307081,.176344417682688,.0809567429191864},
			{.00364806967735219,.0408955272571209,.133798154858527,.263969016188599,.378810386491097,.419123490411159,.352935042100669,.196618858880805,.0108448509340889,-.12764342224366,-.164205771668619,-.0945806353648701},
			{.0103610778435682,.108125437139903,.307955438564179,.483026940855864,.472843368743711,.239137336348973,-.0793880641607704,-.27182160500271,-.22986569210915,-.0356103255868759,.117953212229241,.102027743185548},
			{.0226865677588688,.215421745563086,.505154411463456,.545232207417426,.190640374422976,-.243134723730331,-.328583643091346,-.0445678763884421,.21845031514984,.167634786669938,-.0493628928459047,-.104126145189406},
			{.0423834491801445,.357182911283934,.639335222899324,.330888029929621,-.266835281031443,-.343200873908896,.100663608902278,.288130580163515,-.0121394313148915,-.210652677442557,-.0271294217835161,.101373879524963},
			{-1.09819036077875e-005,-.000673114492428329,-.00580264969117775,-.0236894983394228,-.0638865371768165,-.130530126573866,-.215979045934882,-.299094787988234,-.350471390607008,-.343322274484695,-.265457752901419,-.126823694851883},
			{-7.52664687184791e-005,-.00440263843607072,-.0348058941414186,-.12461549318528,-.279396567870308,-.442872552083127,-.514165451219659,-.41878595436223,-.17983881909766,.0771696694116863,.20365787226593,.136992257494756},
			{-.000290452578522664,-.0159876456159174,-.112620864614273,-.335729924086801,-.568763773603908,-.571848564179541,-.247826724944201,.177857092386487,.347802556495951,.16145683275228,-.108184799649367,-.140077285802758},
			{-.000834112099439185,-.0425778277151948,-.25867132148952,-.599018269179214,-.641790422483938,-.155582109307832,.369774378274552,.314199679218665,-.135279963454564,-.277088021773013,-.000869944358352441,.136743471798207},
			{1.38333382588974e-007,4.40892574312313e-005,.00091209760345402,.00667843693448509,.0275896119309721,.0780021077175621,.166018936281095,.279476204285956,.380031758191769,.415101934174787,.345296751982027,.171676621539384},
			{1.22445802447951e-006,.000373077615955666,.00710215476841621,.0458675410488779,.159019225472198,.354378192390965,.545048556365818,.576420350287669,.36415878271881,.0164466917246055,-.219482093320422,-.176013639831463},
			{5.95944125415059e-006,.00171777040786467,.0294616144122425,.161714728611881,.440875613276901,.683749515477967,.56405239778266,.0593265355668802,-.365421489831386,-.291300866426529,.0708710049647147,.172472435241796},
			{-1.61685195173712e-009,-2.67960316688328e-006,-.000133030165518613,-.00174697885911607,-.0110554312917647,-.043250962033283,-.118412483673806,-.242311866185891,-.382367582308209,-.465694799308523,-.416757503309762,-.215633270360082},
			{-1.7532099583758e-008,-2.78056059144079e-005,-.00127284947854753,-.0147935274439646,-.0789984495974143,-.245839671006818,-.494369541269015,-.657473828481949,-.531851437886183,-.13546099516431,.215778930430848,.212258370983052},
			{1.80533017556786e-011,1.55577451647998e-007,1.85352060386969e-005,.000436554972610718,.00423198634381799,.0229099254307547,.0806818455205034,.200697978560845,.367519772428026,.499098757454024,.480521508528172,.258737186194613}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 12; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,33.98823325,-144.1116465,-.4082482906,1.750000000,-7.678758582,34.53410150,-159.0586197,.3227486122,-2.459674775,14.75000000,-81.71498915,-.2645751311,2.958039892,-22.92947780,.2236067977,-3.324307508,-.1934294858},
			{0.,-.7500000000,2.857738034,-12.37436867,55.33985905,-254.3227936,-.7216878367,2.828427124,-13.26806944,63.17066828,-305.0714286,.9128709293,-5.850213671,33.43604921,-182.9477455,-.9354143468,8.844691708,-63.03769304,.9035079028,-11.53669556,-.8548456319},
			{0.,0.,1.166666667,-7.216878367,40.02082790,-212.3677366,0.,1.587713241,-9.500000000,52.88300766,-284.1991313,.9316949905,-6.067673910,36.47616523,-208.0853992,-1.636634177,13.68699853,-93.25932554,2.074812507,-23.01464213,-2.326595118},
			{0.,0.,0.,-1.875000000,15.87608264,-106.7569280,0.,0.,-2.901185103,23.88339730,-158.7252693,0.,-2.627379873,21.21428571,-141.7297322,-1.218935425,11.40958244,-83.44725167,2.747169230,-28.60654093,-4.158733944},
			{0.,0.,0.,0.,3.100000000,-33.04592764,0.,0.,0.,5.080682370,-53.17144007,0.,0.,5.494338457,-55.95917640,0.,4.138710979,-42.40452863,1.621428571,-20.30485767,-4.462835264},
			{0.,0.,0.,0.,0.,-5.250000000,0.,0.,0.,0.,-8.845830912,0.,0.,0.,-10.42167396,0.,0.,-9.706337648,0.,-6.398809524,-2.185359011},
			{0.,.4330127020,-.7071067810,2.245365598,-7.850689990,29.50000000,-.2500000000,0.,.4949747467,-3.162277660,16.25653401,.3162277660,-1.004158022,3.091022771,-9.096005092,-.3240370350,2.070196679,-10.27118455,.3129843186,-3.053571429,-.2961272134},
			{0.,0.,-.8660254040,3.500000000,-14.98165545,64.91147819,0.,-.2500000000,1.558845727,-8.714212528,44.99219432,.5809475019,-1.565247584,4.964285714,-14.04517129,-1.020504077,5.704791219,-27.63572997,1.293725044,-11.16850681,-1.450721144},
			{0.,0.,0.,1.659882024,-10.71525393,59.18483757,0.,0.,1.275000000,-8.720665110,51.70172994,0.,-.5486726408,1.773290113,-2.439471337,-.9983468483,5.709015429,-29.57691077,2.250018896,-17.91000460,-3.406135253},
			{0.,0.,0.,0.,-3.117691454,27.51181564,0.,0.,0.,-3.300000000,29.34227987,0.,0.,-.4979550017,6.515483910,0.,1.996693696,-13.84473139,1.554721797,-14.02696019,-4.279230906},
			{0.,0.,0.,0.,0.,5.802370207,0.,0.,0.,0.,7.221428571,0.,0.,0.,3.471852928,0.,0.,-1.980327354,0.,-4.377552220,-2.322071589},
			{0.,0.,.3726779962,-.6454972245,2.020725943,-6.851601598,0.,-.3227486122,.2236067977,-.2500000000,-.4433944513,.8333333333e-1,.2886751347,-1.152673615,4.225804810,-.1463850110,.1636634177,.5511331920,.1855768722,-.8658617402,-.2080969936},
			{0.,0.,0.,-.8385254914,3.500000000,-14.87979615,0.,0.,.9682458365e-1,.2886751347,-3.546268517,0.,.6250000000,-2.395787117,9.681316602,-.2746751327,-.1889822365,3.496332169,.6190476190,-2.256095297,-.9371298718},
			{0.,0.,0.,0.,1.788854382,-11.83920042,0.,0.,0.,.6454972245,-5.353808485,0.,0.,-1.357142857,7.368172496,0.,-.6338656908,5.580637822,.6016087651,-1.545511386,-1.655873628},
			{0.,0.,0.,0.,0.,-3.689512162,0.,0.,0.,0.,-2.517439175,0.,0.,0.,1.767857143,0.,0.,2.833614404,0.,.9849347042e-1,-1.114302550},
			{0.,0.,0.,.3307189139,-.5916079782,1.836209865,0.,0.,-.3436931772,.3415650256,-.6681531049,0.,.1479019946,.1889822365,-.9660917834,-.2500000000e-1,-.2236067977,.5443310539,.5634361697e-1,.1080123450,-.8529438595e-1},
			{0.,0.,0.,0.,-.7937253933,3.381321241,0.,0.,0.,.3055050464,-.5976143047,0.,0.,.5070925528,-2.391701925,0.,-.4500000000,.7607257745,.1070899340,.6555622816,-.2947553424},
			{0.,0.,0.,0.,0.,1.807930062,0.,0.,0.,0.,.1200198396,0.,0.,0.,-1.598750131,0.,0.,.3055555556e-1,0.,.8094739131,-.2864842269},
			{0.,0.,0.,0.,.3000000000,-.5477225576,0.,0.,0.,-.3464101616,.4065785564,0.,0.,.1916629694,.8748177653e-1,0.,-.5669467095e-1,-.2415229459,.7142857143e-2,.1173691195,-.1966006724e-1},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.4330127020,0.,0.,0.,.3593680677,0.,0.,-.5197011503,0.,.2321428571,-.3768891806e-1},
			{0.,0.,0.,0.,0.,.2763853992,0.,0.,0.,0.,-.3419382528,0.,0.,0.,.2207201930,0.,0.,-.8705321812e-1,0.,.1974181422e-1,-.1984126984e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,33.98823325,-144.1116465,-.4082482906,1.750000000,-7.678758582,34.53410150,-159.0586197,.3227486122,-2.459674775,14.75000000,-81.71498915,-.2645751311,2.958039892,-22.92947780,.2236067977,-3.324307508,-.1934294858},
			{0.,-.7500000000,2.857738034,-12.37436867,55.33985905,-254.3227936,-.7216878367,2.828427124,-13.26806944,63.17066828,-305.0714286,.9128709293,-5.850213671,33.43604921,-182.9477455,-.9354143468,8.844691708,-63.03769304,.9035079028,-11.53669556,-.8548456319},
			{0.,0.,1.166666667,-7.216878367,40.02082790,-212.3677366,0.,1.587713241,-9.500000000,52.88300766,-284.1991313,.9316949905,-6.067673910,36.47616523,-208.0853992,-1.636634177,13.68699853,-93.25932554,2.074812507,-23.01464213,-2.326595118},
			{0.,0.,0.,-1.875000000,15.87608264,-106.7569280,0.,0.,-2.901185103,23.88339730,-158.7252693,0.,-2.627379873,21.21428571,-141.7297322,-1.218935425,11.40958244,-83.44725167,2.747169230,-28.60654093,-4.158733944},
			{0.,0.,0.,0.,3.100000000,-33.04592764,0.,0.,0.,5.080682370,-53.17144007,0.,0.,5.494338457,-55.95917640,0.,4.138710979,-42.40452863,1.621428571,-20.30485767,-4.462835264},
			{0.,0.,0.,0.,0.,-5.250000000,0.,0.,0.,0.,-8.845830912,0.,0.,0.,-10.42167396,0.,0.,-9.706337648,0.,-6.398809524,-2.185359011},
			{0.,.4330127020,-.7071067810,2.245365598,-7.850689990,29.50000000,-.2500000000,0.,.4949747467,-3.162277660,16.25653401,.3162277660,-1.004158022,3.091022771,-9.096005092,-.3240370350,2.070196679,-10.27118455,.3129843186,-3.053571429,-.2961272134},
			{0.,0.,-.8660254040,3.500000000,-14.98165545,64.91147819,0.,-.2500000000,1.558845727,-8.714212528,44.99219432,.5809475019,-1.565247584,4.964285714,-14.04517129,-1.020504077,5.704791219,-27.63572997,1.293725044,-11.16850681,-1.450721144},
			{0.,0.,0.,1.659882024,-10.71525393,59.18483757,0.,0.,1.275000000,-8.720665110,51.70172994,0.,-.5486726408,1.773290113,-2.439471337,-.9983468483,5.709015429,-29.57691077,2.250018896,-17.91000460,-3.406135253},
			{0.,0.,0.,0.,-3.117691454,27.51181564,0.,0.,0.,-3.300000000,29.34227987,0.,0.,-.4979550017,6.515483910,0.,1.996693696,-13.84473139,1.554721797,-14.02696019,-4.279230906},
			{0.,0.,0.,0.,0.,5.802370207,0.,0.,0.,0.,7.221428571,0.,0.,0.,3.471852928,0.,0.,-1.980327354,0.,-4.377552220,-2.322071589},
			{0.,0.,.3726779962,-.6454972245,2.020725943,-6.851601598,0.,-.3227486122,.2236067977,-.2500000000,-.4433944513,.8333333333e-1,.2886751347,-1.152673615,4.225804810,-.1463850110,.1636634177,.5511331920,.1855768722,-.8658617402,-.2080969936},
			{0.,0.,0.,-.8385254914,3.500000000,-14.87979615,0.,0.,.9682458365e-1,.2886751347,-3.546268517,0.,.6250000000,-2.395787117,9.681316602,-.2746751327,-.1889822365,3.496332169,.6190476190,-2.256095297,-.9371298718},
			{0.,0.,0.,0.,1.788854382,-11.83920042,0.,0.,0.,.6454972245,-5.353808485,0.,0.,-1.357142857,7.368172496,0.,-.6338656908,5.580637822,.6016087651,-1.545511386,-1.655873628},
			{0.,0.,0.,0.,0.,-3.689512162,0.,0.,0.,0.,-2.517439175,0.,0.,0.,1.767857143,0.,0.,2.833614404,0.,.9849347042e-1,-1.114302550},
			{0.,0.,0.,.3307189139,-.5916079782,1.836209865,0.,0.,-.3436931772,.3415650256,-.6681531049,0.,.1479019946,.1889822365,-.9660917834,-.2500000000e-1,-.2236067977,.5443310539,.5634361697e-1,.1080123450,-.8529438595e-1},
			{0.,0.,0.,0.,-.7937253933,3.381321241,0.,0.,0.,.3055050464,-.5976143047,0.,0.,.5070925528,-2.391701925,0.,-.4500000000,.7607257745,.1070899340,.6555622816,-.2947553424},
			{0.,0.,0.,0.,0.,1.807930062,0.,0.,0.,0.,.1200198396,0.,0.,0.,-1.598750131,0.,0.,.3055555556e-1,0.,.8094739131,-.2864842269},
			{0.,0.,0.,0.,.3000000000,-.5477225576,0.,0.,0.,-.3464101616,.4065785564,0.,0.,.1916629694,.8748177653e-1,0.,-.5669467095e-1,-.2415229459,.7142857143e-2,.1173691195,-.1966006724e-1},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.4330127020,0.,0.,0.,.3593680677,0.,0.,-.5197011503,0.,.2321428571,-.3768891806e-1},
			{0.,0.,0.,0.,0.,.2763853992,0.,0.,0.,0.,-.3419382528,0.,0.,0.,.2207201930,0.,0.,-.8705321812e-1,0.,.1974181422e-1,-.1984126984e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,-.6454972245e-1,-.4040610177e-1,2.904737510,1.341640786,.7500000000,.4694764777,-11.64130577,-6.507687761,-4.073608438,50.31152948,31.49343956,-228.6336522},
			{0.,-.2500000000,0.,0.,0.,0.,.7216878367,.3535533905,.1632993162,.9128709293e-1,.5714285714e-1,-2.464751509,-1.581138830,-.9596449171,-.6224437520,10.85080642,7.410417377,5.024466921,-50.14468862,-35.90912479,239.3225832},
			{0.,0.,.1666666667,0.,0.,0.,0.,-.4330127020,-.2000000000,-.1118033988,-.6998542123e-1,1.155301788,1.290994449,.9691236663,.6776309270,-6.197388082,-6.026182953,-4.798824078,33.05867928,29.90051531,-173.7315107},
			{0.,0.,0.,-.1250000000,0.,0.,0.,0.,.3031088914,.1290994449,.8081220354e-1,0.,-.7267220925,-.7857142857,-.6455301571,1.974864371,3.507390157,3.710995566,-14.50782199,-18.91064403,91.34707467},
			{0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,-.2309401077,-.9035079028e-1,0.,0.,.5111012519,.5248906592,0.,-1.266180985,-2.208209791,3.502380952,8.554904710,-32.70648785},
			{0.,0.,0.,0.,0.,-.8333333333e-1,0.,0.,0.,0.,.1855768722,0.,0.,0.,-.3859879246,0.,0.,.8871666102,0.,-2.267857143,6.356265942},
			{0.,-.4330127020,0.,0.,0.,0.,-.7500000000,.6123724358,.2828427124,.1581138830,.9897433188e-1,3.320391543,-1.278019301,-1.195584279,-.8864412337,-15.76980237,1.794170455,4.167058315,75.89869728,4.660714286,-371.5212018},
			{0.,0.,.2886751347,0.,0.,0.,0.,-.2500000000,-.3464101616,-.1936491673,-.1212183053,-1.484643616,.4472135954,1.107142857,.9389529552,8.919961562,-.8451542550e-1,-3.528403268,-49.83237207,-7.640075151,269.1248912},
			{0.,0.,0.,-.2165063510,0.,0.,0.,0.,.3250000000,.2236067977,.1399708425,0.,.3227486122e-1,-.7010681843,-.8470386588,-2.776822654,-.7563225568,2.151452997,21.71637090,7.490281115,-141.0307509},
			{0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,-.3000000000,-.1564921593,0.,0.,.2766416676,.6060915266,0.,.5237229368,-.8366600265,-5.171408842,-4.562714910,50.23385756},
			{0.,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,.2642857143,0.,0.,0.,-.3365806955,0.,0.,.1400231463,0.,1.482553013,-9.676023611},
			{0.,0.,.3726779962,0.,0.,0.,0.,.9682458365,-.4472135954,-.2500000000,-.1564921593,1.083333333,-4.041451885,-.4610694458e-1,.6060915266,-8.002380602,16.80277755,5.405089377,48.86857636,-68.44072363,-278.3505387},
			{0.,0.,0.,-.2795084971,0.,0.,0.,0.,-.9682458365e-1,.2886751347,.1807015806,0.,1.375000000,.7985957060,-.3936679944,2.387560769,-7.275816105,-5.503272838,-21.01190476,34.70865684,144.9163216},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,-.1290994449,-.2020305089,0.,0.,-.6428571429,0.,0.,1.774823935,3.548977048,4.876757779,-11.83920042,-51.10055324},
			{0.,0.,0.,0.,0.,-.1863389981,0.,0.,0.,0.,.1936491673,0.,0.,0.,.2083333333,0.,0.,-1.312336745,0.,2.180166278,9.676640803},
			{0.,0.,0.,-.3307189139,0.,0.,0.,0.,-1.031079532,.3415650256,.2138089935,0.,-1.922725929,3.968626966,.7763237544,-1.575000000,13.75181806,-13.04353288,16.26921940,-78.67156297,-122.2865611},
			{0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,.3055050464,-.2390457218,0.,0.,-.8451542550,-1.388730150,0.,-3.450000000,3.286335346,-3.647357166,28.56871416,42.50718808},
			{0.,0.,0.,0.,0.,-.2204792759,0.,0.,0.,0.,-.3273268354e-1,0.,0.,0.,.7535958771,0.,0.,-.1750000000,0.,-5.735610878,-7.860430760},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,1.039230485,-.2710523708,0.,0.,2.491618602,-3.674234614,0.,3.571764270,-16.97561276,2.292857143,-36.14968880,-30.41019201},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.2795084971,0.,0.,3.354434698,0.,7.125000000,5.464893121},
			{0.,0.,0.,0.,0.,-.2763853992,0.,0.,0.,0.,-1.025814759,0.,0.,0.,-2.869362509,0.,0.,-5.484352741,0.,-6.337122368,-3.339285714}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,-.6454972245e-1,-.4040610177e-1,2.904737510,1.341640786,.7500000000,.4694764777,-11.64130577,-6.507687761,-4.073608438,50.31152948,31.49343956,-228.6336522},
			{0.,-.2500000000,0.,0.,0.,0.,.7216878367,.3535533905,.1632993162,.9128709293e-1,.5714285714e-1,-2.464751509,-1.581138830,-.9596449171,-.6224437520,10.85080642,7.410417377,5.024466921,-50.14468862,-35.90912479,239.3225832},
			{0.,0.,.1666666667,0.,0.,0.,0.,-.4330127020,-.2000000000,-.1118033988,-.6998542123e-1,1.155301788,1.290994449,.9691236663,.6776309270,-6.197388082,-6.026182953,-4.798824078,33.05867928,29.90051531,-173.7315107},
			{0.,0.,0.,-.1250000000,0.,0.,0.,0.,.3031088914,.1290994449,.8081220354e-1,0.,-.7267220925,-.7857142857,-.6455301571,1.974864371,3.507390157,3.710995566,-14.50782199,-18.91064403,91.34707467},
			{0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,-.2309401077,-.9035079028e-1,0.,0.,.5111012519,.5248906592,0.,-1.266180985,-2.208209791,3.502380952,8.554904710,-32.70648785},
			{0.,0.,0.,0.,0.,-.8333333333e-1,0.,0.,0.,0.,.1855768722,0.,0.,0.,-.3859879246,0.,0.,.8871666102,0.,-2.267857143,6.356265942},
			{0.,-.4330127020,0.,0.,0.,0.,-.7500000000,.6123724358,.2828427124,.1581138830,.9897433188e-1,3.320391543,-1.278019301,-1.195584279,-.8864412337,-15.76980237,1.794170455,4.167058315,75.89869728,4.660714286,-371.5212018},
			{0.,0.,.2886751347,0.,0.,0.,0.,-.2500000000,-.3464101616,-.1936491673,-.1212183053,-1.484643616,.4472135954,1.107142857,.9389529552,8.919961562,-.8451542550e-1,-3.528403268,-49.83237207,-7.640075151,269.1248912},
			{0.,0.,0.,-.2165063510,0.,0.,0.,0.,.3250000000,.2236067977,.1399708425,0.,.3227486122e-1,-.7010681843,-.8470386588,-2.776822654,-.7563225568,2.151452997,21.71637090,7.490281115,-141.0307509},
			{0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,-.3000000000,-.1564921593,0.,0.,.2766416676,.6060915266,0.,.5237229368,-.8366600265,-5.171408842,-4.562714910,50.23385756},
			{0.,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,.2642857143,0.,0.,0.,-.3365806955,0.,0.,.1400231463,0.,1.482553013,-9.676023611},
			{0.,0.,.3726779962,0.,0.,0.,0.,.9682458365,-.4472135954,-.2500000000,-.1564921593,1.083333333,-4.041451885,-.4610694458e-1,.6060915266,-8.002380602,16.80277755,5.405089377,48.86857636,-68.44072363,-278.3505387},
			{0.,0.,0.,-.2795084971,0.,0.,0.,0.,-.9682458365e-1,.2886751347,.1807015806,0.,1.375000000,.7985957060,-.3936679944,2.387560769,-7.275816105,-5.503272838,-21.01190476,34.70865684,144.9163216},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,-.1290994449,-.2020305089,0.,0.,-.6428571429,0.,0.,1.774823935,3.548977048,4.876757779,-11.83920042,-51.10055324},
			{0.,0.,0.,0.,0.,-.1863389981,0.,0.,0.,0.,.1936491673,0.,0.,0.,.2083333333,0.,0.,-1.312336745,0.,2.180166278,9.676640803},
			{0.,0.,0.,-.3307189139,0.,0.,0.,0.,-1.031079532,.3415650256,.2138089935,0.,-1.922725929,3.968626966,.7763237544,-1.575000000,13.75181806,-13.04353288,16.26921940,-78.67156297,-122.2865611},
			{0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,.3055050464,-.2390457218,0.,0.,-.8451542550,-1.388730150,0.,-3.450000000,3.286335346,-3.647357166,28.56871416,42.50718808},
			{0.,0.,0.,0.,0.,-.2204792759,0.,0.,0.,0.,-.3273268354e-1,0.,0.,0.,.7535958771,0.,0.,-.1750000000,0.,-5.735610878,-7.860430760},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,1.039230485,-.2710523708,0.,0.,2.491618602,-3.674234614,0.,3.571764270,-16.97561276,2.292857143,-36.14968880,-30.41019201},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.2795084971,0.,0.,3.354434698,0.,7.125000000,5.464893121},
			{0.,0.,0.,0.,0.,-.2763853992,0.,0.,0.,0.,-1.025814759,0.,0.,0.,-2.869362509,0.,0.,-5.484352741,0.,-6.337122368,-3.339285714}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,-.5590169942e-1,-.5103103630e-1,.4082482906,.1250000000,-.2453738645,.3227486122e-1,.1035406358,-.1613743061,.5590169942e-1,.1250000000,-.1189992461,.6614378278e-1,-.7395099729e-1,-.2507429436e-1,-.2795084971e-1,.5467611033e-1,.1208934286e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,0.,.9923207755e-1,.7216878367,-.7071067810,.3163924252,.9128709293e-1,-.1638392857,.3651483717,.3952847075e-1,-.2525381361,.1276586028,-.1403121520,.8964214566e-1,.8558232884e-1,.5646924394e-1,-.8312297405e-1,-.2350825488e-1},
			{0.,0.,.8333333333e-1,-.2165063510,.2420614591,-.1031197389,0.,-.2165063510,.4250000000,-.3913118960,.1159133539,.5776508940,-.6132223630,.3608439183,.1364673394e-1,.3218713882,-.3049687729e-2,-.2191913380,-.1210307296,.9785852773e-1,.4769519990e-1},
			{0.,0.,0.,-.3125000000e-1,.1118033988,-.1786086271,0.,0.,.7577722285e-1,-.2259240285,.3169353608,0.,-.1816805231,.3750000000,-.4018262214,.4937160930,-.5387858374,.3603626151,.2888254471,-.2369298115e-1,-.1067891953},
			{0.,0.,0.,0.,.1250000000e-1,-.5705443310e-1,0.,0.,0.,-.2886751347e-1,.1157619501,0.,0.,.6388765648e-1,-.1986565342,0.,-.1582726231,.3313752322,.4377976190,-.4833325546,.2635268179},
			{0.,0.,0.,0.,0.,-.5208333333e-2,0.,0.,0.,0.,.1159855452e-1,0.,0.,0.,-.2412424530e-1,0.,0.,.5544791314e-1,0.,-.1417410714,.3972666214},
			{0.,-.4330127020,.7071067810,-.4592793268,0.,.1718750000,-.7500000000,0.,.1237436867,.1581138830,-.2157331140,-.3162277660,-.1141088661,.2916059217e-1,.1746928107e-1,.1890216038,-.1725163898e-1,-.7854574205e-1,-.9780759959e-1,.7031250000e-1,.4812067217e-1},
			{0.,0.,.1443375673,-.3750000000,.4192627457,-.1786086271,0.,-.1250000000,.3897114318,-.4357106264,.1300571401,-.7423218081,.2795084971,.8928571429e-2,.5297911641e-1,-.3496171374,-.8451542550e-1,.1102626021,.1796840339,-.5649864734e-1,-.8664029051e-1},
			{0.,0.,0.,-.5412658775e-1,.1936491673,-.3093592167,0.,0.,.8125000000e-1,-.2795084971,.4264736607,0.,.8068715303e-2,.1340277411,-.2555233287,-.6942056633,.3964594048,-.9743899444e-1,-.3527181261,-.5113897256e-1,.1668168700},
			{0.,0.,0.,0.,.2165063510e-1,-.9882117688e-1,0.,0.,0.,-.3750000000e-1,.1613825392,0.,0.,.3458020845e-1,-.1546796083,0.,.6546536711e-1,-.8715208612e-2,-.6464261051,.4446952959,-.3454883043},
			{0.,0.,0.,0.,0.,-.9021097958e-2,0.,0.,0.,0.,.1651785714e-1,0.,0.,0.,-.2103629347e-1,0.,0.,.8751446639e-2,0.,.9265956332e-1,-.6047514756},
			{0.,0.,.1863389981,-.4841229182,.5412658775,-.2305827461,0.,.4841229182,-.3913118960,.6250000000e-1,-.1467113994e-1,.5416666667,.3608439183,-.1959545145,-.1210078569,.1341862601,.9547032698e-1,.1373682881,-.1546473936,-.6352789941e-1,.1066497092},
			{0.,0.,0.,-.6987712428e-1,.2500000000,-.3993810316,0.,0.,-.2420614591e-1,-.7216878367e-1,.2343473623,0.,.3437500000,-.5190872089,.3481045690,.5968901920,.1653594569,-.2206772153,.1994047619,.1132938028,-.1630966411},
			{0.,0.,0.,0.,.2795084971e-1,-.1275775908,0.,0.,0.,-.1613743061e-1,.1073287078,0.,0.,-.8035714286e-1,.1426360826,0.,.2218529919,-.4548734170,.6095947223,.1275775908e-1,.2375739044},
			{0.,0.,0.,0.,0.,-.1164618738e-1,0.,0.,0.,0.,.1210307296e-1,0.,0.,0.,.1302083333e-1,0.,0.,-.8202104657e-1,0.,.1362603923,.6047900501},
			{0.,0.,0.,-.8267972847e-1,.2958039891,-.4725540094,0.,0.,-.2577698829,.4269562821,-.2839650695,0.,-.4806814823,.4724555913e-1,.2566181299,-.3937500000,-.4192627457,.3827327723e-1,-.1056442818e-1,-.2025231469e-1,.1055518026},
			{0.,0.,0.,0.,.3307189139e-1,-.1509518412,0.,0.,0.,.3818813080e-1,-.5229125166e-1,0.,0.,-.1056442819,.3423605578,0.,-.4312500000,.2909776087,-.4559196456,-.3558150542,-.6827053884e-1},
			{0.,0.,0.,0.,0.,-.1377995474e-1,0.,0.,0.,0.,-.2045792721e-2,0.,0.,0.,.4709974231e-1,0.,0.,-.1093750000e-1,0.,-.3584756798,-.4912769225},
			{0.,0.,0.,0.,.3750000000e-1,-.1711632992,0.,0.,0.,.1299038106,-.3303450769,0.,0.,.3114523254,-.2679129406,0.,.4464705337,.1746728448,.2866071429,.3594429284,-.5971745424e-1},
			{0.,0.,0.,0.,0.,-.1562500000e-1,0.,0.,0.,0.,-.2706329388e-1,0.,0.,0.,.1746928107e-1,0.,0.,.2096521686,0.,.4453125000,.3415558198},
			{0.,0.,0.,0.,0.,-.1727408745e-1,0.,0.,0.,0.,-.6411342242e-1,0.,0.,0.,-.1793351568,0.,0.,-.3427720463,0.,-.3960701481,-.2087053571}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,-.5590169942e-1,-.5103103630e-1,.4082482906,.1250000000,-.2453738645,.3227486122e-1,.1035406358,-.1613743061,.5590169942e-1,.1250000000,-.1189992461,.6614378278e-1,-.7395099729e-1,-.2507429436e-1,-.2795084971e-1,.5467611033e-1,.1208934286e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,0.,.9923207755e-1,.7216878367,-.7071067810,.3163924252,.9128709293e-1,-.1638392857,.3651483717,.3952847075e-1,-.2525381361,.1276586028,-.1403121520,.8964214566e-1,.8558232884e-1,.5646924394e-1,-.8312297405e-1,-.2350825488e-1},
			{0.,0.,.8333333333e-1,-.2165063510,.2420614591,-.1031197389,0.,-.2165063510,.4250000000,-.3913118960,.1159133539,.5776508940,-.6132223630,.3608439183,.1364673394e-1,.3218713882,-.3049687729e-2,-.2191913380,-.1210307296,.9785852773e-1,.4769519990e-1},
			{0.,0.,0.,-.3125000000e-1,.1118033988,-.1786086271,0.,0.,.7577722285e-1,-.2259240285,.3169353608,0.,-.1816805231,.3750000000,-.4018262214,.4937160930,-.5387858374,.3603626151,.2888254471,-.2369298115e-1,-.1067891953},
			{0.,0.,0.,0.,.1250000000e-1,-.5705443310e-1,0.,0.,0.,-.2886751347e-1,.1157619501,0.,0.,.6388765648e-1,-.1986565342,0.,-.1582726231,.3313752322,.4377976190,-.4833325546,.2635268179},
			{0.,0.,0.,0.,0.,-.5208333333e-2,0.,0.,0.,0.,.1159855452e-1,0.,0.,0.,-.2412424530e-1,0.,0.,.5544791314e-1,0.,-.1417410714,.3972666214},
			{0.,-.4330127020,.7071067810,-.4592793268,0.,.1718750000,-.7500000000,0.,.1237436867,.1581138830,-.2157331140,-.3162277660,-.1141088661,.2916059217e-1,.1746928107e-1,.1890216038,-.1725163898e-1,-.7854574205e-1,-.9780759959e-1,.7031250000e-1,.4812067217e-1},
			{0.,0.,.1443375673,-.3750000000,.4192627457,-.1786086271,0.,-.1250000000,.3897114318,-.4357106264,.1300571401,-.7423218081,.2795084971,.8928571429e-2,.5297911641e-1,-.3496171374,-.8451542550e-1,.1102626021,.1796840339,-.5649864734e-1,-.8664029051e-1},
			{0.,0.,0.,-.5412658775e-1,.1936491673,-.3093592167,0.,0.,.8125000000e-1,-.2795084971,.4264736607,0.,.8068715303e-2,.1340277411,-.2555233287,-.6942056633,.3964594048,-.9743899444e-1,-.3527181261,-.5113897256e-1,.1668168700},
			{0.,0.,0.,0.,.2165063510e-1,-.9882117688e-1,0.,0.,0.,-.3750000000e-1,.1613825392,0.,0.,.3458020845e-1,-.1546796083,0.,.6546536711e-1,-.8715208612e-2,-.6464261051,.4446952959,-.3454883043},
			{0.,0.,0.,0.,0.,-.9021097958e-2,0.,0.,0.,0.,.1651785714e-1,0.,0.,0.,-.2103629347e-1,0.,0.,.8751446639e-2,0.,.9265956332e-1,-.6047514756},
			{0.,0.,.1863389981,-.4841229182,.5412658775,-.2305827461,0.,.4841229182,-.3913118960,.6250000000e-1,-.1467113994e-1,.5416666667,.3608439183,-.1959545145,-.1210078569,.1341862601,.9547032698e-1,.1373682881,-.1546473936,-.6352789941e-1,.1066497092},
			{0.,0.,0.,-.6987712428e-1,.2500000000,-.3993810316,0.,0.,-.2420614591e-1,-.7216878367e-1,.2343473623,0.,.3437500000,-.5190872089,.3481045690,.5968901920,.1653594569,-.2206772153,.1994047619,.1132938028,-.1630966411},
			{0.,0.,0.,0.,.2795084971e-1,-.1275775908,0.,0.,0.,-.1613743061e-1,.1073287078,0.,0.,-.8035714286e-1,.1426360826,0.,.2218529919,-.4548734170,.6095947223,.1275775908e-1,.2375739044},
			{0.,0.,0.,0.,0.,-.1164618738e-1,0.,0.,0.,0.,.1210307296e-1,0.,0.,0.,.1302083333e-1,0.,0.,-.8202104657e-1,0.,.1362603923,.6047900501},
			{0.,0.,0.,-.8267972847e-1,.2958039891,-.4725540094,0.,0.,-.2577698829,.4269562821,-.2839650695,0.,-.4806814823,.4724555913e-1,.2566181299,-.3937500000,-.4192627457,.3827327723e-1,-.1056442818e-1,-.2025231469e-1,.1055518026},
			{0.,0.,0.,0.,.3307189139e-1,-.1509518412,0.,0.,0.,.3818813080e-1,-.5229125166e-1,0.,0.,-.1056442819,.3423605578,0.,-.4312500000,.2909776087,-.4559196456,-.3558150542,-.6827053884e-1},
			{0.,0.,0.,0.,0.,-.1377995474e-1,0.,0.,0.,0.,-.2045792721e-2,0.,0.,0.,.4709974231e-1,0.,0.,-.1093750000e-1,0.,-.3584756798,-.4912769225},
			{0.,0.,0.,0.,.3750000000e-1,-.1711632992,0.,0.,0.,.1299038106,-.3303450769,0.,0.,.3114523254,-.2679129406,0.,.4464705337,.1746728448,.2866071429,.3594429284,-.5971745424e-1},
			{0.,0.,0.,0.,0.,-.1562500000e-1,0.,0.,0.,0.,-.2706329388e-1,0.,0.,0.,.1746928107e-1,0.,0.,.2096521686,0.,.4453125000,.3415558198},
			{0.,0.,0.,0.,0.,-.1727408745e-1,0.,0.,0.,0.,-.6411342242e-1,0.,0.,0.,-.1793351568,0.,0.,-.3427720463,0.,-.3960701481,-.2087053571}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,.5590169942e-1,.5103103630e-1,-.4082482906,-.1250000000,.2453738645,-.3227486122e-1,-.1035406358,.1613743061,-.5590169942e-1,-.1250000000,.1189992461,-.6614378278e-1,.7395099729e-1,.2507429436e-1,.2795084971e-1,-.5467611033e-1,-.1208934286e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,0.,-.9923207755e-1,-.7216878367,-.3535533905,.5103103630e-1,-.9128709293e-1,.1049107143,.4564354646,.1185854122,-.1515228816,.4870046022e-1,-.2338535867,.5976143047e-1,.1108137724,.1129384878,-.1024538982,-.5342785200e-1},
			{0.,0.,.5833333333,.2165063510,-.2420614591,.1031197389,0.,.7938566203,.3750000000,-.1677050983,.4738596230e-1,.4658474951,.9682458365e-1,.3092947872e-1,.1162325270,-.4091585443,-.1189378215,.3433564566e-1,.2593515633,-.1788811797e-1,-.1454121949},
			{0.,0.,0.,-.4687500000,-.1118033988,.1786086271,0.,0.,-.7252962758,-.2581988897,.2133947249,0.,-.6568449682,-.2857142857,.3504772316e-1,-.3047338563,.6338656908e-1,-.1639473093e-1,.3433961537,.9294938757e-1,-.2599208715},
			{0.,0.,0.,0.,.3875000000,.5705443310e-1,0.,0.,0.,.6350852963,.1552904208,0.,0.,.6867923071,.2642678667,0.,.5173388725,.1832986642,.2026785714,-.1589373493,-.2789272040},
			{0.,0.,0.,0.,0.,-.3281250000,0.,0.,0.,0.,-.5528644321,0.,0.,0.,-.6513546229,0.,0.,-.6066461029,0.,-.3999255952,-.1365849382},
			{0.,.4330127020,.7071067810,.2551551816,-.1825741858,-.1093750000,-.2500000000,-.6123724358,-.4065863991,.1581138830,.2095472184,.1581138830,.2966830521,.2916059217e-1,-.2131252291,-.8100925875e-1,-.1035098339,.8090801999e-1,.3912303983e-1,.2343750000e-1,-.1850795084e-1},
			{0.,0.,-.4330127020,-.6250000000,-.5590169942e-1,.2296396634,0.,-.1250000000,-.4330127020e-1,.4841229182e-1,-.1224809960,.2904737510,.6149186937,.2767857143,-.1149239294,-.2551260193,-.4278593414,-.8679563432e-2,.1617156305,.1877213121,-.9067007150e-1},
			{0.,0.,0.,.4149705060,.5163977793,-.7954951286e-1,0.,0.,.3187500000,.3354101966,-.6561133241e-2,0.,-.1371681602,-.3917733971,-.1143502189,-.2495867120,-.4025587803,-.7962157829e-1,.2812523620,.4211073418,-.2128834533},
			{0.,0.,0.,0.,-.3897114318,-.4150489429,0.,0.,0.,-.4125000000,-.4254630581,0.,0.,-.6224437520e-1,.6629126072e-1,0.,.2495867120,.5017470100,.1943402247,.1877602361,-.2674519317},
			{0.,0.,0.,0.,0.,.3626481379,0.,0.,0.,0.,.4513392857,0.,0.,0.,.2169908080,0.,0.,-.1237704596,0.,-.2735970138,-.1451294743},
			{0.,0.,.1863389981,.4841229182,.4330127020,.3294039230e-1,0.,-.1613743061,-.5031152948,-.5625000000,-.1189992461,.4166666667e-1,.2165063510,.3803822929,.1967692977,-.3659625275e-1,-.1432054905,-.1672490034,.2319710903e-1,.7293944007e-1,-.1300606210e-1},
			{0.,0.,0.,-.2096313728,-.5000000000,-.3309157119,0.,0.,.2420614591e-1,.1443375673,.1835250428,0.,.1562500000,.3992978531,.3043636809,-.6866878320e-1,-.3307189139,-.4995787124,.7738095238e-1,.2828269754,-.5857061698e-1},
			{0.,0.,0.,0.,.2236067977,.4847948451,0.,0.,0.,.8068715303e-1,.1073287078,0.,0.,-.1696428571,-.4686614144,0.,-.7923321136e-1,-.8518830777e-1,.7520109565e-1,.3152989029,-.1034921017},
			{0.,0.,0.,0.,0.,-.2305945101,0.,0.,0.,0.,-.1573399484,0.,0.,0.,.1104910714,0.,0.,.1771009002,0.,.6155841900e-2,-.6964390938e-1},
			{0.,0.,0.,.8267972847e-1,.2958039891,.4185478369,0.,0.,-.8592329430e-1,-.3415650256,-.5445447805,0.,.3697549864e-1,.1889822365,.3860054223,-.6250000000e-2,-.5590169942e-1,-.1726550062,.7042952119e-2,.4532660906e-1,-.5330899122e-2},
			{0.,0.,0.,0.,-.9921567416e-1,-.3320940506,0.,0.,0.,.3818813080e-1,.1718141126,0.,0.,.6338656908e-1,.1977011672,0.,-.5625000000e-1,-.2605485777,.1338624175e-1,.1142921083,-.1842220890e-1},
			{0.,0.,0.,0.,0.,.1129956289,0.,0.,0.,0.,.7501239977e-2,0.,0.,0.,-.9992188321e-1,0.,0.,.1909722222e-2,0.,.5059211956e-1,-.1790526418e-1},
			{0.,0.,0.,0.,.3750000000e-1,.1711632992,0.,0.,0.,-.4330127020e-1,-.2117596647,0.,0.,.2395787117e-1,.1366902758,0.,-.7086833870e-2,-.5391137183e-1,.8928571429e-3,.1222594995e-1,-.1228754203e-2},
			{0.,0.,0.,0.,0.,-.4687500000e-1,0.,0.,0.,0.,.2706329388e-1,0.,0.,0.,.2246050424e-1,0.,0.,-.3248132189e-1,0.,.1450892857e-1,-.2355557379e-2},
			{0.,0.,0.,0.,0.,.1727408745e-1,0.,0.,0.,0.,-.2137114080e-1,0.,0.,0.,.1379501207e-1,0.,0.,-.5440826132e-2,0.,.1233863389e-2,-.1240079365e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,.5590169942e-1,.5103103630e-1,-.4082482906,-.1250000000,.2453738645,-.3227486122e-1,-.1035406358,.1613743061,-.5590169942e-1,-.1250000000,.1189992461,-.6614378278e-1,.7395099729e-1,.2507429436e-1,.2795084971e-1,-.5467611033e-1,-.1208934286e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,0.,-.9923207755e-1,-.7216878367,-.3535533905,.5103103630e-1,-.9128709293e-1,.1049107143,.4564354646,.1185854122,-.1515228816,.4870046022e-1,-.2338535867,.5976143047e-1,.1108137724,.1129384878,-.1024538982,-.5342785200e-1},
			{0.,0.,.5833333333,.2165063510,-.2420614591,.1031197389,0.,.7938566203,.3750000000,-.1677050983,.4738596230e-1,.4658474951,.9682458365e-1,.3092947872e-1,.1162325270,-.4091585443,-.1189378215,.3433564566e-1,.2593515633,-.1788811797e-1,-.1454121949},
			{0.,0.,0.,-.4687500000,-.1118033988,.1786086271,0.,0.,-.7252962758,-.2581988897,.2133947249,0.,-.6568449682,-.2857142857,.3504772316e-1,-.3047338563,.6338656908e-1,-.1639473093e-1,.3433961537,.9294938757e-1,-.2599208715},
			{0.,0.,0.,0.,.3875000000,.5705443310e-1,0.,0.,0.,.6350852963,.1552904208,0.,0.,.6867923071,.2642678667,0.,.5173388725,.1832986642,.2026785714,-.1589373493,-.2789272040},
			{0.,0.,0.,0.,0.,-.3281250000,0.,0.,0.,0.,-.5528644321,0.,0.,0.,-.6513546229,0.,0.,-.6066461029,0.,-.3999255952,-.1365849382},
			{0.,.4330127020,.7071067810,.2551551816,-.1825741858,-.1093750000,-.2500000000,-.6123724358,-.4065863991,.1581138830,.2095472184,.1581138830,.2966830521,.2916059217e-1,-.2131252291,-.8100925875e-1,-.1035098339,.8090801999e-1,.3912303983e-1,.2343750000e-1,-.1850795084e-1},
			{0.,0.,-.4330127020,-.6250000000,-.5590169942e-1,.2296396634,0.,-.1250000000,-.4330127020e-1,.4841229182e-1,-.1224809960,.2904737510,.6149186937,.2767857143,-.1149239294,-.2551260193,-.4278593414,-.8679563432e-2,.1617156305,.1877213121,-.9067007150e-1},
			{0.,0.,0.,.4149705060,.5163977793,-.7954951286e-1,0.,0.,.3187500000,.3354101966,-.6561133241e-2,0.,-.1371681602,-.3917733971,-.1143502189,-.2495867120,-.4025587803,-.7962157829e-1,.2812523620,.4211073418,-.2128834533},
			{0.,0.,0.,0.,-.3897114318,-.4150489429,0.,0.,0.,-.4125000000,-.4254630581,0.,0.,-.6224437520e-1,.6629126072e-1,0.,.2495867120,.5017470100,.1943402247,.1877602361,-.2674519317},
			{0.,0.,0.,0.,0.,.3626481379,0.,0.,0.,0.,.4513392857,0.,0.,0.,.2169908080,0.,0.,-.1237704596,0.,-.2735970138,-.1451294743},
			{0.,0.,.1863389981,.4841229182,.4330127020,.3294039230e-1,0.,-.1613743061,-.5031152948,-.5625000000,-.1189992461,.4166666667e-1,.2165063510,.3803822929,.1967692977,-.3659625275e-1,-.1432054905,-.1672490034,.2319710903e-1,.7293944007e-1,-.1300606210e-1},
			{0.,0.,0.,-.2096313728,-.5000000000,-.3309157119,0.,0.,.2420614591e-1,.1443375673,.1835250428,0.,.1562500000,.3992978531,.3043636809,-.6866878320e-1,-.3307189139,-.4995787124,.7738095238e-1,.2828269754,-.5857061698e-1},
			{0.,0.,0.,0.,.2236067977,.4847948451,0.,0.,0.,.8068715303e-1,.1073287078,0.,0.,-.1696428571,-.4686614144,0.,-.7923321136e-1,-.8518830777e-1,.7520109565e-1,.3152989029,-.1034921017},
			{0.,0.,0.,0.,0.,-.2305945101,0.,0.,0.,0.,-.1573399484,0.,0.,0.,.1104910714,0.,0.,.1771009002,0.,.6155841900e-2,-.6964390938e-1},
			{0.,0.,0.,.8267972847e-1,.2958039891,.4185478369,0.,0.,-.8592329430e-1,-.3415650256,-.5445447805,0.,.3697549864e-1,.1889822365,.3860054223,-.6250000000e-2,-.5590169942e-1,-.1726550062,.7042952119e-2,.4532660906e-1,-.5330899122e-2},
			{0.,0.,0.,0.,-.9921567416e-1,-.3320940506,0.,0.,0.,.3818813080e-1,.1718141126,0.,0.,.6338656908e-1,.1977011672,0.,-.5625000000e-1,-.2605485777,.1338624175e-1,.1142921083,-.1842220890e-1},
			{0.,0.,0.,0.,0.,.1129956289,0.,0.,0.,0.,.7501239977e-2,0.,0.,0.,-.9992188321e-1,0.,0.,.1909722222e-2,0.,.5059211956e-1,-.1790526418e-1},
			{0.,0.,0.,0.,.3750000000e-1,.1711632992,0.,0.,0.,-.4330127020e-1,-.2117596647,0.,0.,.2395787117e-1,.1366902758,0.,-.7086833870e-2,-.5391137183e-1,.8928571429e-3,.1222594995e-1,-.1228754203e-2},
			{0.,0.,0.,0.,0.,-.4687500000e-1,0.,0.,0.,0.,.2706329388e-1,0.,0.,0.,.2246050424e-1,0.,0.,-.3248132189e-1,0.,.1450892857e-1,-.2355557379e-2},
			{0.,0.,0.,0.,0.,.1727408745e-1,0.,0.,0.,0.,-.2137114080e-1,0.,0.,0.,.1379501207e-1,0.,0.,-.5440826132e-2,0.,.1233863389e-2,-.1240079365e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.3367175148e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.6734350295e-2,.1178511302,.2357022604e-1,.7856742012e-2,.3367175148e-2,.7071067810e-1,.1178511302e-1,.3367175148e-2,.4714045206e-1,.6734350295e-2,.3367175148e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,-.2976190476e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.5952380952e-3,.1000000000,.1190476190e-1,.2380952381e-2,.7619047619e-1,.8333333333e-2,.5952380952e-1},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,.1620032899e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,-.1944039478e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,-.1944039478e-2,.5248906592e-1,0.,-.1296026319e-2,.5248906592e-1,.1944039478e-2,.4860098696e-1},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,-.5611958579e-2,0.,0.,.3367175148e-2,.3086577219e-2,.2244783431e-2,0.,-.5050762721e-2,-.7482611439e-3,.2805979290e-3,.1010152544e-1,-.3928371006e-2,-.1122391716e-2,.1795826746e-1,-.2244783431e-2,.2244783431e-1},
			{0.,0.,0.,0.,.5019488348e-3,.1140792807e-2,0.,0.,0.,-.6274360436e-3,-.8670025331e-3,0.,0.,.8365813917e-3,.4563171227e-3,0.,-.1254872087e-2,.2281585613e-3,.2509744174e-2,-.1597109929e-2,.5703964034e-2},
			{0.,0.,0.,0.,0.,-.1041396590e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,-.1562094885e-3,0.,0.,.2082793179e-3,0.,-.3124189770e-3,.6248379539e-3},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,.5154913119e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,.8247860991e-2,0.,.9622504490e-2,.5498573994e-2,.3092947872e-2,0.,.4123930495e-2,.2061965247e-2,0.,.2061965247e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,-.2805979290e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,.1122391716e-2,0.,.1346870059e-1,.5892556509e-2,.2525381361e-2,0.,.7576144082e-2,.3086577219e-2,0.,.4489566864e-2,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,.9720197392e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,-.3110463166e-2,0.,.4860098696e-2,0.,-.8748177653e-3,0.,.4860098696e-2,.9720197392e-3,0.,.3888078956e-2,0.},
			{0.,0.,0.,0.,-.8694008848e-3,-.1975911102e-2,0.,0.,0.,.9780759957e-3,.1343619549e-2,0.,0.,-.1086751106e-2,-.5334959976e-3,0.,.1086751106e-2,-.4939777755e-3,0.,.1580728882e-2,0.},
			{0.,0.,0.,0.,0.,.1803751804e-3,0.,0.,0.,0.,-.2020202020e-3,0.,0.,0.,.2272727273e-3,0.,0.,-.2525252525e-3,0.,.2525252525e-3,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,.3622503687e-1,0.,0.,.4347004424e-2,.4890379978e-2,.4347004424e-2,0.,0.,.1086751106e-2,.1086751106e-2,0.,0.,.3622503687e-3,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,-.1254872087e-1,0.,0.,.3764616260e-2,.3137180218e-2,.2007795340e-2,0.,0.,.1673162783e-2,.1380359296e-2,0.,0.,.7529232524e-3,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,.2550890264e-2,0.,0.,0.,-.9820927513e-3,-.1326462937e-2,0.,0.,.6547285010e-3,.1785623185e-3,0.,0.,.5356869553e-3,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2328633565e-3,0.,0.,0.,0.,.2235488223e-3,0.,0.,0.,-.1956052195e-3,0.,0.,.1304034797e-3,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,.1484784677e-1,0.,0.,0.,.7423923386e-3,.1187827742e-2,0.,0.,0.,.1484784677e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,-.3018254064e-2,0.,0.,0.,.6640158940e-3,.8451111378e-3,0.,0.,0.,.2414603251e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2755276392e-3,0.,0.,0.,0.,-.1983799002e-3,0.,0.,0.,.9918995009e-4,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,.3422378419e-2,0.,0.,0.,0.,.1368951368e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.3124189770e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.3453921745e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.3367175148e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.6734350295e-2,.1178511302,.2357022604e-1,.7856742012e-2,.3367175148e-2,.7071067810e-1,.1178511302e-1,.3367175148e-2,.4714045206e-1,.6734350295e-2,.3367175148e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,-.2976190476e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.5952380952e-3,.1000000000,.1190476190e-1,.2380952381e-2,.7619047619e-1,.8333333333e-2,.5952380952e-1},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,.1620032899e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,-.1944039478e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,-.1944039478e-2,.5248906592e-1,0.,-.1296026319e-2,.5248906592e-1,.1944039478e-2,.4860098696e-1},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,-.5611958579e-2,0.,0.,.3367175148e-2,.3086577219e-2,.2244783431e-2,0.,-.5050762721e-2,-.7482611439e-3,.2805979290e-3,.1010152544e-1,-.3928371006e-2,-.1122391716e-2,.1795826746e-1,-.2244783431e-2,.2244783431e-1},
			{0.,0.,0.,0.,.5019488348e-3,.1140792807e-2,0.,0.,0.,-.6274360436e-3,-.8670025331e-3,0.,0.,.8365813917e-3,.4563171227e-3,0.,-.1254872087e-2,.2281585613e-3,.2509744174e-2,-.1597109929e-2,.5703964034e-2},
			{0.,0.,0.,0.,0.,-.1041396590e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,-.1562094885e-3,0.,0.,.2082793179e-3,0.,-.3124189770e-3,.6248379539e-3},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,.5154913119e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,.8247860991e-2,0.,.9622504490e-2,.5498573994e-2,.3092947872e-2,0.,.4123930495e-2,.2061965247e-2,0.,.2061965247e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,-.2805979290e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,.1122391716e-2,0.,.1346870059e-1,.5892556509e-2,.2525381361e-2,0.,.7576144082e-2,.3086577219e-2,0.,.4489566864e-2,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,.9720197392e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,-.3110463166e-2,0.,.4860098696e-2,0.,-.8748177653e-3,0.,.4860098696e-2,.9720197392e-3,0.,.3888078956e-2,0.},
			{0.,0.,0.,0.,-.8694008848e-3,-.1975911102e-2,0.,0.,0.,.9780759957e-3,.1343619549e-2,0.,0.,-.1086751106e-2,-.5334959976e-3,0.,.1086751106e-2,-.4939777755e-3,0.,.1580728882e-2,0.},
			{0.,0.,0.,0.,0.,.1803751804e-3,0.,0.,0.,0.,-.2020202020e-3,0.,0.,0.,.2272727273e-3,0.,0.,-.2525252525e-3,0.,.2525252525e-3,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,.3622503687e-1,0.,0.,.4347004424e-2,.4890379978e-2,.4347004424e-2,0.,0.,.1086751106e-2,.1086751106e-2,0.,0.,.3622503687e-3,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,-.1254872087e-1,0.,0.,.3764616260e-2,.3137180218e-2,.2007795340e-2,0.,0.,.1673162783e-2,.1380359296e-2,0.,0.,.7529232524e-3,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,.2550890264e-2,0.,0.,0.,-.9820927513e-3,-.1326462937e-2,0.,0.,.6547285010e-3,.1785623185e-3,0.,0.,.5356869553e-3,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2328633565e-3,0.,0.,0.,0.,.2235488223e-3,0.,0.,0.,-.1956052195e-3,0.,0.,.1304034797e-3,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,.1484784677e-1,0.,0.,0.,.7423923386e-3,.1187827742e-2,0.,0.,0.,.1484784677e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,-.3018254064e-2,0.,0.,0.,.6640158940e-3,.8451111378e-3,0.,0.,0.,.2414603251e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2755276392e-3,0.,0.,0.,0.,-.1983799002e-3,0.,0.,0.,.9918995009e-4,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,.3422378419e-2,0.,0.,0.,0.,.1368951368e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.3124189770e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.3453921745e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.,5.477225575,-6.324555320,7.071067810,-7.745966692,-7.483314774,8.366600265,-9.165151390,9.486832980,-10.39230485,-11.48912529},
			{0.,6.,-19.59591794,42.42640686,-75.89466384,121.2435566,3.464101616,-25.45584412,63.68673332,-120.4989626,198.,-10.95445115,56.92099788,-127.2792206,224.6330341,22.44994432,-100.3992032,210.7984820,-37.94733192,155.8845727,57.44562647},
			{0.,0.,24.49489743,-127.2792206,398.4469852,-969.9484525,0.,21.21320343,-161.6663230,575.1086854,-1488.,5.477225575,-94.86832980,487.9036789,-1471.733671,-22.44994432,250.9980080,-1081.487864,56.92099788,-519.6152424,-114.8912529},
			{0.,0.,0.,98.99494934,-708.3501958,2909.845357,0.,0.,102.8785692,-920.1738966,4176.,0.,44.27188724,-622.2539673,3578.636612,7.483314774,-234.2648074,2071.324214,-37.94733192,727.4613394,114.8912529},
			{0.,0.,0.,0.,398.4469852,-3637.306697,0.,0.,0.,460.0869483,-4860.,0.,0.,254.5584412,-3601.874512,0.,75.29940238,-1695.553007,9.486832980,-467.6537182,-57.44562647},
			{0.,0.,0.,0.,0.,1600.414947,0.,0.,0.,0.,1980.,0.,0.,0.,1278.084504,0.,0.,504.0833264,0.,114.3153533,11.48912529},
			{0.,0.,0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-10.95445115,12.,-32.86335345,37.94733192,-42.42640686,46.47580015,89.79977729,-100.3992032,109.9818167,-189.7366596,207.8460970,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,42.42640686,-117.5755077,230.0434742,-384.,32.86335345,-303.5786554,721.2489166,-1301.322404,-179.5995546,1104.391235,-2419.599967,569.2099788,-2909.845357,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-920.1738966,2592.,0.,265.6313234,-2206.173157,7529.079625,89.79977729,-1907.584860,10558.25440,-569.2099788,7482.459491,2068.042553},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-5760.,0.,0.,1527.350647,-13942.74005,0.,903.5928286,-14297.63617,189.7366596,-7066.767297,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3960.,0.,0.,0.,7668.507025,0.,0.,6048.999917,0.,2286.307067,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,42.42640686,-46.47580015,-224.4994432,250.9980080,-274.9545417,853.8149682,-935.3074363,-2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-678.8225098,1254.846604,224.4994432,-2509.980080,5774.045376,-1707.629936,12158.99667,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-6274.233021,0.,2258.982072,-20621.59063,853.8149682,-21512.07104,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,15122.49979,0.,10288.38180,2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-167.3320053,183.3030278,-1328.156617,1454.922679,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,-3666.060556,1328.156617,-17459.07214,-12867.82033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10081.66653,0.,16004.14947,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086,-727.4613394,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,8002.074733,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2895.259574}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[21],
		T o_output[21]
	) {
		static const T matrix[21][21] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.,5.477225575,-6.324555320,7.071067810,-7.745966692,-7.483314774,8.366600265,-9.165151390,9.486832980,-10.39230485,-11.48912529},
			{0.,6.,-19.59591794,42.42640686,-75.89466384,121.2435566,3.464101616,-25.45584412,63.68673332,-120.4989626,198.,-10.95445115,56.92099788,-127.2792206,224.6330341,22.44994432,-100.3992032,210.7984820,-37.94733192,155.8845727,57.44562647},
			{0.,0.,24.49489743,-127.2792206,398.4469852,-969.9484525,0.,21.21320343,-161.6663230,575.1086854,-1488.,5.477225575,-94.86832980,487.9036789,-1471.733671,-22.44994432,250.9980080,-1081.487864,56.92099788,-519.6152424,-114.8912529},
			{0.,0.,0.,98.99494934,-708.3501958,2909.845357,0.,0.,102.8785692,-920.1738966,4176.,0.,44.27188724,-622.2539673,3578.636612,7.483314774,-234.2648074,2071.324214,-37.94733192,727.4613394,114.8912529},
			{0.,0.,0.,0.,398.4469852,-3637.306697,0.,0.,0.,460.0869483,-4860.,0.,0.,254.5584412,-3601.874512,0.,75.29940238,-1695.553007,9.486832980,-467.6537182,-57.44562647},
			{0.,0.,0.,0.,0.,1600.414947,0.,0.,0.,0.,1980.,0.,0.,0.,1278.084504,0.,0.,504.0833264,0.,114.3153533,11.48912529},
			{0.,0.,0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-10.95445115,12.,-32.86335345,37.94733192,-42.42640686,46.47580015,89.79977729,-100.3992032,109.9818167,-189.7366596,207.8460970,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,42.42640686,-117.5755077,230.0434742,-384.,32.86335345,-303.5786554,721.2489166,-1301.322404,-179.5995546,1104.391235,-2419.599967,569.2099788,-2909.845357,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-920.1738966,2592.,0.,265.6313234,-2206.173157,7529.079625,89.79977729,-1907.584860,10558.25440,-569.2099788,7482.459491,2068.042553},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-5760.,0.,0.,1527.350647,-13942.74005,0.,903.5928286,-14297.63617,189.7366596,-7066.767297,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3960.,0.,0.,0.,7668.507025,0.,0.,6048.999917,0.,2286.307067,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,42.42640686,-46.47580015,-224.4994432,250.9980080,-274.9545417,853.8149682,-935.3074363,-2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-678.8225098,1254.846604,224.4994432,-2509.980080,5774.045376,-1707.629936,12158.99667,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-6274.233021,0.,2258.982072,-20621.59063,853.8149682,-21512.07104,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,15122.49979,0.,10288.38180,2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-167.3320053,183.3030278,-1328.156617,1454.922679,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,-3666.060556,1328.156617,-17459.07214,-12867.82033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10081.66653,0.,16004.14947,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086,-727.4613394,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,8002.074733,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2895.259574}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 21; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[21],
		T o_output[1]
	) {
		static const T matrix[1][21] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[21],
		T o_output[1]
	) {
		static const T matrix[1][21] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 21; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 14175;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[21][21] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,7.838367178,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,7.668115805,10.73312629,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.828427124,7.314285714,11.33763824,12.92995257,9.938586932,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,4.381780461,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,2.449489743,9.142857143,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.533303756,6.337932448,13.55261854,16.66751699,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,-2.993325910,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,3.346640106,-5.074680381,1.521277658,0.,0.,0.,0.,0.,0.,0.,4.364357806,17.00840129,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.320493798,5.935526613,-.3207134902,-5.966396197,2.530240385,0.,0.,0.,0.,0.,0.,1.792842914,9.200874127,19.80226324,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,10.95445115,-3.833259389,.5532833353,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,6.857142857,9.972922524,-7.273098319,1.490788040,0.,0.,0.,0.,0.,4.320493800,21.25401924,0.,0.,0.,0.},
			{5.416025602,-4.376809636,2.010178182,-.5158119621,.5766953058e-1,0.,0.,0.,0.,0.,0.,15.73213272,-4.036867140,.4513354669,0.,0.,0.,0.,21.79908256,0.,0.}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[21][21] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,-2.828427124,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,8.854377448,-3.098386677,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.715476066,-4.618802155,12.72792206,-3.265986323,20.08316045,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,3.919183589,4.200000000,0.,0.,0.,-5.091168823,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,3.834057902,5.366563145,4.647580015,0.,0.,6.640783086,-6.196773354,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.414213562,3.657142857,5.668819119,6.464976283,4.969293464,0.,-6.730254569,9.455027815,-6.998542123,17.21413752,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,-.9128709293,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,2.529822128,-1.807392228,0.,0.,0.,2.190890230,6.260990336,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,3.535533905,1.649572198,-2.571428571,0.,0.,1.224744872,4.571428571,7.423074892,0.,0.,-7.377111136,11.49977817,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.825741857,3.836097790,3.659207005,.5216405308,-3.207665141,0.,.7666518779,3.168966225,6.776309270,8.333758491,0.,8.636804253,-8.748177653,15.06237033,0.,0.,0.,0.,0.,0.,0.},
			{2.645751311,-1.496662955,.3055050464,0.,0.,0.,5.184592560,-1.058300524,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,1.673320053,-2.537340190,.7606388294,0.,0.,2.898275350,5.747048932,-2.195775165,0.,0.,2.182178903,8.504200642,0.,0.,10.06230590,0.,0.,0.,0.,0.},
			{2.160246899,2.967763307,-.1603567451,-2.983198099,1.265120192,0.,1.814229470,5.416047582,5.345224839,-3.286878676,0.,.8964214566,4.600437064,9.901131622,0.,-9.525793445,13.38877363,0.,0.,0.,0.},
			{2.683281572,-1.897366596,.6639400021,-.9583148473e-1,0.,0.,5.477225576,-1.916629694,.2766416676,0.,0.,7.423074892,-1.071428571,0.,0.,8.874119673,0.,0.,0.,0.,0.},
			{2.449489743,.9897433188,-2.727411870,1.516350793,-.2869022921,0.,3.428571429,4.986461263,-3.636549159,.7453940198,0.,3.049339172,8.346248498,-2.245365598,0.,2.160246900,10.62700962,0.,12.04989627,0.,0.},
			{2.708012801,-2.188404818,1.005089091,-.2579059811,.2883476528e-1,0.,5.685642496,-2.611298058,.6700593942,-.7491491772e-1,0.,7.866066360,-2.018433570,.2256677334,0.,9.552971226,-1.068054652,0.,10.89954128,0.,0.}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[12][21] = 
		{
			{1.414213562,-1.94468190274016,2.27090372268066,-2.44801058376252,2.49586968540814,-2.4264683064468,3.43216369764802,-4.00974905232829,4.32546942586304,-4.4144312457315,4.2977488551218,5.37669458746301,-5.80779112496784,5.9385809984625,-5.79724384742923,7.27823584467214,-7.46210387346585,7.3119743919632,9.14177937313724,-8.99870040642757,10.9691712594618},
			{1.414213562,-1.71235176911142,1.5663330178964,-1.07607439139602,.363617646552178,.420994417051474,3.29802783242322,-3.07100849759875,2.20598407348367,-.923194488569861,-.507379758856801,4.96464287844228,-3.80884940873901,2.02329330850018,.0162698697481229,6.45780747384475,-4.10479708970297,1.32544341327556,7.79428043908965,-4.03553981216457,8.98680735137078},
			{1.414213562,-1.30970802258292,.519224820053661,.518731848959601,-1.30425364587756,1.47252576661999,3.06556135691961,-1.5947619510804,-.444911232504622,2.09440936915946,-2.63186985254377,4.2894282422279,-.964148801340397,-2.01722273728899,3.46190525493913,5.18623489409548,.205483038001612,-3.63947886955732,5.81833621896757,1.69243379052584,6.23568135536058},
			{1.414213562,-.761953862859853,-.55104212557113,1.37646295656649,-1.03415883911242,-.174795669385988,2.74931534527504,.106757773657142,-2.26282036321899,2.15044627554921,-.119501749587716,3.45007465931297,1.77035093065653,-3.42380239520456,1.33170919606025,3.7410698100085,3.58481468643821,-3.59724489879787,3.76406389114812,5.2355975909823,3.61790059797515},
			{1.414213562,-.103494496994541,-1.29720437850508,.991754894783212,.589436063368126,-1.41318416081787,2.36915365304765,1.68415540504794,-2.32831086437971,-.381749900992996,2.46800676549626,2.56192161158569,3.58714638367868,-1.52344991306454,-2.69130383032704,2.3938769796025,4.93737708601548,.508166439915615,2.07554270972352,5.63164748396125,1.71909544842},
			{1.414213562,.624299774465593,-1.43546196207946,-.337640235075449,1.50389998701775,-.0107903422167659,1.94896143440088,2.83314651369184,-.637278791777431,-2.63766838605638,.808065843843811,1.73374759666501,4.12740556141282,1.98940127224618,-2.96832239843811,1.33269841309056,4.37531622836173,4.52786743647191,.950542947793622,3.96849611187728,.64766447670966},
			{1.414213562,1.37570022553441,-.821946062819915,-1.61733246050546,.359308464240942,1.71165917287845,1.51514018159912,3.36446686788906,1.91964817870762,-1.95173675811295,-2.75500656898115,1.04781596872157,3.55510980002441,4.5904502888745,1.24891795372526,.626153181688305,2.8447312409815,5.48872263996492,.347192252628357,1.9734504193592,.183906859961432},
			{1.414213562,2.10349449699454,.504794589391635,-1.67800190212113,-1.79846700964276,.309494957156941,1.09494796295235,3.24473228828999,4.05017972204568,1.63294619261043,-2.28560008620039,.547225517982272,2.39319484717323,4.87179403320197,5.49730994822616,.23632114600406,1.36209257652084,3.7763572186341,.0946962283839243,.676667877319742,.0362494356624604},
			{1.414213562,2.76195386285985,2.32621648423849,.160883133644535,-2.12140648579363,-2.7058281817046,.714786270724957,2.59853682213183,4.7549644652192,5.36731919542456,3.25729414076559,.233201739437629,1.22672710731592,3.30495822058058,5.89679444984458,.0657431866024218,.451524561772794,1.59193561065057,.0171974308138245,.145628931559436,.00429748986296868},
			{1.414213562,3.30970802258292,4.2909622271481,3.64701232940425,1.49376547749553,-1.29952592474851,.398540259080393,1.67165845871656,3.84755914106515,6.3113399865174,7.98777931164405,.07249762486996,.434860760346183,1.40707187748205,3.23180916099,.0113956405123903,.0887334889355507,.368089956036703,.00166206328995187,.015902832140972,.000231576691010652},
			{1.414213562,3.71235176911142,5.99558490979395,7.62488741580146,8.14599447640013,7.34718311089616,.166073783576779,.764836159193308,2.02218309922343,4.02885967082133,6.68010475273478,.0125887190510937,.0823390061849577,.299004079114616,.796298871799331,.000824565926335141,.00697737063218211,.0321260785532915,5.01144631619021e-005,.000520025886860232,2.90963993885675e-006},
			{1.414213562,3.94468190274016,7.0795491341205,10.545740291788,14.1178337840267,17.5926352342147,.0319379183519753,.154660030302817,.43925822557731,.961797763577864,1.79684767953686,.000465578153648566,.00319092434145458,.0123137463285649,.0353622090543065,5.86464816370859e-006,5.19109369727387e-005,.000252650701904178,6.85464935253131e-008,7.43273958578616e-007,7.65349815116598e-010}
		};
		for (int j = 0; j < 12; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[12][21] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.40022577929605,4.1644090826311,-4.80864541143983,5.37622900930936,-5.88936380548032,5.17702934684997,-5.97791857375092,6.68351614602547,-7.32142511479571,-6.67435503772231,7.46215578440288,-8.17438210188192,7.8090615264033,-8.55439830449838,-8.51143945889148},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.13195404884644,3.83584465679206,-4.42925189189755,4.95205415939119,-5.42470353822284,3.97724364248013,-4.59252537503604,5.13459946240545,-5.62467189980939,-3.67768288095471,4.11177446033512,-4.5042232468252,2.21014640326355,-2.42109408164425,.161112776439225},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-2.66702109783921,3.26642040979696,-3.77173740709444,4.21693061735794,-4.61941604516583,2.13132645129358,-2.46104380066281,2.75153061636385,-3.01415077313091,.104423971232213,-.116749549066047,.127892723210502,-2.94703586087065,3.22831603852647,4.81780334972074},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-2.03452907455009,2.49177904847469,-2.87725860980962,3.21687291987534,-3.52390772571971,.0953780462518065,-.110133081351697,.123132528211026,-.134884926556002,2.8024998850757,-3.1332901246654,3.43234736114769,-3.77550789188145,4.13586167809416,1.3916398503358},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-1.2742056900953,1.56057688324205,-1.80199896789671,2.01469609360342,-2.20698899398908,-1.62700976129611,1.87870904723461,-2.10046056933792,2.30093927043685,3.19783168117628,-3.57528450949468,3.91652795089531,-.49602746920004,.543370868210304,-3.52996795992507},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-.433821252801768,.531320354197217,-.613515899259544,.68593162794344,-.751400451068813,-2.60976045384023,3.01349180095816,-3.3691862575809,3.69075862820297,1.36899683098764,-1.53058498738027,1.67667184761831,3.00982432351404,-3.29709735322896,-2.5031295519909},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,.433821252801768,-.531320354197217,.613515899259544,-.68593162794344,.751400451068813,-2.60976045384023,3.01349180095816,-3.36918625758091,3.69075862820298,-1.36899683098765,1.53058498738027,-1.67667184761832,3.00982432351407,-3.297097353229,2.50312955199098},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,1.2742056900953,-1.56057688324205,1.80199896789671,-2.01469609360342,2.20698899398908,-1.62700976129612,1.87870904723461,-2.10046056933793,2.30093927043686,-3.1978316811763,3.5752845094947,-3.91652795089533,-.496027469200023,.543370868210286,3.52996795992515},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,2.03452907455009,-2.49177904847469,2.87725860980962,-3.21687291987534,3.52390772571971,.0953780462518019,-.110133081351691,.12313252821102,-.134884926555995,-2.8024998850757,3.13329012466539,-3.43234736114768,-3.77550789188148,4.1358616780942,-1.39163985033612},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,2.66702109783921,-3.26642040979696,3.77173740709444,-4.21693061735794,4.61941604516583,2.13132645129358,-2.46104380066281,2.75153061636384,-3.0141507731309,-.104423971232209,.116749549066043,-.127892723210497,-2.94703586087065,3.22831603852648,-4.81780334971999},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,3.13195404884644,-3.83584465679206,4.42925189189755,-4.95205415939119,5.42470353822284,3.97724364248012,-4.59252537503603,5.13459946240544,-5.62467189980938,3.67768288095471,-4.11177446033511,4.5042232468252,2.21014640326349,-2.42109408164419,-.161112776439403},
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,3.40022577929605,-4.1644090826311,4.80864541143983,-5.37622900930936,5.88936380548032,5.17702934684997,-5.97791857375093,6.68351614602547,-7.32142511479571,6.67435503772228,-7.46215578440284,8.17438210188188,7.8090615264033,-8.55439830449838,8.51143945889045}
		};
		for (int j = 0; j < 12; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[12][21] = 
		{
			{1.414213562,3.94468190274016,7.0795491341205,10.545740291788,14.1178337840267,17.5926352342147,-.0319379183519753,-.154660030302817,-.43925822557731,-.961797763577864,-1.79684767953686,.000465578153648587,.00319092434145473,.0123137463285655,.0353622090543082,-5.86464816610367e-006,-5.19109369939388e-005,-.000252650702007359,6.85464900528793e-008,7.43273920925785e-007,-7.65370303624714e-010},
			{1.414213562,3.71235176911142,5.99558490979395,7.62488741580146,8.14599447640013,7.34718311089616,-.166073783576779,-.764836159193308,-2.02218309922343,-4.02885967082133,-6.68010475273478,.0125887190510936,.0823390061849571,.299004079114613,.796298871799325,-.000824565926326729,-.00697737063211093,-.0321260785529637,5.01144631645353e-005,.000520025886887555,-2.90963996691883e-006},
			{1.414213562,3.30970802258292,4.2909622271481,3.64701232940425,1.49376547749553,-1.29952592474851,-.398540259080393,-1.67165845871656,-3.84755914106515,-6.3113399865174,-7.98777931164405,.0724976248699598,.434860760346183,1.40707187748205,3.23180916098999,-.0113956405123911,-.0887334889355571,-.36808995603673,.00166206328995187,.015902832140972,-.00023157669095963},
			{1.414213562,2.76195386285985,2.32621648423849,.160883133644535,-2.12140648579363,-2.7058281817046,-.714786270724957,-2.59853682213183,-4.7549644652192,-5.36731919542456,-3.25729414076559,.233201739437629,1.22672710731592,3.30495822058058,5.89679444984459,-.0657431866024218,-.451524561772794,-1.59193561065057,.017197430813833,.145628931559507,-.00429748986298909},
			{1.414213562,2.10349449699454,.504794589391635,-1.67800190212113,-1.79846700964276,.309494957156941,-1.09494796295235,-3.24473228828999,-4.05017972204568,-1.63294619261043,2.28560008620039,.547225517982271,2.39319484717323,4.87179403320196,5.49730994822615,-.236321146004056,-1.36209257652082,-3.77635721863404,.0946962283839159,.676667877319681,-.0362494356624808},
			{1.414213562,1.37570022553441,-.821946062819915,-1.61733246050546,.359308464240942,1.71165917287845,-1.51514018159912,-3.36446686788906,-1.91964817870762,1.95173675811295,2.75500656898115,1.04781596872157,3.55510980002442,4.59045028887451,1.24891795372526,-.626153181688296,-2.84473124098146,-5.48872263996485,.347192252628353,1.97345041935918,-.183906859961412},
			{1.414213562,.624299774465593,-1.43546196207946,-.337640235075449,1.50389998701775,-.0107903422167659,-1.94896143440088,-2.83314651369184,.637278791777431,2.63766838605638,-.808065843843811,1.73374759666501,4.12740556141283,1.98940127224618,-2.96832239843812,-1.33269841309056,-4.37531622836174,-4.52786743647192,.95054294779362,3.96849611187727,-.647664476709678},
			{1.414213562,-.103494496994541,-1.29720437850508,.991754894783212,.589436063368126,-1.41318416081787,-2.36915365304765,-1.68415540504794,2.32831086437971,.381749900992996,-2.46800676549626,2.56192161158569,3.58714638367868,-1.52344991306454,-2.69130383032704,-2.39387697960251,-4.93737708601551,-.508166439915618,2.07554270972356,5.63164748396134,-1.71909544841957},
			{1.414213562,-.761953862859853,-.55104212557113,1.37646295656649,-1.03415883911242,-.174795669385988,-2.74931534527504,-.106757773657142,2.26282036321899,-2.15044627554921,.119501749587716,3.45007465931297,1.77035093065653,-3.42380239520456,1.33170919606025,-3.74106981000851,-3.58481468643821,3.59724489879787,3.76406389114815,5.23559759098235,-3.6179005979755},
			{1.414213562,-1.30970802258292,.519224820053661,.518731848959601,-1.30425364587756,1.47252576661999,-3.06556135691961,1.5947619510804,.444911232504622,-2.09440936915946,2.63186985254377,4.2894282422279,-.964148801340396,-2.01722273728898,3.46190525493913,-5.18623489409549,-.205483038001613,3.63947886955733,5.81833621896758,1.69243379052584,-6.23568135536041},
			{1.414213562,-1.71235176911142,1.5663330178964,-1.07607439139602,.363617646552178,.420994417051474,-3.29802783242322,3.07100849759875,-2.20598407348367,.923194488569861,.507379758856801,4.96464287844228,-3.80884940873902,2.02329330850018,.0162698697481229,-6.45780747384476,4.10479708970298,-1.32544341327556,7.79428043908953,-4.0355398121645,-8.98680735137036},
			{1.414213562,-1.94468190274016,2.27090372268066,-2.44801058376252,2.49586968540814,-2.4264683064468,-3.43216369764802,4.00974905232829,-4.32546942586304,4.4144312457315,-4.2977488551218,5.37669458746301,-5.80779112496784,5.9385809984625,-5.79724384742923,-7.27823584467216,7.46210387346587,-7.31197439196322,9.14177937313726,-8.9987004064276,-10.9691712594614}
		};
		for (int j = 0; j < 12; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[21][12] = 
		{
			{-.0471753389703784,-.106939314428477,-.16007833916176,-.203167424021368,-.233492536723722,-.249147045798746,-.249147045798746,-.233492536723722,-.203167424021368,-.16007833916176,-.106939314428477,-.0471753389703784},
			{.0648707029602942,.129483784537873,.148249097989688,.109463102050698,.0170873715800573,-.109985117239921,-.2423620138609,-.347295683640528,-.396785228458322,-.374634056214685,-.280718813313607,-.13158670712415},
			{-.0757528111090398,-.118442209516973,-.0587723445986148,.079163297629637,.214173834111807,.252890452071526,.144805168577106,-.083343684650135,-.334186733367935,-.485704652268982,-.453371228180024,-.236159614755104},
			{.0816607421928192,.0813700708167378,-.0587165440093283,-.197744131905342,-.163742854978433,.0594832841179061,.284931226398005,.277045088013018,-.0231126420431299,-.412814367136931,-.576575034174456,-.351784826794551},
			{-.083257226162066,-.0274958626198417,.147631703568395,.14856835834851,-.0973183437761787,-.264947422944133,-.063300653277769,.296934377775289,.304763511236575,-.169083017610384,-.615979854522674,-.470942728759879},
			{.0809421348181849,-.0318345513886016,-.166678842080749,.025111331707413,.23332257818713,.00190097306183677,-.301549099493064,-.0510989036970105,.388722153430519,.147096561093246,-.555575724598485,-.586855162769535},
			{-.114490123783502,-.249388667095643,-.346998489897094,-.394969565625618,-.391157114432837,-.343355485054158,-.26692764812862,-.18078045933897,-.102686955626456,-.0451117599561129,-.0125580867203635,-.00106538514698783},
			{.133757217306151,.232222029374965,.180515058932322,-.0153369352769492,-.278061056920914,-.499125523307164,-.592730124604984,-.535718786284836,-.373308562838685,-.189219165271344,-.0578350093026769,-.00515914961555458},
			{-.144289018132812,-.166811032438666,.050360605399021,.325079179263335,.384413800291541,.112271677012304,-.33819126429111,-.668701504869756,-.683103250662319,-.435514757768147,-.152912459621155,-.0146527768052979},
			{.14725660667997,.0698096725195432,-.237071388762475,-.308935398296182,.0630284952156476,.464687444293257,.343844423679318,-.26960620296335,-.771074782139231,-.71439621956824,-.3046523541133,-.0320836517985888},
			{-.143364315256757,.0383667962154668,.297907873310876,.017167748400487,-.407478174094171,-.142359840909465,.485359330372213,.377362073254748,-.467946339429932,-.904156542888517,-.505132917252383,-.0599392486348982},
			{-.179355789242482,-.375413954291231,-.485531016712943,-.49564139368943,-.422983905496475,-.305440495868931,-.184597475250337,-.0903492073046713,-.0335020098285881,-.0082061858887987,-.000951927643970896,-1.55307570187091e-005},
			{.193736308438081,.288015724966542,.109134393057063,-.254330496941975,-.592252776234179,-.727139754226691,-.626316369393508,-.395126415458444,-.176232920497942,-.0492229675323306,-.00622627098487373,-.000106442860825517},
			{-.198099197414935,-.152996410947114,.228334442676318,.491867092554372,.251527912283571,-.350479915627058,-.80871599529964,-.804353442629352,-.474793812063302,-.159269954185708,-.0226099452664103,-.000410761975881334},
			{.193384472331989,-.00123028710976741,-.391861638259147,-.191314758984308,.444345447527085,.522939940635989,-.220026328987702,-.907628719438496,-.84713976015901,-.365816490950941,-.0602141413953284,-.00117961264325513},
			{-.242787406459337,-.488323349567532,-.587042784922414,-.537446066425504,-.395239038357479,-.234786231237537,-.110311638598763,-.0390176033630983,-.00944473608571052,-.00128990080067228,-6.23516257983926e-005,-1.95632945654767e-007},
			{.248920876546637,.31039455298989,-.0232591344869702,-.514998289242352,-.815181476789607,-.77081506036106,-.501166445756355,-.224887145281521,-.0648664986069148,-.0100439635929164,-.000527611424168513,-1.73164514361146e-006},
			{-.243912856990113,-.100226736373669,.411961635922747,.516784026765791,-.0839003911851568,-.797690550622854,-.966967837799883,-.623492271451269,-.228699162378071,-.041665014660205,-.00242929420722927,-8.42792264012015e-006},
			{-.304951495361639,-.589384113245934,-.65859190093498,-.540749420569416,-.34268072739361,-.1674605403216,-.0611660970770167,-.0156347408665742,-.00247060119508803,-.000188133064231527,-3.78953113918537e-006,-2.28657407372626e-009},
			{.300178666892258,.305157489809165,-.191570776483383,-.752151516427833,-.929808404899219,-.699144109902361,-.347669796815464,-.111720678797647,-.0209212071410637,-.00180008099494654,-3.9323064981717e-005,-2.47941342577084e-008},
			{-.365909637710154,-.679560034193271,-.705832231568849,-.519751446470816,-.283829803146802,-.114101360142282,-.0323995272485856,-.00598493248162283,-.000617381962821684,-2.62127397601647e-005,-2.20019739940013e-007,-2.55305407260627e-011}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 12; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{.0471753389703784,.106939314428477,.16007833916176,.203167424021368,.233492536723722,.249147045798746,.249147045798746,.233492536723722,.203167424021368,.16007833916176,.106939314428477,.0471753389703784},
			{-.057777754465245,-.130973376908003,-.196055124926603,-.248828260626037,-.285968786885903,-.305141566591389,-.305141566591389,-.285968786885903,-.248828260626037,-.196055124926603,-.130973376908003,-.057777754465245},
			{.0667160041638562,.151235028775734,.226384958224997,.287322126407624,.33020831206047,.352347131100821,.352347131100821,.33020831206047,.287322126407624,.226384958224997,.151235028775734,.0667160041638562},
			{-.0745907602644774,-.169085902496444,-.253106077890569,-.32123590311126,-.369184116329077,-.393936068492185,-.393936068492185,-.369184116329077,-.32123590311126,-.253106077890569,-.169085902496444,-.0745907602644774},
			{.0817100839813177,.18522432596281,.277263816688425,.35189630093549,.404420936894292,.43153534198653,.43153534198653,.404420936894292,.35189630093549,.277263816688425,.18522432596281,.0817100839813177},
			{.080203401857055,.167464509402566,.213466153925738,.20667501558646,.148758759444076,.0540426417701356,-.0540426417701356,-.148758759444076,-.20667501558646,-.213466153925738,-.167464509402566,-.080203401857055},
			{-.0982287050422223,-.20510129892574,-.261441577102187,-.25312416525451,-.182191527610292,-.0661884483104899,.0661884483104899,.182191527610292,.25312416525451,.261441577102187,.20510129892574,.0982287050422223},
			{.113424738636514,.236830580375279,.301886729940981,.292282609999163,.210376655093866,.0764278369255381,-.0764278369255381,-.210376655093866,-.292282609999163,-.301886729940981,-.236830580375279,-.113424738636514},
			{-.126812712948275,-.264784638408991,-.337519624793518,-.326781892267586,-.235208250811417,-.0854489193610162,.0854489193610162,.235208250811417,.326781892267586,.337519624793518,.264784638408991,.126812712948275},
			{.138916366921706,.290057038677642,.369734224203667,.357971627561736,.257657729363923,.0936046012978199,-.0936046012978199,-.257657729363923,-.357971627561736,-.369734224203667,-.290057038677642,-.138916366921706},
			{-.122114057148622,-.212661854220922,-.170589599267302,-.00968885598258523,.189947318219643,.325107053658344,.325107053658344,.189947318219644,-.00968885598258477,-.170589599267302,-.212661854220921,-.122114057148622},
			{.14100516752701,.245560757550869,.196979902107224,.01118772721888,-.219332270602307,-.375401289873734,-.375401289873734,-.219332270602308,.0111877272188795,.196979902107224,.245560757550869,.14100516752701},
			{-.157648569851374,-.274545273187232,-.22023022561013,-.0125082592849363,.245220933311432,.419711401411007,.419711401411007,.245220933311433,-.0125082592849357,-.220230225610129,-.274545273187232,-.157648569851374},
			{.172695355768364,.300749278425367,.241250124872966,.0137021115338472,-.268626073550765,-.459770804486501,-.459770804486501,-.268626073550765,.0137021115338465,.241250124872965,.300749278425367,.172695355768364},
			{.157432480656601,.196644442987321,-.00835800794176406,-.284688341235506,-.373334915626667,-.170540758074208,.170540758074208,.373334915626668,.284688341235505,.00835800794176374,-.196644442987321,-.1574324806566},
			{-.176014864289488,-.219855170936379,.00934453695618862,.31829124166993,.41740112481547,.190670363974853,-.190670363974853,-.417401124815472,-.31829124166993,-.00934453695618827,.219855170936379,.176014864289487},
			{.192814623264837,.240839273024148,-.0102364273612059,-.348670585855459,-.457240023201953,-.208868918804014,.208868918804014,.457240023201955,.348670585855458,.0102364273612055,-.240839273024148,-.192814623264836},
			{-.184197562274308,-.118175770575784,.235878303029161,.383530106382951,.0579093560340826,-.374944419288365,-.374944419288369,.0579093560340806,.383530106382954,.235878303029161,-.118175770575781,-.184197562274308},
			{.201778319851171,.12945507062894,-.258391734868296,-.420136181623543,-.0634365212000975,.41073103263393,.410731032633934,-.0634365212000954,-.420136181623546,-.258391734868296,.129455070628936,.20177831985117},
			{.20076502079953,-.00861464492903962,-.385612979315631,-.141367941779104,.412110586758183,.311823666565035,-.311823666565045,-.412110586758192,.141367941779136,.38561297931557,.0086146449290491,-.200765020799505}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 12; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[21][12] = 
		{
			{-.0333580020819281,-.0756175143878672,-.113192479112499,-.143661063203812,-.165104156030235,-.176173565550411,-.176173565550411,-.165104156030235,-.143661063203812,-.113192479112499,-.0756175143878672,-.0333580020819281},
			{-.0930458529460421,-.198498176553059,-.264906281682713,-.280569525791552,-.245575133043822,-.171375823548283,-.0777712222504629,.0120825963201004,.0774021017701836,.104827942520953,.0915588621245824,.0458705139756637},
			{-.166990065079791,-.320581869925543,-.343445053363858,-.236305705409389,-.0589328846007312,.102392716678746,.178820553604284,.151443770493135,.055976904589773,-.0415583234228769,-.0837512895502642,-.0535653264432772},
			{-.248749436479337,-.407700116420065,-.291903838296748,-.0163431059155229,.195900460376747,.201476802304676,.042061033555921,-.11578368309555,-.139826216573226,-.041518866425881,.0575373288449655,.0577428645460353},
			{-.33300679714446,-.435563532322195,-.119559748367329,.215500345510458,.209964312147707,-.0447603211980575,-.187346119471122,-.0688144608361341,.105053693685699,.104391378738875,-.0194425109181931,-.0588717492175104},
			{-.41496926527815,-.39285136242986,.104012975865696,.274868070760673,-.0361323813248869,-.213227413168496,.00134419094323235,.164983977283574,.017756392939621,-.117859739546712,-.0225104271688505,.0572347324287539},
			{.000753341062131352,.008879908280122,.0318988313813435,.072610642674515,.127831088725108,.188746350108197,.242788991878333,.276589848169184,.279285658260975,.245364985307081,.176344417682688,.0809567429191864},
			{.00364806967735219,.0408955272571209,.133798154858527,.263969016188599,.378810386491097,.419123490411159,.352935042100669,.196618858880805,.0108448509340889,-.12764342224366,-.164205771668619,-.0945806353648701},
			{.0103610778435682,.108125437139903,.307955438564179,.483026940855864,.472843368743711,.239137336348973,-.0793880641607704,-.27182160500271,-.22986569210915,-.0356103255868759,.117953212229241,.102027743185548},
			{.0226865677588688,.215421745563086,.505154411463456,.545232207417426,.190640374422976,-.243134723730331,-.328583643091346,-.0445678763884421,.21845031514984,.167634786669938,-.0493628928459047,-.104126145189406},
			{.0423834491801445,.357182911283934,.639335222899324,.330888029929621,-.266835281031443,-.343200873908896,.100663608902278,.288130580163515,-.0121394313148915,-.210652677442557,-.0271294217835161,.101373879524963},
			{-1.09819036077875e-005,-.000673114492428329,-.00580264969117775,-.0236894983394228,-.0638865371768165,-.130530126573866,-.215979045934882,-.299094787988234,-.350471390607008,-.343322274484695,-.265457752901419,-.126823694851883},
			{-7.52664687184791e-005,-.00440263843607072,-.0348058941414186,-.12461549318528,-.279396567870308,-.442872552083127,-.514165451219659,-.41878595436223,-.17983881909766,.0771696694116863,.20365787226593,.136992257494756},
			{-.000290452578522664,-.0159876456159174,-.112620864614273,-.335729924086801,-.568763773603908,-.571848564179541,-.247826724944201,.177857092386487,.347802556495951,.16145683275228,-.108184799649367,-.140077285802758},
			{-.000834112099439185,-.0425778277151948,-.25867132148952,-.599018269179214,-.641790422483938,-.155582109307832,.369774378274552,.314199679218665,-.135279963454564,-.277088021773013,-.000869944358352441,.136743471798207},
			{1.38333382588974e-007,4.40892574312313e-005,.00091209760345402,.00667843693448509,.0275896119309721,.0780021077175621,.166018936281095,.279476204285956,.380031758191769,.415101934174787,.345296751982027,.171676621539384},
			{1.22445802447951e-006,.000373077615955666,.00710215476841621,.0458675410488779,.159019225472198,.354378192390965,.545048556365818,.576420350287669,.36415878271881,.0164466917246055,-.219482093320422,-.176013639831463},
			{5.95944125415059e-006,.00171777040786467,.0294616144122425,.161714728611881,.440875613276901,.683749515477967,.56405239778266,.0593265355668802,-.365421489831386,-.291300866426529,.0708710049647147,.172472435241796},
			{-1.61685195173712e-009,-2.67960316688328e-006,-.000133030165518613,-.00174697885911607,-.0110554312917647,-.043250962033283,-.118412483673806,-.242311866185891,-.382367582308209,-.465694799308523,-.416757503309762,-.215633270360082},
			{-1.7532099583758e-008,-2.78056059144079e-005,-.00127284947854753,-.0147935274439646,-.0789984495974143,-.245839671006818,-.494369541269015,-.657473828481949,-.531851437886183,-.13546099516431,.215778930430848,.212258370983052},
			{1.80533017556786e-011,1.55577451647998e-007,1.85352060386969e-005,.000436554972610718,.00423198634381799,.0229099254307547,.0806818455205034,.200697978560845,.367519772428026,.499098757454024,.480521508528172,.258737186194613}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 12; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[21][21] = 
		{
			{.5000000000,-.7071067810,2.309401077,-8.500000000,33.98823325,-144.1116465,-.4082482906,1.750000000,-7.678758582,34.53410150,-159.0586197,.3227486122,-2.459674775,14.75000000,-81.71498915,-.2645751311,2.958039892,-22.92947780,.2236067977,-3.324307508,-.1934294858},
			{0.,-.7500000000,2.857738034,-12.37436867,55.33985905,-254.3227936,-.7216878367,2.828427124,-13.26806944,63.17066828,-305.0714286,.9128709293,-5.850213671,33.43604921,-182.9477455,-.9354143468,8.844691708,-63.03769304,.9035079028,-11.53669556,-.8548456319},
			{0.,0.,1.166666667,-7.216878367,40.02082790,-212.3677366,0.,1.587713241,-9.500000000,52.88300766,-284.1991313,.9316949905,-6.067673910,36.47616523,-208.0853992,-1.636634177,13.68699853,-93.25932554,2.074812507,-23.01464213,-2.326595118},
			{0.,0.,0.,-1.875000000,15.87608264,-106.7569280,0.,0.,-2.901185103,23.88339730,-158.7252693,0.,-2.627379873,21.21428571,-141.7297322,-1.218935425,11.40958244,-83.44725167,2.747169230,-28.60654093,-4.158733944},
			{0.,0.,0.,0.,3.100000000,-33.04592764,0.,0.,0.,5.080682370,-53.17144007,0.,0.,5.494338457,-55.95917640,0.,4.138710979,-42.40452863,1.621428571,-20.30485767,-4.462835264},
			{0.,0.,0.,0.,0.,-5.250000000,0.,0.,0.,0.,-8.845830912,0.,0.,0.,-10.42167396,0.,0.,-9.706337648,0.,-6.398809524,-2.185359011},
			{0.,.4330127020,-.7071067810,2.245365598,-7.850689990,29.50000000,-.2500000000,0.,.4949747467,-3.162277660,16.25653401,.3162277660,-1.004158022,3.091022771,-9.096005092,-.3240370350,2.070196679,-10.27118455,.3129843186,-3.053571429,-.2961272134},
			{0.,0.,-.8660254040,3.500000000,-14.98165545,64.91147819,0.,-.2500000000,1.558845727,-8.714212528,44.99219432,.5809475019,-1.565247584,4.964285714,-14.04517129,-1.020504077,5.704791219,-27.63572997,1.293725044,-11.16850681,-1.450721144},
			{0.,0.,0.,1.659882024,-10.71525393,59.18483757,0.,0.,1.275000000,-8.720665110,51.70172994,0.,-.5486726408,1.773290113,-2.439471337,-.9983468483,5.709015429,-29.57691077,2.250018896,-17.91000460,-3.406135253},
			{0.,0.,0.,0.,-3.117691454,27.51181564,0.,0.,0.,-3.300000000,29.34227987,0.,0.,-.4979550017,6.515483910,0.,1.996693696,-13.84473139,1.554721797,-14.02696019,-4.279230906},
			{0.,0.,0.,0.,0.,5.802370207,0.,0.,0.,0.,7.221428571,0.,0.,0.,3.471852928,0.,0.,-1.980327354,0.,-4.377552220,-2.322071589},
			{0.,0.,.3726779962,-.6454972245,2.020725943,-6.851601598,0.,-.3227486122,.2236067977,-.2500000000,-.4433944513,.8333333333e-1,.2886751347,-1.152673615,4.225804810,-.1463850110,.1636634177,.5511331920,.1855768722,-.8658617402,-.2080969936},
			{0.,0.,0.,-.8385254914,3.500000000,-14.87979615,0.,0.,.9682458365e-1,.2886751347,-3.546268517,0.,.6250000000,-2.395787117,9.681316602,-.2746751327,-.1889822365,3.496332169,.6190476190,-2.256095297,-.9371298718},
			{0.,0.,0.,0.,1.788854382,-11.83920042,0.,0.,0.,.6454972245,-5.353808485,0.,0.,-1.357142857,7.368172496,0.,-.6338656908,5.580637822,.6016087651,-1.545511386,-1.655873628},
			{0.,0.,0.,0.,0.,-3.689512162,0.,0.,0.,0.,-2.517439175,0.,0.,0.,1.767857143,0.,0.,2.833614404,0.,.9849347042e-1,-1.114302550},
			{0.,0.,0.,.3307189139,-.5916079782,1.836209865,0.,0.,-.3436931772,.3415650256,-.6681531049,0.,.1479019946,.1889822365,-.9660917834,-.2500000000e-1,-.2236067977,.5443310539,.5634361697e-1,.1080123450,-.8529438595e-1},
			{0.,0.,0.,0.,-.7937253933,3.381321241,0.,0.,0.,.3055050464,-.5976143047,0.,0.,.5070925528,-2.391701925,0.,-.4500000000,.7607257745,.1070899340,.6555622816,-.2947553424},
			{0.,0.,0.,0.,0.,1.807930062,0.,0.,0.,0.,.1200198396,0.,0.,0.,-1.598750131,0.,0.,.3055555556e-1,0.,.8094739131,-.2864842269},
			{0.,0.,0.,0.,.3000000000,-.5477225576,0.,0.,0.,-.3464101616,.4065785564,0.,0.,.1916629694,.8748177653e-1,0.,-.5669467095e-1,-.2415229459,.7142857143e-2,.1173691195,-.1966006724e-1},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,.4330127020,0.,0.,0.,.3593680677,0.,0.,-.5197011503,0.,.2321428571,-.3768891806e-1},
			{0.,0.,0.,0.,0.,.2763853992,0.,0.,0.,0.,-.3419382528,0.,0.,0.,.2207201930,0.,0.,-.8705321812e-1,0.,.1974181422e-1,-.1984126984e-2}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[21][21] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,-.8164965809,-.2500000000,-.1154700539,-.6454972245e-1,-.4040610177e-1,2.904737510,1.341640786,.7500000000,.4694764777,-11.64130577,-6.507687761,-4.073608438,50.31152948,31.49343956,-228.6336522},
			{0.,-.2500000000,0.,0.,0.,0.,.7216878367,.3535533905,.1632993162,.9128709293e-1,.5714285714e-1,-2.464751509,-1.581138830,-.9596449171,-.6224437520,10.85080642,7.410417377,5.024466921,-50.14468862,-35.90912479,239.3225832},
			{0.,0.,.1666666667,0.,0.,0.,0.,-.4330127020,-.2000000000,-.1118033988,-.6998542123e-1,1.155301788,1.290994449,.9691236663,.6776309270,-6.197388082,-6.026182953,-4.798824078,33.05867928,29.90051531,-173.7315107},
			{0.,0.,0.,-.1250000000,0.,0.,0.,0.,.3031088914,.1290994449,.8081220354e-1,0.,-.7267220925,-.7857142857,-.6455301571,1.974864371,3.507390157,3.710995566,-14.50782199,-18.91064403,91.34707467},
			{0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,-.2309401077,-.9035079028e-1,0.,0.,.5111012519,.5248906592,0.,-1.266180985,-2.208209791,3.502380952,8.554904710,-32.70648785},
			{0.,0.,0.,0.,0.,-.8333333333e-1,0.,0.,0.,0.,.1855768722,0.,0.,0.,-.3859879246,0.,0.,.8871666102,0.,-2.267857143,6.356265942},
			{0.,-.4330127020,0.,0.,0.,0.,-.7500000000,.6123724358,.2828427124,.1581138830,.9897433188e-1,3.320391543,-1.278019301,-1.195584279,-.8864412337,-15.76980237,1.794170455,4.167058315,75.89869728,4.660714286,-371.5212018},
			{0.,0.,.2886751347,0.,0.,0.,0.,-.2500000000,-.3464101616,-.1936491673,-.1212183053,-1.484643616,.4472135954,1.107142857,.9389529552,8.919961562,-.8451542550e-1,-3.528403268,-49.83237207,-7.640075151,269.1248912},
			{0.,0.,0.,-.2165063510,0.,0.,0.,0.,.3250000000,.2236067977,.1399708425,0.,.3227486122e-1,-.7010681843,-.8470386588,-2.776822654,-.7563225568,2.151452997,21.71637090,7.490281115,-141.0307509},
			{0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,-.3000000000,-.1564921593,0.,0.,.2766416676,.6060915266,0.,.5237229368,-.8366600265,-5.171408842,-4.562714910,50.23385756},
			{0.,0.,0.,0.,0.,-.1443375673,0.,0.,0.,0.,.2642857143,0.,0.,0.,-.3365806955,0.,0.,.1400231463,0.,1.482553013,-9.676023611},
			{0.,0.,.3726779962,0.,0.,0.,0.,.9682458365,-.4472135954,-.2500000000,-.1564921593,1.083333333,-4.041451885,-.4610694458e-1,.6060915266,-8.002380602,16.80277755,5.405089377,48.86857636,-68.44072363,-278.3505387},
			{0.,0.,0.,-.2795084971,0.,0.,0.,0.,-.9682458365e-1,.2886751347,.1807015806,0.,1.375000000,.7985957060,-.3936679944,2.387560769,-7.275816105,-5.503272838,-21.01190476,34.70865684,144.9163216},
			{0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,-.1290994449,-.2020305089,0.,0.,-.6428571429,0.,0.,1.774823935,3.548977048,4.876757779,-11.83920042,-51.10055324},
			{0.,0.,0.,0.,0.,-.1863389981,0.,0.,0.,0.,.1936491673,0.,0.,0.,.2083333333,0.,0.,-1.312336745,0.,2.180166278,9.676640803},
			{0.,0.,0.,-.3307189139,0.,0.,0.,0.,-1.031079532,.3415650256,.2138089935,0.,-1.922725929,3.968626966,.7763237544,-1.575000000,13.75181806,-13.04353288,16.26921940,-78.67156297,-122.2865611},
			{0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,.3055050464,-.2390457218,0.,0.,-.8451542550,-1.388730150,0.,-3.450000000,3.286335346,-3.647357166,28.56871416,42.50718808},
			{0.,0.,0.,0.,0.,-.2204792759,0.,0.,0.,0.,-.3273268354e-1,0.,0.,0.,.7535958771,0.,0.,-.1750000000,0.,-5.735610878,-7.860430760},
			{0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,1.039230485,-.2710523708,0.,0.,2.491618602,-3.674234614,0.,3.571764270,-16.97561276,2.292857143,-36.14968880,-30.41019201},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.2795084971,0.,0.,3.354434698,0.,7.125000000,5.464893121},
			{0.,0.,0.,0.,0.,-.2763853992,0.,0.,0.,0.,-1.025814759,0.,0.,0.,-2.869362509,0.,0.,-5.484352741,0.,-6.337122368,-3.339285714}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[21][21] = 
		{
			{1.,-.7071067810,.1443375673,.1250000000,-.5590169942e-1,-.5103103630e-1,.4082482906,.1250000000,-.2453738645,.3227486122e-1,.1035406358,-.1613743061,.5590169942e-1,.1250000000,-.1189992461,.6614378278e-1,-.7395099729e-1,-.2507429436e-1,-.2795084971e-1,.5467611033e-1,.1208934286e-1},
			{0.,-.2500000000,.4082482906,-.2651650429,0.,.9923207755e-1,.7216878367,-.7071067810,.3163924252,.9128709293e-1,-.1638392857,.3651483717,.3952847075e-1,-.2525381361,.1276586028,-.1403121520,.8964214566e-1,.8558232884e-1,.5646924394e-1,-.8312297405e-1,-.2350825488e-1},
			{0.,0.,.8333333333e-1,-.2165063510,.2420614591,-.1031197389,0.,-.2165063510,.4250000000,-.3913118960,.1159133539,.5776508940,-.6132223630,.3608439183,.1364673394e-1,.3218713882,-.3049687729e-2,-.2191913380,-.1210307296,.9785852773e-1,.4769519990e-1},
			{0.,0.,0.,-.3125000000e-1,.1118033988,-.1786086271,0.,0.,.7577722285e-1,-.2259240285,.3169353608,0.,-.1816805231,.3750000000,-.4018262214,.4937160930,-.5387858374,.3603626151,.2888254471,-.2369298115e-1,-.1067891953},
			{0.,0.,0.,0.,.1250000000e-1,-.5705443310e-1,0.,0.,0.,-.2886751347e-1,.1157619501,0.,0.,.6388765648e-1,-.1986565342,0.,-.1582726231,.3313752322,.4377976190,-.4833325546,.2635268179},
			{0.,0.,0.,0.,0.,-.5208333333e-2,0.,0.,0.,0.,.1159855452e-1,0.,0.,0.,-.2412424530e-1,0.,0.,.5544791314e-1,0.,-.1417410714,.3972666214},
			{0.,-.4330127020,.7071067810,-.4592793268,0.,.1718750000,-.7500000000,0.,.1237436867,.1581138830,-.2157331140,-.3162277660,-.1141088661,.2916059217e-1,.1746928107e-1,.1890216038,-.1725163898e-1,-.7854574205e-1,-.9780759959e-1,.7031250000e-1,.4812067217e-1},
			{0.,0.,.1443375673,-.3750000000,.4192627457,-.1786086271,0.,-.1250000000,.3897114318,-.4357106264,.1300571401,-.7423218081,.2795084971,.8928571429e-2,.5297911641e-1,-.3496171374,-.8451542550e-1,.1102626021,.1796840339,-.5649864734e-1,-.8664029051e-1},
			{0.,0.,0.,-.5412658775e-1,.1936491673,-.3093592167,0.,0.,.8125000000e-1,-.2795084971,.4264736607,0.,.8068715303e-2,.1340277411,-.2555233287,-.6942056633,.3964594048,-.9743899444e-1,-.3527181261,-.5113897256e-1,.1668168700},
			{0.,0.,0.,0.,.2165063510e-1,-.9882117688e-1,0.,0.,0.,-.3750000000e-1,.1613825392,0.,0.,.3458020845e-1,-.1546796083,0.,.6546536711e-1,-.8715208612e-2,-.6464261051,.4446952959,-.3454883043},
			{0.,0.,0.,0.,0.,-.9021097958e-2,0.,0.,0.,0.,.1651785714e-1,0.,0.,0.,-.2103629347e-1,0.,0.,.8751446639e-2,0.,.9265956332e-1,-.6047514756},
			{0.,0.,.1863389981,-.4841229182,.5412658775,-.2305827461,0.,.4841229182,-.3913118960,.6250000000e-1,-.1467113994e-1,.5416666667,.3608439183,-.1959545145,-.1210078569,.1341862601,.9547032698e-1,.1373682881,-.1546473936,-.6352789941e-1,.1066497092},
			{0.,0.,0.,-.6987712428e-1,.2500000000,-.3993810316,0.,0.,-.2420614591e-1,-.7216878367e-1,.2343473623,0.,.3437500000,-.5190872089,.3481045690,.5968901920,.1653594569,-.2206772153,.1994047619,.1132938028,-.1630966411},
			{0.,0.,0.,0.,.2795084971e-1,-.1275775908,0.,0.,0.,-.1613743061e-1,.1073287078,0.,0.,-.8035714286e-1,.1426360826,0.,.2218529919,-.4548734170,.6095947223,.1275775908e-1,.2375739044},
			{0.,0.,0.,0.,0.,-.1164618738e-1,0.,0.,0.,0.,.1210307296e-1,0.,0.,0.,.1302083333e-1,0.,0.,-.8202104657e-1,0.,.1362603923,.6047900501},
			{0.,0.,0.,-.8267972847e-1,.2958039891,-.4725540094,0.,0.,-.2577698829,.4269562821,-.2839650695,0.,-.4806814823,.4724555913e-1,.2566181299,-.3937500000,-.4192627457,.3827327723e-1,-.1056442818e-1,-.2025231469e-1,.1055518026},
			{0.,0.,0.,0.,.3307189139e-1,-.1509518412,0.,0.,0.,.3818813080e-1,-.5229125166e-1,0.,0.,-.1056442819,.3423605578,0.,-.4312500000,.2909776087,-.4559196456,-.3558150542,-.6827053884e-1},
			{0.,0.,0.,0.,0.,-.1377995474e-1,0.,0.,0.,0.,-.2045792721e-2,0.,0.,0.,.4709974231e-1,0.,0.,-.1093750000e-1,0.,-.3584756798,-.4912769225},
			{0.,0.,0.,0.,.3750000000e-1,-.1711632992,0.,0.,0.,.1299038106,-.3303450769,0.,0.,.3114523254,-.2679129406,0.,.4464705337,.1746728448,.2866071429,.3594429284,-.5971745424e-1},
			{0.,0.,0.,0.,0.,-.1562500000e-1,0.,0.,0.,0.,-.2706329388e-1,0.,0.,0.,.1746928107e-1,0.,0.,.2096521686,0.,.4453125000,.3415558198},
			{0.,0.,0.,0.,0.,-.1727408745e-1,0.,0.,0.,0.,-.6411342242e-1,0.,0.,0.,-.1793351568,0.,0.,-.3427720463,0.,-.3960701481,-.2087053571}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[21][21] = 
		{
			{1.,.7071067810,-.1443375673,-.1250000000,.5590169942e-1,.5103103630e-1,-.4082482906,-.1250000000,.2453738645,-.3227486122e-1,-.1035406358,.1613743061,-.5590169942e-1,-.1250000000,.1189992461,-.6614378278e-1,.7395099729e-1,.2507429436e-1,.2795084971e-1,-.5467611033e-1,-.1208934286e-1},
			{0.,-.7500000000,-.4082482906,.2651650429,0.,-.9923207755e-1,-.7216878367,-.3535533905,.5103103630e-1,-.9128709293e-1,.1049107143,.4564354646,.1185854122,-.1515228816,.4870046022e-1,-.2338535867,.5976143047e-1,.1108137724,.1129384878,-.1024538982,-.5342785200e-1},
			{0.,0.,.5833333333,.2165063510,-.2420614591,.1031197389,0.,.7938566203,.3750000000,-.1677050983,.4738596230e-1,.4658474951,.9682458365e-1,.3092947872e-1,.1162325270,-.4091585443,-.1189378215,.3433564566e-1,.2593515633,-.1788811797e-1,-.1454121949},
			{0.,0.,0.,-.4687500000,-.1118033988,.1786086271,0.,0.,-.7252962758,-.2581988897,.2133947249,0.,-.6568449682,-.2857142857,.3504772316e-1,-.3047338563,.6338656908e-1,-.1639473093e-1,.3433961537,.9294938757e-1,-.2599208715},
			{0.,0.,0.,0.,.3875000000,.5705443310e-1,0.,0.,0.,.6350852963,.1552904208,0.,0.,.6867923071,.2642678667,0.,.5173388725,.1832986642,.2026785714,-.1589373493,-.2789272040},
			{0.,0.,0.,0.,0.,-.3281250000,0.,0.,0.,0.,-.5528644321,0.,0.,0.,-.6513546229,0.,0.,-.6066461029,0.,-.3999255952,-.1365849382},
			{0.,.4330127020,.7071067810,.2551551816,-.1825741858,-.1093750000,-.2500000000,-.6123724358,-.4065863991,.1581138830,.2095472184,.1581138830,.2966830521,.2916059217e-1,-.2131252291,-.8100925875e-1,-.1035098339,.8090801999e-1,.3912303983e-1,.2343750000e-1,-.1850795084e-1},
			{0.,0.,-.4330127020,-.6250000000,-.5590169942e-1,.2296396634,0.,-.1250000000,-.4330127020e-1,.4841229182e-1,-.1224809960,.2904737510,.6149186937,.2767857143,-.1149239294,-.2551260193,-.4278593414,-.8679563432e-2,.1617156305,.1877213121,-.9067007150e-1},
			{0.,0.,0.,.4149705060,.5163977793,-.7954951286e-1,0.,0.,.3187500000,.3354101966,-.6561133241e-2,0.,-.1371681602,-.3917733971,-.1143502189,-.2495867120,-.4025587803,-.7962157829e-1,.2812523620,.4211073418,-.2128834533},
			{0.,0.,0.,0.,-.3897114318,-.4150489429,0.,0.,0.,-.4125000000,-.4254630581,0.,0.,-.6224437520e-1,.6629126072e-1,0.,.2495867120,.5017470100,.1943402247,.1877602361,-.2674519317},
			{0.,0.,0.,0.,0.,.3626481379,0.,0.,0.,0.,.4513392857,0.,0.,0.,.2169908080,0.,0.,-.1237704596,0.,-.2735970138,-.1451294743},
			{0.,0.,.1863389981,.4841229182,.4330127020,.3294039230e-1,0.,-.1613743061,-.5031152948,-.5625000000,-.1189992461,.4166666667e-1,.2165063510,.3803822929,.1967692977,-.3659625275e-1,-.1432054905,-.1672490034,.2319710903e-1,.7293944007e-1,-.1300606210e-1},
			{0.,0.,0.,-.2096313728,-.5000000000,-.3309157119,0.,0.,.2420614591e-1,.1443375673,.1835250428,0.,.1562500000,.3992978531,.3043636809,-.6866878320e-1,-.3307189139,-.4995787124,.7738095238e-1,.2828269754,-.5857061698e-1},
			{0.,0.,0.,0.,.2236067977,.4847948451,0.,0.,0.,.8068715303e-1,.1073287078,0.,0.,-.1696428571,-.4686614144,0.,-.7923321136e-1,-.8518830777e-1,.7520109565e-1,.3152989029,-.1034921017},
			{0.,0.,0.,0.,0.,-.2305945101,0.,0.,0.,0.,-.1573399484,0.,0.,0.,.1104910714,0.,0.,.1771009002,0.,.6155841900e-2,-.6964390938e-1},
			{0.,0.,0.,.8267972847e-1,.2958039891,.4185478369,0.,0.,-.8592329430e-1,-.3415650256,-.5445447805,0.,.3697549864e-1,.1889822365,.3860054223,-.6250000000e-2,-.5590169942e-1,-.1726550062,.7042952119e-2,.4532660906e-1,-.5330899122e-2},
			{0.,0.,0.,0.,-.9921567416e-1,-.3320940506,0.,0.,0.,.3818813080e-1,.1718141126,0.,0.,.6338656908e-1,.1977011672,0.,-.5625000000e-1,-.2605485777,.1338624175e-1,.1142921083,-.1842220890e-1},
			{0.,0.,0.,0.,0.,.1129956289,0.,0.,0.,0.,.7501239977e-2,0.,0.,0.,-.9992188321e-1,0.,0.,.1909722222e-2,0.,.5059211956e-1,-.1790526418e-1},
			{0.,0.,0.,0.,.3750000000e-1,.1711632992,0.,0.,0.,-.4330127020e-1,-.2117596647,0.,0.,.2395787117e-1,.1366902758,0.,-.7086833870e-2,-.5391137183e-1,.8928571429e-3,.1222594995e-1,-.1228754203e-2},
			{0.,0.,0.,0.,0.,-.4687500000e-1,0.,0.,0.,0.,.2706329388e-1,0.,0.,0.,.2246050424e-1,0.,0.,-.3248132189e-1,0.,.1450892857e-1,-.2355557379e-2},
			{0.,0.,0.,0.,0.,.1727408745e-1,0.,0.,0.,0.,-.2137114080e-1,0.,0.,0.,.1379501207e-1,0.,0.,-.5440826132e-2,0.,.1233863389e-2,-.1240079365e-3}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[21][21] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.3367175148e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.6734350295e-2,.1178511302,.2357022604e-1,.7856742012e-2,.3367175148e-2,.7071067810e-1,.1178511302e-1,.3367175148e-2,.4714045206e-1,.6734350295e-2,.3367175148e-1},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,-.2976190476e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.5952380952e-3,.1000000000,.1190476190e-1,.2380952381e-2,.7619047619e-1,.8333333333e-2,.5952380952e-1},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,.1620032899e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,-.1944039478e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,-.1944039478e-2,.5248906592e-1,0.,-.1296026319e-2,.5248906592e-1,.1944039478e-2,.4860098696e-1},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,-.5611958579e-2,0.,0.,.3367175148e-2,.3086577219e-2,.2244783431e-2,0.,-.5050762721e-2,-.7482611439e-3,.2805979290e-3,.1010152544e-1,-.3928371006e-2,-.1122391716e-2,.1795826746e-1,-.2244783431e-2,.2244783431e-1},
			{0.,0.,0.,0.,.5019488348e-3,.1140792807e-2,0.,0.,0.,-.6274360436e-3,-.8670025331e-3,0.,0.,.8365813917e-3,.4563171227e-3,0.,-.1254872087e-2,.2281585613e-3,.2509744174e-2,-.1597109929e-2,.5703964034e-2},
			{0.,0.,0.,0.,0.,-.1041396590e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,-.1562094885e-3,0.,0.,.2082793179e-3,0.,-.3124189770e-3,.6248379539e-3},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,.5154913119e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,.8247860991e-2,0.,.9622504490e-2,.5498573994e-2,.3092947872e-2,0.,.4123930495e-2,.2061965247e-2,0.,.2061965247e-2,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,-.2805979290e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,.1122391716e-2,0.,.1346870059e-1,.5892556509e-2,.2525381361e-2,0.,.7576144082e-2,.3086577219e-2,0.,.4489566864e-2,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,.9720197392e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,-.3110463166e-2,0.,.4860098696e-2,0.,-.8748177653e-3,0.,.4860098696e-2,.9720197392e-3,0.,.3888078956e-2,0.},
			{0.,0.,0.,0.,-.8694008848e-3,-.1975911102e-2,0.,0.,0.,.9780759957e-3,.1343619549e-2,0.,0.,-.1086751106e-2,-.5334959976e-3,0.,.1086751106e-2,-.4939777755e-3,0.,.1580728882e-2,0.},
			{0.,0.,0.,0.,0.,.1803751804e-3,0.,0.,0.,0.,-.2020202020e-3,0.,0.,0.,.2272727273e-3,0.,0.,-.2525252525e-3,0.,.2525252525e-3,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,.3622503687e-1,0.,0.,.4347004424e-2,.4890379978e-2,.4347004424e-2,0.,0.,.1086751106e-2,.1086751106e-2,0.,0.,.3622503687e-3,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,-.1254872087e-1,0.,0.,.3764616260e-2,.3137180218e-2,.2007795340e-2,0.,0.,.1673162783e-2,.1380359296e-2,0.,0.,.7529232524e-3,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,.2550890264e-2,0.,0.,0.,-.9820927513e-3,-.1326462937e-2,0.,0.,.6547285010e-3,.1785623185e-3,0.,0.,.5356869553e-3,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2328633565e-3,0.,0.,0.,0.,.2235488223e-3,0.,0.,0.,-.1956052195e-3,0.,0.,.1304034797e-3,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,.1484784677e-1,0.,0.,0.,.7423923386e-3,.1187827742e-2,0.,0.,0.,.1484784677e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,-.3018254064e-2,0.,0.,0.,.6640158940e-3,.8451111378e-3,0.,0.,0.,.2414603251e-3,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.2755276392e-3,0.,0.,0.,0.,-.1983799002e-3,0.,0.,0.,.9918995009e-4,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,.3422378419e-2,0.,0.,0.,0.,.1368951368e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.3124189770e-3,0.,0.,0.,0.,.1249675908e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,.3453921745e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[21][21] = 
		{
			{1.414213562,-2.,2.449489743,-2.828427124,3.162277660,-3.464101616,-3.464101616,4.242640686,-4.898979486,5.477225575,-6.,5.477225575,-6.324555320,7.071067810,-7.745966692,-7.483314774,8.366600265,-9.165151390,9.486832980,-10.39230485,-11.48912529},
			{0.,6.,-19.59591794,42.42640686,-75.89466384,121.2435566,3.464101616,-25.45584412,63.68673332,-120.4989626,198.,-10.95445115,56.92099788,-127.2792206,224.6330341,22.44994432,-100.3992032,210.7984820,-37.94733192,155.8845727,57.44562647},
			{0.,0.,24.49489743,-127.2792206,398.4469852,-969.9484525,0.,21.21320343,-161.6663230,575.1086854,-1488.,5.477225575,-94.86832980,487.9036789,-1471.733671,-22.44994432,250.9980080,-1081.487864,56.92099788,-519.6152424,-114.8912529},
			{0.,0.,0.,98.99494934,-708.3501958,2909.845357,0.,0.,102.8785692,-920.1738966,4176.,0.,44.27188724,-622.2539673,3578.636612,7.483314774,-234.2648074,2071.324214,-37.94733192,727.4613394,114.8912529},
			{0.,0.,0.,0.,398.4469852,-3637.306697,0.,0.,0.,460.0869483,-4860.,0.,0.,254.5584412,-3601.874512,0.,75.29940238,-1695.553007,9.486832980,-467.6537182,-57.44562647},
			{0.,0.,0.,0.,0.,1600.414947,0.,0.,0.,0.,1980.,0.,0.,0.,1278.084504,0.,0.,504.0833264,0.,114.3153533,11.48912529},
			{0.,0.,0.,0.,0.,0.,6.928203232,-8.485281372,9.797958972,-10.95445115,12.,-32.86335345,37.94733192,-42.42640686,46.47580015,89.79977729,-100.3992032,109.9818167,-189.7366596,207.8460970,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,42.42640686,-117.5755077,230.0434742,-384.,32.86335345,-303.5786554,721.2489166,-1301.322404,-179.5995546,1104.391235,-2419.599967,569.2099788,-2909.845357,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-920.1738966,2592.,0.,265.6313234,-2206.173157,7529.079625,89.79977729,-1907.584860,10558.25440,-569.2099788,7482.459491,2068.042553},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,-5760.,0.,0.,1527.350647,-13942.74005,0.,903.5928286,-14297.63617,189.7366596,-7066.767297,-1378.695035},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3960.,0.,0.,0.,7668.507025,0.,0.,6048.999917,0.,2286.307067,344.6737588},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-37.94733192,42.42640686,-46.47580015,-224.4994432,250.9980080,-274.9545417,853.8149682,-935.3074363,-2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-678.8225098,1254.846604,224.4994432,-2509.980080,5774.045376,-1707.629936,12158.99667,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,-6274.233021,0.,2258.982072,-20621.59063,853.8149682,-21512.07104,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7668.507025,0.,0.,15122.49979,0.,10288.38180,2412.716312},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-167.3320053,183.3030278,-1328.156617,1454.922679,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,-3666.060556,1328.156617,-17459.07214,-12867.82033},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10081.66653,0.,16004.14947,6433.910165},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086,-727.4613394,-7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,8002.074733,7238.148935},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2895.259574}
		};
		for (int j = 0; j < 21; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][21] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 21; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
