
/*
 * Copyright (C) 2013 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: Apr 19, 2013
 *      Author: Martin Schreiber <martin.schreiber@in.tum.de>
 */

#ifndef CCOMPUTATION3D_MATRICES_CONST_HPP_
#define CCOMPUTATION3D_MATRICES_CONST_HPP_

#include <string.h>
#include <assert.h>

#include "../../subsimulation_generic/CConfig.hpp"
#include "../../subsimulation_generic/types/CTypes.hpp"
#include "libmath/CMatrixOperations.hpp"

#define SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS (4)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS (35)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS (35)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS (100)
#define SIMULATION_HYPERBOLIC_NUMBER_OF_TOP_BOTTOM_FACE_DOFS (27)
#define SIMULATION_HYPERBOLIC_CFL	(0.5/(T)(2*4+1))
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_DEGREE	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS)
#define SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER	(SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS*2)

class CDG_MatrixComputations_3D
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;



public:
        static inline void mul_edge_comm_hyp_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,0>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_right_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,1>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }


public:
        static inline void mul_edge_comm_left_project_to_edge_space(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_EDGE_FACE_DOFS; i++)
                {
                        /*
                         * rotate edge comm data to normal space
                         */
                        CTriangle_VectorProjections::toEdgeSpace<T,2>(
                                        &(io_hu[i]),
                                        &(io_hv[i])
                        );
                }
        }

public:
        static inline void mul_adaptivity_project_momentum_reference_to_left_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_left_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                        );
                }
        }



public:
        static inline void mul_adaptivity_project_momentum_reference_to_right_child(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i], &io_hv[i],
                                        -(T)M_SQRT1_2, (T)M_SQRT1_2
                                );
                }
        }

        static inline void mul_adaptivity_project_momentum_right_child_to_reference(
                        T io_hu[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS],
                        T io_hv[SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS]
        ) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
                for (int i = 0; i < SIMULATION_HYPERBOLIC_NUMBER_OF_CELL_DOFS; i++)
                {
                        CTriangle_VectorProjections::changeFromReferenceElementToBasisWithXAxis(
                                        &io_hu[i],	&io_hv[i],
                                        -(T)M_SQRT1_2, -(T)M_SQRT1_2
                        );
                }
        }



#if CONFIG_SPHERICAL_DISTORTED_GRID_MODE==1

public:
	static inline void mul_edge_comm_transform_to_edge_space(
		const T i_rotationMatrix[2][2],
		const T i_inverseTransformationMatrix[2][2],
		T io_hu[35],
		T io_hv[35]
	) {
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			/*
			 * scale and shear from reference space
			 */
			CTriangle_VectorProjections::matrixTransformation<T>(
				i_inverseTransformationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

			/*
			 * rotate edge comm data to edge space
			 */
			CTriangle_VectorProjections::fromAnyEdgeToEdgeSpace<T>(
				i_rotationMatrix,
				&(io_hu[i]),
				&(io_hv[i])
			);

		}
	}
#endif


	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void mul_timestep_inv_mass(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_inv_mass
	 *********************************************************/
public:
	static inline void madd_timestep_inv_mass(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_u(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,7.838367178,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,7.838367178,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,0.,0.,0.,0.,7.668115805,0.,0.,0.,10.73312629,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780460,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,9.142857143,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,0.,-2.993325910,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.291502622,0.,0.,0.,0.,-2.993325910,0.,0.,0.,.6110100925,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260101,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,0.,0.,0.,0.,3.346640106,0.,0.,0.,-5.074680381,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357806,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,-3.833259389,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_u
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_u(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,7.838367178,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,7.838367178,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,0.,0.,0.,0.,7.668115805,0.,0.,0.,10.73312629,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780460,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,9.142857143,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,0.,-2.993325910,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.291502622,0.,0.,0.,0.,-2.993325910,0.,0.,0.,.6110100925,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260101,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,0.,0.,0.,0.,3.346640106,0.,0.,0.,-5.074680381,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357806,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,-3.833259389,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_v(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,0.,0.,0.,0.,8.854377448,0.,0.,0.,-3.098386677,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,0.,3.919183589,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.732050808,0.,0.,0.,0.,3.919183589,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,0.,0.,0.,0.,3.834057902,0.,0.,0.,5.366563145,0.,0.,4.647580015,0.,0.,6.640783086,0.,0.,0.,-6.196773354,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.581988897,0.,0.,0.,0.,-.9128709292,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416488,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,0.,2.529822128,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.236067977,0.,0.,0.,0.,2.529822128,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,1.649572198,0.,0.,-2.571428571,0.,0.,1.224744872,0.,0.,0.,4.571428571,0.,0.,7.423074892,0.,0.,-7.377111136,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.},
			{2.645751311,0.,0.,0.,0.,-1.496662955,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.645751311,0.,0.,0.,0.,-1.496662955,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,5.184592558,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,6.831300507,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,0.,0.,0.,0.,1.673320053,0.,0.,0.,-2.537340190,0.,0.,.7606388294,0.,0.,2.898275350,0.,0.,0.,5.747048932,0.,0.,-2.195775165,0.,0.,2.182178903,0.,0.,8.504200642,0.,0.,10.06230590,0.,0.,0.},
			{2.683281572,0.,0.,0.,0.,-1.897366596,0.,0.,0.,.6639400021,0.,0.,-.9583148473e-1,0.,0.,5.477225576,0.,0.,0.,-1.916629694,0.,0.,.2766416676,0.,0.,7.423074892,0.,0.,-1.071428571,0.,0.,8.874119673,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_v
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_v(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,0.,0.,0.,0.,8.854377448,0.,0.,0.,-3.098386677,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,0.,3.919183589,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.732050808,0.,0.,0.,0.,3.919183589,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,0.,0.,0.,0.,3.834057902,0.,0.,0.,5.366563145,0.,0.,4.647580015,0.,0.,6.640783086,0.,0.,0.,-6.196773354,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.581988897,0.,0.,0.,0.,-.9128709292,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416488,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,0.,2.529822128,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.236067977,0.,0.,0.,0.,2.529822128,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,1.649572198,0.,0.,-2.571428571,0.,0.,1.224744872,0.,0.,0.,4.571428571,0.,0.,7.423074892,0.,0.,-7.377111136,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.},
			{2.645751311,0.,0.,0.,0.,-1.496662955,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.645751311,0.,0.,0.,0.,-1.496662955,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,5.184592558,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,6.831300507,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,0.,0.,0.,0.,1.673320053,0.,0.,0.,-2.537340190,0.,0.,.7606388294,0.,0.,2.898275350,0.,0.,0.,5.747048932,0.,0.,-2.195775165,0.,0.,2.182178903,0.,0.,8.504200642,0.,0.,10.06230590,0.,0.,0.},
			{2.683281572,0.,0.,0.,0.,-1.897366596,0.,0.,0.,.6639400021,0.,0.,-.9583148473e-1,0.,0.,5.477225576,0.,0.,0.,-1.916629694,0.,0.,.2766416676,0.,0.,7.423074892,0.,0.,-1.071428571,0.,0.,8.874119673,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_hyp_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.92171958555151,3.24166342459793,-3.96510583993544,4.31411318985288,2.19799643560933,-3.70770257117539,4.53515100205683,-2.29634628619806,3.87360456901059,2.23836106649518,3.41890639763464,-5.76720136441661,7.05426838770182,-7.67518283761778,-3.91413586924262,6.60258197030424,-8.07608098333328,4.09536544265748,-6.90829008438974,-4.00097073404889,5.33523806427656,-8.99977614812198,11.0082572774727,-5.59797201662203,9.44297038492047,5.49216459339238,7.19422103343416,-12.1356119587979,-7.09892453011936,9.00134910987289},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.59519010006696,2.69085533682779,-3.29137384514269,3.5810795200105,1.2388862761662,-2.0898222386161,2.55620812016103,-.514957807193037,.868659455579414,-.353775594121333,3.23038451124486,-5.44919216674985,6.66528903908045,-7.25196565099481,-2.62173967966463,4.42249623366662,-5.4094652504454,1.30644820527577,-2.20378945631257,.307822029727901,4.7630793830775,-8.03462707886587,9.82771577392637,-2.90243304119459,4.89598539807522,.518849452524734,6.06854814809742,-10.2367643607743,-2.66498561200528,7.17423567654379},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.03822870489707,1.75134189417476,-2.14218907478874,2.33074387312075,-.0622566343920988,.105017951653011,-.12845482061473,1.10967175893133,-1.87185600940662,-1.41978880536247,2.90882269973526,-4.90676382786187,6.00180690245521,-6.53008402866105,-.707254507004454,1.19303621855815,-1.45928625505293,-1.57948219926848,2.66435837496252,2.54719860772099,3.86201434387853,-6.51466048131366,7.96853804732135,.544351948717181,-.918243127233759,-3.18953813043078,4.43071582310306,-7.47397774950178,2.19278566166382,4.71658607960972},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-.300186182387741,.506370739691845,-.619377558413115,.673894973328731,-1.13611369814234,1.91645974101964,-2.34415629302528,1.22655366483195,-2.06901890572263,.100858250439743,2.48271365058504,-4.18797939686675,5.12261126340126,-5.57350187032702,1.26648273457434,-2.13637348021345,2.61314820693614,-2.5074905585056,4.22977446202813,.461151554581851,2.81340752869054,-4.74581221430261,5.80493569904152,3.19379348794833,-5.38746839617942,-2.33714259619681,2.75487590413567,-4.64707781591112,4.77320941254594,2.50302559510188},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,.553376983055106,-.93346705704166,1.14178901214037,-1.24228891639978,-1.45368745167752,2.45216080196962,-2.99940982448642,-.194424990185747,.327966881493408,1.49978166741586,1.9899087271236,-3.35668865754018,4.10580126963162,-4.46719258573777,2.74863027720701,-4.6365423474252,5.67128006117502,-.854309601049289,1.44109692494208,-2.52312426315122,1.80736427869175,-3.04876253512471,3.72915523810848,4.12997714955087,-6.96667503835439,1.65834138994758,1.41847476579741,-2.39276208655032,4.48820932686104,1.03297875354961},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,1.44662301694489,-2.44024412222762,2.98483369571688,-3.24755780451478,-.724355119014703,1.22188248065577,-1.49457014152702,-1.6918152058931,2.85384793698481,.160475142667942,1.4741928888764,-2.48675051356285,3.04171892525937,-3.30945005334757,3.38025060487184,-5.70199462778367,6.97450945518368,2.16048862424162,-3.64443230995284,-1.70770592621981,.991945941760336,-1.67326955598402,2.04669332477546,3.46032388717692,-5.8370666898335,4.72311662432345,.576746649127291,-.972888308521253,2.68886700736676,.311154445333738},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,2.30018618238774,-3.88008191896113,4.74600026627036,-5.1637416942433,.987081447023724,-1.66506385533551,2.03665635715207,-1.35561938023839,2.28673412920868,-2.20870753342451,.981387965414959,-1.65545977423627,2.02490893148973,-2.20314076875832,3.10522366683209,-5.23806390000324,6.40704325104372,4.42237267757001,-7.45990406612127,2.83495589139377,.43960319187862,-.741547101224612,.907038257317998,2.03901059601618,-3.43951642050873,4.55407145300732,.170154848292172,-.287026656838237,1.03685515398485,.0611113213546687},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.03822870489707,-5.12505307344404,6.26881178264598,-6.82059059403531,3.26615690329693,-5.50953502563763,6.7390986231311,1.66498726415444,-2.80859307349989,-.923099517802526,.555278916264744,-.936675343241156,1.14571329243577,-1.24655861042429,2.17523616991868,-3.66930929238228,4.48818948881768,4.50024407433993,-7.59126186695628,6.26847807110811,.140734880491405,-.237399419743997,.29037987690505,.792696592497175,-1.33716467766395,2.35258131414952,.0308216365902796,-.0519916499446463,.225963663364161,.00626330408481313},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.59519010006696,-6.06456651609707,7.41799655299993,-8.07092624092507,5.47681396366563,-9.23859424244377,11.300372436005,6.3309248170395,-10.6793559162843,5.76936797267538,.233717104755142,-.394247004353173,.482231155810537,-.524676988090528,1.04841335581819,-1.76852192969244,2.1632031816091,2.66785366107377,-4.50028385780377,5.04596922575242,.0249321870529789,-.0420569990798228,.0514428646412205,.15913874220381,-.268444076758866,.559597243722218,.00229823078236368,-.0038767834400574,.0189957623925199,.000196571591117826},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.92171958555151,-6.61537460386721,8.09172854779269,-8.80395991076744,6.96914454009643,-11.7559404153652,14.3795150585562,10.2353232484353,-17.2655122351262,13.4470270492568,.0451952183653594,-.0762378066864121,.0932518071891671,-.101459801467564,.217799593005347,-.367396460919218,.449388373324038,.614368644454251,-1.03635118137081,1.33333501160483,.000932318622845177,-.0015726868798065,.00192366360061394,.00636097475878139,-.0107300458251746,.0244039729141499,1.66187980931003e-005,-2.80335124459201e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.92171958555151,2.87937722321757,-2.67495037465683,1.63105841067898,2.19799643560933,-3.29333213710544,3.05951577632408,-2.29634628619806,3.44069303528758,2.23836106649518,3.41890639763464,-5.1226626812653,4.75896952776425,-2.90179486950883,-3.91413586924262,5.86468169257239,-5.4483074914535,4.09536544265748,-6.13622406614863,-4.00097073404889,5.33523806427656,-7.99396700420041,7.42642015557747,-5.59797201662203,8.38763014060611,5.49216459339238,7.19422103343416,-10.7793438389747,-7.09892453011936,9.00134910987289},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.59519010006696,2.39012709001286,-2.22043548283783,1.35391669466664,1.2388862761662,-1.85626506095126,1.72447600259692,-.514957807193037,.771578638346196,-.353775594121333,3.23038451124486,-4.84019398522881,4.49655523257725,-2.74178702515998,-2.62173967966463,3.92824092358218,-3.64934800797338,1.30644820527577,-1.95749537633152,.307822029727901,4.7630793830775,-7.13668236880667,6.6300000687783,-2.90243304119459,4.3488132461414,.518849452524734,6.06854814809742,-9.09271021667368,-2.66498561200528,7.17423567654379},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.03822870489707,1.55561306022354,-1.44516935978819,.881196025717106,-.0622566343920988,.0932812135041899,-.0866585368463816,1.10967175893133,-1.66265859611778,-1.41978880536247,2.90882269973526,-4.35838708560735,4.04895512766487,-2.46886161968117,-.707254507004454,1.0597032647262,-.984467621360889,-1.57948219926848,2.36659141141637,2.54719860772099,3.86201434387853,-5.78658625215441,5.37575658648661,.544351948717181,-.815620871022689,-3.18953813043078,4.43071582310306,-6.63869084478427,2.19278566166382,4.71658607960972},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-.300186182387741,.449779074319961,-.417846155642134,.25478285241737,-1.13611369814234,1.70227744362135,-1.58142102819233,1.22655366483195,-1.83778669504629,.100858250439743,2.48271365058504,-3.7199335363258,3.45582979910559,-2.10720180544033,1.26648273457434,-1.89761376598671,1.76288907720341,-2.5074905585056,3.75705761211831,.461151554581851,2.81340752869054,-4.21542331996591,3.91614134649752,3.19379348794833,-4.78536842321548,-2.33714259619681,2.75487590413567,-4.12772341388704,4.77320941254594,2.50302559510188},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,.553376983055106,-.829143384311425,.770276776736346,-.469678401195645,-1.45368745167752,2.17810890152299,-2.02346992933989,-.194424990185747,.291313515578485,1.49978166741586,1.9899087271236,-2.9815473107461,2.76986670414993,-1.68893390563516,2.74863027720701,-4.11836538252447,3.82597422031764,-.854309601049289,1.2800408674882,-2.52312426315122,1.80736427869175,-2.70803481201197,2.51577274454146,4.12997714955087,-6.18808395553405,1.65834138994758,1.41847476579741,-2.12534854800543,4.48820932686104,1.03297875354961},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,1.44662301694489,-2.16752402217829,2.01363654211465,-1.22782046694529,-.724355119014703,1.08532568728512,-1.00827093182807,-1.6918152058931,2.53490374291394,.160475142667942,1.4741928888764,-2.20883288938782,2.05201260878721,-1.25122037987572,3.38025060487184,-5.06474340721712,4.70516234202085,2.16048862424162,-3.23713291923203,-1.70770592621981,.991945941760336,-1.48626603590141,1.38074575986725,3.46032388717692,-5.18471990323708,4.72311662432345,.576746649127291,-.864158942299266,2.68886700736676,.311154445333738},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,2.30018618238774,-3.44644648080968,3.20175947449313,-1.95228172055831,.987081447023724,-1.47897739903412,1.37397459375255,-1.35561938023839,2.03117020639316,-2.20870753342451,.981387965414959,-1.47044666380812,1.36604951383154,-.832952480070547,3.10522366683209,-4.65266127660931,4.32233676392885,4.42237267757001,-6.6261900271284,2.83495589139377,.43960319187862,-.658672278252944,.611908590636855,2.03901059601618,-3.05511829665785,4.55407145300732,.170154848292172,-.254948743871771,1.03685515398485,.0611113213546687},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.03822870489707,-4.55228046671326,4.22908267863918,-2.57869489385804,3.26615690329693,-4.89379296535329,4.54634885284358,1.66498726415444,-2.49470653361407,-.923099517802526,.555278916264744,-.831993114526565,.772924185272261,-.471292665829703,2.17523616991868,-3.25922966624351,3.02783447385586,4.50024407433993,-6.74286736535737,6.26847807110811,.140734880491405,-.210867814600684,.195896854176436,.792696592497175,-1.18772402074228,2.35258131414952,.0308216365902796,-.0461810968751895,.225963663364161,.00626330408481313},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.59519010006696,-5.38679449650258,5.00434880168882,-3.05141556280758,5.47681396366563,-8.20609494322841,7.62348767011616,6.3309248170395,-9.48583802706066,5.76936797267538,.233717104755142,-.350186214905109,.325324080359881,-.198367260350894,1.04841335581819,-1.57087306611691,1.45934595309525,2.66785366107377,-3.99733505330865,5.04596922575242,.0249321870529789,-.0373567361461482,.0347045238135923,.15913874220381,-.23844294088706,.559597243722218,.00229823078236368,-.00344351663776842,.0189957623925199,.000196571591117826},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.92171958555151,-5.87604462970729,5.45886369350783,-3.32855727881992,6.96914454009643,-10.4421041409331,9.70074715430413,10.2353232484353,-15.3359297883224,13.4470270492568,.0451952183653594,-.067717518868618,.0629097851728847,-.0383594160020481,.217799593005347,-.326336470786676,.303167594035556,.614368644454251,-.920529245649199,1.33333501160483,.000932318622845177,-.00139692441436285,.00129774711619293,.00636097475878139,-.00953086286461726,.0244039729141499,1.66187980931003e-005,-2.49005052810442e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.92171958555151,2.26142559207414,-.826744191593017,-1.19523987507357,2.19799643560933,-2.5865404220042,.945601429143385,-2.29634628619806,2.70227576356801,2.23836106649518,3.41890639763464,-4.02327294720743,1.47084987158014,2.12643576380386,-3.91413586924262,4.60604507183384,-1.68390285488932,4.09536544265748,-4.81931093640565,-4.00097073404889,5.33523806427656,-6.27835818791885,2.29527612404422,-5.59797201662203,6.58753611853026,5.49216459339238,7.19422103343416,-8.46595709192435,-7.09892453011936,9.00134910987289},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.59519010006696,1.87717487173315,-.686267736267336,-.992150379408995,1.2388862761662,-1.4578865459724,.532982044151549,-.514957807193037,.605987872997064,-.353775594121333,3.23038451124486,-3.80142569043747,1.38974575226927,2.00918198880834,-2.62173967966463,3.08518956279543,-1.12790027705863,1.30644820527577,-1.53739152505215,.307822029727901,4.7630793830775,-5.60505793446857,2.0491273778591,-2.90243304119459,3.41550161982096,.518849452524734,6.06854814809742,-7.14129688166466,-2.66498561200528,7.17423567654379},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.03822870489707,1.22175835711557,-.446657024142499,-.64574059444934,-.0622566343920988,.0732618573826121,-.0267834658423847,1.10967175893133,-1.30583053509787,-1.41978880536247,2.90882269973526,-3.42302078938587,1.25140644309977,1.80918220620515,-.707254507004454,.832277224231996,-.304268406428259,-1.57948219926848,1.8586902544778,2.54719860772099,3.86201434387853,-4.54470992402429,1.66147962048301,.544351948717181,-.64057807233031,-3.18953813043078,4.43071582310306,-5.21394183895337,2.19278566166382,4.71658607960972},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-.300186182387741,.353250661721207,-.129143286331405,-.186704916697285,-1.13611369814234,1.33694666498547,-.488768188714079,1.22655366483195,-1.44337387662839,.100858250439743,2.48271365058504,-2.92158763778148,1.06808980107197,1.54415783407833,1.26648273457434,-1.49036128356923,.544854334050093,-2.5074905585056,2.95074440626681,.461151554581851,2.81340752869054,-3.31073889879025,1.21035782283977,3.19379348794833,-3.75836640385388,-2.33714259619681,2.75487590413567,-3.24186052799879,4.77320941254594,2.50302559510188},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,.553376983055106,-.651198479192506,.238068659934471,.344180410642771,-1.45368745167752,1.71065853147384,-.625391792982139,-.194424990185747,.228793863525153,1.49978166741586,1.9899087271236,-2.34166865603201,.856079885009688,1.23765104741929,2.74863027720701,-3.23451085132423,1.18248996086068,-.854309601049289,1.00532752474992,-2.52312426315122,1.80736427869175,-2.12685548066823,.77754732335343,4.12997714955087,-4.86004102210358,1.65834138994758,1.41847476579741,-1.66922123290476,4.48820932686104,1.03297875354961},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,1.44662301694489,-1.70234530427801,.622352597994011,.899747042727645,-.724355119014703,.852400743178548,-.311625271383912,-1.6918152058931,1.9908805805112,.160475142667942,1.4741928888764,-1.73478875376213,.634213449888056,.916894502821395,3.38025060487184,-3.97778389924568,1.45421973940992,2.16048862424162,-2.54240228389271,-1.70770592621981,.991945941760336,-1.16729410204273,.426745687640423,3.46032388717692,-4.07201188589473,4.72311662432345,.576746649127291,-.67869924509284,2.68886700736676,.311154445333738},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,2.30018618238774,-2.70679444519172,.989564544259887,1.4306323700677,.987081447023724,-1.16156970101249,.42465293021633,-1.35561938023839,1.5952547827211,-2.20870753342451,.981387965414959,-1.15486977201266,.422203533825773,.610387716162354,3.10522366683209,-3.65413992905702,1.33590022736237,4.42237267757001,-5.20412385861342,2.83495589139377,.43960319187862,-.517312679568431,.189121965733585,2.03901059601618,-2.39945035585008,4.55407145300732,.170154848292172,-.200233442638166,1.03685515398485,.0611113213546687},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.03822870489707,-3.57530214058608,1.30707828207098,1.88966804781976,3.26615690329693,-3.843521635486,1.40513541583971,1.66498726415444,-1.95931021275678,-.923099517802526,.555278916264744,-.653436620408273,.238886891797977,.345363344035539,2.17523616991868,-2.55975678291091,.935809721212703,4.50024407433993,-5.29576072944733,6.26847807110811,.140734880491405,-.165612851500527,.0605456414728778,.792696592497175,-.932823068533675,2.35258131414952,.0308216365902796,-.036270035583218,.225963663364161,.00626330408481313},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.59519010006696,-4.23071865520366,1.54668899419582,2.23607783277941,5.47681396366563,-6.44496072482988,2.35618358032458,6.3309248170395,-7.45005437415178,5.76936797267538,.233717104755142,-.275031719356679,.100547582628473,.145363561432352,1.04841335581819,-1.23374336817428,.451038569416845,2.66785366107377,-3.13945519772704,5.04596922575242,.0249321870529789,-.0293394969148429,.0107260918769642,.15913874220381,-.187269998673138,.559597243722218,.00229823078236368,-.00270449338439953,.0189957623925199,.000196571591117826},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.92171958555151,-4.61496937554466,1.6871654495215,2.43916732844399,6.96914454009643,-8.20109339929485,2.99820005631406,10.2353232484353,-12.0446407028288,13.4470270492568,.0451952183653594,-.0531844625867158,.0194434633176024,.028109786436831,.217799593005347,-.256300439106579,.0936996999356596,.614368644454251,-.722971751522571,1.33333501160483,.000932318622845177,-.00109712634918437,.000401093381258235,.00636097475878139,-.00748541629863617,.0244039729141499,1.66187980931003e-005,-1.95565344631038e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.92171958555151,1.44256348674145,.937852501329485,-2.27058656100761,2.19799643560933,-1.64995425135427,-1.07268327325537,-2.29634628619806,1.72378183093533,2.23836106649518,3.41890639763464,-2.56644599370298,-1.66852122478612,4.03957110939073,-3.91413586924262,2.93819635817362,1.91020695459719,4.09536544265748,-3.07423866276144,-4.00097073404889,5.33523806427656,-4.00496497002541,-2.60374427309723,-5.59797201662203,4.20218958584716,5.49216459339238,7.19422103343416,-5.40043440913241,-7.09892453011936,9.00134910987289},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.59519010006696,1.19744993497978,.77849704852463,-1.88477925223671,1.2388862761662,-.929985893147364,-.604610892164935,-.514957807193037,.386559691398865,-.353775594121333,3.23038451124486,-2.42492961864658,-1.57651731119676,3.81682515580894,-2.62173967966463,1.9680425606586,1.27948173914766,1.30644820527577,-.980702122872796,.307822029727901,4.7630793830775,-3.57546670811254,-2.32451495445634,-2.90243304119459,2.17874863666902,.518849452524734,6.06854814809742,-4.55543360151919,-2.66498561200528,7.17423567654379},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.03822870489707,.779359710871422,.506684427405858,-1.22670766448742,-.0622566343920988,.0467337421144522,.030382965722629,1.10967175893133,-.832989356981887,-1.41978880536247,2.90882269973526,-2.18354511527215,-1.41958622119803,3.4368873474629,-.707254507004454,.530909678713955,.345159831718241,-1.57948219926848,1.18565859680248,2.54719860772099,3.86201434387853,-2.89907066463143,-1.88477020319366,.544351948717181,-.408624781106939,-3.18953813043078,4.43071582310306,-3.32597373321893,2.19278566166382,4.71658607960972},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-.300186182387741,.225338612975934,.146499189649512,-.354681669820299,-1.13611369814234,.852838337634571,.554455021295674,1.22655366483195,-.920728260721081,.100858250439743,2.48271365058504,-1.86368088534508,-1.21163317718593,2.93342290468435,1.26648273457434,-.950701529718779,-.618078730089993,-2.5074905585056,1.88227998927126,.461151554581851,2.81340752869054,-2.11192049221879,-1.37302096972311,3.19379348794833,-2.39746209780118,-2.33714259619681,2.75487590413567,-2.06798297639769,4.77320941254594,2.50302559510188},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,.553376983055106,-.41539953912128,-.270063328509738,.65383646518608,-1.45368745167752,1.09122915405036,.709439828333213,-.194424990185747,.145947615817496,1.49978166741586,1.9899087271236,-1.49375054084377,-.971130695151503,2.35115468793597,2.74863027720701,-2.06329461731692,-1.34140787303666,-.854309601049289,.641298473185774,-2.52312426315122,1.80736427869175,-1.35672120663226,-.88204393756181,4.12997714955087,-3.10022038625728,1.65834138994758,1.41847476579741,-1.06479629943225,4.48820932686104,1.03297875354961},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,1.44662301694489,-1.0859261460488,-.705992187998236,1.70924145531706,-.724355119014703,.543746472353678,.353505404956933,-1.6918152058931,1.26998281171566,.160475142667942,1.4741928888764,-1.10662182393172,-.719447050735517,1.74181633275903,3.38025060487184,-2.53742852796533,-1.64965612574837,2.16048862424162,-1.62179853107076,-1.70770592621981,.991945941760336,-.744616960114543,-.484097154422056,3.46032388717692,-2.5975365648804,4.72311662432345,.576746649127291,-.432942278924123,2.68886700736676,.311154445333738},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,2.30018618238774,-1.72666429814602,-1.12255470615749,2.71775959032344,.987081447023724,-.740965364440283,-.481723145865576,-1.35561938023839,1.01761309753837,-2.20870753342451,.981387965414959,-.736691479430405,-.47894456870109,1.15954811601065,3.10522366683209,-2.33097602484761,-1.51543534565912,4.42237267757001,-3.31971084311917,2.83495589139377,.43960319187862,-.329993781528465,-.214538560322786,2.03901059601618,-1.53060948975261,4.55407145300732,.170154848292172,-.12772892205108,1.03685515398485,.0611113213546687},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.03822870489707,-2.2806853960415,-1.48273994391383,3.58978558499055,3.26615690329693,-2.45178262388351,-1.59397523182193,1.66498726415444,-1.24984407270733,-.923099517802526,.555278916264744,-.41682724950333,-.270991524688991,.656083673232097,2.17523616991868,-1.63286896677384,-1.06157563213922,4.50024407433993,-3.3781660071394,6.26847807110811,.140734880491405,-.10564444540962,-.0686825282564419,.792696592497175,-.595047877309364,2.35258131414952,.0308216365902796,-.0231366573273696,.225963663364161,.00626330408481313},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.59519010006696,-2.69877562014986,-1.7545525650326,4.24785717273984,5.47681396366563,-4.11124073580283,-2.67283724139748,6.3309248170395,-4.75239001984257,5.76936797267538,.233717104755142,-.175442746128902,-.114060434690259,.276145864886064,1.04841335581819,-.787004949964004,-.511654819985594,2.66785366107377,-2.00265861161846,5.04596922575242,.0249321870529789,-.0187156663995726,-.0121675993597458,.15913874220381,-.119459540563619,.559597243722218,.00229823078236368,-.00172519645136334,.0189957623925199,.000196571591117826},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.92171958555151,-2.94388917191153,-1.91390801783746,4.63366448151074,6.96914454009643,-5.23147784770941,-3.40113598727833,10.2353232484353,-7.68327684524188,13.4470270492568,.0451952183653594,-.0339263710725032,-.0220565211009044,.0533999113042758,.217799593005347,-.163494061616168,-.10629224716917,.614368644454251,-.461183712763862,1.33333501160483,.000932318622845177,-.000699856947415067,-.000454997367631841,.00636097475878139,-.00477494739305922,.0244039729141499,1.66187980931003e-005,-1.24751142122696e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.92171958555151,.49553061346938,2.00568935114481,-1.09346229146625,2.19799643560933,-.566770786786546,-2.29403815127598,-2.29634628619806,.592131075007482,2.23836106649518,3.41890639763464,-.881592088933561,-3.56829591856781,1.94536458449533,-3.91413586924262,1.00929093051433,4.08516450282659,4.09536544265748,-1.05602241045943,-4.00097073404889,5.33523806427656,-1.37573338488067,-5.56836192711791,-5.59797201662203,1.44348141522225,5.49216459339238,7.19422103343416,-1.855086864706,-7.09892453011936,9.00134910987289},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.59519010006696,.411332399809833,1.66489212099991,-.907666360512686,1.2388862761662,-.31945663701093,-1.29301955934683,-.514957807193037,.132785948612982,-.353775594121333,3.23038451124486,-.832980227624035,-3.37153654597113,1.8380952537413,-2.62173967966463,.676036338352017,2.73629690762123,1.30644820527577,-.336878015452627,.307822029727901,4.7630793830775,-1.22819773798136,-4.97120270417874,-2.90243304119459,.748415344244522,.518849452524734,6.06854814809742,-1.56482319698726,-2.66498561200528,7.17423567654379},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.03822870489707,.26771549341915,1.08359423150038,-.590754211622909,-.0622566343920988,.0160533661863334,.0649769454361842,1.10967175893133,-.286137650697852,-1.41978880536247,2.90882269973526,-.750062968079831,-3.03592405293217,1.65512593926409,-.707254507004454,.182371175485818,.738158077031132,-1.57948219926848,.407281992951262,2.54719860772099,3.86201434387853,-.995850982046505,-4.03076552074204,.544351948717181,-.140365460738301,-3.18953813043078,4.43071582310306,-1.14249516194806,2.19278566166382,4.71658607960972},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-.300186182387741,.0774053843401605,.313302853289657,-.170806538752116,-1.13611369814234,.292955913914665,1.18575632130325,1.22655366483195,-.316276575739705,.100858250439743,2.48271365058504,-.640187375400919,-2.59119611829218,1.41266903727833,1.26648273457434,-.326572602577083,-1.32182184869504,-2.5074905585056,.646576297249579,.461151554581851,2.81340752869054,-.725459410782454,-2.93633970583683,3.19379348794833,-.823544942743958,-2.33714259619681,2.75487590413567,-.71036656781973,4.77320941254594,2.50302559510188},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,.553376983055106,-.142692637341486,-.577556856071558,.314872611220466,-1.45368745167752,.374844821120283,1.51720649776892,-.194424990185747,.0501340234001082,1.49978166741586,1.9899087271236,-.513113723366546,-2.07685802519487,1.13226204929232,2.74863027720701,-.708756083794028,-2.86873200495546,-.854309601049289,.220290495881434,-2.52312426315122,1.80736427869175,-.466043191865359,-1.88633727617264,4.12997714955087,-1.06494731359279,1.65834138994758,1.41847476579741,-.365764951301697,4.48820932686104,1.03297875354961},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,1.44662301694489,-.373023200905715,-1.50983338153088,.823131392784558,-.724355119014703,.186780703583358,.756005902089247,-1.6918152058931,.436247948578379,.160475142667942,1.4741928888764,-.380132310523241,-1.53860792217029,.83881870492831,3.38025060487184,-.871624387178691,-3.52795105816102,2.16048862424162,-.557099100602201,-1.70770592621981,.991945941760336,-.255781116350547,-1.03528913786275,3.46032388717692,-.892271916858127,4.72311662432345,.576746649127291,-.148718690750111,2.68886700736676,.311154445333738},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,2.30018618238774,-.593121222587361,-2.40069309089209,1.30881054275714,.987081447023724,-.254526767782017,-1.03021208824726,-1.35561938023839,.349557192548049,-2.20870753342451,.981387965414959,-.253058658488868,-1.02426982907298,.558411716942304,3.10522366683209,-.800706513225493,-3.2409067852655,4.42237267757001,-1.14034381554164,2.83495589139377,.43960319187862,-.113355164264732,-.458811705720641,2.03901059601618,-.525775029251871,4.55407145300732,.170154848292172,-.0438757750930783,1.03685515398485,.0611113213546687},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.03822870489707,-.78343133166635,-3.17098446910282,1.72875821562793,3.26615690329693,-.842204422108959,-3.40887201763794,1.66498726415444,-.429330151346565,-.923099517802526,.555278916264744,-.143183065809956,-.579541894432988,.315954814956539,2.17523616991868,-.560901872435633,-2.27028337376948,4.50024407433993,-1.16042357185991,6.26847807110811,.140734880491405,-.0362896034209995,-.14688430786117,.792696592497175,-.204403093795768,2.35258131414952,.0308216365902796,-.00794760307219474,.225963663364161,.00626330408481313},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.59519010006696,-.927048238057034,-3.75228235860234,2.04567036451771,5.47681396366563,-1.4122398512488,-5.7161240011779,6.3309248170395,-1.63247909961861,5.76936797267538,.233717104755142,-.060265806265752,-.243929401394036,.132985500479333,1.04841335581819,-.27034168634063,-1.09422390242844,2.66785366107377,-.68792719315715,5.04596922575242,.0249321870529789,-.0064289618707996,-.0260216019365979,.15913874220381,-.0410351849082053,.559597243722218,.00229823078236368,-.000592617006985495,.0189957623925199,.000196571591117826},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.92171958555151,-1.01124645171658,-4.09307958874724,2.23146629547127,6.96914454009643,-1.7970491081004,-7.2736621396321,10.2353232484353,-2.63925915467191,13.4470270492568,.0451952183653594,-.0116539449562267,-.0471700287973566,.0257161697253048,.217799593005347,-.0561613498441344,-.227316372195247,.614368644454251,-.158419820138511,1.33333501160483,.000932318622845177,-.000240405739976666,-.000973056396143785,.00636097475878139,-.00164022771456071,.0244039729141499,1.66187980931003e-005,-4.28528869302687e-006,.000146461329045194,2.74870868673343e-007},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.92171958555151,-.49553061346938,2.00568935114481,1.09346229146626,2.19799643560933,.566770786786546,-2.29403815127598,-2.29634628619806,-.592131075007482,2.23836106649518,3.41890639763464,.881592088933561,-3.56829591856781,-1.94536458449534,-3.91413586924262,-1.00929093051433,4.08516450282659,4.09536544265748,1.05602241045943,-4.00097073404889,5.33523806427656,1.37573338488067,-5.56836192711792,-5.59797201662203,-1.44348141522225,5.49216459339238,7.19422103343416,1.855086864706,-7.09892453011936,9.00134910987289},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.59519010006696,-.411332399809833,1.66489212099991,.907666360512694,1.2388862761662,.31945663701093,-1.29301955934683,-.514957807193037,-.132785948612982,-.353775594121333,3.23038451124486,.832980227624035,-3.37153654597113,-1.83809525374131,-2.62173967966463,-.676036338352017,2.73629690762123,1.30644820527577,.336878015452627,.307822029727901,4.7630793830775,1.22819773798136,-4.97120270417875,-2.90243304119459,-.748415344244522,.518849452524734,6.06854814809742,1.56482319698726,-2.66498561200528,7.17423567654379},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.03822870489707,-.26771549341915,1.08359423150038,.590754211622914,-.0622566343920988,-.0160533661863334,.0649769454361843,1.10967175893133,.286137650697852,-1.41978880536247,2.90882269973526,.750062968079831,-3.03592405293218,-1.65512593926411,-.707254507004454,-.182371175485818,.738158077031133,-1.57948219926848,-.407281992951262,2.54719860772099,3.86201434387853,.995850982046505,-4.03076552074204,.544351948717181,.140365460738301,-3.18953813043078,4.43071582310306,1.14249516194806,2.19278566166382,4.71658607960972},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-.300186182387741,-.0774053843401605,.313302853289658,.170806538752117,-1.13611369814234,-.292955913914665,1.18575632130325,1.22655366483195,.316276575739705,.100858250439743,2.48271365058504,.640187375400919,-2.59119611829218,-1.41266903727834,1.26648273457434,.326572602577083,-1.32182184869504,-2.5074905585056,-.646576297249579,.461151554581851,2.81340752869054,.725459410782454,-2.93633970583683,3.19379348794833,.823544942743958,-2.33714259619681,2.75487590413567,.71036656781973,4.77320941254594,2.50302559510188},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,.553376983055106,.142692637341486,-.577556856071558,-.314872611220468,-1.45368745167752,-.374844821120283,1.51720649776892,-.194424990185747,-.0501340234001082,1.49978166741586,1.9899087271236,.513113723366546,-2.07685802519487,-1.13226204929233,2.74863027720701,.708756083794028,-2.86873200495547,-.854309601049289,-.220290495881434,-2.52312426315122,1.80736427869175,.466043191865359,-1.88633727617264,4.12997714955087,1.06494731359279,1.65834138994758,1.41847476579741,.365764951301697,4.48820932686104,1.03297875354961},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,1.44662301694489,.373023200905715,-1.50983338153088,-.823131392784565,-.724355119014703,-.186780703583358,.756005902089247,-1.6918152058931,-.436247948578379,.160475142667942,1.4741928888764,.380132310523241,-1.53860792217029,-.838818704928317,3.38025060487184,.871624387178691,-3.52795105816103,2.16048862424162,.557099100602201,-1.70770592621981,.991945941760336,.255781116350547,-1.03528913786275,3.46032388717692,.892271916858127,4.72311662432345,.576746649127291,.148718690750111,2.68886700736676,.311154445333738},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,2.30018618238774,.593121222587361,-2.40069309089209,-1.30881054275715,.987081447023724,.254526767782017,-1.03021208824726,-1.35561938023839,-.349557192548049,-2.20870753342451,.981387965414959,.253058658488868,-1.02426982907298,-.558411716942308,3.10522366683209,.800706513225493,-3.2409067852655,4.42237267757001,1.14034381554164,2.83495589139377,.43960319187862,.113355164264732,-.458811705720641,2.03901059601618,.525775029251871,4.55407145300732,.170154848292172,.0438757750930783,1.03685515398485,.0611113213546687},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.03822870489707,.78343133166635,-3.17098446910282,-1.72875821562795,3.26615690329693,.842204422108959,-3.40887201763794,1.66498726415444,.429330151346565,-.923099517802526,.555278916264744,.143183065809956,-.579541894432988,-.315954814956541,2.17523616991868,.560901872435633,-2.27028337376948,4.50024407433993,1.16042357185991,6.26847807110811,.140734880491405,.0362896034209995,-.14688430786117,.792696592497175,.204403093795768,2.35258131414952,.0308216365902796,.00794760307219474,.225963663364161,.00626330408481313},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.59519010006696,.927048238057034,-3.75228235860234,-2.04567036451773,5.47681396366563,1.4122398512488,-5.7161240011779,6.3309248170395,1.63247909961861,5.76936797267538,.233717104755142,.060265806265752,-.243929401394036,-.132985500479334,1.04841335581819,.27034168634063,-1.09422390242844,2.66785366107377,.68792719315715,5.04596922575242,.0249321870529789,.0064289618707996,-.0260216019365979,.15913874220381,.0410351849082053,.559597243722218,.00229823078236368,.000592617006985495,.0189957623925199,.000196571591117826},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.92171958555151,1.01124645171658,-4.09307958874725,-2.23146629547129,6.96914454009643,1.7970491081004,-7.2736621396321,10.2353232484353,2.63925915467191,13.4470270492568,.0451952183653594,.0116539449562267,-.0471700287973566,-.025716169725305,.217799593005347,.0561613498441344,-.227316372195247,.614368644454251,.158419820138511,1.33333501160483,.000932318622845177,.000240405739976666,-.000973056396143785,.00636097475878139,.00164022771456071,.0244039729141499,1.66187980931003e-005,4.28528869302687e-006,.000146461329045194,2.74870868673343e-007},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.92171958555151,-1.44256348674145,.937852501329483,2.27058656100761,2.19799643560933,1.64995425135427,-1.07268327325536,-2.29634628619806,-1.72378183093533,2.23836106649518,3.41890639763464,2.56644599370298,-1.66852122478611,-4.03957110939074,-3.91413586924262,-2.93819635817362,1.91020695459718,4.09536544265748,3.07423866276144,-4.00097073404889,5.33523806427656,4.00496497002541,-2.60374427309723,-5.59797201662203,-4.20218958584716,5.49216459339238,7.19422103343416,5.40043440913241,-7.09892453011936,9.00134910987289},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.59519010006696,-1.19744993497978,.778497048524628,1.88477925223671,1.2388862761662,.929985893147364,-.604610892164934,-.514957807193037,-.386559691398865,-.353775594121333,3.23038451124486,2.42492961864658,-1.57651731119676,-3.81682515580895,-2.62173967966463,-1.9680425606586,1.27948173914766,1.30644820527577,.980702122872796,.307822029727901,4.7630793830775,3.57546670811254,-2.32451495445633,-2.90243304119459,-2.17874863666902,.518849452524734,6.06854814809742,4.55543360151919,-2.66498561200528,7.17423567654379},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.03822870489707,-.779359710871422,.506684427405857,1.22670766448742,-.0622566343920988,-.0467337421144522,.030382965722629,1.10967175893133,.832989356981887,-1.41978880536247,2.90882269973526,2.18354511527215,-1.41958622119803,-3.43688734746291,-.707254507004454,-.530909678713955,.34515983171824,-1.57948219926848,-1.18565859680248,2.54719860772099,3.86201434387853,2.89907066463143,-1.88477020319366,.544351948717181,.408624781106939,-3.18953813043078,4.43071582310306,3.32597373321893,2.19278566166382,4.71658607960972},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-.300186182387741,-.225338612975934,.146499189649511,.3546816698203,-1.13611369814234,-.852838337634571,.554455021295673,1.22655366483195,.920728260721081,.100858250439743,2.48271365058504,1.86368088534508,-1.21163317718593,-2.93342290468436,1.26648273457434,.950701529718779,-.618078730089992,-2.5074905585056,-1.88227998927126,.461151554581851,2.81340752869054,2.11192049221879,-1.3730209697231,3.19379348794833,2.39746209780118,-2.33714259619681,2.75487590413567,2.06798297639769,4.77320941254594,2.50302559510188},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,.553376983055106,.41539953912128,-.270063328509737,-.653836465186081,-1.45368745167752,-1.09122915405036,.709439828333211,-.194424990185747,-.145947615817496,1.49978166741586,1.9899087271236,1.49375054084377,-.971130695151501,-2.35115468793598,2.74863027720701,2.06329461731692,-1.34140787303665,-.854309601049289,-.641298473185774,-2.52312426315122,1.80736427869175,1.35672120663226,-.882043937561808,4.12997714955087,3.10022038625728,1.65834138994758,1.41847476579741,1.06479629943225,4.48820932686104,1.03297875354961},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,1.44662301694489,1.0859261460488,-.705992187998234,-1.70924145531706,-.724355119014703,-.543746472353678,.353505404956932,-1.6918152058931,-1.26998281171566,.160475142667942,1.4741928888764,1.10662182393172,-.719447050735515,-1.74181633275904,3.38025060487184,2.53742852796533,-1.64965612574837,2.16048862424162,1.62179853107076,-1.70770592621981,.991945941760336,.744616960114543,-.484097154422054,3.46032388717692,2.5975365648804,4.72311662432345,.576746649127291,.432942278924123,2.68886700736676,.311154445333738},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,2.30018618238774,1.72666429814602,-1.12255470615748,-2.71775959032344,.987081447023724,.740965364440283,-.481723145865575,-1.35561938023839,-1.01761309753837,-2.20870753342451,.981387965414959,.736691479430405,-.478944568701089,-1.15954811601066,3.10522366683209,2.33097602484761,-1.51543534565911,4.42237267757001,3.31971084311917,2.83495589139377,.43960319187862,.329993781528465,-.214538560322786,2.03901059601618,1.53060948975261,4.55407145300732,.170154848292172,.12772892205108,1.03685515398485,.0611113213546687},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.03822870489707,2.2806853960415,-1.48273994391383,-3.58978558499056,3.26615690329693,2.45178262388351,-1.59397523182193,1.66498726415444,1.24984407270733,-.923099517802526,.555278916264744,.41682724950333,-.270991524688991,-.656083673232099,2.17523616991868,1.63286896677384,-1.06157563213921,4.50024407433993,3.3781660071394,6.26847807110811,.140734880491405,.10564444540962,-.0686825282564418,.792696592497175,.595047877309364,2.35258131414952,.0308216365902796,.0231366573273696,.225963663364161,.00626330408481313},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.59519010006696,2.69877562014986,-1.7545525650326,-4.24785717273985,5.47681396366563,4.11124073580283,-2.67283724139748,6.3309248170395,4.75239001984257,5.76936797267538,.233717104755142,.175442746128902,-.114060434690259,-.276145864886065,1.04841335581819,.787004949964004,-.511654819985593,2.66785366107377,2.00265861161846,5.04596922575242,.0249321870529789,.0187156663995726,-.0121675993597458,.15913874220381,.119459540563619,.559597243722218,.00229823078236368,.00172519645136334,.0189957623925199,.000196571591117826},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.92171958555151,2.94388917191153,-1.91390801783745,-4.63366448151075,6.96914454009643,5.23147784770941,-3.40113598727832,10.2353232484353,7.68327684524188,13.4470270492568,.0451952183653594,.0339263710725032,-.0220565211009043,-.0533999113042759,.217799593005347,.163494061616168,-.10629224716917,.614368644454251,.461183712763862,1.33333501160483,.000932318622845177,.000699856947415067,-.000454997367631839,.00636097475878139,.00477494739305922,.0244039729141499,1.66187980931003e-005,1.24751142122696e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.92171958555151,-2.26142559207414,-.826744191593013,1.19523987507356,2.19799643560933,2.5865404220042,.94560142914338,-2.29634628619806,-2.70227576356801,2.23836106649518,3.41890639763464,4.02327294720743,1.47084987158013,-2.12643576380383,-3.91413586924262,-4.60604507183384,-1.68390285488931,4.09536544265748,4.81931093640565,-4.00097073404889,5.33523806427656,6.27835818791885,2.29527612404421,-5.59797201662203,-6.58753611853026,5.49216459339238,7.19422103343416,8.46595709192435,-7.09892453011936,9.00134910987289},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.59519010006696,-1.87717487173315,-.686267736267333,.992150379408985,1.2388862761662,1.4578865459724,.532982044151546,-.514957807193037,-.605987872997064,-.353775594121333,3.23038451124486,3.80142569043747,1.38974575226926,-2.00918198880831,-2.62173967966463,-3.08518956279543,-1.12790027705863,1.30644820527577,1.53739152505215,.307822029727901,4.7630793830775,5.60505793446857,2.04912737785909,-2.90243304119459,-3.41550161982096,.518849452524734,6.06854814809742,7.14129688166466,-2.66498561200528,7.17423567654379},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.03822870489707,-1.22175835711557,-.446657024142497,.645740594449334,-.0622566343920988,-.0732618573826121,-.0267834658423845,1.10967175893133,1.30583053509787,-1.41978880536247,2.90882269973526,3.42302078938587,1.25140644309976,-1.80918220620513,-.707254507004454,-.832277224231996,-.304268406428257,-1.57948219926848,-1.8586902544778,2.54719860772099,3.86201434387853,4.54470992402429,1.661479620483,.544351948717181,.64057807233031,-3.18953813043078,4.43071582310306,5.21394183895337,2.19278566166382,4.71658607960972},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-.300186182387741,-.353250661721207,-.129143286331404,.186704916697283,-1.13611369814234,-1.33694666498547,-.488768188714077,1.22655366483195,1.44337387662839,.100858250439743,2.48271365058504,2.92158763778148,1.06808980107197,-1.54415783407832,1.26648273457434,1.49036128356923,.54485433405009,-2.5074905585056,-2.95074440626681,.461151554581851,2.81340752869054,3.31073889879025,1.21035782283977,3.19379348794833,3.75836640385388,-2.33714259619681,2.75487590413567,3.24186052799879,4.77320941254594,2.50302559510188},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,.553376983055106,.651198479192506,.23806865993447,-.344180410642768,-1.45368745167752,-1.71065853147384,-.625391792982136,-.194424990185747,-.228793863525153,1.49978166741586,1.9899087271236,2.34166865603201,.856079885009684,-1.23765104741928,2.74863027720701,3.23451085132423,1.18248996086068,-.854309601049289,-1.00532752474992,-2.52312426315122,1.80736427869175,2.12685548066823,.777547323353426,4.12997714955087,4.86004102210358,1.65834138994758,1.41847476579741,1.66922123290476,4.48820932686104,1.03297875354961},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,1.44662301694489,1.70234530427801,.622352597994008,-.899747042727636,-.724355119014703,-.852400743178548,-.31162527138391,-1.6918152058931,-1.9908805805112,.160475142667942,1.4741928888764,1.73478875376213,.634213449888053,-.916894502821386,3.38025060487184,3.97778389924568,1.45421973940992,2.16048862424162,2.54240228389271,-1.70770592621981,.991945941760336,1.16729410204273,.426745687640421,3.46032388717692,4.07201188589473,4.72311662432345,.576746649127291,.67869924509284,2.68886700736676,.311154445333738},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,2.30018618238774,2.70679444519172,.989564544259882,-1.43063237006769,.987081447023724,1.16156970101249,.424652930216328,-1.35561938023839,-1.5952547827211,-2.20870753342451,.981387965414959,1.15486977201266,.422203533825771,-.610387716162348,3.10522366683209,3.65413992905702,1.33590022736237,4.42237267757001,5.20412385861342,2.83495589139377,.43960319187862,.517312679568431,.189121965733584,2.03901059601618,2.39945035585008,4.55407145300732,.170154848292172,.200233442638166,1.03685515398485,.0611113213546687},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.03822870489707,3.57530214058608,1.30707828207097,-1.88966804781974,3.26615690329693,3.843521635486,1.4051354158397,1.66498726415444,1.95931021275678,-.923099517802526,.555278916264744,.653436620408273,.238886891797976,-.345363344035536,2.17523616991868,2.55975678291091,.935809721212699,4.50024407433993,5.29576072944733,6.26847807110811,.140734880491405,.165612851500527,.0605456414728775,.792696592497175,.932823068533675,2.35258131414952,.0308216365902796,.036270035583218,.225963663364161,.00626330408481313},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.59519010006696,4.23071865520366,1.54668899419581,-2.23607783277939,5.47681396366563,6.44496072482988,2.35618358032457,6.3309248170395,7.45005437415178,5.76936797267538,.233717104755142,.275031719356679,.100547582628472,-.145363561432351,1.04841335581819,1.23374336817428,.451038569416842,2.66785366107377,3.13945519772704,5.04596922575242,.0249321870529789,.0293394969148429,.0107260918769641,.15913874220381,.187269998673138,.559597243722218,.00229823078236368,.00270449338439953,.0189957623925199,.000196571591117826},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.92171958555151,4.61496937554466,1.68716544952149,-2.43916732844396,6.96914454009643,8.20109339929485,2.99820005631404,10.2353232484353,12.0446407028288,13.4470270492568,.0451952183653594,.0531844625867158,.0194434633176023,-.0281097864368307,.217799593005347,.256300439106579,.0936996999356592,.614368644454251,.722971751522571,1.33333501160483,.000932318622845177,.00109712634918437,.000401093381258233,.00636097475878139,.00748541629863617,.0244039729141499,1.66187980931003e-005,1.95565344631038e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.92171958555151,-2.87937722321757,-2.67495037465683,-1.63105841067899,2.19799643560933,3.29333213710544,3.05951577632408,-2.29634628619806,-3.44069303528758,2.23836106649518,3.41890639763464,5.1226626812653,4.75896952776425,2.90179486950885,-3.91413586924262,-5.86468169257239,-5.4483074914535,4.09536544265748,6.13622406614863,-4.00097073404889,5.33523806427656,7.99396700420041,7.42642015557747,-5.59797201662203,-8.38763014060611,5.49216459339238,7.19422103343416,10.7793438389747,-7.09892453011936,9.00134910987289},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.59519010006696,-2.39012709001286,-2.22043548283783,-1.35391669466665,1.2388862761662,1.85626506095126,1.72447600259692,-.514957807193037,-.771578638346196,-.353775594121333,3.23038451124486,4.84019398522881,4.49655523257725,2.74178702516,-2.62173967966463,-3.92824092358218,-3.64934800797338,1.30644820527577,1.95749537633152,.307822029727901,4.7630793830775,7.13668236880667,6.6300000687783,-2.90243304119459,-4.3488132461414,.518849452524734,6.06854814809742,9.09271021667368,-2.66498561200528,7.17423567654379},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.03822870489707,-1.55561306022354,-1.44516935978819,-.881196025717112,-.0622566343920988,-.0932812135041899,-.0866585368463816,1.10967175893133,1.66265859611778,-1.41978880536247,2.90882269973526,4.35838708560735,4.04895512766487,2.46886161968119,-.707254507004454,-1.0597032647262,-.984467621360889,-1.57948219926848,-2.36659141141637,2.54719860772099,3.86201434387853,5.78658625215441,5.37575658648661,.544351948717181,.815620871022689,-3.18953813043078,4.43071582310306,6.63869084478427,2.19278566166382,4.71658607960972},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-.300186182387741,-.449779074319961,-.417846155642134,-.254782852417372,-1.13611369814234,-1.70227744362135,-1.58142102819233,1.22655366483195,1.83778669504629,.100858250439743,2.48271365058504,3.7199335363258,3.45582979910559,2.10720180544034,1.26648273457434,1.89761376598671,1.76288907720341,-2.5074905585056,-3.75705761211831,.461151554581851,2.81340752869054,4.21542331996591,3.91614134649752,3.19379348794833,4.78536842321548,-2.33714259619681,2.75487590413567,4.12772341388704,4.77320941254594,2.50302559510188},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,.553376983055106,.829143384311425,.770276776736346,.469678401195648,-1.45368745167752,-2.17810890152299,-2.02346992933989,-.194424990185747,-.291313515578485,1.49978166741586,1.9899087271236,2.9815473107461,2.76986670414993,1.68893390563517,2.74863027720701,4.11836538252447,3.82597422031764,-.854309601049289,-1.2800408674882,-2.52312426315122,1.80736427869175,2.70803481201197,2.51577274454146,4.12997714955087,6.18808395553405,1.65834138994758,1.41847476579741,2.12534854800543,4.48820932686104,1.03297875354961},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,1.44662301694489,2.16752402217829,2.01363654211465,1.2278204669453,-.724355119014703,-1.08532568728512,-1.00827093182807,-1.6918152058931,-2.53490374291394,.160475142667942,1.4741928888764,2.20883288938782,2.05201260878721,1.25122037987573,3.38025060487184,5.06474340721712,4.70516234202085,2.16048862424162,3.23713291923203,-1.70770592621981,.991945941760336,1.48626603590141,1.38074575986725,3.46032388717692,5.18471990323708,4.72311662432345,.576746649127291,.864158942299266,2.68886700736676,.311154445333738},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,2.30018618238774,3.44644648080968,3.20175947449313,1.95228172055832,.987081447023724,1.47897739903412,1.37397459375255,-1.35561938023839,-2.03117020639316,-2.20870753342451,.981387965414959,1.47044666380812,1.36604951383154,.832952480070553,3.10522366683209,4.65266127660931,4.32233676392885,4.42237267757001,6.6261900271284,2.83495589139377,.43960319187862,.658672278252944,.611908590636855,2.03901059601618,3.05511829665785,4.55407145300732,.170154848292172,.254948743871771,1.03685515398485,.0611113213546687},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.03822870489707,4.55228046671326,4.22908267863918,2.57869489385806,3.26615690329693,4.89379296535329,4.54634885284358,1.66498726415444,2.49470653361407,-.923099517802526,.555278916264744,.831993114526565,.772924185272261,.471292665829706,2.17523616991868,3.25922966624351,3.02783447385586,4.50024407433993,6.74286736535737,6.26847807110811,.140734880491405,.210867814600684,.195896854176436,.792696592497175,1.18772402074228,2.35258131414952,.0308216365902796,.0461810968751895,.225963663364161,.00626330408481313},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.59519010006696,5.38679449650258,5.00434880168882,3.0514155628076,5.47681396366563,8.20609494322841,7.62348767011616,6.3309248170395,9.48583802706066,5.76936797267538,.233717104755142,.350186214905109,.325324080359881,.198367260350895,1.04841335581819,1.57087306611691,1.45934595309525,2.66785366107377,3.99733505330865,5.04596922575242,.0249321870529789,.0373567361461482,.0347045238135923,.15913874220381,.23844294088706,.559597243722218,.00229823078236368,.00344351663776842,.0189957623925199,.000196571591117826},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.92171958555151,5.87604462970729,5.45886369350783,3.32855727881994,6.96914454009643,10.4421041409331,9.70074715430413,10.2353232484353,15.3359297883224,13.4470270492568,.0451952183653594,.067717518868618,.0629097851728847,.0383594160020484,.217799593005347,.326336470786676,.303167594035556,.614368644454251,.920529245649199,1.33333501160483,.000932318622845177,.00139692441436285,.00129774711619293,.00636097475878139,.00953086286461726,.0244039729141499,1.66187980931003e-005,2.49005052810442e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.92171958555151,-3.24166342459793,-3.96510583993544,-4.31411318985288,2.19799643560933,3.70770257117539,4.53515100205683,-2.29634628619806,-3.87360456901059,2.23836106649518,3.41890639763464,5.76720136441661,7.05426838770182,7.67518283761778,-3.91413586924262,-6.60258197030424,-8.07608098333328,4.09536544265748,6.90829008438974,-4.00097073404889,5.33523806427656,8.99977614812198,11.0082572774727,-5.59797201662203,-9.44297038492047,5.49216459339238,7.19422103343416,12.1356119587979,-7.09892453011936,9.00134910987289},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.59519010006696,-2.69085533682779,-3.29137384514269,-3.58107952001051,1.2388862761662,2.0898222386161,2.55620812016103,-.514957807193037,-.868659455579414,-.353775594121333,3.23038451124486,5.44919216674985,6.66528903908045,7.25196565099482,-2.62173967966463,-4.42249623366662,-5.4094652504454,1.30644820527577,2.20378945631257,.307822029727901,4.7630793830775,8.03462707886587,9.82771577392637,-2.90243304119459,-4.89598539807522,.518849452524734,6.06854814809742,10.2367643607743,-2.66498561200528,7.17423567654379},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.03822870489707,-1.75134189417476,-2.14218907478874,-2.33074387312075,-.0622566343920988,-.105017951653011,-.12845482061473,1.10967175893133,1.87185600940662,-1.41978880536247,2.90882269973526,4.90676382786187,6.00180690245521,6.53008402866105,-.707254507004454,-1.19303621855815,-1.45928625505293,-1.57948219926848,-2.66435837496252,2.54719860772099,3.86201434387853,6.51466048131366,7.96853804732135,.544351948717181,.918243127233759,-3.18953813043078,4.43071582310306,7.47397774950178,2.19278566166382,4.71658607960972},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-.300186182387741,-.506370739691845,-.619377558413115,-.673894973328732,-1.13611369814234,-1.91645974101964,-2.34415629302528,1.22655366483195,2.06901890572263,.100858250439743,2.48271365058504,4.18797939686675,5.12261126340126,5.57350187032702,1.26648273457434,2.13637348021345,2.61314820693614,-2.5074905585056,-4.22977446202813,.461151554581851,2.81340752869054,4.74581221430261,5.80493569904152,3.19379348794833,5.38746839617942,-2.33714259619681,2.75487590413567,4.64707781591112,4.77320941254594,2.50302559510188},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,.553376983055106,.93346705704166,1.14178901214037,1.24228891639978,-1.45368745167752,-2.45216080196962,-2.99940982448642,-.194424990185747,-.327966881493408,1.49978166741586,1.9899087271236,3.35668865754018,4.10580126963162,4.46719258573777,2.74863027720701,4.6365423474252,5.67128006117502,-.854309601049289,-1.44109692494208,-2.52312426315122,1.80736427869175,3.04876253512471,3.72915523810848,4.12997714955087,6.96667503835439,1.65834138994758,1.41847476579741,2.39276208655032,4.48820932686104,1.03297875354961},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,1.44662301694489,2.44024412222762,2.98483369571688,3.24755780451479,-.724355119014703,-1.22188248065577,-1.49457014152702,-1.6918152058931,-2.85384793698481,.160475142667942,1.4741928888764,2.48675051356285,3.04171892525937,3.30945005334757,3.38025060487184,5.70199462778367,6.97450945518368,2.16048862424162,3.64443230995284,-1.70770592621981,.991945941760336,1.67326955598402,2.04669332477546,3.46032388717692,5.8370666898335,4.72311662432345,.576746649127291,.972888308521253,2.68886700736676,.311154445333738},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,2.30018618238774,3.88008191896113,4.74600026627036,5.1637416942433,.987081447023724,1.66506385533551,2.03665635715207,-1.35561938023839,-2.28673412920868,-2.20870753342451,.981387965414959,1.65545977423627,2.02490893148973,2.20314076875832,3.10522366683209,5.23806390000324,6.40704325104372,4.42237267757001,7.45990406612127,2.83495589139377,.43960319187862,.741547101224612,.907038257317998,2.03901059601618,3.43951642050873,4.55407145300732,.170154848292172,.287026656838237,1.03685515398485,.0611113213546687},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.03822870489707,5.12505307344404,6.26881178264598,6.82059059403532,3.26615690329693,5.50953502563763,6.7390986231311,1.66498726415444,2.80859307349989,-.923099517802526,.555278916264744,.936675343241156,1.14571329243577,1.24655861042429,2.17523616991868,3.66930929238228,4.48818948881768,4.50024407433993,7.59126186695628,6.26847807110811,.140734880491405,.237399419743997,.29037987690505,.792696592497175,1.33716467766395,2.35258131414952,.0308216365902796,.0519916499446463,.225963663364161,.00626330408481313},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.59519010006696,6.06456651609707,7.41799655299993,8.07092624092508,5.47681396366563,9.23859424244377,11.300372436005,6.3309248170395,10.6793559162843,5.76936797267538,.233717104755142,.394247004353173,.482231155810537,.524676988090528,1.04841335581819,1.76852192969244,2.1632031816091,2.66785366107377,4.50028385780377,5.04596922575242,.0249321870529789,.0420569990798228,.0514428646412205,.15913874220381,.268444076758866,.559597243722218,.00229823078236368,.0038767834400574,.0189957623925199,.000196571591117826},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.92171958555151,6.61537460386721,8.09172854779269,8.80395991076745,6.96914454009643,11.7559404153652,14.3795150585562,10.2353232484353,17.2655122351262,13.4470270492568,.0451952183653594,.0762378066864121,.0932518071891671,.101459801467564,.217799593005347,.367396460919218,.449388373324038,.614368644454251,1.03635118137081,1.33333501160483,.000932318622845177,.0015726868798065,.00192366360061394,.00636097475878139,.0107300458251746,.0244039729141499,1.66187980931003e-005,2.80335124459201e-005,.000146461329045194,2.74870868673343e-007}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_hyp_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_hyp_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.92171958555151,3.24166342459793,-3.96510583993544,4.31411318985288,2.19799643560933,-3.70770257117539,4.53515100205683,-2.29634628619806,3.87360456901059,2.23836106649518,3.41890639763464,-5.76720136441661,7.05426838770182,-7.67518283761778,-3.91413586924262,6.60258197030424,-8.07608098333328,4.09536544265748,-6.90829008438974,-4.00097073404889,5.33523806427656,-8.99977614812198,11.0082572774727,-5.59797201662203,9.44297038492047,5.49216459339238,7.19422103343416,-12.1356119587979,-7.09892453011936,9.00134910987289},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.59519010006696,2.69085533682779,-3.29137384514269,3.5810795200105,1.2388862761662,-2.0898222386161,2.55620812016103,-.514957807193037,.868659455579414,-.353775594121333,3.23038451124486,-5.44919216674985,6.66528903908045,-7.25196565099481,-2.62173967966463,4.42249623366662,-5.4094652504454,1.30644820527577,-2.20378945631257,.307822029727901,4.7630793830775,-8.03462707886587,9.82771577392637,-2.90243304119459,4.89598539807522,.518849452524734,6.06854814809742,-10.2367643607743,-2.66498561200528,7.17423567654379},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.03822870489707,1.75134189417476,-2.14218907478874,2.33074387312075,-.0622566343920988,.105017951653011,-.12845482061473,1.10967175893133,-1.87185600940662,-1.41978880536247,2.90882269973526,-4.90676382786187,6.00180690245521,-6.53008402866105,-.707254507004454,1.19303621855815,-1.45928625505293,-1.57948219926848,2.66435837496252,2.54719860772099,3.86201434387853,-6.51466048131366,7.96853804732135,.544351948717181,-.918243127233759,-3.18953813043078,4.43071582310306,-7.47397774950178,2.19278566166382,4.71658607960972},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-.300186182387741,.506370739691845,-.619377558413115,.673894973328731,-1.13611369814234,1.91645974101964,-2.34415629302528,1.22655366483195,-2.06901890572263,.100858250439743,2.48271365058504,-4.18797939686675,5.12261126340126,-5.57350187032702,1.26648273457434,-2.13637348021345,2.61314820693614,-2.5074905585056,4.22977446202813,.461151554581851,2.81340752869054,-4.74581221430261,5.80493569904152,3.19379348794833,-5.38746839617942,-2.33714259619681,2.75487590413567,-4.64707781591112,4.77320941254594,2.50302559510188},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,.553376983055106,-.93346705704166,1.14178901214037,-1.24228891639978,-1.45368745167752,2.45216080196962,-2.99940982448642,-.194424990185747,.327966881493408,1.49978166741586,1.9899087271236,-3.35668865754018,4.10580126963162,-4.46719258573777,2.74863027720701,-4.6365423474252,5.67128006117502,-.854309601049289,1.44109692494208,-2.52312426315122,1.80736427869175,-3.04876253512471,3.72915523810848,4.12997714955087,-6.96667503835439,1.65834138994758,1.41847476579741,-2.39276208655032,4.48820932686104,1.03297875354961},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,1.44662301694489,-2.44024412222762,2.98483369571688,-3.24755780451478,-.724355119014703,1.22188248065577,-1.49457014152702,-1.6918152058931,2.85384793698481,.160475142667942,1.4741928888764,-2.48675051356285,3.04171892525937,-3.30945005334757,3.38025060487184,-5.70199462778367,6.97450945518368,2.16048862424162,-3.64443230995284,-1.70770592621981,.991945941760336,-1.67326955598402,2.04669332477546,3.46032388717692,-5.8370666898335,4.72311662432345,.576746649127291,-.972888308521253,2.68886700736676,.311154445333738},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,2.30018618238774,-3.88008191896113,4.74600026627036,-5.1637416942433,.987081447023724,-1.66506385533551,2.03665635715207,-1.35561938023839,2.28673412920868,-2.20870753342451,.981387965414959,-1.65545977423627,2.02490893148973,-2.20314076875832,3.10522366683209,-5.23806390000324,6.40704325104372,4.42237267757001,-7.45990406612127,2.83495589139377,.43960319187862,-.741547101224612,.907038257317998,2.03901059601618,-3.43951642050873,4.55407145300732,.170154848292172,-.287026656838237,1.03685515398485,.0611113213546687},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.03822870489707,-5.12505307344404,6.26881178264598,-6.82059059403531,3.26615690329693,-5.50953502563763,6.7390986231311,1.66498726415444,-2.80859307349989,-.923099517802526,.555278916264744,-.936675343241156,1.14571329243577,-1.24655861042429,2.17523616991868,-3.66930929238228,4.48818948881768,4.50024407433993,-7.59126186695628,6.26847807110811,.140734880491405,-.237399419743997,.29037987690505,.792696592497175,-1.33716467766395,2.35258131414952,.0308216365902796,-.0519916499446463,.225963663364161,.00626330408481313},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.59519010006696,-6.06456651609707,7.41799655299993,-8.07092624092507,5.47681396366563,-9.23859424244377,11.300372436005,6.3309248170395,-10.6793559162843,5.76936797267538,.233717104755142,-.394247004353173,.482231155810537,-.524676988090528,1.04841335581819,-1.76852192969244,2.1632031816091,2.66785366107377,-4.50028385780377,5.04596922575242,.0249321870529789,-.0420569990798228,.0514428646412205,.15913874220381,-.268444076758866,.559597243722218,.00229823078236368,-.0038767834400574,.0189957623925199,.000196571591117826},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.92171958555151,-6.61537460386721,8.09172854779269,-8.80395991076744,6.96914454009643,-11.7559404153652,14.3795150585562,10.2353232484353,-17.2655122351262,13.4470270492568,.0451952183653594,-.0762378066864121,.0932518071891671,-.101459801467564,.217799593005347,-.367396460919218,.449388373324038,.614368644454251,-1.03635118137081,1.33333501160483,.000932318622845177,-.0015726868798065,.00192366360061394,.00636097475878139,-.0107300458251746,.0244039729141499,1.66187980931003e-005,-2.80335124459201e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.92171958555151,2.87937722321757,-2.67495037465683,1.63105841067898,2.19799643560933,-3.29333213710544,3.05951577632408,-2.29634628619806,3.44069303528758,2.23836106649518,3.41890639763464,-5.1226626812653,4.75896952776425,-2.90179486950883,-3.91413586924262,5.86468169257239,-5.4483074914535,4.09536544265748,-6.13622406614863,-4.00097073404889,5.33523806427656,-7.99396700420041,7.42642015557747,-5.59797201662203,8.38763014060611,5.49216459339238,7.19422103343416,-10.7793438389747,-7.09892453011936,9.00134910987289},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.59519010006696,2.39012709001286,-2.22043548283783,1.35391669466664,1.2388862761662,-1.85626506095126,1.72447600259692,-.514957807193037,.771578638346196,-.353775594121333,3.23038451124486,-4.84019398522881,4.49655523257725,-2.74178702515998,-2.62173967966463,3.92824092358218,-3.64934800797338,1.30644820527577,-1.95749537633152,.307822029727901,4.7630793830775,-7.13668236880667,6.6300000687783,-2.90243304119459,4.3488132461414,.518849452524734,6.06854814809742,-9.09271021667368,-2.66498561200528,7.17423567654379},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.03822870489707,1.55561306022354,-1.44516935978819,.881196025717106,-.0622566343920988,.0932812135041899,-.0866585368463816,1.10967175893133,-1.66265859611778,-1.41978880536247,2.90882269973526,-4.35838708560735,4.04895512766487,-2.46886161968117,-.707254507004454,1.0597032647262,-.984467621360889,-1.57948219926848,2.36659141141637,2.54719860772099,3.86201434387853,-5.78658625215441,5.37575658648661,.544351948717181,-.815620871022689,-3.18953813043078,4.43071582310306,-6.63869084478427,2.19278566166382,4.71658607960972},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-.300186182387741,.449779074319961,-.417846155642134,.25478285241737,-1.13611369814234,1.70227744362135,-1.58142102819233,1.22655366483195,-1.83778669504629,.100858250439743,2.48271365058504,-3.7199335363258,3.45582979910559,-2.10720180544033,1.26648273457434,-1.89761376598671,1.76288907720341,-2.5074905585056,3.75705761211831,.461151554581851,2.81340752869054,-4.21542331996591,3.91614134649752,3.19379348794833,-4.78536842321548,-2.33714259619681,2.75487590413567,-4.12772341388704,4.77320941254594,2.50302559510188},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,.553376983055106,-.829143384311425,.770276776736346,-.469678401195645,-1.45368745167752,2.17810890152299,-2.02346992933989,-.194424990185747,.291313515578485,1.49978166741586,1.9899087271236,-2.9815473107461,2.76986670414993,-1.68893390563516,2.74863027720701,-4.11836538252447,3.82597422031764,-.854309601049289,1.2800408674882,-2.52312426315122,1.80736427869175,-2.70803481201197,2.51577274454146,4.12997714955087,-6.18808395553405,1.65834138994758,1.41847476579741,-2.12534854800543,4.48820932686104,1.03297875354961},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,1.44662301694489,-2.16752402217829,2.01363654211465,-1.22782046694529,-.724355119014703,1.08532568728512,-1.00827093182807,-1.6918152058931,2.53490374291394,.160475142667942,1.4741928888764,-2.20883288938782,2.05201260878721,-1.25122037987572,3.38025060487184,-5.06474340721712,4.70516234202085,2.16048862424162,-3.23713291923203,-1.70770592621981,.991945941760336,-1.48626603590141,1.38074575986725,3.46032388717692,-5.18471990323708,4.72311662432345,.576746649127291,-.864158942299266,2.68886700736676,.311154445333738},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,2.30018618238774,-3.44644648080968,3.20175947449313,-1.95228172055831,.987081447023724,-1.47897739903412,1.37397459375255,-1.35561938023839,2.03117020639316,-2.20870753342451,.981387965414959,-1.47044666380812,1.36604951383154,-.832952480070547,3.10522366683209,-4.65266127660931,4.32233676392885,4.42237267757001,-6.6261900271284,2.83495589139377,.43960319187862,-.658672278252944,.611908590636855,2.03901059601618,-3.05511829665785,4.55407145300732,.170154848292172,-.254948743871771,1.03685515398485,.0611113213546687},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.03822870489707,-4.55228046671326,4.22908267863918,-2.57869489385804,3.26615690329693,-4.89379296535329,4.54634885284358,1.66498726415444,-2.49470653361407,-.923099517802526,.555278916264744,-.831993114526565,.772924185272261,-.471292665829703,2.17523616991868,-3.25922966624351,3.02783447385586,4.50024407433993,-6.74286736535737,6.26847807110811,.140734880491405,-.210867814600684,.195896854176436,.792696592497175,-1.18772402074228,2.35258131414952,.0308216365902796,-.0461810968751895,.225963663364161,.00626330408481313},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.59519010006696,-5.38679449650258,5.00434880168882,-3.05141556280758,5.47681396366563,-8.20609494322841,7.62348767011616,6.3309248170395,-9.48583802706066,5.76936797267538,.233717104755142,-.350186214905109,.325324080359881,-.198367260350894,1.04841335581819,-1.57087306611691,1.45934595309525,2.66785366107377,-3.99733505330865,5.04596922575242,.0249321870529789,-.0373567361461482,.0347045238135923,.15913874220381,-.23844294088706,.559597243722218,.00229823078236368,-.00344351663776842,.0189957623925199,.000196571591117826},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.92171958555151,-5.87604462970729,5.45886369350783,-3.32855727881992,6.96914454009643,-10.4421041409331,9.70074715430413,10.2353232484353,-15.3359297883224,13.4470270492568,.0451952183653594,-.067717518868618,.0629097851728847,-.0383594160020481,.217799593005347,-.326336470786676,.303167594035556,.614368644454251,-.920529245649199,1.33333501160483,.000932318622845177,-.00139692441436285,.00129774711619293,.00636097475878139,-.00953086286461726,.0244039729141499,1.66187980931003e-005,-2.49005052810442e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.92171958555151,2.26142559207414,-.826744191593017,-1.19523987507357,2.19799643560933,-2.5865404220042,.945601429143385,-2.29634628619806,2.70227576356801,2.23836106649518,3.41890639763464,-4.02327294720743,1.47084987158014,2.12643576380386,-3.91413586924262,4.60604507183384,-1.68390285488932,4.09536544265748,-4.81931093640565,-4.00097073404889,5.33523806427656,-6.27835818791885,2.29527612404422,-5.59797201662203,6.58753611853026,5.49216459339238,7.19422103343416,-8.46595709192435,-7.09892453011936,9.00134910987289},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.59519010006696,1.87717487173315,-.686267736267336,-.992150379408995,1.2388862761662,-1.4578865459724,.532982044151549,-.514957807193037,.605987872997064,-.353775594121333,3.23038451124486,-3.80142569043747,1.38974575226927,2.00918198880834,-2.62173967966463,3.08518956279543,-1.12790027705863,1.30644820527577,-1.53739152505215,.307822029727901,4.7630793830775,-5.60505793446857,2.0491273778591,-2.90243304119459,3.41550161982096,.518849452524734,6.06854814809742,-7.14129688166466,-2.66498561200528,7.17423567654379},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.03822870489707,1.22175835711557,-.446657024142499,-.64574059444934,-.0622566343920988,.0732618573826121,-.0267834658423847,1.10967175893133,-1.30583053509787,-1.41978880536247,2.90882269973526,-3.42302078938587,1.25140644309977,1.80918220620515,-.707254507004454,.832277224231996,-.304268406428259,-1.57948219926848,1.8586902544778,2.54719860772099,3.86201434387853,-4.54470992402429,1.66147962048301,.544351948717181,-.64057807233031,-3.18953813043078,4.43071582310306,-5.21394183895337,2.19278566166382,4.71658607960972},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-.300186182387741,.353250661721207,-.129143286331405,-.186704916697285,-1.13611369814234,1.33694666498547,-.488768188714079,1.22655366483195,-1.44337387662839,.100858250439743,2.48271365058504,-2.92158763778148,1.06808980107197,1.54415783407833,1.26648273457434,-1.49036128356923,.544854334050093,-2.5074905585056,2.95074440626681,.461151554581851,2.81340752869054,-3.31073889879025,1.21035782283977,3.19379348794833,-3.75836640385388,-2.33714259619681,2.75487590413567,-3.24186052799879,4.77320941254594,2.50302559510188},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,.553376983055106,-.651198479192506,.238068659934471,.344180410642771,-1.45368745167752,1.71065853147384,-.625391792982139,-.194424990185747,.228793863525153,1.49978166741586,1.9899087271236,-2.34166865603201,.856079885009688,1.23765104741929,2.74863027720701,-3.23451085132423,1.18248996086068,-.854309601049289,1.00532752474992,-2.52312426315122,1.80736427869175,-2.12685548066823,.77754732335343,4.12997714955087,-4.86004102210358,1.65834138994758,1.41847476579741,-1.66922123290476,4.48820932686104,1.03297875354961},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,1.44662301694489,-1.70234530427801,.622352597994011,.899747042727645,-.724355119014703,.852400743178548,-.311625271383912,-1.6918152058931,1.9908805805112,.160475142667942,1.4741928888764,-1.73478875376213,.634213449888056,.916894502821395,3.38025060487184,-3.97778389924568,1.45421973940992,2.16048862424162,-2.54240228389271,-1.70770592621981,.991945941760336,-1.16729410204273,.426745687640423,3.46032388717692,-4.07201188589473,4.72311662432345,.576746649127291,-.67869924509284,2.68886700736676,.311154445333738},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,2.30018618238774,-2.70679444519172,.989564544259887,1.4306323700677,.987081447023724,-1.16156970101249,.42465293021633,-1.35561938023839,1.5952547827211,-2.20870753342451,.981387965414959,-1.15486977201266,.422203533825773,.610387716162354,3.10522366683209,-3.65413992905702,1.33590022736237,4.42237267757001,-5.20412385861342,2.83495589139377,.43960319187862,-.517312679568431,.189121965733585,2.03901059601618,-2.39945035585008,4.55407145300732,.170154848292172,-.200233442638166,1.03685515398485,.0611113213546687},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.03822870489707,-3.57530214058608,1.30707828207098,1.88966804781976,3.26615690329693,-3.843521635486,1.40513541583971,1.66498726415444,-1.95931021275678,-.923099517802526,.555278916264744,-.653436620408273,.238886891797977,.345363344035539,2.17523616991868,-2.55975678291091,.935809721212703,4.50024407433993,-5.29576072944733,6.26847807110811,.140734880491405,-.165612851500527,.0605456414728778,.792696592497175,-.932823068533675,2.35258131414952,.0308216365902796,-.036270035583218,.225963663364161,.00626330408481313},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.59519010006696,-4.23071865520366,1.54668899419582,2.23607783277941,5.47681396366563,-6.44496072482988,2.35618358032458,6.3309248170395,-7.45005437415178,5.76936797267538,.233717104755142,-.275031719356679,.100547582628473,.145363561432352,1.04841335581819,-1.23374336817428,.451038569416845,2.66785366107377,-3.13945519772704,5.04596922575242,.0249321870529789,-.0293394969148429,.0107260918769642,.15913874220381,-.187269998673138,.559597243722218,.00229823078236368,-.00270449338439953,.0189957623925199,.000196571591117826},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.92171958555151,-4.61496937554466,1.6871654495215,2.43916732844399,6.96914454009643,-8.20109339929485,2.99820005631406,10.2353232484353,-12.0446407028288,13.4470270492568,.0451952183653594,-.0531844625867158,.0194434633176024,.028109786436831,.217799593005347,-.256300439106579,.0936996999356596,.614368644454251,-.722971751522571,1.33333501160483,.000932318622845177,-.00109712634918437,.000401093381258235,.00636097475878139,-.00748541629863617,.0244039729141499,1.66187980931003e-005,-1.95565344631038e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.92171958555151,1.44256348674145,.937852501329485,-2.27058656100761,2.19799643560933,-1.64995425135427,-1.07268327325537,-2.29634628619806,1.72378183093533,2.23836106649518,3.41890639763464,-2.56644599370298,-1.66852122478612,4.03957110939073,-3.91413586924262,2.93819635817362,1.91020695459719,4.09536544265748,-3.07423866276144,-4.00097073404889,5.33523806427656,-4.00496497002541,-2.60374427309723,-5.59797201662203,4.20218958584716,5.49216459339238,7.19422103343416,-5.40043440913241,-7.09892453011936,9.00134910987289},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.59519010006696,1.19744993497978,.77849704852463,-1.88477925223671,1.2388862761662,-.929985893147364,-.604610892164935,-.514957807193037,.386559691398865,-.353775594121333,3.23038451124486,-2.42492961864658,-1.57651731119676,3.81682515580894,-2.62173967966463,1.9680425606586,1.27948173914766,1.30644820527577,-.980702122872796,.307822029727901,4.7630793830775,-3.57546670811254,-2.32451495445634,-2.90243304119459,2.17874863666902,.518849452524734,6.06854814809742,-4.55543360151919,-2.66498561200528,7.17423567654379},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.03822870489707,.779359710871422,.506684427405858,-1.22670766448742,-.0622566343920988,.0467337421144522,.030382965722629,1.10967175893133,-.832989356981887,-1.41978880536247,2.90882269973526,-2.18354511527215,-1.41958622119803,3.4368873474629,-.707254507004454,.530909678713955,.345159831718241,-1.57948219926848,1.18565859680248,2.54719860772099,3.86201434387853,-2.89907066463143,-1.88477020319366,.544351948717181,-.408624781106939,-3.18953813043078,4.43071582310306,-3.32597373321893,2.19278566166382,4.71658607960972},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-.300186182387741,.225338612975934,.146499189649512,-.354681669820299,-1.13611369814234,.852838337634571,.554455021295674,1.22655366483195,-.920728260721081,.100858250439743,2.48271365058504,-1.86368088534508,-1.21163317718593,2.93342290468435,1.26648273457434,-.950701529718779,-.618078730089993,-2.5074905585056,1.88227998927126,.461151554581851,2.81340752869054,-2.11192049221879,-1.37302096972311,3.19379348794833,-2.39746209780118,-2.33714259619681,2.75487590413567,-2.06798297639769,4.77320941254594,2.50302559510188},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,.553376983055106,-.41539953912128,-.270063328509738,.65383646518608,-1.45368745167752,1.09122915405036,.709439828333213,-.194424990185747,.145947615817496,1.49978166741586,1.9899087271236,-1.49375054084377,-.971130695151503,2.35115468793597,2.74863027720701,-2.06329461731692,-1.34140787303666,-.854309601049289,.641298473185774,-2.52312426315122,1.80736427869175,-1.35672120663226,-.88204393756181,4.12997714955087,-3.10022038625728,1.65834138994758,1.41847476579741,-1.06479629943225,4.48820932686104,1.03297875354961},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,1.44662301694489,-1.0859261460488,-.705992187998236,1.70924145531706,-.724355119014703,.543746472353678,.353505404956933,-1.6918152058931,1.26998281171566,.160475142667942,1.4741928888764,-1.10662182393172,-.719447050735517,1.74181633275903,3.38025060487184,-2.53742852796533,-1.64965612574837,2.16048862424162,-1.62179853107076,-1.70770592621981,.991945941760336,-.744616960114543,-.484097154422056,3.46032388717692,-2.5975365648804,4.72311662432345,.576746649127291,-.432942278924123,2.68886700736676,.311154445333738},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,2.30018618238774,-1.72666429814602,-1.12255470615749,2.71775959032344,.987081447023724,-.740965364440283,-.481723145865576,-1.35561938023839,1.01761309753837,-2.20870753342451,.981387965414959,-.736691479430405,-.47894456870109,1.15954811601065,3.10522366683209,-2.33097602484761,-1.51543534565912,4.42237267757001,-3.31971084311917,2.83495589139377,.43960319187862,-.329993781528465,-.214538560322786,2.03901059601618,-1.53060948975261,4.55407145300732,.170154848292172,-.12772892205108,1.03685515398485,.0611113213546687},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.03822870489707,-2.2806853960415,-1.48273994391383,3.58978558499055,3.26615690329693,-2.45178262388351,-1.59397523182193,1.66498726415444,-1.24984407270733,-.923099517802526,.555278916264744,-.41682724950333,-.270991524688991,.656083673232097,2.17523616991868,-1.63286896677384,-1.06157563213922,4.50024407433993,-3.3781660071394,6.26847807110811,.140734880491405,-.10564444540962,-.0686825282564419,.792696592497175,-.595047877309364,2.35258131414952,.0308216365902796,-.0231366573273696,.225963663364161,.00626330408481313},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.59519010006696,-2.69877562014986,-1.7545525650326,4.24785717273984,5.47681396366563,-4.11124073580283,-2.67283724139748,6.3309248170395,-4.75239001984257,5.76936797267538,.233717104755142,-.175442746128902,-.114060434690259,.276145864886064,1.04841335581819,-.787004949964004,-.511654819985594,2.66785366107377,-2.00265861161846,5.04596922575242,.0249321870529789,-.0187156663995726,-.0121675993597458,.15913874220381,-.119459540563619,.559597243722218,.00229823078236368,-.00172519645136334,.0189957623925199,.000196571591117826},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.92171958555151,-2.94388917191153,-1.91390801783746,4.63366448151074,6.96914454009643,-5.23147784770941,-3.40113598727833,10.2353232484353,-7.68327684524188,13.4470270492568,.0451952183653594,-.0339263710725032,-.0220565211009044,.0533999113042758,.217799593005347,-.163494061616168,-.10629224716917,.614368644454251,-.461183712763862,1.33333501160483,.000932318622845177,-.000699856947415067,-.000454997367631841,.00636097475878139,-.00477494739305922,.0244039729141499,1.66187980931003e-005,-1.24751142122696e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.92171958555151,.49553061346938,2.00568935114481,-1.09346229146625,2.19799643560933,-.566770786786546,-2.29403815127598,-2.29634628619806,.592131075007482,2.23836106649518,3.41890639763464,-.881592088933561,-3.56829591856781,1.94536458449533,-3.91413586924262,1.00929093051433,4.08516450282659,4.09536544265748,-1.05602241045943,-4.00097073404889,5.33523806427656,-1.37573338488067,-5.56836192711791,-5.59797201662203,1.44348141522225,5.49216459339238,7.19422103343416,-1.855086864706,-7.09892453011936,9.00134910987289},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.59519010006696,.411332399809833,1.66489212099991,-.907666360512686,1.2388862761662,-.31945663701093,-1.29301955934683,-.514957807193037,.132785948612982,-.353775594121333,3.23038451124486,-.832980227624035,-3.37153654597113,1.8380952537413,-2.62173967966463,.676036338352017,2.73629690762123,1.30644820527577,-.336878015452627,.307822029727901,4.7630793830775,-1.22819773798136,-4.97120270417874,-2.90243304119459,.748415344244522,.518849452524734,6.06854814809742,-1.56482319698726,-2.66498561200528,7.17423567654379},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.03822870489707,.26771549341915,1.08359423150038,-.590754211622909,-.0622566343920988,.0160533661863334,.0649769454361842,1.10967175893133,-.286137650697852,-1.41978880536247,2.90882269973526,-.750062968079831,-3.03592405293217,1.65512593926409,-.707254507004454,.182371175485818,.738158077031132,-1.57948219926848,.407281992951262,2.54719860772099,3.86201434387853,-.995850982046505,-4.03076552074204,.544351948717181,-.140365460738301,-3.18953813043078,4.43071582310306,-1.14249516194806,2.19278566166382,4.71658607960972},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-.300186182387741,.0774053843401605,.313302853289657,-.170806538752116,-1.13611369814234,.292955913914665,1.18575632130325,1.22655366483195,-.316276575739705,.100858250439743,2.48271365058504,-.640187375400919,-2.59119611829218,1.41266903727833,1.26648273457434,-.326572602577083,-1.32182184869504,-2.5074905585056,.646576297249579,.461151554581851,2.81340752869054,-.725459410782454,-2.93633970583683,3.19379348794833,-.823544942743958,-2.33714259619681,2.75487590413567,-.71036656781973,4.77320941254594,2.50302559510188},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,.553376983055106,-.142692637341486,-.577556856071558,.314872611220466,-1.45368745167752,.374844821120283,1.51720649776892,-.194424990185747,.0501340234001082,1.49978166741586,1.9899087271236,-.513113723366546,-2.07685802519487,1.13226204929232,2.74863027720701,-.708756083794028,-2.86873200495546,-.854309601049289,.220290495881434,-2.52312426315122,1.80736427869175,-.466043191865359,-1.88633727617264,4.12997714955087,-1.06494731359279,1.65834138994758,1.41847476579741,-.365764951301697,4.48820932686104,1.03297875354961},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,1.44662301694489,-.373023200905715,-1.50983338153088,.823131392784558,-.724355119014703,.186780703583358,.756005902089247,-1.6918152058931,.436247948578379,.160475142667942,1.4741928888764,-.380132310523241,-1.53860792217029,.83881870492831,3.38025060487184,-.871624387178691,-3.52795105816102,2.16048862424162,-.557099100602201,-1.70770592621981,.991945941760336,-.255781116350547,-1.03528913786275,3.46032388717692,-.892271916858127,4.72311662432345,.576746649127291,-.148718690750111,2.68886700736676,.311154445333738},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,2.30018618238774,-.593121222587361,-2.40069309089209,1.30881054275714,.987081447023724,-.254526767782017,-1.03021208824726,-1.35561938023839,.349557192548049,-2.20870753342451,.981387965414959,-.253058658488868,-1.02426982907298,.558411716942304,3.10522366683209,-.800706513225493,-3.2409067852655,4.42237267757001,-1.14034381554164,2.83495589139377,.43960319187862,-.113355164264732,-.458811705720641,2.03901059601618,-.525775029251871,4.55407145300732,.170154848292172,-.0438757750930783,1.03685515398485,.0611113213546687},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.03822870489707,-.78343133166635,-3.17098446910282,1.72875821562793,3.26615690329693,-.842204422108959,-3.40887201763794,1.66498726415444,-.429330151346565,-.923099517802526,.555278916264744,-.143183065809956,-.579541894432988,.315954814956539,2.17523616991868,-.560901872435633,-2.27028337376948,4.50024407433993,-1.16042357185991,6.26847807110811,.140734880491405,-.0362896034209995,-.14688430786117,.792696592497175,-.204403093795768,2.35258131414952,.0308216365902796,-.00794760307219474,.225963663364161,.00626330408481313},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.59519010006696,-.927048238057034,-3.75228235860234,2.04567036451771,5.47681396366563,-1.4122398512488,-5.7161240011779,6.3309248170395,-1.63247909961861,5.76936797267538,.233717104755142,-.060265806265752,-.243929401394036,.132985500479333,1.04841335581819,-.27034168634063,-1.09422390242844,2.66785366107377,-.68792719315715,5.04596922575242,.0249321870529789,-.0064289618707996,-.0260216019365979,.15913874220381,-.0410351849082053,.559597243722218,.00229823078236368,-.000592617006985495,.0189957623925199,.000196571591117826},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.92171958555151,-1.01124645171658,-4.09307958874724,2.23146629547127,6.96914454009643,-1.7970491081004,-7.2736621396321,10.2353232484353,-2.63925915467191,13.4470270492568,.0451952183653594,-.0116539449562267,-.0471700287973566,.0257161697253048,.217799593005347,-.0561613498441344,-.227316372195247,.614368644454251,-.158419820138511,1.33333501160483,.000932318622845177,-.000240405739976666,-.000973056396143785,.00636097475878139,-.00164022771456071,.0244039729141499,1.66187980931003e-005,-4.28528869302687e-006,.000146461329045194,2.74870868673343e-007},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.92171958555151,-.49553061346938,2.00568935114481,1.09346229146626,2.19799643560933,.566770786786546,-2.29403815127598,-2.29634628619806,-.592131075007482,2.23836106649518,3.41890639763464,.881592088933561,-3.56829591856781,-1.94536458449534,-3.91413586924262,-1.00929093051433,4.08516450282659,4.09536544265748,1.05602241045943,-4.00097073404889,5.33523806427656,1.37573338488067,-5.56836192711792,-5.59797201662203,-1.44348141522225,5.49216459339238,7.19422103343416,1.855086864706,-7.09892453011936,9.00134910987289},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.59519010006696,-.411332399809833,1.66489212099991,.907666360512694,1.2388862761662,.31945663701093,-1.29301955934683,-.514957807193037,-.132785948612982,-.353775594121333,3.23038451124486,.832980227624035,-3.37153654597113,-1.83809525374131,-2.62173967966463,-.676036338352017,2.73629690762123,1.30644820527577,.336878015452627,.307822029727901,4.7630793830775,1.22819773798136,-4.97120270417875,-2.90243304119459,-.748415344244522,.518849452524734,6.06854814809742,1.56482319698726,-2.66498561200528,7.17423567654379},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.03822870489707,-.26771549341915,1.08359423150038,.590754211622914,-.0622566343920988,-.0160533661863334,.0649769454361843,1.10967175893133,.286137650697852,-1.41978880536247,2.90882269973526,.750062968079831,-3.03592405293218,-1.65512593926411,-.707254507004454,-.182371175485818,.738158077031133,-1.57948219926848,-.407281992951262,2.54719860772099,3.86201434387853,.995850982046505,-4.03076552074204,.544351948717181,.140365460738301,-3.18953813043078,4.43071582310306,1.14249516194806,2.19278566166382,4.71658607960972},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-.300186182387741,-.0774053843401605,.313302853289658,.170806538752117,-1.13611369814234,-.292955913914665,1.18575632130325,1.22655366483195,.316276575739705,.100858250439743,2.48271365058504,.640187375400919,-2.59119611829218,-1.41266903727834,1.26648273457434,.326572602577083,-1.32182184869504,-2.5074905585056,-.646576297249579,.461151554581851,2.81340752869054,.725459410782454,-2.93633970583683,3.19379348794833,.823544942743958,-2.33714259619681,2.75487590413567,.71036656781973,4.77320941254594,2.50302559510188},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,.553376983055106,.142692637341486,-.577556856071558,-.314872611220468,-1.45368745167752,-.374844821120283,1.51720649776892,-.194424990185747,-.0501340234001082,1.49978166741586,1.9899087271236,.513113723366546,-2.07685802519487,-1.13226204929233,2.74863027720701,.708756083794028,-2.86873200495547,-.854309601049289,-.220290495881434,-2.52312426315122,1.80736427869175,.466043191865359,-1.88633727617264,4.12997714955087,1.06494731359279,1.65834138994758,1.41847476579741,.365764951301697,4.48820932686104,1.03297875354961},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,1.44662301694489,.373023200905715,-1.50983338153088,-.823131392784565,-.724355119014703,-.186780703583358,.756005902089247,-1.6918152058931,-.436247948578379,.160475142667942,1.4741928888764,.380132310523241,-1.53860792217029,-.838818704928317,3.38025060487184,.871624387178691,-3.52795105816103,2.16048862424162,.557099100602201,-1.70770592621981,.991945941760336,.255781116350547,-1.03528913786275,3.46032388717692,.892271916858127,4.72311662432345,.576746649127291,.148718690750111,2.68886700736676,.311154445333738},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,2.30018618238774,.593121222587361,-2.40069309089209,-1.30881054275715,.987081447023724,.254526767782017,-1.03021208824726,-1.35561938023839,-.349557192548049,-2.20870753342451,.981387965414959,.253058658488868,-1.02426982907298,-.558411716942308,3.10522366683209,.800706513225493,-3.2409067852655,4.42237267757001,1.14034381554164,2.83495589139377,.43960319187862,.113355164264732,-.458811705720641,2.03901059601618,.525775029251871,4.55407145300732,.170154848292172,.0438757750930783,1.03685515398485,.0611113213546687},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.03822870489707,.78343133166635,-3.17098446910282,-1.72875821562795,3.26615690329693,.842204422108959,-3.40887201763794,1.66498726415444,.429330151346565,-.923099517802526,.555278916264744,.143183065809956,-.579541894432988,-.315954814956541,2.17523616991868,.560901872435633,-2.27028337376948,4.50024407433993,1.16042357185991,6.26847807110811,.140734880491405,.0362896034209995,-.14688430786117,.792696592497175,.204403093795768,2.35258131414952,.0308216365902796,.00794760307219474,.225963663364161,.00626330408481313},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.59519010006696,.927048238057034,-3.75228235860234,-2.04567036451773,5.47681396366563,1.4122398512488,-5.7161240011779,6.3309248170395,1.63247909961861,5.76936797267538,.233717104755142,.060265806265752,-.243929401394036,-.132985500479334,1.04841335581819,.27034168634063,-1.09422390242844,2.66785366107377,.68792719315715,5.04596922575242,.0249321870529789,.0064289618707996,-.0260216019365979,.15913874220381,.0410351849082053,.559597243722218,.00229823078236368,.000592617006985495,.0189957623925199,.000196571591117826},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.92171958555151,1.01124645171658,-4.09307958874725,-2.23146629547129,6.96914454009643,1.7970491081004,-7.2736621396321,10.2353232484353,2.63925915467191,13.4470270492568,.0451952183653594,.0116539449562267,-.0471700287973566,-.025716169725305,.217799593005347,.0561613498441344,-.227316372195247,.614368644454251,.158419820138511,1.33333501160483,.000932318622845177,.000240405739976666,-.000973056396143785,.00636097475878139,.00164022771456071,.0244039729141499,1.66187980931003e-005,4.28528869302687e-006,.000146461329045194,2.74870868673343e-007},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.92171958555151,-1.44256348674145,.937852501329483,2.27058656100761,2.19799643560933,1.64995425135427,-1.07268327325536,-2.29634628619806,-1.72378183093533,2.23836106649518,3.41890639763464,2.56644599370298,-1.66852122478611,-4.03957110939074,-3.91413586924262,-2.93819635817362,1.91020695459718,4.09536544265748,3.07423866276144,-4.00097073404889,5.33523806427656,4.00496497002541,-2.60374427309723,-5.59797201662203,-4.20218958584716,5.49216459339238,7.19422103343416,5.40043440913241,-7.09892453011936,9.00134910987289},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.59519010006696,-1.19744993497978,.778497048524628,1.88477925223671,1.2388862761662,.929985893147364,-.604610892164934,-.514957807193037,-.386559691398865,-.353775594121333,3.23038451124486,2.42492961864658,-1.57651731119676,-3.81682515580895,-2.62173967966463,-1.9680425606586,1.27948173914766,1.30644820527577,.980702122872796,.307822029727901,4.7630793830775,3.57546670811254,-2.32451495445633,-2.90243304119459,-2.17874863666902,.518849452524734,6.06854814809742,4.55543360151919,-2.66498561200528,7.17423567654379},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.03822870489707,-.779359710871422,.506684427405857,1.22670766448742,-.0622566343920988,-.0467337421144522,.030382965722629,1.10967175893133,.832989356981887,-1.41978880536247,2.90882269973526,2.18354511527215,-1.41958622119803,-3.43688734746291,-.707254507004454,-.530909678713955,.34515983171824,-1.57948219926848,-1.18565859680248,2.54719860772099,3.86201434387853,2.89907066463143,-1.88477020319366,.544351948717181,.408624781106939,-3.18953813043078,4.43071582310306,3.32597373321893,2.19278566166382,4.71658607960972},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-.300186182387741,-.225338612975934,.146499189649511,.3546816698203,-1.13611369814234,-.852838337634571,.554455021295673,1.22655366483195,.920728260721081,.100858250439743,2.48271365058504,1.86368088534508,-1.21163317718593,-2.93342290468436,1.26648273457434,.950701529718779,-.618078730089992,-2.5074905585056,-1.88227998927126,.461151554581851,2.81340752869054,2.11192049221879,-1.3730209697231,3.19379348794833,2.39746209780118,-2.33714259619681,2.75487590413567,2.06798297639769,4.77320941254594,2.50302559510188},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,.553376983055106,.41539953912128,-.270063328509737,-.653836465186081,-1.45368745167752,-1.09122915405036,.709439828333211,-.194424990185747,-.145947615817496,1.49978166741586,1.9899087271236,1.49375054084377,-.971130695151501,-2.35115468793598,2.74863027720701,2.06329461731692,-1.34140787303665,-.854309601049289,-.641298473185774,-2.52312426315122,1.80736427869175,1.35672120663226,-.882043937561808,4.12997714955087,3.10022038625728,1.65834138994758,1.41847476579741,1.06479629943225,4.48820932686104,1.03297875354961},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,1.44662301694489,1.0859261460488,-.705992187998234,-1.70924145531706,-.724355119014703,-.543746472353678,.353505404956932,-1.6918152058931,-1.26998281171566,.160475142667942,1.4741928888764,1.10662182393172,-.719447050735515,-1.74181633275904,3.38025060487184,2.53742852796533,-1.64965612574837,2.16048862424162,1.62179853107076,-1.70770592621981,.991945941760336,.744616960114543,-.484097154422054,3.46032388717692,2.5975365648804,4.72311662432345,.576746649127291,.432942278924123,2.68886700736676,.311154445333738},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,2.30018618238774,1.72666429814602,-1.12255470615748,-2.71775959032344,.987081447023724,.740965364440283,-.481723145865575,-1.35561938023839,-1.01761309753837,-2.20870753342451,.981387965414959,.736691479430405,-.478944568701089,-1.15954811601066,3.10522366683209,2.33097602484761,-1.51543534565911,4.42237267757001,3.31971084311917,2.83495589139377,.43960319187862,.329993781528465,-.214538560322786,2.03901059601618,1.53060948975261,4.55407145300732,.170154848292172,.12772892205108,1.03685515398485,.0611113213546687},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.03822870489707,2.2806853960415,-1.48273994391383,-3.58978558499056,3.26615690329693,2.45178262388351,-1.59397523182193,1.66498726415444,1.24984407270733,-.923099517802526,.555278916264744,.41682724950333,-.270991524688991,-.656083673232099,2.17523616991868,1.63286896677384,-1.06157563213921,4.50024407433993,3.3781660071394,6.26847807110811,.140734880491405,.10564444540962,-.0686825282564418,.792696592497175,.595047877309364,2.35258131414952,.0308216365902796,.0231366573273696,.225963663364161,.00626330408481313},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.59519010006696,2.69877562014986,-1.7545525650326,-4.24785717273985,5.47681396366563,4.11124073580283,-2.67283724139748,6.3309248170395,4.75239001984257,5.76936797267538,.233717104755142,.175442746128902,-.114060434690259,-.276145864886065,1.04841335581819,.787004949964004,-.511654819985593,2.66785366107377,2.00265861161846,5.04596922575242,.0249321870529789,.0187156663995726,-.0121675993597458,.15913874220381,.119459540563619,.559597243722218,.00229823078236368,.00172519645136334,.0189957623925199,.000196571591117826},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.92171958555151,2.94388917191153,-1.91390801783745,-4.63366448151075,6.96914454009643,5.23147784770941,-3.40113598727832,10.2353232484353,7.68327684524188,13.4470270492568,.0451952183653594,.0339263710725032,-.0220565211009043,-.0533999113042759,.217799593005347,.163494061616168,-.10629224716917,.614368644454251,.461183712763862,1.33333501160483,.000932318622845177,.000699856947415067,-.000454997367631839,.00636097475878139,.00477494739305922,.0244039729141499,1.66187980931003e-005,1.24751142122696e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.92171958555151,-2.26142559207414,-.826744191593013,1.19523987507356,2.19799643560933,2.5865404220042,.94560142914338,-2.29634628619806,-2.70227576356801,2.23836106649518,3.41890639763464,4.02327294720743,1.47084987158013,-2.12643576380383,-3.91413586924262,-4.60604507183384,-1.68390285488931,4.09536544265748,4.81931093640565,-4.00097073404889,5.33523806427656,6.27835818791885,2.29527612404421,-5.59797201662203,-6.58753611853026,5.49216459339238,7.19422103343416,8.46595709192435,-7.09892453011936,9.00134910987289},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.59519010006696,-1.87717487173315,-.686267736267333,.992150379408985,1.2388862761662,1.4578865459724,.532982044151546,-.514957807193037,-.605987872997064,-.353775594121333,3.23038451124486,3.80142569043747,1.38974575226926,-2.00918198880831,-2.62173967966463,-3.08518956279543,-1.12790027705863,1.30644820527577,1.53739152505215,.307822029727901,4.7630793830775,5.60505793446857,2.04912737785909,-2.90243304119459,-3.41550161982096,.518849452524734,6.06854814809742,7.14129688166466,-2.66498561200528,7.17423567654379},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.03822870489707,-1.22175835711557,-.446657024142497,.645740594449334,-.0622566343920988,-.0732618573826121,-.0267834658423845,1.10967175893133,1.30583053509787,-1.41978880536247,2.90882269973526,3.42302078938587,1.25140644309976,-1.80918220620513,-.707254507004454,-.832277224231996,-.304268406428257,-1.57948219926848,-1.8586902544778,2.54719860772099,3.86201434387853,4.54470992402429,1.661479620483,.544351948717181,.64057807233031,-3.18953813043078,4.43071582310306,5.21394183895337,2.19278566166382,4.71658607960972},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-.300186182387741,-.353250661721207,-.129143286331404,.186704916697283,-1.13611369814234,-1.33694666498547,-.488768188714077,1.22655366483195,1.44337387662839,.100858250439743,2.48271365058504,2.92158763778148,1.06808980107197,-1.54415783407832,1.26648273457434,1.49036128356923,.54485433405009,-2.5074905585056,-2.95074440626681,.461151554581851,2.81340752869054,3.31073889879025,1.21035782283977,3.19379348794833,3.75836640385388,-2.33714259619681,2.75487590413567,3.24186052799879,4.77320941254594,2.50302559510188},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,.553376983055106,.651198479192506,.23806865993447,-.344180410642768,-1.45368745167752,-1.71065853147384,-.625391792982136,-.194424990185747,-.228793863525153,1.49978166741586,1.9899087271236,2.34166865603201,.856079885009684,-1.23765104741928,2.74863027720701,3.23451085132423,1.18248996086068,-.854309601049289,-1.00532752474992,-2.52312426315122,1.80736427869175,2.12685548066823,.777547323353426,4.12997714955087,4.86004102210358,1.65834138994758,1.41847476579741,1.66922123290476,4.48820932686104,1.03297875354961},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,1.44662301694489,1.70234530427801,.622352597994008,-.899747042727636,-.724355119014703,-.852400743178548,-.31162527138391,-1.6918152058931,-1.9908805805112,.160475142667942,1.4741928888764,1.73478875376213,.634213449888053,-.916894502821386,3.38025060487184,3.97778389924568,1.45421973940992,2.16048862424162,2.54240228389271,-1.70770592621981,.991945941760336,1.16729410204273,.426745687640421,3.46032388717692,4.07201188589473,4.72311662432345,.576746649127291,.67869924509284,2.68886700736676,.311154445333738},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,2.30018618238774,2.70679444519172,.989564544259882,-1.43063237006769,.987081447023724,1.16156970101249,.424652930216328,-1.35561938023839,-1.5952547827211,-2.20870753342451,.981387965414959,1.15486977201266,.422203533825771,-.610387716162348,3.10522366683209,3.65413992905702,1.33590022736237,4.42237267757001,5.20412385861342,2.83495589139377,.43960319187862,.517312679568431,.189121965733584,2.03901059601618,2.39945035585008,4.55407145300732,.170154848292172,.200233442638166,1.03685515398485,.0611113213546687},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.03822870489707,3.57530214058608,1.30707828207097,-1.88966804781974,3.26615690329693,3.843521635486,1.4051354158397,1.66498726415444,1.95931021275678,-.923099517802526,.555278916264744,.653436620408273,.238886891797976,-.345363344035536,2.17523616991868,2.55975678291091,.935809721212699,4.50024407433993,5.29576072944733,6.26847807110811,.140734880491405,.165612851500527,.0605456414728775,.792696592497175,.932823068533675,2.35258131414952,.0308216365902796,.036270035583218,.225963663364161,.00626330408481313},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.59519010006696,4.23071865520366,1.54668899419581,-2.23607783277939,5.47681396366563,6.44496072482988,2.35618358032457,6.3309248170395,7.45005437415178,5.76936797267538,.233717104755142,.275031719356679,.100547582628472,-.145363561432351,1.04841335581819,1.23374336817428,.451038569416842,2.66785366107377,3.13945519772704,5.04596922575242,.0249321870529789,.0293394969148429,.0107260918769641,.15913874220381,.187269998673138,.559597243722218,.00229823078236368,.00270449338439953,.0189957623925199,.000196571591117826},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.92171958555151,4.61496937554466,1.68716544952149,-2.43916732844396,6.96914454009643,8.20109339929485,2.99820005631404,10.2353232484353,12.0446407028288,13.4470270492568,.0451952183653594,.0531844625867158,.0194434633176023,-.0281097864368307,.217799593005347,.256300439106579,.0936996999356592,.614368644454251,.722971751522571,1.33333501160483,.000932318622845177,.00109712634918437,.000401093381258233,.00636097475878139,.00748541629863617,.0244039729141499,1.66187980931003e-005,1.95565344631038e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.92171958555151,-2.87937722321757,-2.67495037465683,-1.63105841067899,2.19799643560933,3.29333213710544,3.05951577632408,-2.29634628619806,-3.44069303528758,2.23836106649518,3.41890639763464,5.1226626812653,4.75896952776425,2.90179486950885,-3.91413586924262,-5.86468169257239,-5.4483074914535,4.09536544265748,6.13622406614863,-4.00097073404889,5.33523806427656,7.99396700420041,7.42642015557747,-5.59797201662203,-8.38763014060611,5.49216459339238,7.19422103343416,10.7793438389747,-7.09892453011936,9.00134910987289},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.59519010006696,-2.39012709001286,-2.22043548283783,-1.35391669466665,1.2388862761662,1.85626506095126,1.72447600259692,-.514957807193037,-.771578638346196,-.353775594121333,3.23038451124486,4.84019398522881,4.49655523257725,2.74178702516,-2.62173967966463,-3.92824092358218,-3.64934800797338,1.30644820527577,1.95749537633152,.307822029727901,4.7630793830775,7.13668236880667,6.6300000687783,-2.90243304119459,-4.3488132461414,.518849452524734,6.06854814809742,9.09271021667368,-2.66498561200528,7.17423567654379},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.03822870489707,-1.55561306022354,-1.44516935978819,-.881196025717112,-.0622566343920988,-.0932812135041899,-.0866585368463816,1.10967175893133,1.66265859611778,-1.41978880536247,2.90882269973526,4.35838708560735,4.04895512766487,2.46886161968119,-.707254507004454,-1.0597032647262,-.984467621360889,-1.57948219926848,-2.36659141141637,2.54719860772099,3.86201434387853,5.78658625215441,5.37575658648661,.544351948717181,.815620871022689,-3.18953813043078,4.43071582310306,6.63869084478427,2.19278566166382,4.71658607960972},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-.300186182387741,-.449779074319961,-.417846155642134,-.254782852417372,-1.13611369814234,-1.70227744362135,-1.58142102819233,1.22655366483195,1.83778669504629,.100858250439743,2.48271365058504,3.7199335363258,3.45582979910559,2.10720180544034,1.26648273457434,1.89761376598671,1.76288907720341,-2.5074905585056,-3.75705761211831,.461151554581851,2.81340752869054,4.21542331996591,3.91614134649752,3.19379348794833,4.78536842321548,-2.33714259619681,2.75487590413567,4.12772341388704,4.77320941254594,2.50302559510188},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,.553376983055106,.829143384311425,.770276776736346,.469678401195648,-1.45368745167752,-2.17810890152299,-2.02346992933989,-.194424990185747,-.291313515578485,1.49978166741586,1.9899087271236,2.9815473107461,2.76986670414993,1.68893390563517,2.74863027720701,4.11836538252447,3.82597422031764,-.854309601049289,-1.2800408674882,-2.52312426315122,1.80736427869175,2.70803481201197,2.51577274454146,4.12997714955087,6.18808395553405,1.65834138994758,1.41847476579741,2.12534854800543,4.48820932686104,1.03297875354961},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,1.44662301694489,2.16752402217829,2.01363654211465,1.2278204669453,-.724355119014703,-1.08532568728512,-1.00827093182807,-1.6918152058931,-2.53490374291394,.160475142667942,1.4741928888764,2.20883288938782,2.05201260878721,1.25122037987573,3.38025060487184,5.06474340721712,4.70516234202085,2.16048862424162,3.23713291923203,-1.70770592621981,.991945941760336,1.48626603590141,1.38074575986725,3.46032388717692,5.18471990323708,4.72311662432345,.576746649127291,.864158942299266,2.68886700736676,.311154445333738},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,2.30018618238774,3.44644648080968,3.20175947449313,1.95228172055832,.987081447023724,1.47897739903412,1.37397459375255,-1.35561938023839,-2.03117020639316,-2.20870753342451,.981387965414959,1.47044666380812,1.36604951383154,.832952480070553,3.10522366683209,4.65266127660931,4.32233676392885,4.42237267757001,6.6261900271284,2.83495589139377,.43960319187862,.658672278252944,.611908590636855,2.03901059601618,3.05511829665785,4.55407145300732,.170154848292172,.254948743871771,1.03685515398485,.0611113213546687},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.03822870489707,4.55228046671326,4.22908267863918,2.57869489385806,3.26615690329693,4.89379296535329,4.54634885284358,1.66498726415444,2.49470653361407,-.923099517802526,.555278916264744,.831993114526565,.772924185272261,.471292665829706,2.17523616991868,3.25922966624351,3.02783447385586,4.50024407433993,6.74286736535737,6.26847807110811,.140734880491405,.210867814600684,.195896854176436,.792696592497175,1.18772402074228,2.35258131414952,.0308216365902796,.0461810968751895,.225963663364161,.00626330408481313},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.59519010006696,5.38679449650258,5.00434880168882,3.0514155628076,5.47681396366563,8.20609494322841,7.62348767011616,6.3309248170395,9.48583802706066,5.76936797267538,.233717104755142,.350186214905109,.325324080359881,.198367260350895,1.04841335581819,1.57087306611691,1.45934595309525,2.66785366107377,3.99733505330865,5.04596922575242,.0249321870529789,.0373567361461482,.0347045238135923,.15913874220381,.23844294088706,.559597243722218,.00229823078236368,.00344351663776842,.0189957623925199,.000196571591117826},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.92171958555151,5.87604462970729,5.45886369350783,3.32855727881994,6.96914454009643,10.4421041409331,9.70074715430413,10.2353232484353,15.3359297883224,13.4470270492568,.0451952183653594,.067717518868618,.0629097851728847,.0383594160020484,.217799593005347,.326336470786676,.303167594035556,.614368644454251,.920529245649199,1.33333501160483,.000932318622845177,.00139692441436285,.00129774711619293,.00636097475878139,.00953086286461726,.0244039729141499,1.66187980931003e-005,2.49005052810442e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.92171958555151,-3.24166342459793,-3.96510583993544,-4.31411318985288,2.19799643560933,3.70770257117539,4.53515100205683,-2.29634628619806,-3.87360456901059,2.23836106649518,3.41890639763464,5.76720136441661,7.05426838770182,7.67518283761778,-3.91413586924262,-6.60258197030424,-8.07608098333328,4.09536544265748,6.90829008438974,-4.00097073404889,5.33523806427656,8.99977614812198,11.0082572774727,-5.59797201662203,-9.44297038492047,5.49216459339238,7.19422103343416,12.1356119587979,-7.09892453011936,9.00134910987289},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.59519010006696,-2.69085533682779,-3.29137384514269,-3.58107952001051,1.2388862761662,2.0898222386161,2.55620812016103,-.514957807193037,-.868659455579414,-.353775594121333,3.23038451124486,5.44919216674985,6.66528903908045,7.25196565099482,-2.62173967966463,-4.42249623366662,-5.4094652504454,1.30644820527577,2.20378945631257,.307822029727901,4.7630793830775,8.03462707886587,9.82771577392637,-2.90243304119459,-4.89598539807522,.518849452524734,6.06854814809742,10.2367643607743,-2.66498561200528,7.17423567654379},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.03822870489707,-1.75134189417476,-2.14218907478874,-2.33074387312075,-.0622566343920988,-.105017951653011,-.12845482061473,1.10967175893133,1.87185600940662,-1.41978880536247,2.90882269973526,4.90676382786187,6.00180690245521,6.53008402866105,-.707254507004454,-1.19303621855815,-1.45928625505293,-1.57948219926848,-2.66435837496252,2.54719860772099,3.86201434387853,6.51466048131366,7.96853804732135,.544351948717181,.918243127233759,-3.18953813043078,4.43071582310306,7.47397774950178,2.19278566166382,4.71658607960972},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-.300186182387741,-.506370739691845,-.619377558413115,-.673894973328732,-1.13611369814234,-1.91645974101964,-2.34415629302528,1.22655366483195,2.06901890572263,.100858250439743,2.48271365058504,4.18797939686675,5.12261126340126,5.57350187032702,1.26648273457434,2.13637348021345,2.61314820693614,-2.5074905585056,-4.22977446202813,.461151554581851,2.81340752869054,4.74581221430261,5.80493569904152,3.19379348794833,5.38746839617942,-2.33714259619681,2.75487590413567,4.64707781591112,4.77320941254594,2.50302559510188},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,.553376983055106,.93346705704166,1.14178901214037,1.24228891639978,-1.45368745167752,-2.45216080196962,-2.99940982448642,-.194424990185747,-.327966881493408,1.49978166741586,1.9899087271236,3.35668865754018,4.10580126963162,4.46719258573777,2.74863027720701,4.6365423474252,5.67128006117502,-.854309601049289,-1.44109692494208,-2.52312426315122,1.80736427869175,3.04876253512471,3.72915523810848,4.12997714955087,6.96667503835439,1.65834138994758,1.41847476579741,2.39276208655032,4.48820932686104,1.03297875354961},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,1.44662301694489,2.44024412222762,2.98483369571688,3.24755780451479,-.724355119014703,-1.22188248065577,-1.49457014152702,-1.6918152058931,-2.85384793698481,.160475142667942,1.4741928888764,2.48675051356285,3.04171892525937,3.30945005334757,3.38025060487184,5.70199462778367,6.97450945518368,2.16048862424162,3.64443230995284,-1.70770592621981,.991945941760336,1.67326955598402,2.04669332477546,3.46032388717692,5.8370666898335,4.72311662432345,.576746649127291,.972888308521253,2.68886700736676,.311154445333738},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,2.30018618238774,3.88008191896113,4.74600026627036,5.1637416942433,.987081447023724,1.66506385533551,2.03665635715207,-1.35561938023839,-2.28673412920868,-2.20870753342451,.981387965414959,1.65545977423627,2.02490893148973,2.20314076875832,3.10522366683209,5.23806390000324,6.40704325104372,4.42237267757001,7.45990406612127,2.83495589139377,.43960319187862,.741547101224612,.907038257317998,2.03901059601618,3.43951642050873,4.55407145300732,.170154848292172,.287026656838237,1.03685515398485,.0611113213546687},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.03822870489707,5.12505307344404,6.26881178264598,6.82059059403532,3.26615690329693,5.50953502563763,6.7390986231311,1.66498726415444,2.80859307349989,-.923099517802526,.555278916264744,.936675343241156,1.14571329243577,1.24655861042429,2.17523616991868,3.66930929238228,4.48818948881768,4.50024407433993,7.59126186695628,6.26847807110811,.140734880491405,.237399419743997,.29037987690505,.792696592497175,1.33716467766395,2.35258131414952,.0308216365902796,.0519916499446463,.225963663364161,.00626330408481313},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.59519010006696,6.06456651609707,7.41799655299993,8.07092624092508,5.47681396366563,9.23859424244377,11.300372436005,6.3309248170395,10.6793559162843,5.76936797267538,.233717104755142,.394247004353173,.482231155810537,.524676988090528,1.04841335581819,1.76852192969244,2.1632031816091,2.66785366107377,4.50028385780377,5.04596922575242,.0249321870529789,.0420569990798228,.0514428646412205,.15913874220381,.268444076758866,.559597243722218,.00229823078236368,.0038767834400574,.0189957623925199,.000196571591117826},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.92171958555151,6.61537460386721,8.09172854779269,8.80395991076745,6.96914454009643,11.7559404153652,14.3795150585562,10.2353232484353,17.2655122351262,13.4470270492568,.0451952183653594,.0762378066864121,.0932518071891671,.101459801467564,.217799593005347,.367396460919218,.449388373324038,.614368644454251,1.03635118137081,1.33333501160483,.000932318622845177,.0015726868798065,.00192366360061394,.00636097475878139,.0107300458251746,.0244039729141499,1.66187980931003e-005,2.80335124459201e-005,.000146461329045194,2.74870868673343e-007}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_left_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-3.37371117926928,5.6909635577302,-6.96101658051265,7.57372303615021,4.13193546224797,-6.96997843122346,8.52546935665732,-4.77114810448709,8.04823784418982,5.33430574565371,5.05405999869821,-8.52546935665732,10.4280993821027,-5.83592580124217,9.84436405649934,6.52476340023727,-6.34960212623238,10.7108618346039,7.09907099144837,7.15379270875817},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-2.99666740648972,5.05494516239668,-6.18305788326991,6.72728866290428,3.67015303548283,-6.19101816335907,7.5726684320545,-4.23792768749944,7.14877200382768,4.73814719602452,3.40958356039142,-5.75147508548906,7.03503247458549,-3.93704797266052,6.64123137798004,4.40175344784644,-2.40062592164153,4.04950925303536,2.68398137439781,.178027784911327},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-2.35354378347051,3.97008848462071,-4.85609361001944,5.28352541823675,2.88249067692314,-4.86234551094043,5.94747574387061,-3.32841353768903,5.61455297729064,3.72127946338712,1.05379652310981,-1.77760255482498,2.17431033156193,-1.21681941252376,2.05259862717067,1.36044546108577,1.7591790753139,-2.96748105525688,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-1.50132568517008,2.53251962263048,-3.09770233191153,3.3703611015687,1.83874093225775,-3.10169041978979,3.79389504410759,-2.12319514516607,3.58152359635514,2.37380433681192,-1.19541898829253,2.01649920176636,-2.46652157203873,1.38035094930968,-2.32845271402296,-1.54327927740643,3.34189684440344,-5.63729736986577,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-.515715838247201,.869938143977334,-1.06408234437226,1.1577425323902,.631620327664824,-1.06545228035847,1.30322939400866,-.729332332662821,1.23027832131385,.81541833693141,-2.55652048864375,4.31248087502183,-5.27489775246548,2.95201558448519,-4.97962398835612,-3.30045375807884,1.60938069663531,-2.71479282295876,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,.515715838247201,-.869938143977334,1.06408234437226,-1.1577425323902,-.631620327664824,1.06545228035847,-1.30322939400866,.729332332662821,-1.23027832131385,-.81541833693141,-2.55652048864375,4.31248087502184,-5.27489775246549,2.95201558448519,-4.97962398835612,-3.30045375807885,-1.60938069663533,2.71479282295879,1.79934231949429,2.78946777547417},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,1.50132568517008,-2.53251962263048,3.09770233191153,-3.3703611015687,-1.83874093225775,3.10169041978979,-3.79389504410759,2.12319514516607,-3.58152359635514,-2.37380433681192,-1.19541898829252,2.01649920176635,-2.46652157203872,1.38035094930968,-2.32845271402296,-1.54327927740643,-3.34189684440345,5.63729736986578,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,2.35354378347051,-3.97008848462071,4.85609361001944,-5.28352541823675,-2.88249067692314,4.86234551094043,-5.94747574387061,3.32841353768903,-5.61455297729064,-3.72127946338712,1.0537965231098,-1.77760255482497,2.17431033156192,-1.21681941252376,2.05259862717066,1.36044546108577,-1.75917907531388,2.96748105525684,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,2.99666740648972,-5.05494516239668,6.18305788326991,-6.72728866290428,-3.67015303548283,6.19101816335907,-7.5726684320545,4.23792768749944,-7.14877200382768,-4.73814719602452,3.40958356039142,-5.75147508548906,7.03503247458549,-3.93704797266052,6.64123137798004,4.40175344784644,2.40062592164155,-4.0495092530354,-2.68398137439783,.178027784911158},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,3.37371117926928,-5.6909635577302,6.96101658051265,-7.57372303615021,-4.13193546224797,6.96997843122346,-8.52546935665732,4.77114810448709,-8.04823784418982,-5.33430574565371,5.05405999869821,-8.52546935665732,10.4280993821027,-5.83592580124217,9.84436405649934,6.52476340023727,6.34960212623239,-10.7108618346039,-7.09907099144838,7.15379270875832},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-3.37371117926928,5.05494516239668,-4.69605974259137,2.86343545350678,4.13193546224797,-6.19101816335907,5.75147508548906,-4.77114810448709,7.14877200382768,5.33430574565371,5.05405999869821,-7.5726684320545,7.03503247458549,-5.83592580124217,8.74416431584401,6.52476340023727,-6.34960212623238,9.51382286438758,7.09907099144837,7.15379270875817},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-2.99666740648972,4.4900077703237,-4.17123115221737,2.54341976480909,3.67015303548283,-5.4991139896991,5.10869396106863,-4.23792768749944,6.34982988227715,4.73814719602452,3.40958356039142,-5.10869396106863,4.74599254428019,-3.93704797266052,5.89901166751924,4.40175344784644,-2.40062592164153,3.59693872593977,2.68398137439781,.178027784911327},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-2.35354378347051,3.52639397108079,-3.27603094239261,1.99756895385147,2.88249067692314,-4.31893293096971,4.01230209521675,-3.32841353768903,4.98707417885748,3.72127946338712,1.05379652310981,-1.57893884647548,1.46683908849366,-1.21681941252376,1.82320153617303,1.36044546108577,1.7591790753139,-2.63583729760472,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-1.50132568517008,2.24948687251768,-2.08978028527405,1.27424932536978,1.83874093225775,-2.7550475106226,2.55944768672544,-2.12319514516607,3.18125484265547,2.37380433681192,-1.19541898829253,1.79113655913325,-1.66397142209237,1.38035094930968,-2.06822634923255,-1.54327927740643,3.34189684440344,-5.0072766728733,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-.515715838247201,.772714421358283,-.717854095362718,.437713525759442,.631620327664824,-.946378024692648,.879188121703209,-.729332332662821,1.09278321423787,.81541833693141,-2.55652048864375,3.83052081004967,-3.55856572026934,2.95201558448519,-4.42310444183107,-3.30045375807884,1.60938069663531,-2.41138933822272,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,.515715838247201,-.772714421358283,.717854095362718,-.437713525759442,-.631620327664824,.946378024692648,-.879188121703209,.729332332662821,-1.09278321423787,-.81541833693141,-2.55652048864375,3.83052081004967,-3.55856572026934,2.95201558448519,-4.42310444183107,-3.30045375807885,-1.60938069663533,2.41138933822274,1.79934231949429,2.78946777547417},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,1.50132568517008,-2.24948687251768,2.08978028527405,-1.27424932536978,-1.83874093225775,2.7550475106226,-2.55944768672544,2.12319514516607,-3.18125484265547,-2.37380433681192,-1.19541898829252,1.79113655913325,-1.66397142209237,1.38035094930968,-2.06822634923254,-1.54327927740643,-3.34189684440345,5.00727667287331,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,2.35354378347051,-3.52639397108079,3.27603094239261,-1.99756895385147,-2.88249067692314,4.31893293096971,-4.01230209521675,3.32841353768903,-4.98707417885748,-3.72127946338712,1.0537965231098,-1.57893884647547,1.46683908849365,-1.21681941252376,1.82320153617302,1.36044546108577,-1.75917907531388,2.6358372976047,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,2.99666740648972,-4.4900077703237,4.17123115221737,-2.54341976480909,-3.67015303548283,5.4991139896991,-5.10869396106863,4.23792768749944,-6.34982988227715,-4.73814719602452,3.40958356039142,-5.10869396106863,4.74599254428019,-3.93704797266052,5.89901166751924,4.40175344784644,2.40062592164155,-3.59693872593979,-2.68398137439783,.178027784911158},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,3.37371117926928,-5.05494516239668,4.69605974259137,-2.86343545350678,-4.13193546224797,6.19101816335907,-5.75147508548906,4.77114810448709,-7.14877200382768,-5.33430574565371,5.05405999869821,-7.5726684320545,7.03503247458549,-5.83592580124217,8.74416431584401,6.52476340023727,6.34960212623239,-9.51382286438759,-7.09907099144838,7.15379270875832},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-3.37371117926928,3.97008848462071,-1.45140640826254,-2.09832597736703,4.13193546224797,-4.86234551094043,1.77760255482498,-4.77114810448709,5.61455297729064,5.33430574565371,5.05405999869821,-5.94747574387061,2.17431033156193,-5.83592580124217,6.86755344374583,6.52476340023727,-6.34960212623238,7.47203330307115,7.09907099144837,7.15379270875817},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-2.99666740648972,3.52639397108079,-1.2891981696408,-1.86381842737598,3.67015303548283,-4.31893293096971,1.57893884647548,-4.23792768749944,4.98707417885748,4.73814719602452,3.40958356039142,-4.01230209521675,1.46683908849366,-3.93704797266052,4.63300738968997,4.40175344784644,-2.40062592164153,2.82498910610717,2.68398137439781,.178027784911327},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-2.35354378347051,2.7695841689776,-1.01251955130179,-1.46381886216961,2.88249067692314,-3.3920340071429,1.24007812764096,-3.32841353768903,3.91678349286862,3.72127946338712,1.05379652310981,-1.24007812764096,.453354465153136,-1.21681941252376,1.43191888168209,1.36044546108577,1.7591790753139,-2.07015248758758,-1.96682199829962,-4.02051566640726},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-1.50132568517008,1.76671786576882,-.645886267246198,-.933770117915978,1.83874093225775,-2.16377864548779,.791045893312282,-2.12319514516607,2.49851636599797,2.37380433681192,-1.19541898829253,1.40673546385746,-.514281954994446,1.38035094930968,-1.62435819754456,-1.54327927740643,3.34189684440344,-3.93265027010875,-3.73635425856115,-1.66034116472927},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-.515715838247201,.606879902269876,-.221866435121632,-.320756544597897,.631620327664824,-.743273047921452,.271729778549238,-.729332332662821,.858257788295293,.81541833693141,-2.55652048864375,3.00844145080061,-1.09984228773296,2.95201558448519,-3.47384896307412,-3.30045375807884,1.60938069663531,-1.89387396619673,-1.79934231949427,2.78946777547416},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,.515715838247201,-.606879902269876,.221866435121632,.320756544597897,-.631620327664824,.743273047921452,-.271729778549238,.729332332662821,-.858257788295293,-.81541833693141,-2.55652048864375,3.00844145080061,-1.09984228773296,2.95201558448519,-3.47384896307413,-3.30045375807885,-1.60938069663533,1.89387396619674,1.79934231949429,2.78946777547417},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,1.50132568517008,-1.76671786576882,.645886267246198,.933770117915978,-1.83874093225775,2.16377864548779,-.791045893312282,2.12319514516607,-2.49851636599797,-2.37380433681192,-1.19541898829252,1.40673546385745,-.514281954994445,1.38035094930968,-1.62435819754456,-1.54327927740643,-3.34189684440345,3.93265027010876,3.73635425856116,-1.66034116472922},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,2.35354378347051,-2.7695841689776,1.01251955130179,1.46381886216961,-2.88249067692314,3.3920340071429,-1.24007812764096,3.32841353768903,-3.91678349286862,-3.72127946338712,1.0537965231098,-1.24007812764096,.453354465153134,-1.21681941252376,1.43191888168208,1.36044546108577,-1.75917907531388,2.07015248758756,1.9668219982996,-4.02051566640715},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,2.99666740648972,-3.52639397108079,1.2891981696408,1.86381842737598,-3.67015303548283,4.31893293096971,-1.57893884647548,4.23792768749944,-4.98707417885748,-4.73814719602452,3.40958356039142,-4.01230209521675,1.46683908849366,-3.93704797266052,4.63300738968997,4.40175344784644,2.40062592164155,-2.82498910610719,-2.68398137439783,.178027784911158},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,3.37371117926928,-3.97008848462071,1.45140640826254,2.09832597736703,-4.13193546224797,4.86234551094043,-1.77760255482498,4.77114810448709,-5.61455297729064,-5.33430574565371,5.05405999869821,-5.94747574387061,2.17431033156193,-5.83592580124217,6.86755344374583,6.52476340023727,6.34960212623239,-7.47203330307115,-7.09907099144838,7.15379270875832},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-3.37371117926928,2.53251962263048,1.64646470368521,-3.98617119808645,4.13193546224797,-3.10169041978979,-2.01649920176636,-4.77114810448709,3.58152359635514,5.33430574565371,5.05405999869821,-3.79389504410759,-2.46652157203873,-5.83592580124217,4.38081265017735,6.52476340023727,-6.34960212623238,4.76641037958728,7.09907099144837,7.15379270875817},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-2.99666740648972,2.24948687251768,1.4624568765065,-3.54067929092287,3.67015303548283,-2.7550475106226,-1.79113655913325,-4.23792768749944,3.18125484265547,4.73814719602452,3.40958356039142,-2.55944768672544,-1.66397142209237,-3.93704797266052,2.95539562194488,4.40175344784644,-2.40062592164153,1.80206067765193,2.68398137439781,.178027784911327},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-2.35354378347051,1.76671786576882,1.14859469650904,-2.78080367423081,2.88249067692314,-2.16377864548779,-1.40673546385746,-3.32841353768903,2.49851636599797,3.72127946338712,1.05379652310981,-.791045893312282,-.514281954994446,-1.21681941252376,.913421118930409,1.36044546108577,1.7591790753139,-1.32055036479967,-1.96682199829962,-4.02051566640726},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-1.50132568517008,1.12698940591467,.732688608484839,-1.7738747886737,1.83874093225775,-1.38027449512883,-.897356615569124,-2.12319514516607,1.59380370207485,2.37380433681192,-1.19541898829253,.897356615569124,.58339764921819,1.38035094930968,-1.03617816716125,-1.54327927740643,3.34189684440344,-2.508637783912,-3.73635425856115,-1.66034116472927},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-.515715838247201,.387128716912051,.251683644415986,-.609338355176941,.631620327664824,-.47413391064841,-.308248252711719,-.729332332662821,.547482681696682,.81541833693141,-2.55652048864375,1.91908493657046,1.24765296340427,2.95201558448519,-2.21596840955051,-3.30045375807884,1.60938069663531,-1.20810228808803,-1.79934231949427,2.78946777547416},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,.515715838247201,-.387128716912051,-.251683644415986,.609338355176941,-.631620327664824,.47413391064841,.308248252711719,.729332332662821,-.547482681696682,-.81541833693141,-2.55652048864375,1.91908493657046,1.24765296340427,2.95201558448519,-2.21596840955051,-3.30045375807885,-1.60938069663533,1.20810228808804,1.79934231949429,2.78946777547417},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,1.50132568517008,-1.12698940591467,-.732688608484839,1.7738747886737,-1.83874093225775,1.38027449512883,.897356615569124,2.12319514516607,-1.59380370207485,-2.37380433681192,-1.19541898829252,.897356615569122,.583397649218189,1.38035094930968,-1.03617816716125,-1.54327927740643,-3.34189684440345,2.508637783912,3.73635425856116,-1.66034116472922},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,2.35354378347051,-1.76671786576882,-1.14859469650904,2.78080367423081,-2.88249067692314,2.16377864548779,1.40673546385746,3.32841353768903,-2.49851636599797,-3.72127946338712,1.0537965231098,-.791045893312278,-.514281954994444,-1.21681941252376,.913421118930405,1.36044546108577,-1.75917907531388,1.32055036479965,1.9668219982996,-4.02051566640715},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,2.99666740648972,-2.24948687251768,-1.4624568765065,3.54067929092287,-3.67015303548283,2.7550475106226,1.79113655913325,4.23792768749944,-3.18125484265547,-4.73814719602452,3.40958356039142,-2.55944768672544,-1.66397142209237,-3.93704797266052,2.95539562194488,4.40175344784644,2.40062592164155,-1.80206067765195,-2.68398137439783,.178027784911158},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,3.37371117926928,-2.53251962263048,-1.64646470368521,3.98617119808645,-4.13193546224797,3.10169041978979,2.01649920176636,4.77114810448709,-3.58152359635514,-5.33430574565371,5.05405999869821,-3.79389504410759,-2.46652157203873,-5.83592580124217,4.38081265017735,6.52476340023727,6.34960212623239,-4.76641037958728,-7.09907099144838,7.15379270875832},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-3.37371117926928,.869938143977334,3.52112588977045,-1.91964841477002,4.13193546224797,-1.06545228035847,-4.31248087502183,-4.77114810448709,1.23027832131385,5.33430574565371,5.05405999869821,-1.30322939400866,-5.27489775246548,-5.83592580124217,1.5048396829594,6.52476340023727,-6.34960212623238,1.63729519092357,7.09907099144837,7.15379270875817},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-2.99666740648972,.772714421358283,3.12760714457709,-1.70510975326197,3.67015303548283,-.946378024692648,-3.83052081004967,-4.23792768749944,1.09278321423787,4.73814719602452,3.40958356039142,-.879188121703209,-3.55856572026934,-3.93704797266052,1.01519899751182,4.40175344784644,-2.40062592164153,.619020404518225,2.68398137439781,.178027784911327},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-2.35354378347051,.606879902269876,2.45638215849919,-1.33917112430755,2.88249067692314,-.743273047921452,-3.00844145080061,-3.32841353768903,.858257788295293,3.72127946338712,1.05379652310981,-.271729778549238,-1.09984228773296,-1.21681941252376,.313766521598227,1.36044546108577,1.7591790753139,-.453618255557357,-1.96682199829962,-4.02051566640726},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-1.50132568517008,.387128716912051,1.5669262892192,-.854257320336024,1.83874093225775,-.47413391064841,-1.91908493657046,-2.12319514516607,.547482681696682,2.37380433681192,-1.19541898829253,.308248252711719,1.24765296340427,1.38035094930968,-.35593442337628,-1.54327927740643,3.34189684440344,-.861734565902807,-3.73635425856115,-1.66034116472927},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-.515715838247201,.132981412843306,.538250103024586,-.293443344364012,.631620327664824,-.162868303384663,-.659219053205559,-.729332332662821,.188064117536924,.81541833693141,-2.55652048864375,.659219053205559,2.66822795597052,2.95201558448519,-.761200595679671,-3.30045375807884,1.60938069663531,-.414991557357587,-1.79934231949427,2.78946777547416},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,.515715838247201,-.132981412843306,-.538250103024586,.293443344364012,-.631620327664824,.162868303384663,.659219053205559,.729332332662821,-.188064117536924,-.81541833693141,-2.55652048864375,.659219053205559,2.66822795597052,2.95201558448519,-.761200595679671,-3.30045375807885,-1.60938069663533,.414991557357591,1.79934231949429,2.78946777547417},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,1.50132568517008,-.387128716912051,-1.5669262892192,.854257320336024,-1.83874093225775,.47413391064841,1.91908493657046,2.12319514516607,-.547482681696682,-2.37380433681192,-1.19541898829252,.308248252711718,1.24765296340427,1.38035094930968,-.355934423376279,-1.54327927740643,-3.34189684440345,.861734565902809,3.73635425856116,-1.66034116472922},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,2.35354378347051,-.606879902269876,-2.45638215849919,1.33917112430755,-2.88249067692314,.743273047921452,3.00844145080061,3.32841353768903,-.858257788295293,-3.72127946338712,1.0537965231098,-.271729778549237,-1.09984228773295,-1.21681941252376,.313766521598226,1.36044546108577,-1.75917907531388,.453618255557353,1.9668219982996,-4.02051566640715},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,2.99666740648972,-.772714421358283,-3.12760714457709,1.70510975326197,-3.67015303548283,.946378024692648,3.83052081004967,4.23792768749944,-1.09278321423787,-4.73814719602452,3.40958356039142,-.879188121703209,-3.55856572026934,-3.93704797266052,1.01519899751182,4.40175344784644,2.40062592164155,-.61902040451823,-2.68398137439783,.178027784911158},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,3.37371117926928,-.869938143977334,-3.52112588977045,1.91964841477002,-4.13193546224797,1.06545228035847,4.31248087502183,4.77114810448709,-1.23027832131385,-5.33430574565371,5.05405999869821,-1.30322939400866,-5.27489775246548,-5.83592580124217,1.5048396829594,6.52476340023727,6.34960212623239,-1.63729519092357,-7.09907099144838,7.15379270875832},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-3.37371117926928,-.869938143977334,3.52112588977045,1.91964841477004,4.13193546224797,1.06545228035847,-4.31248087502184,-4.77114810448709,-1.23027832131385,5.33430574565371,5.05405999869821,1.30322939400866,-5.27489775246549,-5.83592580124217,-1.5048396829594,6.52476340023727,-6.34960212623238,-1.63729519092357,7.09907099144837,7.15379270875817},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-2.99666740648972,-.772714421358283,3.12760714457709,1.70510975326198,3.67015303548283,.946378024692648,-3.83052081004967,-4.23792768749944,-1.09278321423787,4.73814719602452,3.40958356039142,.879188121703209,-3.55856572026934,-3.93704797266052,-1.01519899751182,4.40175344784644,-2.40062592164153,-.619020404518225,2.68398137439781,.178027784911327},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-2.35354378347051,-.606879902269876,2.45638215849919,1.33917112430757,2.88249067692314,.743273047921452,-3.00844145080061,-3.32841353768903,-.858257788295293,3.72127946338712,1.05379652310981,.271729778549238,-1.09984228773296,-1.21681941252376,-.313766521598227,1.36044546108577,1.7591790753139,.453618255557357,-1.96682199829962,-4.02051566640726},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-1.50132568517008,-.387128716912051,1.5669262892192,.854257320336031,1.83874093225775,.47413391064841,-1.91908493657046,-2.12319514516607,-.547482681696682,2.37380433681192,-1.19541898829253,-.308248252711719,1.24765296340427,1.38035094930968,.35593442337628,-1.54327927740643,3.34189684440344,.861734565902807,-3.73635425856115,-1.66034116472927},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-.515715838247201,-.132981412843306,.538250103024586,.293443344364014,.631620327664824,.162868303384663,-.659219053205559,-.729332332662821,-.188064117536924,.81541833693141,-2.55652048864375,-.659219053205559,2.66822795597052,2.95201558448519,.761200595679671,-3.30045375807884,1.60938069663531,.414991557357587,-1.79934231949427,2.78946777547416},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,.515715838247201,.132981412843306,-.538250103024586,-.293443344364014,-.631620327664824,-.162868303384663,.659219053205559,.729332332662821,.188064117536924,-.81541833693141,-2.55652048864375,-.659219053205559,2.66822795597052,2.95201558448519,.761200595679671,-3.30045375807885,-1.60938069663533,-.414991557357591,1.79934231949429,2.78946777547417},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,1.50132568517008,.387128716912051,-1.5669262892192,-.854257320336031,-1.83874093225775,-.47413391064841,1.91908493657046,2.12319514516607,.547482681696682,-2.37380433681192,-1.19541898829252,-.308248252711718,1.24765296340427,1.38035094930968,.355934423376279,-1.54327927740643,-3.34189684440345,-.861734565902809,3.73635425856116,-1.66034116472922},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,2.35354378347051,.606879902269876,-2.45638215849919,-1.33917112430757,-2.88249067692314,-.743273047921452,3.00844145080061,3.32841353768903,.858257788295293,-3.72127946338712,1.0537965231098,.271729778549237,-1.09984228773295,-1.21681941252376,-.313766521598226,1.36044546108577,-1.75917907531388,-.453618255557353,1.9668219982996,-4.02051566640715},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,2.99666740648972,.772714421358283,-3.12760714457709,-1.70510975326198,-3.67015303548283,-.946378024692648,3.83052081004967,4.23792768749944,1.09278321423787,-4.73814719602452,3.40958356039142,.879188121703209,-3.55856572026934,-3.93704797266052,-1.01519899751182,4.40175344784644,2.40062592164155,.61902040451823,-2.68398137439783,.178027784911158},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,3.37371117926928,.869938143977334,-3.52112588977045,-1.91964841477004,-4.13193546224797,-1.06545228035847,4.31248087502184,4.77114810448709,1.23027832131385,-5.33430574565371,5.05405999869821,1.30322939400866,-5.27489775246549,-5.83592580124217,-1.5048396829594,6.52476340023727,6.34960212623239,1.63729519092357,-7.09907099144838,7.15379270875832},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-3.37371117926928,-2.53251962263048,1.64646470368521,3.98617119808646,4.13193546224797,3.10169041978979,-2.01649920176635,-4.77114810448709,-3.58152359635514,5.33430574565371,5.05405999869821,3.79389504410759,-2.46652157203872,-5.83592580124217,-4.38081265017735,6.52476340023727,-6.34960212623238,-4.76641037958728,7.09907099144837,7.15379270875817},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-2.99666740648972,-2.24948687251768,1.4624568765065,3.54067929092288,3.67015303548283,2.7550475106226,-1.79113655913325,-4.23792768749944,-3.18125484265547,4.73814719602452,3.40958356039142,2.55944768672544,-1.66397142209237,-3.93704797266052,-2.95539562194488,4.40175344784644,-2.40062592164153,-1.80206067765193,2.68398137439781,.178027784911327},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-2.35354378347051,-1.76671786576882,1.14859469650904,2.78080367423081,2.88249067692314,2.16377864548779,-1.40673546385745,-3.32841353768903,-2.49851636599797,3.72127946338712,1.05379652310981,.791045893312282,-.514281954994445,-1.21681941252376,-.913421118930409,1.36044546108577,1.7591790753139,1.32055036479967,-1.96682199829962,-4.02051566640726},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-1.50132568517008,-1.12698940591467,.732688608484838,1.7738747886737,1.83874093225775,1.38027449512883,-.897356615569122,-2.12319514516607,-1.59380370207485,2.37380433681192,-1.19541898829253,-.897356615569124,.583397649218189,1.38035094930968,1.03617816716125,-1.54327927740643,3.34189684440344,2.508637783912,-3.73635425856115,-1.66034116472927},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-.515715838247201,-.387128716912051,.251683644415986,.609338355176942,.631620327664824,.47413391064841,-.308248252711718,-.729332332662821,-.547482681696682,.81541833693141,-2.55652048864375,-1.91908493657046,1.24765296340427,2.95201558448519,2.21596840955051,-3.30045375807884,1.60938069663531,1.20810228808803,-1.79934231949427,2.78946777547416},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,.515715838247201,.387128716912051,-.251683644415986,-.609338355176942,-.631620327664824,-.47413391064841,.308248252711718,.729332332662821,.547482681696682,-.81541833693141,-2.55652048864375,-1.91908493657046,1.24765296340427,2.95201558448519,2.21596840955051,-3.30045375807885,-1.60938069663533,-1.20810228808804,1.79934231949429,2.78946777547417},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,1.50132568517008,1.12698940591467,-.732688608484838,-1.7738747886737,-1.83874093225775,-1.38027449512883,.897356615569122,2.12319514516607,1.59380370207485,-2.37380433681192,-1.19541898829252,-.897356615569122,.583397649218187,1.38035094930968,1.03617816716125,-1.54327927740643,-3.34189684440345,-2.508637783912,3.73635425856116,-1.66034116472922},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,2.35354378347051,1.76671786576882,-1.14859469650904,-2.78080367423081,-2.88249067692314,-2.16377864548779,1.40673546385745,3.32841353768903,2.49851636599797,-3.72127946338712,1.0537965231098,.791045893312278,-.514281954994442,-1.21681941252376,-.913421118930405,1.36044546108577,-1.75917907531388,-1.32055036479965,1.9668219982996,-4.02051566640715},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,2.99666740648972,2.24948687251768,-1.4624568765065,-3.54067929092288,-3.67015303548283,-2.7550475106226,1.79113655913325,4.23792768749944,3.18125484265547,-4.73814719602452,3.40958356039142,2.55944768672544,-1.66397142209237,-3.93704797266052,-2.95539562194488,4.40175344784644,2.40062592164155,1.80206067765195,-2.68398137439783,.178027784911158},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,3.37371117926928,2.53251962263048,-1.64646470368521,-3.98617119808646,-4.13193546224797,-3.10169041978979,2.01649920176635,4.77114810448709,3.58152359635514,-5.33430574565371,5.05405999869821,3.79389504410759,-2.46652157203872,-5.83592580124217,-4.38081265017735,6.52476340023727,6.34960212623239,4.76641037958728,-7.09907099144838,7.15379270875832},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-3.37371117926928,-3.97008848462071,-1.45140640826253,2.098325977367,4.13193546224797,4.86234551094043,1.77760255482497,-4.77114810448709,-5.61455297729064,5.33430574565371,5.05405999869821,5.94747574387061,2.17431033156192,-5.83592580124217,-6.86755344374583,6.52476340023727,-6.34960212623238,-7.47203330307115,7.09907099144837,7.15379270875817},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-2.99666740648972,-3.52639397108079,-1.28919816964079,1.86381842737596,3.67015303548283,4.31893293096971,1.57893884647547,-4.23792768749944,-4.98707417885748,4.73814719602452,3.40958356039142,4.01230209521675,1.46683908849365,-3.93704797266052,-4.63300738968997,4.40175344784644,-2.40062592164153,-2.82498910610717,2.68398137439781,.178027784911327},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-2.35354378347051,-2.7695841689776,-1.01251955130178,1.46381886216959,2.88249067692314,3.3920340071429,1.24007812764096,-3.32841353768903,-3.91678349286862,3.72127946338712,1.05379652310981,1.24007812764096,.453354465153134,-1.21681941252376,-1.43191888168209,1.36044546108577,1.7591790753139,2.07015248758758,-1.96682199829962,-4.02051566640726},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-1.50132568517008,-1.76671786576882,-.645886267246195,.933770117915969,1.83874093225775,2.16377864548779,.791045893312278,-2.12319514516607,-2.49851636599797,2.37380433681192,-1.19541898829253,-1.40673546385746,-.514281954994444,1.38035094930968,1.62435819754456,-1.54327927740643,3.34189684440344,3.93265027010875,-3.73635425856115,-1.66034116472927},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-.515715838247201,-.606879902269876,-.221866435121631,.320756544597894,.631620327664824,.743273047921452,.271729778549237,-.729332332662821,-.858257788295293,.81541833693141,-2.55652048864375,-3.00844145080061,-1.09984228773295,2.95201558448519,3.47384896307412,-3.30045375807884,1.60938069663531,1.89387396619673,-1.79934231949427,2.78946777547416},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,.515715838247201,.606879902269876,.221866435121631,-.320756544597894,-.631620327664824,-.743273047921452,-.271729778549237,.729332332662821,.858257788295293,-.81541833693141,-2.55652048864375,-3.00844145080061,-1.09984228773295,2.95201558448519,3.47384896307413,-3.30045375807885,-1.60938069663533,-1.89387396619674,1.79934231949429,2.78946777547417},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,1.50132568517008,1.76671786576882,.645886267246195,-.933770117915969,-1.83874093225775,-2.16377864548779,-.791045893312278,2.12319514516607,2.49851636599797,-2.37380433681192,-1.19541898829252,-1.40673546385745,-.514281954994442,1.38035094930968,1.62435819754456,-1.54327927740643,-3.34189684440345,-3.93265027010876,3.73635425856116,-1.66034116472922},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,2.35354378347051,2.7695841689776,1.01251955130178,-1.46381886216959,-2.88249067692314,-3.3920340071429,-1.24007812764096,3.32841353768903,3.91678349286862,-3.72127946338712,1.0537965231098,1.24007812764096,.453354465153132,-1.21681941252376,-1.43191888168208,1.36044546108577,-1.75917907531388,-2.07015248758756,1.9668219982996,-4.02051566640715},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,2.99666740648972,3.52639397108079,1.28919816964079,-1.86381842737596,-3.67015303548283,-4.31893293096971,-1.57893884647547,4.23792768749944,4.98707417885748,-4.73814719602452,3.40958356039142,4.01230209521675,1.46683908849365,-3.93704797266052,-4.63300738968997,4.40175344784644,2.40062592164155,2.82498910610719,-2.68398137439783,.178027784911158},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,3.37371117926928,3.97008848462071,1.45140640826253,-2.098325977367,-4.13193546224797,-4.86234551094043,-1.77760255482497,4.77114810448709,5.61455297729064,-5.33430574565371,5.05405999869821,5.94747574387061,2.17431033156192,-5.83592580124217,-6.86755344374583,6.52476340023727,6.34960212623239,7.47203330307115,-7.09907099144838,7.15379270875832},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-3.37371117926928,-5.05494516239668,-4.69605974259137,-2.8634354535068,4.13193546224797,6.19101816335907,5.75147508548906,-4.77114810448709,-7.14877200382768,5.33430574565371,5.05405999869821,7.5726684320545,7.03503247458549,-5.83592580124217,-8.74416431584401,6.52476340023727,-6.34960212623238,-9.51382286438758,7.09907099144837,7.15379270875817},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-2.99666740648972,-4.4900077703237,-4.17123115221737,-2.54341976480911,3.67015303548283,5.4991139896991,5.10869396106863,-4.23792768749944,-6.34982988227715,4.73814719602452,3.40958356039142,5.10869396106863,4.74599254428019,-3.93704797266052,-5.89901166751924,4.40175344784644,-2.40062592164153,-3.59693872593977,2.68398137439781,.178027784911327},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-2.35354378347051,-3.52639397108079,-3.27603094239261,-1.99756895385148,2.88249067692314,4.31893293096971,4.01230209521675,-3.32841353768903,-4.98707417885748,3.72127946338712,1.05379652310981,1.57893884647548,1.46683908849366,-1.21681941252376,-1.82320153617303,1.36044546108577,1.7591790753139,2.63583729760472,-1.96682199829962,-4.02051566640726},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-1.50132568517008,-2.24948687251768,-2.08978028527405,-1.27424932536979,1.83874093225775,2.7550475106226,2.55944768672544,-2.12319514516607,-3.18125484265547,2.37380433681192,-1.19541898829253,-1.79113655913325,-1.66397142209237,1.38035094930968,2.06822634923255,-1.54327927740643,3.34189684440344,5.0072766728733,-3.73635425856115,-1.66034116472927},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-.515715838247201,-.772714421358283,-.717854095362718,-.437713525759445,.631620327664824,.946378024692648,.879188121703209,-.729332332662821,-1.09278321423787,.81541833693141,-2.55652048864375,-3.83052081004967,-3.55856572026934,2.95201558448519,4.42310444183107,-3.30045375807884,1.60938069663531,2.41138933822272,-1.79934231949427,2.78946777547416},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,.515715838247201,.772714421358283,.717854095362718,.437713525759445,-.631620327664824,-.946378024692648,-.879188121703209,.729332332662821,1.09278321423787,-.81541833693141,-2.55652048864375,-3.83052081004967,-3.55856572026934,2.95201558448519,4.42310444183107,-3.30045375807885,-1.60938069663533,-2.41138933822274,1.79934231949429,2.78946777547417},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,1.50132568517008,2.24948687251768,2.08978028527405,1.27424932536979,-1.83874093225775,-2.7550475106226,-2.55944768672544,2.12319514516607,3.18125484265547,-2.37380433681192,-1.19541898829252,-1.79113655913325,-1.66397142209237,1.38035094930968,2.06822634923254,-1.54327927740643,-3.34189684440345,-5.00727667287331,3.73635425856116,-1.66034116472922},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,2.35354378347051,3.52639397108079,3.27603094239261,1.99756895385148,-2.88249067692314,-4.31893293096971,-4.01230209521675,3.32841353768903,4.98707417885748,-3.72127946338712,1.0537965231098,1.57893884647547,1.46683908849365,-1.21681941252376,-1.82320153617302,1.36044546108577,-1.75917907531388,-2.6358372976047,1.9668219982996,-4.02051566640715},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,2.99666740648972,4.4900077703237,4.17123115221737,2.54341976480911,-3.67015303548283,-5.4991139896991,-5.10869396106863,4.23792768749944,6.34982988227715,-4.73814719602452,3.40958356039142,5.10869396106863,4.74599254428019,-3.93704797266052,-5.89901166751924,4.40175344784644,2.40062592164155,3.59693872593979,-2.68398137439783,.178027784911158},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,3.37371117926928,5.05494516239668,4.69605974259137,2.8634354535068,-4.13193546224797,-6.19101816335907,-5.75147508548906,4.77114810448709,7.14877200382768,-5.33430574565371,5.05405999869821,7.5726684320545,7.03503247458549,-5.83592580124217,-8.74416431584401,6.52476340023727,6.34960212623239,9.51382286438759,-7.09907099144838,7.15379270875832},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-3.37371117926928,-5.6909635577302,-6.96101658051265,-7.57372303615022,4.13193546224797,6.96997843122346,8.52546935665732,-4.77114810448709,-8.04823784418982,5.33430574565371,5.05405999869821,8.52546935665732,10.4280993821027,-5.83592580124217,-9.84436405649934,6.52476340023727,-6.34960212623238,-10.7108618346039,7.09907099144837,7.15379270875817},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-2.99666740648972,-5.05494516239668,-6.18305788326991,-6.72728866290429,3.67015303548283,6.19101816335907,7.5726684320545,-4.23792768749944,-7.14877200382768,4.73814719602452,3.40958356039142,5.75147508548906,7.03503247458549,-3.93704797266052,-6.64123137798004,4.40175344784644,-2.40062592164153,-4.04950925303536,2.68398137439781,.178027784911327},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-2.35354378347051,-3.97008848462071,-4.85609361001944,-5.28352541823676,2.88249067692314,4.86234551094043,5.94747574387061,-3.32841353768903,-5.61455297729064,3.72127946338712,1.05379652310981,1.77760255482498,2.17431033156193,-1.21681941252376,-2.05259862717067,1.36044546108577,1.7591790753139,2.96748105525688,-1.96682199829962,-4.02051566640726},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-1.50132568517008,-2.53251962263048,-3.09770233191153,-3.3703611015687,1.83874093225775,3.10169041978979,3.79389504410759,-2.12319514516607,-3.58152359635514,2.37380433681192,-1.19541898829253,-2.01649920176636,-2.46652157203873,1.38035094930968,2.32845271402296,-1.54327927740643,3.34189684440344,5.63729736986577,-3.73635425856115,-1.66034116472927},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-.515715838247201,-.869938143977334,-1.06408234437226,-1.1577425323902,.631620327664824,1.06545228035847,1.30322939400866,-.729332332662821,-1.23027832131385,.81541833693141,-2.55652048864375,-4.31248087502183,-5.27489775246548,2.95201558448519,4.97962398835612,-3.30045375807884,1.60938069663531,2.71479282295876,-1.79934231949427,2.78946777547416},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,.515715838247201,.869938143977334,1.06408234437226,1.1577425323902,-.631620327664824,-1.06545228035847,-1.30322939400866,.729332332662821,1.23027832131385,-.81541833693141,-2.55652048864375,-4.31248087502184,-5.27489775246549,2.95201558448519,4.97962398835612,-3.30045375807885,-1.60938069663533,-2.71479282295879,1.79934231949429,2.78946777547417},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,1.50132568517008,2.53251962263048,3.09770233191153,3.3703611015687,-1.83874093225775,-3.10169041978979,-3.79389504410759,2.12319514516607,3.58152359635514,-2.37380433681192,-1.19541898829252,-2.01649920176635,-2.46652157203872,1.38035094930968,2.32845271402296,-1.54327927740643,-3.34189684440345,-5.63729736986578,3.73635425856116,-1.66034116472922},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,2.35354378347051,3.97008848462071,4.85609361001944,5.28352541823676,-2.88249067692314,-4.86234551094043,-5.94747574387061,3.32841353768903,5.61455297729064,-3.72127946338712,1.0537965231098,1.77760255482497,2.17431033156192,-1.21681941252376,-2.05259862717066,1.36044546108577,-1.75917907531388,-2.96748105525684,1.9668219982996,-4.02051566640715},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,2.99666740648972,5.05494516239668,6.18305788326991,6.72728866290429,-3.67015303548283,-6.19101816335907,-7.5726684320545,4.23792768749944,7.14877200382768,-4.73814719602452,3.40958356039142,5.75147508548906,7.03503247458549,-3.93704797266052,-6.64123137798004,4.40175344784644,2.40062592164155,4.0495092530354,-2.68398137439783,.178027784911158},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,3.37371117926928,5.6909635577302,6.96101658051265,7.57372303615022,-4.13193546224797,-6.96997843122346,-8.52546935665732,4.77114810448709,8.04823784418982,-5.33430574565371,5.05405999869821,8.52546935665732,10.4280993821027,-5.83592580124217,-9.84436405649934,6.52476340023727,6.34960212623239,10.7108618346039,-7.09907099144838,7.15379270875832}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_left_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_left_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-3.37371117926928,5.6909635577302,-6.96101658051265,7.57372303615021,4.13193546224797,-6.96997843122346,8.52546935665732,-4.77114810448709,8.04823784418982,5.33430574565371,5.05405999869821,-8.52546935665732,10.4280993821027,-5.83592580124217,9.84436405649934,6.52476340023727,-6.34960212623238,10.7108618346039,7.09907099144837,7.15379270875817},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-2.99666740648972,5.05494516239668,-6.18305788326991,6.72728866290428,3.67015303548283,-6.19101816335907,7.5726684320545,-4.23792768749944,7.14877200382768,4.73814719602452,3.40958356039142,-5.75147508548906,7.03503247458549,-3.93704797266052,6.64123137798004,4.40175344784644,-2.40062592164153,4.04950925303536,2.68398137439781,.178027784911327},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-2.35354378347051,3.97008848462071,-4.85609361001944,5.28352541823675,2.88249067692314,-4.86234551094043,5.94747574387061,-3.32841353768903,5.61455297729064,3.72127946338712,1.05379652310981,-1.77760255482498,2.17431033156193,-1.21681941252376,2.05259862717067,1.36044546108577,1.7591790753139,-2.96748105525688,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-1.50132568517008,2.53251962263048,-3.09770233191153,3.3703611015687,1.83874093225775,-3.10169041978979,3.79389504410759,-2.12319514516607,3.58152359635514,2.37380433681192,-1.19541898829253,2.01649920176636,-2.46652157203873,1.38035094930968,-2.32845271402296,-1.54327927740643,3.34189684440344,-5.63729736986577,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-.515715838247201,.869938143977334,-1.06408234437226,1.1577425323902,.631620327664824,-1.06545228035847,1.30322939400866,-.729332332662821,1.23027832131385,.81541833693141,-2.55652048864375,4.31248087502183,-5.27489775246548,2.95201558448519,-4.97962398835612,-3.30045375807884,1.60938069663531,-2.71479282295876,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,.515715838247201,-.869938143977334,1.06408234437226,-1.1577425323902,-.631620327664824,1.06545228035847,-1.30322939400866,.729332332662821,-1.23027832131385,-.81541833693141,-2.55652048864375,4.31248087502184,-5.27489775246549,2.95201558448519,-4.97962398835612,-3.30045375807885,-1.60938069663533,2.71479282295879,1.79934231949429,2.78946777547417},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,1.50132568517008,-2.53251962263048,3.09770233191153,-3.3703611015687,-1.83874093225775,3.10169041978979,-3.79389504410759,2.12319514516607,-3.58152359635514,-2.37380433681192,-1.19541898829252,2.01649920176635,-2.46652157203872,1.38035094930968,-2.32845271402296,-1.54327927740643,-3.34189684440345,5.63729736986578,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,2.35354378347051,-3.97008848462071,4.85609361001944,-5.28352541823675,-2.88249067692314,4.86234551094043,-5.94747574387061,3.32841353768903,-5.61455297729064,-3.72127946338712,1.0537965231098,-1.77760255482497,2.17431033156192,-1.21681941252376,2.05259862717066,1.36044546108577,-1.75917907531388,2.96748105525684,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,2.99666740648972,-5.05494516239668,6.18305788326991,-6.72728866290428,-3.67015303548283,6.19101816335907,-7.5726684320545,4.23792768749944,-7.14877200382768,-4.73814719602452,3.40958356039142,-5.75147508548906,7.03503247458549,-3.93704797266052,6.64123137798004,4.40175344784644,2.40062592164155,-4.0495092530354,-2.68398137439783,.178027784911158},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,3.37371117926928,-5.6909635577302,6.96101658051265,-7.57372303615021,-4.13193546224797,6.96997843122346,-8.52546935665732,4.77114810448709,-8.04823784418982,-5.33430574565371,5.05405999869821,-8.52546935665732,10.4280993821027,-5.83592580124217,9.84436405649934,6.52476340023727,6.34960212623239,-10.7108618346039,-7.09907099144838,7.15379270875832},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-3.37371117926928,5.05494516239668,-4.69605974259137,2.86343545350678,4.13193546224797,-6.19101816335907,5.75147508548906,-4.77114810448709,7.14877200382768,5.33430574565371,5.05405999869821,-7.5726684320545,7.03503247458549,-5.83592580124217,8.74416431584401,6.52476340023727,-6.34960212623238,9.51382286438758,7.09907099144837,7.15379270875817},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-2.99666740648972,4.4900077703237,-4.17123115221737,2.54341976480909,3.67015303548283,-5.4991139896991,5.10869396106863,-4.23792768749944,6.34982988227715,4.73814719602452,3.40958356039142,-5.10869396106863,4.74599254428019,-3.93704797266052,5.89901166751924,4.40175344784644,-2.40062592164153,3.59693872593977,2.68398137439781,.178027784911327},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-2.35354378347051,3.52639397108079,-3.27603094239261,1.99756895385147,2.88249067692314,-4.31893293096971,4.01230209521675,-3.32841353768903,4.98707417885748,3.72127946338712,1.05379652310981,-1.57893884647548,1.46683908849366,-1.21681941252376,1.82320153617303,1.36044546108577,1.7591790753139,-2.63583729760472,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-1.50132568517008,2.24948687251768,-2.08978028527405,1.27424932536978,1.83874093225775,-2.7550475106226,2.55944768672544,-2.12319514516607,3.18125484265547,2.37380433681192,-1.19541898829253,1.79113655913325,-1.66397142209237,1.38035094930968,-2.06822634923255,-1.54327927740643,3.34189684440344,-5.0072766728733,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-.515715838247201,.772714421358283,-.717854095362718,.437713525759442,.631620327664824,-.946378024692648,.879188121703209,-.729332332662821,1.09278321423787,.81541833693141,-2.55652048864375,3.83052081004967,-3.55856572026934,2.95201558448519,-4.42310444183107,-3.30045375807884,1.60938069663531,-2.41138933822272,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,.515715838247201,-.772714421358283,.717854095362718,-.437713525759442,-.631620327664824,.946378024692648,-.879188121703209,.729332332662821,-1.09278321423787,-.81541833693141,-2.55652048864375,3.83052081004967,-3.55856572026934,2.95201558448519,-4.42310444183107,-3.30045375807885,-1.60938069663533,2.41138933822274,1.79934231949429,2.78946777547417},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,1.50132568517008,-2.24948687251768,2.08978028527405,-1.27424932536978,-1.83874093225775,2.7550475106226,-2.55944768672544,2.12319514516607,-3.18125484265547,-2.37380433681192,-1.19541898829252,1.79113655913325,-1.66397142209237,1.38035094930968,-2.06822634923254,-1.54327927740643,-3.34189684440345,5.00727667287331,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,2.35354378347051,-3.52639397108079,3.27603094239261,-1.99756895385147,-2.88249067692314,4.31893293096971,-4.01230209521675,3.32841353768903,-4.98707417885748,-3.72127946338712,1.0537965231098,-1.57893884647547,1.46683908849365,-1.21681941252376,1.82320153617302,1.36044546108577,-1.75917907531388,2.6358372976047,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,2.99666740648972,-4.4900077703237,4.17123115221737,-2.54341976480909,-3.67015303548283,5.4991139896991,-5.10869396106863,4.23792768749944,-6.34982988227715,-4.73814719602452,3.40958356039142,-5.10869396106863,4.74599254428019,-3.93704797266052,5.89901166751924,4.40175344784644,2.40062592164155,-3.59693872593979,-2.68398137439783,.178027784911158},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,3.37371117926928,-5.05494516239668,4.69605974259137,-2.86343545350678,-4.13193546224797,6.19101816335907,-5.75147508548906,4.77114810448709,-7.14877200382768,-5.33430574565371,5.05405999869821,-7.5726684320545,7.03503247458549,-5.83592580124217,8.74416431584401,6.52476340023727,6.34960212623239,-9.51382286438759,-7.09907099144838,7.15379270875832},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-3.37371117926928,3.97008848462071,-1.45140640826254,-2.09832597736703,4.13193546224797,-4.86234551094043,1.77760255482498,-4.77114810448709,5.61455297729064,5.33430574565371,5.05405999869821,-5.94747574387061,2.17431033156193,-5.83592580124217,6.86755344374583,6.52476340023727,-6.34960212623238,7.47203330307115,7.09907099144837,7.15379270875817},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-2.99666740648972,3.52639397108079,-1.2891981696408,-1.86381842737598,3.67015303548283,-4.31893293096971,1.57893884647548,-4.23792768749944,4.98707417885748,4.73814719602452,3.40958356039142,-4.01230209521675,1.46683908849366,-3.93704797266052,4.63300738968997,4.40175344784644,-2.40062592164153,2.82498910610717,2.68398137439781,.178027784911327},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-2.35354378347051,2.7695841689776,-1.01251955130179,-1.46381886216961,2.88249067692314,-3.3920340071429,1.24007812764096,-3.32841353768903,3.91678349286862,3.72127946338712,1.05379652310981,-1.24007812764096,.453354465153136,-1.21681941252376,1.43191888168209,1.36044546108577,1.7591790753139,-2.07015248758758,-1.96682199829962,-4.02051566640726},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-1.50132568517008,1.76671786576882,-.645886267246198,-.933770117915978,1.83874093225775,-2.16377864548779,.791045893312282,-2.12319514516607,2.49851636599797,2.37380433681192,-1.19541898829253,1.40673546385746,-.514281954994446,1.38035094930968,-1.62435819754456,-1.54327927740643,3.34189684440344,-3.93265027010875,-3.73635425856115,-1.66034116472927},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-.515715838247201,.606879902269876,-.221866435121632,-.320756544597897,.631620327664824,-.743273047921452,.271729778549238,-.729332332662821,.858257788295293,.81541833693141,-2.55652048864375,3.00844145080061,-1.09984228773296,2.95201558448519,-3.47384896307412,-3.30045375807884,1.60938069663531,-1.89387396619673,-1.79934231949427,2.78946777547416},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,.515715838247201,-.606879902269876,.221866435121632,.320756544597897,-.631620327664824,.743273047921452,-.271729778549238,.729332332662821,-.858257788295293,-.81541833693141,-2.55652048864375,3.00844145080061,-1.09984228773296,2.95201558448519,-3.47384896307413,-3.30045375807885,-1.60938069663533,1.89387396619674,1.79934231949429,2.78946777547417},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,1.50132568517008,-1.76671786576882,.645886267246198,.933770117915978,-1.83874093225775,2.16377864548779,-.791045893312282,2.12319514516607,-2.49851636599797,-2.37380433681192,-1.19541898829252,1.40673546385745,-.514281954994445,1.38035094930968,-1.62435819754456,-1.54327927740643,-3.34189684440345,3.93265027010876,3.73635425856116,-1.66034116472922},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,2.35354378347051,-2.7695841689776,1.01251955130179,1.46381886216961,-2.88249067692314,3.3920340071429,-1.24007812764096,3.32841353768903,-3.91678349286862,-3.72127946338712,1.0537965231098,-1.24007812764096,.453354465153134,-1.21681941252376,1.43191888168208,1.36044546108577,-1.75917907531388,2.07015248758756,1.9668219982996,-4.02051566640715},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,2.99666740648972,-3.52639397108079,1.2891981696408,1.86381842737598,-3.67015303548283,4.31893293096971,-1.57893884647548,4.23792768749944,-4.98707417885748,-4.73814719602452,3.40958356039142,-4.01230209521675,1.46683908849366,-3.93704797266052,4.63300738968997,4.40175344784644,2.40062592164155,-2.82498910610719,-2.68398137439783,.178027784911158},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,3.37371117926928,-3.97008848462071,1.45140640826254,2.09832597736703,-4.13193546224797,4.86234551094043,-1.77760255482498,4.77114810448709,-5.61455297729064,-5.33430574565371,5.05405999869821,-5.94747574387061,2.17431033156193,-5.83592580124217,6.86755344374583,6.52476340023727,6.34960212623239,-7.47203330307115,-7.09907099144838,7.15379270875832},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-3.37371117926928,2.53251962263048,1.64646470368521,-3.98617119808645,4.13193546224797,-3.10169041978979,-2.01649920176636,-4.77114810448709,3.58152359635514,5.33430574565371,5.05405999869821,-3.79389504410759,-2.46652157203873,-5.83592580124217,4.38081265017735,6.52476340023727,-6.34960212623238,4.76641037958728,7.09907099144837,7.15379270875817},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-2.99666740648972,2.24948687251768,1.4624568765065,-3.54067929092287,3.67015303548283,-2.7550475106226,-1.79113655913325,-4.23792768749944,3.18125484265547,4.73814719602452,3.40958356039142,-2.55944768672544,-1.66397142209237,-3.93704797266052,2.95539562194488,4.40175344784644,-2.40062592164153,1.80206067765193,2.68398137439781,.178027784911327},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-2.35354378347051,1.76671786576882,1.14859469650904,-2.78080367423081,2.88249067692314,-2.16377864548779,-1.40673546385746,-3.32841353768903,2.49851636599797,3.72127946338712,1.05379652310981,-.791045893312282,-.514281954994446,-1.21681941252376,.913421118930409,1.36044546108577,1.7591790753139,-1.32055036479967,-1.96682199829962,-4.02051566640726},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-1.50132568517008,1.12698940591467,.732688608484839,-1.7738747886737,1.83874093225775,-1.38027449512883,-.897356615569124,-2.12319514516607,1.59380370207485,2.37380433681192,-1.19541898829253,.897356615569124,.58339764921819,1.38035094930968,-1.03617816716125,-1.54327927740643,3.34189684440344,-2.508637783912,-3.73635425856115,-1.66034116472927},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-.515715838247201,.387128716912051,.251683644415986,-.609338355176941,.631620327664824,-.47413391064841,-.308248252711719,-.729332332662821,.547482681696682,.81541833693141,-2.55652048864375,1.91908493657046,1.24765296340427,2.95201558448519,-2.21596840955051,-3.30045375807884,1.60938069663531,-1.20810228808803,-1.79934231949427,2.78946777547416},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,.515715838247201,-.387128716912051,-.251683644415986,.609338355176941,-.631620327664824,.47413391064841,.308248252711719,.729332332662821,-.547482681696682,-.81541833693141,-2.55652048864375,1.91908493657046,1.24765296340427,2.95201558448519,-2.21596840955051,-3.30045375807885,-1.60938069663533,1.20810228808804,1.79934231949429,2.78946777547417},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,1.50132568517008,-1.12698940591467,-.732688608484839,1.7738747886737,-1.83874093225775,1.38027449512883,.897356615569124,2.12319514516607,-1.59380370207485,-2.37380433681192,-1.19541898829252,.897356615569122,.583397649218189,1.38035094930968,-1.03617816716125,-1.54327927740643,-3.34189684440345,2.508637783912,3.73635425856116,-1.66034116472922},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,2.35354378347051,-1.76671786576882,-1.14859469650904,2.78080367423081,-2.88249067692314,2.16377864548779,1.40673546385746,3.32841353768903,-2.49851636599797,-3.72127946338712,1.0537965231098,-.791045893312278,-.514281954994444,-1.21681941252376,.913421118930405,1.36044546108577,-1.75917907531388,1.32055036479965,1.9668219982996,-4.02051566640715},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,2.99666740648972,-2.24948687251768,-1.4624568765065,3.54067929092287,-3.67015303548283,2.7550475106226,1.79113655913325,4.23792768749944,-3.18125484265547,-4.73814719602452,3.40958356039142,-2.55944768672544,-1.66397142209237,-3.93704797266052,2.95539562194488,4.40175344784644,2.40062592164155,-1.80206067765195,-2.68398137439783,.178027784911158},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,3.37371117926928,-2.53251962263048,-1.64646470368521,3.98617119808645,-4.13193546224797,3.10169041978979,2.01649920176636,4.77114810448709,-3.58152359635514,-5.33430574565371,5.05405999869821,-3.79389504410759,-2.46652157203873,-5.83592580124217,4.38081265017735,6.52476340023727,6.34960212623239,-4.76641037958728,-7.09907099144838,7.15379270875832},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-3.37371117926928,.869938143977334,3.52112588977045,-1.91964841477002,4.13193546224797,-1.06545228035847,-4.31248087502183,-4.77114810448709,1.23027832131385,5.33430574565371,5.05405999869821,-1.30322939400866,-5.27489775246548,-5.83592580124217,1.5048396829594,6.52476340023727,-6.34960212623238,1.63729519092357,7.09907099144837,7.15379270875817},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-2.99666740648972,.772714421358283,3.12760714457709,-1.70510975326197,3.67015303548283,-.946378024692648,-3.83052081004967,-4.23792768749944,1.09278321423787,4.73814719602452,3.40958356039142,-.879188121703209,-3.55856572026934,-3.93704797266052,1.01519899751182,4.40175344784644,-2.40062592164153,.619020404518225,2.68398137439781,.178027784911327},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-2.35354378347051,.606879902269876,2.45638215849919,-1.33917112430755,2.88249067692314,-.743273047921452,-3.00844145080061,-3.32841353768903,.858257788295293,3.72127946338712,1.05379652310981,-.271729778549238,-1.09984228773296,-1.21681941252376,.313766521598227,1.36044546108577,1.7591790753139,-.453618255557357,-1.96682199829962,-4.02051566640726},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-1.50132568517008,.387128716912051,1.5669262892192,-.854257320336024,1.83874093225775,-.47413391064841,-1.91908493657046,-2.12319514516607,.547482681696682,2.37380433681192,-1.19541898829253,.308248252711719,1.24765296340427,1.38035094930968,-.35593442337628,-1.54327927740643,3.34189684440344,-.861734565902807,-3.73635425856115,-1.66034116472927},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-.515715838247201,.132981412843306,.538250103024586,-.293443344364012,.631620327664824,-.162868303384663,-.659219053205559,-.729332332662821,.188064117536924,.81541833693141,-2.55652048864375,.659219053205559,2.66822795597052,2.95201558448519,-.761200595679671,-3.30045375807884,1.60938069663531,-.414991557357587,-1.79934231949427,2.78946777547416},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,.515715838247201,-.132981412843306,-.538250103024586,.293443344364012,-.631620327664824,.162868303384663,.659219053205559,.729332332662821,-.188064117536924,-.81541833693141,-2.55652048864375,.659219053205559,2.66822795597052,2.95201558448519,-.761200595679671,-3.30045375807885,-1.60938069663533,.414991557357591,1.79934231949429,2.78946777547417},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,1.50132568517008,-.387128716912051,-1.5669262892192,.854257320336024,-1.83874093225775,.47413391064841,1.91908493657046,2.12319514516607,-.547482681696682,-2.37380433681192,-1.19541898829252,.308248252711718,1.24765296340427,1.38035094930968,-.355934423376279,-1.54327927740643,-3.34189684440345,.861734565902809,3.73635425856116,-1.66034116472922},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,2.35354378347051,-.606879902269876,-2.45638215849919,1.33917112430755,-2.88249067692314,.743273047921452,3.00844145080061,3.32841353768903,-.858257788295293,-3.72127946338712,1.0537965231098,-.271729778549237,-1.09984228773295,-1.21681941252376,.313766521598226,1.36044546108577,-1.75917907531388,.453618255557353,1.9668219982996,-4.02051566640715},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,2.99666740648972,-.772714421358283,-3.12760714457709,1.70510975326197,-3.67015303548283,.946378024692648,3.83052081004967,4.23792768749944,-1.09278321423787,-4.73814719602452,3.40958356039142,-.879188121703209,-3.55856572026934,-3.93704797266052,1.01519899751182,4.40175344784644,2.40062592164155,-.61902040451823,-2.68398137439783,.178027784911158},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,3.37371117926928,-.869938143977334,-3.52112588977045,1.91964841477002,-4.13193546224797,1.06545228035847,4.31248087502183,4.77114810448709,-1.23027832131385,-5.33430574565371,5.05405999869821,-1.30322939400866,-5.27489775246548,-5.83592580124217,1.5048396829594,6.52476340023727,6.34960212623239,-1.63729519092357,-7.09907099144838,7.15379270875832},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-3.37371117926928,-.869938143977334,3.52112588977045,1.91964841477004,4.13193546224797,1.06545228035847,-4.31248087502184,-4.77114810448709,-1.23027832131385,5.33430574565371,5.05405999869821,1.30322939400866,-5.27489775246549,-5.83592580124217,-1.5048396829594,6.52476340023727,-6.34960212623238,-1.63729519092357,7.09907099144837,7.15379270875817},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-2.99666740648972,-.772714421358283,3.12760714457709,1.70510975326198,3.67015303548283,.946378024692648,-3.83052081004967,-4.23792768749944,-1.09278321423787,4.73814719602452,3.40958356039142,.879188121703209,-3.55856572026934,-3.93704797266052,-1.01519899751182,4.40175344784644,-2.40062592164153,-.619020404518225,2.68398137439781,.178027784911327},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-2.35354378347051,-.606879902269876,2.45638215849919,1.33917112430757,2.88249067692314,.743273047921452,-3.00844145080061,-3.32841353768903,-.858257788295293,3.72127946338712,1.05379652310981,.271729778549238,-1.09984228773296,-1.21681941252376,-.313766521598227,1.36044546108577,1.7591790753139,.453618255557357,-1.96682199829962,-4.02051566640726},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-1.50132568517008,-.387128716912051,1.5669262892192,.854257320336031,1.83874093225775,.47413391064841,-1.91908493657046,-2.12319514516607,-.547482681696682,2.37380433681192,-1.19541898829253,-.308248252711719,1.24765296340427,1.38035094930968,.35593442337628,-1.54327927740643,3.34189684440344,.861734565902807,-3.73635425856115,-1.66034116472927},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-.515715838247201,-.132981412843306,.538250103024586,.293443344364014,.631620327664824,.162868303384663,-.659219053205559,-.729332332662821,-.188064117536924,.81541833693141,-2.55652048864375,-.659219053205559,2.66822795597052,2.95201558448519,.761200595679671,-3.30045375807884,1.60938069663531,.414991557357587,-1.79934231949427,2.78946777547416},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,.515715838247201,.132981412843306,-.538250103024586,-.293443344364014,-.631620327664824,-.162868303384663,.659219053205559,.729332332662821,.188064117536924,-.81541833693141,-2.55652048864375,-.659219053205559,2.66822795597052,2.95201558448519,.761200595679671,-3.30045375807885,-1.60938069663533,-.414991557357591,1.79934231949429,2.78946777547417},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,1.50132568517008,.387128716912051,-1.5669262892192,-.854257320336031,-1.83874093225775,-.47413391064841,1.91908493657046,2.12319514516607,.547482681696682,-2.37380433681192,-1.19541898829252,-.308248252711718,1.24765296340427,1.38035094930968,.355934423376279,-1.54327927740643,-3.34189684440345,-.861734565902809,3.73635425856116,-1.66034116472922},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,2.35354378347051,.606879902269876,-2.45638215849919,-1.33917112430757,-2.88249067692314,-.743273047921452,3.00844145080061,3.32841353768903,.858257788295293,-3.72127946338712,1.0537965231098,.271729778549237,-1.09984228773295,-1.21681941252376,-.313766521598226,1.36044546108577,-1.75917907531388,-.453618255557353,1.9668219982996,-4.02051566640715},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,2.99666740648972,.772714421358283,-3.12760714457709,-1.70510975326198,-3.67015303548283,-.946378024692648,3.83052081004967,4.23792768749944,1.09278321423787,-4.73814719602452,3.40958356039142,.879188121703209,-3.55856572026934,-3.93704797266052,-1.01519899751182,4.40175344784644,2.40062592164155,.61902040451823,-2.68398137439783,.178027784911158},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,3.37371117926928,.869938143977334,-3.52112588977045,-1.91964841477004,-4.13193546224797,-1.06545228035847,4.31248087502184,4.77114810448709,1.23027832131385,-5.33430574565371,5.05405999869821,1.30322939400866,-5.27489775246549,-5.83592580124217,-1.5048396829594,6.52476340023727,6.34960212623239,1.63729519092357,-7.09907099144838,7.15379270875832},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-3.37371117926928,-2.53251962263048,1.64646470368521,3.98617119808646,4.13193546224797,3.10169041978979,-2.01649920176635,-4.77114810448709,-3.58152359635514,5.33430574565371,5.05405999869821,3.79389504410759,-2.46652157203872,-5.83592580124217,-4.38081265017735,6.52476340023727,-6.34960212623238,-4.76641037958728,7.09907099144837,7.15379270875817},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-2.99666740648972,-2.24948687251768,1.4624568765065,3.54067929092288,3.67015303548283,2.7550475106226,-1.79113655913325,-4.23792768749944,-3.18125484265547,4.73814719602452,3.40958356039142,2.55944768672544,-1.66397142209237,-3.93704797266052,-2.95539562194488,4.40175344784644,-2.40062592164153,-1.80206067765193,2.68398137439781,.178027784911327},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-2.35354378347051,-1.76671786576882,1.14859469650904,2.78080367423081,2.88249067692314,2.16377864548779,-1.40673546385745,-3.32841353768903,-2.49851636599797,3.72127946338712,1.05379652310981,.791045893312282,-.514281954994445,-1.21681941252376,-.913421118930409,1.36044546108577,1.7591790753139,1.32055036479967,-1.96682199829962,-4.02051566640726},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-1.50132568517008,-1.12698940591467,.732688608484838,1.7738747886737,1.83874093225775,1.38027449512883,-.897356615569122,-2.12319514516607,-1.59380370207485,2.37380433681192,-1.19541898829253,-.897356615569124,.583397649218189,1.38035094930968,1.03617816716125,-1.54327927740643,3.34189684440344,2.508637783912,-3.73635425856115,-1.66034116472927},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-.515715838247201,-.387128716912051,.251683644415986,.609338355176942,.631620327664824,.47413391064841,-.308248252711718,-.729332332662821,-.547482681696682,.81541833693141,-2.55652048864375,-1.91908493657046,1.24765296340427,2.95201558448519,2.21596840955051,-3.30045375807884,1.60938069663531,1.20810228808803,-1.79934231949427,2.78946777547416},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,.515715838247201,.387128716912051,-.251683644415986,-.609338355176942,-.631620327664824,-.47413391064841,.308248252711718,.729332332662821,.547482681696682,-.81541833693141,-2.55652048864375,-1.91908493657046,1.24765296340427,2.95201558448519,2.21596840955051,-3.30045375807885,-1.60938069663533,-1.20810228808804,1.79934231949429,2.78946777547417},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,1.50132568517008,1.12698940591467,-.732688608484838,-1.7738747886737,-1.83874093225775,-1.38027449512883,.897356615569122,2.12319514516607,1.59380370207485,-2.37380433681192,-1.19541898829252,-.897356615569122,.583397649218187,1.38035094930968,1.03617816716125,-1.54327927740643,-3.34189684440345,-2.508637783912,3.73635425856116,-1.66034116472922},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,2.35354378347051,1.76671786576882,-1.14859469650904,-2.78080367423081,-2.88249067692314,-2.16377864548779,1.40673546385745,3.32841353768903,2.49851636599797,-3.72127946338712,1.0537965231098,.791045893312278,-.514281954994442,-1.21681941252376,-.913421118930405,1.36044546108577,-1.75917907531388,-1.32055036479965,1.9668219982996,-4.02051566640715},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,2.99666740648972,2.24948687251768,-1.4624568765065,-3.54067929092288,-3.67015303548283,-2.7550475106226,1.79113655913325,4.23792768749944,3.18125484265547,-4.73814719602452,3.40958356039142,2.55944768672544,-1.66397142209237,-3.93704797266052,-2.95539562194488,4.40175344784644,2.40062592164155,1.80206067765195,-2.68398137439783,.178027784911158},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,3.37371117926928,2.53251962263048,-1.64646470368521,-3.98617119808646,-4.13193546224797,-3.10169041978979,2.01649920176635,4.77114810448709,3.58152359635514,-5.33430574565371,5.05405999869821,3.79389504410759,-2.46652157203872,-5.83592580124217,-4.38081265017735,6.52476340023727,6.34960212623239,4.76641037958728,-7.09907099144838,7.15379270875832},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-3.37371117926928,-3.97008848462071,-1.45140640826253,2.098325977367,4.13193546224797,4.86234551094043,1.77760255482497,-4.77114810448709,-5.61455297729064,5.33430574565371,5.05405999869821,5.94747574387061,2.17431033156192,-5.83592580124217,-6.86755344374583,6.52476340023727,-6.34960212623238,-7.47203330307115,7.09907099144837,7.15379270875817},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-2.99666740648972,-3.52639397108079,-1.28919816964079,1.86381842737596,3.67015303548283,4.31893293096971,1.57893884647547,-4.23792768749944,-4.98707417885748,4.73814719602452,3.40958356039142,4.01230209521675,1.46683908849365,-3.93704797266052,-4.63300738968997,4.40175344784644,-2.40062592164153,-2.82498910610717,2.68398137439781,.178027784911327},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-2.35354378347051,-2.7695841689776,-1.01251955130178,1.46381886216959,2.88249067692314,3.3920340071429,1.24007812764096,-3.32841353768903,-3.91678349286862,3.72127946338712,1.05379652310981,1.24007812764096,.453354465153134,-1.21681941252376,-1.43191888168209,1.36044546108577,1.7591790753139,2.07015248758758,-1.96682199829962,-4.02051566640726},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-1.50132568517008,-1.76671786576882,-.645886267246195,.933770117915969,1.83874093225775,2.16377864548779,.791045893312278,-2.12319514516607,-2.49851636599797,2.37380433681192,-1.19541898829253,-1.40673546385746,-.514281954994444,1.38035094930968,1.62435819754456,-1.54327927740643,3.34189684440344,3.93265027010875,-3.73635425856115,-1.66034116472927},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-.515715838247201,-.606879902269876,-.221866435121631,.320756544597894,.631620327664824,.743273047921452,.271729778549237,-.729332332662821,-.858257788295293,.81541833693141,-2.55652048864375,-3.00844145080061,-1.09984228773295,2.95201558448519,3.47384896307412,-3.30045375807884,1.60938069663531,1.89387396619673,-1.79934231949427,2.78946777547416},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,.515715838247201,.606879902269876,.221866435121631,-.320756544597894,-.631620327664824,-.743273047921452,-.271729778549237,.729332332662821,.858257788295293,-.81541833693141,-2.55652048864375,-3.00844145080061,-1.09984228773295,2.95201558448519,3.47384896307413,-3.30045375807885,-1.60938069663533,-1.89387396619674,1.79934231949429,2.78946777547417},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,1.50132568517008,1.76671786576882,.645886267246195,-.933770117915969,-1.83874093225775,-2.16377864548779,-.791045893312278,2.12319514516607,2.49851636599797,-2.37380433681192,-1.19541898829252,-1.40673546385745,-.514281954994442,1.38035094930968,1.62435819754456,-1.54327927740643,-3.34189684440345,-3.93265027010876,3.73635425856116,-1.66034116472922},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,2.35354378347051,2.7695841689776,1.01251955130178,-1.46381886216959,-2.88249067692314,-3.3920340071429,-1.24007812764096,3.32841353768903,3.91678349286862,-3.72127946338712,1.0537965231098,1.24007812764096,.453354465153132,-1.21681941252376,-1.43191888168208,1.36044546108577,-1.75917907531388,-2.07015248758756,1.9668219982996,-4.02051566640715},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,2.99666740648972,3.52639397108079,1.28919816964079,-1.86381842737596,-3.67015303548283,-4.31893293096971,-1.57893884647547,4.23792768749944,4.98707417885748,-4.73814719602452,3.40958356039142,4.01230209521675,1.46683908849365,-3.93704797266052,-4.63300738968997,4.40175344784644,2.40062592164155,2.82498910610719,-2.68398137439783,.178027784911158},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,3.37371117926928,3.97008848462071,1.45140640826253,-2.098325977367,-4.13193546224797,-4.86234551094043,-1.77760255482497,4.77114810448709,5.61455297729064,-5.33430574565371,5.05405999869821,5.94747574387061,2.17431033156192,-5.83592580124217,-6.86755344374583,6.52476340023727,6.34960212623239,7.47203330307115,-7.09907099144838,7.15379270875832},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-3.37371117926928,-5.05494516239668,-4.69605974259137,-2.8634354535068,4.13193546224797,6.19101816335907,5.75147508548906,-4.77114810448709,-7.14877200382768,5.33430574565371,5.05405999869821,7.5726684320545,7.03503247458549,-5.83592580124217,-8.74416431584401,6.52476340023727,-6.34960212623238,-9.51382286438758,7.09907099144837,7.15379270875817},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-2.99666740648972,-4.4900077703237,-4.17123115221737,-2.54341976480911,3.67015303548283,5.4991139896991,5.10869396106863,-4.23792768749944,-6.34982988227715,4.73814719602452,3.40958356039142,5.10869396106863,4.74599254428019,-3.93704797266052,-5.89901166751924,4.40175344784644,-2.40062592164153,-3.59693872593977,2.68398137439781,.178027784911327},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-2.35354378347051,-3.52639397108079,-3.27603094239261,-1.99756895385148,2.88249067692314,4.31893293096971,4.01230209521675,-3.32841353768903,-4.98707417885748,3.72127946338712,1.05379652310981,1.57893884647548,1.46683908849366,-1.21681941252376,-1.82320153617303,1.36044546108577,1.7591790753139,2.63583729760472,-1.96682199829962,-4.02051566640726},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-1.50132568517008,-2.24948687251768,-2.08978028527405,-1.27424932536979,1.83874093225775,2.7550475106226,2.55944768672544,-2.12319514516607,-3.18125484265547,2.37380433681192,-1.19541898829253,-1.79113655913325,-1.66397142209237,1.38035094930968,2.06822634923255,-1.54327927740643,3.34189684440344,5.0072766728733,-3.73635425856115,-1.66034116472927},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-.515715838247201,-.772714421358283,-.717854095362718,-.437713525759445,.631620327664824,.946378024692648,.879188121703209,-.729332332662821,-1.09278321423787,.81541833693141,-2.55652048864375,-3.83052081004967,-3.55856572026934,2.95201558448519,4.42310444183107,-3.30045375807884,1.60938069663531,2.41138933822272,-1.79934231949427,2.78946777547416},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,.515715838247201,.772714421358283,.717854095362718,.437713525759445,-.631620327664824,-.946378024692648,-.879188121703209,.729332332662821,1.09278321423787,-.81541833693141,-2.55652048864375,-3.83052081004967,-3.55856572026934,2.95201558448519,4.42310444183107,-3.30045375807885,-1.60938069663533,-2.41138933822274,1.79934231949429,2.78946777547417},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,1.50132568517008,2.24948687251768,2.08978028527405,1.27424932536979,-1.83874093225775,-2.7550475106226,-2.55944768672544,2.12319514516607,3.18125484265547,-2.37380433681192,-1.19541898829252,-1.79113655913325,-1.66397142209237,1.38035094930968,2.06822634923254,-1.54327927740643,-3.34189684440345,-5.00727667287331,3.73635425856116,-1.66034116472922},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,2.35354378347051,3.52639397108079,3.27603094239261,1.99756895385148,-2.88249067692314,-4.31893293096971,-4.01230209521675,3.32841353768903,4.98707417885748,-3.72127946338712,1.0537965231098,1.57893884647547,1.46683908849365,-1.21681941252376,-1.82320153617302,1.36044546108577,-1.75917907531388,-2.6358372976047,1.9668219982996,-4.02051566640715},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,2.99666740648972,4.4900077703237,4.17123115221737,2.54341976480911,-3.67015303548283,-5.4991139896991,-5.10869396106863,4.23792768749944,6.34982988227715,-4.73814719602452,3.40958356039142,5.10869396106863,4.74599254428019,-3.93704797266052,-5.89901166751924,4.40175344784644,2.40062592164155,3.59693872593979,-2.68398137439783,.178027784911158},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,3.37371117926928,5.05494516239668,4.69605974259137,2.8634354535068,-4.13193546224797,-6.19101816335907,-5.75147508548906,4.77114810448709,7.14877200382768,-5.33430574565371,5.05405999869821,7.5726684320545,7.03503247458549,-5.83592580124217,-8.74416431584401,6.52476340023727,6.34960212623239,9.51382286438759,-7.09907099144838,7.15379270875832},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-3.37371117926928,-5.6909635577302,-6.96101658051265,-7.57372303615022,4.13193546224797,6.96997843122346,8.52546935665732,-4.77114810448709,-8.04823784418982,5.33430574565371,5.05405999869821,8.52546935665732,10.4280993821027,-5.83592580124217,-9.84436405649934,6.52476340023727,-6.34960212623238,-10.7108618346039,7.09907099144837,7.15379270875817},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-2.99666740648972,-5.05494516239668,-6.18305788326991,-6.72728866290429,3.67015303548283,6.19101816335907,7.5726684320545,-4.23792768749944,-7.14877200382768,4.73814719602452,3.40958356039142,5.75147508548906,7.03503247458549,-3.93704797266052,-6.64123137798004,4.40175344784644,-2.40062592164153,-4.04950925303536,2.68398137439781,.178027784911327},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-2.35354378347051,-3.97008848462071,-4.85609361001944,-5.28352541823676,2.88249067692314,4.86234551094043,5.94747574387061,-3.32841353768903,-5.61455297729064,3.72127946338712,1.05379652310981,1.77760255482498,2.17431033156193,-1.21681941252376,-2.05259862717067,1.36044546108577,1.7591790753139,2.96748105525688,-1.96682199829962,-4.02051566640726},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-1.50132568517008,-2.53251962263048,-3.09770233191153,-3.3703611015687,1.83874093225775,3.10169041978979,3.79389504410759,-2.12319514516607,-3.58152359635514,2.37380433681192,-1.19541898829253,-2.01649920176636,-2.46652157203873,1.38035094930968,2.32845271402296,-1.54327927740643,3.34189684440344,5.63729736986577,-3.73635425856115,-1.66034116472927},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-.515715838247201,-.869938143977334,-1.06408234437226,-1.1577425323902,.631620327664824,1.06545228035847,1.30322939400866,-.729332332662821,-1.23027832131385,.81541833693141,-2.55652048864375,-4.31248087502183,-5.27489775246548,2.95201558448519,4.97962398835612,-3.30045375807884,1.60938069663531,2.71479282295876,-1.79934231949427,2.78946777547416},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,.515715838247201,.869938143977334,1.06408234437226,1.1577425323902,-.631620327664824,-1.06545228035847,-1.30322939400866,.729332332662821,1.23027832131385,-.81541833693141,-2.55652048864375,-4.31248087502184,-5.27489775246549,2.95201558448519,4.97962398835612,-3.30045375807885,-1.60938069663533,-2.71479282295879,1.79934231949429,2.78946777547417},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,1.50132568517008,2.53251962263048,3.09770233191153,3.3703611015687,-1.83874093225775,-3.10169041978979,-3.79389504410759,2.12319514516607,3.58152359635514,-2.37380433681192,-1.19541898829252,-2.01649920176635,-2.46652157203872,1.38035094930968,2.32845271402296,-1.54327927740643,-3.34189684440345,-5.63729736986578,3.73635425856116,-1.66034116472922},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,2.35354378347051,3.97008848462071,4.85609361001944,5.28352541823676,-2.88249067692314,-4.86234551094043,-5.94747574387061,3.32841353768903,5.61455297729064,-3.72127946338712,1.0537965231098,1.77760255482497,2.17431033156192,-1.21681941252376,-2.05259862717066,1.36044546108577,-1.75917907531388,-2.96748105525684,1.9668219982996,-4.02051566640715},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,2.99666740648972,5.05494516239668,6.18305788326991,6.72728866290429,-3.67015303548283,-6.19101816335907,-7.5726684320545,4.23792768749944,7.14877200382768,-4.73814719602452,3.40958356039142,5.75147508548906,7.03503247458549,-3.93704797266052,-6.64123137798004,4.40175344784644,2.40062592164155,4.0495092530354,-2.68398137439783,.178027784911158},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,3.37371117926928,5.6909635577302,6.96101658051265,7.57372303615022,-4.13193546224797,-6.96997843122346,-8.52546935665732,4.77114810448709,8.04823784418982,-5.33430574565371,5.05405999869821,8.52546935665732,10.4280993821027,-5.83592580124217,-9.84436405649934,6.52476340023727,6.34960212623239,10.7108618346039,-7.09907099144838,7.15379270875832}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_right_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.92171958555151,-6.61537460386721,8.09172854779269,-8.80395991076744,6.96914454009643,-11.7559404153652,14.3795150585562,10.2353232484353,-17.2655122351262,13.4470270492568,-.0451952183653594,.0762378066864121,-.0932518071891671,.101459801467564,-.217799593005347,.367396460919218,-.449388373324038,-.614368644454251,1.03635118137081,-1.33333501160483,.000932318622845132,-.00157268687980642,.00192366360061384,.00636097475878108,-.0107300458251741,.0244039729141487,-1.66187980893617e-005,2.80335124396135e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.59519010006696,-6.06456651609707,7.41799655299993,-8.07092624092507,5.47681396366563,-9.23859424244377,11.300372436005,6.3309248170395,-10.6793559162843,5.76936797267538,-.233717104755142,.394247004353173,-.482231155810537,.524676988090528,-1.04841335581819,1.76852192969244,-2.1632031816091,-2.66785366107377,4.50028385780377,-5.04596922575242,.0249321870529788,-.0420569990798227,.0514428646412204,.159138742203809,-.268444076758865,.559597243722216,-.00229823078236368,.0038767834400574,-.0189957623925199,.00019657159111697},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.03822870489707,-5.12505307344404,6.26881178264598,-6.82059059403531,3.26615690329693,-5.50953502563763,6.7390986231311,1.66498726415444,-2.80859307349989,-.923099517802526,-.555278916264744,.936675343241156,-1.14571329243577,1.24655861042429,-2.17523616991868,3.66930929238228,-4.48818948881768,-4.50024407433993,7.59126186695628,-6.26847807110811,.140734880491405,-.237399419743997,.29037987690505,.792696592497175,-1.33716467766395,2.35258131414951,-.0308216365902829,.0519916499446519,-.225963663364185,.00626330408481366},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,2.30018618238774,-3.88008191896113,4.74600026627036,-5.1637416942433,.987081447023724,-1.66506385533551,2.03665635715207,-1.35561938023839,2.28673412920868,-2.20870753342451,-.981387965414959,1.65545977423627,-2.02490893148973,2.20314076875832,-3.10522366683209,5.23806390000324,-6.40704325104372,-4.42237267757001,7.45990406612127,-2.83495589139377,.439603191878621,-.741547101224613,.907038257318,2.03901059601619,-3.43951642050874,4.55407145300733,-.170154848292169,.287026656838231,-1.03685515398483,.0611113213546687},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,1.44662301694489,-2.44024412222762,2.98483369571688,-3.24755780451478,-.724355119014703,1.22188248065577,-1.49457014152702,-1.6918152058931,2.85384793698481,.160475142667942,-1.4741928888764,2.48675051356285,-3.04171892525937,3.30945005334757,-3.38025060487184,5.70199462778367,-6.97450945518368,-2.16048862424162,3.64443230995284,1.70770592621981,.991945941760336,-1.67326955598402,2.04669332477546,3.46032388717692,-5.8370666898335,4.72311662432345,-.576746649127296,.972888308521261,-2.68886700736678,.311154445333742},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,.553376983055106,-.93346705704166,1.14178901214037,-1.24228891639978,-1.45368745167752,2.45216080196962,-2.99940982448642,-.194424990185747,.327966881493408,1.49978166741586,-1.9899087271236,3.35668865754018,-4.10580126963162,4.46719258573777,-2.74863027720701,4.6365423474252,-5.67128006117502,.854309601049289,-1.44109692494208,2.52312426315122,1.80736427869175,-3.04876253512471,3.72915523810848,4.12997714955087,-6.96667503835439,1.65834138994758,-1.41847476579742,2.39276208655033,-4.48820932686106,1.03297875354965},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-.300186182387741,.506370739691845,-.619377558413115,.673894973328731,-1.13611369814234,1.91645974101964,-2.34415629302528,1.22655366483195,-2.06901890572263,.100858250439743,-2.48271365058504,4.18797939686675,-5.12261126340126,5.57350187032702,-1.26648273457434,2.13637348021345,-2.61314820693614,2.5074905585056,-4.22977446202813,-.461151554581851,2.81340752869054,-4.7458122143026,5.80493569904152,3.19379348794833,-5.38746839617942,-2.33714259619681,-2.75487590413568,4.64707781591113,-4.77320941254596,2.50302559510192},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.03822870489707,1.75134189417476,-2.14218907478874,2.33074387312075,-.0622566343920988,.105017951653011,-.12845482061473,1.10967175893133,-1.87185600940662,-1.41978880536247,-2.90882269973526,4.90676382786187,-6.00180690245521,6.53008402866105,.707254507004454,-1.19303621855815,1.45928625505293,1.57948219926848,-2.66435837496252,-2.54719860772099,3.86201434387853,-6.51466048131367,7.96853804732135,.544351948717181,-.918243127233759,-3.18953813043078,-4.43071582310304,7.47397774950175,-2.19278566166381,4.71658607960963},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.59519010006696,2.69085533682779,-3.29137384514269,3.5810795200105,1.2388862761662,-2.0898222386161,2.55620812016103,-.514957807193037,.868659455579414,-.353775594121333,-3.23038451124486,5.44919216674985,-6.66528903908045,7.25196565099481,2.62173967966463,-4.42249623366662,5.4094652504454,-1.30644820527577,2.20378945631257,-.307822029727901,4.7630793830775,-8.03462707886587,9.82771577392637,-2.90243304119458,4.89598539807521,.518849452524734,-6.0685481480974,10.2367643607743,2.66498561200528,7.17423567654393},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.92171958555151,3.24166342459793,-3.96510583993544,4.31411318985288,2.19799643560933,-3.70770257117539,4.53515100205683,-2.29634628619806,3.87360456901059,2.23836106649518,-3.41890639763464,5.76720136441661,-7.05426838770182,7.67518283761778,3.91413586924262,-6.60258197030424,8.07608098333328,-4.09536544265748,6.90829008438974,4.00097073404889,5.33523806427656,-8.99977614812197,11.0082572774727,-5.59797201662203,9.44297038492046,5.49216459339238,-7.19422103343416,12.1356119587979,7.09892453011937,9.00134910987265},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.92171958555151,-5.87604462970729,5.45886369350783,-3.32855727881992,6.96914454009643,-10.4421041409331,9.70074715430413,10.2353232484353,-15.3359297883224,13.4470270492568,-.0451952183653594,.067717518868618,-.0629097851728847,.0383594160020481,-.217799593005347,.326336470786676,-.303167594035556,-.614368644454251,.920529245649199,-1.33333501160483,.000932318622845132,-.00139692441436278,.00129774711619287,.00636097475878108,-.0095308628646168,.0244039729141487,-1.66187980893617e-005,2.49005052754424e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.59519010006696,-5.38679449650258,5.00434880168882,-3.05141556280758,5.47681396366563,-8.20609494322841,7.62348767011616,6.3309248170395,-9.48583802706066,5.76936797267538,-.233717104755142,.350186214905109,-.325324080359881,.198367260350894,-1.04841335581819,1.57087306611691,-1.45934595309525,-2.66785366107377,3.99733505330865,-5.04596922575242,.0249321870529788,-.0373567361461481,.0347045238135922,.159138742203809,-.238442940887059,.559597243722216,-.00229823078236368,.00344351663776842,-.0189957623925199,.00019657159111697},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.03822870489707,-4.55228046671326,4.22908267863918,-2.57869489385804,3.26615690329693,-4.89379296535329,4.54634885284358,1.66498726415444,-2.49470653361407,-.923099517802526,-.555278916264744,.831993114526565,-.772924185272261,.471292665829703,-2.17523616991868,3.25922966624351,-3.02783447385586,-4.50024407433993,6.74286736535737,-6.26847807110811,.140734880491405,-.210867814600683,.195896854176436,.792696592497175,-1.18772402074228,2.35258131414951,-.0308216365902829,.0461810968751945,-.225963663364185,.00626330408481366},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,2.30018618238774,-3.44644648080968,3.20175947449313,-1.95228172055831,.987081447023724,-1.47897739903412,1.37397459375255,-1.35561938023839,2.03117020639316,-2.20870753342451,-.981387965414959,1.47044666380812,-1.36604951383154,.832952480070547,-3.10522366683209,4.65266127660931,-4.32233676392885,-4.42237267757001,6.6261900271284,-2.83495589139377,.439603191878621,-.658672278252945,.611908590636856,2.03901059601619,-3.05511829665785,4.55407145300733,-.170154848292169,.254948743871766,-1.03685515398483,.0611113213546687},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,1.44662301694489,-2.16752402217829,2.01363654211465,-1.22782046694529,-.724355119014703,1.08532568728512,-1.00827093182807,-1.6918152058931,2.53490374291394,.160475142667942,-1.4741928888764,2.20883288938782,-2.05201260878721,1.25122037987572,-3.38025060487184,5.06474340721712,-4.70516234202085,-2.16048862424162,3.23713291923203,1.70770592621981,.991945941760336,-1.48626603590141,1.38074575986725,3.46032388717692,-5.18471990323707,4.72311662432345,-.576746649127296,.864158942299273,-2.68886700736678,.311154445333742},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,.553376983055106,-.829143384311425,.770276776736346,-.469678401195645,-1.45368745167752,2.17810890152299,-2.02346992933989,-.194424990185747,.291313515578485,1.49978166741586,-1.9899087271236,2.9815473107461,-2.76986670414993,1.68893390563516,-2.74863027720701,4.11836538252447,-3.82597422031764,.854309601049289,-1.2800408674882,2.52312426315122,1.80736427869175,-2.70803481201197,2.51577274454146,4.12997714955087,-6.18808395553405,1.65834138994758,-1.41847476579742,2.12534854800545,-4.48820932686106,1.03297875354965},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-.300186182387741,.449779074319961,-.417846155642134,.25478285241737,-1.13611369814234,1.70227744362135,-1.58142102819233,1.22655366483195,-1.83778669504629,.100858250439743,-2.48271365058504,3.7199335363258,-3.45582979910559,2.10720180544033,-1.26648273457434,1.89761376598671,-1.76288907720341,2.5074905585056,-3.75705761211831,-.461151554581851,2.81340752869054,-4.21542331996591,3.91614134649752,3.19379348794833,-4.78536842321548,-2.33714259619681,-2.75487590413568,4.12772341388705,-4.77320941254596,2.50302559510192},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.03822870489707,1.55561306022354,-1.44516935978819,.881196025717106,-.0622566343920988,.0932812135041899,-.0866585368463816,1.10967175893133,-1.66265859611778,-1.41978880536247,-2.90882269973526,4.35838708560735,-4.04895512766487,2.46886161968117,.707254507004454,-1.0597032647262,.984467621360889,1.57948219926848,-2.36659141141637,-2.54719860772099,3.86201434387853,-5.78658625215441,5.37575658648661,.544351948717181,-.815620871022689,-3.18953813043078,-4.43071582310304,6.63869084478425,-2.19278566166381,4.71658607960963},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.59519010006696,2.39012709001286,-2.22043548283783,1.35391669466664,1.2388862761662,-1.85626506095126,1.72447600259692,-.514957807193037,.771578638346196,-.353775594121333,-3.23038451124486,4.84019398522881,-4.49655523257725,2.74178702515998,2.62173967966463,-3.92824092358218,3.64934800797338,-1.30644820527577,1.95749537633152,-.307822029727901,4.7630793830775,-7.13668236880667,6.6300000687783,-2.90243304119458,4.3488132461414,.518849452524734,-6.0685481480974,9.09271021667365,2.66498561200528,7.17423567654393},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.92171958555151,2.87937722321757,-2.67495037465683,1.63105841067898,2.19799643560933,-3.29333213710544,3.05951577632408,-2.29634628619806,3.44069303528758,2.23836106649518,-3.41890639763464,5.1226626812653,-4.75896952776425,2.90179486950883,3.91413586924262,-5.86468169257239,5.4483074914535,-4.09536544265748,6.13622406614863,4.00097073404889,5.33523806427656,-7.99396700420041,7.42642015557747,-5.59797201662203,8.38763014060611,5.49216459339238,-7.19422103343416,10.7793438389747,7.09892453011937,9.00134910987265},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.92171958555151,-4.61496937554466,1.6871654495215,2.43916732844399,6.96914454009643,-8.20109339929485,2.99820005631406,10.2353232484353,-12.0446407028288,13.4470270492568,-.0451952183653594,.0531844625867158,-.0194434633176024,-.028109786436831,-.217799593005347,.256300439106579,-.0936996999356596,-.614368644454251,.722971751522571,-1.33333501160483,.000932318622845132,-.00109712634918432,.000401093381258216,.00636097475878108,-.00748541629863581,.0244039729141487,-1.66187980893617e-005,1.95565344587043e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.59519010006696,-4.23071865520366,1.54668899419582,2.23607783277941,5.47681396366563,-6.44496072482988,2.35618358032458,6.3309248170395,-7.45005437415178,5.76936797267538,-.233717104755142,.275031719356679,-.100547582628473,-.145363561432352,-1.04841335581819,1.23374336817428,-.451038569416845,-2.66785366107377,3.13945519772704,-5.04596922575242,.0249321870529788,-.0293394969148428,.0107260918769642,.159138742203809,-.187269998673138,.559597243722216,-.00229823078236368,.00270449338439953,-.0189957623925199,.00019657159111697},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.03822870489707,-3.57530214058608,1.30707828207098,1.88966804781976,3.26615690329693,-3.843521635486,1.40513541583971,1.66498726415444,-1.95931021275678,-.923099517802526,-.555278916264744,.653436620408273,-.238886891797977,-.345363344035539,-2.17523616991868,2.55975678291091,-.935809721212703,-4.50024407433993,5.29576072944733,-6.26847807110811,.140734880491405,-.165612851500527,.0605456414728777,.792696592497175,-.932823068533675,2.35258131414951,-.0308216365902829,.0362700355832219,-.225963663364185,.00626330408481366},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,2.30018618238774,-2.70679444519172,.989564544259887,1.4306323700677,.987081447023724,-1.16156970101249,.42465293021633,-1.35561938023839,1.5952547827211,-2.20870753342451,-.981387965414959,1.15486977201266,-.422203533825773,-.610387716162354,-3.10522366683209,3.65413992905702,-1.33590022736237,-4.42237267757001,5.20412385861342,-2.83495589139377,.439603191878621,-.517312679568432,.189121965733585,2.03901059601619,-2.39945035585009,4.55407145300733,-.170154848292169,.200233442638163,-1.03685515398483,.0611113213546687},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,1.44662301694489,-1.70234530427801,.622352597994011,.899747042727645,-.724355119014703,.852400743178548,-.311625271383912,-1.6918152058931,1.9908805805112,.160475142667942,-1.4741928888764,1.73478875376213,-.634213449888056,-.916894502821395,-3.38025060487184,3.97778389924568,-1.45421973940992,-2.16048862424162,2.54240228389271,1.70770592621981,.991945941760336,-1.16729410204273,.426745687640423,3.46032388717692,-4.07201188589473,4.72311662432345,-.576746649127296,.678699245092846,-2.68886700736678,.311154445333742},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,.553376983055106,-.651198479192506,.238068659934471,.344180410642771,-1.45368745167752,1.71065853147384,-.625391792982139,-.194424990185747,.228793863525153,1.49978166741586,-1.9899087271236,2.34166865603201,-.856079885009688,-1.23765104741929,-2.74863027720701,3.23451085132423,-1.18248996086068,.854309601049289,-1.00532752474992,2.52312426315122,1.80736427869175,-2.12685548066823,.77754732335343,4.12997714955087,-4.86004102210358,1.65834138994758,-1.41847476579742,1.66922123290477,-4.48820932686106,1.03297875354965},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-.300186182387741,.353250661721207,-.129143286331405,-.186704916697285,-1.13611369814234,1.33694666498547,-.488768188714079,1.22655366483195,-1.44337387662839,.100858250439743,-2.48271365058504,2.92158763778148,-1.06808980107197,-1.54415783407833,-1.26648273457434,1.49036128356923,-.544854334050093,2.5074905585056,-2.95074440626681,-.461151554581851,2.81340752869054,-3.31073889879025,1.21035782283977,3.19379348794833,-3.75836640385388,-2.33714259619681,-2.75487590413568,3.2418605279988,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.03822870489707,1.22175835711557,-.446657024142499,-.64574059444934,-.0622566343920988,.0732618573826121,-.0267834658423847,1.10967175893133,-1.30583053509787,-1.41978880536247,-2.90882269973526,3.42302078938587,-1.25140644309977,-1.80918220620515,.707254507004454,-.832277224231996,.304268406428259,1.57948219926848,-1.8586902544778,-2.54719860772099,3.86201434387853,-4.54470992402429,1.66147962048301,.544351948717181,-.640578072330311,-3.18953813043078,-4.43071582310304,5.21394183895335,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.59519010006696,1.87717487173315,-.686267736267336,-.992150379408995,1.2388862761662,-1.4578865459724,.532982044151549,-.514957807193037,.605987872997064,-.353775594121333,-3.23038451124486,3.80142569043747,-1.38974575226927,-2.00918198880834,2.62173967966463,-3.08518956279543,1.12790027705863,-1.30644820527577,1.53739152505215,-.307822029727901,4.7630793830775,-5.60505793446857,2.0491273778591,-2.90243304119458,3.41550161982096,.518849452524734,-6.0685481480974,7.14129688166464,2.66498561200528,7.17423567654393},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.92171958555151,2.26142559207414,-.826744191593017,-1.19523987507357,2.19799643560933,-2.5865404220042,.945601429143385,-2.29634628619806,2.70227576356801,2.23836106649518,-3.41890639763464,4.02327294720743,-1.47084987158014,-2.12643576380386,3.91413586924262,-4.60604507183384,1.68390285488932,-4.09536544265748,4.81931093640565,4.00097073404889,5.33523806427656,-6.27835818791885,2.29527612404422,-5.59797201662203,6.58753611853026,5.49216459339238,-7.19422103343416,8.46595709192435,7.09892453011937,9.00134910987265},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.92171958555151,-2.94388917191153,-1.91390801783746,4.63366448151074,6.96914454009643,-5.23147784770941,-3.40113598727833,10.2353232484353,-7.68327684524188,13.4470270492568,-.0451952183653594,.0339263710725032,.0220565211009044,-.0533999113042758,-.217799593005347,.163494061616168,.10629224716917,-.614368644454251,.461183712763862,-1.33333501160483,.000932318622845132,-.000699856947415033,-.000454997367631818,.00636097475878108,-.00477494739305899,.0244039729141487,-1.66187980893617e-005,1.24751142094631e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.59519010006696,-2.69877562014986,-1.7545525650326,4.24785717273984,5.47681396366563,-4.11124073580283,-2.67283724139748,6.3309248170395,-4.75239001984257,5.76936797267538,-.233717104755142,.175442746128902,.114060434690259,-.276145864886064,-1.04841335581819,.787004949964004,.511654819985594,-2.66785366107377,2.00265861161846,-5.04596922575242,.0249321870529788,-.0187156663995725,-.0121675993597458,.159138742203809,-.119459540563619,.559597243722216,-.00229823078236368,.00172519645136334,-.0189957623925199,.00019657159111697},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.03822870489707,-2.2806853960415,-1.48273994391383,3.58978558499055,3.26615690329693,-2.45178262388351,-1.59397523182193,1.66498726415444,-1.24984407270733,-.923099517802526,-.555278916264744,.41682724950333,.270991524688991,-.656083673232097,-2.17523616991868,1.63286896677384,1.06157563213922,-4.50024407433993,3.3781660071394,-6.26847807110811,.140734880491405,-.105644445409619,-.0686825282564419,.792696592497175,-.595047877309364,2.35258131414951,-.0308216365902829,.0231366573273721,-.225963663364185,.00626330408481366},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,2.30018618238774,-1.72666429814602,-1.12255470615749,2.71775959032344,.987081447023724,-.740965364440283,-.481723145865576,-1.35561938023839,1.01761309753837,-2.20870753342451,-.981387965414959,.736691479430405,.47894456870109,-1.15954811601065,-3.10522366683209,2.33097602484761,1.51543534565912,-4.42237267757001,3.31971084311917,-2.83495589139377,.439603191878621,-.329993781528466,-.214538560322787,2.03901059601619,-1.53060948975261,4.55407145300733,-.170154848292169,.127728922051078,-1.03685515398483,.0611113213546687},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,1.44662301694489,-1.0859261460488,-.705992187998236,1.70924145531706,-.724355119014703,.543746472353678,.353505404956933,-1.6918152058931,1.26998281171566,.160475142667942,-1.4741928888764,1.10662182393172,.719447050735517,-1.74181633275903,-3.38025060487184,2.53742852796533,1.64965612574837,-2.16048862424162,1.62179853107076,1.70770592621981,.991945941760336,-.744616960114543,-.484097154422055,3.46032388717692,-2.5975365648804,4.72311662432345,-.576746649127296,.432942278924126,-2.68886700736678,.311154445333742},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,.553376983055106,-.41539953912128,-.270063328509738,.65383646518608,-1.45368745167752,1.09122915405036,.709439828333213,-.194424990185747,.145947615817496,1.49978166741586,-1.9899087271236,1.49375054084377,.971130695151503,-2.35115468793597,-2.74863027720701,2.06329461731692,1.34140787303666,.854309601049289,-.641298473185774,2.52312426315122,1.80736427869175,-1.35672120663226,-.88204393756181,4.12997714955087,-3.10022038625728,1.65834138994758,-1.41847476579742,1.06479629943226,-4.48820932686106,1.03297875354965},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-.300186182387741,.225338612975934,.146499189649512,-.354681669820299,-1.13611369814234,.852838337634571,.554455021295674,1.22655366483195,-.920728260721081,.100858250439743,-2.48271365058504,1.86368088534508,1.21163317718593,-2.93342290468435,-1.26648273457434,.950701529718779,.618078730089993,2.5074905585056,-1.88227998927126,-.461151554581851,2.81340752869054,-2.11192049221879,-1.37302096972311,3.19379348794833,-2.39746209780118,-2.33714259619681,-2.75487590413568,2.0679829763977,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.03822870489707,.779359710871422,.506684427405858,-1.22670766448742,-.0622566343920988,.0467337421144522,.030382965722629,1.10967175893133,-.832989356981887,-1.41978880536247,-2.90882269973526,2.18354511527215,1.41958622119803,-3.4368873474629,.707254507004454,-.530909678713955,-.345159831718241,1.57948219926848,-1.18565859680248,-2.54719860772099,3.86201434387853,-2.89907066463143,-1.88477020319366,.544351948717181,-.408624781106939,-3.18953813043078,-4.43071582310304,3.32597373321892,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.59519010006696,1.19744993497978,.77849704852463,-1.88477925223671,1.2388862761662,-.929985893147364,-.604610892164935,-.514957807193037,.386559691398865,-.353775594121333,-3.23038451124486,2.42492961864658,1.57651731119676,-3.81682515580894,2.62173967966463,-1.9680425606586,-1.27948173914766,-1.30644820527577,.980702122872796,-.307822029727901,4.7630793830775,-3.57546670811253,-2.32451495445634,-2.90243304119458,2.17874863666902,.518849452524734,-6.0685481480974,4.55543360151918,2.66498561200528,7.17423567654393},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.92171958555151,1.44256348674145,.937852501329485,-2.27058656100761,2.19799643560933,-1.64995425135427,-1.07268327325537,-2.29634628619806,1.72378183093533,2.23836106649518,-3.41890639763464,2.56644599370298,1.66852122478612,-4.03957110939073,3.91413586924262,-2.93819635817362,-1.91020695459719,-4.09536544265748,3.07423866276144,4.00097073404889,5.33523806427656,-4.00496497002541,-2.60374427309723,-5.59797201662203,4.20218958584716,5.49216459339238,-7.19422103343416,5.40043440913241,7.09892453011937,9.00134910987265},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.92171958555151,-1.01124645171658,-4.09307958874724,2.23146629547127,6.96914454009643,-1.7970491081004,-7.2736621396321,10.2353232484353,-2.63925915467191,13.4470270492568,-.0451952183653594,.0116539449562267,.0471700287973566,-.0257161697253048,-.217799593005347,.0561613498441344,.227316372195247,-.614368644454251,.158419820138511,-1.33333501160483,.000932318622845132,-.000240405739976654,-.000973056396143737,.00636097475878108,-.00164022771456063,.0244039729141487,-1.66187980893617e-005,4.28528869206282e-006,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.59519010006696,-.927048238057034,-3.75228235860234,2.04567036451771,5.47681396366563,-1.4122398512488,-5.7161240011779,6.3309248170395,-1.63247909961861,5.76936797267538,-.233717104755142,.060265806265752,.243929401394036,-.132985500479333,-1.04841335581819,.27034168634063,1.09422390242844,-2.66785366107377,.68792719315715,-5.04596922575242,.0249321870529788,-.00642896187079958,-.0260216019365979,.159138742203809,-.0410351849082052,.559597243722216,-.00229823078236368,.000592617006985495,-.0189957623925199,.00019657159111697},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.03822870489707,-.78343133166635,-3.17098446910282,1.72875821562793,3.26615690329693,-.842204422108959,-3.40887201763794,1.66498726415444,-.429330151346565,-.923099517802526,-.555278916264744,.143183065809956,.579541894432988,-.315954814956539,-2.17523616991868,.560901872435633,2.27028337376948,-4.50024407433993,1.16042357185991,-6.26847807110811,.140734880491405,-.0362896034209994,-.14688430786117,.792696592497175,-.204403093795768,2.35258131414951,-.0308216365902829,.0079476030721956,-.225963663364185,.00626330408481366},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,2.30018618238774,-.593121222587361,-2.40069309089209,1.30881054275714,.987081447023724,-.254526767782017,-1.03021208824726,-1.35561938023839,.349557192548049,-2.20870753342451,-.981387965414959,.253058658488868,1.02426982907298,-.558411716942304,-3.10522366683209,.800706513225493,3.2409067852655,-4.42237267757001,1.14034381554164,-2.83495589139377,.439603191878621,-.113355164264732,-.458811705720642,2.03901059601619,-.525775029251873,4.55407145300733,-.170154848292169,.0438757750930774,-1.03685515398483,.0611113213546687},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,1.44662301694489,-.373023200905715,-1.50983338153088,.823131392784558,-.724355119014703,.186780703583358,.756005902089247,-1.6918152058931,.436247948578379,.160475142667942,-1.4741928888764,.380132310523241,1.53860792217029,-.83881870492831,-3.38025060487184,.871624387178691,3.52795105816102,-2.16048862424162,.557099100602201,1.70770592621981,.991945941760336,-.255781116350547,-1.03528913786274,3.46032388717692,-.892271916858127,4.72311662432345,-.576746649127296,.148718690750112,-2.68886700736678,.311154445333742},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,.553376983055106,-.142692637341486,-.577556856071558,.314872611220466,-1.45368745167752,.374844821120283,1.51720649776892,-.194424990185747,.0501340234001082,1.49978166741586,-1.9899087271236,.513113723366546,2.07685802519487,-1.13226204929232,-2.74863027720701,.708756083794028,2.86873200495546,.854309601049289,-.220290495881434,2.52312426315122,1.80736427869175,-.466043191865359,-1.88633727617264,4.12997714955087,-1.06494731359279,1.65834138994758,-1.41847476579742,.365764951301699,-4.48820932686106,1.03297875354965},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-.300186182387741,.0774053843401605,.313302853289657,-.170806538752116,-1.13611369814234,.292955913914665,1.18575632130325,1.22655366483195,-.316276575739705,.100858250439743,-2.48271365058504,.640187375400919,2.59119611829218,-1.41266903727833,-1.26648273457434,.326572602577083,1.32182184869504,2.5074905585056,-.646576297249579,-.461151554581851,2.81340752869054,-.725459410782454,-2.93633970583683,3.19379348794833,-.823544942743957,-2.33714259619681,-2.75487590413568,.710366567819732,-4.77320941254596,2.50302559510192},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.03822870489707,.26771549341915,1.08359423150038,-.590754211622909,-.0622566343920988,.0160533661863334,.0649769454361842,1.10967175893133,-.286137650697852,-1.41978880536247,-2.90882269973526,.750062968079831,3.03592405293217,-1.65512593926409,.707254507004454,-.182371175485818,-.738158077031132,1.57948219926848,-.407281992951262,-2.54719860772099,3.86201434387853,-.995850982046505,-4.03076552074204,.544351948717181,-.140365460738301,-3.18953813043078,-4.43071582310304,1.14249516194805,-2.19278566166381,4.71658607960963},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.59519010006696,.411332399809833,1.66489212099991,-.907666360512686,1.2388862761662,-.31945663701093,-1.29301955934683,-.514957807193037,.132785948612982,-.353775594121333,-3.23038451124486,.832980227624035,3.37153654597113,-1.8380952537413,2.62173967966463,-.676036338352017,-2.73629690762123,-1.30644820527577,.336878015452627,-.307822029727901,4.7630793830775,-1.22819773798136,-4.97120270417874,-2.90243304119458,.748415344244522,.518849452524734,-6.0685481480974,1.56482319698725,2.66498561200528,7.17423567654393},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.92171958555151,.49553061346938,2.00568935114481,-1.09346229146625,2.19799643560933,-.566770786786546,-2.29403815127598,-2.29634628619806,.592131075007482,2.23836106649518,-3.41890639763464,.881592088933561,3.56829591856781,-1.94536458449533,3.91413586924262,-1.00929093051433,-4.08516450282659,-4.09536544265748,1.05602241045943,4.00097073404889,5.33523806427656,-1.37573338488067,-5.56836192711791,-5.59797201662203,1.44348141522225,5.49216459339238,-7.19422103343416,1.855086864706,7.09892453011937,9.00134910987265},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.92171958555151,1.01124645171658,-4.09307958874725,-2.23146629547129,6.96914454009643,1.7970491081004,-7.2736621396321,10.2353232484353,2.63925915467191,13.4470270492568,-.0451952183653594,-.0116539449562267,.0471700287973566,.025716169725305,-.217799593005347,-.0561613498441344,.227316372195247,-.614368644454251,-.158419820138511,-1.33333501160483,.000932318622845132,.000240405739976654,-.000973056396143738,.00636097475878108,.00164022771456063,.0244039729141487,-1.66187980893617e-005,-4.28528869206282e-006,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.59519010006696,.927048238057034,-3.75228235860234,-2.04567036451773,5.47681396366563,1.4122398512488,-5.7161240011779,6.3309248170395,1.63247909961861,5.76936797267538,-.233717104755142,-.060265806265752,.243929401394036,.132985500479334,-1.04841335581819,-.27034168634063,1.09422390242844,-2.66785366107377,-.68792719315715,-5.04596922575242,.0249321870529788,.00642896187079958,-.0260216019365979,.159138742203809,.0410351849082052,.559597243722216,-.00229823078236368,-.000592617006985495,-.0189957623925199,.00019657159111697},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.03822870489707,.78343133166635,-3.17098446910282,-1.72875821562795,3.26615690329693,.842204422108959,-3.40887201763794,1.66498726415444,.429330151346565,-.923099517802526,-.555278916264744,-.143183065809956,.579541894432988,.315954814956541,-2.17523616991868,-.560901872435633,2.27028337376948,-4.50024407433993,-1.16042357185991,-6.26847807110811,.140734880491405,.0362896034209994,-.14688430786117,.792696592497175,.204403093795768,2.35258131414951,-.0308216365902829,-.0079476030721956,-.225963663364185,.00626330408481366},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,2.30018618238774,.593121222587361,-2.40069309089209,-1.30881054275715,.987081447023724,.254526767782017,-1.03021208824726,-1.35561938023839,-.349557192548049,-2.20870753342451,-.981387965414959,-.253058658488868,1.02426982907298,.558411716942308,-3.10522366683209,-.800706513225493,3.2409067852655,-4.42237267757001,-1.14034381554164,-2.83495589139377,.439603191878621,.113355164264732,-.458811705720642,2.03901059601619,.525775029251873,4.55407145300733,-.170154848292169,-.0438757750930774,-1.03685515398483,.0611113213546687},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,1.44662301694489,.373023200905715,-1.50983338153088,-.823131392784565,-.724355119014703,-.186780703583358,.756005902089247,-1.6918152058931,-.436247948578379,.160475142667942,-1.4741928888764,-.380132310523241,1.53860792217029,.838818704928317,-3.38025060487184,-.871624387178691,3.52795105816103,-2.16048862424162,-.557099100602201,1.70770592621981,.991945941760336,.255781116350547,-1.03528913786275,3.46032388717692,.892271916858127,4.72311662432345,-.576746649127296,-.148718690750112,-2.68886700736678,.311154445333742},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,.553376983055106,.142692637341486,-.577556856071558,-.314872611220468,-1.45368745167752,-.374844821120283,1.51720649776892,-.194424990185747,-.0501340234001082,1.49978166741586,-1.9899087271236,-.513113723366546,2.07685802519487,1.13226204929233,-2.74863027720701,-.708756083794028,2.86873200495547,.854309601049289,.220290495881434,2.52312426315122,1.80736427869175,.466043191865359,-1.88633727617264,4.12997714955087,1.06494731359279,1.65834138994758,-1.41847476579742,-.365764951301699,-4.48820932686106,1.03297875354965},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-.300186182387741,-.0774053843401605,.313302853289658,.170806538752117,-1.13611369814234,-.292955913914665,1.18575632130325,1.22655366483195,.316276575739705,.100858250439743,-2.48271365058504,-.640187375400919,2.59119611829218,1.41266903727834,-1.26648273457434,-.326572602577083,1.32182184869504,2.5074905585056,.646576297249579,-.461151554581851,2.81340752869054,.725459410782454,-2.93633970583683,3.19379348794833,.823544942743957,-2.33714259619681,-2.75487590413568,-.710366567819732,-4.77320941254596,2.50302559510192},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.03822870489707,-.26771549341915,1.08359423150038,.590754211622914,-.0622566343920988,-.0160533661863334,.0649769454361843,1.10967175893133,.286137650697852,-1.41978880536247,-2.90882269973526,-.750062968079831,3.03592405293218,1.65512593926411,.707254507004454,.182371175485818,-.738158077031133,1.57948219926848,.407281992951262,-2.54719860772099,3.86201434387853,.995850982046505,-4.03076552074204,.544351948717181,.140365460738301,-3.18953813043078,-4.43071582310304,-1.14249516194805,-2.19278566166381,4.71658607960963},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.59519010006696,-.411332399809833,1.66489212099991,.907666360512694,1.2388862761662,.31945663701093,-1.29301955934683,-.514957807193037,-.132785948612982,-.353775594121333,-3.23038451124486,-.832980227624035,3.37153654597113,1.83809525374131,2.62173967966463,.676036338352017,-2.73629690762123,-1.30644820527577,-.336878015452627,-.307822029727901,4.7630793830775,1.22819773798136,-4.97120270417875,-2.90243304119458,-.748415344244522,.518849452524734,-6.0685481480974,-1.56482319698725,2.66498561200528,7.17423567654393},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.92171958555151,-.49553061346938,2.00568935114481,1.09346229146626,2.19799643560933,.566770786786546,-2.29403815127598,-2.29634628619806,-.592131075007482,2.23836106649518,-3.41890639763464,-.881592088933561,3.56829591856781,1.94536458449534,3.91413586924262,1.00929093051433,-4.08516450282659,-4.09536544265748,-1.05602241045943,4.00097073404889,5.33523806427656,1.37573338488067,-5.56836192711792,-5.59797201662203,-1.44348141522225,5.49216459339238,-7.19422103343416,-1.855086864706,7.09892453011937,9.00134910987265},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.92171958555151,2.94388917191153,-1.91390801783745,-4.63366448151075,6.96914454009643,5.23147784770941,-3.40113598727832,10.2353232484353,7.68327684524188,13.4470270492568,-.0451952183653594,-.0339263710725032,.0220565211009043,.0533999113042759,-.217799593005347,-.163494061616168,.10629224716917,-.614368644454251,-.461183712763862,-1.33333501160483,.000932318622845132,.000699856947415033,-.000454997367631817,.00636097475878108,.00477494739305899,.0244039729141487,-1.66187980893617e-005,-1.24751142094631e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.59519010006696,2.69877562014986,-1.7545525650326,-4.24785717273985,5.47681396366563,4.11124073580283,-2.67283724139748,6.3309248170395,4.75239001984257,5.76936797267538,-.233717104755142,-.175442746128902,.114060434690259,.276145864886065,-1.04841335581819,-.787004949964004,.511654819985593,-2.66785366107377,-2.00265861161846,-5.04596922575242,.0249321870529788,.0187156663995725,-.0121675993597457,.159138742203809,.119459540563619,.559597243722216,-.00229823078236368,-.00172519645136334,-.0189957623925199,.00019657159111697},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.03822870489707,2.2806853960415,-1.48273994391383,-3.58978558499056,3.26615690329693,2.45178262388351,-1.59397523182193,1.66498726415444,1.24984407270733,-.923099517802526,-.555278916264744,-.41682724950333,.270991524688991,.656083673232099,-2.17523616991868,-1.63286896677384,1.06157563213921,-4.50024407433993,-3.3781660071394,-6.26847807110811,.140734880491405,.105644445409619,-.0686825282564417,.792696592497175,.595047877309364,2.35258131414951,-.0308216365902829,-.0231366573273721,-.225963663364185,.00626330408481366},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,2.30018618238774,1.72666429814602,-1.12255470615748,-2.71775959032344,.987081447023724,.740965364440283,-.481723145865575,-1.35561938023839,-1.01761309753837,-2.20870753342451,-.981387965414959,-.736691479430405,.478944568701089,1.15954811601066,-3.10522366683209,-2.33097602484761,1.51543534565911,-4.42237267757001,-3.31971084311917,-2.83495589139377,.439603191878621,.329993781528466,-.214538560322786,2.03901059601619,1.53060948975261,4.55407145300733,-.170154848292169,-.127728922051078,-1.03685515398483,.0611113213546687},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,1.44662301694489,1.0859261460488,-.705992187998234,-1.70924145531706,-.724355119014703,-.543746472353678,.353505404956932,-1.6918152058931,-1.26998281171566,.160475142667942,-1.4741928888764,-1.10662182393172,.719447050735515,1.74181633275904,-3.38025060487184,-2.53742852796533,1.64965612574837,-2.16048862424162,-1.62179853107076,1.70770592621981,.991945941760336,.744616960114543,-.484097154422054,3.46032388717692,2.5975365648804,4.72311662432345,-.576746649127296,-.432942278924126,-2.68886700736678,.311154445333742},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,.553376983055106,.41539953912128,-.270063328509737,-.653836465186081,-1.45368745167752,-1.09122915405036,.709439828333211,-.194424990185747,-.145947615817496,1.49978166741586,-1.9899087271236,-1.49375054084377,.971130695151501,2.35115468793598,-2.74863027720701,-2.06329461731692,1.34140787303665,.854309601049289,.641298473185774,2.52312426315122,1.80736427869175,1.35672120663226,-.882043937561808,4.12997714955087,3.10022038625728,1.65834138994758,-1.41847476579742,-1.06479629943226,-4.48820932686106,1.03297875354965},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-.300186182387741,-.225338612975934,.146499189649511,.3546816698203,-1.13611369814234,-.852838337634571,.554455021295673,1.22655366483195,.920728260721081,.100858250439743,-2.48271365058504,-1.86368088534508,1.21163317718593,2.93342290468436,-1.26648273457434,-.950701529718779,.618078730089992,2.5074905585056,1.88227998927126,-.461151554581851,2.81340752869054,2.11192049221879,-1.3730209697231,3.19379348794833,2.39746209780118,-2.33714259619681,-2.75487590413568,-2.0679829763977,-4.77320941254596,2.50302559510192},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.03822870489707,-.779359710871422,.506684427405857,1.22670766448742,-.0622566343920988,-.0467337421144522,.030382965722629,1.10967175893133,.832989356981887,-1.41978880536247,-2.90882269973526,-2.18354511527215,1.41958622119803,3.43688734746291,.707254507004454,.530909678713955,-.34515983171824,1.57948219926848,1.18565859680248,-2.54719860772099,3.86201434387853,2.89907066463143,-1.88477020319366,.544351948717181,.408624781106939,-3.18953813043078,-4.43071582310304,-3.32597373321892,-2.19278566166381,4.71658607960963},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.59519010006696,-1.19744993497978,.778497048524628,1.88477925223671,1.2388862761662,.929985893147364,-.604610892164934,-.514957807193037,-.386559691398865,-.353775594121333,-3.23038451124486,-2.42492961864658,1.57651731119676,3.81682515580895,2.62173967966463,1.9680425606586,-1.27948173914766,-1.30644820527577,-.980702122872796,-.307822029727901,4.7630793830775,3.57546670811253,-2.32451495445633,-2.90243304119458,-2.17874863666902,.518849452524734,-6.0685481480974,-4.55543360151918,2.66498561200528,7.17423567654393},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.92171958555151,-1.44256348674145,.937852501329483,2.27058656100761,2.19799643560933,1.64995425135427,-1.07268327325536,-2.29634628619806,-1.72378183093533,2.23836106649518,-3.41890639763464,-2.56644599370298,1.66852122478611,4.03957110939074,3.91413586924262,2.93819635817362,-1.91020695459718,-4.09536544265748,-3.07423866276144,4.00097073404889,5.33523806427656,4.00496497002541,-2.60374427309723,-5.59797201662203,-4.20218958584716,5.49216459339238,-7.19422103343416,-5.40043440913241,7.09892453011937,9.00134910987265},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.92171958555151,4.61496937554466,1.68716544952149,-2.43916732844396,6.96914454009643,8.20109339929485,2.99820005631404,10.2353232484353,12.0446407028288,13.4470270492568,-.0451952183653594,-.0531844625867158,-.0194434633176023,.0281097864368307,-.217799593005347,-.256300439106579,-.0936996999356592,-.614368644454251,-.722971751522571,-1.33333501160483,.000932318622845132,.00109712634918432,.000401093381258214,.00636097475878108,.00748541629863581,.0244039729141487,-1.66187980893617e-005,-1.95565344587043e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.59519010006696,4.23071865520366,1.54668899419581,-2.23607783277939,5.47681396366563,6.44496072482988,2.35618358032457,6.3309248170395,7.45005437415178,5.76936797267538,-.233717104755142,-.275031719356679,-.100547582628472,.145363561432351,-1.04841335581819,-1.23374336817428,-.451038569416842,-2.66785366107377,-3.13945519772704,-5.04596922575242,.0249321870529788,.0293394969148428,.0107260918769641,.159138742203809,.187269998673138,.559597243722216,-.00229823078236368,-.00270449338439953,-.0189957623925199,.00019657159111697},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.03822870489707,3.57530214058608,1.30707828207097,-1.88966804781974,3.26615690329693,3.843521635486,1.4051354158397,1.66498726415444,1.95931021275678,-.923099517802526,-.555278916264744,-.653436620408273,-.238886891797976,.345363344035536,-2.17523616991868,-2.55975678291091,-.935809721212699,-4.50024407433993,-5.29576072944733,-6.26847807110811,.140734880491405,.165612851500527,.0605456414728774,.792696592497175,.932823068533675,2.35258131414951,-.0308216365902829,-.0362700355832219,-.225963663364185,.00626330408481366},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,2.30018618238774,2.70679444519172,.989564544259882,-1.43063237006769,.987081447023724,1.16156970101249,.424652930216328,-1.35561938023839,-1.5952547827211,-2.20870753342451,-.981387965414959,-1.15486977201266,-.422203533825771,.610387716162348,-3.10522366683209,-3.65413992905702,-1.33590022736237,-4.42237267757001,-5.20412385861342,-2.83495589139377,.439603191878621,.517312679568432,.189121965733584,2.03901059601619,2.39945035585009,4.55407145300733,-.170154848292169,-.200233442638163,-1.03685515398483,.0611113213546687},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,1.44662301694489,1.70234530427801,.622352597994008,-.899747042727636,-.724355119014703,-.852400743178548,-.31162527138391,-1.6918152058931,-1.9908805805112,.160475142667942,-1.4741928888764,-1.73478875376213,-.634213449888053,.916894502821386,-3.38025060487184,-3.97778389924568,-1.45421973940992,-2.16048862424162,-2.54240228389271,1.70770592621981,.991945941760336,1.16729410204273,.426745687640421,3.46032388717692,4.07201188589473,4.72311662432345,-.576746649127296,-.678699245092846,-2.68886700736678,.311154445333742},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,.553376983055106,.651198479192506,.23806865993447,-.344180410642768,-1.45368745167752,-1.71065853147384,-.625391792982136,-.194424990185747,-.228793863525153,1.49978166741586,-1.9899087271236,-2.34166865603201,-.856079885009684,1.23765104741928,-2.74863027720701,-3.23451085132423,-1.18248996086068,.854309601049289,1.00532752474992,2.52312426315122,1.80736427869175,2.12685548066823,.777547323353426,4.12997714955087,4.86004102210358,1.65834138994758,-1.41847476579742,-1.66922123290477,-4.48820932686106,1.03297875354965},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-.300186182387741,-.353250661721207,-.129143286331404,.186704916697283,-1.13611369814234,-1.33694666498547,-.488768188714077,1.22655366483195,1.44337387662839,.100858250439743,-2.48271365058504,-2.92158763778148,-1.06808980107197,1.54415783407832,-1.26648273457434,-1.49036128356923,-.54485433405009,2.5074905585056,2.95074440626681,-.461151554581851,2.81340752869054,3.31073889879025,1.21035782283977,3.19379348794833,3.75836640385388,-2.33714259619681,-2.75487590413568,-3.2418605279988,-4.77320941254596,2.50302559510192},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.03822870489707,-1.22175835711557,-.446657024142497,.645740594449334,-.0622566343920988,-.0732618573826121,-.0267834658423845,1.10967175893133,1.30583053509787,-1.41978880536247,-2.90882269973526,-3.42302078938587,-1.25140644309976,1.80918220620513,.707254507004454,.832277224231996,.304268406428257,1.57948219926848,1.8586902544778,-2.54719860772099,3.86201434387853,4.54470992402429,1.661479620483,.544351948717181,.640578072330311,-3.18953813043078,-4.43071582310304,-5.21394183895335,-2.19278566166381,4.71658607960963},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.59519010006696,-1.87717487173315,-.686267736267333,.992150379408985,1.2388862761662,1.4578865459724,.532982044151546,-.514957807193037,-.605987872997064,-.353775594121333,-3.23038451124486,-3.80142569043747,-1.38974575226926,2.00918198880831,2.62173967966463,3.08518956279543,1.12790027705863,-1.30644820527577,-1.53739152505215,-.307822029727901,4.7630793830775,5.60505793446857,2.04912737785909,-2.90243304119458,-3.41550161982096,.518849452524734,-6.0685481480974,-7.14129688166464,2.66498561200528,7.17423567654393},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.92171958555151,-2.26142559207414,-.826744191593013,1.19523987507356,2.19799643560933,2.5865404220042,.94560142914338,-2.29634628619806,-2.70227576356801,2.23836106649518,-3.41890639763464,-4.02327294720743,-1.47084987158013,2.12643576380383,3.91413586924262,4.60604507183384,1.68390285488931,-4.09536544265748,-4.81931093640565,4.00097073404889,5.33523806427656,6.27835818791885,2.29527612404421,-5.59797201662203,-6.58753611853026,5.49216459339238,-7.19422103343416,-8.46595709192435,7.09892453011937,9.00134910987265},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.92171958555151,5.87604462970729,5.45886369350783,3.32855727881994,6.96914454009643,10.4421041409331,9.70074715430413,10.2353232484353,15.3359297883224,13.4470270492568,-.0451952183653594,-.067717518868618,-.0629097851728847,-.0383594160020484,-.217799593005347,-.326336470786676,-.303167594035556,-.614368644454251,-.920529245649199,-1.33333501160483,.000932318622845132,.00139692441436278,.00129774711619287,.00636097475878108,.0095308628646168,.0244039729141487,-1.66187980893617e-005,-2.49005052754424e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.59519010006696,5.38679449650258,5.00434880168882,3.0514155628076,5.47681396366563,8.20609494322841,7.62348767011616,6.3309248170395,9.48583802706066,5.76936797267538,-.233717104755142,-.350186214905109,-.325324080359881,-.198367260350895,-1.04841335581819,-1.57087306611691,-1.45934595309525,-2.66785366107377,-3.99733505330865,-5.04596922575242,.0249321870529788,.0373567361461481,.0347045238135922,.159138742203809,.238442940887059,.559597243722216,-.00229823078236368,-.00344351663776842,-.0189957623925199,.00019657159111697},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.03822870489707,4.55228046671326,4.22908267863918,2.57869489385806,3.26615690329693,4.89379296535329,4.54634885284358,1.66498726415444,2.49470653361407,-.923099517802526,-.555278916264744,-.831993114526565,-.772924185272261,-.471292665829706,-2.17523616991868,-3.25922966624351,-3.02783447385586,-4.50024407433993,-6.74286736535737,-6.26847807110811,.140734880491405,.210867814600683,.195896854176436,.792696592497175,1.18772402074228,2.35258131414951,-.0308216365902829,-.0461810968751945,-.225963663364185,.00626330408481366},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,2.30018618238774,3.44644648080968,3.20175947449313,1.95228172055832,.987081447023724,1.47897739903412,1.37397459375255,-1.35561938023839,-2.03117020639316,-2.20870753342451,-.981387965414959,-1.47044666380812,-1.36604951383154,-.832952480070553,-3.10522366683209,-4.65266127660931,-4.32233676392885,-4.42237267757001,-6.6261900271284,-2.83495589139377,.439603191878621,.658672278252945,.611908590636856,2.03901059601619,3.05511829665785,4.55407145300733,-.170154848292169,-.254948743871766,-1.03685515398483,.0611113213546687},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,1.44662301694489,2.16752402217829,2.01363654211465,1.2278204669453,-.724355119014703,-1.08532568728512,-1.00827093182807,-1.6918152058931,-2.53490374291394,.160475142667942,-1.4741928888764,-2.20883288938782,-2.05201260878721,-1.25122037987573,-3.38025060487184,-5.06474340721712,-4.70516234202085,-2.16048862424162,-3.23713291923203,1.70770592621981,.991945941760336,1.48626603590141,1.38074575986725,3.46032388717692,5.18471990323707,4.72311662432345,-.576746649127296,-.864158942299273,-2.68886700736678,.311154445333742},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,.553376983055106,.829143384311425,.770276776736346,.469678401195648,-1.45368745167752,-2.17810890152299,-2.02346992933989,-.194424990185747,-.291313515578485,1.49978166741586,-1.9899087271236,-2.9815473107461,-2.76986670414993,-1.68893390563517,-2.74863027720701,-4.11836538252447,-3.82597422031764,.854309601049289,1.2800408674882,2.52312426315122,1.80736427869175,2.70803481201197,2.51577274454146,4.12997714955087,6.18808395553405,1.65834138994758,-1.41847476579742,-2.12534854800545,-4.48820932686106,1.03297875354965},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-.300186182387741,-.449779074319961,-.417846155642134,-.254782852417372,-1.13611369814234,-1.70227744362135,-1.58142102819233,1.22655366483195,1.83778669504629,.100858250439743,-2.48271365058504,-3.7199335363258,-3.45582979910559,-2.10720180544034,-1.26648273457434,-1.89761376598671,-1.76288907720341,2.5074905585056,3.75705761211831,-.461151554581851,2.81340752869054,4.21542331996591,3.91614134649752,3.19379348794833,4.78536842321548,-2.33714259619681,-2.75487590413568,-4.12772341388705,-4.77320941254596,2.50302559510192},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.03822870489707,-1.55561306022354,-1.44516935978819,-.881196025717112,-.0622566343920988,-.0932812135041899,-.0866585368463816,1.10967175893133,1.66265859611778,-1.41978880536247,-2.90882269973526,-4.35838708560735,-4.04895512766487,-2.46886161968119,.707254507004454,1.0597032647262,.984467621360889,1.57948219926848,2.36659141141637,-2.54719860772099,3.86201434387853,5.78658625215441,5.37575658648661,.544351948717181,.815620871022689,-3.18953813043078,-4.43071582310304,-6.63869084478425,-2.19278566166381,4.71658607960963},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.59519010006696,-2.39012709001286,-2.22043548283783,-1.35391669466665,1.2388862761662,1.85626506095126,1.72447600259692,-.514957807193037,-.771578638346196,-.353775594121333,-3.23038451124486,-4.84019398522881,-4.49655523257725,-2.74178702516,2.62173967966463,3.92824092358218,3.64934800797338,-1.30644820527577,-1.95749537633152,-.307822029727901,4.7630793830775,7.13668236880667,6.6300000687783,-2.90243304119458,-4.3488132461414,.518849452524734,-6.0685481480974,-9.09271021667365,2.66498561200528,7.17423567654393},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.92171958555151,-2.87937722321757,-2.67495037465683,-1.63105841067899,2.19799643560933,3.29333213710544,3.05951577632408,-2.29634628619806,-3.44069303528758,2.23836106649518,-3.41890639763464,-5.1226626812653,-4.75896952776425,-2.90179486950885,3.91413586924262,5.86468169257239,5.4483074914535,-4.09536544265748,-6.13622406614863,4.00097073404889,5.33523806427656,7.99396700420041,7.42642015557747,-5.59797201662203,-8.38763014060611,5.49216459339238,-7.19422103343416,-10.7793438389747,7.09892453011937,9.00134910987265},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.92171958555151,6.61537460386721,8.09172854779269,8.80395991076745,6.96914454009643,11.7559404153652,14.3795150585562,10.2353232484353,17.2655122351262,13.4470270492568,-.0451952183653594,-.0762378066864121,-.0932518071891671,-.101459801467564,-.217799593005347,-.367396460919218,-.449388373324038,-.614368644454251,-1.03635118137081,-1.33333501160483,.000932318622845132,.00157268687980642,.00192366360061384,.00636097475878108,.0107300458251741,.0244039729141487,-1.66187980893617e-005,-2.80335124396135e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.59519010006696,6.06456651609707,7.41799655299993,8.07092624092508,5.47681396366563,9.23859424244377,11.300372436005,6.3309248170395,10.6793559162843,5.76936797267538,-.233717104755142,-.394247004353173,-.482231155810537,-.524676988090528,-1.04841335581819,-1.76852192969244,-2.1632031816091,-2.66785366107377,-4.50028385780377,-5.04596922575242,.0249321870529788,.0420569990798227,.0514428646412204,.159138742203809,.268444076758865,.559597243722216,-.00229823078236368,-.0038767834400574,-.0189957623925199,.00019657159111697},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.03822870489707,5.12505307344404,6.26881178264598,6.82059059403532,3.26615690329693,5.50953502563763,6.7390986231311,1.66498726415444,2.80859307349989,-.923099517802526,-.555278916264744,-.936675343241156,-1.14571329243577,-1.24655861042429,-2.17523616991868,-3.66930929238228,-4.48818948881768,-4.50024407433993,-7.59126186695628,-6.26847807110811,.140734880491405,.237399419743997,.29037987690505,.792696592497175,1.33716467766395,2.35258131414951,-.0308216365902829,-.0519916499446519,-.225963663364185,.00626330408481366},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,2.30018618238774,3.88008191896113,4.74600026627036,5.1637416942433,.987081447023724,1.66506385533551,2.03665635715207,-1.35561938023839,-2.28673412920868,-2.20870753342451,-.981387965414959,-1.65545977423627,-2.02490893148973,-2.20314076875832,-3.10522366683209,-5.23806390000324,-6.40704325104372,-4.42237267757001,-7.45990406612127,-2.83495589139377,.439603191878621,.741547101224613,.907038257318,2.03901059601619,3.43951642050874,4.55407145300733,-.170154848292169,-.287026656838231,-1.03685515398483,.0611113213546687},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,1.44662301694489,2.44024412222762,2.98483369571688,3.24755780451479,-.724355119014703,-1.22188248065577,-1.49457014152702,-1.6918152058931,-2.85384793698481,.160475142667942,-1.4741928888764,-2.48675051356285,-3.04171892525937,-3.30945005334757,-3.38025060487184,-5.70199462778367,-6.97450945518368,-2.16048862424162,-3.64443230995284,1.70770592621981,.991945941760336,1.67326955598402,2.04669332477546,3.46032388717692,5.8370666898335,4.72311662432345,-.576746649127296,-.972888308521261,-2.68886700736678,.311154445333742},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,.553376983055106,.93346705704166,1.14178901214037,1.24228891639978,-1.45368745167752,-2.45216080196962,-2.99940982448642,-.194424990185747,-.327966881493408,1.49978166741586,-1.9899087271236,-3.35668865754018,-4.10580126963162,-4.46719258573777,-2.74863027720701,-4.6365423474252,-5.67128006117502,.854309601049289,1.44109692494208,2.52312426315122,1.80736427869175,3.04876253512471,3.72915523810848,4.12997714955087,6.96667503835439,1.65834138994758,-1.41847476579742,-2.39276208655033,-4.48820932686106,1.03297875354965},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-.300186182387741,-.506370739691845,-.619377558413115,-.673894973328732,-1.13611369814234,-1.91645974101964,-2.34415629302528,1.22655366483195,2.06901890572263,.100858250439743,-2.48271365058504,-4.18797939686675,-5.12261126340126,-5.57350187032702,-1.26648273457434,-2.13637348021345,-2.61314820693614,2.5074905585056,4.22977446202813,-.461151554581851,2.81340752869054,4.7458122143026,5.80493569904152,3.19379348794833,5.38746839617942,-2.33714259619681,-2.75487590413568,-4.64707781591113,-4.77320941254596,2.50302559510192},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.03822870489707,-1.75134189417476,-2.14218907478874,-2.33074387312075,-.0622566343920988,-.105017951653011,-.12845482061473,1.10967175893133,1.87185600940662,-1.41978880536247,-2.90882269973526,-4.90676382786187,-6.00180690245521,-6.53008402866105,.707254507004454,1.19303621855815,1.45928625505293,1.57948219926848,2.66435837496252,-2.54719860772099,3.86201434387853,6.51466048131367,7.96853804732135,.544351948717181,.918243127233759,-3.18953813043078,-4.43071582310304,-7.47397774950175,-2.19278566166381,4.71658607960963},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.59519010006696,-2.69085533682779,-3.29137384514269,-3.58107952001051,1.2388862761662,2.0898222386161,2.55620812016103,-.514957807193037,-.868659455579414,-.353775594121333,-3.23038451124486,-5.44919216674985,-6.66528903908045,-7.25196565099482,2.62173967966463,4.42249623366662,5.4094652504454,-1.30644820527577,-2.20378945631257,-.307822029727901,4.7630793830775,8.03462707886587,9.82771577392637,-2.90243304119458,-4.89598539807521,.518849452524734,-6.0685481480974,-10.2367643607743,2.66498561200528,7.17423567654393},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.92171958555151,-3.24166342459793,-3.96510583993544,-4.31411318985288,2.19799643560933,3.70770257117539,4.53515100205683,-2.29634628619806,-3.87360456901059,2.23836106649518,-3.41890639763464,-5.76720136441661,-7.05426838770182,-7.67518283761778,3.91413586924262,6.60258197030424,8.07608098333328,-4.09536544265748,-6.90829008438974,4.00097073404889,5.33523806427656,8.99977614812197,11.0082572774727,-5.59797201662203,-9.44297038492046,5.49216459339238,-7.19422103343416,-12.1356119587979,7.09892453011937,9.00134910987265}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_right_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_right_face(
		const T i_input[35],
		T o_output[100]
	) {
		static const T matrix[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.92171958555151,-6.61537460386721,8.09172854779269,-8.80395991076744,6.96914454009643,-11.7559404153652,14.3795150585562,10.2353232484353,-17.2655122351262,13.4470270492568,-.0451952183653594,.0762378066864121,-.0932518071891671,.101459801467564,-.217799593005347,.367396460919218,-.449388373324038,-.614368644454251,1.03635118137081,-1.33333501160483,.000932318622845132,-.00157268687980642,.00192366360061384,.00636097475878108,-.0107300458251741,.0244039729141487,-1.66187980893617e-005,2.80335124396135e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.59519010006696,-6.06456651609707,7.41799655299993,-8.07092624092507,5.47681396366563,-9.23859424244377,11.300372436005,6.3309248170395,-10.6793559162843,5.76936797267538,-.233717104755142,.394247004353173,-.482231155810537,.524676988090528,-1.04841335581819,1.76852192969244,-2.1632031816091,-2.66785366107377,4.50028385780377,-5.04596922575242,.0249321870529788,-.0420569990798227,.0514428646412204,.159138742203809,-.268444076758865,.559597243722216,-.00229823078236368,.0038767834400574,-.0189957623925199,.00019657159111697},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.03822870489707,-5.12505307344404,6.26881178264598,-6.82059059403531,3.26615690329693,-5.50953502563763,6.7390986231311,1.66498726415444,-2.80859307349989,-.923099517802526,-.555278916264744,.936675343241156,-1.14571329243577,1.24655861042429,-2.17523616991868,3.66930929238228,-4.48818948881768,-4.50024407433993,7.59126186695628,-6.26847807110811,.140734880491405,-.237399419743997,.29037987690505,.792696592497175,-1.33716467766395,2.35258131414951,-.0308216365902829,.0519916499446519,-.225963663364185,.00626330408481366},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,2.30018618238774,-3.88008191896113,4.74600026627036,-5.1637416942433,.987081447023724,-1.66506385533551,2.03665635715207,-1.35561938023839,2.28673412920868,-2.20870753342451,-.981387965414959,1.65545977423627,-2.02490893148973,2.20314076875832,-3.10522366683209,5.23806390000324,-6.40704325104372,-4.42237267757001,7.45990406612127,-2.83495589139377,.439603191878621,-.741547101224613,.907038257318,2.03901059601619,-3.43951642050874,4.55407145300733,-.170154848292169,.287026656838231,-1.03685515398483,.0611113213546687},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,1.44662301694489,-2.44024412222762,2.98483369571688,-3.24755780451478,-.724355119014703,1.22188248065577,-1.49457014152702,-1.6918152058931,2.85384793698481,.160475142667942,-1.4741928888764,2.48675051356285,-3.04171892525937,3.30945005334757,-3.38025060487184,5.70199462778367,-6.97450945518368,-2.16048862424162,3.64443230995284,1.70770592621981,.991945941760336,-1.67326955598402,2.04669332477546,3.46032388717692,-5.8370666898335,4.72311662432345,-.576746649127296,.972888308521261,-2.68886700736678,.311154445333742},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,.553376983055106,-.93346705704166,1.14178901214037,-1.24228891639978,-1.45368745167752,2.45216080196962,-2.99940982448642,-.194424990185747,.327966881493408,1.49978166741586,-1.9899087271236,3.35668865754018,-4.10580126963162,4.46719258573777,-2.74863027720701,4.6365423474252,-5.67128006117502,.854309601049289,-1.44109692494208,2.52312426315122,1.80736427869175,-3.04876253512471,3.72915523810848,4.12997714955087,-6.96667503835439,1.65834138994758,-1.41847476579742,2.39276208655033,-4.48820932686106,1.03297875354965},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-.300186182387741,.506370739691845,-.619377558413115,.673894973328731,-1.13611369814234,1.91645974101964,-2.34415629302528,1.22655366483195,-2.06901890572263,.100858250439743,-2.48271365058504,4.18797939686675,-5.12261126340126,5.57350187032702,-1.26648273457434,2.13637348021345,-2.61314820693614,2.5074905585056,-4.22977446202813,-.461151554581851,2.81340752869054,-4.7458122143026,5.80493569904152,3.19379348794833,-5.38746839617942,-2.33714259619681,-2.75487590413568,4.64707781591113,-4.77320941254596,2.50302559510192},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.03822870489707,1.75134189417476,-2.14218907478874,2.33074387312075,-.0622566343920988,.105017951653011,-.12845482061473,1.10967175893133,-1.87185600940662,-1.41978880536247,-2.90882269973526,4.90676382786187,-6.00180690245521,6.53008402866105,.707254507004454,-1.19303621855815,1.45928625505293,1.57948219926848,-2.66435837496252,-2.54719860772099,3.86201434387853,-6.51466048131367,7.96853804732135,.544351948717181,-.918243127233759,-3.18953813043078,-4.43071582310304,7.47397774950175,-2.19278566166381,4.71658607960963},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.59519010006696,2.69085533682779,-3.29137384514269,3.5810795200105,1.2388862761662,-2.0898222386161,2.55620812016103,-.514957807193037,.868659455579414,-.353775594121333,-3.23038451124486,5.44919216674985,-6.66528903908045,7.25196565099481,2.62173967966463,-4.42249623366662,5.4094652504454,-1.30644820527577,2.20378945631257,-.307822029727901,4.7630793830775,-8.03462707886587,9.82771577392637,-2.90243304119458,4.89598539807521,.518849452524734,-6.0685481480974,10.2367643607743,2.66498561200528,7.17423567654393},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.92171958555151,3.24166342459793,-3.96510583993544,4.31411318985288,2.19799643560933,-3.70770257117539,4.53515100205683,-2.29634628619806,3.87360456901059,2.23836106649518,-3.41890639763464,5.76720136441661,-7.05426838770182,7.67518283761778,3.91413586924262,-6.60258197030424,8.07608098333328,-4.09536544265748,6.90829008438974,4.00097073404889,5.33523806427656,-8.99977614812197,11.0082572774727,-5.59797201662203,9.44297038492046,5.49216459339238,-7.19422103343416,12.1356119587979,7.09892453011937,9.00134910987265},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.92171958555151,-5.87604462970729,5.45886369350783,-3.32855727881992,6.96914454009643,-10.4421041409331,9.70074715430413,10.2353232484353,-15.3359297883224,13.4470270492568,-.0451952183653594,.067717518868618,-.0629097851728847,.0383594160020481,-.217799593005347,.326336470786676,-.303167594035556,-.614368644454251,.920529245649199,-1.33333501160483,.000932318622845132,-.00139692441436278,.00129774711619287,.00636097475878108,-.0095308628646168,.0244039729141487,-1.66187980893617e-005,2.49005052754424e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.59519010006696,-5.38679449650258,5.00434880168882,-3.05141556280758,5.47681396366563,-8.20609494322841,7.62348767011616,6.3309248170395,-9.48583802706066,5.76936797267538,-.233717104755142,.350186214905109,-.325324080359881,.198367260350894,-1.04841335581819,1.57087306611691,-1.45934595309525,-2.66785366107377,3.99733505330865,-5.04596922575242,.0249321870529788,-.0373567361461481,.0347045238135922,.159138742203809,-.238442940887059,.559597243722216,-.00229823078236368,.00344351663776842,-.0189957623925199,.00019657159111697},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.03822870489707,-4.55228046671326,4.22908267863918,-2.57869489385804,3.26615690329693,-4.89379296535329,4.54634885284358,1.66498726415444,-2.49470653361407,-.923099517802526,-.555278916264744,.831993114526565,-.772924185272261,.471292665829703,-2.17523616991868,3.25922966624351,-3.02783447385586,-4.50024407433993,6.74286736535737,-6.26847807110811,.140734880491405,-.210867814600683,.195896854176436,.792696592497175,-1.18772402074228,2.35258131414951,-.0308216365902829,.0461810968751945,-.225963663364185,.00626330408481366},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,2.30018618238774,-3.44644648080968,3.20175947449313,-1.95228172055831,.987081447023724,-1.47897739903412,1.37397459375255,-1.35561938023839,2.03117020639316,-2.20870753342451,-.981387965414959,1.47044666380812,-1.36604951383154,.832952480070547,-3.10522366683209,4.65266127660931,-4.32233676392885,-4.42237267757001,6.6261900271284,-2.83495589139377,.439603191878621,-.658672278252945,.611908590636856,2.03901059601619,-3.05511829665785,4.55407145300733,-.170154848292169,.254948743871766,-1.03685515398483,.0611113213546687},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,1.44662301694489,-2.16752402217829,2.01363654211465,-1.22782046694529,-.724355119014703,1.08532568728512,-1.00827093182807,-1.6918152058931,2.53490374291394,.160475142667942,-1.4741928888764,2.20883288938782,-2.05201260878721,1.25122037987572,-3.38025060487184,5.06474340721712,-4.70516234202085,-2.16048862424162,3.23713291923203,1.70770592621981,.991945941760336,-1.48626603590141,1.38074575986725,3.46032388717692,-5.18471990323707,4.72311662432345,-.576746649127296,.864158942299273,-2.68886700736678,.311154445333742},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,.553376983055106,-.829143384311425,.770276776736346,-.469678401195645,-1.45368745167752,2.17810890152299,-2.02346992933989,-.194424990185747,.291313515578485,1.49978166741586,-1.9899087271236,2.9815473107461,-2.76986670414993,1.68893390563516,-2.74863027720701,4.11836538252447,-3.82597422031764,.854309601049289,-1.2800408674882,2.52312426315122,1.80736427869175,-2.70803481201197,2.51577274454146,4.12997714955087,-6.18808395553405,1.65834138994758,-1.41847476579742,2.12534854800545,-4.48820932686106,1.03297875354965},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-.300186182387741,.449779074319961,-.417846155642134,.25478285241737,-1.13611369814234,1.70227744362135,-1.58142102819233,1.22655366483195,-1.83778669504629,.100858250439743,-2.48271365058504,3.7199335363258,-3.45582979910559,2.10720180544033,-1.26648273457434,1.89761376598671,-1.76288907720341,2.5074905585056,-3.75705761211831,-.461151554581851,2.81340752869054,-4.21542331996591,3.91614134649752,3.19379348794833,-4.78536842321548,-2.33714259619681,-2.75487590413568,4.12772341388705,-4.77320941254596,2.50302559510192},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.03822870489707,1.55561306022354,-1.44516935978819,.881196025717106,-.0622566343920988,.0932812135041899,-.0866585368463816,1.10967175893133,-1.66265859611778,-1.41978880536247,-2.90882269973526,4.35838708560735,-4.04895512766487,2.46886161968117,.707254507004454,-1.0597032647262,.984467621360889,1.57948219926848,-2.36659141141637,-2.54719860772099,3.86201434387853,-5.78658625215441,5.37575658648661,.544351948717181,-.815620871022689,-3.18953813043078,-4.43071582310304,6.63869084478425,-2.19278566166381,4.71658607960963},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.59519010006696,2.39012709001286,-2.22043548283783,1.35391669466664,1.2388862761662,-1.85626506095126,1.72447600259692,-.514957807193037,.771578638346196,-.353775594121333,-3.23038451124486,4.84019398522881,-4.49655523257725,2.74178702515998,2.62173967966463,-3.92824092358218,3.64934800797338,-1.30644820527577,1.95749537633152,-.307822029727901,4.7630793830775,-7.13668236880667,6.6300000687783,-2.90243304119458,4.3488132461414,.518849452524734,-6.0685481480974,9.09271021667365,2.66498561200528,7.17423567654393},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.92171958555151,2.87937722321757,-2.67495037465683,1.63105841067898,2.19799643560933,-3.29333213710544,3.05951577632408,-2.29634628619806,3.44069303528758,2.23836106649518,-3.41890639763464,5.1226626812653,-4.75896952776425,2.90179486950883,3.91413586924262,-5.86468169257239,5.4483074914535,-4.09536544265748,6.13622406614863,4.00097073404889,5.33523806427656,-7.99396700420041,7.42642015557747,-5.59797201662203,8.38763014060611,5.49216459339238,-7.19422103343416,10.7793438389747,7.09892453011937,9.00134910987265},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.92171958555151,-4.61496937554466,1.6871654495215,2.43916732844399,6.96914454009643,-8.20109339929485,2.99820005631406,10.2353232484353,-12.0446407028288,13.4470270492568,-.0451952183653594,.0531844625867158,-.0194434633176024,-.028109786436831,-.217799593005347,.256300439106579,-.0936996999356596,-.614368644454251,.722971751522571,-1.33333501160483,.000932318622845132,-.00109712634918432,.000401093381258216,.00636097475878108,-.00748541629863581,.0244039729141487,-1.66187980893617e-005,1.95565344587043e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.59519010006696,-4.23071865520366,1.54668899419582,2.23607783277941,5.47681396366563,-6.44496072482988,2.35618358032458,6.3309248170395,-7.45005437415178,5.76936797267538,-.233717104755142,.275031719356679,-.100547582628473,-.145363561432352,-1.04841335581819,1.23374336817428,-.451038569416845,-2.66785366107377,3.13945519772704,-5.04596922575242,.0249321870529788,-.0293394969148428,.0107260918769642,.159138742203809,-.187269998673138,.559597243722216,-.00229823078236368,.00270449338439953,-.0189957623925199,.00019657159111697},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.03822870489707,-3.57530214058608,1.30707828207098,1.88966804781976,3.26615690329693,-3.843521635486,1.40513541583971,1.66498726415444,-1.95931021275678,-.923099517802526,-.555278916264744,.653436620408273,-.238886891797977,-.345363344035539,-2.17523616991868,2.55975678291091,-.935809721212703,-4.50024407433993,5.29576072944733,-6.26847807110811,.140734880491405,-.165612851500527,.0605456414728777,.792696592497175,-.932823068533675,2.35258131414951,-.0308216365902829,.0362700355832219,-.225963663364185,.00626330408481366},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,2.30018618238774,-2.70679444519172,.989564544259887,1.4306323700677,.987081447023724,-1.16156970101249,.42465293021633,-1.35561938023839,1.5952547827211,-2.20870753342451,-.981387965414959,1.15486977201266,-.422203533825773,-.610387716162354,-3.10522366683209,3.65413992905702,-1.33590022736237,-4.42237267757001,5.20412385861342,-2.83495589139377,.439603191878621,-.517312679568432,.189121965733585,2.03901059601619,-2.39945035585009,4.55407145300733,-.170154848292169,.200233442638163,-1.03685515398483,.0611113213546687},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,1.44662301694489,-1.70234530427801,.622352597994011,.899747042727645,-.724355119014703,.852400743178548,-.311625271383912,-1.6918152058931,1.9908805805112,.160475142667942,-1.4741928888764,1.73478875376213,-.634213449888056,-.916894502821395,-3.38025060487184,3.97778389924568,-1.45421973940992,-2.16048862424162,2.54240228389271,1.70770592621981,.991945941760336,-1.16729410204273,.426745687640423,3.46032388717692,-4.07201188589473,4.72311662432345,-.576746649127296,.678699245092846,-2.68886700736678,.311154445333742},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,.553376983055106,-.651198479192506,.238068659934471,.344180410642771,-1.45368745167752,1.71065853147384,-.625391792982139,-.194424990185747,.228793863525153,1.49978166741586,-1.9899087271236,2.34166865603201,-.856079885009688,-1.23765104741929,-2.74863027720701,3.23451085132423,-1.18248996086068,.854309601049289,-1.00532752474992,2.52312426315122,1.80736427869175,-2.12685548066823,.77754732335343,4.12997714955087,-4.86004102210358,1.65834138994758,-1.41847476579742,1.66922123290477,-4.48820932686106,1.03297875354965},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-.300186182387741,.353250661721207,-.129143286331405,-.186704916697285,-1.13611369814234,1.33694666498547,-.488768188714079,1.22655366483195,-1.44337387662839,.100858250439743,-2.48271365058504,2.92158763778148,-1.06808980107197,-1.54415783407833,-1.26648273457434,1.49036128356923,-.544854334050093,2.5074905585056,-2.95074440626681,-.461151554581851,2.81340752869054,-3.31073889879025,1.21035782283977,3.19379348794833,-3.75836640385388,-2.33714259619681,-2.75487590413568,3.2418605279988,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.03822870489707,1.22175835711557,-.446657024142499,-.64574059444934,-.0622566343920988,.0732618573826121,-.0267834658423847,1.10967175893133,-1.30583053509787,-1.41978880536247,-2.90882269973526,3.42302078938587,-1.25140644309977,-1.80918220620515,.707254507004454,-.832277224231996,.304268406428259,1.57948219926848,-1.8586902544778,-2.54719860772099,3.86201434387853,-4.54470992402429,1.66147962048301,.544351948717181,-.640578072330311,-3.18953813043078,-4.43071582310304,5.21394183895335,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.59519010006696,1.87717487173315,-.686267736267336,-.992150379408995,1.2388862761662,-1.4578865459724,.532982044151549,-.514957807193037,.605987872997064,-.353775594121333,-3.23038451124486,3.80142569043747,-1.38974575226927,-2.00918198880834,2.62173967966463,-3.08518956279543,1.12790027705863,-1.30644820527577,1.53739152505215,-.307822029727901,4.7630793830775,-5.60505793446857,2.0491273778591,-2.90243304119458,3.41550161982096,.518849452524734,-6.0685481480974,7.14129688166464,2.66498561200528,7.17423567654393},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.92171958555151,2.26142559207414,-.826744191593017,-1.19523987507357,2.19799643560933,-2.5865404220042,.945601429143385,-2.29634628619806,2.70227576356801,2.23836106649518,-3.41890639763464,4.02327294720743,-1.47084987158014,-2.12643576380386,3.91413586924262,-4.60604507183384,1.68390285488932,-4.09536544265748,4.81931093640565,4.00097073404889,5.33523806427656,-6.27835818791885,2.29527612404422,-5.59797201662203,6.58753611853026,5.49216459339238,-7.19422103343416,8.46595709192435,7.09892453011937,9.00134910987265},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.92171958555151,-2.94388917191153,-1.91390801783746,4.63366448151074,6.96914454009643,-5.23147784770941,-3.40113598727833,10.2353232484353,-7.68327684524188,13.4470270492568,-.0451952183653594,.0339263710725032,.0220565211009044,-.0533999113042758,-.217799593005347,.163494061616168,.10629224716917,-.614368644454251,.461183712763862,-1.33333501160483,.000932318622845132,-.000699856947415033,-.000454997367631818,.00636097475878108,-.00477494739305899,.0244039729141487,-1.66187980893617e-005,1.24751142094631e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.59519010006696,-2.69877562014986,-1.7545525650326,4.24785717273984,5.47681396366563,-4.11124073580283,-2.67283724139748,6.3309248170395,-4.75239001984257,5.76936797267538,-.233717104755142,.175442746128902,.114060434690259,-.276145864886064,-1.04841335581819,.787004949964004,.511654819985594,-2.66785366107377,2.00265861161846,-5.04596922575242,.0249321870529788,-.0187156663995725,-.0121675993597458,.159138742203809,-.119459540563619,.559597243722216,-.00229823078236368,.00172519645136334,-.0189957623925199,.00019657159111697},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.03822870489707,-2.2806853960415,-1.48273994391383,3.58978558499055,3.26615690329693,-2.45178262388351,-1.59397523182193,1.66498726415444,-1.24984407270733,-.923099517802526,-.555278916264744,.41682724950333,.270991524688991,-.656083673232097,-2.17523616991868,1.63286896677384,1.06157563213922,-4.50024407433993,3.3781660071394,-6.26847807110811,.140734880491405,-.105644445409619,-.0686825282564419,.792696592497175,-.595047877309364,2.35258131414951,-.0308216365902829,.0231366573273721,-.225963663364185,.00626330408481366},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,2.30018618238774,-1.72666429814602,-1.12255470615749,2.71775959032344,.987081447023724,-.740965364440283,-.481723145865576,-1.35561938023839,1.01761309753837,-2.20870753342451,-.981387965414959,.736691479430405,.47894456870109,-1.15954811601065,-3.10522366683209,2.33097602484761,1.51543534565912,-4.42237267757001,3.31971084311917,-2.83495589139377,.439603191878621,-.329993781528466,-.214538560322787,2.03901059601619,-1.53060948975261,4.55407145300733,-.170154848292169,.127728922051078,-1.03685515398483,.0611113213546687},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,1.44662301694489,-1.0859261460488,-.705992187998236,1.70924145531706,-.724355119014703,.543746472353678,.353505404956933,-1.6918152058931,1.26998281171566,.160475142667942,-1.4741928888764,1.10662182393172,.719447050735517,-1.74181633275903,-3.38025060487184,2.53742852796533,1.64965612574837,-2.16048862424162,1.62179853107076,1.70770592621981,.991945941760336,-.744616960114543,-.484097154422055,3.46032388717692,-2.5975365648804,4.72311662432345,-.576746649127296,.432942278924126,-2.68886700736678,.311154445333742},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,.553376983055106,-.41539953912128,-.270063328509738,.65383646518608,-1.45368745167752,1.09122915405036,.709439828333213,-.194424990185747,.145947615817496,1.49978166741586,-1.9899087271236,1.49375054084377,.971130695151503,-2.35115468793597,-2.74863027720701,2.06329461731692,1.34140787303666,.854309601049289,-.641298473185774,2.52312426315122,1.80736427869175,-1.35672120663226,-.88204393756181,4.12997714955087,-3.10022038625728,1.65834138994758,-1.41847476579742,1.06479629943226,-4.48820932686106,1.03297875354965},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-.300186182387741,.225338612975934,.146499189649512,-.354681669820299,-1.13611369814234,.852838337634571,.554455021295674,1.22655366483195,-.920728260721081,.100858250439743,-2.48271365058504,1.86368088534508,1.21163317718593,-2.93342290468435,-1.26648273457434,.950701529718779,.618078730089993,2.5074905585056,-1.88227998927126,-.461151554581851,2.81340752869054,-2.11192049221879,-1.37302096972311,3.19379348794833,-2.39746209780118,-2.33714259619681,-2.75487590413568,2.0679829763977,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.03822870489707,.779359710871422,.506684427405858,-1.22670766448742,-.0622566343920988,.0467337421144522,.030382965722629,1.10967175893133,-.832989356981887,-1.41978880536247,-2.90882269973526,2.18354511527215,1.41958622119803,-3.4368873474629,.707254507004454,-.530909678713955,-.345159831718241,1.57948219926848,-1.18565859680248,-2.54719860772099,3.86201434387853,-2.89907066463143,-1.88477020319366,.544351948717181,-.408624781106939,-3.18953813043078,-4.43071582310304,3.32597373321892,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.59519010006696,1.19744993497978,.77849704852463,-1.88477925223671,1.2388862761662,-.929985893147364,-.604610892164935,-.514957807193037,.386559691398865,-.353775594121333,-3.23038451124486,2.42492961864658,1.57651731119676,-3.81682515580894,2.62173967966463,-1.9680425606586,-1.27948173914766,-1.30644820527577,.980702122872796,-.307822029727901,4.7630793830775,-3.57546670811253,-2.32451495445634,-2.90243304119458,2.17874863666902,.518849452524734,-6.0685481480974,4.55543360151918,2.66498561200528,7.17423567654393},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.92171958555151,1.44256348674145,.937852501329485,-2.27058656100761,2.19799643560933,-1.64995425135427,-1.07268327325537,-2.29634628619806,1.72378183093533,2.23836106649518,-3.41890639763464,2.56644599370298,1.66852122478612,-4.03957110939073,3.91413586924262,-2.93819635817362,-1.91020695459719,-4.09536544265748,3.07423866276144,4.00097073404889,5.33523806427656,-4.00496497002541,-2.60374427309723,-5.59797201662203,4.20218958584716,5.49216459339238,-7.19422103343416,5.40043440913241,7.09892453011937,9.00134910987265},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.92171958555151,-1.01124645171658,-4.09307958874724,2.23146629547127,6.96914454009643,-1.7970491081004,-7.2736621396321,10.2353232484353,-2.63925915467191,13.4470270492568,-.0451952183653594,.0116539449562267,.0471700287973566,-.0257161697253048,-.217799593005347,.0561613498441344,.227316372195247,-.614368644454251,.158419820138511,-1.33333501160483,.000932318622845132,-.000240405739976654,-.000973056396143737,.00636097475878108,-.00164022771456063,.0244039729141487,-1.66187980893617e-005,4.28528869206282e-006,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.59519010006696,-.927048238057034,-3.75228235860234,2.04567036451771,5.47681396366563,-1.4122398512488,-5.7161240011779,6.3309248170395,-1.63247909961861,5.76936797267538,-.233717104755142,.060265806265752,.243929401394036,-.132985500479333,-1.04841335581819,.27034168634063,1.09422390242844,-2.66785366107377,.68792719315715,-5.04596922575242,.0249321870529788,-.00642896187079958,-.0260216019365979,.159138742203809,-.0410351849082052,.559597243722216,-.00229823078236368,.000592617006985495,-.0189957623925199,.00019657159111697},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.03822870489707,-.78343133166635,-3.17098446910282,1.72875821562793,3.26615690329693,-.842204422108959,-3.40887201763794,1.66498726415444,-.429330151346565,-.923099517802526,-.555278916264744,.143183065809956,.579541894432988,-.315954814956539,-2.17523616991868,.560901872435633,2.27028337376948,-4.50024407433993,1.16042357185991,-6.26847807110811,.140734880491405,-.0362896034209994,-.14688430786117,.792696592497175,-.204403093795768,2.35258131414951,-.0308216365902829,.0079476030721956,-.225963663364185,.00626330408481366},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,2.30018618238774,-.593121222587361,-2.40069309089209,1.30881054275714,.987081447023724,-.254526767782017,-1.03021208824726,-1.35561938023839,.349557192548049,-2.20870753342451,-.981387965414959,.253058658488868,1.02426982907298,-.558411716942304,-3.10522366683209,.800706513225493,3.2409067852655,-4.42237267757001,1.14034381554164,-2.83495589139377,.439603191878621,-.113355164264732,-.458811705720642,2.03901059601619,-.525775029251873,4.55407145300733,-.170154848292169,.0438757750930774,-1.03685515398483,.0611113213546687},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,1.44662301694489,-.373023200905715,-1.50983338153088,.823131392784558,-.724355119014703,.186780703583358,.756005902089247,-1.6918152058931,.436247948578379,.160475142667942,-1.4741928888764,.380132310523241,1.53860792217029,-.83881870492831,-3.38025060487184,.871624387178691,3.52795105816102,-2.16048862424162,.557099100602201,1.70770592621981,.991945941760336,-.255781116350547,-1.03528913786274,3.46032388717692,-.892271916858127,4.72311662432345,-.576746649127296,.148718690750112,-2.68886700736678,.311154445333742},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,.553376983055106,-.142692637341486,-.577556856071558,.314872611220466,-1.45368745167752,.374844821120283,1.51720649776892,-.194424990185747,.0501340234001082,1.49978166741586,-1.9899087271236,.513113723366546,2.07685802519487,-1.13226204929232,-2.74863027720701,.708756083794028,2.86873200495546,.854309601049289,-.220290495881434,2.52312426315122,1.80736427869175,-.466043191865359,-1.88633727617264,4.12997714955087,-1.06494731359279,1.65834138994758,-1.41847476579742,.365764951301699,-4.48820932686106,1.03297875354965},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-.300186182387741,.0774053843401605,.313302853289657,-.170806538752116,-1.13611369814234,.292955913914665,1.18575632130325,1.22655366483195,-.316276575739705,.100858250439743,-2.48271365058504,.640187375400919,2.59119611829218,-1.41266903727833,-1.26648273457434,.326572602577083,1.32182184869504,2.5074905585056,-.646576297249579,-.461151554581851,2.81340752869054,-.725459410782454,-2.93633970583683,3.19379348794833,-.823544942743957,-2.33714259619681,-2.75487590413568,.710366567819732,-4.77320941254596,2.50302559510192},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.03822870489707,.26771549341915,1.08359423150038,-.590754211622909,-.0622566343920988,.0160533661863334,.0649769454361842,1.10967175893133,-.286137650697852,-1.41978880536247,-2.90882269973526,.750062968079831,3.03592405293217,-1.65512593926409,.707254507004454,-.182371175485818,-.738158077031132,1.57948219926848,-.407281992951262,-2.54719860772099,3.86201434387853,-.995850982046505,-4.03076552074204,.544351948717181,-.140365460738301,-3.18953813043078,-4.43071582310304,1.14249516194805,-2.19278566166381,4.71658607960963},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.59519010006696,.411332399809833,1.66489212099991,-.907666360512686,1.2388862761662,-.31945663701093,-1.29301955934683,-.514957807193037,.132785948612982,-.353775594121333,-3.23038451124486,.832980227624035,3.37153654597113,-1.8380952537413,2.62173967966463,-.676036338352017,-2.73629690762123,-1.30644820527577,.336878015452627,-.307822029727901,4.7630793830775,-1.22819773798136,-4.97120270417874,-2.90243304119458,.748415344244522,.518849452524734,-6.0685481480974,1.56482319698725,2.66498561200528,7.17423567654393},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.92171958555151,.49553061346938,2.00568935114481,-1.09346229146625,2.19799643560933,-.566770786786546,-2.29403815127598,-2.29634628619806,.592131075007482,2.23836106649518,-3.41890639763464,.881592088933561,3.56829591856781,-1.94536458449533,3.91413586924262,-1.00929093051433,-4.08516450282659,-4.09536544265748,1.05602241045943,4.00097073404889,5.33523806427656,-1.37573338488067,-5.56836192711791,-5.59797201662203,1.44348141522225,5.49216459339238,-7.19422103343416,1.855086864706,7.09892453011937,9.00134910987265},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.92171958555151,1.01124645171658,-4.09307958874725,-2.23146629547129,6.96914454009643,1.7970491081004,-7.2736621396321,10.2353232484353,2.63925915467191,13.4470270492568,-.0451952183653594,-.0116539449562267,.0471700287973566,.025716169725305,-.217799593005347,-.0561613498441344,.227316372195247,-.614368644454251,-.158419820138511,-1.33333501160483,.000932318622845132,.000240405739976654,-.000973056396143738,.00636097475878108,.00164022771456063,.0244039729141487,-1.66187980893617e-005,-4.28528869206282e-006,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.59519010006696,.927048238057034,-3.75228235860234,-2.04567036451773,5.47681396366563,1.4122398512488,-5.7161240011779,6.3309248170395,1.63247909961861,5.76936797267538,-.233717104755142,-.060265806265752,.243929401394036,.132985500479334,-1.04841335581819,-.27034168634063,1.09422390242844,-2.66785366107377,-.68792719315715,-5.04596922575242,.0249321870529788,.00642896187079958,-.0260216019365979,.159138742203809,.0410351849082052,.559597243722216,-.00229823078236368,-.000592617006985495,-.0189957623925199,.00019657159111697},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.03822870489707,.78343133166635,-3.17098446910282,-1.72875821562795,3.26615690329693,.842204422108959,-3.40887201763794,1.66498726415444,.429330151346565,-.923099517802526,-.555278916264744,-.143183065809956,.579541894432988,.315954814956541,-2.17523616991868,-.560901872435633,2.27028337376948,-4.50024407433993,-1.16042357185991,-6.26847807110811,.140734880491405,.0362896034209994,-.14688430786117,.792696592497175,.204403093795768,2.35258131414951,-.0308216365902829,-.0079476030721956,-.225963663364185,.00626330408481366},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,2.30018618238774,.593121222587361,-2.40069309089209,-1.30881054275715,.987081447023724,.254526767782017,-1.03021208824726,-1.35561938023839,-.349557192548049,-2.20870753342451,-.981387965414959,-.253058658488868,1.02426982907298,.558411716942308,-3.10522366683209,-.800706513225493,3.2409067852655,-4.42237267757001,-1.14034381554164,-2.83495589139377,.439603191878621,.113355164264732,-.458811705720642,2.03901059601619,.525775029251873,4.55407145300733,-.170154848292169,-.0438757750930774,-1.03685515398483,.0611113213546687},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,1.44662301694489,.373023200905715,-1.50983338153088,-.823131392784565,-.724355119014703,-.186780703583358,.756005902089247,-1.6918152058931,-.436247948578379,.160475142667942,-1.4741928888764,-.380132310523241,1.53860792217029,.838818704928317,-3.38025060487184,-.871624387178691,3.52795105816103,-2.16048862424162,-.557099100602201,1.70770592621981,.991945941760336,.255781116350547,-1.03528913786275,3.46032388717692,.892271916858127,4.72311662432345,-.576746649127296,-.148718690750112,-2.68886700736678,.311154445333742},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,.553376983055106,.142692637341486,-.577556856071558,-.314872611220468,-1.45368745167752,-.374844821120283,1.51720649776892,-.194424990185747,-.0501340234001082,1.49978166741586,-1.9899087271236,-.513113723366546,2.07685802519487,1.13226204929233,-2.74863027720701,-.708756083794028,2.86873200495547,.854309601049289,.220290495881434,2.52312426315122,1.80736427869175,.466043191865359,-1.88633727617264,4.12997714955087,1.06494731359279,1.65834138994758,-1.41847476579742,-.365764951301699,-4.48820932686106,1.03297875354965},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-.300186182387741,-.0774053843401605,.313302853289658,.170806538752117,-1.13611369814234,-.292955913914665,1.18575632130325,1.22655366483195,.316276575739705,.100858250439743,-2.48271365058504,-.640187375400919,2.59119611829218,1.41266903727834,-1.26648273457434,-.326572602577083,1.32182184869504,2.5074905585056,.646576297249579,-.461151554581851,2.81340752869054,.725459410782454,-2.93633970583683,3.19379348794833,.823544942743957,-2.33714259619681,-2.75487590413568,-.710366567819732,-4.77320941254596,2.50302559510192},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.03822870489707,-.26771549341915,1.08359423150038,.590754211622914,-.0622566343920988,-.0160533661863334,.0649769454361843,1.10967175893133,.286137650697852,-1.41978880536247,-2.90882269973526,-.750062968079831,3.03592405293218,1.65512593926411,.707254507004454,.182371175485818,-.738158077031133,1.57948219926848,.407281992951262,-2.54719860772099,3.86201434387853,.995850982046505,-4.03076552074204,.544351948717181,.140365460738301,-3.18953813043078,-4.43071582310304,-1.14249516194805,-2.19278566166381,4.71658607960963},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.59519010006696,-.411332399809833,1.66489212099991,.907666360512694,1.2388862761662,.31945663701093,-1.29301955934683,-.514957807193037,-.132785948612982,-.353775594121333,-3.23038451124486,-.832980227624035,3.37153654597113,1.83809525374131,2.62173967966463,.676036338352017,-2.73629690762123,-1.30644820527577,-.336878015452627,-.307822029727901,4.7630793830775,1.22819773798136,-4.97120270417875,-2.90243304119458,-.748415344244522,.518849452524734,-6.0685481480974,-1.56482319698725,2.66498561200528,7.17423567654393},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.92171958555151,-.49553061346938,2.00568935114481,1.09346229146626,2.19799643560933,.566770786786546,-2.29403815127598,-2.29634628619806,-.592131075007482,2.23836106649518,-3.41890639763464,-.881592088933561,3.56829591856781,1.94536458449534,3.91413586924262,1.00929093051433,-4.08516450282659,-4.09536544265748,-1.05602241045943,4.00097073404889,5.33523806427656,1.37573338488067,-5.56836192711792,-5.59797201662203,-1.44348141522225,5.49216459339238,-7.19422103343416,-1.855086864706,7.09892453011937,9.00134910987265},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.92171958555151,2.94388917191153,-1.91390801783745,-4.63366448151075,6.96914454009643,5.23147784770941,-3.40113598727832,10.2353232484353,7.68327684524188,13.4470270492568,-.0451952183653594,-.0339263710725032,.0220565211009043,.0533999113042759,-.217799593005347,-.163494061616168,.10629224716917,-.614368644454251,-.461183712763862,-1.33333501160483,.000932318622845132,.000699856947415033,-.000454997367631817,.00636097475878108,.00477494739305899,.0244039729141487,-1.66187980893617e-005,-1.24751142094631e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.59519010006696,2.69877562014986,-1.7545525650326,-4.24785717273985,5.47681396366563,4.11124073580283,-2.67283724139748,6.3309248170395,4.75239001984257,5.76936797267538,-.233717104755142,-.175442746128902,.114060434690259,.276145864886065,-1.04841335581819,-.787004949964004,.511654819985593,-2.66785366107377,-2.00265861161846,-5.04596922575242,.0249321870529788,.0187156663995725,-.0121675993597457,.159138742203809,.119459540563619,.559597243722216,-.00229823078236368,-.00172519645136334,-.0189957623925199,.00019657159111697},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.03822870489707,2.2806853960415,-1.48273994391383,-3.58978558499056,3.26615690329693,2.45178262388351,-1.59397523182193,1.66498726415444,1.24984407270733,-.923099517802526,-.555278916264744,-.41682724950333,.270991524688991,.656083673232099,-2.17523616991868,-1.63286896677384,1.06157563213921,-4.50024407433993,-3.3781660071394,-6.26847807110811,.140734880491405,.105644445409619,-.0686825282564417,.792696592497175,.595047877309364,2.35258131414951,-.0308216365902829,-.0231366573273721,-.225963663364185,.00626330408481366},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,2.30018618238774,1.72666429814602,-1.12255470615748,-2.71775959032344,.987081447023724,.740965364440283,-.481723145865575,-1.35561938023839,-1.01761309753837,-2.20870753342451,-.981387965414959,-.736691479430405,.478944568701089,1.15954811601066,-3.10522366683209,-2.33097602484761,1.51543534565911,-4.42237267757001,-3.31971084311917,-2.83495589139377,.439603191878621,.329993781528466,-.214538560322786,2.03901059601619,1.53060948975261,4.55407145300733,-.170154848292169,-.127728922051078,-1.03685515398483,.0611113213546687},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,1.44662301694489,1.0859261460488,-.705992187998234,-1.70924145531706,-.724355119014703,-.543746472353678,.353505404956932,-1.6918152058931,-1.26998281171566,.160475142667942,-1.4741928888764,-1.10662182393172,.719447050735515,1.74181633275904,-3.38025060487184,-2.53742852796533,1.64965612574837,-2.16048862424162,-1.62179853107076,1.70770592621981,.991945941760336,.744616960114543,-.484097154422054,3.46032388717692,2.5975365648804,4.72311662432345,-.576746649127296,-.432942278924126,-2.68886700736678,.311154445333742},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,.553376983055106,.41539953912128,-.270063328509737,-.653836465186081,-1.45368745167752,-1.09122915405036,.709439828333211,-.194424990185747,-.145947615817496,1.49978166741586,-1.9899087271236,-1.49375054084377,.971130695151501,2.35115468793598,-2.74863027720701,-2.06329461731692,1.34140787303665,.854309601049289,.641298473185774,2.52312426315122,1.80736427869175,1.35672120663226,-.882043937561808,4.12997714955087,3.10022038625728,1.65834138994758,-1.41847476579742,-1.06479629943226,-4.48820932686106,1.03297875354965},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-.300186182387741,-.225338612975934,.146499189649511,.3546816698203,-1.13611369814234,-.852838337634571,.554455021295673,1.22655366483195,.920728260721081,.100858250439743,-2.48271365058504,-1.86368088534508,1.21163317718593,2.93342290468436,-1.26648273457434,-.950701529718779,.618078730089992,2.5074905585056,1.88227998927126,-.461151554581851,2.81340752869054,2.11192049221879,-1.3730209697231,3.19379348794833,2.39746209780118,-2.33714259619681,-2.75487590413568,-2.0679829763977,-4.77320941254596,2.50302559510192},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.03822870489707,-.779359710871422,.506684427405857,1.22670766448742,-.0622566343920988,-.0467337421144522,.030382965722629,1.10967175893133,.832989356981887,-1.41978880536247,-2.90882269973526,-2.18354511527215,1.41958622119803,3.43688734746291,.707254507004454,.530909678713955,-.34515983171824,1.57948219926848,1.18565859680248,-2.54719860772099,3.86201434387853,2.89907066463143,-1.88477020319366,.544351948717181,.408624781106939,-3.18953813043078,-4.43071582310304,-3.32597373321892,-2.19278566166381,4.71658607960963},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.59519010006696,-1.19744993497978,.778497048524628,1.88477925223671,1.2388862761662,.929985893147364,-.604610892164934,-.514957807193037,-.386559691398865,-.353775594121333,-3.23038451124486,-2.42492961864658,1.57651731119676,3.81682515580895,2.62173967966463,1.9680425606586,-1.27948173914766,-1.30644820527577,-.980702122872796,-.307822029727901,4.7630793830775,3.57546670811253,-2.32451495445633,-2.90243304119458,-2.17874863666902,.518849452524734,-6.0685481480974,-4.55543360151918,2.66498561200528,7.17423567654393},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.92171958555151,-1.44256348674145,.937852501329483,2.27058656100761,2.19799643560933,1.64995425135427,-1.07268327325536,-2.29634628619806,-1.72378183093533,2.23836106649518,-3.41890639763464,-2.56644599370298,1.66852122478611,4.03957110939074,3.91413586924262,2.93819635817362,-1.91020695459718,-4.09536544265748,-3.07423866276144,4.00097073404889,5.33523806427656,4.00496497002541,-2.60374427309723,-5.59797201662203,-4.20218958584716,5.49216459339238,-7.19422103343416,-5.40043440913241,7.09892453011937,9.00134910987265},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.92171958555151,4.61496937554466,1.68716544952149,-2.43916732844396,6.96914454009643,8.20109339929485,2.99820005631404,10.2353232484353,12.0446407028288,13.4470270492568,-.0451952183653594,-.0531844625867158,-.0194434633176023,.0281097864368307,-.217799593005347,-.256300439106579,-.0936996999356592,-.614368644454251,-.722971751522571,-1.33333501160483,.000932318622845132,.00109712634918432,.000401093381258214,.00636097475878108,.00748541629863581,.0244039729141487,-1.66187980893617e-005,-1.95565344587043e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.59519010006696,4.23071865520366,1.54668899419581,-2.23607783277939,5.47681396366563,6.44496072482988,2.35618358032457,6.3309248170395,7.45005437415178,5.76936797267538,-.233717104755142,-.275031719356679,-.100547582628472,.145363561432351,-1.04841335581819,-1.23374336817428,-.451038569416842,-2.66785366107377,-3.13945519772704,-5.04596922575242,.0249321870529788,.0293394969148428,.0107260918769641,.159138742203809,.187269998673138,.559597243722216,-.00229823078236368,-.00270449338439953,-.0189957623925199,.00019657159111697},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.03822870489707,3.57530214058608,1.30707828207097,-1.88966804781974,3.26615690329693,3.843521635486,1.4051354158397,1.66498726415444,1.95931021275678,-.923099517802526,-.555278916264744,-.653436620408273,-.238886891797976,.345363344035536,-2.17523616991868,-2.55975678291091,-.935809721212699,-4.50024407433993,-5.29576072944733,-6.26847807110811,.140734880491405,.165612851500527,.0605456414728774,.792696592497175,.932823068533675,2.35258131414951,-.0308216365902829,-.0362700355832219,-.225963663364185,.00626330408481366},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,2.30018618238774,2.70679444519172,.989564544259882,-1.43063237006769,.987081447023724,1.16156970101249,.424652930216328,-1.35561938023839,-1.5952547827211,-2.20870753342451,-.981387965414959,-1.15486977201266,-.422203533825771,.610387716162348,-3.10522366683209,-3.65413992905702,-1.33590022736237,-4.42237267757001,-5.20412385861342,-2.83495589139377,.439603191878621,.517312679568432,.189121965733584,2.03901059601619,2.39945035585009,4.55407145300733,-.170154848292169,-.200233442638163,-1.03685515398483,.0611113213546687},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,1.44662301694489,1.70234530427801,.622352597994008,-.899747042727636,-.724355119014703,-.852400743178548,-.31162527138391,-1.6918152058931,-1.9908805805112,.160475142667942,-1.4741928888764,-1.73478875376213,-.634213449888053,.916894502821386,-3.38025060487184,-3.97778389924568,-1.45421973940992,-2.16048862424162,-2.54240228389271,1.70770592621981,.991945941760336,1.16729410204273,.426745687640421,3.46032388717692,4.07201188589473,4.72311662432345,-.576746649127296,-.678699245092846,-2.68886700736678,.311154445333742},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,.553376983055106,.651198479192506,.23806865993447,-.344180410642768,-1.45368745167752,-1.71065853147384,-.625391792982136,-.194424990185747,-.228793863525153,1.49978166741586,-1.9899087271236,-2.34166865603201,-.856079885009684,1.23765104741928,-2.74863027720701,-3.23451085132423,-1.18248996086068,.854309601049289,1.00532752474992,2.52312426315122,1.80736427869175,2.12685548066823,.777547323353426,4.12997714955087,4.86004102210358,1.65834138994758,-1.41847476579742,-1.66922123290477,-4.48820932686106,1.03297875354965},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-.300186182387741,-.353250661721207,-.129143286331404,.186704916697283,-1.13611369814234,-1.33694666498547,-.488768188714077,1.22655366483195,1.44337387662839,.100858250439743,-2.48271365058504,-2.92158763778148,-1.06808980107197,1.54415783407832,-1.26648273457434,-1.49036128356923,-.54485433405009,2.5074905585056,2.95074440626681,-.461151554581851,2.81340752869054,3.31073889879025,1.21035782283977,3.19379348794833,3.75836640385388,-2.33714259619681,-2.75487590413568,-3.2418605279988,-4.77320941254596,2.50302559510192},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.03822870489707,-1.22175835711557,-.446657024142497,.645740594449334,-.0622566343920988,-.0732618573826121,-.0267834658423845,1.10967175893133,1.30583053509787,-1.41978880536247,-2.90882269973526,-3.42302078938587,-1.25140644309976,1.80918220620513,.707254507004454,.832277224231996,.304268406428257,1.57948219926848,1.8586902544778,-2.54719860772099,3.86201434387853,4.54470992402429,1.661479620483,.544351948717181,.640578072330311,-3.18953813043078,-4.43071582310304,-5.21394183895335,-2.19278566166381,4.71658607960963},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.59519010006696,-1.87717487173315,-.686267736267333,.992150379408985,1.2388862761662,1.4578865459724,.532982044151546,-.514957807193037,-.605987872997064,-.353775594121333,-3.23038451124486,-3.80142569043747,-1.38974575226926,2.00918198880831,2.62173967966463,3.08518956279543,1.12790027705863,-1.30644820527577,-1.53739152505215,-.307822029727901,4.7630793830775,5.60505793446857,2.04912737785909,-2.90243304119458,-3.41550161982096,.518849452524734,-6.0685481480974,-7.14129688166464,2.66498561200528,7.17423567654393},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.92171958555151,-2.26142559207414,-.826744191593013,1.19523987507356,2.19799643560933,2.5865404220042,.94560142914338,-2.29634628619806,-2.70227576356801,2.23836106649518,-3.41890639763464,-4.02327294720743,-1.47084987158013,2.12643576380383,3.91413586924262,4.60604507183384,1.68390285488931,-4.09536544265748,-4.81931093640565,4.00097073404889,5.33523806427656,6.27835818791885,2.29527612404421,-5.59797201662203,-6.58753611853026,5.49216459339238,-7.19422103343416,-8.46595709192435,7.09892453011937,9.00134910987265},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.92171958555151,5.87604462970729,5.45886369350783,3.32855727881994,6.96914454009643,10.4421041409331,9.70074715430413,10.2353232484353,15.3359297883224,13.4470270492568,-.0451952183653594,-.067717518868618,-.0629097851728847,-.0383594160020484,-.217799593005347,-.326336470786676,-.303167594035556,-.614368644454251,-.920529245649199,-1.33333501160483,.000932318622845132,.00139692441436278,.00129774711619287,.00636097475878108,.0095308628646168,.0244039729141487,-1.66187980893617e-005,-2.49005052754424e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.59519010006696,5.38679449650258,5.00434880168882,3.0514155628076,5.47681396366563,8.20609494322841,7.62348767011616,6.3309248170395,9.48583802706066,5.76936797267538,-.233717104755142,-.350186214905109,-.325324080359881,-.198367260350895,-1.04841335581819,-1.57087306611691,-1.45934595309525,-2.66785366107377,-3.99733505330865,-5.04596922575242,.0249321870529788,.0373567361461481,.0347045238135922,.159138742203809,.238442940887059,.559597243722216,-.00229823078236368,-.00344351663776842,-.0189957623925199,.00019657159111697},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.03822870489707,4.55228046671326,4.22908267863918,2.57869489385806,3.26615690329693,4.89379296535329,4.54634885284358,1.66498726415444,2.49470653361407,-.923099517802526,-.555278916264744,-.831993114526565,-.772924185272261,-.471292665829706,-2.17523616991868,-3.25922966624351,-3.02783447385586,-4.50024407433993,-6.74286736535737,-6.26847807110811,.140734880491405,.210867814600683,.195896854176436,.792696592497175,1.18772402074228,2.35258131414951,-.0308216365902829,-.0461810968751945,-.225963663364185,.00626330408481366},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,2.30018618238774,3.44644648080968,3.20175947449313,1.95228172055832,.987081447023724,1.47897739903412,1.37397459375255,-1.35561938023839,-2.03117020639316,-2.20870753342451,-.981387965414959,-1.47044666380812,-1.36604951383154,-.832952480070553,-3.10522366683209,-4.65266127660931,-4.32233676392885,-4.42237267757001,-6.6261900271284,-2.83495589139377,.439603191878621,.658672278252945,.611908590636856,2.03901059601619,3.05511829665785,4.55407145300733,-.170154848292169,-.254948743871766,-1.03685515398483,.0611113213546687},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,1.44662301694489,2.16752402217829,2.01363654211465,1.2278204669453,-.724355119014703,-1.08532568728512,-1.00827093182807,-1.6918152058931,-2.53490374291394,.160475142667942,-1.4741928888764,-2.20883288938782,-2.05201260878721,-1.25122037987573,-3.38025060487184,-5.06474340721712,-4.70516234202085,-2.16048862424162,-3.23713291923203,1.70770592621981,.991945941760336,1.48626603590141,1.38074575986725,3.46032388717692,5.18471990323707,4.72311662432345,-.576746649127296,-.864158942299273,-2.68886700736678,.311154445333742},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,.553376983055106,.829143384311425,.770276776736346,.469678401195648,-1.45368745167752,-2.17810890152299,-2.02346992933989,-.194424990185747,-.291313515578485,1.49978166741586,-1.9899087271236,-2.9815473107461,-2.76986670414993,-1.68893390563517,-2.74863027720701,-4.11836538252447,-3.82597422031764,.854309601049289,1.2800408674882,2.52312426315122,1.80736427869175,2.70803481201197,2.51577274454146,4.12997714955087,6.18808395553405,1.65834138994758,-1.41847476579742,-2.12534854800545,-4.48820932686106,1.03297875354965},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-.300186182387741,-.449779074319961,-.417846155642134,-.254782852417372,-1.13611369814234,-1.70227744362135,-1.58142102819233,1.22655366483195,1.83778669504629,.100858250439743,-2.48271365058504,-3.7199335363258,-3.45582979910559,-2.10720180544034,-1.26648273457434,-1.89761376598671,-1.76288907720341,2.5074905585056,3.75705761211831,-.461151554581851,2.81340752869054,4.21542331996591,3.91614134649752,3.19379348794833,4.78536842321548,-2.33714259619681,-2.75487590413568,-4.12772341388705,-4.77320941254596,2.50302559510192},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.03822870489707,-1.55561306022354,-1.44516935978819,-.881196025717112,-.0622566343920988,-.0932812135041899,-.0866585368463816,1.10967175893133,1.66265859611778,-1.41978880536247,-2.90882269973526,-4.35838708560735,-4.04895512766487,-2.46886161968119,.707254507004454,1.0597032647262,.984467621360889,1.57948219926848,2.36659141141637,-2.54719860772099,3.86201434387853,5.78658625215441,5.37575658648661,.544351948717181,.815620871022689,-3.18953813043078,-4.43071582310304,-6.63869084478425,-2.19278566166381,4.71658607960963},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.59519010006696,-2.39012709001286,-2.22043548283783,-1.35391669466665,1.2388862761662,1.85626506095126,1.72447600259692,-.514957807193037,-.771578638346196,-.353775594121333,-3.23038451124486,-4.84019398522881,-4.49655523257725,-2.74178702516,2.62173967966463,3.92824092358218,3.64934800797338,-1.30644820527577,-1.95749537633152,-.307822029727901,4.7630793830775,7.13668236880667,6.6300000687783,-2.90243304119458,-4.3488132461414,.518849452524734,-6.0685481480974,-9.09271021667365,2.66498561200528,7.17423567654393},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.92171958555151,-2.87937722321757,-2.67495037465683,-1.63105841067899,2.19799643560933,3.29333213710544,3.05951577632408,-2.29634628619806,-3.44069303528758,2.23836106649518,-3.41890639763464,-5.1226626812653,-4.75896952776425,-2.90179486950885,3.91413586924262,5.86468169257239,5.4483074914535,-4.09536544265748,-6.13622406614863,4.00097073404889,5.33523806427656,7.99396700420041,7.42642015557747,-5.59797201662203,-8.38763014060611,5.49216459339238,-7.19422103343416,-10.7793438389747,7.09892453011937,9.00134910987265},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.92171958555151,6.61537460386721,8.09172854779269,8.80395991076745,6.96914454009643,11.7559404153652,14.3795150585562,10.2353232484353,17.2655122351262,13.4470270492568,-.0451952183653594,-.0762378066864121,-.0932518071891671,-.101459801467564,-.217799593005347,-.367396460919218,-.449388373324038,-.614368644454251,-1.03635118137081,-1.33333501160483,.000932318622845132,.00157268687980642,.00192366360061384,.00636097475878108,.0107300458251741,.0244039729141487,-1.66187980893617e-005,-2.80335124396135e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.59519010006696,6.06456651609707,7.41799655299993,8.07092624092508,5.47681396366563,9.23859424244377,11.300372436005,6.3309248170395,10.6793559162843,5.76936797267538,-.233717104755142,-.394247004353173,-.482231155810537,-.524676988090528,-1.04841335581819,-1.76852192969244,-2.1632031816091,-2.66785366107377,-4.50028385780377,-5.04596922575242,.0249321870529788,.0420569990798227,.0514428646412204,.159138742203809,.268444076758865,.559597243722216,-.00229823078236368,-.0038767834400574,-.0189957623925199,.00019657159111697},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.03822870489707,5.12505307344404,6.26881178264598,6.82059059403532,3.26615690329693,5.50953502563763,6.7390986231311,1.66498726415444,2.80859307349989,-.923099517802526,-.555278916264744,-.936675343241156,-1.14571329243577,-1.24655861042429,-2.17523616991868,-3.66930929238228,-4.48818948881768,-4.50024407433993,-7.59126186695628,-6.26847807110811,.140734880491405,.237399419743997,.29037987690505,.792696592497175,1.33716467766395,2.35258131414951,-.0308216365902829,-.0519916499446519,-.225963663364185,.00626330408481366},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,2.30018618238774,3.88008191896113,4.74600026627036,5.1637416942433,.987081447023724,1.66506385533551,2.03665635715207,-1.35561938023839,-2.28673412920868,-2.20870753342451,-.981387965414959,-1.65545977423627,-2.02490893148973,-2.20314076875832,-3.10522366683209,-5.23806390000324,-6.40704325104372,-4.42237267757001,-7.45990406612127,-2.83495589139377,.439603191878621,.741547101224613,.907038257318,2.03901059601619,3.43951642050874,4.55407145300733,-.170154848292169,-.287026656838231,-1.03685515398483,.0611113213546687},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,1.44662301694489,2.44024412222762,2.98483369571688,3.24755780451479,-.724355119014703,-1.22188248065577,-1.49457014152702,-1.6918152058931,-2.85384793698481,.160475142667942,-1.4741928888764,-2.48675051356285,-3.04171892525937,-3.30945005334757,-3.38025060487184,-5.70199462778367,-6.97450945518368,-2.16048862424162,-3.64443230995284,1.70770592621981,.991945941760336,1.67326955598402,2.04669332477546,3.46032388717692,5.8370666898335,4.72311662432345,-.576746649127296,-.972888308521261,-2.68886700736678,.311154445333742},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,.553376983055106,.93346705704166,1.14178901214037,1.24228891639978,-1.45368745167752,-2.45216080196962,-2.99940982448642,-.194424990185747,-.327966881493408,1.49978166741586,-1.9899087271236,-3.35668865754018,-4.10580126963162,-4.46719258573777,-2.74863027720701,-4.6365423474252,-5.67128006117502,.854309601049289,1.44109692494208,2.52312426315122,1.80736427869175,3.04876253512471,3.72915523810848,4.12997714955087,6.96667503835439,1.65834138994758,-1.41847476579742,-2.39276208655033,-4.48820932686106,1.03297875354965},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-.300186182387741,-.506370739691845,-.619377558413115,-.673894973328732,-1.13611369814234,-1.91645974101964,-2.34415629302528,1.22655366483195,2.06901890572263,.100858250439743,-2.48271365058504,-4.18797939686675,-5.12261126340126,-5.57350187032702,-1.26648273457434,-2.13637348021345,-2.61314820693614,2.5074905585056,4.22977446202813,-.461151554581851,2.81340752869054,4.7458122143026,5.80493569904152,3.19379348794833,5.38746839617942,-2.33714259619681,-2.75487590413568,-4.64707781591113,-4.77320941254596,2.50302559510192},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.03822870489707,-1.75134189417476,-2.14218907478874,-2.33074387312075,-.0622566343920988,-.105017951653011,-.12845482061473,1.10967175893133,1.87185600940662,-1.41978880536247,-2.90882269973526,-4.90676382786187,-6.00180690245521,-6.53008402866105,.707254507004454,1.19303621855815,1.45928625505293,1.57948219926848,2.66435837496252,-2.54719860772099,3.86201434387853,6.51466048131367,7.96853804732135,.544351948717181,.918243127233759,-3.18953813043078,-4.43071582310304,-7.47397774950175,-2.19278566166381,4.71658607960963},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.59519010006696,-2.69085533682779,-3.29137384514269,-3.58107952001051,1.2388862761662,2.0898222386161,2.55620812016103,-.514957807193037,-.868659455579414,-.353775594121333,-3.23038451124486,-5.44919216674985,-6.66528903908045,-7.25196565099482,2.62173967966463,4.42249623366662,5.4094652504454,-1.30644820527577,-2.20378945631257,-.307822029727901,4.7630793830775,8.03462707886587,9.82771577392637,-2.90243304119458,-4.89598539807521,.518849452524734,-6.0685481480974,-10.2367643607743,2.66498561200528,7.17423567654393},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.92171958555151,-3.24166342459793,-3.96510583993544,-4.31411318985288,2.19799643560933,3.70770257117539,4.53515100205683,-2.29634628619806,-3.87360456901059,2.23836106649518,-3.41890639763464,-5.76720136441661,-7.05426838770182,-7.67518283761778,3.91413586924262,6.60258197030424,8.07608098333328,-4.09536544265748,-6.90829008438974,4.00097073404889,5.33523806427656,8.99977614812197,11.0082572774727,-5.59797201662203,-9.44297038492046,5.49216459339238,-7.19422103343416,-12.1356119587979,7.09892453011937,9.00134910987265}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 100; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_hyp_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00222253422114756,-.00498206115568338,-.00730339144757066,-.00897618736857064,-.00985149899052509,-.00985149899052509,-.00897618736857064,-.00730339144757066,-.00498206115568338,-.00222253422114756,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256888,-.0397874192256888,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256889,-.0397874192256889,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00222253422114756,-.00498206115568339,-.00730339144757066,-.00897618736857064,-.0098514989905251,-.0098514989905251,-.00897618736857064,-.00730339144757066,-.00498206115568339,-.00222253422114756},
			{.00374909427250663,.00840401770480107,.0123197666813982,.0151415318298581,.0166180561313977,.0166180561313977,.0151415318298581,.0123197666813982,.00840401770480107,.00374909427250663,.00746479013903255,.0167331691152994,.0245297840368778,.0301481769402938,.0330880720843031,.0330880720843031,.0301481769402938,.0245297840368778,.0167331691152994,.00746479013903255,.00859442576735982,.0192653747976306,.0282418399000882,.0347104558746731,.0380952409936669,.0380952409936669,.0347104558746731,.0282418399000882,.0192653747976306,.00859442576735982,.00673809032372206,.0151041895084832,.0221418013612503,.0272132418374432,.0298669372064565,.0298669372064565,.0272132418374432,.0221418013612503,.0151041895084832,.00673809032372206,.00254028702921173,.00569433991706437,.00834754776191493,.0102595011258728,.0112599549045422,.0112599549045422,.0102595011258728,.00834754776191493,.00569433991706437,.00254028702921173,-.00254028702921173,-.00569433991706437,-.00834754776191493,-.0102595011258728,-.0112599549045422,-.0112599549045422,-.0102595011258728,-.00834754776191493,-.00569433991706437,-.00254028702921173,-.00673809032372206,-.0151041895084832,-.0221418013612503,-.0272132418374432,-.0298669372064565,-.0298669372064565,-.0272132418374432,-.0221418013612503,-.0151041895084832,-.00673809032372206,-.00859442576735982,-.0192653747976306,-.0282418399000882,-.0347104558746731,-.0380952409936669,-.0380952409936669,-.0347104558746731,-.0282418399000882,-.0192653747976306,-.00859442576735982,-.00746479013903255,-.0167331691152994,-.0245297840368778,-.0301481769402938,-.0330880720843031,-.0330880720843031,-.0301481769402938,-.0245297840368778,-.0167331691152994,-.00746479013903255,-.00374909427250663,-.00840401770480107,-.0123197666813982,-.0151415318298581,-.0166180561313977,-.0166180561313977,-.0151415318298581,-.0123197666813982,-.00840401770480107,-.00374909427250663},
			{-.00458578009255966,-.0102795433475266,-.015069170494548,-.0185206693108299,-.0203267097184652,-.0203267097184652,-.0185206693108299,-.015069170494548,-.0102795433475266,-.00458578009255966,-.00693481320266981,-.0155451660317305,-.0227882454870494,-.0280077499283953,-.0307389216397789,-.0307389216397789,-.0280077499283953,-.0227882454870494,-.0155451660317305,-.00693481320266981,-.00314199662793749,-.00704313985467146,-.0103248044877915,-.0126896360809338,-.0139270641206943,-.0139270641206943,-.0126896360809338,-.0103248044877915,-.00704313985467146,-.0031419966279375,.00438062859874147,.00981967313331039,.0143950293941988,.017692120427338,.019417365009608,.019417365009608,.017692120427338,.0143950293941988,.00981967313331039,.00438062859874147,.0102819614083243,.0230481762886867,.0337871913508869,.0415259352316643,.0455753308412161,.0455753308412161,.0415259352316643,.0337871913508869,.0230481762886867,.0102819614083243,.0102819614083243,.0230481762886867,.0337871913508869,.0415259352316643,.0455753308412161,.0455753308412161,.0415259352316643,.0337871913508869,.0230481762886867,.0102819614083243,.00438062859874146,.00981967313331036,.0143950293941988,.017692120427338,.0194173650096079,.0194173650096079,.017692120427338,.0143950293941988,.00981967313331036,.00438062859874146,-.00314199662793748,-.00704313985467143,-.0103248044877915,-.0126896360809337,-.0139270641206943,-.0139270641206943,-.0126896360809337,-.0103248044877915,-.00704313985467143,-.00314199662793748,-.00693481320266981,-.0155451660317305,-.0227882454870495,-.0280077499283953,-.0307389216397789,-.0307389216397789,-.0280077499283953,-.0227882454870495,-.0155451660317305,-.00693481320266981,-.00458578009255966,-.0102795433475266,-.015069170494548,-.0185206693108299,-.0203267097184652,-.0203267097184652,-.0185206693108299,-.015069170494548,-.0102795433475266,-.00458578009255966},
			{.00498941899157684,.0111843454696186,.0163955540682807,.0201508527079391,.0221158602153926,.0221158602153926,.0201508527079391,.0163955540682807,.0111843454696186,.00498941899157684,.00422852158563388,.00947870810595568,.01389519589419,.0170778032048017,.0187431427313533,.0187431427313533,.0170778032048017,.01389519589419,.00947870810595568,.00422852158563388,-.00454244456135888,-.0101824018661483,-.0149267671313349,-.0183456493520488,-.0201346227135425,-.0201346227135425,-.0183456493520488,-.0149267671313349,-.0101824018661483,-.00454244456135888,-.0106057150885859,-.0237739066819889,-.0348510669200656,-.0428334848152169,-.0470103859345993,-.0470103859345993,-.0428334848152169,-.0348510669200656,-.0237739066819889,-.0106057150885859,-.00560552264733118,-.0125654113096862,-.0184201105981369,-.022639121189878,-.0248467718409463,-.0248467718409463,-.022639121189878,-.0184201105981369,-.0125654113096862,-.00560552264733118,.00560552264733123,.0125654113096863,.0184201105981371,.0226391211898782,.0248467718409465,.0248467718409465,.0226391211898782,.0184201105981371,.0125654113096863,.00560552264733123,.0106057150885859,.023773906681989,.0348510669200657,.042833484815217,.0470103859345994,.0470103859345994,.042833484815217,.0348510669200657,.023773906681989,.0106057150885859,.00454244456135883,.0101824018661482,.0149267671313347,.0183456493520486,.0201346227135423,.0201346227135423,.0183456493520486,.0149267671313347,.0101824018661482,.00454244456135883,-.00422852158563391,-.00947870810595576,-.0138951958941901,-.0170778032048018,-.0187431427313534,-.0187431427313534,-.0170778032048018,-.0138951958941901,-.00947870810595576,-.00422852158563391,-.00498941899157684,-.0111843454696186,-.0163955540682807,-.0201508527079391,-.0221158602153927,-.0221158602153927,-.0201508527079391,-.0163955540682807,-.0111843454696186,-.00498941899157684},
			{-.00502787889479981,-.0112705576809201,-.0165219358020692,-.020306181383746,-.0222863357447103,-.0222863357447103,-.020306181383746,-.0165219358020692,-.0112705576809201,-.00502787889479981,-.000280476734556914,-.000628720237921658,-.000921664721701268,-.00113276623502688,-.00124322777173102,-.00124322777173102,-.00113276623502688,-.000921664721701268,-.000628720237921658,-.000280476734556914,.009285522300677,.0208145456318548,.0305128278843669,.0375015993872448,.0411585624649621,.0411585624649621,.0375015993872448,.030512827884367,.0208145456318548,.009285522300677,.00471291107004208,.010564521773763,.0154869311233076,.0190341153866534,.020890224414624,.020890224414624,.0190341153866534,.0154869311233076,.010564521773763,.00471291107004208,-.00869007782643158,-.0194797896774683,-.028556158763447,-.0350967674986958,-.0385192237402185,-.0385192237402185,-.0350967674986958,-.028556158763447,-.0194797896774683,-.00869007782643158,-.00869007782643163,-.0194797896774684,-.0285561587634472,-.0350967674986961,-.0385192237402188,-.0385192237402188,-.0350967674986961,-.0285561587634472,-.0194797896774684,-.00869007782643163,.00471291107004194,.0105645217737627,.0154869311233072,.0190341153866528,.0208902244146234,.0208902244146234,.0190341153866528,.0154869311233072,.0105645217737627,.00471291107004194,.00928552230067672,.0208145456318542,.0305128278843661,.0375015993872437,.0411585624649609,.0411585624649609,.0375015993872437,.030512827884366,.0208145456318542,.00928552230067672,-.000280476734556647,-.00062872023792106,-.000921664721700392,-.0011327662350258,-.00124322777172983,-.00124322777172983,-.0011327662350258,-.000921664721700392,-.00062872023792106,-.000280476734556647,-.00502787889479991,-.0112705576809204,-.0165219358020696,-.0203061813837464,-.0222863357447108,-.0222863357447108,-.0203061813837464,-.0165219358020696,-.0112705576809204,-.00502787889479991},
			{.00302011496343164,.00561961421020581,.00536170126189668,.00190531860926006,-.00385485822902238,-.0100772652494075,-.0145995645209458,-.0156902564956459,-.0126653126632867,-.00616325300098763,.00676992835558772,.0125969991373797,.0120188581714983,.00427098658012745,-.00864109954336841,-.0225893267590133,-.032726570685666,-.0351714797755983,-.0283907269656191,-.0138156268086686,.00992429344157325,.0184664163867283,.0176188977285336,.00626100039466762,-.0126673138949028,-.0331145465105248,-.0479751149143058,-.0515591994085577,-.0416190379908282,-.0202528486753225,.0121973904961468,.0226960330283889,.0216543955468261,.007695043194751,-.0155686825488407,-.040699225317108,-.0589635135390092,-.0633685100664642,-.0511516171339274,-.0248916364078325,.0133868172784104,.024909233429215,.0237660208182528,.00844542422661216,-.0170868603913883,-.044668004428086,-.0647133320925609,-.0695478812236805,-.0561396597315967,-.0273189407568403,.0133868172784104,.024909233429215,.0237660208182528,.00844542422661216,-.0170868603913883,-.044668004428086,-.0647133320925609,-.0695478812236805,-.0561396597315967,-.0273189407568403,.0121973904961468,.0226960330283889,.0216543955468261,.007695043194751,-.0155686825488407,-.040699225317108,-.0589635135390092,-.0633685100664642,-.0511516171339274,-.0248916364078325,.00992429344157325,.0184664163867283,.0176188977285336,.00626100039466762,-.0126673138949028,-.0331145465105248,-.0479751149143058,-.0515591994085576,-.0416190379908282,-.0202528486753225,.00676992835558772,.0125969991373797,.0120188581714983,.00427098658012745,-.00864109954336841,-.0225893267590133,-.032726570685666,-.0351714797755983,-.0283907269656191,-.0138156268086686,.00302011496343164,.00561961421020581,.00536170126189668,.00190531860926006,-.00385485822902238,-.0100772652494075,-.0145995645209458,-.0156902564956459,-.0126653126632867,-.00616325300098763},
			{-.00509449780740388,-.00947947764207594,-.00904441574358152,-.00321399734606523,.00650258915087549,.016998891214194,.0246273570183889,.0264671968724815,.0213645534605355,.0103965177750484,-.0101436118237301,-.0188745083672825,-.0180082602728758,-.00639936313911146,.0129472506789226,.0338463496166405,.0490353238499585,.0526986135407738,.0425387830722098,.0207004192788813,-.0116786295673759,-.0217307597449812,-.0207334236102965,-.0073677692785882,.014906538935309,.0389682675411454,.0564557667339239,.0606734166243637,.0489761140686684,.0238329830486871,-.00915612782195732,-.0170370686684944,-.0162551501156413,-.00577638299838646,.0116868314974169,.0305513961675494,.0442617186819942,.0475683858968708,.0383976183205756,.0186852265424968,-.00345189684709876,-.0064230430990214,-.00612825667404284,-.00217771951719025,.00440598226487886,.01151799867323,.0166868451529422,.0179334719317936,.0144760558386965,.00704440521621975,.00345189684709876,.0064230430990214,.00612825667404284,.00217771951719025,-.00440598226487886,-.01151799867323,-.0166868451529422,-.0179334719317936,-.0144760558386965,-.00704440521621975,.00915612782195732,.0170370686684944,.0162551501156413,.00577638299838646,-.0116868314974169,-.0305513961675494,-.0442617186819942,-.0475683858968708,-.0383976183205756,-.0186852265424968,.0116786295673759,.0217307597449812,.0207334236102965,.0073677692785882,-.014906538935309,-.0389682675411454,-.0564557667339239,-.0606734166243637,-.0489761140686684,-.0238329830486871,.0101436118237301,.0188745083672825,.0180082602728758,.00639936313911146,-.0129472506789226,-.0338463496166405,-.0490353238499585,-.0526986135407738,-.0425387830722098,-.0207004192788813,.00509449780740388,.00947947764207594,.00904441574358152,.00321399734606523,-.00650258915087549,-.016998891214194,-.0246273570183889,-.0264671968724815,-.0213645534605355,-.0103965177750484},
			{.00623143749421823,.0115950138046163,.0110628590900448,.00393126551933778,-.00795377275172705,-.0207925358056529,-.0301234472384809,-.0323738843735185,-.0261324834192153,-.0127167098940724,.00942344685839383,.017534476838053,.0167297296705077,.00594502821252533,-.0120280360541503,-.0314433638141471,-.045553968006072,-.0489571754954932,-.0395186614657246,-.0192307537404636,.00426953652352492,.00794444860844995,.00757983707344909,.00269354891773532,-.00544961307801359,-.0142462298821585,-.0206394043619152,-.0221813156064483,-.0179049525109209,-.00871299076692808,-.00595266514038301,-.0110762941201031,-.0105679461150626,-.00375539468000193,.00759794924327869,.0198623516941821,.0287758313312213,.0309255919116317,.0249634090409368,.0121478095153878,-.0139717558497607,-.0259976453431518,-.0248044799213386,-.00881444804152065,.0178334727861298,.0466197781881845,.0675409888263679,.0725867841561221,.0585926888330286,.0285126451187338,-.0139717558497607,-.0259976453431518,-.0248044799213386,-.00881444804152065,.0178334727861298,.0466197781881845,.0675409888263679,.0725867841561221,.0585926888330287,.0285126451187339,-.00595266514038299,-.011076294120103,-.0105679461150626,-.00375539468000192,.00759794924327867,.0198623516941821,.0287758313312212,.0309255919116316,.0249634090409367,.0121478095153878,.0042695365235249,.00794444860844991,.00757983707344906,.0026935489177353,-.00544961307801357,-.0142462298821584,-.0206394043619151,-.0221813156064482,-.0179049525109208,-.00871299076692804,.00942344685839383,.017534476838053,.0167297296705077,.00594502821252533,-.0120280360541503,-.0314433638141471,-.045553968006072,-.0489571754954932,-.0395186614657246,-.0192307537404636,.00623143749421823,.0115950138046163,.0110628590900448,.00393126551933778,-.00795377275172705,-.0207925358056529,-.0301234472384809,-.0323738843735185,-.0261324834192153,-.0127167098940724},
			{-.00677992663267428,-.0126156032172487,-.0120366084146251,-.00427729425504189,.00865386128958332,.0226226881679193,.0327749034455745,.0352234233386421,.0284326562653077,.0138360306383256,-.00574597286050269,-.0106916958888372,-.0102009990712324,-.00362499744280574,.00733412834718052,.0191726783027454,.0277766583485262,.02985177355496,.0240966134449184,.0117260054351862,.00617255053363847,.0114854411544103,.0109583152913243,.00389411513824503,-.00787860975716505,-.0205960467551767,-.0298387812602527,-.0320679518090523,-.0258855319698288,-.0125965372380952,.0144116920845997,.0268162472661978,.025585511999273,.009091993335417,-.018395004991244,-.0480877203642201,-.0696676884796602,-.0748723634966211,-.0604376285236068,-.0294104382002705,.00761712583185733,.0141734036895712,.0135229134252192,.00480546329270283,-.00972245777063737,-.025416183945038,-.0368220155169205,-.0395728836513071,-.0319435787790183,-.0155445319832301,-.0076171258318574,-.0141734036895713,-.0135229134252194,-.00480546329270287,.00972245777063746,.0254161839450382,.0368220155169208,.0395728836513074,.0319435787790186,.0155445319832303,-.0144116920845997,-.0268162472661979,-.0255855119992731,-.00909199333541702,.018395004991244,.0480877203642202,.0696676884796604,.0748723634966213,.060437628523607,.0294104382002706,-.0061725505336384,-.0114854411544102,-.0109583152913242,-.00389411513824499,.00787860975716497,.0205960467551765,.0298387812602524,.0320679518090519,.0258855319698285,.012596537238095,.00574597286050274,.0106916958888373,.0102009990712325,.00362499744280577,-.00733412834718057,-.0191726783027455,-.0277766583485264,-.0298517735549602,-.0240966134449186,-.0117260054351863,.00677992663267429,.0126156032172487,.0120366084146252,.0042772942550419,-.00865386128958333,-.0226226881679194,-.0327749034455745,-.0352234233386422,-.0284326562653077,-.0138360306383256},
			{-.00345430310158809,-.00436440955976362,.000321510543493063,.00721105333392678,.0101264765379077,.00504590248017281,-.00626512731169085,-.0168673409888969,-.0192939898358052,-.0109524916466014,-.00774321004312494,-.00978331632795496,.00072070214921253,.0161643894454719,.022699639413156,.0113109595805708,-.0140439895674001,-.0378100474408264,-.043249654554018,-.0245512454527271,-.0113510637943875,-.0143417325812942,.0010565044764254,.0236959884558197,.0332762579929111,.016581162471344,-.0205876142606789,-.0554271753160955,-.0634013006484006,-.035990597157184,-.0139509536333168,-.0176266163143223,.00129849018831963,.0291234057204582,.0408979757984384,.0203789735494738,-.0253030779470516,-.0681224215516175,-.0779229701867376,-.0442340172930329,-.015311378873029,-.0193454732724202,.00142511227252541,.0319633704461579,.0448861360340558,.0223662262280236,-.0277705039586591,-.074765369704482,-.0855216174320572,-.0485474911358245,-.015311378873029,-.0193454732724202,.00142511227252541,.0319633704461579,.0448861360340558,.0223662262280236,-.0277705039586591,-.074765369704482,-.0855216174320572,-.0485474911358245,-.0139509536333168,-.0176266163143223,.00129849018831963,.0291234057204581,.0408979757984384,.0203789735494738,-.0253030779470516,-.0681224215516175,-.0779229701867376,-.0442340172930329,-.0113510637943875,-.0143417325812942,.0010565044764254,.0236959884558197,.0332762579929111,.016581162471344,-.0205876142606789,-.0554271753160955,-.0634013006484006,-.035990597157184,-.00774321004312494,-.00978331632795496,.00072070214921253,.0161643894454719,.022699639413156,.0113109595805708,-.0140439895674001,-.0378100474408264,-.043249654554018,-.0245512454527271,-.00345430310158809,-.00436440955976362,.000321510543493063,.00721105333392678,.0101264765379077,.00504590248017282,-.00626512731169085,-.0168673409888969,-.0192939898358052,-.0109524916466014},
			{.00582691049477723,.00736212866080022,-.000542341857377823,-.0121640056225925,-.0170819035500163,-.00851170880280428,.0105683650247204,.0284527684272998,.0325461745984863,.0184752717531363,.0116019125780642,.0146586725826016,-.00107985232008665,-.0242196494967433,-.0340116347817334,-.0169475919543621,.0210425828953059,.0566520683977071,.0648024150672126,.0367859585157503,.0133576128135457,.0168769477792085,-.00124326477125837,-.0278847731596392,-.0391585650653061,-.0195122459271369,.0242269257780715,.0652251419454623,.0746088684970023,.0423527230982234,.0104724625103868,.0132316459066911,-.000974728335761049,-.0218618585222674,-.0307005907665702,-.0152977382124873,.0189940804164865,.0511369706019496,.0584938782987402,.0332048331576998,.0039481602948217,.00498838348212032,-.000367476485083857,-.00824200819081637,-.0115742456343885,-.00576730855338134,.00716084436326645,.0192788426530853,.0220524263094917,.0125183550420337,-.0039481602948217,-.00498838348212032,.000367476485083857,.00824200819081637,.0115742456343885,.00576730855338134,-.00716084436326645,-.0192788426530853,-.0220524263094917,-.0125183550420337,-.0104724625103868,-.0132316459066911,.000974728335761049,.0218618585222673,.0307005907665701,.0152977382124873,-.0189940804164865,-.0511369706019496,-.0584938782987402,-.0332048331576998,-.0133576128135457,-.0168769477792085,.00124326477125837,.0278847731596392,.0391585650653061,.0195122459271369,-.0242269257780715,-.0652251419454623,-.0746088684970023,-.0423527230982234,-.0116019125780642,-.0146586725826016,.00107985232008665,.0242196494967433,.0340116347817334,.0169475919543621,-.0210425828953059,-.0566520683977071,-.0648024150672126,-.0367859585157503,-.00582691049477723,-.00736212866080022,.000542341857377823,.0121640056225925,.0170819035500163,.00851170880280428,-.0105683650247204,-.0284527684272998,-.0325461745984863,-.0184752717531363},
			{-.00712730281031248,-.00900513579898562,.000663376354880294,.0148786482195442,.0208940740185803,.0104112678794644,-.0129269083176432,-.0348025761773465,-.039809508295799,-.0225984003710785,-.0107782127862115,-.013617952315673,.00100318615618461,.0225001295369971,.0315969142516674,.0157443655146451,-.0195486248056994,-.0526299473350607,-.0602016446818767,-.0341742696094045,-.00488334829500983,-.00616996579501025,.000454519455363577,.0101942660987876,.0143157998825042,.00713339233665088,-.00885701048107196,-.0238453599574428,-.0272759134325242,-.0154835374417349,.00680845262802778,.008602278046562,-.000633699255799995,-.0142130304082612,-.0199593474485439,-.00994550477811579,.012348604408976,.0332456326808434,.0380285724615008,.0215874283007975,.015980411393903,.0201907760278596,-.00148738272276524,-.0333600137192788,-.0468474410864738,-.0233435211431231,.0289839394319417,.0780322514257342,.0892584946770308,.0506687795348675,.015980411393903,.0201907760278596,-.00148738272276524,-.0333600137192788,-.0468474410864738,-.0233435211431231,.0289839394319417,.0780322514257343,.0892584946770308,.0506687795348676,.00680845262802777,.00860227804656198,-.000633699255799993,-.0142130304082611,-.0199593474485438,-.00994550477811577,.012348604408976,.0332456326808434,.0380285724615007,.0215874283007975,-.00488334829500981,-.00616996579501022,.000454519455363575,.0101942660987876,.0143157998825041,.00713339233665084,-.00885701048107192,-.0238453599574427,-.0272759134325241,-.0154835374417348,-.0107782127862115,-.013617952315673,.00100318615618461,.0225001295369971,.0315969142516674,.0157443655146451,-.0195486248056994,-.0526299473350607,-.0602016446818768,-.0341742696094045,-.00712730281031249,-.00900513579898562,.000663376354880294,.0148786482195442,.0208940740185803,.0104112678794644,-.0129269083176432,-.0348025761773465,-.039809508295799,-.0225984003710785},
			{.00360886668167895,.00181411871372813,-.00573065302975065,-.00778508692673576,.00135437648599479,.0117852891818123,.00860428289223646,-.00859845646519095,-.022302893606761,-.0160855169229064,.00808968174247022,.00406655172706256,-.0128459051764245,-.0174511504939989,.00303598766527052,.0264180550663856,.0192874706549125,-.0192744100614766,-.0499944517570629,-.0360575006081668,.011858969732663,.00596131162915412,-.0188312971419022,-.0255823000331391,.00445056888235353,.0387272238144227,.0282742310510711,-.0282550850343082,-.0732887533851304,-.0528580507812608,.0145751922353826,.00732671260055841,-.0231444874278468,-.0314417651121289,.00546994372019093,.047597451090816,.0347502660152623,-.0347267347236618,-.0900750818460811,-.0649648551848937,.015996490012713,.00804117592743191,-.0254014187950224,-.0345078043209106,.00600334449639179,.0522389096972928,.0381389332143976,-.0381131072722766,-.0988587405143971,-.0712998937070431,.015996490012713,.00804117592743191,-.0254014187950224,-.0345078043209106,.00600334449639179,.0522389096972928,.0381389332143976,-.0381131072722766,-.0988587405143971,-.0712998937070431,.0145751922353826,.00732671260055841,-.0231444874278468,-.0314417651121289,.00546994372019093,.0475974510908161,.0347502660152623,-.0347267347236618,-.0900750818460811,-.0649648551848937,.011858969732663,.00596131162915412,-.0188312971419022,-.0255823000331391,.00445056888235353,.0387272238144227,.0282742310510711,-.0282550850343082,-.0732887533851304,-.0528580507812608,.00808968174247022,.00406655172706256,-.0128459051764245,-.0174511504939989,.00303598766527052,.0264180550663856,.0192874706549125,-.0192744100614766,-.0499944517570629,-.0360575006081668,.00360886668167895,.00181411871372813,-.00573065302975065,-.00778508692673576,.00135437648599479,.0117852891818123,.00860428289223646,-.00859845646519095,-.022302893606761,-.0160855169229064},
			{-.00608763693165392,-.00306015629121496,.00966678409139083,.0131323173925849,-.00228463754490087,-.0198800809233674,-.0145141826854097,.0145043543443846,.0376217607296319,.0271339441220568,-.0121210427981259,-.00609305150606168,.0192474526663595,.0261476469344647,-.00454892263958079,-.0395830622633527,-.0288990673203423,.0288794981930212,.0749083720110797,.0540261683930699,-.0139553022404171,-.0070151039600872,.0221601411521044,.0301045315921654,-.00523730436076351,-.0455731084104202,-.0332723203472142,.0332497898528585,.086244144927353,.0622018683849159,-.0109410552299944,-.00549989090520571,.0173737067153911,.0236021647649496,-.00410608350003688,-.0357296379204776,-.0260857334565373,.0260680693903023,.0676160169559761,.0487667028405305,-.00412482162620984,-.00207348089107479,.00654995699049303,.00889811061193973,-.00154800991896502,-.0134702165461154,-.00983442595208666,.00982776652839477,.0254915091054086,.0183852422172328,.00412482162620984,.00207348089107479,-.00654995699049303,-.00889811061193973,.00154800991896502,.0134702165461154,.00983442595208666,-.00982776652839477,-.0254915091054086,-.0183852422172328,.0109410552299944,.00549989090520571,-.0173737067153911,-.0236021647649496,.00410608350003688,.0357296379204776,.0260857334565373,-.0260680693903023,-.0676160169559761,-.0487667028405305,.0139553022404171,.0070151039600872,-.0221601411521044,-.0301045315921654,.00523730436076351,.0455731084104202,.0332723203472142,-.0332497898528585,-.086244144927353,-.0622018683849159,.0121210427981259,.00609305150606168,-.0192474526663595,-.0261476469344647,.00454892263958079,.0395830622633527,.0288990673203423,-.0288794981930212,-.0749083720110797,-.0540261683930699,.00608763693165392,.00306015629121496,-.00966678409139083,-.0131323173925849,.00228463754490087,.0198800809233674,.0145141826854097,-.0145043543443846,-.0376217607296319,-.0271339441220568},
			{-.0035177389049427,.00124629807811922,.00733218354822258,-.000640159716652238,-.0104475716905616,-.00111787975143421,.0140189382881418,.00476714217793534,-.0203246135001665,-.0211329311052817,-.00788540855182074,.00279371772291301,.0164359164841472,-.00143499021406278,-.0234194104169471,-.00250585355822768,.0314250314909489,.0106860869193045,-.0455599137506981,-.0473718488397425,-.0115595179541847,.00409541623172844,.0240940301879616,-.00210360630454689,-.0343313974681262,-.00367342529008241,.0460671394948289,.0156651380574512,-.0667879967830832,-.0694441808039846,-.0142071529086165,.00503344558650323,.0296126164102953,-.00258542411082186,-.0421947883405973,-.00451480027120738,.0566185283382572,.0192531394993187,-.0820853677911323,-.0853499340721777,-.0155925613838605,.0055242813098938,.0325002864461468,-.00283754137163211,-.0463094070649962,-.00495506037117232,.0621396759967787,.0211306066321764,-.0900899106409799,-.0936728206340164,-.0155925613838605,.0055242813098938,.0325002864461468,-.00283754137163211,-.0463094070649962,-.00495506037117232,.0621396759967787,.0211306066321764,-.0900899106409799,-.0936728206340164,-.0142071529086165,.00503344558650323,.0296126164102953,-.00258542411082186,-.0421947883405973,-.00451480027120738,.0566185283382572,.0192531394993187,-.0820853677911323,-.0853499340721777,-.0115595179541847,.00409541623172844,.0240940301879616,-.00210360630454689,-.0343313974681262,-.00367342529008241,.0460671394948289,.0156651380574512,-.0667879967830832,-.0694441808039846,-.00788540855182074,.00279371772291301,.0164359164841472,-.00143499021406278,-.0234194104169471,-.00250585355822768,.0314250314909489,.0106860869193045,-.0455599137506981,-.0473718488397425,-.0035177389049427,.00124629807811922,.00733218354822258,-.000640159716652238,-.0104475716905616,-.00111787975143421,.0140189382881418,.00476714217793534,-.0203246135001665,-.0211329311052817},
			{-.00537304737318019,-.0113801575780441,-.0150219679612532,-.0157580888060425,-.0138618270499793,-.0102693186804868,-.00622899008451743,-.002867614478685,-.000823350121801073,-7.1027404987251e-005,-.012044291759767,-.0255099068782191,-.0336734356436761,-.035323533550749,-.0310728489286636,-.0230198361881556,-.0139629838964585,-.00642808131717099,-.0018456321708443,-.000159215940078199,-.0176561817853201,-.0373959351161665,-.0493631599864637,-.0517821007753423,-.0455508617870686,-.0337456465280969,-.0204688649908742,-.00942316696826571,-.00270558184468062,-.000233400654618126,-.0217002192825492,-.0459612390812321,-.0606694816133146,-.0636424655907344,-.0559840004655092,-.0414748748282566,-.0251571299031463,-.0115814841529019,-.00332527836597537,-.000286859608010771,-.0238163130489831,-.050443142708611,-.066585657398596,-.0698485514815503,-.0614432722296599,-.0455192912899041,-.0276103238122155,-.0127108509165695,-.00364954240820819,-.000314832681482957,-.0238163130489831,-.050443142708611,-.066585657398596,-.0698485514815503,-.0614432722296599,-.0455192912899041,-.0276103238122155,-.0127108509165695,-.00364954240820819,-.000314832681482957,-.0217002192825492,-.0459612390812321,-.0606694816133146,-.0636424655907344,-.0559840004655092,-.0414748748282566,-.0251571299031463,-.0115814841529019,-.00332527836597537,-.000286859608010771,-.0176561817853201,-.0373959351161665,-.0493631599864637,-.0517821007753423,-.0455508617870686,-.0337456465280969,-.0204688649908742,-.00942316696826571,-.00270558184468062,-.000233400654618126,-.012044291759767,-.0255099068782191,-.0336734356436761,-.035323533550749,-.0310728489286636,-.0230198361881556,-.0139629838964585,-.00642808131717099,-.0018456321708443,-.000159215940078199,-.00537304737318019,-.0113801575780441,-.0150219679612533,-.0157580888060425,-.0138618270499793,-.0102693186804868,-.00622899008451743,-.002867614478685,-.000823350121801073,-7.1027404987251e-005},
			{.00906355498937132,.0191966824099048,.0253398906075171,.0265816201684498,.0233829004277474,.0173228576075033,.0105074067355295,.00483725150937865,.00138887275435145,.000119812975047951,.0180463682645229,.0382223532192926,.0504540434986315,.052926440804966,.0465574967776602,.0344913963831515,.0209212093573502,.00963141087892507,.00276537288370632,.000238558608969547,.0207772984278405,.0440064852834015,.0580891791243752,.060935720650788,.0536029737631114,.0397109282788219,.0240871849625023,.0110889180127161,.0031838526636951,.000274659329702082,.0162895482815631,.0345013943567046,.0455423254986296,.0477740341007375,.0420251089034541,.031133647415719,.0188845226330227,.00869378981034385,.00249616290908881,.000215334848642728,.00614122492031466,.0130071638050768,.0171696390499558,.0180110021280012,.0158436343117591,.0117375097249386,.0071195406402661,.00327759356566627,.000941063410568089,8.11821000704882e-005,-.00614122492031466,-.0130071638050768,-.0171696390499558,-.0180110021280012,-.0158436343117591,-.0117375097249386,-.0071195406402661,-.00327759356566627,-.000941063410568089,-8.11821000704882e-005,-.0162895482815631,-.0345013943567046,-.0455423254986296,-.0477740341007375,-.0420251089034541,-.031133647415719,-.0188845226330227,-.00869378981034385,-.00249616290908881,-.000215334848642728,-.0207772984278405,-.0440064852834015,-.0580891791243752,-.060935720650788,-.0536029737631114,-.0397109282788219,-.0240871849625023,-.0110889180127161,-.0031838526636951,-.000274659329702082,-.0180463682645229,-.0382223532192926,-.0504540434986315,-.052926440804966,-.0465574967776602,-.0344913963831515,-.0209212093573502,-.00963141087892507,-.00276537288370632,-.000238558608969547,-.00906355498937132,-.0191966824099048,-.0253398906075171,-.0265816201684498,-.0233829004277474,-.0173228576075033,-.0105074067355295,-.00483725150937865,-.00138887275435145,-.000119812975047951},
			{-.0110862696476229,-.0234808083346496,-.0309949970453799,-.0325138435419332,-.0286012651315648,-.0211888018254831,-.0128523459618144,-.0059167815111433,-.00169882765416366,-.00014655165111517,-.0167651321190546,-.035508684751954,-.0468719629787211,-.0491688277486249,-.0432520589832132,-.0320426142733104,-.0194358684154698,-.00894761059462107,-.00256903999044115,-.000221621688025442,-.00759587706914969,-.0160881287631775,-.0212365561003511,-.0222772101383125,-.019596464894583,-.0145177358141793,-.00880592327979835,-.00405394658728139,-.00116397006683541,-.000100411442399873,.0105903093675451,.0224303604700153,.0296083910990837,.0310592898045613,.0273217462387268,.0202408901814494,.0122773777077953,.00565208574704063,.00162282814612863,.000139995451397573,.0248569696711101,.052647261809357,.0694951255924348,.0729005922191637,.0641280432939063,.0475082621191814,.0288167601844263,.0132662530542551,.00380900959639235,.000328589332824135,.0248569696711101,.0526472618093571,.0694951255924348,.0729005922191637,.0641280432939063,.0475082621191814,.0288167601844263,.0132662530542551,.00380900959639235,.000328589332824135,.0105903093675451,.0224303604700152,.0296083910990836,.0310592898045612,.0273217462387267,.0202408901814493,.0122773777077953,.00565208574704061,.00162282814612863,.000139995451397573,-.00759587706914966,-.0160881287631774,-.021236556100351,-.0222772101383123,-.0195964648945829,-.0145177358141792,-.0088059232797983,-.00405394658728137,-.00116397006683541,-.000100411442399872,-.0167651321190546,-.035508684751954,-.0468719629787211,-.0491688277486249,-.0432520589832132,-.0320426142733104,-.0194358684154698,-.00894761059462107,-.00256903999044115,-.000221621688025442,-.0110862696476229,-.0234808083346496,-.0309949970453799,-.0325138435419332,-.0286012651315648,-.0211888018254831,-.0128523459618144,-.0059167815111433,-.00169882765416366,-.00014655165111517},
			{.0120620795615023,.0255475815854486,.0337231667867706,.0353757016668926,.0311187393543622,.0230538333953191,.0139836053488558,.00643757473017433,.00184835792174656,.00015945108064368,.0102225858120058,.0216515190199873,.0285803094377376,.0299808291021325,.0263730629357406,.0195380729315951,.01185107467672,.00545583037857288,.00156647926006563,.00013513443899835,-.0109815046191396,-.0232589151611855,-.0307020949375013,-.0322065883647599,-.0283309837428373,-.0209885680681119,-.0127308915471588,-.00586086804311059,-.00168277376649596,-.000145166740916215,-.0256396545211207,-.0543049946206344,-.0716833562042298,-.0751960526007667,-.0661472776821311,-.0490041804673553,-.0297241291003144,-.0136839747403398,-.00392894594197619,-.000338935802891862,-.0135515298013773,-.028702249180423,-.0378873723537952,-.039743965618796,-.0349613448984257,-.025900567866404,-.0157103295206683,-.00723249961664868,-.0020765969360793,-.000179140426009222,.0135515298013774,.0287022491804232,.0378873723537955,.0397439656187963,.034961344898426,.0259005678664042,.0157103295206684,.00723249961664874,.00207659693607932,.000179140426009223,.0256396545211207,.0543049946206346,.07168335620423,.0751960526007669,.0661472776821312,.0490041804673554,.0297241291003145,.0136839747403399,.0039289459419762,.000338935802891863,.0109815046191395,.0232589151611852,.030702094937501,.0322065883647595,.028330983742837,.0209885680681117,.0127308915471586,.00586086804311052,.00168277376649595,.000145166740916214,-.0102225858120059,-.0216515190199875,-.0285803094377378,-.0299808291021328,-.0263730629357408,-.0195380729315952,-.0118510746767201,-.00545583037857292,-.00156647926006564,-.000135134438998351,-.0120620795615023,-.0255475815854486,-.0337231667867706,-.0353757016668926,-.0311187393543623,-.0230538333953191,-.0139836053488558,-.00643757473017434,-.00184835792174656,-.00015945108064368},
			{.00615133467063503,.00923599360757013,.00365245862188397,-.00803852164203684,-.0191471282194023,-.0235470344199801,-.0197092364291758,-.0112335234696543,-.00369340218158754,-.000342287092850262,.013788910526811,.0207035214794833,.00818739799355415,-.0180192529954539,-.0429204476750506,-.0527833337273658,-.0441804766323359,-.025181209984126,-.00827917760101021,-.000767275128164547,.0202136842696012,.0303500733898785,.0120022156724767,-.0264151029275269,-.0629187038618097,-.0773770807046292,-.0647658278578351,-.0369140859356472,-.0121367588623754,-.00112477756371539,.0248435016410661,.037301566998444,.0147512477576403,-.0324653162766497,-.0773298376385226,-.095099814849551,-.0796000337796347,-.0453690253735618,-.0149166072197978,-.00138240079726687,.0272661121352065,.040939023946499,.016189713563936,-.0356311669302222,-.0848706456485974,-.104373459634974,-.0873622196403312,-.0497931793662997,-.0163711980302944,-.00151720541245167,.0272661121352065,.040939023946499,.016189713563936,-.0356311669302222,-.0848706456485974,-.104373459634974,-.0873622196403312,-.0497931793662997,-.0163711980302944,-.00151720541245167,.0248435016410661,.037301566998444,.0147512477576403,-.0324653162766497,-.0773298376385226,-.095099814849551,-.0796000337796348,-.0453690253735618,-.0149166072197978,-.00138240079726687,.0202136842696012,.0303500733898785,.0120022156724767,-.0264151029275269,-.0629187038618097,-.0773770807046292,-.0647658278578351,-.0369140859356472,-.0121367588623754,-.00112477756371539,.013788910526811,.0207035214794833,.00818739799355415,-.0180192529954539,-.0429204476750506,-.0527833337273658,-.0441804766323359,-.025181209984126,-.00827917760101021,-.000767275128164547,.00615133467063503,.00923599360757013,.00365245862188397,-.00803852164203684,-.0191471282194022,-.0235470344199801,-.0197092364291758,-.0112335234696543,-.00369340218158754,-.000342287092850262},
			{-.0103764132698787,-.015579787438262,-.00616117024045563,.0135598251603545,.0322984402530261,.0397204466191965,.0332466356283884,.0189493318506081,.0062302361129812,.000577388895667578,-.0206603893673846,-.0310207839996088,-.0122674543520802,.0269988540625906,.0643091532913213,.0790870478705028,.0661970971445412,.0377298555968102,.0124049708311625,.00114963418385859,-.0237868954700111,-.0357151132670011,-.0141238700378378,.031084550633434,.0740409621476184,.0910551736114528,.0762146057460633,.0434394587256295,.0142821966818892,.00132360662105264,-.0186490935562653,-.0280009003078451,-.0110732135702299,.0243705065946171,.0580486357216159,.0713878973236672,.0597527876096367,.0340568415423994,.0111973427741056,.00103771691176861,-.00703078293574547,-.0105564515227481,-.00417464584959855,.00918777855792046,.0218845680753062,.0269135231054363,.0225270401599717,.0128395656142358,.00422144305643347,.000391223430424914,.00703078293574547,.0105564515227481,.00417464584959855,-.00918777855792046,-.0218845680753062,-.0269135231054363,-.0225270401599717,-.0128395656142358,-.00422144305643347,-.000391223430424914,.0186490935562653,.0280009003078451,.0110732135702299,-.0243705065946171,-.0580486357216159,-.0713878973236672,-.0597527876096368,-.0340568415423994,-.0111973427741056,-.00103771691176861,.0237868954700111,.0357151132670011,.0141238700378378,-.031084550633434,-.0740409621476184,-.0910551736114528,-.0762146057460633,-.0434394587256295,-.0142821966818892,-.00132360662105264,.0206603893673846,.0310207839996088,.0122674543520802,-.0269988540625906,-.0643091532913213,-.0790870478705028,-.0661970971445413,-.0377298555968102,-.0124049708311625,-.00114963418385859,.0103764132698787,.015579787438262,.00616117024045563,-.0135598251603545,-.0322984402530261,-.0397204466191965,-.0332466356283884,-.0189493318506081,-.0062302361129812,-.000577388895667578},
			{.0126921186677361,.0190567304753116,.00753615934428767,-.0165859729728156,-.0395064870502198,-.0485848634700931,-.040666291301582,-.0231782565195614,-.00762063865589431,-.000706244844981167,.0191935658340169,.0288184045969256,.0113965031604947,-.0250820192048949,-.0597434029668103,-.0734721128884853,-.0614973086649725,-.0350511529297959,-.0115242563962926,-.00106801372426019,.00869614182330973,.0130569241621705,.00516348075342068,-.0113640580446065,-.027068295162004,-.0332884425573554,-.027862947538109,-.0158808321280148,-.00522136266376985,-.000483891263130864,-.0121243184130687,-.0182042001216115,-.00719901837461035,.0158439755235,.0377391073088729,.0464113494513956,.0388470260424168,.0221413437472276,.00727971838223315,.000674649962098675,-.0284575081442016,-.0427278594614471,-.0168971250214698,.0371880750022642,.0885790785929488,.108934070353419,.0911795222062535,.0519688982542005,.0170865394731464,.00158349988319459,-.0284575081442016,-.0427278594614471,-.0168971250214699,.0371880750022643,.0885790785929488,.108934070353419,.0911795222062535,.0519688982542005,.0170865394731464,.00158349988319459,-.0121243184130687,-.0182042001216115,-.00719901837461033,.0158439755235,.0377391073088728,.0464113494513955,.0388470260424167,.0221413437472276,.00727971838223313,.000674649962098674,.00869614182330968,.0130569241621705,.00516348075342066,-.0113640580446064,-.0270682951620039,-.0332884425573552,-.0278629475381089,-.0158808321280148,-.00522136266376982,-.000483891263130862,.0191935658340169,.0288184045969256,.0113965031604947,-.0250820192048949,-.0597434029668103,-.0734721128884853,-.0614973086649725,-.0350511529297959,-.0115242563962926,-.00106801372426019,.0126921186677361,.0190567304753116,.00753615934428767,-.0165859729728156,-.0395064870502198,-.0485848634700931,-.040666291301582,-.0231782565195614,-.00762063865589432,-.000706244844981167},
			{-.00643614945112683,-.00460242005009466,.00815688456903964,.0159153508876072,.00595117342612003,-.0150500971453785,-.0280693431449012,-.0232405097467362,-.00939844620771058,-.000965522727863794,-.0144273549840217,-.010316843689307,.0182845768475155,.0356760542455353,.0133402265193015,-.0337364903828284,-.0629205988448425,-.0520962240967035,-.021067677307141,-.00216432810421968,-.0211496040913043,-.0151238504731025,.0268040511743329,.0522988741642954,.0195559414517781,-.049455594307975,-.0922376801728386,-.0763698207687696,-.0308839031591627,-.00317277023949471,-.0259937880171798,-.0185878733948966,.0329433506754726,.0642776026773378,.024035106963676,-.0607830874353671,-.113364141250993,-.093861848354391,-.0379576671221214,-.00389947332709414,-.028528568522066,-.0204004671991369,.036155816784784,.0705456238699261,.0263788869669642,-.0667103414760682,-.124418829202062,-.103014773014854,-.0416591035794981,-.00427972990848616,-.028528568522066,-.0204004671991369,.036155816784784,.0705456238699261,.0263788869669642,-.0667103414760682,-.124418829202062,-.103014773014854,-.0416591035794981,-.00427972990848616,-.0259937880171798,-.0185878733948966,.0329433506754726,.0642776026773378,.024035106963676,-.0607830874353671,-.113364141250993,-.093861848354391,-.0379576671221214,-.00389947332709414,-.0211496040913043,-.0151238504731025,.0268040511743329,.0522988741642954,.019555941451778,-.049455594307975,-.0922376801728386,-.0763698207687696,-.0308839031591627,-.00317277023949471,-.0144273549840217,-.010316843689307,.0182845768475155,.0356760542455353,.0133402265193015,-.0337364903828284,-.0629205988448425,-.0520962240967035,-.021067677307141,-.00216432810421968,-.00643614945112683,-.00460242005009466,.00815688456903964,.0159153508876072,.00595117342612003,-.0150500971453785,-.0280693431449012,-.0232405097467362,-.00939844620771058,-.000965522727863794},
			{.0108568546750866,.00776361798572502,-.0137594863264114,-.026846898600142,-.0100387701566362,.0253873404889165,.0473489283714467,.0392033837640413,.0158538215160414,.0016286974100757,.0216169922167163,.0154580746075647,-.0273963977344734,-.0534546344636977,-.0199881109986099,.050548520559216,.0942760538578598,.0780575283595627,.0315664109517744,.00324288574275418,.0248882596107707,.0177973221278318,-.0315422539999915,-.0615438450730705,-.0230128822120598,.0581979532585155,.108542709363562,.0898698584387518,.0363433091371754,.00373362683599065,.0195125707984483,.0139532258773057,-.0247293492571541,-.0482508079302309,-.0180422617114151,.0456276051857851,.0850982485008847,.0704586018812573,.0284934102938292,.00292718973168883,.00735631731313645,.00526042202001852,-.00932306367838869,-.0181907477705709,-.00680200490067545,.0172017898334461,.0320823803911286,.0265631750030447,.0107421297593173,.00110356224838248,-.00735631731313645,-.00526042202001852,.00932306367838869,.0181907477705709,.00680200490067545,-.0172017898334461,-.0320823803911286,-.0265631750030447,-.0107421297593173,-.00110356224838248,-.0195125707984483,-.0139532258773057,.0247293492571541,.0482508079302309,.0180422617114151,-.0456276051857851,-.0850982485008847,-.0704586018812573,-.0284934102938292,-.00292718973168883,-.0248882596107707,-.0177973221278318,.0315422539999915,.0615438450730705,.0230128822120598,-.0581979532585155,-.108542709363562,-.0898698584387518,-.0363433091371754,-.00373362683599065,-.0216169922167163,-.0154580746075647,.0273963977344734,.0534546344636977,.0199881109986099,-.050548520559216,-.0942760538578598,-.0780575283595627,-.0315664109517744,-.00324288574275418,-.0108568546750866,-.00776361798572502,.0137594863264114,.026846898600142,.0100387701566362,-.0253873404889165,-.0473489283714467,-.0392033837640413,-.0158538215160414,-.0016286974100757},
			{.00628780165042657,-.00108441059929728,-.0131544407550812,-.00292698561809638,.0175762393921458,.0118959849140124,-.0179938136163745,-.0323721609932466,-.0177761887834042,-.00209542799580947,.0140948166553123,-.00243082867843575,-.0294871627563918,-.00656116842314922,.0393991231745305,.0266661920364473,-.0403351628999706,-.0725658504041882,-.0398473322890685,-.00469713821417048,.0206621236068235,-.00356344348771085,-.0432263445908294,-.00961826437899792,.0577566613983443,.0390909771623532,-.0591288373535847,-.106377018399916,-.0584137080526403,-.00688571215596096,.0253946531907593,-.00437962782806131,-.0531270672113226,-.0118212673996889,.070985461783933,.0480445198574855,-.0726719260198027,-.130742006055988,-.0717930008459142,-.00846284125966016,.0278710091491036,-.00480670660664831,-.0583077455395459,-.0129740165922258,.0779075988938198,.0527295743105294,-.0797585184474156,-.143491293997453,-.0787938850114395,-.00928809400953851,.0278710091491036,-.00480670660664831,-.0583077455395459,-.0129740165922258,.0779075988938198,.0527295743105294,-.0797585184474156,-.143491293997453,-.0787938850114395,-.00928809400953851,.0253946531907593,-.00437962782806131,-.0531270672113226,-.0118212673996889,.070985461783933,.0480445198574855,-.0726719260198027,-.130742006055988,-.0717930008459142,-.00846284125966016,.0206621236068235,-.00356344348771085,-.0432263445908294,-.00961826437899792,.0577566613983443,.0390909771623532,-.0591288373535847,-.106377018399916,-.0584137080526403,-.00688571215596096,.0140948166553123,-.00243082867843575,-.0294871627563918,-.00656116842314922,.0393991231745305,.0266661920364473,-.0403351628999706,-.0725658504041882,-.0398473322890685,-.00469713821417048,.00628780165042657,-.00108441059929729,-.0131544407550812,-.00292698561809638,.0175762393921458,.0118959849140124,-.0179938136163745,-.0323721609932466,-.0177761887834042,-.00209542799580947},
			{-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.000726793993235288,-8.78323358755477e-005,-1.46520306329324e-006,-.0187952392881457,-.0376133897003122,-.0447078783667129,-.0400285772779044,-.0282223784536199,-.0154894473150464,-.00625458279978263,-.00162919071725786,-.000196886088238751,-3.28441793937936e-006,-.0275526505160409,-.0551388873133415,-.0655389659619086,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461161,-.00238829215039021,-.00028862274630013,-4.81475219575472e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030351,-.00293531546086303,-.00035472997281053,-5.91754093322042e-006,-.0371655977491602,-.0743764998227977,-.088405100787603,-.0791522778068628,-.0558067683550777,-.0306287437708179,-.0123677759491016,-.00322155232641602,-.000389321415157626,-6.49458908735937e-006,-.0371655977491602,-.0743764998227977,-.088405100787603,-.0791522778068628,-.0558067683550778,-.0306287437708179,-.0123677759491016,-.00322155232641602,-.000389321415157626,-6.49458908735937e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030351,-.00293531546086303,-.00035472997281053,-5.91754093322042e-006,-.0275526505160409,-.0551388873133415,-.0655389659619087,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461161,-.00238829215039021,-.00028862274630013,-4.81475219575472e-006,-.0187952392881457,-.0376133897003122,-.044707878366713,-.0400285772779045,-.0282223784536199,-.0154894473150464,-.00625458279978263,-.00162919071725786,-.000196886088238751,-3.28441793937936e-006,-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.000726793993235288,-8.78323358755477e-005,-1.46520306329324e-006},
			{.0141437693703256,.0283047797903233,.0336435153051858,.0301222536495194,.0212378680478842,.0116560990333334,.00470669062902145,.00122599650964295,.000148160466678965,2.47158597654252e-006,.028161540477737,.0563574094647636,.0669873209378089,.0599761664108691,.042286540810412,.023208360949976,.00937145220590666,.00244107135996656,.000295001061621808,4.92115409267359e-006,.0324231846605908,.0648858927129044,.0771244129347917,.0690522707843786,.0486857003592358,.0267204478158996,.0107896201789892,.00281047507100845,.000339643135062007,5.66586504797623e-006,.0254200050985947,.0508709968146246,.0604660828524796,.054137466562375,.0381699319272815,.020949019253574,.00845913820104087,.00220343224709027,.000266282609662119,4.44207809674427e-006,.00958344369577791,.0191785694703894,.0227959553223306,.0204100416431963,.0143902171568408,.00789786413180114,.00318912896948871,.000830702778994263,.000100389609953414,1.67468122713911e-006,-.00958344369577791,-.0191785694703894,-.0227959553223306,-.0204100416431963,-.0143902171568408,-.00789786413180114,-.00318912896948871,-.000830702778994263,-.000100389609953414,-1.67468122713911e-006,-.0254200050985947,-.0508709968146246,-.0604660828524796,-.054137466562375,-.0381699319272815,-.020949019253574,-.00845913820104087,-.00220343224709027,-.000266282609662119,-4.44207809674427e-006,-.0324231846605908,-.0648858927129044,-.0771244129347917,-.0690522707843785,-.0486857003592358,-.0267204478158996,-.0107896201789892,-.00281047507100845,-.000339643135062007,-5.66586504797623e-006,-.028161540477737,-.0563574094647636,-.066987320937809,-.0599761664108691,-.042286540810412,-.023208360949976,-.00937145220590666,-.00244107135996656,-.000295001061621808,-4.92115409267359e-006,-.0141437693703256,-.0283047797903233,-.0336435153051858,-.0301222536495194,-.0212378680478842,-.0116560990333334,-.00470669062902145,-.00122599650964295,-.000148160466678965,-2.47158597654252e-006},
			{-.0173002360880135,-.0346215609128356,-.0411517426769954,-.0368446406325781,-.0259775256237805,-.0142573920616249,-.00575708334485075,-.00149960229869156,-.000181225455912347,-3.02317011727578e-006,-.0261621585010403,-.0523562082937465,-.0622314290413362,-.0557180447272185,-.0392843276459713,-.0215606393479301,-.00870610818288429,-.00226776286909028,-.000274056901760265,-4.57176742452532e-006,-.0118534431119483,-.0237213353990866,-.0281955597772025,-.0252445024158391,-.0177987815081903,-.00976860575008604,-.00394452766839492,-.00102746866850002,-.000124168573259081,-2.07135757110846e-006,.0165262850469198,.0330727153955572,.0393107600496111,.0351963424341469,.0248153835062233,.0136195670415941,.0054995319088043,.00143251542459375,.000173118073471871,2.88792423694412e-006,.0387895529705973,.0776263898435274,.0922679721988114,.0826108460155768,.0582452517469885,.031967070378892,.0129081873926539,.00336231843912396,.000406332860775417,6.7783709312756e-006,.0387895529705973,.0776263898435274,.0922679721988115,.0826108460155769,.0582452517469885,.031967070378892,.0129081873926539,.00336231843912396,.000406332860775417,6.77837093127561e-006,.0165262850469198,.0330727153955571,.039310760049611,.0351963424341468,.0248153835062232,.013619567041594,.00549953190880429,.00143251542459375,.000173118073471871,2.88792423694411e-006,-.0118534431119483,-.0237213353990865,-.0281955597772024,-.025244502415839,-.0177987815081902,-.00976860575008599,-.0039445276683949,-.00102746866850002,-.00012416857325908,-2.07135757110845e-006,-.0261621585010403,-.0523562082937465,-.0622314290413362,-.0557180447272185,-.0392843276459713,-.0215606393479301,-.00870610818288428,-.00226776286909028,-.000274056901760265,-4.57176742452532e-006,-.0173002360880135,-.0346215609128356,-.0411517426769954,-.0368446406325781,-.0259775256237805,-.0142573920616249,-.00575708334485075,-.00149960229869156,-.000181225455912347,-3.02317011727578e-006},
			{.00879759938930934,.0102248339942577,-.00281118458446222,-.0202713999635782,-.0287696758053183,-.0241048298357547,-.0129418509611811,-.00409370526960594,-.000560621794885306,-9.99671086012965e-006,.0197208113889518,.0229201187461582,-.00630158734405165,-.0454406296060175,-.0644904734997843,-.0540336950704248,-.0290106187535114,-.00917650212647326,-.00125669699064766,-2.24087550090162e-005,.0289094815853316,.0335994669670076,-.00923773467978503,-.0666131336541757,-.0945390186691718,-.0792100325801501,-.0425277607544874,-.0134521807450406,-.00184223954039861,-3.28498395682792e-005,.0355310166928675,.0412952137572032,-.0113535797638883,-.081870453361282,-.116192586869701,-.0973525928350061,-.052268477136742,-.0165333182193771,-.00226419293161506,-4.03738889129983e-005,.0389958186821867,.0453221105953125,-.0124607224637962,-.0898540388613718,-.127523090289016,-.106845916942104,-.0573654301771659,-.0181455623707819,-.0024849853801249,-4.43109428912591e-005,.0389958186821867,.0453221105953125,-.0124607224637962,-.0898540388613718,-.127523090289016,-.106845916942104,-.0573654301771659,-.0181455623707819,-.0024849853801249,-4.43109428912591e-005,.0355310166928675,.0412952137572032,-.0113535797638883,-.081870453361282,-.116192586869701,-.0973525928350061,-.052268477136742,-.0165333182193771,-.00226419293161506,-4.03738889129983e-005,.0289094815853316,.0335994669670076,-.00923773467978504,-.0666131336541757,-.0945390186691718,-.0792100325801501,-.0425277607544874,-.0134521807450406,-.00184223954039861,-3.28498395682792e-005,.0197208113889518,.0229201187461582,-.00630158734405165,-.0454406296060175,-.0644904734997843,-.0540336950704248,-.0290106187535114,-.00917650212647326,-.00125669699064766,-2.24087550090162e-005,.00879759938930934,.0102248339942577,-.00281118458446222,-.0202713999635783,-.0287696758053183,-.0241048298357547,-.0129418509611811,-.00409370526960595,-.000560621794885306,-9.99671086012965e-006},
			{-.0148402797021791,-.0172478183727625,.00474206242882225,.0341949243312682,.0485302884342244,.0406613669372952,.0218310336296094,.00690548961493529,.000945688008179328,1.68630075889125e-005,-.0295483563533413,-.0343419863926998,.00944188069959736,.0680852268213982,.0966282499631283,.0809605064182685,.0434675878214585,.0137494624111731,.0018829514554676,3.35757928708797e-005,-.0340198653272789,-.0395389082959526,.0108707065122495,.0783884632885607,.111250859819299,.0932121398646372,.0500454734640585,.0158301481800593,.00216789570853973,3.86567678440353e-005,-.0266718139851343,-.0309987825377814,.00852271045732846,.0614571072313015,.0872214575290303,.0730789740555617,.0392360036170109,.0124109476493752,.00169964550185853,3.03071782114281e-005,-.0100553806578476,-.0116866651260001,.00321309596463221,.0231695754808925,.0328828386953925,.0275510658038934,.0147921304520804,.00467897695329769,.000640773159040156,1.14259275260014e-005,.0100553806578476,.0116866651260001,-.00321309596463221,-.0231695754808925,-.0328828386953925,-.0275510658038934,-.0147921304520804,-.00467897695329769,-.000640773159040156,-1.14259275260014e-005,.0266718139851343,.0309987825377814,-.00852271045732846,-.0614571072313014,-.0872214575290303,-.0730789740555617,-.0392360036170109,-.0124109476493752,-.00169964550185853,-3.03071782114281e-005,.0340198653272789,.0395389082959526,-.0108707065122495,-.0783884632885607,-.111250859819299,-.0932121398646372,-.0500454734640585,-.0158301481800593,-.00216789570853973,-3.86567678440353e-005,.0295483563533413,.0343419863926998,-.00944188069959736,-.0680852268213982,-.0966282499631283,-.0809605064182685,-.0434675878214585,-.0137494624111731,-.0018829514554676,-3.35757928708797e-005,.0148402797021791,.0172478183727625,-.00474206242882225,-.0341949243312682,-.0485302884342244,-.0406613669372952,-.0218310336296094,-.00690548961493529,-.000945688008179328,-1.68630075889125e-005},
			{-.00863131572555909,-.00182782839348211,.0164716603838428,.014834131431226,-.0115521085131657,-.0329015220239097,-.0289052514771828,-.0121493830289503,-.00197137671826113,-3.83525277872114e-005,-.0193480678000737,-.00409728352066521,.036923084732338,.0332523788739027,-.0258953542951368,-.0737524728656581,-.0647943816610287,-.0272342124940597,-.00441906328272248,-8.59715171504879e-005,-.0283630626928129,-.00600636252506896,.0541269432119395,.0487459169789592,-.0379609770295909,-.108116533043712,-.0949845291107984,-.0399236597855762,-.00647807162071696,-.000126028891149856,-.0348594439863812,-.00738208212111014,.0665243794538717,.0599108629733206,-.046655700294547,-.132879945598894,-.116740138678666,-.0490679232034869,-.0079618332211835,-.000154895016778896,-.0382587576596884,-.00810194479882727,.0730114947611087,.0657530621708869,-.0512053242073936,-.145837714407803,-.128124036533984,-.05385277468682,-.00873823024415192,-.000169999582091795,-.0382587576596884,-.00810194479882727,.0730114947611087,.0657530621708869,-.0512053242073936,-.145837714407803,-.128124036533984,-.05385277468682,-.00873823024415192,-.000169999582091795,-.0348594439863812,-.00738208212111014,.0665243794538717,.0599108629733206,-.046655700294547,-.132879945598894,-.116740138678666,-.0490679232034869,-.0079618332211835,-.000154895016778896,-.0283630626928129,-.00600636252506896,.0541269432119395,.0487459169789592,-.0379609770295909,-.108116533043712,-.0949845291107984,-.0399236597855762,-.00647807162071696,-.000126028891149856,-.0193480678000737,-.0040972835206652,.036923084732338,.0332523788739027,-.0258953542951368,-.0737524728656581,-.0647943816610287,-.0272342124940597,-.00441906328272248,-8.59715171504879e-005,-.00863131572555909,-.00182782839348211,.0164716603838428,.014834131431226,-.0115521085131657,-.0329015220239097,-.0289052514771828,-.0121493830289503,-.00197137671826113,-3.83525277872114e-005},
			{-.0113062148903861,-.0213785801497483,-.0228814465519833,-.0174855360916378,-.00988118279406322,-.00401765276570605,-.0010799937436432,-.000159171487958625,-8.09632053406128e-006,-2.61175882124433e-008,-.0253441560031943,-.0479224988818084,-.0512913434438016,-.039195801495227,-.0221497858175121,-.00900602186035772,-.00242092779827153,-.000356800844596912,-1.81488157315668e-005,-5.85455200083099e-008,-.0371529546538126,-.0702513994795805,-.0751899158472868,-.0574586044762535,-.0324702068581793,-.0132022672898219,-.00354892941386567,-.000523047822073078,-2.66050338314805e-005,-8.58240870115567e-008,-.0456626054707174,-.0863420411133481,-.0924116936244469,-.0706191367966456,-.0399073037160309,-.0162261636575445,-.0043617893967995,-.000642848639204797,-3.26987496607437e-005,-1.05481554875234e-007,-.0501153878844611,-.0947616728508056,-.101423206654712,-.077505551778014,-.0437988587058273,-.0178084556759703,-.00478712866332422,-.0007055359319215,-3.5887363537237e-005,-1.15767573548085e-007,-.0501153878844611,-.0947616728508056,-.101423206654712,-.077505551778014,-.0437988587058273,-.0178084556759703,-.00478712866332422,-.0007055359319215,-3.5887363537237e-005,-1.15767573548085e-007,-.0456626054707174,-.0863420411133481,-.0924116936244469,-.0706191367966455,-.0399073037160309,-.0162261636575445,-.0043617893967995,-.000642848639204797,-3.26987496607437e-005,-1.05481554875234e-007,-.0371529546538126,-.0702513994795805,-.0751899158472868,-.0574586044762535,-.0324702068581793,-.0132022672898219,-.00354892941386567,-.000523047822073078,-2.66050338314805e-005,-8.58240870115567e-008,-.0253441560031943,-.0479224988818084,-.0512913434438016,-.039195801495227,-.0221497858175121,-.00900602186035772,-.00242092779827153,-.000356800844596912,-1.81488157315668e-005,-5.85455200083099e-008,-.0113062148903861,-.0213785801497483,-.0228814465519833,-.0174855360916378,-.00988118279406322,-.00401765276570605,-.0010799937436432,-.000159171487958625,-8.09632053406128e-006,-2.61175882124433e-008},
			{.0190719517783585,.0360625774106304,.038597696000771,.0294955742829575,.0166681284221623,.00677720002251945,.0018217934825568,.000268499314073518,1.36573235432709e-005,4.40565996475352e-008,.0379740031057454,.0718038951915978,.0768515485379464,.0587283903841316,.0331877705976788,.0134940260805107,.00362735771192228,.000534606730606772,2.71929822744642e-005,8.7720725769792e-008,.0437205527152459,.0826698722318604,.0884813794780883,.0676156706608181,.0382100267353107,.0155360570480548,.00417628037843494,.000615507974819813,3.1308055979897e-005,1.00995373191444e-007,.0342772212097279,.0648137619926128,.0693700245984015,.0530111619489379,.0299569300362286,.0121803781308882,.00327423322613994,.00048256258667776,2.45457363601731e-005,7.91810837934576e-008,.0129226496311487,.0244350477648839,.0261527770090912,.0199854202945654,.0112938825616689,.00459205132170114,.00123439903524221,.000181928027204484,9.25384088109926e-006,2.98515856059849e-008,-.0129226496311487,-.0244350477648839,-.0261527770090912,-.0199854202945654,-.0112938825616689,-.00459205132170114,-.00123439903524221,-.000181928027204484,-9.25384088109926e-006,-2.98515856059849e-008,-.0342772212097279,-.0648137619926128,-.0693700245984015,-.0530111619489379,-.0299569300362286,-.0121803781308882,-.00327423322613994,-.00048256258667776,-2.45457363601731e-005,-7.91810837934576e-008,-.0437205527152459,-.0826698722318604,-.0884813794780884,-.0676156706608181,-.0382100267353107,-.0155360570480548,-.00417628037843494,-.000615507974819813,-3.1308055979897e-005,-1.00995373191444e-007,-.0379740031057454,-.0718038951915978,-.0768515485379464,-.0587283903841316,-.0331877705976788,-.0134940260805107,-.00362735771192228,-.000534606730606772,-2.71929822744642e-005,-8.7720725769792e-008,-.0190719517783585,-.0360625774106304,-.038597696000771,-.0294955742829575,-.0166681284221623,-.00677720002251945,-.0018217934825568,-.000268499314073518,-1.36573235432709e-005,-4.40565996475352e-008},
			{.0111564498581787,.00938834250199416,-.0113241539111341,-.0302961470354155,-.0312651432694344,-.018730813581855,-.00658104714974118,-.00116693908893853,-6.69192068520344e-005,-2.30174087176476e-007,.0250084407901998,.0210450287110927,-.0253843683416991,-.0679122309460246,-.0700843453060978,-.0419872314552365,-.0147521595197384,-.00261582559707332,-.000150006950558683,-5.15961179744944e-007,.0366608170547776,.0308507017275645,-.0372119034382505,-.0995551020305782,-.102739286436345,-.0615506669900124,-.0216257473168005,-.00383463745165151,-.000219900849377703,-7.56366963325805e-007,.0450577468469165,.0379168611112332,-.0457350561038325,-.122357572607023,-.126271074443454,-.075648460517151,-.0265789888566817,-.00471293706542479,-.000270267757204588,-9.29608063726004e-007,.049451546558839,.0416143183763193,-.050194903530658,-.134289253725007,-.138584380130904,-.0830253092787393,-.0291708351373767,-.00517251844641853,-.000296622878728574,-1.0202586605363e-006,.049451546558839,.0416143183763193,-.050194903530658,-.134289253725007,-.138584380130904,-.0830253092787393,-.0291708351373767,-.00517251844641853,-.000296622878728574,-1.0202586605363e-006,.0450577468469165,.0379168611112332,-.0457350561038325,-.122357572607023,-.126271074443454,-.075648460517151,-.0265789888566817,-.00471293706542479,-.000270267757204588,-9.29608063726004e-007,.0366608170547776,.0308507017275645,-.0372119034382505,-.0995551020305782,-.102739286436345,-.0615506669900124,-.0216257473168005,-.00383463745165151,-.000219900849377703,-7.56366963325805e-007,.0250084407901998,.0210450287110927,-.0253843683416991,-.0679122309460246,-.0700843453060978,-.0419872314552365,-.0147521595197384,-.00261582559707332,-.000150006950558683,-5.15961179744944e-007,.0111564498581787,.00938834250199416,-.0113241539111341,-.0302961470354155,-.0312651432694344,-.018730813581855,-.00658104714974118,-.00116693908893853,-6.69192068520345e-005,-2.30174087176476e-007},
			{-.0141462413882811,-.025273750604131,-.0243577599189909,-.0158870112137333,-.0071957937725239,-.002167521076626,-.000387881070632107,-3.23454410928798e-005,-6.92492077729253e-007,-4.31978541371621e-010,-.031710395749536,-.0566539628254887,-.0546006663821845,-.0356125276698685,-.0161301834173286,-.00485874298682688,-.000869479172304184,-7.2505954733546e-005,-1.55229910443945e-006,-9.68328627104381e-010,-.0464854657297909,-.0830511819588261,-.0800411772208778,-.0522057481598326,-.0236458445483905,-.00712261469039873,-.00127460232871753,-.000106289215076142,-2.27557383363344e-006,-1.41950947462063e-009,-.0571326695688298,-.102073533343423,-.0983741059332855,-.0641631467466999,-.0290617766661061,-.00875400483105958,-.00156654198328513,-.000130634091932606,-2.79677972190642e-006,-1.74463920044904e-009,-.0627039536356021,-.112027219286109,-.107967042743438,-.070420006785652,-.0318957316434227,-.00960765035478176,-.00171930309977544,-.00014337285664383,-3.06950729477359e-006,-1.91476744149015e-009,-.0627039536356021,-.112027219286109,-.107967042743438,-.070420006785652,-.0318957316434227,-.00960765035478176,-.00171930309977544,-.00014337285664383,-3.06950729477359e-006,-1.91476744149015e-009,-.0571326695688298,-.102073533343423,-.0983741059332855,-.0641631467466999,-.0290617766661061,-.00875400483105958,-.00156654198328513,-.000130634091932606,-2.79677972190642e-006,-1.74463920044904e-009,-.0464854657297909,-.0830511819588261,-.0800411772208778,-.0522057481598326,-.0236458445483905,-.00712261469039873,-.00127460232871753,-.000106289215076142,-2.27557383363344e-006,-1.41950947462063e-009,-.031710395749536,-.0566539628254887,-.0546006663821845,-.0356125276698685,-.0161301834173286,-.00485874298682688,-.000869479172304184,-7.2505954733546e-005,-1.55229910443945e-006,-9.68328627104381e-010,-.0141462413882811,-.025273750604131,-.0243577599189909,-.0158870112137333,-.0071957937725239,-.002167521076626,-.000387881070632107,-3.23454410928798e-005,-6.92492077729254e-007,-4.31978541371621e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_hyp_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_hyp_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00222253422114756,-.00498206115568338,-.00730339144757066,-.00897618736857064,-.00985149899052509,-.00985149899052509,-.00897618736857064,-.00730339144757066,-.00498206115568338,-.00222253422114756,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256888,-.0397874192256888,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256889,-.0397874192256889,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00222253422114756,-.00498206115568339,-.00730339144757066,-.00897618736857064,-.0098514989905251,-.0098514989905251,-.00897618736857064,-.00730339144757066,-.00498206115568339,-.00222253422114756},
			{.00374909427250663,.00840401770480107,.0123197666813982,.0151415318298581,.0166180561313977,.0166180561313977,.0151415318298581,.0123197666813982,.00840401770480107,.00374909427250663,.00746479013903255,.0167331691152994,.0245297840368778,.0301481769402938,.0330880720843031,.0330880720843031,.0301481769402938,.0245297840368778,.0167331691152994,.00746479013903255,.00859442576735982,.0192653747976306,.0282418399000882,.0347104558746731,.0380952409936669,.0380952409936669,.0347104558746731,.0282418399000882,.0192653747976306,.00859442576735982,.00673809032372206,.0151041895084832,.0221418013612503,.0272132418374432,.0298669372064565,.0298669372064565,.0272132418374432,.0221418013612503,.0151041895084832,.00673809032372206,.00254028702921173,.00569433991706437,.00834754776191493,.0102595011258728,.0112599549045422,.0112599549045422,.0102595011258728,.00834754776191493,.00569433991706437,.00254028702921173,-.00254028702921173,-.00569433991706437,-.00834754776191493,-.0102595011258728,-.0112599549045422,-.0112599549045422,-.0102595011258728,-.00834754776191493,-.00569433991706437,-.00254028702921173,-.00673809032372206,-.0151041895084832,-.0221418013612503,-.0272132418374432,-.0298669372064565,-.0298669372064565,-.0272132418374432,-.0221418013612503,-.0151041895084832,-.00673809032372206,-.00859442576735982,-.0192653747976306,-.0282418399000882,-.0347104558746731,-.0380952409936669,-.0380952409936669,-.0347104558746731,-.0282418399000882,-.0192653747976306,-.00859442576735982,-.00746479013903255,-.0167331691152994,-.0245297840368778,-.0301481769402938,-.0330880720843031,-.0330880720843031,-.0301481769402938,-.0245297840368778,-.0167331691152994,-.00746479013903255,-.00374909427250663,-.00840401770480107,-.0123197666813982,-.0151415318298581,-.0166180561313977,-.0166180561313977,-.0151415318298581,-.0123197666813982,-.00840401770480107,-.00374909427250663},
			{-.00458578009255966,-.0102795433475266,-.015069170494548,-.0185206693108299,-.0203267097184652,-.0203267097184652,-.0185206693108299,-.015069170494548,-.0102795433475266,-.00458578009255966,-.00693481320266981,-.0155451660317305,-.0227882454870494,-.0280077499283953,-.0307389216397789,-.0307389216397789,-.0280077499283953,-.0227882454870494,-.0155451660317305,-.00693481320266981,-.00314199662793749,-.00704313985467146,-.0103248044877915,-.0126896360809338,-.0139270641206943,-.0139270641206943,-.0126896360809338,-.0103248044877915,-.00704313985467146,-.0031419966279375,.00438062859874147,.00981967313331039,.0143950293941988,.017692120427338,.019417365009608,.019417365009608,.017692120427338,.0143950293941988,.00981967313331039,.00438062859874147,.0102819614083243,.0230481762886867,.0337871913508869,.0415259352316643,.0455753308412161,.0455753308412161,.0415259352316643,.0337871913508869,.0230481762886867,.0102819614083243,.0102819614083243,.0230481762886867,.0337871913508869,.0415259352316643,.0455753308412161,.0455753308412161,.0415259352316643,.0337871913508869,.0230481762886867,.0102819614083243,.00438062859874146,.00981967313331036,.0143950293941988,.017692120427338,.0194173650096079,.0194173650096079,.017692120427338,.0143950293941988,.00981967313331036,.00438062859874146,-.00314199662793748,-.00704313985467143,-.0103248044877915,-.0126896360809337,-.0139270641206943,-.0139270641206943,-.0126896360809337,-.0103248044877915,-.00704313985467143,-.00314199662793748,-.00693481320266981,-.0155451660317305,-.0227882454870495,-.0280077499283953,-.0307389216397789,-.0307389216397789,-.0280077499283953,-.0227882454870495,-.0155451660317305,-.00693481320266981,-.00458578009255966,-.0102795433475266,-.015069170494548,-.0185206693108299,-.0203267097184652,-.0203267097184652,-.0185206693108299,-.015069170494548,-.0102795433475266,-.00458578009255966},
			{.00498941899157684,.0111843454696186,.0163955540682807,.0201508527079391,.0221158602153926,.0221158602153926,.0201508527079391,.0163955540682807,.0111843454696186,.00498941899157684,.00422852158563388,.00947870810595568,.01389519589419,.0170778032048017,.0187431427313533,.0187431427313533,.0170778032048017,.01389519589419,.00947870810595568,.00422852158563388,-.00454244456135888,-.0101824018661483,-.0149267671313349,-.0183456493520488,-.0201346227135425,-.0201346227135425,-.0183456493520488,-.0149267671313349,-.0101824018661483,-.00454244456135888,-.0106057150885859,-.0237739066819889,-.0348510669200656,-.0428334848152169,-.0470103859345993,-.0470103859345993,-.0428334848152169,-.0348510669200656,-.0237739066819889,-.0106057150885859,-.00560552264733118,-.0125654113096862,-.0184201105981369,-.022639121189878,-.0248467718409463,-.0248467718409463,-.022639121189878,-.0184201105981369,-.0125654113096862,-.00560552264733118,.00560552264733123,.0125654113096863,.0184201105981371,.0226391211898782,.0248467718409465,.0248467718409465,.0226391211898782,.0184201105981371,.0125654113096863,.00560552264733123,.0106057150885859,.023773906681989,.0348510669200657,.042833484815217,.0470103859345994,.0470103859345994,.042833484815217,.0348510669200657,.023773906681989,.0106057150885859,.00454244456135883,.0101824018661482,.0149267671313347,.0183456493520486,.0201346227135423,.0201346227135423,.0183456493520486,.0149267671313347,.0101824018661482,.00454244456135883,-.00422852158563391,-.00947870810595576,-.0138951958941901,-.0170778032048018,-.0187431427313534,-.0187431427313534,-.0170778032048018,-.0138951958941901,-.00947870810595576,-.00422852158563391,-.00498941899157684,-.0111843454696186,-.0163955540682807,-.0201508527079391,-.0221158602153927,-.0221158602153927,-.0201508527079391,-.0163955540682807,-.0111843454696186,-.00498941899157684},
			{-.00502787889479981,-.0112705576809201,-.0165219358020692,-.020306181383746,-.0222863357447103,-.0222863357447103,-.020306181383746,-.0165219358020692,-.0112705576809201,-.00502787889479981,-.000280476734556914,-.000628720237921658,-.000921664721701268,-.00113276623502688,-.00124322777173102,-.00124322777173102,-.00113276623502688,-.000921664721701268,-.000628720237921658,-.000280476734556914,.009285522300677,.0208145456318548,.0305128278843669,.0375015993872448,.0411585624649621,.0411585624649621,.0375015993872448,.030512827884367,.0208145456318548,.009285522300677,.00471291107004208,.010564521773763,.0154869311233076,.0190341153866534,.020890224414624,.020890224414624,.0190341153866534,.0154869311233076,.010564521773763,.00471291107004208,-.00869007782643158,-.0194797896774683,-.028556158763447,-.0350967674986958,-.0385192237402185,-.0385192237402185,-.0350967674986958,-.028556158763447,-.0194797896774683,-.00869007782643158,-.00869007782643163,-.0194797896774684,-.0285561587634472,-.0350967674986961,-.0385192237402188,-.0385192237402188,-.0350967674986961,-.0285561587634472,-.0194797896774684,-.00869007782643163,.00471291107004194,.0105645217737627,.0154869311233072,.0190341153866528,.0208902244146234,.0208902244146234,.0190341153866528,.0154869311233072,.0105645217737627,.00471291107004194,.00928552230067672,.0208145456318542,.0305128278843661,.0375015993872437,.0411585624649609,.0411585624649609,.0375015993872437,.030512827884366,.0208145456318542,.00928552230067672,-.000280476734556647,-.00062872023792106,-.000921664721700392,-.0011327662350258,-.00124322777172983,-.00124322777172983,-.0011327662350258,-.000921664721700392,-.00062872023792106,-.000280476734556647,-.00502787889479991,-.0112705576809204,-.0165219358020696,-.0203061813837464,-.0222863357447108,-.0222863357447108,-.0203061813837464,-.0165219358020696,-.0112705576809204,-.00502787889479991},
			{.00302011496343164,.00561961421020581,.00536170126189668,.00190531860926006,-.00385485822902238,-.0100772652494075,-.0145995645209458,-.0156902564956459,-.0126653126632867,-.00616325300098763,.00676992835558772,.0125969991373797,.0120188581714983,.00427098658012745,-.00864109954336841,-.0225893267590133,-.032726570685666,-.0351714797755983,-.0283907269656191,-.0138156268086686,.00992429344157325,.0184664163867283,.0176188977285336,.00626100039466762,-.0126673138949028,-.0331145465105248,-.0479751149143058,-.0515591994085577,-.0416190379908282,-.0202528486753225,.0121973904961468,.0226960330283889,.0216543955468261,.007695043194751,-.0155686825488407,-.040699225317108,-.0589635135390092,-.0633685100664642,-.0511516171339274,-.0248916364078325,.0133868172784104,.024909233429215,.0237660208182528,.00844542422661216,-.0170868603913883,-.044668004428086,-.0647133320925609,-.0695478812236805,-.0561396597315967,-.0273189407568403,.0133868172784104,.024909233429215,.0237660208182528,.00844542422661216,-.0170868603913883,-.044668004428086,-.0647133320925609,-.0695478812236805,-.0561396597315967,-.0273189407568403,.0121973904961468,.0226960330283889,.0216543955468261,.007695043194751,-.0155686825488407,-.040699225317108,-.0589635135390092,-.0633685100664642,-.0511516171339274,-.0248916364078325,.00992429344157325,.0184664163867283,.0176188977285336,.00626100039466762,-.0126673138949028,-.0331145465105248,-.0479751149143058,-.0515591994085576,-.0416190379908282,-.0202528486753225,.00676992835558772,.0125969991373797,.0120188581714983,.00427098658012745,-.00864109954336841,-.0225893267590133,-.032726570685666,-.0351714797755983,-.0283907269656191,-.0138156268086686,.00302011496343164,.00561961421020581,.00536170126189668,.00190531860926006,-.00385485822902238,-.0100772652494075,-.0145995645209458,-.0156902564956459,-.0126653126632867,-.00616325300098763},
			{-.00509449780740388,-.00947947764207594,-.00904441574358152,-.00321399734606523,.00650258915087549,.016998891214194,.0246273570183889,.0264671968724815,.0213645534605355,.0103965177750484,-.0101436118237301,-.0188745083672825,-.0180082602728758,-.00639936313911146,.0129472506789226,.0338463496166405,.0490353238499585,.0526986135407738,.0425387830722098,.0207004192788813,-.0116786295673759,-.0217307597449812,-.0207334236102965,-.0073677692785882,.014906538935309,.0389682675411454,.0564557667339239,.0606734166243637,.0489761140686684,.0238329830486871,-.00915612782195732,-.0170370686684944,-.0162551501156413,-.00577638299838646,.0116868314974169,.0305513961675494,.0442617186819942,.0475683858968708,.0383976183205756,.0186852265424968,-.00345189684709876,-.0064230430990214,-.00612825667404284,-.00217771951719025,.00440598226487886,.01151799867323,.0166868451529422,.0179334719317936,.0144760558386965,.00704440521621975,.00345189684709876,.0064230430990214,.00612825667404284,.00217771951719025,-.00440598226487886,-.01151799867323,-.0166868451529422,-.0179334719317936,-.0144760558386965,-.00704440521621975,.00915612782195732,.0170370686684944,.0162551501156413,.00577638299838646,-.0116868314974169,-.0305513961675494,-.0442617186819942,-.0475683858968708,-.0383976183205756,-.0186852265424968,.0116786295673759,.0217307597449812,.0207334236102965,.0073677692785882,-.014906538935309,-.0389682675411454,-.0564557667339239,-.0606734166243637,-.0489761140686684,-.0238329830486871,.0101436118237301,.0188745083672825,.0180082602728758,.00639936313911146,-.0129472506789226,-.0338463496166405,-.0490353238499585,-.0526986135407738,-.0425387830722098,-.0207004192788813,.00509449780740388,.00947947764207594,.00904441574358152,.00321399734606523,-.00650258915087549,-.016998891214194,-.0246273570183889,-.0264671968724815,-.0213645534605355,-.0103965177750484},
			{.00623143749421823,.0115950138046163,.0110628590900448,.00393126551933778,-.00795377275172705,-.0207925358056529,-.0301234472384809,-.0323738843735185,-.0261324834192153,-.0127167098940724,.00942344685839383,.017534476838053,.0167297296705077,.00594502821252533,-.0120280360541503,-.0314433638141471,-.045553968006072,-.0489571754954932,-.0395186614657246,-.0192307537404636,.00426953652352492,.00794444860844995,.00757983707344909,.00269354891773532,-.00544961307801359,-.0142462298821585,-.0206394043619152,-.0221813156064483,-.0179049525109209,-.00871299076692808,-.00595266514038301,-.0110762941201031,-.0105679461150626,-.00375539468000193,.00759794924327869,.0198623516941821,.0287758313312213,.0309255919116317,.0249634090409368,.0121478095153878,-.0139717558497607,-.0259976453431518,-.0248044799213386,-.00881444804152065,.0178334727861298,.0466197781881845,.0675409888263679,.0725867841561221,.0585926888330286,.0285126451187338,-.0139717558497607,-.0259976453431518,-.0248044799213386,-.00881444804152065,.0178334727861298,.0466197781881845,.0675409888263679,.0725867841561221,.0585926888330287,.0285126451187339,-.00595266514038299,-.011076294120103,-.0105679461150626,-.00375539468000192,.00759794924327867,.0198623516941821,.0287758313312212,.0309255919116316,.0249634090409367,.0121478095153878,.0042695365235249,.00794444860844991,.00757983707344906,.0026935489177353,-.00544961307801357,-.0142462298821584,-.0206394043619151,-.0221813156064482,-.0179049525109208,-.00871299076692804,.00942344685839383,.017534476838053,.0167297296705077,.00594502821252533,-.0120280360541503,-.0314433638141471,-.045553968006072,-.0489571754954932,-.0395186614657246,-.0192307537404636,.00623143749421823,.0115950138046163,.0110628590900448,.00393126551933778,-.00795377275172705,-.0207925358056529,-.0301234472384809,-.0323738843735185,-.0261324834192153,-.0127167098940724},
			{-.00677992663267428,-.0126156032172487,-.0120366084146251,-.00427729425504189,.00865386128958332,.0226226881679193,.0327749034455745,.0352234233386421,.0284326562653077,.0138360306383256,-.00574597286050269,-.0106916958888372,-.0102009990712324,-.00362499744280574,.00733412834718052,.0191726783027454,.0277766583485262,.02985177355496,.0240966134449184,.0117260054351862,.00617255053363847,.0114854411544103,.0109583152913243,.00389411513824503,-.00787860975716505,-.0205960467551767,-.0298387812602527,-.0320679518090523,-.0258855319698288,-.0125965372380952,.0144116920845997,.0268162472661978,.025585511999273,.009091993335417,-.018395004991244,-.0480877203642201,-.0696676884796602,-.0748723634966211,-.0604376285236068,-.0294104382002705,.00761712583185733,.0141734036895712,.0135229134252192,.00480546329270283,-.00972245777063737,-.025416183945038,-.0368220155169205,-.0395728836513071,-.0319435787790183,-.0155445319832301,-.0076171258318574,-.0141734036895713,-.0135229134252194,-.00480546329270287,.00972245777063746,.0254161839450382,.0368220155169208,.0395728836513074,.0319435787790186,.0155445319832303,-.0144116920845997,-.0268162472661979,-.0255855119992731,-.00909199333541702,.018395004991244,.0480877203642202,.0696676884796604,.0748723634966213,.060437628523607,.0294104382002706,-.0061725505336384,-.0114854411544102,-.0109583152913242,-.00389411513824499,.00787860975716497,.0205960467551765,.0298387812602524,.0320679518090519,.0258855319698285,.012596537238095,.00574597286050274,.0106916958888373,.0102009990712325,.00362499744280577,-.00733412834718057,-.0191726783027455,-.0277766583485264,-.0298517735549602,-.0240966134449186,-.0117260054351863,.00677992663267429,.0126156032172487,.0120366084146252,.0042772942550419,-.00865386128958333,-.0226226881679194,-.0327749034455745,-.0352234233386422,-.0284326562653077,-.0138360306383256},
			{-.00345430310158809,-.00436440955976362,.000321510543493063,.00721105333392678,.0101264765379077,.00504590248017281,-.00626512731169085,-.0168673409888969,-.0192939898358052,-.0109524916466014,-.00774321004312494,-.00978331632795496,.00072070214921253,.0161643894454719,.022699639413156,.0113109595805708,-.0140439895674001,-.0378100474408264,-.043249654554018,-.0245512454527271,-.0113510637943875,-.0143417325812942,.0010565044764254,.0236959884558197,.0332762579929111,.016581162471344,-.0205876142606789,-.0554271753160955,-.0634013006484006,-.035990597157184,-.0139509536333168,-.0176266163143223,.00129849018831963,.0291234057204582,.0408979757984384,.0203789735494738,-.0253030779470516,-.0681224215516175,-.0779229701867376,-.0442340172930329,-.015311378873029,-.0193454732724202,.00142511227252541,.0319633704461579,.0448861360340558,.0223662262280236,-.0277705039586591,-.074765369704482,-.0855216174320572,-.0485474911358245,-.015311378873029,-.0193454732724202,.00142511227252541,.0319633704461579,.0448861360340558,.0223662262280236,-.0277705039586591,-.074765369704482,-.0855216174320572,-.0485474911358245,-.0139509536333168,-.0176266163143223,.00129849018831963,.0291234057204581,.0408979757984384,.0203789735494738,-.0253030779470516,-.0681224215516175,-.0779229701867376,-.0442340172930329,-.0113510637943875,-.0143417325812942,.0010565044764254,.0236959884558197,.0332762579929111,.016581162471344,-.0205876142606789,-.0554271753160955,-.0634013006484006,-.035990597157184,-.00774321004312494,-.00978331632795496,.00072070214921253,.0161643894454719,.022699639413156,.0113109595805708,-.0140439895674001,-.0378100474408264,-.043249654554018,-.0245512454527271,-.00345430310158809,-.00436440955976362,.000321510543493063,.00721105333392678,.0101264765379077,.00504590248017282,-.00626512731169085,-.0168673409888969,-.0192939898358052,-.0109524916466014},
			{.00582691049477723,.00736212866080022,-.000542341857377823,-.0121640056225925,-.0170819035500163,-.00851170880280428,.0105683650247204,.0284527684272998,.0325461745984863,.0184752717531363,.0116019125780642,.0146586725826016,-.00107985232008665,-.0242196494967433,-.0340116347817334,-.0169475919543621,.0210425828953059,.0566520683977071,.0648024150672126,.0367859585157503,.0133576128135457,.0168769477792085,-.00124326477125837,-.0278847731596392,-.0391585650653061,-.0195122459271369,.0242269257780715,.0652251419454623,.0746088684970023,.0423527230982234,.0104724625103868,.0132316459066911,-.000974728335761049,-.0218618585222674,-.0307005907665702,-.0152977382124873,.0189940804164865,.0511369706019496,.0584938782987402,.0332048331576998,.0039481602948217,.00498838348212032,-.000367476485083857,-.00824200819081637,-.0115742456343885,-.00576730855338134,.00716084436326645,.0192788426530853,.0220524263094917,.0125183550420337,-.0039481602948217,-.00498838348212032,.000367476485083857,.00824200819081637,.0115742456343885,.00576730855338134,-.00716084436326645,-.0192788426530853,-.0220524263094917,-.0125183550420337,-.0104724625103868,-.0132316459066911,.000974728335761049,.0218618585222673,.0307005907665701,.0152977382124873,-.0189940804164865,-.0511369706019496,-.0584938782987402,-.0332048331576998,-.0133576128135457,-.0168769477792085,.00124326477125837,.0278847731596392,.0391585650653061,.0195122459271369,-.0242269257780715,-.0652251419454623,-.0746088684970023,-.0423527230982234,-.0116019125780642,-.0146586725826016,.00107985232008665,.0242196494967433,.0340116347817334,.0169475919543621,-.0210425828953059,-.0566520683977071,-.0648024150672126,-.0367859585157503,-.00582691049477723,-.00736212866080022,.000542341857377823,.0121640056225925,.0170819035500163,.00851170880280428,-.0105683650247204,-.0284527684272998,-.0325461745984863,-.0184752717531363},
			{-.00712730281031248,-.00900513579898562,.000663376354880294,.0148786482195442,.0208940740185803,.0104112678794644,-.0129269083176432,-.0348025761773465,-.039809508295799,-.0225984003710785,-.0107782127862115,-.013617952315673,.00100318615618461,.0225001295369971,.0315969142516674,.0157443655146451,-.0195486248056994,-.0526299473350607,-.0602016446818767,-.0341742696094045,-.00488334829500983,-.00616996579501025,.000454519455363577,.0101942660987876,.0143157998825042,.00713339233665088,-.00885701048107196,-.0238453599574428,-.0272759134325242,-.0154835374417349,.00680845262802778,.008602278046562,-.000633699255799995,-.0142130304082612,-.0199593474485439,-.00994550477811579,.012348604408976,.0332456326808434,.0380285724615008,.0215874283007975,.015980411393903,.0201907760278596,-.00148738272276524,-.0333600137192788,-.0468474410864738,-.0233435211431231,.0289839394319417,.0780322514257342,.0892584946770308,.0506687795348675,.015980411393903,.0201907760278596,-.00148738272276524,-.0333600137192788,-.0468474410864738,-.0233435211431231,.0289839394319417,.0780322514257343,.0892584946770308,.0506687795348676,.00680845262802777,.00860227804656198,-.000633699255799993,-.0142130304082611,-.0199593474485438,-.00994550477811577,.012348604408976,.0332456326808434,.0380285724615007,.0215874283007975,-.00488334829500981,-.00616996579501022,.000454519455363575,.0101942660987876,.0143157998825041,.00713339233665084,-.00885701048107192,-.0238453599574427,-.0272759134325241,-.0154835374417348,-.0107782127862115,-.013617952315673,.00100318615618461,.0225001295369971,.0315969142516674,.0157443655146451,-.0195486248056994,-.0526299473350607,-.0602016446818768,-.0341742696094045,-.00712730281031249,-.00900513579898562,.000663376354880294,.0148786482195442,.0208940740185803,.0104112678794644,-.0129269083176432,-.0348025761773465,-.039809508295799,-.0225984003710785},
			{.00360886668167895,.00181411871372813,-.00573065302975065,-.00778508692673576,.00135437648599479,.0117852891818123,.00860428289223646,-.00859845646519095,-.022302893606761,-.0160855169229064,.00808968174247022,.00406655172706256,-.0128459051764245,-.0174511504939989,.00303598766527052,.0264180550663856,.0192874706549125,-.0192744100614766,-.0499944517570629,-.0360575006081668,.011858969732663,.00596131162915412,-.0188312971419022,-.0255823000331391,.00445056888235353,.0387272238144227,.0282742310510711,-.0282550850343082,-.0732887533851304,-.0528580507812608,.0145751922353826,.00732671260055841,-.0231444874278468,-.0314417651121289,.00546994372019093,.047597451090816,.0347502660152623,-.0347267347236618,-.0900750818460811,-.0649648551848937,.015996490012713,.00804117592743191,-.0254014187950224,-.0345078043209106,.00600334449639179,.0522389096972928,.0381389332143976,-.0381131072722766,-.0988587405143971,-.0712998937070431,.015996490012713,.00804117592743191,-.0254014187950224,-.0345078043209106,.00600334449639179,.0522389096972928,.0381389332143976,-.0381131072722766,-.0988587405143971,-.0712998937070431,.0145751922353826,.00732671260055841,-.0231444874278468,-.0314417651121289,.00546994372019093,.0475974510908161,.0347502660152623,-.0347267347236618,-.0900750818460811,-.0649648551848937,.011858969732663,.00596131162915412,-.0188312971419022,-.0255823000331391,.00445056888235353,.0387272238144227,.0282742310510711,-.0282550850343082,-.0732887533851304,-.0528580507812608,.00808968174247022,.00406655172706256,-.0128459051764245,-.0174511504939989,.00303598766527052,.0264180550663856,.0192874706549125,-.0192744100614766,-.0499944517570629,-.0360575006081668,.00360886668167895,.00181411871372813,-.00573065302975065,-.00778508692673576,.00135437648599479,.0117852891818123,.00860428289223646,-.00859845646519095,-.022302893606761,-.0160855169229064},
			{-.00608763693165392,-.00306015629121496,.00966678409139083,.0131323173925849,-.00228463754490087,-.0198800809233674,-.0145141826854097,.0145043543443846,.0376217607296319,.0271339441220568,-.0121210427981259,-.00609305150606168,.0192474526663595,.0261476469344647,-.00454892263958079,-.0395830622633527,-.0288990673203423,.0288794981930212,.0749083720110797,.0540261683930699,-.0139553022404171,-.0070151039600872,.0221601411521044,.0301045315921654,-.00523730436076351,-.0455731084104202,-.0332723203472142,.0332497898528585,.086244144927353,.0622018683849159,-.0109410552299944,-.00549989090520571,.0173737067153911,.0236021647649496,-.00410608350003688,-.0357296379204776,-.0260857334565373,.0260680693903023,.0676160169559761,.0487667028405305,-.00412482162620984,-.00207348089107479,.00654995699049303,.00889811061193973,-.00154800991896502,-.0134702165461154,-.00983442595208666,.00982776652839477,.0254915091054086,.0183852422172328,.00412482162620984,.00207348089107479,-.00654995699049303,-.00889811061193973,.00154800991896502,.0134702165461154,.00983442595208666,-.00982776652839477,-.0254915091054086,-.0183852422172328,.0109410552299944,.00549989090520571,-.0173737067153911,-.0236021647649496,.00410608350003688,.0357296379204776,.0260857334565373,-.0260680693903023,-.0676160169559761,-.0487667028405305,.0139553022404171,.0070151039600872,-.0221601411521044,-.0301045315921654,.00523730436076351,.0455731084104202,.0332723203472142,-.0332497898528585,-.086244144927353,-.0622018683849159,.0121210427981259,.00609305150606168,-.0192474526663595,-.0261476469344647,.00454892263958079,.0395830622633527,.0288990673203423,-.0288794981930212,-.0749083720110797,-.0540261683930699,.00608763693165392,.00306015629121496,-.00966678409139083,-.0131323173925849,.00228463754490087,.0198800809233674,.0145141826854097,-.0145043543443846,-.0376217607296319,-.0271339441220568},
			{-.0035177389049427,.00124629807811922,.00733218354822258,-.000640159716652238,-.0104475716905616,-.00111787975143421,.0140189382881418,.00476714217793534,-.0203246135001665,-.0211329311052817,-.00788540855182074,.00279371772291301,.0164359164841472,-.00143499021406278,-.0234194104169471,-.00250585355822768,.0314250314909489,.0106860869193045,-.0455599137506981,-.0473718488397425,-.0115595179541847,.00409541623172844,.0240940301879616,-.00210360630454689,-.0343313974681262,-.00367342529008241,.0460671394948289,.0156651380574512,-.0667879967830832,-.0694441808039846,-.0142071529086165,.00503344558650323,.0296126164102953,-.00258542411082186,-.0421947883405973,-.00451480027120738,.0566185283382572,.0192531394993187,-.0820853677911323,-.0853499340721777,-.0155925613838605,.0055242813098938,.0325002864461468,-.00283754137163211,-.0463094070649962,-.00495506037117232,.0621396759967787,.0211306066321764,-.0900899106409799,-.0936728206340164,-.0155925613838605,.0055242813098938,.0325002864461468,-.00283754137163211,-.0463094070649962,-.00495506037117232,.0621396759967787,.0211306066321764,-.0900899106409799,-.0936728206340164,-.0142071529086165,.00503344558650323,.0296126164102953,-.00258542411082186,-.0421947883405973,-.00451480027120738,.0566185283382572,.0192531394993187,-.0820853677911323,-.0853499340721777,-.0115595179541847,.00409541623172844,.0240940301879616,-.00210360630454689,-.0343313974681262,-.00367342529008241,.0460671394948289,.0156651380574512,-.0667879967830832,-.0694441808039846,-.00788540855182074,.00279371772291301,.0164359164841472,-.00143499021406278,-.0234194104169471,-.00250585355822768,.0314250314909489,.0106860869193045,-.0455599137506981,-.0473718488397425,-.0035177389049427,.00124629807811922,.00733218354822258,-.000640159716652238,-.0104475716905616,-.00111787975143421,.0140189382881418,.00476714217793534,-.0203246135001665,-.0211329311052817},
			{-.00537304737318019,-.0113801575780441,-.0150219679612532,-.0157580888060425,-.0138618270499793,-.0102693186804868,-.00622899008451743,-.002867614478685,-.000823350121801073,-7.1027404987251e-005,-.012044291759767,-.0255099068782191,-.0336734356436761,-.035323533550749,-.0310728489286636,-.0230198361881556,-.0139629838964585,-.00642808131717099,-.0018456321708443,-.000159215940078199,-.0176561817853201,-.0373959351161665,-.0493631599864637,-.0517821007753423,-.0455508617870686,-.0337456465280969,-.0204688649908742,-.00942316696826571,-.00270558184468062,-.000233400654618126,-.0217002192825492,-.0459612390812321,-.0606694816133146,-.0636424655907344,-.0559840004655092,-.0414748748282566,-.0251571299031463,-.0115814841529019,-.00332527836597537,-.000286859608010771,-.0238163130489831,-.050443142708611,-.066585657398596,-.0698485514815503,-.0614432722296599,-.0455192912899041,-.0276103238122155,-.0127108509165695,-.00364954240820819,-.000314832681482957,-.0238163130489831,-.050443142708611,-.066585657398596,-.0698485514815503,-.0614432722296599,-.0455192912899041,-.0276103238122155,-.0127108509165695,-.00364954240820819,-.000314832681482957,-.0217002192825492,-.0459612390812321,-.0606694816133146,-.0636424655907344,-.0559840004655092,-.0414748748282566,-.0251571299031463,-.0115814841529019,-.00332527836597537,-.000286859608010771,-.0176561817853201,-.0373959351161665,-.0493631599864637,-.0517821007753423,-.0455508617870686,-.0337456465280969,-.0204688649908742,-.00942316696826571,-.00270558184468062,-.000233400654618126,-.012044291759767,-.0255099068782191,-.0336734356436761,-.035323533550749,-.0310728489286636,-.0230198361881556,-.0139629838964585,-.00642808131717099,-.0018456321708443,-.000159215940078199,-.00537304737318019,-.0113801575780441,-.0150219679612533,-.0157580888060425,-.0138618270499793,-.0102693186804868,-.00622899008451743,-.002867614478685,-.000823350121801073,-7.1027404987251e-005},
			{.00906355498937132,.0191966824099048,.0253398906075171,.0265816201684498,.0233829004277474,.0173228576075033,.0105074067355295,.00483725150937865,.00138887275435145,.000119812975047951,.0180463682645229,.0382223532192926,.0504540434986315,.052926440804966,.0465574967776602,.0344913963831515,.0209212093573502,.00963141087892507,.00276537288370632,.000238558608969547,.0207772984278405,.0440064852834015,.0580891791243752,.060935720650788,.0536029737631114,.0397109282788219,.0240871849625023,.0110889180127161,.0031838526636951,.000274659329702082,.0162895482815631,.0345013943567046,.0455423254986296,.0477740341007375,.0420251089034541,.031133647415719,.0188845226330227,.00869378981034385,.00249616290908881,.000215334848642728,.00614122492031466,.0130071638050768,.0171696390499558,.0180110021280012,.0158436343117591,.0117375097249386,.0071195406402661,.00327759356566627,.000941063410568089,8.11821000704882e-005,-.00614122492031466,-.0130071638050768,-.0171696390499558,-.0180110021280012,-.0158436343117591,-.0117375097249386,-.0071195406402661,-.00327759356566627,-.000941063410568089,-8.11821000704882e-005,-.0162895482815631,-.0345013943567046,-.0455423254986296,-.0477740341007375,-.0420251089034541,-.031133647415719,-.0188845226330227,-.00869378981034385,-.00249616290908881,-.000215334848642728,-.0207772984278405,-.0440064852834015,-.0580891791243752,-.060935720650788,-.0536029737631114,-.0397109282788219,-.0240871849625023,-.0110889180127161,-.0031838526636951,-.000274659329702082,-.0180463682645229,-.0382223532192926,-.0504540434986315,-.052926440804966,-.0465574967776602,-.0344913963831515,-.0209212093573502,-.00963141087892507,-.00276537288370632,-.000238558608969547,-.00906355498937132,-.0191966824099048,-.0253398906075171,-.0265816201684498,-.0233829004277474,-.0173228576075033,-.0105074067355295,-.00483725150937865,-.00138887275435145,-.000119812975047951},
			{-.0110862696476229,-.0234808083346496,-.0309949970453799,-.0325138435419332,-.0286012651315648,-.0211888018254831,-.0128523459618144,-.0059167815111433,-.00169882765416366,-.00014655165111517,-.0167651321190546,-.035508684751954,-.0468719629787211,-.0491688277486249,-.0432520589832132,-.0320426142733104,-.0194358684154698,-.00894761059462107,-.00256903999044115,-.000221621688025442,-.00759587706914969,-.0160881287631775,-.0212365561003511,-.0222772101383125,-.019596464894583,-.0145177358141793,-.00880592327979835,-.00405394658728139,-.00116397006683541,-.000100411442399873,.0105903093675451,.0224303604700153,.0296083910990837,.0310592898045613,.0273217462387268,.0202408901814494,.0122773777077953,.00565208574704063,.00162282814612863,.000139995451397573,.0248569696711101,.052647261809357,.0694951255924348,.0729005922191637,.0641280432939063,.0475082621191814,.0288167601844263,.0132662530542551,.00380900959639235,.000328589332824135,.0248569696711101,.0526472618093571,.0694951255924348,.0729005922191637,.0641280432939063,.0475082621191814,.0288167601844263,.0132662530542551,.00380900959639235,.000328589332824135,.0105903093675451,.0224303604700152,.0296083910990836,.0310592898045612,.0273217462387267,.0202408901814493,.0122773777077953,.00565208574704061,.00162282814612863,.000139995451397573,-.00759587706914966,-.0160881287631774,-.021236556100351,-.0222772101383123,-.0195964648945829,-.0145177358141792,-.0088059232797983,-.00405394658728137,-.00116397006683541,-.000100411442399872,-.0167651321190546,-.035508684751954,-.0468719629787211,-.0491688277486249,-.0432520589832132,-.0320426142733104,-.0194358684154698,-.00894761059462107,-.00256903999044115,-.000221621688025442,-.0110862696476229,-.0234808083346496,-.0309949970453799,-.0325138435419332,-.0286012651315648,-.0211888018254831,-.0128523459618144,-.0059167815111433,-.00169882765416366,-.00014655165111517},
			{.0120620795615023,.0255475815854486,.0337231667867706,.0353757016668926,.0311187393543622,.0230538333953191,.0139836053488558,.00643757473017433,.00184835792174656,.00015945108064368,.0102225858120058,.0216515190199873,.0285803094377376,.0299808291021325,.0263730629357406,.0195380729315951,.01185107467672,.00545583037857288,.00156647926006563,.00013513443899835,-.0109815046191396,-.0232589151611855,-.0307020949375013,-.0322065883647599,-.0283309837428373,-.0209885680681119,-.0127308915471588,-.00586086804311059,-.00168277376649596,-.000145166740916215,-.0256396545211207,-.0543049946206344,-.0716833562042298,-.0751960526007667,-.0661472776821311,-.0490041804673553,-.0297241291003144,-.0136839747403398,-.00392894594197619,-.000338935802891862,-.0135515298013773,-.028702249180423,-.0378873723537952,-.039743965618796,-.0349613448984257,-.025900567866404,-.0157103295206683,-.00723249961664868,-.0020765969360793,-.000179140426009222,.0135515298013774,.0287022491804232,.0378873723537955,.0397439656187963,.034961344898426,.0259005678664042,.0157103295206684,.00723249961664874,.00207659693607932,.000179140426009223,.0256396545211207,.0543049946206346,.07168335620423,.0751960526007669,.0661472776821312,.0490041804673554,.0297241291003145,.0136839747403399,.0039289459419762,.000338935802891863,.0109815046191395,.0232589151611852,.030702094937501,.0322065883647595,.028330983742837,.0209885680681117,.0127308915471586,.00586086804311052,.00168277376649595,.000145166740916214,-.0102225858120059,-.0216515190199875,-.0285803094377378,-.0299808291021328,-.0263730629357408,-.0195380729315952,-.0118510746767201,-.00545583037857292,-.00156647926006564,-.000135134438998351,-.0120620795615023,-.0255475815854486,-.0337231667867706,-.0353757016668926,-.0311187393543623,-.0230538333953191,-.0139836053488558,-.00643757473017434,-.00184835792174656,-.00015945108064368},
			{.00615133467063503,.00923599360757013,.00365245862188397,-.00803852164203684,-.0191471282194023,-.0235470344199801,-.0197092364291758,-.0112335234696543,-.00369340218158754,-.000342287092850262,.013788910526811,.0207035214794833,.00818739799355415,-.0180192529954539,-.0429204476750506,-.0527833337273658,-.0441804766323359,-.025181209984126,-.00827917760101021,-.000767275128164547,.0202136842696012,.0303500733898785,.0120022156724767,-.0264151029275269,-.0629187038618097,-.0773770807046292,-.0647658278578351,-.0369140859356472,-.0121367588623754,-.00112477756371539,.0248435016410661,.037301566998444,.0147512477576403,-.0324653162766497,-.0773298376385226,-.095099814849551,-.0796000337796347,-.0453690253735618,-.0149166072197978,-.00138240079726687,.0272661121352065,.040939023946499,.016189713563936,-.0356311669302222,-.0848706456485974,-.104373459634974,-.0873622196403312,-.0497931793662997,-.0163711980302944,-.00151720541245167,.0272661121352065,.040939023946499,.016189713563936,-.0356311669302222,-.0848706456485974,-.104373459634974,-.0873622196403312,-.0497931793662997,-.0163711980302944,-.00151720541245167,.0248435016410661,.037301566998444,.0147512477576403,-.0324653162766497,-.0773298376385226,-.095099814849551,-.0796000337796348,-.0453690253735618,-.0149166072197978,-.00138240079726687,.0202136842696012,.0303500733898785,.0120022156724767,-.0264151029275269,-.0629187038618097,-.0773770807046292,-.0647658278578351,-.0369140859356472,-.0121367588623754,-.00112477756371539,.013788910526811,.0207035214794833,.00818739799355415,-.0180192529954539,-.0429204476750506,-.0527833337273658,-.0441804766323359,-.025181209984126,-.00827917760101021,-.000767275128164547,.00615133467063503,.00923599360757013,.00365245862188397,-.00803852164203684,-.0191471282194022,-.0235470344199801,-.0197092364291758,-.0112335234696543,-.00369340218158754,-.000342287092850262},
			{-.0103764132698787,-.015579787438262,-.00616117024045563,.0135598251603545,.0322984402530261,.0397204466191965,.0332466356283884,.0189493318506081,.0062302361129812,.000577388895667578,-.0206603893673846,-.0310207839996088,-.0122674543520802,.0269988540625906,.0643091532913213,.0790870478705028,.0661970971445412,.0377298555968102,.0124049708311625,.00114963418385859,-.0237868954700111,-.0357151132670011,-.0141238700378378,.031084550633434,.0740409621476184,.0910551736114528,.0762146057460633,.0434394587256295,.0142821966818892,.00132360662105264,-.0186490935562653,-.0280009003078451,-.0110732135702299,.0243705065946171,.0580486357216159,.0713878973236672,.0597527876096367,.0340568415423994,.0111973427741056,.00103771691176861,-.00703078293574547,-.0105564515227481,-.00417464584959855,.00918777855792046,.0218845680753062,.0269135231054363,.0225270401599717,.0128395656142358,.00422144305643347,.000391223430424914,.00703078293574547,.0105564515227481,.00417464584959855,-.00918777855792046,-.0218845680753062,-.0269135231054363,-.0225270401599717,-.0128395656142358,-.00422144305643347,-.000391223430424914,.0186490935562653,.0280009003078451,.0110732135702299,-.0243705065946171,-.0580486357216159,-.0713878973236672,-.0597527876096368,-.0340568415423994,-.0111973427741056,-.00103771691176861,.0237868954700111,.0357151132670011,.0141238700378378,-.031084550633434,-.0740409621476184,-.0910551736114528,-.0762146057460633,-.0434394587256295,-.0142821966818892,-.00132360662105264,.0206603893673846,.0310207839996088,.0122674543520802,-.0269988540625906,-.0643091532913213,-.0790870478705028,-.0661970971445413,-.0377298555968102,-.0124049708311625,-.00114963418385859,.0103764132698787,.015579787438262,.00616117024045563,-.0135598251603545,-.0322984402530261,-.0397204466191965,-.0332466356283884,-.0189493318506081,-.0062302361129812,-.000577388895667578},
			{.0126921186677361,.0190567304753116,.00753615934428767,-.0165859729728156,-.0395064870502198,-.0485848634700931,-.040666291301582,-.0231782565195614,-.00762063865589431,-.000706244844981167,.0191935658340169,.0288184045969256,.0113965031604947,-.0250820192048949,-.0597434029668103,-.0734721128884853,-.0614973086649725,-.0350511529297959,-.0115242563962926,-.00106801372426019,.00869614182330973,.0130569241621705,.00516348075342068,-.0113640580446065,-.027068295162004,-.0332884425573554,-.027862947538109,-.0158808321280148,-.00522136266376985,-.000483891263130864,-.0121243184130687,-.0182042001216115,-.00719901837461035,.0158439755235,.0377391073088729,.0464113494513956,.0388470260424168,.0221413437472276,.00727971838223315,.000674649962098675,-.0284575081442016,-.0427278594614471,-.0168971250214698,.0371880750022642,.0885790785929488,.108934070353419,.0911795222062535,.0519688982542005,.0170865394731464,.00158349988319459,-.0284575081442016,-.0427278594614471,-.0168971250214699,.0371880750022643,.0885790785929488,.108934070353419,.0911795222062535,.0519688982542005,.0170865394731464,.00158349988319459,-.0121243184130687,-.0182042001216115,-.00719901837461033,.0158439755235,.0377391073088728,.0464113494513955,.0388470260424167,.0221413437472276,.00727971838223313,.000674649962098674,.00869614182330968,.0130569241621705,.00516348075342066,-.0113640580446064,-.0270682951620039,-.0332884425573552,-.0278629475381089,-.0158808321280148,-.00522136266376982,-.000483891263130862,.0191935658340169,.0288184045969256,.0113965031604947,-.0250820192048949,-.0597434029668103,-.0734721128884853,-.0614973086649725,-.0350511529297959,-.0115242563962926,-.00106801372426019,.0126921186677361,.0190567304753116,.00753615934428767,-.0165859729728156,-.0395064870502198,-.0485848634700931,-.040666291301582,-.0231782565195614,-.00762063865589432,-.000706244844981167},
			{-.00643614945112683,-.00460242005009466,.00815688456903964,.0159153508876072,.00595117342612003,-.0150500971453785,-.0280693431449012,-.0232405097467362,-.00939844620771058,-.000965522727863794,-.0144273549840217,-.010316843689307,.0182845768475155,.0356760542455353,.0133402265193015,-.0337364903828284,-.0629205988448425,-.0520962240967035,-.021067677307141,-.00216432810421968,-.0211496040913043,-.0151238504731025,.0268040511743329,.0522988741642954,.0195559414517781,-.049455594307975,-.0922376801728386,-.0763698207687696,-.0308839031591627,-.00317277023949471,-.0259937880171798,-.0185878733948966,.0329433506754726,.0642776026773378,.024035106963676,-.0607830874353671,-.113364141250993,-.093861848354391,-.0379576671221214,-.00389947332709414,-.028528568522066,-.0204004671991369,.036155816784784,.0705456238699261,.0263788869669642,-.0667103414760682,-.124418829202062,-.103014773014854,-.0416591035794981,-.00427972990848616,-.028528568522066,-.0204004671991369,.036155816784784,.0705456238699261,.0263788869669642,-.0667103414760682,-.124418829202062,-.103014773014854,-.0416591035794981,-.00427972990848616,-.0259937880171798,-.0185878733948966,.0329433506754726,.0642776026773378,.024035106963676,-.0607830874353671,-.113364141250993,-.093861848354391,-.0379576671221214,-.00389947332709414,-.0211496040913043,-.0151238504731025,.0268040511743329,.0522988741642954,.019555941451778,-.049455594307975,-.0922376801728386,-.0763698207687696,-.0308839031591627,-.00317277023949471,-.0144273549840217,-.010316843689307,.0182845768475155,.0356760542455353,.0133402265193015,-.0337364903828284,-.0629205988448425,-.0520962240967035,-.021067677307141,-.00216432810421968,-.00643614945112683,-.00460242005009466,.00815688456903964,.0159153508876072,.00595117342612003,-.0150500971453785,-.0280693431449012,-.0232405097467362,-.00939844620771058,-.000965522727863794},
			{.0108568546750866,.00776361798572502,-.0137594863264114,-.026846898600142,-.0100387701566362,.0253873404889165,.0473489283714467,.0392033837640413,.0158538215160414,.0016286974100757,.0216169922167163,.0154580746075647,-.0273963977344734,-.0534546344636977,-.0199881109986099,.050548520559216,.0942760538578598,.0780575283595627,.0315664109517744,.00324288574275418,.0248882596107707,.0177973221278318,-.0315422539999915,-.0615438450730705,-.0230128822120598,.0581979532585155,.108542709363562,.0898698584387518,.0363433091371754,.00373362683599065,.0195125707984483,.0139532258773057,-.0247293492571541,-.0482508079302309,-.0180422617114151,.0456276051857851,.0850982485008847,.0704586018812573,.0284934102938292,.00292718973168883,.00735631731313645,.00526042202001852,-.00932306367838869,-.0181907477705709,-.00680200490067545,.0172017898334461,.0320823803911286,.0265631750030447,.0107421297593173,.00110356224838248,-.00735631731313645,-.00526042202001852,.00932306367838869,.0181907477705709,.00680200490067545,-.0172017898334461,-.0320823803911286,-.0265631750030447,-.0107421297593173,-.00110356224838248,-.0195125707984483,-.0139532258773057,.0247293492571541,.0482508079302309,.0180422617114151,-.0456276051857851,-.0850982485008847,-.0704586018812573,-.0284934102938292,-.00292718973168883,-.0248882596107707,-.0177973221278318,.0315422539999915,.0615438450730705,.0230128822120598,-.0581979532585155,-.108542709363562,-.0898698584387518,-.0363433091371754,-.00373362683599065,-.0216169922167163,-.0154580746075647,.0273963977344734,.0534546344636977,.0199881109986099,-.050548520559216,-.0942760538578598,-.0780575283595627,-.0315664109517744,-.00324288574275418,-.0108568546750866,-.00776361798572502,.0137594863264114,.026846898600142,.0100387701566362,-.0253873404889165,-.0473489283714467,-.0392033837640413,-.0158538215160414,-.0016286974100757},
			{.00628780165042657,-.00108441059929728,-.0131544407550812,-.00292698561809638,.0175762393921458,.0118959849140124,-.0179938136163745,-.0323721609932466,-.0177761887834042,-.00209542799580947,.0140948166553123,-.00243082867843575,-.0294871627563918,-.00656116842314922,.0393991231745305,.0266661920364473,-.0403351628999706,-.0725658504041882,-.0398473322890685,-.00469713821417048,.0206621236068235,-.00356344348771085,-.0432263445908294,-.00961826437899792,.0577566613983443,.0390909771623532,-.0591288373535847,-.106377018399916,-.0584137080526403,-.00688571215596096,.0253946531907593,-.00437962782806131,-.0531270672113226,-.0118212673996889,.070985461783933,.0480445198574855,-.0726719260198027,-.130742006055988,-.0717930008459142,-.00846284125966016,.0278710091491036,-.00480670660664831,-.0583077455395459,-.0129740165922258,.0779075988938198,.0527295743105294,-.0797585184474156,-.143491293997453,-.0787938850114395,-.00928809400953851,.0278710091491036,-.00480670660664831,-.0583077455395459,-.0129740165922258,.0779075988938198,.0527295743105294,-.0797585184474156,-.143491293997453,-.0787938850114395,-.00928809400953851,.0253946531907593,-.00437962782806131,-.0531270672113226,-.0118212673996889,.070985461783933,.0480445198574855,-.0726719260198027,-.130742006055988,-.0717930008459142,-.00846284125966016,.0206621236068235,-.00356344348771085,-.0432263445908294,-.00961826437899792,.0577566613983443,.0390909771623532,-.0591288373535847,-.106377018399916,-.0584137080526403,-.00688571215596096,.0140948166553123,-.00243082867843575,-.0294871627563918,-.00656116842314922,.0393991231745305,.0266661920364473,-.0403351628999706,-.0725658504041882,-.0398473322890685,-.00469713821417048,.00628780165042657,-.00108441059929729,-.0131544407550812,-.00292698561809638,.0175762393921458,.0118959849140124,-.0179938136163745,-.0323721609932466,-.0177761887834042,-.00209542799580947},
			{-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.000726793993235288,-8.78323358755477e-005,-1.46520306329324e-006,-.0187952392881457,-.0376133897003122,-.0447078783667129,-.0400285772779044,-.0282223784536199,-.0154894473150464,-.00625458279978263,-.00162919071725786,-.000196886088238751,-3.28441793937936e-006,-.0275526505160409,-.0551388873133415,-.0655389659619086,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461161,-.00238829215039021,-.00028862274630013,-4.81475219575472e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030351,-.00293531546086303,-.00035472997281053,-5.91754093322042e-006,-.0371655977491602,-.0743764998227977,-.088405100787603,-.0791522778068628,-.0558067683550777,-.0306287437708179,-.0123677759491016,-.00322155232641602,-.000389321415157626,-6.49458908735937e-006,-.0371655977491602,-.0743764998227977,-.088405100787603,-.0791522778068628,-.0558067683550778,-.0306287437708179,-.0123677759491016,-.00322155232641602,-.000389321415157626,-6.49458908735937e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030351,-.00293531546086303,-.00035472997281053,-5.91754093322042e-006,-.0275526505160409,-.0551388873133415,-.0655389659619087,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461161,-.00238829215039021,-.00028862274630013,-4.81475219575472e-006,-.0187952392881457,-.0376133897003122,-.044707878366713,-.0400285772779045,-.0282223784536199,-.0154894473150464,-.00625458279978263,-.00162919071725786,-.000196886088238751,-3.28441793937936e-006,-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.000726793993235288,-8.78323358755477e-005,-1.46520306329324e-006},
			{.0141437693703256,.0283047797903233,.0336435153051858,.0301222536495194,.0212378680478842,.0116560990333334,.00470669062902145,.00122599650964295,.000148160466678965,2.47158597654252e-006,.028161540477737,.0563574094647636,.0669873209378089,.0599761664108691,.042286540810412,.023208360949976,.00937145220590666,.00244107135996656,.000295001061621808,4.92115409267359e-006,.0324231846605908,.0648858927129044,.0771244129347917,.0690522707843786,.0486857003592358,.0267204478158996,.0107896201789892,.00281047507100845,.000339643135062007,5.66586504797623e-006,.0254200050985947,.0508709968146246,.0604660828524796,.054137466562375,.0381699319272815,.020949019253574,.00845913820104087,.00220343224709027,.000266282609662119,4.44207809674427e-006,.00958344369577791,.0191785694703894,.0227959553223306,.0204100416431963,.0143902171568408,.00789786413180114,.00318912896948871,.000830702778994263,.000100389609953414,1.67468122713911e-006,-.00958344369577791,-.0191785694703894,-.0227959553223306,-.0204100416431963,-.0143902171568408,-.00789786413180114,-.00318912896948871,-.000830702778994263,-.000100389609953414,-1.67468122713911e-006,-.0254200050985947,-.0508709968146246,-.0604660828524796,-.054137466562375,-.0381699319272815,-.020949019253574,-.00845913820104087,-.00220343224709027,-.000266282609662119,-4.44207809674427e-006,-.0324231846605908,-.0648858927129044,-.0771244129347917,-.0690522707843785,-.0486857003592358,-.0267204478158996,-.0107896201789892,-.00281047507100845,-.000339643135062007,-5.66586504797623e-006,-.028161540477737,-.0563574094647636,-.066987320937809,-.0599761664108691,-.042286540810412,-.023208360949976,-.00937145220590666,-.00244107135996656,-.000295001061621808,-4.92115409267359e-006,-.0141437693703256,-.0283047797903233,-.0336435153051858,-.0301222536495194,-.0212378680478842,-.0116560990333334,-.00470669062902145,-.00122599650964295,-.000148160466678965,-2.47158597654252e-006},
			{-.0173002360880135,-.0346215609128356,-.0411517426769954,-.0368446406325781,-.0259775256237805,-.0142573920616249,-.00575708334485075,-.00149960229869156,-.000181225455912347,-3.02317011727578e-006,-.0261621585010403,-.0523562082937465,-.0622314290413362,-.0557180447272185,-.0392843276459713,-.0215606393479301,-.00870610818288429,-.00226776286909028,-.000274056901760265,-4.57176742452532e-006,-.0118534431119483,-.0237213353990866,-.0281955597772025,-.0252445024158391,-.0177987815081903,-.00976860575008604,-.00394452766839492,-.00102746866850002,-.000124168573259081,-2.07135757110846e-006,.0165262850469198,.0330727153955572,.0393107600496111,.0351963424341469,.0248153835062233,.0136195670415941,.0054995319088043,.00143251542459375,.000173118073471871,2.88792423694412e-006,.0387895529705973,.0776263898435274,.0922679721988114,.0826108460155768,.0582452517469885,.031967070378892,.0129081873926539,.00336231843912396,.000406332860775417,6.7783709312756e-006,.0387895529705973,.0776263898435274,.0922679721988115,.0826108460155769,.0582452517469885,.031967070378892,.0129081873926539,.00336231843912396,.000406332860775417,6.77837093127561e-006,.0165262850469198,.0330727153955571,.039310760049611,.0351963424341468,.0248153835062232,.013619567041594,.00549953190880429,.00143251542459375,.000173118073471871,2.88792423694411e-006,-.0118534431119483,-.0237213353990865,-.0281955597772024,-.025244502415839,-.0177987815081902,-.00976860575008599,-.0039445276683949,-.00102746866850002,-.00012416857325908,-2.07135757110845e-006,-.0261621585010403,-.0523562082937465,-.0622314290413362,-.0557180447272185,-.0392843276459713,-.0215606393479301,-.00870610818288428,-.00226776286909028,-.000274056901760265,-4.57176742452532e-006,-.0173002360880135,-.0346215609128356,-.0411517426769954,-.0368446406325781,-.0259775256237805,-.0142573920616249,-.00575708334485075,-.00149960229869156,-.000181225455912347,-3.02317011727578e-006},
			{.00879759938930934,.0102248339942577,-.00281118458446222,-.0202713999635782,-.0287696758053183,-.0241048298357547,-.0129418509611811,-.00409370526960594,-.000560621794885306,-9.99671086012965e-006,.0197208113889518,.0229201187461582,-.00630158734405165,-.0454406296060175,-.0644904734997843,-.0540336950704248,-.0290106187535114,-.00917650212647326,-.00125669699064766,-2.24087550090162e-005,.0289094815853316,.0335994669670076,-.00923773467978503,-.0666131336541757,-.0945390186691718,-.0792100325801501,-.0425277607544874,-.0134521807450406,-.00184223954039861,-3.28498395682792e-005,.0355310166928675,.0412952137572032,-.0113535797638883,-.081870453361282,-.116192586869701,-.0973525928350061,-.052268477136742,-.0165333182193771,-.00226419293161506,-4.03738889129983e-005,.0389958186821867,.0453221105953125,-.0124607224637962,-.0898540388613718,-.127523090289016,-.106845916942104,-.0573654301771659,-.0181455623707819,-.0024849853801249,-4.43109428912591e-005,.0389958186821867,.0453221105953125,-.0124607224637962,-.0898540388613718,-.127523090289016,-.106845916942104,-.0573654301771659,-.0181455623707819,-.0024849853801249,-4.43109428912591e-005,.0355310166928675,.0412952137572032,-.0113535797638883,-.081870453361282,-.116192586869701,-.0973525928350061,-.052268477136742,-.0165333182193771,-.00226419293161506,-4.03738889129983e-005,.0289094815853316,.0335994669670076,-.00923773467978504,-.0666131336541757,-.0945390186691718,-.0792100325801501,-.0425277607544874,-.0134521807450406,-.00184223954039861,-3.28498395682792e-005,.0197208113889518,.0229201187461582,-.00630158734405165,-.0454406296060175,-.0644904734997843,-.0540336950704248,-.0290106187535114,-.00917650212647326,-.00125669699064766,-2.24087550090162e-005,.00879759938930934,.0102248339942577,-.00281118458446222,-.0202713999635783,-.0287696758053183,-.0241048298357547,-.0129418509611811,-.00409370526960595,-.000560621794885306,-9.99671086012965e-006},
			{-.0148402797021791,-.0172478183727625,.00474206242882225,.0341949243312682,.0485302884342244,.0406613669372952,.0218310336296094,.00690548961493529,.000945688008179328,1.68630075889125e-005,-.0295483563533413,-.0343419863926998,.00944188069959736,.0680852268213982,.0966282499631283,.0809605064182685,.0434675878214585,.0137494624111731,.0018829514554676,3.35757928708797e-005,-.0340198653272789,-.0395389082959526,.0108707065122495,.0783884632885607,.111250859819299,.0932121398646372,.0500454734640585,.0158301481800593,.00216789570853973,3.86567678440353e-005,-.0266718139851343,-.0309987825377814,.00852271045732846,.0614571072313015,.0872214575290303,.0730789740555617,.0392360036170109,.0124109476493752,.00169964550185853,3.03071782114281e-005,-.0100553806578476,-.0116866651260001,.00321309596463221,.0231695754808925,.0328828386953925,.0275510658038934,.0147921304520804,.00467897695329769,.000640773159040156,1.14259275260014e-005,.0100553806578476,.0116866651260001,-.00321309596463221,-.0231695754808925,-.0328828386953925,-.0275510658038934,-.0147921304520804,-.00467897695329769,-.000640773159040156,-1.14259275260014e-005,.0266718139851343,.0309987825377814,-.00852271045732846,-.0614571072313014,-.0872214575290303,-.0730789740555617,-.0392360036170109,-.0124109476493752,-.00169964550185853,-3.03071782114281e-005,.0340198653272789,.0395389082959526,-.0108707065122495,-.0783884632885607,-.111250859819299,-.0932121398646372,-.0500454734640585,-.0158301481800593,-.00216789570853973,-3.86567678440353e-005,.0295483563533413,.0343419863926998,-.00944188069959736,-.0680852268213982,-.0966282499631283,-.0809605064182685,-.0434675878214585,-.0137494624111731,-.0018829514554676,-3.35757928708797e-005,.0148402797021791,.0172478183727625,-.00474206242882225,-.0341949243312682,-.0485302884342244,-.0406613669372952,-.0218310336296094,-.00690548961493529,-.000945688008179328,-1.68630075889125e-005},
			{-.00863131572555909,-.00182782839348211,.0164716603838428,.014834131431226,-.0115521085131657,-.0329015220239097,-.0289052514771828,-.0121493830289503,-.00197137671826113,-3.83525277872114e-005,-.0193480678000737,-.00409728352066521,.036923084732338,.0332523788739027,-.0258953542951368,-.0737524728656581,-.0647943816610287,-.0272342124940597,-.00441906328272248,-8.59715171504879e-005,-.0283630626928129,-.00600636252506896,.0541269432119395,.0487459169789592,-.0379609770295909,-.108116533043712,-.0949845291107984,-.0399236597855762,-.00647807162071696,-.000126028891149856,-.0348594439863812,-.00738208212111014,.0665243794538717,.0599108629733206,-.046655700294547,-.132879945598894,-.116740138678666,-.0490679232034869,-.0079618332211835,-.000154895016778896,-.0382587576596884,-.00810194479882727,.0730114947611087,.0657530621708869,-.0512053242073936,-.145837714407803,-.128124036533984,-.05385277468682,-.00873823024415192,-.000169999582091795,-.0382587576596884,-.00810194479882727,.0730114947611087,.0657530621708869,-.0512053242073936,-.145837714407803,-.128124036533984,-.05385277468682,-.00873823024415192,-.000169999582091795,-.0348594439863812,-.00738208212111014,.0665243794538717,.0599108629733206,-.046655700294547,-.132879945598894,-.116740138678666,-.0490679232034869,-.0079618332211835,-.000154895016778896,-.0283630626928129,-.00600636252506896,.0541269432119395,.0487459169789592,-.0379609770295909,-.108116533043712,-.0949845291107984,-.0399236597855762,-.00647807162071696,-.000126028891149856,-.0193480678000737,-.0040972835206652,.036923084732338,.0332523788739027,-.0258953542951368,-.0737524728656581,-.0647943816610287,-.0272342124940597,-.00441906328272248,-8.59715171504879e-005,-.00863131572555909,-.00182782839348211,.0164716603838428,.014834131431226,-.0115521085131657,-.0329015220239097,-.0289052514771828,-.0121493830289503,-.00197137671826113,-3.83525277872114e-005},
			{-.0113062148903861,-.0213785801497483,-.0228814465519833,-.0174855360916378,-.00988118279406322,-.00401765276570605,-.0010799937436432,-.000159171487958625,-8.09632053406128e-006,-2.61175882124433e-008,-.0253441560031943,-.0479224988818084,-.0512913434438016,-.039195801495227,-.0221497858175121,-.00900602186035772,-.00242092779827153,-.000356800844596912,-1.81488157315668e-005,-5.85455200083099e-008,-.0371529546538126,-.0702513994795805,-.0751899158472868,-.0574586044762535,-.0324702068581793,-.0132022672898219,-.00354892941386567,-.000523047822073078,-2.66050338314805e-005,-8.58240870115567e-008,-.0456626054707174,-.0863420411133481,-.0924116936244469,-.0706191367966456,-.0399073037160309,-.0162261636575445,-.0043617893967995,-.000642848639204797,-3.26987496607437e-005,-1.05481554875234e-007,-.0501153878844611,-.0947616728508056,-.101423206654712,-.077505551778014,-.0437988587058273,-.0178084556759703,-.00478712866332422,-.0007055359319215,-3.5887363537237e-005,-1.15767573548085e-007,-.0501153878844611,-.0947616728508056,-.101423206654712,-.077505551778014,-.0437988587058273,-.0178084556759703,-.00478712866332422,-.0007055359319215,-3.5887363537237e-005,-1.15767573548085e-007,-.0456626054707174,-.0863420411133481,-.0924116936244469,-.0706191367966455,-.0399073037160309,-.0162261636575445,-.0043617893967995,-.000642848639204797,-3.26987496607437e-005,-1.05481554875234e-007,-.0371529546538126,-.0702513994795805,-.0751899158472868,-.0574586044762535,-.0324702068581793,-.0132022672898219,-.00354892941386567,-.000523047822073078,-2.66050338314805e-005,-8.58240870115567e-008,-.0253441560031943,-.0479224988818084,-.0512913434438016,-.039195801495227,-.0221497858175121,-.00900602186035772,-.00242092779827153,-.000356800844596912,-1.81488157315668e-005,-5.85455200083099e-008,-.0113062148903861,-.0213785801497483,-.0228814465519833,-.0174855360916378,-.00988118279406322,-.00401765276570605,-.0010799937436432,-.000159171487958625,-8.09632053406128e-006,-2.61175882124433e-008},
			{.0190719517783585,.0360625774106304,.038597696000771,.0294955742829575,.0166681284221623,.00677720002251945,.0018217934825568,.000268499314073518,1.36573235432709e-005,4.40565996475352e-008,.0379740031057454,.0718038951915978,.0768515485379464,.0587283903841316,.0331877705976788,.0134940260805107,.00362735771192228,.000534606730606772,2.71929822744642e-005,8.7720725769792e-008,.0437205527152459,.0826698722318604,.0884813794780883,.0676156706608181,.0382100267353107,.0155360570480548,.00417628037843494,.000615507974819813,3.1308055979897e-005,1.00995373191444e-007,.0342772212097279,.0648137619926128,.0693700245984015,.0530111619489379,.0299569300362286,.0121803781308882,.00327423322613994,.00048256258667776,2.45457363601731e-005,7.91810837934576e-008,.0129226496311487,.0244350477648839,.0261527770090912,.0199854202945654,.0112938825616689,.00459205132170114,.00123439903524221,.000181928027204484,9.25384088109926e-006,2.98515856059849e-008,-.0129226496311487,-.0244350477648839,-.0261527770090912,-.0199854202945654,-.0112938825616689,-.00459205132170114,-.00123439903524221,-.000181928027204484,-9.25384088109926e-006,-2.98515856059849e-008,-.0342772212097279,-.0648137619926128,-.0693700245984015,-.0530111619489379,-.0299569300362286,-.0121803781308882,-.00327423322613994,-.00048256258667776,-2.45457363601731e-005,-7.91810837934576e-008,-.0437205527152459,-.0826698722318604,-.0884813794780884,-.0676156706608181,-.0382100267353107,-.0155360570480548,-.00417628037843494,-.000615507974819813,-3.1308055979897e-005,-1.00995373191444e-007,-.0379740031057454,-.0718038951915978,-.0768515485379464,-.0587283903841316,-.0331877705976788,-.0134940260805107,-.00362735771192228,-.000534606730606772,-2.71929822744642e-005,-8.7720725769792e-008,-.0190719517783585,-.0360625774106304,-.038597696000771,-.0294955742829575,-.0166681284221623,-.00677720002251945,-.0018217934825568,-.000268499314073518,-1.36573235432709e-005,-4.40565996475352e-008},
			{.0111564498581787,.00938834250199416,-.0113241539111341,-.0302961470354155,-.0312651432694344,-.018730813581855,-.00658104714974118,-.00116693908893853,-6.69192068520344e-005,-2.30174087176476e-007,.0250084407901998,.0210450287110927,-.0253843683416991,-.0679122309460246,-.0700843453060978,-.0419872314552365,-.0147521595197384,-.00261582559707332,-.000150006950558683,-5.15961179744944e-007,.0366608170547776,.0308507017275645,-.0372119034382505,-.0995551020305782,-.102739286436345,-.0615506669900124,-.0216257473168005,-.00383463745165151,-.000219900849377703,-7.56366963325805e-007,.0450577468469165,.0379168611112332,-.0457350561038325,-.122357572607023,-.126271074443454,-.075648460517151,-.0265789888566817,-.00471293706542479,-.000270267757204588,-9.29608063726004e-007,.049451546558839,.0416143183763193,-.050194903530658,-.134289253725007,-.138584380130904,-.0830253092787393,-.0291708351373767,-.00517251844641853,-.000296622878728574,-1.0202586605363e-006,.049451546558839,.0416143183763193,-.050194903530658,-.134289253725007,-.138584380130904,-.0830253092787393,-.0291708351373767,-.00517251844641853,-.000296622878728574,-1.0202586605363e-006,.0450577468469165,.0379168611112332,-.0457350561038325,-.122357572607023,-.126271074443454,-.075648460517151,-.0265789888566817,-.00471293706542479,-.000270267757204588,-9.29608063726004e-007,.0366608170547776,.0308507017275645,-.0372119034382505,-.0995551020305782,-.102739286436345,-.0615506669900124,-.0216257473168005,-.00383463745165151,-.000219900849377703,-7.56366963325805e-007,.0250084407901998,.0210450287110927,-.0253843683416991,-.0679122309460246,-.0700843453060978,-.0419872314552365,-.0147521595197384,-.00261582559707332,-.000150006950558683,-5.15961179744944e-007,.0111564498581787,.00938834250199416,-.0113241539111341,-.0302961470354155,-.0312651432694344,-.018730813581855,-.00658104714974118,-.00116693908893853,-6.69192068520345e-005,-2.30174087176476e-007},
			{-.0141462413882811,-.025273750604131,-.0243577599189909,-.0158870112137333,-.0071957937725239,-.002167521076626,-.000387881070632107,-3.23454410928798e-005,-6.92492077729253e-007,-4.31978541371621e-010,-.031710395749536,-.0566539628254887,-.0546006663821845,-.0356125276698685,-.0161301834173286,-.00485874298682688,-.000869479172304184,-7.2505954733546e-005,-1.55229910443945e-006,-9.68328627104381e-010,-.0464854657297909,-.0830511819588261,-.0800411772208778,-.0522057481598326,-.0236458445483905,-.00712261469039873,-.00127460232871753,-.000106289215076142,-2.27557383363344e-006,-1.41950947462063e-009,-.0571326695688298,-.102073533343423,-.0983741059332855,-.0641631467466999,-.0290617766661061,-.00875400483105958,-.00156654198328513,-.000130634091932606,-2.79677972190642e-006,-1.74463920044904e-009,-.0627039536356021,-.112027219286109,-.107967042743438,-.070420006785652,-.0318957316434227,-.00960765035478176,-.00171930309977544,-.00014337285664383,-3.06950729477359e-006,-1.91476744149015e-009,-.0627039536356021,-.112027219286109,-.107967042743438,-.070420006785652,-.0318957316434227,-.00960765035478176,-.00171930309977544,-.00014337285664383,-3.06950729477359e-006,-1.91476744149015e-009,-.0571326695688298,-.102073533343423,-.0983741059332855,-.0641631467466999,-.0290617766661061,-.00875400483105958,-.00156654198328513,-.000130634091932606,-2.79677972190642e-006,-1.74463920044904e-009,-.0464854657297909,-.0830511819588261,-.0800411772208778,-.0522057481598326,-.0236458445483905,-.00712261469039873,-.00127460232871753,-.000106289215076142,-2.27557383363344e-006,-1.41950947462063e-009,-.031710395749536,-.0566539628254887,-.0546006663821845,-.0356125276698685,-.0161301834173286,-.00485874298682688,-.000869479172304184,-7.2505954733546e-005,-1.55229910443945e-006,-9.68328627104381e-010,-.0141462413882811,-.025273750604131,-.0243577599189909,-.0158870112137333,-.0071957937725239,-.002167521076626,-.000387881070632107,-3.23454410928798e-005,-6.92492077729254e-007,-4.31978541371621e-010}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_left_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{.00265100998409647,.00594253790984439,.008711390565351,.0107066798372692,.0117507401837501,.0117507401837501,.0107066798372692,.008711390565351,.00594253790984439,.00265100998409647,.00527840372812152,.0118321373536874,.0173451766357429,.0213179803576284,.0233968001502014,.0233968001502014,.0213179803576284,.0173451766357429,.0118321373536874,.00527840372812152,.00607717674128413,.0136226771632526,.019969996509099,.0245439987302064,.0269374032410133,.0269374032410133,.0245439987302064,.019969996509099,.0136226771632526,.00607717674128413,.00476454936076254,.0106802748271453,.0156566178922341,.0192426678437941,.0211191138346674,.0211191138346674,.0192426678437941,.0156566178922341,.0106802748271453,.00476454936076254,.00179625418474627,.00402650637025399,.00590260762948584,.00725456281862631,.00796199046987792,.00796199046987792,.00725456281862631,.00590260762948584,.00402650637025399,.00179625418474628,-.00179625418474627,-.00402650637025399,-.00590260762948584,-.00725456281862631,-.00796199046987792,-.00796199046987792,-.00725456281862631,-.00590260762948584,-.00402650637025399,-.00179625418474628,-.00476454936076254,-.0106802748271453,-.0156566178922341,-.0192426678437941,-.0211191138346674,-.0211191138346674,-.0192426678437941,-.0156566178922341,-.0106802748271453,-.00476454936076254,-.00607717674128413,-.0136226771632526,-.019969996509099,-.0245439987302064,-.0269374032410133,-.0269374032410133,-.0245439987302064,-.019969996509099,-.0136226771632526,-.00607717674128413,-.00527840372812152,-.0118321373536874,-.0173451766357429,-.0213179803576284,-.0233968001502014,-.0233968001502014,-.0213179803576284,-.0173451766357429,-.0118321373536874,-.00527840372812152,-.00265100998409647,-.00594253790984439,-.008711390565351,-.0107066798372692,-.0117507401837501,-.0117507401837501,-.0107066798372692,-.008711390565351,-.00594253790984439,-.00265100998409647},
			{-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.002221727122666,-.00498025195339731,-.00730073926966875,-.00897292772598497,-.00984792148436088,-.00984792148436088,-.00897292772598497,-.00730073926966875,-.00498025195339731,-.002221727122666,.00309757218884701,.00694355746343096,.010178822902703,.0125102183310402,.0137301504746905,.0137301504746905,.0125102183310402,.010178822902703,.00694355746343096,.00309757218884701,.00727044463764254,.0162975217520129,.0238911521277625,.0293632704051698,.0322266255011462,.0322266255011462,.0293632704051698,.0238911521277625,.0162975217520129,.00727044463764254,.00727044463764255,.016297521752013,.0238911521277625,.0293632704051698,.0322266255011463,.0322266255011463,.0293632704051698,.0238911521277625,.016297521752013,.00727044463764255,.003097572188847,.00694355746343095,.010178822902703,.0125102183310401,.0137301504746904,.0137301504746904,.0125102183310401,.010178822902703,.00694355746343095,.003097572188847,-.00222172712266599,-.00498025195339728,-.00730073926966871,-.00897292772598493,-.00984792148436084,-.00984792148436084,-.00897292772598493,-.00730073926966871,-.00498025195339728,-.00222172712266599,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468},
			{.00352805200405569,.00790852652678675,.0115934074660505,.0142488046002341,.0156382747342036,.0156382747342036,.0142488046002341,.0115934074660505,.00790852652678675,.00352805200405569,.00299001628838423,.00670245878037739,.00982538724528936,.0120758304570704,.0132534033295837,.0132534033295837,.0120758304570704,.00982538724528936,.00670245878037739,.00299001628838423,-.0032119933533482,-.00720004541021951,-.0105548182625439,-.0129723330655266,-.0142373282611347,-.0142373282611347,-.0129723330655266,-.0105548182625439,-.00720004541021951,-.0032119933533482,-.00749937306045005,-.0168106906345655,-.0246434257572659,-.0302878475826814,-.0332413626893215,-.0332413626893215,-.0302878475826814,-.0246434257572659,-.0168106906345655,-.00749937306045005,-.00396370307706834,-.0088850875478213,-.0130249851175851,-.0160082761176901,-.0175693208639632,-.0175693208639632,-.0160082761176901,-.0130249851175851,-.0088850875478213,-.00396370307706834,.00396370307706838,.00888508754782138,.0130249851175852,.0160082761176902,.0175693208639633,.0175693208639633,.0160082761176902,.0130249851175852,.00888508754782138,.00396370307706837,.00749937306045007,.0168106906345656,.024643425757266,.0302878475826815,.0332413626893216,.0332413626893216,.0302878475826815,.024643425757266,.0168106906345656,.00749937306045007,.00321199335334816,.00720004541021943,.0105548182625438,.0129723330655265,.0142373282611345,.0142373282611345,.0129723330655265,.0105548182625438,.00720004541021943,.00321199335334816,-.00299001628838425,-.00670245878037744,-.00982538724528943,-.0120758304570705,-.0132534033295838,-.0132534033295838,-.0120758304570705,-.00982538724528943,-.00670245878037744,-.00299001628838425,-.0035280520040557,-.00790852652678675,-.0115934074660505,-.0142488046002341,-.0156382747342036,-.0156382747342036,-.0142488046002341,-.0115934074660505,-.00790852652678675,-.0035280520040557},
			{-.00355524726055973,-.00796948776183025,-.0116827728408898,-.0143586385526627,-.0157588191287273,-.0157588191287273,-.0143586385526627,-.0116827728408898,-.00796948776183025,-.00355524726055973,-.000198327000917931,-.000444572343586338,-.000651715374523445,-.000800986686075343,-.000879094787718522,-.000879094787718522,-.000800986686075343,-.000651715374523445,-.000444572343586338,-.000198327000917931,.00656585578393543,.0147181063597184,.0215758275045217,.0265176352250662,.0291034986151868,.0291034986151868,.0265176352250662,.0215758275045218,.0147181063597184,.00656585578393543,.00333253137587672,.00747024498425,.0109509140141708,.0134591520602391,.0147716193401924,.0147716193401924,.0134591520602391,.0109509140141708,.00747024498425,.00333253137587672,-.00614481295848751,-.0137742913733916,-.020192253500946,-.0248171622895082,-.0272372043055647,-.0272372043055647,-.0248171622895082,-.020192253500946,-.0137742913733916,-.00614481295848751,-.00614481295848755,-.0137742913733917,-.0201922535009461,-.0248171622895084,-.0272372043055649,-.0272372043055649,-.0248171622895084,-.0201922535009461,-.0137742913733917,-.00614481295848755,.00333253137587662,.00747024498424977,.0109509140141705,.0134591520602387,.014771619340192,.014771619340192,.0134591520602387,.0109509140141704,.00747024498424977,.00333253137587662,.00656585578393523,.014718106359718,.0215758275045211,.0265176352250655,.0291034986151859,.0291034986151859,.0265176352250655,.0215758275045211,.014718106359718,.00656585578393523,-.000198327000917742,-.000444572343585915,-.000651715374522825,-.000800986686074581,-.000879094787717686,-.000879094787717686,-.000800986686074581,-.000651715374522825,-.000444572343585915,-.000198327000917742,-.0035552472605598,-.00796948776183042,-.0116827728408901,-.014358638552663,-.0157588191287277,-.0157588191287277,-.014358638552663,-.0116827728408901,-.00796948776183042,-.0035552472605598},
			{.00222253422114756,.00498206115568338,.00730339144757066,.00897618736857064,.00985149899052509,.00985149899052509,.00897618736857064,.00730339144757066,.00498206115568338,.00222253422114756,.00498206115568338,.011167852050509,.0163713757428243,.0201211364889587,.0220832462235868,.0220832462235868,.0201211364889587,.0163713757428243,.011167852050509,.00498206115568339,.00730339144757066,.0163713757428243,.0239994174797936,.02949633325581,.0323726639565543,.0323726639565543,.02949633325581,.0239994174797936,.0163713757428243,.00730339144757066,.00897618736857064,.0201211364889587,.02949633325581,.0362522830510504,.0397874192256888,.0397874192256888,.0362522830510504,.02949633325581,.0201211364889587,.00897618736857064,.00985149899052509,.0220832462235868,.0323726639565543,.0397874192256888,.0436672836966291,.0436672836966291,.0397874192256889,.0323726639565543,.0220832462235868,.0098514989905251,.00985149899052509,.0220832462235868,.0323726639565543,.0397874192256888,.0436672836966291,.0436672836966291,.0397874192256889,.0323726639565543,.0220832462235868,.0098514989905251,.00897618736857064,.0201211364889587,.02949633325581,.0362522830510504,.0397874192256889,.0397874192256889,.0362522830510504,.02949633325581,.0201211364889587,.00897618736857064,.00730339144757066,.0163713757428243,.0239994174797936,.02949633325581,.0323726639565543,.0323726639565543,.02949633325581,.0239994174797936,.0163713757428243,.00730339144757066,.00498206115568338,.011167852050509,.0163713757428243,.0201211364889587,.0220832462235868,.0220832462235868,.0201211364889587,.0163713757428243,.011167852050509,.00498206115568339,.00222253422114756,.00498206115568339,.00730339144757066,.00897618736857064,.0098514989905251,.0098514989905251,.00897618736857064,.00730339144757066,.00498206115568339,.00222253422114756},
			{-.00374909427409703,-.00840401770836613,-.0123197666866244,-.0151415318362812,-.0166180561384473,-.0166180561384473,-.0151415318362812,-.0123197666866244,-.00840401770836613,-.00374909427409703,-.00746479014118745,-.0167331691201299,-.024529784043959,-.0301481769489968,-.0330880720938548,-.0330880720938548,-.0301481769489968,-.024529784043959,-.0167331691201299,-.00746479014118745,-.00859442576984082,-.019265374803192,-.0282418399082409,-.0347104558846931,-.0380952410046641,-.0380952410046641,-.0347104558846931,-.0282418399082409,-.019265374803192,-.00859442576984082,-.00673809032566718,-.0151041895128434,-.022141801367642,-.027213241845299,-.0298669372150783,-.0298669372150783,-.027213241845299,-.022141801367642,-.0151041895128434,-.00673809032566718,-.00254028702994505,-.00569433991870819,-.00834754776432466,-.0102595011288345,-.0112599549077927,-.0112599549077927,-.0102595011288345,-.00834754776432466,-.00569433991870818,-.00254028702994505,.00254028702994505,.00569433991870819,.00834754776432466,.0102595011288345,.0112599549077927,.0112599549077927,.0102595011288345,.00834754776432466,.00569433991870818,.00254028702994505,.00673809032566718,.0151041895128434,.022141801367642,.027213241845299,.0298669372150783,.0298669372150783,.027213241845299,.022141801367642,.0151041895128434,.00673809032566718,.00859442576984082,.019265374803192,.0282418399082409,.0347104558846931,.0380952410046641,.0380952410046641,.0347104558846931,.0282418399082409,.019265374803192,.00859442576984082,.00746479014118745,.0167331691201299,.024529784043959,.0301481769489968,.0330880720938548,.0330880720938548,.0301481769489968,.024529784043959,.0167331691201299,.00746479014118745,.00374909427409703,.00840401770836614,.0123197666866244,.0151415318362812,.0166180561384473,.0166180561384473,.0151415318362812,.0123197666866244,.00840401770836614,.00374909427409703},
			{.00458578009298867,.0102795433484883,.0150691704959577,.0185206693125625,.0203267097203668,.0203267097203668,.0185206693125625,.0150691704959577,.0102795433484883,.00458578009298867,.00693481320331858,.0155451660331847,.0227882454891813,.0280077499310155,.0307389216426545,.0307389216426545,.0280077499310155,.0227882454891813,.0155451660331847,.00693481320331858,.00314199662823143,.00704313985533036,.0103248044887574,.0126896360821209,.0139270641219972,.0139270641219972,.0126896360821209,.0103248044887574,.00704313985533036,.00314199662823143,-.00438062859915128,-.00981967313422903,-.0143950293955455,-.0176921204289931,-.0194173650114245,-.0194173650114245,-.0176921204289931,-.0143950293955455,-.00981967313422903,-.00438062859915128,-.0102819614092862,-.0230481762908429,-.0337871913540478,-.0415259352355491,-.0455753308454797,-.0455753308454797,-.0415259352355491,-.0337871913540478,-.0230481762908429,-.0102819614092862,-.0102819614092862,-.0230481762908429,-.0337871913540478,-.0415259352355491,-.0455753308454797,-.0455753308454797,-.0415259352355492,-.0337871913540478,-.0230481762908429,-.0102819614092862,-.00438062859915127,-.009819673134229,-.0143950293955455,-.0176921204289931,-.0194173650114245,-.0194173650114245,-.0176921204289931,-.0143950293955455,-.009819673134229,-.00438062859915127,.00314199662823142,.00704313985533033,.0103248044887574,.0126896360821208,.0139270641219972,.0139270641219972,.0126896360821208,.0103248044887574,.00704313985533033,.00314199662823142,.00693481320331858,.0155451660331847,.0227882454891813,.0280077499310155,.0307389216426545,.0307389216426545,.0280077499310155,.0227882454891813,.0155451660331847,.00693481320331858,.00458578009298867,.0102795433484883,.0150691704959577,.0185206693125625,.0203267097203668,.0203267097203668,.0185206693125625,.0150691704959577,.0102795433484883,.00458578009298867},
			{-.00498941899246989,-.0111843454716204,-.0163955540712153,-.0201508527115458,-.0221158602193511,-.0221158602193511,-.0201508527115458,-.0163955540712153,-.0111843454716204,-.00498941899246989,-.00422852158639074,-.00947870810765226,-.0138951958966771,-.0170778032078584,-.0187431427347081,-.0187431427347081,-.0170778032078584,-.0138951958966771,-.00947870810765226,-.00422852158639074,.00454244456217192,.0101824018679708,.0149267671340066,.0183456493553325,.0201346227171464,.0201346227171464,.0183456493553325,.0149267671340066,.0101824018679708,.00454244456217192,.0106057150904842,.0237739066862442,.0348510669263035,.0428334848228836,.0470103859430136,.0470103859430136,.0428334848228836,.0348510669263035,.0237739066862442,.0106057150904842,.0056055226483345,.0125654113119353,.0184201106014339,.0226391211939302,.0248467718453936,.0248467718453936,.0226391211939302,.0184201106014339,.0125654113119353,.0056055226483345,-.00560552264833455,-.0125654113119354,-.0184201106014341,-.0226391211939304,-.0248467718453938,-.0248467718453938,-.0226391211939304,-.0184201106014341,-.0125654113119354,-.00560552264833455,-.0106057150904842,-.0237739066862443,-.0348510669263036,-.0428334848228837,-.0470103859430138,-.0470103859430138,-.0428334848228837,-.0348510669263036,-.0237739066862442,-.0106057150904842,-.00454244456217188,-.0101824018679707,-.0149267671340064,-.0183456493553323,-.0201346227171462,-.0201346227171462,-.0183456493553323,-.0149267671340064,-.0101824018679707,-.00454244456217188,.00422852158639077,.00947870810765234,.0138951958966772,.0170778032078585,.0187431427347082,.0187431427347082,.0170778032078585,.0138951958966772,.00947870810765234,.00422852158639077,.00498941899246989,.0111843454716204,.0163955540712153,.0201508527115458,.0221158602193511,.0221158602193511,.0201508527115458,.0163955540712153,.0111843454716204,.00498941899246989},
			{-.00272203738908372,-.00610175384992259,-.00894479121996913,-.01099353944528,-.012065572865233,-.012065572865233,-.01099353944528,-.00894479121996913,-.00610175384992259,-.00272203738908372,-.00610175384992259,-.0136777695245317,-.0200507584804236,-.0246432587236037,-.0270463425584096,-.0270463425584096,-.0246432587236037,-.0200507584804236,-.0136777695245317,-.00610175384992259,-.00894479121996913,-.0200507584804236,-.0293931634773647,-.0361254828831082,-.0396482541575827,-.0396482541575827,-.0361254828831082,-.0293931634773647,-.0200507584804236,-.00894479121996913,-.01099353944528,-.0246432587236037,-.0361254828831082,-.0443997977469403,-.0487294376468829,-.0487294376468829,-.0443997977469403,-.0361254828831082,-.0246432587236037,-.01099353944528,-.012065572865233,-.0270463425584096,-.0396482541575827,-.0487294376468829,-.053481281759782,-.053481281759782,-.0487294376468829,-.0396482541575827,-.0270463425584096,-.012065572865233,-.012065572865233,-.0270463425584096,-.0396482541575827,-.0487294376468829,-.053481281759782,-.053481281759782,-.0487294376468829,-.0396482541575827,-.0270463425584096,-.012065572865233,-.01099353944528,-.0246432587236037,-.0361254828831082,-.0443997977469403,-.0487294376468829,-.0487294376468829,-.0443997977469403,-.0361254828831082,-.0246432587236037,-.01099353944528,-.00894479121996913,-.0200507584804236,-.0293931634773647,-.0361254828831082,-.0396482541575827,-.0396482541575827,-.0361254828831082,-.0293931634773647,-.0200507584804236,-.00894479121996913,-.00610175384992259,-.0136777695245317,-.0200507584804236,-.0246432587236037,-.0270463425584096,-.0270463425584096,-.0246432587236037,-.0200507584804236,-.0136777695245317,-.00610175384992259,-.00272203738908372,-.00610175384992259,-.00894479121996913,-.01099353944528,-.012065572865233,-.012065572865233,-.01099353944528,-.00894479121996913,-.00610175384992259,-.00272203738908372},
			{.00459168398220963,.0102927775821281,.0150885710584479,.0185445134519897,.0203528790176253,.0203528790176253,.0185445134519897,.0150885710584479,.0102927775821281,.00459168398220963,.00914246343674623,.0204938630514994,.0300427271887783,.0369238250811581,.0405244465804058,.0405244465804058,.0369238250811581,.0300427271887783,.0204938630514994,.00914246343674623,.0105259788787713,.0235951689735483,.0345890485685456,.0425114528066451,.0466569510209666,.0466569510209666,.0425114528066451,.0345890485685456,.0235951689735483,.0105259788787713,.00825244156510291,.0184987786328967,.0271180576544867,.0333292783668801,.0365793781595865,.0365793781595865,.0333292783668801,.0271180576544867,.0184987786328967,.00825244156510291,.00311120351019257,.00697411360782243,.010223616307811,.0125652713841337,.0137905720183488,.0137905720183489,.0125652713841337,.010223616307811,.00697411360782243,.00311120351019257,-.00311120351019257,-.00697411360782243,-.010223616307811,-.0125652713841337,-.0137905720183489,-.0137905720183488,-.0125652713841337,-.010223616307811,-.00697411360782243,-.00311120351019257,-.00825244156510291,-.0184987786328967,-.0271180576544867,-.0333292783668801,-.0365793781595865,-.0365793781595865,-.0333292783668801,-.0271180576544867,-.0184987786328967,-.00825244156510291,-.0105259788787713,-.0235951689735483,-.0345890485685456,-.0425114528066451,-.0466569510209666,-.0466569510209666,-.0425114528066451,-.0345890485685456,-.0235951689735483,-.0105259788787713,-.00914246343674623,-.0204938630514994,-.0300427271887783,-.0369238250811581,-.0405244465804058,-.0405244465804058,-.0369238250811581,-.0300427271887783,-.0204938630514994,-.00914246343674623,-.00459168398220963,-.0102927775821281,-.0150885710584479,-.0185445134519897,-.0203528790176254,-.0203528790176254,-.0185445134519897,-.0150885710584479,-.0102927775821281,-.00459168398220963},
			{-.00561641065141988,-.0125898179990038,-.0184558892850007,-.0226830947601565,-.0248950334876143,-.0248950334876143,-.0226830947601565,-.0184558892850007,-.0125898179990038,-.00561641065141988,-.00849337690664137,-.0190388623781496,-.0279097867968623,-.0343023480945728,-.0376473366420087,-.0376473366420087,-.0343023480945728,-.0279097867968623,-.0190388623781496,-.00849337690664138,-.00384814425717993,-.00862604941815623,-.0126452513484336,-.0155415667147304,-.0170571003606095,-.0170571003606095,-.0155415667147304,-.0126452513484336,-.00862604941815623,-.00384814425717993,.00536515241143043,.0120265943124634,.0176302384294998,.0216683337640905,.0237813182187718,.0237813182187718,.0216683337640905,.0176302384294998,.0120265943124634,.00536515241143043,.0125927795065653,.0282281357131822,.041380689338455,.050858676220366,.0558181527269258,.0558181527269258,.050858676220366,.041380689338455,.0282281357131822,.0125927795065653,.0125927795065653,.0282281357131822,.0413806893384551,.050858676220366,.0558181527269258,.0558181527269258,.050858676220366,.0413806893384551,.0282281357131822,.0125927795065653,.00536515241143042,.0120265943124634,.0176302384294998,.0216683337640904,.0237813182187717,.0237813182187717,.0216683337640904,.0176302384294998,.0120265943124634,.00536515241143042,-.00384814425717991,-.00862604941815619,-.0126452513484336,-.0155415667147303,-.0170571003606094,-.0170571003606094,-.0155415667147303,-.0126452513484336,-.00862604941815619,-.00384814425717991,-.00849337690664137,-.0190388623781496,-.0279097867968623,-.0343023480945728,-.0376473366420086,-.0376473366420086,-.0343023480945728,-.0279097867968623,-.0190388623781496,-.00849337690664138,-.00561641065141988,-.0125898179990038,-.0184558892850007,-.0226830947601565,-.0248950334876143,-.0248950334876143,-.0226830947601565,-.0184558892850007,-.0125898179990038,-.00561641065141988},
			{.00314313803755599,.00704569845308084,.0103285552337492,.0126942459116857,.0139321234784299,.0139321234784299,.0126942459116857,.0103285552337492,.00704569845308084,.00314313803755599,.00704569845308084,.0157937278282394,.0231526216040999,.0284555841055385,.0312304263023817,.0312304263023817,.0284555841055385,.0231526216040999,.0157937278282394,.00704569845308084,.0103285552337492,.0231526216040999,.033940301680024,.0417141145196382,.0457818604054276,.0457818604054276,.0417141145196382,.033940301680024,.0231526216040999,.0103285552337492,.0126942459116857,.0284555841055385,.0417141145196382,.0512684703442582,.0562679078659487,.0562679078659487,.0512684703442582,.0417141145196382,.0284555841055385,.0126942459116857,.0139321234784299,.0312304263023817,.0457818604054276,.0562679078659487,.0617548648194743,.0617548648194743,.0562679078659487,.0457818604054276,.0312304263023817,.0139321234784299,.0139321234784299,.0312304263023817,.0457818604054276,.0562679078659487,.0617548648194743,.0617548648194743,.0562679078659487,.0457818604054276,.0312304263023817,.0139321234784299,.0126942459116857,.0284555841055385,.0417141145196382,.0512684703442582,.0562679078659487,.0562679078659487,.0512684703442582,.0417141145196382,.0284555841055385,.0126942459116857,.0103285552337492,.0231526216040999,.033940301680024,.0417141145196382,.0457818604054276,.0457818604054276,.0417141145196382,.033940301680024,.0231526216040999,.0103285552337492,.00704569845308084,.0157937278282394,.0231526216040999,.0284555841055385,.0312304263023817,.0312304263023817,.0284555841055385,.0231526216040999,.0157937278282394,.00704569845308084,.00314313803755599,.00704569845308084,.0103285552337492,.0126942459116857,.0139321234784299,.0139321234784299,.0126942459116857,.0103285552337492,.00704569845308084,.00314313803755599},
			{-.00530201996819294,-.0118850758196888,-.017422781130702,-.0214133596745384,-.0235014803675002,-.0235014803675002,-.0214133596745384,-.017422781130702,-.0118850758196888,-.00530201996819294,-.010556807456243,-.0236642747073748,-.0346903532714859,-.0426359607152567,-.0467936003004029,-.0467936003004029,-.0426359607152567,-.0346903532714859,-.0236642747073748,-.010556807456243,-.0121543534825683,-.0272453543265051,-.039939993018198,-.0490879974604127,-.0538748064820265,-.0538748064820265,-.0490879974604127,-.039939993018198,-.0272453543265051,-.0121543534825683,-.00952909872152508,-.0213605496542905,-.0313132357844681,-.0384853356875881,-.0422382276693348,-.0422382276693348,-.0384853356875881,-.0313132357844681,-.0213605496542905,-.00952909872152508,-.00359250836949255,-.00805301274050799,-.0118052152589717,-.0145091256372526,-.0159239809397558,-.0159239809397558,-.0145091256372526,-.0118052152589717,-.00805301274050799,-.00359250836949255,.00359250836949255,.00805301274050799,.0118052152589717,.0145091256372526,.0159239809397558,.0159239809397558,.0145091256372526,.0118052152589717,.00805301274050799,.00359250836949255,.00952909872152508,.0213605496542905,.0313132357844681,.0384853356875881,.0422382276693348,.0422382276693348,.0384853356875881,.0313132357844681,.0213605496542905,.00952909872152508,.0121543534825683,.0272453543265051,.039939993018198,.0490879974604127,.0538748064820265,.0538748064820265,.0490879974604127,.039939993018198,.0272453543265051,.0121543534825683,.010556807456243,.0236642747073748,.0346903532714859,.0426359607152567,.0467936003004029,.0467936003004029,.0426359607152567,.0346903532714859,.0236642747073748,.010556807456243,.00530201996819294,.0118850758196888,.017422781130702,.0214133596745384,.0235014803675002,.0235014803675002,.0214133596745384,.017422781130702,.0118850758196888,.00530201996819294},
			{-.00351413515806021,-.00787733034668568,-.0115476758084439,-.0141925983938026,-.015576587587625,-.015576587587625,-.0141925983938026,-.0115476758084439,-.00787733034668568,-.00351413515806021,-.00787733034668568,-.017657924524755,-.0258854178874996,-.0318143102064225,-.0349166780965639,-.0349166780965639,-.0318143102064225,-.0258854178874996,-.017657924524755,-.00787733034668568,-.0115476758084439,-.0258854178874996,-.0379464108746825,-.0466377978533816,-.0511856760122494,-.0511856760122494,-.0466377978533816,-.0379464108746825,-.0258854178874996,-.0115476758084439,-.0141925983938026,-.0318143102064225,-.0466377978533816,-.0573198924081667,-.0629094334832252,-.0629094334832252,-.0573198924081666,-.0466377978533816,-.0318143102064225,-.0141925983938026,-.015576587587625,-.0349166780965639,-.0511856760122494,-.0629094334832252,-.0690440378533661,-.0690440378533661,-.0629094334832252,-.0511856760122494,-.0349166780965639,-.015576587587625,-.015576587587625,-.0349166780965639,-.0511856760122494,-.0629094334832252,-.0690440378533661,-.0690440378533661,-.0629094334832252,-.0511856760122494,-.0349166780965639,-.015576587587625,-.0141925983938026,-.0318143102064225,-.0466377978533816,-.0573198924081666,-.0629094334832252,-.0629094334832252,-.0573198924081667,-.0466377978533816,-.0318143102064225,-.0141925983938026,-.0115476758084439,-.0258854178874996,-.0379464108746825,-.0466377978533816,-.0511856760122494,-.0511856760122494,-.0466377978533816,-.0379464108746825,-.0258854178874996,-.0115476758084439,-.00787733034668568,-.017657924524755,-.0258854178874996,-.0318143102064225,-.0349166780965639,-.0349166780965639,-.0318143102064225,-.0258854178874996,-.017657924524755,-.00787733034668568,-.00351413515806021,-.00787733034668568,-.0115476758084439,-.0141925983938026,-.015576587587625,-.015576587587625,-.0141925983938026,-.0115476758084439,-.00787733034668568,-.00351413515806021},
			{.00374909427409703,.00746479014118745,.00859442576984082,.00673809032566718,.00254028702994505,-.00254028702994505,-.00673809032566718,-.00859442576984082,-.00746479014118745,-.00374909427409703,.00840401770836613,.0167331691201299,.019265374803192,.0151041895128434,.00569433991870819,-.00569433991870819,-.0151041895128434,-.019265374803192,-.0167331691201299,-.00840401770836614,.0123197666866244,.024529784043959,.0282418399082409,.022141801367642,.00834754776432466,-.00834754776432466,-.022141801367642,-.0282418399082409,-.024529784043959,-.0123197666866244,.0151415318362812,.0301481769489968,.0347104558846931,.027213241845299,.0102595011288345,-.0102595011288345,-.027213241845299,-.0347104558846931,-.0301481769489968,-.0151415318362812,.0166180561384473,.0330880720938548,.0380952410046641,.0298669372150783,.0112599549077927,-.0112599549077927,-.0298669372150783,-.0380952410046641,-.0330880720938548,-.0166180561384473,.0166180561384473,.0330880720938548,.0380952410046641,.0298669372150783,.0112599549077927,-.0112599549077927,-.0298669372150783,-.0380952410046641,-.0330880720938548,-.0166180561384473,.0151415318362812,.0301481769489968,.0347104558846931,.027213241845299,.0102595011288345,-.0102595011288345,-.027213241845299,-.0347104558846931,-.0301481769489968,-.0151415318362812,.0123197666866244,.024529784043959,.0282418399082409,.022141801367642,.00834754776432466,-.00834754776432466,-.022141801367642,-.0282418399082409,-.024529784043959,-.0123197666866244,.00840401770836613,.0167331691201299,.019265374803192,.0151041895128434,.00569433991870818,-.00569433991870818,-.0151041895128434,-.019265374803192,-.0167331691201299,-.00840401770836614,.00374909427409703,.00746479014118745,.00859442576984082,.00673809032566718,.00254028702994505,-.00254028702994505,-.00673809032566718,-.00859442576984082,-.00746479014118745,-.00374909427409703},
			{-.00632418062917951,-.0125920229688431,-.0144975551423388,-.0113661853236565,-.00428509737360599,.00428509737360599,.0113661853236565,.0144975551423388,.0125920229688431,.00632418062917952,-.0125920229688431,-.0250718712423055,-.0288659603588969,-.0226311161960246,-.00853202141368567,.00853202141368567,.0226311161960246,.0288659603588969,.0250718712423055,.0125920229688431,-.0144975551423388,-.0288659603588969,-.0332342033583603,-.0260558494688553,-.00982315956908459,.00982315956908459,.0260558494688553,.0332342033583603,.0288659603588969,.0144975551423388,-.0113661853236565,-.0226311161960246,-.0260558494688553,-.0204279694693769,-.0077014262770414,.0077014262770414,.0204279694693769,.0260558494688553,.0226311161960246,.0113661853236565,-.00428509737360599,-.00853202141368567,-.00982315956908459,-.0077014262770414,-.00290346854050359,.00290346854050359,.0077014262770414,.00982315956908459,.00853202141368567,.00428509737360599,.00428509737360599,.00853202141368567,.00982315956908459,.0077014262770414,.00290346854050359,-.00290346854050359,-.0077014262770414,-.00982315956908459,-.00853202141368567,-.00428509737360599,.0113661853236565,.0226311161960246,.0260558494688553,.0204279694693769,.0077014262770414,-.0077014262770414,-.0204279694693769,-.0260558494688553,-.0226311161960246,-.0113661853236565,.0144975551423388,.0288659603588969,.0332342033583603,.0260558494688553,.00982315956908459,-.00982315956908459,-.0260558494688553,-.0332342033583603,-.0288659603588969,-.0144975551423388,.0125920229688431,.0250718712423055,.0288659603588969,.0226311161960246,.00853202141368567,-.00853202141368567,-.0226311161960246,-.0288659603588969,-.0250718712423055,-.0125920229688431,.00632418062917952,.0125920229688431,.0144975551423388,.0113661853236565,.00428509737360599,-.00428509737360599,-.0113661853236565,-.0144975551423388,-.0125920229688431,-.00632418062917951},
			{.00773554878208247,.0154021862517905,.0177329762700093,.0139027782716481,.00524140307070943,-.00524140307070943,-.0139027782716481,-.0177329762700093,-.0154021862517905,-.00773554878208247,.0116980284141665,.023291846188219,.0268165667515141,.0210243771759672,.00792627437025251,-.00792627437025251,-.0210243771759672,-.0268165667515141,-.023291846188219,-.0116980284141665,.00530009457452694,.0105529738210754,.0121499397090725,.00952563829202252,.00359120377371563,-.00359120377371563,-.00952563829202252,-.0121499397090725,-.0105529738210754,-.00530009457452694,-.0073894878380083,-.0147131472107018,-.0169396659720433,-.0132808174115363,-.00500692133631402,.00500692133631402,.0132808174115363,.0169396659720433,.0147131472107018,.0073894878380083,-.0173441840742927,-.0345338593321725,-.0397598170829848,-.0311719765824586,-.0117519599742572,.0117519599742572,.0311719765824586,.0397598170829848,.0345338593321725,.0173441840742927,-.0173441840742927,-.0345338593321725,-.0397598170829848,-.0311719765824586,-.0117519599742572,.0117519599742572,.0311719765824586,.0397598170829848,.0345338593321725,.0173441840742927,-.00738948783800828,-.0147131472107018,-.0169396659720432,-.0132808174115363,-.00500692133631401,.00500692133631401,.0132808174115363,.0169396659720432,.0147131472107018,.00738948783800828,.00530009457452691,.0105529738210754,.0121499397090724,.00952563829202247,.00359120377371561,-.00359120377371561,-.00952563829202247,-.0121499397090724,-.0105529738210754,-.00530009457452691,.0116980284141665,.023291846188219,.0268165667515141,.0210243771759672,.00792627437025251,-.00792627437025251,-.0210243771759672,-.0268165667515141,-.023291846188219,-.0116980284141665,.00773554878208247,.0154021862517905,.0177329762700093,.0139027782716481,.00524140307070943,-.00524140307070943,-.0139027782716481,-.0177329762700093,-.0154021862517905,-.00773554878208247},
			{-.00841642931466872,-.0167578817652623,-.0192938271762863,-.0151264963737114,-.005702749694565,.005702749694565,.0151264963737114,.0192938271762863,.0167578817652623,.00841642931466872,-.00713290527236138,-.0142022678178642,-.0163514759578514,-.0128196722983645,-.00483306778237002,.00483306778237002,.0128196722983645,.0163514759578514,.0142022678178642,.00713290527236138,.00766244799865884,.015256635895486,.0175653999940025,.0137713972911834,.00519187191506662,-.00519187191506662,-.0137713972911834,-.0175653999940025,-.015256635895486,-.00766244799865884,.0178903097786119,.0356212456381444,.0410117559470465,.0321535054680605,.0121220002938583,-.0121220002938583,-.0321535054680605,-.0410117559470465,-.0356212456381444,-.0178903097786118,.00945570721013499,.0188271792597616,.0216762683937647,.0169943470654115,.00640693688361546,-.00640693688361546,-.0169943470654115,-.0216762683937647,-.0188271792597616,-.00945570721013499,-.00945570721013506,-.0188271792597618,-.0216762683937649,-.0169943470654116,-.00640693688361551,.00640693688361551,.0169943470654116,.0216762683937649,.0188271792597618,.00945570721013506,-.0178903097786119,-.0356212456381445,-.0410117559470466,-.0321535054680606,-.0121220002938583,.0121220002938583,.0321535054680606,.0410117559470466,.0356212456381445,.0178903097786119,-.00766244799865876,-.0152566358954859,-.0175653999940023,-.0137713972911832,-.00519187191506656,.00519187191506656,.0137713972911832,.0175653999940023,.0152566358954859,.00766244799865876,.00713290527236143,.0142022678178643,.0163514759578515,.0128196722983646,.00483306778237005,-.00483306778237005,-.0128196722983646,-.0163514759578515,-.0142022678178643,-.00713290527236143,.00841642931466873,.0167578817652623,.0192938271762863,.0151264963737114,.00570274969456501,-.00570274969456501,-.0151264963737114,-.0192938271762863,-.0167578817652623,-.00841642931466873},
			{-.00459168398220963,-.00914246343674623,-.0105259788787713,-.00825244156510291,-.00311120351019257,.00311120351019257,.00825244156510291,.0105259788787713,.00914246343674623,.00459168398220963,-.0102927775821281,-.0204938630514994,-.0235951689735483,-.0184987786328967,-.00697411360782243,.00697411360782243,.0184987786328967,.0235951689735483,.0204938630514994,.0102927775821281,-.0150885710584479,-.0300427271887783,-.0345890485685456,-.0271180576544867,-.010223616307811,.010223616307811,.0271180576544867,.0345890485685456,.0300427271887783,.0150885710584479,-.0185445134519897,-.0369238250811581,-.0425114528066451,-.0333292783668801,-.0125652713841337,.0125652713841337,.0333292783668801,.0425114528066451,.0369238250811581,.0185445134519897,-.0203528790176253,-.0405244465804058,-.0466569510209666,-.0365793781595865,-.0137905720183488,.0137905720183489,.0365793781595865,.0466569510209666,.0405244465804058,.0203528790176254,-.0203528790176253,-.0405244465804058,-.0466569510209666,-.0365793781595865,-.0137905720183489,.0137905720183488,.0365793781595865,.0466569510209666,.0405244465804058,.0203528790176254,-.0185445134519897,-.0369238250811581,-.0425114528066451,-.0333292783668801,-.0125652713841337,.0125652713841337,.0333292783668801,.0425114528066451,.0369238250811581,.0185445134519897,-.0150885710584479,-.0300427271887783,-.0345890485685456,-.0271180576544867,-.010223616307811,.010223616307811,.0271180576544867,.0345890485685456,.0300427271887783,.0150885710584479,-.0102927775821281,-.0204938630514994,-.0235951689735483,-.0184987786328967,-.00697411360782243,.00697411360782243,.0184987786328967,.0235951689735483,.0204938630514994,.0102927775821281,-.00459168398220963,-.00914246343674623,-.0105259788787713,-.00825244156510291,-.00311120351019257,.00311120351019257,.00825244156510291,.0105259788787713,.00914246343674623,.00459168398220963},
			{.00774550779202725,.0154220155529008,.0177558063098679,.0139206771836668,.00524815103220206,-.00524815103220206,-.0139206771836668,-.0177558063098679,-.0154220155529008,-.00774550779202725,.0154220155529008,.030706645722922,.0353534369104813,.0277173434974017,.0104495494699397,-.0104495494699397,-.0277173434974017,-.0353534369104813,-.030706645722922,-.0154220155529008,.0177558063098679,.0353534369104813,.0407034201215399,.0319117680095566,.0120308643041625,-.0120308643041625,-.0319117680095566,-.0407034201215399,-.0353534369104813,-.0177558063098679,.0139206771836668,.0277173434974017,.0319117680095566,.025019050842778,.0094322823360418,-.0094322823360418,-.025019050842778,-.0319117680095566,-.0277173434974017,-.0139206771836669,.00524815103220206,.0104495494699397,.0120308643041625,.0094322823360418,.00355600820454337,-.00355600820454337,-.0094322823360418,-.0120308643041625,-.0104495494699397,-.00524815103220206,-.00524815103220206,-.0104495494699397,-.0120308643041625,-.0094322823360418,-.00355600820454337,.00355600820454337,.0094322823360418,.0120308643041625,.0104495494699397,.00524815103220206,-.0139206771836668,-.0277173434974017,-.0319117680095566,-.025019050842778,-.0094322823360418,.0094322823360418,.025019050842778,.0319117680095566,.0277173434974017,.0139206771836669,-.0177558063098679,-.0353534369104813,-.0407034201215399,-.0319117680095566,-.0120308643041625,.0120308643041625,.0319117680095566,.0407034201215399,.0353534369104813,.0177558063098679,-.0154220155529008,-.030706645722922,-.0353534369104813,-.0277173434974017,-.0104495494699397,.0104495494699397,.0277173434974017,.0353534369104813,.030706645722922,.0154220155529008,-.00774550779202725,-.0154220155529008,-.0177558063098679,-.0139206771836669,-.00524815103220206,.00524815103220206,.0139206771836669,.0177558063098679,.0154220155529008,.00774550779202725},
			{-.00947407369825788,-.0188637486201043,-.0217183717412093,-.0170273563863006,-.00641938152974948,.00641938152974948,.0170273563863006,.0217183717412093,.0188637486201043,.00947407369825788,-.0143271003056479,-.0285265691642717,-.0328434525972573,-.0257494981204761,-.00970766388421236,.00970766388421236,.0257494981204761,.0328434525972573,.0285265691642717,.0143271003056479,-.00649126364804426,-.0129247005652958,-.0148805763464081,-.0116664766448895,-.00439830840398169,.00439830840398169,.0116664766448895,.0148805763464081,.0129247005652958,.00649126364804426,.00905023733181398,.0180198515883421,.020746769022353,.0162656130126723,.00613220122811362,-.00613220122811362,-.0162656130126722,-.020746769022353,-.0180198515883421,-.00905023733181398,.0212422004934682,.0422951671064499,.0486956320598683,.0381777184505166,.0143931527072752,-.0143931527072752,-.0381777184505166,-.0486956320598683,-.0422951671064499,-.0212422004934682,.0212422004934682,.0422951671064499,.0486956320598684,.0381777184505167,.0143931527072752,-.0143931527072752,-.0381777184505167,-.0486956320598684,-.0422951671064499,-.0212422004934682,.00905023733181396,.018019851588342,.020746769022353,.0162656130126722,.0061322012281136,-.0061322012281136,-.0162656130126722,-.020746769022353,-.018019851588342,-.00905023733181396,-.00649126364804423,-.0129247005652958,-.0148805763464081,-.0116664766448895,-.00439830840398167,.00439830840398167,.0116664766448895,.0148805763464081,.0129247005652958,.00649126364804423,-.0143271003056479,-.0285265691642717,-.0328434525972573,-.0257494981204761,-.00970766388421236,.00970766388421236,.0257494981204761,.0328434525972573,.0285265691642717,.0143271003056479,-.00947407369825788,-.0188637486201043,-.0217183717412093,-.0170273563863006,-.00641938152974948,.00641938152974948,.0170273563863006,.0217183717412093,.0188637486201043,.00947407369825788},
			{.00530201996819294,.010556807456243,.0121543534825683,.00952909872152508,.00359250836949255,-.00359250836949255,-.00952909872152508,-.0121543534825683,-.010556807456243,-.00530201996819294,.0118850758196888,.0236642747073748,.0272453543265051,.0213605496542905,.00805301274050799,-.00805301274050799,-.0213605496542905,-.0272453543265051,-.0236642747073748,-.0118850758196888,.017422781130702,.0346903532714859,.039939993018198,.0313132357844681,.0118052152589717,-.0118052152589717,-.0313132357844681,-.039939993018198,-.0346903532714859,-.017422781130702,.0214133596745384,.0426359607152567,.0490879974604127,.0384853356875881,.0145091256372526,-.0145091256372526,-.0384853356875881,-.0490879974604127,-.0426359607152567,-.0214133596745384,.0235014803675002,.0467936003004029,.0538748064820265,.0422382276693348,.0159239809397558,-.0159239809397558,-.0422382276693348,-.0538748064820265,-.0467936003004029,-.0235014803675002,.0235014803675002,.0467936003004029,.0538748064820265,.0422382276693348,.0159239809397558,-.0159239809397558,-.0422382276693348,-.0538748064820265,-.0467936003004029,-.0235014803675002,.0214133596745384,.0426359607152567,.0490879974604127,.0384853356875881,.0145091256372526,-.0145091256372526,-.0384853356875881,-.0490879974604127,-.0426359607152567,-.0214133596745384,.017422781130702,.0346903532714859,.039939993018198,.0313132357844681,.0118052152589717,-.0118052152589717,-.0313132357844681,-.039939993018198,-.0346903532714859,-.017422781130702,.0118850758196888,.0236642747073748,.0272453543265051,.0213605496542905,.00805301274050799,-.00805301274050799,-.0213605496542905,-.0272453543265051,-.0236642747073748,-.0118850758196888,.00530201996819294,.010556807456243,.0121543534825683,.00952909872152508,.00359250836949255,-.00359250836949255,-.00952909872152508,-.0121543534825683,-.010556807456243,-.00530201996819294},
			{-.00894374201432336,-.0178078096555534,-.0205026390981384,-.0160742134329204,-.00606004282024417,.00606004282024417,.0160742134329204,.0205026390981384,.0178078096555534,.00894374201432336,-.0178078096555534,-.0354569803355862,-.0408226326197064,-.0320052314476158,-.0120661003945087,.0120661003945087,.0320052314476158,.0408226326197064,.0354569803355862,.0178078096555534,-.0205026390981384,-.0408226326197064,-.0470002611116591,-.0368485356882857,-.0138920454842895,.0138920454842895,.0368485356882857,.0470002611116591,.0408226326197064,.0205026390981384,-.0160742134329204,-.0320052314476158,-.0368485356882857,-.0288895114677148,-.0108914614877351,.0108914614877351,.0288895114677148,.0368485356882857,.0320052314476158,.0160742134329204,-.00606004282024417,-.0120661003945087,-.0138920454842895,-.0108914614877351,-.00410612458682053,.00410612458682053,.0108914614877351,.0138920454842895,.0120661003945087,.00606004282024417,.00606004282024417,.0120661003945087,.0138920454842895,.0108914614877351,.00410612458682053,-.00410612458682053,-.0108914614877351,-.0138920454842895,-.0120661003945087,-.00606004282024417,.0160742134329204,.0320052314476158,.0368485356882857,.0288895114677148,.0108914614877351,-.0108914614877351,-.0288895114677148,-.0368485356882857,-.0320052314476158,-.0160742134329204,.0205026390981384,.0408226326197064,.0470002611116591,.0368485356882857,.0138920454842895,-.0138920454842895,-.0368485356882857,-.0470002611116591,-.0408226326197064,-.0205026390981384,.0178078096555534,.0354569803355862,.0408226326197064,.0320052314476158,.0120661003945087,-.0120661003945087,-.0320052314476158,-.0408226326197064,-.0354569803355862,-.0178078096555534,.00894374201432336,.0178078096555534,.0205026390981384,.0160742134329204,.00606004282024417,-.00606004282024417,-.0160742134329204,-.0205026390981384,-.0178078096555534,-.00894374201432336},
			{-.00592783853288971,-.0118028695476119,-.0135889803034609,-.0106538562517747,-.00401654646156772,.00401654646156772,.0106538562517747,.0135889803034609,.0118028695476119,.00592783853288971,-.01328791872398,-.026457463439368,-.0304612321695831,-.0238818205295374,-.00900354195484198,.00900354195484198,.0238818205295374,.0304612321695831,.026457463439368,.01328791872398,-.0194792614807672,-.0387849940354634,-.044654269700405,-.0350092619013457,-.0131986319027464,.0131986319027464,.0350092619013457,.044654269700405,.0387849940354634,.0194792614807672,-.0239408639271153,-.0476684532179932,-.0548820495950343,-.0430279133629583,-.016221695607902,.016221695607902,.0430279133629583,.0548820495950343,.0476684532179932,.0239408639271153,-.0262754538342299,-.0523168355867033,-.060233864778329,-.047223774154247,-.0178035519251087,.0178035519251087,.047223774154247,.060233864778329,.0523168355867033,.0262754538342299,-.0262754538342299,-.0523168355867033,-.060233864778329,-.047223774154247,-.0178035519251087,.0178035519251087,.047223774154247,.060233864778329,.0523168355867033,.0262754538342299,-.0239408639271153,-.0476684532179932,-.0548820495950343,-.0430279133629583,-.016221695607902,.016221695607902,.0430279133629583,.0548820495950343,.0476684532179932,.0239408639271153,-.0194792614807672,-.0387849940354634,-.044654269700405,-.0350092619013457,-.0131986319027464,.0131986319027464,.0350092619013457,.044654269700405,.0387849940354634,.0194792614807672,-.01328791872398,-.026457463439368,-.0304612321695831,-.0238818205295374,-.00900354195484198,.00900354195484198,.0238818205295374,.0304612321695831,.026457463439368,.01328791872398,-.00592783853288971,-.0118028695476119,-.0135889803034609,-.0106538562517747,-.00401654646156772,.00401654646156772,.0106538562517747,.0135889803034609,.0118028695476119,.00592783853288971},
			{-.00561641065141988,-.00849337690664137,-.00384814425717993,.00536515241143043,.0125927795065653,.0125927795065653,.00536515241143042,-.00384814425717991,-.00849337690664137,-.00561641065141988,-.0125898179990038,-.0190388623781496,-.00862604941815623,.0120265943124634,.0282281357131822,.0282281357131822,.0120265943124634,-.00862604941815619,-.0190388623781496,-.0125898179990038,-.0184558892850007,-.0279097867968623,-.0126452513484336,.0176302384294998,.041380689338455,.0413806893384551,.0176302384294998,-.0126452513484336,-.0279097867968623,-.0184558892850007,-.0226830947601565,-.0343023480945728,-.0155415667147304,.0216683337640905,.050858676220366,.050858676220366,.0216683337640904,-.0155415667147303,-.0343023480945728,-.0226830947601565,-.0248950334876143,-.0376473366420087,-.0170571003606095,.0237813182187718,.0558181527269258,.0558181527269258,.0237813182187717,-.0170571003606094,-.0376473366420086,-.0248950334876143,-.0248950334876143,-.0376473366420087,-.0170571003606095,.0237813182187718,.0558181527269258,.0558181527269258,.0237813182187717,-.0170571003606094,-.0376473366420086,-.0248950334876143,-.0226830947601565,-.0343023480945728,-.0155415667147304,.0216683337640905,.050858676220366,.050858676220366,.0216683337640904,-.0155415667147303,-.0343023480945728,-.0226830947601565,-.0184558892850007,-.0279097867968623,-.0126452513484336,.0176302384294998,.041380689338455,.0413806893384551,.0176302384294998,-.0126452513484336,-.0279097867968623,-.0184558892850007,-.0125898179990038,-.0190388623781496,-.00862604941815623,.0120265943124634,.0282281357131822,.0282281357131822,.0120265943124634,-.00862604941815619,-.0190388623781496,-.0125898179990038,-.00561641065141988,-.00849337690664138,-.00384814425717993,.00536515241143043,.0125927795065653,.0125927795065653,.00536515241143042,-.00384814425717991,-.00849337690664138,-.00561641065141988},
			{.00947407369825788,.0143271003056479,.00649126364804426,-.00905023733181398,-.0212422004934682,-.0212422004934682,-.00905023733181396,.00649126364804423,.0143271003056479,.00947407369825788,.0188637486201043,.0285265691642717,.0129247005652958,-.0180198515883421,-.0422951671064499,-.0422951671064499,-.018019851588342,.0129247005652958,.0285265691642717,.0188637486201043,.0217183717412093,.0328434525972573,.0148805763464081,-.020746769022353,-.0486956320598683,-.0486956320598684,-.020746769022353,.0148805763464081,.0328434525972573,.0217183717412093,.0170273563863006,.0257494981204761,.0116664766448895,-.0162656130126723,-.0381777184505166,-.0381777184505167,-.0162656130126722,.0116664766448895,.0257494981204761,.0170273563863006,.00641938152974948,.00970766388421236,.00439830840398169,-.00613220122811362,-.0143931527072752,-.0143931527072752,-.0061322012281136,.00439830840398167,.00970766388421236,.00641938152974948,-.00641938152974948,-.00970766388421236,-.00439830840398169,.00613220122811362,.0143931527072752,.0143931527072752,.0061322012281136,-.00439830840398167,-.00970766388421236,-.00641938152974948,-.0170273563863006,-.0257494981204761,-.0116664766448895,.0162656130126722,.0381777184505166,.0381777184505167,.0162656130126722,-.0116664766448895,-.0257494981204761,-.0170273563863006,-.0217183717412093,-.0328434525972573,-.0148805763464081,.020746769022353,.0486956320598683,.0486956320598684,.020746769022353,-.0148805763464081,-.0328434525972573,-.0217183717412093,-.0188637486201043,-.0285265691642717,-.0129247005652958,.0180198515883421,.0422951671064499,.0422951671064499,.018019851588342,-.0129247005652958,-.0285265691642717,-.0188637486201043,-.00947407369825788,-.0143271003056479,-.00649126364804426,.00905023733181398,.0212422004934682,.0212422004934682,.00905023733181396,-.00649126364804423,-.0143271003056479,-.00947407369825788},
			{-.0115884038691255,-.0175244810103018,-.00793991973994695,.0110699798896205,.0259828249417684,.0259828249417684,.0110699798896205,-.00793991973994691,-.0175244810103018,-.0115884038691255,-.0175244810103018,-.0265012712836701,-.0120070869359958,.0167404980488237,.0392923415017616,.0392923415017616,.0167404980488237,-.0120070869359958,-.02650127128367,-.0175244810103018,-.00793991973994695,-.0120070869359958,-.00544012153776933,.00758471596598284,.0178024123929934,.0178024123929934,.00758471596598282,-.00544012153776931,-.0120070869359958,-.00793991973994695,.0110699798896205,.0167404980488237,.00758471596598284,-.0105747483553876,-.0248204457515694,-.0248204457515694,-.0105747483553876,.00758471596598281,.0167404980488237,.0110699798896205,.0259828249417684,.0392923415017616,.0178024123929934,-.0248204457515694,-.0582571335603207,-.0582571335603207,-.0248204457515693,.0178024123929933,.0392923415017616,.0259828249417684,.0259828249417684,.0392923415017616,.0178024123929934,-.0248204457515694,-.0582571335603207,-.0582571335603207,-.0248204457515693,.0178024123929933,.0392923415017616,.0259828249417684,.0110699798896205,.0167404980488237,.00758471596598282,-.0105747483553876,-.0248204457515693,-.0248204457515693,-.0105747483553876,.00758471596598279,.0167404980488237,.0110699798896205,-.00793991973994691,-.0120070869359958,-.00544012153776931,.00758471596598281,.0178024123929933,.0178024123929933,.00758471596598279,-.00544012153776928,-.0120070869359958,-.00793991973994691,-.0175244810103018,-.02650127128367,-.0120070869359958,.0167404980488237,.0392923415017616,.0392923415017616,.0167404980488237,-.0120070869359958,-.0265012712836701,-.0175244810103018,-.0115884038691255,-.0175244810103018,-.00793991973994695,.0110699798896205,.0259828249417684,.0259828249417684,.0110699798896205,-.00793991973994691,-.0175244810103018,-.0115884038691255},
			{.00648527240266935,.009807306886327,.004443454245332,-.00619514437769402,-.0145408892752851,-.0145408892752851,-.00619514437769401,.00444345424533198,.009807306886327,.00648527240266935,.0145374696209095,.0219841846372146,.00996050390679461,-.0138871149268619,-.0325950435040259,-.0325950435040259,-.0138871149268619,.00996050390679457,.0219841846372146,.0145374696209095,.0213110252927245,.032227445838975,.0146014785393375,-.020357645805406,-.0477823042555251,-.0477823042555251,-.020357645805406,.0146014785393374,.032227445838975,.0213110252927245,.0261921817305127,.0396089398107403,.0179458554519699,-.0250204366620803,-.0587265408103395,-.0587265408103396,-.0250204366620803,.0179458554519699,.0396089398107403,.0261921817305127,.0287463085698583,.0434713998871676,.0196958429687218,-.0274603009493809,-.0644532510022925,-.0644532510022925,-.0274603009493808,.0196958429687217,.0434713998871676,.0287463085698583,.0287463085698583,.0434713998871676,.0196958429687218,-.0274603009493809,-.0644532510022925,-.0644532510022925,-.0274603009493808,.0196958429687217,.0434713998871676,.0287463085698583,.0261921817305127,.0396089398107403,.0179458554519699,-.0250204366620803,-.0587265408103395,-.0587265408103396,-.0250204366620803,.0179458554519699,.0396089398107403,.0261921817305127,.0213110252927245,.032227445838975,.0146014785393375,-.020357645805406,-.0477823042555251,-.0477823042555251,-.020357645805406,.0146014785393374,.032227445838975,.0213110252927245,.0145374696209095,.0219841846372146,.00996050390679461,-.0138871149268619,-.0325950435040259,-.0325950435040259,-.0138871149268619,.00996050390679457,.0219841846372146,.0145374696209095,.00648527240266935,.009807306886327,.004443454245332,-.00619514437769402,-.0145408892752851,-.0145408892752851,-.00619514437769401,.00444345424533198,.009807306886327,.00648527240266935},
			{-.0109397180005024,-.01654351043707,-.00749546562948678,.0104503139199635,.0245283803472424,.0245283803472424,.0104503139199635,-.00749546562948674,-.01654351043707,-.0109397180005024,-.0217819806884396,-.0329396447735407,-.0149241587017916,.0208075323314845,.0488382522307979,.0488382522307979,.0208075323314844,-.0149241587017915,-.0329396447735407,-.0217819806884396,-.0250782155433939,-.037924352397948,-.0171826095193439,.0239563053607907,.0562288725587115,.0562288725587116,.0239563053607906,-.0171826095193438,-.037924352397948,-.0250782155433939,-.0196614975872982,-.029732959344012,-.0134712868634331,.0187819121036241,.0440838320508345,.0440838320508346,.0187819121036241,-.0134712868634331,-.029732959344012,-.0196614975872982,-.00741246330878832,-.011209444713996,-.00507872908225817,.00708085605986294,.0166197811807936,.0166197811807936,.00708085605986292,-.00507872908225814,-.011209444713996,-.00741246330878832,.00741246330878832,.011209444713996,.00507872908225817,-.00708085605986294,-.0166197811807936,-.0166197811807936,-.00708085605986292,.00507872908225814,.011209444713996,.00741246330878832,.0196614975872982,.029732959344012,.0134712868634331,-.0187819121036241,-.0440838320508345,-.0440838320508346,-.0187819121036241,.0134712868634331,.029732959344012,.0196614975872982,.0250782155433939,.037924352397948,.0171826095193439,-.0239563053607907,-.0562288725587116,-.0562288725587116,-.0239563053607906,.0171826095193438,.037924352397948,.0250782155433939,.0217819806884396,.0329396447735407,.0149241587017916,-.0208075323314845,-.0488382522307979,-.0488382522307979,-.0208075323314844,.0149241587017915,.0329396447735407,.0217819806884396,.0109397180005024,.01654351043707,.00749546562948678,-.0104503139199635,-.0245283803472424,-.0245283803472424,-.0104503139199634,.00749546562948674,.01654351043707,.0109397180005024},
			{-.00725075497095922,-.0109649024347056,-.00496793287269008,.00692638197801623,.0162572084329942,.0162572084329943,.00692638197801622,-.00496793287269006,-.0109649024347056,-.00725075497095922,-.0162533851431734,-.0245790656341836,-.0111361819105275,.0155262664906382,.0364423664946087,.0364423664946087,.0155262664906382,-.0111361819105275,-.0245790656341835,-.0162533851431734,-.0238264506073575,-.0360313798109832,-.0163249492895439,.022760539936583,.0534222402072165,.0534222402072165,.0227605399365829,-.0163249492895438,-.0360313798109832,-.0238264506073575,-.0292837494080609,-.0442841409574315,-.0200640763482701,.0279736985956793,.0656582686538416,.0656582686538417,.0279736985956792,-.02006407634827,-.0442841409574315,-.0292837494080609,-.0321393500254263,-.0486025026021575,-.0220206218714746,.030701549796244,.0720609252908172,.0720609252908173,.0307015497962439,-.0220206218714745,-.0486025026021575,-.0321393500254263,-.0321393500254263,-.0486025026021575,-.0220206218714746,.030701549796244,.0720609252908172,.0720609252908173,.0307015497962439,-.0220206218714745,-.0486025026021575,-.0321393500254263,-.0292837494080609,-.0442841409574315,-.0200640763482701,.0279736985956793,.0656582686538416,.0656582686538416,.0279736985956792,-.02006407634827,-.0442841409574315,-.0292837494080609,-.0238264506073575,-.0360313798109832,-.0163249492895439,.022760539936583,.0534222402072165,.0534222402072165,.0227605399365829,-.0163249492895438,-.0360313798109832,-.0238264506073575,-.0162533851431734,-.0245790656341835,-.0111361819105275,.0155262664906383,.0364423664946087,.0364423664946087,.0155262664906382,-.0111361819105275,-.0245790656341836,-.0162533851431734,-.00725075497095922,-.0109649024347056,-.00496793287269008,.00692638197801623,.0162572084329942,.0162572084329943,.00692638197801622,-.00496793287269006,-.0109649024347056,-.00725075497095922},
			{.00705610400811139,.00598003257676846,-.0064239867066964,-.0149987461209001,-.00792740615413668,.00792740615413675,.0149987461209001,.00642398670669633,-.0059800325767685,-.00705610400811139,.0158170530535735,.0134049175607548,-.014400090820439,-.033621381269131,-.0177701750956426,.0177701750956428,.0336213812691311,.0144000908204389,-.0134049175607549,-.0158170530535735,.023186814932101,.0196507744905787,-.0211096365250878,-.0492868515145319,-.0260499702351701,.0260499702351703,.049286851514532,.0211096365250876,-.0196507744905789,-.023186814932101,.0284976092004682,.0241516609141409,-.0259446661310533,-.0605756951653628,-.0320165522353802,.0320165522353805,.0605756951653629,.025944666131053,-.024151660914141,-.0284976092004682,.0312765494684072,.0265068066591675,-.0284746565222694,-.0664827253786431,-.0351386417279264,.0351386417279267,.0664827253786433,.0284746565222691,-.0265068066591676,-.0312765494684072,.0312765494684072,.0265068066591675,-.0284746565222694,-.0664827253786431,-.0351386417279264,.0351386417279267,.0664827253786433,.0284746565222691,-.0265068066591676,-.0312765494684072,.0284976092004682,.0241516609141409,-.0259446661310533,-.0605756951653627,-.0320165522353802,.0320165522353805,.0605756951653629,.025944666131053,-.024151660914141,-.0284976092004682,.023186814932101,.0196507744905787,-.0211096365250878,-.0492868515145319,-.0260499702351701,.0260499702351703,.049286851514532,.0211096365250876,-.0196507744905789,-.023186814932101,.0158170530535735,.0134049175607548,-.014400090820439,-.033621381269131,-.0177701750956426,.0177701750956428,.0336213812691311,.0144000908204389,-.0134049175607549,-.0158170530535735,.00705610400811139,.00598003257676846,-.0064239867066964,-.0149987461209001,-.00792740615413669,.00792740615413675,.0149987461209001,.00642398670669633,-.0059800325767685,-.00705610400811139},
			{-.0119026284826952,-.010087451374564,.0108363378798955,.0253007187221328,.0133723893774315,-.0133723893774316,-.0253007187221329,-.0108363378798954,.010087451374564,.0119026284826952,-.0236992236673589,-.0200850397630209,.0215761413980188,.0503760486864314,.0266256522484521,-.0266256522484523,-.0503760486864316,-.0215761413980185,.020085039763021,.0236992236673589,-.0272855920608065,-.0231244790627329,.0248412268982238,.0579993814728395,.0306548727418767,-.0306548727418769,-.0579993814728396,-.0248412268982236,.0231244790627331,.0272855920608065,-.0213920963213377,-.01812975442821,.0194756968206063,.0454719235074688,.0240336361018362,-.0240336361018364,-.0454719235074689,-.019475696820606,.0181297544282102,.0213920963213377,-.00806490596028758,-.00683499000520012,.00734241567585834,.017143097217421,.00906077703341984,-.00906077703341992,-.017143097217421,-.00734241567585826,.00683499000520017,.00806490596028758,.00806490596028758,.00683499000520012,-.00734241567585834,-.017143097217421,-.00906077703341984,.00906077703341992,.017143097217421,.00734241567585826,-.00683499000520017,-.00806490596028758,.0213920963213377,.01812975442821,-.0194756968206063,-.0454719235074688,-.0240336361018362,.0240336361018364,.0454719235074689,.019475696820606,-.0181297544282102,-.0213920963213377,.0272855920608065,.0231244790627329,-.0248412268982238,-.0579993814728395,-.0306548727418767,.0306548727418769,.0579993814728396,.0248412268982236,-.0231244790627331,-.0272855920608065,.0236992236673589,.0200850397630209,-.0215761413980188,-.0503760486864314,-.0266256522484521,.0266256522484523,.0503760486864316,.0215761413980185,-.020085039763021,-.0236992236673589,.0119026284826952,.010087451374564,-.0108363378798955,-.0253007187221328,-.0133723893774315,.0133723893774316,.0253007187221328,.0108363378798954,-.010087451374564,-.0119026284826952},
			{-.00788896410842497,-.00668587967398251,.00718223548063763,.0167691079501009,.00886310952205344,-.00886310952205352,-.0167691079501009,-.00718223548063755,.00668587967398256,.00788896410842497,-.0176840029139668,-.0149871534477983,.0160997909767078,.0375898470038056,.0198676597409558,-.019867659740956,-.0375898470038057,-.0160997909767076,.0149871534477984,.0176840029139668,-.0259236471823205,-.0219702337835042,.0236012911228172,.0551043751861423,.0291247521258975,-.0291247521258977,-.0551043751861425,-.023601291122817,.0219702337835044,.0259236471823205,-.0318612956810125,-.0270023777840407,.0290070185583519,.0677256860801781,.0357955935981209,-.0357955935981212,-.0677256860801783,-.0290070185583516,.0270023777840409,.0318612956810126,-.0349682453529598,-.0296355107751738,.031835633806656,.0743299466305301,.0392861957663534,-.0392861957663538,-.0743299466305303,-.0318356338066557,.029635510775174,.0349682453529598,-.0349682453529598,-.0296355107751738,.031835633806656,.0743299466305301,.0392861957663534,-.0392861957663538,-.0743299466305303,-.0318356338066557,.029635510775174,.0349682453529598,-.0318612956810125,-.0270023777840407,.0290070185583519,.0677256860801781,.035795593598121,-.0357955935981213,-.0677256860801783,-.0290070185583516,.0270023777840409,.0318612956810126,-.0259236471823205,-.0219702337835042,.0236012911228172,.0551043751861423,.0291247521258975,-.0291247521258977,-.0551043751861425,-.023601291122817,.0219702337835044,.0259236471823205,-.0176840029139668,-.0149871534477983,.0160997909767078,.0375898470038056,.0198676597409558,-.019867659740956,-.0375898470038057,-.0160997909767076,.0149871534477984,.0176840029139668,-.00788896410842497,-.00668587967398251,.00718223548063763,.0167691079501009,.00886310952205344,-.00886310952205352,-.0167691079501009,-.00718223548063755,.00668587967398256,.00788896410842497},
			{-.00794977455310546,-.000443472655919539,.0146816998664313,.00745176669518039,-.013740219487093,-.013740219487093,.00745176669518016,.0146816998664309,-.000443472655919117,-.00794977455310563,-.0178203163850575,-.000994093981384771,.0329106863273325,.0167039755968772,-.0308002518592783,-.0308002518592785,.0167039755968767,.0329106863273315,-.000994093981383826,-.0178203163850579,-.0261234742434189,-.00145727987972302,.0482450169810793,.0244869881565972,-.0451512514565309,-.0451512514565312,.0244869881565965,.0482450169810779,-.00145727987972163,-.0261234742434194,-.0321068918748639,-.0017910606795139,.0592952349782769,.0300955789325382,-.05549286189967,-.0554928618996703,.0300955789325372,.0592952349782752,-.0017910606795122,-.0321068918748646,-.0352377908243784,-.00196571570441829,.0650774013003321,.0330303449893761,-.0609042403571174,-.0609042403571177,.0330303449893751,.0650774013003302,-.00196571570441642,-.0352377908243792,-.0352377908243784,-.00196571570441829,.0650774013003321,.0330303449893761,-.0609042403571174,-.0609042403571177,.0330303449893751,.0650774013003302,-.00196571570441642,-.0352377908243792,-.0321068918748639,-.0017910606795139,.0592952349782769,.0300955789325382,-.05549286189967,-.0554928618996703,.0300955789325373,.0592952349782752,-.0017910606795122,-.0321068918748646,-.0261234742434189,-.00145727987972302,.0482450169810793,.0244869881565972,-.0451512514565309,-.0451512514565312,.0244869881565965,.0482450169810779,-.00145727987972163,-.0261234742434194,-.0178203163850575,-.000994093981384771,.0329106863273325,.0167039755968772,-.0308002518592783,-.0308002518592785,.0167039755968767,.0329106863273315,-.000994093981383826,-.0178203163850579,-.00794977455310546,-.000443472655919539,.0146816998664313,.00745176669518039,-.013740219487093,-.013740219487093,.00745176669518016,.0146816998664309,-.000443472655919117,-.00794977455310563}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_left_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_left_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{.00265100998409647,.00594253790984439,.008711390565351,.0107066798372692,.0117507401837501,.0117507401837501,.0107066798372692,.008711390565351,.00594253790984439,.00265100998409647,.00527840372812152,.0118321373536874,.0173451766357429,.0213179803576284,.0233968001502014,.0233968001502014,.0213179803576284,.0173451766357429,.0118321373536874,.00527840372812152,.00607717674128413,.0136226771632526,.019969996509099,.0245439987302064,.0269374032410133,.0269374032410133,.0245439987302064,.019969996509099,.0136226771632526,.00607717674128413,.00476454936076254,.0106802748271453,.0156566178922341,.0192426678437941,.0211191138346674,.0211191138346674,.0192426678437941,.0156566178922341,.0106802748271453,.00476454936076254,.00179625418474627,.00402650637025399,.00590260762948584,.00725456281862631,.00796199046987792,.00796199046987792,.00725456281862631,.00590260762948584,.00402650637025399,.00179625418474628,-.00179625418474627,-.00402650637025399,-.00590260762948584,-.00725456281862631,-.00796199046987792,-.00796199046987792,-.00725456281862631,-.00590260762948584,-.00402650637025399,-.00179625418474628,-.00476454936076254,-.0106802748271453,-.0156566178922341,-.0192426678437941,-.0211191138346674,-.0211191138346674,-.0192426678437941,-.0156566178922341,-.0106802748271453,-.00476454936076254,-.00607717674128413,-.0136226771632526,-.019969996509099,-.0245439987302064,-.0269374032410133,-.0269374032410133,-.0245439987302064,-.019969996509099,-.0136226771632526,-.00607717674128413,-.00527840372812152,-.0118321373536874,-.0173451766357429,-.0213179803576284,-.0233968001502014,-.0233968001502014,-.0213179803576284,-.0173451766357429,-.0118321373536874,-.00527840372812152,-.00265100998409647,-.00594253790984439,-.008711390565351,-.0107066798372692,-.0117507401837501,-.0117507401837501,-.0107066798372692,-.008711390565351,-.00594253790984439,-.00265100998409647},
			{-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.002221727122666,-.00498025195339731,-.00730073926966875,-.00897292772598497,-.00984792148436088,-.00984792148436088,-.00897292772598497,-.00730073926966875,-.00498025195339731,-.002221727122666,.00309757218884701,.00694355746343096,.010178822902703,.0125102183310402,.0137301504746905,.0137301504746905,.0125102183310402,.010178822902703,.00694355746343096,.00309757218884701,.00727044463764254,.0162975217520129,.0238911521277625,.0293632704051698,.0322266255011462,.0322266255011462,.0293632704051698,.0238911521277625,.0162975217520129,.00727044463764254,.00727044463764255,.016297521752013,.0238911521277625,.0293632704051698,.0322266255011463,.0322266255011463,.0293632704051698,.0238911521277625,.016297521752013,.00727044463764255,.003097572188847,.00694355746343095,.010178822902703,.0125102183310401,.0137301504746904,.0137301504746904,.0125102183310401,.010178822902703,.00694355746343095,.003097572188847,-.00222172712266599,-.00498025195339728,-.00730073926966871,-.00897292772598493,-.00984792148436084,-.00984792148436084,-.00897292772598493,-.00730073926966871,-.00498025195339728,-.00222172712266599,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468},
			{.00352805200405569,.00790852652678675,.0115934074660505,.0142488046002341,.0156382747342036,.0156382747342036,.0142488046002341,.0115934074660505,.00790852652678675,.00352805200405569,.00299001628838423,.00670245878037739,.00982538724528936,.0120758304570704,.0132534033295837,.0132534033295837,.0120758304570704,.00982538724528936,.00670245878037739,.00299001628838423,-.0032119933533482,-.00720004541021951,-.0105548182625439,-.0129723330655266,-.0142373282611347,-.0142373282611347,-.0129723330655266,-.0105548182625439,-.00720004541021951,-.0032119933533482,-.00749937306045005,-.0168106906345655,-.0246434257572659,-.0302878475826814,-.0332413626893215,-.0332413626893215,-.0302878475826814,-.0246434257572659,-.0168106906345655,-.00749937306045005,-.00396370307706834,-.0088850875478213,-.0130249851175851,-.0160082761176901,-.0175693208639632,-.0175693208639632,-.0160082761176901,-.0130249851175851,-.0088850875478213,-.00396370307706834,.00396370307706838,.00888508754782138,.0130249851175852,.0160082761176902,.0175693208639633,.0175693208639633,.0160082761176902,.0130249851175852,.00888508754782138,.00396370307706837,.00749937306045007,.0168106906345656,.024643425757266,.0302878475826815,.0332413626893216,.0332413626893216,.0302878475826815,.024643425757266,.0168106906345656,.00749937306045007,.00321199335334816,.00720004541021943,.0105548182625438,.0129723330655265,.0142373282611345,.0142373282611345,.0129723330655265,.0105548182625438,.00720004541021943,.00321199335334816,-.00299001628838425,-.00670245878037744,-.00982538724528943,-.0120758304570705,-.0132534033295838,-.0132534033295838,-.0120758304570705,-.00982538724528943,-.00670245878037744,-.00299001628838425,-.0035280520040557,-.00790852652678675,-.0115934074660505,-.0142488046002341,-.0156382747342036,-.0156382747342036,-.0142488046002341,-.0115934074660505,-.00790852652678675,-.0035280520040557},
			{-.00355524726055973,-.00796948776183025,-.0116827728408898,-.0143586385526627,-.0157588191287273,-.0157588191287273,-.0143586385526627,-.0116827728408898,-.00796948776183025,-.00355524726055973,-.000198327000917931,-.000444572343586338,-.000651715374523445,-.000800986686075343,-.000879094787718522,-.000879094787718522,-.000800986686075343,-.000651715374523445,-.000444572343586338,-.000198327000917931,.00656585578393543,.0147181063597184,.0215758275045217,.0265176352250662,.0291034986151868,.0291034986151868,.0265176352250662,.0215758275045218,.0147181063597184,.00656585578393543,.00333253137587672,.00747024498425,.0109509140141708,.0134591520602391,.0147716193401924,.0147716193401924,.0134591520602391,.0109509140141708,.00747024498425,.00333253137587672,-.00614481295848751,-.0137742913733916,-.020192253500946,-.0248171622895082,-.0272372043055647,-.0272372043055647,-.0248171622895082,-.020192253500946,-.0137742913733916,-.00614481295848751,-.00614481295848755,-.0137742913733917,-.0201922535009461,-.0248171622895084,-.0272372043055649,-.0272372043055649,-.0248171622895084,-.0201922535009461,-.0137742913733917,-.00614481295848755,.00333253137587662,.00747024498424977,.0109509140141705,.0134591520602387,.014771619340192,.014771619340192,.0134591520602387,.0109509140141704,.00747024498424977,.00333253137587662,.00656585578393523,.014718106359718,.0215758275045211,.0265176352250655,.0291034986151859,.0291034986151859,.0265176352250655,.0215758275045211,.014718106359718,.00656585578393523,-.000198327000917742,-.000444572343585915,-.000651715374522825,-.000800986686074581,-.000879094787717686,-.000879094787717686,-.000800986686074581,-.000651715374522825,-.000444572343585915,-.000198327000917742,-.0035552472605598,-.00796948776183042,-.0116827728408901,-.014358638552663,-.0157588191287277,-.0157588191287277,-.014358638552663,-.0116827728408901,-.00796948776183042,-.0035552472605598},
			{.00222253422114756,.00498206115568338,.00730339144757066,.00897618736857064,.00985149899052509,.00985149899052509,.00897618736857064,.00730339144757066,.00498206115568338,.00222253422114756,.00498206115568338,.011167852050509,.0163713757428243,.0201211364889587,.0220832462235868,.0220832462235868,.0201211364889587,.0163713757428243,.011167852050509,.00498206115568339,.00730339144757066,.0163713757428243,.0239994174797936,.02949633325581,.0323726639565543,.0323726639565543,.02949633325581,.0239994174797936,.0163713757428243,.00730339144757066,.00897618736857064,.0201211364889587,.02949633325581,.0362522830510504,.0397874192256888,.0397874192256888,.0362522830510504,.02949633325581,.0201211364889587,.00897618736857064,.00985149899052509,.0220832462235868,.0323726639565543,.0397874192256888,.0436672836966291,.0436672836966291,.0397874192256889,.0323726639565543,.0220832462235868,.0098514989905251,.00985149899052509,.0220832462235868,.0323726639565543,.0397874192256888,.0436672836966291,.0436672836966291,.0397874192256889,.0323726639565543,.0220832462235868,.0098514989905251,.00897618736857064,.0201211364889587,.02949633325581,.0362522830510504,.0397874192256889,.0397874192256889,.0362522830510504,.02949633325581,.0201211364889587,.00897618736857064,.00730339144757066,.0163713757428243,.0239994174797936,.02949633325581,.0323726639565543,.0323726639565543,.02949633325581,.0239994174797936,.0163713757428243,.00730339144757066,.00498206115568338,.011167852050509,.0163713757428243,.0201211364889587,.0220832462235868,.0220832462235868,.0201211364889587,.0163713757428243,.011167852050509,.00498206115568339,.00222253422114756,.00498206115568339,.00730339144757066,.00897618736857064,.0098514989905251,.0098514989905251,.00897618736857064,.00730339144757066,.00498206115568339,.00222253422114756},
			{-.00374909427409703,-.00840401770836613,-.0123197666866244,-.0151415318362812,-.0166180561384473,-.0166180561384473,-.0151415318362812,-.0123197666866244,-.00840401770836613,-.00374909427409703,-.00746479014118745,-.0167331691201299,-.024529784043959,-.0301481769489968,-.0330880720938548,-.0330880720938548,-.0301481769489968,-.024529784043959,-.0167331691201299,-.00746479014118745,-.00859442576984082,-.019265374803192,-.0282418399082409,-.0347104558846931,-.0380952410046641,-.0380952410046641,-.0347104558846931,-.0282418399082409,-.019265374803192,-.00859442576984082,-.00673809032566718,-.0151041895128434,-.022141801367642,-.027213241845299,-.0298669372150783,-.0298669372150783,-.027213241845299,-.022141801367642,-.0151041895128434,-.00673809032566718,-.00254028702994505,-.00569433991870819,-.00834754776432466,-.0102595011288345,-.0112599549077927,-.0112599549077927,-.0102595011288345,-.00834754776432466,-.00569433991870818,-.00254028702994505,.00254028702994505,.00569433991870819,.00834754776432466,.0102595011288345,.0112599549077927,.0112599549077927,.0102595011288345,.00834754776432466,.00569433991870818,.00254028702994505,.00673809032566718,.0151041895128434,.022141801367642,.027213241845299,.0298669372150783,.0298669372150783,.027213241845299,.022141801367642,.0151041895128434,.00673809032566718,.00859442576984082,.019265374803192,.0282418399082409,.0347104558846931,.0380952410046641,.0380952410046641,.0347104558846931,.0282418399082409,.019265374803192,.00859442576984082,.00746479014118745,.0167331691201299,.024529784043959,.0301481769489968,.0330880720938548,.0330880720938548,.0301481769489968,.024529784043959,.0167331691201299,.00746479014118745,.00374909427409703,.00840401770836614,.0123197666866244,.0151415318362812,.0166180561384473,.0166180561384473,.0151415318362812,.0123197666866244,.00840401770836614,.00374909427409703},
			{.00458578009298867,.0102795433484883,.0150691704959577,.0185206693125625,.0203267097203668,.0203267097203668,.0185206693125625,.0150691704959577,.0102795433484883,.00458578009298867,.00693481320331858,.0155451660331847,.0227882454891813,.0280077499310155,.0307389216426545,.0307389216426545,.0280077499310155,.0227882454891813,.0155451660331847,.00693481320331858,.00314199662823143,.00704313985533036,.0103248044887574,.0126896360821209,.0139270641219972,.0139270641219972,.0126896360821209,.0103248044887574,.00704313985533036,.00314199662823143,-.00438062859915128,-.00981967313422903,-.0143950293955455,-.0176921204289931,-.0194173650114245,-.0194173650114245,-.0176921204289931,-.0143950293955455,-.00981967313422903,-.00438062859915128,-.0102819614092862,-.0230481762908429,-.0337871913540478,-.0415259352355491,-.0455753308454797,-.0455753308454797,-.0415259352355491,-.0337871913540478,-.0230481762908429,-.0102819614092862,-.0102819614092862,-.0230481762908429,-.0337871913540478,-.0415259352355491,-.0455753308454797,-.0455753308454797,-.0415259352355492,-.0337871913540478,-.0230481762908429,-.0102819614092862,-.00438062859915127,-.009819673134229,-.0143950293955455,-.0176921204289931,-.0194173650114245,-.0194173650114245,-.0176921204289931,-.0143950293955455,-.009819673134229,-.00438062859915127,.00314199662823142,.00704313985533033,.0103248044887574,.0126896360821208,.0139270641219972,.0139270641219972,.0126896360821208,.0103248044887574,.00704313985533033,.00314199662823142,.00693481320331858,.0155451660331847,.0227882454891813,.0280077499310155,.0307389216426545,.0307389216426545,.0280077499310155,.0227882454891813,.0155451660331847,.00693481320331858,.00458578009298867,.0102795433484883,.0150691704959577,.0185206693125625,.0203267097203668,.0203267097203668,.0185206693125625,.0150691704959577,.0102795433484883,.00458578009298867},
			{-.00498941899246989,-.0111843454716204,-.0163955540712153,-.0201508527115458,-.0221158602193511,-.0221158602193511,-.0201508527115458,-.0163955540712153,-.0111843454716204,-.00498941899246989,-.00422852158639074,-.00947870810765226,-.0138951958966771,-.0170778032078584,-.0187431427347081,-.0187431427347081,-.0170778032078584,-.0138951958966771,-.00947870810765226,-.00422852158639074,.00454244456217192,.0101824018679708,.0149267671340066,.0183456493553325,.0201346227171464,.0201346227171464,.0183456493553325,.0149267671340066,.0101824018679708,.00454244456217192,.0106057150904842,.0237739066862442,.0348510669263035,.0428334848228836,.0470103859430136,.0470103859430136,.0428334848228836,.0348510669263035,.0237739066862442,.0106057150904842,.0056055226483345,.0125654113119353,.0184201106014339,.0226391211939302,.0248467718453936,.0248467718453936,.0226391211939302,.0184201106014339,.0125654113119353,.0056055226483345,-.00560552264833455,-.0125654113119354,-.0184201106014341,-.0226391211939304,-.0248467718453938,-.0248467718453938,-.0226391211939304,-.0184201106014341,-.0125654113119354,-.00560552264833455,-.0106057150904842,-.0237739066862443,-.0348510669263036,-.0428334848228837,-.0470103859430138,-.0470103859430138,-.0428334848228837,-.0348510669263036,-.0237739066862442,-.0106057150904842,-.00454244456217188,-.0101824018679707,-.0149267671340064,-.0183456493553323,-.0201346227171462,-.0201346227171462,-.0183456493553323,-.0149267671340064,-.0101824018679707,-.00454244456217188,.00422852158639077,.00947870810765234,.0138951958966772,.0170778032078585,.0187431427347082,.0187431427347082,.0170778032078585,.0138951958966772,.00947870810765234,.00422852158639077,.00498941899246989,.0111843454716204,.0163955540712153,.0201508527115458,.0221158602193511,.0221158602193511,.0201508527115458,.0163955540712153,.0111843454716204,.00498941899246989},
			{-.00272203738908372,-.00610175384992259,-.00894479121996913,-.01099353944528,-.012065572865233,-.012065572865233,-.01099353944528,-.00894479121996913,-.00610175384992259,-.00272203738908372,-.00610175384992259,-.0136777695245317,-.0200507584804236,-.0246432587236037,-.0270463425584096,-.0270463425584096,-.0246432587236037,-.0200507584804236,-.0136777695245317,-.00610175384992259,-.00894479121996913,-.0200507584804236,-.0293931634773647,-.0361254828831082,-.0396482541575827,-.0396482541575827,-.0361254828831082,-.0293931634773647,-.0200507584804236,-.00894479121996913,-.01099353944528,-.0246432587236037,-.0361254828831082,-.0443997977469403,-.0487294376468829,-.0487294376468829,-.0443997977469403,-.0361254828831082,-.0246432587236037,-.01099353944528,-.012065572865233,-.0270463425584096,-.0396482541575827,-.0487294376468829,-.053481281759782,-.053481281759782,-.0487294376468829,-.0396482541575827,-.0270463425584096,-.012065572865233,-.012065572865233,-.0270463425584096,-.0396482541575827,-.0487294376468829,-.053481281759782,-.053481281759782,-.0487294376468829,-.0396482541575827,-.0270463425584096,-.012065572865233,-.01099353944528,-.0246432587236037,-.0361254828831082,-.0443997977469403,-.0487294376468829,-.0487294376468829,-.0443997977469403,-.0361254828831082,-.0246432587236037,-.01099353944528,-.00894479121996913,-.0200507584804236,-.0293931634773647,-.0361254828831082,-.0396482541575827,-.0396482541575827,-.0361254828831082,-.0293931634773647,-.0200507584804236,-.00894479121996913,-.00610175384992259,-.0136777695245317,-.0200507584804236,-.0246432587236037,-.0270463425584096,-.0270463425584096,-.0246432587236037,-.0200507584804236,-.0136777695245317,-.00610175384992259,-.00272203738908372,-.00610175384992259,-.00894479121996913,-.01099353944528,-.012065572865233,-.012065572865233,-.01099353944528,-.00894479121996913,-.00610175384992259,-.00272203738908372},
			{.00459168398220963,.0102927775821281,.0150885710584479,.0185445134519897,.0203528790176253,.0203528790176253,.0185445134519897,.0150885710584479,.0102927775821281,.00459168398220963,.00914246343674623,.0204938630514994,.0300427271887783,.0369238250811581,.0405244465804058,.0405244465804058,.0369238250811581,.0300427271887783,.0204938630514994,.00914246343674623,.0105259788787713,.0235951689735483,.0345890485685456,.0425114528066451,.0466569510209666,.0466569510209666,.0425114528066451,.0345890485685456,.0235951689735483,.0105259788787713,.00825244156510291,.0184987786328967,.0271180576544867,.0333292783668801,.0365793781595865,.0365793781595865,.0333292783668801,.0271180576544867,.0184987786328967,.00825244156510291,.00311120351019257,.00697411360782243,.010223616307811,.0125652713841337,.0137905720183488,.0137905720183489,.0125652713841337,.010223616307811,.00697411360782243,.00311120351019257,-.00311120351019257,-.00697411360782243,-.010223616307811,-.0125652713841337,-.0137905720183489,-.0137905720183488,-.0125652713841337,-.010223616307811,-.00697411360782243,-.00311120351019257,-.00825244156510291,-.0184987786328967,-.0271180576544867,-.0333292783668801,-.0365793781595865,-.0365793781595865,-.0333292783668801,-.0271180576544867,-.0184987786328967,-.00825244156510291,-.0105259788787713,-.0235951689735483,-.0345890485685456,-.0425114528066451,-.0466569510209666,-.0466569510209666,-.0425114528066451,-.0345890485685456,-.0235951689735483,-.0105259788787713,-.00914246343674623,-.0204938630514994,-.0300427271887783,-.0369238250811581,-.0405244465804058,-.0405244465804058,-.0369238250811581,-.0300427271887783,-.0204938630514994,-.00914246343674623,-.00459168398220963,-.0102927775821281,-.0150885710584479,-.0185445134519897,-.0203528790176254,-.0203528790176254,-.0185445134519897,-.0150885710584479,-.0102927775821281,-.00459168398220963},
			{-.00561641065141988,-.0125898179990038,-.0184558892850007,-.0226830947601565,-.0248950334876143,-.0248950334876143,-.0226830947601565,-.0184558892850007,-.0125898179990038,-.00561641065141988,-.00849337690664137,-.0190388623781496,-.0279097867968623,-.0343023480945728,-.0376473366420087,-.0376473366420087,-.0343023480945728,-.0279097867968623,-.0190388623781496,-.00849337690664138,-.00384814425717993,-.00862604941815623,-.0126452513484336,-.0155415667147304,-.0170571003606095,-.0170571003606095,-.0155415667147304,-.0126452513484336,-.00862604941815623,-.00384814425717993,.00536515241143043,.0120265943124634,.0176302384294998,.0216683337640905,.0237813182187718,.0237813182187718,.0216683337640905,.0176302384294998,.0120265943124634,.00536515241143043,.0125927795065653,.0282281357131822,.041380689338455,.050858676220366,.0558181527269258,.0558181527269258,.050858676220366,.041380689338455,.0282281357131822,.0125927795065653,.0125927795065653,.0282281357131822,.0413806893384551,.050858676220366,.0558181527269258,.0558181527269258,.050858676220366,.0413806893384551,.0282281357131822,.0125927795065653,.00536515241143042,.0120265943124634,.0176302384294998,.0216683337640904,.0237813182187717,.0237813182187717,.0216683337640904,.0176302384294998,.0120265943124634,.00536515241143042,-.00384814425717991,-.00862604941815619,-.0126452513484336,-.0155415667147303,-.0170571003606094,-.0170571003606094,-.0155415667147303,-.0126452513484336,-.00862604941815619,-.00384814425717991,-.00849337690664137,-.0190388623781496,-.0279097867968623,-.0343023480945728,-.0376473366420086,-.0376473366420086,-.0343023480945728,-.0279097867968623,-.0190388623781496,-.00849337690664138,-.00561641065141988,-.0125898179990038,-.0184558892850007,-.0226830947601565,-.0248950334876143,-.0248950334876143,-.0226830947601565,-.0184558892850007,-.0125898179990038,-.00561641065141988},
			{.00314313803755599,.00704569845308084,.0103285552337492,.0126942459116857,.0139321234784299,.0139321234784299,.0126942459116857,.0103285552337492,.00704569845308084,.00314313803755599,.00704569845308084,.0157937278282394,.0231526216040999,.0284555841055385,.0312304263023817,.0312304263023817,.0284555841055385,.0231526216040999,.0157937278282394,.00704569845308084,.0103285552337492,.0231526216040999,.033940301680024,.0417141145196382,.0457818604054276,.0457818604054276,.0417141145196382,.033940301680024,.0231526216040999,.0103285552337492,.0126942459116857,.0284555841055385,.0417141145196382,.0512684703442582,.0562679078659487,.0562679078659487,.0512684703442582,.0417141145196382,.0284555841055385,.0126942459116857,.0139321234784299,.0312304263023817,.0457818604054276,.0562679078659487,.0617548648194743,.0617548648194743,.0562679078659487,.0457818604054276,.0312304263023817,.0139321234784299,.0139321234784299,.0312304263023817,.0457818604054276,.0562679078659487,.0617548648194743,.0617548648194743,.0562679078659487,.0457818604054276,.0312304263023817,.0139321234784299,.0126942459116857,.0284555841055385,.0417141145196382,.0512684703442582,.0562679078659487,.0562679078659487,.0512684703442582,.0417141145196382,.0284555841055385,.0126942459116857,.0103285552337492,.0231526216040999,.033940301680024,.0417141145196382,.0457818604054276,.0457818604054276,.0417141145196382,.033940301680024,.0231526216040999,.0103285552337492,.00704569845308084,.0157937278282394,.0231526216040999,.0284555841055385,.0312304263023817,.0312304263023817,.0284555841055385,.0231526216040999,.0157937278282394,.00704569845308084,.00314313803755599,.00704569845308084,.0103285552337492,.0126942459116857,.0139321234784299,.0139321234784299,.0126942459116857,.0103285552337492,.00704569845308084,.00314313803755599},
			{-.00530201996819294,-.0118850758196888,-.017422781130702,-.0214133596745384,-.0235014803675002,-.0235014803675002,-.0214133596745384,-.017422781130702,-.0118850758196888,-.00530201996819294,-.010556807456243,-.0236642747073748,-.0346903532714859,-.0426359607152567,-.0467936003004029,-.0467936003004029,-.0426359607152567,-.0346903532714859,-.0236642747073748,-.010556807456243,-.0121543534825683,-.0272453543265051,-.039939993018198,-.0490879974604127,-.0538748064820265,-.0538748064820265,-.0490879974604127,-.039939993018198,-.0272453543265051,-.0121543534825683,-.00952909872152508,-.0213605496542905,-.0313132357844681,-.0384853356875881,-.0422382276693348,-.0422382276693348,-.0384853356875881,-.0313132357844681,-.0213605496542905,-.00952909872152508,-.00359250836949255,-.00805301274050799,-.0118052152589717,-.0145091256372526,-.0159239809397558,-.0159239809397558,-.0145091256372526,-.0118052152589717,-.00805301274050799,-.00359250836949255,.00359250836949255,.00805301274050799,.0118052152589717,.0145091256372526,.0159239809397558,.0159239809397558,.0145091256372526,.0118052152589717,.00805301274050799,.00359250836949255,.00952909872152508,.0213605496542905,.0313132357844681,.0384853356875881,.0422382276693348,.0422382276693348,.0384853356875881,.0313132357844681,.0213605496542905,.00952909872152508,.0121543534825683,.0272453543265051,.039939993018198,.0490879974604127,.0538748064820265,.0538748064820265,.0490879974604127,.039939993018198,.0272453543265051,.0121543534825683,.010556807456243,.0236642747073748,.0346903532714859,.0426359607152567,.0467936003004029,.0467936003004029,.0426359607152567,.0346903532714859,.0236642747073748,.010556807456243,.00530201996819294,.0118850758196888,.017422781130702,.0214133596745384,.0235014803675002,.0235014803675002,.0214133596745384,.017422781130702,.0118850758196888,.00530201996819294},
			{-.00351413515806021,-.00787733034668568,-.0115476758084439,-.0141925983938026,-.015576587587625,-.015576587587625,-.0141925983938026,-.0115476758084439,-.00787733034668568,-.00351413515806021,-.00787733034668568,-.017657924524755,-.0258854178874996,-.0318143102064225,-.0349166780965639,-.0349166780965639,-.0318143102064225,-.0258854178874996,-.017657924524755,-.00787733034668568,-.0115476758084439,-.0258854178874996,-.0379464108746825,-.0466377978533816,-.0511856760122494,-.0511856760122494,-.0466377978533816,-.0379464108746825,-.0258854178874996,-.0115476758084439,-.0141925983938026,-.0318143102064225,-.0466377978533816,-.0573198924081667,-.0629094334832252,-.0629094334832252,-.0573198924081666,-.0466377978533816,-.0318143102064225,-.0141925983938026,-.015576587587625,-.0349166780965639,-.0511856760122494,-.0629094334832252,-.0690440378533661,-.0690440378533661,-.0629094334832252,-.0511856760122494,-.0349166780965639,-.015576587587625,-.015576587587625,-.0349166780965639,-.0511856760122494,-.0629094334832252,-.0690440378533661,-.0690440378533661,-.0629094334832252,-.0511856760122494,-.0349166780965639,-.015576587587625,-.0141925983938026,-.0318143102064225,-.0466377978533816,-.0573198924081666,-.0629094334832252,-.0629094334832252,-.0573198924081667,-.0466377978533816,-.0318143102064225,-.0141925983938026,-.0115476758084439,-.0258854178874996,-.0379464108746825,-.0466377978533816,-.0511856760122494,-.0511856760122494,-.0466377978533816,-.0379464108746825,-.0258854178874996,-.0115476758084439,-.00787733034668568,-.017657924524755,-.0258854178874996,-.0318143102064225,-.0349166780965639,-.0349166780965639,-.0318143102064225,-.0258854178874996,-.017657924524755,-.00787733034668568,-.00351413515806021,-.00787733034668568,-.0115476758084439,-.0141925983938026,-.015576587587625,-.015576587587625,-.0141925983938026,-.0115476758084439,-.00787733034668568,-.00351413515806021},
			{.00374909427409703,.00746479014118745,.00859442576984082,.00673809032566718,.00254028702994505,-.00254028702994505,-.00673809032566718,-.00859442576984082,-.00746479014118745,-.00374909427409703,.00840401770836613,.0167331691201299,.019265374803192,.0151041895128434,.00569433991870819,-.00569433991870819,-.0151041895128434,-.019265374803192,-.0167331691201299,-.00840401770836614,.0123197666866244,.024529784043959,.0282418399082409,.022141801367642,.00834754776432466,-.00834754776432466,-.022141801367642,-.0282418399082409,-.024529784043959,-.0123197666866244,.0151415318362812,.0301481769489968,.0347104558846931,.027213241845299,.0102595011288345,-.0102595011288345,-.027213241845299,-.0347104558846931,-.0301481769489968,-.0151415318362812,.0166180561384473,.0330880720938548,.0380952410046641,.0298669372150783,.0112599549077927,-.0112599549077927,-.0298669372150783,-.0380952410046641,-.0330880720938548,-.0166180561384473,.0166180561384473,.0330880720938548,.0380952410046641,.0298669372150783,.0112599549077927,-.0112599549077927,-.0298669372150783,-.0380952410046641,-.0330880720938548,-.0166180561384473,.0151415318362812,.0301481769489968,.0347104558846931,.027213241845299,.0102595011288345,-.0102595011288345,-.027213241845299,-.0347104558846931,-.0301481769489968,-.0151415318362812,.0123197666866244,.024529784043959,.0282418399082409,.022141801367642,.00834754776432466,-.00834754776432466,-.022141801367642,-.0282418399082409,-.024529784043959,-.0123197666866244,.00840401770836613,.0167331691201299,.019265374803192,.0151041895128434,.00569433991870818,-.00569433991870818,-.0151041895128434,-.019265374803192,-.0167331691201299,-.00840401770836614,.00374909427409703,.00746479014118745,.00859442576984082,.00673809032566718,.00254028702994505,-.00254028702994505,-.00673809032566718,-.00859442576984082,-.00746479014118745,-.00374909427409703},
			{-.00632418062917951,-.0125920229688431,-.0144975551423388,-.0113661853236565,-.00428509737360599,.00428509737360599,.0113661853236565,.0144975551423388,.0125920229688431,.00632418062917952,-.0125920229688431,-.0250718712423055,-.0288659603588969,-.0226311161960246,-.00853202141368567,.00853202141368567,.0226311161960246,.0288659603588969,.0250718712423055,.0125920229688431,-.0144975551423388,-.0288659603588969,-.0332342033583603,-.0260558494688553,-.00982315956908459,.00982315956908459,.0260558494688553,.0332342033583603,.0288659603588969,.0144975551423388,-.0113661853236565,-.0226311161960246,-.0260558494688553,-.0204279694693769,-.0077014262770414,.0077014262770414,.0204279694693769,.0260558494688553,.0226311161960246,.0113661853236565,-.00428509737360599,-.00853202141368567,-.00982315956908459,-.0077014262770414,-.00290346854050359,.00290346854050359,.0077014262770414,.00982315956908459,.00853202141368567,.00428509737360599,.00428509737360599,.00853202141368567,.00982315956908459,.0077014262770414,.00290346854050359,-.00290346854050359,-.0077014262770414,-.00982315956908459,-.00853202141368567,-.00428509737360599,.0113661853236565,.0226311161960246,.0260558494688553,.0204279694693769,.0077014262770414,-.0077014262770414,-.0204279694693769,-.0260558494688553,-.0226311161960246,-.0113661853236565,.0144975551423388,.0288659603588969,.0332342033583603,.0260558494688553,.00982315956908459,-.00982315956908459,-.0260558494688553,-.0332342033583603,-.0288659603588969,-.0144975551423388,.0125920229688431,.0250718712423055,.0288659603588969,.0226311161960246,.00853202141368567,-.00853202141368567,-.0226311161960246,-.0288659603588969,-.0250718712423055,-.0125920229688431,.00632418062917952,.0125920229688431,.0144975551423388,.0113661853236565,.00428509737360599,-.00428509737360599,-.0113661853236565,-.0144975551423388,-.0125920229688431,-.00632418062917951},
			{.00773554878208247,.0154021862517905,.0177329762700093,.0139027782716481,.00524140307070943,-.00524140307070943,-.0139027782716481,-.0177329762700093,-.0154021862517905,-.00773554878208247,.0116980284141665,.023291846188219,.0268165667515141,.0210243771759672,.00792627437025251,-.00792627437025251,-.0210243771759672,-.0268165667515141,-.023291846188219,-.0116980284141665,.00530009457452694,.0105529738210754,.0121499397090725,.00952563829202252,.00359120377371563,-.00359120377371563,-.00952563829202252,-.0121499397090725,-.0105529738210754,-.00530009457452694,-.0073894878380083,-.0147131472107018,-.0169396659720433,-.0132808174115363,-.00500692133631402,.00500692133631402,.0132808174115363,.0169396659720433,.0147131472107018,.0073894878380083,-.0173441840742927,-.0345338593321725,-.0397598170829848,-.0311719765824586,-.0117519599742572,.0117519599742572,.0311719765824586,.0397598170829848,.0345338593321725,.0173441840742927,-.0173441840742927,-.0345338593321725,-.0397598170829848,-.0311719765824586,-.0117519599742572,.0117519599742572,.0311719765824586,.0397598170829848,.0345338593321725,.0173441840742927,-.00738948783800828,-.0147131472107018,-.0169396659720432,-.0132808174115363,-.00500692133631401,.00500692133631401,.0132808174115363,.0169396659720432,.0147131472107018,.00738948783800828,.00530009457452691,.0105529738210754,.0121499397090724,.00952563829202247,.00359120377371561,-.00359120377371561,-.00952563829202247,-.0121499397090724,-.0105529738210754,-.00530009457452691,.0116980284141665,.023291846188219,.0268165667515141,.0210243771759672,.00792627437025251,-.00792627437025251,-.0210243771759672,-.0268165667515141,-.023291846188219,-.0116980284141665,.00773554878208247,.0154021862517905,.0177329762700093,.0139027782716481,.00524140307070943,-.00524140307070943,-.0139027782716481,-.0177329762700093,-.0154021862517905,-.00773554878208247},
			{-.00841642931466872,-.0167578817652623,-.0192938271762863,-.0151264963737114,-.005702749694565,.005702749694565,.0151264963737114,.0192938271762863,.0167578817652623,.00841642931466872,-.00713290527236138,-.0142022678178642,-.0163514759578514,-.0128196722983645,-.00483306778237002,.00483306778237002,.0128196722983645,.0163514759578514,.0142022678178642,.00713290527236138,.00766244799865884,.015256635895486,.0175653999940025,.0137713972911834,.00519187191506662,-.00519187191506662,-.0137713972911834,-.0175653999940025,-.015256635895486,-.00766244799865884,.0178903097786119,.0356212456381444,.0410117559470465,.0321535054680605,.0121220002938583,-.0121220002938583,-.0321535054680605,-.0410117559470465,-.0356212456381444,-.0178903097786118,.00945570721013499,.0188271792597616,.0216762683937647,.0169943470654115,.00640693688361546,-.00640693688361546,-.0169943470654115,-.0216762683937647,-.0188271792597616,-.00945570721013499,-.00945570721013506,-.0188271792597618,-.0216762683937649,-.0169943470654116,-.00640693688361551,.00640693688361551,.0169943470654116,.0216762683937649,.0188271792597618,.00945570721013506,-.0178903097786119,-.0356212456381445,-.0410117559470466,-.0321535054680606,-.0121220002938583,.0121220002938583,.0321535054680606,.0410117559470466,.0356212456381445,.0178903097786119,-.00766244799865876,-.0152566358954859,-.0175653999940023,-.0137713972911832,-.00519187191506656,.00519187191506656,.0137713972911832,.0175653999940023,.0152566358954859,.00766244799865876,.00713290527236143,.0142022678178643,.0163514759578515,.0128196722983646,.00483306778237005,-.00483306778237005,-.0128196722983646,-.0163514759578515,-.0142022678178643,-.00713290527236143,.00841642931466873,.0167578817652623,.0192938271762863,.0151264963737114,.00570274969456501,-.00570274969456501,-.0151264963737114,-.0192938271762863,-.0167578817652623,-.00841642931466873},
			{-.00459168398220963,-.00914246343674623,-.0105259788787713,-.00825244156510291,-.00311120351019257,.00311120351019257,.00825244156510291,.0105259788787713,.00914246343674623,.00459168398220963,-.0102927775821281,-.0204938630514994,-.0235951689735483,-.0184987786328967,-.00697411360782243,.00697411360782243,.0184987786328967,.0235951689735483,.0204938630514994,.0102927775821281,-.0150885710584479,-.0300427271887783,-.0345890485685456,-.0271180576544867,-.010223616307811,.010223616307811,.0271180576544867,.0345890485685456,.0300427271887783,.0150885710584479,-.0185445134519897,-.0369238250811581,-.0425114528066451,-.0333292783668801,-.0125652713841337,.0125652713841337,.0333292783668801,.0425114528066451,.0369238250811581,.0185445134519897,-.0203528790176253,-.0405244465804058,-.0466569510209666,-.0365793781595865,-.0137905720183488,.0137905720183489,.0365793781595865,.0466569510209666,.0405244465804058,.0203528790176254,-.0203528790176253,-.0405244465804058,-.0466569510209666,-.0365793781595865,-.0137905720183489,.0137905720183488,.0365793781595865,.0466569510209666,.0405244465804058,.0203528790176254,-.0185445134519897,-.0369238250811581,-.0425114528066451,-.0333292783668801,-.0125652713841337,.0125652713841337,.0333292783668801,.0425114528066451,.0369238250811581,.0185445134519897,-.0150885710584479,-.0300427271887783,-.0345890485685456,-.0271180576544867,-.010223616307811,.010223616307811,.0271180576544867,.0345890485685456,.0300427271887783,.0150885710584479,-.0102927775821281,-.0204938630514994,-.0235951689735483,-.0184987786328967,-.00697411360782243,.00697411360782243,.0184987786328967,.0235951689735483,.0204938630514994,.0102927775821281,-.00459168398220963,-.00914246343674623,-.0105259788787713,-.00825244156510291,-.00311120351019257,.00311120351019257,.00825244156510291,.0105259788787713,.00914246343674623,.00459168398220963},
			{.00774550779202725,.0154220155529008,.0177558063098679,.0139206771836668,.00524815103220206,-.00524815103220206,-.0139206771836668,-.0177558063098679,-.0154220155529008,-.00774550779202725,.0154220155529008,.030706645722922,.0353534369104813,.0277173434974017,.0104495494699397,-.0104495494699397,-.0277173434974017,-.0353534369104813,-.030706645722922,-.0154220155529008,.0177558063098679,.0353534369104813,.0407034201215399,.0319117680095566,.0120308643041625,-.0120308643041625,-.0319117680095566,-.0407034201215399,-.0353534369104813,-.0177558063098679,.0139206771836668,.0277173434974017,.0319117680095566,.025019050842778,.0094322823360418,-.0094322823360418,-.025019050842778,-.0319117680095566,-.0277173434974017,-.0139206771836669,.00524815103220206,.0104495494699397,.0120308643041625,.0094322823360418,.00355600820454337,-.00355600820454337,-.0094322823360418,-.0120308643041625,-.0104495494699397,-.00524815103220206,-.00524815103220206,-.0104495494699397,-.0120308643041625,-.0094322823360418,-.00355600820454337,.00355600820454337,.0094322823360418,.0120308643041625,.0104495494699397,.00524815103220206,-.0139206771836668,-.0277173434974017,-.0319117680095566,-.025019050842778,-.0094322823360418,.0094322823360418,.025019050842778,.0319117680095566,.0277173434974017,.0139206771836669,-.0177558063098679,-.0353534369104813,-.0407034201215399,-.0319117680095566,-.0120308643041625,.0120308643041625,.0319117680095566,.0407034201215399,.0353534369104813,.0177558063098679,-.0154220155529008,-.030706645722922,-.0353534369104813,-.0277173434974017,-.0104495494699397,.0104495494699397,.0277173434974017,.0353534369104813,.030706645722922,.0154220155529008,-.00774550779202725,-.0154220155529008,-.0177558063098679,-.0139206771836669,-.00524815103220206,.00524815103220206,.0139206771836669,.0177558063098679,.0154220155529008,.00774550779202725},
			{-.00947407369825788,-.0188637486201043,-.0217183717412093,-.0170273563863006,-.00641938152974948,.00641938152974948,.0170273563863006,.0217183717412093,.0188637486201043,.00947407369825788,-.0143271003056479,-.0285265691642717,-.0328434525972573,-.0257494981204761,-.00970766388421236,.00970766388421236,.0257494981204761,.0328434525972573,.0285265691642717,.0143271003056479,-.00649126364804426,-.0129247005652958,-.0148805763464081,-.0116664766448895,-.00439830840398169,.00439830840398169,.0116664766448895,.0148805763464081,.0129247005652958,.00649126364804426,.00905023733181398,.0180198515883421,.020746769022353,.0162656130126723,.00613220122811362,-.00613220122811362,-.0162656130126722,-.020746769022353,-.0180198515883421,-.00905023733181398,.0212422004934682,.0422951671064499,.0486956320598683,.0381777184505166,.0143931527072752,-.0143931527072752,-.0381777184505166,-.0486956320598683,-.0422951671064499,-.0212422004934682,.0212422004934682,.0422951671064499,.0486956320598684,.0381777184505167,.0143931527072752,-.0143931527072752,-.0381777184505167,-.0486956320598684,-.0422951671064499,-.0212422004934682,.00905023733181396,.018019851588342,.020746769022353,.0162656130126722,.0061322012281136,-.0061322012281136,-.0162656130126722,-.020746769022353,-.018019851588342,-.00905023733181396,-.00649126364804423,-.0129247005652958,-.0148805763464081,-.0116664766448895,-.00439830840398167,.00439830840398167,.0116664766448895,.0148805763464081,.0129247005652958,.00649126364804423,-.0143271003056479,-.0285265691642717,-.0328434525972573,-.0257494981204761,-.00970766388421236,.00970766388421236,.0257494981204761,.0328434525972573,.0285265691642717,.0143271003056479,-.00947407369825788,-.0188637486201043,-.0217183717412093,-.0170273563863006,-.00641938152974948,.00641938152974948,.0170273563863006,.0217183717412093,.0188637486201043,.00947407369825788},
			{.00530201996819294,.010556807456243,.0121543534825683,.00952909872152508,.00359250836949255,-.00359250836949255,-.00952909872152508,-.0121543534825683,-.010556807456243,-.00530201996819294,.0118850758196888,.0236642747073748,.0272453543265051,.0213605496542905,.00805301274050799,-.00805301274050799,-.0213605496542905,-.0272453543265051,-.0236642747073748,-.0118850758196888,.017422781130702,.0346903532714859,.039939993018198,.0313132357844681,.0118052152589717,-.0118052152589717,-.0313132357844681,-.039939993018198,-.0346903532714859,-.017422781130702,.0214133596745384,.0426359607152567,.0490879974604127,.0384853356875881,.0145091256372526,-.0145091256372526,-.0384853356875881,-.0490879974604127,-.0426359607152567,-.0214133596745384,.0235014803675002,.0467936003004029,.0538748064820265,.0422382276693348,.0159239809397558,-.0159239809397558,-.0422382276693348,-.0538748064820265,-.0467936003004029,-.0235014803675002,.0235014803675002,.0467936003004029,.0538748064820265,.0422382276693348,.0159239809397558,-.0159239809397558,-.0422382276693348,-.0538748064820265,-.0467936003004029,-.0235014803675002,.0214133596745384,.0426359607152567,.0490879974604127,.0384853356875881,.0145091256372526,-.0145091256372526,-.0384853356875881,-.0490879974604127,-.0426359607152567,-.0214133596745384,.017422781130702,.0346903532714859,.039939993018198,.0313132357844681,.0118052152589717,-.0118052152589717,-.0313132357844681,-.039939993018198,-.0346903532714859,-.017422781130702,.0118850758196888,.0236642747073748,.0272453543265051,.0213605496542905,.00805301274050799,-.00805301274050799,-.0213605496542905,-.0272453543265051,-.0236642747073748,-.0118850758196888,.00530201996819294,.010556807456243,.0121543534825683,.00952909872152508,.00359250836949255,-.00359250836949255,-.00952909872152508,-.0121543534825683,-.010556807456243,-.00530201996819294},
			{-.00894374201432336,-.0178078096555534,-.0205026390981384,-.0160742134329204,-.00606004282024417,.00606004282024417,.0160742134329204,.0205026390981384,.0178078096555534,.00894374201432336,-.0178078096555534,-.0354569803355862,-.0408226326197064,-.0320052314476158,-.0120661003945087,.0120661003945087,.0320052314476158,.0408226326197064,.0354569803355862,.0178078096555534,-.0205026390981384,-.0408226326197064,-.0470002611116591,-.0368485356882857,-.0138920454842895,.0138920454842895,.0368485356882857,.0470002611116591,.0408226326197064,.0205026390981384,-.0160742134329204,-.0320052314476158,-.0368485356882857,-.0288895114677148,-.0108914614877351,.0108914614877351,.0288895114677148,.0368485356882857,.0320052314476158,.0160742134329204,-.00606004282024417,-.0120661003945087,-.0138920454842895,-.0108914614877351,-.00410612458682053,.00410612458682053,.0108914614877351,.0138920454842895,.0120661003945087,.00606004282024417,.00606004282024417,.0120661003945087,.0138920454842895,.0108914614877351,.00410612458682053,-.00410612458682053,-.0108914614877351,-.0138920454842895,-.0120661003945087,-.00606004282024417,.0160742134329204,.0320052314476158,.0368485356882857,.0288895114677148,.0108914614877351,-.0108914614877351,-.0288895114677148,-.0368485356882857,-.0320052314476158,-.0160742134329204,.0205026390981384,.0408226326197064,.0470002611116591,.0368485356882857,.0138920454842895,-.0138920454842895,-.0368485356882857,-.0470002611116591,-.0408226326197064,-.0205026390981384,.0178078096555534,.0354569803355862,.0408226326197064,.0320052314476158,.0120661003945087,-.0120661003945087,-.0320052314476158,-.0408226326197064,-.0354569803355862,-.0178078096555534,.00894374201432336,.0178078096555534,.0205026390981384,.0160742134329204,.00606004282024417,-.00606004282024417,-.0160742134329204,-.0205026390981384,-.0178078096555534,-.00894374201432336},
			{-.00592783853288971,-.0118028695476119,-.0135889803034609,-.0106538562517747,-.00401654646156772,.00401654646156772,.0106538562517747,.0135889803034609,.0118028695476119,.00592783853288971,-.01328791872398,-.026457463439368,-.0304612321695831,-.0238818205295374,-.00900354195484198,.00900354195484198,.0238818205295374,.0304612321695831,.026457463439368,.01328791872398,-.0194792614807672,-.0387849940354634,-.044654269700405,-.0350092619013457,-.0131986319027464,.0131986319027464,.0350092619013457,.044654269700405,.0387849940354634,.0194792614807672,-.0239408639271153,-.0476684532179932,-.0548820495950343,-.0430279133629583,-.016221695607902,.016221695607902,.0430279133629583,.0548820495950343,.0476684532179932,.0239408639271153,-.0262754538342299,-.0523168355867033,-.060233864778329,-.047223774154247,-.0178035519251087,.0178035519251087,.047223774154247,.060233864778329,.0523168355867033,.0262754538342299,-.0262754538342299,-.0523168355867033,-.060233864778329,-.047223774154247,-.0178035519251087,.0178035519251087,.047223774154247,.060233864778329,.0523168355867033,.0262754538342299,-.0239408639271153,-.0476684532179932,-.0548820495950343,-.0430279133629583,-.016221695607902,.016221695607902,.0430279133629583,.0548820495950343,.0476684532179932,.0239408639271153,-.0194792614807672,-.0387849940354634,-.044654269700405,-.0350092619013457,-.0131986319027464,.0131986319027464,.0350092619013457,.044654269700405,.0387849940354634,.0194792614807672,-.01328791872398,-.026457463439368,-.0304612321695831,-.0238818205295374,-.00900354195484198,.00900354195484198,.0238818205295374,.0304612321695831,.026457463439368,.01328791872398,-.00592783853288971,-.0118028695476119,-.0135889803034609,-.0106538562517747,-.00401654646156772,.00401654646156772,.0106538562517747,.0135889803034609,.0118028695476119,.00592783853288971},
			{-.00561641065141988,-.00849337690664137,-.00384814425717993,.00536515241143043,.0125927795065653,.0125927795065653,.00536515241143042,-.00384814425717991,-.00849337690664137,-.00561641065141988,-.0125898179990038,-.0190388623781496,-.00862604941815623,.0120265943124634,.0282281357131822,.0282281357131822,.0120265943124634,-.00862604941815619,-.0190388623781496,-.0125898179990038,-.0184558892850007,-.0279097867968623,-.0126452513484336,.0176302384294998,.041380689338455,.0413806893384551,.0176302384294998,-.0126452513484336,-.0279097867968623,-.0184558892850007,-.0226830947601565,-.0343023480945728,-.0155415667147304,.0216683337640905,.050858676220366,.050858676220366,.0216683337640904,-.0155415667147303,-.0343023480945728,-.0226830947601565,-.0248950334876143,-.0376473366420087,-.0170571003606095,.0237813182187718,.0558181527269258,.0558181527269258,.0237813182187717,-.0170571003606094,-.0376473366420086,-.0248950334876143,-.0248950334876143,-.0376473366420087,-.0170571003606095,.0237813182187718,.0558181527269258,.0558181527269258,.0237813182187717,-.0170571003606094,-.0376473366420086,-.0248950334876143,-.0226830947601565,-.0343023480945728,-.0155415667147304,.0216683337640905,.050858676220366,.050858676220366,.0216683337640904,-.0155415667147303,-.0343023480945728,-.0226830947601565,-.0184558892850007,-.0279097867968623,-.0126452513484336,.0176302384294998,.041380689338455,.0413806893384551,.0176302384294998,-.0126452513484336,-.0279097867968623,-.0184558892850007,-.0125898179990038,-.0190388623781496,-.00862604941815623,.0120265943124634,.0282281357131822,.0282281357131822,.0120265943124634,-.00862604941815619,-.0190388623781496,-.0125898179990038,-.00561641065141988,-.00849337690664138,-.00384814425717993,.00536515241143043,.0125927795065653,.0125927795065653,.00536515241143042,-.00384814425717991,-.00849337690664138,-.00561641065141988},
			{.00947407369825788,.0143271003056479,.00649126364804426,-.00905023733181398,-.0212422004934682,-.0212422004934682,-.00905023733181396,.00649126364804423,.0143271003056479,.00947407369825788,.0188637486201043,.0285265691642717,.0129247005652958,-.0180198515883421,-.0422951671064499,-.0422951671064499,-.018019851588342,.0129247005652958,.0285265691642717,.0188637486201043,.0217183717412093,.0328434525972573,.0148805763464081,-.020746769022353,-.0486956320598683,-.0486956320598684,-.020746769022353,.0148805763464081,.0328434525972573,.0217183717412093,.0170273563863006,.0257494981204761,.0116664766448895,-.0162656130126723,-.0381777184505166,-.0381777184505167,-.0162656130126722,.0116664766448895,.0257494981204761,.0170273563863006,.00641938152974948,.00970766388421236,.00439830840398169,-.00613220122811362,-.0143931527072752,-.0143931527072752,-.0061322012281136,.00439830840398167,.00970766388421236,.00641938152974948,-.00641938152974948,-.00970766388421236,-.00439830840398169,.00613220122811362,.0143931527072752,.0143931527072752,.0061322012281136,-.00439830840398167,-.00970766388421236,-.00641938152974948,-.0170273563863006,-.0257494981204761,-.0116664766448895,.0162656130126722,.0381777184505166,.0381777184505167,.0162656130126722,-.0116664766448895,-.0257494981204761,-.0170273563863006,-.0217183717412093,-.0328434525972573,-.0148805763464081,.020746769022353,.0486956320598683,.0486956320598684,.020746769022353,-.0148805763464081,-.0328434525972573,-.0217183717412093,-.0188637486201043,-.0285265691642717,-.0129247005652958,.0180198515883421,.0422951671064499,.0422951671064499,.018019851588342,-.0129247005652958,-.0285265691642717,-.0188637486201043,-.00947407369825788,-.0143271003056479,-.00649126364804426,.00905023733181398,.0212422004934682,.0212422004934682,.00905023733181396,-.00649126364804423,-.0143271003056479,-.00947407369825788},
			{-.0115884038691255,-.0175244810103018,-.00793991973994695,.0110699798896205,.0259828249417684,.0259828249417684,.0110699798896205,-.00793991973994691,-.0175244810103018,-.0115884038691255,-.0175244810103018,-.0265012712836701,-.0120070869359958,.0167404980488237,.0392923415017616,.0392923415017616,.0167404980488237,-.0120070869359958,-.02650127128367,-.0175244810103018,-.00793991973994695,-.0120070869359958,-.00544012153776933,.00758471596598284,.0178024123929934,.0178024123929934,.00758471596598282,-.00544012153776931,-.0120070869359958,-.00793991973994695,.0110699798896205,.0167404980488237,.00758471596598284,-.0105747483553876,-.0248204457515694,-.0248204457515694,-.0105747483553876,.00758471596598281,.0167404980488237,.0110699798896205,.0259828249417684,.0392923415017616,.0178024123929934,-.0248204457515694,-.0582571335603207,-.0582571335603207,-.0248204457515693,.0178024123929933,.0392923415017616,.0259828249417684,.0259828249417684,.0392923415017616,.0178024123929934,-.0248204457515694,-.0582571335603207,-.0582571335603207,-.0248204457515693,.0178024123929933,.0392923415017616,.0259828249417684,.0110699798896205,.0167404980488237,.00758471596598282,-.0105747483553876,-.0248204457515693,-.0248204457515693,-.0105747483553876,.00758471596598279,.0167404980488237,.0110699798896205,-.00793991973994691,-.0120070869359958,-.00544012153776931,.00758471596598281,.0178024123929933,.0178024123929933,.00758471596598279,-.00544012153776928,-.0120070869359958,-.00793991973994691,-.0175244810103018,-.02650127128367,-.0120070869359958,.0167404980488237,.0392923415017616,.0392923415017616,.0167404980488237,-.0120070869359958,-.0265012712836701,-.0175244810103018,-.0115884038691255,-.0175244810103018,-.00793991973994695,.0110699798896205,.0259828249417684,.0259828249417684,.0110699798896205,-.00793991973994691,-.0175244810103018,-.0115884038691255},
			{.00648527240266935,.009807306886327,.004443454245332,-.00619514437769402,-.0145408892752851,-.0145408892752851,-.00619514437769401,.00444345424533198,.009807306886327,.00648527240266935,.0145374696209095,.0219841846372146,.00996050390679461,-.0138871149268619,-.0325950435040259,-.0325950435040259,-.0138871149268619,.00996050390679457,.0219841846372146,.0145374696209095,.0213110252927245,.032227445838975,.0146014785393375,-.020357645805406,-.0477823042555251,-.0477823042555251,-.020357645805406,.0146014785393374,.032227445838975,.0213110252927245,.0261921817305127,.0396089398107403,.0179458554519699,-.0250204366620803,-.0587265408103395,-.0587265408103396,-.0250204366620803,.0179458554519699,.0396089398107403,.0261921817305127,.0287463085698583,.0434713998871676,.0196958429687218,-.0274603009493809,-.0644532510022925,-.0644532510022925,-.0274603009493808,.0196958429687217,.0434713998871676,.0287463085698583,.0287463085698583,.0434713998871676,.0196958429687218,-.0274603009493809,-.0644532510022925,-.0644532510022925,-.0274603009493808,.0196958429687217,.0434713998871676,.0287463085698583,.0261921817305127,.0396089398107403,.0179458554519699,-.0250204366620803,-.0587265408103395,-.0587265408103396,-.0250204366620803,.0179458554519699,.0396089398107403,.0261921817305127,.0213110252927245,.032227445838975,.0146014785393375,-.020357645805406,-.0477823042555251,-.0477823042555251,-.020357645805406,.0146014785393374,.032227445838975,.0213110252927245,.0145374696209095,.0219841846372146,.00996050390679461,-.0138871149268619,-.0325950435040259,-.0325950435040259,-.0138871149268619,.00996050390679457,.0219841846372146,.0145374696209095,.00648527240266935,.009807306886327,.004443454245332,-.00619514437769402,-.0145408892752851,-.0145408892752851,-.00619514437769401,.00444345424533198,.009807306886327,.00648527240266935},
			{-.0109397180005024,-.01654351043707,-.00749546562948678,.0104503139199635,.0245283803472424,.0245283803472424,.0104503139199635,-.00749546562948674,-.01654351043707,-.0109397180005024,-.0217819806884396,-.0329396447735407,-.0149241587017916,.0208075323314845,.0488382522307979,.0488382522307979,.0208075323314844,-.0149241587017915,-.0329396447735407,-.0217819806884396,-.0250782155433939,-.037924352397948,-.0171826095193439,.0239563053607907,.0562288725587115,.0562288725587116,.0239563053607906,-.0171826095193438,-.037924352397948,-.0250782155433939,-.0196614975872982,-.029732959344012,-.0134712868634331,.0187819121036241,.0440838320508345,.0440838320508346,.0187819121036241,-.0134712868634331,-.029732959344012,-.0196614975872982,-.00741246330878832,-.011209444713996,-.00507872908225817,.00708085605986294,.0166197811807936,.0166197811807936,.00708085605986292,-.00507872908225814,-.011209444713996,-.00741246330878832,.00741246330878832,.011209444713996,.00507872908225817,-.00708085605986294,-.0166197811807936,-.0166197811807936,-.00708085605986292,.00507872908225814,.011209444713996,.00741246330878832,.0196614975872982,.029732959344012,.0134712868634331,-.0187819121036241,-.0440838320508345,-.0440838320508346,-.0187819121036241,.0134712868634331,.029732959344012,.0196614975872982,.0250782155433939,.037924352397948,.0171826095193439,-.0239563053607907,-.0562288725587116,-.0562288725587116,-.0239563053607906,.0171826095193438,.037924352397948,.0250782155433939,.0217819806884396,.0329396447735407,.0149241587017916,-.0208075323314845,-.0488382522307979,-.0488382522307979,-.0208075323314844,.0149241587017915,.0329396447735407,.0217819806884396,.0109397180005024,.01654351043707,.00749546562948678,-.0104503139199635,-.0245283803472424,-.0245283803472424,-.0104503139199634,.00749546562948674,.01654351043707,.0109397180005024},
			{-.00725075497095922,-.0109649024347056,-.00496793287269008,.00692638197801623,.0162572084329942,.0162572084329943,.00692638197801622,-.00496793287269006,-.0109649024347056,-.00725075497095922,-.0162533851431734,-.0245790656341836,-.0111361819105275,.0155262664906382,.0364423664946087,.0364423664946087,.0155262664906382,-.0111361819105275,-.0245790656341835,-.0162533851431734,-.0238264506073575,-.0360313798109832,-.0163249492895439,.022760539936583,.0534222402072165,.0534222402072165,.0227605399365829,-.0163249492895438,-.0360313798109832,-.0238264506073575,-.0292837494080609,-.0442841409574315,-.0200640763482701,.0279736985956793,.0656582686538416,.0656582686538417,.0279736985956792,-.02006407634827,-.0442841409574315,-.0292837494080609,-.0321393500254263,-.0486025026021575,-.0220206218714746,.030701549796244,.0720609252908172,.0720609252908173,.0307015497962439,-.0220206218714745,-.0486025026021575,-.0321393500254263,-.0321393500254263,-.0486025026021575,-.0220206218714746,.030701549796244,.0720609252908172,.0720609252908173,.0307015497962439,-.0220206218714745,-.0486025026021575,-.0321393500254263,-.0292837494080609,-.0442841409574315,-.0200640763482701,.0279736985956793,.0656582686538416,.0656582686538416,.0279736985956792,-.02006407634827,-.0442841409574315,-.0292837494080609,-.0238264506073575,-.0360313798109832,-.0163249492895439,.022760539936583,.0534222402072165,.0534222402072165,.0227605399365829,-.0163249492895438,-.0360313798109832,-.0238264506073575,-.0162533851431734,-.0245790656341835,-.0111361819105275,.0155262664906383,.0364423664946087,.0364423664946087,.0155262664906382,-.0111361819105275,-.0245790656341836,-.0162533851431734,-.00725075497095922,-.0109649024347056,-.00496793287269008,.00692638197801623,.0162572084329942,.0162572084329943,.00692638197801622,-.00496793287269006,-.0109649024347056,-.00725075497095922},
			{.00705610400811139,.00598003257676846,-.0064239867066964,-.0149987461209001,-.00792740615413668,.00792740615413675,.0149987461209001,.00642398670669633,-.0059800325767685,-.00705610400811139,.0158170530535735,.0134049175607548,-.014400090820439,-.033621381269131,-.0177701750956426,.0177701750956428,.0336213812691311,.0144000908204389,-.0134049175607549,-.0158170530535735,.023186814932101,.0196507744905787,-.0211096365250878,-.0492868515145319,-.0260499702351701,.0260499702351703,.049286851514532,.0211096365250876,-.0196507744905789,-.023186814932101,.0284976092004682,.0241516609141409,-.0259446661310533,-.0605756951653628,-.0320165522353802,.0320165522353805,.0605756951653629,.025944666131053,-.024151660914141,-.0284976092004682,.0312765494684072,.0265068066591675,-.0284746565222694,-.0664827253786431,-.0351386417279264,.0351386417279267,.0664827253786433,.0284746565222691,-.0265068066591676,-.0312765494684072,.0312765494684072,.0265068066591675,-.0284746565222694,-.0664827253786431,-.0351386417279264,.0351386417279267,.0664827253786433,.0284746565222691,-.0265068066591676,-.0312765494684072,.0284976092004682,.0241516609141409,-.0259446661310533,-.0605756951653627,-.0320165522353802,.0320165522353805,.0605756951653629,.025944666131053,-.024151660914141,-.0284976092004682,.023186814932101,.0196507744905787,-.0211096365250878,-.0492868515145319,-.0260499702351701,.0260499702351703,.049286851514532,.0211096365250876,-.0196507744905789,-.023186814932101,.0158170530535735,.0134049175607548,-.014400090820439,-.033621381269131,-.0177701750956426,.0177701750956428,.0336213812691311,.0144000908204389,-.0134049175607549,-.0158170530535735,.00705610400811139,.00598003257676846,-.0064239867066964,-.0149987461209001,-.00792740615413669,.00792740615413675,.0149987461209001,.00642398670669633,-.0059800325767685,-.00705610400811139},
			{-.0119026284826952,-.010087451374564,.0108363378798955,.0253007187221328,.0133723893774315,-.0133723893774316,-.0253007187221329,-.0108363378798954,.010087451374564,.0119026284826952,-.0236992236673589,-.0200850397630209,.0215761413980188,.0503760486864314,.0266256522484521,-.0266256522484523,-.0503760486864316,-.0215761413980185,.020085039763021,.0236992236673589,-.0272855920608065,-.0231244790627329,.0248412268982238,.0579993814728395,.0306548727418767,-.0306548727418769,-.0579993814728396,-.0248412268982236,.0231244790627331,.0272855920608065,-.0213920963213377,-.01812975442821,.0194756968206063,.0454719235074688,.0240336361018362,-.0240336361018364,-.0454719235074689,-.019475696820606,.0181297544282102,.0213920963213377,-.00806490596028758,-.00683499000520012,.00734241567585834,.017143097217421,.00906077703341984,-.00906077703341992,-.017143097217421,-.00734241567585826,.00683499000520017,.00806490596028758,.00806490596028758,.00683499000520012,-.00734241567585834,-.017143097217421,-.00906077703341984,.00906077703341992,.017143097217421,.00734241567585826,-.00683499000520017,-.00806490596028758,.0213920963213377,.01812975442821,-.0194756968206063,-.0454719235074688,-.0240336361018362,.0240336361018364,.0454719235074689,.019475696820606,-.0181297544282102,-.0213920963213377,.0272855920608065,.0231244790627329,-.0248412268982238,-.0579993814728395,-.0306548727418767,.0306548727418769,.0579993814728396,.0248412268982236,-.0231244790627331,-.0272855920608065,.0236992236673589,.0200850397630209,-.0215761413980188,-.0503760486864314,-.0266256522484521,.0266256522484523,.0503760486864316,.0215761413980185,-.020085039763021,-.0236992236673589,.0119026284826952,.010087451374564,-.0108363378798955,-.0253007187221328,-.0133723893774315,.0133723893774316,.0253007187221328,.0108363378798954,-.010087451374564,-.0119026284826952},
			{-.00788896410842497,-.00668587967398251,.00718223548063763,.0167691079501009,.00886310952205344,-.00886310952205352,-.0167691079501009,-.00718223548063755,.00668587967398256,.00788896410842497,-.0176840029139668,-.0149871534477983,.0160997909767078,.0375898470038056,.0198676597409558,-.019867659740956,-.0375898470038057,-.0160997909767076,.0149871534477984,.0176840029139668,-.0259236471823205,-.0219702337835042,.0236012911228172,.0551043751861423,.0291247521258975,-.0291247521258977,-.0551043751861425,-.023601291122817,.0219702337835044,.0259236471823205,-.0318612956810125,-.0270023777840407,.0290070185583519,.0677256860801781,.0357955935981209,-.0357955935981212,-.0677256860801783,-.0290070185583516,.0270023777840409,.0318612956810126,-.0349682453529598,-.0296355107751738,.031835633806656,.0743299466305301,.0392861957663534,-.0392861957663538,-.0743299466305303,-.0318356338066557,.029635510775174,.0349682453529598,-.0349682453529598,-.0296355107751738,.031835633806656,.0743299466305301,.0392861957663534,-.0392861957663538,-.0743299466305303,-.0318356338066557,.029635510775174,.0349682453529598,-.0318612956810125,-.0270023777840407,.0290070185583519,.0677256860801781,.035795593598121,-.0357955935981213,-.0677256860801783,-.0290070185583516,.0270023777840409,.0318612956810126,-.0259236471823205,-.0219702337835042,.0236012911228172,.0551043751861423,.0291247521258975,-.0291247521258977,-.0551043751861425,-.023601291122817,.0219702337835044,.0259236471823205,-.0176840029139668,-.0149871534477983,.0160997909767078,.0375898470038056,.0198676597409558,-.019867659740956,-.0375898470038057,-.0160997909767076,.0149871534477984,.0176840029139668,-.00788896410842497,-.00668587967398251,.00718223548063763,.0167691079501009,.00886310952205344,-.00886310952205352,-.0167691079501009,-.00718223548063755,.00668587967398256,.00788896410842497},
			{-.00794977455310546,-.000443472655919539,.0146816998664313,.00745176669518039,-.013740219487093,-.013740219487093,.00745176669518016,.0146816998664309,-.000443472655919117,-.00794977455310563,-.0178203163850575,-.000994093981384771,.0329106863273325,.0167039755968772,-.0308002518592783,-.0308002518592785,.0167039755968767,.0329106863273315,-.000994093981383826,-.0178203163850579,-.0261234742434189,-.00145727987972302,.0482450169810793,.0244869881565972,-.0451512514565309,-.0451512514565312,.0244869881565965,.0482450169810779,-.00145727987972163,-.0261234742434194,-.0321068918748639,-.0017910606795139,.0592952349782769,.0300955789325382,-.05549286189967,-.0554928618996703,.0300955789325372,.0592952349782752,-.0017910606795122,-.0321068918748646,-.0352377908243784,-.00196571570441829,.0650774013003321,.0330303449893761,-.0609042403571174,-.0609042403571177,.0330303449893751,.0650774013003302,-.00196571570441642,-.0352377908243792,-.0352377908243784,-.00196571570441829,.0650774013003321,.0330303449893761,-.0609042403571174,-.0609042403571177,.0330303449893751,.0650774013003302,-.00196571570441642,-.0352377908243792,-.0321068918748639,-.0017910606795139,.0592952349782769,.0300955789325382,-.05549286189967,-.0554928618996703,.0300955789325373,.0592952349782752,-.0017910606795122,-.0321068918748646,-.0261234742434189,-.00145727987972302,.0482450169810793,.0244869881565972,-.0451512514565309,-.0451512514565312,.0244869881565965,.0482450169810779,-.00145727987972163,-.0261234742434194,-.0178203163850575,-.000994093981384771,.0329106863273325,.0167039755968772,-.0308002518592783,-.0308002518592785,.0167039755968767,.0329106863273315,-.000994093981383826,-.0178203163850579,-.00794977455310546,-.000443472655919539,.0146816998664313,.00745176669518039,-.013740219487093,-.013740219487093,.00745176669518016,.0146816998664309,-.000443472655919117,-.00794977455310563}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_right_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{.00265100998409647,.00594253790984439,.008711390565351,.0107066798372692,.0117507401837501,.0117507401837501,.0107066798372692,.008711390565351,.00594253790984439,.00265100998409647,.00527840372812152,.0118321373536874,.0173451766357429,.0213179803576284,.0233968001502014,.0233968001502014,.0213179803576284,.0173451766357429,.0118321373536874,.00527840372812152,.00607717674128413,.0136226771632526,.019969996509099,.0245439987302064,.0269374032410133,.0269374032410133,.0245439987302064,.019969996509099,.0136226771632526,.00607717674128413,.00476454936076254,.0106802748271453,.0156566178922341,.0192426678437941,.0211191138346674,.0211191138346674,.0192426678437941,.0156566178922341,.0106802748271453,.00476454936076254,.00179625418474627,.00402650637025399,.00590260762948584,.00725456281862631,.00796199046987792,.00796199046987792,.00725456281862631,.00590260762948584,.00402650637025399,.00179625418474628,-.00179625418474627,-.00402650637025399,-.00590260762948584,-.00725456281862631,-.00796199046987792,-.00796199046987792,-.00725456281862631,-.00590260762948584,-.00402650637025399,-.00179625418474628,-.00476454936076254,-.0106802748271453,-.0156566178922341,-.0192426678437941,-.0211191138346674,-.0211191138346674,-.0192426678437941,-.0156566178922341,-.0106802748271453,-.00476454936076254,-.00607717674128413,-.0136226771632526,-.019969996509099,-.0245439987302064,-.0269374032410133,-.0269374032410133,-.0245439987302064,-.019969996509099,-.0136226771632526,-.00607717674128413,-.00527840372812152,-.0118321373536874,-.0173451766357429,-.0213179803576284,-.0233968001502014,-.0233968001502014,-.0213179803576284,-.0173451766357429,-.0118321373536874,-.00527840372812152,-.00265100998409647,-.00594253790984439,-.008711390565351,-.0107066798372692,-.0117507401837501,-.0117507401837501,-.0107066798372692,-.008711390565351,-.00594253790984439,-.00265100998409647},
			{-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.002221727122666,-.00498025195339731,-.00730073926966875,-.00897292772598497,-.00984792148436088,-.00984792148436088,-.00897292772598497,-.00730073926966875,-.00498025195339731,-.002221727122666,.00309757218884701,.00694355746343096,.010178822902703,.0125102183310402,.0137301504746905,.0137301504746905,.0125102183310402,.010178822902703,.00694355746343096,.00309757218884701,.00727044463764254,.0162975217520129,.0238911521277625,.0293632704051698,.0322266255011462,.0322266255011462,.0293632704051698,.0238911521277625,.0162975217520129,.00727044463764254,.00727044463764255,.016297521752013,.0238911521277625,.0293632704051698,.0322266255011463,.0322266255011463,.0293632704051698,.0238911521277625,.016297521752013,.00727044463764255,.003097572188847,.00694355746343095,.010178822902703,.0125102183310401,.0137301504746904,.0137301504746904,.0125102183310401,.010178822902703,.00694355746343095,.003097572188847,-.00222172712266599,-.00498025195339728,-.00730073926966871,-.00897292772598493,-.00984792148436084,-.00984792148436084,-.00897292772598493,-.00730073926966871,-.00498025195339728,-.00222172712266599,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468},
			{.00352805200405569,.00790852652678675,.0115934074660505,.0142488046002341,.0156382747342036,.0156382747342036,.0142488046002341,.0115934074660505,.00790852652678675,.00352805200405569,.00299001628838423,.00670245878037739,.00982538724528936,.0120758304570704,.0132534033295837,.0132534033295837,.0120758304570704,.00982538724528936,.00670245878037739,.00299001628838423,-.0032119933533482,-.00720004541021951,-.0105548182625439,-.0129723330655266,-.0142373282611347,-.0142373282611347,-.0129723330655266,-.0105548182625439,-.00720004541021951,-.0032119933533482,-.00749937306045005,-.0168106906345655,-.0246434257572659,-.0302878475826814,-.0332413626893215,-.0332413626893215,-.0302878475826814,-.0246434257572659,-.0168106906345655,-.00749937306045005,-.00396370307706834,-.0088850875478213,-.0130249851175851,-.0160082761176901,-.0175693208639632,-.0175693208639632,-.0160082761176901,-.0130249851175851,-.0088850875478213,-.00396370307706834,.00396370307706838,.00888508754782138,.0130249851175852,.0160082761176902,.0175693208639633,.0175693208639633,.0160082761176902,.0130249851175852,.00888508754782138,.00396370307706837,.00749937306045007,.0168106906345656,.024643425757266,.0302878475826815,.0332413626893216,.0332413626893216,.0302878475826815,.024643425757266,.0168106906345656,.00749937306045007,.00321199335334816,.00720004541021943,.0105548182625438,.0129723330655265,.0142373282611345,.0142373282611345,.0129723330655265,.0105548182625438,.00720004541021943,.00321199335334816,-.00299001628838425,-.00670245878037744,-.00982538724528943,-.0120758304570705,-.0132534033295838,-.0132534033295838,-.0120758304570705,-.00982538724528943,-.00670245878037744,-.00299001628838425,-.0035280520040557,-.00790852652678675,-.0115934074660505,-.0142488046002341,-.0156382747342036,-.0156382747342036,-.0142488046002341,-.0115934074660505,-.00790852652678675,-.0035280520040557},
			{-.00355524726055973,-.00796948776183025,-.0116827728408898,-.0143586385526627,-.0157588191287273,-.0157588191287273,-.0143586385526627,-.0116827728408898,-.00796948776183025,-.00355524726055973,-.000198327000917931,-.000444572343586338,-.000651715374523445,-.000800986686075343,-.000879094787718522,-.000879094787718522,-.000800986686075343,-.000651715374523445,-.000444572343586338,-.000198327000917931,.00656585578393543,.0147181063597184,.0215758275045217,.0265176352250662,.0291034986151868,.0291034986151868,.0265176352250662,.0215758275045218,.0147181063597184,.00656585578393543,.00333253137587672,.00747024498425,.0109509140141708,.0134591520602391,.0147716193401924,.0147716193401924,.0134591520602391,.0109509140141708,.00747024498425,.00333253137587672,-.00614481295848751,-.0137742913733916,-.020192253500946,-.0248171622895082,-.0272372043055647,-.0272372043055647,-.0248171622895082,-.020192253500946,-.0137742913733916,-.00614481295848751,-.00614481295848755,-.0137742913733917,-.0201922535009461,-.0248171622895084,-.0272372043055649,-.0272372043055649,-.0248171622895084,-.0201922535009461,-.0137742913733917,-.00614481295848755,.00333253137587662,.00747024498424977,.0109509140141705,.0134591520602387,.014771619340192,.014771619340192,.0134591520602387,.0109509140141704,.00747024498424977,.00333253137587662,.00656585578393523,.014718106359718,.0215758275045211,.0265176352250655,.0291034986151859,.0291034986151859,.0265176352250655,.0215758275045211,.014718106359718,.00656585578393523,-.000198327000917742,-.000444572343585915,-.000651715374522825,-.000800986686074581,-.000879094787717686,-.000879094787717686,-.000800986686074581,-.000651715374522825,-.000444572343585915,-.000198327000917742,-.0035552472605598,-.00796948776183042,-.0116827728408901,-.014358638552663,-.0157588191287277,-.0157588191287277,-.014358638552663,-.0116827728408901,-.00796948776183042,-.0035552472605598},
			{-.00435807799231643,-.00895572847242053,-.0110946867695545,-.0103234510778548,-.00712570259555149,-.0027257963949736,.00134726370928414,.00379129532198382,.00397366731673714,.00213554377116887,-.00976912340532947,-.0200752755655013,-.0248699918602522,-.0231411800629203,-.015973066137951,-.00611018008563575,.0030200435739616,.00849861611742791,.00890742351499224,.00478706224964608,-.0143209266404436,-.0294291039975391,-.0364578595439588,-.0339235290930591,-.0234155203996869,-.00895714355686739,.0044271958372491,.0124584420641652,.0130577282547148,.00701753519287297,-.0176010449034518,-.0361696553536002,-.0448083031935061,-.0416935002770177,-.0287786982183586,-.0110087210073302,.00544121722596732,.0153119699376961,.0160485188646415,.00862485753488115,-.0193174082690916,-.0396967341001901,-.049177778443395,-.0457592359678989,-.0315850488415031,-.012082234855126,.00597181674221008,.0168051144868407,.0176134878766033,.00946590927856652,-.0193174082690916,-.0396967341001901,-.049177778443395,-.0457592359678989,-.0315850488415031,-.012082234855126,.00597181674221008,.0168051144868407,.0176134878766033,.00946590927856652,-.0176010449034518,-.0361696553536002,-.0448083031935061,-.0416935002770177,-.0287786982183586,-.0110087210073302,.00544121722596732,.0153119699376961,.0160485188646415,.00862485753488115,-.0143209266404436,-.0294291039975391,-.0364578595439588,-.0339235290930591,-.0234155203996869,-.00895714355686739,.0044271958372491,.0124584420641652,.0130577282547148,.00701753519287297,-.00976912340532947,-.0200752755655013,-.0248699918602522,-.0231411800629204,-.015973066137951,-.00611018008563575,.0030200435739616,.00849861611742791,.00890742351499224,.00478706224964608,-.00435807799231643,-.00895572847242053,-.0110946867695545,-.0103234510778548,-.00712570259555149,-.0027257963949736,.00134726370928414,.00379129532198382,.00397366731673714,.00213554377116887},
			{.00735144822140268,.0151070206329527,.0187151343924685,.0174141711549991,.0120200312533801,.00459802488506717,-.00227263931871786,-.00639536770584407,-.00670300292458653,-.00360235394730565,.0146374068493633,.0300794619817186,.0372635470036412,.0346732100211315,.0239329833386512,.0091550887552036,-.00452503307213473,-.0127337629596822,-.0133462928615887,-.00717261670817588,.0168524639340767,.0346313423832577,.0429025843441626,.0399202555051752,.027554726236705,.0105405147679591,-.00520979962048207,-.0146607444359217,-.0153659675800657,-.00825803816423586,.0132124503996921,.0271512163030548,.0336359282466497,.0312977614352663,.0216030994104901,.00826383780458816,-.00408451958996732,-.0114941268790076,-.0120470267902114,-.00647436007402489,.00498114669912899,.0102361172510779,.0126808796165353,.0117993813673682,.00814445496968726,.00311549993810543,-.00153988023853377,-.00433333185221062,-.00454177733236969,-.00244085966918394,-.00498114669912899,-.0102361172510779,-.0126808796165353,-.0117993813673682,-.00814445496968726,-.00311549993810543,.00153988023853377,.00433333185221062,.00454177733236969,.00244085966918394,-.0132124503996921,-.0271512163030548,-.0336359282466497,-.0312977614352663,-.0216030994104901,-.00826383780458816,.00408451958996732,.0114941268790076,.0120470267902114,.00647436007402489,-.0168524639340767,-.0346313423832577,-.0429025843441627,-.0399202555051752,-.027554726236705,-.0105405147679591,.00520979962048207,.0146607444359217,.0153659675800657,.00825803816423586,-.0146374068493633,-.0300794619817185,-.0372635470036412,-.0346732100211315,-.0239329833386512,-.0091550887552036,.00452503307213473,.0127337629596822,.0133462928615887,.00717261670817588,-.00735144822140268,-.0151070206329527,-.0187151343924685,-.0174141711549991,-.0120200312533801,-.00459802488506717,.00227263931871786,.00639536770584407,.00670300292458653,.00360235394730565},
			{-.00899207180285294,-.0184784562398471,-.0228917931799034,-.0213004938206645,-.0147025430701201,-.00562416665024674,.00277982450810196,.00782262268394569,.00819891289135883,.00440629170986428,-.0135981963807978,-.0279439135132015,-.034617950789736,-.0322115196955465,-.0222338157821648,-.00850510586048974,.00420376976453103,.0118297053005546,.0123987474800167,.00666338317747926,-.00616101485733601,-.0126607133406354,-.0156845586850965,-.0145942627878117,-.0100736057586743,-.00385345836332292,.0019046267056908,.00535975419633905,.00561757348530501,.00301901822910458,.00858979848715933,.0176517958190368,.0218676957586917,.0203475854739549,.0140448035769736,.00537256143445094,-.00265546504496181,-.00747266636314624,-.00783212268480781,-.00420916988800805,.0201614847183412,.0414312876127182,.051326607314859,.0477586912197692,.0329651613029748,.0126101695425049,-.00623275598422004,-.0175394159608113,-.0183831113218753,-.00987952330905502,.0201614847183412,.0414312876127182,.0513266073148591,.0477586912197692,.0329651613029748,.0126101695425049,-.00623275598422005,-.0175394159608113,-.0183831113218753,-.00987952330905502,.00858979848715931,.0176517958190368,.0218676957586917,.0203475854739549,.0140448035769735,.00537256143445092,-.0026554650449618,-.00747266636314622,-.00783212268480779,-.00420916988800804,-.00616101485733598,-.0126607133406353,-.0156845586850964,-.0145942627878116,-.0100736057586743,-.0038534583633229,.0019046267056908,.00535975419633902,.00561757348530498,.00301901822910456,-.0135981963807978,-.0279439135132015,-.034617950789736,-.0322115196955465,-.0222338157821648,-.00850510586048974,.00420376976453103,.0118297053005546,.0123987474800167,.00666338317747926,-.00899207180285294,-.0184784562398471,-.0228917931799034,-.0213004938206645,-.0147025430701201,-.00562416665024674,.00277982450810196,.00782262268394569,.00819891289135883,.00440629170986428},
			{.00978355109164593,.0201049240576492,.0249067215059292,.0231753564852141,.0159966562164246,.00611920400292648,-.0030245037736683,-.0085111674347139,-.00892057858602882,-.00479413209917604,.00829153796163795,.0170388787750279,.0211083915167262,.0196410634821265,.0135571308449561,.00518601188975199,-.00256326027426807,-.0072131956200491,-.00756017066737565,-.00406301637524721,-.0089070969028758,-.018303835195316,-.0226754661889265,-.0210992045770331,-.0145636043300628,-.00557101838708364,.00275355522170069,.00774869905491992,.00812143332734517,.00436465234070388,-.0207963202945656,-.0427358569791504,-.0529427559658922,-.049262494966556,-.0340031531703131,-.0130072327727005,.00642901014367235,.0180916890395887,.0189619502929062,.0101906052040814,-.010991643978613,-.0225875211759695,-.0279822543883277,-.0260370968758398,-.0179719560241624,-.00687481582123123,.00339797568190965,.00956214378689379,.0100221098640343,.0053861213302785,.0109916439786131,.0225875211759697,.027982254388328,.0260370968758401,.0179719560241625,.00687481582123129,-.00339797568190968,-.00956214378689387,-.0100221098640343,-.00538612133027854,.0207963202945656,.0427358569791505,.0529427559658923,.0492624949665561,.0340031531703132,.0130072327727006,-.00642901014367236,-.0180916890395887,-.0189619502929062,-.0101906052040814,.00890709690287571,.0183038351953158,.0226754661889263,.0210992045770329,.0145636043300626,.00557101838708358,-.00275355522170066,-.00774869905491983,-.00812143332734509,-.00436465234070383,-.00829153796163801,-.017038878775028,-.0211083915167264,-.0196410634821266,-.0135571308449562,-.00518601188975203,.00256326027426809,.00721319562004915,.00756017066737571,.00406301637524725,-.00978355109164594,-.0201049240576493,-.0249067215059292,-.0231753564852141,-.0159966562164247,-.00611920400292649,.00302450377366831,.0085111674347139,.00892057858602882,.00479413209917604},
			{-.00774458111624399,-.0136429110526415,-.0119270111969813,-.00443011400826239,.00356799186187752,.00716050023137007,.00509898471326269,.000227342285586894,-.00308610359639842,-.00244256114805105,-.0173603521507787,-.0305821240271899,-.0267357409494468,-.00993060026064162,.00799805622325859,.0160510689637666,.0114299493936489,.000509613377058344,-.00691784931981508,-.0054752763310899,-.025449195315512,-.0448314896363585,-.0391929315383665,-.0145576416560195,.0117246524265364,.0235298676855081,.0167555941284487,.000747061479831429,-.0101411363648726,-.00802641418481,-.0312781735952783,-.0550998606437757,-.0481698262427054,-.0178919780059722,.0144101103942559,.0289192360315085,.0205933576816159,.000918171217707236,-.0124638999285189,-.0098648139207399,-.0343282602007917,-.0604729156402031,-.0528670999299058,-.0196367116713162,.0158153102395603,.0317392911793161,.0226015159980186,.00100770655212074,-.0136793153398003,-.0108267798332916,-.0343282602007917,-.0604729156402031,-.0528670999299058,-.0196367116713162,.0158153102395603,.0317392911793161,.0226015159980186,.00100770655212074,-.0136793153398003,-.0108267798332916,-.0312781735952783,-.0550998606437757,-.0481698262427054,-.0178919780059722,.0144101103942559,.0289192360315085,.0205933576816159,.000918171217707236,-.0124638999285189,-.0098648139207399,-.025449195315512,-.0448314896363585,-.0391929315383665,-.0145576416560195,.0117246524265364,.0235298676855081,.0167555941284487,.000747061479831429,-.0101411363648726,-.00802641418481,-.0173603521507787,-.0305821240271899,-.0267357409494468,-.00993060026064162,.00799805622325859,.0160510689637666,.0114299493936489,.000509613377058344,-.00691784931981508,-.0054752763310899,-.00774458111624399,-.0136429110526415,-.0119270111969814,-.00443011400826239,.00356799186187752,.00716050023137007,.0050989847132627,.000227342285586894,-.00308610359639842,-.00244256114805105},
			{.0130639899374604,.0230136207541996,.0201191454931664,.00747296257306306,-.0060186870123603,-.0120787298326045,-.00860125085985731,-.000383493604971994,.00520581109864628,.00412024792313705,.0260116007120718,.0458222271192026,.0400590617216945,.0148793530550254,-.0119837571925505,-.0240498575870592,-.0171258783925904,-.000763570898011854,.0103652467836164,.00820379105651837,.0299478976965691,.0527564368369676,.0461211401613239,.0171310235004581,-.0137972414076181,-.0276892868919076,-.0197175121878277,-.000879120950335205,.0119338042172612,.00944525859843065,.0234793626877832,.041361417992028,.0361592986724362,.0134308430613569,-.0108171344240126,-.0217085959117477,-.0154586684063579,-.000689237015849483,.00935618654441216,.0074051492548628,.00885181373718756,.0155934201809443,.0136322003698287,.00506348160695133,-.00407810298621525,-.00818422757303578,-.0058279798807838,-.000259845114460841,.00352731978643567,.00279177091694338,-.00885181373718756,-.0155934201809443,-.0136322003698287,-.00506348160695133,.00407810298621525,.00818422757303578,.0058279798807838,.000259845114460841,-.00352731978643567,-.00279177091694338,-.0234793626877832,-.041361417992028,-.0361592986724362,-.0134308430613569,.0108171344240126,.0217085959117477,.0154586684063579,.000689237015849483,-.00935618654441216,-.0074051492548628,-.0299478976965691,-.0527564368369676,-.0461211401613239,-.0171310235004581,.0137972414076181,.0276892868919076,.0197175121878277,.000879120950335204,-.0119338042172612,-.00944525859843065,-.0260116007120718,-.0458222271192026,-.0400590617216945,-.0148793530550254,.0119837571925505,.0240498575870592,.0171258783925904,.000763570898011854,-.0103652467836164,-.00820379105651837,-.0130639899374604,-.0230136207541996,-.0201191454931664,-.00747296257306306,.0060186870123603,.0120787298326045,.00860125085985731,.000383493604971994,-.00520581109864628,-.00412024792313705},
			{-.0159794821505739,-.028149573279088,-.0246091376242554,-.00914070453359377,.00736187812026119,.0147743414290495,.0105207930537044,.000469077919138421,-.00636759259064836,-.00503976415007148,-.0241648577892824,-.0425689912043686,-.0372149926639303,-.0138229651666283,.0111329476238222,.0223423923378182,.0159099941773837,.000709359734017749,-.00962934643082784,-.00762134735221235,-.010948514324695,-.0192869833562834,-.0168612157301903,-.00626285217385856,.0050440700954407,.0101227991776989,.00720843468957457,.000321393789153591,-.00436282465449179,-.0034530486952082,.0152646169438994,.0268902614734654,.0235082123196634,.00873178191808065,-.0070325338727842,-.0141133899326471,-.0100501301855435,-.00044809304112726,.00608272914198089,.00481430302393592,.0358282376130031,.0631152868812828,.055177134148947,.020494740123233,-.0165063621014286,-.0331261432822222,-.0235890919276015,-.00105173840976448,.0142770346504849,.0112998572657607,.0358282376130031,.0631152868812828,.0551771341489471,.020494740123233,-.0165063621014286,-.0331261432822222,-.0235890919276015,-.00105173840976448,.0142770346504849,.0112998572657607,.0152646169438994,.0268902614734653,.0235082123196633,.00873178191808063,-.00703253387278418,-.0141133899326471,-.0100501301855435,-.000448093041127259,.00608272914198088,.0048143030239359,-.0109485143246949,-.0192869833562833,-.0168612157301902,-.00626285217385853,.00504407009544068,.0101227991776988,.00720843468957453,.00032139378915359,-.00436282465449177,-.00345304869520818,-.0241648577892824,-.0425689912043686,-.0372149926639303,-.0138229651666283,.0111329476238222,.0223423923378182,.0159099941773837,.000709359734017748,-.00962934643082784,-.00762134735221235,-.0159794821505739,-.028149573279088,-.0246091376242554,-.00914070453359377,.00736187812026119,.0147743414290495,.0105207930537044,.000469077919138421,-.00636759259064836,-.00503976415007148},
			{-.0113741780920774,-.0157705273052622,-.00608002687266981,.00608414677874269,.00833345789650543,.000957688797273868,-.00550488775656931,-.00405218361689488,.00128277564401616,.00255185410234015,-.0254965031859464,-.0353514158497966,-.0136290660542447,.0136383012884271,.0186803858782727,.00214676746514514,-.0123398268505581,-.00908342665833277,.00287548630149319,.00572026881623259,-.0373762861378719,-.0518229744896624,-.019979362225491,.0199929005037731,.0273842825684829,.00314702743601977,-.0180894178270092,-.0133157379040649,.00421528387662904,.0083855579136398,-.0459370896279213,-.063692701172494,-.0245555096050894,.0245721487409458,.0336564804246319,.00386783429623537,-.0222326853173956,-.0163656240029997,.005180768162293,.0103062172640176,-.0504166383248294,-.0699036857788496,-.0269500365972072,.0269682982960067,.0369384872790024,.00424500560207766,-.024400702432737,-.0179615154769812,.00568597002550107,.0113112265601881,-.0504166383248294,-.0699036857788496,-.0269500365972072,.0269682982960067,.0369384872790024,.00424500560207766,-.024400702432737,-.0179615154769812,.00568597002550107,.0113112265601881,-.0459370896279213,-.063692701172494,-.0245555096050894,.0245721487409458,.0336564804246319,.00386783429623537,-.0222326853173956,-.0163656240029997,.005180768162293,.0103062172640176,-.0373762861378719,-.0518229744896624,-.019979362225491,.0199929005037731,.0273842825684829,.00314702743601977,-.0180894178270092,-.0133157379040649,.00421528387662904,.0083855579136398,-.0254965031859464,-.0353514158497966,-.0136290660542447,.0136383012884271,.0186803858782727,.00214676746514514,-.0123398268505581,-.00908342665833277,.0028754863014932,.00572026881623259,-.0113741780920774,-.0157705273052622,-.00608002687266981,.0060841467787427,.00833345789650543,.000957688797273868,-.00550488775656931,-.00405218361689488,.00128277564401616,.00255185410234015},
			{.019186595894105,.0266026021391187,.0102561273163527,-.0102630770029412,-.014057340035159,-.00161548270097898,.00928595068344064,.00683544858509201,-.00216385726557964,-.00430460935690978,.0382022700423445,.052968217830653,.0204208890149373,-.0204347264775717,-.0279894517539301,-.00321657404638918,.0184891784643094,.0136100043045404,-.00430843803919177,-.00857087155986831,.0439833629490508,.0609838197318553,.0235111518841866,-.0235270833495331,-.0322250540052594,-.00370333342961076,.0212871184388807,.015669586084838,-.00496042758221491,-.00986788885031005,.0344832662837457,.0478117441190089,.0184329086431872,-.0184453990242086,-.025264669269575,-.00290343948776025,.016689250760381,.0122850658360403,-.00388901015588316,-.00773649434850662,.0130003294489922,.0180252189558684,.00694928035797946,-.00695398928163203,-.00952488146632226,-.00109460831133298,.00629191435511048,.00463151900567974,-.00146617239912651,-.00291668934384738,-.0130003294489922,-.0180252189558684,-.00694928035797946,.00695398928163203,.00952488146632226,.00109460831133298,-.00629191435511048,-.00463151900567974,.00146617239912651,.00291668934384738,-.0344832662837457,-.0478117441190089,-.0184329086431872,.0184453990242085,.025264669269575,.00290343948776025,-.016689250760381,-.0122850658360403,.00388901015588316,.00773649434850662,-.0439833629490508,-.0609838197318553,-.0235111518841866,.0235270833495331,.0322250540052594,.00370333342961076,-.0212871184388807,-.015669586084838,.00496042758221491,.00986788885031005,-.0382022700423445,-.052968217830653,-.0204208890149373,.0204347264775717,.0279894517539301,.00321657404638918,-.0184891784643094,-.0136100043045404,.00430843803919177,.00857087155986831,-.019186595894105,-.0266026021391187,-.0102561273163527,.0102630770029412,.014057340035159,.00161548270097898,-.00928595068344064,-.00683544858509201,.00216385726557964,.00430460935690978},
			{-.0149432388948351,-.0143716720347549,.00337087856178778,.00991288633119597,-.000790460352998803,-.0073875487912777,-.000452661276806676,.00518463670922043,.000881265822650354,-.00248741703478494,-.0334969555607631,-.032215723971892,.00755620452698254,.022220852872113,-.00177190604415069,-.016560023921583,-.00101469131156782,.0116219480040223,.00197545674711399,-.00557582586088984,-.0491044511733968,-.0472262454396426,.0110769253515695,.0325743867352538,-.0025975039334847,-.0242759639637272,-.0014874742832843,.0170370521365157,.00289589659000063,-.00817381353480803,-.0603515171721834,-.0580431202166143,.0136140255026904,.0400353453393464,-.00319244588831583,-.0298362209742388,-.00182817092148764,.02093928187792,.00355918350788891,-.0100459741656872,-.0662366867006586,-.063703186747533,.0149415952441392,.0439393862896491,-.00350375679056903,-.0327456957770259,-.00200644474630778,.0229811729426385,.00390625677643855,-.011025605893504,-.0662366867006586,-.063703186747533,.0149415952441392,.0439393862896491,-.00350375679056903,-.0327456957770259,-.00200644474630778,.0229811729426385,.00390625677643855,-.011025605893504,-.0603515171721834,-.0580431202166143,.0136140255026904,.0400353453393464,-.00319244588831583,-.0298362209742388,-.00182817092148764,.02093928187792,.00355918350788891,-.0100459741656872,-.0491044511733968,-.0472262454396426,.0110769253515695,.0325743867352538,-.0025975039334847,-.0242759639637272,-.0014874742832843,.0170370521365157,.00289589659000063,-.00817381353480803,-.0334969555607631,-.032215723971892,.00755620452698254,.022220852872113,-.00177190604415069,-.016560023921583,-.00101469131156782,.0116219480040223,.00197545674711399,-.00557582586088984,-.0149432388948351,-.0143716720347549,.00337087856178778,.00991288633119597,-.000790460352998802,-.0073875487912777,-.000452661276806676,.00518463670922043,.000881265822650354,-.00248741703478494},
			{5.02239597246239e-005,.000582196454509688,.00202770964403212,.0044045611294125,.00726150487830256,.00980179190824761,.0111426514550797,.0106221354138729,.00804698659569713,.00379931823382166,.000112582670920343,.00130505902378937,.0045453398901192,.00987332060036796,.0162774822730591,.0219718221917673,.0249775101132113,.0238107146933112,.0180382281439193,.00851660037928648,.000165039185640327,.00191313526973573,.00666318526458248,.0144736732405605,.0238617754993688,.0322093232636935,.0366154746082026,.0349050251728234,.0264429193136947,.0124848058722647,.000202840374105465,.00235132688229124,.00818934598203497,.0177887771525588,.0293271652446273,.0395866663734618,.0450020189978578,.0428998018667281,.032499503831288,.0153443722103867,.00022262032405145,.00258061618548581,.0089879288791991,.0195234472015054,.0321869995510595,.0434469544588521,.0493903844165837,.0470831698838632,.0356686882793406,.0168406764624987,.00022262032405145,.00258061618548581,.0089879288791991,.0195234472015054,.0321869995510595,.0434469544588521,.0493903844165837,.0470831698838632,.0356686882793406,.0168406764624987,.000202840374105465,.00235132688229124,.00818934598203497,.0177887771525588,.0293271652446273,.0395866663734618,.0450020189978578,.0428998018667281,.032499503831288,.0153443722103867,.000165039185640327,.00191313526973573,.00666318526458248,.0144736732405605,.0238617754993688,.0322093232636935,.0366154746082026,.0349050251728234,.0264429193136947,.0124848058722647,.000112582670920343,.00130505902378937,.0045453398901192,.00987332060036795,.0162774822730591,.0219718221917673,.0249775101132113,.0238107146933112,.0180382281439193,.00851660037928648,5.02239597246239e-005,.000582196454509688,.00202770964403212,.0044045611294125,.00726150487830256,.00980179190824761,.0111426514550797,.0106221354138729,.00804698659569713,.00379931823382166},
			{-8.47205671528915e-005,-.00098208134306624,-.00342045334548888,-.00742985855733822,-.0122491100870261,-.0165342074606321,-.0187960438809947,-.0179180084878277,-.0135741043119093,-.00640890119633241,-.000168686410157299,-.00195541391909401,-.00681043594667852,-.0147935290111086,-.0243891002815539,-.0329211216952395,-.0374246452071332,-.0356763963055754,-.0270272851613995,-.0127607093790004,-.000194213474599731,-.0022513238093916,-.0078410491248818,-.0170322118311734,-.0280798666805756,-.0379030262496602,-.0430880613000287,-.0410752524832421,-.0311172841682885,-.0146917686169386,-.000152264731741222,-.00176505372042869,-.00614743773072645,-.0133533740168041,-.0220148132165338,-.0297162394935752,-.033781343486181,-.0322032871995818,-.0243961699164533,-.0115184500553977,-5.74044134859514e-005,-.000665432319314789,-.00231760863686745,-.00503427546699351,-.00829967272293673,-.0112031412634403,-.0127357017440349,-.012140768205952,-.00919745373300046,-.00434250178709194,5.74044134859514e-005,.000665432319314789,.00231760863686745,.00503427546699351,.00829967272293673,.0112031412634403,.0127357017440349,.012140768205952,.00919745373300046,.00434250178709194,.000152264731741222,.00176505372042869,.00614743773072645,.013353374016804,.0220148132165338,.0297162394935752,.033781343486181,.0322032871995818,.0243961699164533,.0115184500553977,.000194213474599731,.0022513238093916,.0078410491248818,.0170322118311734,.0280798666805756,.0379030262496602,.0430880613000287,.0410752524832421,.0311172841682885,.0146917686169386,.000168686410157299,.00195541391909401,.00681043594667852,.0147935290111086,.0243891002815539,.0329211216952395,.0374246452071332,.0356763963055754,.0270272851613995,.0127607093790004,8.47205671528915e-005,.00098208134306624,.00342045334548889,.00742985855733822,.0122491100870261,.0165342074606321,.0187960438809947,.0179180084878277,.0135741043119093,.00640890119633241},
			{.000103627666330889,.00120125255471199,.00418379633067172,.00908798098667198,.0149827454608269,.0202241485315363,.02299075925832,.021916772600681,.0166034388065025,.00783917644841335,.000156710198511108,.00181658559896353,.00632691612890426,.0137432343592401,.0226575498468763,.0305838242171288,.0347676115352073,.0331434828804184,.0251084317871825,.0118547386126776,7.10016118524656e-005,.000823051127621699,.00286657312335497,.00622672806775283,.0102655894449765,.0138567932186921,.0157523663597753,.0150165128324275,.0113760249486971,.0053710961863794,-9.89917330502748e-005,-.00114751278719633,-.00399662816086328,-.00868141703515759,-.0143124707091497,-.0193193920454638,-.0219622344466939,-.0209362941329066,-.0158606599978981,-.00748847957105857,-.000232347745540099,-.00269337651607831,-.00938065749861198,-.0203765265447757,-.0335934143176455,-.0453453742919027,-.0515485031272343,-.0491404745815968,-.0372272358482508,-.0175765318198328,-.000232347745540099,-.00269337651607831,-.00938065749861198,-.0203765265447757,-.0335934143176455,-.0453453742919027,-.0515485031272343,-.0491404745815968,-.0372272358482508,-.0175765318198328,-9.89917330502745e-005,-.00114751278719633,-.00399662816086327,-.00868141703515757,-.0143124707091497,-.0193193920454637,-.0219622344466939,-.0209362941329065,-.0158606599978981,-.00748847957105856,7.10016118524653e-005,.000823051127621695,.00286657312335496,.0062267280677528,.0102655894449765,.0138567932186921,.0157523663597753,.0150165128324274,.0113760249486971,.00537109618637938,.000156710198511108,.00181658559896353,.00632691612890426,.0137432343592401,.0226575498468763,.0305838242171288,.0347676115352073,.0331434828804184,.0251084317871825,.0118547386126776,.000103627666330889,.00120125255471199,.00418379633067172,.00908798098667198,.0149827454608269,.0202241485315363,.02299075925832,.021916772600681,.0166034388065025,.00783917644841335},
			{-.000112748940416249,-.00130698642082339,-.00455205274713417,-.00988790216985572,-.0163015219298734,-.0220042716244384,-.0250143985435671,-.0238458799234204,-.0180648681860857,-.00852917825508497,-9.55544782092518e-005,-.00110766810763179,-.0038578546585677,-.00837997527015808,-.0138155038643826,-.0186485716467526,-.0211996475685226,-.0202093306164191,-.015309935925496,-.00722845975057063,.000102648386927895,.00118990074176208,.00414425953786325,.0090020997455888,.0148411588117245,.0200330307267911,.0227734970367722,.0217096595318657,.0164465366372481,.00776509638558674,.000239663804666116,.00277818431911703,.00967603133467496,.0210181332564652,.0346511883228178,.0467731886166761,.0531716387245257,.0506877872817214,.0383994299572614,.018129973583278,.000126671410044506,.00146837577562602,.00511414952502166,.0111088805412601,.0183144671790717,.0247214040626872,.0281032276066716,.0267904179187864,.0202955550353876,.00958237862017949,-.000126671410044507,-.00146837577562603,-.0051141495250217,-.0111088805412602,-.0183144671790719,-.0247214040626874,-.0281032276066718,-.0267904179187866,-.0202955550353878,-.00958237862017957,-.000239663804666117,-.00277818431911704,-.00967603133467498,-.0210181332564652,-.0346511883228178,-.0467731886166762,-.0531716387245259,-.0506877872817215,-.0383994299572615,-.018129973583278,-.000102648386927894,-.00118990074176207,-.00414425953786321,-.0090020997455887,-.0148411588117243,-.0200330307267909,-.0227734970367719,-.0217096595318655,-.016446536637248,-.00776509638558665,9.55544782092525e-005,.0011076681076318,.00385785465856773,.00837997527015815,.0138155038643827,.0186485716467528,.0211996475685227,.0202093306164193,.0153099359254961,.00722845975057069,.000112748940416249,.00130698642082339,.00455205274713418,.00988790216985573,.0163015219298734,.0220042716244384,.0250143985435671,.0238458799234204,.0180648681860857,.00852917825508498},
			{.000242033524403197,.00261162972756074,.00794330061991523,.0139365347274024,.0166502677108084,.0135390642006158,.00568409316229951,-.00258267825885607,-.00653083370918549,-.00434965045780644,.000542545446017795,.00585426262277764,.0178058043335604,.0312403146145367,.0373234532024064,.030349339594584,.01274153598164,-.00578936463998794,-.0146396004287218,-.00975023213611034,.000795337842419811,.0085819844909475,.0261022004795129,.0457963560553539,.0547138584602276,.0444902421524166,.0186782984008672,-.00848684808903276,-.0214607426978308,-.0142932332160281,.000977504977807211,.0105476341146326,.0320807454890066,.0562857236534088,.067245723951962,.0546804525678283,.0229564452865287,-.0104307073176385,-.0263761909665255,-.0175670084741824,.00107282623531448,.011576185140315,.0352090947774598,.0617744179108896,.0738031810643202,.0600126090459714,.0251950397513031,-.0114478562435068,-.0289482614400908,-.0192800527823109,.00107282623531448,.011576185140315,.0352090947774598,.0617744179108896,.0738031810643202,.0600126090459713,.0251950397513031,-.0114478562435068,-.0289482614400908,-.0192800527823109,.000977504977807211,.0105476341146326,.0320807454890066,.0562857236534088,.067245723951962,.0546804525678283,.0229564452865287,-.0104307073176385,-.0263761909665255,-.0175670084741824,.000795337842419811,.0085819844909475,.0261022004795129,.0457963560553539,.0547138584602276,.0444902421524166,.0186782984008672,-.00848684808903276,-.0214607426978308,-.0142932332160281,.000542545446017795,.00585426262277763,.0178058043335604,.0312403146145367,.0373234532024064,.030349339594584,.01274153598164,-.00578936463998794,-.0146396004287218,-.00975023213611034,.000242033524403197,.00261162972756074,.00794330061991523,.0139365347274024,.0166502677108084,.0135390642006158,.0056840931622995,-.00258267825885607,-.00653083370918549,-.00434965045780644},
			{-.000408275603560732,-.00440544220444747,-.0133992010522381,-.0235089215074875,-.0280865971597952,-.0228384461275931,-.00958824432382064,.00435660525762981,.0110165733484533,.00733723218846652,-.000812914127394553,-.00877163899626159,-.0266790367491162,-.0468084162917745,-.0559229878604316,-.045473438390492,-.0190910727943728,.00867440016136512,.0219350067266604,.0146091014255062,-.000935931217489798,-.0100990381252994,-.0307163358399061,-.053891864555414,-.0643857307310363,-.0523548664268737,-.0219800965458574,.00998708428163375,.0252543987851819,.0168198750923781,-.00073377666535768,-.00791771700785594,-.0240818236035156,-.0422516013189939,-.0504788662986896,-.0410465839626478,-.0172325504762159,.00782994440604093,.0197996264895458,.0131869005183092,-.000276636740648008,-.0029850110119799,-.00907894391448041,-.0159290228592211,-.0190307346959162,-.0154747264913728,-.00649674052317934,.00295192038968209,.0074645384579598,.00497151429155405,.000276636740648008,.0029850110119799,.00907894391448041,.0159290228592211,.0190307346959162,.0154747264913728,.00649674052317934,-.00295192038968209,-.0074645384579598,-.00497151429155405,.00073377666535768,.00791771700785594,.0240818236035156,.0422516013189938,.0504788662986896,.0410465839626478,.0172325504762159,-.00782994440604093,-.0197996264895458,-.0131869005183092,.000935931217489798,.0100990381252994,.0307163358399061,.053891864555414,.0643857307310362,.0523548664268737,.0219800965458574,-.00998708428163375,-.0252543987851819,-.0168198750923781,.000812914127394553,.00877163899626159,.0266790367491162,.0468084162917745,.0559229878604317,.0454734383904919,.0190910727943728,-.00867440016136512,-.0219350067266604,-.0146091014255062,.000408275603560732,.00440544220444747,.0133992010522381,.0235089215074875,.0280865971597952,.0228384461275931,.00958824432382064,-.00435660525762981,-.0110165733484533,-.00733723218846652},
			{.000499390519149255,.00538860527147271,.0163895023638538,.0287554103499522,.0343546864285749,.0279353048988254,.0117280539636516,-.00532886937735551,-.0134751433486315,-.00897468317910862,.000755199746953267,.00814887984733841,.0247849079292855,.0434851639891283,.0519526292603973,.042244965376185,.0177356658686522,-.0080585446679718,-.0203776893169333,-.0135719005586946,.000342162793575017,.00369206094721455,.0112294440905165,.0197020791513965,.0235384834714527,.019140175067471,.00803560250650699,-.00365113225589165,-.00923263961808127,-.00614910085446924,-.000477049563208446,-.00514753823408188,-.0156562943109128,-.0274689955482004,-.0328177799266881,-.0266855786985745,-.0112033825355281,.00509047471144027,.0128723133542602,.00857318776860553,-.00111970350560565,-.0120820079305306,-.0367475603725958,-.0644736584683689,-.0770280198622701,-.0626348671549949,-.0262959400178523,.0119480716872726,.0302131591759193,.0201224969878625,-.00111970350560565,-.0120820079305306,-.0367475603725958,-.064473658468369,-.0770280198622701,-.0626348671549949,-.0262959400178523,.0119480716872726,.0302131591759193,.0201224969878625,-.000477049563208445,-.00514753823408186,-.0156562943109127,-.0274689955482003,-.0328177799266881,-.0266855786985745,-.0112033825355281,.00509047471144026,.0128723133542602,.00857318776860551,.000342162793575015,.00369206094721453,.0112294440905164,.0197020791513964,.0235384834714526,.0191401750674709,.00803560250650695,-.00365113225589163,-.00923263961808122,-.00614910085446921,.000755199746953267,.00814887984733841,.0247849079292855,.0434851639891283,.0519526292603973,.042244965376185,.0177356658686522,-.0080585446679718,-.0203776893169333,-.0135719005586946,.000499390519149255,.00538860527147271,.0163895023638538,.0287554103499522,.0343546864285749,.0279353048988254,.0117280539636516,-.00532886937735551,-.0134751433486316,-.00897468317910862},
			{.000682727668349805,.00664570504694168,.0164335220422574,.0198480228837579,.0106420257503786,-.00420811508616648,-.0112538525390341,-.00576778839286375,.00325440242770835,.00455104492220568,.00153041107940269,.0148970974896404,.0368375933377187,.0444915821252141,.0238553011261929,-.00943296463557281,-.0252267798862333,-.0129291482816634,.00729511013408645,.010201680545096,.00224348735178138,.0218382173561541,.0540016181505257,.065221889139497,.0349703861072661,-.0138281388148133,-.0369808885747392,-.0189533263510734,.0106941772285541,.0149550284742905,.00275734403299806,.0268401238235168,.0663703494746075,.0801605530322499,.0429801333125165,-.0169953871227395,-.0454511287373907,-.0232944666606214,.0131436113270545,.0183803837780314,.00302622604032566,.0294574346429948,.0728424445705406,.0879773978473551,.0471713348390493,-.0186526898568866,-.0498832890278595,-.0255660232311389,.014425308697734,.0201727442620858,.00302622604032566,.0294574346429948,.0728424445705406,.0879773978473551,.0471713348390493,-.0186526898568866,-.0498832890278595,-.0255660232311389,.014425308697734,.0201727442620858,.00275734403299806,.0268401238235168,.0663703494746075,.0801605530322499,.0429801333125165,-.0169953871227395,-.0454511287373907,-.0232944666606214,.0131436113270545,.0183803837780314,.00224348735178138,.0218382173561541,.0540016181505257,.065221889139497,.0349703861072661,-.0138281388148133,-.0369808885747392,-.0189533263510734,.0106941772285541,.0149550284742905,.00153041107940269,.0148970974896404,.0368375933377187,.0444915821252141,.0238553011261929,-.00943296463557281,-.0252267798862333,-.0129291482816634,.00729511013408645,.010201680545096,.000682727668349805,.00664570504694168,.0164335220422574,.0198480228837579,.0106420257503786,-.00420811508616648,-.0112538525390341,-.00576778839286375,.00325440242770835,.00455104492220568},
			{-.00115166298286166,-.0112103446987566,-.0277209784976989,-.0334807483245332,-.0179515606112687,.00709848245065787,.0189836240489798,.00972942608448226,-.00548970692279972,-.00767695556108528,-.0022930664987097,-.0223208232358323,-.0551950076111466,-.0666632369688139,-.0357432016569395,.0141337288264978,.0377981345051569,.0193721786130192,-.0109305093762129,-.0152855217812643,-.00264007285345256,-.025698600334876,-.0635475863095515,-.0767512858190868,-.0411521673894173,.0162725650625018,.0435180701799816,.0223037416914183,-.0125846071602312,-.0175986571380644,-.00206983570855075,-.0201478836325818,-.0498217551700164,-.0601735485661986,-.0322635890276594,.0127578056007183,.0341184734761949,.017486290549451,-.00986642063466756,-.0137974711263254,-.000780336349086855,-.00759583279519514,-.0187830011695426,-.0226856687251885,-.0121635022365666,.00480974378966284,.0128628011000313,.00659240154668345,-.00371968008127681,-.00520170185530649,.000780336349086855,.00759583279519514,.0187830011695426,.0226856687251885,.0121635022365666,-.00480974378966284,-.0128628011000313,-.00659240154668345,.00371968008127681,.00520170185530649,.00206983570855075,.0201478836325818,.0498217551700164,.0601735485661986,.0322635890276594,-.0127578056007183,-.0341184734761949,-.017486290549451,.00986642063466756,.0137974711263254,.00264007285345256,.025698600334876,.0635475863095515,.0767512858190868,.0411521673894173,-.0162725650625018,-.0435180701799816,-.0223037416914183,.0125846071602312,.0175986571380644,.0022930664987097,.0223208232358323,.0551950076111466,.0666632369688139,.0357432016569395,-.0141337288264978,-.0377981345051569,-.0193721786130192,.0109305093762129,.0152855217812643,.00115166298286166,.0112103446987566,.0277209784976989,.0334807483245332,.0179515606112687,-.00709848245065787,-.0189836240489798,-.00972942608448226,.00548970692279973,.00767695556108528},
			{.00148169134577296,.0125696636361974,.0228905745669076,.0127235476313918,-.00841173160413407,-.0124282780657018,.00206969137961716,.00930159426344669,.000766794088585497,-.00444614718711676,.00332137828441454,.0281763188823123,.0513118049188827,.0285212672154559,-.0188558452230951,-.0278593871779371,.00463944668591847,.0208505727492996,.0017188554429443,-.0099665404395655,.00486893376025062,.0413047290907605,.0752199110957276,.0418104018690363,-.027641495043065,-.0408401269458114,.00680113996769058,.0305656413953226,.00251973505529715,-.0146103277205166,.00598413244462014,.0507653177552249,.0924485590960711,.0513868117060249,-.0339726058003504,-.0501943014082524,.00835889834306659,.0375665095010369,.00309686453723171,-.0179567314824952,.00656767426042838,.055715690424466,.101463667057506,.0563977892686102,-.0372854395753275,-.0550889915004361,.00917401511436313,.0412298022791773,.00339885483776274,-.0197077795738015,.00656767426042838,.055715690424466,.101463667057506,.0563977892686102,-.0372854395753275,-.0550889915004361,.00917401511436313,.0412298022791773,.00339885483776274,-.0197077795738015,.00598413244462014,.0507653177552249,.0924485590960711,.0513868117060249,-.0339726058003504,-.0501943014082523,.0083588983430666,.0375665095010369,.00309686453723171,-.0179567314824952,.00486893376025062,.0413047290907605,.0752199110957276,.0418104018690363,-.027641495043065,-.0408401269458114,.00680113996769058,.0305656413953226,.00251973505529715,-.0146103277205166,.00332137828441454,.0281763188823123,.0513118049188827,.0285212672154559,-.0188558452230951,-.0278593871779371,.00463944668591847,.0208505727492996,.0017188554429443,-.0099665404395655,.00148169134577296,.0125696636361974,.0228905745669076,.0127235476313918,-.00841173160413407,-.0124282780657018,.00206969137961716,.00930159426344669,.000766794088585497,-.00444614718711676},
			{-1.03605502214324e-006,-6.2106840321439e-005,-.000513920961277903,-.00197298030906211,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186,-2.32243419779848e-006,-.000139219488151642,-.00115201180432313,-.00442265791238583,-.0109526932361906,-.0199562351910325,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-3.404543928249e-006,-.000204087101167348,-.00168877757544104,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-4.18433332293283e-006,-.000250831969330619,-.00207558146784554,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-4.59236798590328e-006,-.000275291812791727,-.00227798149655704,-.00874533824411281,-.0216577924252842,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.59236798590328e-006,-.000275291812791727,-.00227798149655704,-.00874533824411281,-.0216577924252843,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.18433332293283e-006,-.000250831969330619,-.00207558146784554,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-3.404543928249e-006,-.000204087101167348,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-2.32243419779848e-006,-.000139219488151642,-.00115201180432313,-.00442265791238583,-.0109526932361906,-.0199562351910326,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-1.03605502214324e-006,-6.21068403214391e-005,-.000513920961277903,-.00197298030906211,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186},
			{1.74767520475977e-006,.000104765270720098,.000866910445908273,.00332813286160627,.00824210667082646,.0150174405185659,.0212996498258157,.0237895578215264,.0200145017350098,.0100011552359343,3.47978143111129e-006,.000208597251185043,.001726098112448,.00662661740611046,.0164107894122825,.0299010997678524,.0424095539898868,.0473671889011406,.0398507064131546,.0199132162457208,4.00637159776859e-006,.000240163964049163,.00198730598159011,.00762941359700327,.0188942098519485,.0344259888799142,.0488273289408453,.0545351953956101,.0458812547527418,.0229266537472159,3.14102354559662e-006,.000188290239053809,.00155806188421368,.00598151398652822,.0148131935773177,.0269902177103302,.0382809697326146,.0427559772280557,.0359712268228302,.017974657987755,1.18417845234822e-006,7.09861739774591e-005,.000587395568332318,.00225505472099886,.00558463328595991,.0101754201370336,.0144320788540114,.016119174596298,.0135612965295474,.0067765180261918,-1.18417845234822e-006,-7.09861739774591e-005,-.000587395568332318,-.00225505472099886,-.00558463328595991,-.0101754201370336,-.0144320788540114,-.016119174596298,-.0135612965295474,-.0067765180261918,-3.14102354559662e-006,-.000188290239053809,-.00155806188421368,-.00598151398652822,-.0148131935773177,-.0269902177103302,-.0382809697326146,-.0427559772280557,-.0359712268228302,-.017974657987755,-4.00637159776859e-006,-.000240163964049163,-.00198730598159011,-.00762941359700327,-.0188942098519485,-.0344259888799142,-.0488273289408453,-.0545351953956101,-.0458812547527418,-.0229266537472159,-3.47978143111129e-006,-.000208597251185043,-.001726098112448,-.00662661740611046,-.0164107894122825,-.0299010997678524,-.0424095539898868,-.0473671889011406,-.0398507064131546,-.0199132162457208,-1.74767520475977e-006,-.000104765270720098,-.000866910445908273,-.00332813286160627,-.00824210667082646,-.0150174405185659,-.0212996498258157,-.0237895578215264,-.0200145017350098,-.0100011552359343},
			{-2.1377040911701e-006,-.000128145748833051,-.00106037895476748,-.00407087267407408,-.0100814986114699,-.0183688845318685,-.0260530952485506,-.0290986763122241,-.0244811405031877,-.0122331142571899,-3.23272774874231e-006,-.000193787493716783,-.00160355050327985,-.00615614813548528,-.015245674293661,-.0277782144801489,-.0393986072714656,-.0440042654896674,-.0370214299314904,-.0184994396914433,-1.46467098517923e-006,-8.78004401849179e-005,-.00072653006314475,-.00278920226363585,-.00690744737044604,-.0125856391046194,-.0178505588506293,-.0199372715230704,-.0167735171239198,-.00838165000707885,2.04207081203481e-006,.000122412963730206,.0010129413711518,.00388875630709325,.00963048821447668,.0175471259596245,.0248875724147149,.0277969050120105,.0233859413346208,.0116858482275818,4.79303205216701e-006,.000287320721349026,.00237751816944035,.00912746684058028,.0226041322455455,.0411856124930787,.0584146894325829,.0652433088453238,.0548901466718698,.0274283559519402,4.79303205216701e-006,.000287320721349026,.00237751816944035,.00912746684058028,.0226041322455455,.0411856124930787,.0584146894325829,.0652433088453238,.0548901466718698,.0274283559519403,2.04207081203481e-006,.000122412963730205,.0010129413711518,.00388875630709324,.00963048821447665,.0175471259596245,.0248875724147148,.0277969050120105,.0233859413346208,.0116858482275817,-1.46467098517922e-006,-8.78004401849175e-005,-.000726530063144746,-.00278920226363584,-.00690744737044601,-.0125856391046194,-.0178505588506292,-.0199372715230703,-.0167735171239197,-.00838165000707881,-3.23272774874231e-006,-.000193787493716783,-.00160355050327985,-.00615614813548528,-.015245674293661,-.0277782144801489,-.0393986072714656,-.0440042654896674,-.0370214299314904,-.0184994396914433,-2.1377040911701e-006,-.000128145748833051,-.00106037895476748,-.00407087267407408,-.0100814986114699,-.0183688845318685,-.0260530952485506,-.0290986763122241,-.0244811405031877,-.0122331142571899},
			{-7.0687420406234e-006,-.000396419472948955,-.00289468675708114,-.0091512705781711,-.0170446886407066,-.020343232859846,-.0143340443821725,-.00198780768336474,.00723004945575377,.00622084218798443,-1.58453826290029e-005,-.00088861896421812,-.00648876688291386,-.0205136052524374,-.0382075922069433,-.0456016511456594,-.032131377344278,-.00445589514439379,.0162069713952851,.0139447194673076,-2.32283443257473e-005,-.00130266007192951,-.00951212822907477,-.0300716680261107,-.0560099511902079,-.066849181205329,-.04710259853538,-.00653206483660142,.0237584109428924,.0204420904749686,-2.85486506407837e-005,-.00160102617628203,-.0116908214315209,-.0369593946354349,-.0688386785778866,-.0821605661208479,-.0578911527658522,-.00802819324390579,.0292001256859699,.0251242228526073,-3.13325682074437e-005,-.00175715001389931,-.0128308502042083,-.0405634846946589,-.0755514724317883,-.0901724419250166,-.0635364002126376,-.00881106135495838,.0320475717480869,.0275742078353698,-3.13325682074437e-005,-.00175715001389931,-.0128308502042083,-.0405634846946589,-.0755514724317883,-.0901724419250165,-.0635364002126376,-.00881106135495838,.0320475717480869,.0275742078353698,-2.85486506407837e-005,-.00160102617628203,-.0116908214315209,-.0369593946354349,-.0688386785778866,-.0821605661208479,-.0578911527658522,-.00802819324390579,.0292001256859699,.0251242228526073,-2.32283443257473e-005,-.00130266007192951,-.00951212822907477,-.0300716680261107,-.0560099511902079,-.066849181205329,-.04710259853538,-.00653206483660142,.0237584109428924,.0204420904749686,-1.58453826290029e-005,-.00088861896421812,-.00648876688291386,-.0205136052524374,-.0382075922069433,-.0456016511456594,-.032131377344278,-.00445589514439379,.0162069713952851,.0139447194673076,-7.0687420406234e-006,-.000396419472948955,-.00289468675708114,-.0091512705781711,-.0170446886407066,-.020343232859846,-.0143340443821725,-.00198780768336474,.00723004945575377,.00622084218798443},
			{1.19239470204654e-005,.000668702403646816,.00488291853542222,.0154368719238809,.0287519283012612,.0343160960538323,.0241794628831796,.00335314450111479,-.0121960493352718,-.0104936624148843,2.37416708289763e-005,.00133144774315747,.00972233811117492,.0307362261184839,.0572477231117576,.0683264908211426,.048143525596736,.00667641787160104,-.0242834514641005,-.0208938431558764,2.73344626884814e-005,.00153293375681807,.0111936051282609,.0353874936634631,.0659109362045824,.0786662374118139,.0554290139727572,.00768675029312828,-.027958230184157,-.0240556774743184,2.14304112375827e-005,.00120183086029444,.00877586524613968,.0277440442315677,.0516746381304762,.0616748841000227,.0434567372868267,.00602646636005635,-.0219194493467751,-.0188598205404102,8.07935083711268e-006,.000453095046079154,.00330853633356522,.0104596157536215,.0194815454639894,.0232516782323097,.0163833639440749,.00227200194579396,-.00826372016223111,-.00711022785245187,-8.07935083711268e-006,-.000453095046079154,-.00330853633356522,-.0104596157536215,-.0194815454639894,-.0232516782323097,-.0163833639440749,-.00227200194579396,.00826372016223111,.00711022785245187,-2.14304112375827e-005,-.00120183086029444,-.00877586524613968,-.0277440442315677,-.0516746381304761,-.0616748841000227,-.0434567372868267,-.00602646636005635,.0219194493467751,.0188598205404102,-2.73344626884814e-005,-.00153293375681807,-.0111936051282609,-.0353874936634631,-.0659109362045824,-.0786662374118139,-.0554290139727572,-.00768675029312828,.027958230184157,.0240556774743184,-2.37416708289763e-005,-.00133144774315747,-.00972233811117492,-.0307362261184839,-.0572477231117576,-.0683264908211426,-.048143525596736,-.00667641787160104,.0242834514641005,.0208938431558764,-1.19239470204654e-005,-.000668702403646816,-.00488291853542223,-.0154368719238809,-.0287519283012612,-.0343160960538323,-.0241794628831796,-.00335314450111479,.0121960493352718,.0104936624148843},
			{-2.71193324668268e-005,-.00139397384538797,-.00859091112473706,-.0204390993260263,-.0232648893283274,-.00816857426450726,.0104893149252651,.0116472227517443,-.00129246985153553,-.00610326187849476,-6.07910427499648e-005,-.00312474961288118,-.0192574963297445,-.0458165466432155,-.0521508736788253,-.0183107806184887,.0235129825861177,.0261085635896738,-.00289721696114192,-.0136811499406799,-8.91158835339698e-005,-.00458068837081261,-.0282302905567179,-.0671642046243376,-.076449933653419,-.026842464271009,.0344685684418851,.0382735285799644,-.00424713967062055,-.0200557139600161,-.000109527316707461,-.00562986625988991,-.0346962612267728,-.082547743674565,-.0939603105918888,-.0329905620505779,.0423633774639968,.0470398398136498,-.00521992032573585,-.0246493492246598,-.000120207857264269,-.0061788618595791,-.0380796621567155,-.0905973750422723,-.103122836783299,-.0362076319703515,.0464944361325487,.051626923036526,-.00572894010653844,-.0270530269738013,-.000120207857264269,-.0061788618595791,-.0380796621567155,-.0905973750422723,-.103122836783299,-.0362076319703515,.0464944361325487,.051626923036526,-.00572894010653844,-.0270530269738013,-.000109527316707461,-.00562986625988991,-.0346962612267728,-.082547743674565,-.0939603105918888,-.0329905620505779,.0423633774639968,.0470398398136498,-.00521992032573585,-.0246493492246598,-8.91158835339698e-005,-.00458068837081261,-.0282302905567179,-.0671642046243376,-.076449933653419,-.026842464271009,.0344685684418851,.0382735285799644,-.00424713967062055,-.0200557139600161,-6.07910427499648e-005,-.00312474961288118,-.0192574963297445,-.0458165466432156,-.0521508736788254,-.0183107806184887,.0235129825861177,.0261085635896738,-.00289721696114192,-.0136811499406799,-2.71193324668268e-005,-.00139397384538797,-.00859091112473705,-.0204390993260263,-.0232648893283274,-.00816857426450726,.0104893149252651,.0116472227517443,-.00129246985153553,-.00610326187849476},
			{1.8467923733974e-008,5.72496315380497e-006,.000112551238536802,.000763670899970611,.00284090951583314,.00698705136166929,.0123641411463411,.0161796260245334,.0151169390000152,.00799470122065349,4.1397934207577e-008,1.28331506776816e-005,.000252296296814152,.0017118544633724,.00636821913065334,.0156622637575245,.0277156170386288,.0362684567748484,.0338863239396712,.0179210245780364,6.06867939173738e-008,1.88125998409004e-005,.000369850661970342,.00250947205515881,.00933541273013333,.0229599034620059,.0406293688733935,.0531672993864887,.0496752409729615,.026271106183758,7.45867227452771e-008,2.3121507627533e-005,.000454562632178227,.00308425086139813,.0114736303579195,.0282187250839213,.0499352705236225,.065344935240019,.0610530427888428,.0322883379835085,8.18600363005434e-008,2.53761981227818e-005,.000498889241964133,.00338501114114201,.0125924797742609,.0309704700072926,.0548047012562974,.0717170372141512,.0670066214870632,.0354369306242455,8.18600363005434e-008,2.53761981227818e-005,.000498889241964133,.00338501114114201,.0125924797742609,.0309704700072926,.0548047012562974,.0717170372141512,.0670066214870632,.0354369306242455,7.45867227452771e-008,2.3121507627533e-005,.000454562632178227,.00308425086139813,.0114736303579195,.0282187250839213,.0499352705236225,.065344935240019,.0610530427888428,.0322883379835085,6.06867939173738e-008,1.88125998409004e-005,.000369850661970342,.00250947205515881,.00933541273013333,.0229599034620059,.0406293688733935,.0531672993864886,.0496752409729615,.026271106183758,4.1397934207577e-008,1.28331506776816e-005,.000252296296814152,.0017118544633724,.00636821913065334,.0156622637575245,.0277156170386288,.0362684567748484,.0338863239396712,.0179210245780364,1.8467923733974e-008,5.72496315380497e-006,.000112551238536802,.000763670899970611,.00284090951583315,.00698705136166929,.0123641411463411,.0161796260245334,.0151169390000152,.00799470122065349},
			{-3.11527203680034e-008,-9.65718609285329e-006,-.000189857685775429,-.0012882025257772,-.00479220409464544,-.0117861466401087,-.0208565205959732,-.0272926925875222,-.0255000930408489,-.013485906436498,-6.20279200448354e-008,-1.9228342172032e-005,-.000378024044579789,-.00256492923656619,-.00954171734955457,-.0234672976482734,-.041527243099746,-.0543422511302053,-.0507730212189817,-.0268516751119555,-7.14145132549108e-008,-2.21381386949937e-005,-.000435229862984357,-.00295307617650668,-.0109856512944789,-.0270185690209856,-.0478114992513544,-.0625657834542022,-.0584564272703957,-.0309150993103302,-5.59894812942294e-008,-1.73564566340746e-005,-.000341223277476823,-.00231523251800061,-.00861282797603967,-.0211827483777367,-.0374845521025614,-.0490520148175479,-.0458302506312781,-.0242376555640243,-2.11082586119828e-008,-6.54345364077286e-006,-.000128642541758133,-.000872851928740155,-.00324707062998819,-.0079859809473875,-.0141318262188787,-.0184928059748667,-.0172781879777248,-.00913769318749374,2.11082586119828e-008,6.54345364077286e-006,.000128642541758133,.000872851928740155,.00324707062998819,.00798598094738749,.0141318262188787,.0184928059748667,.0172781879777248,.00913769318749374,5.59894812942294e-008,1.73564566340746e-005,.000341223277476823,.00231523251800061,.00861282797603967,.0211827483777367,.0374845521025615,.0490520148175479,.0458302506312781,.0242376555640243,7.14145132549108e-008,2.21381386949937e-005,.000435229862984357,.00295307617650668,.0109856512944789,.0270185690209856,.0478114992513544,.0625657834542022,.0584564272703957,.0309150993103302,6.20279200448354e-008,1.9228342172032e-005,.000378024044579789,.00256492923656619,.00954171734955457,.0234672976482734,.041527243099746,.0543422511302053,.0507730212189817,.0268516751119555,3.11527203680034e-008,9.65718609285329e-006,.000189857685775429,.0012882025257772,.00479220409464544,.0117861466401087,.0208565205959732,.0272926925875222,.0255000930408489,.013485906436498},
			{1.62757657902233e-007,4.73190249691825e-005,.000825150543237864,.00465350306811798,.013244685304365,.0221077948264185,.0214226110182188,.00800738602387552,-.0066385606490133,-.00788880135076707,3.64839649040836e-007,.000106070931993143,.00184966801858007,.0104313520363045,.0296894560930798,.0495571158340357,.0480211990401099,.0179494589952879,-.0148810825158051,-.0176836380743178,5.34832208853932e-007,.000155493381824678,.00271149814617032,.0152917125799703,.0435228940266753,.0726476461525729,.0703960877661124,.0263127892689876,-.0218147404016795,-.0259231123501117,6.57332165731892e-007,.000191108163905862,.00333254975914684,.0187941832625992,.0534915394321128,.0892871330302339,.086519869342777,.0323395683174988,-.0268112696201347,-.0318606383488469,7.21431817457549e-007,.000209744049059384,.00365752287024036,.0206268953439552,.0587077592165956,.0979939549829493,.094956841974485,.0354931566768965,-.0294257667261143,-.0349675239211424,7.21431817457549e-007,.000209744049059384,.00365752287024036,.0206268953439552,.0587077592165956,.0979939549829493,.094956841974485,.0354931566768965,-.0294257667261143,-.0349675239211424,6.57332165731892e-007,.000191108163905862,.00333254975914684,.0187941832625992,.0534915394321128,.0892871330302339,.086519869342777,.0323395683174988,-.0268112696201347,-.0318606383488469,5.34832208853932e-007,.000155493381824678,.00271149814617032,.0152917125799703,.0435228940266753,.0726476461525729,.0703960877661124,.0263127892689876,-.0218147404016795,-.0259231123501117,3.64839649040836e-007,.000106070931993143,.00184966801858007,.0104313520363045,.0296894560930798,.0495571158340357,.0480211990401099,.0179494589952879,-.0148810825158051,-.0176836380743178,1.62757657902233e-007,4.73190249691825e-005,.000825150543237864,.00465350306811798,.013244685304365,.0221077948264185,.0214226110182188,.00800738602387552,-.0066385606490133,-.00788880135076707},
			{-3.0545495487195e-010,-4.89665844207368e-007,-2.28716807432812e-005,-.000274273335410219,-.00153266885205138,-.00508819457391411,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594,-6.84711736268708e-010,-1.09764122346374e-006,-5.12694522820253e-005,-.000614814618998954,-.0034356501149343,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-1.00374477198287e-009,-1.60907368927037e-006,-7.51578247671699e-005,-.000901279950190104,-.00503644914868864,-.0167201370314615,-.0369150385504742,-.0565976592019674,-.058726053964138,-.0328701880528208,-1.23364620508441e-009,-1.97762190735818e-006,-9.237225228407e-005,-.00110771245968658,-.00619001618021515,-.0205497793593547,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.35394503757885e-009,-2.17046942359914e-006,-.000101379919197693,-.001215730881087,-.00679363471892789,-.0225536881419213,-.0497944643424746,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.35394503757885e-009,-2.17046942359914e-006,-.000101379919197693,-.001215730881087,-.00679363471892789,-.0225536881419213,-.0497944643424746,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.23364620508441e-009,-1.97762190735818e-006,-9.237225228407e-005,-.00110771245968658,-.00619001618021515,-.0205497793593547,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.00374477198287e-009,-1.60907368927037e-006,-7.51578247671699e-005,-.000901279950190104,-.00503644914868864,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-6.84711736268708e-010,-1.09764122346374e-006,-5.12694522820253e-005,-.000614814618998954,-.0034356501149343,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-3.0545495487195e-010,-4.89665844207368e-007,-2.28716807432812e-005,-.000274273335410219,-.00153266885205138,-.00508819457391411,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_right_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_right_fluxes_to_dofs(
		const T i_input[100],
		T o_output[35]
	) {
		static const T matrix[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{.00265100998409647,.00594253790984439,.008711390565351,.0107066798372692,.0117507401837501,.0117507401837501,.0107066798372692,.008711390565351,.00594253790984439,.00265100998409647,.00527840372812152,.0118321373536874,.0173451766357429,.0213179803576284,.0233968001502014,.0233968001502014,.0213179803576284,.0173451766357429,.0118321373536874,.00527840372812152,.00607717674128413,.0136226771632526,.019969996509099,.0245439987302064,.0269374032410133,.0269374032410133,.0245439987302064,.019969996509099,.0136226771632526,.00607717674128413,.00476454936076254,.0106802748271453,.0156566178922341,.0192426678437941,.0211191138346674,.0211191138346674,.0192426678437941,.0156566178922341,.0106802748271453,.00476454936076254,.00179625418474627,.00402650637025399,.00590260762948584,.00725456281862631,.00796199046987792,.00796199046987792,.00725456281862631,.00590260762948584,.00402650637025399,.00179625418474628,-.00179625418474627,-.00402650637025399,-.00590260762948584,-.00725456281862631,-.00796199046987792,-.00796199046987792,-.00725456281862631,-.00590260762948584,-.00402650637025399,-.00179625418474628,-.00476454936076254,-.0106802748271453,-.0156566178922341,-.0192426678437941,-.0211191138346674,-.0211191138346674,-.0192426678437941,-.0156566178922341,-.0106802748271453,-.00476454936076254,-.00607717674128413,-.0136226771632526,-.019969996509099,-.0245439987302064,-.0269374032410133,-.0269374032410133,-.0245439987302064,-.019969996509099,-.0136226771632526,-.00607717674128413,-.00527840372812152,-.0118321373536874,-.0173451766357429,-.0213179803576284,-.0233968001502014,-.0233968001502014,-.0213179803576284,-.0173451766357429,-.0118321373536874,-.00527840372812152,-.00265100998409647,-.00594253790984439,-.008711390565351,-.0107066798372692,-.0117507401837501,-.0117507401837501,-.0107066798372692,-.008711390565351,-.00594253790984439,-.00265100998409647},
			{-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.002221727122666,-.00498025195339731,-.00730073926966875,-.00897292772598497,-.00984792148436088,-.00984792148436088,-.00897292772598497,-.00730073926966875,-.00498025195339731,-.002221727122666,.00309757218884701,.00694355746343096,.010178822902703,.0125102183310402,.0137301504746905,.0137301504746905,.0125102183310402,.010178822902703,.00694355746343096,.00309757218884701,.00727044463764254,.0162975217520129,.0238911521277625,.0293632704051698,.0322266255011462,.0322266255011462,.0293632704051698,.0238911521277625,.0162975217520129,.00727044463764254,.00727044463764255,.016297521752013,.0238911521277625,.0293632704051698,.0322266255011463,.0322266255011463,.0293632704051698,.0238911521277625,.016297521752013,.00727044463764255,.003097572188847,.00694355746343095,.010178822902703,.0125102183310401,.0137301504746904,.0137301504746904,.0125102183310401,.010178822902703,.00694355746343095,.003097572188847,-.00222172712266599,-.00498025195339728,-.00730073926966871,-.00897292772598493,-.00984792148436084,-.00984792148436084,-.00897292772598493,-.00730073926966871,-.00498025195339728,-.00222172712266599,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468},
			{.00352805200405569,.00790852652678675,.0115934074660505,.0142488046002341,.0156382747342036,.0156382747342036,.0142488046002341,.0115934074660505,.00790852652678675,.00352805200405569,.00299001628838423,.00670245878037739,.00982538724528936,.0120758304570704,.0132534033295837,.0132534033295837,.0120758304570704,.00982538724528936,.00670245878037739,.00299001628838423,-.0032119933533482,-.00720004541021951,-.0105548182625439,-.0129723330655266,-.0142373282611347,-.0142373282611347,-.0129723330655266,-.0105548182625439,-.00720004541021951,-.0032119933533482,-.00749937306045005,-.0168106906345655,-.0246434257572659,-.0302878475826814,-.0332413626893215,-.0332413626893215,-.0302878475826814,-.0246434257572659,-.0168106906345655,-.00749937306045005,-.00396370307706834,-.0088850875478213,-.0130249851175851,-.0160082761176901,-.0175693208639632,-.0175693208639632,-.0160082761176901,-.0130249851175851,-.0088850875478213,-.00396370307706834,.00396370307706838,.00888508754782138,.0130249851175852,.0160082761176902,.0175693208639633,.0175693208639633,.0160082761176902,.0130249851175852,.00888508754782138,.00396370307706837,.00749937306045007,.0168106906345656,.024643425757266,.0302878475826815,.0332413626893216,.0332413626893216,.0302878475826815,.024643425757266,.0168106906345656,.00749937306045007,.00321199335334816,.00720004541021943,.0105548182625438,.0129723330655265,.0142373282611345,.0142373282611345,.0129723330655265,.0105548182625438,.00720004541021943,.00321199335334816,-.00299001628838425,-.00670245878037744,-.00982538724528943,-.0120758304570705,-.0132534033295838,-.0132534033295838,-.0120758304570705,-.00982538724528943,-.00670245878037744,-.00299001628838425,-.0035280520040557,-.00790852652678675,-.0115934074660505,-.0142488046002341,-.0156382747342036,-.0156382747342036,-.0142488046002341,-.0115934074660505,-.00790852652678675,-.0035280520040557},
			{-.00355524726055973,-.00796948776183025,-.0116827728408898,-.0143586385526627,-.0157588191287273,-.0157588191287273,-.0143586385526627,-.0116827728408898,-.00796948776183025,-.00355524726055973,-.000198327000917931,-.000444572343586338,-.000651715374523445,-.000800986686075343,-.000879094787718522,-.000879094787718522,-.000800986686075343,-.000651715374523445,-.000444572343586338,-.000198327000917931,.00656585578393543,.0147181063597184,.0215758275045217,.0265176352250662,.0291034986151868,.0291034986151868,.0265176352250662,.0215758275045218,.0147181063597184,.00656585578393543,.00333253137587672,.00747024498425,.0109509140141708,.0134591520602391,.0147716193401924,.0147716193401924,.0134591520602391,.0109509140141708,.00747024498425,.00333253137587672,-.00614481295848751,-.0137742913733916,-.020192253500946,-.0248171622895082,-.0272372043055647,-.0272372043055647,-.0248171622895082,-.020192253500946,-.0137742913733916,-.00614481295848751,-.00614481295848755,-.0137742913733917,-.0201922535009461,-.0248171622895084,-.0272372043055649,-.0272372043055649,-.0248171622895084,-.0201922535009461,-.0137742913733917,-.00614481295848755,.00333253137587662,.00747024498424977,.0109509140141705,.0134591520602387,.014771619340192,.014771619340192,.0134591520602387,.0109509140141704,.00747024498424977,.00333253137587662,.00656585578393523,.014718106359718,.0215758275045211,.0265176352250655,.0291034986151859,.0291034986151859,.0265176352250655,.0215758275045211,.014718106359718,.00656585578393523,-.000198327000917742,-.000444572343585915,-.000651715374522825,-.000800986686074581,-.000879094787717686,-.000879094787717686,-.000800986686074581,-.000651715374522825,-.000444572343585915,-.000198327000917742,-.0035552472605598,-.00796948776183042,-.0116827728408901,-.014358638552663,-.0157588191287277,-.0157588191287277,-.014358638552663,-.0116827728408901,-.00796948776183042,-.0035552472605598},
			{-.00435807799231643,-.00895572847242053,-.0110946867695545,-.0103234510778548,-.00712570259555149,-.0027257963949736,.00134726370928414,.00379129532198382,.00397366731673714,.00213554377116887,-.00976912340532947,-.0200752755655013,-.0248699918602522,-.0231411800629203,-.015973066137951,-.00611018008563575,.0030200435739616,.00849861611742791,.00890742351499224,.00478706224964608,-.0143209266404436,-.0294291039975391,-.0364578595439588,-.0339235290930591,-.0234155203996869,-.00895714355686739,.0044271958372491,.0124584420641652,.0130577282547148,.00701753519287297,-.0176010449034518,-.0361696553536002,-.0448083031935061,-.0416935002770177,-.0287786982183586,-.0110087210073302,.00544121722596732,.0153119699376961,.0160485188646415,.00862485753488115,-.0193174082690916,-.0396967341001901,-.049177778443395,-.0457592359678989,-.0315850488415031,-.012082234855126,.00597181674221008,.0168051144868407,.0176134878766033,.00946590927856652,-.0193174082690916,-.0396967341001901,-.049177778443395,-.0457592359678989,-.0315850488415031,-.012082234855126,.00597181674221008,.0168051144868407,.0176134878766033,.00946590927856652,-.0176010449034518,-.0361696553536002,-.0448083031935061,-.0416935002770177,-.0287786982183586,-.0110087210073302,.00544121722596732,.0153119699376961,.0160485188646415,.00862485753488115,-.0143209266404436,-.0294291039975391,-.0364578595439588,-.0339235290930591,-.0234155203996869,-.00895714355686739,.0044271958372491,.0124584420641652,.0130577282547148,.00701753519287297,-.00976912340532947,-.0200752755655013,-.0248699918602522,-.0231411800629204,-.015973066137951,-.00611018008563575,.0030200435739616,.00849861611742791,.00890742351499224,.00478706224964608,-.00435807799231643,-.00895572847242053,-.0110946867695545,-.0103234510778548,-.00712570259555149,-.0027257963949736,.00134726370928414,.00379129532198382,.00397366731673714,.00213554377116887},
			{.00735144822140268,.0151070206329527,.0187151343924685,.0174141711549991,.0120200312533801,.00459802488506717,-.00227263931871786,-.00639536770584407,-.00670300292458653,-.00360235394730565,.0146374068493633,.0300794619817186,.0372635470036412,.0346732100211315,.0239329833386512,.0091550887552036,-.00452503307213473,-.0127337629596822,-.0133462928615887,-.00717261670817588,.0168524639340767,.0346313423832577,.0429025843441626,.0399202555051752,.027554726236705,.0105405147679591,-.00520979962048207,-.0146607444359217,-.0153659675800657,-.00825803816423586,.0132124503996921,.0271512163030548,.0336359282466497,.0312977614352663,.0216030994104901,.00826383780458816,-.00408451958996732,-.0114941268790076,-.0120470267902114,-.00647436007402489,.00498114669912899,.0102361172510779,.0126808796165353,.0117993813673682,.00814445496968726,.00311549993810543,-.00153988023853377,-.00433333185221062,-.00454177733236969,-.00244085966918394,-.00498114669912899,-.0102361172510779,-.0126808796165353,-.0117993813673682,-.00814445496968726,-.00311549993810543,.00153988023853377,.00433333185221062,.00454177733236969,.00244085966918394,-.0132124503996921,-.0271512163030548,-.0336359282466497,-.0312977614352663,-.0216030994104901,-.00826383780458816,.00408451958996732,.0114941268790076,.0120470267902114,.00647436007402489,-.0168524639340767,-.0346313423832577,-.0429025843441627,-.0399202555051752,-.027554726236705,-.0105405147679591,.00520979962048207,.0146607444359217,.0153659675800657,.00825803816423586,-.0146374068493633,-.0300794619817185,-.0372635470036412,-.0346732100211315,-.0239329833386512,-.0091550887552036,.00452503307213473,.0127337629596822,.0133462928615887,.00717261670817588,-.00735144822140268,-.0151070206329527,-.0187151343924685,-.0174141711549991,-.0120200312533801,-.00459802488506717,.00227263931871786,.00639536770584407,.00670300292458653,.00360235394730565},
			{-.00899207180285294,-.0184784562398471,-.0228917931799034,-.0213004938206645,-.0147025430701201,-.00562416665024674,.00277982450810196,.00782262268394569,.00819891289135883,.00440629170986428,-.0135981963807978,-.0279439135132015,-.034617950789736,-.0322115196955465,-.0222338157821648,-.00850510586048974,.00420376976453103,.0118297053005546,.0123987474800167,.00666338317747926,-.00616101485733601,-.0126607133406354,-.0156845586850965,-.0145942627878117,-.0100736057586743,-.00385345836332292,.0019046267056908,.00535975419633905,.00561757348530501,.00301901822910458,.00858979848715933,.0176517958190368,.0218676957586917,.0203475854739549,.0140448035769736,.00537256143445094,-.00265546504496181,-.00747266636314624,-.00783212268480781,-.00420916988800805,.0201614847183412,.0414312876127182,.051326607314859,.0477586912197692,.0329651613029748,.0126101695425049,-.00623275598422004,-.0175394159608113,-.0183831113218753,-.00987952330905502,.0201614847183412,.0414312876127182,.0513266073148591,.0477586912197692,.0329651613029748,.0126101695425049,-.00623275598422005,-.0175394159608113,-.0183831113218753,-.00987952330905502,.00858979848715931,.0176517958190368,.0218676957586917,.0203475854739549,.0140448035769735,.00537256143445092,-.0026554650449618,-.00747266636314622,-.00783212268480779,-.00420916988800804,-.00616101485733598,-.0126607133406353,-.0156845586850964,-.0145942627878116,-.0100736057586743,-.0038534583633229,.0019046267056908,.00535975419633902,.00561757348530498,.00301901822910456,-.0135981963807978,-.0279439135132015,-.034617950789736,-.0322115196955465,-.0222338157821648,-.00850510586048974,.00420376976453103,.0118297053005546,.0123987474800167,.00666338317747926,-.00899207180285294,-.0184784562398471,-.0228917931799034,-.0213004938206645,-.0147025430701201,-.00562416665024674,.00277982450810196,.00782262268394569,.00819891289135883,.00440629170986428},
			{.00978355109164593,.0201049240576492,.0249067215059292,.0231753564852141,.0159966562164246,.00611920400292648,-.0030245037736683,-.0085111674347139,-.00892057858602882,-.00479413209917604,.00829153796163795,.0170388787750279,.0211083915167262,.0196410634821265,.0135571308449561,.00518601188975199,-.00256326027426807,-.0072131956200491,-.00756017066737565,-.00406301637524721,-.0089070969028758,-.018303835195316,-.0226754661889265,-.0210992045770331,-.0145636043300628,-.00557101838708364,.00275355522170069,.00774869905491992,.00812143332734517,.00436465234070388,-.0207963202945656,-.0427358569791504,-.0529427559658922,-.049262494966556,-.0340031531703131,-.0130072327727005,.00642901014367235,.0180916890395887,.0189619502929062,.0101906052040814,-.010991643978613,-.0225875211759695,-.0279822543883277,-.0260370968758398,-.0179719560241624,-.00687481582123123,.00339797568190965,.00956214378689379,.0100221098640343,.0053861213302785,.0109916439786131,.0225875211759697,.027982254388328,.0260370968758401,.0179719560241625,.00687481582123129,-.00339797568190968,-.00956214378689387,-.0100221098640343,-.00538612133027854,.0207963202945656,.0427358569791505,.0529427559658923,.0492624949665561,.0340031531703132,.0130072327727006,-.00642901014367236,-.0180916890395887,-.0189619502929062,-.0101906052040814,.00890709690287571,.0183038351953158,.0226754661889263,.0210992045770329,.0145636043300626,.00557101838708358,-.00275355522170066,-.00774869905491983,-.00812143332734509,-.00436465234070383,-.00829153796163801,-.017038878775028,-.0211083915167264,-.0196410634821266,-.0135571308449562,-.00518601188975203,.00256326027426809,.00721319562004915,.00756017066737571,.00406301637524725,-.00978355109164594,-.0201049240576493,-.0249067215059292,-.0231753564852141,-.0159966562164247,-.00611920400292649,.00302450377366831,.0085111674347139,.00892057858602882,.00479413209917604},
			{-.00774458111624399,-.0136429110526415,-.0119270111969813,-.00443011400826239,.00356799186187752,.00716050023137007,.00509898471326269,.000227342285586894,-.00308610359639842,-.00244256114805105,-.0173603521507787,-.0305821240271899,-.0267357409494468,-.00993060026064162,.00799805622325859,.0160510689637666,.0114299493936489,.000509613377058344,-.00691784931981508,-.0054752763310899,-.025449195315512,-.0448314896363585,-.0391929315383665,-.0145576416560195,.0117246524265364,.0235298676855081,.0167555941284487,.000747061479831429,-.0101411363648726,-.00802641418481,-.0312781735952783,-.0550998606437757,-.0481698262427054,-.0178919780059722,.0144101103942559,.0289192360315085,.0205933576816159,.000918171217707236,-.0124638999285189,-.0098648139207399,-.0343282602007917,-.0604729156402031,-.0528670999299058,-.0196367116713162,.0158153102395603,.0317392911793161,.0226015159980186,.00100770655212074,-.0136793153398003,-.0108267798332916,-.0343282602007917,-.0604729156402031,-.0528670999299058,-.0196367116713162,.0158153102395603,.0317392911793161,.0226015159980186,.00100770655212074,-.0136793153398003,-.0108267798332916,-.0312781735952783,-.0550998606437757,-.0481698262427054,-.0178919780059722,.0144101103942559,.0289192360315085,.0205933576816159,.000918171217707236,-.0124638999285189,-.0098648139207399,-.025449195315512,-.0448314896363585,-.0391929315383665,-.0145576416560195,.0117246524265364,.0235298676855081,.0167555941284487,.000747061479831429,-.0101411363648726,-.00802641418481,-.0173603521507787,-.0305821240271899,-.0267357409494468,-.00993060026064162,.00799805622325859,.0160510689637666,.0114299493936489,.000509613377058344,-.00691784931981508,-.0054752763310899,-.00774458111624399,-.0136429110526415,-.0119270111969814,-.00443011400826239,.00356799186187752,.00716050023137007,.0050989847132627,.000227342285586894,-.00308610359639842,-.00244256114805105},
			{.0130639899374604,.0230136207541996,.0201191454931664,.00747296257306306,-.0060186870123603,-.0120787298326045,-.00860125085985731,-.000383493604971994,.00520581109864628,.00412024792313705,.0260116007120718,.0458222271192026,.0400590617216945,.0148793530550254,-.0119837571925505,-.0240498575870592,-.0171258783925904,-.000763570898011854,.0103652467836164,.00820379105651837,.0299478976965691,.0527564368369676,.0461211401613239,.0171310235004581,-.0137972414076181,-.0276892868919076,-.0197175121878277,-.000879120950335205,.0119338042172612,.00944525859843065,.0234793626877832,.041361417992028,.0361592986724362,.0134308430613569,-.0108171344240126,-.0217085959117477,-.0154586684063579,-.000689237015849483,.00935618654441216,.0074051492548628,.00885181373718756,.0155934201809443,.0136322003698287,.00506348160695133,-.00407810298621525,-.00818422757303578,-.0058279798807838,-.000259845114460841,.00352731978643567,.00279177091694338,-.00885181373718756,-.0155934201809443,-.0136322003698287,-.00506348160695133,.00407810298621525,.00818422757303578,.0058279798807838,.000259845114460841,-.00352731978643567,-.00279177091694338,-.0234793626877832,-.041361417992028,-.0361592986724362,-.0134308430613569,.0108171344240126,.0217085959117477,.0154586684063579,.000689237015849483,-.00935618654441216,-.0074051492548628,-.0299478976965691,-.0527564368369676,-.0461211401613239,-.0171310235004581,.0137972414076181,.0276892868919076,.0197175121878277,.000879120950335204,-.0119338042172612,-.00944525859843065,-.0260116007120718,-.0458222271192026,-.0400590617216945,-.0148793530550254,.0119837571925505,.0240498575870592,.0171258783925904,.000763570898011854,-.0103652467836164,-.00820379105651837,-.0130639899374604,-.0230136207541996,-.0201191454931664,-.00747296257306306,.0060186870123603,.0120787298326045,.00860125085985731,.000383493604971994,-.00520581109864628,-.00412024792313705},
			{-.0159794821505739,-.028149573279088,-.0246091376242554,-.00914070453359377,.00736187812026119,.0147743414290495,.0105207930537044,.000469077919138421,-.00636759259064836,-.00503976415007148,-.0241648577892824,-.0425689912043686,-.0372149926639303,-.0138229651666283,.0111329476238222,.0223423923378182,.0159099941773837,.000709359734017749,-.00962934643082784,-.00762134735221235,-.010948514324695,-.0192869833562834,-.0168612157301903,-.00626285217385856,.0050440700954407,.0101227991776989,.00720843468957457,.000321393789153591,-.00436282465449179,-.0034530486952082,.0152646169438994,.0268902614734654,.0235082123196634,.00873178191808065,-.0070325338727842,-.0141133899326471,-.0100501301855435,-.00044809304112726,.00608272914198089,.00481430302393592,.0358282376130031,.0631152868812828,.055177134148947,.020494740123233,-.0165063621014286,-.0331261432822222,-.0235890919276015,-.00105173840976448,.0142770346504849,.0112998572657607,.0358282376130031,.0631152868812828,.0551771341489471,.020494740123233,-.0165063621014286,-.0331261432822222,-.0235890919276015,-.00105173840976448,.0142770346504849,.0112998572657607,.0152646169438994,.0268902614734653,.0235082123196633,.00873178191808063,-.00703253387278418,-.0141133899326471,-.0100501301855435,-.000448093041127259,.00608272914198088,.0048143030239359,-.0109485143246949,-.0192869833562833,-.0168612157301902,-.00626285217385853,.00504407009544068,.0101227991776988,.00720843468957453,.00032139378915359,-.00436282465449177,-.00345304869520818,-.0241648577892824,-.0425689912043686,-.0372149926639303,-.0138229651666283,.0111329476238222,.0223423923378182,.0159099941773837,.000709359734017748,-.00962934643082784,-.00762134735221235,-.0159794821505739,-.028149573279088,-.0246091376242554,-.00914070453359377,.00736187812026119,.0147743414290495,.0105207930537044,.000469077919138421,-.00636759259064836,-.00503976415007148},
			{-.0113741780920774,-.0157705273052622,-.00608002687266981,.00608414677874269,.00833345789650543,.000957688797273868,-.00550488775656931,-.00405218361689488,.00128277564401616,.00255185410234015,-.0254965031859464,-.0353514158497966,-.0136290660542447,.0136383012884271,.0186803858782727,.00214676746514514,-.0123398268505581,-.00908342665833277,.00287548630149319,.00572026881623259,-.0373762861378719,-.0518229744896624,-.019979362225491,.0199929005037731,.0273842825684829,.00314702743601977,-.0180894178270092,-.0133157379040649,.00421528387662904,.0083855579136398,-.0459370896279213,-.063692701172494,-.0245555096050894,.0245721487409458,.0336564804246319,.00386783429623537,-.0222326853173956,-.0163656240029997,.005180768162293,.0103062172640176,-.0504166383248294,-.0699036857788496,-.0269500365972072,.0269682982960067,.0369384872790024,.00424500560207766,-.024400702432737,-.0179615154769812,.00568597002550107,.0113112265601881,-.0504166383248294,-.0699036857788496,-.0269500365972072,.0269682982960067,.0369384872790024,.00424500560207766,-.024400702432737,-.0179615154769812,.00568597002550107,.0113112265601881,-.0459370896279213,-.063692701172494,-.0245555096050894,.0245721487409458,.0336564804246319,.00386783429623537,-.0222326853173956,-.0163656240029997,.005180768162293,.0103062172640176,-.0373762861378719,-.0518229744896624,-.019979362225491,.0199929005037731,.0273842825684829,.00314702743601977,-.0180894178270092,-.0133157379040649,.00421528387662904,.0083855579136398,-.0254965031859464,-.0353514158497966,-.0136290660542447,.0136383012884271,.0186803858782727,.00214676746514514,-.0123398268505581,-.00908342665833277,.0028754863014932,.00572026881623259,-.0113741780920774,-.0157705273052622,-.00608002687266981,.0060841467787427,.00833345789650543,.000957688797273868,-.00550488775656931,-.00405218361689488,.00128277564401616,.00255185410234015},
			{.019186595894105,.0266026021391187,.0102561273163527,-.0102630770029412,-.014057340035159,-.00161548270097898,.00928595068344064,.00683544858509201,-.00216385726557964,-.00430460935690978,.0382022700423445,.052968217830653,.0204208890149373,-.0204347264775717,-.0279894517539301,-.00321657404638918,.0184891784643094,.0136100043045404,-.00430843803919177,-.00857087155986831,.0439833629490508,.0609838197318553,.0235111518841866,-.0235270833495331,-.0322250540052594,-.00370333342961076,.0212871184388807,.015669586084838,-.00496042758221491,-.00986788885031005,.0344832662837457,.0478117441190089,.0184329086431872,-.0184453990242086,-.025264669269575,-.00290343948776025,.016689250760381,.0122850658360403,-.00388901015588316,-.00773649434850662,.0130003294489922,.0180252189558684,.00694928035797946,-.00695398928163203,-.00952488146632226,-.00109460831133298,.00629191435511048,.00463151900567974,-.00146617239912651,-.00291668934384738,-.0130003294489922,-.0180252189558684,-.00694928035797946,.00695398928163203,.00952488146632226,.00109460831133298,-.00629191435511048,-.00463151900567974,.00146617239912651,.00291668934384738,-.0344832662837457,-.0478117441190089,-.0184329086431872,.0184453990242085,.025264669269575,.00290343948776025,-.016689250760381,-.0122850658360403,.00388901015588316,.00773649434850662,-.0439833629490508,-.0609838197318553,-.0235111518841866,.0235270833495331,.0322250540052594,.00370333342961076,-.0212871184388807,-.015669586084838,.00496042758221491,.00986788885031005,-.0382022700423445,-.052968217830653,-.0204208890149373,.0204347264775717,.0279894517539301,.00321657404638918,-.0184891784643094,-.0136100043045404,.00430843803919177,.00857087155986831,-.019186595894105,-.0266026021391187,-.0102561273163527,.0102630770029412,.014057340035159,.00161548270097898,-.00928595068344064,-.00683544858509201,.00216385726557964,.00430460935690978},
			{-.0149432388948351,-.0143716720347549,.00337087856178778,.00991288633119597,-.000790460352998803,-.0073875487912777,-.000452661276806676,.00518463670922043,.000881265822650354,-.00248741703478494,-.0334969555607631,-.032215723971892,.00755620452698254,.022220852872113,-.00177190604415069,-.016560023921583,-.00101469131156782,.0116219480040223,.00197545674711399,-.00557582586088984,-.0491044511733968,-.0472262454396426,.0110769253515695,.0325743867352538,-.0025975039334847,-.0242759639637272,-.0014874742832843,.0170370521365157,.00289589659000063,-.00817381353480803,-.0603515171721834,-.0580431202166143,.0136140255026904,.0400353453393464,-.00319244588831583,-.0298362209742388,-.00182817092148764,.02093928187792,.00355918350788891,-.0100459741656872,-.0662366867006586,-.063703186747533,.0149415952441392,.0439393862896491,-.00350375679056903,-.0327456957770259,-.00200644474630778,.0229811729426385,.00390625677643855,-.011025605893504,-.0662366867006586,-.063703186747533,.0149415952441392,.0439393862896491,-.00350375679056903,-.0327456957770259,-.00200644474630778,.0229811729426385,.00390625677643855,-.011025605893504,-.0603515171721834,-.0580431202166143,.0136140255026904,.0400353453393464,-.00319244588831583,-.0298362209742388,-.00182817092148764,.02093928187792,.00355918350788891,-.0100459741656872,-.0491044511733968,-.0472262454396426,.0110769253515695,.0325743867352538,-.0025975039334847,-.0242759639637272,-.0014874742832843,.0170370521365157,.00289589659000063,-.00817381353480803,-.0334969555607631,-.032215723971892,.00755620452698254,.022220852872113,-.00177190604415069,-.016560023921583,-.00101469131156782,.0116219480040223,.00197545674711399,-.00557582586088984,-.0149432388948351,-.0143716720347549,.00337087856178778,.00991288633119597,-.000790460352998802,-.0073875487912777,-.000452661276806676,.00518463670922043,.000881265822650354,-.00248741703478494},
			{5.02239597246239e-005,.000582196454509688,.00202770964403212,.0044045611294125,.00726150487830256,.00980179190824761,.0111426514550797,.0106221354138729,.00804698659569713,.00379931823382166,.000112582670920343,.00130505902378937,.0045453398901192,.00987332060036796,.0162774822730591,.0219718221917673,.0249775101132113,.0238107146933112,.0180382281439193,.00851660037928648,.000165039185640327,.00191313526973573,.00666318526458248,.0144736732405605,.0238617754993688,.0322093232636935,.0366154746082026,.0349050251728234,.0264429193136947,.0124848058722647,.000202840374105465,.00235132688229124,.00818934598203497,.0177887771525588,.0293271652446273,.0395866663734618,.0450020189978578,.0428998018667281,.032499503831288,.0153443722103867,.00022262032405145,.00258061618548581,.0089879288791991,.0195234472015054,.0321869995510595,.0434469544588521,.0493903844165837,.0470831698838632,.0356686882793406,.0168406764624987,.00022262032405145,.00258061618548581,.0089879288791991,.0195234472015054,.0321869995510595,.0434469544588521,.0493903844165837,.0470831698838632,.0356686882793406,.0168406764624987,.000202840374105465,.00235132688229124,.00818934598203497,.0177887771525588,.0293271652446273,.0395866663734618,.0450020189978578,.0428998018667281,.032499503831288,.0153443722103867,.000165039185640327,.00191313526973573,.00666318526458248,.0144736732405605,.0238617754993688,.0322093232636935,.0366154746082026,.0349050251728234,.0264429193136947,.0124848058722647,.000112582670920343,.00130505902378937,.0045453398901192,.00987332060036795,.0162774822730591,.0219718221917673,.0249775101132113,.0238107146933112,.0180382281439193,.00851660037928648,5.02239597246239e-005,.000582196454509688,.00202770964403212,.0044045611294125,.00726150487830256,.00980179190824761,.0111426514550797,.0106221354138729,.00804698659569713,.00379931823382166},
			{-8.47205671528915e-005,-.00098208134306624,-.00342045334548888,-.00742985855733822,-.0122491100870261,-.0165342074606321,-.0187960438809947,-.0179180084878277,-.0135741043119093,-.00640890119633241,-.000168686410157299,-.00195541391909401,-.00681043594667852,-.0147935290111086,-.0243891002815539,-.0329211216952395,-.0374246452071332,-.0356763963055754,-.0270272851613995,-.0127607093790004,-.000194213474599731,-.0022513238093916,-.0078410491248818,-.0170322118311734,-.0280798666805756,-.0379030262496602,-.0430880613000287,-.0410752524832421,-.0311172841682885,-.0146917686169386,-.000152264731741222,-.00176505372042869,-.00614743773072645,-.0133533740168041,-.0220148132165338,-.0297162394935752,-.033781343486181,-.0322032871995818,-.0243961699164533,-.0115184500553977,-5.74044134859514e-005,-.000665432319314789,-.00231760863686745,-.00503427546699351,-.00829967272293673,-.0112031412634403,-.0127357017440349,-.012140768205952,-.00919745373300046,-.00434250178709194,5.74044134859514e-005,.000665432319314789,.00231760863686745,.00503427546699351,.00829967272293673,.0112031412634403,.0127357017440349,.012140768205952,.00919745373300046,.00434250178709194,.000152264731741222,.00176505372042869,.00614743773072645,.013353374016804,.0220148132165338,.0297162394935752,.033781343486181,.0322032871995818,.0243961699164533,.0115184500553977,.000194213474599731,.0022513238093916,.0078410491248818,.0170322118311734,.0280798666805756,.0379030262496602,.0430880613000287,.0410752524832421,.0311172841682885,.0146917686169386,.000168686410157299,.00195541391909401,.00681043594667852,.0147935290111086,.0243891002815539,.0329211216952395,.0374246452071332,.0356763963055754,.0270272851613995,.0127607093790004,8.47205671528915e-005,.00098208134306624,.00342045334548889,.00742985855733822,.0122491100870261,.0165342074606321,.0187960438809947,.0179180084878277,.0135741043119093,.00640890119633241},
			{.000103627666330889,.00120125255471199,.00418379633067172,.00908798098667198,.0149827454608269,.0202241485315363,.02299075925832,.021916772600681,.0166034388065025,.00783917644841335,.000156710198511108,.00181658559896353,.00632691612890426,.0137432343592401,.0226575498468763,.0305838242171288,.0347676115352073,.0331434828804184,.0251084317871825,.0118547386126776,7.10016118524656e-005,.000823051127621699,.00286657312335497,.00622672806775283,.0102655894449765,.0138567932186921,.0157523663597753,.0150165128324275,.0113760249486971,.0053710961863794,-9.89917330502748e-005,-.00114751278719633,-.00399662816086328,-.00868141703515759,-.0143124707091497,-.0193193920454638,-.0219622344466939,-.0209362941329066,-.0158606599978981,-.00748847957105857,-.000232347745540099,-.00269337651607831,-.00938065749861198,-.0203765265447757,-.0335934143176455,-.0453453742919027,-.0515485031272343,-.0491404745815968,-.0372272358482508,-.0175765318198328,-.000232347745540099,-.00269337651607831,-.00938065749861198,-.0203765265447757,-.0335934143176455,-.0453453742919027,-.0515485031272343,-.0491404745815968,-.0372272358482508,-.0175765318198328,-9.89917330502745e-005,-.00114751278719633,-.00399662816086327,-.00868141703515757,-.0143124707091497,-.0193193920454637,-.0219622344466939,-.0209362941329065,-.0158606599978981,-.00748847957105856,7.10016118524653e-005,.000823051127621695,.00286657312335496,.0062267280677528,.0102655894449765,.0138567932186921,.0157523663597753,.0150165128324274,.0113760249486971,.00537109618637938,.000156710198511108,.00181658559896353,.00632691612890426,.0137432343592401,.0226575498468763,.0305838242171288,.0347676115352073,.0331434828804184,.0251084317871825,.0118547386126776,.000103627666330889,.00120125255471199,.00418379633067172,.00908798098667198,.0149827454608269,.0202241485315363,.02299075925832,.021916772600681,.0166034388065025,.00783917644841335},
			{-.000112748940416249,-.00130698642082339,-.00455205274713417,-.00988790216985572,-.0163015219298734,-.0220042716244384,-.0250143985435671,-.0238458799234204,-.0180648681860857,-.00852917825508497,-9.55544782092518e-005,-.00110766810763179,-.0038578546585677,-.00837997527015808,-.0138155038643826,-.0186485716467526,-.0211996475685226,-.0202093306164191,-.015309935925496,-.00722845975057063,.000102648386927895,.00118990074176208,.00414425953786325,.0090020997455888,.0148411588117245,.0200330307267911,.0227734970367722,.0217096595318657,.0164465366372481,.00776509638558674,.000239663804666116,.00277818431911703,.00967603133467496,.0210181332564652,.0346511883228178,.0467731886166761,.0531716387245257,.0506877872817214,.0383994299572614,.018129973583278,.000126671410044506,.00146837577562602,.00511414952502166,.0111088805412601,.0183144671790717,.0247214040626872,.0281032276066716,.0267904179187864,.0202955550353876,.00958237862017949,-.000126671410044507,-.00146837577562603,-.0051141495250217,-.0111088805412602,-.0183144671790719,-.0247214040626874,-.0281032276066718,-.0267904179187866,-.0202955550353878,-.00958237862017957,-.000239663804666117,-.00277818431911704,-.00967603133467498,-.0210181332564652,-.0346511883228178,-.0467731886166762,-.0531716387245259,-.0506877872817215,-.0383994299572615,-.018129973583278,-.000102648386927894,-.00118990074176207,-.00414425953786321,-.0090020997455887,-.0148411588117243,-.0200330307267909,-.0227734970367719,-.0217096595318655,-.016446536637248,-.00776509638558665,9.55544782092525e-005,.0011076681076318,.00385785465856773,.00837997527015815,.0138155038643827,.0186485716467528,.0211996475685227,.0202093306164193,.0153099359254961,.00722845975057069,.000112748940416249,.00130698642082339,.00455205274713418,.00988790216985573,.0163015219298734,.0220042716244384,.0250143985435671,.0238458799234204,.0180648681860857,.00852917825508498},
			{.000242033524403197,.00261162972756074,.00794330061991523,.0139365347274024,.0166502677108084,.0135390642006158,.00568409316229951,-.00258267825885607,-.00653083370918549,-.00434965045780644,.000542545446017795,.00585426262277764,.0178058043335604,.0312403146145367,.0373234532024064,.030349339594584,.01274153598164,-.00578936463998794,-.0146396004287218,-.00975023213611034,.000795337842419811,.0085819844909475,.0261022004795129,.0457963560553539,.0547138584602276,.0444902421524166,.0186782984008672,-.00848684808903276,-.0214607426978308,-.0142932332160281,.000977504977807211,.0105476341146326,.0320807454890066,.0562857236534088,.067245723951962,.0546804525678283,.0229564452865287,-.0104307073176385,-.0263761909665255,-.0175670084741824,.00107282623531448,.011576185140315,.0352090947774598,.0617744179108896,.0738031810643202,.0600126090459714,.0251950397513031,-.0114478562435068,-.0289482614400908,-.0192800527823109,.00107282623531448,.011576185140315,.0352090947774598,.0617744179108896,.0738031810643202,.0600126090459713,.0251950397513031,-.0114478562435068,-.0289482614400908,-.0192800527823109,.000977504977807211,.0105476341146326,.0320807454890066,.0562857236534088,.067245723951962,.0546804525678283,.0229564452865287,-.0104307073176385,-.0263761909665255,-.0175670084741824,.000795337842419811,.0085819844909475,.0261022004795129,.0457963560553539,.0547138584602276,.0444902421524166,.0186782984008672,-.00848684808903276,-.0214607426978308,-.0142932332160281,.000542545446017795,.00585426262277763,.0178058043335604,.0312403146145367,.0373234532024064,.030349339594584,.01274153598164,-.00578936463998794,-.0146396004287218,-.00975023213611034,.000242033524403197,.00261162972756074,.00794330061991523,.0139365347274024,.0166502677108084,.0135390642006158,.0056840931622995,-.00258267825885607,-.00653083370918549,-.00434965045780644},
			{-.000408275603560732,-.00440544220444747,-.0133992010522381,-.0235089215074875,-.0280865971597952,-.0228384461275931,-.00958824432382064,.00435660525762981,.0110165733484533,.00733723218846652,-.000812914127394553,-.00877163899626159,-.0266790367491162,-.0468084162917745,-.0559229878604316,-.045473438390492,-.0190910727943728,.00867440016136512,.0219350067266604,.0146091014255062,-.000935931217489798,-.0100990381252994,-.0307163358399061,-.053891864555414,-.0643857307310363,-.0523548664268737,-.0219800965458574,.00998708428163375,.0252543987851819,.0168198750923781,-.00073377666535768,-.00791771700785594,-.0240818236035156,-.0422516013189939,-.0504788662986896,-.0410465839626478,-.0172325504762159,.00782994440604093,.0197996264895458,.0131869005183092,-.000276636740648008,-.0029850110119799,-.00907894391448041,-.0159290228592211,-.0190307346959162,-.0154747264913728,-.00649674052317934,.00295192038968209,.0074645384579598,.00497151429155405,.000276636740648008,.0029850110119799,.00907894391448041,.0159290228592211,.0190307346959162,.0154747264913728,.00649674052317934,-.00295192038968209,-.0074645384579598,-.00497151429155405,.00073377666535768,.00791771700785594,.0240818236035156,.0422516013189938,.0504788662986896,.0410465839626478,.0172325504762159,-.00782994440604093,-.0197996264895458,-.0131869005183092,.000935931217489798,.0100990381252994,.0307163358399061,.053891864555414,.0643857307310362,.0523548664268737,.0219800965458574,-.00998708428163375,-.0252543987851819,-.0168198750923781,.000812914127394553,.00877163899626159,.0266790367491162,.0468084162917745,.0559229878604317,.0454734383904919,.0190910727943728,-.00867440016136512,-.0219350067266604,-.0146091014255062,.000408275603560732,.00440544220444747,.0133992010522381,.0235089215074875,.0280865971597952,.0228384461275931,.00958824432382064,-.00435660525762981,-.0110165733484533,-.00733723218846652},
			{.000499390519149255,.00538860527147271,.0163895023638538,.0287554103499522,.0343546864285749,.0279353048988254,.0117280539636516,-.00532886937735551,-.0134751433486315,-.00897468317910862,.000755199746953267,.00814887984733841,.0247849079292855,.0434851639891283,.0519526292603973,.042244965376185,.0177356658686522,-.0080585446679718,-.0203776893169333,-.0135719005586946,.000342162793575017,.00369206094721455,.0112294440905165,.0197020791513965,.0235384834714527,.019140175067471,.00803560250650699,-.00365113225589165,-.00923263961808127,-.00614910085446924,-.000477049563208446,-.00514753823408188,-.0156562943109128,-.0274689955482004,-.0328177799266881,-.0266855786985745,-.0112033825355281,.00509047471144027,.0128723133542602,.00857318776860553,-.00111970350560565,-.0120820079305306,-.0367475603725958,-.0644736584683689,-.0770280198622701,-.0626348671549949,-.0262959400178523,.0119480716872726,.0302131591759193,.0201224969878625,-.00111970350560565,-.0120820079305306,-.0367475603725958,-.064473658468369,-.0770280198622701,-.0626348671549949,-.0262959400178523,.0119480716872726,.0302131591759193,.0201224969878625,-.000477049563208445,-.00514753823408186,-.0156562943109127,-.0274689955482003,-.0328177799266881,-.0266855786985745,-.0112033825355281,.00509047471144026,.0128723133542602,.00857318776860551,.000342162793575015,.00369206094721453,.0112294440905164,.0197020791513964,.0235384834714526,.0191401750674709,.00803560250650695,-.00365113225589163,-.00923263961808122,-.00614910085446921,.000755199746953267,.00814887984733841,.0247849079292855,.0434851639891283,.0519526292603973,.042244965376185,.0177356658686522,-.0080585446679718,-.0203776893169333,-.0135719005586946,.000499390519149255,.00538860527147271,.0163895023638538,.0287554103499522,.0343546864285749,.0279353048988254,.0117280539636516,-.00532886937735551,-.0134751433486316,-.00897468317910862},
			{.000682727668349805,.00664570504694168,.0164335220422574,.0198480228837579,.0106420257503786,-.00420811508616648,-.0112538525390341,-.00576778839286375,.00325440242770835,.00455104492220568,.00153041107940269,.0148970974896404,.0368375933377187,.0444915821252141,.0238553011261929,-.00943296463557281,-.0252267798862333,-.0129291482816634,.00729511013408645,.010201680545096,.00224348735178138,.0218382173561541,.0540016181505257,.065221889139497,.0349703861072661,-.0138281388148133,-.0369808885747392,-.0189533263510734,.0106941772285541,.0149550284742905,.00275734403299806,.0268401238235168,.0663703494746075,.0801605530322499,.0429801333125165,-.0169953871227395,-.0454511287373907,-.0232944666606214,.0131436113270545,.0183803837780314,.00302622604032566,.0294574346429948,.0728424445705406,.0879773978473551,.0471713348390493,-.0186526898568866,-.0498832890278595,-.0255660232311389,.014425308697734,.0201727442620858,.00302622604032566,.0294574346429948,.0728424445705406,.0879773978473551,.0471713348390493,-.0186526898568866,-.0498832890278595,-.0255660232311389,.014425308697734,.0201727442620858,.00275734403299806,.0268401238235168,.0663703494746075,.0801605530322499,.0429801333125165,-.0169953871227395,-.0454511287373907,-.0232944666606214,.0131436113270545,.0183803837780314,.00224348735178138,.0218382173561541,.0540016181505257,.065221889139497,.0349703861072661,-.0138281388148133,-.0369808885747392,-.0189533263510734,.0106941772285541,.0149550284742905,.00153041107940269,.0148970974896404,.0368375933377187,.0444915821252141,.0238553011261929,-.00943296463557281,-.0252267798862333,-.0129291482816634,.00729511013408645,.010201680545096,.000682727668349805,.00664570504694168,.0164335220422574,.0198480228837579,.0106420257503786,-.00420811508616648,-.0112538525390341,-.00576778839286375,.00325440242770835,.00455104492220568},
			{-.00115166298286166,-.0112103446987566,-.0277209784976989,-.0334807483245332,-.0179515606112687,.00709848245065787,.0189836240489798,.00972942608448226,-.00548970692279972,-.00767695556108528,-.0022930664987097,-.0223208232358323,-.0551950076111466,-.0666632369688139,-.0357432016569395,.0141337288264978,.0377981345051569,.0193721786130192,-.0109305093762129,-.0152855217812643,-.00264007285345256,-.025698600334876,-.0635475863095515,-.0767512858190868,-.0411521673894173,.0162725650625018,.0435180701799816,.0223037416914183,-.0125846071602312,-.0175986571380644,-.00206983570855075,-.0201478836325818,-.0498217551700164,-.0601735485661986,-.0322635890276594,.0127578056007183,.0341184734761949,.017486290549451,-.00986642063466756,-.0137974711263254,-.000780336349086855,-.00759583279519514,-.0187830011695426,-.0226856687251885,-.0121635022365666,.00480974378966284,.0128628011000313,.00659240154668345,-.00371968008127681,-.00520170185530649,.000780336349086855,.00759583279519514,.0187830011695426,.0226856687251885,.0121635022365666,-.00480974378966284,-.0128628011000313,-.00659240154668345,.00371968008127681,.00520170185530649,.00206983570855075,.0201478836325818,.0498217551700164,.0601735485661986,.0322635890276594,-.0127578056007183,-.0341184734761949,-.017486290549451,.00986642063466756,.0137974711263254,.00264007285345256,.025698600334876,.0635475863095515,.0767512858190868,.0411521673894173,-.0162725650625018,-.0435180701799816,-.0223037416914183,.0125846071602312,.0175986571380644,.0022930664987097,.0223208232358323,.0551950076111466,.0666632369688139,.0357432016569395,-.0141337288264978,-.0377981345051569,-.0193721786130192,.0109305093762129,.0152855217812643,.00115166298286166,.0112103446987566,.0277209784976989,.0334807483245332,.0179515606112687,-.00709848245065787,-.0189836240489798,-.00972942608448226,.00548970692279973,.00767695556108528},
			{.00148169134577296,.0125696636361974,.0228905745669076,.0127235476313918,-.00841173160413407,-.0124282780657018,.00206969137961716,.00930159426344669,.000766794088585497,-.00444614718711676,.00332137828441454,.0281763188823123,.0513118049188827,.0285212672154559,-.0188558452230951,-.0278593871779371,.00463944668591847,.0208505727492996,.0017188554429443,-.0099665404395655,.00486893376025062,.0413047290907605,.0752199110957276,.0418104018690363,-.027641495043065,-.0408401269458114,.00680113996769058,.0305656413953226,.00251973505529715,-.0146103277205166,.00598413244462014,.0507653177552249,.0924485590960711,.0513868117060249,-.0339726058003504,-.0501943014082524,.00835889834306659,.0375665095010369,.00309686453723171,-.0179567314824952,.00656767426042838,.055715690424466,.101463667057506,.0563977892686102,-.0372854395753275,-.0550889915004361,.00917401511436313,.0412298022791773,.00339885483776274,-.0197077795738015,.00656767426042838,.055715690424466,.101463667057506,.0563977892686102,-.0372854395753275,-.0550889915004361,.00917401511436313,.0412298022791773,.00339885483776274,-.0197077795738015,.00598413244462014,.0507653177552249,.0924485590960711,.0513868117060249,-.0339726058003504,-.0501943014082523,.0083588983430666,.0375665095010369,.00309686453723171,-.0179567314824952,.00486893376025062,.0413047290907605,.0752199110957276,.0418104018690363,-.027641495043065,-.0408401269458114,.00680113996769058,.0305656413953226,.00251973505529715,-.0146103277205166,.00332137828441454,.0281763188823123,.0513118049188827,.0285212672154559,-.0188558452230951,-.0278593871779371,.00463944668591847,.0208505727492996,.0017188554429443,-.0099665404395655,.00148169134577296,.0125696636361974,.0228905745669076,.0127235476313918,-.00841173160413407,-.0124282780657018,.00206969137961716,.00930159426344669,.000766794088585497,-.00444614718711676},
			{-1.03605502214324e-006,-6.2106840321439e-005,-.000513920961277903,-.00197298030906211,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186,-2.32243419779848e-006,-.000139219488151642,-.00115201180432313,-.00442265791238583,-.0109526932361906,-.0199562351910325,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-3.404543928249e-006,-.000204087101167348,-.00168877757544104,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-4.18433332293283e-006,-.000250831969330619,-.00207558146784554,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-4.59236798590328e-006,-.000275291812791727,-.00227798149655704,-.00874533824411281,-.0216577924252842,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.59236798590328e-006,-.000275291812791727,-.00227798149655704,-.00874533824411281,-.0216577924252843,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.18433332293283e-006,-.000250831969330619,-.00207558146784554,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-3.404543928249e-006,-.000204087101167348,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-2.32243419779848e-006,-.000139219488151642,-.00115201180432313,-.00442265791238583,-.0109526932361906,-.0199562351910326,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-1.03605502214324e-006,-6.21068403214391e-005,-.000513920961277903,-.00197298030906211,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186},
			{1.74767520475977e-006,.000104765270720098,.000866910445908273,.00332813286160627,.00824210667082646,.0150174405185659,.0212996498258157,.0237895578215264,.0200145017350098,.0100011552359343,3.47978143111129e-006,.000208597251185043,.001726098112448,.00662661740611046,.0164107894122825,.0299010997678524,.0424095539898868,.0473671889011406,.0398507064131546,.0199132162457208,4.00637159776859e-006,.000240163964049163,.00198730598159011,.00762941359700327,.0188942098519485,.0344259888799142,.0488273289408453,.0545351953956101,.0458812547527418,.0229266537472159,3.14102354559662e-006,.000188290239053809,.00155806188421368,.00598151398652822,.0148131935773177,.0269902177103302,.0382809697326146,.0427559772280557,.0359712268228302,.017974657987755,1.18417845234822e-006,7.09861739774591e-005,.000587395568332318,.00225505472099886,.00558463328595991,.0101754201370336,.0144320788540114,.016119174596298,.0135612965295474,.0067765180261918,-1.18417845234822e-006,-7.09861739774591e-005,-.000587395568332318,-.00225505472099886,-.00558463328595991,-.0101754201370336,-.0144320788540114,-.016119174596298,-.0135612965295474,-.0067765180261918,-3.14102354559662e-006,-.000188290239053809,-.00155806188421368,-.00598151398652822,-.0148131935773177,-.0269902177103302,-.0382809697326146,-.0427559772280557,-.0359712268228302,-.017974657987755,-4.00637159776859e-006,-.000240163964049163,-.00198730598159011,-.00762941359700327,-.0188942098519485,-.0344259888799142,-.0488273289408453,-.0545351953956101,-.0458812547527418,-.0229266537472159,-3.47978143111129e-006,-.000208597251185043,-.001726098112448,-.00662661740611046,-.0164107894122825,-.0299010997678524,-.0424095539898868,-.0473671889011406,-.0398507064131546,-.0199132162457208,-1.74767520475977e-006,-.000104765270720098,-.000866910445908273,-.00332813286160627,-.00824210667082646,-.0150174405185659,-.0212996498258157,-.0237895578215264,-.0200145017350098,-.0100011552359343},
			{-2.1377040911701e-006,-.000128145748833051,-.00106037895476748,-.00407087267407408,-.0100814986114699,-.0183688845318685,-.0260530952485506,-.0290986763122241,-.0244811405031877,-.0122331142571899,-3.23272774874231e-006,-.000193787493716783,-.00160355050327985,-.00615614813548528,-.015245674293661,-.0277782144801489,-.0393986072714656,-.0440042654896674,-.0370214299314904,-.0184994396914433,-1.46467098517923e-006,-8.78004401849179e-005,-.00072653006314475,-.00278920226363585,-.00690744737044604,-.0125856391046194,-.0178505588506293,-.0199372715230704,-.0167735171239198,-.00838165000707885,2.04207081203481e-006,.000122412963730206,.0010129413711518,.00388875630709325,.00963048821447668,.0175471259596245,.0248875724147149,.0277969050120105,.0233859413346208,.0116858482275818,4.79303205216701e-006,.000287320721349026,.00237751816944035,.00912746684058028,.0226041322455455,.0411856124930787,.0584146894325829,.0652433088453238,.0548901466718698,.0274283559519402,4.79303205216701e-006,.000287320721349026,.00237751816944035,.00912746684058028,.0226041322455455,.0411856124930787,.0584146894325829,.0652433088453238,.0548901466718698,.0274283559519403,2.04207081203481e-006,.000122412963730205,.0010129413711518,.00388875630709324,.00963048821447665,.0175471259596245,.0248875724147148,.0277969050120105,.0233859413346208,.0116858482275817,-1.46467098517922e-006,-8.78004401849175e-005,-.000726530063144746,-.00278920226363584,-.00690744737044601,-.0125856391046194,-.0178505588506292,-.0199372715230703,-.0167735171239197,-.00838165000707881,-3.23272774874231e-006,-.000193787493716783,-.00160355050327985,-.00615614813548528,-.015245674293661,-.0277782144801489,-.0393986072714656,-.0440042654896674,-.0370214299314904,-.0184994396914433,-2.1377040911701e-006,-.000128145748833051,-.00106037895476748,-.00407087267407408,-.0100814986114699,-.0183688845318685,-.0260530952485506,-.0290986763122241,-.0244811405031877,-.0122331142571899},
			{-7.0687420406234e-006,-.000396419472948955,-.00289468675708114,-.0091512705781711,-.0170446886407066,-.020343232859846,-.0143340443821725,-.00198780768336474,.00723004945575377,.00622084218798443,-1.58453826290029e-005,-.00088861896421812,-.00648876688291386,-.0205136052524374,-.0382075922069433,-.0456016511456594,-.032131377344278,-.00445589514439379,.0162069713952851,.0139447194673076,-2.32283443257473e-005,-.00130266007192951,-.00951212822907477,-.0300716680261107,-.0560099511902079,-.066849181205329,-.04710259853538,-.00653206483660142,.0237584109428924,.0204420904749686,-2.85486506407837e-005,-.00160102617628203,-.0116908214315209,-.0369593946354349,-.0688386785778866,-.0821605661208479,-.0578911527658522,-.00802819324390579,.0292001256859699,.0251242228526073,-3.13325682074437e-005,-.00175715001389931,-.0128308502042083,-.0405634846946589,-.0755514724317883,-.0901724419250166,-.0635364002126376,-.00881106135495838,.0320475717480869,.0275742078353698,-3.13325682074437e-005,-.00175715001389931,-.0128308502042083,-.0405634846946589,-.0755514724317883,-.0901724419250165,-.0635364002126376,-.00881106135495838,.0320475717480869,.0275742078353698,-2.85486506407837e-005,-.00160102617628203,-.0116908214315209,-.0369593946354349,-.0688386785778866,-.0821605661208479,-.0578911527658522,-.00802819324390579,.0292001256859699,.0251242228526073,-2.32283443257473e-005,-.00130266007192951,-.00951212822907477,-.0300716680261107,-.0560099511902079,-.066849181205329,-.04710259853538,-.00653206483660142,.0237584109428924,.0204420904749686,-1.58453826290029e-005,-.00088861896421812,-.00648876688291386,-.0205136052524374,-.0382075922069433,-.0456016511456594,-.032131377344278,-.00445589514439379,.0162069713952851,.0139447194673076,-7.0687420406234e-006,-.000396419472948955,-.00289468675708114,-.0091512705781711,-.0170446886407066,-.020343232859846,-.0143340443821725,-.00198780768336474,.00723004945575377,.00622084218798443},
			{1.19239470204654e-005,.000668702403646816,.00488291853542222,.0154368719238809,.0287519283012612,.0343160960538323,.0241794628831796,.00335314450111479,-.0121960493352718,-.0104936624148843,2.37416708289763e-005,.00133144774315747,.00972233811117492,.0307362261184839,.0572477231117576,.0683264908211426,.048143525596736,.00667641787160104,-.0242834514641005,-.0208938431558764,2.73344626884814e-005,.00153293375681807,.0111936051282609,.0353874936634631,.0659109362045824,.0786662374118139,.0554290139727572,.00768675029312828,-.027958230184157,-.0240556774743184,2.14304112375827e-005,.00120183086029444,.00877586524613968,.0277440442315677,.0516746381304762,.0616748841000227,.0434567372868267,.00602646636005635,-.0219194493467751,-.0188598205404102,8.07935083711268e-006,.000453095046079154,.00330853633356522,.0104596157536215,.0194815454639894,.0232516782323097,.0163833639440749,.00227200194579396,-.00826372016223111,-.00711022785245187,-8.07935083711268e-006,-.000453095046079154,-.00330853633356522,-.0104596157536215,-.0194815454639894,-.0232516782323097,-.0163833639440749,-.00227200194579396,.00826372016223111,.00711022785245187,-2.14304112375827e-005,-.00120183086029444,-.00877586524613968,-.0277440442315677,-.0516746381304761,-.0616748841000227,-.0434567372868267,-.00602646636005635,.0219194493467751,.0188598205404102,-2.73344626884814e-005,-.00153293375681807,-.0111936051282609,-.0353874936634631,-.0659109362045824,-.0786662374118139,-.0554290139727572,-.00768675029312828,.027958230184157,.0240556774743184,-2.37416708289763e-005,-.00133144774315747,-.00972233811117492,-.0307362261184839,-.0572477231117576,-.0683264908211426,-.048143525596736,-.00667641787160104,.0242834514641005,.0208938431558764,-1.19239470204654e-005,-.000668702403646816,-.00488291853542223,-.0154368719238809,-.0287519283012612,-.0343160960538323,-.0241794628831796,-.00335314450111479,.0121960493352718,.0104936624148843},
			{-2.71193324668268e-005,-.00139397384538797,-.00859091112473706,-.0204390993260263,-.0232648893283274,-.00816857426450726,.0104893149252651,.0116472227517443,-.00129246985153553,-.00610326187849476,-6.07910427499648e-005,-.00312474961288118,-.0192574963297445,-.0458165466432155,-.0521508736788253,-.0183107806184887,.0235129825861177,.0261085635896738,-.00289721696114192,-.0136811499406799,-8.91158835339698e-005,-.00458068837081261,-.0282302905567179,-.0671642046243376,-.076449933653419,-.026842464271009,.0344685684418851,.0382735285799644,-.00424713967062055,-.0200557139600161,-.000109527316707461,-.00562986625988991,-.0346962612267728,-.082547743674565,-.0939603105918888,-.0329905620505779,.0423633774639968,.0470398398136498,-.00521992032573585,-.0246493492246598,-.000120207857264269,-.0061788618595791,-.0380796621567155,-.0905973750422723,-.103122836783299,-.0362076319703515,.0464944361325487,.051626923036526,-.00572894010653844,-.0270530269738013,-.000120207857264269,-.0061788618595791,-.0380796621567155,-.0905973750422723,-.103122836783299,-.0362076319703515,.0464944361325487,.051626923036526,-.00572894010653844,-.0270530269738013,-.000109527316707461,-.00562986625988991,-.0346962612267728,-.082547743674565,-.0939603105918888,-.0329905620505779,.0423633774639968,.0470398398136498,-.00521992032573585,-.0246493492246598,-8.91158835339698e-005,-.00458068837081261,-.0282302905567179,-.0671642046243376,-.076449933653419,-.026842464271009,.0344685684418851,.0382735285799644,-.00424713967062055,-.0200557139600161,-6.07910427499648e-005,-.00312474961288118,-.0192574963297445,-.0458165466432156,-.0521508736788254,-.0183107806184887,.0235129825861177,.0261085635896738,-.00289721696114192,-.0136811499406799,-2.71193324668268e-005,-.00139397384538797,-.00859091112473705,-.0204390993260263,-.0232648893283274,-.00816857426450726,.0104893149252651,.0116472227517443,-.00129246985153553,-.00610326187849476},
			{1.8467923733974e-008,5.72496315380497e-006,.000112551238536802,.000763670899970611,.00284090951583314,.00698705136166929,.0123641411463411,.0161796260245334,.0151169390000152,.00799470122065349,4.1397934207577e-008,1.28331506776816e-005,.000252296296814152,.0017118544633724,.00636821913065334,.0156622637575245,.0277156170386288,.0362684567748484,.0338863239396712,.0179210245780364,6.06867939173738e-008,1.88125998409004e-005,.000369850661970342,.00250947205515881,.00933541273013333,.0229599034620059,.0406293688733935,.0531672993864887,.0496752409729615,.026271106183758,7.45867227452771e-008,2.3121507627533e-005,.000454562632178227,.00308425086139813,.0114736303579195,.0282187250839213,.0499352705236225,.065344935240019,.0610530427888428,.0322883379835085,8.18600363005434e-008,2.53761981227818e-005,.000498889241964133,.00338501114114201,.0125924797742609,.0309704700072926,.0548047012562974,.0717170372141512,.0670066214870632,.0354369306242455,8.18600363005434e-008,2.53761981227818e-005,.000498889241964133,.00338501114114201,.0125924797742609,.0309704700072926,.0548047012562974,.0717170372141512,.0670066214870632,.0354369306242455,7.45867227452771e-008,2.3121507627533e-005,.000454562632178227,.00308425086139813,.0114736303579195,.0282187250839213,.0499352705236225,.065344935240019,.0610530427888428,.0322883379835085,6.06867939173738e-008,1.88125998409004e-005,.000369850661970342,.00250947205515881,.00933541273013333,.0229599034620059,.0406293688733935,.0531672993864886,.0496752409729615,.026271106183758,4.1397934207577e-008,1.28331506776816e-005,.000252296296814152,.0017118544633724,.00636821913065334,.0156622637575245,.0277156170386288,.0362684567748484,.0338863239396712,.0179210245780364,1.8467923733974e-008,5.72496315380497e-006,.000112551238536802,.000763670899970611,.00284090951583315,.00698705136166929,.0123641411463411,.0161796260245334,.0151169390000152,.00799470122065349},
			{-3.11527203680034e-008,-9.65718609285329e-006,-.000189857685775429,-.0012882025257772,-.00479220409464544,-.0117861466401087,-.0208565205959732,-.0272926925875222,-.0255000930408489,-.013485906436498,-6.20279200448354e-008,-1.9228342172032e-005,-.000378024044579789,-.00256492923656619,-.00954171734955457,-.0234672976482734,-.041527243099746,-.0543422511302053,-.0507730212189817,-.0268516751119555,-7.14145132549108e-008,-2.21381386949937e-005,-.000435229862984357,-.00295307617650668,-.0109856512944789,-.0270185690209856,-.0478114992513544,-.0625657834542022,-.0584564272703957,-.0309150993103302,-5.59894812942294e-008,-1.73564566340746e-005,-.000341223277476823,-.00231523251800061,-.00861282797603967,-.0211827483777367,-.0374845521025614,-.0490520148175479,-.0458302506312781,-.0242376555640243,-2.11082586119828e-008,-6.54345364077286e-006,-.000128642541758133,-.000872851928740155,-.00324707062998819,-.0079859809473875,-.0141318262188787,-.0184928059748667,-.0172781879777248,-.00913769318749374,2.11082586119828e-008,6.54345364077286e-006,.000128642541758133,.000872851928740155,.00324707062998819,.00798598094738749,.0141318262188787,.0184928059748667,.0172781879777248,.00913769318749374,5.59894812942294e-008,1.73564566340746e-005,.000341223277476823,.00231523251800061,.00861282797603967,.0211827483777367,.0374845521025615,.0490520148175479,.0458302506312781,.0242376555640243,7.14145132549108e-008,2.21381386949937e-005,.000435229862984357,.00295307617650668,.0109856512944789,.0270185690209856,.0478114992513544,.0625657834542022,.0584564272703957,.0309150993103302,6.20279200448354e-008,1.9228342172032e-005,.000378024044579789,.00256492923656619,.00954171734955457,.0234672976482734,.041527243099746,.0543422511302053,.0507730212189817,.0268516751119555,3.11527203680034e-008,9.65718609285329e-006,.000189857685775429,.0012882025257772,.00479220409464544,.0117861466401087,.0208565205959732,.0272926925875222,.0255000930408489,.013485906436498},
			{1.62757657902233e-007,4.73190249691825e-005,.000825150543237864,.00465350306811798,.013244685304365,.0221077948264185,.0214226110182188,.00800738602387552,-.0066385606490133,-.00788880135076707,3.64839649040836e-007,.000106070931993143,.00184966801858007,.0104313520363045,.0296894560930798,.0495571158340357,.0480211990401099,.0179494589952879,-.0148810825158051,-.0176836380743178,5.34832208853932e-007,.000155493381824678,.00271149814617032,.0152917125799703,.0435228940266753,.0726476461525729,.0703960877661124,.0263127892689876,-.0218147404016795,-.0259231123501117,6.57332165731892e-007,.000191108163905862,.00333254975914684,.0187941832625992,.0534915394321128,.0892871330302339,.086519869342777,.0323395683174988,-.0268112696201347,-.0318606383488469,7.21431817457549e-007,.000209744049059384,.00365752287024036,.0206268953439552,.0587077592165956,.0979939549829493,.094956841974485,.0354931566768965,-.0294257667261143,-.0349675239211424,7.21431817457549e-007,.000209744049059384,.00365752287024036,.0206268953439552,.0587077592165956,.0979939549829493,.094956841974485,.0354931566768965,-.0294257667261143,-.0349675239211424,6.57332165731892e-007,.000191108163905862,.00333254975914684,.0187941832625992,.0534915394321128,.0892871330302339,.086519869342777,.0323395683174988,-.0268112696201347,-.0318606383488469,5.34832208853932e-007,.000155493381824678,.00271149814617032,.0152917125799703,.0435228940266753,.0726476461525729,.0703960877661124,.0263127892689876,-.0218147404016795,-.0259231123501117,3.64839649040836e-007,.000106070931993143,.00184966801858007,.0104313520363045,.0296894560930798,.0495571158340357,.0480211990401099,.0179494589952879,-.0148810825158051,-.0176836380743178,1.62757657902233e-007,4.73190249691825e-005,.000825150543237864,.00465350306811798,.013244685304365,.0221077948264185,.0214226110182188,.00800738602387552,-.0066385606490133,-.00788880135076707},
			{-3.0545495487195e-010,-4.89665844207368e-007,-2.28716807432812e-005,-.000274273335410219,-.00153266885205138,-.00508819457391411,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594,-6.84711736268708e-010,-1.09764122346374e-006,-5.12694522820253e-005,-.000614814618998954,-.0034356501149343,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-1.00374477198287e-009,-1.60907368927037e-006,-7.51578247671699e-005,-.000901279950190104,-.00503644914868864,-.0167201370314615,-.0369150385504742,-.0565976592019674,-.058726053964138,-.0328701880528208,-1.23364620508441e-009,-1.97762190735818e-006,-9.237225228407e-005,-.00110771245968658,-.00619001618021515,-.0205497793593547,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.35394503757885e-009,-2.17046942359914e-006,-.000101379919197693,-.001215730881087,-.00679363471892789,-.0225536881419213,-.0497944643424746,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.35394503757885e-009,-2.17046942359914e-006,-.000101379919197693,-.001215730881087,-.00679363471892789,-.0225536881419213,-.0497944643424746,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.23364620508441e-009,-1.97762190735818e-006,-9.237225228407e-005,-.00110771245968658,-.00619001618021515,-.0205497793593547,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.00374477198287e-009,-1.60907368927037e-006,-7.51578247671699e-005,-.000901279950190104,-.00503644914868864,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-6.84711736268708e-010,-1.09764122346374e-006,-5.12694522820253e-005,-.000614814618998954,-.0034356501149343,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-3.0545495487195e-010,-4.89665844207368e-007,-2.28716807432812e-005,-.000274273335410219,-.00153266885205138,-.00508819457391411,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 100; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_left(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,-8.500000000,0.,33.98823325,-.4082482906,0.,0.,0.,1.750000000,0.,0.,-7.678758582,0.,34.53410150,.3227486122,0.,0.,-2.459674775,0.,14.75000000,-.2645751311,0.,2.958039892,.2236067977},
			{0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,-8.500000000,0.,0.,-.4082482906,0.,0.,0.,1.750000000,0.,0.,-7.678758582,0.,0.,.3227486122,0.,0.,-2.459674775,0.,0.,-.2645751311,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,1.750000000,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,-12.37436867,0.,55.33985905,-.7216878367,0.,0.,0.,2.828427124,0.,0.,-13.26806944,0.,63.17066828,.9128709293,0.,0.,-5.850213671,0.,33.43604921,-.9354143468,0.,8.844691708,.9035079028},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,-12.37436867,0.,0.,-.7216878367,0.,0.,0.,2.828427124,0.,0.,-13.26806944,0.,0.,.9128709295,0.,0.,-5.850213671,0.,0.,-.9354143468,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,2.828427124,0.,0.,0.,0.,0.,.9128709292,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,-7.216878367,0.,40.02082790,0.,0.,0.,0.,1.587713241,0.,0.,-9.500000000,0.,52.88300766,.9316949905,0.,0.,-6.067673910,0.,36.47616523,-1.636634177,0.,13.68699853,2.074812507},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,-7.216878367,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,-9.500000000,0.,0.,.9316949905,0.,0.,-6.067673910,0.,0.,-1.636634177,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,15.87608264,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,23.88339730,0.,0.,0.,-2.627379873,0.,21.21428571,-1.218935425,0.,11.40958244,2.747169230},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,0.,0.,0.,-2.627379873,0.,0.,-1.218935425,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,0.,0.,0.,0.,5.494338457,0.,0.,4.138710979,1.621428571},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,2.245365598,0.,-7.850689990,-.2500000000,0.,0.,0.,0.,0.,0.,.4949747467,0.,-3.162277660,.3162277660,0.,0.,-1.004158022,0.,3.091022771,-.3240370350,0.,2.070196679,.3129843186},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,2.245365598,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,.3162277660,0.,0.,-1.004158022,0.,0.,-.3240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277659,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,3.500000000,0.,-14.98165545,0.,0.,0.,0.,-.2500000000,0.,0.,1.558845727,0.,-8.714212528,.5809475019,0.,0.,-1.565247584,0.,4.964285714,-1.020504077,0.,5.704791219,1.293725044},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,1.558845727,0.,0.,.5809475019,0.,0.,-1.565247584,0.,0.,-1.020504077,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,-10.71525393,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,-8.720665110,0.,0.,0.,-.5486726408,0.,1.773290113,-.9983468483,0.,5.709015429,2.250018896},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,0.,0.,0.,-.5486726408,0.,0.,-.9983468480,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,0.,0.,0.,0.,-.4979550017,0.,0.,1.996693696,1.554721797},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,-.6454972245,0.,2.020725943,0.,0.,0.,0.,-.3227486122,0.,0.,.2236067977,0.,-.2500000000,.8333333333e-1,0.,0.,.2886751347,0.,-1.152673615,-.1463850110,0.,.1636634177,.1855768722},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,-.6454972245,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,.2236067977,0.,0.,.8333333333e-1,0.,0.,.2886751347,0.,0.,-.1463850109,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,.2886751347,0.,0.,0.,.6250000000,0.,-2.395787117,-.2746751327,0.,-.1889822365,.6190476190},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,.6250000000,0.,0.,-.2746751327,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,0.,0.,0.,0.,-1.357142857,0.,0.,-.6338656908,.6016087651},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,-.5916079782,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,.3415650256,0.,0.,0.,.1479019946,0.,.1889822365,-.2500000000e-1,0.,-.2236067977,.5634361697e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931771,0.,0.,0.,0.,0.,.1479019946,0.,0.,-.2500000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,.5070925528,0.,0.,-.4500000000,.1070899340},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,.1916629694,0.,0.,-.5669467095e-1,.7142857143e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_left(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,-8.500000000,0.,33.98823325,-.4082482906,0.,0.,0.,1.750000000,0.,0.,-7.678758582,0.,34.53410150,.3227486122,0.,0.,-2.459674775,0.,14.75000000,-.2645751311,0.,2.958039892,.2236067977},
			{0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,-8.500000000,0.,0.,-.4082482906,0.,0.,0.,1.750000000,0.,0.,-7.678758582,0.,0.,.3227486122,0.,0.,-2.459674775,0.,0.,-.2645751311,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,1.750000000,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,-12.37436867,0.,55.33985905,-.7216878367,0.,0.,0.,2.828427124,0.,0.,-13.26806944,0.,63.17066828,.9128709293,0.,0.,-5.850213671,0.,33.43604921,-.9354143468,0.,8.844691708,.9035079028},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,-12.37436867,0.,0.,-.7216878367,0.,0.,0.,2.828427124,0.,0.,-13.26806944,0.,0.,.9128709295,0.,0.,-5.850213671,0.,0.,-.9354143468,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,2.828427124,0.,0.,0.,0.,0.,.9128709292,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,-7.216878367,0.,40.02082790,0.,0.,0.,0.,1.587713241,0.,0.,-9.500000000,0.,52.88300766,.9316949905,0.,0.,-6.067673910,0.,36.47616523,-1.636634177,0.,13.68699853,2.074812507},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,-7.216878367,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,-9.500000000,0.,0.,.9316949905,0.,0.,-6.067673910,0.,0.,-1.636634177,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,15.87608264,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,23.88339730,0.,0.,0.,-2.627379873,0.,21.21428571,-1.218935425,0.,11.40958244,2.747169230},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,0.,0.,0.,-2.627379873,0.,0.,-1.218935425,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,0.,0.,0.,0.,5.494338457,0.,0.,4.138710979,1.621428571},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,2.245365598,0.,-7.850689990,-.2500000000,0.,0.,0.,0.,0.,0.,.4949747467,0.,-3.162277660,.3162277660,0.,0.,-1.004158022,0.,3.091022771,-.3240370350,0.,2.070196679,.3129843186},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,2.245365598,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,.3162277660,0.,0.,-1.004158022,0.,0.,-.3240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277659,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,3.500000000,0.,-14.98165545,0.,0.,0.,0.,-.2500000000,0.,0.,1.558845727,0.,-8.714212528,.5809475019,0.,0.,-1.565247584,0.,4.964285714,-1.020504077,0.,5.704791219,1.293725044},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,1.558845727,0.,0.,.5809475019,0.,0.,-1.565247584,0.,0.,-1.020504077,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,-10.71525393,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,-8.720665110,0.,0.,0.,-.5486726408,0.,1.773290113,-.9983468483,0.,5.709015429,2.250018896},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,0.,0.,0.,-.5486726408,0.,0.,-.9983468480,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,0.,0.,0.,0.,-.4979550017,0.,0.,1.996693696,1.554721797},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,-.6454972245,0.,2.020725943,0.,0.,0.,0.,-.3227486122,0.,0.,.2236067977,0.,-.2500000000,.8333333333e-1,0.,0.,.2886751347,0.,-1.152673615,-.1463850110,0.,.1636634177,.1855768722},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,-.6454972245,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,.2236067977,0.,0.,.8333333333e-1,0.,0.,.2886751347,0.,0.,-.1463850109,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,.2886751347,0.,0.,0.,.6250000000,0.,-2.395787117,-.2746751327,0.,-.1889822365,.6190476190},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,.6250000000,0.,0.,-.2746751327,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,0.,0.,0.,0.,-1.357142857,0.,0.,-.6338656908,.6016087651},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,-.5916079782,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,.3415650256,0.,0.,0.,.1479019946,0.,.1889822365,-.2500000000e-1,0.,-.2236067977,.5634361697e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931771,0.,0.,0.,0.,0.,.1479019946,0.,0.,-.2500000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,.5070925528,0.,0.,-.4500000000,.1070899340},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,.1916629694,0.,0.,-.5669467095e-1,.7142857143e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_coarsen_right(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,-.1154700539,0.,-.6454972245e-1,2.904737510,0.,0.,1.341640786,0.,.7500000000,-11.64130577,0.,-6.507687761,50.31152948},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,-.1154700539,0.,0.,2.904737510,0.,0.,1.341640786,0.,0.,-11.64130577,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,.1632993162,0.,.9128709293e-1,-2.464751509,0.,0.,-1.581138830,0.,-.9596449171,10.85080642,0.,7.410417377,-50.14468862},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,.1632993162,0.,0.,-2.464751509,0.,0.,-1.581138830,0.,0.,10.85080642,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,0.,0.,0.,-2.464751508,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.2000000000,0.,-.1118033988,1.155301788,0.,0.,1.290994449,0.,.9691236663,-6.197388082,0.,-6.026182953,33.05867928},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.2000000000,0.,0.,1.155301788,0.,0.,1.290994449,0.,0.,-6.197388081,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,.1290994449,0.,0.,0.,-.7267220925,0.,-.7857142857,1.974864371,0.,3.507390157,-14.50782199},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,0.,0.,0.,-.7267220925,0.,0.,1.974864371,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,0.,0.,0.,0.,.5111012519,0.,0.,-1.266180985,3.502380952},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,.2828427124,0.,.1581138830,3.320391543,0.,0.,-1.278019301,0.,-1.195584279,-15.76980237,0.,1.794170455,75.89869728},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,.2828427124,0.,0.,3.320391543,0.,0.,-1.278019301,0.,0.,-15.76980236,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,3.320391542,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-.3464101616,0.,-.1936491673,-1.484643616,0.,0.,.4472135954,0.,1.107142857,8.919961562,0.,-.8451542550e-1,-49.83237207},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-.3464101616,0.,0.,-1.484643616,0.,0.,.4472135954,0.,0.,8.919961562,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,.2236067977,0.,0.,0.,.3227486122e-1,0.,-.7010681843,-2.776822654,0.,-.7563225568,21.71637090},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,-2.776822653,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,0.,0.,0.,0.,.2766416676,0.,0.,.5237229368,-5.171408842},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,-.4472135954,0.,-.2500000000,1.083333333,0.,0.,-4.041451885,0.,-.4610694458e-1,-8.002380602,0.,16.80277755,48.86857636},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,-.4472135954,0.,0.,1.083333333,0.,0.,-4.041451885,0.,0.,-8.002380594,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,.2886751347,0.,0.,0.,1.375000000,0.,.7985957060,2.387560769,0.,-7.275816105,-21.01190476},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,1.375000000,0.,0.,2.387560769,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,0.,0.,0.,0.,-.6428571429,0.,0.,1.774823935,4.876757779},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,.3415650256,0.,0.,0.,-1.922725929,0.,3.968626966,-1.575000000,0.,13.75181806,16.26921940},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079531,0.,0.,0.,0.,0.,-1.922725929,0.,0.,-1.575000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,-.8451542550,0.,0.,-3.450000000,-3.647357166},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,0.,0.,0.,0.,2.491618602,0.,0.,3.571764270,2.292857143}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_coarsen_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_coarsen_right(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,-.1154700539,0.,-.6454972245e-1,2.904737510,0.,0.,1.341640786,0.,.7500000000,-11.64130577,0.,-6.507687761,50.31152948},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,-.1154700539,0.,0.,2.904737510,0.,0.,1.341640786,0.,0.,-11.64130577,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,.1632993162,0.,.9128709293e-1,-2.464751509,0.,0.,-1.581138830,0.,-.9596449171,10.85080642,0.,7.410417377,-50.14468862},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,.1632993162,0.,0.,-2.464751509,0.,0.,-1.581138830,0.,0.,10.85080642,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,0.,0.,0.,-2.464751508,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.2000000000,0.,-.1118033988,1.155301788,0.,0.,1.290994449,0.,.9691236663,-6.197388082,0.,-6.026182953,33.05867928},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.2000000000,0.,0.,1.155301788,0.,0.,1.290994449,0.,0.,-6.197388081,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,.1290994449,0.,0.,0.,-.7267220925,0.,-.7857142857,1.974864371,0.,3.507390157,-14.50782199},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,0.,0.,0.,-.7267220925,0.,0.,1.974864371,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,0.,0.,0.,0.,.5111012519,0.,0.,-1.266180985,3.502380952},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,.2828427124,0.,.1581138830,3.320391543,0.,0.,-1.278019301,0.,-1.195584279,-15.76980237,0.,1.794170455,75.89869728},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,.2828427124,0.,0.,3.320391543,0.,0.,-1.278019301,0.,0.,-15.76980236,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,3.320391542,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-.3464101616,0.,-.1936491673,-1.484643616,0.,0.,.4472135954,0.,1.107142857,8.919961562,0.,-.8451542550e-1,-49.83237207},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-.3464101616,0.,0.,-1.484643616,0.,0.,.4472135954,0.,0.,8.919961562,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,.2236067977,0.,0.,0.,.3227486122e-1,0.,-.7010681843,-2.776822654,0.,-.7563225568,21.71637090},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,-2.776822653,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,0.,0.,0.,0.,.2766416676,0.,0.,.5237229368,-5.171408842},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,-.4472135954,0.,-.2500000000,1.083333333,0.,0.,-4.041451885,0.,-.4610694458e-1,-8.002380602,0.,16.80277755,48.86857636},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,-.4472135954,0.,0.,1.083333333,0.,0.,-4.041451885,0.,0.,-8.002380594,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,.2886751347,0.,0.,0.,1.375000000,0.,.7985957060,2.387560769,0.,-7.275816105,-21.01190476},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,1.375000000,0.,0.,2.387560769,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,0.,0.,0.,0.,-.6428571429,0.,0.,1.774823935,4.876757779},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,.3415650256,0.,0.,0.,-1.922725929,0.,3.968626966,-1.575000000,0.,13.75181806,16.26921940},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079531,0.,0.,0.,0.,0.,-1.922725929,0.,0.,-1.575000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,-.8451542550,0.,0.,-3.450000000,-3.647357166},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,0.,0.,0.,0.,2.491618602,0.,0.,3.571764270,2.292857143}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_left(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,.1250000000,0.,-.5590169942e-1,.4082482906,0.,0.,0.,.1250000000,0.,0.,-.2453738645,0.,.3227486122e-1,-.1613743061,0.,0.,.5590169942e-1,0.,.1250000000,.6614378278e-1,0.,-.7395099729e-1,-.2795084971e-1},
			{0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,.1250000000,0.,0.,.4082482906,0.,0.,0.,.1250000000,0.,0.,-.2453738645,0.,0.,-.1613743061,0.,0.,.5590169942e-1,0.,0.,.6614378278e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,-.2651650429,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,.3163924252,0.,.9128709293e-1,.3651483717,0.,0.,.3952847075e-1,0.,-.2525381361,-.1403121520,0.,.8964214566e-1,.5646924394e-1},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,-.2651650429,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,.3163924252,0.,0.,.3651483718,0.,0.,.3952847075e-1,0.,0.,-.1403121520,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,.3651483716,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,0.,.2420614591,0.,0.,0.,0.,-.2165063510,0.,0.,.4250000000,0.,-.3913118960,.5776508940,0.,0.,-.6132223630,0.,.3608439183,.3218713882,0.,-.3049687729e-2,-.1210307296},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,.4250000000,0.,0.,.5776508940,0.,0.,-.6132223630,0.,0.,.3218713881,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,.1118033988,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,-.2259240285,0.,0.,0.,-.1816805231,0.,.3750000000,.4937160930,0.,-.5387858374,.2888254471},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,0.,0.,0.,-.1816805231,0.,0.,.4937160930,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,0.,0.,0.,0.,.6388765648e-1,0.,0.,-.1582726231,.4377976190},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,-.4592793268,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,.1237436867,0.,.1581138830,-.3162277660,0.,0.,-.1141088661,0.,.2916059217e-1,.1890216038,0.,-.1725163898e-1,-.9780759959e-1},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,-.4592793268,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,-.3162277660,0.,0.,-.1141088661,0.,0.,.1890216037,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3162277659,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,-.3750000000,0.,.4192627457,0.,0.,0.,0.,-.1250000000,0.,0.,.3897114318,0.,-.4357106264,-.7423218081,0.,0.,.2795084971,0.,.8928571429e-2,-.3496171374,0.,-.8451542550e-1,.1796840339},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,-.3750000000,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.3897114318,0.,0.,-.7423218081,0.,0.,.2795084971,0.,0.,-.3496171374,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,-.2795084971,0.,0.,0.,.8068715303e-2,0.,.1340277411,-.6942056633,0.,.3964594048,-.3527181261},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,-.6942056631,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,0.,0.,0.,0.,.3458020845e-1,0.,0.,.6546536711e-1,-.6464261051},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,-.4841229182,0.,.5412658775,0.,0.,0.,0.,.4841229182,0.,0.,-.3913118960,0.,.6250000000e-1,.5416666667,0.,0.,.3608439183,0.,-.1959545145,.1341862601,0.,.9547032698e-1,-.1546473936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.3913118960,0.,0.,.5416666667,0.,0.,.3608439183,0.,0.,.1341862600,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,-.7216878367e-1,0.,0.,0.,.3437500000,0.,-.5190872089,.5968901920,0.,.1653594569,.1994047619},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,0.,0.,0.,.3437500000,0.,0.,.5968901920,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,0.,0.,0.,0.,-.8035714286e-1,0.,0.,.2218529919,.6095947223},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,.4269562821,0.,0.,0.,-.4806814823,0.,.4724555913e-1,-.3937500000,0.,-.4192627457,-.1056442818e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698828,0.,0.,0.,0.,0.,-.4806814823,0.,0.,-.3937500000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,0.,0.,0.,0.,-.1056442819,0.,0.,-.4312500000,-.4559196456},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,0.,0.,0.,0.,.3114523254,0.,0.,.4464705337,.2866071429}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_left
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_left(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,.1250000000,0.,-.5590169942e-1,.4082482906,0.,0.,0.,.1250000000,0.,0.,-.2453738645,0.,.3227486122e-1,-.1613743061,0.,0.,.5590169942e-1,0.,.1250000000,.6614378278e-1,0.,-.7395099729e-1,-.2795084971e-1},
			{0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,.1250000000,0.,0.,.4082482906,0.,0.,0.,.1250000000,0.,0.,-.2453738645,0.,0.,-.1613743061,0.,0.,.5590169942e-1,0.,0.,.6614378278e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,-.2651650429,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,.3163924252,0.,.9128709293e-1,.3651483717,0.,0.,.3952847075e-1,0.,-.2525381361,-.1403121520,0.,.8964214566e-1,.5646924394e-1},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,-.2651650429,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,.3163924252,0.,0.,.3651483718,0.,0.,.3952847075e-1,0.,0.,-.1403121520,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,.3651483716,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,0.,.2420614591,0.,0.,0.,0.,-.2165063510,0.,0.,.4250000000,0.,-.3913118960,.5776508940,0.,0.,-.6132223630,0.,.3608439183,.3218713882,0.,-.3049687729e-2,-.1210307296},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,.4250000000,0.,0.,.5776508940,0.,0.,-.6132223630,0.,0.,.3218713881,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,.1118033988,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,-.2259240285,0.,0.,0.,-.1816805231,0.,.3750000000,.4937160930,0.,-.5387858374,.2888254471},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,0.,0.,0.,-.1816805231,0.,0.,.4937160930,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,0.,0.,0.,0.,.6388765648e-1,0.,0.,-.1582726231,.4377976190},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,-.4592793268,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,.1237436867,0.,.1581138830,-.3162277660,0.,0.,-.1141088661,0.,.2916059217e-1,.1890216038,0.,-.1725163898e-1,-.9780759959e-1},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,-.4592793268,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,-.3162277660,0.,0.,-.1141088661,0.,0.,.1890216037,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3162277659,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,-.3750000000,0.,.4192627457,0.,0.,0.,0.,-.1250000000,0.,0.,.3897114318,0.,-.4357106264,-.7423218081,0.,0.,.2795084971,0.,.8928571429e-2,-.3496171374,0.,-.8451542550e-1,.1796840339},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,-.3750000000,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.3897114318,0.,0.,-.7423218081,0.,0.,.2795084971,0.,0.,-.3496171374,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,-.2795084971,0.,0.,0.,.8068715303e-2,0.,.1340277411,-.6942056633,0.,.3964594048,-.3527181261},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,-.6942056631,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,0.,0.,0.,0.,.3458020845e-1,0.,0.,.6546536711e-1,-.6464261051},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,-.4841229182,0.,.5412658775,0.,0.,0.,0.,.4841229182,0.,0.,-.3913118960,0.,.6250000000e-1,.5416666667,0.,0.,.3608439183,0.,-.1959545145,.1341862601,0.,.9547032698e-1,-.1546473936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.3913118960,0.,0.,.5416666667,0.,0.,.3608439183,0.,0.,.1341862600,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,-.7216878367e-1,0.,0.,0.,.3437500000,0.,-.5190872089,.5968901920,0.,.1653594569,.1994047619},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,0.,0.,0.,.3437500000,0.,0.,.5968901920,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,0.,0.,0.,0.,-.8035714286e-1,0.,0.,.2218529919,.6095947223},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,.4269562821,0.,0.,0.,-.4806814823,0.,.4724555913e-1,-.3937500000,0.,-.4192627457,-.1056442818e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698828,0.,0.,0.,0.,0.,-.4806814823,0.,0.,-.3937500000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,0.,0.,0.,0.,-.1056442819,0.,0.,-.4312500000,-.4559196456},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,0.,0.,0.,0.,.3114523254,0.,0.,.4464705337,.2866071429}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void mul_adaptivity_refine_right(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,-.1250000000,0.,.5590169942e-1,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,.2453738645,0.,-.3227486122e-1,.1613743061,0.,0.,-.5590169942e-1,0.,-.1250000000,-.6614378278e-1,0.,.7395099729e-1,.2795084971e-1},
			{0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,-.1250000000,0.,0.,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,.2453738645,0.,0.,.1613743061,0.,0.,-.5590169942e-1,0.,0.,-.6614378278e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,.2651650429,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,.5103103630e-1,0.,-.9128709293e-1,.4564354646,0.,0.,.1185854122,0.,-.1515228816,-.2338535867,0.,.5976143047e-1,.1129384878},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,.2651650429,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,.5103103630e-1,0.,0.,.4564354646,0.,0.,.1185854122,0.,0.,-.2338535867,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,0.,0.,0.,.4564354645,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,.2165063510,0.,-.2420614591,0.,0.,0.,0.,.7938566203,0.,0.,.3750000000,0.,-.1677050983,.4658474951,0.,0.,.9682458365e-1,0.,.3092947872e-1,-.4091585443,0.,-.1189378215,.2593515633},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,.2165063510,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,.3750000000,0.,0.,.4658474951,0.,0.,.9682458365e-1,0.,0.,-.4091585442,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,-.1118033988,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,-.2581988897,0.,0.,0.,-.6568449682,0.,-.2857142857,-.3047338563,0.,.6338656908e-1,.3433961537},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,0.,0.,0.,-.6568449682,0.,0.,-.3047338563,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,0.,0.,0.,0.,.6867923071,0.,0.,.5173388725,.2026785714},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,.2551551816,0.,-.1825741858,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,-.4065863991,0.,.1581138830,.1581138830,0.,0.,.2966830521,0.,.2916059217e-1,-.8100925875e-1,0.,-.1035098339,.3912303983e-1},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,.2551551816,0.,0.,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,-.4065863991,0.,0.,.1581138830,0.,0.,.2966830520,0.,0.,-.8100925872e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.6250000000,0.,-.5590169942e-1,0.,0.,0.,0.,-.1250000000,0.,0.,-.4330127020e-1,0.,.4841229182e-1,.2904737510,0.,0.,.6149186937,0.,.2767857143,-.2551260193,0.,-.4278593414,.1617156305},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.6250000000,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.4330127020e-1,0.,0.,.2904737510,0.,0.,.6149186937,0.,0.,-.2551260193,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,.5163977793,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,.3354101966,0.,0.,0.,-.1371681602,0.,-.3917733971,-.2495867120,0.,-.4025587803,.2812523620},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,0.,0.,0.,-.1371681602,0.,0.,-.2495867119,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,0.,0.,0.,0.,-.6224437520e-1,0.,0.,.2495867120,.1943402247},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,0.,.4330127020,0.,0.,0.,0.,-.1613743061,0.,0.,-.5031152948,0.,-.5625000000,.4166666667e-1,0.,0.,.2165063510,0.,.3803822929,-.3659625275e-1,0.,-.1432054905,.2319710903e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,-.5031152948,0.,0.,.4166666667e-1,0.,0.,.2165063510,0.,0.,-.3659625271e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,.1443375673,0.,0.,0.,.1562500000,0.,.3992978531,-.6866878320e-1,0.,-.3307189139,.7738095238e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,0.,0.,0.,.1562500000,0.,0.,-.6866878320e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,0.,0.,0.,0.,-.1696428571,0.,0.,-.7923321136e-1,.7520109565e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,-.3415650256,0.,0.,0.,.3697549864e-1,0.,.1889822365,-.6250000000e-2,0.,-.5590169942e-1,.7042952119e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329428e-1,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,-.6250000000e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,0.,0.,0.,0.,.6338656908e-1,0.,0.,-.5625000000e-1,.1338624175e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,.2395787117e-1,0.,0.,-.7086833870e-2,.8928571429e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * adaptivity_refine_right
	 *********************************************************/
public:
	static inline void madd_adaptivity_refine_right(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,-.1250000000,0.,.5590169942e-1,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,.2453738645,0.,-.3227486122e-1,.1613743061,0.,0.,-.5590169942e-1,0.,-.1250000000,-.6614378278e-1,0.,.7395099729e-1,.2795084971e-1},
			{0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,-.1250000000,0.,0.,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,.2453738645,0.,0.,.1613743061,0.,0.,-.5590169942e-1,0.,0.,-.6614378278e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,.2651650429,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,.5103103630e-1,0.,-.9128709293e-1,.4564354646,0.,0.,.1185854122,0.,-.1515228816,-.2338535867,0.,.5976143047e-1,.1129384878},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,.2651650429,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,.5103103630e-1,0.,0.,.4564354646,0.,0.,.1185854122,0.,0.,-.2338535867,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,0.,0.,0.,.4564354645,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,.2165063510,0.,-.2420614591,0.,0.,0.,0.,.7938566203,0.,0.,.3750000000,0.,-.1677050983,.4658474951,0.,0.,.9682458365e-1,0.,.3092947872e-1,-.4091585443,0.,-.1189378215,.2593515633},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,.2165063510,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,.3750000000,0.,0.,.4658474951,0.,0.,.9682458365e-1,0.,0.,-.4091585442,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,-.1118033988,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,-.2581988897,0.,0.,0.,-.6568449682,0.,-.2857142857,-.3047338563,0.,.6338656908e-1,.3433961537},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,0.,0.,0.,-.6568449682,0.,0.,-.3047338563,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,0.,0.,0.,0.,.6867923071,0.,0.,.5173388725,.2026785714},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,.2551551816,0.,-.1825741858,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,-.4065863991,0.,.1581138830,.1581138830,0.,0.,.2966830521,0.,.2916059217e-1,-.8100925875e-1,0.,-.1035098339,.3912303983e-1},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,.2551551816,0.,0.,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,-.4065863991,0.,0.,.1581138830,0.,0.,.2966830520,0.,0.,-.8100925872e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.6250000000,0.,-.5590169942e-1,0.,0.,0.,0.,-.1250000000,0.,0.,-.4330127020e-1,0.,.4841229182e-1,.2904737510,0.,0.,.6149186937,0.,.2767857143,-.2551260193,0.,-.4278593414,.1617156305},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.6250000000,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.4330127020e-1,0.,0.,.2904737510,0.,0.,.6149186937,0.,0.,-.2551260193,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,.5163977793,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,.3354101966,0.,0.,0.,-.1371681602,0.,-.3917733971,-.2495867120,0.,-.4025587803,.2812523620},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,0.,0.,0.,-.1371681602,0.,0.,-.2495867119,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,0.,0.,0.,0.,-.6224437520e-1,0.,0.,.2495867120,.1943402247},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,0.,.4330127020,0.,0.,0.,0.,-.1613743061,0.,0.,-.5031152948,0.,-.5625000000,.4166666667e-1,0.,0.,.2165063510,0.,.3803822929,-.3659625275e-1,0.,-.1432054905,.2319710903e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,-.5031152948,0.,0.,.4166666667e-1,0.,0.,.2165063510,0.,0.,-.3659625271e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,.1443375673,0.,0.,0.,.1562500000,0.,.3992978531,-.6866878320e-1,0.,-.3307189139,.7738095238e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,0.,0.,0.,.1562500000,0.,0.,-.6866878320e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,0.,0.,0.,0.,-.1696428571,0.,0.,-.7923321136e-1,.7520109565e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,-.3415650256,0.,0.,0.,.3697549864e-1,0.,.1889822365,-.6250000000e-2,0.,-.5590169942e-1,.7042952119e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329428e-1,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,-.6250000000e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,0.,0.,0.,0.,.6338656908e-1,0.,0.,-.5625000000e-1,.1338624175e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,.2395787117e-1,0.,0.,-.7086833870e-2,.8928571429e-3}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void mul_convert_poly_to_dofs(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.1178511302,.2357022604e-1,.7856742012e-2,.7071067810e-1,.1178511302e-1,.4714045206e-1,.3535533905,.1178511302,.5892556509e-1,.3535533905e-1,.1178511302,.2946278254e-1,.1178511302e-1,.5892556509e-1,.1178511302e-1,.3535533905e-1,.2357022604,.7856742012e-1,.3928371006e-1,.7856742012e-1,.1964185503e-1,.3928371006e-1,.1767766952,.5892556509e-1,.5892556509e-1,.1414213562},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-1,.6804138176e-2,.2041241452e-1,.2041241452,.6804138176e-1,.3402069088e-1,.6804138176e-1,.1701034544e-1,.3402069088e-1,.1837117307,.6123724358e-1,.6123724358e-1,.1632993162},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.8784104612e-2,.1756820922e-1,.4392052306e-2,.8784104612e-2,.7905694150e-1,.2635231383e-1,.2635231383e-1,.9035079028e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1,.4454354030e-2,.4454354030e-2,.2672612419e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3367175148e-2},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.1000000000,.1190476190e-1,.7619047619e-1,0.,-.4166666667e-1,-.3333333333e-1,-.2500000000e-1,.8333333333e-1,.8333333333e-2,0.,.6666666667e-1,.8333333333e-2,.5000000000e-1,0.,-.2777777778e-1,-.2222222222e-1,.5555555556e-1,.5555555556e-2,.4444444444e-1,0.,-.2083333333e-1,.4166666667e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,.4811252245e-1,.4811252245e-2,0.,.3849001795e-1,.4811252245e-2,.2886751347e-1,0.,-.2405626122e-1,-.1924500898e-1,.4811252245e-1,.4811252245e-2,.3849001795e-1,0.,-.2165063510e-1,.4330127020e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,-.4969039948e-2,.1242259987e-1,.1242259987e-2,.9938079897e-2,0.,-.9316949905e-2,.1863389981e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1574851971e-2,.3149703940e-2,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,.5248906592e-1,0.,.5248906592e-1,0.,0.,.6804138176e-2,.8748177653e-2,0.,-.1020620726e-1,-.4860098696e-2,.2041241452e-1,-.2916059217e-2,.2624453295e-1,0.,0.,.4536092117e-2,0.,-.6804138176e-2,.1360827635e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,.1178511302e-1,-.1683587573e-2,.1515228816e-1,0.,0.,.3928371006e-2,0.,-.5892556509e-2,.1178511302e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1014301032e-2,0.,-.1521451549e-2,.3042903097e-2,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,0.,0.,.3367175148e-2,.3086577219e-2,0.,-.5050762721e-2,-.7482611439e-3,.1010152544e-1,-.3928371006e-2,.1795826746e-1,0.,0.,0.,-.1262690680e-2,0.,0.,.1683587573e-2,0.,-.2525381361e-2,.5050762721e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7290148044e-3,0.,0.,.9720197392e-3,0.,-.1458029609e-2,.2916059217e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5019488348e-3,0.,0.,0.,-.6274360436e-3,0.,0.,.8365813917e-3,0.,-.1254872087e-2,.2509744174e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,0.,.9622504490e-2,.5498573994e-2,0.,.4123930495e-2,0.,0.,.7216878367e-1,.5773502693e-1,.4330127020e-1,0.,.1443375673e-1,.9622504490e-2,0.,.4811252245e-2,0.,0.,.4811252245e-1,.3849001795e-1,0.,.9622504490e-2,0.,0.,.3608439183e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,0.,.8333333333e-2,.5555555556e-2,0.,.2777777778e-2,0.,0.,.4166666667e-1,.3333333333e-1,0.,.8333333333e-2,0.,0.,.3750000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,.8606629657e-2,0.,.2151657415e-2,0.,0.,.1613743061e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2727723628e-2,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,0.,.1346870059e-1,.5892556509e-2,0.,.7576144082e-2,0.,0.,0.,-.1178511302e-1,-.1515228816e-1,0.,.1178511302e-1,.5050762721e-2,0.,.6734350295e-2,0.,0.,0.,-.7856742012e-2,0.,.7856742012e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,0.,.3888078956e-2,0.,0.,0.,-.6804138176e-2,0.,.6804138176e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1756820922e-2,0.,.1756820922e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,0.,.4860098696e-2,0.,0.,.4860098696e-2,0.,0.,0.,0.,.2187044413e-2,0.,0.,-.2430049348e-2,0.,.2430049348e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1262690680e-2,0.,0.,-.1402989645e-2,0.,.1402989645e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.8694008848e-3,0.,0.,0.,.9780759957e-3,0.,0.,-.1086751106e-2,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,0.,0.,.4347004424e-2,.4890379978e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,.1521451549e-1,.1956151991e-1,0.,0.,.2173502212e-2,0.,0.,0.,0.,0.,.1014301032e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,.8784104612e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2268046058e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,0.,0.,.3764616260e-2,.3137180218e-2,0.,0.,.1673162783e-2,0.,0.,0.,0.,0.,0.,-.2823462197e-2,0.,0.,.1882308131e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1630126659e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,0.,0.,0.,-.9820927513e-3,0.,0.,.6547285010e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,0.,0.,0.,.7423923386e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3340765524e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,0.,0.,0.,.6640158940e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_poly_to_dofs
	 *********************************************************/
public:
	static inline void madd_convert_poly_to_dofs(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.1178511302,.2357022604e-1,.7856742012e-2,.7071067810e-1,.1178511302e-1,.4714045206e-1,.3535533905,.1178511302,.5892556509e-1,.3535533905e-1,.1178511302,.2946278254e-1,.1178511302e-1,.5892556509e-1,.1178511302e-1,.3535533905e-1,.2357022604,.7856742012e-1,.3928371006e-1,.7856742012e-1,.1964185503e-1,.3928371006e-1,.1767766952,.5892556509e-1,.5892556509e-1,.1414213562},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-1,.6804138176e-2,.2041241452e-1,.2041241452,.6804138176e-1,.3402069088e-1,.6804138176e-1,.1701034544e-1,.3402069088e-1,.1837117307,.6123724358e-1,.6123724358e-1,.1632993162},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.8784104612e-2,.1756820922e-1,.4392052306e-2,.8784104612e-2,.7905694150e-1,.2635231383e-1,.2635231383e-1,.9035079028e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1,.4454354030e-2,.4454354030e-2,.2672612419e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3367175148e-2},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.1000000000,.1190476190e-1,.7619047619e-1,0.,-.4166666667e-1,-.3333333333e-1,-.2500000000e-1,.8333333333e-1,.8333333333e-2,0.,.6666666667e-1,.8333333333e-2,.5000000000e-1,0.,-.2777777778e-1,-.2222222222e-1,.5555555556e-1,.5555555556e-2,.4444444444e-1,0.,-.2083333333e-1,.4166666667e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,.4811252245e-1,.4811252245e-2,0.,.3849001795e-1,.4811252245e-2,.2886751347e-1,0.,-.2405626122e-1,-.1924500898e-1,.4811252245e-1,.4811252245e-2,.3849001795e-1,0.,-.2165063510e-1,.4330127020e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,-.4969039948e-2,.1242259987e-1,.1242259987e-2,.9938079897e-2,0.,-.9316949905e-2,.1863389981e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1574851971e-2,.3149703940e-2,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,.5248906592e-1,0.,.5248906592e-1,0.,0.,.6804138176e-2,.8748177653e-2,0.,-.1020620726e-1,-.4860098696e-2,.2041241452e-1,-.2916059217e-2,.2624453295e-1,0.,0.,.4536092117e-2,0.,-.6804138176e-2,.1360827635e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,.1178511302e-1,-.1683587573e-2,.1515228816e-1,0.,0.,.3928371006e-2,0.,-.5892556509e-2,.1178511302e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1014301032e-2,0.,-.1521451549e-2,.3042903097e-2,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,0.,0.,.3367175148e-2,.3086577219e-2,0.,-.5050762721e-2,-.7482611439e-3,.1010152544e-1,-.3928371006e-2,.1795826746e-1,0.,0.,0.,-.1262690680e-2,0.,0.,.1683587573e-2,0.,-.2525381361e-2,.5050762721e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7290148044e-3,0.,0.,.9720197392e-3,0.,-.1458029609e-2,.2916059217e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5019488348e-3,0.,0.,0.,-.6274360436e-3,0.,0.,.8365813917e-3,0.,-.1254872087e-2,.2509744174e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,0.,.9622504490e-2,.5498573994e-2,0.,.4123930495e-2,0.,0.,.7216878367e-1,.5773502693e-1,.4330127020e-1,0.,.1443375673e-1,.9622504490e-2,0.,.4811252245e-2,0.,0.,.4811252245e-1,.3849001795e-1,0.,.9622504490e-2,0.,0.,.3608439183e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,0.,.8333333333e-2,.5555555556e-2,0.,.2777777778e-2,0.,0.,.4166666667e-1,.3333333333e-1,0.,.8333333333e-2,0.,0.,.3750000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,.8606629657e-2,0.,.2151657415e-2,0.,0.,.1613743061e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2727723628e-2,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,0.,.1346870059e-1,.5892556509e-2,0.,.7576144082e-2,0.,0.,0.,-.1178511302e-1,-.1515228816e-1,0.,.1178511302e-1,.5050762721e-2,0.,.6734350295e-2,0.,0.,0.,-.7856742012e-2,0.,.7856742012e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,0.,.3888078956e-2,0.,0.,0.,-.6804138176e-2,0.,.6804138176e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1756820922e-2,0.,.1756820922e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,0.,.4860098696e-2,0.,0.,.4860098696e-2,0.,0.,0.,0.,.2187044413e-2,0.,0.,-.2430049348e-2,0.,.2430049348e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1262690680e-2,0.,0.,-.1402989645e-2,0.,.1402989645e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.8694008848e-3,0.,0.,0.,.9780759957e-3,0.,0.,-.1086751106e-2,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,0.,0.,.4347004424e-2,.4890379978e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,.1521451549e-1,.1956151991e-1,0.,0.,.2173502212e-2,0.,0.,0.,0.,0.,.1014301032e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,.8784104612e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2268046058e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,0.,0.,.3764616260e-2,.3137180218e-2,0.,0.,.1673162783e-2,0.,0.,0.,0.,0.,0.,-.2823462197e-2,0.,0.,.1882308131e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1630126659e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,0.,0.,0.,-.9820927513e-3,0.,0.,.6547285010e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,0.,0.,0.,.7423923386e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3340765524e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,0.,0.,0.,.6640158940e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_poly(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-2.,3.464101616,-4.472135954,5.291502622,2.449489743,-4.242640686,5.477225575,-2.828427124,4.898979486,3.162277660,-3.464101616,6.,-7.745966692,9.165151390,4.242640686,-7.348469229,9.486832980,-4.898979486,8.485281372,5.477225575,5.477225575,-9.486832980,12.24744872,-6.324555320,10.95445115,7.071067810,-7.483314774,12.96148140,8.366600265,9.486832980},
			{0.,4.898979486,-18.97366596,44.89988864,-84.85281372,0.,-6.928203232,26.83281572,-63.49803146,0.,8.485281372,-32.86335345,0.,-9.797958972,0.,0.,-12.,46.47580015,-109.9818167,0.,14.69693846,-56.92099788,0.,-16.97056274,0.,0.,18.97366596,-73.48469229,0.,-21.90890230,0.,0.,-25.92296279,0.,0.},
			{0.,0.,18.97366596,-112.2497216,381.8376617,0.,0.,-26.83281572,158.7450787,0.,0.,32.86335345,0.,0.,0.,0.,0.,-46.47580015,274.9545417,0.,0.,56.92099788,0.,0.,0.,0.,0.,73.48469229,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,-593.9696960,0.,0.,0.,-105.8300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,-183.3030278,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,296.9848480,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,6.,-10.39230485,13.41640786,-15.87450787,-19.59591794,33.94112549,-43.81780460,42.42640686,-73.48469229,-75.89466384,3.464101616,-6.,7.745966692,-9.165151390,-25.45584412,44.09081537,-56.92099788,63.68673332,-110.3086578,-120.4989626,-10.95445115,18.97366596,-24.49489743,56.92099788,-98.59006035,-127.2792206,22.44994432,-38.88444419,-100.3992032,-37.94733192},
			{0.,0.,0.,0.,0.,0.,20.78460970,-80.49844717,190.4940944,0.,-67.88225098,262.9068276,0.,146.9693846,0.,0.,12.,-46.47580015,109.9818167,0.,-88.18163075,341.5259873,0.,220.6173157,0.,0.,-37.94733192,146.9693846,0.,197.1801207,0.,0.,77.76888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,80.49844717,-476.2352360,0.,0.,-262.9068276,0.,0.,0.,0.,0.,46.47580015,-274.9545417,0.,0.,-341.5259873,0.,0.,0.,0.,0.,-146.9693846,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,317.4901573,0.,0.,0.,0.,0.,0.,0.,0.,0.,183.3030278,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,54.77225575,-127.2792206,220.4540769,398.4469852,0.,0.,0.,0.,21.21320343,-36.74234614,47.43416490,-161.6663230,280.0142853,575.1086854,5.477225575,-9.486832980,12.24744872,-94.86832980,164.3167672,487.9036789,-22.44994432,38.88444419,250.9980080,56.92099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,-328.6335345,0.,-440.9081537,0.,0.,0.,0.,0.,0.,73.48469229,-284.6049894,0.,-560.0285706,0.,0.,18.97366596,-73.48469229,0.,-328.6335345,0.,0.,-77.76888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,328.6335345,0.,0.,0.,0.,0.,0.,0.,0.,0.,284.6049894,0.,0.,0.,0.,0.,73.48469229,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,-171.4642820,-708.3501958,0.,0.,0.,0.,0.,0.,0.,102.8785692,-178.1909088,-920.1738966,0.,0.,0.,44.27188724,-76.68115805,-622.2539673,7.483314774,-12.96148140,-234.2648074,-37.94733192},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,342.9285640,0.,0.,0.,0.,0.,0.,0.,0.,0.,356.3818176,0.,0.,0.,0.,0.,153.3623161,0.,0.,25.92296279,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,398.4469852,0.,0.,0.,0.,0.,0.,0.,0.,0.,460.0869483,0.,0.,0.,0.,0.,254.5584412,0.,0.,75.29940238,9.486832980},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-18.33030278,-8.485281372,14.69693846,-18.97366596,9.797958972,-16.97056274,-10.95445115,-32.86335345,56.92099788,-73.48469229,37.94733192,-65.72670690,-42.42640686,89.79977729,-155.5377768,-100.3992032,-189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,219.9636334,0.,-29.39387692,113.8419958,0.,33.94112549,0.,0.,-113.8419958,440.9081537,0.,131.4534138,0.,0.,311.0755535,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,-549.9090834,0.,0.,-113.8419958,0.,0.,0.,0.,0.,-440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,366.6060556,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,94.86832980,-117.5755077,203.6467529,230.0434742,32.86335345,-56.92099788,73.48469229,-303.5786554,525.8136552,721.2489166,-179.5995546,311.0755535,1104.391235,569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,-569.2099788,0.,-407.2935059,0.,0.,113.8419958,-440.9081537,0.,-1051.627310,0.,0.,-622.1511070,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,569.2099788,0.,0.,0.,0.,0.,440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-356.3818176,-920.1738966,0.,0.,0.,265.6313234,-460.0869483,-2206.173157,89.79977729,-155.5377768,-1907.584860,-569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,712.7636352,0.,0.,0.,0.,0.,920.1738966,0.,0.,311.0755535,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,0.,0.,0.,1527.350647,0.,0.,903.5928286,189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,73.48469229,-37.94733192,65.72670690,42.42640686,-224.4994432,388.8444419,250.9980080,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,-440.9081537,0.,-131.4534138,0.,0.,-777.6888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-460.0869483,-678.8225098,224.4994432,-388.8444419,-2509.980080,-1707.629936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,777.6888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,0.,0.,2258.982072,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-259.2296279,-167.3320053,-1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,518.4592558,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_poly
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_poly(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-2.,3.464101616,-4.472135954,5.291502622,2.449489743,-4.242640686,5.477225575,-2.828427124,4.898979486,3.162277660,-3.464101616,6.,-7.745966692,9.165151390,4.242640686,-7.348469229,9.486832980,-4.898979486,8.485281372,5.477225575,5.477225575,-9.486832980,12.24744872,-6.324555320,10.95445115,7.071067810,-7.483314774,12.96148140,8.366600265,9.486832980},
			{0.,4.898979486,-18.97366596,44.89988864,-84.85281372,0.,-6.928203232,26.83281572,-63.49803146,0.,8.485281372,-32.86335345,0.,-9.797958972,0.,0.,-12.,46.47580015,-109.9818167,0.,14.69693846,-56.92099788,0.,-16.97056274,0.,0.,18.97366596,-73.48469229,0.,-21.90890230,0.,0.,-25.92296279,0.,0.},
			{0.,0.,18.97366596,-112.2497216,381.8376617,0.,0.,-26.83281572,158.7450787,0.,0.,32.86335345,0.,0.,0.,0.,0.,-46.47580015,274.9545417,0.,0.,56.92099788,0.,0.,0.,0.,0.,73.48469229,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,-593.9696960,0.,0.,0.,-105.8300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,-183.3030278,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,296.9848480,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,6.,-10.39230485,13.41640786,-15.87450787,-19.59591794,33.94112549,-43.81780460,42.42640686,-73.48469229,-75.89466384,3.464101616,-6.,7.745966692,-9.165151390,-25.45584412,44.09081537,-56.92099788,63.68673332,-110.3086578,-120.4989626,-10.95445115,18.97366596,-24.49489743,56.92099788,-98.59006035,-127.2792206,22.44994432,-38.88444419,-100.3992032,-37.94733192},
			{0.,0.,0.,0.,0.,0.,20.78460970,-80.49844717,190.4940944,0.,-67.88225098,262.9068276,0.,146.9693846,0.,0.,12.,-46.47580015,109.9818167,0.,-88.18163075,341.5259873,0.,220.6173157,0.,0.,-37.94733192,146.9693846,0.,197.1801207,0.,0.,77.76888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,80.49844717,-476.2352360,0.,0.,-262.9068276,0.,0.,0.,0.,0.,46.47580015,-274.9545417,0.,0.,-341.5259873,0.,0.,0.,0.,0.,-146.9693846,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,317.4901573,0.,0.,0.,0.,0.,0.,0.,0.,0.,183.3030278,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,54.77225575,-127.2792206,220.4540769,398.4469852,0.,0.,0.,0.,21.21320343,-36.74234614,47.43416490,-161.6663230,280.0142853,575.1086854,5.477225575,-9.486832980,12.24744872,-94.86832980,164.3167672,487.9036789,-22.44994432,38.88444419,250.9980080,56.92099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,-328.6335345,0.,-440.9081537,0.,0.,0.,0.,0.,0.,73.48469229,-284.6049894,0.,-560.0285706,0.,0.,18.97366596,-73.48469229,0.,-328.6335345,0.,0.,-77.76888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,328.6335345,0.,0.,0.,0.,0.,0.,0.,0.,0.,284.6049894,0.,0.,0.,0.,0.,73.48469229,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,-171.4642820,-708.3501958,0.,0.,0.,0.,0.,0.,0.,102.8785692,-178.1909088,-920.1738966,0.,0.,0.,44.27188724,-76.68115805,-622.2539673,7.483314774,-12.96148140,-234.2648074,-37.94733192},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,342.9285640,0.,0.,0.,0.,0.,0.,0.,0.,0.,356.3818176,0.,0.,0.,0.,0.,153.3623161,0.,0.,25.92296279,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,398.4469852,0.,0.,0.,0.,0.,0.,0.,0.,0.,460.0869483,0.,0.,0.,0.,0.,254.5584412,0.,0.,75.29940238,9.486832980},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-18.33030278,-8.485281372,14.69693846,-18.97366596,9.797958972,-16.97056274,-10.95445115,-32.86335345,56.92099788,-73.48469229,37.94733192,-65.72670690,-42.42640686,89.79977729,-155.5377768,-100.3992032,-189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,219.9636334,0.,-29.39387692,113.8419958,0.,33.94112549,0.,0.,-113.8419958,440.9081537,0.,131.4534138,0.,0.,311.0755535,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,-549.9090834,0.,0.,-113.8419958,0.,0.,0.,0.,0.,-440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,366.6060556,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,94.86832980,-117.5755077,203.6467529,230.0434742,32.86335345,-56.92099788,73.48469229,-303.5786554,525.8136552,721.2489166,-179.5995546,311.0755535,1104.391235,569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,-569.2099788,0.,-407.2935059,0.,0.,113.8419958,-440.9081537,0.,-1051.627310,0.,0.,-622.1511070,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,569.2099788,0.,0.,0.,0.,0.,440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-356.3818176,-920.1738966,0.,0.,0.,265.6313234,-460.0869483,-2206.173157,89.79977729,-155.5377768,-1907.584860,-569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,712.7636352,0.,0.,0.,0.,0.,920.1738966,0.,0.,311.0755535,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,0.,0.,0.,1527.350647,0.,0.,903.5928286,189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,73.48469229,-37.94733192,65.72670690,42.42640686,-224.4994432,388.8444419,250.9980080,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,-440.9081537,0.,-131.4534138,0.,0.,-777.6888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-460.0869483,-678.8225098,224.4994432,-388.8444419,-2509.980080,-1707.629936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,777.6888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,0.,0.,2258.982072,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-259.2296279,-167.3320053,-1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,518.4592558,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void mul_convert_dofs_to_0th_order(
		const T i_input[35],
		T o_output[1]
	) {
		static const T matrix[1][35] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * convert_dofs_to_0th_order
	 *********************************************************/
public:
	static inline void madd_convert_dofs_to_0th_order(
		const T i_input[35],
		T o_output[1]
	) {
		static const T matrix[1][35] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 1; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void mul_timestep_stiffness_w(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_stiffness_w
	 *********************************************************/
public:
	static inline void madd_timestep_stiffness_w(
		const T i_input[35],
		T o_output[35]
	) {
		static const T matrix[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_bottom_face(
		const T i_input[35],
		T o_output[27]
	) {
		static const T matrix[27][35] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.207666290,2.091739373,2.700423918,3.195184670,-1.025843012,-1.776812217,-2.293854709,-1.398249481,-2.421839143,.7821963046,-2.091739372,-3.622998868,-4.677271425,-5.534222184,-4.286111665,-7.423763173,-9.584037044,-1.735425509,-3.005845154,3.129410556,2.402465932,4.161193056,5.372077139,7.607430278,13.17645576,8.483868382,-2.650906245,-4.591504302,-11.29653454,2.875267208},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.207666290,2.091739373,2.700423918,3.195184670,-1.025843012,-1.776812217,-2.293854709,-1.398249481,-2.421839143,.7821963046,2.091739375,3.622998872,4.677271433,5.534222193,4.286111669,7.423763181,9.584037053,1.735425510,3.005845155,-3.129410559,2.402465935,4.161193062,5.372077144,7.607430290,13.17645578,8.483868396,2.650906248,4.591504307,11.29653455,2.875267158},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-2.415332579,-4.183478747,-5.400847836,-6.390369340,3.923332432,6.795411104,8.772838015,-6.407991400,-11.09896668,10.56920547,.1593486743e-8,.276e-8,.3563144678e-8,.4215969639e-8,-.2627089028e-8,-.4550251674e-8,-.5874349651e-8,.4352222914e-8,.7538271208e-8,-.7266379083e-8,-3.130880433,-5.422843982,-7.000861478,5.367005593,9.295926372,-9.215865385,-.1571496103e-7,-.2721911093e-7,.2851586381e-7,4.649692528},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.393815818,.6821090058,.8805989395,1.041938717,-1.469667825,-2.545539342,-3.286277160,.1253720207,.2171507097,1.416680365,-.6821090051,-1.181447453,-1.525242103,-1.804690794,-.8311042347,-1.439514761,-1.858405566,.4291959607,.7433892102,.7095745382,-.6707402935,-1.161756267,-1.499820892,-1.388519566,-2.404986437,-.3003409923,.6556076472,1.135545755,1.898978815,.2835315068e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.393815818,.6821090058,.8805989395,1.041938717,-1.469667825,-2.545539342,-3.286277160,.1253720207,.2171507097,1.416680365,.6821090058,1.181447454,1.525242105,1.804690796,.8311042356,1.439514763,1.858405568,-.4291959612,-.7433892109,-.7095745387,-.6707402924,-1.161756265,-1.499820889,-1.388519564,-2.404986432,-.3003409919,-.6556076472,-1.135545755,-1.898978815,.2835315542e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.787631635,-1.364218011,-1.761197878,-2.083877432,-.5099913165,-.8833308713,-1.140375252,1.364354035,2.363130509,-1.084617593,.3464101616e-9,.6e-9,.7745966692e-9,.9165151390e-9,.4372877027e-11,.7574045189e-11,.9778050292e-11,-.2779346762e-9,-.4813969803e-9,.2775108880e-9,-1.743691247,-3.020161833,-3.899012161,-.8344299601,-1.445275086,1.717913014,0.,0.,0.,1.442216374},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.780140598,-1.351243152,-1.744447408,-2.064058009,-.5220598641,-.9042342088,-1.167361344,1.368139104,2.369686441,-1.070141386,1.351243151,2.340421790,3.021471539,3.575053338,.2738823620e-1,.4743781665e-1,.6124195795e-1,-1.092460490,-1.892197074,1.075289136,-.4881593780,-.8455168449,-1.091557553,-.2385311826,-.4131481275,.4820493399,-1.668760973,-2.890378790,-1.548163659,-1.041619966},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.780140598,-1.351243152,-1.744447408,-2.064058009,-.5220598641,-.9042342088,-1.167361344,1.368139104,2.369686441,-1.070141386,-1.351243152,-2.340421793,-3.021471542,-3.575053341,-.2738823623e-1,-.4743781670e-1,-.6124195801e-1,1.092460491,1.892197076,-1.075289137,-.4881593725,-.8455168354,-1.091557541,-.2385311799,-.4131481229,.4820493345,1.668760962,2.890378772,1.548163649,-1.041620004},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.560281195,2.702486303,3.488894813,4.128116015,-.5536850767,-.9590106839,-1.238077470,-1.785522687,-3.092616012,-.1767011503,-.3464101616e-9,-.6e-9,-.7745966692e-9,-.9165151390e-9,-.8344854191e-9,-.1445371145e-8,-.1865966124e-8,-.6239040877e-9,-.1080633579e-8,.3096646864e-9,-.4528013083,-.7842748716,-1.012494506,-1.648891590,-2.855964011,-2.444365371,-.7483314774e-8,-.1296148140e-7,-.3631457412e-7,.9725382146e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.283894526,-2.223770551,-2.870875435,-3.396865625,.4596200614,.7960852982,1.027741701,.5904589114,1.022704835,-1.343510775,2.223770553,3.851683580,4.972502121,5.883543854,-1.098259750,-1.902241688,-2.455783459,-.4185010573,-.7248650942,1.591483172,1.261805030,2.185510421,2.821481821,-.2397411885,-.4152439192,-.6463928632,-.6392709500,-1.107249765,-.5300284972e-1,-2.181654023},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.283894526,-2.223770551,-2.870875435,-3.396865625,.4596200614,.7960852982,1.027741701,.5904589114,1.022704835,-1.343510775,-2.223770551,-3.851683579,-4.972502117,-5.883543850,1.098259750,1.902241687,2.455783458,.4185010573,.7248650939,-1.591483171,1.261805025,2.185510412,2.821481811,-.2397411877,-.4152439177,-.6463928607,.6392709245,1.107249721,.5300284761e-1,-2.181654030},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.567789052,4.447541102,5.741750871,6.793731251,1.727781187,2.992604800,3.863436185,-.6176495990,-1.069800487,-2.390029691,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1560422168,-.2702730475,-.3489210041,-.7800237319,-1.351040735,-1.950823367,0.,0.,0.,.1154979035e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.805810311,-3.127755208,-4.037914609,-4.777724998,1.840927032,3.188579152,4.116437986,-1.585266239,-2.745761671,1.099742715,3.127755210,5.417430937,6.993873264,8.275262446,-3.210799769,-5.561268334,-7.179566546,2.802690869,4.854402980,-2.005254469,4.133650004,7.159691827,9.243122405,-3.691753629,-6.394304856,2.774298724,4.281181012,7.415223028,-3.392270752,3.547918793},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.805810311,-3.127755208,-4.037914609,-4.777724998,1.840927032,3.188579152,4.116437986,-1.585266239,-2.745761671,1.099742715,-3.127755208,-5.417430934,-6.993873261,-8.275262441,3.210799767,5.561268332,7.179566543,-2.802690867,-4.854402979,2.005254469,4.133650001,7.159691821,9.243122398,-3.691753626,-6.394304850,2.774298722,-4.281181016,-7.415223034,3.392270755,3.547919252},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,3.611620623,6.255510417,8.075829218,9.555449995,5.548439256,9.610178692,12.40668735,6.505405380,11.26769265,6.079065903,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1147467860e-1,-.1987472634e-1,-.2565816138e-1,-.7349529016e-1,-.1272975767,-.2596251910,-.5986651819e-8,-.1036918512e-7,-.4964694369e-7,.6246716172e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.139723890,.2420088766,.3124321161,.3696746651,-1.423600127,-2.465747748,-3.183266656,.6044092113,1.046867463,1.088285796,-1.881087112,-3.258138451,-4.206238652,-4.976888691,-1.804153808,-3.124886061,-4.034210557,1.619309591,2.804726483,1.142629197,1.289024480,2.232655892,2.882346363,2.227214966,3.857649481,-.2121957890,-.4725155092,-.8184208692,-1.167298565,-.1238885768},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.559207282,2.700626232,3.486493472,4.125274710,-.5553799448,-.9619462815,-1.241867310,-1.784795277,-3.091356102,-.1734310073,1.061547995,1.838655062,2.373693478,2.808592000,2.556054550,4.427216349,5.715511728,1.908436179,3.305508424,-.9519572496,.3183248011,.5513547288,.7117958945,1.158730221,2.006979616,1.716453276,-.3086735638e-1,-.5346382954e-1,-.1497356346,-.8941606665e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.698931170,-2.942635106,-3.798925584,-4.494949370,1.527877431,2.646361336,3.416437796,-1.007511332,-1.745060817,.2702851370,.8195391220,1.419483395,1.832545187,2.168296706,-.7519007415,-1.302330287,-1.681301170,.5224058015,.9048333903,-.1906718870,-2.010827754,-3.482855835,-4.496347549,1.506342189,2.609061204,-.7470992591,-2.148063678,-3.720555426,1.317034145,1.229138747},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.698931170,-2.942635106,-3.798925584,-4.494949370,1.527877431,2.646361336,3.416437796,-1.007511332,-1.745060817,.2702851370,-.8195391179,-1.419483391,-1.832545177,-2.168296695,.7519007376,1.302330280,1.681301162,-.5224057986,-.9048333851,.1906718861,-2.010827749,-3.482855828,-4.496347539,1.506342185,2.609061198,-.7470992577,2.148063733,3.720555523,-1.317034179,1.229138974},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.139723890,.2420088766,.3124321161,.3696746651,-1.423600127,-2.465747748,-3.183266656,.6044092113,1.046867463,1.088285796,1.881087116,3.258138455,4.206238662,4.976888702,1.804153811,3.124886068,4.034210565,-1.619309594,-2.804726489,-1.142629199,1.289024475,2.232655882,2.882346350,2.227214957,3.857649465,-.2121957881,.4725155698,.8184209742,1.167298714,-.1238887416},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.559207282,2.700626232,3.486493472,4.125274710,-.5553799448,-.9619462815,-1.241867310,-1.784795277,-3.091356102,-.1734310073,-1.061547993,-1.838655058,-2.373693472,-2.808591992,-2.556054543,-4.427216337,-5.715511713,-1.908436174,-3.305508415,.9519572471,.3183248000,.5513547270,.7117958920,1.158730217,2.006979608,1.716453270,.3086736086e-1,.5346383732e-1,.1497356563,-.8941604766e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.971066118,-1.681935855,-2.171369850,-2.569199455,-.1906367772,-.3301925839,-.4262767929,1.203401099,2.084351846,-1.362876347,-2.724401559,-4.718801920,-6.091947082,-7.208088995,.4756631459,.8238727362,1.063615129,1.696386578,2.938227741,-2.388976129,3.201877769,5.545814974,7.159616348,.7410054061,1.283459012,-2.832011274,-3.040858372,-5.266921197,-1.847444888,2.450578880},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.844934019,4.927570268,6.361465859,7.526987913,2.597617435,4.499205374,5.808449163,.5616969466,.9728876503,-1.873883554,.5212328525,.9028017828,1.165512090,1.379052502,1.939035835,3.358508584,4.335815938,3.687840215,6.387526618,4.461246382,.8451479178e-1,.1463839133,.1889808195,.4540265923,.7863971259,1.260587466,.1137306696e-2,.1969872981e-2,.7969298371e-2,-.3689094461e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.873867901,-3.245634412,-4.190096007,-4.957788456,2.048368944,3.547879081,4.580292201,-1.991866711,-3.450014347,1.736399667,2.203168706,3.816000135,4.926434990,5.829036490,-2.414698978,-4.182381316,-5.399431060,2.358674124,4.085343421,-2.072270251,.6985932727,1.209999042,1.562102047,-.6879618021,-1.191584795,.6128789301,-2.029197433,-3.514673052,1.839475646,-3.732912624},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.873867901,-3.245634412,-4.190096007,-4.957788456,2.048368944,3.547879081,4.580292201,-1.991866711,-3.450014347,1.736399667,-2.203168707,-3.816000137,-4.926434994,-5.829036495,2.414698980,4.182381320,5.399431066,-2.358674126,-4.085343424,2.072270252,.6985932694,1.209999036,1.562102039,-.6879617990,-1.191584790,.6128789271,2.029197438,3.514673060,-1.839475650,-3.732912697},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.971066118,-1.681935855,-2.171369850,-2.569199455,-.1906367772,-.3301925839,-.4262767929,1.203401099,2.084351846,-1.362876347,2.724401559,4.718801918,6.091947082,7.208088995,-.4756631459,-.8238727362,-1.063615129,-1.696386578,-2.938227741,2.388976128,3.201877775,5.545814986,7.159616362,.7410054080,1.283459015,-2.832011281,3.040858346,5.266921153,1.847444873,2.450579118},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.844934019,4.927570268,6.361465859,7.526987913,2.597617435,4.499205374,5.808449163,.5616969466,.9728876503,-1.873883554,-.5212328521,-.9028017822,-1.165512089,-1.379052502,-1.939035833,-3.358508582,-4.335815935,-3.687840212,-6.387526614,-4.461246379,.8451479167e-1,.1463839132,.1889808193,.4540265917,.7863971248,1.260587465,-.1137294723e-2,-.1969852243e-2,-.7969214472e-2,-.3689141724e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 27; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_bottom_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_bottom_face(
		const T i_input[35],
		T o_output[27]
	) {
		static const T matrix[27][35] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.207666290,2.091739373,2.700423918,3.195184670,-1.025843012,-1.776812217,-2.293854709,-1.398249481,-2.421839143,.7821963046,-2.091739372,-3.622998868,-4.677271425,-5.534222184,-4.286111665,-7.423763173,-9.584037044,-1.735425509,-3.005845154,3.129410556,2.402465932,4.161193056,5.372077139,7.607430278,13.17645576,8.483868382,-2.650906245,-4.591504302,-11.29653454,2.875267208},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.207666290,2.091739373,2.700423918,3.195184670,-1.025843012,-1.776812217,-2.293854709,-1.398249481,-2.421839143,.7821963046,2.091739375,3.622998872,4.677271433,5.534222193,4.286111669,7.423763181,9.584037053,1.735425510,3.005845155,-3.129410559,2.402465935,4.161193062,5.372077144,7.607430290,13.17645578,8.483868396,2.650906248,4.591504307,11.29653455,2.875267158},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-2.415332579,-4.183478747,-5.400847836,-6.390369340,3.923332432,6.795411104,8.772838015,-6.407991400,-11.09896668,10.56920547,.1593486743e-8,.276e-8,.3563144678e-8,.4215969639e-8,-.2627089028e-8,-.4550251674e-8,-.5874349651e-8,.4352222914e-8,.7538271208e-8,-.7266379083e-8,-3.130880433,-5.422843982,-7.000861478,5.367005593,9.295926372,-9.215865385,-.1571496103e-7,-.2721911093e-7,.2851586381e-7,4.649692528},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.393815818,.6821090058,.8805989395,1.041938717,-1.469667825,-2.545539342,-3.286277160,.1253720207,.2171507097,1.416680365,-.6821090051,-1.181447453,-1.525242103,-1.804690794,-.8311042347,-1.439514761,-1.858405566,.4291959607,.7433892102,.7095745382,-.6707402935,-1.161756267,-1.499820892,-1.388519566,-2.404986437,-.3003409923,.6556076472,1.135545755,1.898978815,.2835315068e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.393815818,.6821090058,.8805989395,1.041938717,-1.469667825,-2.545539342,-3.286277160,.1253720207,.2171507097,1.416680365,.6821090058,1.181447454,1.525242105,1.804690796,.8311042356,1.439514763,1.858405568,-.4291959612,-.7433892109,-.7095745387,-.6707402924,-1.161756265,-1.499820889,-1.388519564,-2.404986432,-.3003409919,-.6556076472,-1.135545755,-1.898978815,.2835315542e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.787631635,-1.364218011,-1.761197878,-2.083877432,-.5099913165,-.8833308713,-1.140375252,1.364354035,2.363130509,-1.084617593,.3464101616e-9,.6e-9,.7745966692e-9,.9165151390e-9,.4372877027e-11,.7574045189e-11,.9778050292e-11,-.2779346762e-9,-.4813969803e-9,.2775108880e-9,-1.743691247,-3.020161833,-3.899012161,-.8344299601,-1.445275086,1.717913014,0.,0.,0.,1.442216374},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.780140598,-1.351243152,-1.744447408,-2.064058009,-.5220598641,-.9042342088,-1.167361344,1.368139104,2.369686441,-1.070141386,1.351243151,2.340421790,3.021471539,3.575053338,.2738823620e-1,.4743781665e-1,.6124195795e-1,-1.092460490,-1.892197074,1.075289136,-.4881593780,-.8455168449,-1.091557553,-.2385311826,-.4131481275,.4820493399,-1.668760973,-2.890378790,-1.548163659,-1.041619966},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.780140598,-1.351243152,-1.744447408,-2.064058009,-.5220598641,-.9042342088,-1.167361344,1.368139104,2.369686441,-1.070141386,-1.351243152,-2.340421793,-3.021471542,-3.575053341,-.2738823623e-1,-.4743781670e-1,-.6124195801e-1,1.092460491,1.892197076,-1.075289137,-.4881593725,-.8455168354,-1.091557541,-.2385311799,-.4131481229,.4820493345,1.668760962,2.890378772,1.548163649,-1.041620004},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.560281195,2.702486303,3.488894813,4.128116015,-.5536850767,-.9590106839,-1.238077470,-1.785522687,-3.092616012,-.1767011503,-.3464101616e-9,-.6e-9,-.7745966692e-9,-.9165151390e-9,-.8344854191e-9,-.1445371145e-8,-.1865966124e-8,-.6239040877e-9,-.1080633579e-8,.3096646864e-9,-.4528013083,-.7842748716,-1.012494506,-1.648891590,-2.855964011,-2.444365371,-.7483314774e-8,-.1296148140e-7,-.3631457412e-7,.9725382146e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.283894526,-2.223770551,-2.870875435,-3.396865625,.4596200614,.7960852982,1.027741701,.5904589114,1.022704835,-1.343510775,2.223770553,3.851683580,4.972502121,5.883543854,-1.098259750,-1.902241688,-2.455783459,-.4185010573,-.7248650942,1.591483172,1.261805030,2.185510421,2.821481821,-.2397411885,-.4152439192,-.6463928632,-.6392709500,-1.107249765,-.5300284972e-1,-2.181654023},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.283894526,-2.223770551,-2.870875435,-3.396865625,.4596200614,.7960852982,1.027741701,.5904589114,1.022704835,-1.343510775,-2.223770551,-3.851683579,-4.972502117,-5.883543850,1.098259750,1.902241687,2.455783458,.4185010573,.7248650939,-1.591483171,1.261805025,2.185510412,2.821481811,-.2397411877,-.4152439177,-.6463928607,.6392709245,1.107249721,.5300284761e-1,-2.181654030},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.567789052,4.447541102,5.741750871,6.793731251,1.727781187,2.992604800,3.863436185,-.6176495990,-1.069800487,-2.390029691,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1560422168,-.2702730475,-.3489210041,-.7800237319,-1.351040735,-1.950823367,0.,0.,0.,.1154979035e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.805810311,-3.127755208,-4.037914609,-4.777724998,1.840927032,3.188579152,4.116437986,-1.585266239,-2.745761671,1.099742715,3.127755210,5.417430937,6.993873264,8.275262446,-3.210799769,-5.561268334,-7.179566546,2.802690869,4.854402980,-2.005254469,4.133650004,7.159691827,9.243122405,-3.691753629,-6.394304856,2.774298724,4.281181012,7.415223028,-3.392270752,3.547918793},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.805810311,-3.127755208,-4.037914609,-4.777724998,1.840927032,3.188579152,4.116437986,-1.585266239,-2.745761671,1.099742715,-3.127755208,-5.417430934,-6.993873261,-8.275262441,3.210799767,5.561268332,7.179566543,-2.802690867,-4.854402979,2.005254469,4.133650001,7.159691821,9.243122398,-3.691753626,-6.394304850,2.774298722,-4.281181016,-7.415223034,3.392270755,3.547919252},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,3.611620623,6.255510417,8.075829218,9.555449995,5.548439256,9.610178692,12.40668735,6.505405380,11.26769265,6.079065903,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1147467860e-1,-.1987472634e-1,-.2565816138e-1,-.7349529016e-1,-.1272975767,-.2596251910,-.5986651819e-8,-.1036918512e-7,-.4964694369e-7,.6246716172e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.139723890,.2420088766,.3124321161,.3696746651,-1.423600127,-2.465747748,-3.183266656,.6044092113,1.046867463,1.088285796,-1.881087112,-3.258138451,-4.206238652,-4.976888691,-1.804153808,-3.124886061,-4.034210557,1.619309591,2.804726483,1.142629197,1.289024480,2.232655892,2.882346363,2.227214966,3.857649481,-.2121957890,-.4725155092,-.8184208692,-1.167298565,-.1238885768},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.559207282,2.700626232,3.486493472,4.125274710,-.5553799448,-.9619462815,-1.241867310,-1.784795277,-3.091356102,-.1734310073,1.061547995,1.838655062,2.373693478,2.808592000,2.556054550,4.427216349,5.715511728,1.908436179,3.305508424,-.9519572496,.3183248011,.5513547288,.7117958945,1.158730221,2.006979616,1.716453276,-.3086735638e-1,-.5346382954e-1,-.1497356346,-.8941606665e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.698931170,-2.942635106,-3.798925584,-4.494949370,1.527877431,2.646361336,3.416437796,-1.007511332,-1.745060817,.2702851370,.8195391220,1.419483395,1.832545187,2.168296706,-.7519007415,-1.302330287,-1.681301170,.5224058015,.9048333903,-.1906718870,-2.010827754,-3.482855835,-4.496347549,1.506342189,2.609061204,-.7470992591,-2.148063678,-3.720555426,1.317034145,1.229138747},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.698931170,-2.942635106,-3.798925584,-4.494949370,1.527877431,2.646361336,3.416437796,-1.007511332,-1.745060817,.2702851370,-.8195391179,-1.419483391,-1.832545177,-2.168296695,.7519007376,1.302330280,1.681301162,-.5224057986,-.9048333851,.1906718861,-2.010827749,-3.482855828,-4.496347539,1.506342185,2.609061198,-.7470992577,2.148063733,3.720555523,-1.317034179,1.229138974},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.139723890,.2420088766,.3124321161,.3696746651,-1.423600127,-2.465747748,-3.183266656,.6044092113,1.046867463,1.088285796,1.881087116,3.258138455,4.206238662,4.976888702,1.804153811,3.124886068,4.034210565,-1.619309594,-2.804726489,-1.142629199,1.289024475,2.232655882,2.882346350,2.227214957,3.857649465,-.2121957881,.4725155698,.8184209742,1.167298714,-.1238887416},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.559207282,2.700626232,3.486493472,4.125274710,-.5553799448,-.9619462815,-1.241867310,-1.784795277,-3.091356102,-.1734310073,-1.061547993,-1.838655058,-2.373693472,-2.808591992,-2.556054543,-4.427216337,-5.715511713,-1.908436174,-3.305508415,.9519572471,.3183248000,.5513547270,.7117958920,1.158730217,2.006979608,1.716453270,.3086736086e-1,.5346383732e-1,.1497356563,-.8941604766e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.971066118,-1.681935855,-2.171369850,-2.569199455,-.1906367772,-.3301925839,-.4262767929,1.203401099,2.084351846,-1.362876347,-2.724401559,-4.718801920,-6.091947082,-7.208088995,.4756631459,.8238727362,1.063615129,1.696386578,2.938227741,-2.388976129,3.201877769,5.545814974,7.159616348,.7410054061,1.283459012,-2.832011274,-3.040858372,-5.266921197,-1.847444888,2.450578880},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.844934019,4.927570268,6.361465859,7.526987913,2.597617435,4.499205374,5.808449163,.5616969466,.9728876503,-1.873883554,.5212328525,.9028017828,1.165512090,1.379052502,1.939035835,3.358508584,4.335815938,3.687840215,6.387526618,4.461246382,.8451479178e-1,.1463839133,.1889808195,.4540265923,.7863971259,1.260587466,.1137306696e-2,.1969872981e-2,.7969298371e-2,-.3689094461e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.873867901,-3.245634412,-4.190096007,-4.957788456,2.048368944,3.547879081,4.580292201,-1.991866711,-3.450014347,1.736399667,2.203168706,3.816000135,4.926434990,5.829036490,-2.414698978,-4.182381316,-5.399431060,2.358674124,4.085343421,-2.072270251,.6985932727,1.209999042,1.562102047,-.6879618021,-1.191584795,.6128789301,-2.029197433,-3.514673052,1.839475646,-3.732912624},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.873867901,-3.245634412,-4.190096007,-4.957788456,2.048368944,3.547879081,4.580292201,-1.991866711,-3.450014347,1.736399667,-2.203168707,-3.816000137,-4.926434994,-5.829036495,2.414698980,4.182381320,5.399431066,-2.358674126,-4.085343424,2.072270252,.6985932694,1.209999036,1.562102039,-.6879617990,-1.191584790,.6128789271,2.029197438,3.514673060,-1.839475650,-3.732912697},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.971066118,-1.681935855,-2.171369850,-2.569199455,-.1906367772,-.3301925839,-.4262767929,1.203401099,2.084351846,-1.362876347,2.724401559,4.718801918,6.091947082,7.208088995,-.4756631459,-.8238727362,-1.063615129,-1.696386578,-2.938227741,2.388976128,3.201877775,5.545814986,7.159616362,.7410054080,1.283459015,-2.832011281,3.040858346,5.266921153,1.847444873,2.450579118},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.844934019,4.927570268,6.361465859,7.526987913,2.597617435,4.499205374,5.808449163,.5616969466,.9728876503,-1.873883554,-.5212328521,-.9028017822,-1.165512089,-1.379052502,-1.939035833,-3.358508582,-4.335815935,-3.687840212,-6.387526614,-4.461246379,.8451479167e-1,.1463839132,.1889808193,.4540265917,.7863971248,1.260587465,-.1137294723e-2,-.1969852243e-2,-.7969214472e-2,-.3689141724e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 27; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void mul_timestep_dofs_to_top_face(
		const T i_input[35],
		T o_output[27]
	) {
		static const T matrix[27][35] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.207666290,-2.091739373,2.700423918,-3.195184670,-1.025843012,1.776812217,-2.293854709,-1.398249481,2.421839143,.7821963046,-2.091739372,3.622998868,-4.677271425,5.534222184,-4.286111665,7.423763173,-9.584037044,-1.735425509,3.005845154,3.129410556,2.402465932,-4.161193056,5.372077139,7.607430278,-13.17645576,8.483868382,-2.650906245,4.591504302,-11.29653454,2.875267208},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.207666290,-2.091739373,2.700423918,-3.195184670,-1.025843012,1.776812217,-2.293854709,-1.398249481,2.421839143,.7821963046,2.091739375,-3.622998872,4.677271433,-5.534222193,4.286111669,-7.423763181,9.584037053,1.735425510,-3.005845155,-3.129410559,2.402465935,-4.161193062,5.372077144,7.607430290,-13.17645578,8.483868396,2.650906248,-4.591504307,11.29653455,2.875267158},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-2.415332579,4.183478747,-5.400847836,6.390369340,3.923332432,-6.795411104,8.772838015,-6.407991400,11.09896668,10.56920547,.1593486743e-8,-.276e-8,.3563144678e-8,-.4215969639e-8,-.2627089028e-8,.4550251674e-8,-.5874349651e-8,.4352222914e-8,-.7538271208e-8,-.7266379083e-8,-3.130880433,5.422843982,-7.000861478,5.367005593,-9.295926372,-9.215865385,-.1571496103e-7,.2721911093e-7,.2851586381e-7,4.649692528},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.393815818,-.6821090058,.8805989395,-1.041938717,-1.469667825,2.545539342,-3.286277160,.1253720207,-.2171507097,1.416680365,-.6821090051,1.181447453,-1.525242103,1.804690794,-.8311042347,1.439514761,-1.858405566,.4291959607,-.7433892102,.7095745382,-.6707402935,1.161756267,-1.499820892,-1.388519566,2.404986437,-.3003409923,.6556076472,-1.135545755,1.898978815,.2835315068e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.393815818,-.6821090058,.8805989395,-1.041938717,-1.469667825,2.545539342,-3.286277160,.1253720207,-.2171507097,1.416680365,.6821090058,-1.181447454,1.525242105,-1.804690796,.8311042356,-1.439514763,1.858405568,-.4291959612,.7433892109,-.7095745387,-.6707402924,1.161756265,-1.499820889,-1.388519564,2.404986432,-.3003409919,-.6556076472,1.135545755,-1.898978815,.2835315542e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.787631635,1.364218011,-1.761197878,2.083877432,-.5099913165,.8833308713,-1.140375252,1.364354035,-2.363130509,-1.084617593,.3464101616e-9,-.6e-9,.7745966692e-9,-.9165151390e-9,.4372877027e-11,-.7574045189e-11,.9778050292e-11,-.2779346762e-9,.4813969803e-9,.2775108880e-9,-1.743691247,3.020161833,-3.899012161,-.8344299601,1.445275086,1.717913014,0.,-0.,0.,1.442216374},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.780140598,1.351243152,-1.744447408,2.064058009,-.5220598641,.9042342088,-1.167361344,1.368139104,-2.369686441,-1.070141386,1.351243151,-2.340421790,3.021471539,-3.575053338,.2738823620e-1,-.4743781665e-1,.6124195795e-1,-1.092460490,1.892197074,1.075289136,-.4881593780,.8455168449,-1.091557553,-.2385311826,.4131481275,.4820493399,-1.668760973,2.890378790,-1.548163659,-1.041619966},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.780140598,1.351243152,-1.744447408,2.064058009,-.5220598641,.9042342088,-1.167361344,1.368139104,-2.369686441,-1.070141386,-1.351243152,2.340421793,-3.021471542,3.575053341,-.2738823623e-1,.4743781670e-1,-.6124195801e-1,1.092460491,-1.892197076,-1.075289137,-.4881593725,.8455168354,-1.091557541,-.2385311799,.4131481229,.4820493345,1.668760962,-2.890378772,1.548163649,-1.041620004},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.560281195,-2.702486303,3.488894813,-4.128116015,-.5536850767,.9590106839,-1.238077470,-1.785522687,3.092616012,-.1767011503,-.3464101616e-9,.6e-9,-.7745966692e-9,.9165151390e-9,-.8344854191e-9,.1445371145e-8,-.1865966124e-8,-.6239040877e-9,.1080633579e-8,.3096646864e-9,-.4528013083,.7842748716,-1.012494506,-1.648891590,2.855964011,-2.444365371,-.7483314774e-8,.1296148140e-7,-.3631457412e-7,.9725382146e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.283894526,2.223770551,-2.870875435,3.396865625,.4596200614,-.7960852982,1.027741701,.5904589114,-1.022704835,-1.343510775,2.223770553,-3.851683580,4.972502121,-5.883543854,-1.098259750,1.902241688,-2.455783459,-.4185010573,.7248650942,1.591483172,1.261805030,-2.185510421,2.821481821,-.2397411885,.4152439192,-.6463928632,-.6392709500,1.107249765,-.5300284972e-1,-2.181654023},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.283894526,2.223770551,-2.870875435,3.396865625,.4596200614,-.7960852982,1.027741701,.5904589114,-1.022704835,-1.343510775,-2.223770551,3.851683579,-4.972502117,5.883543850,1.098259750,-1.902241687,2.455783458,.4185010573,-.7248650939,-1.591483171,1.261805025,-2.185510412,2.821481811,-.2397411877,.4152439177,-.6463928607,.6392709245,-1.107249721,.5300284761e-1,-2.181654030},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.567789052,-4.447541102,5.741750871,-6.793731251,1.727781187,-2.992604800,3.863436185,-.6176495990,1.069800487,-2.390029691,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.1560422168,.2702730475,-.3489210041,-.7800237319,1.351040735,-1.950823367,0.,-0.,0.,.1154979035e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.805810311,3.127755208,-4.037914609,4.777724998,1.840927032,-3.188579152,4.116437986,-1.585266239,2.745761671,1.099742715,3.127755210,-5.417430937,6.993873264,-8.275262446,-3.210799769,5.561268334,-7.179566546,2.802690869,-4.854402980,-2.005254469,4.133650004,-7.159691827,9.243122405,-3.691753629,6.394304856,2.774298724,4.281181012,-7.415223028,-3.392270752,3.547918793},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.805810311,3.127755208,-4.037914609,4.777724998,1.840927032,-3.188579152,4.116437986,-1.585266239,2.745761671,1.099742715,-3.127755208,5.417430934,-6.993873261,8.275262441,3.210799767,-5.561268332,7.179566543,-2.802690867,4.854402979,2.005254469,4.133650001,-7.159691821,9.243122398,-3.691753626,6.394304850,2.774298722,-4.281181016,7.415223034,3.392270755,3.547919252},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,3.611620623,-6.255510417,8.075829218,-9.555449995,5.548439256,-9.610178692,12.40668735,6.505405380,-11.26769265,6.079065903,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.1147467860e-1,.1987472634e-1,-.2565816138e-1,-.7349529016e-1,.1272975767,-.2596251910,-.5986651819e-8,.1036918512e-7,-.4964694369e-7,.6246716172e-4},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.139723890,-.2420088766,.3124321161,-.3696746651,-1.423600127,2.465747748,-3.183266656,.6044092113,-1.046867463,1.088285796,-1.881087112,3.258138451,-4.206238652,4.976888691,-1.804153808,3.124886061,-4.034210557,1.619309591,-2.804726483,1.142629197,1.289024480,-2.232655892,2.882346363,2.227214966,-3.857649481,-.2121957890,-.4725155092,.8184208692,-1.167298565,-.1238885768},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.559207282,-2.700626232,3.486493472,-4.125274710,-.5553799448,.9619462815,-1.241867310,-1.784795277,3.091356102,-.1734310073,1.061547995,-1.838655062,2.373693478,-2.808592000,2.556054550,-4.427216349,5.715511728,1.908436179,-3.305508424,-.9519572496,.3183248011,-.5513547288,.7117958945,1.158730221,-2.006979616,1.716453276,-.3086735638e-1,.5346382954e-1,-.1497356346,-.8941606665e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.698931170,2.942635106,-3.798925584,4.494949370,1.527877431,-2.646361336,3.416437796,-1.007511332,1.745060817,.2702851370,.8195391220,-1.419483395,1.832545187,-2.168296706,-.7519007415,1.302330287,-1.681301170,.5224058015,-.9048333903,-.1906718870,-2.010827754,3.482855835,-4.496347549,1.506342189,-2.609061204,-.7470992591,-2.148063678,3.720555426,1.317034145,1.229138747},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.698931170,2.942635106,-3.798925584,4.494949370,1.527877431,-2.646361336,3.416437796,-1.007511332,1.745060817,.2702851370,-.8195391179,1.419483391,-1.832545177,2.168296695,.7519007376,-1.302330280,1.681301162,-.5224057986,.9048333851,.1906718861,-2.010827749,3.482855828,-4.496347539,1.506342185,-2.609061198,-.7470992577,2.148063733,-3.720555523,-1.317034179,1.229138974},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.139723890,-.2420088766,.3124321161,-.3696746651,-1.423600127,2.465747748,-3.183266656,.6044092113,-1.046867463,1.088285796,1.881087116,-3.258138455,4.206238662,-4.976888702,1.804153811,-3.124886068,4.034210565,-1.619309594,2.804726489,-1.142629199,1.289024475,-2.232655882,2.882346350,2.227214957,-3.857649465,-.2121957881,.4725155698,-.8184209742,1.167298714,-.1238887416},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.559207282,-2.700626232,3.486493472,-4.125274710,-.5553799448,.9619462815,-1.241867310,-1.784795277,3.091356102,-.1734310073,-1.061547993,1.838655058,-2.373693472,2.808591992,-2.556054543,4.427216337,-5.715511713,-1.908436174,3.305508415,.9519572471,.3183248000,-.5513547270,.7117958920,1.158730217,-2.006979608,1.716453270,.3086736086e-1,-.5346383732e-1,.1497356563,-.8941604766e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.971066118,1.681935855,-2.171369850,2.569199455,-.1906367772,.3301925839,-.4262767929,1.203401099,-2.084351846,-1.362876347,-2.724401559,4.718801920,-6.091947082,7.208088995,.4756631459,-.8238727362,1.063615129,1.696386578,-2.938227741,-2.388976129,3.201877769,-5.545814974,7.159616348,.7410054061,-1.283459012,-2.832011274,-3.040858372,5.266921197,-1.847444888,2.450578880},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.844934019,-4.927570268,6.361465859,-7.526987913,2.597617435,-4.499205374,5.808449163,.5616969466,-.9728876503,-1.873883554,.5212328525,-.9028017828,1.165512090,-1.379052502,1.939035835,-3.358508584,4.335815938,3.687840215,-6.387526618,4.461246382,.8451479178e-1,-.1463839133,.1889808195,.4540265923,-.7863971259,1.260587466,.1137306696e-2,-.1969872981e-2,.7969298371e-2,-.3689094461e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.873867901,3.245634412,-4.190096007,4.957788456,2.048368944,-3.547879081,4.580292201,-1.991866711,3.450014347,1.736399667,2.203168706,-3.816000135,4.926434990,-5.829036490,-2.414698978,4.182381316,-5.399431060,2.358674124,-4.085343421,-2.072270251,.6985932727,-1.209999042,1.562102047,-.6879618021,1.191584795,.6128789301,-2.029197433,3.514673052,1.839475646,-3.732912624},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.873867901,3.245634412,-4.190096007,4.957788456,2.048368944,-3.547879081,4.580292201,-1.991866711,3.450014347,1.736399667,-2.203168707,3.816000137,-4.926434994,5.829036495,2.414698980,-4.182381320,5.399431066,-2.358674126,4.085343424,2.072270252,.6985932694,-1.209999036,1.562102039,-.6879617990,1.191584790,.6128789271,2.029197438,-3.514673060,-1.839475650,-3.732912697},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.971066118,1.681935855,-2.171369850,2.569199455,-.1906367772,.3301925839,-.4262767929,1.203401099,-2.084351846,-1.362876347,2.724401559,-4.718801918,6.091947082,-7.208088995,-.4756631459,.8238727362,-1.063615129,-1.696386578,2.938227741,2.388976128,3.201877775,-5.545814986,7.159616362,.7410054080,-1.283459015,-2.832011281,3.040858346,-5.266921153,1.847444873,2.450579118},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.844934019,-4.927570268,6.361465859,-7.526987913,2.597617435,-4.499205374,5.808449163,.5616969466,-.9728876503,-1.873883554,-.5212328521,.9028017822,-1.165512089,1.379052502,-1.939035833,3.358508582,-4.335815935,-3.687840212,6.387526614,-4.461246379,.8451479167e-1,-.1463839132,.1889808193,.4540265917,-.7863971248,1.260587465,-.1137294723e-2,.1969852243e-2,-.7969214472e-2,-.3689141724e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 27; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_dofs_to_top_face
	 *********************************************************/
public:
	static inline void madd_timestep_dofs_to_top_face(
		const T i_input[35],
		T o_output[27]
	) {
		static const T matrix[27][35] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.207666290,-2.091739373,2.700423918,-3.195184670,-1.025843012,1.776812217,-2.293854709,-1.398249481,2.421839143,.7821963046,-2.091739372,3.622998868,-4.677271425,5.534222184,-4.286111665,7.423763173,-9.584037044,-1.735425509,3.005845154,3.129410556,2.402465932,-4.161193056,5.372077139,7.607430278,-13.17645576,8.483868382,-2.650906245,4.591504302,-11.29653454,2.875267208},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.207666290,-2.091739373,2.700423918,-3.195184670,-1.025843012,1.776812217,-2.293854709,-1.398249481,2.421839143,.7821963046,2.091739375,-3.622998872,4.677271433,-5.534222193,4.286111669,-7.423763181,9.584037053,1.735425510,-3.005845155,-3.129410559,2.402465935,-4.161193062,5.372077144,7.607430290,-13.17645578,8.483868396,2.650906248,-4.591504307,11.29653455,2.875267158},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-2.415332579,4.183478747,-5.400847836,6.390369340,3.923332432,-6.795411104,8.772838015,-6.407991400,11.09896668,10.56920547,.1593486743e-8,-.276e-8,.3563144678e-8,-.4215969639e-8,-.2627089028e-8,.4550251674e-8,-.5874349651e-8,.4352222914e-8,-.7538271208e-8,-.7266379083e-8,-3.130880433,5.422843982,-7.000861478,5.367005593,-9.295926372,-9.215865385,-.1571496103e-7,.2721911093e-7,.2851586381e-7,4.649692528},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.393815818,-.6821090058,.8805989395,-1.041938717,-1.469667825,2.545539342,-3.286277160,.1253720207,-.2171507097,1.416680365,-.6821090051,1.181447453,-1.525242103,1.804690794,-.8311042347,1.439514761,-1.858405566,.4291959607,-.7433892102,.7095745382,-.6707402935,1.161756267,-1.499820892,-1.388519566,2.404986437,-.3003409923,.6556076472,-1.135545755,1.898978815,.2835315068e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.393815818,-.6821090058,.8805989395,-1.041938717,-1.469667825,2.545539342,-3.286277160,.1253720207,-.2171507097,1.416680365,.6821090058,-1.181447454,1.525242105,-1.804690796,.8311042356,-1.439514763,1.858405568,-.4291959612,.7433892109,-.7095745387,-.6707402924,1.161756265,-1.499820889,-1.388519564,2.404986432,-.3003409919,-.6556076472,1.135545755,-1.898978815,.2835315542e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.787631635,1.364218011,-1.761197878,2.083877432,-.5099913165,.8833308713,-1.140375252,1.364354035,-2.363130509,-1.084617593,.3464101616e-9,-.6e-9,.7745966692e-9,-.9165151390e-9,.4372877027e-11,-.7574045189e-11,.9778050292e-11,-.2779346762e-9,.4813969803e-9,.2775108880e-9,-1.743691247,3.020161833,-3.899012161,-.8344299601,1.445275086,1.717913014,0.,-0.,0.,1.442216374},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.780140598,1.351243152,-1.744447408,2.064058009,-.5220598641,.9042342088,-1.167361344,1.368139104,-2.369686441,-1.070141386,1.351243151,-2.340421790,3.021471539,-3.575053338,.2738823620e-1,-.4743781665e-1,.6124195795e-1,-1.092460490,1.892197074,1.075289136,-.4881593780,.8455168449,-1.091557553,-.2385311826,.4131481275,.4820493399,-1.668760973,2.890378790,-1.548163659,-1.041619966},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.780140598,1.351243152,-1.744447408,2.064058009,-.5220598641,.9042342088,-1.167361344,1.368139104,-2.369686441,-1.070141386,-1.351243152,2.340421793,-3.021471542,3.575053341,-.2738823623e-1,.4743781670e-1,-.6124195801e-1,1.092460491,-1.892197076,-1.075289137,-.4881593725,.8455168354,-1.091557541,-.2385311799,.4131481229,.4820493345,1.668760962,-2.890378772,1.548163649,-1.041620004},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.560281195,-2.702486303,3.488894813,-4.128116015,-.5536850767,.9590106839,-1.238077470,-1.785522687,3.092616012,-.1767011503,-.3464101616e-9,.6e-9,-.7745966692e-9,.9165151390e-9,-.8344854191e-9,.1445371145e-8,-.1865966124e-8,-.6239040877e-9,.1080633579e-8,.3096646864e-9,-.4528013083,.7842748716,-1.012494506,-1.648891590,2.855964011,-2.444365371,-.7483314774e-8,.1296148140e-7,-.3631457412e-7,.9725382146e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.283894526,2.223770551,-2.870875435,3.396865625,.4596200614,-.7960852982,1.027741701,.5904589114,-1.022704835,-1.343510775,2.223770553,-3.851683580,4.972502121,-5.883543854,-1.098259750,1.902241688,-2.455783459,-.4185010573,.7248650942,1.591483172,1.261805030,-2.185510421,2.821481821,-.2397411885,.4152439192,-.6463928632,-.6392709500,1.107249765,-.5300284972e-1,-2.181654023},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.283894526,2.223770551,-2.870875435,3.396865625,.4596200614,-.7960852982,1.027741701,.5904589114,-1.022704835,-1.343510775,-2.223770551,3.851683579,-4.972502117,5.883543850,1.098259750,-1.902241687,2.455783458,.4185010573,-.7248650939,-1.591483171,1.261805025,-2.185510412,2.821481811,-.2397411877,.4152439177,-.6463928607,.6392709245,-1.107249721,.5300284761e-1,-2.181654030},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.567789052,-4.447541102,5.741750871,-6.793731251,1.727781187,-2.992604800,3.863436185,-.6176495990,1.069800487,-2.390029691,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.1560422168,.2702730475,-.3489210041,-.7800237319,1.351040735,-1.950823367,0.,-0.,0.,.1154979035e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.805810311,3.127755208,-4.037914609,4.777724998,1.840927032,-3.188579152,4.116437986,-1.585266239,2.745761671,1.099742715,3.127755210,-5.417430937,6.993873264,-8.275262446,-3.210799769,5.561268334,-7.179566546,2.802690869,-4.854402980,-2.005254469,4.133650004,-7.159691827,9.243122405,-3.691753629,6.394304856,2.774298724,4.281181012,-7.415223028,-3.392270752,3.547918793},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.805810311,3.127755208,-4.037914609,4.777724998,1.840927032,-3.188579152,4.116437986,-1.585266239,2.745761671,1.099742715,-3.127755208,5.417430934,-6.993873261,8.275262441,3.210799767,-5.561268332,7.179566543,-2.802690867,4.854402979,2.005254469,4.133650001,-7.159691821,9.243122398,-3.691753626,6.394304850,2.774298722,-4.281181016,7.415223034,3.392270755,3.547919252},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,3.611620623,-6.255510417,8.075829218,-9.555449995,5.548439256,-9.610178692,12.40668735,6.505405380,-11.26769265,6.079065903,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.1147467860e-1,.1987472634e-1,-.2565816138e-1,-.7349529016e-1,.1272975767,-.2596251910,-.5986651819e-8,.1036918512e-7,-.4964694369e-7,.6246716172e-4},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.139723890,-.2420088766,.3124321161,-.3696746651,-1.423600127,2.465747748,-3.183266656,.6044092113,-1.046867463,1.088285796,-1.881087112,3.258138451,-4.206238652,4.976888691,-1.804153808,3.124886061,-4.034210557,1.619309591,-2.804726483,1.142629197,1.289024480,-2.232655892,2.882346363,2.227214966,-3.857649481,-.2121957890,-.4725155092,.8184208692,-1.167298565,-.1238885768},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.559207282,-2.700626232,3.486493472,-4.125274710,-.5553799448,.9619462815,-1.241867310,-1.784795277,3.091356102,-.1734310073,1.061547995,-1.838655062,2.373693478,-2.808592000,2.556054550,-4.427216349,5.715511728,1.908436179,-3.305508424,-.9519572496,.3183248011,-.5513547288,.7117958945,1.158730221,-2.006979616,1.716453276,-.3086735638e-1,.5346382954e-1,-.1497356346,-.8941606665e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.698931170,2.942635106,-3.798925584,4.494949370,1.527877431,-2.646361336,3.416437796,-1.007511332,1.745060817,.2702851370,.8195391220,-1.419483395,1.832545187,-2.168296706,-.7519007415,1.302330287,-1.681301170,.5224058015,-.9048333903,-.1906718870,-2.010827754,3.482855835,-4.496347549,1.506342189,-2.609061204,-.7470992591,-2.148063678,3.720555426,1.317034145,1.229138747},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.698931170,2.942635106,-3.798925584,4.494949370,1.527877431,-2.646361336,3.416437796,-1.007511332,1.745060817,.2702851370,-.8195391179,1.419483391,-1.832545177,2.168296695,.7519007376,-1.302330280,1.681301162,-.5224057986,.9048333851,.1906718861,-2.010827749,3.482855828,-4.496347539,1.506342185,-2.609061198,-.7470992577,2.148063733,-3.720555523,-1.317034179,1.229138974},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.139723890,-.2420088766,.3124321161,-.3696746651,-1.423600127,2.465747748,-3.183266656,.6044092113,-1.046867463,1.088285796,1.881087116,-3.258138455,4.206238662,-4.976888702,1.804153811,-3.124886068,4.034210565,-1.619309594,2.804726489,-1.142629199,1.289024475,-2.232655882,2.882346350,2.227214957,-3.857649465,-.2121957881,.4725155698,-.8184209742,1.167298714,-.1238887416},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.559207282,-2.700626232,3.486493472,-4.125274710,-.5553799448,.9619462815,-1.241867310,-1.784795277,3.091356102,-.1734310073,-1.061547993,1.838655058,-2.373693472,2.808591992,-2.556054543,4.427216337,-5.715511713,-1.908436174,3.305508415,.9519572471,.3183248000,-.5513547270,.7117958920,1.158730217,-2.006979608,1.716453270,.3086736086e-1,-.5346383732e-1,.1497356563,-.8941604766e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.971066118,1.681935855,-2.171369850,2.569199455,-.1906367772,.3301925839,-.4262767929,1.203401099,-2.084351846,-1.362876347,-2.724401559,4.718801920,-6.091947082,7.208088995,.4756631459,-.8238727362,1.063615129,1.696386578,-2.938227741,-2.388976129,3.201877769,-5.545814974,7.159616348,.7410054061,-1.283459012,-2.832011274,-3.040858372,5.266921197,-1.847444888,2.450578880},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.844934019,-4.927570268,6.361465859,-7.526987913,2.597617435,-4.499205374,5.808449163,.5616969466,-.9728876503,-1.873883554,.5212328525,-.9028017828,1.165512090,-1.379052502,1.939035835,-3.358508584,4.335815938,3.687840215,-6.387526618,4.461246382,.8451479178e-1,-.1463839133,.1889808195,.4540265923,-.7863971259,1.260587466,.1137306696e-2,-.1969872981e-2,.7969298371e-2,-.3689094461e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.873867901,3.245634412,-4.190096007,4.957788456,2.048368944,-3.547879081,4.580292201,-1.991866711,3.450014347,1.736399667,2.203168706,-3.816000135,4.926434990,-5.829036490,-2.414698978,4.182381316,-5.399431060,2.358674124,-4.085343421,-2.072270251,.6985932727,-1.209999042,1.562102047,-.6879618021,1.191584795,.6128789301,-2.029197433,3.514673052,1.839475646,-3.732912624},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.873867901,3.245634412,-4.190096007,4.957788456,2.048368944,-3.547879081,4.580292201,-1.991866711,3.450014347,1.736399667,-2.203168707,3.816000137,-4.926434994,5.829036495,2.414698980,-4.182381320,5.399431066,-2.358674126,4.085343424,2.072270252,.6985932694,-1.209999036,1.562102039,-.6879617990,1.191584790,.6128789271,2.029197438,-3.514673060,-1.839475650,-3.732912697},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.971066118,1.681935855,-2.171369850,2.569199455,-.1906367772,.3301925839,-.4262767929,1.203401099,-2.084351846,-1.362876347,2.724401559,-4.718801918,6.091947082,-7.208088995,-.4756631459,.8238727362,-1.063615129,-1.696386578,2.938227741,2.388976128,3.201877775,-5.545814986,7.159616362,.7410054080,-1.283459015,-2.832011281,3.040858346,-5.266921153,1.847444873,2.450579118},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.844934019,-4.927570268,6.361465859,-7.526987913,2.597617435,-4.499205374,5.808449163,.5616969466,-.9728876503,-1.873883554,-.5212328521,.9028017822,-1.165512089,1.379052502,-1.939035833,3.358508582,-4.335815935,-3.687840212,6.387526614,-4.461246379,.8451479167e-1,-.1463839132,.1889808193,.4540265917,-.7863971248,1.260587465,-.1137294723e-2,.1969852243e-2,-.7969214472e-2,-.3689141724e-2}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 27; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 35; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_bottom_fluxes_to_dofs(
		const T i_input[27],
		T o_output[35]
	) {
		static const T matrix[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-.2270692495e-2,-.2270692495e-2,-.2270692495e-2,-.1889769944,-.1889769944,-.1889769944,-.1453110246,-.1453110246,-.1453110246,-.8863366081e-1,-.8863366081e-1,-.8863366081e-1,-.3345937098e-1,-.3345937098e-1,-.3345937098e-1,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{-.3468540079e-2,-.3468540079e-2,-.3468540079e-2,-.2886671272,-.2886671272,-.2886671272,-.2219662565,-.2219662565,-.2219662565,-.1353901533,-.1353901533,-.1353901533,-.5111003340e-1,-.5111003340e-1,-.5111003340e-1,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1119514294e-2,-.1119514294e-2,.2239028587e-2,-.3038270720e-1,-.3038270720e-1,.6076541432e-1,.4628026305e-1,.4628026305e-1,-.9256052604e-1,.4645713347e-1,.4645713347e-1,-.9291426695e-1,.2466688309e-1,.2466688309e-1,-.4933376618e-1,-.7312744874e-2,-.816044061e-1,.8891715086e-1,.8891715086e-1,-.7312744874e-2,-.816044061e-1,.2010850666e-1,-.5891192536e-1,.3880341870e-1,.3880341870e-1,.2010850666e-1,-.5891192536e-1},
			{-.1939055637e-2,-.1939055637e-2,.3878111275e-2,-.5262439255e-1,-.5262439255e-1,.1052487850,.8015976696e-1,.8015976696e-1,-.1603195338,.8046611556e-1,.8046611556e-1,-.1609322311,.4272429479e-1,.4272429479e-1,-.8544858956e-1,-.1266604567e-1,-.1413429774,.1540090230,.1540090230,-.1266604567e-1,-.1413429774,.3482895521e-1,-.1020384480,.6720949271e-1,.6720949271e-1,.3482895521e-1,-.1020384480},
			{-.2503310063e-2,-.2503310063e-2,.5006620125e-2,-.6793779862e-1,-.6793779862e-1,.1358755971,.1034858141,.1034858141,-.2069716281,.1038813085,.1038813085,-.2077626169,.5515682737e-1,.5515682737e-1,-.1103136547,-.1635179464e-1,-.1824729992,.1988247936,.1988247936,-.1635179464e-1,-.1824729992,.4496398781e-1,-.1317310698,.8676708195e-1,.8676708195e-1,.4496398781e-1,-.1317310698},
			{-.2961956411e-2,-.2961956411e-2,.5923912822e-2,-.8038508741e-1,-.8038508741e-1,.1607701747,.1224460666,.1224460666,-.2448921330,.1229140218,.1229140218,-.2458280436,.6526243827e-1,.6526243827e-1,-.1305248765,-.1934770434e-1,-.2159049643,.2352526685,.2352526685,-.1934770434e-1,-.2159049643,.5320210786e-1,-.1558663038,.1026641959,.1026641959,.5320210786e-1,-.1558663038},
			{.9509629649e-3,.9509629649e-3,-.3636953995e-2,.1133841892,.1133841892,.3934559288e-1,.3097014551e-1,.3097014551e-1,.3284624728e-1,-.1663114073e-1,-.1663114073e-1,-.6251896835e-1,-.2514656806e-1,-.2514656806e-1,-.7579018773e-1,.7450711923e-1,.2906698235e-1,-.7996469216e-1,-.7996469216e-1,.7450711923e-1,.2906698235e-1,.3947641499e-2,-.5379057771e-1,-.4241692689e-1,-.4241692689e-1,.3947641499e-2,-.5379057771e-1},
			{.1647116171e-2,.1647116171e-2,-.6299389101e-2,.1963871764,.1963871764,.6814856589e-1,.5364186551e-1,.5364186551e-1,.5689136911e-1,-.2880598072e-1,-.2880598072e-1,-.1082860296,-.4355513348e-1,-.4355513348e-1,-.1312724558,.1290501160,.5034549024e-1,-.1385029096,-.1385029096,.1290501160,.5034549024e-1,.6837515642e-2,-.9316801354e-1,-.7346827246e-1,-.7346827246e-1,.6837515642e-2,-.9316801354e-1},
			{.2126417833e-2,.2126417833e-2,-.8132476364e-2,.2535347545,.2535347545,.8797942029e-1,.6925135063e-1,.6925135063e-1,.7344644172e-1,-.3718836120e-1,-.3718836120e-1,-.1397966631,-.5622943557e-1,-.5622943557e-1,-.1694720118,.1666029834,.6499574844e-1,-.1788064875,-.1788064875,.1666029834,.6499574844e-1,.8827194741e-2,-.1202793883,-.9484713192e-1,-.9484713192e-1,.8827194741e-2,-.1202793883},
			{.1296186118e-2,.1296186118e-2,.5940248584e-2,-.9672393084e-2,-.9672393084e-2,-.1052592792,-.8116208511e-1,-.8116208511e-1,.1059225219,-.2136548439e-1,-.2136548439e-1,.2234936694e-1,.2165431039e-1,.2165431039e-1,-.8886208754e-1,-.3163303256e-1,.9341103024e-1,.5273023340e-1,.5273023340e-1,-.3163303256e-1,.9341103024e-1,-.2491962037e-1,-.1163142919e-1,.4124689791e-1,.4124689791e-1,-.2491962037e-1,-.1163142919e-1},
			{.2245060214e-2,.2245060214e-2,.1028881236e-1,-.1675307626e-1,-.1675307626e-1,-.1823144197,-.1405768551,-.1405768551,.1834631897,-.3700610451e-1,-.3700610451e-1,.3871023906e-1,.3750636580e-1,.3750636580e-1,-.1539136505,-.5479001961e-1,.1617926504,.9133144337e-1,.9133144337e-1,-.5479001961e-1,.1617926504,-.4316204859e-1,-.2014622633e-1,.7144172285e-1,.7144172285e-1,-.4316204859e-1,-.2014622633e-1},
			{-.7251009248e-3,-.7251009248e-3,-.9797720361e-2,-.1092962313,-.1092962313,.8367774286e-1,.6348397322e-1,.6348397322e-1,.1048243834e-1,.4861432003e-1,.4861432003e-1,.8648212617e-1,-.1502218966e-1,-.1502218966e-1,-.8303840497e-1,-.5695773555e-1,.9076878046e-2,-.1414594348e-1,-.1414594348e-1,-.5695773555e-1,.9076878046e-2,.2822197953e-1,.3880374283e-1,-.3595677329e-1,-.3595677329e-1,.2822197953e-1,.3880374283e-1},
			{.1939055636e-2,-.1939055639e-2,-.1477172296e-11,.5262439250e-1,-.5262439255e-1,-.2672538285e-10,-.8015976689e-1,.8015976696e-1,.2055008219e-10,-.8046611563e-1,.8046611556e-1,-0.,-.4272429480e-1,.4272429479e-1,-0.,.9845066677e-1,-.5555835628e-1,-.4289231078e-1,.4289231057e-1,-.9845066699e-1,.5555835614e-1,.5641598021e-1,-.1079351250e-1,-.4562246768e-1,.4562246772e-1,-.5641598021e-1,.1079351250e-1},
			{.3358542881e-2,-.3358542884e-2,-.2558537468e-11,.9114812152e-1,-.9114812160e-1,-.4628972095e-10,-.1388407889,.1388407891,.3559378643e-10,-.1393714005,.1393714005,-0.,-.7400064931e-1,.7400064927e-1,-0.,.1705215569,-.9622989583e-1,-.742916614e-1,.7429166116e-1,-.1705215571,.962298956e-1,.9771534407e-1,-.1869491204e-1,-.7902043198e-1,.7902043202e-1,-.9771534403e-1,.1869491203e-1},
			{.4335860212e-2,-.4335860220e-2,-.3303057668e-11,.1176717189,-.1176717190,-.5975977276e-10,-.1792426878,.1792426879,.4595138070e-10,-.1799277044,.1799277042,-0.,-.9553442744e-1,.9553442741e-1,-0.,.2201423833,-.1242322613,-.9591012258e-1,.9591012213e-1,-.2201423837,.1242322610,.1261499667,-.2413502766e-1,-.1020149390,.1020149391,-.1261499667,.2413502765e-1},
			{.5130258989e-2,-.5130258998e-2,-.3908230539e-11,.1392310554,-.1392310556,-.7070871670e-10,-.2120828083,.2120828085,.5437040688e-10,-.2128933309,.2128933307,-0.,-.1130378590,.1130378589,-0.,.2604759806,-.1469935939,-.1134823874,.1134823869,-.2604759812,.1469935935,.1492626536,-.2855694985e-1,-.1207057037,.1207057037,-.1492626536,.2855694983e-1},
			{.3973252640e-2,-.3973252644e-2,.2435328156e-11,.6411930517e-1,-.6411930524e-1,-.3373654288e-12,-.1624751717e-2,.1624751718e-2,.4950415964e-10,.3974002441e-1,-.3974002439e-1,-0.,.4385866113e-1,-.4385866112e-1,-0.,.9442419982e-1,-.1337765132,.3935231328e-1,-.3935231309e-1,-.9442420005e-1,.1337765128,-.9849870528e-2,-.4015289410e-1,.5000276456e-1,-.5000276462e-1,.9849870528e-2,.4015289406e-1},
			{.6881875447e-2,-.6881875454e-2,.4218112101e-11,.1110578943,-.1110578945,-.5843340636e-12,-.2814152524e-2,.2814152527e-2,.8574371972e-10,.6883174139e-1,-.6883174137e-1,-0.,.7596542946e-1,-.7596542944e-1,-0.,.1635475116,-.2317077178,.6816020601e-1,-.6816020569e-1,-.1635475120,.2317077171,-.1706047621e-1,-.6954685268e-1,.8660732877e-1,-.8660732887e-1,.1706047621e-1,.6954685260e-1},
			{.8884462997e-2,-.8884463007e-2,.5445559306e-11,.1433751251,-.1433751252,-.7543720322e-12,-.3633055286e-2,.3633055289e-2,.1106946662e-9,.8886139602e-1,-.8886139599e-1,-0.,.9807094772e-1,-.9807094769e-1,-0.,.2111389296,-.2991333773,.8799444758e-1,-.8799444716e-1,-.2111389301,.2991333765,-.2202498007e-1,-.8978460072e-1,.1118095806,-.1118095808,.2202498007e-1,.8978460063e-1},
			{.1608750430e-2,-.1608750432e-2,-.4034538186e-11,-.3311226875e-1,.3311226880e-1,.2144253100e-10,.6480800897e-1,-.6480800902e-1,.3701184809e-10,.1514326846e-1,-.1514326846e-1,-0.,-.3828400335e-1,.3828400333e-1,-0.,-.8474998733e-1,-.9988203793e-1,-.2734121092e-1,.2734121078e-1,.8474998753e-1,.9988203767e-1,-.3512819587e-1,-.7636653997e-1,-.4884262097e-1,.4884262099e-1,.3512819587e-1,.7636653992e-1},
			{.2786437481e-2,-.2786437483e-2,-.6988025119e-11,-.5735213183e-1,.5735213188e-1,.3713955314e-10,.1122507642,-.1122507643,.6410640134e-10,.2622891035e-1,-.2622891035e-1,-0.,-.6630983887e-1,.6630983886e-1,-0.,-.1467912839,-.1730007645,-.4735636645e-1,.4735636618e-1,.1467912843,.1730007640,-.6084382000e-1,-.1322707272,-.8459790107e-1,.8459790113e-1,.6084382000e-1,.1322707271},
			{-.2900983391e-2,.2900983395e-2,.6735979401e-11,-.5474334561e-1,.5474334565e-1,-.2140983594e-10,-.6378935306e-1,.6378935317e-1,-.1837023118e-10,-.5758708731e-1,.5758708725e-1,-0.,.2739123664e-1,-.2739123663e-1,-0.,-.5980191219e-1,.4982269315e-1,.9979215896e-2,-.9979215849e-2,.5980191234e-1,-.4982269302e-1,.4947010457e-1,-.9238197165e-1,.4291186704e-1,-.4291186706e-1,-.4947010455e-1,.9238197160e-1},
			{-.2227101124e-2,-.2227101127e-2,.2902345977e-2,.5174730169e-1,.5174730161e-1,.1345249687,.2895906774e-1,.2895906742e-1,.2686152177e-1,-.4565783520e-1,-.4565783503e-1,.5646315914e-2,-.5646454713e-1,-.5646454707e-1,.1567410230e-3,-.6746381852e-1,-.1666020075e-1,.1052409173,.1052409171,-.6746381827e-1,-.1666020070e-1,-.6630339505e-1,-.1750103542e-2,-.1446623172e-1,-.1446623165e-1,-.6630339516e-1,-.1750103540e-2},
			{-.3857452299e-2,-.3857452305e-2,.5027010692e-2,.8962895569e-1,.8962895553e-1,.2330040807,.5015857668e-1,.5015857611e-1,.4652552046e-1,-.7908169034e-1,-.7908169002e-1,.9779706035e-2,-.9779946441e-1,-.9779946432e-1,.2714834155e-3,-.1168507614,-.2885631417e-1,.1822826158,.1822826154,-.1168507609,-.2885631408e-1,-.1148408489,-.3031268254e-2,-.2505624833e-1,-.2505624821e-1,-.1148408492,-.3031268250e-2},
			{-.4979949508e-2,-.4979949513e-2,.6489842900e-2,.1157104843,.1157104841,.3008069748,.6475444405e-1,.6475444331e-1,.6006418864e-1,-.1020940232,-.1020940229,.1262554620e-1,-.1262585657,-.1262585656,.3504835826e-3,-.1508536843,-.3725334142e-1,.2353258451,.2353258446,-.1508536837,-.3725334129e-1,-.1482588985,-.3913350489e-2,-.3234747750e-1,-.3234747736e-1,-.1482588988,-.3913350484e-2},
			{-.7052136016e-2,-.7052136025e-2,-.4975248153e-2,.1071236388,.1071236386,.6437588332e-1,.1415037995e-1,.1415037979e-1,.9781715849e-1,.8674924746e-2,.8674924715e-2,.2822480032e-1,.5042836149e-1,.5042836146e-1,.1003925893e-2,-.1165659990,-.6064459330e-1,-.7883759978e-1,-.7883759957e-1,-.1165659985,-.6064459310e-1,-.1534448774e-1,-.9401828142e-2,.1424607884e-1,.1424607878e-1,-.1534448778e-1,-.9401828126e-2},
			{-.1221465788e-1,-.1221465790e-1,-.8617382582e-2,.1855435851,.1855435847,.1115023007,.2450917702e-1,.2450917675e-1,.1694242884,.1502541041e-1,.1502541036e-1,.4888678819e-1,.8734448425e-1,.8734448420e-1,.1738850654e-2,-.2018982328,-.1050395168,-.1365507284,-.1365507280,-.2018982319,-.1050395164,-.2657743238e-1,-.1628444403e-1,.2467493236e-1,.2467493225e-1,-.2657743245e-1,-.1628444400e-1},
			{-.7864599685e-2,-.7864599698e-2,.8543165539e-2,.2317116788e-1,.2317116784e-1,-.1325361900,-.2859660209e-1,-.2859660176e-1,.1450070316,.2338942873e-1,.2338942865e-1,.7058964713e-1,-.3789617428e-1,-.3789617426e-1,.3546410271e-2,.1110571477e-1,-.8983420728e-1,.3910101757e-1,.3910101750e-1,.1110571472e-1,-.8983420697e-1,.5864432557e-1,-.2610381621e-1,-.1269128828e-1,-.1269128822e-1,.5864432571e-1,-.2610381617e-1},
			{.2457406867e-2,-.2457406870e-2,.1456786833e-10,-.5057982506e-1,.5057982506e-1,-0.,.9899586944e-1,-.9899586881e-1,.4439325132e-9,.2313172558e-1,-.2313172466e-1,-0.,-.5847978098e-1,.5847978105e-1,.8177605344e-10,.2473009711e-1,.1615508286e-2,.1124234492,-.1124234521,-.2473010027e-1,-.1615508521e-2,.6296906016e-1,-.2355095997e-4,.4201992979e-1,-.4201992986e-1,-.6296905964e-1,.2355071203e-4},
			{.4256353547e-2,-.4256353552e-2,.2523228810e-10,-.8760682683e-1,.8760682683e-1,-0.,.1714658756,-.1714658745,.7689136679e-9,.4006532397e-1,-.4006532237e-1,-0.,-.1012899519,.1012899520,.1416402794e-9,.4283378466e-1,.2798142432e-2,.1947231258,-.1947231310,-.4283379015e-1,-.2798142839e-2,.1090656115,-.4079145923e-4,.7278065332e-1,-.7278065345e-1,-.1090656106,.4079102978e-4},
			{.1047195901e-1,-.1047195902e-1,-.2643438623e-10,-.1465053324,.1465053324,-0.,.9184167768e-1,-.9184167710e-1,.2154288659e-8,.1917883762e-2,-.1917883686e-2,-0.,.4633750594e-1,-.4633750599e-1,.6781638958e-9,.6109303566e-1,.7836730672e-2,-.6892976350e-1,.6892976527e-1,-.6109304346e-1,-.7836731808e-2,.3825625994e-1,-.1650255183e-3,-.3809123559e-1,.3809123567e-1,-.3825625962e-1,.1650237808e-3},
			{-.2665390899e-2,-.2665390853e-2,-.4310294403e-2,-.2187432388e-2,-.2187432754e-2,-.1112663225,.6179199770e-1,.6179199994e-1,-.5769386249e-2,.7894214835e-1,.7894214861e-1,-.4179238568e-3,-.4846361631e-1,-.4846362257e-1,-.8532846255e-6,.6483970316e-2,.4679778692e-2,-.6432957221e-1,-.6432958410e-1,.6483978942e-2,.4679777698e-2,-.5074575336e-1,.7639251247e-4,.7729988409e-1,.7729988561e-1,-.5074575829e-1,.7639349119e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 27; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_bottom_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_bottom_fluxes_to_dofs(
		const T i_input[27],
		T o_output[35]
	) {
		static const T matrix[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-.2270692495e-2,-.2270692495e-2,-.2270692495e-2,-.1889769944,-.1889769944,-.1889769944,-.1453110246,-.1453110246,-.1453110246,-.8863366081e-1,-.8863366081e-1,-.8863366081e-1,-.3345937098e-1,-.3345937098e-1,-.3345937098e-1,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{-.3468540079e-2,-.3468540079e-2,-.3468540079e-2,-.2886671272,-.2886671272,-.2886671272,-.2219662565,-.2219662565,-.2219662565,-.1353901533,-.1353901533,-.1353901533,-.5111003340e-1,-.5111003340e-1,-.5111003340e-1,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1119514294e-2,-.1119514294e-2,.2239028587e-2,-.3038270720e-1,-.3038270720e-1,.6076541432e-1,.4628026305e-1,.4628026305e-1,-.9256052604e-1,.4645713347e-1,.4645713347e-1,-.9291426695e-1,.2466688309e-1,.2466688309e-1,-.4933376618e-1,-.7312744874e-2,-.816044061e-1,.8891715086e-1,.8891715086e-1,-.7312744874e-2,-.816044061e-1,.2010850666e-1,-.5891192536e-1,.3880341870e-1,.3880341870e-1,.2010850666e-1,-.5891192536e-1},
			{-.1939055637e-2,-.1939055637e-2,.3878111275e-2,-.5262439255e-1,-.5262439255e-1,.1052487850,.8015976696e-1,.8015976696e-1,-.1603195338,.8046611556e-1,.8046611556e-1,-.1609322311,.4272429479e-1,.4272429479e-1,-.8544858956e-1,-.1266604567e-1,-.1413429774,.1540090230,.1540090230,-.1266604567e-1,-.1413429774,.3482895521e-1,-.1020384480,.6720949271e-1,.6720949271e-1,.3482895521e-1,-.1020384480},
			{-.2503310063e-2,-.2503310063e-2,.5006620125e-2,-.6793779862e-1,-.6793779862e-1,.1358755971,.1034858141,.1034858141,-.2069716281,.1038813085,.1038813085,-.2077626169,.5515682737e-1,.5515682737e-1,-.1103136547,-.1635179464e-1,-.1824729992,.1988247936,.1988247936,-.1635179464e-1,-.1824729992,.4496398781e-1,-.1317310698,.8676708195e-1,.8676708195e-1,.4496398781e-1,-.1317310698},
			{-.2961956411e-2,-.2961956411e-2,.5923912822e-2,-.8038508741e-1,-.8038508741e-1,.1607701747,.1224460666,.1224460666,-.2448921330,.1229140218,.1229140218,-.2458280436,.6526243827e-1,.6526243827e-1,-.1305248765,-.1934770434e-1,-.2159049643,.2352526685,.2352526685,-.1934770434e-1,-.2159049643,.5320210786e-1,-.1558663038,.1026641959,.1026641959,.5320210786e-1,-.1558663038},
			{.9509629649e-3,.9509629649e-3,-.3636953995e-2,.1133841892,.1133841892,.3934559288e-1,.3097014551e-1,.3097014551e-1,.3284624728e-1,-.1663114073e-1,-.1663114073e-1,-.6251896835e-1,-.2514656806e-1,-.2514656806e-1,-.7579018773e-1,.7450711923e-1,.2906698235e-1,-.7996469216e-1,-.7996469216e-1,.7450711923e-1,.2906698235e-1,.3947641499e-2,-.5379057771e-1,-.4241692689e-1,-.4241692689e-1,.3947641499e-2,-.5379057771e-1},
			{.1647116171e-2,.1647116171e-2,-.6299389101e-2,.1963871764,.1963871764,.6814856589e-1,.5364186551e-1,.5364186551e-1,.5689136911e-1,-.2880598072e-1,-.2880598072e-1,-.1082860296,-.4355513348e-1,-.4355513348e-1,-.1312724558,.1290501160,.5034549024e-1,-.1385029096,-.1385029096,.1290501160,.5034549024e-1,.6837515642e-2,-.9316801354e-1,-.7346827246e-1,-.7346827246e-1,.6837515642e-2,-.9316801354e-1},
			{.2126417833e-2,.2126417833e-2,-.8132476364e-2,.2535347545,.2535347545,.8797942029e-1,.6925135063e-1,.6925135063e-1,.7344644172e-1,-.3718836120e-1,-.3718836120e-1,-.1397966631,-.5622943557e-1,-.5622943557e-1,-.1694720118,.1666029834,.6499574844e-1,-.1788064875,-.1788064875,.1666029834,.6499574844e-1,.8827194741e-2,-.1202793883,-.9484713192e-1,-.9484713192e-1,.8827194741e-2,-.1202793883},
			{.1296186118e-2,.1296186118e-2,.5940248584e-2,-.9672393084e-2,-.9672393084e-2,-.1052592792,-.8116208511e-1,-.8116208511e-1,.1059225219,-.2136548439e-1,-.2136548439e-1,.2234936694e-1,.2165431039e-1,.2165431039e-1,-.8886208754e-1,-.3163303256e-1,.9341103024e-1,.5273023340e-1,.5273023340e-1,-.3163303256e-1,.9341103024e-1,-.2491962037e-1,-.1163142919e-1,.4124689791e-1,.4124689791e-1,-.2491962037e-1,-.1163142919e-1},
			{.2245060214e-2,.2245060214e-2,.1028881236e-1,-.1675307626e-1,-.1675307626e-1,-.1823144197,-.1405768551,-.1405768551,.1834631897,-.3700610451e-1,-.3700610451e-1,.3871023906e-1,.3750636580e-1,.3750636580e-1,-.1539136505,-.5479001961e-1,.1617926504,.9133144337e-1,.9133144337e-1,-.5479001961e-1,.1617926504,-.4316204859e-1,-.2014622633e-1,.7144172285e-1,.7144172285e-1,-.4316204859e-1,-.2014622633e-1},
			{-.7251009248e-3,-.7251009248e-3,-.9797720361e-2,-.1092962313,-.1092962313,.8367774286e-1,.6348397322e-1,.6348397322e-1,.1048243834e-1,.4861432003e-1,.4861432003e-1,.8648212617e-1,-.1502218966e-1,-.1502218966e-1,-.8303840497e-1,-.5695773555e-1,.9076878046e-2,-.1414594348e-1,-.1414594348e-1,-.5695773555e-1,.9076878046e-2,.2822197953e-1,.3880374283e-1,-.3595677329e-1,-.3595677329e-1,.2822197953e-1,.3880374283e-1},
			{.1939055636e-2,-.1939055639e-2,-.1477172296e-11,.5262439250e-1,-.5262439255e-1,-.2672538285e-10,-.8015976689e-1,.8015976696e-1,.2055008219e-10,-.8046611563e-1,.8046611556e-1,-0.,-.4272429480e-1,.4272429479e-1,-0.,.9845066677e-1,-.5555835628e-1,-.4289231078e-1,.4289231057e-1,-.9845066699e-1,.5555835614e-1,.5641598021e-1,-.1079351250e-1,-.4562246768e-1,.4562246772e-1,-.5641598021e-1,.1079351250e-1},
			{.3358542881e-2,-.3358542884e-2,-.2558537468e-11,.9114812152e-1,-.9114812160e-1,-.4628972095e-10,-.1388407889,.1388407891,.3559378643e-10,-.1393714005,.1393714005,-0.,-.7400064931e-1,.7400064927e-1,-0.,.1705215569,-.9622989583e-1,-.742916614e-1,.7429166116e-1,-.1705215571,.962298956e-1,.9771534407e-1,-.1869491204e-1,-.7902043198e-1,.7902043202e-1,-.9771534403e-1,.1869491203e-1},
			{.4335860212e-2,-.4335860220e-2,-.3303057668e-11,.1176717189,-.1176717190,-.5975977276e-10,-.1792426878,.1792426879,.4595138070e-10,-.1799277044,.1799277042,-0.,-.9553442744e-1,.9553442741e-1,-0.,.2201423833,-.1242322613,-.9591012258e-1,.9591012213e-1,-.2201423837,.1242322610,.1261499667,-.2413502766e-1,-.1020149390,.1020149391,-.1261499667,.2413502765e-1},
			{.5130258989e-2,-.5130258998e-2,-.3908230539e-11,.1392310554,-.1392310556,-.7070871670e-10,-.2120828083,.2120828085,.5437040688e-10,-.2128933309,.2128933307,-0.,-.1130378590,.1130378589,-0.,.2604759806,-.1469935939,-.1134823874,.1134823869,-.2604759812,.1469935935,.1492626536,-.2855694985e-1,-.1207057037,.1207057037,-.1492626536,.2855694983e-1},
			{.3973252640e-2,-.3973252644e-2,.2435328156e-11,.6411930517e-1,-.6411930524e-1,-.3373654288e-12,-.1624751717e-2,.1624751718e-2,.4950415964e-10,.3974002441e-1,-.3974002439e-1,-0.,.4385866113e-1,-.4385866112e-1,-0.,.9442419982e-1,-.1337765132,.3935231328e-1,-.3935231309e-1,-.9442420005e-1,.1337765128,-.9849870528e-2,-.4015289410e-1,.5000276456e-1,-.5000276462e-1,.9849870528e-2,.4015289406e-1},
			{.6881875447e-2,-.6881875454e-2,.4218112101e-11,.1110578943,-.1110578945,-.5843340636e-12,-.2814152524e-2,.2814152527e-2,.8574371972e-10,.6883174139e-1,-.6883174137e-1,-0.,.7596542946e-1,-.7596542944e-1,-0.,.1635475116,-.2317077178,.6816020601e-1,-.6816020569e-1,-.1635475120,.2317077171,-.1706047621e-1,-.6954685268e-1,.8660732877e-1,-.8660732887e-1,.1706047621e-1,.6954685260e-1},
			{.8884462997e-2,-.8884463007e-2,.5445559306e-11,.1433751251,-.1433751252,-.7543720322e-12,-.3633055286e-2,.3633055289e-2,.1106946662e-9,.8886139602e-1,-.8886139599e-1,-0.,.9807094772e-1,-.9807094769e-1,-0.,.2111389296,-.2991333773,.8799444758e-1,-.8799444716e-1,-.2111389301,.2991333765,-.2202498007e-1,-.8978460072e-1,.1118095806,-.1118095808,.2202498007e-1,.8978460063e-1},
			{.1608750430e-2,-.1608750432e-2,-.4034538186e-11,-.3311226875e-1,.3311226880e-1,.2144253100e-10,.6480800897e-1,-.6480800902e-1,.3701184809e-10,.1514326846e-1,-.1514326846e-1,-0.,-.3828400335e-1,.3828400333e-1,-0.,-.8474998733e-1,-.9988203793e-1,-.2734121092e-1,.2734121078e-1,.8474998753e-1,.9988203767e-1,-.3512819587e-1,-.7636653997e-1,-.4884262097e-1,.4884262099e-1,.3512819587e-1,.7636653992e-1},
			{.2786437481e-2,-.2786437483e-2,-.6988025119e-11,-.5735213183e-1,.5735213188e-1,.3713955314e-10,.1122507642,-.1122507643,.6410640134e-10,.2622891035e-1,-.2622891035e-1,-0.,-.6630983887e-1,.6630983886e-1,-0.,-.1467912839,-.1730007645,-.4735636645e-1,.4735636618e-1,.1467912843,.1730007640,-.6084382000e-1,-.1322707272,-.8459790107e-1,.8459790113e-1,.6084382000e-1,.1322707271},
			{-.2900983391e-2,.2900983395e-2,.6735979401e-11,-.5474334561e-1,.5474334565e-1,-.2140983594e-10,-.6378935306e-1,.6378935317e-1,-.1837023118e-10,-.5758708731e-1,.5758708725e-1,-0.,.2739123664e-1,-.2739123663e-1,-0.,-.5980191219e-1,.4982269315e-1,.9979215896e-2,-.9979215849e-2,.5980191234e-1,-.4982269302e-1,.4947010457e-1,-.9238197165e-1,.4291186704e-1,-.4291186706e-1,-.4947010455e-1,.9238197160e-1},
			{-.2227101124e-2,-.2227101127e-2,.2902345977e-2,.5174730169e-1,.5174730161e-1,.1345249687,.2895906774e-1,.2895906742e-1,.2686152177e-1,-.4565783520e-1,-.4565783503e-1,.5646315914e-2,-.5646454713e-1,-.5646454707e-1,.1567410230e-3,-.6746381852e-1,-.1666020075e-1,.1052409173,.1052409171,-.6746381827e-1,-.1666020070e-1,-.6630339505e-1,-.1750103542e-2,-.1446623172e-1,-.1446623165e-1,-.6630339516e-1,-.1750103540e-2},
			{-.3857452299e-2,-.3857452305e-2,.5027010692e-2,.8962895569e-1,.8962895553e-1,.2330040807,.5015857668e-1,.5015857611e-1,.4652552046e-1,-.7908169034e-1,-.7908169002e-1,.9779706035e-2,-.9779946441e-1,-.9779946432e-1,.2714834155e-3,-.1168507614,-.2885631417e-1,.1822826158,.1822826154,-.1168507609,-.2885631408e-1,-.1148408489,-.3031268254e-2,-.2505624833e-1,-.2505624821e-1,-.1148408492,-.3031268250e-2},
			{-.4979949508e-2,-.4979949513e-2,.6489842900e-2,.1157104843,.1157104841,.3008069748,.6475444405e-1,.6475444331e-1,.6006418864e-1,-.1020940232,-.1020940229,.1262554620e-1,-.1262585657,-.1262585656,.3504835826e-3,-.1508536843,-.3725334142e-1,.2353258451,.2353258446,-.1508536837,-.3725334129e-1,-.1482588985,-.3913350489e-2,-.3234747750e-1,-.3234747736e-1,-.1482588988,-.3913350484e-2},
			{-.7052136016e-2,-.7052136025e-2,-.4975248153e-2,.1071236388,.1071236386,.6437588332e-1,.1415037995e-1,.1415037979e-1,.9781715849e-1,.8674924746e-2,.8674924715e-2,.2822480032e-1,.5042836149e-1,.5042836146e-1,.1003925893e-2,-.1165659990,-.6064459330e-1,-.7883759978e-1,-.7883759957e-1,-.1165659985,-.6064459310e-1,-.1534448774e-1,-.9401828142e-2,.1424607884e-1,.1424607878e-1,-.1534448778e-1,-.9401828126e-2},
			{-.1221465788e-1,-.1221465790e-1,-.8617382582e-2,.1855435851,.1855435847,.1115023007,.2450917702e-1,.2450917675e-1,.1694242884,.1502541041e-1,.1502541036e-1,.4888678819e-1,.8734448425e-1,.8734448420e-1,.1738850654e-2,-.2018982328,-.1050395168,-.1365507284,-.1365507280,-.2018982319,-.1050395164,-.2657743238e-1,-.1628444403e-1,.2467493236e-1,.2467493225e-1,-.2657743245e-1,-.1628444400e-1},
			{-.7864599685e-2,-.7864599698e-2,.8543165539e-2,.2317116788e-1,.2317116784e-1,-.1325361900,-.2859660209e-1,-.2859660176e-1,.1450070316,.2338942873e-1,.2338942865e-1,.7058964713e-1,-.3789617428e-1,-.3789617426e-1,.3546410271e-2,.1110571477e-1,-.8983420728e-1,.3910101757e-1,.3910101750e-1,.1110571472e-1,-.8983420697e-1,.5864432557e-1,-.2610381621e-1,-.1269128828e-1,-.1269128822e-1,.5864432571e-1,-.2610381617e-1},
			{.2457406867e-2,-.2457406870e-2,.1456786833e-10,-.5057982506e-1,.5057982506e-1,-0.,.9899586944e-1,-.9899586881e-1,.4439325132e-9,.2313172558e-1,-.2313172466e-1,-0.,-.5847978098e-1,.5847978105e-1,.8177605344e-10,.2473009711e-1,.1615508286e-2,.1124234492,-.1124234521,-.2473010027e-1,-.1615508521e-2,.6296906016e-1,-.2355095997e-4,.4201992979e-1,-.4201992986e-1,-.6296905964e-1,.2355071203e-4},
			{.4256353547e-2,-.4256353552e-2,.2523228810e-10,-.8760682683e-1,.8760682683e-1,-0.,.1714658756,-.1714658745,.7689136679e-9,.4006532397e-1,-.4006532237e-1,-0.,-.1012899519,.1012899520,.1416402794e-9,.4283378466e-1,.2798142432e-2,.1947231258,-.1947231310,-.4283379015e-1,-.2798142839e-2,.1090656115,-.4079145923e-4,.7278065332e-1,-.7278065345e-1,-.1090656106,.4079102978e-4},
			{.1047195901e-1,-.1047195902e-1,-.2643438623e-10,-.1465053324,.1465053324,-0.,.9184167768e-1,-.9184167710e-1,.2154288659e-8,.1917883762e-2,-.1917883686e-2,-0.,.4633750594e-1,-.4633750599e-1,.6781638958e-9,.6109303566e-1,.7836730672e-2,-.6892976350e-1,.6892976527e-1,-.6109304346e-1,-.7836731808e-2,.3825625994e-1,-.1650255183e-3,-.3809123559e-1,.3809123567e-1,-.3825625962e-1,.1650237808e-3},
			{-.2665390899e-2,-.2665390853e-2,-.4310294403e-2,-.2187432388e-2,-.2187432754e-2,-.1112663225,.6179199770e-1,.6179199994e-1,-.5769386249e-2,.7894214835e-1,.7894214861e-1,-.4179238568e-3,-.4846361631e-1,-.4846362257e-1,-.8532846255e-6,.6483970316e-2,.4679778692e-2,-.6432957221e-1,-.6432958410e-1,.6483978942e-2,.4679777698e-2,-.5074575336e-1,.7639251247e-4,.7729988409e-1,.7729988561e-1,-.5074575829e-1,.7639349119e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 27; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}


	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void mul_timestep_top_fluxes_to_dofs(
		const T i_input[27],
		T o_output[35]
	) {
		static const T matrix[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{.2270692495e-2,.2270692495e-2,.2270692495e-2,.1889769944,.1889769944,.1889769944,.1453110246,.1453110246,.1453110246,.8863366081e-1,.8863366081e-1,.8863366081e-1,.3345937098e-1,.3345937098e-1,.3345937098e-1,.1281992189,.1281992189,.1281992189,.1281992189,.1281992189,.1281992189,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{.3468540079e-2,.3468540079e-2,.3468540079e-2,.2886671272,.2886671272,.2886671272,.2219662565,.2219662565,.2219662565,.1353901533,.1353901533,.1353901533,.5111003340e-1,.5111003340e-1,.5111003340e-1,.1958275416,.1958275416,.1958275416,.1958275416,.1958275416,.1958275416,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1119514294e-2,-.1119514294e-2,.2239028587e-2,-.3038270720e-1,-.3038270720e-1,.6076541432e-1,.4628026305e-1,.4628026305e-1,-.9256052604e-1,.4645713347e-1,.4645713347e-1,-.9291426695e-1,.2466688309e-1,.2466688309e-1,-.4933376618e-1,-.7312744874e-2,-.816044061e-1,.8891715086e-1,.8891715086e-1,-.7312744874e-2,-.816044061e-1,.2010850666e-1,-.5891192536e-1,.3880341870e-1,.3880341870e-1,.2010850666e-1,-.5891192536e-1},
			{.1939055637e-2,.1939055637e-2,-.3878111275e-2,.5262439255e-1,.5262439255e-1,-.1052487850,-.8015976696e-1,-.8015976696e-1,.1603195338,-.8046611556e-1,-.8046611556e-1,.1609322311,-.4272429479e-1,-.4272429479e-1,.8544858956e-1,.1266604567e-1,.1413429774,-.1540090230,-.1540090230,.1266604567e-1,.1413429774,-.3482895521e-1,.1020384480,-.6720949271e-1,-.6720949271e-1,-.3482895521e-1,.1020384480},
			{-.2503310063e-2,-.2503310063e-2,.5006620125e-2,-.6793779862e-1,-.6793779862e-1,.1358755971,.1034858141,.1034858141,-.2069716281,.1038813085,.1038813085,-.2077626169,.5515682737e-1,.5515682737e-1,-.1103136547,-.1635179464e-1,-.1824729992,.1988247936,.1988247936,-.1635179464e-1,-.1824729992,.4496398781e-1,-.1317310698,.8676708195e-1,.8676708195e-1,.4496398781e-1,-.1317310698},
			{.2961956411e-2,.2961956411e-2,-.5923912822e-2,.8038508741e-1,.8038508741e-1,-.1607701747,-.1224460666,-.1224460666,.2448921330,-.1229140218,-.1229140218,.2458280436,-.6526243827e-1,-.6526243827e-1,.1305248765,.1934770434e-1,.2159049643,-.2352526685,-.2352526685,.1934770434e-1,.2159049643,-.5320210786e-1,.1558663038,-.1026641959,-.1026641959,-.5320210786e-1,.1558663038},
			{.9509629649e-3,.9509629649e-3,-.3636953995e-2,.1133841892,.1133841892,.3934559288e-1,.3097014551e-1,.3097014551e-1,.3284624728e-1,-.1663114073e-1,-.1663114073e-1,-.6251896835e-1,-.2514656806e-1,-.2514656806e-1,-.7579018773e-1,.7450711923e-1,.2906698235e-1,-.7996469216e-1,-.7996469216e-1,.7450711923e-1,.2906698235e-1,.3947641499e-2,-.5379057771e-1,-.4241692689e-1,-.4241692689e-1,.3947641499e-2,-.5379057771e-1},
			{-.1647116171e-2,-.1647116171e-2,.6299389101e-2,-.1963871764,-.1963871764,-.6814856589e-1,-.5364186551e-1,-.5364186551e-1,-.5689136911e-1,.2880598072e-1,.2880598072e-1,.1082860296,.4355513348e-1,.4355513348e-1,.1312724558,-.1290501160,-.5034549024e-1,.1385029096,.1385029096,-.1290501160,-.5034549024e-1,-.6837515642e-2,.9316801354e-1,.7346827246e-1,.7346827246e-1,-.6837515642e-2,.9316801354e-1},
			{.2126417833e-2,.2126417833e-2,-.8132476364e-2,.2535347545,.2535347545,.8797942029e-1,.6925135063e-1,.6925135063e-1,.7344644172e-1,-.3718836120e-1,-.3718836120e-1,-.1397966631,-.5622943557e-1,-.5622943557e-1,-.1694720118,.1666029834,.6499574844e-1,-.1788064875,-.1788064875,.1666029834,.6499574844e-1,.8827194741e-2,-.1202793883,-.9484713192e-1,-.9484713192e-1,.8827194741e-2,-.1202793883},
			{.1296186118e-2,.1296186118e-2,.5940248584e-2,-.9672393084e-2,-.9672393084e-2,-.1052592792,-.8116208511e-1,-.8116208511e-1,.1059225219,-.2136548439e-1,-.2136548439e-1,.2234936694e-1,.2165431039e-1,.2165431039e-1,-.8886208754e-1,-.3163303256e-1,.9341103024e-1,.5273023340e-1,.5273023340e-1,-.3163303256e-1,.9341103024e-1,-.2491962037e-1,-.1163142919e-1,.4124689791e-1,.4124689791e-1,-.2491962037e-1,-.1163142919e-1},
			{-.2245060214e-2,-.2245060214e-2,-.1028881236e-1,.1675307626e-1,.1675307626e-1,.1823144197,.1405768551,.1405768551,-.1834631897,.3700610451e-1,.3700610451e-1,-.3871023906e-1,-.3750636580e-1,-.3750636580e-1,.1539136505,.5479001961e-1,-.1617926504,-.9133144337e-1,-.9133144337e-1,.5479001961e-1,-.1617926504,.4316204859e-1,.2014622633e-1,-.7144172285e-1,-.7144172285e-1,.4316204859e-1,.2014622633e-1},
			{-.7251009248e-3,-.7251009248e-3,-.9797720361e-2,-.1092962313,-.1092962313,.8367774286e-1,.6348397322e-1,.6348397322e-1,.1048243834e-1,.4861432003e-1,.4861432003e-1,.8648212617e-1,-.1502218966e-1,-.1502218966e-1,-.8303840497e-1,-.5695773555e-1,.9076878046e-2,-.1414594348e-1,-.1414594348e-1,-.5695773555e-1,.9076878046e-2,.2822197953e-1,.3880374283e-1,-.3595677329e-1,-.3595677329e-1,.2822197953e-1,.3880374283e-1},
			{.1939055636e-2,-.1939055639e-2,-.1477172296e-11,.5262439250e-1,-.5262439255e-1,-.2672538285e-10,-.8015976689e-1,.8015976696e-1,.2055008219e-10,-.8046611563e-1,.8046611556e-1,-0.,-.4272429480e-1,.4272429479e-1,-0.,.9845066677e-1,-.5555835628e-1,-.4289231078e-1,.4289231057e-1,-.9845066699e-1,.5555835614e-1,.5641598021e-1,-.1079351250e-1,-.4562246768e-1,.4562246772e-1,-.5641598021e-1,.1079351250e-1},
			{-.3358542881e-2,.3358542884e-2,.2558537468e-11,-.9114812152e-1,.9114812160e-1,.4628972095e-10,.1388407889,-.1388407891,-.3559378643e-10,.1393714005,-.1393714005,0.,.7400064931e-1,-.7400064927e-1,0.,-.1705215569,.9622989583e-1,.742916614e-1,-.7429166116e-1,.1705215571,-.962298956e-1,-.9771534407e-1,.1869491204e-1,.7902043198e-1,-.7902043202e-1,.9771534403e-1,-.1869491203e-1},
			{.4335860212e-2,-.4335860220e-2,-.3303057668e-11,.1176717189,-.1176717190,-.5975977276e-10,-.1792426878,.1792426879,.4595138070e-10,-.1799277044,.1799277042,-0.,-.9553442744e-1,.9553442741e-1,-0.,.2201423833,-.1242322613,-.9591012258e-1,.9591012213e-1,-.2201423837,.1242322610,.1261499667,-.2413502766e-1,-.1020149390,.1020149391,-.1261499667,.2413502765e-1},
			{-.5130258989e-2,.5130258998e-2,.3908230539e-11,-.1392310554,.1392310556,.7070871670e-10,.2120828083,-.2120828085,-.5437040688e-10,.2128933309,-.2128933307,0.,.1130378590,-.1130378589,0.,-.2604759806,.1469935939,.1134823874,-.1134823869,.2604759812,-.1469935935,-.1492626536,.2855694985e-1,.1207057037,-.1207057037,.1492626536,-.2855694983e-1},
			{.3973252640e-2,-.3973252644e-2,.2435328156e-11,.6411930517e-1,-.6411930524e-1,-.3373654288e-12,-.1624751717e-2,.1624751718e-2,.4950415964e-10,.3974002441e-1,-.3974002439e-1,-0.,.4385866113e-1,-.4385866112e-1,-0.,.9442419982e-1,-.1337765132,.3935231328e-1,-.3935231309e-1,-.9442420005e-1,.1337765128,-.9849870528e-2,-.4015289410e-1,.5000276456e-1,-.5000276462e-1,.9849870528e-2,.4015289406e-1},
			{-.6881875447e-2,.6881875454e-2,-.4218112101e-11,-.1110578943,.1110578945,.5843340636e-12,.2814152524e-2,-.2814152527e-2,-.8574371972e-10,-.6883174139e-1,.6883174137e-1,0.,-.7596542946e-1,.7596542944e-1,0.,-.1635475116,.2317077178,-.6816020601e-1,.6816020569e-1,.1635475120,-.2317077171,.1706047621e-1,.6954685268e-1,-.8660732877e-1,.8660732887e-1,-.1706047621e-1,-.6954685260e-1},
			{.8884462997e-2,-.8884463007e-2,.5445559306e-11,.1433751251,-.1433751252,-.7543720322e-12,-.3633055286e-2,.3633055289e-2,.1106946662e-9,.8886139602e-1,-.8886139599e-1,-0.,.9807094772e-1,-.9807094769e-1,-0.,.2111389296,-.2991333773,.8799444758e-1,-.8799444716e-1,-.2111389301,.2991333765,-.2202498007e-1,-.8978460072e-1,.1118095806,-.1118095808,.2202498007e-1,.8978460063e-1},
			{.1608750430e-2,-.1608750432e-2,-.4034538186e-11,-.3311226875e-1,.3311226880e-1,.2144253100e-10,.6480800897e-1,-.6480800902e-1,.3701184809e-10,.1514326846e-1,-.1514326846e-1,-0.,-.3828400335e-1,.3828400333e-1,-0.,-.8474998733e-1,-.9988203793e-1,-.2734121092e-1,.2734121078e-1,.8474998753e-1,.9988203767e-1,-.3512819587e-1,-.7636653997e-1,-.4884262097e-1,.4884262099e-1,.3512819587e-1,.7636653992e-1},
			{-.2786437481e-2,.2786437483e-2,.6988025119e-11,.5735213183e-1,-.5735213188e-1,-.3713955314e-10,-.1122507642,.1122507643,-.6410640134e-10,-.2622891035e-1,.2622891035e-1,0.,.6630983887e-1,-.6630983886e-1,0.,.1467912839,.1730007645,.4735636645e-1,-.4735636618e-1,-.1467912843,-.1730007640,.6084382000e-1,.1322707272,.8459790107e-1,-.8459790113e-1,-.6084382000e-1,-.1322707271},
			{-.2900983391e-2,.2900983395e-2,.6735979401e-11,-.5474334561e-1,.5474334565e-1,-.2140983594e-10,-.6378935306e-1,.6378935317e-1,-.1837023118e-10,-.5758708731e-1,.5758708725e-1,-0.,.2739123664e-1,-.2739123663e-1,-0.,-.5980191219e-1,.4982269315e-1,.9979215896e-2,-.9979215849e-2,.5980191234e-1,-.4982269302e-1,.4947010457e-1,-.9238197165e-1,.4291186704e-1,-.4291186706e-1,-.4947010455e-1,.9238197160e-1},
			{-.2227101124e-2,-.2227101127e-2,.2902345977e-2,.5174730169e-1,.5174730161e-1,.1345249687,.2895906774e-1,.2895906742e-1,.2686152177e-1,-.4565783520e-1,-.4565783503e-1,.5646315914e-2,-.5646454713e-1,-.5646454707e-1,.1567410230e-3,-.6746381852e-1,-.1666020075e-1,.1052409173,.1052409171,-.6746381827e-1,-.1666020070e-1,-.6630339505e-1,-.1750103542e-2,-.1446623172e-1,-.1446623165e-1,-.6630339516e-1,-.1750103540e-2},
			{.3857452299e-2,.3857452305e-2,-.5027010692e-2,-.8962895569e-1,-.8962895553e-1,-.2330040807,-.5015857668e-1,-.5015857611e-1,-.4652552046e-1,.7908169034e-1,.7908169002e-1,-.9779706035e-2,.9779946441e-1,.9779946432e-1,-.2714834155e-3,.1168507614,.2885631417e-1,-.1822826158,-.1822826154,.1168507609,.2885631408e-1,.1148408489,.3031268254e-2,.2505624833e-1,.2505624821e-1,.1148408492,.3031268250e-2},
			{-.4979949508e-2,-.4979949513e-2,.6489842900e-2,.1157104843,.1157104841,.3008069748,.6475444405e-1,.6475444331e-1,.6006418864e-1,-.1020940232,-.1020940229,.1262554620e-1,-.1262585657,-.1262585656,.3504835826e-3,-.1508536843,-.3725334142e-1,.2353258451,.2353258446,-.1508536837,-.3725334129e-1,-.1482588985,-.3913350489e-2,-.3234747750e-1,-.3234747736e-1,-.1482588988,-.3913350484e-2},
			{-.7052136016e-2,-.7052136025e-2,-.4975248153e-2,.1071236388,.1071236386,.6437588332e-1,.1415037995e-1,.1415037979e-1,.9781715849e-1,.8674924746e-2,.8674924715e-2,.2822480032e-1,.5042836149e-1,.5042836146e-1,.1003925893e-2,-.1165659990,-.6064459330e-1,-.7883759978e-1,-.7883759957e-1,-.1165659985,-.6064459310e-1,-.1534448774e-1,-.9401828142e-2,.1424607884e-1,.1424607878e-1,-.1534448778e-1,-.9401828126e-2},
			{.1221465788e-1,.1221465790e-1,.8617382582e-2,-.1855435851,-.1855435847,-.1115023007,-.2450917702e-1,-.2450917675e-1,-.1694242884,-.1502541041e-1,-.1502541036e-1,-.4888678819e-1,-.8734448425e-1,-.8734448420e-1,-.1738850654e-2,.2018982328,.1050395168,.1365507284,.1365507280,.2018982319,.1050395164,.2657743238e-1,.1628444403e-1,-.2467493236e-1,-.2467493225e-1,.2657743245e-1,.1628444400e-1},
			{-.7864599685e-2,-.7864599698e-2,.8543165539e-2,.2317116788e-1,.2317116784e-1,-.1325361900,-.2859660209e-1,-.2859660176e-1,.1450070316,.2338942873e-1,.2338942865e-1,.7058964713e-1,-.3789617428e-1,-.3789617426e-1,.3546410271e-2,.1110571477e-1,-.8983420728e-1,.3910101757e-1,.3910101750e-1,.1110571472e-1,-.8983420697e-1,.5864432557e-1,-.2610381621e-1,-.1269128828e-1,-.1269128822e-1,.5864432571e-1,-.2610381617e-1},
			{.2457406867e-2,-.2457406870e-2,.1456786833e-10,-.5057982506e-1,.5057982506e-1,-0.,.9899586944e-1,-.9899586881e-1,.4439325132e-9,.2313172558e-1,-.2313172466e-1,-0.,-.5847978098e-1,.5847978105e-1,.8177605344e-10,.2473009711e-1,.1615508286e-2,.1124234492,-.1124234521,-.2473010027e-1,-.1615508521e-2,.6296906016e-1,-.2355095997e-4,.4201992979e-1,-.4201992986e-1,-.6296905964e-1,.2355071203e-4},
			{-.4256353547e-2,.4256353552e-2,-.2523228810e-10,.8760682683e-1,-.8760682683e-1,0.,-.1714658756,.1714658745,-.7689136679e-9,-.4006532397e-1,.4006532237e-1,0.,.1012899519,-.1012899520,-.1416402794e-9,-.4283378466e-1,-.2798142432e-2,-.1947231258,.1947231310,.4283379015e-1,.2798142839e-2,-.1090656115,.4079145923e-4,-.7278065332e-1,.7278065345e-1,.1090656106,-.4079102978e-4},
			{.1047195901e-1,-.1047195902e-1,-.2643438623e-10,-.1465053324,.1465053324,-0.,.9184167768e-1,-.9184167710e-1,.2154288659e-8,.1917883762e-2,-.1917883686e-2,-0.,.4633750594e-1,-.4633750599e-1,.6781638958e-9,.6109303566e-1,.7836730672e-2,-.6892976350e-1,.6892976527e-1,-.6109304346e-1,-.7836731808e-2,.3825625994e-1,-.1650255183e-3,-.3809123559e-1,.3809123567e-1,-.3825625962e-1,.1650237808e-3},
			{-.2665390899e-2,-.2665390853e-2,-.4310294403e-2,-.2187432388e-2,-.2187432754e-2,-.1112663225,.6179199770e-1,.6179199994e-1,-.5769386249e-2,.7894214835e-1,.7894214861e-1,-.4179238568e-3,-.4846361631e-1,-.4846362257e-1,-.8532846255e-6,.6483970316e-2,.4679778692e-2,-.6432957221e-1,-.6432958410e-1,.6483978942e-2,.4679777698e-2,-.5074575336e-1,.7639251247e-4,.7729988409e-1,.7729988561e-1,-.5074575829e-1,.7639349119e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
			o_output[i] = 0;

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 27; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

	/*********************************************************
	 * timestep_top_fluxes_to_dofs
	 *********************************************************/
public:
	static inline void madd_timestep_top_fluxes_to_dofs(
		const T i_input[27],
		T o_output[35]
	) {
		static const T matrix[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{.2270692495e-2,.2270692495e-2,.2270692495e-2,.1889769944,.1889769944,.1889769944,.1453110246,.1453110246,.1453110246,.8863366081e-1,.8863366081e-1,.8863366081e-1,.3345937098e-1,.3345937098e-1,.3345937098e-1,.1281992189,.1281992189,.1281992189,.1281992189,.1281992189,.1281992189,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{.3468540079e-2,.3468540079e-2,.3468540079e-2,.2886671272,.2886671272,.2886671272,.2219662565,.2219662565,.2219662565,.1353901533,.1353901533,.1353901533,.5111003340e-1,.5111003340e-1,.5111003340e-1,.1958275416,.1958275416,.1958275416,.1958275416,.1958275416,.1958275416,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1119514294e-2,-.1119514294e-2,.2239028587e-2,-.3038270720e-1,-.3038270720e-1,.6076541432e-1,.4628026305e-1,.4628026305e-1,-.9256052604e-1,.4645713347e-1,.4645713347e-1,-.9291426695e-1,.2466688309e-1,.2466688309e-1,-.4933376618e-1,-.7312744874e-2,-.816044061e-1,.8891715086e-1,.8891715086e-1,-.7312744874e-2,-.816044061e-1,.2010850666e-1,-.5891192536e-1,.3880341870e-1,.3880341870e-1,.2010850666e-1,-.5891192536e-1},
			{.1939055637e-2,.1939055637e-2,-.3878111275e-2,.5262439255e-1,.5262439255e-1,-.1052487850,-.8015976696e-1,-.8015976696e-1,.1603195338,-.8046611556e-1,-.8046611556e-1,.1609322311,-.4272429479e-1,-.4272429479e-1,.8544858956e-1,.1266604567e-1,.1413429774,-.1540090230,-.1540090230,.1266604567e-1,.1413429774,-.3482895521e-1,.1020384480,-.6720949271e-1,-.6720949271e-1,-.3482895521e-1,.1020384480},
			{-.2503310063e-2,-.2503310063e-2,.5006620125e-2,-.6793779862e-1,-.6793779862e-1,.1358755971,.1034858141,.1034858141,-.2069716281,.1038813085,.1038813085,-.2077626169,.5515682737e-1,.5515682737e-1,-.1103136547,-.1635179464e-1,-.1824729992,.1988247936,.1988247936,-.1635179464e-1,-.1824729992,.4496398781e-1,-.1317310698,.8676708195e-1,.8676708195e-1,.4496398781e-1,-.1317310698},
			{.2961956411e-2,.2961956411e-2,-.5923912822e-2,.8038508741e-1,.8038508741e-1,-.1607701747,-.1224460666,-.1224460666,.2448921330,-.1229140218,-.1229140218,.2458280436,-.6526243827e-1,-.6526243827e-1,.1305248765,.1934770434e-1,.2159049643,-.2352526685,-.2352526685,.1934770434e-1,.2159049643,-.5320210786e-1,.1558663038,-.1026641959,-.1026641959,-.5320210786e-1,.1558663038},
			{.9509629649e-3,.9509629649e-3,-.3636953995e-2,.1133841892,.1133841892,.3934559288e-1,.3097014551e-1,.3097014551e-1,.3284624728e-1,-.1663114073e-1,-.1663114073e-1,-.6251896835e-1,-.2514656806e-1,-.2514656806e-1,-.7579018773e-1,.7450711923e-1,.2906698235e-1,-.7996469216e-1,-.7996469216e-1,.7450711923e-1,.2906698235e-1,.3947641499e-2,-.5379057771e-1,-.4241692689e-1,-.4241692689e-1,.3947641499e-2,-.5379057771e-1},
			{-.1647116171e-2,-.1647116171e-2,.6299389101e-2,-.1963871764,-.1963871764,-.6814856589e-1,-.5364186551e-1,-.5364186551e-1,-.5689136911e-1,.2880598072e-1,.2880598072e-1,.1082860296,.4355513348e-1,.4355513348e-1,.1312724558,-.1290501160,-.5034549024e-1,.1385029096,.1385029096,-.1290501160,-.5034549024e-1,-.6837515642e-2,.9316801354e-1,.7346827246e-1,.7346827246e-1,-.6837515642e-2,.9316801354e-1},
			{.2126417833e-2,.2126417833e-2,-.8132476364e-2,.2535347545,.2535347545,.8797942029e-1,.6925135063e-1,.6925135063e-1,.7344644172e-1,-.3718836120e-1,-.3718836120e-1,-.1397966631,-.5622943557e-1,-.5622943557e-1,-.1694720118,.1666029834,.6499574844e-1,-.1788064875,-.1788064875,.1666029834,.6499574844e-1,.8827194741e-2,-.1202793883,-.9484713192e-1,-.9484713192e-1,.8827194741e-2,-.1202793883},
			{.1296186118e-2,.1296186118e-2,.5940248584e-2,-.9672393084e-2,-.9672393084e-2,-.1052592792,-.8116208511e-1,-.8116208511e-1,.1059225219,-.2136548439e-1,-.2136548439e-1,.2234936694e-1,.2165431039e-1,.2165431039e-1,-.8886208754e-1,-.3163303256e-1,.9341103024e-1,.5273023340e-1,.5273023340e-1,-.3163303256e-1,.9341103024e-1,-.2491962037e-1,-.1163142919e-1,.4124689791e-1,.4124689791e-1,-.2491962037e-1,-.1163142919e-1},
			{-.2245060214e-2,-.2245060214e-2,-.1028881236e-1,.1675307626e-1,.1675307626e-1,.1823144197,.1405768551,.1405768551,-.1834631897,.3700610451e-1,.3700610451e-1,-.3871023906e-1,-.3750636580e-1,-.3750636580e-1,.1539136505,.5479001961e-1,-.1617926504,-.9133144337e-1,-.9133144337e-1,.5479001961e-1,-.1617926504,.4316204859e-1,.2014622633e-1,-.7144172285e-1,-.7144172285e-1,.4316204859e-1,.2014622633e-1},
			{-.7251009248e-3,-.7251009248e-3,-.9797720361e-2,-.1092962313,-.1092962313,.8367774286e-1,.6348397322e-1,.6348397322e-1,.1048243834e-1,.4861432003e-1,.4861432003e-1,.8648212617e-1,-.1502218966e-1,-.1502218966e-1,-.8303840497e-1,-.5695773555e-1,.9076878046e-2,-.1414594348e-1,-.1414594348e-1,-.5695773555e-1,.9076878046e-2,.2822197953e-1,.3880374283e-1,-.3595677329e-1,-.3595677329e-1,.2822197953e-1,.3880374283e-1},
			{.1939055636e-2,-.1939055639e-2,-.1477172296e-11,.5262439250e-1,-.5262439255e-1,-.2672538285e-10,-.8015976689e-1,.8015976696e-1,.2055008219e-10,-.8046611563e-1,.8046611556e-1,-0.,-.4272429480e-1,.4272429479e-1,-0.,.9845066677e-1,-.5555835628e-1,-.4289231078e-1,.4289231057e-1,-.9845066699e-1,.5555835614e-1,.5641598021e-1,-.1079351250e-1,-.4562246768e-1,.4562246772e-1,-.5641598021e-1,.1079351250e-1},
			{-.3358542881e-2,.3358542884e-2,.2558537468e-11,-.9114812152e-1,.9114812160e-1,.4628972095e-10,.1388407889,-.1388407891,-.3559378643e-10,.1393714005,-.1393714005,0.,.7400064931e-1,-.7400064927e-1,0.,-.1705215569,.9622989583e-1,.742916614e-1,-.7429166116e-1,.1705215571,-.962298956e-1,-.9771534407e-1,.1869491204e-1,.7902043198e-1,-.7902043202e-1,.9771534403e-1,-.1869491203e-1},
			{.4335860212e-2,-.4335860220e-2,-.3303057668e-11,.1176717189,-.1176717190,-.5975977276e-10,-.1792426878,.1792426879,.4595138070e-10,-.1799277044,.1799277042,-0.,-.9553442744e-1,.9553442741e-1,-0.,.2201423833,-.1242322613,-.9591012258e-1,.9591012213e-1,-.2201423837,.1242322610,.1261499667,-.2413502766e-1,-.1020149390,.1020149391,-.1261499667,.2413502765e-1},
			{-.5130258989e-2,.5130258998e-2,.3908230539e-11,-.1392310554,.1392310556,.7070871670e-10,.2120828083,-.2120828085,-.5437040688e-10,.2128933309,-.2128933307,0.,.1130378590,-.1130378589,0.,-.2604759806,.1469935939,.1134823874,-.1134823869,.2604759812,-.1469935935,-.1492626536,.2855694985e-1,.1207057037,-.1207057037,.1492626536,-.2855694983e-1},
			{.3973252640e-2,-.3973252644e-2,.2435328156e-11,.6411930517e-1,-.6411930524e-1,-.3373654288e-12,-.1624751717e-2,.1624751718e-2,.4950415964e-10,.3974002441e-1,-.3974002439e-1,-0.,.4385866113e-1,-.4385866112e-1,-0.,.9442419982e-1,-.1337765132,.3935231328e-1,-.3935231309e-1,-.9442420005e-1,.1337765128,-.9849870528e-2,-.4015289410e-1,.5000276456e-1,-.5000276462e-1,.9849870528e-2,.4015289406e-1},
			{-.6881875447e-2,.6881875454e-2,-.4218112101e-11,-.1110578943,.1110578945,.5843340636e-12,.2814152524e-2,-.2814152527e-2,-.8574371972e-10,-.6883174139e-1,.6883174137e-1,0.,-.7596542946e-1,.7596542944e-1,0.,-.1635475116,.2317077178,-.6816020601e-1,.6816020569e-1,.1635475120,-.2317077171,.1706047621e-1,.6954685268e-1,-.8660732877e-1,.8660732887e-1,-.1706047621e-1,-.6954685260e-1},
			{.8884462997e-2,-.8884463007e-2,.5445559306e-11,.1433751251,-.1433751252,-.7543720322e-12,-.3633055286e-2,.3633055289e-2,.1106946662e-9,.8886139602e-1,-.8886139599e-1,-0.,.9807094772e-1,-.9807094769e-1,-0.,.2111389296,-.2991333773,.8799444758e-1,-.8799444716e-1,-.2111389301,.2991333765,-.2202498007e-1,-.8978460072e-1,.1118095806,-.1118095808,.2202498007e-1,.8978460063e-1},
			{.1608750430e-2,-.1608750432e-2,-.4034538186e-11,-.3311226875e-1,.3311226880e-1,.2144253100e-10,.6480800897e-1,-.6480800902e-1,.3701184809e-10,.1514326846e-1,-.1514326846e-1,-0.,-.3828400335e-1,.3828400333e-1,-0.,-.8474998733e-1,-.9988203793e-1,-.2734121092e-1,.2734121078e-1,.8474998753e-1,.9988203767e-1,-.3512819587e-1,-.7636653997e-1,-.4884262097e-1,.4884262099e-1,.3512819587e-1,.7636653992e-1},
			{-.2786437481e-2,.2786437483e-2,.6988025119e-11,.5735213183e-1,-.5735213188e-1,-.3713955314e-10,-.1122507642,.1122507643,-.6410640134e-10,-.2622891035e-1,.2622891035e-1,0.,.6630983887e-1,-.6630983886e-1,0.,.1467912839,.1730007645,.4735636645e-1,-.4735636618e-1,-.1467912843,-.1730007640,.6084382000e-1,.1322707272,.8459790107e-1,-.8459790113e-1,-.6084382000e-1,-.1322707271},
			{-.2900983391e-2,.2900983395e-2,.6735979401e-11,-.5474334561e-1,.5474334565e-1,-.2140983594e-10,-.6378935306e-1,.6378935317e-1,-.1837023118e-10,-.5758708731e-1,.5758708725e-1,-0.,.2739123664e-1,-.2739123663e-1,-0.,-.5980191219e-1,.4982269315e-1,.9979215896e-2,-.9979215849e-2,.5980191234e-1,-.4982269302e-1,.4947010457e-1,-.9238197165e-1,.4291186704e-1,-.4291186706e-1,-.4947010455e-1,.9238197160e-1},
			{-.2227101124e-2,-.2227101127e-2,.2902345977e-2,.5174730169e-1,.5174730161e-1,.1345249687,.2895906774e-1,.2895906742e-1,.2686152177e-1,-.4565783520e-1,-.4565783503e-1,.5646315914e-2,-.5646454713e-1,-.5646454707e-1,.1567410230e-3,-.6746381852e-1,-.1666020075e-1,.1052409173,.1052409171,-.6746381827e-1,-.1666020070e-1,-.6630339505e-1,-.1750103542e-2,-.1446623172e-1,-.1446623165e-1,-.6630339516e-1,-.1750103540e-2},
			{.3857452299e-2,.3857452305e-2,-.5027010692e-2,-.8962895569e-1,-.8962895553e-1,-.2330040807,-.5015857668e-1,-.5015857611e-1,-.4652552046e-1,.7908169034e-1,.7908169002e-1,-.9779706035e-2,.9779946441e-1,.9779946432e-1,-.2714834155e-3,.1168507614,.2885631417e-1,-.1822826158,-.1822826154,.1168507609,.2885631408e-1,.1148408489,.3031268254e-2,.2505624833e-1,.2505624821e-1,.1148408492,.3031268250e-2},
			{-.4979949508e-2,-.4979949513e-2,.6489842900e-2,.1157104843,.1157104841,.3008069748,.6475444405e-1,.6475444331e-1,.6006418864e-1,-.1020940232,-.1020940229,.1262554620e-1,-.1262585657,-.1262585656,.3504835826e-3,-.1508536843,-.3725334142e-1,.2353258451,.2353258446,-.1508536837,-.3725334129e-1,-.1482588985,-.3913350489e-2,-.3234747750e-1,-.3234747736e-1,-.1482588988,-.3913350484e-2},
			{-.7052136016e-2,-.7052136025e-2,-.4975248153e-2,.1071236388,.1071236386,.6437588332e-1,.1415037995e-1,.1415037979e-1,.9781715849e-1,.8674924746e-2,.8674924715e-2,.2822480032e-1,.5042836149e-1,.5042836146e-1,.1003925893e-2,-.1165659990,-.6064459330e-1,-.7883759978e-1,-.7883759957e-1,-.1165659985,-.6064459310e-1,-.1534448774e-1,-.9401828142e-2,.1424607884e-1,.1424607878e-1,-.1534448778e-1,-.9401828126e-2},
			{.1221465788e-1,.1221465790e-1,.8617382582e-2,-.1855435851,-.1855435847,-.1115023007,-.2450917702e-1,-.2450917675e-1,-.1694242884,-.1502541041e-1,-.1502541036e-1,-.4888678819e-1,-.8734448425e-1,-.8734448420e-1,-.1738850654e-2,.2018982328,.1050395168,.1365507284,.1365507280,.2018982319,.1050395164,.2657743238e-1,.1628444403e-1,-.2467493236e-1,-.2467493225e-1,.2657743245e-1,.1628444400e-1},
			{-.7864599685e-2,-.7864599698e-2,.8543165539e-2,.2317116788e-1,.2317116784e-1,-.1325361900,-.2859660209e-1,-.2859660176e-1,.1450070316,.2338942873e-1,.2338942865e-1,.7058964713e-1,-.3789617428e-1,-.3789617426e-1,.3546410271e-2,.1110571477e-1,-.8983420728e-1,.3910101757e-1,.3910101750e-1,.1110571472e-1,-.8983420697e-1,.5864432557e-1,-.2610381621e-1,-.1269128828e-1,-.1269128822e-1,.5864432571e-1,-.2610381617e-1},
			{.2457406867e-2,-.2457406870e-2,.1456786833e-10,-.5057982506e-1,.5057982506e-1,-0.,.9899586944e-1,-.9899586881e-1,.4439325132e-9,.2313172558e-1,-.2313172466e-1,-0.,-.5847978098e-1,.5847978105e-1,.8177605344e-10,.2473009711e-1,.1615508286e-2,.1124234492,-.1124234521,-.2473010027e-1,-.1615508521e-2,.6296906016e-1,-.2355095997e-4,.4201992979e-1,-.4201992986e-1,-.6296905964e-1,.2355071203e-4},
			{-.4256353547e-2,.4256353552e-2,-.2523228810e-10,.8760682683e-1,-.8760682683e-1,0.,-.1714658756,.1714658745,-.7689136679e-9,-.4006532397e-1,.4006532237e-1,0.,.1012899519,-.1012899520,-.1416402794e-9,-.4283378466e-1,-.2798142432e-2,-.1947231258,.1947231310,.4283379015e-1,.2798142839e-2,-.1090656115,.4079145923e-4,-.7278065332e-1,.7278065345e-1,.1090656106,-.4079102978e-4},
			{.1047195901e-1,-.1047195902e-1,-.2643438623e-10,-.1465053324,.1465053324,-0.,.9184167768e-1,-.9184167710e-1,.2154288659e-8,.1917883762e-2,-.1917883686e-2,-0.,.4633750594e-1,-.4633750599e-1,.6781638958e-9,.6109303566e-1,.7836730672e-2,-.6892976350e-1,.6892976527e-1,-.6109304346e-1,-.7836731808e-2,.3825625994e-1,-.1650255183e-3,-.3809123559e-1,.3809123567e-1,-.3825625962e-1,.1650237808e-3},
			{-.2665390899e-2,-.2665390853e-2,-.4310294403e-2,-.2187432388e-2,-.2187432754e-2,-.1112663225,.6179199770e-1,.6179199994e-1,-.5769386249e-2,.7894214835e-1,.7894214861e-1,-.4179238568e-3,-.4846361631e-1,-.4846362257e-1,-.8532846255e-6,.6483970316e-2,.4679778692e-2,-.6432957221e-1,-.6432958410e-1,.6483978942e-2,.4679777698e-2,-.5074575336e-1,.7639251247e-4,.7729988409e-1,.7729988561e-1,-.5074575829e-1,.7639349119e-4}
		};

#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
		for (int i = 0; i < 35; i++)
		{
#if CONFIG_ENABLE_INTEL_SIMD_PRAGMA
#pragma simd
#endif
			for (int j = 0; j < 27; j++)
				o_output[i] += matrix[i][j]*i_input[j];
		}
	}

public:
	static inline int getNumberOfFlops()
	{
		return 148400;
	}


	/**
	 * setup the matrices for the static variables
	 *
	 * this method has to be invoked once during startup
	 */
public:
	static void setup(int i_verbosity_level)
	{
	}


public:
	static void debugOutput(int i_verbosity_level)
	{
		std::cout << "**********************************************" << std::endl;
		std::cout << "* COMPUTATION MATRICES CONST 0" << std::endl;
		std::cout << "**********************************************" << std::endl;
		std::cout << std::endl;
		std::cout << "Information about System DOFs and matrices:" << std::endl;
		std::cout << " + Basis function degree: " << SIMULATION_HYPERBOLIC_DEGREE_OF_BASIS_FUNCTIONS << std::endl;
		std::cout << " + Number of basis functions: " << SIMULATION_HYPERBOLIC_NUMBER_OF_BASIS_FUNCTIONS << std::endl;
//		std::cout << " + Integration cell degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_CELL_ORDER << std::endl;
//		std::cout << " + Integration edge degree: " << SIMULATION_HYPERBOLIC_INTEGRATION_EDGE_ORDER << std::endl;
		std::cout << std::endl;

		std::cout << std::endl;
		std::cout << "timestep_inv_mass:" << std::endl;
		static const T timestep_inv_mass[35][35] = 
		{
			{1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_inv_mass[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_inv_mass[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_u:" << std::endl;
		static const T timestep_stiffness_u[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.898979486,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.,0.,0.,0.,0.,7.071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,7.838367178,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,3.464101616,0.,0.,0.,0.,7.838367178,0.,0.,0.,8.400000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.098386677,0.,0.,0.,0.,7.668115805,0.,0.,0.,10.73312629,0.,0.,9.295160030,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832980,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,9.486832977,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780461,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.381780460,0.,0.,0.,12.52198067,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,2.449489743,0.,0.,0.,9.142857143,0.,0.,14.84614978,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,0.,0.,0.,-2.993325910,0.,0.,0.,.6110100926,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260102,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.291502622,0.,0.,0.,0.,-2.993325910,0.,0.,0.,.6110100925,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,13.66260101,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.732863826,0.,0.,0.,0.,3.346640106,0.,0.,0.,-5.074680381,0.,0.,1.521277658,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.364357806,0.,0.,17.00840129,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,10.95445115,0.,0.,0.,-3.833259389,0.,0.,.5532833353,0.,0.,0.,0.,0.,0.,0.,0.,17.74823935,0.,0.,0.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_stiffness_u[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_stiffness_u[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_v:" << std::endl;
		static const T timestep_stiffness_v[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,4.242640686,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,-2.309401077,0.,0.,0.,0.,8.164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,5.,0.,0.,0.,0.,-2.828427124,0.,0.,0.,12.12435566,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{-3.577708763,0.,0.,0.,0.,8.854377448,0.,0.,0.,-3.098386677,0.,0.,16.09968943,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,2.449489743,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.123724358,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.732050808,0.,0.,0.,0.,3.919183589,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,1.732050808,0.,0.,0.,0.,3.919183589,0.,0.,0.,4.200000000,0.,0.,0.,0.,0.,-5.091168823,0.,0.,0.,9.699484525,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{1.549193338,0.,0.,0.,0.,3.834057902,0.,0.,0.,5.366563145,0.,0.,4.647580015,0.,0.,6.640783086,0.,0.,0.,-6.196773354,0.,0.,13.41640786,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.581988897,0.,0.,0.,0.,-.9128709293,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.581988897,0.,0.,0.,0.,-.9128709295,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416490,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,2.581988897,0.,0.,0.,0.,-.9128709292,0.,0.,0.,0.,0.,0.,0.,0.,0.,4.743416488,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.236067977,0.,0.,0.,0.,2.529822128,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.236067977,0.,0.,0.,0.,2.529822128,0.,0.,0.,-1.807392228,0.,0.,0.,0.,0.,2.190890230,0.,0.,0.,6.260990336,0.,0.,0.,0.,0.,8.082903771,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.,0.,0.,0.,0.,3.535533905,0.,0.,0.,1.649572198,0.,0.,-2.571428571,0.,0.,1.224744872,0.,0.,0.,4.571428571,0.,0.,7.423074892,0.,0.,-7.377111136,0.,0.,11.49977817,0.,0.,0.,0.,0.,0.},
			{2.645751311,0.,0.,0.,0.,-1.496662955,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,5.184592560,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,6.831300514,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,2.645751311,0.,0.,0.,0.,-1.496662955,0.,0.,0.,.3055050463,0.,0.,0.,0.,0.,5.184592558,0.,0.,0.,-1.058300524,0.,0.,0.,0.,0.,6.831300507,0.,0.,0.,0.,0.,0.,0.,0.},
			{2.366431913,0.,0.,0.,0.,1.673320053,0.,0.,0.,-2.537340190,0.,0.,.7606388294,0.,0.,2.898275350,0.,0.,0.,5.747048932,0.,0.,-2.195775165,0.,0.,2.182178903,0.,0.,8.504200642,0.,0.,10.06230590,0.,0.,0.},
			{2.683281572,0.,0.,0.,0.,-1.897366596,0.,0.,0.,.6639400021,0.,0.,-.9583148473e-1,0.,0.,5.477225576,0.,0.,0.,-1.916629694,0.,0.,.2766416676,0.,0.,7.423074892,0.,0.,-1.071428571,0.,0.,8.874119673,0.,0.,0.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_stiffness_v[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_stiffness_v[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_hyp_face:" << std::endl;
		static const T timestep_dofs_to_hyp_face[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.92171958555151,3.24166342459793,-3.96510583993544,4.31411318985288,2.19799643560933,-3.70770257117539,4.53515100205683,-2.29634628619806,3.87360456901059,2.23836106649518,3.41890639763464,-5.76720136441661,7.05426838770182,-7.67518283761778,-3.91413586924262,6.60258197030424,-8.07608098333328,4.09536544265748,-6.90829008438974,-4.00097073404889,5.33523806427656,-8.99977614812198,11.0082572774727,-5.59797201662203,9.44297038492047,5.49216459339238,7.19422103343416,-12.1356119587979,-7.09892453011936,9.00134910987289},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.59519010006696,2.69085533682779,-3.29137384514269,3.5810795200105,1.2388862761662,-2.0898222386161,2.55620812016103,-.514957807193037,.868659455579414,-.353775594121333,3.23038451124486,-5.44919216674985,6.66528903908045,-7.25196565099481,-2.62173967966463,4.42249623366662,-5.4094652504454,1.30644820527577,-2.20378945631257,.307822029727901,4.7630793830775,-8.03462707886587,9.82771577392637,-2.90243304119459,4.89598539807522,.518849452524734,6.06854814809742,-10.2367643607743,-2.66498561200528,7.17423567654379},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.03822870489707,1.75134189417476,-2.14218907478874,2.33074387312075,-.0622566343920988,.105017951653011,-.12845482061473,1.10967175893133,-1.87185600940662,-1.41978880536247,2.90882269973526,-4.90676382786187,6.00180690245521,-6.53008402866105,-.707254507004454,1.19303621855815,-1.45928625505293,-1.57948219926848,2.66435837496252,2.54719860772099,3.86201434387853,-6.51466048131366,7.96853804732135,.544351948717181,-.918243127233759,-3.18953813043078,4.43071582310306,-7.47397774950178,2.19278566166382,4.71658607960972},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-.300186182387741,.506370739691845,-.619377558413115,.673894973328731,-1.13611369814234,1.91645974101964,-2.34415629302528,1.22655366483195,-2.06901890572263,.100858250439743,2.48271365058504,-4.18797939686675,5.12261126340126,-5.57350187032702,1.26648273457434,-2.13637348021345,2.61314820693614,-2.5074905585056,4.22977446202813,.461151554581851,2.81340752869054,-4.74581221430261,5.80493569904152,3.19379348794833,-5.38746839617942,-2.33714259619681,2.75487590413567,-4.64707781591112,4.77320941254594,2.50302559510188},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,.553376983055106,-.93346705704166,1.14178901214037,-1.24228891639978,-1.45368745167752,2.45216080196962,-2.99940982448642,-.194424990185747,.327966881493408,1.49978166741586,1.9899087271236,-3.35668865754018,4.10580126963162,-4.46719258573777,2.74863027720701,-4.6365423474252,5.67128006117502,-.854309601049289,1.44109692494208,-2.52312426315122,1.80736427869175,-3.04876253512471,3.72915523810848,4.12997714955087,-6.96667503835439,1.65834138994758,1.41847476579741,-2.39276208655032,4.48820932686104,1.03297875354961},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,1.44662301694489,-2.44024412222762,2.98483369571688,-3.24755780451478,-.724355119014703,1.22188248065577,-1.49457014152702,-1.6918152058931,2.85384793698481,.160475142667942,1.4741928888764,-2.48675051356285,3.04171892525937,-3.30945005334757,3.38025060487184,-5.70199462778367,6.97450945518368,2.16048862424162,-3.64443230995284,-1.70770592621981,.991945941760336,-1.67326955598402,2.04669332477546,3.46032388717692,-5.8370666898335,4.72311662432345,.576746649127291,-.972888308521253,2.68886700736676,.311154445333738},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,2.30018618238774,-3.88008191896113,4.74600026627036,-5.1637416942433,.987081447023724,-1.66506385533551,2.03665635715207,-1.35561938023839,2.28673412920868,-2.20870753342451,.981387965414959,-1.65545977423627,2.02490893148973,-2.20314076875832,3.10522366683209,-5.23806390000324,6.40704325104372,4.42237267757001,-7.45990406612127,2.83495589139377,.43960319187862,-.741547101224612,.907038257317998,2.03901059601618,-3.43951642050873,4.55407145300732,.170154848292172,-.287026656838237,1.03685515398485,.0611113213546687},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.03822870489707,-5.12505307344404,6.26881178264598,-6.82059059403531,3.26615690329693,-5.50953502563763,6.7390986231311,1.66498726415444,-2.80859307349989,-.923099517802526,.555278916264744,-.936675343241156,1.14571329243577,-1.24655861042429,2.17523616991868,-3.66930929238228,4.48818948881768,4.50024407433993,-7.59126186695628,6.26847807110811,.140734880491405,-.237399419743997,.29037987690505,.792696592497175,-1.33716467766395,2.35258131414952,.0308216365902796,-.0519916499446463,.225963663364161,.00626330408481313},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.59519010006696,-6.06456651609707,7.41799655299993,-8.07092624092507,5.47681396366563,-9.23859424244377,11.300372436005,6.3309248170395,-10.6793559162843,5.76936797267538,.233717104755142,-.394247004353173,.482231155810537,-.524676988090528,1.04841335581819,-1.76852192969244,2.1632031816091,2.66785366107377,-4.50028385780377,5.04596922575242,.0249321870529789,-.0420569990798228,.0514428646412205,.15913874220381,-.268444076758866,.559597243722218,.00229823078236368,-.0038767834400574,.0189957623925199,.000196571591117826},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.92171958555151,-6.61537460386721,8.09172854779269,-8.80395991076744,6.96914454009643,-11.7559404153652,14.3795150585562,10.2353232484353,-17.2655122351262,13.4470270492568,.0451952183653594,-.0762378066864121,.0932518071891671,-.101459801467564,.217799593005347,-.367396460919218,.449388373324038,.614368644454251,-1.03635118137081,1.33333501160483,.000932318622845177,-.0015726868798065,.00192366360061394,.00636097475878139,-.0107300458251746,.0244039729141499,1.66187980931003e-005,-2.80335124459201e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.92171958555151,2.87937722321757,-2.67495037465683,1.63105841067898,2.19799643560933,-3.29333213710544,3.05951577632408,-2.29634628619806,3.44069303528758,2.23836106649518,3.41890639763464,-5.1226626812653,4.75896952776425,-2.90179486950883,-3.91413586924262,5.86468169257239,-5.4483074914535,4.09536544265748,-6.13622406614863,-4.00097073404889,5.33523806427656,-7.99396700420041,7.42642015557747,-5.59797201662203,8.38763014060611,5.49216459339238,7.19422103343416,-10.7793438389747,-7.09892453011936,9.00134910987289},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.59519010006696,2.39012709001286,-2.22043548283783,1.35391669466664,1.2388862761662,-1.85626506095126,1.72447600259692,-.514957807193037,.771578638346196,-.353775594121333,3.23038451124486,-4.84019398522881,4.49655523257725,-2.74178702515998,-2.62173967966463,3.92824092358218,-3.64934800797338,1.30644820527577,-1.95749537633152,.307822029727901,4.7630793830775,-7.13668236880667,6.6300000687783,-2.90243304119459,4.3488132461414,.518849452524734,6.06854814809742,-9.09271021667368,-2.66498561200528,7.17423567654379},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.03822870489707,1.55561306022354,-1.44516935978819,.881196025717106,-.0622566343920988,.0932812135041899,-.0866585368463816,1.10967175893133,-1.66265859611778,-1.41978880536247,2.90882269973526,-4.35838708560735,4.04895512766487,-2.46886161968117,-.707254507004454,1.0597032647262,-.984467621360889,-1.57948219926848,2.36659141141637,2.54719860772099,3.86201434387853,-5.78658625215441,5.37575658648661,.544351948717181,-.815620871022689,-3.18953813043078,4.43071582310306,-6.63869084478427,2.19278566166382,4.71658607960972},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-.300186182387741,.449779074319961,-.417846155642134,.25478285241737,-1.13611369814234,1.70227744362135,-1.58142102819233,1.22655366483195,-1.83778669504629,.100858250439743,2.48271365058504,-3.7199335363258,3.45582979910559,-2.10720180544033,1.26648273457434,-1.89761376598671,1.76288907720341,-2.5074905585056,3.75705761211831,.461151554581851,2.81340752869054,-4.21542331996591,3.91614134649752,3.19379348794833,-4.78536842321548,-2.33714259619681,2.75487590413567,-4.12772341388704,4.77320941254594,2.50302559510188},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,.553376983055106,-.829143384311425,.770276776736346,-.469678401195645,-1.45368745167752,2.17810890152299,-2.02346992933989,-.194424990185747,.291313515578485,1.49978166741586,1.9899087271236,-2.9815473107461,2.76986670414993,-1.68893390563516,2.74863027720701,-4.11836538252447,3.82597422031764,-.854309601049289,1.2800408674882,-2.52312426315122,1.80736427869175,-2.70803481201197,2.51577274454146,4.12997714955087,-6.18808395553405,1.65834138994758,1.41847476579741,-2.12534854800543,4.48820932686104,1.03297875354961},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,1.44662301694489,-2.16752402217829,2.01363654211465,-1.22782046694529,-.724355119014703,1.08532568728512,-1.00827093182807,-1.6918152058931,2.53490374291394,.160475142667942,1.4741928888764,-2.20883288938782,2.05201260878721,-1.25122037987572,3.38025060487184,-5.06474340721712,4.70516234202085,2.16048862424162,-3.23713291923203,-1.70770592621981,.991945941760336,-1.48626603590141,1.38074575986725,3.46032388717692,-5.18471990323708,4.72311662432345,.576746649127291,-.864158942299266,2.68886700736676,.311154445333738},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,2.30018618238774,-3.44644648080968,3.20175947449313,-1.95228172055831,.987081447023724,-1.47897739903412,1.37397459375255,-1.35561938023839,2.03117020639316,-2.20870753342451,.981387965414959,-1.47044666380812,1.36604951383154,-.832952480070547,3.10522366683209,-4.65266127660931,4.32233676392885,4.42237267757001,-6.6261900271284,2.83495589139377,.43960319187862,-.658672278252944,.611908590636855,2.03901059601618,-3.05511829665785,4.55407145300732,.170154848292172,-.254948743871771,1.03685515398485,.0611113213546687},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.03822870489707,-4.55228046671326,4.22908267863918,-2.57869489385804,3.26615690329693,-4.89379296535329,4.54634885284358,1.66498726415444,-2.49470653361407,-.923099517802526,.555278916264744,-.831993114526565,.772924185272261,-.471292665829703,2.17523616991868,-3.25922966624351,3.02783447385586,4.50024407433993,-6.74286736535737,6.26847807110811,.140734880491405,-.210867814600684,.195896854176436,.792696592497175,-1.18772402074228,2.35258131414952,.0308216365902796,-.0461810968751895,.225963663364161,.00626330408481313},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.59519010006696,-5.38679449650258,5.00434880168882,-3.05141556280758,5.47681396366563,-8.20609494322841,7.62348767011616,6.3309248170395,-9.48583802706066,5.76936797267538,.233717104755142,-.350186214905109,.325324080359881,-.198367260350894,1.04841335581819,-1.57087306611691,1.45934595309525,2.66785366107377,-3.99733505330865,5.04596922575242,.0249321870529789,-.0373567361461482,.0347045238135923,.15913874220381,-.23844294088706,.559597243722218,.00229823078236368,-.00344351663776842,.0189957623925199,.000196571591117826},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.92171958555151,-5.87604462970729,5.45886369350783,-3.32855727881992,6.96914454009643,-10.4421041409331,9.70074715430413,10.2353232484353,-15.3359297883224,13.4470270492568,.0451952183653594,-.067717518868618,.0629097851728847,-.0383594160020481,.217799593005347,-.326336470786676,.303167594035556,.614368644454251,-.920529245649199,1.33333501160483,.000932318622845177,-.00139692441436285,.00129774711619293,.00636097475878139,-.00953086286461726,.0244039729141499,1.66187980931003e-005,-2.49005052810442e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.92171958555151,2.26142559207414,-.826744191593017,-1.19523987507357,2.19799643560933,-2.5865404220042,.945601429143385,-2.29634628619806,2.70227576356801,2.23836106649518,3.41890639763464,-4.02327294720743,1.47084987158014,2.12643576380386,-3.91413586924262,4.60604507183384,-1.68390285488932,4.09536544265748,-4.81931093640565,-4.00097073404889,5.33523806427656,-6.27835818791885,2.29527612404422,-5.59797201662203,6.58753611853026,5.49216459339238,7.19422103343416,-8.46595709192435,-7.09892453011936,9.00134910987289},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.59519010006696,1.87717487173315,-.686267736267336,-.992150379408995,1.2388862761662,-1.4578865459724,.532982044151549,-.514957807193037,.605987872997064,-.353775594121333,3.23038451124486,-3.80142569043747,1.38974575226927,2.00918198880834,-2.62173967966463,3.08518956279543,-1.12790027705863,1.30644820527577,-1.53739152505215,.307822029727901,4.7630793830775,-5.60505793446857,2.0491273778591,-2.90243304119459,3.41550161982096,.518849452524734,6.06854814809742,-7.14129688166466,-2.66498561200528,7.17423567654379},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.03822870489707,1.22175835711557,-.446657024142499,-.64574059444934,-.0622566343920988,.0732618573826121,-.0267834658423847,1.10967175893133,-1.30583053509787,-1.41978880536247,2.90882269973526,-3.42302078938587,1.25140644309977,1.80918220620515,-.707254507004454,.832277224231996,-.304268406428259,-1.57948219926848,1.8586902544778,2.54719860772099,3.86201434387853,-4.54470992402429,1.66147962048301,.544351948717181,-.64057807233031,-3.18953813043078,4.43071582310306,-5.21394183895337,2.19278566166382,4.71658607960972},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-.300186182387741,.353250661721207,-.129143286331405,-.186704916697285,-1.13611369814234,1.33694666498547,-.488768188714079,1.22655366483195,-1.44337387662839,.100858250439743,2.48271365058504,-2.92158763778148,1.06808980107197,1.54415783407833,1.26648273457434,-1.49036128356923,.544854334050093,-2.5074905585056,2.95074440626681,.461151554581851,2.81340752869054,-3.31073889879025,1.21035782283977,3.19379348794833,-3.75836640385388,-2.33714259619681,2.75487590413567,-3.24186052799879,4.77320941254594,2.50302559510188},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,.553376983055106,-.651198479192506,.238068659934471,.344180410642771,-1.45368745167752,1.71065853147384,-.625391792982139,-.194424990185747,.228793863525153,1.49978166741586,1.9899087271236,-2.34166865603201,.856079885009688,1.23765104741929,2.74863027720701,-3.23451085132423,1.18248996086068,-.854309601049289,1.00532752474992,-2.52312426315122,1.80736427869175,-2.12685548066823,.77754732335343,4.12997714955087,-4.86004102210358,1.65834138994758,1.41847476579741,-1.66922123290476,4.48820932686104,1.03297875354961},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,1.44662301694489,-1.70234530427801,.622352597994011,.899747042727645,-.724355119014703,.852400743178548,-.311625271383912,-1.6918152058931,1.9908805805112,.160475142667942,1.4741928888764,-1.73478875376213,.634213449888056,.916894502821395,3.38025060487184,-3.97778389924568,1.45421973940992,2.16048862424162,-2.54240228389271,-1.70770592621981,.991945941760336,-1.16729410204273,.426745687640423,3.46032388717692,-4.07201188589473,4.72311662432345,.576746649127291,-.67869924509284,2.68886700736676,.311154445333738},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,2.30018618238774,-2.70679444519172,.989564544259887,1.4306323700677,.987081447023724,-1.16156970101249,.42465293021633,-1.35561938023839,1.5952547827211,-2.20870753342451,.981387965414959,-1.15486977201266,.422203533825773,.610387716162354,3.10522366683209,-3.65413992905702,1.33590022736237,4.42237267757001,-5.20412385861342,2.83495589139377,.43960319187862,-.517312679568431,.189121965733585,2.03901059601618,-2.39945035585008,4.55407145300732,.170154848292172,-.200233442638166,1.03685515398485,.0611113213546687},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.03822870489707,-3.57530214058608,1.30707828207098,1.88966804781976,3.26615690329693,-3.843521635486,1.40513541583971,1.66498726415444,-1.95931021275678,-.923099517802526,.555278916264744,-.653436620408273,.238886891797977,.345363344035539,2.17523616991868,-2.55975678291091,.935809721212703,4.50024407433993,-5.29576072944733,6.26847807110811,.140734880491405,-.165612851500527,.0605456414728778,.792696592497175,-.932823068533675,2.35258131414952,.0308216365902796,-.036270035583218,.225963663364161,.00626330408481313},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.59519010006696,-4.23071865520366,1.54668899419582,2.23607783277941,5.47681396366563,-6.44496072482988,2.35618358032458,6.3309248170395,-7.45005437415178,5.76936797267538,.233717104755142,-.275031719356679,.100547582628473,.145363561432352,1.04841335581819,-1.23374336817428,.451038569416845,2.66785366107377,-3.13945519772704,5.04596922575242,.0249321870529789,-.0293394969148429,.0107260918769642,.15913874220381,-.187269998673138,.559597243722218,.00229823078236368,-.00270449338439953,.0189957623925199,.000196571591117826},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.92171958555151,-4.61496937554466,1.6871654495215,2.43916732844399,6.96914454009643,-8.20109339929485,2.99820005631406,10.2353232484353,-12.0446407028288,13.4470270492568,.0451952183653594,-.0531844625867158,.0194434633176024,.028109786436831,.217799593005347,-.256300439106579,.0936996999356596,.614368644454251,-.722971751522571,1.33333501160483,.000932318622845177,-.00109712634918437,.000401093381258235,.00636097475878139,-.00748541629863617,.0244039729141499,1.66187980931003e-005,-1.95565344631038e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.92171958555151,1.44256348674145,.937852501329485,-2.27058656100761,2.19799643560933,-1.64995425135427,-1.07268327325537,-2.29634628619806,1.72378183093533,2.23836106649518,3.41890639763464,-2.56644599370298,-1.66852122478612,4.03957110939073,-3.91413586924262,2.93819635817362,1.91020695459719,4.09536544265748,-3.07423866276144,-4.00097073404889,5.33523806427656,-4.00496497002541,-2.60374427309723,-5.59797201662203,4.20218958584716,5.49216459339238,7.19422103343416,-5.40043440913241,-7.09892453011936,9.00134910987289},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.59519010006696,1.19744993497978,.77849704852463,-1.88477925223671,1.2388862761662,-.929985893147364,-.604610892164935,-.514957807193037,.386559691398865,-.353775594121333,3.23038451124486,-2.42492961864658,-1.57651731119676,3.81682515580894,-2.62173967966463,1.9680425606586,1.27948173914766,1.30644820527577,-.980702122872796,.307822029727901,4.7630793830775,-3.57546670811254,-2.32451495445634,-2.90243304119459,2.17874863666902,.518849452524734,6.06854814809742,-4.55543360151919,-2.66498561200528,7.17423567654379},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.03822870489707,.779359710871422,.506684427405858,-1.22670766448742,-.0622566343920988,.0467337421144522,.030382965722629,1.10967175893133,-.832989356981887,-1.41978880536247,2.90882269973526,-2.18354511527215,-1.41958622119803,3.4368873474629,-.707254507004454,.530909678713955,.345159831718241,-1.57948219926848,1.18565859680248,2.54719860772099,3.86201434387853,-2.89907066463143,-1.88477020319366,.544351948717181,-.408624781106939,-3.18953813043078,4.43071582310306,-3.32597373321893,2.19278566166382,4.71658607960972},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-.300186182387741,.225338612975934,.146499189649512,-.354681669820299,-1.13611369814234,.852838337634571,.554455021295674,1.22655366483195,-.920728260721081,.100858250439743,2.48271365058504,-1.86368088534508,-1.21163317718593,2.93342290468435,1.26648273457434,-.950701529718779,-.618078730089993,-2.5074905585056,1.88227998927126,.461151554581851,2.81340752869054,-2.11192049221879,-1.37302096972311,3.19379348794833,-2.39746209780118,-2.33714259619681,2.75487590413567,-2.06798297639769,4.77320941254594,2.50302559510188},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,.553376983055106,-.41539953912128,-.270063328509738,.65383646518608,-1.45368745167752,1.09122915405036,.709439828333213,-.194424990185747,.145947615817496,1.49978166741586,1.9899087271236,-1.49375054084377,-.971130695151503,2.35115468793597,2.74863027720701,-2.06329461731692,-1.34140787303666,-.854309601049289,.641298473185774,-2.52312426315122,1.80736427869175,-1.35672120663226,-.88204393756181,4.12997714955087,-3.10022038625728,1.65834138994758,1.41847476579741,-1.06479629943225,4.48820932686104,1.03297875354961},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,1.44662301694489,-1.0859261460488,-.705992187998236,1.70924145531706,-.724355119014703,.543746472353678,.353505404956933,-1.6918152058931,1.26998281171566,.160475142667942,1.4741928888764,-1.10662182393172,-.719447050735517,1.74181633275903,3.38025060487184,-2.53742852796533,-1.64965612574837,2.16048862424162,-1.62179853107076,-1.70770592621981,.991945941760336,-.744616960114543,-.484097154422056,3.46032388717692,-2.5975365648804,4.72311662432345,.576746649127291,-.432942278924123,2.68886700736676,.311154445333738},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,2.30018618238774,-1.72666429814602,-1.12255470615749,2.71775959032344,.987081447023724,-.740965364440283,-.481723145865576,-1.35561938023839,1.01761309753837,-2.20870753342451,.981387965414959,-.736691479430405,-.47894456870109,1.15954811601065,3.10522366683209,-2.33097602484761,-1.51543534565912,4.42237267757001,-3.31971084311917,2.83495589139377,.43960319187862,-.329993781528465,-.214538560322786,2.03901059601618,-1.53060948975261,4.55407145300732,.170154848292172,-.12772892205108,1.03685515398485,.0611113213546687},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.03822870489707,-2.2806853960415,-1.48273994391383,3.58978558499055,3.26615690329693,-2.45178262388351,-1.59397523182193,1.66498726415444,-1.24984407270733,-.923099517802526,.555278916264744,-.41682724950333,-.270991524688991,.656083673232097,2.17523616991868,-1.63286896677384,-1.06157563213922,4.50024407433993,-3.3781660071394,6.26847807110811,.140734880491405,-.10564444540962,-.0686825282564419,.792696592497175,-.595047877309364,2.35258131414952,.0308216365902796,-.0231366573273696,.225963663364161,.00626330408481313},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.59519010006696,-2.69877562014986,-1.7545525650326,4.24785717273984,5.47681396366563,-4.11124073580283,-2.67283724139748,6.3309248170395,-4.75239001984257,5.76936797267538,.233717104755142,-.175442746128902,-.114060434690259,.276145864886064,1.04841335581819,-.787004949964004,-.511654819985594,2.66785366107377,-2.00265861161846,5.04596922575242,.0249321870529789,-.0187156663995726,-.0121675993597458,.15913874220381,-.119459540563619,.559597243722218,.00229823078236368,-.00172519645136334,.0189957623925199,.000196571591117826},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.92171958555151,-2.94388917191153,-1.91390801783746,4.63366448151074,6.96914454009643,-5.23147784770941,-3.40113598727833,10.2353232484353,-7.68327684524188,13.4470270492568,.0451952183653594,-.0339263710725032,-.0220565211009044,.0533999113042758,.217799593005347,-.163494061616168,-.10629224716917,.614368644454251,-.461183712763862,1.33333501160483,.000932318622845177,-.000699856947415067,-.000454997367631841,.00636097475878139,-.00477494739305922,.0244039729141499,1.66187980931003e-005,-1.24751142122696e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.92171958555151,.49553061346938,2.00568935114481,-1.09346229146625,2.19799643560933,-.566770786786546,-2.29403815127598,-2.29634628619806,.592131075007482,2.23836106649518,3.41890639763464,-.881592088933561,-3.56829591856781,1.94536458449533,-3.91413586924262,1.00929093051433,4.08516450282659,4.09536544265748,-1.05602241045943,-4.00097073404889,5.33523806427656,-1.37573338488067,-5.56836192711791,-5.59797201662203,1.44348141522225,5.49216459339238,7.19422103343416,-1.855086864706,-7.09892453011936,9.00134910987289},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.59519010006696,.411332399809833,1.66489212099991,-.907666360512686,1.2388862761662,-.31945663701093,-1.29301955934683,-.514957807193037,.132785948612982,-.353775594121333,3.23038451124486,-.832980227624035,-3.37153654597113,1.8380952537413,-2.62173967966463,.676036338352017,2.73629690762123,1.30644820527577,-.336878015452627,.307822029727901,4.7630793830775,-1.22819773798136,-4.97120270417874,-2.90243304119459,.748415344244522,.518849452524734,6.06854814809742,-1.56482319698726,-2.66498561200528,7.17423567654379},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.03822870489707,.26771549341915,1.08359423150038,-.590754211622909,-.0622566343920988,.0160533661863334,.0649769454361842,1.10967175893133,-.286137650697852,-1.41978880536247,2.90882269973526,-.750062968079831,-3.03592405293217,1.65512593926409,-.707254507004454,.182371175485818,.738158077031132,-1.57948219926848,.407281992951262,2.54719860772099,3.86201434387853,-.995850982046505,-4.03076552074204,.544351948717181,-.140365460738301,-3.18953813043078,4.43071582310306,-1.14249516194806,2.19278566166382,4.71658607960972},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-.300186182387741,.0774053843401605,.313302853289657,-.170806538752116,-1.13611369814234,.292955913914665,1.18575632130325,1.22655366483195,-.316276575739705,.100858250439743,2.48271365058504,-.640187375400919,-2.59119611829218,1.41266903727833,1.26648273457434,-.326572602577083,-1.32182184869504,-2.5074905585056,.646576297249579,.461151554581851,2.81340752869054,-.725459410782454,-2.93633970583683,3.19379348794833,-.823544942743958,-2.33714259619681,2.75487590413567,-.71036656781973,4.77320941254594,2.50302559510188},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,.553376983055106,-.142692637341486,-.577556856071558,.314872611220466,-1.45368745167752,.374844821120283,1.51720649776892,-.194424990185747,.0501340234001082,1.49978166741586,1.9899087271236,-.513113723366546,-2.07685802519487,1.13226204929232,2.74863027720701,-.708756083794028,-2.86873200495546,-.854309601049289,.220290495881434,-2.52312426315122,1.80736427869175,-.466043191865359,-1.88633727617264,4.12997714955087,-1.06494731359279,1.65834138994758,1.41847476579741,-.365764951301697,4.48820932686104,1.03297875354961},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,1.44662301694489,-.373023200905715,-1.50983338153088,.823131392784558,-.724355119014703,.186780703583358,.756005902089247,-1.6918152058931,.436247948578379,.160475142667942,1.4741928888764,-.380132310523241,-1.53860792217029,.83881870492831,3.38025060487184,-.871624387178691,-3.52795105816102,2.16048862424162,-.557099100602201,-1.70770592621981,.991945941760336,-.255781116350547,-1.03528913786275,3.46032388717692,-.892271916858127,4.72311662432345,.576746649127291,-.148718690750111,2.68886700736676,.311154445333738},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,2.30018618238774,-.593121222587361,-2.40069309089209,1.30881054275714,.987081447023724,-.254526767782017,-1.03021208824726,-1.35561938023839,.349557192548049,-2.20870753342451,.981387965414959,-.253058658488868,-1.02426982907298,.558411716942304,3.10522366683209,-.800706513225493,-3.2409067852655,4.42237267757001,-1.14034381554164,2.83495589139377,.43960319187862,-.113355164264732,-.458811705720641,2.03901059601618,-.525775029251871,4.55407145300732,.170154848292172,-.0438757750930783,1.03685515398485,.0611113213546687},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.03822870489707,-.78343133166635,-3.17098446910282,1.72875821562793,3.26615690329693,-.842204422108959,-3.40887201763794,1.66498726415444,-.429330151346565,-.923099517802526,.555278916264744,-.143183065809956,-.579541894432988,.315954814956539,2.17523616991868,-.560901872435633,-2.27028337376948,4.50024407433993,-1.16042357185991,6.26847807110811,.140734880491405,-.0362896034209995,-.14688430786117,.792696592497175,-.204403093795768,2.35258131414952,.0308216365902796,-.00794760307219474,.225963663364161,.00626330408481313},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.59519010006696,-.927048238057034,-3.75228235860234,2.04567036451771,5.47681396366563,-1.4122398512488,-5.7161240011779,6.3309248170395,-1.63247909961861,5.76936797267538,.233717104755142,-.060265806265752,-.243929401394036,.132985500479333,1.04841335581819,-.27034168634063,-1.09422390242844,2.66785366107377,-.68792719315715,5.04596922575242,.0249321870529789,-.0064289618707996,-.0260216019365979,.15913874220381,-.0410351849082053,.559597243722218,.00229823078236368,-.000592617006985495,.0189957623925199,.000196571591117826},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.92171958555151,-1.01124645171658,-4.09307958874724,2.23146629547127,6.96914454009643,-1.7970491081004,-7.2736621396321,10.2353232484353,-2.63925915467191,13.4470270492568,.0451952183653594,-.0116539449562267,-.0471700287973566,.0257161697253048,.217799593005347,-.0561613498441344,-.227316372195247,.614368644454251,-.158419820138511,1.33333501160483,.000932318622845177,-.000240405739976666,-.000973056396143785,.00636097475878139,-.00164022771456071,.0244039729141499,1.66187980931003e-005,-4.28528869302687e-006,.000146461329045194,2.74870868673343e-007},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.92171958555151,-.49553061346938,2.00568935114481,1.09346229146626,2.19799643560933,.566770786786546,-2.29403815127598,-2.29634628619806,-.592131075007482,2.23836106649518,3.41890639763464,.881592088933561,-3.56829591856781,-1.94536458449534,-3.91413586924262,-1.00929093051433,4.08516450282659,4.09536544265748,1.05602241045943,-4.00097073404889,5.33523806427656,1.37573338488067,-5.56836192711792,-5.59797201662203,-1.44348141522225,5.49216459339238,7.19422103343416,1.855086864706,-7.09892453011936,9.00134910987289},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.59519010006696,-.411332399809833,1.66489212099991,.907666360512694,1.2388862761662,.31945663701093,-1.29301955934683,-.514957807193037,-.132785948612982,-.353775594121333,3.23038451124486,.832980227624035,-3.37153654597113,-1.83809525374131,-2.62173967966463,-.676036338352017,2.73629690762123,1.30644820527577,.336878015452627,.307822029727901,4.7630793830775,1.22819773798136,-4.97120270417875,-2.90243304119459,-.748415344244522,.518849452524734,6.06854814809742,1.56482319698726,-2.66498561200528,7.17423567654379},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.03822870489707,-.26771549341915,1.08359423150038,.590754211622914,-.0622566343920988,-.0160533661863334,.0649769454361843,1.10967175893133,.286137650697852,-1.41978880536247,2.90882269973526,.750062968079831,-3.03592405293218,-1.65512593926411,-.707254507004454,-.182371175485818,.738158077031133,-1.57948219926848,-.407281992951262,2.54719860772099,3.86201434387853,.995850982046505,-4.03076552074204,.544351948717181,.140365460738301,-3.18953813043078,4.43071582310306,1.14249516194806,2.19278566166382,4.71658607960972},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-.300186182387741,-.0774053843401605,.313302853289658,.170806538752117,-1.13611369814234,-.292955913914665,1.18575632130325,1.22655366483195,.316276575739705,.100858250439743,2.48271365058504,.640187375400919,-2.59119611829218,-1.41266903727834,1.26648273457434,.326572602577083,-1.32182184869504,-2.5074905585056,-.646576297249579,.461151554581851,2.81340752869054,.725459410782454,-2.93633970583683,3.19379348794833,.823544942743958,-2.33714259619681,2.75487590413567,.71036656781973,4.77320941254594,2.50302559510188},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,.553376983055106,.142692637341486,-.577556856071558,-.314872611220468,-1.45368745167752,-.374844821120283,1.51720649776892,-.194424990185747,-.0501340234001082,1.49978166741586,1.9899087271236,.513113723366546,-2.07685802519487,-1.13226204929233,2.74863027720701,.708756083794028,-2.86873200495547,-.854309601049289,-.220290495881434,-2.52312426315122,1.80736427869175,.466043191865359,-1.88633727617264,4.12997714955087,1.06494731359279,1.65834138994758,1.41847476579741,.365764951301697,4.48820932686104,1.03297875354961},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,1.44662301694489,.373023200905715,-1.50983338153088,-.823131392784565,-.724355119014703,-.186780703583358,.756005902089247,-1.6918152058931,-.436247948578379,.160475142667942,1.4741928888764,.380132310523241,-1.53860792217029,-.838818704928317,3.38025060487184,.871624387178691,-3.52795105816103,2.16048862424162,.557099100602201,-1.70770592621981,.991945941760336,.255781116350547,-1.03528913786275,3.46032388717692,.892271916858127,4.72311662432345,.576746649127291,.148718690750111,2.68886700736676,.311154445333738},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,2.30018618238774,.593121222587361,-2.40069309089209,-1.30881054275715,.987081447023724,.254526767782017,-1.03021208824726,-1.35561938023839,-.349557192548049,-2.20870753342451,.981387965414959,.253058658488868,-1.02426982907298,-.558411716942308,3.10522366683209,.800706513225493,-3.2409067852655,4.42237267757001,1.14034381554164,2.83495589139377,.43960319187862,.113355164264732,-.458811705720641,2.03901059601618,.525775029251871,4.55407145300732,.170154848292172,.0438757750930783,1.03685515398485,.0611113213546687},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.03822870489707,.78343133166635,-3.17098446910282,-1.72875821562795,3.26615690329693,.842204422108959,-3.40887201763794,1.66498726415444,.429330151346565,-.923099517802526,.555278916264744,.143183065809956,-.579541894432988,-.315954814956541,2.17523616991868,.560901872435633,-2.27028337376948,4.50024407433993,1.16042357185991,6.26847807110811,.140734880491405,.0362896034209995,-.14688430786117,.792696592497175,.204403093795768,2.35258131414952,.0308216365902796,.00794760307219474,.225963663364161,.00626330408481313},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.59519010006696,.927048238057034,-3.75228235860234,-2.04567036451773,5.47681396366563,1.4122398512488,-5.7161240011779,6.3309248170395,1.63247909961861,5.76936797267538,.233717104755142,.060265806265752,-.243929401394036,-.132985500479334,1.04841335581819,.27034168634063,-1.09422390242844,2.66785366107377,.68792719315715,5.04596922575242,.0249321870529789,.0064289618707996,-.0260216019365979,.15913874220381,.0410351849082053,.559597243722218,.00229823078236368,.000592617006985495,.0189957623925199,.000196571591117826},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.92171958555151,1.01124645171658,-4.09307958874725,-2.23146629547129,6.96914454009643,1.7970491081004,-7.2736621396321,10.2353232484353,2.63925915467191,13.4470270492568,.0451952183653594,.0116539449562267,-.0471700287973566,-.025716169725305,.217799593005347,.0561613498441344,-.227316372195247,.614368644454251,.158419820138511,1.33333501160483,.000932318622845177,.000240405739976666,-.000973056396143785,.00636097475878139,.00164022771456071,.0244039729141499,1.66187980931003e-005,4.28528869302687e-006,.000146461329045194,2.74870868673343e-007},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.92171958555151,-1.44256348674145,.937852501329483,2.27058656100761,2.19799643560933,1.64995425135427,-1.07268327325536,-2.29634628619806,-1.72378183093533,2.23836106649518,3.41890639763464,2.56644599370298,-1.66852122478611,-4.03957110939074,-3.91413586924262,-2.93819635817362,1.91020695459718,4.09536544265748,3.07423866276144,-4.00097073404889,5.33523806427656,4.00496497002541,-2.60374427309723,-5.59797201662203,-4.20218958584716,5.49216459339238,7.19422103343416,5.40043440913241,-7.09892453011936,9.00134910987289},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.59519010006696,-1.19744993497978,.778497048524628,1.88477925223671,1.2388862761662,.929985893147364,-.604610892164934,-.514957807193037,-.386559691398865,-.353775594121333,3.23038451124486,2.42492961864658,-1.57651731119676,-3.81682515580895,-2.62173967966463,-1.9680425606586,1.27948173914766,1.30644820527577,.980702122872796,.307822029727901,4.7630793830775,3.57546670811254,-2.32451495445633,-2.90243304119459,-2.17874863666902,.518849452524734,6.06854814809742,4.55543360151919,-2.66498561200528,7.17423567654379},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.03822870489707,-.779359710871422,.506684427405857,1.22670766448742,-.0622566343920988,-.0467337421144522,.030382965722629,1.10967175893133,.832989356981887,-1.41978880536247,2.90882269973526,2.18354511527215,-1.41958622119803,-3.43688734746291,-.707254507004454,-.530909678713955,.34515983171824,-1.57948219926848,-1.18565859680248,2.54719860772099,3.86201434387853,2.89907066463143,-1.88477020319366,.544351948717181,.408624781106939,-3.18953813043078,4.43071582310306,3.32597373321893,2.19278566166382,4.71658607960972},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-.300186182387741,-.225338612975934,.146499189649511,.3546816698203,-1.13611369814234,-.852838337634571,.554455021295673,1.22655366483195,.920728260721081,.100858250439743,2.48271365058504,1.86368088534508,-1.21163317718593,-2.93342290468436,1.26648273457434,.950701529718779,-.618078730089992,-2.5074905585056,-1.88227998927126,.461151554581851,2.81340752869054,2.11192049221879,-1.3730209697231,3.19379348794833,2.39746209780118,-2.33714259619681,2.75487590413567,2.06798297639769,4.77320941254594,2.50302559510188},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,.553376983055106,.41539953912128,-.270063328509737,-.653836465186081,-1.45368745167752,-1.09122915405036,.709439828333211,-.194424990185747,-.145947615817496,1.49978166741586,1.9899087271236,1.49375054084377,-.971130695151501,-2.35115468793598,2.74863027720701,2.06329461731692,-1.34140787303665,-.854309601049289,-.641298473185774,-2.52312426315122,1.80736427869175,1.35672120663226,-.882043937561808,4.12997714955087,3.10022038625728,1.65834138994758,1.41847476579741,1.06479629943225,4.48820932686104,1.03297875354961},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,1.44662301694489,1.0859261460488,-.705992187998234,-1.70924145531706,-.724355119014703,-.543746472353678,.353505404956932,-1.6918152058931,-1.26998281171566,.160475142667942,1.4741928888764,1.10662182393172,-.719447050735515,-1.74181633275904,3.38025060487184,2.53742852796533,-1.64965612574837,2.16048862424162,1.62179853107076,-1.70770592621981,.991945941760336,.744616960114543,-.484097154422054,3.46032388717692,2.5975365648804,4.72311662432345,.576746649127291,.432942278924123,2.68886700736676,.311154445333738},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,2.30018618238774,1.72666429814602,-1.12255470615748,-2.71775959032344,.987081447023724,.740965364440283,-.481723145865575,-1.35561938023839,-1.01761309753837,-2.20870753342451,.981387965414959,.736691479430405,-.478944568701089,-1.15954811601066,3.10522366683209,2.33097602484761,-1.51543534565911,4.42237267757001,3.31971084311917,2.83495589139377,.43960319187862,.329993781528465,-.214538560322786,2.03901059601618,1.53060948975261,4.55407145300732,.170154848292172,.12772892205108,1.03685515398485,.0611113213546687},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.03822870489707,2.2806853960415,-1.48273994391383,-3.58978558499056,3.26615690329693,2.45178262388351,-1.59397523182193,1.66498726415444,1.24984407270733,-.923099517802526,.555278916264744,.41682724950333,-.270991524688991,-.656083673232099,2.17523616991868,1.63286896677384,-1.06157563213921,4.50024407433993,3.3781660071394,6.26847807110811,.140734880491405,.10564444540962,-.0686825282564418,.792696592497175,.595047877309364,2.35258131414952,.0308216365902796,.0231366573273696,.225963663364161,.00626330408481313},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.59519010006696,2.69877562014986,-1.7545525650326,-4.24785717273985,5.47681396366563,4.11124073580283,-2.67283724139748,6.3309248170395,4.75239001984257,5.76936797267538,.233717104755142,.175442746128902,-.114060434690259,-.276145864886065,1.04841335581819,.787004949964004,-.511654819985593,2.66785366107377,2.00265861161846,5.04596922575242,.0249321870529789,.0187156663995726,-.0121675993597458,.15913874220381,.119459540563619,.559597243722218,.00229823078236368,.00172519645136334,.0189957623925199,.000196571591117826},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.92171958555151,2.94388917191153,-1.91390801783745,-4.63366448151075,6.96914454009643,5.23147784770941,-3.40113598727832,10.2353232484353,7.68327684524188,13.4470270492568,.0451952183653594,.0339263710725032,-.0220565211009043,-.0533999113042759,.217799593005347,.163494061616168,-.10629224716917,.614368644454251,.461183712763862,1.33333501160483,.000932318622845177,.000699856947415067,-.000454997367631839,.00636097475878139,.00477494739305922,.0244039729141499,1.66187980931003e-005,1.24751142122696e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.92171958555151,-2.26142559207414,-.826744191593013,1.19523987507356,2.19799643560933,2.5865404220042,.94560142914338,-2.29634628619806,-2.70227576356801,2.23836106649518,3.41890639763464,4.02327294720743,1.47084987158013,-2.12643576380383,-3.91413586924262,-4.60604507183384,-1.68390285488931,4.09536544265748,4.81931093640565,-4.00097073404889,5.33523806427656,6.27835818791885,2.29527612404421,-5.59797201662203,-6.58753611853026,5.49216459339238,7.19422103343416,8.46595709192435,-7.09892453011936,9.00134910987289},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.59519010006696,-1.87717487173315,-.686267736267333,.992150379408985,1.2388862761662,1.4578865459724,.532982044151546,-.514957807193037,-.605987872997064,-.353775594121333,3.23038451124486,3.80142569043747,1.38974575226926,-2.00918198880831,-2.62173967966463,-3.08518956279543,-1.12790027705863,1.30644820527577,1.53739152505215,.307822029727901,4.7630793830775,5.60505793446857,2.04912737785909,-2.90243304119459,-3.41550161982096,.518849452524734,6.06854814809742,7.14129688166466,-2.66498561200528,7.17423567654379},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.03822870489707,-1.22175835711557,-.446657024142497,.645740594449334,-.0622566343920988,-.0732618573826121,-.0267834658423845,1.10967175893133,1.30583053509787,-1.41978880536247,2.90882269973526,3.42302078938587,1.25140644309976,-1.80918220620513,-.707254507004454,-.832277224231996,-.304268406428257,-1.57948219926848,-1.8586902544778,2.54719860772099,3.86201434387853,4.54470992402429,1.661479620483,.544351948717181,.64057807233031,-3.18953813043078,4.43071582310306,5.21394183895337,2.19278566166382,4.71658607960972},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-.300186182387741,-.353250661721207,-.129143286331404,.186704916697283,-1.13611369814234,-1.33694666498547,-.488768188714077,1.22655366483195,1.44337387662839,.100858250439743,2.48271365058504,2.92158763778148,1.06808980107197,-1.54415783407832,1.26648273457434,1.49036128356923,.54485433405009,-2.5074905585056,-2.95074440626681,.461151554581851,2.81340752869054,3.31073889879025,1.21035782283977,3.19379348794833,3.75836640385388,-2.33714259619681,2.75487590413567,3.24186052799879,4.77320941254594,2.50302559510188},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,.553376983055106,.651198479192506,.23806865993447,-.344180410642768,-1.45368745167752,-1.71065853147384,-.625391792982136,-.194424990185747,-.228793863525153,1.49978166741586,1.9899087271236,2.34166865603201,.856079885009684,-1.23765104741928,2.74863027720701,3.23451085132423,1.18248996086068,-.854309601049289,-1.00532752474992,-2.52312426315122,1.80736427869175,2.12685548066823,.777547323353426,4.12997714955087,4.86004102210358,1.65834138994758,1.41847476579741,1.66922123290476,4.48820932686104,1.03297875354961},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,1.44662301694489,1.70234530427801,.622352597994008,-.899747042727636,-.724355119014703,-.852400743178548,-.31162527138391,-1.6918152058931,-1.9908805805112,.160475142667942,1.4741928888764,1.73478875376213,.634213449888053,-.916894502821386,3.38025060487184,3.97778389924568,1.45421973940992,2.16048862424162,2.54240228389271,-1.70770592621981,.991945941760336,1.16729410204273,.426745687640421,3.46032388717692,4.07201188589473,4.72311662432345,.576746649127291,.67869924509284,2.68886700736676,.311154445333738},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,2.30018618238774,2.70679444519172,.989564544259882,-1.43063237006769,.987081447023724,1.16156970101249,.424652930216328,-1.35561938023839,-1.5952547827211,-2.20870753342451,.981387965414959,1.15486977201266,.422203533825771,-.610387716162348,3.10522366683209,3.65413992905702,1.33590022736237,4.42237267757001,5.20412385861342,2.83495589139377,.43960319187862,.517312679568431,.189121965733584,2.03901059601618,2.39945035585008,4.55407145300732,.170154848292172,.200233442638166,1.03685515398485,.0611113213546687},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.03822870489707,3.57530214058608,1.30707828207097,-1.88966804781974,3.26615690329693,3.843521635486,1.4051354158397,1.66498726415444,1.95931021275678,-.923099517802526,.555278916264744,.653436620408273,.238886891797976,-.345363344035536,2.17523616991868,2.55975678291091,.935809721212699,4.50024407433993,5.29576072944733,6.26847807110811,.140734880491405,.165612851500527,.0605456414728775,.792696592497175,.932823068533675,2.35258131414952,.0308216365902796,.036270035583218,.225963663364161,.00626330408481313},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.59519010006696,4.23071865520366,1.54668899419581,-2.23607783277939,5.47681396366563,6.44496072482988,2.35618358032457,6.3309248170395,7.45005437415178,5.76936797267538,.233717104755142,.275031719356679,.100547582628472,-.145363561432351,1.04841335581819,1.23374336817428,.451038569416842,2.66785366107377,3.13945519772704,5.04596922575242,.0249321870529789,.0293394969148429,.0107260918769641,.15913874220381,.187269998673138,.559597243722218,.00229823078236368,.00270449338439953,.0189957623925199,.000196571591117826},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.92171958555151,4.61496937554466,1.68716544952149,-2.43916732844396,6.96914454009643,8.20109339929485,2.99820005631404,10.2353232484353,12.0446407028288,13.4470270492568,.0451952183653594,.0531844625867158,.0194434633176023,-.0281097864368307,.217799593005347,.256300439106579,.0936996999356592,.614368644454251,.722971751522571,1.33333501160483,.000932318622845177,.00109712634918437,.000401093381258233,.00636097475878139,.00748541629863617,.0244039729141499,1.66187980931003e-005,1.95565344631038e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.92171958555151,-2.87937722321757,-2.67495037465683,-1.63105841067899,2.19799643560933,3.29333213710544,3.05951577632408,-2.29634628619806,-3.44069303528758,2.23836106649518,3.41890639763464,5.1226626812653,4.75896952776425,2.90179486950885,-3.91413586924262,-5.86468169257239,-5.4483074914535,4.09536544265748,6.13622406614863,-4.00097073404889,5.33523806427656,7.99396700420041,7.42642015557747,-5.59797201662203,-8.38763014060611,5.49216459339238,7.19422103343416,10.7793438389747,-7.09892453011936,9.00134910987289},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.59519010006696,-2.39012709001286,-2.22043548283783,-1.35391669466665,1.2388862761662,1.85626506095126,1.72447600259692,-.514957807193037,-.771578638346196,-.353775594121333,3.23038451124486,4.84019398522881,4.49655523257725,2.74178702516,-2.62173967966463,-3.92824092358218,-3.64934800797338,1.30644820527577,1.95749537633152,.307822029727901,4.7630793830775,7.13668236880667,6.6300000687783,-2.90243304119459,-4.3488132461414,.518849452524734,6.06854814809742,9.09271021667368,-2.66498561200528,7.17423567654379},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.03822870489707,-1.55561306022354,-1.44516935978819,-.881196025717112,-.0622566343920988,-.0932812135041899,-.0866585368463816,1.10967175893133,1.66265859611778,-1.41978880536247,2.90882269973526,4.35838708560735,4.04895512766487,2.46886161968119,-.707254507004454,-1.0597032647262,-.984467621360889,-1.57948219926848,-2.36659141141637,2.54719860772099,3.86201434387853,5.78658625215441,5.37575658648661,.544351948717181,.815620871022689,-3.18953813043078,4.43071582310306,6.63869084478427,2.19278566166382,4.71658607960972},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-.300186182387741,-.449779074319961,-.417846155642134,-.254782852417372,-1.13611369814234,-1.70227744362135,-1.58142102819233,1.22655366483195,1.83778669504629,.100858250439743,2.48271365058504,3.7199335363258,3.45582979910559,2.10720180544034,1.26648273457434,1.89761376598671,1.76288907720341,-2.5074905585056,-3.75705761211831,.461151554581851,2.81340752869054,4.21542331996591,3.91614134649752,3.19379348794833,4.78536842321548,-2.33714259619681,2.75487590413567,4.12772341388704,4.77320941254594,2.50302559510188},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,.553376983055106,.829143384311425,.770276776736346,.469678401195648,-1.45368745167752,-2.17810890152299,-2.02346992933989,-.194424990185747,-.291313515578485,1.49978166741586,1.9899087271236,2.9815473107461,2.76986670414993,1.68893390563517,2.74863027720701,4.11836538252447,3.82597422031764,-.854309601049289,-1.2800408674882,-2.52312426315122,1.80736427869175,2.70803481201197,2.51577274454146,4.12997714955087,6.18808395553405,1.65834138994758,1.41847476579741,2.12534854800543,4.48820932686104,1.03297875354961},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,1.44662301694489,2.16752402217829,2.01363654211465,1.2278204669453,-.724355119014703,-1.08532568728512,-1.00827093182807,-1.6918152058931,-2.53490374291394,.160475142667942,1.4741928888764,2.20883288938782,2.05201260878721,1.25122037987573,3.38025060487184,5.06474340721712,4.70516234202085,2.16048862424162,3.23713291923203,-1.70770592621981,.991945941760336,1.48626603590141,1.38074575986725,3.46032388717692,5.18471990323708,4.72311662432345,.576746649127291,.864158942299266,2.68886700736676,.311154445333738},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,2.30018618238774,3.44644648080968,3.20175947449313,1.95228172055832,.987081447023724,1.47897739903412,1.37397459375255,-1.35561938023839,-2.03117020639316,-2.20870753342451,.981387965414959,1.47044666380812,1.36604951383154,.832952480070553,3.10522366683209,4.65266127660931,4.32233676392885,4.42237267757001,6.6261900271284,2.83495589139377,.43960319187862,.658672278252944,.611908590636855,2.03901059601618,3.05511829665785,4.55407145300732,.170154848292172,.254948743871771,1.03685515398485,.0611113213546687},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.03822870489707,4.55228046671326,4.22908267863918,2.57869489385806,3.26615690329693,4.89379296535329,4.54634885284358,1.66498726415444,2.49470653361407,-.923099517802526,.555278916264744,.831993114526565,.772924185272261,.471292665829706,2.17523616991868,3.25922966624351,3.02783447385586,4.50024407433993,6.74286736535737,6.26847807110811,.140734880491405,.210867814600684,.195896854176436,.792696592497175,1.18772402074228,2.35258131414952,.0308216365902796,.0461810968751895,.225963663364161,.00626330408481313},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.59519010006696,5.38679449650258,5.00434880168882,3.0514155628076,5.47681396366563,8.20609494322841,7.62348767011616,6.3309248170395,9.48583802706066,5.76936797267538,.233717104755142,.350186214905109,.325324080359881,.198367260350895,1.04841335581819,1.57087306611691,1.45934595309525,2.66785366107377,3.99733505330865,5.04596922575242,.0249321870529789,.0373567361461482,.0347045238135923,.15913874220381,.23844294088706,.559597243722218,.00229823078236368,.00344351663776842,.0189957623925199,.000196571591117826},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.92171958555151,5.87604462970729,5.45886369350783,3.32855727881994,6.96914454009643,10.4421041409331,9.70074715430413,10.2353232484353,15.3359297883224,13.4470270492568,.0451952183653594,.067717518868618,.0629097851728847,.0383594160020484,.217799593005347,.326336470786676,.303167594035556,.614368644454251,.920529245649199,1.33333501160483,.000932318622845177,.00139692441436285,.00129774711619293,.00636097475878139,.00953086286461726,.0244039729141499,1.66187980931003e-005,2.49005052810442e-005,.000146461329045194,2.74870868673343e-007},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.92171958555151,-3.24166342459793,-3.96510583993544,-4.31411318985288,2.19799643560933,3.70770257117539,4.53515100205683,-2.29634628619806,-3.87360456901059,2.23836106649518,3.41890639763464,5.76720136441661,7.05426838770182,7.67518283761778,-3.91413586924262,-6.60258197030424,-8.07608098333328,4.09536544265748,6.90829008438974,-4.00097073404889,5.33523806427656,8.99977614812198,11.0082572774727,-5.59797201662203,-9.44297038492047,5.49216459339238,7.19422103343416,12.1356119587979,-7.09892453011936,9.00134910987289},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.59519010006696,-2.69085533682779,-3.29137384514269,-3.58107952001051,1.2388862761662,2.0898222386161,2.55620812016103,-.514957807193037,-.868659455579414,-.353775594121333,3.23038451124486,5.44919216674985,6.66528903908045,7.25196565099482,-2.62173967966463,-4.42249623366662,-5.4094652504454,1.30644820527577,2.20378945631257,.307822029727901,4.7630793830775,8.03462707886587,9.82771577392637,-2.90243304119459,-4.89598539807522,.518849452524734,6.06854814809742,10.2367643607743,-2.66498561200528,7.17423567654379},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.03822870489707,-1.75134189417476,-2.14218907478874,-2.33074387312075,-.0622566343920988,-.105017951653011,-.12845482061473,1.10967175893133,1.87185600940662,-1.41978880536247,2.90882269973526,4.90676382786187,6.00180690245521,6.53008402866105,-.707254507004454,-1.19303621855815,-1.45928625505293,-1.57948219926848,-2.66435837496252,2.54719860772099,3.86201434387853,6.51466048131366,7.96853804732135,.544351948717181,.918243127233759,-3.18953813043078,4.43071582310306,7.47397774950178,2.19278566166382,4.71658607960972},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-.300186182387741,-.506370739691845,-.619377558413115,-.673894973328732,-1.13611369814234,-1.91645974101964,-2.34415629302528,1.22655366483195,2.06901890572263,.100858250439743,2.48271365058504,4.18797939686675,5.12261126340126,5.57350187032702,1.26648273457434,2.13637348021345,2.61314820693614,-2.5074905585056,-4.22977446202813,.461151554581851,2.81340752869054,4.74581221430261,5.80493569904152,3.19379348794833,5.38746839617942,-2.33714259619681,2.75487590413567,4.64707781591112,4.77320941254594,2.50302559510188},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,.553376983055106,.93346705704166,1.14178901214037,1.24228891639978,-1.45368745167752,-2.45216080196962,-2.99940982448642,-.194424990185747,-.327966881493408,1.49978166741586,1.9899087271236,3.35668865754018,4.10580126963162,4.46719258573777,2.74863027720701,4.6365423474252,5.67128006117502,-.854309601049289,-1.44109692494208,-2.52312426315122,1.80736427869175,3.04876253512471,3.72915523810848,4.12997714955087,6.96667503835439,1.65834138994758,1.41847476579741,2.39276208655032,4.48820932686104,1.03297875354961},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,1.44662301694489,2.44024412222762,2.98483369571688,3.24755780451479,-.724355119014703,-1.22188248065577,-1.49457014152702,-1.6918152058931,-2.85384793698481,.160475142667942,1.4741928888764,2.48675051356285,3.04171892525937,3.30945005334757,3.38025060487184,5.70199462778367,6.97450945518368,2.16048862424162,3.64443230995284,-1.70770592621981,.991945941760336,1.67326955598402,2.04669332477546,3.46032388717692,5.8370666898335,4.72311662432345,.576746649127291,.972888308521253,2.68886700736676,.311154445333738},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,2.30018618238774,3.88008191896113,4.74600026627036,5.1637416942433,.987081447023724,1.66506385533551,2.03665635715207,-1.35561938023839,-2.28673412920868,-2.20870753342451,.981387965414959,1.65545977423627,2.02490893148973,2.20314076875832,3.10522366683209,5.23806390000324,6.40704325104372,4.42237267757001,7.45990406612127,2.83495589139377,.43960319187862,.741547101224612,.907038257317998,2.03901059601618,3.43951642050873,4.55407145300732,.170154848292172,.287026656838237,1.03685515398485,.0611113213546687},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.03822870489707,5.12505307344404,6.26881178264598,6.82059059403532,3.26615690329693,5.50953502563763,6.7390986231311,1.66498726415444,2.80859307349989,-.923099517802526,.555278916264744,.936675343241156,1.14571329243577,1.24655861042429,2.17523616991868,3.66930929238228,4.48818948881768,4.50024407433993,7.59126186695628,6.26847807110811,.140734880491405,.237399419743997,.29037987690505,.792696592497175,1.33716467766395,2.35258131414952,.0308216365902796,.0519916499446463,.225963663364161,.00626330408481313},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.59519010006696,6.06456651609707,7.41799655299993,8.07092624092508,5.47681396366563,9.23859424244377,11.300372436005,6.3309248170395,10.6793559162843,5.76936797267538,.233717104755142,.394247004353173,.482231155810537,.524676988090528,1.04841335581819,1.76852192969244,2.1632031816091,2.66785366107377,4.50028385780377,5.04596922575242,.0249321870529789,.0420569990798228,.0514428646412205,.15913874220381,.268444076758866,.559597243722218,.00229823078236368,.0038767834400574,.0189957623925199,.000196571591117826},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.92171958555151,6.61537460386721,8.09172854779269,8.80395991076745,6.96914454009643,11.7559404153652,14.3795150585562,10.2353232484353,17.2655122351262,13.4470270492568,.0451952183653594,.0762378066864121,.0932518071891671,.101459801467564,.217799593005347,.367396460919218,.449388373324038,.614368644454251,1.03635118137081,1.33333501160483,.000932318622845177,.0015726868798065,.00192366360061394,.00636097475878139,.0107300458251746,.0244039729141499,1.66187980931003e-005,2.80335124459201e-005,.000146461329045194,2.74870868673343e-007}
		};
		for (int j = 0; j < 100; j++)
		{
			std::cout << timestep_dofs_to_hyp_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_hyp_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_left_face:" << std::endl;
		static const T timestep_dofs_to_left_face[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-3.37371117926928,5.6909635577302,-6.96101658051265,7.57372303615021,4.13193546224797,-6.96997843122346,8.52546935665732,-4.77114810448709,8.04823784418982,5.33430574565371,5.05405999869821,-8.52546935665732,10.4280993821027,-5.83592580124217,9.84436405649934,6.52476340023727,-6.34960212623238,10.7108618346039,7.09907099144837,7.15379270875817},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-2.99666740648972,5.05494516239668,-6.18305788326991,6.72728866290428,3.67015303548283,-6.19101816335907,7.5726684320545,-4.23792768749944,7.14877200382768,4.73814719602452,3.40958356039142,-5.75147508548906,7.03503247458549,-3.93704797266052,6.64123137798004,4.40175344784644,-2.40062592164153,4.04950925303536,2.68398137439781,.178027784911327},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-2.35354378347051,3.97008848462071,-4.85609361001944,5.28352541823675,2.88249067692314,-4.86234551094043,5.94747574387061,-3.32841353768903,5.61455297729064,3.72127946338712,1.05379652310981,-1.77760255482498,2.17431033156193,-1.21681941252376,2.05259862717067,1.36044546108577,1.7591790753139,-2.96748105525688,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-1.50132568517008,2.53251962263048,-3.09770233191153,3.3703611015687,1.83874093225775,-3.10169041978979,3.79389504410759,-2.12319514516607,3.58152359635514,2.37380433681192,-1.19541898829253,2.01649920176636,-2.46652157203873,1.38035094930968,-2.32845271402296,-1.54327927740643,3.34189684440344,-5.63729736986577,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,-.515715838247201,.869938143977334,-1.06408234437226,1.1577425323902,.631620327664824,-1.06545228035847,1.30322939400866,-.729332332662821,1.23027832131385,.81541833693141,-2.55652048864375,4.31248087502183,-5.27489775246548,2.95201558448519,-4.97962398835612,-3.30045375807884,1.60938069663531,-2.71479282295876,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,.515715838247201,-.869938143977334,1.06408234437226,-1.1577425323902,-.631620327664824,1.06545228035847,-1.30322939400866,.729332332662821,-1.23027832131385,-.81541833693141,-2.55652048864375,4.31248087502184,-5.27489775246549,2.95201558448519,-4.97962398835612,-3.30045375807885,-1.60938069663533,2.71479282295879,1.79934231949429,2.78946777547417},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,1.50132568517008,-2.53251962263048,3.09770233191153,-3.3703611015687,-1.83874093225775,3.10169041978979,-3.79389504410759,2.12319514516607,-3.58152359635514,-2.37380433681192,-1.19541898829252,2.01649920176635,-2.46652157203872,1.38035094930968,-2.32845271402296,-1.54327927740643,-3.34189684440345,5.63729736986578,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,2.35354378347051,-3.97008848462071,4.85609361001944,-5.28352541823675,-2.88249067692314,4.86234551094043,-5.94747574387061,3.32841353768903,-5.61455297729064,-3.72127946338712,1.0537965231098,-1.77760255482497,2.17431033156192,-1.21681941252376,2.05259862717066,1.36044546108577,-1.75917907531388,2.96748105525684,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,2.99666740648972,-5.05494516239668,6.18305788326991,-6.72728866290428,-3.67015303548283,6.19101816335907,-7.5726684320545,4.23792768749944,-7.14877200382768,-4.73814719602452,3.40958356039142,-5.75147508548906,7.03503247458549,-3.93704797266052,6.64123137798004,4.40175344784644,2.40062592164155,-4.0495092530354,-2.68398137439783,.178027784911158},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-2.,3.37371117926928,-4.12662270785725,4.48984672091457,2.449489743,-4.13193546224797,5.05405999869821,-2.828427124,4.77114810448709,3.162277660,3.37371117926928,-5.6909635577302,6.96101658051265,-7.57372303615021,-4.13193546224797,6.96997843122346,-8.52546935665732,4.77114810448709,-8.04823784418982,-5.33430574565371,5.05405999869821,-8.52546935665732,10.4280993821027,-5.83592580124217,9.84436405649934,6.52476340023727,6.34960212623239,-10.7108618346039,-7.09907099144838,7.15379270875832},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-3.37371117926928,5.05494516239668,-4.69605974259137,2.86343545350678,4.13193546224797,-6.19101816335907,5.75147508548906,-4.77114810448709,7.14877200382768,5.33430574565371,5.05405999869821,-7.5726684320545,7.03503247458549,-5.83592580124217,8.74416431584401,6.52476340023727,-6.34960212623238,9.51382286438758,7.09907099144837,7.15379270875817},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-2.99666740648972,4.4900077703237,-4.17123115221737,2.54341976480909,3.67015303548283,-5.4991139896991,5.10869396106863,-4.23792768749944,6.34982988227715,4.73814719602452,3.40958356039142,-5.10869396106863,4.74599254428019,-3.93704797266052,5.89901166751924,4.40175344784644,-2.40062592164153,3.59693872593977,2.68398137439781,.178027784911327},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-2.35354378347051,3.52639397108079,-3.27603094239261,1.99756895385147,2.88249067692314,-4.31893293096971,4.01230209521675,-3.32841353768903,4.98707417885748,3.72127946338712,1.05379652310981,-1.57893884647548,1.46683908849366,-1.21681941252376,1.82320153617303,1.36044546108577,1.7591790753139,-2.63583729760472,-1.96682199829962,-4.02051566640726},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-1.50132568517008,2.24948687251768,-2.08978028527405,1.27424932536978,1.83874093225775,-2.7550475106226,2.55944768672544,-2.12319514516607,3.18125484265547,2.37380433681192,-1.19541898829253,1.79113655913325,-1.66397142209237,1.38035094930968,-2.06822634923255,-1.54327927740643,3.34189684440344,-5.0072766728733,-3.73635425856115,-1.66034116472927},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,-.515715838247201,.772714421358283,-.717854095362718,.437713525759442,.631620327664824,-.946378024692648,.879188121703209,-.729332332662821,1.09278321423787,.81541833693141,-2.55652048864375,3.83052081004967,-3.55856572026934,2.95201558448519,-4.42310444183107,-3.30045375807884,1.60938069663531,-2.41138933822272,-1.79934231949427,2.78946777547416},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,.515715838247201,-.772714421358283,.717854095362718,-.437713525759442,-.631620327664824,.946378024692648,-.879188121703209,.729332332662821,-1.09278321423787,-.81541833693141,-2.55652048864375,3.83052081004967,-3.55856572026934,2.95201558448519,-4.42310444183107,-3.30045375807885,-1.60938069663533,2.41138933822274,1.79934231949429,2.78946777547417},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,1.50132568517008,-2.24948687251768,2.08978028527405,-1.27424932536978,-1.83874093225775,2.7550475106226,-2.55944768672544,2.12319514516607,-3.18125484265547,-2.37380433681192,-1.19541898829252,1.79113655913325,-1.66397142209237,1.38035094930968,-2.06822634923254,-1.54327927740643,-3.34189684440345,5.00727667287331,3.73635425856116,-1.66034116472922},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,2.35354378347051,-3.52639397108079,3.27603094239261,-1.99756895385147,-2.88249067692314,4.31893293096971,-4.01230209521675,3.32841353768903,-4.98707417885748,-3.72127946338712,1.0537965231098,-1.57893884647547,1.46683908849365,-1.21681941252376,1.82320153617302,1.36044546108577,-1.75917907531388,2.6358372976047,1.9668219982996,-4.02051566640715},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,2.99666740648972,-4.4900077703237,4.17123115221737,-2.54341976480909,-3.67015303548283,5.4991139896991,-5.10869396106863,4.23792768749944,-6.34982988227715,-4.73814719602452,3.40958356039142,-5.10869396106863,4.74599254428019,-3.93704797266052,5.89901166751924,4.40175344784644,2.40062592164155,-3.59693872593979,-2.68398137439783,.178027784911158},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-2.,2.99666740648972,-2.78391331885099,1.69749886814094,2.449489743,-3.67015303548283,3.40958356039142,-2.828427124,4.23792768749944,3.162277660,3.37371117926928,-5.05494516239668,4.69605974259137,-2.86343545350678,-4.13193546224797,6.19101816335907,-5.75147508548906,4.77114810448709,-7.14877200382768,-5.33430574565371,5.05405999869821,-7.5726684320545,7.03503247458549,-5.83592580124217,8.74416431584401,6.52476340023727,6.34960212623239,-9.51382286438759,-7.09907099144838,7.15379270875832},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-3.37371117926928,3.97008848462071,-1.45140640826254,-2.09832597736703,4.13193546224797,-4.86234551094043,1.77760255482498,-4.77114810448709,5.61455297729064,5.33430574565371,5.05405999869821,-5.94747574387061,2.17431033156193,-5.83592580124217,6.86755344374583,6.52476340023727,-6.34960212623238,7.47203330307115,7.09907099144837,7.15379270875817},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-2.99666740648972,3.52639397108079,-1.2891981696408,-1.86381842737598,3.67015303548283,-4.31893293096971,1.57893884647548,-4.23792768749944,4.98707417885748,4.73814719602452,3.40958356039142,-4.01230209521675,1.46683908849366,-3.93704797266052,4.63300738968997,4.40175344784644,-2.40062592164153,2.82498910610717,2.68398137439781,.178027784911327},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-2.35354378347051,2.7695841689776,-1.01251955130179,-1.46381886216961,2.88249067692314,-3.3920340071429,1.24007812764096,-3.32841353768903,3.91678349286862,3.72127946338712,1.05379652310981,-1.24007812764096,.453354465153136,-1.21681941252376,1.43191888168209,1.36044546108577,1.7591790753139,-2.07015248758758,-1.96682199829962,-4.02051566640726},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-1.50132568517008,1.76671786576882,-.645886267246198,-.933770117915978,1.83874093225775,-2.16377864548779,.791045893312282,-2.12319514516607,2.49851636599797,2.37380433681192,-1.19541898829253,1.40673546385746,-.514281954994446,1.38035094930968,-1.62435819754456,-1.54327927740643,3.34189684440344,-3.93265027010875,-3.73635425856115,-1.66034116472927},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,-.515715838247201,.606879902269876,-.221866435121632,-.320756544597897,.631620327664824,-.743273047921452,.271729778549238,-.729332332662821,.858257788295293,.81541833693141,-2.55652048864375,3.00844145080061,-1.09984228773296,2.95201558448519,-3.47384896307412,-3.30045375807884,1.60938069663531,-1.89387396619673,-1.79934231949427,2.78946777547416},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,.515715838247201,-.606879902269876,.221866435121632,.320756544597897,-.631620327664824,.743273047921452,-.271729778549238,.729332332662821,-.858257788295293,-.81541833693141,-2.55652048864375,3.00844145080061,-1.09984228773296,2.95201558448519,-3.47384896307413,-3.30045375807885,-1.60938069663533,1.89387396619674,1.79934231949429,2.78946777547417},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,1.50132568517008,-1.76671786576882,.645886267246198,.933770117915978,-1.83874093225775,2.16377864548779,-.791045893312282,2.12319514516607,-2.49851636599797,-2.37380433681192,-1.19541898829252,1.40673546385745,-.514281954994445,1.38035094930968,-1.62435819754456,-1.54327927740643,-3.34189684440345,3.93265027010876,3.73635425856116,-1.66034116472922},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,2.35354378347051,-2.7695841689776,1.01251955130179,1.46381886216961,-2.88249067692314,3.3920340071429,-1.24007812764096,3.32841353768903,-3.91678349286862,-3.72127946338712,1.0537965231098,-1.24007812764096,.453354465153134,-1.21681941252376,1.43191888168208,1.36044546108577,-1.75917907531388,2.07015248758756,1.9668219982996,-4.02051566640715},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,2.99666740648972,-3.52639397108079,1.2891981696408,1.86381842737598,-3.67015303548283,4.31893293096971,-1.57893884647548,4.23792768749944,-4.98707417885748,-4.73814719602452,3.40958356039142,-4.01230209521675,1.46683908849366,-3.93704797266052,4.63300738968997,4.40175344784644,2.40062592164155,-2.82498910610719,-2.68398137439783,.178027784911158},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-2.,2.35354378347051,-.860421257928482,-1.24392745337042,2.449489743,-2.88249067692314,1.05379652310981,-2.828427124,3.32841353768903,3.162277660,3.37371117926928,-3.97008848462071,1.45140640826254,2.09832597736703,-4.13193546224797,4.86234551094043,-1.77760255482498,4.77114810448709,-5.61455297729064,-5.33430574565371,5.05405999869821,-5.94747574387061,2.17431033156193,-5.83592580124217,6.86755344374583,6.52476340023727,6.34960212623239,-7.47203330307115,-7.09907099144838,7.15379270875832},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-3.37371117926928,2.53251962263048,1.64646470368521,-3.98617119808645,4.13193546224797,-3.10169041978979,-2.01649920176636,-4.77114810448709,3.58152359635514,5.33430574565371,5.05405999869821,-3.79389504410759,-2.46652157203873,-5.83592580124217,4.38081265017735,6.52476340023727,-6.34960212623238,4.76641037958728,7.09907099144837,7.15379270875817},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-2.99666740648972,2.24948687251768,1.4624568765065,-3.54067929092287,3.67015303548283,-2.7550475106226,-1.79113655913325,-4.23792768749944,3.18125484265547,4.73814719602452,3.40958356039142,-2.55944768672544,-1.66397142209237,-3.93704797266052,2.95539562194488,4.40175344784644,-2.40062592164153,1.80206067765193,2.68398137439781,.178027784911327},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-2.35354378347051,1.76671786576882,1.14859469650904,-2.78080367423081,2.88249067692314,-2.16377864548779,-1.40673546385746,-3.32841353768903,2.49851636599797,3.72127946338712,1.05379652310981,-.791045893312282,-.514281954994446,-1.21681941252376,.913421118930409,1.36044546108577,1.7591790753139,-1.32055036479967,-1.96682199829962,-4.02051566640726},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-1.50132568517008,1.12698940591467,.732688608484839,-1.7738747886737,1.83874093225775,-1.38027449512883,-.897356615569124,-2.12319514516607,1.59380370207485,2.37380433681192,-1.19541898829253,.897356615569124,.58339764921819,1.38035094930968,-1.03617816716125,-1.54327927740643,3.34189684440344,-2.508637783912,-3.73635425856115,-1.66034116472927},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,-.515715838247201,.387128716912051,.251683644415986,-.609338355176941,.631620327664824,-.47413391064841,-.308248252711719,-.729332332662821,.547482681696682,.81541833693141,-2.55652048864375,1.91908493657046,1.24765296340427,2.95201558448519,-2.21596840955051,-3.30045375807884,1.60938069663531,-1.20810228808803,-1.79934231949427,2.78946777547416},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,.515715838247201,-.387128716912051,-.251683644415986,.609338355176941,-.631620327664824,.47413391064841,.308248252711719,.729332332662821,-.547482681696682,-.81541833693141,-2.55652048864375,1.91908493657046,1.24765296340427,2.95201558448519,-2.21596840955051,-3.30045375807885,-1.60938069663533,1.20810228808804,1.79934231949429,2.78946777547417},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,1.50132568517008,-1.12698940591467,-.732688608484839,1.7738747886737,-1.83874093225775,1.38027449512883,.897356615569124,2.12319514516607,-1.59380370207485,-2.37380433681192,-1.19541898829252,.897356615569122,.583397649218189,1.38035094930968,-1.03617816716125,-1.54327927740643,-3.34189684440345,2.508637783912,3.73635425856116,-1.66034116472922},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,2.35354378347051,-1.76671786576882,-1.14859469650904,2.78080367423081,-2.88249067692314,2.16377864548779,1.40673546385746,3.32841353768903,-2.49851636599797,-3.72127946338712,1.0537965231098,-.791045893312278,-.514281954994444,-1.21681941252376,.913421118930405,1.36044546108577,-1.75917907531388,1.32055036479965,1.9668219982996,-4.02051566640715},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,2.99666740648972,-2.24948687251768,-1.4624568765065,3.54067929092287,-3.67015303548283,2.7550475106226,1.79113655913325,4.23792768749944,-3.18125484265547,-4.73814719602452,3.40958356039142,-2.55944768672544,-1.66397142209237,-3.93704797266052,2.95539562194488,4.40175344784644,2.40062592164155,-1.80206067765195,-2.68398137439783,.178027784911158},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-2.,1.50132568517008,.976055516507974,-2.36307792050314,2.449489743,-1.83874093225775,-1.19541898829253,-2.828427124,2.12319514516607,3.162277660,3.37371117926928,-2.53251962263048,-1.64646470368521,3.98617119808645,-4.13193546224797,3.10169041978979,2.01649920176636,4.77114810448709,-3.58152359635514,-5.33430574565371,5.05405999869821,-3.79389504410759,-2.46652157203873,-5.83592580124217,4.38081265017735,6.52476340023727,6.34960212623239,-4.76641037958728,-7.09907099144838,7.15379270875832},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-3.37371117926928,.869938143977334,3.52112588977045,-1.91964841477002,4.13193546224797,-1.06545228035847,-4.31248087502183,-4.77114810448709,1.23027832131385,5.33430574565371,5.05405999869821,-1.30322939400866,-5.27489775246548,-5.83592580124217,1.5048396829594,6.52476340023727,-6.34960212623238,1.63729519092357,7.09907099144837,7.15379270875817},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-2.99666740648972,.772714421358283,3.12760714457709,-1.70510975326197,3.67015303548283,-.946378024692648,-3.83052081004967,-4.23792768749944,1.09278321423787,4.73814719602452,3.40958356039142,-.879188121703209,-3.55856572026934,-3.93704797266052,1.01519899751182,4.40175344784644,-2.40062592164153,.619020404518225,2.68398137439781,.178027784911327},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-2.35354378347051,.606879902269876,2.45638215849919,-1.33917112430755,2.88249067692314,-.743273047921452,-3.00844145080061,-3.32841353768903,.858257788295293,3.72127946338712,1.05379652310981,-.271729778549238,-1.09984228773296,-1.21681941252376,.313766521598227,1.36044546108577,1.7591790753139,-.453618255557357,-1.96682199829962,-4.02051566640726},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-1.50132568517008,.387128716912051,1.5669262892192,-.854257320336024,1.83874093225775,-.47413391064841,-1.91908493657046,-2.12319514516607,.547482681696682,2.37380433681192,-1.19541898829253,.308248252711719,1.24765296340427,1.38035094930968,-.35593442337628,-1.54327927740643,3.34189684440344,-.861734565902807,-3.73635425856115,-1.66034116472927},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,-.515715838247201,.132981412843306,.538250103024586,-.293443344364012,.631620327664824,-.162868303384663,-.659219053205559,-.729332332662821,.188064117536924,.81541833693141,-2.55652048864375,.659219053205559,2.66822795597052,2.95201558448519,-.761200595679671,-3.30045375807884,1.60938069663531,-.414991557357587,-1.79934231949427,2.78946777547416},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,.515715838247201,-.132981412843306,-.538250103024586,.293443344364012,-.631620327664824,.162868303384663,.659219053205559,.729332332662821,-.188064117536924,-.81541833693141,-2.55652048864375,.659219053205559,2.66822795597052,2.95201558448519,-.761200595679671,-3.30045375807885,-1.60938069663533,.414991557357591,1.79934231949429,2.78946777547417},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,1.50132568517008,-.387128716912051,-1.5669262892192,.854257320336024,-1.83874093225775,.47413391064841,1.91908493657046,2.12319514516607,-.547482681696682,-2.37380433681192,-1.19541898829252,.308248252711718,1.24765296340427,1.38035094930968,-.355934423376279,-1.54327927740643,-3.34189684440345,.861734565902809,3.73635425856116,-1.66034116472922},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,2.35354378347051,-.606879902269876,-2.45638215849919,1.33917112430755,-2.88249067692314,.743273047921452,3.00844145080061,3.32841353768903,-.858257788295293,-3.72127946338712,1.0537965231098,-.271729778549237,-1.09984228773295,-1.21681941252376,.313766521598226,1.36044546108577,-1.75917907531388,.453618255557353,1.9668219982996,-4.02051566640715},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,2.99666740648972,-.772714421358283,-3.12760714457709,1.70510975326197,-3.67015303548283,.946378024692648,3.83052081004967,4.23792768749944,-1.09278321423787,-4.73814719602452,3.40958356039142,-.879188121703209,-3.55856572026934,-3.93704797266052,1.01519899751182,4.40175344784644,2.40062592164155,-.61902040451823,-2.68398137439783,.178027784911158},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-2.,.515715838247201,2.08739023760243,-1.13800400400502,2.449489743,-.631620327664824,-2.55652048864375,-2.828427124,.729332332662821,3.162277660,3.37371117926928,-.869938143977334,-3.52112588977045,1.91964841477002,-4.13193546224797,1.06545228035847,4.31248087502183,4.77114810448709,-1.23027832131385,-5.33430574565371,5.05405999869821,-1.30322939400866,-5.27489775246548,-5.83592580124217,1.5048396829594,6.52476340023727,6.34960212623239,-1.63729519092357,-7.09907099144838,7.15379270875832},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-3.37371117926928,-.869938143977334,3.52112588977045,1.91964841477004,4.13193546224797,1.06545228035847,-4.31248087502184,-4.77114810448709,-1.23027832131385,5.33430574565371,5.05405999869821,1.30322939400866,-5.27489775246549,-5.83592580124217,-1.5048396829594,6.52476340023727,-6.34960212623238,-1.63729519092357,7.09907099144837,7.15379270875817},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-2.99666740648972,-.772714421358283,3.12760714457709,1.70510975326198,3.67015303548283,.946378024692648,-3.83052081004967,-4.23792768749944,-1.09278321423787,4.73814719602452,3.40958356039142,.879188121703209,-3.55856572026934,-3.93704797266052,-1.01519899751182,4.40175344784644,-2.40062592164153,-.619020404518225,2.68398137439781,.178027784911327},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-2.35354378347051,-.606879902269876,2.45638215849919,1.33917112430757,2.88249067692314,.743273047921452,-3.00844145080061,-3.32841353768903,-.858257788295293,3.72127946338712,1.05379652310981,.271729778549238,-1.09984228773296,-1.21681941252376,-.313766521598227,1.36044546108577,1.7591790753139,.453618255557357,-1.96682199829962,-4.02051566640726},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-1.50132568517008,-.387128716912051,1.5669262892192,.854257320336031,1.83874093225775,.47413391064841,-1.91908493657046,-2.12319514516607,-.547482681696682,2.37380433681192,-1.19541898829253,-.308248252711719,1.24765296340427,1.38035094930968,.35593442337628,-1.54327927740643,3.34189684440344,.861734565902807,-3.73635425856115,-1.66034116472927},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,-.515715838247201,-.132981412843306,.538250103024586,.293443344364014,.631620327664824,.162868303384663,-.659219053205559,-.729332332662821,-.188064117536924,.81541833693141,-2.55652048864375,-.659219053205559,2.66822795597052,2.95201558448519,.761200595679671,-3.30045375807884,1.60938069663531,.414991557357587,-1.79934231949427,2.78946777547416},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,.515715838247201,.132981412843306,-.538250103024586,-.293443344364014,-.631620327664824,-.162868303384663,.659219053205559,.729332332662821,.188064117536924,-.81541833693141,-2.55652048864375,-.659219053205559,2.66822795597052,2.95201558448519,.761200595679671,-3.30045375807885,-1.60938069663533,-.414991557357591,1.79934231949429,2.78946777547417},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,1.50132568517008,.387128716912051,-1.5669262892192,-.854257320336031,-1.83874093225775,-.47413391064841,1.91908493657046,2.12319514516607,.547482681696682,-2.37380433681192,-1.19541898829252,-.308248252711718,1.24765296340427,1.38035094930968,.355934423376279,-1.54327927740643,-3.34189684440345,-.861734565902809,3.73635425856116,-1.66034116472922},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,2.35354378347051,.606879902269876,-2.45638215849919,-1.33917112430757,-2.88249067692314,-.743273047921452,3.00844145080061,3.32841353768903,.858257788295293,-3.72127946338712,1.0537965231098,.271729778549237,-1.09984228773295,-1.21681941252376,-.313766521598226,1.36044546108577,-1.75917907531388,-.453618255557353,1.9668219982996,-4.02051566640715},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,2.99666740648972,.772714421358283,-3.12760714457709,-1.70510975326198,-3.67015303548283,-.946378024692648,3.83052081004967,4.23792768749944,1.09278321423787,-4.73814719602452,3.40958356039142,.879188121703209,-3.55856572026934,-3.93704797266052,-1.01519899751182,4.40175344784644,2.40062592164155,.61902040451823,-2.68398137439783,.178027784911158},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-2.,-.515715838247201,2.08739023760243,1.13800400400503,2.449489743,.631620327664824,-2.55652048864375,-2.828427124,-.729332332662821,3.162277660,3.37371117926928,.869938143977334,-3.52112588977045,-1.91964841477004,-4.13193546224797,-1.06545228035847,4.31248087502184,4.77114810448709,1.23027832131385,-5.33430574565371,5.05405999869821,1.30322939400866,-5.27489775246549,-5.83592580124217,-1.5048396829594,6.52476340023727,6.34960212623239,1.63729519092357,-7.09907099144838,7.15379270875832},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-3.37371117926928,-2.53251962263048,1.64646470368521,3.98617119808646,4.13193546224797,3.10169041978979,-2.01649920176635,-4.77114810448709,-3.58152359635514,5.33430574565371,5.05405999869821,3.79389504410759,-2.46652157203872,-5.83592580124217,-4.38081265017735,6.52476340023727,-6.34960212623238,-4.76641037958728,7.09907099144837,7.15379270875817},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-2.99666740648972,-2.24948687251768,1.4624568765065,3.54067929092288,3.67015303548283,2.7550475106226,-1.79113655913325,-4.23792768749944,-3.18125484265547,4.73814719602452,3.40958356039142,2.55944768672544,-1.66397142209237,-3.93704797266052,-2.95539562194488,4.40175344784644,-2.40062592164153,-1.80206067765193,2.68398137439781,.178027784911327},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-2.35354378347051,-1.76671786576882,1.14859469650904,2.78080367423081,2.88249067692314,2.16377864548779,-1.40673546385745,-3.32841353768903,-2.49851636599797,3.72127946338712,1.05379652310981,.791045893312282,-.514281954994445,-1.21681941252376,-.913421118930409,1.36044546108577,1.7591790753139,1.32055036479967,-1.96682199829962,-4.02051566640726},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-1.50132568517008,-1.12698940591467,.732688608484838,1.7738747886737,1.83874093225775,1.38027449512883,-.897356615569122,-2.12319514516607,-1.59380370207485,2.37380433681192,-1.19541898829253,-.897356615569124,.583397649218189,1.38035094930968,1.03617816716125,-1.54327927740643,3.34189684440344,2.508637783912,-3.73635425856115,-1.66034116472927},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,-.515715838247201,-.387128716912051,.251683644415986,.609338355176942,.631620327664824,.47413391064841,-.308248252711718,-.729332332662821,-.547482681696682,.81541833693141,-2.55652048864375,-1.91908493657046,1.24765296340427,2.95201558448519,2.21596840955051,-3.30045375807884,1.60938069663531,1.20810228808803,-1.79934231949427,2.78946777547416},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,.515715838247201,.387128716912051,-.251683644415986,-.609338355176942,-.631620327664824,-.47413391064841,.308248252711718,.729332332662821,.547482681696682,-.81541833693141,-2.55652048864375,-1.91908493657046,1.24765296340427,2.95201558448519,2.21596840955051,-3.30045375807885,-1.60938069663533,-1.20810228808804,1.79934231949429,2.78946777547417},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,1.50132568517008,1.12698940591467,-.732688608484838,-1.7738747886737,-1.83874093225775,-1.38027449512883,.897356615569122,2.12319514516607,1.59380370207485,-2.37380433681192,-1.19541898829252,-.897356615569122,.583397649218187,1.38035094930968,1.03617816716125,-1.54327927740643,-3.34189684440345,-2.508637783912,3.73635425856116,-1.66034116472922},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,2.35354378347051,1.76671786576882,-1.14859469650904,-2.78080367423081,-2.88249067692314,-2.16377864548779,1.40673546385745,3.32841353768903,2.49851636599797,-3.72127946338712,1.0537965231098,.791045893312278,-.514281954994442,-1.21681941252376,-.913421118930405,1.36044546108577,-1.75917907531388,-1.32055036479965,1.9668219982996,-4.02051566640715},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,2.99666740648972,2.24948687251768,-1.4624568765065,-3.54067929092288,-3.67015303548283,-2.7550475106226,1.79113655913325,4.23792768749944,3.18125484265547,-4.73814719602452,3.40958356039142,2.55944768672544,-1.66397142209237,-3.93704797266052,-2.95539562194488,4.40175344784644,2.40062592164155,1.80206067765195,-2.68398137439783,.178027784911158},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-2.,-1.50132568517008,.976055516507971,2.36307792050314,2.449489743,1.83874093225775,-1.19541898829252,-2.828427124,-2.12319514516607,3.162277660,3.37371117926928,2.53251962263048,-1.64646470368521,-3.98617119808646,-4.13193546224797,-3.10169041978979,2.01649920176635,4.77114810448709,3.58152359635514,-5.33430574565371,5.05405999869821,3.79389504410759,-2.46652157203872,-5.83592580124217,-4.38081265017735,6.52476340023727,6.34960212623239,4.76641037958728,-7.09907099144838,7.15379270875832},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-3.37371117926928,-3.97008848462071,-1.45140640826253,2.098325977367,4.13193546224797,4.86234551094043,1.77760255482497,-4.77114810448709,-5.61455297729064,5.33430574565371,5.05405999869821,5.94747574387061,2.17431033156192,-5.83592580124217,-6.86755344374583,6.52476340023727,-6.34960212623238,-7.47203330307115,7.09907099144837,7.15379270875817},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-2.99666740648972,-3.52639397108079,-1.28919816964079,1.86381842737596,3.67015303548283,4.31893293096971,1.57893884647547,-4.23792768749944,-4.98707417885748,4.73814719602452,3.40958356039142,4.01230209521675,1.46683908849365,-3.93704797266052,-4.63300738968997,4.40175344784644,-2.40062592164153,-2.82498910610717,2.68398137439781,.178027784911327},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-2.35354378347051,-2.7695841689776,-1.01251955130178,1.46381886216959,2.88249067692314,3.3920340071429,1.24007812764096,-3.32841353768903,-3.91678349286862,3.72127946338712,1.05379652310981,1.24007812764096,.453354465153134,-1.21681941252376,-1.43191888168209,1.36044546108577,1.7591790753139,2.07015248758758,-1.96682199829962,-4.02051566640726},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-1.50132568517008,-1.76671786576882,-.645886267246195,.933770117915969,1.83874093225775,2.16377864548779,.791045893312278,-2.12319514516607,-2.49851636599797,2.37380433681192,-1.19541898829253,-1.40673546385746,-.514281954994444,1.38035094930968,1.62435819754456,-1.54327927740643,3.34189684440344,3.93265027010875,-3.73635425856115,-1.66034116472927},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,-.515715838247201,-.606879902269876,-.221866435121631,.320756544597894,.631620327664824,.743273047921452,.271729778549237,-.729332332662821,-.858257788295293,.81541833693141,-2.55652048864375,-3.00844145080061,-1.09984228773295,2.95201558448519,3.47384896307412,-3.30045375807884,1.60938069663531,1.89387396619673,-1.79934231949427,2.78946777547416},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,.515715838247201,.606879902269876,.221866435121631,-.320756544597894,-.631620327664824,-.743273047921452,-.271729778549237,.729332332662821,.858257788295293,-.81541833693141,-2.55652048864375,-3.00844145080061,-1.09984228773295,2.95201558448519,3.47384896307413,-3.30045375807885,-1.60938069663533,-1.89387396619674,1.79934231949429,2.78946777547417},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,1.50132568517008,1.76671786576882,.645886267246195,-.933770117915969,-1.83874093225775,-2.16377864548779,-.791045893312278,2.12319514516607,2.49851636599797,-2.37380433681192,-1.19541898829252,-1.40673546385745,-.514281954994442,1.38035094930968,1.62435819754456,-1.54327927740643,-3.34189684440345,-3.93265027010876,3.73635425856116,-1.66034116472922},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,2.35354378347051,2.7695841689776,1.01251955130178,-1.46381886216959,-2.88249067692314,-3.3920340071429,-1.24007812764096,3.32841353768903,3.91678349286862,-3.72127946338712,1.0537965231098,1.24007812764096,.453354465153132,-1.21681941252376,-1.43191888168208,1.36044546108577,-1.75917907531388,-2.07015248758756,1.9668219982996,-4.02051566640715},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,2.99666740648972,3.52639397108079,1.28919816964079,-1.86381842737596,-3.67015303548283,-4.31893293096971,-1.57893884647547,4.23792768749944,4.98707417885748,-4.73814719602452,3.40958356039142,4.01230209521675,1.46683908849365,-3.93704797266052,-4.63300738968997,4.40175344784644,2.40062592164155,2.82498910610719,-2.68398137439783,.178027784911158},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-2.,-2.35354378347051,-.860421257928478,1.2439274533704,2.449489743,2.88249067692314,1.0537965231098,-2.828427124,-3.32841353768903,3.162277660,3.37371117926928,3.97008848462071,1.45140640826253,-2.098325977367,-4.13193546224797,-4.86234551094043,-1.77760255482497,4.77114810448709,5.61455297729064,-5.33430574565371,5.05405999869821,5.94747574387061,2.17431033156192,-5.83592580124217,-6.86755344374583,6.52476340023727,6.34960212623239,7.47203330307115,-7.09907099144838,7.15379270875832},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-3.37371117926928,-5.05494516239668,-4.69605974259137,-2.8634354535068,4.13193546224797,6.19101816335907,5.75147508548906,-4.77114810448709,-7.14877200382768,5.33430574565371,5.05405999869821,7.5726684320545,7.03503247458549,-5.83592580124217,-8.74416431584401,6.52476340023727,-6.34960212623238,-9.51382286438758,7.09907099144837,7.15379270875817},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-2.99666740648972,-4.4900077703237,-4.17123115221737,-2.54341976480911,3.67015303548283,5.4991139896991,5.10869396106863,-4.23792768749944,-6.34982988227715,4.73814719602452,3.40958356039142,5.10869396106863,4.74599254428019,-3.93704797266052,-5.89901166751924,4.40175344784644,-2.40062592164153,-3.59693872593977,2.68398137439781,.178027784911327},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-2.35354378347051,-3.52639397108079,-3.27603094239261,-1.99756895385148,2.88249067692314,4.31893293096971,4.01230209521675,-3.32841353768903,-4.98707417885748,3.72127946338712,1.05379652310981,1.57893884647548,1.46683908849366,-1.21681941252376,-1.82320153617303,1.36044546108577,1.7591790753139,2.63583729760472,-1.96682199829962,-4.02051566640726},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-1.50132568517008,-2.24948687251768,-2.08978028527405,-1.27424932536979,1.83874093225775,2.7550475106226,2.55944768672544,-2.12319514516607,-3.18125484265547,2.37380433681192,-1.19541898829253,-1.79113655913325,-1.66397142209237,1.38035094930968,2.06822634923255,-1.54327927740643,3.34189684440344,5.0072766728733,-3.73635425856115,-1.66034116472927},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,-.515715838247201,-.772714421358283,-.717854095362718,-.437713525759445,.631620327664824,.946378024692648,.879188121703209,-.729332332662821,-1.09278321423787,.81541833693141,-2.55652048864375,-3.83052081004967,-3.55856572026934,2.95201558448519,4.42310444183107,-3.30045375807884,1.60938069663531,2.41138933822272,-1.79934231949427,2.78946777547416},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,.515715838247201,.772714421358283,.717854095362718,.437713525759445,-.631620327664824,-.946378024692648,-.879188121703209,.729332332662821,1.09278321423787,-.81541833693141,-2.55652048864375,-3.83052081004967,-3.55856572026934,2.95201558448519,4.42310444183107,-3.30045375807885,-1.60938069663533,-2.41138933822274,1.79934231949429,2.78946777547417},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,1.50132568517008,2.24948687251768,2.08978028527405,1.27424932536979,-1.83874093225775,-2.7550475106226,-2.55944768672544,2.12319514516607,3.18125484265547,-2.37380433681192,-1.19541898829252,-1.79113655913325,-1.66397142209237,1.38035094930968,2.06822634923254,-1.54327927740643,-3.34189684440345,-5.00727667287331,3.73635425856116,-1.66034116472922},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,2.35354378347051,3.52639397108079,3.27603094239261,1.99756895385148,-2.88249067692314,-4.31893293096971,-4.01230209521675,3.32841353768903,4.98707417885748,-3.72127946338712,1.0537965231098,1.57893884647547,1.46683908849365,-1.21681941252376,-1.82320153617302,1.36044546108577,-1.75917907531388,-2.6358372976047,1.9668219982996,-4.02051566640715},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,2.99666740648972,4.4900077703237,4.17123115221737,2.54341976480911,-3.67015303548283,-5.4991139896991,-5.10869396106863,4.23792768749944,6.34982988227715,-4.73814719602452,3.40958356039142,5.10869396106863,4.74599254428019,-3.93704797266052,-5.89901166751924,4.40175344784644,2.40062592164155,3.59693872593979,-2.68398137439783,.178027784911158},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-2.,-2.99666740648972,-2.78391331885099,-1.69749886814095,2.449489743,3.67015303548283,3.40958356039142,-2.828427124,-4.23792768749944,3.162277660,3.37371117926928,5.05494516239668,4.69605974259137,2.8634354535068,-4.13193546224797,-6.19101816335907,-5.75147508548906,4.77114810448709,7.14877200382768,-5.33430574565371,5.05405999869821,7.5726684320545,7.03503247458549,-5.83592580124217,-8.74416431584401,6.52476340023727,6.34960212623239,9.51382286438759,-7.09907099144838,7.15379270875832},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-3.37371117926928,-5.6909635577302,-6.96101658051265,-7.57372303615022,4.13193546224797,6.96997843122346,8.52546935665732,-4.77114810448709,-8.04823784418982,5.33430574565371,5.05405999869821,8.52546935665732,10.4280993821027,-5.83592580124217,-9.84436405649934,6.52476340023727,-6.34960212623238,-10.7108618346039,7.09907099144837,7.15379270875817},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-2.99666740648972,-5.05494516239668,-6.18305788326991,-6.72728866290429,3.67015303548283,6.19101816335907,7.5726684320545,-4.23792768749944,-7.14877200382768,4.73814719602452,3.40958356039142,5.75147508548906,7.03503247458549,-3.93704797266052,-6.64123137798004,4.40175344784644,-2.40062592164153,-4.04950925303536,2.68398137439781,.178027784911327},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-2.35354378347051,-3.97008848462071,-4.85609361001944,-5.28352541823676,2.88249067692314,4.86234551094043,5.94747574387061,-3.32841353768903,-5.61455297729064,3.72127946338712,1.05379652310981,1.77760255482498,2.17431033156193,-1.21681941252376,-2.05259862717067,1.36044546108577,1.7591790753139,2.96748105525688,-1.96682199829962,-4.02051566640726},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-1.50132568517008,-2.53251962263048,-3.09770233191153,-3.3703611015687,1.83874093225775,3.10169041978979,3.79389504410759,-2.12319514516607,-3.58152359635514,2.37380433681192,-1.19541898829253,-2.01649920176636,-2.46652157203873,1.38035094930968,2.32845271402296,-1.54327927740643,3.34189684440344,5.63729736986577,-3.73635425856115,-1.66034116472927},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,-.515715838247201,-.869938143977334,-1.06408234437226,-1.1577425323902,.631620327664824,1.06545228035847,1.30322939400866,-.729332332662821,-1.23027832131385,.81541833693141,-2.55652048864375,-4.31248087502183,-5.27489775246548,2.95201558448519,4.97962398835612,-3.30045375807884,1.60938069663531,2.71479282295876,-1.79934231949427,2.78946777547416},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,.515715838247201,.869938143977334,1.06408234437226,1.1577425323902,-.631620327664824,-1.06545228035847,-1.30322939400866,.729332332662821,1.23027832131385,-.81541833693141,-2.55652048864375,-4.31248087502184,-5.27489775246549,2.95201558448519,4.97962398835612,-3.30045375807885,-1.60938069663533,-2.71479282295879,1.79934231949429,2.78946777547417},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,1.50132568517008,2.53251962263048,3.09770233191153,3.3703611015687,-1.83874093225775,-3.10169041978979,-3.79389504410759,2.12319514516607,3.58152359635514,-2.37380433681192,-1.19541898829252,-2.01649920176635,-2.46652157203872,1.38035094930968,2.32845271402296,-1.54327927740643,-3.34189684440345,-5.63729736986578,3.73635425856116,-1.66034116472922},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,2.35354378347051,3.97008848462071,4.85609361001944,5.28352541823676,-2.88249067692314,-4.86234551094043,-5.94747574387061,3.32841353768903,5.61455297729064,-3.72127946338712,1.0537965231098,1.77760255482497,2.17431033156192,-1.21681941252376,-2.05259862717066,1.36044546108577,-1.75917907531388,-2.96748105525684,1.9668219982996,-4.02051566640715},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,2.99666740648972,5.05494516239668,6.18305788326991,6.72728866290429,-3.67015303548283,-6.19101816335907,-7.5726684320545,4.23792768749944,7.14877200382768,-4.73814719602452,3.40958356039142,5.75147508548906,7.03503247458549,-3.93704797266052,-6.64123137798004,4.40175344784644,2.40062592164155,4.0495092530354,-2.68398137439783,.178027784911158},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-2.,-3.37371117926928,-4.12662270785725,-4.48984672091457,2.449489743,4.13193546224797,5.05405999869821,-2.828427124,-4.77114810448709,3.162277660,3.37371117926928,5.6909635577302,6.96101658051265,7.57372303615022,-4.13193546224797,-6.96997843122346,-8.52546935665732,4.77114810448709,8.04823784418982,-5.33430574565371,5.05405999869821,8.52546935665732,10.4280993821027,-5.83592580124217,-9.84436405649934,6.52476340023727,6.34960212623239,10.7108618346039,-7.09907099144838,7.15379270875832}
		};
		for (int j = 0; j < 100; j++)
		{
			std::cout << timestep_dofs_to_left_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_left_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_right_face:" << std::endl;
		static const T timestep_dofs_to_right_face[100][35] = 
		{
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.92171958555151,-6.61537460386721,8.09172854779269,-8.80395991076744,6.96914454009643,-11.7559404153652,14.3795150585562,10.2353232484353,-17.2655122351262,13.4470270492568,-.0451952183653594,.0762378066864121,-.0932518071891671,.101459801467564,-.217799593005347,.367396460919218,-.449388373324038,-.614368644454251,1.03635118137081,-1.33333501160483,.000932318622845132,-.00157268687980642,.00192366360061384,.00636097475878108,-.0107300458251741,.0244039729141487,-1.66187980893617e-005,2.80335124396135e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.59519010006696,-6.06456651609707,7.41799655299993,-8.07092624092507,5.47681396366563,-9.23859424244377,11.300372436005,6.3309248170395,-10.6793559162843,5.76936797267538,-.233717104755142,.394247004353173,-.482231155810537,.524676988090528,-1.04841335581819,1.76852192969244,-2.1632031816091,-2.66785366107377,4.50028385780377,-5.04596922575242,.0249321870529788,-.0420569990798227,.0514428646412204,.159138742203809,-.268444076758865,.559597243722216,-.00229823078236368,.0038767834400574,-.0189957623925199,.00019657159111697},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,3.03822870489707,-5.12505307344404,6.26881178264598,-6.82059059403531,3.26615690329693,-5.50953502563763,6.7390986231311,1.66498726415444,-2.80859307349989,-.923099517802526,-.555278916264744,.936675343241156,-1.14571329243577,1.24655861042429,-2.17523616991868,3.66930929238228,-4.48818948881768,-4.50024407433993,7.59126186695628,-6.26847807110811,.140734880491405,-.237399419743997,.29037987690505,.792696592497175,-1.33716467766395,2.35258131414951,-.0308216365902829,.0519916499446519,-.225963663364185,.00626330408481366},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,2.30018618238774,-3.88008191896113,4.74600026627036,-5.1637416942433,.987081447023724,-1.66506385533551,2.03665635715207,-1.35561938023839,2.28673412920868,-2.20870753342451,-.981387965414959,1.65545977423627,-2.02490893148973,2.20314076875832,-3.10522366683209,5.23806390000324,-6.40704325104372,-4.42237267757001,7.45990406612127,-2.83495589139377,.439603191878621,-.741547101224613,.907038257318,2.03901059601619,-3.43951642050874,4.55407145300733,-.170154848292169,.287026656838231,-1.03685515398483,.0611113213546687},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,1.44662301694489,-2.44024412222762,2.98483369571688,-3.24755780451478,-.724355119014703,1.22188248065577,-1.49457014152702,-1.6918152058931,2.85384793698481,.160475142667942,-1.4741928888764,2.48675051356285,-3.04171892525937,3.30945005334757,-3.38025060487184,5.70199462778367,-6.97450945518368,-2.16048862424162,3.64443230995284,1.70770592621981,.991945941760336,-1.67326955598402,2.04669332477546,3.46032388717692,-5.8370666898335,4.72311662432345,-.576746649127296,.972888308521261,-2.68886700736678,.311154445333742},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,.553376983055106,-.93346705704166,1.14178901214037,-1.24228891639978,-1.45368745167752,2.45216080196962,-2.99940982448642,-.194424990185747,.327966881493408,1.49978166741586,-1.9899087271236,3.35668865754018,-4.10580126963162,4.46719258573777,-2.74863027720701,4.6365423474252,-5.67128006117502,.854309601049289,-1.44109692494208,2.52312426315122,1.80736427869175,-3.04876253512471,3.72915523810848,4.12997714955087,-6.96667503835439,1.65834138994758,-1.41847476579742,2.39276208655033,-4.48820932686106,1.03297875354965},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-.300186182387741,.506370739691845,-.619377558413115,.673894973328731,-1.13611369814234,1.91645974101964,-2.34415629302528,1.22655366483195,-2.06901890572263,.100858250439743,-2.48271365058504,4.18797939686675,-5.12261126340126,5.57350187032702,-1.26648273457434,2.13637348021345,-2.61314820693614,2.5074905585056,-4.22977446202813,-.461151554581851,2.81340752869054,-4.7458122143026,5.80493569904152,3.19379348794833,-5.38746839617942,-2.33714259619681,-2.75487590413568,4.64707781591113,-4.77320941254596,2.50302559510192},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.03822870489707,1.75134189417476,-2.14218907478874,2.33074387312075,-.0622566343920988,.105017951653011,-.12845482061473,1.10967175893133,-1.87185600940662,-1.41978880536247,-2.90882269973526,4.90676382786187,-6.00180690245521,6.53008402866105,.707254507004454,-1.19303621855815,1.45928625505293,1.57948219926848,-2.66435837496252,-2.54719860772099,3.86201434387853,-6.51466048131367,7.96853804732135,.544351948717181,-.918243127233759,-3.18953813043078,-4.43071582310304,7.47397774950175,-2.19278566166381,4.71658607960963},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.59519010006696,2.69085533682779,-3.29137384514269,3.5810795200105,1.2388862761662,-2.0898222386161,2.55620812016103,-.514957807193037,.868659455579414,-.353775594121333,-3.23038451124486,5.44919216674985,-6.66528903908045,7.25196565099481,2.62173967966463,-4.42249623366662,5.4094652504454,-1.30644820527577,2.20378945631257,-.307822029727901,4.7630793830775,-8.03462707886587,9.82771577392637,-2.90243304119458,4.89598539807521,.518849452524734,-6.0685481480974,10.2367643607743,2.66498561200528,7.17423567654393},
			{1.414213562,-2.38557405224355,2.91796290062108,-3.17480106311619,3.19927335807145,-1.92171958555151,3.24166342459793,-3.96510583993544,4.31411318985288,2.19799643560933,-3.70770257117539,4.53515100205683,-2.29634628619806,3.87360456901059,2.23836106649518,-3.41890639763464,5.76720136441661,-7.05426838770182,7.67518283761778,3.91413586924262,-6.60258197030424,8.07608098333328,-4.09536544265748,6.90829008438974,4.00097073404889,5.33523806427656,-8.99977614812197,11.0082572774727,-5.59797201662203,9.44297038492046,5.49216459339238,-7.19422103343416,12.1356119587979,7.09892453011937,9.00134910987265},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.92171958555151,-5.87604462970729,5.45886369350783,-3.32855727881992,6.96914454009643,-10.4421041409331,9.70074715430413,10.2353232484353,-15.3359297883224,13.4470270492568,-.0451952183653594,.067717518868618,-.0629097851728847,.0383594160020481,-.217799593005347,.326336470786676,-.303167594035556,-.614368644454251,.920529245649199,-1.33333501160483,.000932318622845132,-.00139692441436278,.00129774711619287,.00636097475878108,-.0095308628646168,.0244039729141487,-1.66187980893617e-005,2.49005052754424e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.59519010006696,-5.38679449650258,5.00434880168882,-3.05141556280758,5.47681396366563,-8.20609494322841,7.62348767011616,6.3309248170395,-9.48583802706066,5.76936797267538,-.233717104755142,.350186214905109,-.325324080359881,.198367260350894,-1.04841335581819,1.57087306611691,-1.45934595309525,-2.66785366107377,3.99733505330865,-5.04596922575242,.0249321870529788,-.0373567361461481,.0347045238135922,.159138742203809,-.238442940887059,.559597243722216,-.00229823078236368,.00344351663776842,-.0189957623925199,.00019657159111697},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,3.03822870489707,-4.55228046671326,4.22908267863918,-2.57869489385804,3.26615690329693,-4.89379296535329,4.54634885284358,1.66498726415444,-2.49470653361407,-.923099517802526,-.555278916264744,.831993114526565,-.772924185272261,.471292665829703,-2.17523616991868,3.25922966624351,-3.02783447385586,-4.50024407433993,6.74286736535737,-6.26847807110811,.140734880491405,-.210867814600683,.195896854176436,.792696592497175,-1.18772402074228,2.35258131414951,-.0308216365902829,.0461810968751945,-.225963663364185,.00626330408481366},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,2.30018618238774,-3.44644648080968,3.20175947449313,-1.95228172055831,.987081447023724,-1.47897739903412,1.37397459375255,-1.35561938023839,2.03117020639316,-2.20870753342451,-.981387965414959,1.47044666380812,-1.36604951383154,.832952480070547,-3.10522366683209,4.65266127660931,-4.32233676392885,-4.42237267757001,6.6261900271284,-2.83495589139377,.439603191878621,-.658672278252945,.611908590636856,2.03901059601619,-3.05511829665785,4.55407145300733,-.170154848292169,.254948743871766,-1.03685515398483,.0611113213546687},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,1.44662301694489,-2.16752402217829,2.01363654211465,-1.22782046694529,-.724355119014703,1.08532568728512,-1.00827093182807,-1.6918152058931,2.53490374291394,.160475142667942,-1.4741928888764,2.20883288938782,-2.05201260878721,1.25122037987572,-3.38025060487184,5.06474340721712,-4.70516234202085,-2.16048862424162,3.23713291923203,1.70770592621981,.991945941760336,-1.48626603590141,1.38074575986725,3.46032388717692,-5.18471990323707,4.72311662432345,-.576746649127296,.864158942299273,-2.68886700736678,.311154445333742},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,.553376983055106,-.829143384311425,.770276776736346,-.469678401195645,-1.45368745167752,2.17810890152299,-2.02346992933989,-.194424990185747,.291313515578485,1.49978166741586,-1.9899087271236,2.9815473107461,-2.76986670414993,1.68893390563516,-2.74863027720701,4.11836538252447,-3.82597422031764,.854309601049289,-1.2800408674882,2.52312426315122,1.80736427869175,-2.70803481201197,2.51577274454146,4.12997714955087,-6.18808395553405,1.65834138994758,-1.41847476579742,2.12534854800545,-4.48820932686106,1.03297875354965},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-.300186182387741,.449779074319961,-.417846155642134,.25478285241737,-1.13611369814234,1.70227744362135,-1.58142102819233,1.22655366483195,-1.83778669504629,.100858250439743,-2.48271365058504,3.7199335363258,-3.45582979910559,2.10720180544033,-1.26648273457434,1.89761376598671,-1.76288907720341,2.5074905585056,-3.75705761211831,-.461151554581851,2.81340752869054,-4.21542331996591,3.91614134649752,3.19379348794833,-4.78536842321548,-2.33714259619681,-2.75487590413568,4.12772341388705,-4.77320941254596,2.50302559510192},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.03822870489707,1.55561306022354,-1.44516935978819,.881196025717106,-.0622566343920988,.0932812135041899,-.0866585368463816,1.10967175893133,-1.66265859611778,-1.41978880536247,-2.90882269973526,4.35838708560735,-4.04895512766487,2.46886161968117,.707254507004454,-1.0597032647262,.984467621360889,1.57948219926848,-2.36659141141637,-2.54719860772099,3.86201434387853,-5.78658625215441,5.37575658648661,.544351948717181,-.815620871022689,-3.18953813043078,-4.43071582310304,6.63869084478425,-2.19278566166381,4.71658607960963},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.59519010006696,2.39012709001286,-2.22043548283783,1.35391669466664,1.2388862761662,-1.85626506095126,1.72447600259692,-.514957807193037,.771578638346196,-.353775594121333,-3.23038451124486,4.84019398522881,-4.49655523257725,2.74178702515998,2.62173967966463,-3.92824092358218,3.64934800797338,-1.30644820527577,1.95749537633152,-.307822029727901,4.7630793830775,-7.13668236880667,6.6300000687783,-2.90243304119458,4.3488132461414,.518849452524734,-6.0685481480974,9.09271021667365,2.66498561200528,7.17423567654393},
			{1.414213562,-2.11896384374972,1.96852398633026,-1.20031296082077,.0796164457723227,-1.92171958555151,2.87937722321757,-2.67495037465683,1.63105841067898,2.19799643560933,-3.29333213710544,3.05951577632408,-2.29634628619806,3.44069303528758,2.23836106649518,-3.41890639763464,5.1226626812653,-4.75896952776425,2.90179486950883,3.91413586924262,-5.86468169257239,5.4483074914535,-4.09536544265748,6.13622406614863,4.00097073404889,5.33523806427656,-7.99396700420041,7.42642015557747,-5.59797201662203,8.38763014060611,5.49216459339238,-7.19422103343416,10.7793438389747,7.09892453011937,9.00134910987265},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.92171958555151,-4.61496937554466,1.6871654495215,2.43916732844399,6.96914454009643,-8.20109339929485,2.99820005631406,10.2353232484353,-12.0446407028288,13.4470270492568,-.0451952183653594,.0531844625867158,-.0194434633176024,-.028109786436831,-.217799593005347,.256300439106579,-.0936996999356596,-.614368644454251,.722971751522571,-1.33333501160483,.000932318622845132,-.00109712634918432,.000401093381258216,.00636097475878108,-.00748541629863581,.0244039729141487,-1.66187980893617e-005,1.95565344587043e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.59519010006696,-4.23071865520366,1.54668899419582,2.23607783277941,5.47681396366563,-6.44496072482988,2.35618358032458,6.3309248170395,-7.45005437415178,5.76936797267538,-.233717104755142,.275031719356679,-.100547582628473,-.145363561432352,-1.04841335581819,1.23374336817428,-.451038569416845,-2.66785366107377,3.13945519772704,-5.04596922575242,.0249321870529788,-.0293394969148428,.0107260918769642,.159138742203809,-.187269998673138,.559597243722216,-.00229823078236368,.00270449338439953,-.0189957623925199,.00019657159111697},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,3.03822870489707,-3.57530214058608,1.30707828207098,1.88966804781976,3.26615690329693,-3.843521635486,1.40513541583971,1.66498726415444,-1.95931021275678,-.923099517802526,-.555278916264744,.653436620408273,-.238886891797977,-.345363344035539,-2.17523616991868,2.55975678291091,-.935809721212703,-4.50024407433993,5.29576072944733,-6.26847807110811,.140734880491405,-.165612851500527,.0605456414728777,.792696592497175,-.932823068533675,2.35258131414951,-.0308216365902829,.0362700355832219,-.225963663364185,.00626330408481366},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,2.30018618238774,-2.70679444519172,.989564544259887,1.4306323700677,.987081447023724,-1.16156970101249,.42465293021633,-1.35561938023839,1.5952547827211,-2.20870753342451,-.981387965414959,1.15486977201266,-.422203533825773,-.610387716162354,-3.10522366683209,3.65413992905702,-1.33590022736237,-4.42237267757001,5.20412385861342,-2.83495589139377,.439603191878621,-.517312679568432,.189121965733585,2.03901059601619,-2.39945035585009,4.55407145300733,-.170154848292169,.200233442638163,-1.03685515398483,.0611113213546687},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,1.44662301694489,-1.70234530427801,.622352597994011,.899747042727645,-.724355119014703,.852400743178548,-.311625271383912,-1.6918152058931,1.9908805805112,.160475142667942,-1.4741928888764,1.73478875376213,-.634213449888056,-.916894502821395,-3.38025060487184,3.97778389924568,-1.45421973940992,-2.16048862424162,2.54240228389271,1.70770592621981,.991945941760336,-1.16729410204273,.426745687640423,3.46032388717692,-4.07201188589473,4.72311662432345,-.576746649127296,.678699245092846,-2.68886700736678,.311154445333742},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,.553376983055106,-.651198479192506,.238068659934471,.344180410642771,-1.45368745167752,1.71065853147384,-.625391792982139,-.194424990185747,.228793863525153,1.49978166741586,-1.9899087271236,2.34166865603201,-.856079885009688,-1.23765104741929,-2.74863027720701,3.23451085132423,-1.18248996086068,.854309601049289,-1.00532752474992,2.52312426315122,1.80736427869175,-2.12685548066823,.77754732335343,4.12997714955087,-4.86004102210358,1.65834138994758,-1.41847476579742,1.66922123290477,-4.48820932686106,1.03297875354965},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-.300186182387741,.353250661721207,-.129143286331405,-.186704916697285,-1.13611369814234,1.33694666498547,-.488768188714079,1.22655366483195,-1.44337387662839,.100858250439743,-2.48271365058504,2.92158763778148,-1.06808980107197,-1.54415783407833,-1.26648273457434,1.49036128356923,-.544854334050093,2.5074905585056,-2.95074440626681,-.461151554581851,2.81340752869054,-3.31073889879025,1.21035782283977,3.19379348794833,-3.75836640385388,-2.33714259619681,-2.75487590413568,3.2418605279988,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.03822870489707,1.22175835711557,-.446657024142499,-.64574059444934,-.0622566343920988,.0732618573826121,-.0267834658423847,1.10967175893133,-1.30583053509787,-1.41978880536247,-2.90882269973526,3.42302078938587,-1.25140644309977,-1.80918220620515,.707254507004454,-.832277224231996,.304268406428259,1.57948219926848,-1.8586902544778,-2.54719860772099,3.86201434387853,-4.54470992402429,1.66147962048301,.544351948717181,-.640578072330311,-3.18953813043078,-4.43071582310304,5.21394183895335,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.59519010006696,1.87717487173315,-.686267736267336,-.992150379408995,1.2388862761662,-1.4578865459724,.532982044151549,-.514957807193037,.605987872997064,-.353775594121333,-3.23038451124486,3.80142569043747,-1.38974575226927,-2.00918198880834,2.62173967966463,-3.08518956279543,1.12790027705863,-1.30644820527577,1.53739152505215,-.307822029727901,4.7630793830775,-5.60505793446857,2.0491273778591,-2.90243304119458,3.41550161982096,.518849452524734,-6.0685481480974,7.14129688166464,2.66498561200528,7.17423567654393},
			{1.414213562,-1.66420676884452,.608409706261881,.879589537656949,-1.79802926655929,-1.92171958555151,2.26142559207414,-.826744191593017,-1.19523987507357,2.19799643560933,-2.5865404220042,.945601429143385,-2.29634628619806,2.70227576356801,2.23836106649518,-3.41890639763464,4.02327294720743,-1.47084987158014,-2.12643576380386,3.91413586924262,-4.60604507183384,1.68390285488932,-4.09536544265748,4.81931093640565,4.00097073404889,5.33523806427656,-6.27835818791885,2.29527612404422,-5.59797201662203,6.58753611853026,5.49216459339238,-7.19422103343416,8.46595709192435,7.09892453011937,9.00134910987265},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.92171958555151,-2.94388917191153,-1.91390801783746,4.63366448151074,6.96914454009643,-5.23147784770941,-3.40113598727833,10.2353232484353,-7.68327684524188,13.4470270492568,-.0451952183653594,.0339263710725032,.0220565211009044,-.0533999113042758,-.217799593005347,.163494061616168,.10629224716917,-.614368644454251,.461183712763862,-1.33333501160483,.000932318622845132,-.000699856947415033,-.000454997367631818,.00636097475878108,-.00477494739305899,.0244039729141487,-1.66187980893617e-005,1.24751142094631e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.59519010006696,-2.69877562014986,-1.7545525650326,4.24785717273984,5.47681396366563,-4.11124073580283,-2.67283724139748,6.3309248170395,-4.75239001984257,5.76936797267538,-.233717104755142,.175442746128902,.114060434690259,-.276145864886064,-1.04841335581819,.787004949964004,.511654819985594,-2.66785366107377,2.00265861161846,-5.04596922575242,.0249321870529788,-.0187156663995725,-.0121675993597458,.159138742203809,-.119459540563619,.559597243722216,-.00229823078236368,.00172519645136334,-.0189957623925199,.00019657159111697},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,3.03822870489707,-2.2806853960415,-1.48273994391383,3.58978558499055,3.26615690329693,-2.45178262388351,-1.59397523182193,1.66498726415444,-1.24984407270733,-.923099517802526,-.555278916264744,.41682724950333,.270991524688991,-.656083673232097,-2.17523616991868,1.63286896677384,1.06157563213922,-4.50024407433993,3.3781660071394,-6.26847807110811,.140734880491405,-.105644445409619,-.0686825282564419,.792696592497175,-.595047877309364,2.35258131414951,-.0308216365902829,.0231366573273721,-.225963663364185,.00626330408481366},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,2.30018618238774,-1.72666429814602,-1.12255470615749,2.71775959032344,.987081447023724,-.740965364440283,-.481723145865576,-1.35561938023839,1.01761309753837,-2.20870753342451,-.981387965414959,.736691479430405,.47894456870109,-1.15954811601065,-3.10522366683209,2.33097602484761,1.51543534565912,-4.42237267757001,3.31971084311917,-2.83495589139377,.439603191878621,-.329993781528466,-.214538560322787,2.03901059601619,-1.53060948975261,4.55407145300733,-.170154848292169,.127728922051078,-1.03685515398483,.0611113213546687},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,1.44662301694489,-1.0859261460488,-.705992187998236,1.70924145531706,-.724355119014703,.543746472353678,.353505404956933,-1.6918152058931,1.26998281171566,.160475142667942,-1.4741928888764,1.10662182393172,.719447050735517,-1.74181633275903,-3.38025060487184,2.53742852796533,1.64965612574837,-2.16048862424162,1.62179853107076,1.70770592621981,.991945941760336,-.744616960114543,-.484097154422055,3.46032388717692,-2.5975365648804,4.72311662432345,-.576746649127296,.432942278924126,-2.68886700736678,.311154445333742},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,.553376983055106,-.41539953912128,-.270063328509738,.65383646518608,-1.45368745167752,1.09122915405036,.709439828333213,-.194424990185747,.145947615817496,1.49978166741586,-1.9899087271236,1.49375054084377,.971130695151503,-2.35115468793597,-2.74863027720701,2.06329461731692,1.34140787303666,.854309601049289,-.641298473185774,2.52312426315122,1.80736427869175,-1.35672120663226,-.88204393756181,4.12997714955087,-3.10022038625728,1.65834138994758,-1.41847476579742,1.06479629943226,-4.48820932686106,1.03297875354965},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-.300186182387741,.225338612975934,.146499189649512,-.354681669820299,-1.13611369814234,.852838337634571,.554455021295674,1.22655366483195,-.920728260721081,.100858250439743,-2.48271365058504,1.86368088534508,1.21163317718593,-2.93342290468435,-1.26648273457434,.950701529718779,.618078730089993,2.5074905585056,-1.88227998927126,-.461151554581851,2.81340752869054,-2.11192049221879,-1.37302096972311,3.19379348794833,-2.39746209780118,-2.33714259619681,-2.75487590413568,2.0679829763977,-4.77320941254596,2.50302559510192},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.03822870489707,.779359710871422,.506684427405858,-1.22670766448742,-.0622566343920988,.0467337421144522,.030382965722629,1.10967175893133,-.832989356981887,-1.41978880536247,-2.90882269973526,2.18354511527215,1.41958622119803,-3.4368873474629,.707254507004454,-.530909678713955,-.345159831718241,1.57948219926848,-1.18565859680248,-2.54719860772099,3.86201434387853,-2.89907066463143,-1.88477020319366,.544351948717181,-.408624781106939,-3.18953813043078,-4.43071582310304,3.32597373321892,-2.19278566166381,4.71658607960963},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.59519010006696,1.19744993497978,.77849704852463,-1.88477925223671,1.2388862761662,-.929985893147364,-.604610892164935,-.514957807193037,.386559691398865,-.353775594121333,-3.23038451124486,2.42492961864658,1.57651731119676,-3.81682515580894,2.62173967966463,-1.9680425606586,-1.27948173914766,-1.30644820527577,.980702122872796,-.307822029727901,4.7630793830775,-3.57546670811253,-2.32451495445634,-2.90243304119458,2.17874863666902,.518849452524734,-6.0685481480974,4.55543360151918,2.66498561200528,7.17423567654393},
			{1.414213562,-1.06159757258303,-.69017547465484,1.67094842220172,-.742527141878811,-1.92171958555151,1.44256348674145,.937852501329485,-2.27058656100761,2.19799643560933,-1.64995425135427,-1.07268327325537,-2.29634628619806,1.72378183093533,2.23836106649518,-3.41890639763464,2.56644599370298,1.66852122478612,-4.03957110939073,3.91413586924262,-2.93819635817362,-1.91020695459719,-4.09536544265748,3.07423866276144,4.00097073404889,5.33523806427656,-4.00496497002541,-2.60374427309723,-5.59797201662203,4.20218958584716,5.49216459339238,-7.19422103343416,5.40043440913241,7.09892453011937,9.00134910987265},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.92171958555151,-1.01124645171658,-4.09307958874724,2.23146629547127,6.96914454009643,-1.7970491081004,-7.2736621396321,10.2353232484353,-2.63925915467191,13.4470270492568,-.0451952183653594,.0116539449562267,.0471700287973566,-.0257161697253048,-.217799593005347,.0561613498441344,.227316372195247,-.614368644454251,.158419820138511,-1.33333501160483,.000932318622845132,-.000240405739976654,-.000973056396143737,.00636097475878108,-.00164022771456063,.0244039729141487,-1.66187980893617e-005,4.28528869206282e-006,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.59519010006696,-.927048238057034,-3.75228235860234,2.04567036451771,5.47681396366563,-1.4122398512488,-5.7161240011779,6.3309248170395,-1.63247909961861,5.76936797267538,-.233717104755142,.060265806265752,.243929401394036,-.132985500479333,-1.04841335581819,.27034168634063,1.09422390242844,-2.66785366107377,.68792719315715,-5.04596922575242,.0249321870529788,-.00642896187079958,-.0260216019365979,.159138742203809,-.0410351849082052,.559597243722216,-.00229823078236368,.000592617006985495,-.0189957623925199,.00019657159111697},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,3.03822870489707,-.78343133166635,-3.17098446910282,1.72875821562793,3.26615690329693,-.842204422108959,-3.40887201763794,1.66498726415444,-.429330151346565,-.923099517802526,-.555278916264744,.143183065809956,.579541894432988,-.315954814956539,-2.17523616991868,.560901872435633,2.27028337376948,-4.50024407433993,1.16042357185991,-6.26847807110811,.140734880491405,-.0362896034209994,-.14688430786117,.792696592497175,-.204403093795768,2.35258131414951,-.0308216365902829,.0079476030721956,-.225963663364185,.00626330408481366},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,2.30018618238774,-.593121222587361,-2.40069309089209,1.30881054275714,.987081447023724,-.254526767782017,-1.03021208824726,-1.35561938023839,.349557192548049,-2.20870753342451,-.981387965414959,.253058658488868,1.02426982907298,-.558411716942304,-3.10522366683209,.800706513225493,3.2409067852655,-4.42237267757001,1.14034381554164,-2.83495589139377,.439603191878621,-.113355164264732,-.458811705720642,2.03901059601619,-.525775029251873,4.55407145300733,-.170154848292169,.0438757750930774,-1.03685515398483,.0611113213546687},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,1.44662301694489,-.373023200905715,-1.50983338153088,.823131392784558,-.724355119014703,.186780703583358,.756005902089247,-1.6918152058931,.436247948578379,.160475142667942,-1.4741928888764,.380132310523241,1.53860792217029,-.83881870492831,-3.38025060487184,.871624387178691,3.52795105816102,-2.16048862424162,.557099100602201,1.70770592621981,.991945941760336,-.255781116350547,-1.03528913786274,3.46032388717692,-.892271916858127,4.72311662432345,-.576746649127296,.148718690750112,-2.68886700736678,.311154445333742},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,.553376983055106,-.142692637341486,-.577556856071558,.314872611220466,-1.45368745167752,.374844821120283,1.51720649776892,-.194424990185747,.0501340234001082,1.49978166741586,-1.9899087271236,.513113723366546,2.07685802519487,-1.13226204929232,-2.74863027720701,.708756083794028,2.86873200495546,.854309601049289,-.220290495881434,2.52312426315122,1.80736427869175,-.466043191865359,-1.88633727617264,4.12997714955087,-1.06494731359279,1.65834138994758,-1.41847476579742,.365764951301699,-4.48820932686106,1.03297875354965},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-.300186182387741,.0774053843401605,.313302853289657,-.170806538752116,-1.13611369814234,.292955913914665,1.18575632130325,1.22655366483195,-.316276575739705,.100858250439743,-2.48271365058504,.640187375400919,2.59119611829218,-1.41266903727833,-1.26648273457434,.326572602577083,1.32182184869504,2.5074905585056,-.646576297249579,-.461151554581851,2.81340752869054,-.725459410782454,-2.93633970583683,3.19379348794833,-.823544942743957,-2.33714259619681,-2.75487590413568,.710366567819732,-4.77320941254596,2.50302559510192},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.03822870489707,.26771549341915,1.08359423150038,-.590754211622909,-.0622566343920988,.0160533661863334,.0649769454361842,1.10967175893133,-.286137650697852,-1.41978880536247,-2.90882269973526,.750062968079831,3.03592405293217,-1.65512593926409,.707254507004454,-.182371175485818,-.738158077031132,1.57948219926848,-.407281992951262,-2.54719860772099,3.86201434387853,-.995850982046505,-4.03076552074204,.544351948717181,-.140365460738301,-3.18953813043078,-4.43071582310304,1.14249516194805,-2.19278566166381,4.71658607960963},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.59519010006696,.411332399809833,1.66489212099991,-.907666360512686,1.2388862761662,-.31945663701093,-1.29301955934683,-.514957807193037,.132785948612982,-.353775594121333,-3.23038451124486,.832980227624035,3.37153654597113,-1.8380952537413,2.62173967966463,-.676036338352017,-2.73629690762123,-1.30644820527577,.336878015452627,-.307822029727901,4.7630793830775,-1.22819773798136,-4.97120270417874,-2.90243304119458,.748415344244522,.518849452524734,-6.0685481480974,1.56482319698725,2.66498561200528,7.17423567654393},
			{1.414213562,-.364666166331411,-1.47600779224259,.804690348317657,1.24748791313838,-1.92171958555151,.49553061346938,2.00568935114481,-1.09346229146625,2.19799643560933,-.566770786786546,-2.29403815127598,-2.29634628619806,.592131075007482,2.23836106649518,-3.41890639763464,.881592088933561,3.56829591856781,-1.94536458449533,3.91413586924262,-1.00929093051433,-4.08516450282659,-4.09536544265748,1.05602241045943,4.00097073404889,5.33523806427656,-1.37573338488067,-5.56836192711791,-5.59797201662203,1.44348141522225,5.49216459339238,-7.19422103343416,1.855086864706,7.09892453011937,9.00134910987265},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.92171958555151,1.01124645171658,-4.09307958874725,-2.23146629547129,6.96914454009643,1.7970491081004,-7.2736621396321,10.2353232484353,2.63925915467191,13.4470270492568,-.0451952183653594,-.0116539449562267,.0471700287973566,.025716169725305,-.217799593005347,-.0561613498441344,.227316372195247,-.614368644454251,-.158419820138511,-1.33333501160483,.000932318622845132,.000240405739976654,-.000973056396143738,.00636097475878108,.00164022771456063,.0244039729141487,-1.66187980893617e-005,-4.28528869206282e-006,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.59519010006696,.927048238057034,-3.75228235860234,-2.04567036451773,5.47681396366563,1.4122398512488,-5.7161240011779,6.3309248170395,1.63247909961861,5.76936797267538,-.233717104755142,-.060265806265752,.243929401394036,.132985500479334,-1.04841335581819,-.27034168634063,1.09422390242844,-2.66785366107377,-.68792719315715,-5.04596922575242,.0249321870529788,.00642896187079958,-.0260216019365979,.159138742203809,.0410351849082052,.559597243722216,-.00229823078236368,-.000592617006985495,-.0189957623925199,.00019657159111697},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,3.03822870489707,.78343133166635,-3.17098446910282,-1.72875821562795,3.26615690329693,.842204422108959,-3.40887201763794,1.66498726415444,.429330151346565,-.923099517802526,-.555278916264744,-.143183065809956,.579541894432988,.315954814956541,-2.17523616991868,-.560901872435633,2.27028337376948,-4.50024407433993,-1.16042357185991,-6.26847807110811,.140734880491405,.0362896034209994,-.14688430786117,.792696592497175,.204403093795768,2.35258131414951,-.0308216365902829,-.0079476030721956,-.225963663364185,.00626330408481366},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,2.30018618238774,.593121222587361,-2.40069309089209,-1.30881054275715,.987081447023724,.254526767782017,-1.03021208824726,-1.35561938023839,-.349557192548049,-2.20870753342451,-.981387965414959,-.253058658488868,1.02426982907298,.558411716942308,-3.10522366683209,-.800706513225493,3.2409067852655,-4.42237267757001,-1.14034381554164,-2.83495589139377,.439603191878621,.113355164264732,-.458811705720642,2.03901059601619,.525775029251873,4.55407145300733,-.170154848292169,-.0438757750930774,-1.03685515398483,.0611113213546687},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,1.44662301694489,.373023200905715,-1.50983338153088,-.823131392784565,-.724355119014703,-.186780703583358,.756005902089247,-1.6918152058931,-.436247948578379,.160475142667942,-1.4741928888764,-.380132310523241,1.53860792217029,.838818704928317,-3.38025060487184,-.871624387178691,3.52795105816103,-2.16048862424162,-.557099100602201,1.70770592621981,.991945941760336,.255781116350547,-1.03528913786275,3.46032388717692,.892271916858127,4.72311662432345,-.576746649127296,-.148718690750112,-2.68886700736678,.311154445333742},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,.553376983055106,.142692637341486,-.577556856071558,-.314872611220468,-1.45368745167752,-.374844821120283,1.51720649776892,-.194424990185747,-.0501340234001082,1.49978166741586,-1.9899087271236,-.513113723366546,2.07685802519487,1.13226204929233,-2.74863027720701,-.708756083794028,2.86873200495547,.854309601049289,.220290495881434,2.52312426315122,1.80736427869175,.466043191865359,-1.88633727617264,4.12997714955087,1.06494731359279,1.65834138994758,-1.41847476579742,-.365764951301699,-4.48820932686106,1.03297875354965},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-.300186182387741,-.0774053843401605,.313302853289658,.170806538752117,-1.13611369814234,-.292955913914665,1.18575632130325,1.22655366483195,.316276575739705,.100858250439743,-2.48271365058504,-.640187375400919,2.59119611829218,1.41266903727834,-1.26648273457434,-.326572602577083,1.32182184869504,2.5074905585056,.646576297249579,-.461151554581851,2.81340752869054,.725459410782454,-2.93633970583683,3.19379348794833,.823544942743957,-2.33714259619681,-2.75487590413568,-.710366567819732,-4.77320941254596,2.50302559510192},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.03822870489707,-.26771549341915,1.08359423150038,.590754211622914,-.0622566343920988,-.0160533661863334,.0649769454361843,1.10967175893133,.286137650697852,-1.41978880536247,-2.90882269973526,-.750062968079831,3.03592405293218,1.65512593926411,.707254507004454,.182371175485818,-.738158077031133,1.57948219926848,.407281992951262,-2.54719860772099,3.86201434387853,.995850982046505,-4.03076552074204,.544351948717181,.140365460738301,-3.18953813043078,-4.43071582310304,-1.14249516194805,-2.19278566166381,4.71658607960963},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.59519010006696,-.411332399809833,1.66489212099991,.907666360512694,1.2388862761662,.31945663701093,-1.29301955934683,-.514957807193037,-.132785948612982,-.353775594121333,-3.23038451124486,-.832980227624035,3.37153654597113,1.83809525374131,2.62173967966463,.676036338352017,-2.73629690762123,-1.30644820527577,-.336878015452627,-.307822029727901,4.7630793830775,1.22819773798136,-4.97120270417875,-2.90243304119458,-.748415344244522,.518849452524734,-6.0685481480974,-1.56482319698725,2.66498561200528,7.17423567654393},
			{1.414213562,.364666166331411,-1.47600779224259,-.804690348317664,1.24748791313839,-1.92171958555151,-.49553061346938,2.00568935114481,1.09346229146626,2.19799643560933,.566770786786546,-2.29403815127598,-2.29634628619806,-.592131075007482,2.23836106649518,-3.41890639763464,-.881592088933561,3.56829591856781,1.94536458449534,3.91413586924262,1.00929093051433,-4.08516450282659,-4.09536544265748,-1.05602241045943,4.00097073404889,5.33523806427656,1.37573338488067,-5.56836192711792,-5.59797201662203,-1.44348141522225,5.49216459339238,-7.19422103343416,-1.855086864706,7.09892453011937,9.00134910987265},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.92171958555151,2.94388917191153,-1.91390801783745,-4.63366448151075,6.96914454009643,5.23147784770941,-3.40113598727832,10.2353232484353,7.68327684524188,13.4470270492568,-.0451952183653594,-.0339263710725032,.0220565211009043,.0533999113042759,-.217799593005347,-.163494061616168,.10629224716917,-.614368644454251,-.461183712763862,-1.33333501160483,.000932318622845132,.000699856947415033,-.000454997367631817,.00636097475878108,.00477494739305899,.0244039729141487,-1.66187980893617e-005,-1.24751142094631e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.59519010006696,2.69877562014986,-1.7545525650326,-4.24785717273985,5.47681396366563,4.11124073580283,-2.67283724139748,6.3309248170395,4.75239001984257,5.76936797267538,-.233717104755142,-.175442746128902,.114060434690259,.276145864886065,-1.04841335581819,-.787004949964004,.511654819985593,-2.66785366107377,-2.00265861161846,-5.04596922575242,.0249321870529788,.0187156663995725,-.0121675993597457,.159138742203809,.119459540563619,.559597243722216,-.00229823078236368,-.00172519645136334,-.0189957623925199,.00019657159111697},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,3.03822870489707,2.2806853960415,-1.48273994391383,-3.58978558499056,3.26615690329693,2.45178262388351,-1.59397523182193,1.66498726415444,1.24984407270733,-.923099517802526,-.555278916264744,-.41682724950333,.270991524688991,.656083673232099,-2.17523616991868,-1.63286896677384,1.06157563213921,-4.50024407433993,-3.3781660071394,-6.26847807110811,.140734880491405,.105644445409619,-.0686825282564417,.792696592497175,.595047877309364,2.35258131414951,-.0308216365902829,-.0231366573273721,-.225963663364185,.00626330408481366},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,2.30018618238774,1.72666429814602,-1.12255470615748,-2.71775959032344,.987081447023724,.740965364440283,-.481723145865575,-1.35561938023839,-1.01761309753837,-2.20870753342451,-.981387965414959,-.736691479430405,.478944568701089,1.15954811601066,-3.10522366683209,-2.33097602484761,1.51543534565911,-4.42237267757001,-3.31971084311917,-2.83495589139377,.439603191878621,.329993781528466,-.214538560322786,2.03901059601619,1.53060948975261,4.55407145300733,-.170154848292169,-.127728922051078,-1.03685515398483,.0611113213546687},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,1.44662301694489,1.0859261460488,-.705992187998234,-1.70924145531706,-.724355119014703,-.543746472353678,.353505404956932,-1.6918152058931,-1.26998281171566,.160475142667942,-1.4741928888764,-1.10662182393172,.719447050735515,1.74181633275904,-3.38025060487184,-2.53742852796533,1.64965612574837,-2.16048862424162,-1.62179853107076,1.70770592621981,.991945941760336,.744616960114543,-.484097154422054,3.46032388717692,2.5975365648804,4.72311662432345,-.576746649127296,-.432942278924126,-2.68886700736678,.311154445333742},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,.553376983055106,.41539953912128,-.270063328509737,-.653836465186081,-1.45368745167752,-1.09122915405036,.709439828333211,-.194424990185747,-.145947615817496,1.49978166741586,-1.9899087271236,-1.49375054084377,.971130695151501,2.35115468793598,-2.74863027720701,-2.06329461731692,1.34140787303665,.854309601049289,.641298473185774,2.52312426315122,1.80736427869175,1.35672120663226,-.882043937561808,4.12997714955087,3.10022038625728,1.65834138994758,-1.41847476579742,-1.06479629943226,-4.48820932686106,1.03297875354965},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-.300186182387741,-.225338612975934,.146499189649511,.3546816698203,-1.13611369814234,-.852838337634571,.554455021295673,1.22655366483195,.920728260721081,.100858250439743,-2.48271365058504,-1.86368088534508,1.21163317718593,2.93342290468436,-1.26648273457434,-.950701529718779,.618078730089992,2.5074905585056,1.88227998927126,-.461151554581851,2.81340752869054,2.11192049221879,-1.3730209697231,3.19379348794833,2.39746209780118,-2.33714259619681,-2.75487590413568,-2.0679829763977,-4.77320941254596,2.50302559510192},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.03822870489707,-.779359710871422,.506684427405857,1.22670766448742,-.0622566343920988,-.0467337421144522,.030382965722629,1.10967175893133,.832989356981887,-1.41978880536247,-2.90882269973526,-2.18354511527215,1.41958622119803,3.43688734746291,.707254507004454,.530909678713955,-.34515983171824,1.57948219926848,1.18565859680248,-2.54719860772099,3.86201434387853,2.89907066463143,-1.88477020319366,.544351948717181,.408624781106939,-3.18953813043078,-4.43071582310304,-3.32597373321892,-2.19278566166381,4.71658607960963},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.59519010006696,-1.19744993497978,.778497048524628,1.88477925223671,1.2388862761662,.929985893147364,-.604610892164934,-.514957807193037,-.386559691398865,-.353775594121333,-3.23038451124486,-2.42492961864658,1.57651731119676,3.81682515580895,2.62173967966463,1.9680425606586,-1.27948173914766,-1.30644820527577,-.980702122872796,-.307822029727901,4.7630793830775,3.57546670811253,-2.32451495445633,-2.90243304119458,-2.17874863666902,.518849452524734,-6.0685481480974,-4.55543360151918,2.66498561200528,7.17423567654393},
			{1.414213562,1.06159757258303,-.690175474654838,-1.67094842220172,-.742527141878789,-1.92171958555151,-1.44256348674145,.937852501329483,2.27058656100761,2.19799643560933,1.64995425135427,-1.07268327325536,-2.29634628619806,-1.72378183093533,2.23836106649518,-3.41890639763464,-2.56644599370298,1.66852122478611,4.03957110939074,3.91413586924262,2.93819635817362,-1.91020695459718,-4.09536544265748,-3.07423866276144,4.00097073404889,5.33523806427656,4.00496497002541,-2.60374427309723,-5.59797201662203,-4.20218958584716,5.49216459339238,-7.19422103343416,-5.40043440913241,7.09892453011937,9.00134910987265},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.92171958555151,4.61496937554466,1.68716544952149,-2.43916732844396,6.96914454009643,8.20109339929485,2.99820005631404,10.2353232484353,12.0446407028288,13.4470270492568,-.0451952183653594,-.0531844625867158,-.0194434633176023,.0281097864368307,-.217799593005347,-.256300439106579,-.0936996999356592,-.614368644454251,-.722971751522571,-1.33333501160483,.000932318622845132,.00109712634918432,.000401093381258214,.00636097475878108,.00748541629863581,.0244039729141487,-1.66187980893617e-005,-1.95565344587043e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.59519010006696,4.23071865520366,1.54668899419581,-2.23607783277939,5.47681396366563,6.44496072482988,2.35618358032457,6.3309248170395,7.45005437415178,5.76936797267538,-.233717104755142,-.275031719356679,-.100547582628472,.145363561432351,-1.04841335581819,-1.23374336817428,-.451038569416842,-2.66785366107377,-3.13945519772704,-5.04596922575242,.0249321870529788,.0293394969148428,.0107260918769641,.159138742203809,.187269998673138,.559597243722216,-.00229823078236368,-.00270449338439953,-.0189957623925199,.00019657159111697},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,3.03822870489707,3.57530214058608,1.30707828207097,-1.88966804781974,3.26615690329693,3.843521635486,1.4051354158397,1.66498726415444,1.95931021275678,-.923099517802526,-.555278916264744,-.653436620408273,-.238886891797976,.345363344035536,-2.17523616991868,-2.55975678291091,-.935809721212699,-4.50024407433993,-5.29576072944733,-6.26847807110811,.140734880491405,.165612851500527,.0605456414728774,.792696592497175,.932823068533675,2.35258131414951,-.0308216365902829,-.0362700355832219,-.225963663364185,.00626330408481366},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,2.30018618238774,2.70679444519172,.989564544259882,-1.43063237006769,.987081447023724,1.16156970101249,.424652930216328,-1.35561938023839,-1.5952547827211,-2.20870753342451,-.981387965414959,-1.15486977201266,-.422203533825771,.610387716162348,-3.10522366683209,-3.65413992905702,-1.33590022736237,-4.42237267757001,-5.20412385861342,-2.83495589139377,.439603191878621,.517312679568432,.189121965733584,2.03901059601619,2.39945035585009,4.55407145300733,-.170154848292169,-.200233442638163,-1.03685515398483,.0611113213546687},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,1.44662301694489,1.70234530427801,.622352597994008,-.899747042727636,-.724355119014703,-.852400743178548,-.31162527138391,-1.6918152058931,-1.9908805805112,.160475142667942,-1.4741928888764,-1.73478875376213,-.634213449888053,.916894502821386,-3.38025060487184,-3.97778389924568,-1.45421973940992,-2.16048862424162,-2.54240228389271,1.70770592621981,.991945941760336,1.16729410204273,.426745687640421,3.46032388717692,4.07201188589473,4.72311662432345,-.576746649127296,-.678699245092846,-2.68886700736678,.311154445333742},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,.553376983055106,.651198479192506,.23806865993447,-.344180410642768,-1.45368745167752,-1.71065853147384,-.625391792982136,-.194424990185747,-.228793863525153,1.49978166741586,-1.9899087271236,-2.34166865603201,-.856079885009684,1.23765104741928,-2.74863027720701,-3.23451085132423,-1.18248996086068,.854309601049289,1.00532752474992,2.52312426315122,1.80736427869175,2.12685548066823,.777547323353426,4.12997714955087,4.86004102210358,1.65834138994758,-1.41847476579742,-1.66922123290477,-4.48820932686106,1.03297875354965},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-.300186182387741,-.353250661721207,-.129143286331404,.186704916697283,-1.13611369814234,-1.33694666498547,-.488768188714077,1.22655366483195,1.44337387662839,.100858250439743,-2.48271365058504,-2.92158763778148,-1.06808980107197,1.54415783407832,-1.26648273457434,-1.49036128356923,-.54485433405009,2.5074905585056,2.95074440626681,-.461151554581851,2.81340752869054,3.31073889879025,1.21035782283977,3.19379348794833,3.75836640385388,-2.33714259619681,-2.75487590413568,-3.2418605279988,-4.77320941254596,2.50302559510192},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.03822870489707,-1.22175835711557,-.446657024142497,.645740594449334,-.0622566343920988,-.0732618573826121,-.0267834658423845,1.10967175893133,1.30583053509787,-1.41978880536247,-2.90882269973526,-3.42302078938587,-1.25140644309976,1.80918220620513,.707254507004454,.832277224231996,.304268406428257,1.57948219926848,1.8586902544778,-2.54719860772099,3.86201434387853,4.54470992402429,1.661479620483,.544351948717181,.640578072330311,-3.18953813043078,-4.43071582310304,-5.21394183895335,-2.19278566166381,4.71658607960963},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.59519010006696,-1.87717487173315,-.686267736267333,.992150379408985,1.2388862761662,1.4578865459724,.532982044151546,-.514957807193037,-.605987872997064,-.353775594121333,-3.23038451124486,-3.80142569043747,-1.38974575226926,2.00918198880831,2.62173967966463,3.08518956279543,1.12790027705863,-1.30644820527577,-1.53739152505215,-.307822029727901,4.7630793830775,5.60505793446857,2.04912737785909,-2.90243304119458,-3.41550161982096,.518849452524734,-6.0685481480974,-7.14129688166464,2.66498561200528,7.17423567654393},
			{1.414213562,1.66420676884452,.608409706261878,-.87958953765694,-1.79802926655924,-1.92171958555151,-2.26142559207414,-.826744191593013,1.19523987507356,2.19799643560933,2.5865404220042,.94560142914338,-2.29634628619806,-2.70227576356801,2.23836106649518,-3.41890639763464,-4.02327294720743,-1.47084987158013,2.12643576380383,3.91413586924262,4.60604507183384,1.68390285488931,-4.09536544265748,-4.81931093640565,4.00097073404889,5.33523806427656,6.27835818791885,2.29527612404421,-5.59797201662203,-6.58753611853026,5.49216459339238,-7.19422103343416,-8.46595709192435,7.09892453011937,9.00134910987265},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.92171958555151,5.87604462970729,5.45886369350783,3.32855727881994,6.96914454009643,10.4421041409331,9.70074715430413,10.2353232484353,15.3359297883224,13.4470270492568,-.0451952183653594,-.067717518868618,-.0629097851728847,-.0383594160020484,-.217799593005347,-.326336470786676,-.303167594035556,-.614368644454251,-.920529245649199,-1.33333501160483,.000932318622845132,.00139692441436278,.00129774711619287,.00636097475878108,.0095308628646168,.0244039729141487,-1.66187980893617e-005,-2.49005052754424e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.59519010006696,5.38679449650258,5.00434880168882,3.0514155628076,5.47681396366563,8.20609494322841,7.62348767011616,6.3309248170395,9.48583802706066,5.76936797267538,-.233717104755142,-.350186214905109,-.325324080359881,-.198367260350895,-1.04841335581819,-1.57087306611691,-1.45934595309525,-2.66785366107377,-3.99733505330865,-5.04596922575242,.0249321870529788,.0373567361461481,.0347045238135922,.159138742203809,.238442940887059,.559597243722216,-.00229823078236368,-.00344351663776842,-.0189957623925199,.00019657159111697},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,3.03822870489707,4.55228046671326,4.22908267863918,2.57869489385806,3.26615690329693,4.89379296535329,4.54634885284358,1.66498726415444,2.49470653361407,-.923099517802526,-.555278916264744,-.831993114526565,-.772924185272261,-.471292665829706,-2.17523616991868,-3.25922966624351,-3.02783447385586,-4.50024407433993,-6.74286736535737,-6.26847807110811,.140734880491405,.210867814600683,.195896854176436,.792696592497175,1.18772402074228,2.35258131414951,-.0308216365902829,-.0461810968751945,-.225963663364185,.00626330408481366},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,2.30018618238774,3.44644648080968,3.20175947449313,1.95228172055832,.987081447023724,1.47897739903412,1.37397459375255,-1.35561938023839,-2.03117020639316,-2.20870753342451,-.981387965414959,-1.47044666380812,-1.36604951383154,-.832952480070553,-3.10522366683209,-4.65266127660931,-4.32233676392885,-4.42237267757001,-6.6261900271284,-2.83495589139377,.439603191878621,.658672278252945,.611908590636856,2.03901059601619,3.05511829665785,4.55407145300733,-.170154848292169,-.254948743871766,-1.03685515398483,.0611113213546687},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,1.44662301694489,2.16752402217829,2.01363654211465,1.2278204669453,-.724355119014703,-1.08532568728512,-1.00827093182807,-1.6918152058931,-2.53490374291394,.160475142667942,-1.4741928888764,-2.20883288938782,-2.05201260878721,-1.25122037987573,-3.38025060487184,-5.06474340721712,-4.70516234202085,-2.16048862424162,-3.23713291923203,1.70770592621981,.991945941760336,1.48626603590141,1.38074575986725,3.46032388717692,5.18471990323707,4.72311662432345,-.576746649127296,-.864158942299273,-2.68886700736678,.311154445333742},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,.553376983055106,.829143384311425,.770276776736346,.469678401195648,-1.45368745167752,-2.17810890152299,-2.02346992933989,-.194424990185747,-.291313515578485,1.49978166741586,-1.9899087271236,-2.9815473107461,-2.76986670414993,-1.68893390563517,-2.74863027720701,-4.11836538252447,-3.82597422031764,.854309601049289,1.2800408674882,2.52312426315122,1.80736427869175,2.70803481201197,2.51577274454146,4.12997714955087,6.18808395553405,1.65834138994758,-1.41847476579742,-2.12534854800545,-4.48820932686106,1.03297875354965},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-.300186182387741,-.449779074319961,-.417846155642134,-.254782852417372,-1.13611369814234,-1.70227744362135,-1.58142102819233,1.22655366483195,1.83778669504629,.100858250439743,-2.48271365058504,-3.7199335363258,-3.45582979910559,-2.10720180544034,-1.26648273457434,-1.89761376598671,-1.76288907720341,2.5074905585056,3.75705761211831,-.461151554581851,2.81340752869054,4.21542331996591,3.91614134649752,3.19379348794833,4.78536842321548,-2.33714259619681,-2.75487590413568,-4.12772341388705,-4.77320941254596,2.50302559510192},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.03822870489707,-1.55561306022354,-1.44516935978819,-.881196025717112,-.0622566343920988,-.0932812135041899,-.0866585368463816,1.10967175893133,1.66265859611778,-1.41978880536247,-2.90882269973526,-4.35838708560735,-4.04895512766487,-2.46886161968119,.707254507004454,1.0597032647262,.984467621360889,1.57948219926848,2.36659141141637,-2.54719860772099,3.86201434387853,5.78658625215441,5.37575658648661,.544351948717181,.815620871022689,-3.18953813043078,-4.43071582310304,-6.63869084478425,-2.19278566166381,4.71658607960963},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.59519010006696,-2.39012709001286,-2.22043548283783,-1.35391669466665,1.2388862761662,1.85626506095126,1.72447600259692,-.514957807193037,-.771578638346196,-.353775594121333,-3.23038451124486,-4.84019398522881,-4.49655523257725,-2.74178702516,2.62173967966463,3.92824092358218,3.64934800797338,-1.30644820527577,-1.95749537633152,-.307822029727901,4.7630793830775,7.13668236880667,6.6300000687783,-2.90243304119458,-4.3488132461414,.518849452524734,-6.0685481480974,-9.09271021667365,2.66498561200528,7.17423567654393},
			{1.414213562,2.11896384374972,1.96852398633026,1.20031296082078,.0796164457722469,-1.92171958555151,-2.87937722321757,-2.67495037465683,-1.63105841067899,2.19799643560933,3.29333213710544,3.05951577632408,-2.29634628619806,-3.44069303528758,2.23836106649518,-3.41890639763464,-5.1226626812653,-4.75896952776425,-2.90179486950885,3.91413586924262,5.86468169257239,5.4483074914535,-4.09536544265748,-6.13622406614863,4.00097073404889,5.33523806427656,7.99396700420041,7.42642015557747,-5.59797201662203,-8.38763014060611,5.49216459339238,-7.19422103343416,-10.7793438389747,7.09892453011937,9.00134910987265},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.92171958555151,6.61537460386721,8.09172854779269,8.80395991076745,6.96914454009643,11.7559404153652,14.3795150585562,10.2353232484353,17.2655122351262,13.4470270492568,-.0451952183653594,-.0762378066864121,-.0932518071891671,-.101459801467564,-.217799593005347,-.367396460919218,-.449388373324038,-.614368644454251,-1.03635118137081,-1.33333501160483,.000932318622845132,.00157268687980642,.00192366360061384,.00636097475878108,.0107300458251741,.0244039729141487,-1.66187980893617e-005,-2.80335124396135e-005,-.000146461329012245,2.74870867647865e-007},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.59519010006696,6.06456651609707,7.41799655299993,8.07092624092508,5.47681396366563,9.23859424244377,11.300372436005,6.3309248170395,10.6793559162843,5.76936797267538,-.233717104755142,-.394247004353173,-.482231155810537,-.524676988090528,-1.04841335581819,-1.76852192969244,-2.1632031816091,-2.66785366107377,-4.50028385780377,-5.04596922575242,.0249321870529788,.0420569990798227,.0514428646412204,.159138742203809,.268444076758865,.559597243722216,-.00229823078236368,-.0038767834400574,-.0189957623925199,.00019657159111697},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,3.03822870489707,5.12505307344404,6.26881178264598,6.82059059403532,3.26615690329693,5.50953502563763,6.7390986231311,1.66498726415444,2.80859307349989,-.923099517802526,-.555278916264744,-.936675343241156,-1.14571329243577,-1.24655861042429,-2.17523616991868,-3.66930929238228,-4.48818948881768,-4.50024407433993,-7.59126186695628,-6.26847807110811,.140734880491405,.237399419743997,.29037987690505,.792696592497175,1.33716467766395,2.35258131414951,-.0308216365902829,-.0519916499446519,-.225963663364185,.00626330408481366},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,2.30018618238774,3.88008191896113,4.74600026627036,5.1637416942433,.987081447023724,1.66506385533551,2.03665635715207,-1.35561938023839,-2.28673412920868,-2.20870753342451,-.981387965414959,-1.65545977423627,-2.02490893148973,-2.20314076875832,-3.10522366683209,-5.23806390000324,-6.40704325104372,-4.42237267757001,-7.45990406612127,-2.83495589139377,.439603191878621,.741547101224613,.907038257318,2.03901059601619,3.43951642050874,4.55407145300733,-.170154848292169,-.287026656838231,-1.03685515398483,.0611113213546687},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,1.44662301694489,2.44024412222762,2.98483369571688,3.24755780451479,-.724355119014703,-1.22188248065577,-1.49457014152702,-1.6918152058931,-2.85384793698481,.160475142667942,-1.4741928888764,-2.48675051356285,-3.04171892525937,-3.30945005334757,-3.38025060487184,-5.70199462778367,-6.97450945518368,-2.16048862424162,-3.64443230995284,1.70770592621981,.991945941760336,1.67326955598402,2.04669332477546,3.46032388717692,5.8370666898335,4.72311662432345,-.576746649127296,-.972888308521261,-2.68886700736678,.311154445333742},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,.553376983055106,.93346705704166,1.14178901214037,1.24228891639978,-1.45368745167752,-2.45216080196962,-2.99940982448642,-.194424990185747,-.327966881493408,1.49978166741586,-1.9899087271236,-3.35668865754018,-4.10580126963162,-4.46719258573777,-2.74863027720701,-4.6365423474252,-5.67128006117502,.854309601049289,1.44109692494208,2.52312426315122,1.80736427869175,3.04876253512471,3.72915523810848,4.12997714955087,6.96667503835439,1.65834138994758,-1.41847476579742,-2.39276208655033,-4.48820932686106,1.03297875354965},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-.300186182387741,-.506370739691845,-.619377558413115,-.673894973328732,-1.13611369814234,-1.91645974101964,-2.34415629302528,1.22655366483195,2.06901890572263,.100858250439743,-2.48271365058504,-4.18797939686675,-5.12261126340126,-5.57350187032702,-1.26648273457434,-2.13637348021345,-2.61314820693614,2.5074905585056,4.22977446202813,-.461151554581851,2.81340752869054,4.7458122143026,5.80493569904152,3.19379348794833,5.38746839617942,-2.33714259619681,-2.75487590413568,-4.64707781591113,-4.77320941254596,2.50302559510192},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.03822870489707,-1.75134189417476,-2.14218907478874,-2.33074387312075,-.0622566343920988,-.105017951653011,-.12845482061473,1.10967175893133,1.87185600940662,-1.41978880536247,-2.90882269973526,-4.90676382786187,-6.00180690245521,-6.53008402866105,.707254507004454,1.19303621855815,1.45928625505293,1.57948219926848,2.66435837496252,-2.54719860772099,3.86201434387853,6.51466048131367,7.96853804732135,.544351948717181,.918243127233759,-3.18953813043078,-4.43071582310304,-7.47397774950175,-2.19278566166381,4.71658607960963},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.59519010006696,-2.69085533682779,-3.29137384514269,-3.58107952001051,1.2388862761662,2.0898222386161,2.55620812016103,-.514957807193037,-.868659455579414,-.353775594121333,-3.23038451124486,-5.44919216674985,-6.66528903908045,-7.25196565099482,2.62173967966463,4.42249623366662,5.4094652504454,-1.30644820527577,-2.20378945631257,-.307822029727901,4.7630793830775,8.03462707886587,9.82771577392637,-2.90243304119458,-4.89598539807521,.518849452524734,-6.0685481480974,-10.2367643607743,2.66498561200528,7.17423567654393},
			{1.414213562,2.38557405224355,2.91796290062108,3.17480106311619,3.19927335807151,-1.92171958555151,-3.24166342459793,-3.96510583993544,-4.31411318985288,2.19799643560933,3.70770257117539,4.53515100205683,-2.29634628619806,-3.87360456901059,2.23836106649518,-3.41890639763464,-5.76720136441661,-7.05426838770182,-7.67518283761778,3.91413586924262,6.60258197030424,8.07608098333328,-4.09536544265748,-6.90829008438974,4.00097073404889,5.33523806427656,8.99977614812197,11.0082572774727,-5.59797201662203,-9.44297038492046,5.49216459339238,-7.19422103343416,-12.1356119587979,7.09892453011937,9.00134910987265}
		};
		for (int j = 0; j < 100; j++)
		{
			std::cout << timestep_dofs_to_right_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_right_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_hyp_fluxes_to_dofs:" << std::endl;
		static const T timestep_hyp_fluxes_to_dofs[35][100] = 
		{
			{-.00222253422114756,-.00498206115568338,-.00730339144757066,-.00897618736857064,-.00985149899052509,-.00985149899052509,-.00897618736857064,-.00730339144757066,-.00498206115568338,-.00222253422114756,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256888,-.0397874192256888,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00985149899052509,-.0220832462235868,-.0323726639565543,-.0397874192256888,-.0436672836966291,-.0436672836966291,-.0397874192256889,-.0323726639565543,-.0220832462235868,-.0098514989905251,-.00897618736857064,-.0201211364889587,-.02949633325581,-.0362522830510504,-.0397874192256889,-.0397874192256889,-.0362522830510504,-.02949633325581,-.0201211364889587,-.00897618736857064,-.00730339144757066,-.0163713757428243,-.0239994174797936,-.02949633325581,-.0323726639565543,-.0323726639565543,-.02949633325581,-.0239994174797936,-.0163713757428243,-.00730339144757066,-.00498206115568338,-.011167852050509,-.0163713757428243,-.0201211364889587,-.0220832462235868,-.0220832462235868,-.0201211364889587,-.0163713757428243,-.011167852050509,-.00498206115568339,-.00222253422114756,-.00498206115568339,-.00730339144757066,-.00897618736857064,-.0098514989905251,-.0098514989905251,-.00897618736857064,-.00730339144757066,-.00498206115568339,-.00222253422114756},
			{.00374909427250663,.00840401770480107,.0123197666813982,.0151415318298581,.0166180561313977,.0166180561313977,.0151415318298581,.0123197666813982,.00840401770480107,.00374909427250663,.00746479013903255,.0167331691152994,.0245297840368778,.0301481769402938,.0330880720843031,.0330880720843031,.0301481769402938,.0245297840368778,.0167331691152994,.00746479013903255,.00859442576735982,.0192653747976306,.0282418399000882,.0347104558746731,.0380952409936669,.0380952409936669,.0347104558746731,.0282418399000882,.0192653747976306,.00859442576735982,.00673809032372206,.0151041895084832,.0221418013612503,.0272132418374432,.0298669372064565,.0298669372064565,.0272132418374432,.0221418013612503,.0151041895084832,.00673809032372206,.00254028702921173,.00569433991706437,.00834754776191493,.0102595011258728,.0112599549045422,.0112599549045422,.0102595011258728,.00834754776191493,.00569433991706437,.00254028702921173,-.00254028702921173,-.00569433991706437,-.00834754776191493,-.0102595011258728,-.0112599549045422,-.0112599549045422,-.0102595011258728,-.00834754776191493,-.00569433991706437,-.00254028702921173,-.00673809032372206,-.0151041895084832,-.0221418013612503,-.0272132418374432,-.0298669372064565,-.0298669372064565,-.0272132418374432,-.0221418013612503,-.0151041895084832,-.00673809032372206,-.00859442576735982,-.0192653747976306,-.0282418399000882,-.0347104558746731,-.0380952409936669,-.0380952409936669,-.0347104558746731,-.0282418399000882,-.0192653747976306,-.00859442576735982,-.00746479013903255,-.0167331691152994,-.0245297840368778,-.0301481769402938,-.0330880720843031,-.0330880720843031,-.0301481769402938,-.0245297840368778,-.0167331691152994,-.00746479013903255,-.00374909427250663,-.00840401770480107,-.0123197666813982,-.0151415318298581,-.0166180561313977,-.0166180561313977,-.0151415318298581,-.0123197666813982,-.00840401770480107,-.00374909427250663},
			{-.00458578009255966,-.0102795433475266,-.015069170494548,-.0185206693108299,-.0203267097184652,-.0203267097184652,-.0185206693108299,-.015069170494548,-.0102795433475266,-.00458578009255966,-.00693481320266981,-.0155451660317305,-.0227882454870494,-.0280077499283953,-.0307389216397789,-.0307389216397789,-.0280077499283953,-.0227882454870494,-.0155451660317305,-.00693481320266981,-.00314199662793749,-.00704313985467146,-.0103248044877915,-.0126896360809338,-.0139270641206943,-.0139270641206943,-.0126896360809338,-.0103248044877915,-.00704313985467146,-.0031419966279375,.00438062859874147,.00981967313331039,.0143950293941988,.017692120427338,.019417365009608,.019417365009608,.017692120427338,.0143950293941988,.00981967313331039,.00438062859874147,.0102819614083243,.0230481762886867,.0337871913508869,.0415259352316643,.0455753308412161,.0455753308412161,.0415259352316643,.0337871913508869,.0230481762886867,.0102819614083243,.0102819614083243,.0230481762886867,.0337871913508869,.0415259352316643,.0455753308412161,.0455753308412161,.0415259352316643,.0337871913508869,.0230481762886867,.0102819614083243,.00438062859874146,.00981967313331036,.0143950293941988,.017692120427338,.0194173650096079,.0194173650096079,.017692120427338,.0143950293941988,.00981967313331036,.00438062859874146,-.00314199662793748,-.00704313985467143,-.0103248044877915,-.0126896360809337,-.0139270641206943,-.0139270641206943,-.0126896360809337,-.0103248044877915,-.00704313985467143,-.00314199662793748,-.00693481320266981,-.0155451660317305,-.0227882454870495,-.0280077499283953,-.0307389216397789,-.0307389216397789,-.0280077499283953,-.0227882454870495,-.0155451660317305,-.00693481320266981,-.00458578009255966,-.0102795433475266,-.015069170494548,-.0185206693108299,-.0203267097184652,-.0203267097184652,-.0185206693108299,-.015069170494548,-.0102795433475266,-.00458578009255966},
			{.00498941899157684,.0111843454696186,.0163955540682807,.0201508527079391,.0221158602153926,.0221158602153926,.0201508527079391,.0163955540682807,.0111843454696186,.00498941899157684,.00422852158563388,.00947870810595568,.01389519589419,.0170778032048017,.0187431427313533,.0187431427313533,.0170778032048017,.01389519589419,.00947870810595568,.00422852158563388,-.00454244456135888,-.0101824018661483,-.0149267671313349,-.0183456493520488,-.0201346227135425,-.0201346227135425,-.0183456493520488,-.0149267671313349,-.0101824018661483,-.00454244456135888,-.0106057150885859,-.0237739066819889,-.0348510669200656,-.0428334848152169,-.0470103859345993,-.0470103859345993,-.0428334848152169,-.0348510669200656,-.0237739066819889,-.0106057150885859,-.00560552264733118,-.0125654113096862,-.0184201105981369,-.022639121189878,-.0248467718409463,-.0248467718409463,-.022639121189878,-.0184201105981369,-.0125654113096862,-.00560552264733118,.00560552264733123,.0125654113096863,.0184201105981371,.0226391211898782,.0248467718409465,.0248467718409465,.0226391211898782,.0184201105981371,.0125654113096863,.00560552264733123,.0106057150885859,.023773906681989,.0348510669200657,.042833484815217,.0470103859345994,.0470103859345994,.042833484815217,.0348510669200657,.023773906681989,.0106057150885859,.00454244456135883,.0101824018661482,.0149267671313347,.0183456493520486,.0201346227135423,.0201346227135423,.0183456493520486,.0149267671313347,.0101824018661482,.00454244456135883,-.00422852158563391,-.00947870810595576,-.0138951958941901,-.0170778032048018,-.0187431427313534,-.0187431427313534,-.0170778032048018,-.0138951958941901,-.00947870810595576,-.00422852158563391,-.00498941899157684,-.0111843454696186,-.0163955540682807,-.0201508527079391,-.0221158602153927,-.0221158602153927,-.0201508527079391,-.0163955540682807,-.0111843454696186,-.00498941899157684},
			{-.00502787889479981,-.0112705576809201,-.0165219358020692,-.020306181383746,-.0222863357447103,-.0222863357447103,-.020306181383746,-.0165219358020692,-.0112705576809201,-.00502787889479981,-.000280476734556914,-.000628720237921658,-.000921664721701268,-.00113276623502688,-.00124322777173102,-.00124322777173102,-.00113276623502688,-.000921664721701268,-.000628720237921658,-.000280476734556914,.009285522300677,.0208145456318548,.0305128278843669,.0375015993872448,.0411585624649621,.0411585624649621,.0375015993872448,.030512827884367,.0208145456318548,.009285522300677,.00471291107004208,.010564521773763,.0154869311233076,.0190341153866534,.020890224414624,.020890224414624,.0190341153866534,.0154869311233076,.010564521773763,.00471291107004208,-.00869007782643158,-.0194797896774683,-.028556158763447,-.0350967674986958,-.0385192237402185,-.0385192237402185,-.0350967674986958,-.028556158763447,-.0194797896774683,-.00869007782643158,-.00869007782643163,-.0194797896774684,-.0285561587634472,-.0350967674986961,-.0385192237402188,-.0385192237402188,-.0350967674986961,-.0285561587634472,-.0194797896774684,-.00869007782643163,.00471291107004194,.0105645217737627,.0154869311233072,.0190341153866528,.0208902244146234,.0208902244146234,.0190341153866528,.0154869311233072,.0105645217737627,.00471291107004194,.00928552230067672,.0208145456318542,.0305128278843661,.0375015993872437,.0411585624649609,.0411585624649609,.0375015993872437,.030512827884366,.0208145456318542,.00928552230067672,-.000280476734556647,-.00062872023792106,-.000921664721700392,-.0011327662350258,-.00124322777172983,-.00124322777172983,-.0011327662350258,-.000921664721700392,-.00062872023792106,-.000280476734556647,-.00502787889479991,-.0112705576809204,-.0165219358020696,-.0203061813837464,-.0222863357447108,-.0222863357447108,-.0203061813837464,-.0165219358020696,-.0112705576809204,-.00502787889479991},
			{.00302011496343164,.00561961421020581,.00536170126189668,.00190531860926006,-.00385485822902238,-.0100772652494075,-.0145995645209458,-.0156902564956459,-.0126653126632867,-.00616325300098763,.00676992835558772,.0125969991373797,.0120188581714983,.00427098658012745,-.00864109954336841,-.0225893267590133,-.032726570685666,-.0351714797755983,-.0283907269656191,-.0138156268086686,.00992429344157325,.0184664163867283,.0176188977285336,.00626100039466762,-.0126673138949028,-.0331145465105248,-.0479751149143058,-.0515591994085577,-.0416190379908282,-.0202528486753225,.0121973904961468,.0226960330283889,.0216543955468261,.007695043194751,-.0155686825488407,-.040699225317108,-.0589635135390092,-.0633685100664642,-.0511516171339274,-.0248916364078325,.0133868172784104,.024909233429215,.0237660208182528,.00844542422661216,-.0170868603913883,-.044668004428086,-.0647133320925609,-.0695478812236805,-.0561396597315967,-.0273189407568403,.0133868172784104,.024909233429215,.0237660208182528,.00844542422661216,-.0170868603913883,-.044668004428086,-.0647133320925609,-.0695478812236805,-.0561396597315967,-.0273189407568403,.0121973904961468,.0226960330283889,.0216543955468261,.007695043194751,-.0155686825488407,-.040699225317108,-.0589635135390092,-.0633685100664642,-.0511516171339274,-.0248916364078325,.00992429344157325,.0184664163867283,.0176188977285336,.00626100039466762,-.0126673138949028,-.0331145465105248,-.0479751149143058,-.0515591994085576,-.0416190379908282,-.0202528486753225,.00676992835558772,.0125969991373797,.0120188581714983,.00427098658012745,-.00864109954336841,-.0225893267590133,-.032726570685666,-.0351714797755983,-.0283907269656191,-.0138156268086686,.00302011496343164,.00561961421020581,.00536170126189668,.00190531860926006,-.00385485822902238,-.0100772652494075,-.0145995645209458,-.0156902564956459,-.0126653126632867,-.00616325300098763},
			{-.00509449780740388,-.00947947764207594,-.00904441574358152,-.00321399734606523,.00650258915087549,.016998891214194,.0246273570183889,.0264671968724815,.0213645534605355,.0103965177750484,-.0101436118237301,-.0188745083672825,-.0180082602728758,-.00639936313911146,.0129472506789226,.0338463496166405,.0490353238499585,.0526986135407738,.0425387830722098,.0207004192788813,-.0116786295673759,-.0217307597449812,-.0207334236102965,-.0073677692785882,.014906538935309,.0389682675411454,.0564557667339239,.0606734166243637,.0489761140686684,.0238329830486871,-.00915612782195732,-.0170370686684944,-.0162551501156413,-.00577638299838646,.0116868314974169,.0305513961675494,.0442617186819942,.0475683858968708,.0383976183205756,.0186852265424968,-.00345189684709876,-.0064230430990214,-.00612825667404284,-.00217771951719025,.00440598226487886,.01151799867323,.0166868451529422,.0179334719317936,.0144760558386965,.00704440521621975,.00345189684709876,.0064230430990214,.00612825667404284,.00217771951719025,-.00440598226487886,-.01151799867323,-.0166868451529422,-.0179334719317936,-.0144760558386965,-.00704440521621975,.00915612782195732,.0170370686684944,.0162551501156413,.00577638299838646,-.0116868314974169,-.0305513961675494,-.0442617186819942,-.0475683858968708,-.0383976183205756,-.0186852265424968,.0116786295673759,.0217307597449812,.0207334236102965,.0073677692785882,-.014906538935309,-.0389682675411454,-.0564557667339239,-.0606734166243637,-.0489761140686684,-.0238329830486871,.0101436118237301,.0188745083672825,.0180082602728758,.00639936313911146,-.0129472506789226,-.0338463496166405,-.0490353238499585,-.0526986135407738,-.0425387830722098,-.0207004192788813,.00509449780740388,.00947947764207594,.00904441574358152,.00321399734606523,-.00650258915087549,-.016998891214194,-.0246273570183889,-.0264671968724815,-.0213645534605355,-.0103965177750484},
			{.00623143749421823,.0115950138046163,.0110628590900448,.00393126551933778,-.00795377275172705,-.0207925358056529,-.0301234472384809,-.0323738843735185,-.0261324834192153,-.0127167098940724,.00942344685839383,.017534476838053,.0167297296705077,.00594502821252533,-.0120280360541503,-.0314433638141471,-.045553968006072,-.0489571754954932,-.0395186614657246,-.0192307537404636,.00426953652352492,.00794444860844995,.00757983707344909,.00269354891773532,-.00544961307801359,-.0142462298821585,-.0206394043619152,-.0221813156064483,-.0179049525109209,-.00871299076692808,-.00595266514038301,-.0110762941201031,-.0105679461150626,-.00375539468000193,.00759794924327869,.0198623516941821,.0287758313312213,.0309255919116317,.0249634090409368,.0121478095153878,-.0139717558497607,-.0259976453431518,-.0248044799213386,-.00881444804152065,.0178334727861298,.0466197781881845,.0675409888263679,.0725867841561221,.0585926888330286,.0285126451187338,-.0139717558497607,-.0259976453431518,-.0248044799213386,-.00881444804152065,.0178334727861298,.0466197781881845,.0675409888263679,.0725867841561221,.0585926888330287,.0285126451187339,-.00595266514038299,-.011076294120103,-.0105679461150626,-.00375539468000192,.00759794924327867,.0198623516941821,.0287758313312212,.0309255919116316,.0249634090409367,.0121478095153878,.0042695365235249,.00794444860844991,.00757983707344906,.0026935489177353,-.00544961307801357,-.0142462298821584,-.0206394043619151,-.0221813156064482,-.0179049525109208,-.00871299076692804,.00942344685839383,.017534476838053,.0167297296705077,.00594502821252533,-.0120280360541503,-.0314433638141471,-.045553968006072,-.0489571754954932,-.0395186614657246,-.0192307537404636,.00623143749421823,.0115950138046163,.0110628590900448,.00393126551933778,-.00795377275172705,-.0207925358056529,-.0301234472384809,-.0323738843735185,-.0261324834192153,-.0127167098940724},
			{-.00677992663267428,-.0126156032172487,-.0120366084146251,-.00427729425504189,.00865386128958332,.0226226881679193,.0327749034455745,.0352234233386421,.0284326562653077,.0138360306383256,-.00574597286050269,-.0106916958888372,-.0102009990712324,-.00362499744280574,.00733412834718052,.0191726783027454,.0277766583485262,.02985177355496,.0240966134449184,.0117260054351862,.00617255053363847,.0114854411544103,.0109583152913243,.00389411513824503,-.00787860975716505,-.0205960467551767,-.0298387812602527,-.0320679518090523,-.0258855319698288,-.0125965372380952,.0144116920845997,.0268162472661978,.025585511999273,.009091993335417,-.018395004991244,-.0480877203642201,-.0696676884796602,-.0748723634966211,-.0604376285236068,-.0294104382002705,.00761712583185733,.0141734036895712,.0135229134252192,.00480546329270283,-.00972245777063737,-.025416183945038,-.0368220155169205,-.0395728836513071,-.0319435787790183,-.0155445319832301,-.0076171258318574,-.0141734036895713,-.0135229134252194,-.00480546329270287,.00972245777063746,.0254161839450382,.0368220155169208,.0395728836513074,.0319435787790186,.0155445319832303,-.0144116920845997,-.0268162472661979,-.0255855119992731,-.00909199333541702,.018395004991244,.0480877203642202,.0696676884796604,.0748723634966213,.060437628523607,.0294104382002706,-.0061725505336384,-.0114854411544102,-.0109583152913242,-.00389411513824499,.00787860975716497,.0205960467551765,.0298387812602524,.0320679518090519,.0258855319698285,.012596537238095,.00574597286050274,.0106916958888373,.0102009990712325,.00362499744280577,-.00733412834718057,-.0191726783027455,-.0277766583485264,-.0298517735549602,-.0240966134449186,-.0117260054351863,.00677992663267429,.0126156032172487,.0120366084146252,.0042772942550419,-.00865386128958333,-.0226226881679194,-.0327749034455745,-.0352234233386422,-.0284326562653077,-.0138360306383256},
			{-.00345430310158809,-.00436440955976362,.000321510543493063,.00721105333392678,.0101264765379077,.00504590248017281,-.00626512731169085,-.0168673409888969,-.0192939898358052,-.0109524916466014,-.00774321004312494,-.00978331632795496,.00072070214921253,.0161643894454719,.022699639413156,.0113109595805708,-.0140439895674001,-.0378100474408264,-.043249654554018,-.0245512454527271,-.0113510637943875,-.0143417325812942,.0010565044764254,.0236959884558197,.0332762579929111,.016581162471344,-.0205876142606789,-.0554271753160955,-.0634013006484006,-.035990597157184,-.0139509536333168,-.0176266163143223,.00129849018831963,.0291234057204582,.0408979757984384,.0203789735494738,-.0253030779470516,-.0681224215516175,-.0779229701867376,-.0442340172930329,-.015311378873029,-.0193454732724202,.00142511227252541,.0319633704461579,.0448861360340558,.0223662262280236,-.0277705039586591,-.074765369704482,-.0855216174320572,-.0485474911358245,-.015311378873029,-.0193454732724202,.00142511227252541,.0319633704461579,.0448861360340558,.0223662262280236,-.0277705039586591,-.074765369704482,-.0855216174320572,-.0485474911358245,-.0139509536333168,-.0176266163143223,.00129849018831963,.0291234057204581,.0408979757984384,.0203789735494738,-.0253030779470516,-.0681224215516175,-.0779229701867376,-.0442340172930329,-.0113510637943875,-.0143417325812942,.0010565044764254,.0236959884558197,.0332762579929111,.016581162471344,-.0205876142606789,-.0554271753160955,-.0634013006484006,-.035990597157184,-.00774321004312494,-.00978331632795496,.00072070214921253,.0161643894454719,.022699639413156,.0113109595805708,-.0140439895674001,-.0378100474408264,-.043249654554018,-.0245512454527271,-.00345430310158809,-.00436440955976362,.000321510543493063,.00721105333392678,.0101264765379077,.00504590248017282,-.00626512731169085,-.0168673409888969,-.0192939898358052,-.0109524916466014},
			{.00582691049477723,.00736212866080022,-.000542341857377823,-.0121640056225925,-.0170819035500163,-.00851170880280428,.0105683650247204,.0284527684272998,.0325461745984863,.0184752717531363,.0116019125780642,.0146586725826016,-.00107985232008665,-.0242196494967433,-.0340116347817334,-.0169475919543621,.0210425828953059,.0566520683977071,.0648024150672126,.0367859585157503,.0133576128135457,.0168769477792085,-.00124326477125837,-.0278847731596392,-.0391585650653061,-.0195122459271369,.0242269257780715,.0652251419454623,.0746088684970023,.0423527230982234,.0104724625103868,.0132316459066911,-.000974728335761049,-.0218618585222674,-.0307005907665702,-.0152977382124873,.0189940804164865,.0511369706019496,.0584938782987402,.0332048331576998,.0039481602948217,.00498838348212032,-.000367476485083857,-.00824200819081637,-.0115742456343885,-.00576730855338134,.00716084436326645,.0192788426530853,.0220524263094917,.0125183550420337,-.0039481602948217,-.00498838348212032,.000367476485083857,.00824200819081637,.0115742456343885,.00576730855338134,-.00716084436326645,-.0192788426530853,-.0220524263094917,-.0125183550420337,-.0104724625103868,-.0132316459066911,.000974728335761049,.0218618585222673,.0307005907665701,.0152977382124873,-.0189940804164865,-.0511369706019496,-.0584938782987402,-.0332048331576998,-.0133576128135457,-.0168769477792085,.00124326477125837,.0278847731596392,.0391585650653061,.0195122459271369,-.0242269257780715,-.0652251419454623,-.0746088684970023,-.0423527230982234,-.0116019125780642,-.0146586725826016,.00107985232008665,.0242196494967433,.0340116347817334,.0169475919543621,-.0210425828953059,-.0566520683977071,-.0648024150672126,-.0367859585157503,-.00582691049477723,-.00736212866080022,.000542341857377823,.0121640056225925,.0170819035500163,.00851170880280428,-.0105683650247204,-.0284527684272998,-.0325461745984863,-.0184752717531363},
			{-.00712730281031248,-.00900513579898562,.000663376354880294,.0148786482195442,.0208940740185803,.0104112678794644,-.0129269083176432,-.0348025761773465,-.039809508295799,-.0225984003710785,-.0107782127862115,-.013617952315673,.00100318615618461,.0225001295369971,.0315969142516674,.0157443655146451,-.0195486248056994,-.0526299473350607,-.0602016446818767,-.0341742696094045,-.00488334829500983,-.00616996579501025,.000454519455363577,.0101942660987876,.0143157998825042,.00713339233665088,-.00885701048107196,-.0238453599574428,-.0272759134325242,-.0154835374417349,.00680845262802778,.008602278046562,-.000633699255799995,-.0142130304082612,-.0199593474485439,-.00994550477811579,.012348604408976,.0332456326808434,.0380285724615008,.0215874283007975,.015980411393903,.0201907760278596,-.00148738272276524,-.0333600137192788,-.0468474410864738,-.0233435211431231,.0289839394319417,.0780322514257342,.0892584946770308,.0506687795348675,.015980411393903,.0201907760278596,-.00148738272276524,-.0333600137192788,-.0468474410864738,-.0233435211431231,.0289839394319417,.0780322514257343,.0892584946770308,.0506687795348676,.00680845262802777,.00860227804656198,-.000633699255799993,-.0142130304082611,-.0199593474485438,-.00994550477811577,.012348604408976,.0332456326808434,.0380285724615007,.0215874283007975,-.00488334829500981,-.00616996579501022,.000454519455363575,.0101942660987876,.0143157998825041,.00713339233665084,-.00885701048107192,-.0238453599574427,-.0272759134325241,-.0154835374417348,-.0107782127862115,-.013617952315673,.00100318615618461,.0225001295369971,.0315969142516674,.0157443655146451,-.0195486248056994,-.0526299473350607,-.0602016446818768,-.0341742696094045,-.00712730281031249,-.00900513579898562,.000663376354880294,.0148786482195442,.0208940740185803,.0104112678794644,-.0129269083176432,-.0348025761773465,-.039809508295799,-.0225984003710785},
			{.00360886668167895,.00181411871372813,-.00573065302975065,-.00778508692673576,.00135437648599479,.0117852891818123,.00860428289223646,-.00859845646519095,-.022302893606761,-.0160855169229064,.00808968174247022,.00406655172706256,-.0128459051764245,-.0174511504939989,.00303598766527052,.0264180550663856,.0192874706549125,-.0192744100614766,-.0499944517570629,-.0360575006081668,.011858969732663,.00596131162915412,-.0188312971419022,-.0255823000331391,.00445056888235353,.0387272238144227,.0282742310510711,-.0282550850343082,-.0732887533851304,-.0528580507812608,.0145751922353826,.00732671260055841,-.0231444874278468,-.0314417651121289,.00546994372019093,.047597451090816,.0347502660152623,-.0347267347236618,-.0900750818460811,-.0649648551848937,.015996490012713,.00804117592743191,-.0254014187950224,-.0345078043209106,.00600334449639179,.0522389096972928,.0381389332143976,-.0381131072722766,-.0988587405143971,-.0712998937070431,.015996490012713,.00804117592743191,-.0254014187950224,-.0345078043209106,.00600334449639179,.0522389096972928,.0381389332143976,-.0381131072722766,-.0988587405143971,-.0712998937070431,.0145751922353826,.00732671260055841,-.0231444874278468,-.0314417651121289,.00546994372019093,.0475974510908161,.0347502660152623,-.0347267347236618,-.0900750818460811,-.0649648551848937,.011858969732663,.00596131162915412,-.0188312971419022,-.0255823000331391,.00445056888235353,.0387272238144227,.0282742310510711,-.0282550850343082,-.0732887533851304,-.0528580507812608,.00808968174247022,.00406655172706256,-.0128459051764245,-.0174511504939989,.00303598766527052,.0264180550663856,.0192874706549125,-.0192744100614766,-.0499944517570629,-.0360575006081668,.00360886668167895,.00181411871372813,-.00573065302975065,-.00778508692673576,.00135437648599479,.0117852891818123,.00860428289223646,-.00859845646519095,-.022302893606761,-.0160855169229064},
			{-.00608763693165392,-.00306015629121496,.00966678409139083,.0131323173925849,-.00228463754490087,-.0198800809233674,-.0145141826854097,.0145043543443846,.0376217607296319,.0271339441220568,-.0121210427981259,-.00609305150606168,.0192474526663595,.0261476469344647,-.00454892263958079,-.0395830622633527,-.0288990673203423,.0288794981930212,.0749083720110797,.0540261683930699,-.0139553022404171,-.0070151039600872,.0221601411521044,.0301045315921654,-.00523730436076351,-.0455731084104202,-.0332723203472142,.0332497898528585,.086244144927353,.0622018683849159,-.0109410552299944,-.00549989090520571,.0173737067153911,.0236021647649496,-.00410608350003688,-.0357296379204776,-.0260857334565373,.0260680693903023,.0676160169559761,.0487667028405305,-.00412482162620984,-.00207348089107479,.00654995699049303,.00889811061193973,-.00154800991896502,-.0134702165461154,-.00983442595208666,.00982776652839477,.0254915091054086,.0183852422172328,.00412482162620984,.00207348089107479,-.00654995699049303,-.00889811061193973,.00154800991896502,.0134702165461154,.00983442595208666,-.00982776652839477,-.0254915091054086,-.0183852422172328,.0109410552299944,.00549989090520571,-.0173737067153911,-.0236021647649496,.00410608350003688,.0357296379204776,.0260857334565373,-.0260680693903023,-.0676160169559761,-.0487667028405305,.0139553022404171,.0070151039600872,-.0221601411521044,-.0301045315921654,.00523730436076351,.0455731084104202,.0332723203472142,-.0332497898528585,-.086244144927353,-.0622018683849159,.0121210427981259,.00609305150606168,-.0192474526663595,-.0261476469344647,.00454892263958079,.0395830622633527,.0288990673203423,-.0288794981930212,-.0749083720110797,-.0540261683930699,.00608763693165392,.00306015629121496,-.00966678409139083,-.0131323173925849,.00228463754490087,.0198800809233674,.0145141826854097,-.0145043543443846,-.0376217607296319,-.0271339441220568},
			{-.0035177389049427,.00124629807811922,.00733218354822258,-.000640159716652238,-.0104475716905616,-.00111787975143421,.0140189382881418,.00476714217793534,-.0203246135001665,-.0211329311052817,-.00788540855182074,.00279371772291301,.0164359164841472,-.00143499021406278,-.0234194104169471,-.00250585355822768,.0314250314909489,.0106860869193045,-.0455599137506981,-.0473718488397425,-.0115595179541847,.00409541623172844,.0240940301879616,-.00210360630454689,-.0343313974681262,-.00367342529008241,.0460671394948289,.0156651380574512,-.0667879967830832,-.0694441808039846,-.0142071529086165,.00503344558650323,.0296126164102953,-.00258542411082186,-.0421947883405973,-.00451480027120738,.0566185283382572,.0192531394993187,-.0820853677911323,-.0853499340721777,-.0155925613838605,.0055242813098938,.0325002864461468,-.00283754137163211,-.0463094070649962,-.00495506037117232,.0621396759967787,.0211306066321764,-.0900899106409799,-.0936728206340164,-.0155925613838605,.0055242813098938,.0325002864461468,-.00283754137163211,-.0463094070649962,-.00495506037117232,.0621396759967787,.0211306066321764,-.0900899106409799,-.0936728206340164,-.0142071529086165,.00503344558650323,.0296126164102953,-.00258542411082186,-.0421947883405973,-.00451480027120738,.0566185283382572,.0192531394993187,-.0820853677911323,-.0853499340721777,-.0115595179541847,.00409541623172844,.0240940301879616,-.00210360630454689,-.0343313974681262,-.00367342529008241,.0460671394948289,.0156651380574512,-.0667879967830832,-.0694441808039846,-.00788540855182074,.00279371772291301,.0164359164841472,-.00143499021406278,-.0234194104169471,-.00250585355822768,.0314250314909489,.0106860869193045,-.0455599137506981,-.0473718488397425,-.0035177389049427,.00124629807811922,.00733218354822258,-.000640159716652238,-.0104475716905616,-.00111787975143421,.0140189382881418,.00476714217793534,-.0203246135001665,-.0211329311052817},
			{-.00537304737318019,-.0113801575780441,-.0150219679612532,-.0157580888060425,-.0138618270499793,-.0102693186804868,-.00622899008451743,-.002867614478685,-.000823350121801073,-7.1027404987251e-005,-.012044291759767,-.0255099068782191,-.0336734356436761,-.035323533550749,-.0310728489286636,-.0230198361881556,-.0139629838964585,-.00642808131717099,-.0018456321708443,-.000159215940078199,-.0176561817853201,-.0373959351161665,-.0493631599864637,-.0517821007753423,-.0455508617870686,-.0337456465280969,-.0204688649908742,-.00942316696826571,-.00270558184468062,-.000233400654618126,-.0217002192825492,-.0459612390812321,-.0606694816133146,-.0636424655907344,-.0559840004655092,-.0414748748282566,-.0251571299031463,-.0115814841529019,-.00332527836597537,-.000286859608010771,-.0238163130489831,-.050443142708611,-.066585657398596,-.0698485514815503,-.0614432722296599,-.0455192912899041,-.0276103238122155,-.0127108509165695,-.00364954240820819,-.000314832681482957,-.0238163130489831,-.050443142708611,-.066585657398596,-.0698485514815503,-.0614432722296599,-.0455192912899041,-.0276103238122155,-.0127108509165695,-.00364954240820819,-.000314832681482957,-.0217002192825492,-.0459612390812321,-.0606694816133146,-.0636424655907344,-.0559840004655092,-.0414748748282566,-.0251571299031463,-.0115814841529019,-.00332527836597537,-.000286859608010771,-.0176561817853201,-.0373959351161665,-.0493631599864637,-.0517821007753423,-.0455508617870686,-.0337456465280969,-.0204688649908742,-.00942316696826571,-.00270558184468062,-.000233400654618126,-.012044291759767,-.0255099068782191,-.0336734356436761,-.035323533550749,-.0310728489286636,-.0230198361881556,-.0139629838964585,-.00642808131717099,-.0018456321708443,-.000159215940078199,-.00537304737318019,-.0113801575780441,-.0150219679612533,-.0157580888060425,-.0138618270499793,-.0102693186804868,-.00622899008451743,-.002867614478685,-.000823350121801073,-7.1027404987251e-005},
			{.00906355498937132,.0191966824099048,.0253398906075171,.0265816201684498,.0233829004277474,.0173228576075033,.0105074067355295,.00483725150937865,.00138887275435145,.000119812975047951,.0180463682645229,.0382223532192926,.0504540434986315,.052926440804966,.0465574967776602,.0344913963831515,.0209212093573502,.00963141087892507,.00276537288370632,.000238558608969547,.0207772984278405,.0440064852834015,.0580891791243752,.060935720650788,.0536029737631114,.0397109282788219,.0240871849625023,.0110889180127161,.0031838526636951,.000274659329702082,.0162895482815631,.0345013943567046,.0455423254986296,.0477740341007375,.0420251089034541,.031133647415719,.0188845226330227,.00869378981034385,.00249616290908881,.000215334848642728,.00614122492031466,.0130071638050768,.0171696390499558,.0180110021280012,.0158436343117591,.0117375097249386,.0071195406402661,.00327759356566627,.000941063410568089,8.11821000704882e-005,-.00614122492031466,-.0130071638050768,-.0171696390499558,-.0180110021280012,-.0158436343117591,-.0117375097249386,-.0071195406402661,-.00327759356566627,-.000941063410568089,-8.11821000704882e-005,-.0162895482815631,-.0345013943567046,-.0455423254986296,-.0477740341007375,-.0420251089034541,-.031133647415719,-.0188845226330227,-.00869378981034385,-.00249616290908881,-.000215334848642728,-.0207772984278405,-.0440064852834015,-.0580891791243752,-.060935720650788,-.0536029737631114,-.0397109282788219,-.0240871849625023,-.0110889180127161,-.0031838526636951,-.000274659329702082,-.0180463682645229,-.0382223532192926,-.0504540434986315,-.052926440804966,-.0465574967776602,-.0344913963831515,-.0209212093573502,-.00963141087892507,-.00276537288370632,-.000238558608969547,-.00906355498937132,-.0191966824099048,-.0253398906075171,-.0265816201684498,-.0233829004277474,-.0173228576075033,-.0105074067355295,-.00483725150937865,-.00138887275435145,-.000119812975047951},
			{-.0110862696476229,-.0234808083346496,-.0309949970453799,-.0325138435419332,-.0286012651315648,-.0211888018254831,-.0128523459618144,-.0059167815111433,-.00169882765416366,-.00014655165111517,-.0167651321190546,-.035508684751954,-.0468719629787211,-.0491688277486249,-.0432520589832132,-.0320426142733104,-.0194358684154698,-.00894761059462107,-.00256903999044115,-.000221621688025442,-.00759587706914969,-.0160881287631775,-.0212365561003511,-.0222772101383125,-.019596464894583,-.0145177358141793,-.00880592327979835,-.00405394658728139,-.00116397006683541,-.000100411442399873,.0105903093675451,.0224303604700153,.0296083910990837,.0310592898045613,.0273217462387268,.0202408901814494,.0122773777077953,.00565208574704063,.00162282814612863,.000139995451397573,.0248569696711101,.052647261809357,.0694951255924348,.0729005922191637,.0641280432939063,.0475082621191814,.0288167601844263,.0132662530542551,.00380900959639235,.000328589332824135,.0248569696711101,.0526472618093571,.0694951255924348,.0729005922191637,.0641280432939063,.0475082621191814,.0288167601844263,.0132662530542551,.00380900959639235,.000328589332824135,.0105903093675451,.0224303604700152,.0296083910990836,.0310592898045612,.0273217462387267,.0202408901814493,.0122773777077953,.00565208574704061,.00162282814612863,.000139995451397573,-.00759587706914966,-.0160881287631774,-.021236556100351,-.0222772101383123,-.0195964648945829,-.0145177358141792,-.0088059232797983,-.00405394658728137,-.00116397006683541,-.000100411442399872,-.0167651321190546,-.035508684751954,-.0468719629787211,-.0491688277486249,-.0432520589832132,-.0320426142733104,-.0194358684154698,-.00894761059462107,-.00256903999044115,-.000221621688025442,-.0110862696476229,-.0234808083346496,-.0309949970453799,-.0325138435419332,-.0286012651315648,-.0211888018254831,-.0128523459618144,-.0059167815111433,-.00169882765416366,-.00014655165111517},
			{.0120620795615023,.0255475815854486,.0337231667867706,.0353757016668926,.0311187393543622,.0230538333953191,.0139836053488558,.00643757473017433,.00184835792174656,.00015945108064368,.0102225858120058,.0216515190199873,.0285803094377376,.0299808291021325,.0263730629357406,.0195380729315951,.01185107467672,.00545583037857288,.00156647926006563,.00013513443899835,-.0109815046191396,-.0232589151611855,-.0307020949375013,-.0322065883647599,-.0283309837428373,-.0209885680681119,-.0127308915471588,-.00586086804311059,-.00168277376649596,-.000145166740916215,-.0256396545211207,-.0543049946206344,-.0716833562042298,-.0751960526007667,-.0661472776821311,-.0490041804673553,-.0297241291003144,-.0136839747403398,-.00392894594197619,-.000338935802891862,-.0135515298013773,-.028702249180423,-.0378873723537952,-.039743965618796,-.0349613448984257,-.025900567866404,-.0157103295206683,-.00723249961664868,-.0020765969360793,-.000179140426009222,.0135515298013774,.0287022491804232,.0378873723537955,.0397439656187963,.034961344898426,.0259005678664042,.0157103295206684,.00723249961664874,.00207659693607932,.000179140426009223,.0256396545211207,.0543049946206346,.07168335620423,.0751960526007669,.0661472776821312,.0490041804673554,.0297241291003145,.0136839747403399,.0039289459419762,.000338935802891863,.0109815046191395,.0232589151611852,.030702094937501,.0322065883647595,.028330983742837,.0209885680681117,.0127308915471586,.00586086804311052,.00168277376649595,.000145166740916214,-.0102225858120059,-.0216515190199875,-.0285803094377378,-.0299808291021328,-.0263730629357408,-.0195380729315952,-.0118510746767201,-.00545583037857292,-.00156647926006564,-.000135134438998351,-.0120620795615023,-.0255475815854486,-.0337231667867706,-.0353757016668926,-.0311187393543623,-.0230538333953191,-.0139836053488558,-.00643757473017434,-.00184835792174656,-.00015945108064368},
			{.00615133467063503,.00923599360757013,.00365245862188397,-.00803852164203684,-.0191471282194023,-.0235470344199801,-.0197092364291758,-.0112335234696543,-.00369340218158754,-.000342287092850262,.013788910526811,.0207035214794833,.00818739799355415,-.0180192529954539,-.0429204476750506,-.0527833337273658,-.0441804766323359,-.025181209984126,-.00827917760101021,-.000767275128164547,.0202136842696012,.0303500733898785,.0120022156724767,-.0264151029275269,-.0629187038618097,-.0773770807046292,-.0647658278578351,-.0369140859356472,-.0121367588623754,-.00112477756371539,.0248435016410661,.037301566998444,.0147512477576403,-.0324653162766497,-.0773298376385226,-.095099814849551,-.0796000337796347,-.0453690253735618,-.0149166072197978,-.00138240079726687,.0272661121352065,.040939023946499,.016189713563936,-.0356311669302222,-.0848706456485974,-.104373459634974,-.0873622196403312,-.0497931793662997,-.0163711980302944,-.00151720541245167,.0272661121352065,.040939023946499,.016189713563936,-.0356311669302222,-.0848706456485974,-.104373459634974,-.0873622196403312,-.0497931793662997,-.0163711980302944,-.00151720541245167,.0248435016410661,.037301566998444,.0147512477576403,-.0324653162766497,-.0773298376385226,-.095099814849551,-.0796000337796348,-.0453690253735618,-.0149166072197978,-.00138240079726687,.0202136842696012,.0303500733898785,.0120022156724767,-.0264151029275269,-.0629187038618097,-.0773770807046292,-.0647658278578351,-.0369140859356472,-.0121367588623754,-.00112477756371539,.013788910526811,.0207035214794833,.00818739799355415,-.0180192529954539,-.0429204476750506,-.0527833337273658,-.0441804766323359,-.025181209984126,-.00827917760101021,-.000767275128164547,.00615133467063503,.00923599360757013,.00365245862188397,-.00803852164203684,-.0191471282194022,-.0235470344199801,-.0197092364291758,-.0112335234696543,-.00369340218158754,-.000342287092850262},
			{-.0103764132698787,-.015579787438262,-.00616117024045563,.0135598251603545,.0322984402530261,.0397204466191965,.0332466356283884,.0189493318506081,.0062302361129812,.000577388895667578,-.0206603893673846,-.0310207839996088,-.0122674543520802,.0269988540625906,.0643091532913213,.0790870478705028,.0661970971445412,.0377298555968102,.0124049708311625,.00114963418385859,-.0237868954700111,-.0357151132670011,-.0141238700378378,.031084550633434,.0740409621476184,.0910551736114528,.0762146057460633,.0434394587256295,.0142821966818892,.00132360662105264,-.0186490935562653,-.0280009003078451,-.0110732135702299,.0243705065946171,.0580486357216159,.0713878973236672,.0597527876096367,.0340568415423994,.0111973427741056,.00103771691176861,-.00703078293574547,-.0105564515227481,-.00417464584959855,.00918777855792046,.0218845680753062,.0269135231054363,.0225270401599717,.0128395656142358,.00422144305643347,.000391223430424914,.00703078293574547,.0105564515227481,.00417464584959855,-.00918777855792046,-.0218845680753062,-.0269135231054363,-.0225270401599717,-.0128395656142358,-.00422144305643347,-.000391223430424914,.0186490935562653,.0280009003078451,.0110732135702299,-.0243705065946171,-.0580486357216159,-.0713878973236672,-.0597527876096368,-.0340568415423994,-.0111973427741056,-.00103771691176861,.0237868954700111,.0357151132670011,.0141238700378378,-.031084550633434,-.0740409621476184,-.0910551736114528,-.0762146057460633,-.0434394587256295,-.0142821966818892,-.00132360662105264,.0206603893673846,.0310207839996088,.0122674543520802,-.0269988540625906,-.0643091532913213,-.0790870478705028,-.0661970971445413,-.0377298555968102,-.0124049708311625,-.00114963418385859,.0103764132698787,.015579787438262,.00616117024045563,-.0135598251603545,-.0322984402530261,-.0397204466191965,-.0332466356283884,-.0189493318506081,-.0062302361129812,-.000577388895667578},
			{.0126921186677361,.0190567304753116,.00753615934428767,-.0165859729728156,-.0395064870502198,-.0485848634700931,-.040666291301582,-.0231782565195614,-.00762063865589431,-.000706244844981167,.0191935658340169,.0288184045969256,.0113965031604947,-.0250820192048949,-.0597434029668103,-.0734721128884853,-.0614973086649725,-.0350511529297959,-.0115242563962926,-.00106801372426019,.00869614182330973,.0130569241621705,.00516348075342068,-.0113640580446065,-.027068295162004,-.0332884425573554,-.027862947538109,-.0158808321280148,-.00522136266376985,-.000483891263130864,-.0121243184130687,-.0182042001216115,-.00719901837461035,.0158439755235,.0377391073088729,.0464113494513956,.0388470260424168,.0221413437472276,.00727971838223315,.000674649962098675,-.0284575081442016,-.0427278594614471,-.0168971250214698,.0371880750022642,.0885790785929488,.108934070353419,.0911795222062535,.0519688982542005,.0170865394731464,.00158349988319459,-.0284575081442016,-.0427278594614471,-.0168971250214699,.0371880750022643,.0885790785929488,.108934070353419,.0911795222062535,.0519688982542005,.0170865394731464,.00158349988319459,-.0121243184130687,-.0182042001216115,-.00719901837461033,.0158439755235,.0377391073088728,.0464113494513955,.0388470260424167,.0221413437472276,.00727971838223313,.000674649962098674,.00869614182330968,.0130569241621705,.00516348075342066,-.0113640580446064,-.0270682951620039,-.0332884425573552,-.0278629475381089,-.0158808321280148,-.00522136266376982,-.000483891263130862,.0191935658340169,.0288184045969256,.0113965031604947,-.0250820192048949,-.0597434029668103,-.0734721128884853,-.0614973086649725,-.0350511529297959,-.0115242563962926,-.00106801372426019,.0126921186677361,.0190567304753116,.00753615934428767,-.0165859729728156,-.0395064870502198,-.0485848634700931,-.040666291301582,-.0231782565195614,-.00762063865589432,-.000706244844981167},
			{-.00643614945112683,-.00460242005009466,.00815688456903964,.0159153508876072,.00595117342612003,-.0150500971453785,-.0280693431449012,-.0232405097467362,-.00939844620771058,-.000965522727863794,-.0144273549840217,-.010316843689307,.0182845768475155,.0356760542455353,.0133402265193015,-.0337364903828284,-.0629205988448425,-.0520962240967035,-.021067677307141,-.00216432810421968,-.0211496040913043,-.0151238504731025,.0268040511743329,.0522988741642954,.0195559414517781,-.049455594307975,-.0922376801728386,-.0763698207687696,-.0308839031591627,-.00317277023949471,-.0259937880171798,-.0185878733948966,.0329433506754726,.0642776026773378,.024035106963676,-.0607830874353671,-.113364141250993,-.093861848354391,-.0379576671221214,-.00389947332709414,-.028528568522066,-.0204004671991369,.036155816784784,.0705456238699261,.0263788869669642,-.0667103414760682,-.124418829202062,-.103014773014854,-.0416591035794981,-.00427972990848616,-.028528568522066,-.0204004671991369,.036155816784784,.0705456238699261,.0263788869669642,-.0667103414760682,-.124418829202062,-.103014773014854,-.0416591035794981,-.00427972990848616,-.0259937880171798,-.0185878733948966,.0329433506754726,.0642776026773378,.024035106963676,-.0607830874353671,-.113364141250993,-.093861848354391,-.0379576671221214,-.00389947332709414,-.0211496040913043,-.0151238504731025,.0268040511743329,.0522988741642954,.019555941451778,-.049455594307975,-.0922376801728386,-.0763698207687696,-.0308839031591627,-.00317277023949471,-.0144273549840217,-.010316843689307,.0182845768475155,.0356760542455353,.0133402265193015,-.0337364903828284,-.0629205988448425,-.0520962240967035,-.021067677307141,-.00216432810421968,-.00643614945112683,-.00460242005009466,.00815688456903964,.0159153508876072,.00595117342612003,-.0150500971453785,-.0280693431449012,-.0232405097467362,-.00939844620771058,-.000965522727863794},
			{.0108568546750866,.00776361798572502,-.0137594863264114,-.026846898600142,-.0100387701566362,.0253873404889165,.0473489283714467,.0392033837640413,.0158538215160414,.0016286974100757,.0216169922167163,.0154580746075647,-.0273963977344734,-.0534546344636977,-.0199881109986099,.050548520559216,.0942760538578598,.0780575283595627,.0315664109517744,.00324288574275418,.0248882596107707,.0177973221278318,-.0315422539999915,-.0615438450730705,-.0230128822120598,.0581979532585155,.108542709363562,.0898698584387518,.0363433091371754,.00373362683599065,.0195125707984483,.0139532258773057,-.0247293492571541,-.0482508079302309,-.0180422617114151,.0456276051857851,.0850982485008847,.0704586018812573,.0284934102938292,.00292718973168883,.00735631731313645,.00526042202001852,-.00932306367838869,-.0181907477705709,-.00680200490067545,.0172017898334461,.0320823803911286,.0265631750030447,.0107421297593173,.00110356224838248,-.00735631731313645,-.00526042202001852,.00932306367838869,.0181907477705709,.00680200490067545,-.0172017898334461,-.0320823803911286,-.0265631750030447,-.0107421297593173,-.00110356224838248,-.0195125707984483,-.0139532258773057,.0247293492571541,.0482508079302309,.0180422617114151,-.0456276051857851,-.0850982485008847,-.0704586018812573,-.0284934102938292,-.00292718973168883,-.0248882596107707,-.0177973221278318,.0315422539999915,.0615438450730705,.0230128822120598,-.0581979532585155,-.108542709363562,-.0898698584387518,-.0363433091371754,-.00373362683599065,-.0216169922167163,-.0154580746075647,.0273963977344734,.0534546344636977,.0199881109986099,-.050548520559216,-.0942760538578598,-.0780575283595627,-.0315664109517744,-.00324288574275418,-.0108568546750866,-.00776361798572502,.0137594863264114,.026846898600142,.0100387701566362,-.0253873404889165,-.0473489283714467,-.0392033837640413,-.0158538215160414,-.0016286974100757},
			{.00628780165042657,-.00108441059929728,-.0131544407550812,-.00292698561809638,.0175762393921458,.0118959849140124,-.0179938136163745,-.0323721609932466,-.0177761887834042,-.00209542799580947,.0140948166553123,-.00243082867843575,-.0294871627563918,-.00656116842314922,.0393991231745305,.0266661920364473,-.0403351628999706,-.0725658504041882,-.0398473322890685,-.00469713821417048,.0206621236068235,-.00356344348771085,-.0432263445908294,-.00961826437899792,.0577566613983443,.0390909771623532,-.0591288373535847,-.106377018399916,-.0584137080526403,-.00688571215596096,.0253946531907593,-.00437962782806131,-.0531270672113226,-.0118212673996889,.070985461783933,.0480445198574855,-.0726719260198027,-.130742006055988,-.0717930008459142,-.00846284125966016,.0278710091491036,-.00480670660664831,-.0583077455395459,-.0129740165922258,.0779075988938198,.0527295743105294,-.0797585184474156,-.143491293997453,-.0787938850114395,-.00928809400953851,.0278710091491036,-.00480670660664831,-.0583077455395459,-.0129740165922258,.0779075988938198,.0527295743105294,-.0797585184474156,-.143491293997453,-.0787938850114395,-.00928809400953851,.0253946531907593,-.00437962782806131,-.0531270672113226,-.0118212673996889,.070985461783933,.0480445198574855,-.0726719260198027,-.130742006055988,-.0717930008459142,-.00846284125966016,.0206621236068235,-.00356344348771085,-.0432263445908294,-.00961826437899792,.0577566613983443,.0390909771623532,-.0591288373535847,-.106377018399916,-.0584137080526403,-.00688571215596096,.0140948166553123,-.00243082867843575,-.0294871627563918,-.00656116842314922,.0393991231745305,.0266661920364473,-.0403351628999706,-.0725658504041882,-.0398473322890685,-.00469713821417048,.00628780165042657,-.00108441059929729,-.0131544407550812,-.00292698561809638,.0175762393921458,.0118959849140124,-.0179938136163745,-.0323721609932466,-.0177761887834042,-.00209542799580947},
			{-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.000726793993235288,-8.78323358755477e-005,-1.46520306329324e-006,-.0187952392881457,-.0376133897003122,-.0447078783667129,-.0400285772779044,-.0282223784536199,-.0154894473150464,-.00625458279978263,-.00162919071725786,-.000196886088238751,-3.28441793937936e-006,-.0275526505160409,-.0551388873133415,-.0655389659619086,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461161,-.00238829215039021,-.00028862274630013,-4.81475219575472e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030351,-.00293531546086303,-.00035472997281053,-5.91754093322042e-006,-.0371655977491602,-.0743764998227977,-.088405100787603,-.0791522778068628,-.0558067683550777,-.0306287437708179,-.0123677759491016,-.00322155232641602,-.000389321415157626,-6.49458908735937e-006,-.0371655977491602,-.0743764998227977,-.088405100787603,-.0791522778068628,-.0558067683550778,-.0306287437708179,-.0123677759491016,-.00322155232641602,-.000389321415157626,-6.49458908735937e-006,-.0338634119926563,-.0677681029932591,-.0805502543085171,-.0721195502254918,-.050848303356817,-.0279073614294862,-.0112688916030351,-.00293531546086303,-.00035472997281053,-5.91754093322042e-006,-.0275526505160409,-.0551388873133415,-.0655389659619087,-.0586794019211047,-.041372249554411,-.0227065653177011,-.00916882894461161,-.00238829215039021,-.00028862274630013,-4.81475219575472e-006,-.0187952392881457,-.0376133897003122,-.044707878366713,-.0400285772779045,-.0282223784536199,-.0154894473150464,-.00625458279978263,-.00162919071725786,-.000196886088238751,-3.28441793937936e-006,-.00838469484962211,-.0167796105206252,-.0199445142321406,-.0178570435094928,-.0125902111506184,-.00690995667226622,-.00279021551063458,-.000726793993235288,-8.78323358755477e-005,-1.46520306329324e-006},
			{.0141437693703256,.0283047797903233,.0336435153051858,.0301222536495194,.0212378680478842,.0116560990333334,.00470669062902145,.00122599650964295,.000148160466678965,2.47158597654252e-006,.028161540477737,.0563574094647636,.0669873209378089,.0599761664108691,.042286540810412,.023208360949976,.00937145220590666,.00244107135996656,.000295001061621808,4.92115409267359e-006,.0324231846605908,.0648858927129044,.0771244129347917,.0690522707843786,.0486857003592358,.0267204478158996,.0107896201789892,.00281047507100845,.000339643135062007,5.66586504797623e-006,.0254200050985947,.0508709968146246,.0604660828524796,.054137466562375,.0381699319272815,.020949019253574,.00845913820104087,.00220343224709027,.000266282609662119,4.44207809674427e-006,.00958344369577791,.0191785694703894,.0227959553223306,.0204100416431963,.0143902171568408,.00789786413180114,.00318912896948871,.000830702778994263,.000100389609953414,1.67468122713911e-006,-.00958344369577791,-.0191785694703894,-.0227959553223306,-.0204100416431963,-.0143902171568408,-.00789786413180114,-.00318912896948871,-.000830702778994263,-.000100389609953414,-1.67468122713911e-006,-.0254200050985947,-.0508709968146246,-.0604660828524796,-.054137466562375,-.0381699319272815,-.020949019253574,-.00845913820104087,-.00220343224709027,-.000266282609662119,-4.44207809674427e-006,-.0324231846605908,-.0648858927129044,-.0771244129347917,-.0690522707843785,-.0486857003592358,-.0267204478158996,-.0107896201789892,-.00281047507100845,-.000339643135062007,-5.66586504797623e-006,-.028161540477737,-.0563574094647636,-.066987320937809,-.0599761664108691,-.042286540810412,-.023208360949976,-.00937145220590666,-.00244107135996656,-.000295001061621808,-4.92115409267359e-006,-.0141437693703256,-.0283047797903233,-.0336435153051858,-.0301222536495194,-.0212378680478842,-.0116560990333334,-.00470669062902145,-.00122599650964295,-.000148160466678965,-2.47158597654252e-006},
			{-.0173002360880135,-.0346215609128356,-.0411517426769954,-.0368446406325781,-.0259775256237805,-.0142573920616249,-.00575708334485075,-.00149960229869156,-.000181225455912347,-3.02317011727578e-006,-.0261621585010403,-.0523562082937465,-.0622314290413362,-.0557180447272185,-.0392843276459713,-.0215606393479301,-.00870610818288429,-.00226776286909028,-.000274056901760265,-4.57176742452532e-006,-.0118534431119483,-.0237213353990866,-.0281955597772025,-.0252445024158391,-.0177987815081903,-.00976860575008604,-.00394452766839492,-.00102746866850002,-.000124168573259081,-2.07135757110846e-006,.0165262850469198,.0330727153955572,.0393107600496111,.0351963424341469,.0248153835062233,.0136195670415941,.0054995319088043,.00143251542459375,.000173118073471871,2.88792423694412e-006,.0387895529705973,.0776263898435274,.0922679721988114,.0826108460155768,.0582452517469885,.031967070378892,.0129081873926539,.00336231843912396,.000406332860775417,6.7783709312756e-006,.0387895529705973,.0776263898435274,.0922679721988115,.0826108460155769,.0582452517469885,.031967070378892,.0129081873926539,.00336231843912396,.000406332860775417,6.77837093127561e-006,.0165262850469198,.0330727153955571,.039310760049611,.0351963424341468,.0248153835062232,.013619567041594,.00549953190880429,.00143251542459375,.000173118073471871,2.88792423694411e-006,-.0118534431119483,-.0237213353990865,-.0281955597772024,-.025244502415839,-.0177987815081902,-.00976860575008599,-.0039445276683949,-.00102746866850002,-.00012416857325908,-2.07135757110845e-006,-.0261621585010403,-.0523562082937465,-.0622314290413362,-.0557180447272185,-.0392843276459713,-.0215606393479301,-.00870610818288428,-.00226776286909028,-.000274056901760265,-4.57176742452532e-006,-.0173002360880135,-.0346215609128356,-.0411517426769954,-.0368446406325781,-.0259775256237805,-.0142573920616249,-.00575708334485075,-.00149960229869156,-.000181225455912347,-3.02317011727578e-006},
			{.00879759938930934,.0102248339942577,-.00281118458446222,-.0202713999635782,-.0287696758053183,-.0241048298357547,-.0129418509611811,-.00409370526960594,-.000560621794885306,-9.99671086012965e-006,.0197208113889518,.0229201187461582,-.00630158734405165,-.0454406296060175,-.0644904734997843,-.0540336950704248,-.0290106187535114,-.00917650212647326,-.00125669699064766,-2.24087550090162e-005,.0289094815853316,.0335994669670076,-.00923773467978503,-.0666131336541757,-.0945390186691718,-.0792100325801501,-.0425277607544874,-.0134521807450406,-.00184223954039861,-3.28498395682792e-005,.0355310166928675,.0412952137572032,-.0113535797638883,-.081870453361282,-.116192586869701,-.0973525928350061,-.052268477136742,-.0165333182193771,-.00226419293161506,-4.03738889129983e-005,.0389958186821867,.0453221105953125,-.0124607224637962,-.0898540388613718,-.127523090289016,-.106845916942104,-.0573654301771659,-.0181455623707819,-.0024849853801249,-4.43109428912591e-005,.0389958186821867,.0453221105953125,-.0124607224637962,-.0898540388613718,-.127523090289016,-.106845916942104,-.0573654301771659,-.0181455623707819,-.0024849853801249,-4.43109428912591e-005,.0355310166928675,.0412952137572032,-.0113535797638883,-.081870453361282,-.116192586869701,-.0973525928350061,-.052268477136742,-.0165333182193771,-.00226419293161506,-4.03738889129983e-005,.0289094815853316,.0335994669670076,-.00923773467978504,-.0666131336541757,-.0945390186691718,-.0792100325801501,-.0425277607544874,-.0134521807450406,-.00184223954039861,-3.28498395682792e-005,.0197208113889518,.0229201187461582,-.00630158734405165,-.0454406296060175,-.0644904734997843,-.0540336950704248,-.0290106187535114,-.00917650212647326,-.00125669699064766,-2.24087550090162e-005,.00879759938930934,.0102248339942577,-.00281118458446222,-.0202713999635783,-.0287696758053183,-.0241048298357547,-.0129418509611811,-.00409370526960595,-.000560621794885306,-9.99671086012965e-006},
			{-.0148402797021791,-.0172478183727625,.00474206242882225,.0341949243312682,.0485302884342244,.0406613669372952,.0218310336296094,.00690548961493529,.000945688008179328,1.68630075889125e-005,-.0295483563533413,-.0343419863926998,.00944188069959736,.0680852268213982,.0966282499631283,.0809605064182685,.0434675878214585,.0137494624111731,.0018829514554676,3.35757928708797e-005,-.0340198653272789,-.0395389082959526,.0108707065122495,.0783884632885607,.111250859819299,.0932121398646372,.0500454734640585,.0158301481800593,.00216789570853973,3.86567678440353e-005,-.0266718139851343,-.0309987825377814,.00852271045732846,.0614571072313015,.0872214575290303,.0730789740555617,.0392360036170109,.0124109476493752,.00169964550185853,3.03071782114281e-005,-.0100553806578476,-.0116866651260001,.00321309596463221,.0231695754808925,.0328828386953925,.0275510658038934,.0147921304520804,.00467897695329769,.000640773159040156,1.14259275260014e-005,.0100553806578476,.0116866651260001,-.00321309596463221,-.0231695754808925,-.0328828386953925,-.0275510658038934,-.0147921304520804,-.00467897695329769,-.000640773159040156,-1.14259275260014e-005,.0266718139851343,.0309987825377814,-.00852271045732846,-.0614571072313014,-.0872214575290303,-.0730789740555617,-.0392360036170109,-.0124109476493752,-.00169964550185853,-3.03071782114281e-005,.0340198653272789,.0395389082959526,-.0108707065122495,-.0783884632885607,-.111250859819299,-.0932121398646372,-.0500454734640585,-.0158301481800593,-.00216789570853973,-3.86567678440353e-005,.0295483563533413,.0343419863926998,-.00944188069959736,-.0680852268213982,-.0966282499631283,-.0809605064182685,-.0434675878214585,-.0137494624111731,-.0018829514554676,-3.35757928708797e-005,.0148402797021791,.0172478183727625,-.00474206242882225,-.0341949243312682,-.0485302884342244,-.0406613669372952,-.0218310336296094,-.00690548961493529,-.000945688008179328,-1.68630075889125e-005},
			{-.00863131572555909,-.00182782839348211,.0164716603838428,.014834131431226,-.0115521085131657,-.0329015220239097,-.0289052514771828,-.0121493830289503,-.00197137671826113,-3.83525277872114e-005,-.0193480678000737,-.00409728352066521,.036923084732338,.0332523788739027,-.0258953542951368,-.0737524728656581,-.0647943816610287,-.0272342124940597,-.00441906328272248,-8.59715171504879e-005,-.0283630626928129,-.00600636252506896,.0541269432119395,.0487459169789592,-.0379609770295909,-.108116533043712,-.0949845291107984,-.0399236597855762,-.00647807162071696,-.000126028891149856,-.0348594439863812,-.00738208212111014,.0665243794538717,.0599108629733206,-.046655700294547,-.132879945598894,-.116740138678666,-.0490679232034869,-.0079618332211835,-.000154895016778896,-.0382587576596884,-.00810194479882727,.0730114947611087,.0657530621708869,-.0512053242073936,-.145837714407803,-.128124036533984,-.05385277468682,-.00873823024415192,-.000169999582091795,-.0382587576596884,-.00810194479882727,.0730114947611087,.0657530621708869,-.0512053242073936,-.145837714407803,-.128124036533984,-.05385277468682,-.00873823024415192,-.000169999582091795,-.0348594439863812,-.00738208212111014,.0665243794538717,.0599108629733206,-.046655700294547,-.132879945598894,-.116740138678666,-.0490679232034869,-.0079618332211835,-.000154895016778896,-.0283630626928129,-.00600636252506896,.0541269432119395,.0487459169789592,-.0379609770295909,-.108116533043712,-.0949845291107984,-.0399236597855762,-.00647807162071696,-.000126028891149856,-.0193480678000737,-.0040972835206652,.036923084732338,.0332523788739027,-.0258953542951368,-.0737524728656581,-.0647943816610287,-.0272342124940597,-.00441906328272248,-8.59715171504879e-005,-.00863131572555909,-.00182782839348211,.0164716603838428,.014834131431226,-.0115521085131657,-.0329015220239097,-.0289052514771828,-.0121493830289503,-.00197137671826113,-3.83525277872114e-005},
			{-.0113062148903861,-.0213785801497483,-.0228814465519833,-.0174855360916378,-.00988118279406322,-.00401765276570605,-.0010799937436432,-.000159171487958625,-8.09632053406128e-006,-2.61175882124433e-008,-.0253441560031943,-.0479224988818084,-.0512913434438016,-.039195801495227,-.0221497858175121,-.00900602186035772,-.00242092779827153,-.000356800844596912,-1.81488157315668e-005,-5.85455200083099e-008,-.0371529546538126,-.0702513994795805,-.0751899158472868,-.0574586044762535,-.0324702068581793,-.0132022672898219,-.00354892941386567,-.000523047822073078,-2.66050338314805e-005,-8.58240870115567e-008,-.0456626054707174,-.0863420411133481,-.0924116936244469,-.0706191367966456,-.0399073037160309,-.0162261636575445,-.0043617893967995,-.000642848639204797,-3.26987496607437e-005,-1.05481554875234e-007,-.0501153878844611,-.0947616728508056,-.101423206654712,-.077505551778014,-.0437988587058273,-.0178084556759703,-.00478712866332422,-.0007055359319215,-3.5887363537237e-005,-1.15767573548085e-007,-.0501153878844611,-.0947616728508056,-.101423206654712,-.077505551778014,-.0437988587058273,-.0178084556759703,-.00478712866332422,-.0007055359319215,-3.5887363537237e-005,-1.15767573548085e-007,-.0456626054707174,-.0863420411133481,-.0924116936244469,-.0706191367966455,-.0399073037160309,-.0162261636575445,-.0043617893967995,-.000642848639204797,-3.26987496607437e-005,-1.05481554875234e-007,-.0371529546538126,-.0702513994795805,-.0751899158472868,-.0574586044762535,-.0324702068581793,-.0132022672898219,-.00354892941386567,-.000523047822073078,-2.66050338314805e-005,-8.58240870115567e-008,-.0253441560031943,-.0479224988818084,-.0512913434438016,-.039195801495227,-.0221497858175121,-.00900602186035772,-.00242092779827153,-.000356800844596912,-1.81488157315668e-005,-5.85455200083099e-008,-.0113062148903861,-.0213785801497483,-.0228814465519833,-.0174855360916378,-.00988118279406322,-.00401765276570605,-.0010799937436432,-.000159171487958625,-8.09632053406128e-006,-2.61175882124433e-008},
			{.0190719517783585,.0360625774106304,.038597696000771,.0294955742829575,.0166681284221623,.00677720002251945,.0018217934825568,.000268499314073518,1.36573235432709e-005,4.40565996475352e-008,.0379740031057454,.0718038951915978,.0768515485379464,.0587283903841316,.0331877705976788,.0134940260805107,.00362735771192228,.000534606730606772,2.71929822744642e-005,8.7720725769792e-008,.0437205527152459,.0826698722318604,.0884813794780883,.0676156706608181,.0382100267353107,.0155360570480548,.00417628037843494,.000615507974819813,3.1308055979897e-005,1.00995373191444e-007,.0342772212097279,.0648137619926128,.0693700245984015,.0530111619489379,.0299569300362286,.0121803781308882,.00327423322613994,.00048256258667776,2.45457363601731e-005,7.91810837934576e-008,.0129226496311487,.0244350477648839,.0261527770090912,.0199854202945654,.0112938825616689,.00459205132170114,.00123439903524221,.000181928027204484,9.25384088109926e-006,2.98515856059849e-008,-.0129226496311487,-.0244350477648839,-.0261527770090912,-.0199854202945654,-.0112938825616689,-.00459205132170114,-.00123439903524221,-.000181928027204484,-9.25384088109926e-006,-2.98515856059849e-008,-.0342772212097279,-.0648137619926128,-.0693700245984015,-.0530111619489379,-.0299569300362286,-.0121803781308882,-.00327423322613994,-.00048256258667776,-2.45457363601731e-005,-7.91810837934576e-008,-.0437205527152459,-.0826698722318604,-.0884813794780884,-.0676156706608181,-.0382100267353107,-.0155360570480548,-.00417628037843494,-.000615507974819813,-3.1308055979897e-005,-1.00995373191444e-007,-.0379740031057454,-.0718038951915978,-.0768515485379464,-.0587283903841316,-.0331877705976788,-.0134940260805107,-.00362735771192228,-.000534606730606772,-2.71929822744642e-005,-8.7720725769792e-008,-.0190719517783585,-.0360625774106304,-.038597696000771,-.0294955742829575,-.0166681284221623,-.00677720002251945,-.0018217934825568,-.000268499314073518,-1.36573235432709e-005,-4.40565996475352e-008},
			{.0111564498581787,.00938834250199416,-.0113241539111341,-.0302961470354155,-.0312651432694344,-.018730813581855,-.00658104714974118,-.00116693908893853,-6.69192068520344e-005,-2.30174087176476e-007,.0250084407901998,.0210450287110927,-.0253843683416991,-.0679122309460246,-.0700843453060978,-.0419872314552365,-.0147521595197384,-.00261582559707332,-.000150006950558683,-5.15961179744944e-007,.0366608170547776,.0308507017275645,-.0372119034382505,-.0995551020305782,-.102739286436345,-.0615506669900124,-.0216257473168005,-.00383463745165151,-.000219900849377703,-7.56366963325805e-007,.0450577468469165,.0379168611112332,-.0457350561038325,-.122357572607023,-.126271074443454,-.075648460517151,-.0265789888566817,-.00471293706542479,-.000270267757204588,-9.29608063726004e-007,.049451546558839,.0416143183763193,-.050194903530658,-.134289253725007,-.138584380130904,-.0830253092787393,-.0291708351373767,-.00517251844641853,-.000296622878728574,-1.0202586605363e-006,.049451546558839,.0416143183763193,-.050194903530658,-.134289253725007,-.138584380130904,-.0830253092787393,-.0291708351373767,-.00517251844641853,-.000296622878728574,-1.0202586605363e-006,.0450577468469165,.0379168611112332,-.0457350561038325,-.122357572607023,-.126271074443454,-.075648460517151,-.0265789888566817,-.00471293706542479,-.000270267757204588,-9.29608063726004e-007,.0366608170547776,.0308507017275645,-.0372119034382505,-.0995551020305782,-.102739286436345,-.0615506669900124,-.0216257473168005,-.00383463745165151,-.000219900849377703,-7.56366963325805e-007,.0250084407901998,.0210450287110927,-.0253843683416991,-.0679122309460246,-.0700843453060978,-.0419872314552365,-.0147521595197384,-.00261582559707332,-.000150006950558683,-5.15961179744944e-007,.0111564498581787,.00938834250199416,-.0113241539111341,-.0302961470354155,-.0312651432694344,-.018730813581855,-.00658104714974118,-.00116693908893853,-6.69192068520345e-005,-2.30174087176476e-007},
			{-.0141462413882811,-.025273750604131,-.0243577599189909,-.0158870112137333,-.0071957937725239,-.002167521076626,-.000387881070632107,-3.23454410928798e-005,-6.92492077729253e-007,-4.31978541371621e-010,-.031710395749536,-.0566539628254887,-.0546006663821845,-.0356125276698685,-.0161301834173286,-.00485874298682688,-.000869479172304184,-7.2505954733546e-005,-1.55229910443945e-006,-9.68328627104381e-010,-.0464854657297909,-.0830511819588261,-.0800411772208778,-.0522057481598326,-.0236458445483905,-.00712261469039873,-.00127460232871753,-.000106289215076142,-2.27557383363344e-006,-1.41950947462063e-009,-.0571326695688298,-.102073533343423,-.0983741059332855,-.0641631467466999,-.0290617766661061,-.00875400483105958,-.00156654198328513,-.000130634091932606,-2.79677972190642e-006,-1.74463920044904e-009,-.0627039536356021,-.112027219286109,-.107967042743438,-.070420006785652,-.0318957316434227,-.00960765035478176,-.00171930309977544,-.00014337285664383,-3.06950729477359e-006,-1.91476744149015e-009,-.0627039536356021,-.112027219286109,-.107967042743438,-.070420006785652,-.0318957316434227,-.00960765035478176,-.00171930309977544,-.00014337285664383,-3.06950729477359e-006,-1.91476744149015e-009,-.0571326695688298,-.102073533343423,-.0983741059332855,-.0641631467466999,-.0290617766661061,-.00875400483105958,-.00156654198328513,-.000130634091932606,-2.79677972190642e-006,-1.74463920044904e-009,-.0464854657297909,-.0830511819588261,-.0800411772208778,-.0522057481598326,-.0236458445483905,-.00712261469039873,-.00127460232871753,-.000106289215076142,-2.27557383363344e-006,-1.41950947462063e-009,-.031710395749536,-.0566539628254887,-.0546006663821845,-.0356125276698685,-.0161301834173286,-.00485874298682688,-.000869479172304184,-7.2505954733546e-005,-1.55229910443945e-006,-9.68328627104381e-010,-.0141462413882811,-.025273750604131,-.0243577599189909,-.0158870112137333,-.0071957937725239,-.002167521076626,-.000387881070632107,-3.23454410928798e-005,-6.92492077729254e-007,-4.31978541371621e-010}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_hyp_fluxes_to_dofs[j][0];
			for (int i = 1; i < 100; i++)
			{
				std::cout << ", " << timestep_hyp_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_left_fluxes_to_dofs:" << std::endl;
		static const T timestep_left_fluxes_to_dofs[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{.00265100998409647,.00594253790984439,.008711390565351,.0107066798372692,.0117507401837501,.0117507401837501,.0107066798372692,.008711390565351,.00594253790984439,.00265100998409647,.00527840372812152,.0118321373536874,.0173451766357429,.0213179803576284,.0233968001502014,.0233968001502014,.0213179803576284,.0173451766357429,.0118321373536874,.00527840372812152,.00607717674128413,.0136226771632526,.019969996509099,.0245439987302064,.0269374032410133,.0269374032410133,.0245439987302064,.019969996509099,.0136226771632526,.00607717674128413,.00476454936076254,.0106802748271453,.0156566178922341,.0192426678437941,.0211191138346674,.0211191138346674,.0192426678437941,.0156566178922341,.0106802748271453,.00476454936076254,.00179625418474627,.00402650637025399,.00590260762948584,.00725456281862631,.00796199046987792,.00796199046987792,.00725456281862631,.00590260762948584,.00402650637025399,.00179625418474628,-.00179625418474627,-.00402650637025399,-.00590260762948584,-.00725456281862631,-.00796199046987792,-.00796199046987792,-.00725456281862631,-.00590260762948584,-.00402650637025399,-.00179625418474628,-.00476454936076254,-.0106802748271453,-.0156566178922341,-.0192426678437941,-.0211191138346674,-.0211191138346674,-.0192426678437941,-.0156566178922341,-.0106802748271453,-.00476454936076254,-.00607717674128413,-.0136226771632526,-.019969996509099,-.0245439987302064,-.0269374032410133,-.0269374032410133,-.0245439987302064,-.019969996509099,-.0136226771632526,-.00607717674128413,-.00527840372812152,-.0118321373536874,-.0173451766357429,-.0213179803576284,-.0233968001502014,-.0233968001502014,-.0213179803576284,-.0173451766357429,-.0118321373536874,-.00527840372812152,-.00265100998409647,-.00594253790984439,-.008711390565351,-.0107066798372692,-.0117507401837501,-.0117507401837501,-.0107066798372692,-.008711390565351,-.00594253790984439,-.00265100998409647},
			{-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.002221727122666,-.00498025195339731,-.00730073926966875,-.00897292772598497,-.00984792148436088,-.00984792148436088,-.00897292772598497,-.00730073926966875,-.00498025195339731,-.002221727122666,.00309757218884701,.00694355746343096,.010178822902703,.0125102183310402,.0137301504746905,.0137301504746905,.0125102183310402,.010178822902703,.00694355746343096,.00309757218884701,.00727044463764254,.0162975217520129,.0238911521277625,.0293632704051698,.0322266255011462,.0322266255011462,.0293632704051698,.0238911521277625,.0162975217520129,.00727044463764254,.00727044463764255,.016297521752013,.0238911521277625,.0293632704051698,.0322266255011463,.0322266255011463,.0293632704051698,.0238911521277625,.016297521752013,.00727044463764255,.003097572188847,.00694355746343095,.010178822902703,.0125102183310401,.0137301504746904,.0137301504746904,.0125102183310401,.010178822902703,.00694355746343095,.003097572188847,-.00222172712266599,-.00498025195339728,-.00730073926966871,-.00897292772598493,-.00984792148436084,-.00984792148436084,-.00897292772598493,-.00730073926966871,-.00498025195339728,-.00222172712266599,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468},
			{.00352805200405569,.00790852652678675,.0115934074660505,.0142488046002341,.0156382747342036,.0156382747342036,.0142488046002341,.0115934074660505,.00790852652678675,.00352805200405569,.00299001628838423,.00670245878037739,.00982538724528936,.0120758304570704,.0132534033295837,.0132534033295837,.0120758304570704,.00982538724528936,.00670245878037739,.00299001628838423,-.0032119933533482,-.00720004541021951,-.0105548182625439,-.0129723330655266,-.0142373282611347,-.0142373282611347,-.0129723330655266,-.0105548182625439,-.00720004541021951,-.0032119933533482,-.00749937306045005,-.0168106906345655,-.0246434257572659,-.0302878475826814,-.0332413626893215,-.0332413626893215,-.0302878475826814,-.0246434257572659,-.0168106906345655,-.00749937306045005,-.00396370307706834,-.0088850875478213,-.0130249851175851,-.0160082761176901,-.0175693208639632,-.0175693208639632,-.0160082761176901,-.0130249851175851,-.0088850875478213,-.00396370307706834,.00396370307706838,.00888508754782138,.0130249851175852,.0160082761176902,.0175693208639633,.0175693208639633,.0160082761176902,.0130249851175852,.00888508754782138,.00396370307706837,.00749937306045007,.0168106906345656,.024643425757266,.0302878475826815,.0332413626893216,.0332413626893216,.0302878475826815,.024643425757266,.0168106906345656,.00749937306045007,.00321199335334816,.00720004541021943,.0105548182625438,.0129723330655265,.0142373282611345,.0142373282611345,.0129723330655265,.0105548182625438,.00720004541021943,.00321199335334816,-.00299001628838425,-.00670245878037744,-.00982538724528943,-.0120758304570705,-.0132534033295838,-.0132534033295838,-.0120758304570705,-.00982538724528943,-.00670245878037744,-.00299001628838425,-.0035280520040557,-.00790852652678675,-.0115934074660505,-.0142488046002341,-.0156382747342036,-.0156382747342036,-.0142488046002341,-.0115934074660505,-.00790852652678675,-.0035280520040557},
			{-.00355524726055973,-.00796948776183025,-.0116827728408898,-.0143586385526627,-.0157588191287273,-.0157588191287273,-.0143586385526627,-.0116827728408898,-.00796948776183025,-.00355524726055973,-.000198327000917931,-.000444572343586338,-.000651715374523445,-.000800986686075343,-.000879094787718522,-.000879094787718522,-.000800986686075343,-.000651715374523445,-.000444572343586338,-.000198327000917931,.00656585578393543,.0147181063597184,.0215758275045217,.0265176352250662,.0291034986151868,.0291034986151868,.0265176352250662,.0215758275045218,.0147181063597184,.00656585578393543,.00333253137587672,.00747024498425,.0109509140141708,.0134591520602391,.0147716193401924,.0147716193401924,.0134591520602391,.0109509140141708,.00747024498425,.00333253137587672,-.00614481295848751,-.0137742913733916,-.020192253500946,-.0248171622895082,-.0272372043055647,-.0272372043055647,-.0248171622895082,-.020192253500946,-.0137742913733916,-.00614481295848751,-.00614481295848755,-.0137742913733917,-.0201922535009461,-.0248171622895084,-.0272372043055649,-.0272372043055649,-.0248171622895084,-.0201922535009461,-.0137742913733917,-.00614481295848755,.00333253137587662,.00747024498424977,.0109509140141705,.0134591520602387,.014771619340192,.014771619340192,.0134591520602387,.0109509140141704,.00747024498424977,.00333253137587662,.00656585578393523,.014718106359718,.0215758275045211,.0265176352250655,.0291034986151859,.0291034986151859,.0265176352250655,.0215758275045211,.014718106359718,.00656585578393523,-.000198327000917742,-.000444572343585915,-.000651715374522825,-.000800986686074581,-.000879094787717686,-.000879094787717686,-.000800986686074581,-.000651715374522825,-.000444572343585915,-.000198327000917742,-.0035552472605598,-.00796948776183042,-.0116827728408901,-.014358638552663,-.0157588191287277,-.0157588191287277,-.014358638552663,-.0116827728408901,-.00796948776183042,-.0035552472605598},
			{.00222253422114756,.00498206115568338,.00730339144757066,.00897618736857064,.00985149899052509,.00985149899052509,.00897618736857064,.00730339144757066,.00498206115568338,.00222253422114756,.00498206115568338,.011167852050509,.0163713757428243,.0201211364889587,.0220832462235868,.0220832462235868,.0201211364889587,.0163713757428243,.011167852050509,.00498206115568339,.00730339144757066,.0163713757428243,.0239994174797936,.02949633325581,.0323726639565543,.0323726639565543,.02949633325581,.0239994174797936,.0163713757428243,.00730339144757066,.00897618736857064,.0201211364889587,.02949633325581,.0362522830510504,.0397874192256888,.0397874192256888,.0362522830510504,.02949633325581,.0201211364889587,.00897618736857064,.00985149899052509,.0220832462235868,.0323726639565543,.0397874192256888,.0436672836966291,.0436672836966291,.0397874192256889,.0323726639565543,.0220832462235868,.0098514989905251,.00985149899052509,.0220832462235868,.0323726639565543,.0397874192256888,.0436672836966291,.0436672836966291,.0397874192256889,.0323726639565543,.0220832462235868,.0098514989905251,.00897618736857064,.0201211364889587,.02949633325581,.0362522830510504,.0397874192256889,.0397874192256889,.0362522830510504,.02949633325581,.0201211364889587,.00897618736857064,.00730339144757066,.0163713757428243,.0239994174797936,.02949633325581,.0323726639565543,.0323726639565543,.02949633325581,.0239994174797936,.0163713757428243,.00730339144757066,.00498206115568338,.011167852050509,.0163713757428243,.0201211364889587,.0220832462235868,.0220832462235868,.0201211364889587,.0163713757428243,.011167852050509,.00498206115568339,.00222253422114756,.00498206115568339,.00730339144757066,.00897618736857064,.0098514989905251,.0098514989905251,.00897618736857064,.00730339144757066,.00498206115568339,.00222253422114756},
			{-.00374909427409703,-.00840401770836613,-.0123197666866244,-.0151415318362812,-.0166180561384473,-.0166180561384473,-.0151415318362812,-.0123197666866244,-.00840401770836613,-.00374909427409703,-.00746479014118745,-.0167331691201299,-.024529784043959,-.0301481769489968,-.0330880720938548,-.0330880720938548,-.0301481769489968,-.024529784043959,-.0167331691201299,-.00746479014118745,-.00859442576984082,-.019265374803192,-.0282418399082409,-.0347104558846931,-.0380952410046641,-.0380952410046641,-.0347104558846931,-.0282418399082409,-.019265374803192,-.00859442576984082,-.00673809032566718,-.0151041895128434,-.022141801367642,-.027213241845299,-.0298669372150783,-.0298669372150783,-.027213241845299,-.022141801367642,-.0151041895128434,-.00673809032566718,-.00254028702994505,-.00569433991870819,-.00834754776432466,-.0102595011288345,-.0112599549077927,-.0112599549077927,-.0102595011288345,-.00834754776432466,-.00569433991870818,-.00254028702994505,.00254028702994505,.00569433991870819,.00834754776432466,.0102595011288345,.0112599549077927,.0112599549077927,.0102595011288345,.00834754776432466,.00569433991870818,.00254028702994505,.00673809032566718,.0151041895128434,.022141801367642,.027213241845299,.0298669372150783,.0298669372150783,.027213241845299,.022141801367642,.0151041895128434,.00673809032566718,.00859442576984082,.019265374803192,.0282418399082409,.0347104558846931,.0380952410046641,.0380952410046641,.0347104558846931,.0282418399082409,.019265374803192,.00859442576984082,.00746479014118745,.0167331691201299,.024529784043959,.0301481769489968,.0330880720938548,.0330880720938548,.0301481769489968,.024529784043959,.0167331691201299,.00746479014118745,.00374909427409703,.00840401770836614,.0123197666866244,.0151415318362812,.0166180561384473,.0166180561384473,.0151415318362812,.0123197666866244,.00840401770836614,.00374909427409703},
			{.00458578009298867,.0102795433484883,.0150691704959577,.0185206693125625,.0203267097203668,.0203267097203668,.0185206693125625,.0150691704959577,.0102795433484883,.00458578009298867,.00693481320331858,.0155451660331847,.0227882454891813,.0280077499310155,.0307389216426545,.0307389216426545,.0280077499310155,.0227882454891813,.0155451660331847,.00693481320331858,.00314199662823143,.00704313985533036,.0103248044887574,.0126896360821209,.0139270641219972,.0139270641219972,.0126896360821209,.0103248044887574,.00704313985533036,.00314199662823143,-.00438062859915128,-.00981967313422903,-.0143950293955455,-.0176921204289931,-.0194173650114245,-.0194173650114245,-.0176921204289931,-.0143950293955455,-.00981967313422903,-.00438062859915128,-.0102819614092862,-.0230481762908429,-.0337871913540478,-.0415259352355491,-.0455753308454797,-.0455753308454797,-.0415259352355491,-.0337871913540478,-.0230481762908429,-.0102819614092862,-.0102819614092862,-.0230481762908429,-.0337871913540478,-.0415259352355491,-.0455753308454797,-.0455753308454797,-.0415259352355492,-.0337871913540478,-.0230481762908429,-.0102819614092862,-.00438062859915127,-.009819673134229,-.0143950293955455,-.0176921204289931,-.0194173650114245,-.0194173650114245,-.0176921204289931,-.0143950293955455,-.009819673134229,-.00438062859915127,.00314199662823142,.00704313985533033,.0103248044887574,.0126896360821208,.0139270641219972,.0139270641219972,.0126896360821208,.0103248044887574,.00704313985533033,.00314199662823142,.00693481320331858,.0155451660331847,.0227882454891813,.0280077499310155,.0307389216426545,.0307389216426545,.0280077499310155,.0227882454891813,.0155451660331847,.00693481320331858,.00458578009298867,.0102795433484883,.0150691704959577,.0185206693125625,.0203267097203668,.0203267097203668,.0185206693125625,.0150691704959577,.0102795433484883,.00458578009298867},
			{-.00498941899246989,-.0111843454716204,-.0163955540712153,-.0201508527115458,-.0221158602193511,-.0221158602193511,-.0201508527115458,-.0163955540712153,-.0111843454716204,-.00498941899246989,-.00422852158639074,-.00947870810765226,-.0138951958966771,-.0170778032078584,-.0187431427347081,-.0187431427347081,-.0170778032078584,-.0138951958966771,-.00947870810765226,-.00422852158639074,.00454244456217192,.0101824018679708,.0149267671340066,.0183456493553325,.0201346227171464,.0201346227171464,.0183456493553325,.0149267671340066,.0101824018679708,.00454244456217192,.0106057150904842,.0237739066862442,.0348510669263035,.0428334848228836,.0470103859430136,.0470103859430136,.0428334848228836,.0348510669263035,.0237739066862442,.0106057150904842,.0056055226483345,.0125654113119353,.0184201106014339,.0226391211939302,.0248467718453936,.0248467718453936,.0226391211939302,.0184201106014339,.0125654113119353,.0056055226483345,-.00560552264833455,-.0125654113119354,-.0184201106014341,-.0226391211939304,-.0248467718453938,-.0248467718453938,-.0226391211939304,-.0184201106014341,-.0125654113119354,-.00560552264833455,-.0106057150904842,-.0237739066862443,-.0348510669263036,-.0428334848228837,-.0470103859430138,-.0470103859430138,-.0428334848228837,-.0348510669263036,-.0237739066862442,-.0106057150904842,-.00454244456217188,-.0101824018679707,-.0149267671340064,-.0183456493553323,-.0201346227171462,-.0201346227171462,-.0183456493553323,-.0149267671340064,-.0101824018679707,-.00454244456217188,.00422852158639077,.00947870810765234,.0138951958966772,.0170778032078585,.0187431427347082,.0187431427347082,.0170778032078585,.0138951958966772,.00947870810765234,.00422852158639077,.00498941899246989,.0111843454716204,.0163955540712153,.0201508527115458,.0221158602193511,.0221158602193511,.0201508527115458,.0163955540712153,.0111843454716204,.00498941899246989},
			{-.00272203738908372,-.00610175384992259,-.00894479121996913,-.01099353944528,-.012065572865233,-.012065572865233,-.01099353944528,-.00894479121996913,-.00610175384992259,-.00272203738908372,-.00610175384992259,-.0136777695245317,-.0200507584804236,-.0246432587236037,-.0270463425584096,-.0270463425584096,-.0246432587236037,-.0200507584804236,-.0136777695245317,-.00610175384992259,-.00894479121996913,-.0200507584804236,-.0293931634773647,-.0361254828831082,-.0396482541575827,-.0396482541575827,-.0361254828831082,-.0293931634773647,-.0200507584804236,-.00894479121996913,-.01099353944528,-.0246432587236037,-.0361254828831082,-.0443997977469403,-.0487294376468829,-.0487294376468829,-.0443997977469403,-.0361254828831082,-.0246432587236037,-.01099353944528,-.012065572865233,-.0270463425584096,-.0396482541575827,-.0487294376468829,-.053481281759782,-.053481281759782,-.0487294376468829,-.0396482541575827,-.0270463425584096,-.012065572865233,-.012065572865233,-.0270463425584096,-.0396482541575827,-.0487294376468829,-.053481281759782,-.053481281759782,-.0487294376468829,-.0396482541575827,-.0270463425584096,-.012065572865233,-.01099353944528,-.0246432587236037,-.0361254828831082,-.0443997977469403,-.0487294376468829,-.0487294376468829,-.0443997977469403,-.0361254828831082,-.0246432587236037,-.01099353944528,-.00894479121996913,-.0200507584804236,-.0293931634773647,-.0361254828831082,-.0396482541575827,-.0396482541575827,-.0361254828831082,-.0293931634773647,-.0200507584804236,-.00894479121996913,-.00610175384992259,-.0136777695245317,-.0200507584804236,-.0246432587236037,-.0270463425584096,-.0270463425584096,-.0246432587236037,-.0200507584804236,-.0136777695245317,-.00610175384992259,-.00272203738908372,-.00610175384992259,-.00894479121996913,-.01099353944528,-.012065572865233,-.012065572865233,-.01099353944528,-.00894479121996913,-.00610175384992259,-.00272203738908372},
			{.00459168398220963,.0102927775821281,.0150885710584479,.0185445134519897,.0203528790176253,.0203528790176253,.0185445134519897,.0150885710584479,.0102927775821281,.00459168398220963,.00914246343674623,.0204938630514994,.0300427271887783,.0369238250811581,.0405244465804058,.0405244465804058,.0369238250811581,.0300427271887783,.0204938630514994,.00914246343674623,.0105259788787713,.0235951689735483,.0345890485685456,.0425114528066451,.0466569510209666,.0466569510209666,.0425114528066451,.0345890485685456,.0235951689735483,.0105259788787713,.00825244156510291,.0184987786328967,.0271180576544867,.0333292783668801,.0365793781595865,.0365793781595865,.0333292783668801,.0271180576544867,.0184987786328967,.00825244156510291,.00311120351019257,.00697411360782243,.010223616307811,.0125652713841337,.0137905720183488,.0137905720183489,.0125652713841337,.010223616307811,.00697411360782243,.00311120351019257,-.00311120351019257,-.00697411360782243,-.010223616307811,-.0125652713841337,-.0137905720183489,-.0137905720183488,-.0125652713841337,-.010223616307811,-.00697411360782243,-.00311120351019257,-.00825244156510291,-.0184987786328967,-.0271180576544867,-.0333292783668801,-.0365793781595865,-.0365793781595865,-.0333292783668801,-.0271180576544867,-.0184987786328967,-.00825244156510291,-.0105259788787713,-.0235951689735483,-.0345890485685456,-.0425114528066451,-.0466569510209666,-.0466569510209666,-.0425114528066451,-.0345890485685456,-.0235951689735483,-.0105259788787713,-.00914246343674623,-.0204938630514994,-.0300427271887783,-.0369238250811581,-.0405244465804058,-.0405244465804058,-.0369238250811581,-.0300427271887783,-.0204938630514994,-.00914246343674623,-.00459168398220963,-.0102927775821281,-.0150885710584479,-.0185445134519897,-.0203528790176254,-.0203528790176254,-.0185445134519897,-.0150885710584479,-.0102927775821281,-.00459168398220963},
			{-.00561641065141988,-.0125898179990038,-.0184558892850007,-.0226830947601565,-.0248950334876143,-.0248950334876143,-.0226830947601565,-.0184558892850007,-.0125898179990038,-.00561641065141988,-.00849337690664137,-.0190388623781496,-.0279097867968623,-.0343023480945728,-.0376473366420087,-.0376473366420087,-.0343023480945728,-.0279097867968623,-.0190388623781496,-.00849337690664138,-.00384814425717993,-.00862604941815623,-.0126452513484336,-.0155415667147304,-.0170571003606095,-.0170571003606095,-.0155415667147304,-.0126452513484336,-.00862604941815623,-.00384814425717993,.00536515241143043,.0120265943124634,.0176302384294998,.0216683337640905,.0237813182187718,.0237813182187718,.0216683337640905,.0176302384294998,.0120265943124634,.00536515241143043,.0125927795065653,.0282281357131822,.041380689338455,.050858676220366,.0558181527269258,.0558181527269258,.050858676220366,.041380689338455,.0282281357131822,.0125927795065653,.0125927795065653,.0282281357131822,.0413806893384551,.050858676220366,.0558181527269258,.0558181527269258,.050858676220366,.0413806893384551,.0282281357131822,.0125927795065653,.00536515241143042,.0120265943124634,.0176302384294998,.0216683337640904,.0237813182187717,.0237813182187717,.0216683337640904,.0176302384294998,.0120265943124634,.00536515241143042,-.00384814425717991,-.00862604941815619,-.0126452513484336,-.0155415667147303,-.0170571003606094,-.0170571003606094,-.0155415667147303,-.0126452513484336,-.00862604941815619,-.00384814425717991,-.00849337690664137,-.0190388623781496,-.0279097867968623,-.0343023480945728,-.0376473366420086,-.0376473366420086,-.0343023480945728,-.0279097867968623,-.0190388623781496,-.00849337690664138,-.00561641065141988,-.0125898179990038,-.0184558892850007,-.0226830947601565,-.0248950334876143,-.0248950334876143,-.0226830947601565,-.0184558892850007,-.0125898179990038,-.00561641065141988},
			{.00314313803755599,.00704569845308084,.0103285552337492,.0126942459116857,.0139321234784299,.0139321234784299,.0126942459116857,.0103285552337492,.00704569845308084,.00314313803755599,.00704569845308084,.0157937278282394,.0231526216040999,.0284555841055385,.0312304263023817,.0312304263023817,.0284555841055385,.0231526216040999,.0157937278282394,.00704569845308084,.0103285552337492,.0231526216040999,.033940301680024,.0417141145196382,.0457818604054276,.0457818604054276,.0417141145196382,.033940301680024,.0231526216040999,.0103285552337492,.0126942459116857,.0284555841055385,.0417141145196382,.0512684703442582,.0562679078659487,.0562679078659487,.0512684703442582,.0417141145196382,.0284555841055385,.0126942459116857,.0139321234784299,.0312304263023817,.0457818604054276,.0562679078659487,.0617548648194743,.0617548648194743,.0562679078659487,.0457818604054276,.0312304263023817,.0139321234784299,.0139321234784299,.0312304263023817,.0457818604054276,.0562679078659487,.0617548648194743,.0617548648194743,.0562679078659487,.0457818604054276,.0312304263023817,.0139321234784299,.0126942459116857,.0284555841055385,.0417141145196382,.0512684703442582,.0562679078659487,.0562679078659487,.0512684703442582,.0417141145196382,.0284555841055385,.0126942459116857,.0103285552337492,.0231526216040999,.033940301680024,.0417141145196382,.0457818604054276,.0457818604054276,.0417141145196382,.033940301680024,.0231526216040999,.0103285552337492,.00704569845308084,.0157937278282394,.0231526216040999,.0284555841055385,.0312304263023817,.0312304263023817,.0284555841055385,.0231526216040999,.0157937278282394,.00704569845308084,.00314313803755599,.00704569845308084,.0103285552337492,.0126942459116857,.0139321234784299,.0139321234784299,.0126942459116857,.0103285552337492,.00704569845308084,.00314313803755599},
			{-.00530201996819294,-.0118850758196888,-.017422781130702,-.0214133596745384,-.0235014803675002,-.0235014803675002,-.0214133596745384,-.017422781130702,-.0118850758196888,-.00530201996819294,-.010556807456243,-.0236642747073748,-.0346903532714859,-.0426359607152567,-.0467936003004029,-.0467936003004029,-.0426359607152567,-.0346903532714859,-.0236642747073748,-.010556807456243,-.0121543534825683,-.0272453543265051,-.039939993018198,-.0490879974604127,-.0538748064820265,-.0538748064820265,-.0490879974604127,-.039939993018198,-.0272453543265051,-.0121543534825683,-.00952909872152508,-.0213605496542905,-.0313132357844681,-.0384853356875881,-.0422382276693348,-.0422382276693348,-.0384853356875881,-.0313132357844681,-.0213605496542905,-.00952909872152508,-.00359250836949255,-.00805301274050799,-.0118052152589717,-.0145091256372526,-.0159239809397558,-.0159239809397558,-.0145091256372526,-.0118052152589717,-.00805301274050799,-.00359250836949255,.00359250836949255,.00805301274050799,.0118052152589717,.0145091256372526,.0159239809397558,.0159239809397558,.0145091256372526,.0118052152589717,.00805301274050799,.00359250836949255,.00952909872152508,.0213605496542905,.0313132357844681,.0384853356875881,.0422382276693348,.0422382276693348,.0384853356875881,.0313132357844681,.0213605496542905,.00952909872152508,.0121543534825683,.0272453543265051,.039939993018198,.0490879974604127,.0538748064820265,.0538748064820265,.0490879974604127,.039939993018198,.0272453543265051,.0121543534825683,.010556807456243,.0236642747073748,.0346903532714859,.0426359607152567,.0467936003004029,.0467936003004029,.0426359607152567,.0346903532714859,.0236642747073748,.010556807456243,.00530201996819294,.0118850758196888,.017422781130702,.0214133596745384,.0235014803675002,.0235014803675002,.0214133596745384,.017422781130702,.0118850758196888,.00530201996819294},
			{-.00351413515806021,-.00787733034668568,-.0115476758084439,-.0141925983938026,-.015576587587625,-.015576587587625,-.0141925983938026,-.0115476758084439,-.00787733034668568,-.00351413515806021,-.00787733034668568,-.017657924524755,-.0258854178874996,-.0318143102064225,-.0349166780965639,-.0349166780965639,-.0318143102064225,-.0258854178874996,-.017657924524755,-.00787733034668568,-.0115476758084439,-.0258854178874996,-.0379464108746825,-.0466377978533816,-.0511856760122494,-.0511856760122494,-.0466377978533816,-.0379464108746825,-.0258854178874996,-.0115476758084439,-.0141925983938026,-.0318143102064225,-.0466377978533816,-.0573198924081667,-.0629094334832252,-.0629094334832252,-.0573198924081666,-.0466377978533816,-.0318143102064225,-.0141925983938026,-.015576587587625,-.0349166780965639,-.0511856760122494,-.0629094334832252,-.0690440378533661,-.0690440378533661,-.0629094334832252,-.0511856760122494,-.0349166780965639,-.015576587587625,-.015576587587625,-.0349166780965639,-.0511856760122494,-.0629094334832252,-.0690440378533661,-.0690440378533661,-.0629094334832252,-.0511856760122494,-.0349166780965639,-.015576587587625,-.0141925983938026,-.0318143102064225,-.0466377978533816,-.0573198924081666,-.0629094334832252,-.0629094334832252,-.0573198924081667,-.0466377978533816,-.0318143102064225,-.0141925983938026,-.0115476758084439,-.0258854178874996,-.0379464108746825,-.0466377978533816,-.0511856760122494,-.0511856760122494,-.0466377978533816,-.0379464108746825,-.0258854178874996,-.0115476758084439,-.00787733034668568,-.017657924524755,-.0258854178874996,-.0318143102064225,-.0349166780965639,-.0349166780965639,-.0318143102064225,-.0258854178874996,-.017657924524755,-.00787733034668568,-.00351413515806021,-.00787733034668568,-.0115476758084439,-.0141925983938026,-.015576587587625,-.015576587587625,-.0141925983938026,-.0115476758084439,-.00787733034668568,-.00351413515806021},
			{.00374909427409703,.00746479014118745,.00859442576984082,.00673809032566718,.00254028702994505,-.00254028702994505,-.00673809032566718,-.00859442576984082,-.00746479014118745,-.00374909427409703,.00840401770836613,.0167331691201299,.019265374803192,.0151041895128434,.00569433991870819,-.00569433991870819,-.0151041895128434,-.019265374803192,-.0167331691201299,-.00840401770836614,.0123197666866244,.024529784043959,.0282418399082409,.022141801367642,.00834754776432466,-.00834754776432466,-.022141801367642,-.0282418399082409,-.024529784043959,-.0123197666866244,.0151415318362812,.0301481769489968,.0347104558846931,.027213241845299,.0102595011288345,-.0102595011288345,-.027213241845299,-.0347104558846931,-.0301481769489968,-.0151415318362812,.0166180561384473,.0330880720938548,.0380952410046641,.0298669372150783,.0112599549077927,-.0112599549077927,-.0298669372150783,-.0380952410046641,-.0330880720938548,-.0166180561384473,.0166180561384473,.0330880720938548,.0380952410046641,.0298669372150783,.0112599549077927,-.0112599549077927,-.0298669372150783,-.0380952410046641,-.0330880720938548,-.0166180561384473,.0151415318362812,.0301481769489968,.0347104558846931,.027213241845299,.0102595011288345,-.0102595011288345,-.027213241845299,-.0347104558846931,-.0301481769489968,-.0151415318362812,.0123197666866244,.024529784043959,.0282418399082409,.022141801367642,.00834754776432466,-.00834754776432466,-.022141801367642,-.0282418399082409,-.024529784043959,-.0123197666866244,.00840401770836613,.0167331691201299,.019265374803192,.0151041895128434,.00569433991870818,-.00569433991870818,-.0151041895128434,-.019265374803192,-.0167331691201299,-.00840401770836614,.00374909427409703,.00746479014118745,.00859442576984082,.00673809032566718,.00254028702994505,-.00254028702994505,-.00673809032566718,-.00859442576984082,-.00746479014118745,-.00374909427409703},
			{-.00632418062917951,-.0125920229688431,-.0144975551423388,-.0113661853236565,-.00428509737360599,.00428509737360599,.0113661853236565,.0144975551423388,.0125920229688431,.00632418062917952,-.0125920229688431,-.0250718712423055,-.0288659603588969,-.0226311161960246,-.00853202141368567,.00853202141368567,.0226311161960246,.0288659603588969,.0250718712423055,.0125920229688431,-.0144975551423388,-.0288659603588969,-.0332342033583603,-.0260558494688553,-.00982315956908459,.00982315956908459,.0260558494688553,.0332342033583603,.0288659603588969,.0144975551423388,-.0113661853236565,-.0226311161960246,-.0260558494688553,-.0204279694693769,-.0077014262770414,.0077014262770414,.0204279694693769,.0260558494688553,.0226311161960246,.0113661853236565,-.00428509737360599,-.00853202141368567,-.00982315956908459,-.0077014262770414,-.00290346854050359,.00290346854050359,.0077014262770414,.00982315956908459,.00853202141368567,.00428509737360599,.00428509737360599,.00853202141368567,.00982315956908459,.0077014262770414,.00290346854050359,-.00290346854050359,-.0077014262770414,-.00982315956908459,-.00853202141368567,-.00428509737360599,.0113661853236565,.0226311161960246,.0260558494688553,.0204279694693769,.0077014262770414,-.0077014262770414,-.0204279694693769,-.0260558494688553,-.0226311161960246,-.0113661853236565,.0144975551423388,.0288659603588969,.0332342033583603,.0260558494688553,.00982315956908459,-.00982315956908459,-.0260558494688553,-.0332342033583603,-.0288659603588969,-.0144975551423388,.0125920229688431,.0250718712423055,.0288659603588969,.0226311161960246,.00853202141368567,-.00853202141368567,-.0226311161960246,-.0288659603588969,-.0250718712423055,-.0125920229688431,.00632418062917952,.0125920229688431,.0144975551423388,.0113661853236565,.00428509737360599,-.00428509737360599,-.0113661853236565,-.0144975551423388,-.0125920229688431,-.00632418062917951},
			{.00773554878208247,.0154021862517905,.0177329762700093,.0139027782716481,.00524140307070943,-.00524140307070943,-.0139027782716481,-.0177329762700093,-.0154021862517905,-.00773554878208247,.0116980284141665,.023291846188219,.0268165667515141,.0210243771759672,.00792627437025251,-.00792627437025251,-.0210243771759672,-.0268165667515141,-.023291846188219,-.0116980284141665,.00530009457452694,.0105529738210754,.0121499397090725,.00952563829202252,.00359120377371563,-.00359120377371563,-.00952563829202252,-.0121499397090725,-.0105529738210754,-.00530009457452694,-.0073894878380083,-.0147131472107018,-.0169396659720433,-.0132808174115363,-.00500692133631402,.00500692133631402,.0132808174115363,.0169396659720433,.0147131472107018,.0073894878380083,-.0173441840742927,-.0345338593321725,-.0397598170829848,-.0311719765824586,-.0117519599742572,.0117519599742572,.0311719765824586,.0397598170829848,.0345338593321725,.0173441840742927,-.0173441840742927,-.0345338593321725,-.0397598170829848,-.0311719765824586,-.0117519599742572,.0117519599742572,.0311719765824586,.0397598170829848,.0345338593321725,.0173441840742927,-.00738948783800828,-.0147131472107018,-.0169396659720432,-.0132808174115363,-.00500692133631401,.00500692133631401,.0132808174115363,.0169396659720432,.0147131472107018,.00738948783800828,.00530009457452691,.0105529738210754,.0121499397090724,.00952563829202247,.00359120377371561,-.00359120377371561,-.00952563829202247,-.0121499397090724,-.0105529738210754,-.00530009457452691,.0116980284141665,.023291846188219,.0268165667515141,.0210243771759672,.00792627437025251,-.00792627437025251,-.0210243771759672,-.0268165667515141,-.023291846188219,-.0116980284141665,.00773554878208247,.0154021862517905,.0177329762700093,.0139027782716481,.00524140307070943,-.00524140307070943,-.0139027782716481,-.0177329762700093,-.0154021862517905,-.00773554878208247},
			{-.00841642931466872,-.0167578817652623,-.0192938271762863,-.0151264963737114,-.005702749694565,.005702749694565,.0151264963737114,.0192938271762863,.0167578817652623,.00841642931466872,-.00713290527236138,-.0142022678178642,-.0163514759578514,-.0128196722983645,-.00483306778237002,.00483306778237002,.0128196722983645,.0163514759578514,.0142022678178642,.00713290527236138,.00766244799865884,.015256635895486,.0175653999940025,.0137713972911834,.00519187191506662,-.00519187191506662,-.0137713972911834,-.0175653999940025,-.015256635895486,-.00766244799865884,.0178903097786119,.0356212456381444,.0410117559470465,.0321535054680605,.0121220002938583,-.0121220002938583,-.0321535054680605,-.0410117559470465,-.0356212456381444,-.0178903097786118,.00945570721013499,.0188271792597616,.0216762683937647,.0169943470654115,.00640693688361546,-.00640693688361546,-.0169943470654115,-.0216762683937647,-.0188271792597616,-.00945570721013499,-.00945570721013506,-.0188271792597618,-.0216762683937649,-.0169943470654116,-.00640693688361551,.00640693688361551,.0169943470654116,.0216762683937649,.0188271792597618,.00945570721013506,-.0178903097786119,-.0356212456381445,-.0410117559470466,-.0321535054680606,-.0121220002938583,.0121220002938583,.0321535054680606,.0410117559470466,.0356212456381445,.0178903097786119,-.00766244799865876,-.0152566358954859,-.0175653999940023,-.0137713972911832,-.00519187191506656,.00519187191506656,.0137713972911832,.0175653999940023,.0152566358954859,.00766244799865876,.00713290527236143,.0142022678178643,.0163514759578515,.0128196722983646,.00483306778237005,-.00483306778237005,-.0128196722983646,-.0163514759578515,-.0142022678178643,-.00713290527236143,.00841642931466873,.0167578817652623,.0192938271762863,.0151264963737114,.00570274969456501,-.00570274969456501,-.0151264963737114,-.0192938271762863,-.0167578817652623,-.00841642931466873},
			{-.00459168398220963,-.00914246343674623,-.0105259788787713,-.00825244156510291,-.00311120351019257,.00311120351019257,.00825244156510291,.0105259788787713,.00914246343674623,.00459168398220963,-.0102927775821281,-.0204938630514994,-.0235951689735483,-.0184987786328967,-.00697411360782243,.00697411360782243,.0184987786328967,.0235951689735483,.0204938630514994,.0102927775821281,-.0150885710584479,-.0300427271887783,-.0345890485685456,-.0271180576544867,-.010223616307811,.010223616307811,.0271180576544867,.0345890485685456,.0300427271887783,.0150885710584479,-.0185445134519897,-.0369238250811581,-.0425114528066451,-.0333292783668801,-.0125652713841337,.0125652713841337,.0333292783668801,.0425114528066451,.0369238250811581,.0185445134519897,-.0203528790176253,-.0405244465804058,-.0466569510209666,-.0365793781595865,-.0137905720183488,.0137905720183489,.0365793781595865,.0466569510209666,.0405244465804058,.0203528790176254,-.0203528790176253,-.0405244465804058,-.0466569510209666,-.0365793781595865,-.0137905720183489,.0137905720183488,.0365793781595865,.0466569510209666,.0405244465804058,.0203528790176254,-.0185445134519897,-.0369238250811581,-.0425114528066451,-.0333292783668801,-.0125652713841337,.0125652713841337,.0333292783668801,.0425114528066451,.0369238250811581,.0185445134519897,-.0150885710584479,-.0300427271887783,-.0345890485685456,-.0271180576544867,-.010223616307811,.010223616307811,.0271180576544867,.0345890485685456,.0300427271887783,.0150885710584479,-.0102927775821281,-.0204938630514994,-.0235951689735483,-.0184987786328967,-.00697411360782243,.00697411360782243,.0184987786328967,.0235951689735483,.0204938630514994,.0102927775821281,-.00459168398220963,-.00914246343674623,-.0105259788787713,-.00825244156510291,-.00311120351019257,.00311120351019257,.00825244156510291,.0105259788787713,.00914246343674623,.00459168398220963},
			{.00774550779202725,.0154220155529008,.0177558063098679,.0139206771836668,.00524815103220206,-.00524815103220206,-.0139206771836668,-.0177558063098679,-.0154220155529008,-.00774550779202725,.0154220155529008,.030706645722922,.0353534369104813,.0277173434974017,.0104495494699397,-.0104495494699397,-.0277173434974017,-.0353534369104813,-.030706645722922,-.0154220155529008,.0177558063098679,.0353534369104813,.0407034201215399,.0319117680095566,.0120308643041625,-.0120308643041625,-.0319117680095566,-.0407034201215399,-.0353534369104813,-.0177558063098679,.0139206771836668,.0277173434974017,.0319117680095566,.025019050842778,.0094322823360418,-.0094322823360418,-.025019050842778,-.0319117680095566,-.0277173434974017,-.0139206771836669,.00524815103220206,.0104495494699397,.0120308643041625,.0094322823360418,.00355600820454337,-.00355600820454337,-.0094322823360418,-.0120308643041625,-.0104495494699397,-.00524815103220206,-.00524815103220206,-.0104495494699397,-.0120308643041625,-.0094322823360418,-.00355600820454337,.00355600820454337,.0094322823360418,.0120308643041625,.0104495494699397,.00524815103220206,-.0139206771836668,-.0277173434974017,-.0319117680095566,-.025019050842778,-.0094322823360418,.0094322823360418,.025019050842778,.0319117680095566,.0277173434974017,.0139206771836669,-.0177558063098679,-.0353534369104813,-.0407034201215399,-.0319117680095566,-.0120308643041625,.0120308643041625,.0319117680095566,.0407034201215399,.0353534369104813,.0177558063098679,-.0154220155529008,-.030706645722922,-.0353534369104813,-.0277173434974017,-.0104495494699397,.0104495494699397,.0277173434974017,.0353534369104813,.030706645722922,.0154220155529008,-.00774550779202725,-.0154220155529008,-.0177558063098679,-.0139206771836669,-.00524815103220206,.00524815103220206,.0139206771836669,.0177558063098679,.0154220155529008,.00774550779202725},
			{-.00947407369825788,-.0188637486201043,-.0217183717412093,-.0170273563863006,-.00641938152974948,.00641938152974948,.0170273563863006,.0217183717412093,.0188637486201043,.00947407369825788,-.0143271003056479,-.0285265691642717,-.0328434525972573,-.0257494981204761,-.00970766388421236,.00970766388421236,.0257494981204761,.0328434525972573,.0285265691642717,.0143271003056479,-.00649126364804426,-.0129247005652958,-.0148805763464081,-.0116664766448895,-.00439830840398169,.00439830840398169,.0116664766448895,.0148805763464081,.0129247005652958,.00649126364804426,.00905023733181398,.0180198515883421,.020746769022353,.0162656130126723,.00613220122811362,-.00613220122811362,-.0162656130126722,-.020746769022353,-.0180198515883421,-.00905023733181398,.0212422004934682,.0422951671064499,.0486956320598683,.0381777184505166,.0143931527072752,-.0143931527072752,-.0381777184505166,-.0486956320598683,-.0422951671064499,-.0212422004934682,.0212422004934682,.0422951671064499,.0486956320598684,.0381777184505167,.0143931527072752,-.0143931527072752,-.0381777184505167,-.0486956320598684,-.0422951671064499,-.0212422004934682,.00905023733181396,.018019851588342,.020746769022353,.0162656130126722,.0061322012281136,-.0061322012281136,-.0162656130126722,-.020746769022353,-.018019851588342,-.00905023733181396,-.00649126364804423,-.0129247005652958,-.0148805763464081,-.0116664766448895,-.00439830840398167,.00439830840398167,.0116664766448895,.0148805763464081,.0129247005652958,.00649126364804423,-.0143271003056479,-.0285265691642717,-.0328434525972573,-.0257494981204761,-.00970766388421236,.00970766388421236,.0257494981204761,.0328434525972573,.0285265691642717,.0143271003056479,-.00947407369825788,-.0188637486201043,-.0217183717412093,-.0170273563863006,-.00641938152974948,.00641938152974948,.0170273563863006,.0217183717412093,.0188637486201043,.00947407369825788},
			{.00530201996819294,.010556807456243,.0121543534825683,.00952909872152508,.00359250836949255,-.00359250836949255,-.00952909872152508,-.0121543534825683,-.010556807456243,-.00530201996819294,.0118850758196888,.0236642747073748,.0272453543265051,.0213605496542905,.00805301274050799,-.00805301274050799,-.0213605496542905,-.0272453543265051,-.0236642747073748,-.0118850758196888,.017422781130702,.0346903532714859,.039939993018198,.0313132357844681,.0118052152589717,-.0118052152589717,-.0313132357844681,-.039939993018198,-.0346903532714859,-.017422781130702,.0214133596745384,.0426359607152567,.0490879974604127,.0384853356875881,.0145091256372526,-.0145091256372526,-.0384853356875881,-.0490879974604127,-.0426359607152567,-.0214133596745384,.0235014803675002,.0467936003004029,.0538748064820265,.0422382276693348,.0159239809397558,-.0159239809397558,-.0422382276693348,-.0538748064820265,-.0467936003004029,-.0235014803675002,.0235014803675002,.0467936003004029,.0538748064820265,.0422382276693348,.0159239809397558,-.0159239809397558,-.0422382276693348,-.0538748064820265,-.0467936003004029,-.0235014803675002,.0214133596745384,.0426359607152567,.0490879974604127,.0384853356875881,.0145091256372526,-.0145091256372526,-.0384853356875881,-.0490879974604127,-.0426359607152567,-.0214133596745384,.017422781130702,.0346903532714859,.039939993018198,.0313132357844681,.0118052152589717,-.0118052152589717,-.0313132357844681,-.039939993018198,-.0346903532714859,-.017422781130702,.0118850758196888,.0236642747073748,.0272453543265051,.0213605496542905,.00805301274050799,-.00805301274050799,-.0213605496542905,-.0272453543265051,-.0236642747073748,-.0118850758196888,.00530201996819294,.010556807456243,.0121543534825683,.00952909872152508,.00359250836949255,-.00359250836949255,-.00952909872152508,-.0121543534825683,-.010556807456243,-.00530201996819294},
			{-.00894374201432336,-.0178078096555534,-.0205026390981384,-.0160742134329204,-.00606004282024417,.00606004282024417,.0160742134329204,.0205026390981384,.0178078096555534,.00894374201432336,-.0178078096555534,-.0354569803355862,-.0408226326197064,-.0320052314476158,-.0120661003945087,.0120661003945087,.0320052314476158,.0408226326197064,.0354569803355862,.0178078096555534,-.0205026390981384,-.0408226326197064,-.0470002611116591,-.0368485356882857,-.0138920454842895,.0138920454842895,.0368485356882857,.0470002611116591,.0408226326197064,.0205026390981384,-.0160742134329204,-.0320052314476158,-.0368485356882857,-.0288895114677148,-.0108914614877351,.0108914614877351,.0288895114677148,.0368485356882857,.0320052314476158,.0160742134329204,-.00606004282024417,-.0120661003945087,-.0138920454842895,-.0108914614877351,-.00410612458682053,.00410612458682053,.0108914614877351,.0138920454842895,.0120661003945087,.00606004282024417,.00606004282024417,.0120661003945087,.0138920454842895,.0108914614877351,.00410612458682053,-.00410612458682053,-.0108914614877351,-.0138920454842895,-.0120661003945087,-.00606004282024417,.0160742134329204,.0320052314476158,.0368485356882857,.0288895114677148,.0108914614877351,-.0108914614877351,-.0288895114677148,-.0368485356882857,-.0320052314476158,-.0160742134329204,.0205026390981384,.0408226326197064,.0470002611116591,.0368485356882857,.0138920454842895,-.0138920454842895,-.0368485356882857,-.0470002611116591,-.0408226326197064,-.0205026390981384,.0178078096555534,.0354569803355862,.0408226326197064,.0320052314476158,.0120661003945087,-.0120661003945087,-.0320052314476158,-.0408226326197064,-.0354569803355862,-.0178078096555534,.00894374201432336,.0178078096555534,.0205026390981384,.0160742134329204,.00606004282024417,-.00606004282024417,-.0160742134329204,-.0205026390981384,-.0178078096555534,-.00894374201432336},
			{-.00592783853288971,-.0118028695476119,-.0135889803034609,-.0106538562517747,-.00401654646156772,.00401654646156772,.0106538562517747,.0135889803034609,.0118028695476119,.00592783853288971,-.01328791872398,-.026457463439368,-.0304612321695831,-.0238818205295374,-.00900354195484198,.00900354195484198,.0238818205295374,.0304612321695831,.026457463439368,.01328791872398,-.0194792614807672,-.0387849940354634,-.044654269700405,-.0350092619013457,-.0131986319027464,.0131986319027464,.0350092619013457,.044654269700405,.0387849940354634,.0194792614807672,-.0239408639271153,-.0476684532179932,-.0548820495950343,-.0430279133629583,-.016221695607902,.016221695607902,.0430279133629583,.0548820495950343,.0476684532179932,.0239408639271153,-.0262754538342299,-.0523168355867033,-.060233864778329,-.047223774154247,-.0178035519251087,.0178035519251087,.047223774154247,.060233864778329,.0523168355867033,.0262754538342299,-.0262754538342299,-.0523168355867033,-.060233864778329,-.047223774154247,-.0178035519251087,.0178035519251087,.047223774154247,.060233864778329,.0523168355867033,.0262754538342299,-.0239408639271153,-.0476684532179932,-.0548820495950343,-.0430279133629583,-.016221695607902,.016221695607902,.0430279133629583,.0548820495950343,.0476684532179932,.0239408639271153,-.0194792614807672,-.0387849940354634,-.044654269700405,-.0350092619013457,-.0131986319027464,.0131986319027464,.0350092619013457,.044654269700405,.0387849940354634,.0194792614807672,-.01328791872398,-.026457463439368,-.0304612321695831,-.0238818205295374,-.00900354195484198,.00900354195484198,.0238818205295374,.0304612321695831,.026457463439368,.01328791872398,-.00592783853288971,-.0118028695476119,-.0135889803034609,-.0106538562517747,-.00401654646156772,.00401654646156772,.0106538562517747,.0135889803034609,.0118028695476119,.00592783853288971},
			{-.00561641065141988,-.00849337690664137,-.00384814425717993,.00536515241143043,.0125927795065653,.0125927795065653,.00536515241143042,-.00384814425717991,-.00849337690664137,-.00561641065141988,-.0125898179990038,-.0190388623781496,-.00862604941815623,.0120265943124634,.0282281357131822,.0282281357131822,.0120265943124634,-.00862604941815619,-.0190388623781496,-.0125898179990038,-.0184558892850007,-.0279097867968623,-.0126452513484336,.0176302384294998,.041380689338455,.0413806893384551,.0176302384294998,-.0126452513484336,-.0279097867968623,-.0184558892850007,-.0226830947601565,-.0343023480945728,-.0155415667147304,.0216683337640905,.050858676220366,.050858676220366,.0216683337640904,-.0155415667147303,-.0343023480945728,-.0226830947601565,-.0248950334876143,-.0376473366420087,-.0170571003606095,.0237813182187718,.0558181527269258,.0558181527269258,.0237813182187717,-.0170571003606094,-.0376473366420086,-.0248950334876143,-.0248950334876143,-.0376473366420087,-.0170571003606095,.0237813182187718,.0558181527269258,.0558181527269258,.0237813182187717,-.0170571003606094,-.0376473366420086,-.0248950334876143,-.0226830947601565,-.0343023480945728,-.0155415667147304,.0216683337640905,.050858676220366,.050858676220366,.0216683337640904,-.0155415667147303,-.0343023480945728,-.0226830947601565,-.0184558892850007,-.0279097867968623,-.0126452513484336,.0176302384294998,.041380689338455,.0413806893384551,.0176302384294998,-.0126452513484336,-.0279097867968623,-.0184558892850007,-.0125898179990038,-.0190388623781496,-.00862604941815623,.0120265943124634,.0282281357131822,.0282281357131822,.0120265943124634,-.00862604941815619,-.0190388623781496,-.0125898179990038,-.00561641065141988,-.00849337690664138,-.00384814425717993,.00536515241143043,.0125927795065653,.0125927795065653,.00536515241143042,-.00384814425717991,-.00849337690664138,-.00561641065141988},
			{.00947407369825788,.0143271003056479,.00649126364804426,-.00905023733181398,-.0212422004934682,-.0212422004934682,-.00905023733181396,.00649126364804423,.0143271003056479,.00947407369825788,.0188637486201043,.0285265691642717,.0129247005652958,-.0180198515883421,-.0422951671064499,-.0422951671064499,-.018019851588342,.0129247005652958,.0285265691642717,.0188637486201043,.0217183717412093,.0328434525972573,.0148805763464081,-.020746769022353,-.0486956320598683,-.0486956320598684,-.020746769022353,.0148805763464081,.0328434525972573,.0217183717412093,.0170273563863006,.0257494981204761,.0116664766448895,-.0162656130126723,-.0381777184505166,-.0381777184505167,-.0162656130126722,.0116664766448895,.0257494981204761,.0170273563863006,.00641938152974948,.00970766388421236,.00439830840398169,-.00613220122811362,-.0143931527072752,-.0143931527072752,-.0061322012281136,.00439830840398167,.00970766388421236,.00641938152974948,-.00641938152974948,-.00970766388421236,-.00439830840398169,.00613220122811362,.0143931527072752,.0143931527072752,.0061322012281136,-.00439830840398167,-.00970766388421236,-.00641938152974948,-.0170273563863006,-.0257494981204761,-.0116664766448895,.0162656130126722,.0381777184505166,.0381777184505167,.0162656130126722,-.0116664766448895,-.0257494981204761,-.0170273563863006,-.0217183717412093,-.0328434525972573,-.0148805763464081,.020746769022353,.0486956320598683,.0486956320598684,.020746769022353,-.0148805763464081,-.0328434525972573,-.0217183717412093,-.0188637486201043,-.0285265691642717,-.0129247005652958,.0180198515883421,.0422951671064499,.0422951671064499,.018019851588342,-.0129247005652958,-.0285265691642717,-.0188637486201043,-.00947407369825788,-.0143271003056479,-.00649126364804426,.00905023733181398,.0212422004934682,.0212422004934682,.00905023733181396,-.00649126364804423,-.0143271003056479,-.00947407369825788},
			{-.0115884038691255,-.0175244810103018,-.00793991973994695,.0110699798896205,.0259828249417684,.0259828249417684,.0110699798896205,-.00793991973994691,-.0175244810103018,-.0115884038691255,-.0175244810103018,-.0265012712836701,-.0120070869359958,.0167404980488237,.0392923415017616,.0392923415017616,.0167404980488237,-.0120070869359958,-.02650127128367,-.0175244810103018,-.00793991973994695,-.0120070869359958,-.00544012153776933,.00758471596598284,.0178024123929934,.0178024123929934,.00758471596598282,-.00544012153776931,-.0120070869359958,-.00793991973994695,.0110699798896205,.0167404980488237,.00758471596598284,-.0105747483553876,-.0248204457515694,-.0248204457515694,-.0105747483553876,.00758471596598281,.0167404980488237,.0110699798896205,.0259828249417684,.0392923415017616,.0178024123929934,-.0248204457515694,-.0582571335603207,-.0582571335603207,-.0248204457515693,.0178024123929933,.0392923415017616,.0259828249417684,.0259828249417684,.0392923415017616,.0178024123929934,-.0248204457515694,-.0582571335603207,-.0582571335603207,-.0248204457515693,.0178024123929933,.0392923415017616,.0259828249417684,.0110699798896205,.0167404980488237,.00758471596598282,-.0105747483553876,-.0248204457515693,-.0248204457515693,-.0105747483553876,.00758471596598279,.0167404980488237,.0110699798896205,-.00793991973994691,-.0120070869359958,-.00544012153776931,.00758471596598281,.0178024123929933,.0178024123929933,.00758471596598279,-.00544012153776928,-.0120070869359958,-.00793991973994691,-.0175244810103018,-.02650127128367,-.0120070869359958,.0167404980488237,.0392923415017616,.0392923415017616,.0167404980488237,-.0120070869359958,-.0265012712836701,-.0175244810103018,-.0115884038691255,-.0175244810103018,-.00793991973994695,.0110699798896205,.0259828249417684,.0259828249417684,.0110699798896205,-.00793991973994691,-.0175244810103018,-.0115884038691255},
			{.00648527240266935,.009807306886327,.004443454245332,-.00619514437769402,-.0145408892752851,-.0145408892752851,-.00619514437769401,.00444345424533198,.009807306886327,.00648527240266935,.0145374696209095,.0219841846372146,.00996050390679461,-.0138871149268619,-.0325950435040259,-.0325950435040259,-.0138871149268619,.00996050390679457,.0219841846372146,.0145374696209095,.0213110252927245,.032227445838975,.0146014785393375,-.020357645805406,-.0477823042555251,-.0477823042555251,-.020357645805406,.0146014785393374,.032227445838975,.0213110252927245,.0261921817305127,.0396089398107403,.0179458554519699,-.0250204366620803,-.0587265408103395,-.0587265408103396,-.0250204366620803,.0179458554519699,.0396089398107403,.0261921817305127,.0287463085698583,.0434713998871676,.0196958429687218,-.0274603009493809,-.0644532510022925,-.0644532510022925,-.0274603009493808,.0196958429687217,.0434713998871676,.0287463085698583,.0287463085698583,.0434713998871676,.0196958429687218,-.0274603009493809,-.0644532510022925,-.0644532510022925,-.0274603009493808,.0196958429687217,.0434713998871676,.0287463085698583,.0261921817305127,.0396089398107403,.0179458554519699,-.0250204366620803,-.0587265408103395,-.0587265408103396,-.0250204366620803,.0179458554519699,.0396089398107403,.0261921817305127,.0213110252927245,.032227445838975,.0146014785393375,-.020357645805406,-.0477823042555251,-.0477823042555251,-.020357645805406,.0146014785393374,.032227445838975,.0213110252927245,.0145374696209095,.0219841846372146,.00996050390679461,-.0138871149268619,-.0325950435040259,-.0325950435040259,-.0138871149268619,.00996050390679457,.0219841846372146,.0145374696209095,.00648527240266935,.009807306886327,.004443454245332,-.00619514437769402,-.0145408892752851,-.0145408892752851,-.00619514437769401,.00444345424533198,.009807306886327,.00648527240266935},
			{-.0109397180005024,-.01654351043707,-.00749546562948678,.0104503139199635,.0245283803472424,.0245283803472424,.0104503139199635,-.00749546562948674,-.01654351043707,-.0109397180005024,-.0217819806884396,-.0329396447735407,-.0149241587017916,.0208075323314845,.0488382522307979,.0488382522307979,.0208075323314844,-.0149241587017915,-.0329396447735407,-.0217819806884396,-.0250782155433939,-.037924352397948,-.0171826095193439,.0239563053607907,.0562288725587115,.0562288725587116,.0239563053607906,-.0171826095193438,-.037924352397948,-.0250782155433939,-.0196614975872982,-.029732959344012,-.0134712868634331,.0187819121036241,.0440838320508345,.0440838320508346,.0187819121036241,-.0134712868634331,-.029732959344012,-.0196614975872982,-.00741246330878832,-.011209444713996,-.00507872908225817,.00708085605986294,.0166197811807936,.0166197811807936,.00708085605986292,-.00507872908225814,-.011209444713996,-.00741246330878832,.00741246330878832,.011209444713996,.00507872908225817,-.00708085605986294,-.0166197811807936,-.0166197811807936,-.00708085605986292,.00507872908225814,.011209444713996,.00741246330878832,.0196614975872982,.029732959344012,.0134712868634331,-.0187819121036241,-.0440838320508345,-.0440838320508346,-.0187819121036241,.0134712868634331,.029732959344012,.0196614975872982,.0250782155433939,.037924352397948,.0171826095193439,-.0239563053607907,-.0562288725587116,-.0562288725587116,-.0239563053607906,.0171826095193438,.037924352397948,.0250782155433939,.0217819806884396,.0329396447735407,.0149241587017916,-.0208075323314845,-.0488382522307979,-.0488382522307979,-.0208075323314844,.0149241587017915,.0329396447735407,.0217819806884396,.0109397180005024,.01654351043707,.00749546562948678,-.0104503139199635,-.0245283803472424,-.0245283803472424,-.0104503139199634,.00749546562948674,.01654351043707,.0109397180005024},
			{-.00725075497095922,-.0109649024347056,-.00496793287269008,.00692638197801623,.0162572084329942,.0162572084329943,.00692638197801622,-.00496793287269006,-.0109649024347056,-.00725075497095922,-.0162533851431734,-.0245790656341836,-.0111361819105275,.0155262664906382,.0364423664946087,.0364423664946087,.0155262664906382,-.0111361819105275,-.0245790656341835,-.0162533851431734,-.0238264506073575,-.0360313798109832,-.0163249492895439,.022760539936583,.0534222402072165,.0534222402072165,.0227605399365829,-.0163249492895438,-.0360313798109832,-.0238264506073575,-.0292837494080609,-.0442841409574315,-.0200640763482701,.0279736985956793,.0656582686538416,.0656582686538417,.0279736985956792,-.02006407634827,-.0442841409574315,-.0292837494080609,-.0321393500254263,-.0486025026021575,-.0220206218714746,.030701549796244,.0720609252908172,.0720609252908173,.0307015497962439,-.0220206218714745,-.0486025026021575,-.0321393500254263,-.0321393500254263,-.0486025026021575,-.0220206218714746,.030701549796244,.0720609252908172,.0720609252908173,.0307015497962439,-.0220206218714745,-.0486025026021575,-.0321393500254263,-.0292837494080609,-.0442841409574315,-.0200640763482701,.0279736985956793,.0656582686538416,.0656582686538416,.0279736985956792,-.02006407634827,-.0442841409574315,-.0292837494080609,-.0238264506073575,-.0360313798109832,-.0163249492895439,.022760539936583,.0534222402072165,.0534222402072165,.0227605399365829,-.0163249492895438,-.0360313798109832,-.0238264506073575,-.0162533851431734,-.0245790656341835,-.0111361819105275,.0155262664906383,.0364423664946087,.0364423664946087,.0155262664906382,-.0111361819105275,-.0245790656341836,-.0162533851431734,-.00725075497095922,-.0109649024347056,-.00496793287269008,.00692638197801623,.0162572084329942,.0162572084329943,.00692638197801622,-.00496793287269006,-.0109649024347056,-.00725075497095922},
			{.00705610400811139,.00598003257676846,-.0064239867066964,-.0149987461209001,-.00792740615413668,.00792740615413675,.0149987461209001,.00642398670669633,-.0059800325767685,-.00705610400811139,.0158170530535735,.0134049175607548,-.014400090820439,-.033621381269131,-.0177701750956426,.0177701750956428,.0336213812691311,.0144000908204389,-.0134049175607549,-.0158170530535735,.023186814932101,.0196507744905787,-.0211096365250878,-.0492868515145319,-.0260499702351701,.0260499702351703,.049286851514532,.0211096365250876,-.0196507744905789,-.023186814932101,.0284976092004682,.0241516609141409,-.0259446661310533,-.0605756951653628,-.0320165522353802,.0320165522353805,.0605756951653629,.025944666131053,-.024151660914141,-.0284976092004682,.0312765494684072,.0265068066591675,-.0284746565222694,-.0664827253786431,-.0351386417279264,.0351386417279267,.0664827253786433,.0284746565222691,-.0265068066591676,-.0312765494684072,.0312765494684072,.0265068066591675,-.0284746565222694,-.0664827253786431,-.0351386417279264,.0351386417279267,.0664827253786433,.0284746565222691,-.0265068066591676,-.0312765494684072,.0284976092004682,.0241516609141409,-.0259446661310533,-.0605756951653627,-.0320165522353802,.0320165522353805,.0605756951653629,.025944666131053,-.024151660914141,-.0284976092004682,.023186814932101,.0196507744905787,-.0211096365250878,-.0492868515145319,-.0260499702351701,.0260499702351703,.049286851514532,.0211096365250876,-.0196507744905789,-.023186814932101,.0158170530535735,.0134049175607548,-.014400090820439,-.033621381269131,-.0177701750956426,.0177701750956428,.0336213812691311,.0144000908204389,-.0134049175607549,-.0158170530535735,.00705610400811139,.00598003257676846,-.0064239867066964,-.0149987461209001,-.00792740615413669,.00792740615413675,.0149987461209001,.00642398670669633,-.0059800325767685,-.00705610400811139},
			{-.0119026284826952,-.010087451374564,.0108363378798955,.0253007187221328,.0133723893774315,-.0133723893774316,-.0253007187221329,-.0108363378798954,.010087451374564,.0119026284826952,-.0236992236673589,-.0200850397630209,.0215761413980188,.0503760486864314,.0266256522484521,-.0266256522484523,-.0503760486864316,-.0215761413980185,.020085039763021,.0236992236673589,-.0272855920608065,-.0231244790627329,.0248412268982238,.0579993814728395,.0306548727418767,-.0306548727418769,-.0579993814728396,-.0248412268982236,.0231244790627331,.0272855920608065,-.0213920963213377,-.01812975442821,.0194756968206063,.0454719235074688,.0240336361018362,-.0240336361018364,-.0454719235074689,-.019475696820606,.0181297544282102,.0213920963213377,-.00806490596028758,-.00683499000520012,.00734241567585834,.017143097217421,.00906077703341984,-.00906077703341992,-.017143097217421,-.00734241567585826,.00683499000520017,.00806490596028758,.00806490596028758,.00683499000520012,-.00734241567585834,-.017143097217421,-.00906077703341984,.00906077703341992,.017143097217421,.00734241567585826,-.00683499000520017,-.00806490596028758,.0213920963213377,.01812975442821,-.0194756968206063,-.0454719235074688,-.0240336361018362,.0240336361018364,.0454719235074689,.019475696820606,-.0181297544282102,-.0213920963213377,.0272855920608065,.0231244790627329,-.0248412268982238,-.0579993814728395,-.0306548727418767,.0306548727418769,.0579993814728396,.0248412268982236,-.0231244790627331,-.0272855920608065,.0236992236673589,.0200850397630209,-.0215761413980188,-.0503760486864314,-.0266256522484521,.0266256522484523,.0503760486864316,.0215761413980185,-.020085039763021,-.0236992236673589,.0119026284826952,.010087451374564,-.0108363378798955,-.0253007187221328,-.0133723893774315,.0133723893774316,.0253007187221328,.0108363378798954,-.010087451374564,-.0119026284826952},
			{-.00788896410842497,-.00668587967398251,.00718223548063763,.0167691079501009,.00886310952205344,-.00886310952205352,-.0167691079501009,-.00718223548063755,.00668587967398256,.00788896410842497,-.0176840029139668,-.0149871534477983,.0160997909767078,.0375898470038056,.0198676597409558,-.019867659740956,-.0375898470038057,-.0160997909767076,.0149871534477984,.0176840029139668,-.0259236471823205,-.0219702337835042,.0236012911228172,.0551043751861423,.0291247521258975,-.0291247521258977,-.0551043751861425,-.023601291122817,.0219702337835044,.0259236471823205,-.0318612956810125,-.0270023777840407,.0290070185583519,.0677256860801781,.0357955935981209,-.0357955935981212,-.0677256860801783,-.0290070185583516,.0270023777840409,.0318612956810126,-.0349682453529598,-.0296355107751738,.031835633806656,.0743299466305301,.0392861957663534,-.0392861957663538,-.0743299466305303,-.0318356338066557,.029635510775174,.0349682453529598,-.0349682453529598,-.0296355107751738,.031835633806656,.0743299466305301,.0392861957663534,-.0392861957663538,-.0743299466305303,-.0318356338066557,.029635510775174,.0349682453529598,-.0318612956810125,-.0270023777840407,.0290070185583519,.0677256860801781,.035795593598121,-.0357955935981213,-.0677256860801783,-.0290070185583516,.0270023777840409,.0318612956810126,-.0259236471823205,-.0219702337835042,.0236012911228172,.0551043751861423,.0291247521258975,-.0291247521258977,-.0551043751861425,-.023601291122817,.0219702337835044,.0259236471823205,-.0176840029139668,-.0149871534477983,.0160997909767078,.0375898470038056,.0198676597409558,-.019867659740956,-.0375898470038057,-.0160997909767076,.0149871534477984,.0176840029139668,-.00788896410842497,-.00668587967398251,.00718223548063763,.0167691079501009,.00886310952205344,-.00886310952205352,-.0167691079501009,-.00718223548063755,.00668587967398256,.00788896410842497},
			{-.00794977455310546,-.000443472655919539,.0146816998664313,.00745176669518039,-.013740219487093,-.013740219487093,.00745176669518016,.0146816998664309,-.000443472655919117,-.00794977455310563,-.0178203163850575,-.000994093981384771,.0329106863273325,.0167039755968772,-.0308002518592783,-.0308002518592785,.0167039755968767,.0329106863273315,-.000994093981383826,-.0178203163850579,-.0261234742434189,-.00145727987972302,.0482450169810793,.0244869881565972,-.0451512514565309,-.0451512514565312,.0244869881565965,.0482450169810779,-.00145727987972163,-.0261234742434194,-.0321068918748639,-.0017910606795139,.0592952349782769,.0300955789325382,-.05549286189967,-.0554928618996703,.0300955789325372,.0592952349782752,-.0017910606795122,-.0321068918748646,-.0352377908243784,-.00196571570441829,.0650774013003321,.0330303449893761,-.0609042403571174,-.0609042403571177,.0330303449893751,.0650774013003302,-.00196571570441642,-.0352377908243792,-.0352377908243784,-.00196571570441829,.0650774013003321,.0330303449893761,-.0609042403571174,-.0609042403571177,.0330303449893751,.0650774013003302,-.00196571570441642,-.0352377908243792,-.0321068918748639,-.0017910606795139,.0592952349782769,.0300955789325382,-.05549286189967,-.0554928618996703,.0300955789325373,.0592952349782752,-.0017910606795122,-.0321068918748646,-.0261234742434189,-.00145727987972302,.0482450169810793,.0244869881565972,-.0451512514565309,-.0451512514565312,.0244869881565965,.0482450169810779,-.00145727987972163,-.0261234742434194,-.0178203163850575,-.000994093981384771,.0329106863273325,.0167039755968772,-.0308002518592783,-.0308002518592785,.0167039755968767,.0329106863273315,-.000994093981383826,-.0178203163850579,-.00794977455310546,-.000443472655919539,.0146816998664313,.00745176669518039,-.013740219487093,-.013740219487093,.00745176669518016,.0146816998664309,-.000443472655919117,-.00794977455310563}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_left_fluxes_to_dofs[j][0];
			for (int i = 1; i < 100; i++)
			{
				std::cout << ", " << timestep_left_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_right_fluxes_to_dofs:" << std::endl;
		static const T timestep_right_fluxes_to_dofs[35][100] = 
		{
			{-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00696606173921495,-.0156152131511908,-.0228909302027138,-.0281339539329744,-.0308774324097372,-.0308774324097372,-.0281339539329744,-.0228909302027138,-.0156152131511908,-.00696606173921495,-.00634712295584285,-.0142277920527693,-.0208570572598191,-.0256342351721291,-.0281339539329744,-.0281339539329744,-.0256342351721291,-.0208570572598191,-.0142277920527693,-.00634712295584285,-.00516427761687462,-.01157631080205,-.016970150840012,-.0208570572598191,-.0228909302027138,-.0228909302027138,-.0208570572598191,-.016970150840012,-.01157631080205,-.00516427761687462,-.00352284922654042,-.00789686391411969,-.01157631080205,-.0142277920527693,-.0156152131511908,-.0156152131511908,-.0142277920527693,-.01157631080205,-.00789686391411969,-.00352284922654042,-.00157156901877799,-.00352284922654042,-.00516427761687462,-.00634712295584285,-.00696606173921495,-.00696606173921495,-.00634712295584285,-.00516427761687462,-.00352284922654042,-.00157156901877799},
			{.00265100998409647,.00594253790984439,.008711390565351,.0107066798372692,.0117507401837501,.0117507401837501,.0107066798372692,.008711390565351,.00594253790984439,.00265100998409647,.00527840372812152,.0118321373536874,.0173451766357429,.0213179803576284,.0233968001502014,.0233968001502014,.0213179803576284,.0173451766357429,.0118321373536874,.00527840372812152,.00607717674128413,.0136226771632526,.019969996509099,.0245439987302064,.0269374032410133,.0269374032410133,.0245439987302064,.019969996509099,.0136226771632526,.00607717674128413,.00476454936076254,.0106802748271453,.0156566178922341,.0192426678437941,.0211191138346674,.0211191138346674,.0192426678437941,.0156566178922341,.0106802748271453,.00476454936076254,.00179625418474627,.00402650637025399,.00590260762948584,.00725456281862631,.00796199046987792,.00796199046987792,.00725456281862631,.00590260762948584,.00402650637025399,.00179625418474628,-.00179625418474627,-.00402650637025399,-.00590260762948584,-.00725456281862631,-.00796199046987792,-.00796199046987792,-.00725456281862631,-.00590260762948584,-.00402650637025399,-.00179625418474628,-.00476454936076254,-.0106802748271453,-.0156566178922341,-.0192426678437941,-.0211191138346674,-.0211191138346674,-.0192426678437941,-.0156566178922341,-.0106802748271453,-.00476454936076254,-.00607717674128413,-.0136226771632526,-.019969996509099,-.0245439987302064,-.0269374032410133,-.0269374032410133,-.0245439987302064,-.019969996509099,-.0136226771632526,-.00607717674128413,-.00527840372812152,-.0118321373536874,-.0173451766357429,-.0213179803576284,-.0233968001502014,-.0233968001502014,-.0213179803576284,-.0173451766357429,-.0118321373536874,-.00527840372812152,-.00265100998409647,-.00594253790984439,-.008711390565351,-.0107066798372692,-.0117507401837501,-.0117507401837501,-.0107066798372692,-.008711390565351,-.00594253790984439,-.00265100998409647},
			{-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.002221727122666,-.00498025195339731,-.00730073926966875,-.00897292772598497,-.00984792148436088,-.00984792148436088,-.00897292772598497,-.00730073926966875,-.00498025195339731,-.002221727122666,.00309757218884701,.00694355746343096,.010178822902703,.0125102183310402,.0137301504746905,.0137301504746905,.0125102183310402,.010178822902703,.00694355746343096,.00309757218884701,.00727044463764254,.0162975217520129,.0238911521277625,.0293632704051698,.0322266255011462,.0322266255011462,.0293632704051698,.0238911521277625,.0162975217520129,.00727044463764254,.00727044463764255,.016297521752013,.0238911521277625,.0293632704051698,.0322266255011463,.0322266255011463,.0293632704051698,.0238911521277625,.016297521752013,.00727044463764255,.003097572188847,.00694355746343095,.010178822902703,.0125102183310401,.0137301504746904,.0137301504746904,.0125102183310401,.010178822902703,.00694355746343095,.003097572188847,-.00222172712266599,-.00498025195339728,-.00730073926966871,-.00897292772598493,-.00984792148436084,-.00984792148436084,-.00897292772598493,-.00730073926966871,-.00498025195339728,-.00222172712266599,-.0049036534431635,-.0109920923186073,-.0161137229194875,-.0198044699053702,-.0217356999435838,-.0217356999435838,-.0198044699053702,-.0161137229194875,-.0109920923186073,-.0049036534431635,-.00324263620133468,-.00726873481045476,-.0106555126463622,-.0130960908652564,-.0143731542849291,-.0143731542849291,-.0130960908652564,-.0106555126463622,-.00726873481045476,-.00324263620133468},
			{.00352805200405569,.00790852652678675,.0115934074660505,.0142488046002341,.0156382747342036,.0156382747342036,.0142488046002341,.0115934074660505,.00790852652678675,.00352805200405569,.00299001628838423,.00670245878037739,.00982538724528936,.0120758304570704,.0132534033295837,.0132534033295837,.0120758304570704,.00982538724528936,.00670245878037739,.00299001628838423,-.0032119933533482,-.00720004541021951,-.0105548182625439,-.0129723330655266,-.0142373282611347,-.0142373282611347,-.0129723330655266,-.0105548182625439,-.00720004541021951,-.0032119933533482,-.00749937306045005,-.0168106906345655,-.0246434257572659,-.0302878475826814,-.0332413626893215,-.0332413626893215,-.0302878475826814,-.0246434257572659,-.0168106906345655,-.00749937306045005,-.00396370307706834,-.0088850875478213,-.0130249851175851,-.0160082761176901,-.0175693208639632,-.0175693208639632,-.0160082761176901,-.0130249851175851,-.0088850875478213,-.00396370307706834,.00396370307706838,.00888508754782138,.0130249851175852,.0160082761176902,.0175693208639633,.0175693208639633,.0160082761176902,.0130249851175852,.00888508754782138,.00396370307706837,.00749937306045007,.0168106906345656,.024643425757266,.0302878475826815,.0332413626893216,.0332413626893216,.0302878475826815,.024643425757266,.0168106906345656,.00749937306045007,.00321199335334816,.00720004541021943,.0105548182625438,.0129723330655265,.0142373282611345,.0142373282611345,.0129723330655265,.0105548182625438,.00720004541021943,.00321199335334816,-.00299001628838425,-.00670245878037744,-.00982538724528943,-.0120758304570705,-.0132534033295838,-.0132534033295838,-.0120758304570705,-.00982538724528943,-.00670245878037744,-.00299001628838425,-.0035280520040557,-.00790852652678675,-.0115934074660505,-.0142488046002341,-.0156382747342036,-.0156382747342036,-.0142488046002341,-.0115934074660505,-.00790852652678675,-.0035280520040557},
			{-.00355524726055973,-.00796948776183025,-.0116827728408898,-.0143586385526627,-.0157588191287273,-.0157588191287273,-.0143586385526627,-.0116827728408898,-.00796948776183025,-.00355524726055973,-.000198327000917931,-.000444572343586338,-.000651715374523445,-.000800986686075343,-.000879094787718522,-.000879094787718522,-.000800986686075343,-.000651715374523445,-.000444572343586338,-.000198327000917931,.00656585578393543,.0147181063597184,.0215758275045217,.0265176352250662,.0291034986151868,.0291034986151868,.0265176352250662,.0215758275045218,.0147181063597184,.00656585578393543,.00333253137587672,.00747024498425,.0109509140141708,.0134591520602391,.0147716193401924,.0147716193401924,.0134591520602391,.0109509140141708,.00747024498425,.00333253137587672,-.00614481295848751,-.0137742913733916,-.020192253500946,-.0248171622895082,-.0272372043055647,-.0272372043055647,-.0248171622895082,-.020192253500946,-.0137742913733916,-.00614481295848751,-.00614481295848755,-.0137742913733917,-.0201922535009461,-.0248171622895084,-.0272372043055649,-.0272372043055649,-.0248171622895084,-.0201922535009461,-.0137742913733917,-.00614481295848755,.00333253137587662,.00747024498424977,.0109509140141705,.0134591520602387,.014771619340192,.014771619340192,.0134591520602387,.0109509140141704,.00747024498424977,.00333253137587662,.00656585578393523,.014718106359718,.0215758275045211,.0265176352250655,.0291034986151859,.0291034986151859,.0265176352250655,.0215758275045211,.014718106359718,.00656585578393523,-.000198327000917742,-.000444572343585915,-.000651715374522825,-.000800986686074581,-.000879094787717686,-.000879094787717686,-.000800986686074581,-.000651715374522825,-.000444572343585915,-.000198327000917742,-.0035552472605598,-.00796948776183042,-.0116827728408901,-.014358638552663,-.0157588191287277,-.0157588191287277,-.014358638552663,-.0116827728408901,-.00796948776183042,-.0035552472605598},
			{-.00435807799231643,-.00895572847242053,-.0110946867695545,-.0103234510778548,-.00712570259555149,-.0027257963949736,.00134726370928414,.00379129532198382,.00397366731673714,.00213554377116887,-.00976912340532947,-.0200752755655013,-.0248699918602522,-.0231411800629203,-.015973066137951,-.00611018008563575,.0030200435739616,.00849861611742791,.00890742351499224,.00478706224964608,-.0143209266404436,-.0294291039975391,-.0364578595439588,-.0339235290930591,-.0234155203996869,-.00895714355686739,.0044271958372491,.0124584420641652,.0130577282547148,.00701753519287297,-.0176010449034518,-.0361696553536002,-.0448083031935061,-.0416935002770177,-.0287786982183586,-.0110087210073302,.00544121722596732,.0153119699376961,.0160485188646415,.00862485753488115,-.0193174082690916,-.0396967341001901,-.049177778443395,-.0457592359678989,-.0315850488415031,-.012082234855126,.00597181674221008,.0168051144868407,.0176134878766033,.00946590927856652,-.0193174082690916,-.0396967341001901,-.049177778443395,-.0457592359678989,-.0315850488415031,-.012082234855126,.00597181674221008,.0168051144868407,.0176134878766033,.00946590927856652,-.0176010449034518,-.0361696553536002,-.0448083031935061,-.0416935002770177,-.0287786982183586,-.0110087210073302,.00544121722596732,.0153119699376961,.0160485188646415,.00862485753488115,-.0143209266404436,-.0294291039975391,-.0364578595439588,-.0339235290930591,-.0234155203996869,-.00895714355686739,.0044271958372491,.0124584420641652,.0130577282547148,.00701753519287297,-.00976912340532947,-.0200752755655013,-.0248699918602522,-.0231411800629204,-.015973066137951,-.00611018008563575,.0030200435739616,.00849861611742791,.00890742351499224,.00478706224964608,-.00435807799231643,-.00895572847242053,-.0110946867695545,-.0103234510778548,-.00712570259555149,-.0027257963949736,.00134726370928414,.00379129532198382,.00397366731673714,.00213554377116887},
			{.00735144822140268,.0151070206329527,.0187151343924685,.0174141711549991,.0120200312533801,.00459802488506717,-.00227263931871786,-.00639536770584407,-.00670300292458653,-.00360235394730565,.0146374068493633,.0300794619817186,.0372635470036412,.0346732100211315,.0239329833386512,.0091550887552036,-.00452503307213473,-.0127337629596822,-.0133462928615887,-.00717261670817588,.0168524639340767,.0346313423832577,.0429025843441626,.0399202555051752,.027554726236705,.0105405147679591,-.00520979962048207,-.0146607444359217,-.0153659675800657,-.00825803816423586,.0132124503996921,.0271512163030548,.0336359282466497,.0312977614352663,.0216030994104901,.00826383780458816,-.00408451958996732,-.0114941268790076,-.0120470267902114,-.00647436007402489,.00498114669912899,.0102361172510779,.0126808796165353,.0117993813673682,.00814445496968726,.00311549993810543,-.00153988023853377,-.00433333185221062,-.00454177733236969,-.00244085966918394,-.00498114669912899,-.0102361172510779,-.0126808796165353,-.0117993813673682,-.00814445496968726,-.00311549993810543,.00153988023853377,.00433333185221062,.00454177733236969,.00244085966918394,-.0132124503996921,-.0271512163030548,-.0336359282466497,-.0312977614352663,-.0216030994104901,-.00826383780458816,.00408451958996732,.0114941268790076,.0120470267902114,.00647436007402489,-.0168524639340767,-.0346313423832577,-.0429025843441627,-.0399202555051752,-.027554726236705,-.0105405147679591,.00520979962048207,.0146607444359217,.0153659675800657,.00825803816423586,-.0146374068493633,-.0300794619817185,-.0372635470036412,-.0346732100211315,-.0239329833386512,-.0091550887552036,.00452503307213473,.0127337629596822,.0133462928615887,.00717261670817588,-.00735144822140268,-.0151070206329527,-.0187151343924685,-.0174141711549991,-.0120200312533801,-.00459802488506717,.00227263931871786,.00639536770584407,.00670300292458653,.00360235394730565},
			{-.00899207180285294,-.0184784562398471,-.0228917931799034,-.0213004938206645,-.0147025430701201,-.00562416665024674,.00277982450810196,.00782262268394569,.00819891289135883,.00440629170986428,-.0135981963807978,-.0279439135132015,-.034617950789736,-.0322115196955465,-.0222338157821648,-.00850510586048974,.00420376976453103,.0118297053005546,.0123987474800167,.00666338317747926,-.00616101485733601,-.0126607133406354,-.0156845586850965,-.0145942627878117,-.0100736057586743,-.00385345836332292,.0019046267056908,.00535975419633905,.00561757348530501,.00301901822910458,.00858979848715933,.0176517958190368,.0218676957586917,.0203475854739549,.0140448035769736,.00537256143445094,-.00265546504496181,-.00747266636314624,-.00783212268480781,-.00420916988800805,.0201614847183412,.0414312876127182,.051326607314859,.0477586912197692,.0329651613029748,.0126101695425049,-.00623275598422004,-.0175394159608113,-.0183831113218753,-.00987952330905502,.0201614847183412,.0414312876127182,.0513266073148591,.0477586912197692,.0329651613029748,.0126101695425049,-.00623275598422005,-.0175394159608113,-.0183831113218753,-.00987952330905502,.00858979848715931,.0176517958190368,.0218676957586917,.0203475854739549,.0140448035769735,.00537256143445092,-.0026554650449618,-.00747266636314622,-.00783212268480779,-.00420916988800804,-.00616101485733598,-.0126607133406353,-.0156845586850964,-.0145942627878116,-.0100736057586743,-.0038534583633229,.0019046267056908,.00535975419633902,.00561757348530498,.00301901822910456,-.0135981963807978,-.0279439135132015,-.034617950789736,-.0322115196955465,-.0222338157821648,-.00850510586048974,.00420376976453103,.0118297053005546,.0123987474800167,.00666338317747926,-.00899207180285294,-.0184784562398471,-.0228917931799034,-.0213004938206645,-.0147025430701201,-.00562416665024674,.00277982450810196,.00782262268394569,.00819891289135883,.00440629170986428},
			{.00978355109164593,.0201049240576492,.0249067215059292,.0231753564852141,.0159966562164246,.00611920400292648,-.0030245037736683,-.0085111674347139,-.00892057858602882,-.00479413209917604,.00829153796163795,.0170388787750279,.0211083915167262,.0196410634821265,.0135571308449561,.00518601188975199,-.00256326027426807,-.0072131956200491,-.00756017066737565,-.00406301637524721,-.0089070969028758,-.018303835195316,-.0226754661889265,-.0210992045770331,-.0145636043300628,-.00557101838708364,.00275355522170069,.00774869905491992,.00812143332734517,.00436465234070388,-.0207963202945656,-.0427358569791504,-.0529427559658922,-.049262494966556,-.0340031531703131,-.0130072327727005,.00642901014367235,.0180916890395887,.0189619502929062,.0101906052040814,-.010991643978613,-.0225875211759695,-.0279822543883277,-.0260370968758398,-.0179719560241624,-.00687481582123123,.00339797568190965,.00956214378689379,.0100221098640343,.0053861213302785,.0109916439786131,.0225875211759697,.027982254388328,.0260370968758401,.0179719560241625,.00687481582123129,-.00339797568190968,-.00956214378689387,-.0100221098640343,-.00538612133027854,.0207963202945656,.0427358569791505,.0529427559658923,.0492624949665561,.0340031531703132,.0130072327727006,-.00642901014367236,-.0180916890395887,-.0189619502929062,-.0101906052040814,.00890709690287571,.0183038351953158,.0226754661889263,.0210992045770329,.0145636043300626,.00557101838708358,-.00275355522170066,-.00774869905491983,-.00812143332734509,-.00436465234070383,-.00829153796163801,-.017038878775028,-.0211083915167264,-.0196410634821266,-.0135571308449562,-.00518601188975203,.00256326027426809,.00721319562004915,.00756017066737571,.00406301637524725,-.00978355109164594,-.0201049240576493,-.0249067215059292,-.0231753564852141,-.0159966562164247,-.00611920400292649,.00302450377366831,.0085111674347139,.00892057858602882,.00479413209917604},
			{-.00774458111624399,-.0136429110526415,-.0119270111969813,-.00443011400826239,.00356799186187752,.00716050023137007,.00509898471326269,.000227342285586894,-.00308610359639842,-.00244256114805105,-.0173603521507787,-.0305821240271899,-.0267357409494468,-.00993060026064162,.00799805622325859,.0160510689637666,.0114299493936489,.000509613377058344,-.00691784931981508,-.0054752763310899,-.025449195315512,-.0448314896363585,-.0391929315383665,-.0145576416560195,.0117246524265364,.0235298676855081,.0167555941284487,.000747061479831429,-.0101411363648726,-.00802641418481,-.0312781735952783,-.0550998606437757,-.0481698262427054,-.0178919780059722,.0144101103942559,.0289192360315085,.0205933576816159,.000918171217707236,-.0124638999285189,-.0098648139207399,-.0343282602007917,-.0604729156402031,-.0528670999299058,-.0196367116713162,.0158153102395603,.0317392911793161,.0226015159980186,.00100770655212074,-.0136793153398003,-.0108267798332916,-.0343282602007917,-.0604729156402031,-.0528670999299058,-.0196367116713162,.0158153102395603,.0317392911793161,.0226015159980186,.00100770655212074,-.0136793153398003,-.0108267798332916,-.0312781735952783,-.0550998606437757,-.0481698262427054,-.0178919780059722,.0144101103942559,.0289192360315085,.0205933576816159,.000918171217707236,-.0124638999285189,-.0098648139207399,-.025449195315512,-.0448314896363585,-.0391929315383665,-.0145576416560195,.0117246524265364,.0235298676855081,.0167555941284487,.000747061479831429,-.0101411363648726,-.00802641418481,-.0173603521507787,-.0305821240271899,-.0267357409494468,-.00993060026064162,.00799805622325859,.0160510689637666,.0114299493936489,.000509613377058344,-.00691784931981508,-.0054752763310899,-.00774458111624399,-.0136429110526415,-.0119270111969814,-.00443011400826239,.00356799186187752,.00716050023137007,.0050989847132627,.000227342285586894,-.00308610359639842,-.00244256114805105},
			{.0130639899374604,.0230136207541996,.0201191454931664,.00747296257306306,-.0060186870123603,-.0120787298326045,-.00860125085985731,-.000383493604971994,.00520581109864628,.00412024792313705,.0260116007120718,.0458222271192026,.0400590617216945,.0148793530550254,-.0119837571925505,-.0240498575870592,-.0171258783925904,-.000763570898011854,.0103652467836164,.00820379105651837,.0299478976965691,.0527564368369676,.0461211401613239,.0171310235004581,-.0137972414076181,-.0276892868919076,-.0197175121878277,-.000879120950335205,.0119338042172612,.00944525859843065,.0234793626877832,.041361417992028,.0361592986724362,.0134308430613569,-.0108171344240126,-.0217085959117477,-.0154586684063579,-.000689237015849483,.00935618654441216,.0074051492548628,.00885181373718756,.0155934201809443,.0136322003698287,.00506348160695133,-.00407810298621525,-.00818422757303578,-.0058279798807838,-.000259845114460841,.00352731978643567,.00279177091694338,-.00885181373718756,-.0155934201809443,-.0136322003698287,-.00506348160695133,.00407810298621525,.00818422757303578,.0058279798807838,.000259845114460841,-.00352731978643567,-.00279177091694338,-.0234793626877832,-.041361417992028,-.0361592986724362,-.0134308430613569,.0108171344240126,.0217085959117477,.0154586684063579,.000689237015849483,-.00935618654441216,-.0074051492548628,-.0299478976965691,-.0527564368369676,-.0461211401613239,-.0171310235004581,.0137972414076181,.0276892868919076,.0197175121878277,.000879120950335204,-.0119338042172612,-.00944525859843065,-.0260116007120718,-.0458222271192026,-.0400590617216945,-.0148793530550254,.0119837571925505,.0240498575870592,.0171258783925904,.000763570898011854,-.0103652467836164,-.00820379105651837,-.0130639899374604,-.0230136207541996,-.0201191454931664,-.00747296257306306,.0060186870123603,.0120787298326045,.00860125085985731,.000383493604971994,-.00520581109864628,-.00412024792313705},
			{-.0159794821505739,-.028149573279088,-.0246091376242554,-.00914070453359377,.00736187812026119,.0147743414290495,.0105207930537044,.000469077919138421,-.00636759259064836,-.00503976415007148,-.0241648577892824,-.0425689912043686,-.0372149926639303,-.0138229651666283,.0111329476238222,.0223423923378182,.0159099941773837,.000709359734017749,-.00962934643082784,-.00762134735221235,-.010948514324695,-.0192869833562834,-.0168612157301903,-.00626285217385856,.0050440700954407,.0101227991776989,.00720843468957457,.000321393789153591,-.00436282465449179,-.0034530486952082,.0152646169438994,.0268902614734654,.0235082123196634,.00873178191808065,-.0070325338727842,-.0141133899326471,-.0100501301855435,-.00044809304112726,.00608272914198089,.00481430302393592,.0358282376130031,.0631152868812828,.055177134148947,.020494740123233,-.0165063621014286,-.0331261432822222,-.0235890919276015,-.00105173840976448,.0142770346504849,.0112998572657607,.0358282376130031,.0631152868812828,.0551771341489471,.020494740123233,-.0165063621014286,-.0331261432822222,-.0235890919276015,-.00105173840976448,.0142770346504849,.0112998572657607,.0152646169438994,.0268902614734653,.0235082123196633,.00873178191808063,-.00703253387278418,-.0141133899326471,-.0100501301855435,-.000448093041127259,.00608272914198088,.0048143030239359,-.0109485143246949,-.0192869833562833,-.0168612157301902,-.00626285217385853,.00504407009544068,.0101227991776988,.00720843468957453,.00032139378915359,-.00436282465449177,-.00345304869520818,-.0241648577892824,-.0425689912043686,-.0372149926639303,-.0138229651666283,.0111329476238222,.0223423923378182,.0159099941773837,.000709359734017748,-.00962934643082784,-.00762134735221235,-.0159794821505739,-.028149573279088,-.0246091376242554,-.00914070453359377,.00736187812026119,.0147743414290495,.0105207930537044,.000469077919138421,-.00636759259064836,-.00503976415007148},
			{-.0113741780920774,-.0157705273052622,-.00608002687266981,.00608414677874269,.00833345789650543,.000957688797273868,-.00550488775656931,-.00405218361689488,.00128277564401616,.00255185410234015,-.0254965031859464,-.0353514158497966,-.0136290660542447,.0136383012884271,.0186803858782727,.00214676746514514,-.0123398268505581,-.00908342665833277,.00287548630149319,.00572026881623259,-.0373762861378719,-.0518229744896624,-.019979362225491,.0199929005037731,.0273842825684829,.00314702743601977,-.0180894178270092,-.0133157379040649,.00421528387662904,.0083855579136398,-.0459370896279213,-.063692701172494,-.0245555096050894,.0245721487409458,.0336564804246319,.00386783429623537,-.0222326853173956,-.0163656240029997,.005180768162293,.0103062172640176,-.0504166383248294,-.0699036857788496,-.0269500365972072,.0269682982960067,.0369384872790024,.00424500560207766,-.024400702432737,-.0179615154769812,.00568597002550107,.0113112265601881,-.0504166383248294,-.0699036857788496,-.0269500365972072,.0269682982960067,.0369384872790024,.00424500560207766,-.024400702432737,-.0179615154769812,.00568597002550107,.0113112265601881,-.0459370896279213,-.063692701172494,-.0245555096050894,.0245721487409458,.0336564804246319,.00386783429623537,-.0222326853173956,-.0163656240029997,.005180768162293,.0103062172640176,-.0373762861378719,-.0518229744896624,-.019979362225491,.0199929005037731,.0273842825684829,.00314702743601977,-.0180894178270092,-.0133157379040649,.00421528387662904,.0083855579136398,-.0254965031859464,-.0353514158497966,-.0136290660542447,.0136383012884271,.0186803858782727,.00214676746514514,-.0123398268505581,-.00908342665833277,.0028754863014932,.00572026881623259,-.0113741780920774,-.0157705273052622,-.00608002687266981,.0060841467787427,.00833345789650543,.000957688797273868,-.00550488775656931,-.00405218361689488,.00128277564401616,.00255185410234015},
			{.019186595894105,.0266026021391187,.0102561273163527,-.0102630770029412,-.014057340035159,-.00161548270097898,.00928595068344064,.00683544858509201,-.00216385726557964,-.00430460935690978,.0382022700423445,.052968217830653,.0204208890149373,-.0204347264775717,-.0279894517539301,-.00321657404638918,.0184891784643094,.0136100043045404,-.00430843803919177,-.00857087155986831,.0439833629490508,.0609838197318553,.0235111518841866,-.0235270833495331,-.0322250540052594,-.00370333342961076,.0212871184388807,.015669586084838,-.00496042758221491,-.00986788885031005,.0344832662837457,.0478117441190089,.0184329086431872,-.0184453990242086,-.025264669269575,-.00290343948776025,.016689250760381,.0122850658360403,-.00388901015588316,-.00773649434850662,.0130003294489922,.0180252189558684,.00694928035797946,-.00695398928163203,-.00952488146632226,-.00109460831133298,.00629191435511048,.00463151900567974,-.00146617239912651,-.00291668934384738,-.0130003294489922,-.0180252189558684,-.00694928035797946,.00695398928163203,.00952488146632226,.00109460831133298,-.00629191435511048,-.00463151900567974,.00146617239912651,.00291668934384738,-.0344832662837457,-.0478117441190089,-.0184329086431872,.0184453990242085,.025264669269575,.00290343948776025,-.016689250760381,-.0122850658360403,.00388901015588316,.00773649434850662,-.0439833629490508,-.0609838197318553,-.0235111518841866,.0235270833495331,.0322250540052594,.00370333342961076,-.0212871184388807,-.015669586084838,.00496042758221491,.00986788885031005,-.0382022700423445,-.052968217830653,-.0204208890149373,.0204347264775717,.0279894517539301,.00321657404638918,-.0184891784643094,-.0136100043045404,.00430843803919177,.00857087155986831,-.019186595894105,-.0266026021391187,-.0102561273163527,.0102630770029412,.014057340035159,.00161548270097898,-.00928595068344064,-.00683544858509201,.00216385726557964,.00430460935690978},
			{-.0149432388948351,-.0143716720347549,.00337087856178778,.00991288633119597,-.000790460352998803,-.0073875487912777,-.000452661276806676,.00518463670922043,.000881265822650354,-.00248741703478494,-.0334969555607631,-.032215723971892,.00755620452698254,.022220852872113,-.00177190604415069,-.016560023921583,-.00101469131156782,.0116219480040223,.00197545674711399,-.00557582586088984,-.0491044511733968,-.0472262454396426,.0110769253515695,.0325743867352538,-.0025975039334847,-.0242759639637272,-.0014874742832843,.0170370521365157,.00289589659000063,-.00817381353480803,-.0603515171721834,-.0580431202166143,.0136140255026904,.0400353453393464,-.00319244588831583,-.0298362209742388,-.00182817092148764,.02093928187792,.00355918350788891,-.0100459741656872,-.0662366867006586,-.063703186747533,.0149415952441392,.0439393862896491,-.00350375679056903,-.0327456957770259,-.00200644474630778,.0229811729426385,.00390625677643855,-.011025605893504,-.0662366867006586,-.063703186747533,.0149415952441392,.0439393862896491,-.00350375679056903,-.0327456957770259,-.00200644474630778,.0229811729426385,.00390625677643855,-.011025605893504,-.0603515171721834,-.0580431202166143,.0136140255026904,.0400353453393464,-.00319244588831583,-.0298362209742388,-.00182817092148764,.02093928187792,.00355918350788891,-.0100459741656872,-.0491044511733968,-.0472262454396426,.0110769253515695,.0325743867352538,-.0025975039334847,-.0242759639637272,-.0014874742832843,.0170370521365157,.00289589659000063,-.00817381353480803,-.0334969555607631,-.032215723971892,.00755620452698254,.022220852872113,-.00177190604415069,-.016560023921583,-.00101469131156782,.0116219480040223,.00197545674711399,-.00557582586088984,-.0149432388948351,-.0143716720347549,.00337087856178778,.00991288633119597,-.000790460352998802,-.0073875487912777,-.000452661276806676,.00518463670922043,.000881265822650354,-.00248741703478494},
			{5.02239597246239e-005,.000582196454509688,.00202770964403212,.0044045611294125,.00726150487830256,.00980179190824761,.0111426514550797,.0106221354138729,.00804698659569713,.00379931823382166,.000112582670920343,.00130505902378937,.0045453398901192,.00987332060036796,.0162774822730591,.0219718221917673,.0249775101132113,.0238107146933112,.0180382281439193,.00851660037928648,.000165039185640327,.00191313526973573,.00666318526458248,.0144736732405605,.0238617754993688,.0322093232636935,.0366154746082026,.0349050251728234,.0264429193136947,.0124848058722647,.000202840374105465,.00235132688229124,.00818934598203497,.0177887771525588,.0293271652446273,.0395866663734618,.0450020189978578,.0428998018667281,.032499503831288,.0153443722103867,.00022262032405145,.00258061618548581,.0089879288791991,.0195234472015054,.0321869995510595,.0434469544588521,.0493903844165837,.0470831698838632,.0356686882793406,.0168406764624987,.00022262032405145,.00258061618548581,.0089879288791991,.0195234472015054,.0321869995510595,.0434469544588521,.0493903844165837,.0470831698838632,.0356686882793406,.0168406764624987,.000202840374105465,.00235132688229124,.00818934598203497,.0177887771525588,.0293271652446273,.0395866663734618,.0450020189978578,.0428998018667281,.032499503831288,.0153443722103867,.000165039185640327,.00191313526973573,.00666318526458248,.0144736732405605,.0238617754993688,.0322093232636935,.0366154746082026,.0349050251728234,.0264429193136947,.0124848058722647,.000112582670920343,.00130505902378937,.0045453398901192,.00987332060036795,.0162774822730591,.0219718221917673,.0249775101132113,.0238107146933112,.0180382281439193,.00851660037928648,5.02239597246239e-005,.000582196454509688,.00202770964403212,.0044045611294125,.00726150487830256,.00980179190824761,.0111426514550797,.0106221354138729,.00804698659569713,.00379931823382166},
			{-8.47205671528915e-005,-.00098208134306624,-.00342045334548888,-.00742985855733822,-.0122491100870261,-.0165342074606321,-.0187960438809947,-.0179180084878277,-.0135741043119093,-.00640890119633241,-.000168686410157299,-.00195541391909401,-.00681043594667852,-.0147935290111086,-.0243891002815539,-.0329211216952395,-.0374246452071332,-.0356763963055754,-.0270272851613995,-.0127607093790004,-.000194213474599731,-.0022513238093916,-.0078410491248818,-.0170322118311734,-.0280798666805756,-.0379030262496602,-.0430880613000287,-.0410752524832421,-.0311172841682885,-.0146917686169386,-.000152264731741222,-.00176505372042869,-.00614743773072645,-.0133533740168041,-.0220148132165338,-.0297162394935752,-.033781343486181,-.0322032871995818,-.0243961699164533,-.0115184500553977,-5.74044134859514e-005,-.000665432319314789,-.00231760863686745,-.00503427546699351,-.00829967272293673,-.0112031412634403,-.0127357017440349,-.012140768205952,-.00919745373300046,-.00434250178709194,5.74044134859514e-005,.000665432319314789,.00231760863686745,.00503427546699351,.00829967272293673,.0112031412634403,.0127357017440349,.012140768205952,.00919745373300046,.00434250178709194,.000152264731741222,.00176505372042869,.00614743773072645,.013353374016804,.0220148132165338,.0297162394935752,.033781343486181,.0322032871995818,.0243961699164533,.0115184500553977,.000194213474599731,.0022513238093916,.0078410491248818,.0170322118311734,.0280798666805756,.0379030262496602,.0430880613000287,.0410752524832421,.0311172841682885,.0146917686169386,.000168686410157299,.00195541391909401,.00681043594667852,.0147935290111086,.0243891002815539,.0329211216952395,.0374246452071332,.0356763963055754,.0270272851613995,.0127607093790004,8.47205671528915e-005,.00098208134306624,.00342045334548889,.00742985855733822,.0122491100870261,.0165342074606321,.0187960438809947,.0179180084878277,.0135741043119093,.00640890119633241},
			{.000103627666330889,.00120125255471199,.00418379633067172,.00908798098667198,.0149827454608269,.0202241485315363,.02299075925832,.021916772600681,.0166034388065025,.00783917644841335,.000156710198511108,.00181658559896353,.00632691612890426,.0137432343592401,.0226575498468763,.0305838242171288,.0347676115352073,.0331434828804184,.0251084317871825,.0118547386126776,7.10016118524656e-005,.000823051127621699,.00286657312335497,.00622672806775283,.0102655894449765,.0138567932186921,.0157523663597753,.0150165128324275,.0113760249486971,.0053710961863794,-9.89917330502748e-005,-.00114751278719633,-.00399662816086328,-.00868141703515759,-.0143124707091497,-.0193193920454638,-.0219622344466939,-.0209362941329066,-.0158606599978981,-.00748847957105857,-.000232347745540099,-.00269337651607831,-.00938065749861198,-.0203765265447757,-.0335934143176455,-.0453453742919027,-.0515485031272343,-.0491404745815968,-.0372272358482508,-.0175765318198328,-.000232347745540099,-.00269337651607831,-.00938065749861198,-.0203765265447757,-.0335934143176455,-.0453453742919027,-.0515485031272343,-.0491404745815968,-.0372272358482508,-.0175765318198328,-9.89917330502745e-005,-.00114751278719633,-.00399662816086327,-.00868141703515757,-.0143124707091497,-.0193193920454637,-.0219622344466939,-.0209362941329065,-.0158606599978981,-.00748847957105856,7.10016118524653e-005,.000823051127621695,.00286657312335496,.0062267280677528,.0102655894449765,.0138567932186921,.0157523663597753,.0150165128324274,.0113760249486971,.00537109618637938,.000156710198511108,.00181658559896353,.00632691612890426,.0137432343592401,.0226575498468763,.0305838242171288,.0347676115352073,.0331434828804184,.0251084317871825,.0118547386126776,.000103627666330889,.00120125255471199,.00418379633067172,.00908798098667198,.0149827454608269,.0202241485315363,.02299075925832,.021916772600681,.0166034388065025,.00783917644841335},
			{-.000112748940416249,-.00130698642082339,-.00455205274713417,-.00988790216985572,-.0163015219298734,-.0220042716244384,-.0250143985435671,-.0238458799234204,-.0180648681860857,-.00852917825508497,-9.55544782092518e-005,-.00110766810763179,-.0038578546585677,-.00837997527015808,-.0138155038643826,-.0186485716467526,-.0211996475685226,-.0202093306164191,-.015309935925496,-.00722845975057063,.000102648386927895,.00118990074176208,.00414425953786325,.0090020997455888,.0148411588117245,.0200330307267911,.0227734970367722,.0217096595318657,.0164465366372481,.00776509638558674,.000239663804666116,.00277818431911703,.00967603133467496,.0210181332564652,.0346511883228178,.0467731886166761,.0531716387245257,.0506877872817214,.0383994299572614,.018129973583278,.000126671410044506,.00146837577562602,.00511414952502166,.0111088805412601,.0183144671790717,.0247214040626872,.0281032276066716,.0267904179187864,.0202955550353876,.00958237862017949,-.000126671410044507,-.00146837577562603,-.0051141495250217,-.0111088805412602,-.0183144671790719,-.0247214040626874,-.0281032276066718,-.0267904179187866,-.0202955550353878,-.00958237862017957,-.000239663804666117,-.00277818431911704,-.00967603133467498,-.0210181332564652,-.0346511883228178,-.0467731886166762,-.0531716387245259,-.0506877872817215,-.0383994299572615,-.018129973583278,-.000102648386927894,-.00118990074176207,-.00414425953786321,-.0090020997455887,-.0148411588117243,-.0200330307267909,-.0227734970367719,-.0217096595318655,-.016446536637248,-.00776509638558665,9.55544782092525e-005,.0011076681076318,.00385785465856773,.00837997527015815,.0138155038643827,.0186485716467528,.0211996475685227,.0202093306164193,.0153099359254961,.00722845975057069,.000112748940416249,.00130698642082339,.00455205274713418,.00988790216985573,.0163015219298734,.0220042716244384,.0250143985435671,.0238458799234204,.0180648681860857,.00852917825508498},
			{.000242033524403197,.00261162972756074,.00794330061991523,.0139365347274024,.0166502677108084,.0135390642006158,.00568409316229951,-.00258267825885607,-.00653083370918549,-.00434965045780644,.000542545446017795,.00585426262277764,.0178058043335604,.0312403146145367,.0373234532024064,.030349339594584,.01274153598164,-.00578936463998794,-.0146396004287218,-.00975023213611034,.000795337842419811,.0085819844909475,.0261022004795129,.0457963560553539,.0547138584602276,.0444902421524166,.0186782984008672,-.00848684808903276,-.0214607426978308,-.0142932332160281,.000977504977807211,.0105476341146326,.0320807454890066,.0562857236534088,.067245723951962,.0546804525678283,.0229564452865287,-.0104307073176385,-.0263761909665255,-.0175670084741824,.00107282623531448,.011576185140315,.0352090947774598,.0617744179108896,.0738031810643202,.0600126090459714,.0251950397513031,-.0114478562435068,-.0289482614400908,-.0192800527823109,.00107282623531448,.011576185140315,.0352090947774598,.0617744179108896,.0738031810643202,.0600126090459713,.0251950397513031,-.0114478562435068,-.0289482614400908,-.0192800527823109,.000977504977807211,.0105476341146326,.0320807454890066,.0562857236534088,.067245723951962,.0546804525678283,.0229564452865287,-.0104307073176385,-.0263761909665255,-.0175670084741824,.000795337842419811,.0085819844909475,.0261022004795129,.0457963560553539,.0547138584602276,.0444902421524166,.0186782984008672,-.00848684808903276,-.0214607426978308,-.0142932332160281,.000542545446017795,.00585426262277763,.0178058043335604,.0312403146145367,.0373234532024064,.030349339594584,.01274153598164,-.00578936463998794,-.0146396004287218,-.00975023213611034,.000242033524403197,.00261162972756074,.00794330061991523,.0139365347274024,.0166502677108084,.0135390642006158,.0056840931622995,-.00258267825885607,-.00653083370918549,-.00434965045780644},
			{-.000408275603560732,-.00440544220444747,-.0133992010522381,-.0235089215074875,-.0280865971597952,-.0228384461275931,-.00958824432382064,.00435660525762981,.0110165733484533,.00733723218846652,-.000812914127394553,-.00877163899626159,-.0266790367491162,-.0468084162917745,-.0559229878604316,-.045473438390492,-.0190910727943728,.00867440016136512,.0219350067266604,.0146091014255062,-.000935931217489798,-.0100990381252994,-.0307163358399061,-.053891864555414,-.0643857307310363,-.0523548664268737,-.0219800965458574,.00998708428163375,.0252543987851819,.0168198750923781,-.00073377666535768,-.00791771700785594,-.0240818236035156,-.0422516013189939,-.0504788662986896,-.0410465839626478,-.0172325504762159,.00782994440604093,.0197996264895458,.0131869005183092,-.000276636740648008,-.0029850110119799,-.00907894391448041,-.0159290228592211,-.0190307346959162,-.0154747264913728,-.00649674052317934,.00295192038968209,.0074645384579598,.00497151429155405,.000276636740648008,.0029850110119799,.00907894391448041,.0159290228592211,.0190307346959162,.0154747264913728,.00649674052317934,-.00295192038968209,-.0074645384579598,-.00497151429155405,.00073377666535768,.00791771700785594,.0240818236035156,.0422516013189938,.0504788662986896,.0410465839626478,.0172325504762159,-.00782994440604093,-.0197996264895458,-.0131869005183092,.000935931217489798,.0100990381252994,.0307163358399061,.053891864555414,.0643857307310362,.0523548664268737,.0219800965458574,-.00998708428163375,-.0252543987851819,-.0168198750923781,.000812914127394553,.00877163899626159,.0266790367491162,.0468084162917745,.0559229878604317,.0454734383904919,.0190910727943728,-.00867440016136512,-.0219350067266604,-.0146091014255062,.000408275603560732,.00440544220444747,.0133992010522381,.0235089215074875,.0280865971597952,.0228384461275931,.00958824432382064,-.00435660525762981,-.0110165733484533,-.00733723218846652},
			{.000499390519149255,.00538860527147271,.0163895023638538,.0287554103499522,.0343546864285749,.0279353048988254,.0117280539636516,-.00532886937735551,-.0134751433486315,-.00897468317910862,.000755199746953267,.00814887984733841,.0247849079292855,.0434851639891283,.0519526292603973,.042244965376185,.0177356658686522,-.0080585446679718,-.0203776893169333,-.0135719005586946,.000342162793575017,.00369206094721455,.0112294440905165,.0197020791513965,.0235384834714527,.019140175067471,.00803560250650699,-.00365113225589165,-.00923263961808127,-.00614910085446924,-.000477049563208446,-.00514753823408188,-.0156562943109128,-.0274689955482004,-.0328177799266881,-.0266855786985745,-.0112033825355281,.00509047471144027,.0128723133542602,.00857318776860553,-.00111970350560565,-.0120820079305306,-.0367475603725958,-.0644736584683689,-.0770280198622701,-.0626348671549949,-.0262959400178523,.0119480716872726,.0302131591759193,.0201224969878625,-.00111970350560565,-.0120820079305306,-.0367475603725958,-.064473658468369,-.0770280198622701,-.0626348671549949,-.0262959400178523,.0119480716872726,.0302131591759193,.0201224969878625,-.000477049563208445,-.00514753823408186,-.0156562943109127,-.0274689955482003,-.0328177799266881,-.0266855786985745,-.0112033825355281,.00509047471144026,.0128723133542602,.00857318776860551,.000342162793575015,.00369206094721453,.0112294440905164,.0197020791513964,.0235384834714526,.0191401750674709,.00803560250650695,-.00365113225589163,-.00923263961808122,-.00614910085446921,.000755199746953267,.00814887984733841,.0247849079292855,.0434851639891283,.0519526292603973,.042244965376185,.0177356658686522,-.0080585446679718,-.0203776893169333,-.0135719005586946,.000499390519149255,.00538860527147271,.0163895023638538,.0287554103499522,.0343546864285749,.0279353048988254,.0117280539636516,-.00532886937735551,-.0134751433486316,-.00897468317910862},
			{.000682727668349805,.00664570504694168,.0164335220422574,.0198480228837579,.0106420257503786,-.00420811508616648,-.0112538525390341,-.00576778839286375,.00325440242770835,.00455104492220568,.00153041107940269,.0148970974896404,.0368375933377187,.0444915821252141,.0238553011261929,-.00943296463557281,-.0252267798862333,-.0129291482816634,.00729511013408645,.010201680545096,.00224348735178138,.0218382173561541,.0540016181505257,.065221889139497,.0349703861072661,-.0138281388148133,-.0369808885747392,-.0189533263510734,.0106941772285541,.0149550284742905,.00275734403299806,.0268401238235168,.0663703494746075,.0801605530322499,.0429801333125165,-.0169953871227395,-.0454511287373907,-.0232944666606214,.0131436113270545,.0183803837780314,.00302622604032566,.0294574346429948,.0728424445705406,.0879773978473551,.0471713348390493,-.0186526898568866,-.0498832890278595,-.0255660232311389,.014425308697734,.0201727442620858,.00302622604032566,.0294574346429948,.0728424445705406,.0879773978473551,.0471713348390493,-.0186526898568866,-.0498832890278595,-.0255660232311389,.014425308697734,.0201727442620858,.00275734403299806,.0268401238235168,.0663703494746075,.0801605530322499,.0429801333125165,-.0169953871227395,-.0454511287373907,-.0232944666606214,.0131436113270545,.0183803837780314,.00224348735178138,.0218382173561541,.0540016181505257,.065221889139497,.0349703861072661,-.0138281388148133,-.0369808885747392,-.0189533263510734,.0106941772285541,.0149550284742905,.00153041107940269,.0148970974896404,.0368375933377187,.0444915821252141,.0238553011261929,-.00943296463557281,-.0252267798862333,-.0129291482816634,.00729511013408645,.010201680545096,.000682727668349805,.00664570504694168,.0164335220422574,.0198480228837579,.0106420257503786,-.00420811508616648,-.0112538525390341,-.00576778839286375,.00325440242770835,.00455104492220568},
			{-.00115166298286166,-.0112103446987566,-.0277209784976989,-.0334807483245332,-.0179515606112687,.00709848245065787,.0189836240489798,.00972942608448226,-.00548970692279972,-.00767695556108528,-.0022930664987097,-.0223208232358323,-.0551950076111466,-.0666632369688139,-.0357432016569395,.0141337288264978,.0377981345051569,.0193721786130192,-.0109305093762129,-.0152855217812643,-.00264007285345256,-.025698600334876,-.0635475863095515,-.0767512858190868,-.0411521673894173,.0162725650625018,.0435180701799816,.0223037416914183,-.0125846071602312,-.0175986571380644,-.00206983570855075,-.0201478836325818,-.0498217551700164,-.0601735485661986,-.0322635890276594,.0127578056007183,.0341184734761949,.017486290549451,-.00986642063466756,-.0137974711263254,-.000780336349086855,-.00759583279519514,-.0187830011695426,-.0226856687251885,-.0121635022365666,.00480974378966284,.0128628011000313,.00659240154668345,-.00371968008127681,-.00520170185530649,.000780336349086855,.00759583279519514,.0187830011695426,.0226856687251885,.0121635022365666,-.00480974378966284,-.0128628011000313,-.00659240154668345,.00371968008127681,.00520170185530649,.00206983570855075,.0201478836325818,.0498217551700164,.0601735485661986,.0322635890276594,-.0127578056007183,-.0341184734761949,-.017486290549451,.00986642063466756,.0137974711263254,.00264007285345256,.025698600334876,.0635475863095515,.0767512858190868,.0411521673894173,-.0162725650625018,-.0435180701799816,-.0223037416914183,.0125846071602312,.0175986571380644,.0022930664987097,.0223208232358323,.0551950076111466,.0666632369688139,.0357432016569395,-.0141337288264978,-.0377981345051569,-.0193721786130192,.0109305093762129,.0152855217812643,.00115166298286166,.0112103446987566,.0277209784976989,.0334807483245332,.0179515606112687,-.00709848245065787,-.0189836240489798,-.00972942608448226,.00548970692279973,.00767695556108528},
			{.00148169134577296,.0125696636361974,.0228905745669076,.0127235476313918,-.00841173160413407,-.0124282780657018,.00206969137961716,.00930159426344669,.000766794088585497,-.00444614718711676,.00332137828441454,.0281763188823123,.0513118049188827,.0285212672154559,-.0188558452230951,-.0278593871779371,.00463944668591847,.0208505727492996,.0017188554429443,-.0099665404395655,.00486893376025062,.0413047290907605,.0752199110957276,.0418104018690363,-.027641495043065,-.0408401269458114,.00680113996769058,.0305656413953226,.00251973505529715,-.0146103277205166,.00598413244462014,.0507653177552249,.0924485590960711,.0513868117060249,-.0339726058003504,-.0501943014082524,.00835889834306659,.0375665095010369,.00309686453723171,-.0179567314824952,.00656767426042838,.055715690424466,.101463667057506,.0563977892686102,-.0372854395753275,-.0550889915004361,.00917401511436313,.0412298022791773,.00339885483776274,-.0197077795738015,.00656767426042838,.055715690424466,.101463667057506,.0563977892686102,-.0372854395753275,-.0550889915004361,.00917401511436313,.0412298022791773,.00339885483776274,-.0197077795738015,.00598413244462014,.0507653177552249,.0924485590960711,.0513868117060249,-.0339726058003504,-.0501943014082523,.0083588983430666,.0375665095010369,.00309686453723171,-.0179567314824952,.00486893376025062,.0413047290907605,.0752199110957276,.0418104018690363,-.027641495043065,-.0408401269458114,.00680113996769058,.0305656413953226,.00251973505529715,-.0146103277205166,.00332137828441454,.0281763188823123,.0513118049188827,.0285212672154559,-.0188558452230951,-.0278593871779371,.00463944668591847,.0208505727492996,.0017188554429443,-.0099665404395655,.00148169134577296,.0125696636361974,.0228905745669076,.0127235476313918,-.00841173160413407,-.0124282780657018,.00206969137961716,.00930159426344669,.000766794088585497,-.00444614718711676},
			{-1.03605502214324e-006,-6.2106840321439e-005,-.000513920961277903,-.00197298030906211,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186,-2.32243419779848e-006,-.000139219488151642,-.00115201180432313,-.00442265791238583,-.0109526932361906,-.0199562351910325,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-3.404543928249e-006,-.000204087101167348,-.00168877757544104,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-4.18433332293283e-006,-.000250831969330619,-.00207558146784554,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-4.59236798590328e-006,-.000275291812791727,-.00227798149655704,-.00874533824411281,-.0216577924252842,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.59236798590328e-006,-.000275291812791727,-.00227798149655704,-.00874533824411281,-.0216577924252843,-.0394613443503929,-.0559691123983598,-.062511846274886,-.052592127399495,-.0262800462022158,-4.18433332293283e-006,-.000250831969330619,-.00207558146784554,-.00796830967106449,-.0197334845170196,-.0359551801249216,-.0509962230340228,-.0569576310628798,-.0479192851873238,-.0239450482604383,-3.404543928249e-006,-.000204087101167348,-.00168877757544105,-.0064833411239848,-.0160559663178376,-.029254598220584,-.0414926030253305,-.0463430472758461,-.0389890811366307,-.0194826660246954,-2.32243419779848e-006,-.000139219488151642,-.00115201180432313,-.00442265791238583,-.0109526932361906,-.0199562351910326,-.0283044784419232,-.0316132439739062,-.0265966829275197,-.0132902411581778,-1.03605502214324e-006,-6.21068403214391e-005,-.000513920961277903,-.00197298030906211,-.00488607722195371,-.00890262368352143,-.0126268365608368,-.0141029012647388,-.0118649763879334,-.00592887458791186},
			{1.74767520475977e-006,.000104765270720098,.000866910445908273,.00332813286160627,.00824210667082646,.0150174405185659,.0212996498258157,.0237895578215264,.0200145017350098,.0100011552359343,3.47978143111129e-006,.000208597251185043,.001726098112448,.00662661740611046,.0164107894122825,.0299010997678524,.0424095539898868,.0473671889011406,.0398507064131546,.0199132162457208,4.00637159776859e-006,.000240163964049163,.00198730598159011,.00762941359700327,.0188942098519485,.0344259888799142,.0488273289408453,.0545351953956101,.0458812547527418,.0229266537472159,3.14102354559662e-006,.000188290239053809,.00155806188421368,.00598151398652822,.0148131935773177,.0269902177103302,.0382809697326146,.0427559772280557,.0359712268228302,.017974657987755,1.18417845234822e-006,7.09861739774591e-005,.000587395568332318,.00225505472099886,.00558463328595991,.0101754201370336,.0144320788540114,.016119174596298,.0135612965295474,.0067765180261918,-1.18417845234822e-006,-7.09861739774591e-005,-.000587395568332318,-.00225505472099886,-.00558463328595991,-.0101754201370336,-.0144320788540114,-.016119174596298,-.0135612965295474,-.0067765180261918,-3.14102354559662e-006,-.000188290239053809,-.00155806188421368,-.00598151398652822,-.0148131935773177,-.0269902177103302,-.0382809697326146,-.0427559772280557,-.0359712268228302,-.017974657987755,-4.00637159776859e-006,-.000240163964049163,-.00198730598159011,-.00762941359700327,-.0188942098519485,-.0344259888799142,-.0488273289408453,-.0545351953956101,-.0458812547527418,-.0229266537472159,-3.47978143111129e-006,-.000208597251185043,-.001726098112448,-.00662661740611046,-.0164107894122825,-.0299010997678524,-.0424095539898868,-.0473671889011406,-.0398507064131546,-.0199132162457208,-1.74767520475977e-006,-.000104765270720098,-.000866910445908273,-.00332813286160627,-.00824210667082646,-.0150174405185659,-.0212996498258157,-.0237895578215264,-.0200145017350098,-.0100011552359343},
			{-2.1377040911701e-006,-.000128145748833051,-.00106037895476748,-.00407087267407408,-.0100814986114699,-.0183688845318685,-.0260530952485506,-.0290986763122241,-.0244811405031877,-.0122331142571899,-3.23272774874231e-006,-.000193787493716783,-.00160355050327985,-.00615614813548528,-.015245674293661,-.0277782144801489,-.0393986072714656,-.0440042654896674,-.0370214299314904,-.0184994396914433,-1.46467098517923e-006,-8.78004401849179e-005,-.00072653006314475,-.00278920226363585,-.00690744737044604,-.0125856391046194,-.0178505588506293,-.0199372715230704,-.0167735171239198,-.00838165000707885,2.04207081203481e-006,.000122412963730206,.0010129413711518,.00388875630709325,.00963048821447668,.0175471259596245,.0248875724147149,.0277969050120105,.0233859413346208,.0116858482275818,4.79303205216701e-006,.000287320721349026,.00237751816944035,.00912746684058028,.0226041322455455,.0411856124930787,.0584146894325829,.0652433088453238,.0548901466718698,.0274283559519402,4.79303205216701e-006,.000287320721349026,.00237751816944035,.00912746684058028,.0226041322455455,.0411856124930787,.0584146894325829,.0652433088453238,.0548901466718698,.0274283559519403,2.04207081203481e-006,.000122412963730205,.0010129413711518,.00388875630709324,.00963048821447665,.0175471259596245,.0248875724147148,.0277969050120105,.0233859413346208,.0116858482275817,-1.46467098517922e-006,-8.78004401849175e-005,-.000726530063144746,-.00278920226363584,-.00690744737044601,-.0125856391046194,-.0178505588506292,-.0199372715230703,-.0167735171239197,-.00838165000707881,-3.23272774874231e-006,-.000193787493716783,-.00160355050327985,-.00615614813548528,-.015245674293661,-.0277782144801489,-.0393986072714656,-.0440042654896674,-.0370214299314904,-.0184994396914433,-2.1377040911701e-006,-.000128145748833051,-.00106037895476748,-.00407087267407408,-.0100814986114699,-.0183688845318685,-.0260530952485506,-.0290986763122241,-.0244811405031877,-.0122331142571899},
			{-7.0687420406234e-006,-.000396419472948955,-.00289468675708114,-.0091512705781711,-.0170446886407066,-.020343232859846,-.0143340443821725,-.00198780768336474,.00723004945575377,.00622084218798443,-1.58453826290029e-005,-.00088861896421812,-.00648876688291386,-.0205136052524374,-.0382075922069433,-.0456016511456594,-.032131377344278,-.00445589514439379,.0162069713952851,.0139447194673076,-2.32283443257473e-005,-.00130266007192951,-.00951212822907477,-.0300716680261107,-.0560099511902079,-.066849181205329,-.04710259853538,-.00653206483660142,.0237584109428924,.0204420904749686,-2.85486506407837e-005,-.00160102617628203,-.0116908214315209,-.0369593946354349,-.0688386785778866,-.0821605661208479,-.0578911527658522,-.00802819324390579,.0292001256859699,.0251242228526073,-3.13325682074437e-005,-.00175715001389931,-.0128308502042083,-.0405634846946589,-.0755514724317883,-.0901724419250166,-.0635364002126376,-.00881106135495838,.0320475717480869,.0275742078353698,-3.13325682074437e-005,-.00175715001389931,-.0128308502042083,-.0405634846946589,-.0755514724317883,-.0901724419250165,-.0635364002126376,-.00881106135495838,.0320475717480869,.0275742078353698,-2.85486506407837e-005,-.00160102617628203,-.0116908214315209,-.0369593946354349,-.0688386785778866,-.0821605661208479,-.0578911527658522,-.00802819324390579,.0292001256859699,.0251242228526073,-2.32283443257473e-005,-.00130266007192951,-.00951212822907477,-.0300716680261107,-.0560099511902079,-.066849181205329,-.04710259853538,-.00653206483660142,.0237584109428924,.0204420904749686,-1.58453826290029e-005,-.00088861896421812,-.00648876688291386,-.0205136052524374,-.0382075922069433,-.0456016511456594,-.032131377344278,-.00445589514439379,.0162069713952851,.0139447194673076,-7.0687420406234e-006,-.000396419472948955,-.00289468675708114,-.0091512705781711,-.0170446886407066,-.020343232859846,-.0143340443821725,-.00198780768336474,.00723004945575377,.00622084218798443},
			{1.19239470204654e-005,.000668702403646816,.00488291853542222,.0154368719238809,.0287519283012612,.0343160960538323,.0241794628831796,.00335314450111479,-.0121960493352718,-.0104936624148843,2.37416708289763e-005,.00133144774315747,.00972233811117492,.0307362261184839,.0572477231117576,.0683264908211426,.048143525596736,.00667641787160104,-.0242834514641005,-.0208938431558764,2.73344626884814e-005,.00153293375681807,.0111936051282609,.0353874936634631,.0659109362045824,.0786662374118139,.0554290139727572,.00768675029312828,-.027958230184157,-.0240556774743184,2.14304112375827e-005,.00120183086029444,.00877586524613968,.0277440442315677,.0516746381304762,.0616748841000227,.0434567372868267,.00602646636005635,-.0219194493467751,-.0188598205404102,8.07935083711268e-006,.000453095046079154,.00330853633356522,.0104596157536215,.0194815454639894,.0232516782323097,.0163833639440749,.00227200194579396,-.00826372016223111,-.00711022785245187,-8.07935083711268e-006,-.000453095046079154,-.00330853633356522,-.0104596157536215,-.0194815454639894,-.0232516782323097,-.0163833639440749,-.00227200194579396,.00826372016223111,.00711022785245187,-2.14304112375827e-005,-.00120183086029444,-.00877586524613968,-.0277440442315677,-.0516746381304761,-.0616748841000227,-.0434567372868267,-.00602646636005635,.0219194493467751,.0188598205404102,-2.73344626884814e-005,-.00153293375681807,-.0111936051282609,-.0353874936634631,-.0659109362045824,-.0786662374118139,-.0554290139727572,-.00768675029312828,.027958230184157,.0240556774743184,-2.37416708289763e-005,-.00133144774315747,-.00972233811117492,-.0307362261184839,-.0572477231117576,-.0683264908211426,-.048143525596736,-.00667641787160104,.0242834514641005,.0208938431558764,-1.19239470204654e-005,-.000668702403646816,-.00488291853542223,-.0154368719238809,-.0287519283012612,-.0343160960538323,-.0241794628831796,-.00335314450111479,.0121960493352718,.0104936624148843},
			{-2.71193324668268e-005,-.00139397384538797,-.00859091112473706,-.0204390993260263,-.0232648893283274,-.00816857426450726,.0104893149252651,.0116472227517443,-.00129246985153553,-.00610326187849476,-6.07910427499648e-005,-.00312474961288118,-.0192574963297445,-.0458165466432155,-.0521508736788253,-.0183107806184887,.0235129825861177,.0261085635896738,-.00289721696114192,-.0136811499406799,-8.91158835339698e-005,-.00458068837081261,-.0282302905567179,-.0671642046243376,-.076449933653419,-.026842464271009,.0344685684418851,.0382735285799644,-.00424713967062055,-.0200557139600161,-.000109527316707461,-.00562986625988991,-.0346962612267728,-.082547743674565,-.0939603105918888,-.0329905620505779,.0423633774639968,.0470398398136498,-.00521992032573585,-.0246493492246598,-.000120207857264269,-.0061788618595791,-.0380796621567155,-.0905973750422723,-.103122836783299,-.0362076319703515,.0464944361325487,.051626923036526,-.00572894010653844,-.0270530269738013,-.000120207857264269,-.0061788618595791,-.0380796621567155,-.0905973750422723,-.103122836783299,-.0362076319703515,.0464944361325487,.051626923036526,-.00572894010653844,-.0270530269738013,-.000109527316707461,-.00562986625988991,-.0346962612267728,-.082547743674565,-.0939603105918888,-.0329905620505779,.0423633774639968,.0470398398136498,-.00521992032573585,-.0246493492246598,-8.91158835339698e-005,-.00458068837081261,-.0282302905567179,-.0671642046243376,-.076449933653419,-.026842464271009,.0344685684418851,.0382735285799644,-.00424713967062055,-.0200557139600161,-6.07910427499648e-005,-.00312474961288118,-.0192574963297445,-.0458165466432156,-.0521508736788254,-.0183107806184887,.0235129825861177,.0261085635896738,-.00289721696114192,-.0136811499406799,-2.71193324668268e-005,-.00139397384538797,-.00859091112473705,-.0204390993260263,-.0232648893283274,-.00816857426450726,.0104893149252651,.0116472227517443,-.00129246985153553,-.00610326187849476},
			{1.8467923733974e-008,5.72496315380497e-006,.000112551238536802,.000763670899970611,.00284090951583314,.00698705136166929,.0123641411463411,.0161796260245334,.0151169390000152,.00799470122065349,4.1397934207577e-008,1.28331506776816e-005,.000252296296814152,.0017118544633724,.00636821913065334,.0156622637575245,.0277156170386288,.0362684567748484,.0338863239396712,.0179210245780364,6.06867939173738e-008,1.88125998409004e-005,.000369850661970342,.00250947205515881,.00933541273013333,.0229599034620059,.0406293688733935,.0531672993864887,.0496752409729615,.026271106183758,7.45867227452771e-008,2.3121507627533e-005,.000454562632178227,.00308425086139813,.0114736303579195,.0282187250839213,.0499352705236225,.065344935240019,.0610530427888428,.0322883379835085,8.18600363005434e-008,2.53761981227818e-005,.000498889241964133,.00338501114114201,.0125924797742609,.0309704700072926,.0548047012562974,.0717170372141512,.0670066214870632,.0354369306242455,8.18600363005434e-008,2.53761981227818e-005,.000498889241964133,.00338501114114201,.0125924797742609,.0309704700072926,.0548047012562974,.0717170372141512,.0670066214870632,.0354369306242455,7.45867227452771e-008,2.3121507627533e-005,.000454562632178227,.00308425086139813,.0114736303579195,.0282187250839213,.0499352705236225,.065344935240019,.0610530427888428,.0322883379835085,6.06867939173738e-008,1.88125998409004e-005,.000369850661970342,.00250947205515881,.00933541273013333,.0229599034620059,.0406293688733935,.0531672993864886,.0496752409729615,.026271106183758,4.1397934207577e-008,1.28331506776816e-005,.000252296296814152,.0017118544633724,.00636821913065334,.0156622637575245,.0277156170386288,.0362684567748484,.0338863239396712,.0179210245780364,1.8467923733974e-008,5.72496315380497e-006,.000112551238536802,.000763670899970611,.00284090951583315,.00698705136166929,.0123641411463411,.0161796260245334,.0151169390000152,.00799470122065349},
			{-3.11527203680034e-008,-9.65718609285329e-006,-.000189857685775429,-.0012882025257772,-.00479220409464544,-.0117861466401087,-.0208565205959732,-.0272926925875222,-.0255000930408489,-.013485906436498,-6.20279200448354e-008,-1.9228342172032e-005,-.000378024044579789,-.00256492923656619,-.00954171734955457,-.0234672976482734,-.041527243099746,-.0543422511302053,-.0507730212189817,-.0268516751119555,-7.14145132549108e-008,-2.21381386949937e-005,-.000435229862984357,-.00295307617650668,-.0109856512944789,-.0270185690209856,-.0478114992513544,-.0625657834542022,-.0584564272703957,-.0309150993103302,-5.59894812942294e-008,-1.73564566340746e-005,-.000341223277476823,-.00231523251800061,-.00861282797603967,-.0211827483777367,-.0374845521025614,-.0490520148175479,-.0458302506312781,-.0242376555640243,-2.11082586119828e-008,-6.54345364077286e-006,-.000128642541758133,-.000872851928740155,-.00324707062998819,-.0079859809473875,-.0141318262188787,-.0184928059748667,-.0172781879777248,-.00913769318749374,2.11082586119828e-008,6.54345364077286e-006,.000128642541758133,.000872851928740155,.00324707062998819,.00798598094738749,.0141318262188787,.0184928059748667,.0172781879777248,.00913769318749374,5.59894812942294e-008,1.73564566340746e-005,.000341223277476823,.00231523251800061,.00861282797603967,.0211827483777367,.0374845521025615,.0490520148175479,.0458302506312781,.0242376555640243,7.14145132549108e-008,2.21381386949937e-005,.000435229862984357,.00295307617650668,.0109856512944789,.0270185690209856,.0478114992513544,.0625657834542022,.0584564272703957,.0309150993103302,6.20279200448354e-008,1.9228342172032e-005,.000378024044579789,.00256492923656619,.00954171734955457,.0234672976482734,.041527243099746,.0543422511302053,.0507730212189817,.0268516751119555,3.11527203680034e-008,9.65718609285329e-006,.000189857685775429,.0012882025257772,.00479220409464544,.0117861466401087,.0208565205959732,.0272926925875222,.0255000930408489,.013485906436498},
			{1.62757657902233e-007,4.73190249691825e-005,.000825150543237864,.00465350306811798,.013244685304365,.0221077948264185,.0214226110182188,.00800738602387552,-.0066385606490133,-.00788880135076707,3.64839649040836e-007,.000106070931993143,.00184966801858007,.0104313520363045,.0296894560930798,.0495571158340357,.0480211990401099,.0179494589952879,-.0148810825158051,-.0176836380743178,5.34832208853932e-007,.000155493381824678,.00271149814617032,.0152917125799703,.0435228940266753,.0726476461525729,.0703960877661124,.0263127892689876,-.0218147404016795,-.0259231123501117,6.57332165731892e-007,.000191108163905862,.00333254975914684,.0187941832625992,.0534915394321128,.0892871330302339,.086519869342777,.0323395683174988,-.0268112696201347,-.0318606383488469,7.21431817457549e-007,.000209744049059384,.00365752287024036,.0206268953439552,.0587077592165956,.0979939549829493,.094956841974485,.0354931566768965,-.0294257667261143,-.0349675239211424,7.21431817457549e-007,.000209744049059384,.00365752287024036,.0206268953439552,.0587077592165956,.0979939549829493,.094956841974485,.0354931566768965,-.0294257667261143,-.0349675239211424,6.57332165731892e-007,.000191108163905862,.00333254975914684,.0187941832625992,.0534915394321128,.0892871330302339,.086519869342777,.0323395683174988,-.0268112696201347,-.0318606383488469,5.34832208853932e-007,.000155493381824678,.00271149814617032,.0152917125799703,.0435228940266753,.0726476461525729,.0703960877661124,.0263127892689876,-.0218147404016795,-.0259231123501117,3.64839649040836e-007,.000106070931993143,.00184966801858007,.0104313520363045,.0296894560930798,.0495571158340357,.0480211990401099,.0179494589952879,-.0148810825158051,-.0176836380743178,1.62757657902233e-007,4.73190249691825e-005,.000825150543237864,.00465350306811798,.013244685304365,.0221077948264185,.0214226110182188,.00800738602387552,-.0066385606490133,-.00788880135076707},
			{-3.0545495487195e-010,-4.89665844207368e-007,-2.28716807432812e-005,-.000274273335410219,-.00153266885205138,-.00508819457391411,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594,-6.84711736268708e-010,-1.09764122346374e-006,-5.12694522820253e-005,-.000614814618998954,-.0034356501149343,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-1.00374477198287e-009,-1.60907368927037e-006,-7.51578247671699e-005,-.000901279950190104,-.00503644914868864,-.0167201370314615,-.0369150385504742,-.0565976592019674,-.058726053964138,-.0328701880528208,-1.23364620508441e-009,-1.97762190735818e-006,-9.237225228407e-005,-.00110771245968658,-.00619001618021515,-.0205497793593547,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.35394503757885e-009,-2.17046942359914e-006,-.000101379919197693,-.001215730881087,-.00679363471892789,-.0225536881419213,-.0497944643424746,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.35394503757885e-009,-2.17046942359914e-006,-.000101379919197693,-.001215730881087,-.00679363471892789,-.0225536881419213,-.0497944643424746,-.0763442280886821,-.0792152064555801,-.0443383908346372,-1.23364620508441e-009,-1.97762190735818e-006,-9.237225228407e-005,-.00110771245968658,-.00619001618021515,-.0205497793593547,-.0453701961788293,-.0695609974169401,-.0721768876258489,-.0403988980900667,-1.00374477198287e-009,-1.60907368927037e-006,-7.51578247671699e-005,-.000901279950190104,-.00503644914868864,-.0167201370314615,-.0369150385504742,-.0565976592019673,-.058726053964138,-.0328701880528208,-6.84711736268708e-010,-1.09764122346374e-006,-5.12694522820253e-005,-.000614814618998954,-.0034356501149343,-.0114057620791853,-.0251818598172014,-.0386085014663319,-.0400604013055631,-.0224226358745209,-3.0545495487195e-010,-4.89665844207368e-007,-2.28716807432812e-005,-.000274273335410219,-.00153266885205138,-.00508819457391411,-.0112338133649814,-.0172235372177759,-.0178712404429137,-.010002903216594}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_right_fluxes_to_dofs[j][0];
			for (int i = 1; i < 100; i++)
			{
				std::cout << ", " << timestep_right_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_left:" << std::endl;
		static const T adaptivity_coarsen_left[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,-8.500000000,0.,33.98823325,-.4082482906,0.,0.,0.,1.750000000,0.,0.,-7.678758582,0.,34.53410150,.3227486122,0.,0.,-2.459674775,0.,14.75000000,-.2645751311,0.,2.958039892,.2236067977},
			{0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,-8.500000000,0.,0.,-.4082482906,0.,0.,0.,1.750000000,0.,0.,-7.678758582,0.,0.,.3227486122,0.,0.,-2.459674775,0.,0.,-.2645751311,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,2.309401077,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,1.750000000,0.,0.,0.,0.,0.,.3227486122,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,-12.37436867,0.,55.33985905,-.7216878367,0.,0.,0.,2.828427124,0.,0.,-13.26806944,0.,63.17066828,.9128709293,0.,0.,-5.850213671,0.,33.43604921,-.9354143468,0.,8.844691708,.9035079028},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,-12.37436867,0.,0.,-.7216878367,0.,0.,0.,2.828427124,0.,0.,-13.26806944,0.,0.,.9128709295,0.,0.,-5.850213671,0.,0.,-.9354143468,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,2.857738034,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,2.828427124,0.,0.,0.,0.,0.,.9128709292,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,-7.216878367,0.,40.02082790,0.,0.,0.,0.,1.587713241,0.,0.,-9.500000000,0.,52.88300766,.9316949905,0.,0.,-6.067673910,0.,36.47616523,-1.636634177,0.,13.68699853,2.074812507},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,-7.216878367,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,-9.500000000,0.,0.,.9316949905,0.,0.,-6.067673910,0.,0.,-1.636634177,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.166666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.587713241,0.,0.,0.,0.,0.,.9316949905,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,15.87608264,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,23.88339730,0.,0.,0.,-2.627379873,0.,21.21428571,-1.218935425,0.,11.40958244,2.747169230},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-2.901185103,0.,0.,0.,0.,0.,-2.627379873,0.,0.,-1.218935425,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.100000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.080682370,0.,0.,0.,0.,0.,5.494338457,0.,0.,4.138710979,1.621428571},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,2.245365598,0.,-7.850689990,-.2500000000,0.,0.,0.,0.,0.,0.,.4949747467,0.,-3.162277660,.3162277660,0.,0.,-1.004158022,0.,3.091022771,-.3240370350,0.,2.070196679,.3129843186},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,2.245365598,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,.4949747467,0.,0.,.3162277660,0.,0.,-1.004158022,0.,0.,-.3240370349,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3162277659,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,3.500000000,0.,-14.98165545,0.,0.,0.,0.,-.2500000000,0.,0.,1.558845727,0.,-8.714212528,.5809475019,0.,0.,-1.565247584,0.,4.964285714,-1.020504077,0.,5.704791219,1.293725044},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,3.500000000,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,1.558845727,0.,0.,.5809475019,0.,0.,-1.565247584,0.,0.,-1.020504077,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8660254040,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,.5809475019,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,-10.71525393,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,-8.720665110,0.,0.,0.,-.5486726408,0.,1.773290113,-.9983468483,0.,5.709015429,2.250018896},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.659882024,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.275000000,0.,0.,0.,0.,0.,-.5486726408,0.,0.,-.9983468480,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.117691454,0.,0.,0.,0.,0.,0.,0.,0.,0.,-3.300000000,0.,0.,0.,0.,0.,-.4979550017,0.,0.,1.996693696,1.554721797},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,-.6454972245,0.,2.020725943,0.,0.,0.,0.,-.3227486122,0.,0.,.2236067977,0.,-.2500000000,.8333333333e-1,0.,0.,.2886751347,0.,-1.152673615,-.1463850110,0.,.1636634177,.1855768722},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,-.6454972245,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,.2236067977,0.,0.,.8333333333e-1,0.,0.,.2886751347,0.,0.,-.1463850109,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3227486122,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,3.500000000,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,.2886751347,0.,0.,0.,.6250000000,0.,-2.395787117,-.2746751327,0.,-.1889822365,.6190476190},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8385254914,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365e-1,0.,0.,0.,0.,0.,.6250000000,0.,0.,-.2746751327,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.788854382,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6454972245,0.,0.,0.,0.,0.,-1.357142857,0.,0.,-.6338656908,.6016087651},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,-.5916079782,0.,0.,0.,0.,0.,0.,0.,-.3436931772,0.,.3415650256,0.,0.,0.,.1479019946,0.,.1889822365,-.2500000000e-1,0.,-.2236067977,.5634361697e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3436931771,0.,0.,0.,0.,0.,.1479019946,0.,0.,-.2500000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7937253933,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,.5070925528,0.,0.,-.4500000000,.1070899340},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3464101616,0.,0.,0.,0.,0.,.1916629694,0.,0.,-.5669467095e-1,.7142857143e-2}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << adaptivity_coarsen_left[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << adaptivity_coarsen_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_coarsen_right:" << std::endl;
		static const T adaptivity_coarsen_right[35][35] = 
		{
			{.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,-.1154700539,0.,-.6454972245e-1,2.904737510,0.,0.,1.341640786,0.,.7500000000,-11.64130577,0.,-6.507687761,50.31152948},
			{0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,-.1154700539,0.,0.,2.904737510,0.,0.,1.341640786,0.,0.,-11.64130577,0.,0.},
			{0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,2.904737510,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8164965809,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,.1632993162,0.,.9128709293e-1,-2.464751509,0.,0.,-1.581138830,0.,-.9596449171,10.85080642,0.,7.410417377,-50.14468862},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,.1632993162,0.,0.,-2.464751509,0.,0.,-1.581138830,0.,0.,10.85080642,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,.3535533905,0.,0.,0.,0.,0.,-2.464751508,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.2000000000,0.,-.1118033988,1.155301788,0.,0.,1.290994449,0.,.9691236663,-6.197388082,0.,-6.026182953,33.05867928},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.2000000000,0.,0.,1.155301788,0.,0.,1.290994449,0.,0.,-6.197388081,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1666666667,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,1.155301788,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,.1290994449,0.,0.,0.,-.7267220925,0.,-.7857142857,1.974864371,0.,3.507390157,-14.50782199},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3031088914,0.,0.,0.,0.,0.,-.7267220925,0.,0.,1.974864371,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2309401077,0.,0.,0.,0.,0.,.5111012519,0.,0.,-1.266180985,3.502380952},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,.2828427124,0.,.1581138830,3.320391543,0.,0.,-1.278019301,0.,-1.195584279,-15.76980237,0.,1.794170455,75.89869728},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,.2828427124,0.,0.,3.320391543,0.,0.,-1.278019301,0.,0.,-15.76980236,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,.6123724358,0.,0.,0.,0.,0.,3.320391542,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-.3464101616,0.,-.1936491673,-1.484643616,0.,0.,.4472135954,0.,1.107142857,8.919961562,0.,-.8451542550e-1,-49.83237207},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,-.3464101616,0.,0.,-1.484643616,0.,0.,.4472135954,0.,0.,8.919961562,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2886751347,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,-1.484643616,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,.2236067977,0.,0.,0.,.3227486122e-1,0.,-.7010681843,-2.776822654,0.,-.7563225568,21.71637090},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3250000000,0.,0.,0.,0.,0.,.3227486122e-1,0.,0.,-2.776822653,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1732050808,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3000000000,0.,0.,0.,0.,0.,.2766416676,0.,0.,.5237229368,-5.171408842},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,-.4472135954,0.,-.2500000000,1.083333333,0.,0.,-4.041451885,0.,-.4610694458e-1,-8.002380602,0.,16.80277755,48.86857636},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,-.4472135954,0.,0.,1.083333333,0.,0.,-4.041451885,0.,0.,-8.002380594,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3726779962,0.,0.,0.,0.,0.,0.,0.,0.,0.,.9682458365,0.,0.,0.,0.,0.,1.083333333,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,.2886751347,0.,0.,0.,1.375000000,0.,.7985957060,2.387560769,0.,-7.275816105,-21.01190476},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2795084971,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9682458365e-1,0.,0.,0.,0.,0.,1.375000000,0.,0.,2.387560769,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1290994449,0.,0.,0.,0.,0.,-.6428571429,0.,0.,1.774823935,4.876757779},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079532,0.,.3415650256,0.,0.,0.,-1.922725929,0.,3.968626966,-1.575000000,0.,13.75181806,16.26921940},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3307189139,0.,0.,0.,0.,0.,0.,0.,0.,0.,-1.031079531,0.,0.,0.,0.,0.,-1.922725929,0.,0.,-1.575000000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2645751311,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3055050464,0.,0.,0.,0.,0.,-.8451542550,0.,0.,-3.450000000,-3.647357166},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3000000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,1.039230485,0.,0.,0.,0.,0.,2.491618602,0.,0.,3.571764270,2.292857143}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << adaptivity_coarsen_right[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << adaptivity_coarsen_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_left:" << std::endl;
		static const T adaptivity_refine_left[35][35] = 
		{
			{1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,.1250000000,0.,-.5590169942e-1,.4082482906,0.,0.,0.,.1250000000,0.,0.,-.2453738645,0.,.3227486122e-1,-.1613743061,0.,0.,.5590169942e-1,0.,.1250000000,.6614378278e-1,0.,-.7395099729e-1,-.2795084971e-1},
			{0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,.1250000000,0.,0.,.4082482906,0.,0.,0.,.1250000000,0.,0.,-.2453738645,0.,0.,-.1613743061,0.,0.,.5590169942e-1,0.,0.,.6614378278e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,.1250000000,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,-.2651650429,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,.3163924252,0.,.9128709293e-1,.3651483717,0.,0.,.3952847075e-1,0.,-.2525381361,-.1403121520,0.,.8964214566e-1,.5646924394e-1},
			{0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,-.2651650429,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,.3163924252,0.,0.,.3651483718,0.,0.,.3952847075e-1,0.,0.,-.1403121520,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,.4082482906,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,-.7071067810,0.,0.,0.,0.,0.,.3651483716,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,0.,.2420614591,0.,0.,0.,0.,-.2165063510,0.,0.,.4250000000,0.,-.3913118960,.5776508940,0.,0.,-.6132223630,0.,.3608439183,.3218713882,0.,-.3049687729e-2,-.1210307296},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,-.2165063510,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,.4250000000,0.,0.,.5776508940,0.,0.,-.6132223630,0.,0.,.3218713881,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8333333333e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2165063510,0.,0.,0.,0.,0.,.5776508940,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,.1118033988,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,-.2259240285,0.,0.,0.,-.1816805231,0.,.3750000000,.4937160930,0.,-.5387858374,.2888254471},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3125000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7577722285e-1,0.,0.,0.,0.,0.,-.1816805231,0.,0.,.4937160930,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1250000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2886751347e-1,0.,0.,0.,0.,0.,.6388765648e-1,0.,0.,-.1582726231,.4377976190},
			{0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,-.4592793268,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,.1237436867,0.,.1581138830,-.3162277660,0.,0.,-.1141088661,0.,.2916059217e-1,.1890216038,0.,-.1725163898e-1,-.9780759959e-1},
			{0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,-.4592793268,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,.1237436867,0.,0.,-.3162277660,0.,0.,-.1141088661,0.,0.,.1890216037,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3162277659,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,-.3750000000,0.,.4192627457,0.,0.,0.,0.,-.1250000000,0.,0.,.3897114318,0.,-.4357106264,-.7423218081,0.,0.,.2795084971,0.,.8928571429e-2,-.3496171374,0.,-.8451542550e-1,.1796840339},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,-.3750000000,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,.3897114318,0.,0.,-.7423218081,0.,0.,.2795084971,0.,0.,-.3496171374,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1443375673,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,-.7423218081,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,.1936491673,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,-.2795084971,0.,0.,0.,.8068715303e-2,0.,.1340277411,-.6942056633,0.,.3964594048,-.3527181261},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.5412658775e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8125000000e-1,0.,0.,0.,0.,0.,.8068715303e-2,0.,0.,-.6942056631,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2165063510e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3750000000e-1,0.,0.,0.,0.,0.,.3458020845e-1,0.,0.,.6546536711e-1,-.6464261051},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,-.4841229182,0.,.5412658775,0.,0.,0.,0.,.4841229182,0.,0.,-.3913118960,0.,.6250000000e-1,.5416666667,0.,0.,.3608439183,0.,-.1959545145,.1341862601,0.,.9547032698e-1,-.1546473936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,-.4841229182,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,-.3913118960,0.,0.,.5416666667,0.,0.,.3608439183,0.,0.,.1341862600,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4841229182,0.,0.,0.,0.,0.,.5416666667,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,.2500000000,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,-.7216878367e-1,0.,0.,0.,.3437500000,0.,-.5190872089,.5968901920,0.,.1653594569,.1994047619},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6987712428e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2420614591e-1,0.,0.,0.,0.,0.,.3437500000,0.,0.,.5968901920,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2795084971e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061e-1,0.,0.,0.,0.,0.,-.8035714286e-1,0.,0.,.2218529919,.6095947223},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,-.2577698829,0.,.4269562821,0.,0.,0.,-.4806814823,0.,.4724555913e-1,-.3937500000,0.,-.4192627457,-.1056442818e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2577698828,0.,0.,0.,0.,0.,-.4806814823,0.,0.,-.3937500000,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3307189139e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,0.,0.,0.,0.,-.1056442819,0.,0.,-.4312500000,-.4559196456},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1299038106,0.,0.,0.,0.,0.,.3114523254,0.,0.,.4464705337,.2866071429}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << adaptivity_refine_left[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << adaptivity_refine_left[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "adaptivity_refine_right:" << std::endl;
		static const T adaptivity_refine_right[35][35] = 
		{
			{1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,-.1250000000,0.,.5590169942e-1,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,.2453738645,0.,-.3227486122e-1,.1613743061,0.,0.,-.5590169942e-1,0.,-.1250000000,-.6614378278e-1,0.,.7395099729e-1,.2795084971e-1},
			{0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,-.1250000000,0.,0.,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,.2453738645,0.,0.,.1613743061,0.,0.,-.5590169942e-1,0.,0.,-.6614378278e-1,0.,0.},
			{0.,0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,-.1443375673,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.1613743061,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,1.,0.,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,1.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,.2651650429,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,.5103103630e-1,0.,-.9128709293e-1,.4564354646,0.,0.,.1185854122,0.,-.1515228816,-.2338535867,0.,.5976143047e-1,.1129384878},
			{0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,.2651650429,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,.5103103630e-1,0.,0.,.4564354646,0.,0.,.1185854122,0.,0.,-.2338535867,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,-.4082482906,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,-.3535533905,0.,0.,0.,0.,0.,.4564354645,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,-.7500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7216878367,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,.2165063510,0.,-.2420614591,0.,0.,0.,0.,.7938566203,0.,0.,.3750000000,0.,-.1677050983,.4658474951,0.,0.,.9682458365e-1,0.,.3092947872e-1,-.4091585443,0.,-.1189378215,.2593515633},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,.2165063510,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,.3750000000,0.,0.,.4658474951,0.,0.,.9682458365e-1,0.,0.,-.4091585442,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5833333333,0.,0.,0.,0.,0.,0.,0.,0.,0.,.7938566203,0.,0.,0.,0.,0.,.4658474951,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,-.1118033988,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,-.2581988897,0.,0.,0.,-.6568449682,0.,-.2857142857,-.3047338563,0.,.6338656908e-1,.3433961537},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4687500000,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7252962758,0.,0.,0.,0.,0.,-.6568449682,0.,0.,-.3047338563,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3875000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,.6350852963,0.,0.,0.,0.,0.,.6867923071,0.,0.,.5173388725,.2026785714},
			{0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,.2551551816,0.,-.1825741858,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,-.4065863991,0.,.1581138830,.1581138830,0.,0.,.2966830521,0.,.2916059217e-1,-.8100925875e-1,0.,-.1035098339,.3912303983e-1},
			{0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,.2551551816,0.,0.,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,-.4065863991,0.,0.,.1581138830,0.,0.,.2966830520,0.,0.,-.8100925872e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,.7071067810,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,-.6123724358,0.,0.,0.,0.,0.,.1581138830,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2500000000,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.6250000000,0.,-.5590169942e-1,0.,0.,0.,0.,-.1250000000,0.,0.,-.4330127020e-1,0.,.4841229182e-1,.2904737510,0.,0.,.6149186937,0.,.2767857143,-.2551260193,0.,-.4278593414,.1617156305},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,-.6250000000,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,-.4330127020e-1,0.,0.,.2904737510,0.,0.,.6149186937,0.,0.,-.2551260193,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1250000000,0.,0.,0.,0.,0.,.2904737510,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,.5163977793,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,.3354101966,0.,0.,0.,-.1371681602,0.,-.3917733971,-.2495867120,0.,-.4025587803,.2812523620},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4149705060,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3187500000,0.,0.,0.,0.,0.,-.1371681602,0.,0.,-.2495867119,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.3897114318,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4125000000,0.,0.,0.,0.,0.,-.6224437520e-1,0.,0.,.2495867120,.1943402247},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,0.,.4330127020,0.,0.,0.,0.,-.1613743061,0.,0.,-.5031152948,0.,-.5625000000,.4166666667e-1,0.,0.,.2165063510,0.,.3803822929,-.3659625275e-1,0.,-.1432054905,.2319710903e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,.4841229182,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,-.5031152948,0.,0.,.4166666667e-1,0.,0.,.2165063510,0.,0.,-.3659625271e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1863389981,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1613743061,0.,0.,0.,0.,0.,.4166666667e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,-.5000000000,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,.1443375673,0.,0.,0.,.1562500000,0.,.3992978531,-.6866878320e-1,0.,-.3307189139,.7738095238e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2096313728,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2420614591e-1,0.,0.,0.,0.,0.,.1562500000,0.,0.,-.6866878320e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2236067977,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8068715303e-1,0.,0.,0.,0.,0.,-.1696428571,0.,0.,-.7923321136e-1,.7520109565e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,.2958039891,0.,0.,0.,0.,0.,0.,0.,-.8592329430e-1,0.,-.3415650256,0.,0.,0.,.3697549864e-1,0.,.1889822365,-.6250000000e-2,0.,-.5590169942e-1,.7042952119e-2},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8267972847e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.8592329428e-1,0.,0.,0.,0.,0.,.3697549864e-1,0.,0.,-.6250000000e-2,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.9921567416e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3818813080e-1,0.,0.,0.,0.,0.,.6338656908e-1,0.,0.,-.5625000000e-1,.1338624175e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3750000000e-1,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.4330127020e-1,0.,0.,0.,0.,0.,.2395787117e-1,0.,0.,-.7086833870e-2,.8928571429e-3}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << adaptivity_refine_right[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << adaptivity_refine_right[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_poly_to_dofs:" << std::endl;
		static const T convert_poly_to_dofs[35][35] = 
		{
			{.7071067810,.2357022604,.1178511302,.7071067810e-1,.4714045206e-1,.2357022604,.5892556509e-1,.2357022604e-1,.1178511302e-1,.1178511302,.2357022604e-1,.7856742012e-2,.7071067810e-1,.1178511302e-1,.4714045206e-1,.3535533905,.1178511302,.5892556509e-1,.3535533905e-1,.1178511302,.2946278254e-1,.1178511302e-1,.5892556509e-1,.1178511302e-1,.3535533905e-1,.2357022604,.7856742012e-1,.3928371006e-1,.7856742012e-1,.1964185503e-1,.3928371006e-1,.1767766952,.5892556509e-1,.5892556509e-1,.1414213562},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2041241452,.6804138176e-1,.3402069088e-1,.2041241452e-1,.6804138176e-1,.1701034544e-1,.6804138176e-2,.3402069088e-1,.6804138176e-2,.2041241452e-1,.2041241452,.6804138176e-1,.3402069088e-1,.6804138176e-1,.1701034544e-1,.3402069088e-1,.1837117307,.6123724358e-1,.6123724358e-1,.1632993162},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.5270462768e-1,.1756820922e-1,.8784104612e-2,.1756820922e-1,.4392052306e-2,.8784104612e-2,.7905694150e-1,.2635231383e-1,.2635231383e-1,.9035079028e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1336306209e-1,.4454354030e-2,.4454354030e-2,.2672612419e-1},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3367175148e-2},
			{0.,-.8333333333e-1,-.6666666667e-1,-.5000000000e-1,-.3809523810e-1,.1666666667,.1666666667e-1,0.,-.2380952381e-2,.1333333333,.1666666667e-1,.3174603175e-2,.1000000000,.1190476190e-1,.7619047619e-1,0.,-.4166666667e-1,-.3333333333e-1,-.2500000000e-1,.8333333333e-1,.8333333333e-2,0.,.6666666667e-1,.8333333333e-2,.5000000000e-1,0.,-.2777777778e-1,-.2222222222e-1,.5555555556e-1,.5555555556e-2,.4444444444e-1,0.,-.2083333333e-1,.4166666667e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.2405626122e-1,-.1924500898e-1,-.1443375673e-1,.4811252245e-1,.4811252245e-2,0.,.3849001795e-1,.4811252245e-2,.2886751347e-1,0.,-.2405626122e-1,-.1924500898e-1,.4811252245e-1,.4811252245e-2,.3849001795e-1,0.,-.2165063510e-1,.4330127020e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6211299937e-2,-.4969039948e-2,.1242259987e-1,.1242259987e-2,.9938079897e-2,0.,-.9316949905e-2,.1863389981e-1,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1574851971e-2,.3149703940e-2,0.},
			{0.,0.,.1360827635e-1,.1749635531e-1,.1749635531e-1,0.,-.2041241452e-1,-.9720197392e-2,-.4374088827e-2,.4082482906e-1,-.5832118436e-2,-.3888078956e-2,.5248906592e-1,0.,.5248906592e-1,0.,0.,.6804138176e-2,.8748177653e-2,0.,-.1020620726e-1,-.4860098696e-2,.2041241452e-1,-.2916059217e-2,.2624453295e-1,0.,0.,.4536092117e-2,0.,-.6804138176e-2,.1360827635e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3928371006e-2,.5050762721e-2,0.,-.5892556509e-2,-.2805979290e-2,.1178511302e-1,-.1683587573e-2,.1515228816e-1,0.,0.,.3928371006e-2,0.,-.5892556509e-2,.1178511302e-1,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1014301032e-2,0.,-.1521451549e-2,.3042903097e-2,0.,0.,0.,0.},
			{0.,0.,0.,-.2525381361e-2,-.4489566864e-2,0.,0.,.3367175148e-2,.3086577219e-2,0.,-.5050762721e-2,-.7482611439e-3,.1010152544e-1,-.3928371006e-2,.1795826746e-1,0.,0.,0.,-.1262690680e-2,0.,0.,.1683587573e-2,0.,-.2525381361e-2,.5050762721e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.7290148044e-3,0.,0.,.9720197392e-3,0.,-.1458029609e-2,.2916059217e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.5019488348e-3,0.,0.,0.,-.6274360436e-3,0.,0.,.8365813917e-3,0.,-.1254872087e-2,.2509744174e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,.1443375673,.1154700539,.8660254040e-1,.6598288793e-1,0.,.2886751347e-1,.1924500898e-1,.1237179149e-1,0.,.9622504490e-2,.5498573994e-2,0.,.4123930495e-2,0.,0.,.7216878367e-1,.5773502693e-1,.4330127020e-1,0.,.1443375673e-1,.9622504490e-2,0.,.4811252245e-2,0.,0.,.4811252245e-1,.3849001795e-1,0.,.9622504490e-2,0.,0.,.3608439183e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.4166666667e-1,.3333333333e-1,.2500000000e-1,0.,.8333333333e-2,.5555555556e-2,0.,.2777777778e-2,0.,0.,.4166666667e-1,.3333333333e-1,0.,.8333333333e-2,0.,0.,.3750000000e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1075828707e-1,.8606629657e-2,0.,.2151657415e-2,0.,0.,.1613743061e-1,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2727723628e-2,0.,0.},
			{0.,0.,-.2357022604e-1,-.3030457633e-1,-.3030457633e-1,0.,.2357022604e-1,.1010152544e-1,.3788072042e-2,0.,.1346870059e-1,.5892556509e-2,0.,.7576144082e-2,0.,0.,0.,-.1178511302e-1,-.1515228816e-1,0.,.1178511302e-1,.5050762721e-2,0.,.6734350295e-2,0.,0.,0.,-.7856742012e-2,0.,.7856742012e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.6804138176e-2,-.8748177653e-2,0.,.6804138176e-2,.2916059217e-2,0.,.3888078956e-2,0.,0.,0.,-.6804138176e-2,0.,.6804138176e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1756820922e-2,0.,.1756820922e-2,0.,0.,0.,0.,0.},
			{0.,0.,0.,.4374088827e-2,.7776157915e-2,0.,0.,-.4860098696e-2,-.4374088827e-2,0.,.4860098696e-2,0.,0.,.4860098696e-2,0.,0.,0.,0.,.2187044413e-2,0.,0.,-.2430049348e-2,0.,.2430049348e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1262690680e-2,0.,0.,-.1402989645e-2,0.,.1402989645e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.8694008848e-3,0.,0.,0.,.9780759957e-3,0.,0.,-.1086751106e-2,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,.3042903097e-1,.3912303982e-1,.3912303982e-1,0.,0.,.4347004424e-2,.4890379978e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,.1521451549e-1,.1956151991e-1,0.,0.,.2173502212e-2,0.,0.,0.,0.,0.,.1014301032e-1,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.8784104612e-2,.1129384878e-1,0.,0.,.1254872087e-2,0.,0.,0.,0.,0.,.8784104612e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.2268046058e-2,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,-.5646924394e-2,-.1003897670e-1,0.,0.,.3764616260e-2,.3137180218e-2,0.,0.,.1673162783e-2,0.,0.,0.,0.,0.,0.,-.2823462197e-2,0.,0.,.1882308131e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1630126659e-2,0.,0.,.1086751106e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1122391716e-2,0.,0.,0.,-.9820927513e-3,0.,0.,.6547285010e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,.6681531049e-2,.1187827742e-1,0.,0.,0.,.7423923386e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,.3340765524e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,.1928791874e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,-.1328031788e-2,0.,0.,0.,.6640158940e-3,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,.1505846505e-2,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << convert_poly_to_dofs[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << convert_poly_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_poly:" << std::endl;
		static const T convert_dofs_to_poly[35][35] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-2.,3.464101616,-4.472135954,5.291502622,2.449489743,-4.242640686,5.477225575,-2.828427124,4.898979486,3.162277660,-3.464101616,6.,-7.745966692,9.165151390,4.242640686,-7.348469229,9.486832980,-4.898979486,8.485281372,5.477225575,5.477225575,-9.486832980,12.24744872,-6.324555320,10.95445115,7.071067810,-7.483314774,12.96148140,8.366600265,9.486832980},
			{0.,4.898979486,-18.97366596,44.89988864,-84.85281372,0.,-6.928203232,26.83281572,-63.49803146,0.,8.485281372,-32.86335345,0.,-9.797958972,0.,0.,-12.,46.47580015,-109.9818167,0.,14.69693846,-56.92099788,0.,-16.97056274,0.,0.,18.97366596,-73.48469229,0.,-21.90890230,0.,0.,-25.92296279,0.,0.},
			{0.,0.,18.97366596,-112.2497216,381.8376617,0.,0.,-26.83281572,158.7450787,0.,0.,32.86335345,0.,0.,0.,0.,0.,-46.47580015,274.9545417,0.,0.,56.92099788,0.,0.,0.,0.,0.,73.48469229,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,74.83314774,-593.9696960,0.,0.,0.,-105.8300524,0.,0.,0.,0.,0.,0.,0.,0.,0.,-183.3030278,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,296.9848480,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,6.,-10.39230485,13.41640786,-15.87450787,-19.59591794,33.94112549,-43.81780460,42.42640686,-73.48469229,-75.89466384,3.464101616,-6.,7.745966692,-9.165151390,-25.45584412,44.09081537,-56.92099788,63.68673332,-110.3086578,-120.4989626,-10.95445115,18.97366596,-24.49489743,56.92099788,-98.59006035,-127.2792206,22.44994432,-38.88444419,-100.3992032,-37.94733192},
			{0.,0.,0.,0.,0.,0.,20.78460970,-80.49844717,190.4940944,0.,-67.88225098,262.9068276,0.,146.9693846,0.,0.,12.,-46.47580015,109.9818167,0.,-88.18163075,341.5259873,0.,220.6173157,0.,0.,-37.94733192,146.9693846,0.,197.1801207,0.,0.,77.76888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,80.49844717,-476.2352360,0.,0.,-262.9068276,0.,0.,0.,0.,0.,46.47580015,-274.9545417,0.,0.,-341.5259873,0.,0.,0.,0.,0.,-146.9693846,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,317.4901573,0.,0.,0.,0.,0.,0.,0.,0.,0.,183.3030278,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,24.49489743,-42.42640686,54.77225575,-127.2792206,220.4540769,398.4469852,0.,0.,0.,0.,21.21320343,-36.74234614,47.43416490,-161.6663230,280.0142853,575.1086854,5.477225575,-9.486832980,12.24744872,-94.86832980,164.3167672,487.9036789,-22.44994432,38.88444419,250.9980080,56.92099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,84.85281372,-328.6335345,0.,-440.9081537,0.,0.,0.,0.,0.,0.,73.48469229,-284.6049894,0.,-560.0285706,0.,0.,18.97366596,-73.48469229,0.,-328.6335345,0.,0.,-77.76888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,328.6335345,0.,0.,0.,0.,0.,0.,0.,0.,0.,284.6049894,0.,0.,0.,0.,0.,73.48469229,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,98.99494934,-171.4642820,-708.3501958,0.,0.,0.,0.,0.,0.,0.,102.8785692,-178.1909088,-920.1738966,0.,0.,0.,44.27188724,-76.68115805,-622.2539673,7.483314774,-12.96148140,-234.2648074,-37.94733192},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,342.9285640,0.,0.,0.,0.,0.,0.,0.,0.,0.,356.3818176,0.,0.,0.,0.,0.,153.3623161,0.,0.,25.92296279,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,398.4469852,0.,0.,0.,0.,0.,0.,0.,0.,0.,460.0869483,0.,0.,0.,0.,0.,254.5584412,0.,0.,75.29940238,9.486832980},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,6.928203232,-12.,15.49193338,-18.33030278,-8.485281372,14.69693846,-18.97366596,9.797958972,-16.97056274,-10.95445115,-32.86335345,56.92099788,-73.48469229,37.94733192,-65.72670690,-42.42640686,89.79977729,-155.5377768,-100.3992032,-189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,24.,-92.95160030,219.9636334,0.,-29.39387692,113.8419958,0.,33.94112549,0.,0.,-113.8419958,440.9081537,0.,131.4534138,0.,0.,311.0755535,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,92.95160030,-549.9090834,0.,0.,-113.8419958,0.,0.,0.,0.,0.,-440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,366.6060556,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,42.42640686,-73.48469229,94.86832980,-117.5755077,203.6467529,230.0434742,32.86335345,-56.92099788,73.48469229,-303.5786554,525.8136552,721.2489166,-179.5995546,311.0755535,1104.391235,569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,146.9693846,-569.2099788,0.,-407.2935059,0.,0.,113.8419958,-440.9081537,0.,-1051.627310,0.,0.,-622.1511070,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,569.2099788,0.,0.,0.,0.,0.,440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,205.7571384,-356.3818176,-920.1738966,0.,0.,0.,265.6313234,-460.0869483,-2206.173157,89.79977729,-155.5377768,-1907.584860,-569.2099788},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,712.7636352,0.,0.,0.,0.,0.,920.1738966,0.,0.,311.0755535,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,0.,0.,0.,1527.350647,0.,0.,903.5928286,189.7366596},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,32.86335345,-56.92099788,73.48469229,-37.94733192,65.72670690,42.42640686,-224.4994432,388.8444419,250.9980080,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,113.8419958,-440.9081537,0.,-131.4534138,0.,0.,-777.6888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,440.9081537,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,265.6313234,-460.0869483,-678.8225098,224.4994432,-388.8444419,-2509.980080,-1707.629936},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,920.1738966,0.,0.,777.6888838,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1527.350647,0.,0.,2258.982072,853.8149682},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,149.6662955,-259.2296279,-167.3320053,-1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,518.4592558,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,1505.988048,1328.156617},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,664.0783086}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << convert_dofs_to_poly[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << convert_dofs_to_poly[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "convert_dofs_to_0th_order:" << std::endl;
		static const T convert_dofs_to_0th_order[1][35] = 
		{
			{1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562,1.414213562}
		};
		for (int j = 0; j < 1; j++)
		{
			std::cout << convert_dofs_to_0th_order[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << convert_dofs_to_0th_order[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_stiffness_w:" << std::endl;
		static const T timestep_stiffness_w[35][35] = 
		{
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,10.39230485,0.,15.87450787,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,5.291502622,0.,11.83215956,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,7.745966692,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,3.464101616,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.},
			{0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_stiffness_w[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_stiffness_w[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_bottom_face:" << std::endl;
		static const T timestep_dofs_to_bottom_face[27][35] = 
		{
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.207666290,2.091739373,2.700423918,3.195184670,-1.025843012,-1.776812217,-2.293854709,-1.398249481,-2.421839143,.7821963046,-2.091739372,-3.622998868,-4.677271425,-5.534222184,-4.286111665,-7.423763173,-9.584037044,-1.735425509,-3.005845154,3.129410556,2.402465932,4.161193056,5.372077139,7.607430278,13.17645576,8.483868382,-2.650906245,-4.591504302,-11.29653454,2.875267208},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.207666290,2.091739373,2.700423918,3.195184670,-1.025843012,-1.776812217,-2.293854709,-1.398249481,-2.421839143,.7821963046,2.091739375,3.622998872,4.677271433,5.534222193,4.286111669,7.423763181,9.584037053,1.735425510,3.005845155,-3.129410559,2.402465935,4.161193062,5.372077144,7.607430290,13.17645578,8.483868396,2.650906248,4.591504307,11.29653455,2.875267158},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-2.415332579,-4.183478747,-5.400847836,-6.390369340,3.923332432,6.795411104,8.772838015,-6.407991400,-11.09896668,10.56920547,.1593486743e-8,.276e-8,.3563144678e-8,.4215969639e-8,-.2627089028e-8,-.4550251674e-8,-.5874349651e-8,.4352222914e-8,.7538271208e-8,-.7266379083e-8,-3.130880433,-5.422843982,-7.000861478,5.367005593,9.295926372,-9.215865385,-.1571496103e-7,-.2721911093e-7,.2851586381e-7,4.649692528},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.393815818,.6821090058,.8805989395,1.041938717,-1.469667825,-2.545539342,-3.286277160,.1253720207,.2171507097,1.416680365,-.6821090051,-1.181447453,-1.525242103,-1.804690794,-.8311042347,-1.439514761,-1.858405566,.4291959607,.7433892102,.7095745382,-.6707402935,-1.161756267,-1.499820892,-1.388519566,-2.404986437,-.3003409923,.6556076472,1.135545755,1.898978815,.2835315068e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.393815818,.6821090058,.8805989395,1.041938717,-1.469667825,-2.545539342,-3.286277160,.1253720207,.2171507097,1.416680365,.6821090058,1.181447454,1.525242105,1.804690796,.8311042356,1.439514763,1.858405568,-.4291959612,-.7433892109,-.7095745387,-.6707402924,-1.161756265,-1.499820889,-1.388519564,-2.404986432,-.3003409919,-.6556076472,-1.135545755,-1.898978815,.2835315542e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.787631635,-1.364218011,-1.761197878,-2.083877432,-.5099913165,-.8833308713,-1.140375252,1.364354035,2.363130509,-1.084617593,.3464101616e-9,.6e-9,.7745966692e-9,.9165151390e-9,.4372877027e-11,.7574045189e-11,.9778050292e-11,-.2779346762e-9,-.4813969803e-9,.2775108880e-9,-1.743691247,-3.020161833,-3.899012161,-.8344299601,-1.445275086,1.717913014,0.,0.,0.,1.442216374},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.780140598,-1.351243152,-1.744447408,-2.064058009,-.5220598641,-.9042342088,-1.167361344,1.368139104,2.369686441,-1.070141386,1.351243151,2.340421790,3.021471539,3.575053338,.2738823620e-1,.4743781665e-1,.6124195795e-1,-1.092460490,-1.892197074,1.075289136,-.4881593780,-.8455168449,-1.091557553,-.2385311826,-.4131481275,.4820493399,-1.668760973,-2.890378790,-1.548163659,-1.041619966},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.780140598,-1.351243152,-1.744447408,-2.064058009,-.5220598641,-.9042342088,-1.167361344,1.368139104,2.369686441,-1.070141386,-1.351243152,-2.340421793,-3.021471542,-3.575053341,-.2738823623e-1,-.4743781670e-1,-.6124195801e-1,1.092460491,1.892197076,-1.075289137,-.4881593725,-.8455168354,-1.091557541,-.2385311799,-.4131481229,.4820493345,1.668760962,2.890378772,1.548163649,-1.041620004},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.560281195,2.702486303,3.488894813,4.128116015,-.5536850767,-.9590106839,-1.238077470,-1.785522687,-3.092616012,-.1767011503,-.3464101616e-9,-.6e-9,-.7745966692e-9,-.9165151390e-9,-.8344854191e-9,-.1445371145e-8,-.1865966124e-8,-.6239040877e-9,-.1080633579e-8,.3096646864e-9,-.4528013083,-.7842748716,-1.012494506,-1.648891590,-2.855964011,-2.444365371,-.7483314774e-8,-.1296148140e-7,-.3631457412e-7,.9725382146e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.283894526,-2.223770551,-2.870875435,-3.396865625,.4596200614,.7960852982,1.027741701,.5904589114,1.022704835,-1.343510775,2.223770553,3.851683580,4.972502121,5.883543854,-1.098259750,-1.902241688,-2.455783459,-.4185010573,-.7248650942,1.591483172,1.261805030,2.185510421,2.821481821,-.2397411885,-.4152439192,-.6463928632,-.6392709500,-1.107249765,-.5300284972e-1,-2.181654023},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.283894526,-2.223770551,-2.870875435,-3.396865625,.4596200614,.7960852982,1.027741701,.5904589114,1.022704835,-1.343510775,-2.223770551,-3.851683579,-4.972502117,-5.883543850,1.098259750,1.902241687,2.455783458,.4185010573,.7248650939,-1.591483171,1.261805025,2.185510412,2.821481811,-.2397411877,-.4152439177,-.6463928607,.6392709245,1.107249721,.5300284761e-1,-2.181654030},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.567789052,4.447541102,5.741750871,6.793731251,1.727781187,2.992604800,3.863436185,-.6176495990,-1.069800487,-2.390029691,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1560422168,-.2702730475,-.3489210041,-.7800237319,-1.351040735,-1.950823367,0.,0.,0.,.1154979035e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.805810311,-3.127755208,-4.037914609,-4.777724998,1.840927032,3.188579152,4.116437986,-1.585266239,-2.745761671,1.099742715,3.127755210,5.417430937,6.993873264,8.275262446,-3.210799769,-5.561268334,-7.179566546,2.802690869,4.854402980,-2.005254469,4.133650004,7.159691827,9.243122405,-3.691753629,-6.394304856,2.774298724,4.281181012,7.415223028,-3.392270752,3.547918793},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.805810311,-3.127755208,-4.037914609,-4.777724998,1.840927032,3.188579152,4.116437986,-1.585266239,-2.745761671,1.099742715,-3.127755208,-5.417430934,-6.993873261,-8.275262441,3.210799767,5.561268332,7.179566543,-2.802690867,-4.854402979,2.005254469,4.133650001,7.159691821,9.243122398,-3.691753626,-6.394304850,2.774298722,-4.281181016,-7.415223034,3.392270755,3.547919252},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,3.611620623,6.255510417,8.075829218,9.555449995,5.548439256,9.610178692,12.40668735,6.505405380,11.26769265,6.079065903,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,-.1147467860e-1,-.1987472634e-1,-.2565816138e-1,-.7349529016e-1,-.1272975767,-.2596251910,-.5986651819e-8,-.1036918512e-7,-.4964694369e-7,.6246716172e-4},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.139723890,.2420088766,.3124321161,.3696746651,-1.423600127,-2.465747748,-3.183266656,.6044092113,1.046867463,1.088285796,-1.881087112,-3.258138451,-4.206238652,-4.976888691,-1.804153808,-3.124886061,-4.034210557,1.619309591,2.804726483,1.142629197,1.289024480,2.232655892,2.882346363,2.227214966,3.857649481,-.2121957890,-.4725155092,-.8184208692,-1.167298565,-.1238885768},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.559207282,2.700626232,3.486493472,4.125274710,-.5553799448,-.9619462815,-1.241867310,-1.784795277,-3.091356102,-.1734310073,1.061547995,1.838655062,2.373693478,2.808592000,2.556054550,4.427216349,5.715511728,1.908436179,3.305508424,-.9519572496,.3183248011,.5513547288,.7117958945,1.158730221,2.006979616,1.716453276,-.3086735638e-1,-.5346382954e-1,-.1497356346,-.8941606665e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.698931170,-2.942635106,-3.798925584,-4.494949370,1.527877431,2.646361336,3.416437796,-1.007511332,-1.745060817,.2702851370,.8195391220,1.419483395,1.832545187,2.168296706,-.7519007415,-1.302330287,-1.681301170,.5224058015,.9048333903,-.1906718870,-2.010827754,-3.482855835,-4.496347549,1.506342189,2.609061204,-.7470992591,-2.148063678,-3.720555426,1.317034145,1.229138747},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.698931170,-2.942635106,-3.798925584,-4.494949370,1.527877431,2.646361336,3.416437796,-1.007511332,-1.745060817,.2702851370,-.8195391179,-1.419483391,-1.832545177,-2.168296695,.7519007376,1.302330280,1.681301162,-.5224057986,-.9048333851,.1906718861,-2.010827749,-3.482855828,-4.496347539,1.506342185,2.609061198,-.7470992577,2.148063733,3.720555523,-1.317034179,1.229138974},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,.139723890,.2420088766,.3124321161,.3696746651,-1.423600127,-2.465747748,-3.183266656,.6044092113,1.046867463,1.088285796,1.881087116,3.258138455,4.206238662,4.976888702,1.804153811,3.124886068,4.034210565,-1.619309594,-2.804726489,-1.142629199,1.289024475,2.232655882,2.882346350,2.227214957,3.857649465,-.2121957881,.4725155698,.8184209742,1.167298714,-.1238887416},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,1.559207282,2.700626232,3.486493472,4.125274710,-.5553799448,-.9619462815,-1.241867310,-1.784795277,-3.091356102,-.1734310073,-1.061547993,-1.838655058,-2.373693472,-2.808591992,-2.556054543,-4.427216337,-5.715511713,-1.908436174,-3.305508415,.9519572471,.3183248000,.5513547270,.7117958920,1.158730217,2.006979608,1.716453270,.3086736086e-1,.5346383732e-1,.1497356563,-.8941604766e-1},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.971066118,-1.681935855,-2.171369850,-2.569199455,-.1906367772,-.3301925839,-.4262767929,1.203401099,2.084351846,-1.362876347,-2.724401559,-4.718801920,-6.091947082,-7.208088995,.4756631459,.8238727362,1.063615129,1.696386578,2.938227741,-2.388976129,3.201877769,5.545814974,7.159616348,.7410054061,1.283459012,-2.832011274,-3.040858372,-5.266921197,-1.847444888,2.450578880},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.844934019,4.927570268,6.361465859,7.526987913,2.597617435,4.499205374,5.808449163,.5616969466,.9728876503,-1.873883554,.5212328525,.9028017828,1.165512090,1.379052502,1.939035835,3.358508584,4.335815938,3.687840215,6.387526618,4.461246382,.8451479178e-1,.1463839133,.1889808195,.4540265923,.7863971259,1.260587466,.1137306696e-2,.1969872981e-2,.7969298371e-2,-.3689094461e-2},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.873867901,-3.245634412,-4.190096007,-4.957788456,2.048368944,3.547879081,4.580292201,-1.991866711,-3.450014347,1.736399667,2.203168706,3.816000135,4.926434990,5.829036490,-2.414698978,-4.182381316,-5.399431060,2.358674124,4.085343421,-2.072270251,.6985932727,1.209999042,1.562102047,-.6879618021,-1.191584795,.6128789301,-2.029197433,-3.514673052,1.839475646,-3.732912624},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-1.873867901,-3.245634412,-4.190096007,-4.957788456,2.048368944,3.547879081,4.580292201,-1.991866711,-3.450014347,1.736399667,-2.203168707,-3.816000137,-4.926434994,-5.829036495,2.414698980,4.182381320,5.399431066,-2.358674126,-4.085343424,2.072270252,.6985932694,1.209999036,1.562102039,-.6879617990,-1.191584790,.6128789271,2.029197438,3.514673060,-1.839475650,-3.732912697},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,-.971066118,-1.681935855,-2.171369850,-2.569199455,-.1906367772,-.3301925839,-.4262767929,1.203401099,2.084351846,-1.362876347,2.724401559,4.718801918,6.091947082,7.208088995,-.4756631459,-.8238727362,-1.063615129,-1.696386578,-2.938227741,2.388976128,3.201877775,5.545814986,7.159616362,.7410054080,1.283459015,-2.832011281,3.040858346,5.266921153,1.847444873,2.450579118},
			{1.414213562,2.449489743,3.162277660,3.741657387,4.242640686,2.844934019,4.927570268,6.361465859,7.526987913,2.597617435,4.499205374,5.808449163,.5616969466,.9728876503,-1.873883554,-.5212328521,-.9028017822,-1.165512089,-1.379052502,-1.939035833,-3.358508582,-4.335815935,-3.687840212,-6.387526614,-4.461246379,.8451479167e-1,.1463839132,.1889808193,.4540265917,.7863971248,1.260587465,-.1137294723e-2,-.1969852243e-2,-.7969214472e-2,-.3689141724e-2}
		};
		for (int j = 0; j < 27; j++)
		{
			std::cout << timestep_dofs_to_bottom_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_bottom_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_dofs_to_top_face:" << std::endl;
		static const T timestep_dofs_to_top_face[27][35] = 
		{
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.207666290,-2.091739373,2.700423918,-3.195184670,-1.025843012,1.776812217,-2.293854709,-1.398249481,2.421839143,.7821963046,-2.091739372,3.622998868,-4.677271425,5.534222184,-4.286111665,7.423763173,-9.584037044,-1.735425509,3.005845154,3.129410556,2.402465932,-4.161193056,5.372077139,7.607430278,-13.17645576,8.483868382,-2.650906245,4.591504302,-11.29653454,2.875267208},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.207666290,-2.091739373,2.700423918,-3.195184670,-1.025843012,1.776812217,-2.293854709,-1.398249481,2.421839143,.7821963046,2.091739375,-3.622998872,4.677271433,-5.534222193,4.286111669,-7.423763181,9.584037053,1.735425510,-3.005845155,-3.129410559,2.402465935,-4.161193062,5.372077144,7.607430290,-13.17645578,8.483868396,2.650906248,-4.591504307,11.29653455,2.875267158},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-2.415332579,4.183478747,-5.400847836,6.390369340,3.923332432,-6.795411104,8.772838015,-6.407991400,11.09896668,10.56920547,.1593486743e-8,-.276e-8,.3563144678e-8,-.4215969639e-8,-.2627089028e-8,.4550251674e-8,-.5874349651e-8,.4352222914e-8,-.7538271208e-8,-.7266379083e-8,-3.130880433,5.422843982,-7.000861478,5.367005593,-9.295926372,-9.215865385,-.1571496103e-7,.2721911093e-7,.2851586381e-7,4.649692528},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.393815818,-.6821090058,.8805989395,-1.041938717,-1.469667825,2.545539342,-3.286277160,.1253720207,-.2171507097,1.416680365,-.6821090051,1.181447453,-1.525242103,1.804690794,-.8311042347,1.439514761,-1.858405566,.4291959607,-.7433892102,.7095745382,-.6707402935,1.161756267,-1.499820892,-1.388519566,2.404986437,-.3003409923,.6556076472,-1.135545755,1.898978815,.2835315068e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.393815818,-.6821090058,.8805989395,-1.041938717,-1.469667825,2.545539342,-3.286277160,.1253720207,-.2171507097,1.416680365,.6821090058,-1.181447454,1.525242105,-1.804690796,.8311042356,-1.439514763,1.858405568,-.4291959612,.7433892109,-.7095745387,-.6707402924,1.161756265,-1.499820889,-1.388519564,2.404986432,-.3003409919,-.6556076472,1.135545755,-1.898978815,.2835315542e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.787631635,1.364218011,-1.761197878,2.083877432,-.5099913165,.8833308713,-1.140375252,1.364354035,-2.363130509,-1.084617593,.3464101616e-9,-.6e-9,.7745966692e-9,-.9165151390e-9,.4372877027e-11,-.7574045189e-11,.9778050292e-11,-.2779346762e-9,.4813969803e-9,.2775108880e-9,-1.743691247,3.020161833,-3.899012161,-.8344299601,1.445275086,1.717913014,0.,-0.,0.,1.442216374},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.780140598,1.351243152,-1.744447408,2.064058009,-.5220598641,.9042342088,-1.167361344,1.368139104,-2.369686441,-1.070141386,1.351243151,-2.340421790,3.021471539,-3.575053338,.2738823620e-1,-.4743781665e-1,.6124195795e-1,-1.092460490,1.892197074,1.075289136,-.4881593780,.8455168449,-1.091557553,-.2385311826,.4131481275,.4820493399,-1.668760973,2.890378790,-1.548163659,-1.041619966},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.780140598,1.351243152,-1.744447408,2.064058009,-.5220598641,.9042342088,-1.167361344,1.368139104,-2.369686441,-1.070141386,-1.351243152,2.340421793,-3.021471542,3.575053341,-.2738823623e-1,.4743781670e-1,-.6124195801e-1,1.092460491,-1.892197076,-1.075289137,-.4881593725,.8455168354,-1.091557541,-.2385311799,.4131481229,.4820493345,1.668760962,-2.890378772,1.548163649,-1.041620004},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.560281195,-2.702486303,3.488894813,-4.128116015,-.5536850767,.9590106839,-1.238077470,-1.785522687,3.092616012,-.1767011503,-.3464101616e-9,.6e-9,-.7745966692e-9,.9165151390e-9,-.8344854191e-9,.1445371145e-8,-.1865966124e-8,-.6239040877e-9,.1080633579e-8,.3096646864e-9,-.4528013083,.7842748716,-1.012494506,-1.648891590,2.855964011,-2.444365371,-.7483314774e-8,.1296148140e-7,-.3631457412e-7,.9725382146e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.283894526,2.223770551,-2.870875435,3.396865625,.4596200614,-.7960852982,1.027741701,.5904589114,-1.022704835,-1.343510775,2.223770553,-3.851683580,4.972502121,-5.883543854,-1.098259750,1.902241688,-2.455783459,-.4185010573,.7248650942,1.591483172,1.261805030,-2.185510421,2.821481821,-.2397411885,.4152439192,-.6463928632,-.6392709500,1.107249765,-.5300284972e-1,-2.181654023},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.283894526,2.223770551,-2.870875435,3.396865625,.4596200614,-.7960852982,1.027741701,.5904589114,-1.022704835,-1.343510775,-2.223770551,3.851683579,-4.972502117,5.883543850,1.098259750,-1.902241687,2.455783458,.4185010573,-.7248650939,-1.591483171,1.261805025,-2.185510412,2.821481811,-.2397411877,.4152439177,-.6463928607,.6392709245,-1.107249721,.5300284761e-1,-2.181654030},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.567789052,-4.447541102,5.741750871,-6.793731251,1.727781187,-2.992604800,3.863436185,-.6176495990,1.069800487,-2.390029691,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.1560422168,.2702730475,-.3489210041,-.7800237319,1.351040735,-1.950823367,0.,-0.,0.,.1154979035e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.805810311,3.127755208,-4.037914609,4.777724998,1.840927032,-3.188579152,4.116437986,-1.585266239,2.745761671,1.099742715,3.127755210,-5.417430937,6.993873264,-8.275262446,-3.210799769,5.561268334,-7.179566546,2.802690869,-4.854402980,-2.005254469,4.133650004,-7.159691827,9.243122405,-3.691753629,6.394304856,2.774298724,4.281181012,-7.415223028,-3.392270752,3.547918793},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.805810311,3.127755208,-4.037914609,4.777724998,1.840927032,-3.188579152,4.116437986,-1.585266239,2.745761671,1.099742715,-3.127755208,5.417430934,-6.993873261,8.275262441,3.210799767,-5.561268332,7.179566543,-2.802690867,4.854402979,2.005254469,4.133650001,-7.159691821,9.243122398,-3.691753626,6.394304850,2.774298722,-4.281181016,7.415223034,3.392270755,3.547919252},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,3.611620623,-6.255510417,8.075829218,-9.555449995,5.548439256,-9.610178692,12.40668735,6.505405380,-11.26769265,6.079065903,0.,-0.,0.,-0.,0.,-0.,0.,0.,-0.,0.,-.1147467860e-1,.1987472634e-1,-.2565816138e-1,-.7349529016e-1,.1272975767,-.2596251910,-.5986651819e-8,.1036918512e-7,-.4964694369e-7,.6246716172e-4},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.139723890,-.2420088766,.3124321161,-.3696746651,-1.423600127,2.465747748,-3.183266656,.6044092113,-1.046867463,1.088285796,-1.881087112,3.258138451,-4.206238652,4.976888691,-1.804153808,3.124886061,-4.034210557,1.619309591,-2.804726483,1.142629197,1.289024480,-2.232655892,2.882346363,2.227214966,-3.857649481,-.2121957890,-.4725155092,.8184208692,-1.167298565,-.1238885768},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.559207282,-2.700626232,3.486493472,-4.125274710,-.5553799448,.9619462815,-1.241867310,-1.784795277,3.091356102,-.1734310073,1.061547995,-1.838655062,2.373693478,-2.808592000,2.556054550,-4.427216349,5.715511728,1.908436179,-3.305508424,-.9519572496,.3183248011,-.5513547288,.7117958945,1.158730221,-2.006979616,1.716453276,-.3086735638e-1,.5346382954e-1,-.1497356346,-.8941606665e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.698931170,2.942635106,-3.798925584,4.494949370,1.527877431,-2.646361336,3.416437796,-1.007511332,1.745060817,.2702851370,.8195391220,-1.419483395,1.832545187,-2.168296706,-.7519007415,1.302330287,-1.681301170,.5224058015,-.9048333903,-.1906718870,-2.010827754,3.482855835,-4.496347549,1.506342189,-2.609061204,-.7470992591,-2.148063678,3.720555426,1.317034145,1.229138747},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.698931170,2.942635106,-3.798925584,4.494949370,1.527877431,-2.646361336,3.416437796,-1.007511332,1.745060817,.2702851370,-.8195391179,1.419483391,-1.832545177,2.168296695,.7519007376,-1.302330280,1.681301162,-.5224057986,.9048333851,.1906718861,-2.010827749,3.482855828,-4.496347539,1.506342185,-2.609061198,-.7470992577,2.148063733,-3.720555523,-1.317034179,1.229138974},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,.139723890,-.2420088766,.3124321161,-.3696746651,-1.423600127,2.465747748,-3.183266656,.6044092113,-1.046867463,1.088285796,1.881087116,-3.258138455,4.206238662,-4.976888702,1.804153811,-3.124886068,4.034210565,-1.619309594,2.804726489,-1.142629199,1.289024475,-2.232655882,2.882346350,2.227214957,-3.857649465,-.2121957881,.4725155698,-.8184209742,1.167298714,-.1238887416},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,1.559207282,-2.700626232,3.486493472,-4.125274710,-.5553799448,.9619462815,-1.241867310,-1.784795277,3.091356102,-.1734310073,-1.061547993,1.838655058,-2.373693472,2.808591992,-2.556054543,4.427216337,-5.715511713,-1.908436174,3.305508415,.9519572471,.3183248000,-.5513547270,.7117958920,1.158730217,-2.006979608,1.716453270,.3086736086e-1,-.5346383732e-1,.1497356563,-.8941604766e-1},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.971066118,1.681935855,-2.171369850,2.569199455,-.1906367772,.3301925839,-.4262767929,1.203401099,-2.084351846,-1.362876347,-2.724401559,4.718801920,-6.091947082,7.208088995,.4756631459,-.8238727362,1.063615129,1.696386578,-2.938227741,-2.388976129,3.201877769,-5.545814974,7.159616348,.7410054061,-1.283459012,-2.832011274,-3.040858372,5.266921197,-1.847444888,2.450578880},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.844934019,-4.927570268,6.361465859,-7.526987913,2.597617435,-4.499205374,5.808449163,.5616969466,-.9728876503,-1.873883554,.5212328525,-.9028017828,1.165512090,-1.379052502,1.939035835,-3.358508584,4.335815938,3.687840215,-6.387526618,4.461246382,.8451479178e-1,-.1463839133,.1889808195,.4540265923,-.7863971259,1.260587466,.1137306696e-2,-.1969872981e-2,.7969298371e-2,-.3689094461e-2},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.873867901,3.245634412,-4.190096007,4.957788456,2.048368944,-3.547879081,4.580292201,-1.991866711,3.450014347,1.736399667,2.203168706,-3.816000135,4.926434990,-5.829036490,-2.414698978,4.182381316,-5.399431060,2.358674124,-4.085343421,-2.072270251,.6985932727,-1.209999042,1.562102047,-.6879618021,1.191584795,.6128789301,-2.029197433,3.514673052,1.839475646,-3.732912624},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-1.873867901,3.245634412,-4.190096007,4.957788456,2.048368944,-3.547879081,4.580292201,-1.991866711,3.450014347,1.736399667,-2.203168707,3.816000137,-4.926434994,5.829036495,2.414698980,-4.182381320,5.399431066,-2.358674126,4.085343424,2.072270252,.6985932694,-1.209999036,1.562102039,-.6879617990,1.191584790,.6128789271,2.029197438,-3.514673060,-1.839475650,-3.732912697},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,-.971066118,1.681935855,-2.171369850,2.569199455,-.1906367772,.3301925839,-.4262767929,1.203401099,-2.084351846,-1.362876347,2.724401559,-4.718801918,6.091947082,-7.208088995,-.4756631459,.8238727362,-1.063615129,-1.696386578,2.938227741,2.388976128,3.201877775,-5.545814986,7.159616362,.7410054080,-1.283459015,-2.832011281,3.040858346,-5.266921153,1.847444873,2.450579118},
			{1.414213562,-2.449489743,3.162277660,-3.741657387,4.242640686,2.844934019,-4.927570268,6.361465859,-7.526987913,2.597617435,-4.499205374,5.808449163,.5616969466,-.9728876503,-1.873883554,-.5212328521,.9028017822,-1.165512089,1.379052502,-1.939035833,3.358508582,-4.335815935,-3.687840212,6.387526614,-4.461246379,.8451479167e-1,-.1463839132,.1889808193,.4540265917,-.7863971248,1.260587465,-.1137294723e-2,.1969852243e-2,-.7969214472e-2,-.3689141724e-2}
		};
		for (int j = 0; j < 27; j++)
		{
			std::cout << timestep_dofs_to_top_face[j][0];
			for (int i = 1; i < 35; i++)
			{
				std::cout << ", " << timestep_dofs_to_top_face[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_bottom_fluxes_to_dofs:" << std::endl;
		static const T timestep_bottom_fluxes_to_dofs[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{-.2270692495e-2,-.2270692495e-2,-.2270692495e-2,-.1889769944,-.1889769944,-.1889769944,-.1453110246,-.1453110246,-.1453110246,-.8863366081e-1,-.8863366081e-1,-.8863366081e-1,-.3345937098e-1,-.3345937098e-1,-.3345937098e-1,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.1281992189,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1,-.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{-.3468540079e-2,-.3468540079e-2,-.3468540079e-2,-.2886671272,-.2886671272,-.2886671272,-.2219662565,-.2219662565,-.2219662565,-.1353901533,-.1353901533,-.1353901533,-.5111003340e-1,-.5111003340e-1,-.5111003340e-1,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.1958275416,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1,-.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1119514294e-2,-.1119514294e-2,.2239028587e-2,-.3038270720e-1,-.3038270720e-1,.6076541432e-1,.4628026305e-1,.4628026305e-1,-.9256052604e-1,.4645713347e-1,.4645713347e-1,-.9291426695e-1,.2466688309e-1,.2466688309e-1,-.4933376618e-1,-.7312744874e-2,-.816044061e-1,.8891715086e-1,.8891715086e-1,-.7312744874e-2,-.816044061e-1,.2010850666e-1,-.5891192536e-1,.3880341870e-1,.3880341870e-1,.2010850666e-1,-.5891192536e-1},
			{-.1939055637e-2,-.1939055637e-2,.3878111275e-2,-.5262439255e-1,-.5262439255e-1,.1052487850,.8015976696e-1,.8015976696e-1,-.1603195338,.8046611556e-1,.8046611556e-1,-.1609322311,.4272429479e-1,.4272429479e-1,-.8544858956e-1,-.1266604567e-1,-.1413429774,.1540090230,.1540090230,-.1266604567e-1,-.1413429774,.3482895521e-1,-.1020384480,.6720949271e-1,.6720949271e-1,.3482895521e-1,-.1020384480},
			{-.2503310063e-2,-.2503310063e-2,.5006620125e-2,-.6793779862e-1,-.6793779862e-1,.1358755971,.1034858141,.1034858141,-.2069716281,.1038813085,.1038813085,-.2077626169,.5515682737e-1,.5515682737e-1,-.1103136547,-.1635179464e-1,-.1824729992,.1988247936,.1988247936,-.1635179464e-1,-.1824729992,.4496398781e-1,-.1317310698,.8676708195e-1,.8676708195e-1,.4496398781e-1,-.1317310698},
			{-.2961956411e-2,-.2961956411e-2,.5923912822e-2,-.8038508741e-1,-.8038508741e-1,.1607701747,.1224460666,.1224460666,-.2448921330,.1229140218,.1229140218,-.2458280436,.6526243827e-1,.6526243827e-1,-.1305248765,-.1934770434e-1,-.2159049643,.2352526685,.2352526685,-.1934770434e-1,-.2159049643,.5320210786e-1,-.1558663038,.1026641959,.1026641959,.5320210786e-1,-.1558663038},
			{.9509629649e-3,.9509629649e-3,-.3636953995e-2,.1133841892,.1133841892,.3934559288e-1,.3097014551e-1,.3097014551e-1,.3284624728e-1,-.1663114073e-1,-.1663114073e-1,-.6251896835e-1,-.2514656806e-1,-.2514656806e-1,-.7579018773e-1,.7450711923e-1,.2906698235e-1,-.7996469216e-1,-.7996469216e-1,.7450711923e-1,.2906698235e-1,.3947641499e-2,-.5379057771e-1,-.4241692689e-1,-.4241692689e-1,.3947641499e-2,-.5379057771e-1},
			{.1647116171e-2,.1647116171e-2,-.6299389101e-2,.1963871764,.1963871764,.6814856589e-1,.5364186551e-1,.5364186551e-1,.5689136911e-1,-.2880598072e-1,-.2880598072e-1,-.1082860296,-.4355513348e-1,-.4355513348e-1,-.1312724558,.1290501160,.5034549024e-1,-.1385029096,-.1385029096,.1290501160,.5034549024e-1,.6837515642e-2,-.9316801354e-1,-.7346827246e-1,-.7346827246e-1,.6837515642e-2,-.9316801354e-1},
			{.2126417833e-2,.2126417833e-2,-.8132476364e-2,.2535347545,.2535347545,.8797942029e-1,.6925135063e-1,.6925135063e-1,.7344644172e-1,-.3718836120e-1,-.3718836120e-1,-.1397966631,-.5622943557e-1,-.5622943557e-1,-.1694720118,.1666029834,.6499574844e-1,-.1788064875,-.1788064875,.1666029834,.6499574844e-1,.8827194741e-2,-.1202793883,-.9484713192e-1,-.9484713192e-1,.8827194741e-2,-.1202793883},
			{.1296186118e-2,.1296186118e-2,.5940248584e-2,-.9672393084e-2,-.9672393084e-2,-.1052592792,-.8116208511e-1,-.8116208511e-1,.1059225219,-.2136548439e-1,-.2136548439e-1,.2234936694e-1,.2165431039e-1,.2165431039e-1,-.8886208754e-1,-.3163303256e-1,.9341103024e-1,.5273023340e-1,.5273023340e-1,-.3163303256e-1,.9341103024e-1,-.2491962037e-1,-.1163142919e-1,.4124689791e-1,.4124689791e-1,-.2491962037e-1,-.1163142919e-1},
			{.2245060214e-2,.2245060214e-2,.1028881236e-1,-.1675307626e-1,-.1675307626e-1,-.1823144197,-.1405768551,-.1405768551,.1834631897,-.3700610451e-1,-.3700610451e-1,.3871023906e-1,.3750636580e-1,.3750636580e-1,-.1539136505,-.5479001961e-1,.1617926504,.9133144337e-1,.9133144337e-1,-.5479001961e-1,.1617926504,-.4316204859e-1,-.2014622633e-1,.7144172285e-1,.7144172285e-1,-.4316204859e-1,-.2014622633e-1},
			{-.7251009248e-3,-.7251009248e-3,-.9797720361e-2,-.1092962313,-.1092962313,.8367774286e-1,.6348397322e-1,.6348397322e-1,.1048243834e-1,.4861432003e-1,.4861432003e-1,.8648212617e-1,-.1502218966e-1,-.1502218966e-1,-.8303840497e-1,-.5695773555e-1,.9076878046e-2,-.1414594348e-1,-.1414594348e-1,-.5695773555e-1,.9076878046e-2,.2822197953e-1,.3880374283e-1,-.3595677329e-1,-.3595677329e-1,.2822197953e-1,.3880374283e-1},
			{.1939055636e-2,-.1939055639e-2,-.1477172296e-11,.5262439250e-1,-.5262439255e-1,-.2672538285e-10,-.8015976689e-1,.8015976696e-1,.2055008219e-10,-.8046611563e-1,.8046611556e-1,-0.,-.4272429480e-1,.4272429479e-1,-0.,.9845066677e-1,-.5555835628e-1,-.4289231078e-1,.4289231057e-1,-.9845066699e-1,.5555835614e-1,.5641598021e-1,-.1079351250e-1,-.4562246768e-1,.4562246772e-1,-.5641598021e-1,.1079351250e-1},
			{.3358542881e-2,-.3358542884e-2,-.2558537468e-11,.9114812152e-1,-.9114812160e-1,-.4628972095e-10,-.1388407889,.1388407891,.3559378643e-10,-.1393714005,.1393714005,-0.,-.7400064931e-1,.7400064927e-1,-0.,.1705215569,-.9622989583e-1,-.742916614e-1,.7429166116e-1,-.1705215571,.962298956e-1,.9771534407e-1,-.1869491204e-1,-.7902043198e-1,.7902043202e-1,-.9771534403e-1,.1869491203e-1},
			{.4335860212e-2,-.4335860220e-2,-.3303057668e-11,.1176717189,-.1176717190,-.5975977276e-10,-.1792426878,.1792426879,.4595138070e-10,-.1799277044,.1799277042,-0.,-.9553442744e-1,.9553442741e-1,-0.,.2201423833,-.1242322613,-.9591012258e-1,.9591012213e-1,-.2201423837,.1242322610,.1261499667,-.2413502766e-1,-.1020149390,.1020149391,-.1261499667,.2413502765e-1},
			{.5130258989e-2,-.5130258998e-2,-.3908230539e-11,.1392310554,-.1392310556,-.7070871670e-10,-.2120828083,.2120828085,.5437040688e-10,-.2128933309,.2128933307,-0.,-.1130378590,.1130378589,-0.,.2604759806,-.1469935939,-.1134823874,.1134823869,-.2604759812,.1469935935,.1492626536,-.2855694985e-1,-.1207057037,.1207057037,-.1492626536,.2855694983e-1},
			{.3973252640e-2,-.3973252644e-2,.2435328156e-11,.6411930517e-1,-.6411930524e-1,-.3373654288e-12,-.1624751717e-2,.1624751718e-2,.4950415964e-10,.3974002441e-1,-.3974002439e-1,-0.,.4385866113e-1,-.4385866112e-1,-0.,.9442419982e-1,-.1337765132,.3935231328e-1,-.3935231309e-1,-.9442420005e-1,.1337765128,-.9849870528e-2,-.4015289410e-1,.5000276456e-1,-.5000276462e-1,.9849870528e-2,.4015289406e-1},
			{.6881875447e-2,-.6881875454e-2,.4218112101e-11,.1110578943,-.1110578945,-.5843340636e-12,-.2814152524e-2,.2814152527e-2,.8574371972e-10,.6883174139e-1,-.6883174137e-1,-0.,.7596542946e-1,-.7596542944e-1,-0.,.1635475116,-.2317077178,.6816020601e-1,-.6816020569e-1,-.1635475120,.2317077171,-.1706047621e-1,-.6954685268e-1,.8660732877e-1,-.8660732887e-1,.1706047621e-1,.6954685260e-1},
			{.8884462997e-2,-.8884463007e-2,.5445559306e-11,.1433751251,-.1433751252,-.7543720322e-12,-.3633055286e-2,.3633055289e-2,.1106946662e-9,.8886139602e-1,-.8886139599e-1,-0.,.9807094772e-1,-.9807094769e-1,-0.,.2111389296,-.2991333773,.8799444758e-1,-.8799444716e-1,-.2111389301,.2991333765,-.2202498007e-1,-.8978460072e-1,.1118095806,-.1118095808,.2202498007e-1,.8978460063e-1},
			{.1608750430e-2,-.1608750432e-2,-.4034538186e-11,-.3311226875e-1,.3311226880e-1,.2144253100e-10,.6480800897e-1,-.6480800902e-1,.3701184809e-10,.1514326846e-1,-.1514326846e-1,-0.,-.3828400335e-1,.3828400333e-1,-0.,-.8474998733e-1,-.9988203793e-1,-.2734121092e-1,.2734121078e-1,.8474998753e-1,.9988203767e-1,-.3512819587e-1,-.7636653997e-1,-.4884262097e-1,.4884262099e-1,.3512819587e-1,.7636653992e-1},
			{.2786437481e-2,-.2786437483e-2,-.6988025119e-11,-.5735213183e-1,.5735213188e-1,.3713955314e-10,.1122507642,-.1122507643,.6410640134e-10,.2622891035e-1,-.2622891035e-1,-0.,-.6630983887e-1,.6630983886e-1,-0.,-.1467912839,-.1730007645,-.4735636645e-1,.4735636618e-1,.1467912843,.1730007640,-.6084382000e-1,-.1322707272,-.8459790107e-1,.8459790113e-1,.6084382000e-1,.1322707271},
			{-.2900983391e-2,.2900983395e-2,.6735979401e-11,-.5474334561e-1,.5474334565e-1,-.2140983594e-10,-.6378935306e-1,.6378935317e-1,-.1837023118e-10,-.5758708731e-1,.5758708725e-1,-0.,.2739123664e-1,-.2739123663e-1,-0.,-.5980191219e-1,.4982269315e-1,.9979215896e-2,-.9979215849e-2,.5980191234e-1,-.4982269302e-1,.4947010457e-1,-.9238197165e-1,.4291186704e-1,-.4291186706e-1,-.4947010455e-1,.9238197160e-1},
			{-.2227101124e-2,-.2227101127e-2,.2902345977e-2,.5174730169e-1,.5174730161e-1,.1345249687,.2895906774e-1,.2895906742e-1,.2686152177e-1,-.4565783520e-1,-.4565783503e-1,.5646315914e-2,-.5646454713e-1,-.5646454707e-1,.1567410230e-3,-.6746381852e-1,-.1666020075e-1,.1052409173,.1052409171,-.6746381827e-1,-.1666020070e-1,-.6630339505e-1,-.1750103542e-2,-.1446623172e-1,-.1446623165e-1,-.6630339516e-1,-.1750103540e-2},
			{-.3857452299e-2,-.3857452305e-2,.5027010692e-2,.8962895569e-1,.8962895553e-1,.2330040807,.5015857668e-1,.5015857611e-1,.4652552046e-1,-.7908169034e-1,-.7908169002e-1,.9779706035e-2,-.9779946441e-1,-.9779946432e-1,.2714834155e-3,-.1168507614,-.2885631417e-1,.1822826158,.1822826154,-.1168507609,-.2885631408e-1,-.1148408489,-.3031268254e-2,-.2505624833e-1,-.2505624821e-1,-.1148408492,-.3031268250e-2},
			{-.4979949508e-2,-.4979949513e-2,.6489842900e-2,.1157104843,.1157104841,.3008069748,.6475444405e-1,.6475444331e-1,.6006418864e-1,-.1020940232,-.1020940229,.1262554620e-1,-.1262585657,-.1262585656,.3504835826e-3,-.1508536843,-.3725334142e-1,.2353258451,.2353258446,-.1508536837,-.3725334129e-1,-.1482588985,-.3913350489e-2,-.3234747750e-1,-.3234747736e-1,-.1482588988,-.3913350484e-2},
			{-.7052136016e-2,-.7052136025e-2,-.4975248153e-2,.1071236388,.1071236386,.6437588332e-1,.1415037995e-1,.1415037979e-1,.9781715849e-1,.8674924746e-2,.8674924715e-2,.2822480032e-1,.5042836149e-1,.5042836146e-1,.1003925893e-2,-.1165659990,-.6064459330e-1,-.7883759978e-1,-.7883759957e-1,-.1165659985,-.6064459310e-1,-.1534448774e-1,-.9401828142e-2,.1424607884e-1,.1424607878e-1,-.1534448778e-1,-.9401828126e-2},
			{-.1221465788e-1,-.1221465790e-1,-.8617382582e-2,.1855435851,.1855435847,.1115023007,.2450917702e-1,.2450917675e-1,.1694242884,.1502541041e-1,.1502541036e-1,.4888678819e-1,.8734448425e-1,.8734448420e-1,.1738850654e-2,-.2018982328,-.1050395168,-.1365507284,-.1365507280,-.2018982319,-.1050395164,-.2657743238e-1,-.1628444403e-1,.2467493236e-1,.2467493225e-1,-.2657743245e-1,-.1628444400e-1},
			{-.7864599685e-2,-.7864599698e-2,.8543165539e-2,.2317116788e-1,.2317116784e-1,-.1325361900,-.2859660209e-1,-.2859660176e-1,.1450070316,.2338942873e-1,.2338942865e-1,.7058964713e-1,-.3789617428e-1,-.3789617426e-1,.3546410271e-2,.1110571477e-1,-.8983420728e-1,.3910101757e-1,.3910101750e-1,.1110571472e-1,-.8983420697e-1,.5864432557e-1,-.2610381621e-1,-.1269128828e-1,-.1269128822e-1,.5864432571e-1,-.2610381617e-1},
			{.2457406867e-2,-.2457406870e-2,.1456786833e-10,-.5057982506e-1,.5057982506e-1,-0.,.9899586944e-1,-.9899586881e-1,.4439325132e-9,.2313172558e-1,-.2313172466e-1,-0.,-.5847978098e-1,.5847978105e-1,.8177605344e-10,.2473009711e-1,.1615508286e-2,.1124234492,-.1124234521,-.2473010027e-1,-.1615508521e-2,.6296906016e-1,-.2355095997e-4,.4201992979e-1,-.4201992986e-1,-.6296905964e-1,.2355071203e-4},
			{.4256353547e-2,-.4256353552e-2,.2523228810e-10,-.8760682683e-1,.8760682683e-1,-0.,.1714658756,-.1714658745,.7689136679e-9,.4006532397e-1,-.4006532237e-1,-0.,-.1012899519,.1012899520,.1416402794e-9,.4283378466e-1,.2798142432e-2,.1947231258,-.1947231310,-.4283379015e-1,-.2798142839e-2,.1090656115,-.4079145923e-4,.7278065332e-1,-.7278065345e-1,-.1090656106,.4079102978e-4},
			{.1047195901e-1,-.1047195902e-1,-.2643438623e-10,-.1465053324,.1465053324,-0.,.9184167768e-1,-.9184167710e-1,.2154288659e-8,.1917883762e-2,-.1917883686e-2,-0.,.4633750594e-1,-.4633750599e-1,.6781638958e-9,.6109303566e-1,.7836730672e-2,-.6892976350e-1,.6892976527e-1,-.6109304346e-1,-.7836731808e-2,.3825625994e-1,-.1650255183e-3,-.3809123559e-1,.3809123567e-1,-.3825625962e-1,.1650237808e-3},
			{-.2665390899e-2,-.2665390853e-2,-.4310294403e-2,-.2187432388e-2,-.2187432754e-2,-.1112663225,.6179199770e-1,.6179199994e-1,-.5769386249e-2,.7894214835e-1,.7894214861e-1,-.4179238568e-3,-.4846361631e-1,-.4846362257e-1,-.8532846255e-6,.6483970316e-2,.4679778692e-2,-.6432957221e-1,-.6432958410e-1,.6483978942e-2,.4679777698e-2,-.5074575336e-1,.7639251247e-4,.7729988409e-1,.7729988561e-1,-.5074575829e-1,.7639349119e-4}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_bottom_fluxes_to_dofs[j][0];
			for (int i = 1; i < 27; i++)
			{
				std::cout << ", " << timestep_bottom_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << std::endl;
		std::cout << "timestep_top_fluxes_to_dofs:" << std::endl;
		static const T timestep_top_fluxes_to_dofs[35][27] = 
		{
			{-.1310984923e-2,-.1310984923e-2,-.1310984923e-2,-.1091059186,-.1091059186,-.1091059186,-.8389535915e-1,-.8389535915e-1,-.8389535915e-1,-.5117266791e-1,-.5117266791e-1,-.5117266791e-1,-.1931777683e-1,-.1931777683e-1,-.1931777683e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.7401585353e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1,-.2928505310e-1},
			{.2270692495e-2,.2270692495e-2,.2270692495e-2,.1889769944,.1889769944,.1889769944,.1453110246,.1453110246,.1453110246,.8863366081e-1,.8863366081e-1,.8863366081e-1,.3345937098e-1,.3345937098e-1,.3345937098e-1,.1281992189,.1281992189,.1281992189,.1281992189,.1281992189,.1281992189,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1,.5072319989e-1},
			{-.2931451405e-2,-.2931451405e-2,-.2931451405e-2,-.2439682507,-.2439682507,-.2439682507,-.1875957261,-.1875957261,-.1875957261,-.1144255641,-.1144255641,-.1144255641,-.4319586218e-1,-.4319586218e-1,-.4319586218e-1,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.1655044799,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1,-.6548336947e-1},
			{.3468540079e-2,.3468540079e-2,.3468540079e-2,.2886671272,.2886671272,.2886671272,.2219662565,.2219662565,.2219662565,.1353901533,.1353901533,.1353901533,.5111003340e-1,.5111003340e-1,.5111003340e-1,.1958275416,.1958275416,.1958275416,.1958275416,.1958275416,.1958275416,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1,.7748096766e-1},
			{-.3932954768e-2,-.3932954768e-2,-.3932954768e-2,-.3273177557,-.3273177557,-.3273177557,-.2516860774,-.2516860774,-.2516860774,-.1535180037,-.1535180037,-.1535180037,-.5795333050e-1,-.5795333050e-1,-.5795333050e-1,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.2220475606,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1,-.8785515930e-1},
			{-.1119514294e-2,-.1119514294e-2,.2239028587e-2,-.3038270720e-1,-.3038270720e-1,.6076541432e-1,.4628026305e-1,.4628026305e-1,-.9256052604e-1,.4645713347e-1,.4645713347e-1,-.9291426695e-1,.2466688309e-1,.2466688309e-1,-.4933376618e-1,-.7312744874e-2,-.816044061e-1,.8891715086e-1,.8891715086e-1,-.7312744874e-2,-.816044061e-1,.2010850666e-1,-.5891192536e-1,.3880341870e-1,.3880341870e-1,.2010850666e-1,-.5891192536e-1},
			{.1939055637e-2,.1939055637e-2,-.3878111275e-2,.5262439255e-1,.5262439255e-1,-.1052487850,-.8015976696e-1,-.8015976696e-1,.1603195338,-.8046611556e-1,-.8046611556e-1,.1609322311,-.4272429479e-1,-.4272429479e-1,.8544858956e-1,.1266604567e-1,.1413429774,-.1540090230,-.1540090230,.1266604567e-1,.1413429774,-.3482895521e-1,.1020384480,-.6720949271e-1,-.6720949271e-1,-.3482895521e-1,.1020384480},
			{-.2503310063e-2,-.2503310063e-2,.5006620125e-2,-.6793779862e-1,-.6793779862e-1,.1358755971,.1034858141,.1034858141,-.2069716281,.1038813085,.1038813085,-.2077626169,.5515682737e-1,.5515682737e-1,-.1103136547,-.1635179464e-1,-.1824729992,.1988247936,.1988247936,-.1635179464e-1,-.1824729992,.4496398781e-1,-.1317310698,.8676708195e-1,.8676708195e-1,.4496398781e-1,-.1317310698},
			{.2961956411e-2,.2961956411e-2,-.5923912822e-2,.8038508741e-1,.8038508741e-1,-.1607701747,-.1224460666,-.1224460666,.2448921330,-.1229140218,-.1229140218,.2458280436,-.6526243827e-1,-.6526243827e-1,.1305248765,.1934770434e-1,.2159049643,-.2352526685,-.2352526685,.1934770434e-1,.2159049643,-.5320210786e-1,.1558663038,-.1026641959,-.1026641959,-.5320210786e-1,.1558663038},
			{.9509629649e-3,.9509629649e-3,-.3636953995e-2,.1133841892,.1133841892,.3934559288e-1,.3097014551e-1,.3097014551e-1,.3284624728e-1,-.1663114073e-1,-.1663114073e-1,-.6251896835e-1,-.2514656806e-1,-.2514656806e-1,-.7579018773e-1,.7450711923e-1,.2906698235e-1,-.7996469216e-1,-.7996469216e-1,.7450711923e-1,.2906698235e-1,.3947641499e-2,-.5379057771e-1,-.4241692689e-1,-.4241692689e-1,.3947641499e-2,-.5379057771e-1},
			{-.1647116171e-2,-.1647116171e-2,.6299389101e-2,-.1963871764,-.1963871764,-.6814856589e-1,-.5364186551e-1,-.5364186551e-1,-.5689136911e-1,.2880598072e-1,.2880598072e-1,.1082860296,.4355513348e-1,.4355513348e-1,.1312724558,-.1290501160,-.5034549024e-1,.1385029096,.1385029096,-.1290501160,-.5034549024e-1,-.6837515642e-2,.9316801354e-1,.7346827246e-1,.7346827246e-1,-.6837515642e-2,.9316801354e-1},
			{.2126417833e-2,.2126417833e-2,-.8132476364e-2,.2535347545,.2535347545,.8797942029e-1,.6925135063e-1,.6925135063e-1,.7344644172e-1,-.3718836120e-1,-.3718836120e-1,-.1397966631,-.5622943557e-1,-.5622943557e-1,-.1694720118,.1666029834,.6499574844e-1,-.1788064875,-.1788064875,.1666029834,.6499574844e-1,.8827194741e-2,-.1202793883,-.9484713192e-1,-.9484713192e-1,.8827194741e-2,-.1202793883},
			{.1296186118e-2,.1296186118e-2,.5940248584e-2,-.9672393084e-2,-.9672393084e-2,-.1052592792,-.8116208511e-1,-.8116208511e-1,.1059225219,-.2136548439e-1,-.2136548439e-1,.2234936694e-1,.2165431039e-1,.2165431039e-1,-.8886208754e-1,-.3163303256e-1,.9341103024e-1,.5273023340e-1,.5273023340e-1,-.3163303256e-1,.9341103024e-1,-.2491962037e-1,-.1163142919e-1,.4124689791e-1,.4124689791e-1,-.2491962037e-1,-.1163142919e-1},
			{-.2245060214e-2,-.2245060214e-2,-.1028881236e-1,.1675307626e-1,.1675307626e-1,.1823144197,.1405768551,.1405768551,-.1834631897,.3700610451e-1,.3700610451e-1,-.3871023906e-1,-.3750636580e-1,-.3750636580e-1,.1539136505,.5479001961e-1,-.1617926504,-.9133144337e-1,-.9133144337e-1,.5479001961e-1,-.1617926504,.4316204859e-1,.2014622633e-1,-.7144172285e-1,-.7144172285e-1,.4316204859e-1,.2014622633e-1},
			{-.7251009248e-3,-.7251009248e-3,-.9797720361e-2,-.1092962313,-.1092962313,.8367774286e-1,.6348397322e-1,.6348397322e-1,.1048243834e-1,.4861432003e-1,.4861432003e-1,.8648212617e-1,-.1502218966e-1,-.1502218966e-1,-.8303840497e-1,-.5695773555e-1,.9076878046e-2,-.1414594348e-1,-.1414594348e-1,-.5695773555e-1,.9076878046e-2,.2822197953e-1,.3880374283e-1,-.3595677329e-1,-.3595677329e-1,.2822197953e-1,.3880374283e-1},
			{.1939055636e-2,-.1939055639e-2,-.1477172296e-11,.5262439250e-1,-.5262439255e-1,-.2672538285e-10,-.8015976689e-1,.8015976696e-1,.2055008219e-10,-.8046611563e-1,.8046611556e-1,-0.,-.4272429480e-1,.4272429479e-1,-0.,.9845066677e-1,-.5555835628e-1,-.4289231078e-1,.4289231057e-1,-.9845066699e-1,.5555835614e-1,.5641598021e-1,-.1079351250e-1,-.4562246768e-1,.4562246772e-1,-.5641598021e-1,.1079351250e-1},
			{-.3358542881e-2,.3358542884e-2,.2558537468e-11,-.9114812152e-1,.9114812160e-1,.4628972095e-10,.1388407889,-.1388407891,-.3559378643e-10,.1393714005,-.1393714005,0.,.7400064931e-1,-.7400064927e-1,0.,-.1705215569,.9622989583e-1,.742916614e-1,-.7429166116e-1,.1705215571,-.962298956e-1,-.9771534407e-1,.1869491204e-1,.7902043198e-1,-.7902043202e-1,.9771534403e-1,-.1869491203e-1},
			{.4335860212e-2,-.4335860220e-2,-.3303057668e-11,.1176717189,-.1176717190,-.5975977276e-10,-.1792426878,.1792426879,.4595138070e-10,-.1799277044,.1799277042,-0.,-.9553442744e-1,.9553442741e-1,-0.,.2201423833,-.1242322613,-.9591012258e-1,.9591012213e-1,-.2201423837,.1242322610,.1261499667,-.2413502766e-1,-.1020149390,.1020149391,-.1261499667,.2413502765e-1},
			{-.5130258989e-2,.5130258998e-2,.3908230539e-11,-.1392310554,.1392310556,.7070871670e-10,.2120828083,-.2120828085,-.5437040688e-10,.2128933309,-.2128933307,0.,.1130378590,-.1130378589,0.,-.2604759806,.1469935939,.1134823874,-.1134823869,.2604759812,-.1469935935,-.1492626536,.2855694985e-1,.1207057037,-.1207057037,.1492626536,-.2855694983e-1},
			{.3973252640e-2,-.3973252644e-2,.2435328156e-11,.6411930517e-1,-.6411930524e-1,-.3373654288e-12,-.1624751717e-2,.1624751718e-2,.4950415964e-10,.3974002441e-1,-.3974002439e-1,-0.,.4385866113e-1,-.4385866112e-1,-0.,.9442419982e-1,-.1337765132,.3935231328e-1,-.3935231309e-1,-.9442420005e-1,.1337765128,-.9849870528e-2,-.4015289410e-1,.5000276456e-1,-.5000276462e-1,.9849870528e-2,.4015289406e-1},
			{-.6881875447e-2,.6881875454e-2,-.4218112101e-11,-.1110578943,.1110578945,.5843340636e-12,.2814152524e-2,-.2814152527e-2,-.8574371972e-10,-.6883174139e-1,.6883174137e-1,0.,-.7596542946e-1,.7596542944e-1,0.,-.1635475116,.2317077178,-.6816020601e-1,.6816020569e-1,.1635475120,-.2317077171,.1706047621e-1,.6954685268e-1,-.8660732877e-1,.8660732887e-1,-.1706047621e-1,-.6954685260e-1},
			{.8884462997e-2,-.8884463007e-2,.5445559306e-11,.1433751251,-.1433751252,-.7543720322e-12,-.3633055286e-2,.3633055289e-2,.1106946662e-9,.8886139602e-1,-.8886139599e-1,-0.,.9807094772e-1,-.9807094769e-1,-0.,.2111389296,-.2991333773,.8799444758e-1,-.8799444716e-1,-.2111389301,.2991333765,-.2202498007e-1,-.8978460072e-1,.1118095806,-.1118095808,.2202498007e-1,.8978460063e-1},
			{.1608750430e-2,-.1608750432e-2,-.4034538186e-11,-.3311226875e-1,.3311226880e-1,.2144253100e-10,.6480800897e-1,-.6480800902e-1,.3701184809e-10,.1514326846e-1,-.1514326846e-1,-0.,-.3828400335e-1,.3828400333e-1,-0.,-.8474998733e-1,-.9988203793e-1,-.2734121092e-1,.2734121078e-1,.8474998753e-1,.9988203767e-1,-.3512819587e-1,-.7636653997e-1,-.4884262097e-1,.4884262099e-1,.3512819587e-1,.7636653992e-1},
			{-.2786437481e-2,.2786437483e-2,.6988025119e-11,.5735213183e-1,-.5735213188e-1,-.3713955314e-10,-.1122507642,.1122507643,-.6410640134e-10,-.2622891035e-1,.2622891035e-1,0.,.6630983887e-1,-.6630983886e-1,0.,.1467912839,.1730007645,.4735636645e-1,-.4735636618e-1,-.1467912843,-.1730007640,.6084382000e-1,.1322707272,.8459790107e-1,-.8459790113e-1,-.6084382000e-1,-.1322707271},
			{-.2900983391e-2,.2900983395e-2,.6735979401e-11,-.5474334561e-1,.5474334565e-1,-.2140983594e-10,-.6378935306e-1,.6378935317e-1,-.1837023118e-10,-.5758708731e-1,.5758708725e-1,-0.,.2739123664e-1,-.2739123663e-1,-0.,-.5980191219e-1,.4982269315e-1,.9979215896e-2,-.9979215849e-2,.5980191234e-1,-.4982269302e-1,.4947010457e-1,-.9238197165e-1,.4291186704e-1,-.4291186706e-1,-.4947010455e-1,.9238197160e-1},
			{-.2227101124e-2,-.2227101127e-2,.2902345977e-2,.5174730169e-1,.5174730161e-1,.1345249687,.2895906774e-1,.2895906742e-1,.2686152177e-1,-.4565783520e-1,-.4565783503e-1,.5646315914e-2,-.5646454713e-1,-.5646454707e-1,.1567410230e-3,-.6746381852e-1,-.1666020075e-1,.1052409173,.1052409171,-.6746381827e-1,-.1666020070e-1,-.6630339505e-1,-.1750103542e-2,-.1446623172e-1,-.1446623165e-1,-.6630339516e-1,-.1750103540e-2},
			{.3857452299e-2,.3857452305e-2,-.5027010692e-2,-.8962895569e-1,-.8962895553e-1,-.2330040807,-.5015857668e-1,-.5015857611e-1,-.4652552046e-1,.7908169034e-1,.7908169002e-1,-.9779706035e-2,.9779946441e-1,.9779946432e-1,-.2714834155e-3,.1168507614,.2885631417e-1,-.1822826158,-.1822826154,.1168507609,.2885631408e-1,.1148408489,.3031268254e-2,.2505624833e-1,.2505624821e-1,.1148408492,.3031268250e-2},
			{-.4979949508e-2,-.4979949513e-2,.6489842900e-2,.1157104843,.1157104841,.3008069748,.6475444405e-1,.6475444331e-1,.6006418864e-1,-.1020940232,-.1020940229,.1262554620e-1,-.1262585657,-.1262585656,.3504835826e-3,-.1508536843,-.3725334142e-1,.2353258451,.2353258446,-.1508536837,-.3725334129e-1,-.1482588985,-.3913350489e-2,-.3234747750e-1,-.3234747736e-1,-.1482588988,-.3913350484e-2},
			{-.7052136016e-2,-.7052136025e-2,-.4975248153e-2,.1071236388,.1071236386,.6437588332e-1,.1415037995e-1,.1415037979e-1,.9781715849e-1,.8674924746e-2,.8674924715e-2,.2822480032e-1,.5042836149e-1,.5042836146e-1,.1003925893e-2,-.1165659990,-.6064459330e-1,-.7883759978e-1,-.7883759957e-1,-.1165659985,-.6064459310e-1,-.1534448774e-1,-.9401828142e-2,.1424607884e-1,.1424607878e-1,-.1534448778e-1,-.9401828126e-2},
			{.1221465788e-1,.1221465790e-1,.8617382582e-2,-.1855435851,-.1855435847,-.1115023007,-.2450917702e-1,-.2450917675e-1,-.1694242884,-.1502541041e-1,-.1502541036e-1,-.4888678819e-1,-.8734448425e-1,-.8734448420e-1,-.1738850654e-2,.2018982328,.1050395168,.1365507284,.1365507280,.2018982319,.1050395164,.2657743238e-1,.1628444403e-1,-.2467493236e-1,-.2467493225e-1,.2657743245e-1,.1628444400e-1},
			{-.7864599685e-2,-.7864599698e-2,.8543165539e-2,.2317116788e-1,.2317116784e-1,-.1325361900,-.2859660209e-1,-.2859660176e-1,.1450070316,.2338942873e-1,.2338942865e-1,.7058964713e-1,-.3789617428e-1,-.3789617426e-1,.3546410271e-2,.1110571477e-1,-.8983420728e-1,.3910101757e-1,.3910101750e-1,.1110571472e-1,-.8983420697e-1,.5864432557e-1,-.2610381621e-1,-.1269128828e-1,-.1269128822e-1,.5864432571e-1,-.2610381617e-1},
			{.2457406867e-2,-.2457406870e-2,.1456786833e-10,-.5057982506e-1,.5057982506e-1,-0.,.9899586944e-1,-.9899586881e-1,.4439325132e-9,.2313172558e-1,-.2313172466e-1,-0.,-.5847978098e-1,.5847978105e-1,.8177605344e-10,.2473009711e-1,.1615508286e-2,.1124234492,-.1124234521,-.2473010027e-1,-.1615508521e-2,.6296906016e-1,-.2355095997e-4,.4201992979e-1,-.4201992986e-1,-.6296905964e-1,.2355071203e-4},
			{-.4256353547e-2,.4256353552e-2,-.2523228810e-10,.8760682683e-1,-.8760682683e-1,0.,-.1714658756,.1714658745,-.7689136679e-9,-.4006532397e-1,.4006532237e-1,0.,.1012899519,-.1012899520,-.1416402794e-9,-.4283378466e-1,-.2798142432e-2,-.1947231258,.1947231310,.4283379015e-1,.2798142839e-2,-.1090656115,.4079145923e-4,-.7278065332e-1,.7278065345e-1,.1090656106,-.4079102978e-4},
			{.1047195901e-1,-.1047195902e-1,-.2643438623e-10,-.1465053324,.1465053324,-0.,.9184167768e-1,-.9184167710e-1,.2154288659e-8,.1917883762e-2,-.1917883686e-2,-0.,.4633750594e-1,-.4633750599e-1,.6781638958e-9,.6109303566e-1,.7836730672e-2,-.6892976350e-1,.6892976527e-1,-.6109304346e-1,-.7836731808e-2,.3825625994e-1,-.1650255183e-3,-.3809123559e-1,.3809123567e-1,-.3825625962e-1,.1650237808e-3},
			{-.2665390899e-2,-.2665390853e-2,-.4310294403e-2,-.2187432388e-2,-.2187432754e-2,-.1112663225,.6179199770e-1,.6179199994e-1,-.5769386249e-2,.7894214835e-1,.7894214861e-1,-.4179238568e-3,-.4846361631e-1,-.4846362257e-1,-.8532846255e-6,.6483970316e-2,.4679778692e-2,-.6432957221e-1,-.6432958410e-1,.6483978942e-2,.4679777698e-2,-.5074575336e-1,.7639251247e-4,.7729988409e-1,.7729988561e-1,-.5074575829e-1,.7639349119e-4}
		};
		for (int j = 0; j < 35; j++)
		{
			std::cout << timestep_top_fluxes_to_dofs[j][0];
			for (int i = 1; i < 27; i++)
			{
				std::cout << ", " << timestep_top_fluxes_to_dofs[j][i];
			}
			std::cout << std::endl;
		}

	
		std::cout << "Flops per cell update using matrix multiplications for dense matrices: " << getNumberOfFlops() << std::endl;
		std::cout << std::endl;
	}
};

#endif
